/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 2465
#define LARGE_STATE_COUNT 1025
#define SYMBOL_COUNT 252
#define ALIAS_COUNT 1
#define TOKEN_COUNT 160
#define EXTERNAL_TOKEN_COUNT 0
#define FIELD_COUNT 23
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 82
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym_identifier = 1,
  anon_sym_POUND_BANG = 2,
  aux_sym_shebang_token1 = 3,
  anon_sym_break = 4,
  anon_sym_continue = 5,
  anon_sym_SEMI = 6,
  anon_sym_COLON = 7,
  anon_sym_DOT_AMP = 8,
  anon_sym_DOT_AT = 9,
  anon_sym_QMARK_DOT = 10,
  anon_sym_STAR_DOT = 11,
  anon_sym_STAR = 12,
  anon_sym_STAR_COLON = 13,
  anon_sym_DOT = 14,
  anon_sym_import = 15,
  anon_sym_STAR2 = 16,
  anon_sym_as = 17,
  anon_sym_package = 18,
  anon_sym_AT = 19,
  aux_sym_annotation_token1 = 20,
  anon_sym_assert = 21,
  anon_sym_EQ = 22,
  anon_sym_STAR_STAR_EQ = 23,
  anon_sym_STAR_EQ = 24,
  anon_sym_SLASH_EQ = 25,
  anon_sym_PERCENT_EQ = 26,
  anon_sym_PLUS_EQ = 27,
  anon_sym_DASH_EQ = 28,
  anon_sym_LT_LT_EQ = 29,
  anon_sym_GT_GT_EQ = 30,
  anon_sym_GT_GT_GT_EQ = 31,
  anon_sym_AMP_EQ = 32,
  anon_sym_CARET_EQ = 33,
  anon_sym_PIPE_EQ = 34,
  anon_sym_QMARK_EQ = 35,
  anon_sym_PLUS_PLUS = 36,
  anon_sym_DASH_DASH = 37,
  anon_sym_PERCENT = 38,
  anon_sym_SLASH = 39,
  anon_sym_PLUS = 40,
  anon_sym_DASH = 41,
  anon_sym_LT_LT = 42,
  anon_sym_GT_GT = 43,
  anon_sym_GT_GT_GT = 44,
  anon_sym_DOT_DOT = 45,
  anon_sym_DOT_DOT_LT = 46,
  anon_sym_LT_DOT_DOT_LT = 47,
  anon_sym_LT_DOT_DOT = 48,
  anon_sym_LT = 49,
  anon_sym_LT_EQ = 50,
  anon_sym_GT = 51,
  anon_sym_GT_EQ = 52,
  anon_sym_in = 53,
  anon_sym_BANGin = 54,
  anon_sym_instanceof = 55,
  anon_sym_BANGinstanceof = 56,
  anon_sym_EQ_EQ = 57,
  anon_sym_BANG_EQ = 58,
  anon_sym_LT_EQ_GT = 59,
  anon_sym_EQ_EQ_EQ = 60,
  anon_sym_BANG_EQ_EQ = 61,
  anon_sym_EQ_TILDE = 62,
  anon_sym_EQ_EQ_TILDE = 63,
  anon_sym_AMP = 64,
  anon_sym_CARET = 65,
  anon_sym_PIPE = 66,
  anon_sym_AMP_AMP = 67,
  anon_sym_PIPE_PIPE = 68,
  anon_sym_QMARK_COLON = 69,
  anon_sym_STAR_STAR = 70,
  anon_sym_true = 71,
  anon_sym_false = 72,
  anon_sym_ATinterface = 73,
  anon_sym_interface = 74,
  anon_sym_class = 75,
  anon_sym_extends = 76,
  anon_sym_COMMA = 77,
  anon_sym_LBRACE = 78,
  anon_sym_DASH_GT = 79,
  anon_sym_RBRACE = 80,
  aux_sym_comment_token1 = 81,
  anon_sym_SLASH_STAR = 82,
  aux_sym_comment_token2 = 83,
  anon_sym_SLASH_STAR_STAR = 84,
  aux_sym_groovy_doc_token1 = 85,
  aux_sym_groovy_doc_token2 = 86,
  aux_sym_groovy_doc_token3 = 87,
  anon_sym_STAR_SLASH = 88,
  anon_sym_ATparam = 89,
  anon_sym_ATthrows = 90,
  sym_groovy_doc_tag = 91,
  sym_groovy_doc_at_text = 92,
  anon_sym__ = 93,
  anon_sym_def = 94,
  anon_sym_LPAREN = 95,
  anon_sym_RPAREN = 96,
  anon_sym_null = 97,
  anon_sym_this = 98,
  anon_sym_do = 99,
  anon_sym_while = 100,
  anon_sym_for = 101,
  aux_sym__type_identifier_token1 = 102,
  anon_sym_if = 103,
  anon_sym_else = 104,
  anon_sym_LBRACK = 105,
  anon_sym_RBRACK = 106,
  aux_sym_number_literal_token1 = 107,
  aux_sym_number_literal_token2 = 108,
  aux_sym_number_literal_token3 = 109,
  aux_sym_number_literal_token4 = 110,
  aux_sym_number_literal_token5 = 111,
  anon_sym_pipeline = 112,
  anon_sym_return = 113,
  anon_sym_SQUOTE = 114,
  aux_sym__plain_string_token1 = 115,
  anon_sym_SQUOTE_SQUOTE_SQUOTE = 116,
  aux_sym__plain_string_token2 = 117,
  aux_sym__plain_string_token3 = 118,
  aux_sym__plain_string_token4 = 119,
  anon_sym_DQUOTE = 120,
  aux_sym__interpolate_string_token1 = 121,
  anon_sym_DQUOTE_DQUOTE_DQUOTE = 122,
  aux_sym__interpolate_string_token2 = 123,
  aux_sym__interpolate_string_token3 = 124,
  aux_sym__interpolate_string_token4 = 125,
  anon_sym_BSLASH_SLASH = 126,
  aux_sym__interpolate_string_token5 = 127,
  anon_sym_DOLLAR_SLASH = 128,
  aux_sym__interpolate_string_token6 = 129,
  anon_sym_DOLLAR_DOLLAR = 130,
  anon_sym_SLASH_DOLLAR = 131,
  sym_escape_sequence = 132,
  anon_sym_DOLLAR = 133,
  aux_sym_interpolation_token1 = 134,
  anon_sym_switch = 135,
  anon_sym_case = 136,
  anon_sym_default = 137,
  anon_sym_QMARK = 138,
  anon_sym_try = 139,
  anon_sym_catch = 140,
  anon_sym_finally = 141,
  anon_sym_int = 142,
  anon_sym_boolean = 143,
  anon_sym_char = 144,
  anon_sym_short = 145,
  anon_sym_long = 146,
  anon_sym_float = 147,
  anon_sym_double = 148,
  anon_sym_void = 149,
  anon_sym_LBRACK_RBRACK = 150,
  anon_sym_public = 151,
  anon_sym_protected = 152,
  anon_sym_private = 153,
  anon_sym_static = 154,
  anon_sym_final = 155,
  anon_sym_synchronized = 156,
  anon_sym_TILDE = 157,
  anon_sym_BANG = 158,
  anon_sym_new = 159,
  sym_source_file = 160,
  sym_shebang = 161,
  sym__statement = 162,
  sym_label = 163,
  sym_access_op = 164,
  sym_dotted_identifier = 165,
  sym__import_name = 166,
  sym_groovy_import = 167,
  sym_groovy_package = 168,
  sym_annotation = 169,
  sym_assertion = 170,
  sym_assignment = 171,
  sym_increment_op = 172,
  sym_binary_op = 173,
  sym_boolean_literal = 174,
  sym_class_definition = 175,
  sym_generic_parameters = 176,
  sym_generic_param = 177,
  sym_closure = 178,
  sym_comment = 179,
  sym_groovy_doc = 180,
  sym_groovy_doc_param = 181,
  sym_groovy_doc_throws = 182,
  sym_declaration = 183,
  sym_parenthesized_expression = 184,
  sym__expression = 185,
  sym__primary_expression = 186,
  sym__callable_expression = 187,
  sym__juxtable_expression = 188,
  sym_do_while_loop = 189,
  sym_for_parameters = 190,
  sym_for_loop = 191,
  sym_for_in_loop = 192,
  sym_function_call = 193,
  sym___immediately_invoked_closure = 194,
  sym__immediately_invoked_closure = 195,
  sym_argument_list = 196,
  sym__param_list = 197,
  sym_parameter_list = 198,
  sym_parameter = 199,
  sym_function_declaration = 200,
  sym_function_definition = 201,
  sym__type_identifier = 202,
  sym_if_statement = 203,
  sym_index = 204,
  sym_juxt_function_call = 205,
  sym__juxt_argument_list = 206,
  sym_list = 207,
  sym_map_item = 208,
  sym_map = 209,
  sym_number_literal = 210,
  sym_pipeline = 211,
  sym_return = 212,
  sym_string = 213,
  sym__plain_string = 214,
  sym__interpolate_string = 215,
  sym_interpolation = 216,
  sym_switch_statement = 217,
  sym_switch_block = 218,
  sym_case = 219,
  sym_ternary_op = 220,
  sym_try_statement = 221,
  sym_builtintype = 222,
  sym__type = 223,
  sym_array_type = 224,
  sym_access_modifier = 225,
  sym_modifier = 226,
  sym_type_with_generics = 227,
  sym_generics = 228,
  sym_unary_op = 229,
  sym_while_loop = 230,
  aux_sym_source_file_repeat1 = 231,
  aux_sym_dotted_identifier_repeat1 = 232,
  aux_sym_class_definition_repeat1 = 233,
  aux_sym_class_definition_repeat2 = 234,
  aux_sym_generic_parameters_repeat1 = 235,
  aux_sym_groovy_doc_repeat1 = 236,
  aux_sym_for_parameters_repeat1 = 237,
  aux_sym_for_parameters_repeat2 = 238,
  aux_sym_argument_list_repeat1 = 239,
  aux_sym__param_list_repeat1 = 240,
  aux_sym__juxt_argument_list_repeat1 = 241,
  aux_sym_list_repeat1 = 242,
  aux_sym_map_repeat1 = 243,
  aux_sym__plain_string_repeat1 = 244,
  aux_sym__plain_string_repeat2 = 245,
  aux_sym__interpolate_string_repeat1 = 246,
  aux_sym__interpolate_string_repeat2 = 247,
  aux_sym__interpolate_string_repeat3 = 248,
  aux_sym__interpolate_string_repeat4 = 249,
  aux_sym_switch_block_repeat1 = 250,
  aux_sym_generics_repeat1 = 251,
  alias_sym_qualified_name = 252,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [anon_sym_POUND_BANG] = "#!",
  [aux_sym_shebang_token1] = "shebang_token1",
  [anon_sym_break] = "break",
  [anon_sym_continue] = "continue",
  [anon_sym_SEMI] = ";",
  [anon_sym_COLON] = ":",
  [anon_sym_DOT_AMP] = ".&",
  [anon_sym_DOT_AT] = ".@",
  [anon_sym_QMARK_DOT] = "\?.",
  [anon_sym_STAR_DOT] = "*.",
  [anon_sym_STAR] = "*",
  [anon_sym_STAR_COLON] = "*:",
  [anon_sym_DOT] = ".",
  [anon_sym_import] = "import",
  [anon_sym_STAR2] = "wildcard_import",
  [anon_sym_as] = "as",
  [anon_sym_package] = "package",
  [anon_sym_AT] = "@",
  [aux_sym_annotation_token1] = "identifier",
  [anon_sym_assert] = "assert",
  [anon_sym_EQ] = "=",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_GT_GT_GT_EQ] = ">>>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_QMARK_EQ] = "\?=",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_PERCENT] = "%",
  [anon_sym_SLASH] = "/",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_GT_GT_GT] = ">>>",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_DOT_DOT_LT] = "..<",
  [anon_sym_LT_DOT_DOT_LT] = "<..<",
  [anon_sym_LT_DOT_DOT] = "<..",
  [anon_sym_LT] = "<",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_in] = "in",
  [anon_sym_BANGin] = "!in",
  [anon_sym_instanceof] = "instanceof",
  [anon_sym_BANGinstanceof] = "!instanceof",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT_EQ_GT] = "<=>",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_EQ_TILDE] = "=~",
  [anon_sym_EQ_EQ_TILDE] = "==~",
  [anon_sym_AMP] = "&",
  [anon_sym_CARET] = "^",
  [anon_sym_PIPE] = "|",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_QMARK_COLON] = "\?:",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_ATinterface] = "@interface",
  [anon_sym_interface] = "interface",
  [anon_sym_class] = "class",
  [anon_sym_extends] = "extends",
  [anon_sym_COMMA] = ",",
  [anon_sym_LBRACE] = "{",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_RBRACE] = "}",
  [aux_sym_comment_token1] = "comment_token1",
  [anon_sym_SLASH_STAR] = "/*",
  [aux_sym_comment_token2] = "comment_token2",
  [anon_sym_SLASH_STAR_STAR] = "/**",
  [aux_sym_groovy_doc_token1] = "groovy_doc_token1",
  [aux_sym_groovy_doc_token2] = "first_line",
  [aux_sym_groovy_doc_token3] = "groovy_doc_token3",
  [anon_sym_STAR_SLASH] = "*/",
  [anon_sym_ATparam] = "@param",
  [anon_sym_ATthrows] = "@throws",
  [sym_groovy_doc_tag] = "groovy_doc_tag",
  [sym_groovy_doc_at_text] = "groovy_doc_at_text",
  [anon_sym__] = "_",
  [anon_sym_def] = "def",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_null] = "null",
  [anon_sym_this] = "this",
  [anon_sym_do] = "do",
  [anon_sym_while] = "while",
  [anon_sym_for] = "for",
  [aux_sym__type_identifier_token1] = "identifier",
  [anon_sym_if] = "if",
  [anon_sym_else] = "else",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [aux_sym_number_literal_token1] = "number_literal_token1",
  [aux_sym_number_literal_token2] = "number_literal_token2",
  [aux_sym_number_literal_token3] = "number_literal_token3",
  [aux_sym_number_literal_token4] = "number_literal_token4",
  [aux_sym_number_literal_token5] = "number_literal_token5",
  [anon_sym_pipeline] = "pipeline",
  [anon_sym_return] = "return",
  [anon_sym_SQUOTE] = "'",
  [aux_sym__plain_string_token1] = "string_content",
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = "'''",
  [aux_sym__plain_string_token2] = "string_internal_quote",
  [aux_sym__plain_string_token3] = "string_content",
  [aux_sym__plain_string_token4] = "_plain_string_token4",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym__interpolate_string_token1] = "string_content",
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = "\"\"\"",
  [aux_sym__interpolate_string_token2] = "string_content",
  [aux_sym__interpolate_string_token3] = "_interpolate_string_token3",
  [aux_sym__interpolate_string_token4] = "string_content",
  [anon_sym_BSLASH_SLASH] = "escape_sequence",
  [aux_sym__interpolate_string_token5] = "string_content",
  [anon_sym_DOLLAR_SLASH] = "$/",
  [aux_sym__interpolate_string_token6] = "string_content",
  [anon_sym_DOLLAR_DOLLAR] = "escape_sequence",
  [anon_sym_SLASH_DOLLAR] = "/$",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_DOLLAR] = "$",
  [aux_sym_interpolation_token1] = "identifier",
  [anon_sym_switch] = "switch",
  [anon_sym_case] = "case",
  [anon_sym_default] = "default",
  [anon_sym_QMARK] = "\?",
  [anon_sym_try] = "try",
  [anon_sym_catch] = "catch",
  [anon_sym_finally] = "finally",
  [anon_sym_int] = "int",
  [anon_sym_boolean] = "boolean",
  [anon_sym_char] = "char",
  [anon_sym_short] = "short",
  [anon_sym_long] = "long",
  [anon_sym_float] = "float",
  [anon_sym_double] = "double",
  [anon_sym_void] = "void",
  [anon_sym_LBRACK_RBRACK] = "[]",
  [anon_sym_public] = "public",
  [anon_sym_protected] = "protected",
  [anon_sym_private] = "private",
  [anon_sym_static] = "static",
  [anon_sym_final] = "final",
  [anon_sym_synchronized] = "synchronized",
  [anon_sym_TILDE] = "~",
  [anon_sym_BANG] = "!",
  [anon_sym_new] = "new",
  [sym_source_file] = "source_file",
  [sym_shebang] = "shebang",
  [sym__statement] = "_statement",
  [sym_label] = "label",
  [sym_access_op] = "access_op",
  [sym_dotted_identifier] = "dotted_identifier",
  [sym__import_name] = "_import_name",
  [sym_groovy_import] = "groovy_import",
  [sym_groovy_package] = "groovy_package",
  [sym_annotation] = "annotation",
  [sym_assertion] = "assertion",
  [sym_assignment] = "assignment",
  [sym_increment_op] = "increment_op",
  [sym_binary_op] = "binary_op",
  [sym_boolean_literal] = "boolean_literal",
  [sym_class_definition] = "class_definition",
  [sym_generic_parameters] = "generic_parameters",
  [sym_generic_param] = "generic_param",
  [sym_closure] = "closure",
  [sym_comment] = "comment",
  [sym_groovy_doc] = "groovy_doc",
  [sym_groovy_doc_param] = "groovy_doc_param",
  [sym_groovy_doc_throws] = "groovy_doc_throws",
  [sym_declaration] = "declaration",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym__expression] = "_expression",
  [sym__primary_expression] = "_primary_expression",
  [sym__callable_expression] = "_callable_expression",
  [sym__juxtable_expression] = "_juxtable_expression",
  [sym_do_while_loop] = "do_while_loop",
  [sym_for_parameters] = "for_parameters",
  [sym_for_loop] = "for_loop",
  [sym_for_in_loop] = "for_in_loop",
  [sym_function_call] = "function_call",
  [sym___immediately_invoked_closure] = "function_call",
  [sym__immediately_invoked_closure] = "_immediately_invoked_closure",
  [sym_argument_list] = "argument_list",
  [sym__param_list] = "parameter_list",
  [sym_parameter_list] = "parameter_list",
  [sym_parameter] = "parameter",
  [sym_function_declaration] = "function_declaration",
  [sym_function_definition] = "function_definition",
  [sym__type_identifier] = "_type_identifier",
  [sym_if_statement] = "if_statement",
  [sym_index] = "index",
  [sym_juxt_function_call] = "juxt_function_call",
  [sym__juxt_argument_list] = "argument_list",
  [sym_list] = "list",
  [sym_map_item] = "map_item",
  [sym_map] = "map",
  [sym_number_literal] = "number_literal",
  [sym_pipeline] = "pipeline",
  [sym_return] = "return",
  [sym_string] = "string",
  [sym__plain_string] = "_plain_string",
  [sym__interpolate_string] = "_interpolate_string",
  [sym_interpolation] = "interpolation",
  [sym_switch_statement] = "switch_statement",
  [sym_switch_block] = "switch_block",
  [sym_case] = "case",
  [sym_ternary_op] = "ternary_op",
  [sym_try_statement] = "try_statement",
  [sym_builtintype] = "builtintype",
  [sym__type] = "_type",
  [sym_array_type] = "array_type",
  [sym_access_modifier] = "access_modifier",
  [sym_modifier] = "modifier",
  [sym_type_with_generics] = "type_with_generics",
  [sym_generics] = "generics",
  [sym_unary_op] = "unary_op",
  [sym_while_loop] = "while_loop",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_dotted_identifier_repeat1] = "dotted_identifier_repeat1",
  [aux_sym_class_definition_repeat1] = "class_definition_repeat1",
  [aux_sym_class_definition_repeat2] = "class_definition_repeat2",
  [aux_sym_generic_parameters_repeat1] = "generic_parameters_repeat1",
  [aux_sym_groovy_doc_repeat1] = "groovy_doc_repeat1",
  [aux_sym_for_parameters_repeat1] = "for_parameters_repeat1",
  [aux_sym_for_parameters_repeat2] = "for_parameters_repeat2",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym__param_list_repeat1] = "_param_list_repeat1",
  [aux_sym__juxt_argument_list_repeat1] = "_juxt_argument_list_repeat1",
  [aux_sym_list_repeat1] = "list_repeat1",
  [aux_sym_map_repeat1] = "map_repeat1",
  [aux_sym__plain_string_repeat1] = "_plain_string_repeat1",
  [aux_sym__plain_string_repeat2] = "_plain_string_repeat2",
  [aux_sym__interpolate_string_repeat1] = "_interpolate_string_repeat1",
  [aux_sym__interpolate_string_repeat2] = "_interpolate_string_repeat2",
  [aux_sym__interpolate_string_repeat3] = "_interpolate_string_repeat3",
  [aux_sym__interpolate_string_repeat4] = "_interpolate_string_repeat4",
  [aux_sym_switch_block_repeat1] = "switch_block_repeat1",
  [aux_sym_generics_repeat1] = "generics_repeat1",
  [alias_sym_qualified_name] = "qualified_name",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [anon_sym_POUND_BANG] = anon_sym_POUND_BANG,
  [aux_sym_shebang_token1] = aux_sym_shebang_token1,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_DOT_AMP] = anon_sym_DOT_AMP,
  [anon_sym_DOT_AT] = anon_sym_DOT_AT,
  [anon_sym_QMARK_DOT] = anon_sym_QMARK_DOT,
  [anon_sym_STAR_DOT] = anon_sym_STAR_DOT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_STAR_COLON] = anon_sym_STAR_COLON,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_STAR2] = anon_sym_STAR2,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_package] = anon_sym_package,
  [anon_sym_AT] = anon_sym_AT,
  [aux_sym_annotation_token1] = sym_identifier,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_GT_GT_GT_EQ] = anon_sym_GT_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_QMARK_EQ] = anon_sym_QMARK_EQ,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_GT_GT_GT] = anon_sym_GT_GT_GT,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_DOT_DOT_LT] = anon_sym_DOT_DOT_LT,
  [anon_sym_LT_DOT_DOT_LT] = anon_sym_LT_DOT_DOT_LT,
  [anon_sym_LT_DOT_DOT] = anon_sym_LT_DOT_DOT,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_BANGin] = anon_sym_BANGin,
  [anon_sym_instanceof] = anon_sym_instanceof,
  [anon_sym_BANGinstanceof] = anon_sym_BANGinstanceof,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT_EQ_GT] = anon_sym_LT_EQ_GT,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_EQ_TILDE] = anon_sym_EQ_TILDE,
  [anon_sym_EQ_EQ_TILDE] = anon_sym_EQ_EQ_TILDE,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_QMARK_COLON] = anon_sym_QMARK_COLON,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_ATinterface] = anon_sym_ATinterface,
  [anon_sym_interface] = anon_sym_interface,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_extends] = anon_sym_extends,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_comment_token1] = aux_sym_comment_token1,
  [anon_sym_SLASH_STAR] = anon_sym_SLASH_STAR,
  [aux_sym_comment_token2] = aux_sym_comment_token2,
  [anon_sym_SLASH_STAR_STAR] = anon_sym_SLASH_STAR_STAR,
  [aux_sym_groovy_doc_token1] = aux_sym_groovy_doc_token1,
  [aux_sym_groovy_doc_token2] = aux_sym_groovy_doc_token2,
  [aux_sym_groovy_doc_token3] = aux_sym_groovy_doc_token3,
  [anon_sym_STAR_SLASH] = anon_sym_STAR_SLASH,
  [anon_sym_ATparam] = anon_sym_ATparam,
  [anon_sym_ATthrows] = anon_sym_ATthrows,
  [sym_groovy_doc_tag] = sym_groovy_doc_tag,
  [sym_groovy_doc_at_text] = sym_groovy_doc_at_text,
  [anon_sym__] = anon_sym__,
  [anon_sym_def] = anon_sym_def,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_null] = anon_sym_null,
  [anon_sym_this] = anon_sym_this,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_for] = anon_sym_for,
  [aux_sym__type_identifier_token1] = sym_identifier,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [aux_sym_number_literal_token1] = aux_sym_number_literal_token1,
  [aux_sym_number_literal_token2] = aux_sym_number_literal_token2,
  [aux_sym_number_literal_token3] = aux_sym_number_literal_token3,
  [aux_sym_number_literal_token4] = aux_sym_number_literal_token4,
  [aux_sym_number_literal_token5] = aux_sym_number_literal_token5,
  [anon_sym_pipeline] = anon_sym_pipeline,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym__plain_string_token1] = aux_sym__plain_string_token1,
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = anon_sym_SQUOTE_SQUOTE_SQUOTE,
  [aux_sym__plain_string_token2] = aux_sym__plain_string_token2,
  [aux_sym__plain_string_token3] = aux_sym__plain_string_token1,
  [aux_sym__plain_string_token4] = aux_sym__plain_string_token4,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym__interpolate_string_token1] = aux_sym__plain_string_token1,
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = anon_sym_DQUOTE_DQUOTE_DQUOTE,
  [aux_sym__interpolate_string_token2] = aux_sym__plain_string_token1,
  [aux_sym__interpolate_string_token3] = aux_sym__interpolate_string_token3,
  [aux_sym__interpolate_string_token4] = aux_sym__plain_string_token1,
  [anon_sym_BSLASH_SLASH] = sym_escape_sequence,
  [aux_sym__interpolate_string_token5] = aux_sym__plain_string_token1,
  [anon_sym_DOLLAR_SLASH] = anon_sym_DOLLAR_SLASH,
  [aux_sym__interpolate_string_token6] = aux_sym__plain_string_token1,
  [anon_sym_DOLLAR_DOLLAR] = sym_escape_sequence,
  [anon_sym_SLASH_DOLLAR] = anon_sym_SLASH_DOLLAR,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [aux_sym_interpolation_token1] = sym_identifier,
  [anon_sym_switch] = anon_sym_switch,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_catch] = anon_sym_catch,
  [anon_sym_finally] = anon_sym_finally,
  [anon_sym_int] = anon_sym_int,
  [anon_sym_boolean] = anon_sym_boolean,
  [anon_sym_char] = anon_sym_char,
  [anon_sym_short] = anon_sym_short,
  [anon_sym_long] = anon_sym_long,
  [anon_sym_float] = anon_sym_float,
  [anon_sym_double] = anon_sym_double,
  [anon_sym_void] = anon_sym_void,
  [anon_sym_LBRACK_RBRACK] = anon_sym_LBRACK_RBRACK,
  [anon_sym_public] = anon_sym_public,
  [anon_sym_protected] = anon_sym_protected,
  [anon_sym_private] = anon_sym_private,
  [anon_sym_static] = anon_sym_static,
  [anon_sym_final] = anon_sym_final,
  [anon_sym_synchronized] = anon_sym_synchronized,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_new] = anon_sym_new,
  [sym_source_file] = sym_source_file,
  [sym_shebang] = sym_shebang,
  [sym__statement] = sym__statement,
  [sym_label] = sym_label,
  [sym_access_op] = sym_access_op,
  [sym_dotted_identifier] = sym_dotted_identifier,
  [sym__import_name] = sym__import_name,
  [sym_groovy_import] = sym_groovy_import,
  [sym_groovy_package] = sym_groovy_package,
  [sym_annotation] = sym_annotation,
  [sym_assertion] = sym_assertion,
  [sym_assignment] = sym_assignment,
  [sym_increment_op] = sym_increment_op,
  [sym_binary_op] = sym_binary_op,
  [sym_boolean_literal] = sym_boolean_literal,
  [sym_class_definition] = sym_class_definition,
  [sym_generic_parameters] = sym_generic_parameters,
  [sym_generic_param] = sym_generic_param,
  [sym_closure] = sym_closure,
  [sym_comment] = sym_comment,
  [sym_groovy_doc] = sym_groovy_doc,
  [sym_groovy_doc_param] = sym_groovy_doc_param,
  [sym_groovy_doc_throws] = sym_groovy_doc_throws,
  [sym_declaration] = sym_declaration,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym__expression] = sym__expression,
  [sym__primary_expression] = sym__primary_expression,
  [sym__callable_expression] = sym__callable_expression,
  [sym__juxtable_expression] = sym__juxtable_expression,
  [sym_do_while_loop] = sym_do_while_loop,
  [sym_for_parameters] = sym_for_parameters,
  [sym_for_loop] = sym_for_loop,
  [sym_for_in_loop] = sym_for_in_loop,
  [sym_function_call] = sym_function_call,
  [sym___immediately_invoked_closure] = sym_function_call,
  [sym__immediately_invoked_closure] = sym__immediately_invoked_closure,
  [sym_argument_list] = sym_argument_list,
  [sym__param_list] = sym_parameter_list,
  [sym_parameter_list] = sym_parameter_list,
  [sym_parameter] = sym_parameter,
  [sym_function_declaration] = sym_function_declaration,
  [sym_function_definition] = sym_function_definition,
  [sym__type_identifier] = sym__type_identifier,
  [sym_if_statement] = sym_if_statement,
  [sym_index] = sym_index,
  [sym_juxt_function_call] = sym_juxt_function_call,
  [sym__juxt_argument_list] = sym_argument_list,
  [sym_list] = sym_list,
  [sym_map_item] = sym_map_item,
  [sym_map] = sym_map,
  [sym_number_literal] = sym_number_literal,
  [sym_pipeline] = sym_pipeline,
  [sym_return] = sym_return,
  [sym_string] = sym_string,
  [sym__plain_string] = sym__plain_string,
  [sym__interpolate_string] = sym__interpolate_string,
  [sym_interpolation] = sym_interpolation,
  [sym_switch_statement] = sym_switch_statement,
  [sym_switch_block] = sym_switch_block,
  [sym_case] = sym_case,
  [sym_ternary_op] = sym_ternary_op,
  [sym_try_statement] = sym_try_statement,
  [sym_builtintype] = sym_builtintype,
  [sym__type] = sym__type,
  [sym_array_type] = sym_array_type,
  [sym_access_modifier] = sym_access_modifier,
  [sym_modifier] = sym_modifier,
  [sym_type_with_generics] = sym_type_with_generics,
  [sym_generics] = sym_generics,
  [sym_unary_op] = sym_unary_op,
  [sym_while_loop] = sym_while_loop,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_dotted_identifier_repeat1] = aux_sym_dotted_identifier_repeat1,
  [aux_sym_class_definition_repeat1] = aux_sym_class_definition_repeat1,
  [aux_sym_class_definition_repeat2] = aux_sym_class_definition_repeat2,
  [aux_sym_generic_parameters_repeat1] = aux_sym_generic_parameters_repeat1,
  [aux_sym_groovy_doc_repeat1] = aux_sym_groovy_doc_repeat1,
  [aux_sym_for_parameters_repeat1] = aux_sym_for_parameters_repeat1,
  [aux_sym_for_parameters_repeat2] = aux_sym_for_parameters_repeat2,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym__param_list_repeat1] = aux_sym__param_list_repeat1,
  [aux_sym__juxt_argument_list_repeat1] = aux_sym__juxt_argument_list_repeat1,
  [aux_sym_list_repeat1] = aux_sym_list_repeat1,
  [aux_sym_map_repeat1] = aux_sym_map_repeat1,
  [aux_sym__plain_string_repeat1] = aux_sym__plain_string_repeat1,
  [aux_sym__plain_string_repeat2] = aux_sym__plain_string_repeat2,
  [aux_sym__interpolate_string_repeat1] = aux_sym__interpolate_string_repeat1,
  [aux_sym__interpolate_string_repeat2] = aux_sym__interpolate_string_repeat2,
  [aux_sym__interpolate_string_repeat3] = aux_sym__interpolate_string_repeat3,
  [aux_sym__interpolate_string_repeat4] = aux_sym__interpolate_string_repeat4,
  [aux_sym_switch_block_repeat1] = aux_sym_switch_block_repeat1,
  [aux_sym_generics_repeat1] = aux_sym_generics_repeat1,
  [alias_sym_qualified_name] = alias_sym_qualified_name,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_POUND_BANG] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_shebang_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR2] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_package] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_annotation_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DOT_DOT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANGin] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_instanceof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANGinstanceof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATinterface] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_interface] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extends] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comment_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SLASH_STAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comment_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SLASH_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_groovy_doc_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_groovy_doc_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_groovy_doc_token3] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_STAR_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATparam] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATthrows] = {
    .visible = true,
    .named = false,
  },
  [sym_groovy_doc_tag] = {
    .visible = true,
    .named = true,
  },
  [sym_groovy_doc_at_text] = {
    .visible = true,
    .named = true,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_def] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_null] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_this] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__type_identifier_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_number_literal_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_number_literal_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_number_literal_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_number_literal_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_number_literal_token5] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_pipeline] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__plain_string_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__plain_string_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__plain_string_token3] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__plain_string_token4] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__interpolate_string_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__interpolate_string_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolate_string_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolate_string_token4] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_BSLASH_SLASH] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__interpolate_string_token5] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOLLAR_SLASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__interpolate_string_token6] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOLLAR_DOLLAR] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SLASH_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_interpolation_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_switch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_catch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_boolean] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_char] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_short] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_long] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_float] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_double] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_void] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_public] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_protected] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_private] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_final] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_synchronized] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym_shebang] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym_label] = {
    .visible = true,
    .named = true,
  },
  [sym_access_op] = {
    .visible = true,
    .named = true,
  },
  [sym_dotted_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__import_name] = {
    .visible = false,
    .named = true,
  },
  [sym_groovy_import] = {
    .visible = true,
    .named = true,
  },
  [sym_groovy_package] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation] = {
    .visible = true,
    .named = true,
  },
  [sym_assertion] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_increment_op] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_op] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_param] = {
    .visible = true,
    .named = true,
  },
  [sym_closure] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_groovy_doc] = {
    .visible = true,
    .named = true,
  },
  [sym_groovy_doc_param] = {
    .visible = true,
    .named = true,
  },
  [sym_groovy_doc_throws] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym__primary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__callable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__juxtable_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_do_while_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_for_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_for_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_for_in_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_function_call] = {
    .visible = true,
    .named = true,
  },
  [sym___immediately_invoked_closure] = {
    .visible = true,
    .named = true,
  },
  [sym__immediately_invoked_closure] = {
    .visible = false,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym__param_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_function_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym__type_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_index] = {
    .visible = true,
    .named = true,
  },
  [sym_juxt_function_call] = {
    .visible = true,
    .named = true,
  },
  [sym__juxt_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym_map_item] = {
    .visible = true,
    .named = true,
  },
  [sym_map] = {
    .visible = true,
    .named = true,
  },
  [sym_number_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline] = {
    .visible = true,
    .named = true,
  },
  [sym_return] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym__plain_string] = {
    .visible = false,
    .named = true,
  },
  [sym__interpolate_string] = {
    .visible = false,
    .named = true,
  },
  [sym_interpolation] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_block] = {
    .visible = true,
    .named = true,
  },
  [sym_case] = {
    .visible = true,
    .named = true,
  },
  [sym_ternary_op] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_builtintype] = {
    .visible = true,
    .named = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
  },
  [sym_array_type] = {
    .visible = true,
    .named = true,
  },
  [sym_access_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_type_with_generics] = {
    .visible = true,
    .named = true,
  },
  [sym_generics] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_op] = {
    .visible = true,
    .named = true,
  },
  [sym_while_loop] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dotted_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_definition_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_groovy_doc_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_for_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_for_parameters_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__param_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__juxt_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_map_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__plain_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__plain_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolate_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolate_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolate_string_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__interpolate_string_repeat4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generics_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_qualified_name] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_args = 1,
  field_body = 2,
  field_catch_body = 3,
  field_catch_exception = 4,
  field_collection = 5,
  field_condition = 6,
  field_else = 7,
  field_else_body = 8,
  field_finally_body = 9,
  field_function = 10,
  field_generics = 11,
  field_import = 12,
  field_import_alias = 13,
  field_increment = 14,
  field_initializer = 15,
  field_key = 16,
  field_name = 17,
  field_parameters = 18,
  field_superclass = 19,
  field_then = 20,
  field_type = 21,
  field_value = 22,
  field_variable = 23,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_args] = "args",
  [field_body] = "body",
  [field_catch_body] = "catch_body",
  [field_catch_exception] = "catch_exception",
  [field_collection] = "collection",
  [field_condition] = "condition",
  [field_else] = "else",
  [field_else_body] = "else_body",
  [field_finally_body] = "finally_body",
  [field_function] = "function",
  [field_generics] = "generics",
  [field_import] = "import",
  [field_import_alias] = "import_alias",
  [field_increment] = "increment",
  [field_initializer] = "initializer",
  [field_key] = "key",
  [field_name] = "name",
  [field_parameters] = "parameters",
  [field_superclass] = "superclass",
  [field_then] = "then",
  [field_type] = "type",
  [field_value] = "value",
  [field_variable] = "variable",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [4] = {.index = 2, .length = 1},
  [5] = {.index = 3, .length = 2},
  [7] = {.index = 5, .length = 1},
  [8] = {.index = 6, .length = 2},
  [9] = {.index = 8, .length = 2},
  [10] = {.index = 10, .length = 1},
  [11] = {.index = 11, .length = 2},
  [12] = {.index = 13, .length = 3},
  [13] = {.index = 16, .length = 2},
  [14] = {.index = 18, .length = 1},
  [15] = {.index = 19, .length = 2},
  [16] = {.index = 21, .length = 1},
  [17] = {.index = 22, .length = 2},
  [18] = {.index = 24, .length = 2},
  [19] = {.index = 26, .length = 3},
  [20] = {.index = 29, .length = 2},
  [21] = {.index = 31, .length = 2},
  [22] = {.index = 33, .length = 4},
  [23] = {.index = 37, .length = 2},
  [24] = {.index = 39, .length = 2},
  [25] = {.index = 41, .length = 2},
  [26] = {.index = 43, .length = 3},
  [27] = {.index = 46, .length = 2},
  [28] = {.index = 48, .length = 3},
  [29] = {.index = 51, .length = 1},
  [30] = {.index = 52, .length = 2},
  [31] = {.index = 54, .length = 2},
  [32] = {.index = 56, .length = 3},
  [33] = {.index = 59, .length = 3},
  [34] = {.index = 62, .length = 3},
  [35] = {.index = 65, .length = 3},
  [36] = {.index = 68, .length = 2},
  [37] = {.index = 70, .length = 4},
  [38] = {.index = 74, .length = 3},
  [39] = {.index = 77, .length = 2},
  [40] = {.index = 79, .length = 3},
  [41] = {.index = 82, .length = 1},
  [42] = {.index = 83, .length = 2},
  [43] = {.index = 85, .length = 2},
  [44] = {.index = 87, .length = 4},
  [45] = {.index = 91, .length = 1},
  [46] = {.index = 92, .length = 1},
  [47] = {.index = 93, .length = 1},
  [48] = {.index = 94, .length = 1},
  [49] = {.index = 95, .length = 3},
  [50] = {.index = 98, .length = 3},
  [51] = {.index = 101, .length = 2},
  [52] = {.index = 103, .length = 4},
  [53] = {.index = 107, .length = 3},
  [54] = {.index = 110, .length = 2},
  [55] = {.index = 112, .length = 3},
  [56] = {.index = 115, .length = 2},
  [57] = {.index = 117, .length = 2},
  [58] = {.index = 119, .length = 3},
  [59] = {.index = 122, .length = 2},
  [60] = {.index = 124, .length = 2},
  [61] = {.index = 126, .length = 2},
  [62] = {.index = 128, .length = 3},
  [63] = {.index = 131, .length = 4},
  [64] = {.index = 135, .length = 3},
  [65] = {.index = 138, .length = 3},
  [66] = {.index = 141, .length = 2},
  [67] = {.index = 143, .length = 4},
  [68] = {.index = 147, .length = 3},
  [69] = {.index = 150, .length = 3},
  [70] = {.index = 153, .length = 3},
  [71] = {.index = 156, .length = 3},
  [72] = {.index = 159, .length = 3},
  [73] = {.index = 162, .length = 3},
  [74] = {.index = 165, .length = 4},
  [75] = {.index = 169, .length = 3},
  [76] = {.index = 172, .length = 4},
  [77] = {.index = 176, .length = 4},
  [78] = {.index = 180, .length = 4},
  [79] = {.index = 184, .length = 4},
  [80] = {.index = 188, .length = 4},
  [81] = {.index = 192, .length = 5},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_import, 1},
  [1] =
    {field_name, 0},
  [2] =
    {field_name, 1},
  [3] =
    {field_args, 0, .inherited = true},
    {field_function, 0, .inherited = true},
  [5] =
    {field_body, 1},
  [6] =
    {field_args, 1},
    {field_function, 0},
  [8] =
    {field_name, 1},
    {field_type, 0},
  [10] =
    {field_import, 2},
  [11] =
    {field_body, 2},
    {field_name, 1},
  [13] =
    {field_function, 1},
    {field_parameters, 2},
    {field_type, 0},
  [16] =
    {field_body, 2},
    {field_condition, 1},
  [18] =
    {field_body, 2},
  [19] =
    {field_body, 2},
    {field_value, 1},
  [21] =
    {field_name, 2},
  [22] =
    {field_name, 2},
    {field_type, 1},
  [24] =
    {field_import, 1},
    {field_import_alias, 3},
  [26] =
    {field_body, 3},
    {field_generics, 2},
    {field_name, 1},
  [29] =
    {field_name, 0},
    {field_value, 2},
  [31] =
    {field_name, 1},
    {field_value, 3},
  [33] =
    {field_body, 3},
    {field_function, 1},
    {field_parameters, 2},
    {field_type, 0},
  [37] =
    {field_body, 1},
    {field_condition, 3},
  [39] =
    {field_key, 0},
    {field_value, 2},
  [41] =
    {field_body, 1},
    {field_finally_body, 3},
  [43] =
    {field_name, 1},
    {field_type, 0},
    {field_value, 3},
  [46] =
    {field_body, 3},
    {field_name, 2},
  [48] =
    {field_function, 2},
    {field_parameters, 3},
    {field_type, 1},
  [51] =
    {field_name, 3},
  [52] =
    {field_name, 3},
    {field_type, 2},
  [54] =
    {field_import, 2},
    {field_import_alias, 4},
  [56] =
    {field_body, 4},
    {field_name, 1},
    {field_superclass, 3},
  [59] =
    {field_body, 2},
    {field_condition, 1},
    {field_else_body, 4},
  [62] =
    {field_condition, 0},
    {field_else, 4},
    {field_then, 2},
  [65] =
    {field_body, 4},
    {field_generics, 3},
    {field_name, 2},
  [68] =
    {field_name, 2},
    {field_value, 4},
  [70] =
    {field_body, 4},
    {field_function, 2},
    {field_parameters, 3},
    {field_type, 1},
  [74] =
    {field_name, 2},
    {field_type, 1},
    {field_value, 4},
  [77] =
    {field_body, 4},
    {field_name, 3},
  [79] =
    {field_function, 3},
    {field_parameters, 4},
    {field_type, 2},
  [82] =
    {field_name, 4},
  [83] =
    {field_name, 4},
    {field_type, 3},
  [85] =
    {field_name, 0},
    {field_superclass, 2},
  [87] =
    {field_body, 5},
    {field_generics, 2},
    {field_name, 1},
    {field_superclass, 4},
  [91] =
    {field_increment, 3},
  [92] =
    {field_condition, 2},
  [93] =
    {field_initializer, 1},
  [94] =
    {field_value, 1},
  [95] =
    {field_body, 5},
    {field_name, 2},
    {field_superclass, 4},
  [98] =
    {field_body, 5},
    {field_generics, 4},
    {field_name, 3},
  [101] =
    {field_name, 3},
    {field_value, 5},
  [103] =
    {field_body, 5},
    {field_function, 3},
    {field_parameters, 4},
    {field_type, 2},
  [107] =
    {field_name, 3},
    {field_type, 2},
    {field_value, 5},
  [110] =
    {field_body, 5},
    {field_name, 4},
  [112] =
    {field_function, 4},
    {field_parameters, 5},
    {field_type, 3},
  [115] =
    {field_increment, 3},
    {field_increment, 4},
  [117] =
    {field_condition, 2},
    {field_increment, 4},
  [119] =
    {field_body, 6},
    {field_collection, 4},
    {field_variable, 2},
  [122] =
    {field_increment, 4},
    {field_initializer, 1},
  [124] =
    {field_condition, 3},
    {field_initializer, 1},
  [126] =
    {field_initializer, 1},
    {field_initializer, 2},
  [128] =
    {field_body, 1},
    {field_catch_body, 6},
    {field_catch_exception, 4},
  [131] =
    {field_body, 6},
    {field_generics, 3},
    {field_name, 2},
    {field_superclass, 5},
  [135] =
    {field_body, 6},
    {field_name, 3},
    {field_superclass, 5},
  [138] =
    {field_body, 6},
    {field_generics, 5},
    {field_name, 4},
  [141] =
    {field_name, 4},
    {field_value, 6},
  [143] =
    {field_body, 6},
    {field_function, 4},
    {field_parameters, 5},
    {field_type, 3},
  [147] =
    {field_name, 4},
    {field_type, 3},
    {field_value, 6},
  [150] =
    {field_condition, 2},
    {field_increment, 4},
    {field_increment, 5},
  [153] =
    {field_increment, 4},
    {field_increment, 5},
    {field_initializer, 1},
  [156] =
    {field_condition, 3},
    {field_increment, 5},
    {field_initializer, 1},
  [159] =
    {field_increment, 5},
    {field_initializer, 1},
    {field_initializer, 2},
  [162] =
    {field_condition, 4},
    {field_initializer, 1},
    {field_initializer, 2},
  [165] =
    {field_body, 7},
    {field_generics, 4},
    {field_name, 3},
    {field_superclass, 6},
  [169] =
    {field_body, 7},
    {field_name, 4},
    {field_superclass, 6},
  [172] =
    {field_condition, 3},
    {field_increment, 5},
    {field_increment, 6},
    {field_initializer, 1},
  [176] =
    {field_increment, 5},
    {field_increment, 6},
    {field_initializer, 1},
    {field_initializer, 2},
  [180] =
    {field_condition, 4},
    {field_increment, 6},
    {field_initializer, 1},
    {field_initializer, 2},
  [184] =
    {field_body, 1},
    {field_catch_body, 6},
    {field_catch_exception, 4},
    {field_finally_body, 8},
  [188] =
    {field_body, 8},
    {field_generics, 5},
    {field_name, 4},
    {field_superclass, 7},
  [192] =
    {field_condition, 4},
    {field_increment, 6},
    {field_increment, 7},
    {field_initializer, 1},
    {field_initializer, 2},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [1] = alias_sym_qualified_name,
  },
  [2] = {
    [1] = alias_sym_qualified_name,
  },
  [6] = {
    [0] = sym_escape_sequence,
  },
  [10] = {
    [2] = alias_sym_qualified_name,
  },
  [18] = {
    [1] = alias_sym_qualified_name,
  },
  [31] = {
    [2] = alias_sym_qualified_name,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__import_name, 2,
    sym__import_name,
    alias_sym_qualified_name,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 6,
  [7] = 7,
  [8] = 6,
  [9] = 9,
  [10] = 10,
  [11] = 10,
  [12] = 9,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 7,
  [18] = 15,
  [19] = 13,
  [20] = 16,
  [21] = 21,
  [22] = 21,
  [23] = 14,
  [24] = 7,
  [25] = 6,
  [26] = 15,
  [27] = 9,
  [28] = 16,
  [29] = 10,
  [30] = 21,
  [31] = 10,
  [32] = 6,
  [33] = 9,
  [34] = 15,
  [35] = 13,
  [36] = 21,
  [37] = 14,
  [38] = 16,
  [39] = 39,
  [40] = 7,
  [41] = 13,
  [42] = 14,
  [43] = 43,
  [44] = 43,
  [45] = 43,
  [46] = 43,
  [47] = 43,
  [48] = 43,
  [49] = 43,
  [50] = 43,
  [51] = 43,
  [52] = 43,
  [53] = 43,
  [54] = 43,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 55,
  [60] = 60,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 64,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 67,
  [70] = 70,
  [71] = 64,
  [72] = 66,
  [73] = 67,
  [74] = 68,
  [75] = 70,
  [76] = 64,
  [77] = 66,
  [78] = 67,
  [79] = 68,
  [80] = 70,
  [81] = 64,
  [82] = 66,
  [83] = 67,
  [84] = 68,
  [85] = 70,
  [86] = 70,
  [87] = 64,
  [88] = 66,
  [89] = 67,
  [90] = 68,
  [91] = 70,
  [92] = 64,
  [93] = 66,
  [94] = 67,
  [95] = 68,
  [96] = 70,
  [97] = 64,
  [98] = 66,
  [99] = 67,
  [100] = 68,
  [101] = 64,
  [102] = 70,
  [103] = 64,
  [104] = 66,
  [105] = 67,
  [106] = 68,
  [107] = 66,
  [108] = 70,
  [109] = 64,
  [110] = 66,
  [111] = 67,
  [112] = 68,
  [113] = 70,
  [114] = 64,
  [115] = 66,
  [116] = 67,
  [117] = 68,
  [118] = 68,
  [119] = 70,
  [120] = 66,
  [121] = 67,
  [122] = 68,
  [123] = 70,
  [124] = 124,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 130,
  [134] = 134,
  [135] = 135,
  [136] = 130,
  [137] = 137,
  [138] = 134,
  [139] = 137,
  [140] = 132,
  [141] = 130,
  [142] = 134,
  [143] = 135,
  [144] = 137,
  [145] = 132,
  [146] = 135,
  [147] = 147,
  [148] = 132,
  [149] = 149,
  [150] = 131,
  [151] = 149,
  [152] = 130,
  [153] = 135,
  [154] = 134,
  [155] = 135,
  [156] = 130,
  [157] = 134,
  [158] = 135,
  [159] = 132,
  [160] = 135,
  [161] = 135,
  [162] = 135,
  [163] = 135,
  [164] = 149,
  [165] = 137,
  [166] = 132,
  [167] = 135,
  [168] = 135,
  [169] = 149,
  [170] = 135,
  [171] = 149,
  [172] = 130,
  [173] = 149,
  [174] = 134,
  [175] = 135,
  [176] = 149,
  [177] = 149,
  [178] = 149,
  [179] = 137,
  [180] = 149,
  [181] = 132,
  [182] = 130,
  [183] = 134,
  [184] = 135,
  [185] = 137,
  [186] = 149,
  [187] = 130,
  [188] = 134,
  [189] = 132,
  [190] = 130,
  [191] = 134,
  [192] = 132,
  [193] = 130,
  [194] = 134,
  [195] = 132,
  [196] = 130,
  [197] = 134,
  [198] = 135,
  [199] = 132,
  [200] = 130,
  [201] = 134,
  [202] = 132,
  [203] = 130,
  [204] = 134,
  [205] = 132,
  [206] = 149,
  [207] = 130,
  [208] = 134,
  [209] = 132,
  [210] = 149,
  [211] = 149,
  [212] = 130,
  [213] = 134,
  [214] = 137,
  [215] = 132,
  [216] = 149,
  [217] = 131,
  [218] = 134,
  [219] = 132,
  [220] = 149,
  [221] = 131,
  [222] = 131,
  [223] = 131,
  [224] = 137,
  [225] = 2,
  [226] = 226,
  [227] = 226,
  [228] = 226,
  [229] = 226,
  [230] = 226,
  [231] = 226,
  [232] = 2,
  [233] = 226,
  [234] = 226,
  [235] = 226,
  [236] = 226,
  [237] = 226,
  [238] = 226,
  [239] = 226,
  [240] = 226,
  [241] = 226,
  [242] = 226,
  [243] = 226,
  [244] = 226,
  [245] = 226,
  [246] = 226,
  [247] = 2,
  [248] = 14,
  [249] = 249,
  [250] = 250,
  [251] = 249,
  [252] = 250,
  [253] = 253,
  [254] = 254,
  [255] = 255,
  [256] = 254,
  [257] = 257,
  [258] = 257,
  [259] = 259,
  [260] = 259,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 261,
  [269] = 9,
  [270] = 270,
  [271] = 271,
  [272] = 263,
  [273] = 265,
  [274] = 274,
  [275] = 275,
  [276] = 264,
  [277] = 277,
  [278] = 266,
  [279] = 270,
  [280] = 280,
  [281] = 267,
  [282] = 280,
  [283] = 2,
  [284] = 284,
  [285] = 285,
  [286] = 9,
  [287] = 10,
  [288] = 274,
  [289] = 271,
  [290] = 270,
  [291] = 10,
  [292] = 275,
  [293] = 280,
  [294] = 284,
  [295] = 262,
  [296] = 255,
  [297] = 277,
  [298] = 285,
  [299] = 277,
  [300] = 262,
  [301] = 253,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 249,
  [306] = 306,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 310,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 250,
  [317] = 317,
  [318] = 265,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 306,
  [325] = 315,
  [326] = 7,
  [327] = 267,
  [328] = 9,
  [329] = 303,
  [330] = 275,
  [331] = 265,
  [332] = 14,
  [333] = 321,
  [334] = 320,
  [335] = 317,
  [336] = 307,
  [337] = 308,
  [338] = 319,
  [339] = 302,
  [340] = 274,
  [341] = 313,
  [342] = 314,
  [343] = 322,
  [344] = 254,
  [345] = 323,
  [346] = 285,
  [347] = 14,
  [348] = 348,
  [349] = 15,
  [350] = 310,
  [351] = 16,
  [352] = 302,
  [353] = 353,
  [354] = 21,
  [355] = 14,
  [356] = 309,
  [357] = 7,
  [358] = 358,
  [359] = 320,
  [360] = 312,
  [361] = 358,
  [362] = 16,
  [363] = 14,
  [364] = 304,
  [365] = 365,
  [366] = 348,
  [367] = 15,
  [368] = 311,
  [369] = 353,
  [370] = 21,
  [371] = 266,
  [372] = 257,
  [373] = 259,
  [374] = 265,
  [375] = 365,
  [376] = 10,
  [377] = 353,
  [378] = 365,
  [379] = 306,
  [380] = 310,
  [381] = 321,
  [382] = 249,
  [383] = 284,
  [384] = 250,
  [385] = 348,
  [386] = 257,
  [387] = 387,
  [388] = 274,
  [389] = 284,
  [390] = 14,
  [391] = 285,
  [392] = 392,
  [393] = 275,
  [394] = 394,
  [395] = 15,
  [396] = 259,
  [397] = 397,
  [398] = 398,
  [399] = 261,
  [400] = 263,
  [401] = 264,
  [402] = 266,
  [403] = 267,
  [404] = 404,
  [405] = 405,
  [406] = 406,
  [407] = 407,
  [408] = 408,
  [409] = 409,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 307,
  [418] = 418,
  [419] = 308,
  [420] = 420,
  [421] = 255,
  [422] = 422,
  [423] = 7,
  [424] = 16,
  [425] = 21,
  [426] = 426,
  [427] = 427,
  [428] = 270,
  [429] = 280,
  [430] = 262,
  [431] = 277,
  [432] = 9,
  [433] = 10,
  [434] = 309,
  [435] = 426,
  [436] = 392,
  [437] = 358,
  [438] = 394,
  [439] = 303,
  [440] = 397,
  [441] = 427,
  [442] = 398,
  [443] = 265,
  [444] = 313,
  [445] = 314,
  [446] = 253,
  [447] = 404,
  [448] = 405,
  [449] = 406,
  [450] = 407,
  [451] = 408,
  [452] = 409,
  [453] = 410,
  [454] = 411,
  [455] = 412,
  [456] = 413,
  [457] = 414,
  [458] = 415,
  [459] = 416,
  [460] = 271,
  [461] = 418,
  [462] = 420,
  [463] = 322,
  [464] = 323,
  [465] = 422,
  [466] = 304,
  [467] = 255,
  [468] = 254,
  [469] = 261,
  [470] = 312,
  [471] = 263,
  [472] = 264,
  [473] = 253,
  [474] = 271,
  [475] = 387,
  [476] = 476,
  [477] = 306,
  [478] = 405,
  [479] = 310,
  [480] = 387,
  [481] = 15,
  [482] = 16,
  [483] = 483,
  [484] = 21,
  [485] = 406,
  [486] = 486,
  [487] = 407,
  [488] = 408,
  [489] = 409,
  [490] = 410,
  [491] = 491,
  [492] = 411,
  [493] = 412,
  [494] = 494,
  [495] = 413,
  [496] = 7,
  [497] = 14,
  [498] = 414,
  [499] = 415,
  [500] = 14,
  [501] = 501,
  [502] = 416,
  [503] = 503,
  [504] = 504,
  [505] = 307,
  [506] = 313,
  [507] = 314,
  [508] = 392,
  [509] = 509,
  [510] = 476,
  [511] = 308,
  [512] = 512,
  [513] = 513,
  [514] = 358,
  [515] = 515,
  [516] = 397,
  [517] = 517,
  [518] = 518,
  [519] = 322,
  [520] = 315,
  [521] = 483,
  [522] = 486,
  [523] = 501,
  [524] = 323,
  [525] = 317,
  [526] = 319,
  [527] = 515,
  [528] = 509,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 535,
  [536] = 309,
  [537] = 491,
  [538] = 404,
  [539] = 494,
  [540] = 503,
  [541] = 303,
  [542] = 512,
  [543] = 398,
  [544] = 517,
  [545] = 513,
  [546] = 518,
  [547] = 304,
  [548] = 311,
  [549] = 529,
  [550] = 530,
  [551] = 422,
  [552] = 531,
  [553] = 532,
  [554] = 533,
  [555] = 534,
  [556] = 535,
  [557] = 320,
  [558] = 311,
  [559] = 265,
  [560] = 504,
  [561] = 365,
  [562] = 348,
  [563] = 312,
  [564] = 353,
  [565] = 394,
  [566] = 265,
  [567] = 315,
  [568] = 317,
  [569] = 319,
  [570] = 302,
  [571] = 321,
  [572] = 418,
  [573] = 397,
  [574] = 398,
  [575] = 404,
  [576] = 405,
  [577] = 406,
  [578] = 407,
  [579] = 408,
  [580] = 409,
  [581] = 410,
  [582] = 411,
  [583] = 412,
  [584] = 413,
  [585] = 414,
  [586] = 415,
  [587] = 416,
  [588] = 427,
  [589] = 422,
  [590] = 418,
  [591] = 392,
  [592] = 420,
  [593] = 420,
  [594] = 426,
  [595] = 394,
  [596] = 426,
  [597] = 387,
  [598] = 427,
  [599] = 517,
  [600] = 486,
  [601] = 501,
  [602] = 494,
  [603] = 533,
  [604] = 491,
  [605] = 531,
  [606] = 509,
  [607] = 515,
  [608] = 426,
  [609] = 503,
  [610] = 427,
  [611] = 532,
  [612] = 533,
  [613] = 517,
  [614] = 426,
  [615] = 427,
  [616] = 534,
  [617] = 518,
  [618] = 483,
  [619] = 518,
  [620] = 535,
  [621] = 531,
  [622] = 532,
  [623] = 529,
  [624] = 494,
  [625] = 491,
  [626] = 529,
  [627] = 512,
  [628] = 530,
  [629] = 504,
  [630] = 512,
  [631] = 513,
  [632] = 513,
  [633] = 534,
  [634] = 509,
  [635] = 530,
  [636] = 483,
  [637] = 486,
  [638] = 501,
  [639] = 515,
  [640] = 503,
  [641] = 535,
  [642] = 504,
  [643] = 643,
  [644] = 644,
  [645] = 512,
  [646] = 513,
  [647] = 647,
  [648] = 513,
  [649] = 512,
  [650] = 494,
  [651] = 647,
  [652] = 494,
  [653] = 653,
  [654] = 654,
  [655] = 655,
  [656] = 656,
  [657] = 657,
  [658] = 657,
  [659] = 657,
  [660] = 9,
  [661] = 10,
  [662] = 657,
  [663] = 6,
  [664] = 21,
  [665] = 15,
  [666] = 14,
  [667] = 7,
  [668] = 16,
  [669] = 669,
  [670] = 669,
  [671] = 669,
  [672] = 669,
  [673] = 669,
  [674] = 669,
  [675] = 675,
  [676] = 676,
  [677] = 677,
  [678] = 678,
  [679] = 679,
  [680] = 680,
  [681] = 681,
  [682] = 682,
  [683] = 683,
  [684] = 684,
  [685] = 685,
  [686] = 686,
  [687] = 687,
  [688] = 688,
  [689] = 689,
  [690] = 680,
  [691] = 691,
  [692] = 692,
  [693] = 693,
  [694] = 694,
  [695] = 695,
  [696] = 679,
  [697] = 677,
  [698] = 698,
  [699] = 699,
  [700] = 700,
  [701] = 701,
  [702] = 702,
  [703] = 703,
  [704] = 704,
  [705] = 705,
  [706] = 14,
  [707] = 707,
  [708] = 708,
  [709] = 709,
  [710] = 683,
  [711] = 711,
  [712] = 712,
  [713] = 691,
  [714] = 675,
  [715] = 715,
  [716] = 716,
  [717] = 717,
  [718] = 718,
  [719] = 683,
  [720] = 720,
  [721] = 721,
  [722] = 722,
  [723] = 709,
  [724] = 724,
  [725] = 725,
  [726] = 726,
  [727] = 727,
  [728] = 14,
  [729] = 729,
  [730] = 730,
  [731] = 731,
  [732] = 732,
  [733] = 733,
  [734] = 734,
  [735] = 735,
  [736] = 682,
  [737] = 686,
  [738] = 738,
  [739] = 739,
  [740] = 685,
  [741] = 741,
  [742] = 742,
  [743] = 743,
  [744] = 678,
  [745] = 745,
  [746] = 715,
  [747] = 747,
  [748] = 748,
  [749] = 749,
  [750] = 750,
  [751] = 741,
  [752] = 716,
  [753] = 749,
  [754] = 679,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 758,
  [759] = 759,
  [760] = 760,
  [761] = 761,
  [762] = 762,
  [763] = 720,
  [764] = 760,
  [765] = 761,
  [766] = 766,
  [767] = 759,
  [768] = 766,
  [769] = 769,
  [770] = 770,
  [771] = 758,
  [772] = 762,
  [773] = 769,
  [774] = 770,
  [775] = 755,
  [776] = 718,
  [777] = 701,
  [778] = 705,
  [779] = 716,
  [780] = 310,
  [781] = 709,
  [782] = 684,
  [783] = 731,
  [784] = 784,
  [785] = 321,
  [786] = 741,
  [787] = 749,
  [788] = 733,
  [789] = 680,
  [790] = 677,
  [791] = 750,
  [792] = 718,
  [793] = 681,
  [794] = 755,
  [795] = 716,
  [796] = 687,
  [797] = 682,
  [798] = 688,
  [799] = 683,
  [800] = 680,
  [801] = 677,
  [802] = 698,
  [803] = 711,
  [804] = 686,
  [805] = 718,
  [806] = 695,
  [807] = 689,
  [808] = 742,
  [809] = 678,
  [810] = 747,
  [811] = 693,
  [812] = 684,
  [813] = 703,
  [814] = 694,
  [815] = 707,
  [816] = 699,
  [817] = 703,
  [818] = 721,
  [819] = 7,
  [820] = 676,
  [821] = 722,
  [822] = 681,
  [823] = 724,
  [824] = 302,
  [825] = 306,
  [826] = 725,
  [827] = 726,
  [828] = 310,
  [829] = 321,
  [830] = 727,
  [831] = 707,
  [832] = 7,
  [833] = 700,
  [834] = 702,
  [835] = 689,
  [836] = 712,
  [837] = 693,
  [838] = 679,
  [839] = 694,
  [840] = 699,
  [841] = 14,
  [842] = 700,
  [843] = 702,
  [844] = 712,
  [845] = 730,
  [846] = 732,
  [847] = 757,
  [848] = 717,
  [849] = 735,
  [850] = 691,
  [851] = 738,
  [852] = 685,
  [853] = 739,
  [854] = 734,
  [855] = 743,
  [856] = 756,
  [857] = 692,
  [858] = 701,
  [859] = 705,
  [860] = 731,
  [861] = 733,
  [862] = 750,
  [863] = 745,
  [864] = 748,
  [865] = 687,
  [866] = 688,
  [867] = 675,
  [868] = 695,
  [869] = 698,
  [870] = 711,
  [871] = 757,
  [872] = 721,
  [873] = 722,
  [874] = 724,
  [875] = 725,
  [876] = 726,
  [877] = 727,
  [878] = 695,
  [879] = 729,
  [880] = 675,
  [881] = 730,
  [882] = 732,
  [883] = 735,
  [884] = 717,
  [885] = 738,
  [886] = 739,
  [887] = 743,
  [888] = 745,
  [889] = 748,
  [890] = 708,
  [891] = 715,
  [892] = 676,
  [893] = 893,
  [894] = 691,
  [895] = 742,
  [896] = 302,
  [897] = 718,
  [898] = 678,
  [899] = 704,
  [900] = 715,
  [901] = 720,
  [902] = 747,
  [903] = 685,
  [904] = 734,
  [905] = 741,
  [906] = 749,
  [907] = 709,
  [908] = 756,
  [909] = 306,
  [910] = 692,
  [911] = 682,
  [912] = 686,
  [913] = 14,
  [914] = 759,
  [915] = 766,
  [916] = 770,
  [917] = 762,
  [918] = 708,
  [919] = 758,
  [920] = 920,
  [921] = 759,
  [922] = 758,
  [923] = 770,
  [924] = 760,
  [925] = 925,
  [926] = 720,
  [927] = 761,
  [928] = 760,
  [929] = 761,
  [930] = 769,
  [931] = 769,
  [932] = 704,
  [933] = 729,
  [934] = 766,
  [935] = 729,
  [936] = 762,
  [937] = 704,
  [938] = 784,
  [939] = 705,
  [940] = 731,
  [941] = 733,
  [942] = 321,
  [943] = 708,
  [944] = 757,
  [945] = 750,
  [946] = 718,
  [947] = 718,
  [948] = 717,
  [949] = 755,
  [950] = 734,
  [951] = 306,
  [952] = 7,
  [953] = 687,
  [954] = 688,
  [955] = 745,
  [956] = 712,
  [957] = 698,
  [958] = 302,
  [959] = 306,
  [960] = 711,
  [961] = 310,
  [962] = 700,
  [963] = 321,
  [964] = 704,
  [965] = 721,
  [966] = 722,
  [967] = 702,
  [968] = 724,
  [969] = 310,
  [970] = 743,
  [971] = 742,
  [972] = 756,
  [973] = 747,
  [974] = 725,
  [975] = 684,
  [976] = 703,
  [977] = 726,
  [978] = 707,
  [979] = 727,
  [980] = 676,
  [981] = 893,
  [982] = 7,
  [983] = 699,
  [984] = 748,
  [985] = 692,
  [986] = 708,
  [987] = 730,
  [988] = 681,
  [989] = 701,
  [990] = 7,
  [991] = 302,
  [992] = 718,
  [993] = 732,
  [994] = 735,
  [995] = 738,
  [996] = 689,
  [997] = 693,
  [998] = 739,
  [999] = 694,
  [1000] = 13,
  [1001] = 708,
  [1002] = 704,
  [1003] = 1003,
  [1004] = 1003,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 1018,
  [1019] = 1019,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 657,
  [1026] = 280,
  [1027] = 277,
  [1028] = 262,
  [1029] = 270,
  [1030] = 284,
  [1031] = 9,
  [1032] = 275,
  [1033] = 10,
  [1034] = 1034,
  [1035] = 1034,
  [1036] = 321,
  [1037] = 312,
  [1038] = 280,
  [1039] = 262,
  [1040] = 365,
  [1041] = 1034,
  [1042] = 353,
  [1043] = 270,
  [1044] = 21,
  [1045] = 265,
  [1046] = 1034,
  [1047] = 277,
  [1048] = 15,
  [1049] = 7,
  [1050] = 320,
  [1051] = 1034,
  [1052] = 310,
  [1053] = 304,
  [1054] = 348,
  [1055] = 302,
  [1056] = 14,
  [1057] = 16,
  [1058] = 306,
  [1059] = 1034,
  [1060] = 314,
  [1061] = 322,
  [1062] = 323,
  [1063] = 284,
  [1064] = 308,
  [1065] = 358,
  [1066] = 267,
  [1067] = 257,
  [1068] = 259,
  [1069] = 274,
  [1070] = 313,
  [1071] = 307,
  [1072] = 303,
  [1073] = 266,
  [1074] = 309,
  [1075] = 255,
  [1076] = 285,
  [1077] = 9,
  [1078] = 275,
  [1079] = 10,
  [1080] = 1080,
  [1081] = 14,
  [1082] = 15,
  [1083] = 1080,
  [1084] = 16,
  [1085] = 1080,
  [1086] = 1080,
  [1087] = 310,
  [1088] = 254,
  [1089] = 1080,
  [1090] = 304,
  [1091] = 14,
  [1092] = 302,
  [1093] = 312,
  [1094] = 1080,
  [1095] = 348,
  [1096] = 306,
  [1097] = 1080,
  [1098] = 353,
  [1099] = 320,
  [1100] = 1080,
  [1101] = 321,
  [1102] = 7,
  [1103] = 365,
  [1104] = 21,
  [1105] = 1105,
  [1106] = 1106,
  [1107] = 266,
  [1108] = 249,
  [1109] = 285,
  [1110] = 1110,
  [1111] = 1106,
  [1112] = 303,
  [1113] = 358,
  [1114] = 1106,
  [1115] = 257,
  [1116] = 1105,
  [1117] = 259,
  [1118] = 309,
  [1119] = 1106,
  [1120] = 307,
  [1121] = 1121,
  [1122] = 1105,
  [1123] = 1121,
  [1124] = 313,
  [1125] = 308,
  [1126] = 1121,
  [1127] = 314,
  [1128] = 1105,
  [1129] = 1105,
  [1130] = 1121,
  [1131] = 1106,
  [1132] = 1132,
  [1133] = 274,
  [1134] = 1105,
  [1135] = 1121,
  [1136] = 1105,
  [1137] = 1121,
  [1138] = 249,
  [1139] = 1121,
  [1140] = 1105,
  [1141] = 1121,
  [1142] = 1105,
  [1143] = 1105,
  [1144] = 1105,
  [1145] = 1105,
  [1146] = 1105,
  [1147] = 1105,
  [1148] = 1105,
  [1149] = 1105,
  [1150] = 1105,
  [1151] = 322,
  [1152] = 253,
  [1153] = 323,
  [1154] = 1106,
  [1155] = 267,
  [1156] = 426,
  [1157] = 315,
  [1158] = 1158,
  [1159] = 317,
  [1160] = 1160,
  [1161] = 1161,
  [1162] = 261,
  [1163] = 261,
  [1164] = 271,
  [1165] = 427,
  [1166] = 1166,
  [1167] = 263,
  [1168] = 1158,
  [1169] = 250,
  [1170] = 1158,
  [1171] = 1158,
  [1172] = 263,
  [1173] = 250,
  [1174] = 265,
  [1175] = 1158,
  [1176] = 1158,
  [1177] = 1177,
  [1178] = 1178,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 319,
  [1185] = 1185,
  [1186] = 1186,
  [1187] = 1187,
  [1188] = 1188,
  [1189] = 422,
  [1190] = 1190,
  [1191] = 1191,
  [1192] = 1192,
  [1193] = 1193,
  [1194] = 1194,
  [1195] = 1177,
  [1196] = 1196,
  [1197] = 1197,
  [1198] = 1198,
  [1199] = 1199,
  [1200] = 1200,
  [1201] = 1201,
  [1202] = 1202,
  [1203] = 1203,
  [1204] = 1204,
  [1205] = 1205,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1178,
  [1209] = 1179,
  [1210] = 1190,
  [1211] = 1191,
  [1212] = 1181,
  [1213] = 1192,
  [1214] = 1193,
  [1215] = 1194,
  [1216] = 1196,
  [1217] = 1197,
  [1218] = 1198,
  [1219] = 1199,
  [1220] = 1220,
  [1221] = 1200,
  [1222] = 1201,
  [1223] = 1202,
  [1224] = 1203,
  [1225] = 1183,
  [1226] = 1181,
  [1227] = 1227,
  [1228] = 1185,
  [1229] = 1187,
  [1230] = 1185,
  [1231] = 1187,
  [1232] = 1232,
  [1233] = 1204,
  [1234] = 1205,
  [1235] = 1235,
  [1236] = 1178,
  [1237] = 1179,
  [1238] = 1238,
  [1239] = 1190,
  [1240] = 1191,
  [1241] = 1192,
  [1242] = 1193,
  [1243] = 1194,
  [1244] = 1177,
  [1245] = 1196,
  [1246] = 1197,
  [1247] = 1198,
  [1248] = 1199,
  [1249] = 1200,
  [1250] = 1201,
  [1251] = 1202,
  [1252] = 1203,
  [1253] = 1206,
  [1254] = 1220,
  [1255] = 1204,
  [1256] = 1205,
  [1257] = 1257,
  [1258] = 1181,
  [1259] = 1183,
  [1260] = 1185,
  [1261] = 1261,
  [1262] = 1187,
  [1263] = 1178,
  [1264] = 1179,
  [1265] = 1206,
  [1266] = 1204,
  [1267] = 1205,
  [1268] = 1268,
  [1269] = 1186,
  [1270] = 1235,
  [1271] = 1178,
  [1272] = 1179,
  [1273] = 1273,
  [1274] = 1273,
  [1275] = 1268,
  [1276] = 1186,
  [1277] = 1235,
  [1278] = 1183,
  [1279] = 1273,
  [1280] = 1268,
  [1281] = 1186,
  [1282] = 1235,
  [1283] = 387,
  [1284] = 1273,
  [1285] = 1183,
  [1286] = 1268,
  [1287] = 1186,
  [1288] = 1235,
  [1289] = 1190,
  [1290] = 1191,
  [1291] = 1192,
  [1292] = 1220,
  [1293] = 1194,
  [1294] = 1177,
  [1295] = 1196,
  [1296] = 1197,
  [1297] = 1198,
  [1298] = 1199,
  [1299] = 1200,
  [1300] = 1201,
  [1301] = 1202,
  [1302] = 1203,
  [1303] = 249,
  [1304] = 1268,
  [1305] = 392,
  [1306] = 1206,
  [1307] = 394,
  [1308] = 397,
  [1309] = 1227,
  [1310] = 1232,
  [1311] = 1257,
  [1312] = 1188,
  [1313] = 1190,
  [1314] = 1191,
  [1315] = 1192,
  [1316] = 1193,
  [1317] = 1194,
  [1318] = 1177,
  [1319] = 264,
  [1320] = 1196,
  [1321] = 1197,
  [1322] = 1198,
  [1323] = 1199,
  [1324] = 1220,
  [1325] = 1200,
  [1326] = 1201,
  [1327] = 1202,
  [1328] = 1203,
  [1329] = 398,
  [1330] = 264,
  [1331] = 249,
  [1332] = 1181,
  [1333] = 1227,
  [1334] = 1232,
  [1335] = 1257,
  [1336] = 1188,
  [1337] = 1220,
  [1338] = 1206,
  [1339] = 1185,
  [1340] = 1227,
  [1341] = 1232,
  [1342] = 1257,
  [1343] = 1188,
  [1344] = 1187,
  [1345] = 1227,
  [1346] = 1232,
  [1347] = 1188,
  [1348] = 1232,
  [1349] = 1188,
  [1350] = 1273,
  [1351] = 1232,
  [1352] = 1188,
  [1353] = 1232,
  [1354] = 1188,
  [1355] = 1232,
  [1356] = 1188,
  [1357] = 404,
  [1358] = 1232,
  [1359] = 1188,
  [1360] = 405,
  [1361] = 1232,
  [1362] = 1188,
  [1363] = 406,
  [1364] = 1232,
  [1365] = 1188,
  [1366] = 407,
  [1367] = 1188,
  [1368] = 408,
  [1369] = 1188,
  [1370] = 409,
  [1371] = 1188,
  [1372] = 410,
  [1373] = 411,
  [1374] = 412,
  [1375] = 413,
  [1376] = 1204,
  [1377] = 414,
  [1378] = 1205,
  [1379] = 415,
  [1380] = 416,
  [1381] = 1188,
  [1382] = 1193,
  [1383] = 250,
  [1384] = 250,
  [1385] = 1385,
  [1386] = 261,
  [1387] = 263,
  [1388] = 261,
  [1389] = 263,
  [1390] = 264,
  [1391] = 427,
  [1392] = 1392,
  [1393] = 1393,
  [1394] = 426,
  [1395] = 255,
  [1396] = 264,
  [1397] = 418,
  [1398] = 1398,
  [1399] = 533,
  [1400] = 530,
  [1401] = 1385,
  [1402] = 535,
  [1403] = 504,
  [1404] = 509,
  [1405] = 476,
  [1406] = 503,
  [1407] = 311,
  [1408] = 1385,
  [1409] = 494,
  [1410] = 311,
  [1411] = 420,
  [1412] = 534,
  [1413] = 483,
  [1414] = 486,
  [1415] = 1385,
  [1416] = 501,
  [1417] = 515,
  [1418] = 518,
  [1419] = 531,
  [1420] = 532,
  [1421] = 491,
  [1422] = 1385,
  [1423] = 512,
  [1424] = 513,
  [1425] = 517,
  [1426] = 529,
  [1427] = 284,
  [1428] = 1393,
  [1429] = 1393,
  [1430] = 284,
  [1431] = 1393,
  [1432] = 284,
  [1433] = 1393,
  [1434] = 284,
  [1435] = 311,
  [1436] = 265,
  [1437] = 265,
  [1438] = 311,
  [1439] = 265,
  [1440] = 265,
  [1441] = 512,
  [1442] = 513,
  [1443] = 494,
  [1444] = 1444,
  [1445] = 1445,
  [1446] = 1446,
  [1447] = 647,
  [1448] = 1448,
  [1449] = 1449,
  [1450] = 1448,
  [1451] = 1451,
  [1452] = 1449,
  [1453] = 1449,
  [1454] = 1451,
  [1455] = 1449,
  [1456] = 1449,
  [1457] = 1451,
  [1458] = 1451,
  [1459] = 1451,
  [1460] = 1451,
  [1461] = 1461,
  [1462] = 1462,
  [1463] = 1462,
  [1464] = 1462,
  [1465] = 1448,
  [1466] = 1448,
  [1467] = 1449,
  [1468] = 1451,
  [1469] = 1449,
  [1470] = 1448,
  [1471] = 1451,
  [1472] = 1448,
  [1473] = 1462,
  [1474] = 1462,
  [1475] = 1462,
  [1476] = 1449,
  [1477] = 1477,
  [1478] = 1478,
  [1479] = 1479,
  [1480] = 1478,
  [1481] = 1479,
  [1482] = 1479,
  [1483] = 1479,
  [1484] = 1479,
  [1485] = 1479,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1479,
  [1489] = 1479,
  [1490] = 1479,
  [1491] = 1491,
  [1492] = 1478,
  [1493] = 1487,
  [1494] = 1479,
  [1495] = 1486,
  [1496] = 1496,
  [1497] = 1478,
  [1498] = 1491,
  [1499] = 1478,
  [1500] = 1479,
  [1501] = 1491,
  [1502] = 1479,
  [1503] = 1478,
  [1504] = 1486,
  [1505] = 1486,
  [1506] = 1506,
  [1507] = 1491,
  [1508] = 1477,
  [1509] = 1486,
  [1510] = 1487,
  [1511] = 1487,
  [1512] = 1478,
  [1513] = 1479,
  [1514] = 1514,
  [1515] = 1478,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1491,
  [1519] = 1478,
  [1520] = 1478,
  [1521] = 1478,
  [1522] = 1486,
  [1523] = 1479,
  [1524] = 1486,
  [1525] = 1525,
  [1526] = 1477,
  [1527] = 1486,
  [1528] = 1486,
  [1529] = 1478,
  [1530] = 1479,
  [1531] = 1486,
  [1532] = 1486,
  [1533] = 1477,
  [1534] = 1478,
  [1535] = 1479,
  [1536] = 1491,
  [1537] = 1486,
  [1538] = 1478,
  [1539] = 1478,
  [1540] = 1478,
  [1541] = 1478,
  [1542] = 1477,
  [1543] = 1543,
  [1544] = 1543,
  [1545] = 1543,
  [1546] = 1543,
  [1547] = 1543,
  [1548] = 1548,
  [1549] = 1549,
  [1550] = 1550,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1551,
  [1554] = 1549,
  [1555] = 1555,
  [1556] = 1551,
  [1557] = 1557,
  [1558] = 1558,
  [1559] = 1551,
  [1560] = 1552,
  [1561] = 1557,
  [1562] = 1562,
  [1563] = 1563,
  [1564] = 1550,
  [1565] = 1549,
  [1566] = 1555,
  [1567] = 1558,
  [1568] = 1551,
  [1569] = 1562,
  [1570] = 1552,
  [1571] = 1557,
  [1572] = 1562,
  [1573] = 1563,
  [1574] = 1550,
  [1575] = 1549,
  [1576] = 1555,
  [1577] = 1558,
  [1578] = 1552,
  [1579] = 1557,
  [1580] = 1562,
  [1581] = 1563,
  [1582] = 1550,
  [1583] = 1555,
  [1584] = 1558,
  [1585] = 1552,
  [1586] = 1557,
  [1587] = 1562,
  [1588] = 1563,
  [1589] = 1550,
  [1590] = 1549,
  [1591] = 1555,
  [1592] = 1558,
  [1593] = 1552,
  [1594] = 1557,
  [1595] = 1562,
  [1596] = 1563,
  [1597] = 1550,
  [1598] = 1549,
  [1599] = 1555,
  [1600] = 1558,
  [1601] = 1551,
  [1602] = 1563,
  [1603] = 1603,
  [1604] = 1603,
  [1605] = 1605,
  [1606] = 1603,
  [1607] = 1603,
  [1608] = 1603,
  [1609] = 1605,
  [1610] = 1605,
  [1611] = 1605,
  [1612] = 1605,
  [1613] = 1605,
  [1614] = 1603,
  [1615] = 1603,
  [1616] = 1603,
  [1617] = 1603,
  [1618] = 1603,
  [1619] = 1603,
  [1620] = 1603,
  [1621] = 1603,
  [1622] = 1603,
  [1623] = 1603,
  [1624] = 1603,
  [1625] = 1625,
  [1626] = 1626,
  [1627] = 1626,
  [1628] = 1625,
  [1629] = 1629,
  [1630] = 1625,
  [1631] = 1631,
  [1632] = 1626,
  [1633] = 1626,
  [1634] = 1625,
  [1635] = 1629,
  [1636] = 1631,
  [1637] = 1626,
  [1638] = 1631,
  [1639] = 1631,
  [1640] = 1629,
  [1641] = 1625,
  [1642] = 1629,
  [1643] = 1631,
  [1644] = 1626,
  [1645] = 1629,
  [1646] = 1625,
  [1647] = 1631,
  [1648] = 1626,
  [1649] = 1649,
  [1650] = 1631,
  [1651] = 1626,
  [1652] = 1629,
  [1653] = 1631,
  [1654] = 1654,
  [1655] = 1655,
  [1656] = 1656,
  [1657] = 1656,
  [1658] = 1658,
  [1659] = 1658,
  [1660] = 1660,
  [1661] = 1656,
  [1662] = 1656,
  [1663] = 1658,
  [1664] = 1660,
  [1665] = 1658,
  [1666] = 1658,
  [1667] = 1658,
  [1668] = 1656,
  [1669] = 1656,
  [1670] = 1660,
  [1671] = 1660,
  [1672] = 1660,
  [1673] = 1660,
  [1674] = 1674,
  [1675] = 1674,
  [1676] = 1674,
  [1677] = 1677,
  [1678] = 1677,
  [1679] = 1677,
  [1680] = 1674,
  [1681] = 1677,
  [1682] = 1674,
  [1683] = 1677,
  [1684] = 1674,
  [1685] = 1677,
  [1686] = 1686,
  [1687] = 262,
  [1688] = 280,
  [1689] = 277,
  [1690] = 1690,
  [1691] = 270,
  [1692] = 1692,
  [1693] = 1693,
  [1694] = 1693,
  [1695] = 1695,
  [1696] = 1693,
  [1697] = 1693,
  [1698] = 1698,
  [1699] = 1699,
  [1700] = 1693,
  [1701] = 1693,
  [1702] = 320,
  [1703] = 321,
  [1704] = 1704,
  [1705] = 1705,
  [1706] = 348,
  [1707] = 302,
  [1708] = 1708,
  [1709] = 306,
  [1710] = 365,
  [1711] = 1711,
  [1712] = 310,
  [1713] = 353,
  [1714] = 270,
  [1715] = 1690,
  [1716] = 280,
  [1717] = 9,
  [1718] = 1718,
  [1719] = 1686,
  [1720] = 277,
  [1721] = 262,
  [1722] = 10,
  [1723] = 16,
  [1724] = 14,
  [1725] = 1725,
  [1726] = 1725,
  [1727] = 13,
  [1728] = 15,
  [1729] = 1725,
  [1730] = 21,
  [1731] = 1725,
  [1732] = 7,
  [1733] = 302,
  [1734] = 365,
  [1735] = 348,
  [1736] = 306,
  [1737] = 1737,
  [1738] = 1738,
  [1739] = 353,
  [1740] = 1711,
  [1741] = 1741,
  [1742] = 1708,
  [1743] = 1741,
  [1744] = 1741,
  [1745] = 321,
  [1746] = 310,
  [1747] = 320,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1741,
  [1751] = 1718,
  [1752] = 1752,
  [1753] = 1753,
  [1754] = 1753,
  [1755] = 1749,
  [1756] = 1753,
  [1757] = 1757,
  [1758] = 1753,
  [1759] = 1753,
  [1760] = 1757,
  [1761] = 1748,
  [1762] = 1762,
  [1763] = 1757,
  [1764] = 1757,
  [1765] = 1765,
  [1766] = 1753,
  [1767] = 1757,
  [1768] = 1768,
  [1769] = 1757,
  [1770] = 1752,
  [1771] = 1771,
  [1772] = 1772,
  [1773] = 1772,
  [1774] = 1774,
  [1775] = 1775,
  [1776] = 1774,
  [1777] = 1777,
  [1778] = 1774,
  [1779] = 1779,
  [1780] = 1777,
  [1781] = 1777,
  [1782] = 1774,
  [1783] = 1777,
  [1784] = 1784,
  [1785] = 741,
  [1786] = 749,
  [1787] = 1787,
  [1788] = 680,
  [1789] = 677,
  [1790] = 682,
  [1791] = 675,
  [1792] = 715,
  [1793] = 1793,
  [1794] = 686,
  [1795] = 769,
  [1796] = 766,
  [1797] = 679,
  [1798] = 1798,
  [1799] = 1799,
  [1800] = 770,
  [1801] = 761,
  [1802] = 691,
  [1803] = 1803,
  [1804] = 709,
  [1805] = 683,
  [1806] = 1806,
  [1807] = 741,
  [1808] = 749,
  [1809] = 1809,
  [1810] = 680,
  [1811] = 677,
  [1812] = 758,
  [1813] = 675,
  [1814] = 1809,
  [1815] = 682,
  [1816] = 760,
  [1817] = 1809,
  [1818] = 715,
  [1819] = 762,
  [1820] = 1809,
  [1821] = 685,
  [1822] = 759,
  [1823] = 1809,
  [1824] = 1809,
  [1825] = 686,
  [1826] = 1826,
  [1827] = 1827,
  [1828] = 766,
  [1829] = 678,
  [1830] = 1830,
  [1831] = 695,
  [1832] = 1826,
  [1833] = 1833,
  [1834] = 9,
  [1835] = 1835,
  [1836] = 678,
  [1837] = 716,
  [1838] = 1826,
  [1839] = 1826,
  [1840] = 1835,
  [1841] = 770,
  [1842] = 1830,
  [1843] = 1835,
  [1844] = 1830,
  [1845] = 1833,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 716,
  [1849] = 1830,
  [1850] = 1830,
  [1851] = 1830,
  [1852] = 1826,
  [1853] = 1835,
  [1854] = 1826,
  [1855] = 1835,
  [1856] = 691,
  [1857] = 1833,
  [1858] = 769,
  [1859] = 1827,
  [1860] = 275,
  [1861] = 1827,
  [1862] = 761,
  [1863] = 1833,
  [1864] = 1833,
  [1865] = 1827,
  [1866] = 709,
  [1867] = 1835,
  [1868] = 1868,
  [1869] = 10,
  [1870] = 1827,
  [1871] = 1827,
  [1872] = 1833,
  [1873] = 1873,
  [1874] = 1874,
  [1875] = 1874,
  [1876] = 1876,
  [1877] = 1874,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1873,
  [1881] = 1881,
  [1882] = 1873,
  [1883] = 275,
  [1884] = 1879,
  [1885] = 762,
  [1886] = 1876,
  [1887] = 1879,
  [1888] = 760,
  [1889] = 1876,
  [1890] = 1876,
  [1891] = 1881,
  [1892] = 1892,
  [1893] = 1893,
  [1894] = 685,
  [1895] = 1895,
  [1896] = 718,
  [1897] = 1879,
  [1898] = 1898,
  [1899] = 1874,
  [1900] = 1874,
  [1901] = 1876,
  [1902] = 1874,
  [1903] = 1881,
  [1904] = 759,
  [1905] = 729,
  [1906] = 758,
  [1907] = 1876,
  [1908] = 1879,
  [1909] = 1879,
  [1910] = 1881,
  [1911] = 1873,
  [1912] = 1881,
  [1913] = 1873,
  [1914] = 1881,
  [1915] = 1873,
  [1916] = 1916,
  [1917] = 1917,
  [1918] = 1916,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 716,
  [1922] = 1919,
  [1923] = 1923,
  [1924] = 1923,
  [1925] = 1925,
  [1926] = 1920,
  [1927] = 1927,
  [1928] = 1928,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1929,
  [1938] = 1925,
  [1939] = 1939,
  [1940] = 1940,
  [1941] = 1920,
  [1942] = 1928,
  [1943] = 1916,
  [1944] = 1929,
  [1945] = 1930,
  [1946] = 1931,
  [1947] = 1919,
  [1948] = 1923,
  [1949] = 1949,
  [1950] = 1932,
  [1951] = 1933,
  [1952] = 1934,
  [1953] = 1925,
  [1954] = 1920,
  [1955] = 1955,
  [1956] = 1956,
  [1957] = 1928,
  [1958] = 1929,
  [1959] = 1930,
  [1960] = 1931,
  [1961] = 1932,
  [1962] = 1933,
  [1963] = 1934,
  [1964] = 1964,
  [1965] = 1916,
  [1966] = 1966,
  [1967] = 718,
  [1968] = 1968,
  [1969] = 1916,
  [1970] = 1940,
  [1971] = 1934,
  [1972] = 1916,
  [1973] = 1956,
  [1974] = 1916,
  [1975] = 1975,
  [1976] = 1916,
  [1977] = 1930,
  [1978] = 1931,
  [1979] = 1979,
  [1980] = 1980,
  [1981] = 678,
  [1982] = 1925,
  [1983] = 1916,
  [1984] = 1940,
  [1985] = 1916,
  [1986] = 1986,
  [1987] = 1987,
  [1988] = 1956,
  [1989] = 1980,
  [1990] = 1919,
  [1991] = 1923,
  [1992] = 716,
  [1993] = 1936,
  [1994] = 1925,
  [1995] = 1920,
  [1996] = 1936,
  [1997] = 1997,
  [1998] = 1928,
  [1999] = 1929,
  [2000] = 1930,
  [2001] = 1931,
  [2002] = 2002,
  [2003] = 1932,
  [2004] = 1933,
  [2005] = 1934,
  [2006] = 2006,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 1919,
  [2010] = 1940,
  [2011] = 254,
  [2012] = 1940,
  [2013] = 1916,
  [2014] = 2014,
  [2015] = 1980,
  [2016] = 1979,
  [2017] = 1919,
  [2018] = 1923,
  [2019] = 2019,
  [2020] = 1936,
  [2021] = 1916,
  [2022] = 1925,
  [2023] = 1920,
  [2024] = 678,
  [2025] = 1928,
  [2026] = 1929,
  [2027] = 1930,
  [2028] = 1931,
  [2029] = 1932,
  [2030] = 1933,
  [2031] = 1934,
  [2032] = 1923,
  [2033] = 1979,
  [2034] = 1928,
  [2035] = 1932,
  [2036] = 1980,
  [2037] = 1933,
  [2038] = 2038,
  [2039] = 1940,
  [2040] = 1916,
  [2041] = 2041,
  [2042] = 2042,
  [2043] = 2043,
  [2044] = 2044,
  [2045] = 2045,
  [2046] = 2042,
  [2047] = 2047,
  [2048] = 2048,
  [2049] = 2043,
  [2050] = 2048,
  [2051] = 2051,
  [2052] = 2052,
  [2053] = 2053,
  [2054] = 718,
  [2055] = 2055,
  [2056] = 2048,
  [2057] = 2043,
  [2058] = 2043,
  [2059] = 2045,
  [2060] = 2060,
  [2061] = 2061,
  [2062] = 2052,
  [2063] = 2053,
  [2064] = 2060,
  [2065] = 2042,
  [2066] = 2051,
  [2067] = 2067,
  [2068] = 2068,
  [2069] = 2061,
  [2070] = 2061,
  [2071] = 2071,
  [2072] = 2072,
  [2073] = 2073,
  [2074] = 2060,
  [2075] = 2075,
  [2076] = 2053,
  [2077] = 2053,
  [2078] = 1956,
  [2079] = 2045,
  [2080] = 2080,
  [2081] = 2081,
  [2082] = 2045,
  [2083] = 2051,
  [2084] = 2068,
  [2085] = 2052,
  [2086] = 1979,
  [2087] = 2068,
  [2088] = 2068,
  [2089] = 2042,
  [2090] = 2061,
  [2091] = 2048,
  [2092] = 2051,
  [2093] = 2048,
  [2094] = 2094,
  [2095] = 2043,
  [2096] = 2042,
  [2097] = 2097,
  [2098] = 2098,
  [2099] = 2099,
  [2100] = 2043,
  [2101] = 2042,
  [2102] = 2052,
  [2103] = 2103,
  [2104] = 2052,
  [2105] = 2068,
  [2106] = 2060,
  [2107] = 2068,
  [2108] = 2048,
  [2109] = 14,
  [2110] = 2053,
  [2111] = 2051,
  [2112] = 2051,
  [2113] = 2060,
  [2114] = 2061,
  [2115] = 2053,
  [2116] = 2061,
  [2117] = 2045,
  [2118] = 2060,
  [2119] = 2045,
  [2120] = 2052,
  [2121] = 2121,
  [2122] = 2122,
  [2123] = 2123,
  [2124] = 2124,
  [2125] = 2125,
  [2126] = 2126,
  [2127] = 2127,
  [2128] = 2128,
  [2129] = 2126,
  [2130] = 2130,
  [2131] = 2131,
  [2132] = 2132,
  [2133] = 2133,
  [2134] = 2134,
  [2135] = 2124,
  [2136] = 2127,
  [2137] = 2137,
  [2138] = 2138,
  [2139] = 2139,
  [2140] = 2122,
  [2141] = 2122,
  [2142] = 2126,
  [2143] = 2127,
  [2144] = 2144,
  [2145] = 2126,
  [2146] = 2137,
  [2147] = 2147,
  [2148] = 2131,
  [2149] = 2149,
  [2150] = 2150,
  [2151] = 2133,
  [2152] = 2124,
  [2153] = 2153,
  [2154] = 2121,
  [2155] = 2137,
  [2156] = 2153,
  [2157] = 2122,
  [2158] = 2126,
  [2159] = 2131,
  [2160] = 2160,
  [2161] = 2133,
  [2162] = 2124,
  [2163] = 2163,
  [2164] = 2137,
  [2165] = 2130,
  [2166] = 2122,
  [2167] = 2153,
  [2168] = 2168,
  [2169] = 2138,
  [2170] = 2138,
  [2171] = 2127,
  [2172] = 2153,
  [2173] = 2121,
  [2174] = 2138,
  [2175] = 2175,
  [2176] = 2176,
  [2177] = 2177,
  [2178] = 2127,
  [2179] = 2130,
  [2180] = 2127,
  [2181] = 2181,
  [2182] = 2133,
  [2183] = 2183,
  [2184] = 2138,
  [2185] = 2138,
  [2186] = 2124,
  [2187] = 2187,
  [2188] = 2121,
  [2189] = 2130,
  [2190] = 2190,
  [2191] = 2191,
  [2192] = 2192,
  [2193] = 2133,
  [2194] = 2131,
  [2195] = 2130,
  [2196] = 2131,
  [2197] = 2153,
  [2198] = 2198,
  [2199] = 2133,
  [2200] = 2153,
  [2201] = 2121,
  [2202] = 2121,
  [2203] = 2124,
  [2204] = 2131,
  [2205] = 2205,
  [2206] = 2137,
  [2207] = 718,
  [2208] = 2122,
  [2209] = 2137,
  [2210] = 2130,
  [2211] = 2126,
  [2212] = 2212,
  [2213] = 2213,
  [2214] = 2214,
  [2215] = 2215,
  [2216] = 2213,
  [2217] = 2217,
  [2218] = 2215,
  [2219] = 2219,
  [2220] = 2220,
  [2221] = 2217,
  [2222] = 2222,
  [2223] = 2222,
  [2224] = 2219,
  [2225] = 2212,
  [2226] = 2226,
  [2227] = 2215,
  [2228] = 2228,
  [2229] = 2213,
  [2230] = 2217,
  [2231] = 2220,
  [2232] = 2232,
  [2233] = 2233,
  [2234] = 2226,
  [2235] = 2215,
  [2236] = 2217,
  [2237] = 2226,
  [2238] = 2215,
  [2239] = 2217,
  [2240] = 2226,
  [2241] = 2215,
  [2242] = 2226,
  [2243] = 2226,
  [2244] = 2215,
  [2245] = 2226,
  [2246] = 2215,
  [2247] = 2215,
  [2248] = 2217,
  [2249] = 2217,
  [2250] = 2250,
  [2251] = 2212,
  [2252] = 2252,
  [2253] = 2226,
  [2254] = 2215,
  [2255] = 2232,
  [2256] = 2217,
  [2257] = 2252,
  [2258] = 2258,
  [2259] = 2226,
  [2260] = 2215,
  [2261] = 2217,
  [2262] = 2262,
  [2263] = 2228,
  [2264] = 2217,
  [2265] = 2220,
  [2266] = 2222,
  [2267] = 2262,
  [2268] = 2226,
  [2269] = 2215,
  [2270] = 2217,
  [2271] = 2271,
  [2272] = 2252,
  [2273] = 2271,
  [2274] = 2226,
  [2275] = 2217,
  [2276] = 2215,
  [2277] = 2262,
  [2278] = 2217,
  [2279] = 2271,
  [2280] = 2271,
  [2281] = 2281,
  [2282] = 2282,
  [2283] = 2283,
  [2284] = 2284,
  [2285] = 2215,
  [2286] = 2217,
  [2287] = 2252,
  [2288] = 2288,
  [2289] = 2226,
  [2290] = 2290,
  [2291] = 2262,
  [2292] = 2284,
  [2293] = 2228,
  [2294] = 2294,
  [2295] = 2262,
  [2296] = 2232,
  [2297] = 2212,
  [2298] = 2298,
  [2299] = 2281,
  [2300] = 2281,
  [2301] = 2284,
  [2302] = 2302,
  [2303] = 2262,
  [2304] = 2250,
  [2305] = 2212,
  [2306] = 2281,
  [2307] = 2219,
  [2308] = 2213,
  [2309] = 2213,
  [2310] = 2213,
  [2311] = 2222,
  [2312] = 2125,
  [2313] = 2217,
  [2314] = 2314,
  [2315] = 2220,
  [2316] = 2220,
  [2317] = 2220,
  [2318] = 2284,
  [2319] = 2250,
  [2320] = 2226,
  [2321] = 2215,
  [2322] = 1748,
  [2323] = 2228,
  [2324] = 2284,
  [2325] = 2228,
  [2326] = 2212,
  [2327] = 2213,
  [2328] = 2250,
  [2329] = 2226,
  [2330] = 2215,
  [2331] = 2217,
  [2332] = 2212,
  [2333] = 2333,
  [2334] = 2284,
  [2335] = 2212,
  [2336] = 2213,
  [2337] = 2226,
  [2338] = 2338,
  [2339] = 2228,
  [2340] = 2226,
  [2341] = 2341,
  [2342] = 2342,
  [2343] = 2343,
  [2344] = 2344,
  [2345] = 2345,
  [2346] = 2346,
  [2347] = 2345,
  [2348] = 2343,
  [2349] = 2349,
  [2350] = 2350,
  [2351] = 2341,
  [2352] = 2342,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 2355,
  [2356] = 2356,
  [2357] = 2345,
  [2358] = 2358,
  [2359] = 2343,
  [2360] = 2345,
  [2361] = 2361,
  [2362] = 2344,
  [2363] = 2363,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2366,
  [2367] = 2367,
  [2368] = 2364,
  [2369] = 2346,
  [2370] = 2370,
  [2371] = 2371,
  [2372] = 2345,
  [2373] = 2364,
  [2374] = 2343,
  [2375] = 2364,
  [2376] = 2343,
  [2377] = 2344,
  [2378] = 2345,
  [2379] = 2349,
  [2380] = 2365,
  [2381] = 2343,
  [2382] = 2382,
  [2383] = 2383,
  [2384] = 2345,
  [2385] = 2343,
  [2386] = 2342,
  [2387] = 2387,
  [2388] = 2388,
  [2389] = 2343,
  [2390] = 2346,
  [2391] = 2346,
  [2392] = 2392,
  [2393] = 2343,
  [2394] = 2349,
  [2395] = 2345,
  [2396] = 2396,
  [2397] = 2343,
  [2398] = 2398,
  [2399] = 2399,
  [2400] = 2345,
  [2401] = 2342,
  [2402] = 2365,
  [2403] = 2364,
  [2404] = 2345,
  [2405] = 2342,
  [2406] = 2344,
  [2407] = 2349,
  [2408] = 2345,
  [2409] = 2366,
  [2410] = 2410,
  [2411] = 2345,
  [2412] = 2344,
  [2413] = 2365,
  [2414] = 2341,
  [2415] = 2345,
  [2416] = 2416,
  [2417] = 2349,
  [2418] = 2345,
  [2419] = 2365,
  [2420] = 2344,
  [2421] = 2421,
  [2422] = 2341,
  [2423] = 2423,
  [2424] = 2424,
  [2425] = 2341,
  [2426] = 2426,
  [2427] = 2346,
  [2428] = 2341,
  [2429] = 2429,
  [2430] = 2341,
  [2431] = 2341,
  [2432] = 2341,
  [2433] = 2341,
  [2434] = 2341,
  [2435] = 2341,
  [2436] = 2341,
  [2437] = 2341,
  [2438] = 2349,
  [2439] = 2341,
  [2440] = 2366,
  [2441] = 2366,
  [2442] = 2365,
  [2443] = 2345,
  [2444] = 2364,
  [2445] = 2382,
  [2446] = 2346,
  [2447] = 2343,
  [2448] = 2345,
  [2449] = 2449,
  [2450] = 2450,
  [2451] = 2366,
  [2452] = 2366,
  [2453] = 2453,
  [2454] = 2429,
  [2455] = 2343,
  [2456] = 2345,
  [2457] = 2341,
  [2458] = 2382,
  [2459] = 2342,
  [2460] = 2460,
  [2461] = 2461,
  [2462] = 2462,
  [2463] = 2463,
  [2464] = 2464,
};

static const TSCharacterRange aux_sym_annotation_token1_character_set_1[] = {
  {'$', '$'}, {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0xf8},
  {0x100, 0xfffe},
};

static const TSCharacterRange aux_sym_number_literal_token5_character_set_1[] = {
  {'0', '9'}, {'D', 'D'}, {'F', 'G'}, {'I', 'I'}, {'L', 'L'}, {'d', 'd'}, {'f', 'g'}, {'i', 'i'},
  {'l', 'l'},
};

static const TSCharacterRange sym_escape_sequence_character_set_1[] = {
  {'\n', '\n'}, {'"', '"'}, {'$', '$'}, {'\'', '\''}, {'\\', '\\'}, {'b', 'b'}, {'f', 'f'}, {'n', 'n'},
  {'r', 'u'},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '#', 5,
        '$', 496,
        '%', 231,
        '&', 273,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 208,
        '+', 236,
        ',', 288,
        '-', 238,
        '.', 206,
        '/', 233,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 256,
        '=', 214,
        '>', 260,
        '?', 500,
        '@', 210,
        '[', 435,
        '\\', 46,
        ']', 436,
        '^', 275,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '|', 276,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(182);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(15);
      if (lookahead == '"') ADVANCE(467);
      if (lookahead == '$') ADVANCE(495);
      if (lookahead == '/') ADVANCE(470);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(471);
      if (lookahead != 0) ADVANCE(472);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(24);
      if (lookahead == '\'') ADVANCE(448);
      if (lookahead == '/') ADVANCE(451);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(452);
      if (lookahead != 0) ADVANCE(453);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(54);
      if (lookahead == '/') ADVANCE(319);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(320);
      if (lookahead != 0 &&
          lookahead != '.' &&
          lookahead != '/') ADVANCE(321);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '$', 349,
        '%', 231,
        '&', 273,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 198,
        '+', 236,
        ',', 288,
        '-', 239,
        '.', 206,
        '/', 233,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 256,
        '=', 214,
        '>', 260,
        '?', 500,
        '[', 434,
        '^', 275,
        '_', 431,
        '{', 289,
        '|', 276,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(4);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 5:
      if (lookahead == '!') ADVANCE(188);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 533,
        '"', 468,
        '$', 349,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 203,
        '+', 235,
        ',', 288,
        '-', 241,
        '/', 232,
        '0', 439,
        ':', 193,
        ';', 192,
        '[', 434,
        ']', 436,
        '_', 431,
        '{', 289,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 533,
        '"', 468,
        '$', 349,
        '\'', 449,
        '(', 347,
        '*', 203,
        '+', 235,
        '-', 241,
        '/', 232,
        '0', 439,
        '@', 209,
        '[', 434,
        '_', 344,
        'b', 407,
        'c', 380,
        'd', 368,
        'f', 387,
        'i', 402,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 58,
        '%', 231,
        '&', 273,
        '(', 347,
        ')', 348,
        '*', 199,
        '+', 236,
        ',', 288,
        '-', 242,
        '.', 206,
        '/', 233,
        ';', 192,
        '<', 256,
        '=', 214,
        '>', 260,
        '?', 500,
        '[', 434,
        '^', 275,
        '_', 431,
        '{', 289,
        '|', 276,
        '}', 291,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(430);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 58,
        '%', 230,
        '&', 272,
        '(', 347,
        ')', 348,
        '*', 200,
        '+', 235,
        ',', 288,
        '-', 243,
        '.', 206,
        '/', 232,
        ':', 193,
        ';', 192,
        '<', 257,
        '=', 61,
        '>', 261,
        '?', 499,
        '[', 434,
        ']', 436,
        '^', 274,
        '_', 431,
        '{', 289,
        '|', 277,
        '}', 291,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(9);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(430);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 58,
        '%', 230,
        '&', 272,
        '(', 347,
        ')', 348,
        '*', 200,
        '+', 235,
        ',', 288,
        '-', 237,
        '.', 206,
        '/', 232,
        ';', 192,
        '<', 257,
        '=', 61,
        '>', 261,
        '?', 499,
        '[', 435,
        '^', 274,
        '_', 431,
        '|', 277,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(430);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 58,
        '%', 230,
        '&', 272,
        '(', 347,
        ')', 348,
        '*', 200,
        '+', 234,
        ',', 288,
        '-', 244,
        '.', 206,
        '/', 232,
        ':', 193,
        ';', 192,
        '<', 257,
        '=', 61,
        '>', 261,
        '?', 499,
        '[', 434,
        ']', 436,
        '^', 274,
        '_', 431,
        '{', 289,
        '|', 277,
        '}', 291,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(430);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 12:
      if (lookahead == '"') ADVANCE(473);
      END_STATE();
    case 13:
      if (lookahead == '"') ADVANCE(480);
      if (lookahead == '$') ADVANCE(495);
      if (lookahead == '/') ADVANCE(477);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(475);
      if (lookahead != 0) ADVANCE(478);
      END_STATE();
    case 14:
      if (lookahead == '"') ADVANCE(178);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 15:
      if (lookahead == '"') ADVANCE(467);
      if (lookahead == '$') ADVANCE(495);
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '$', 495,
        '(', 347,
        ',', 288,
        '/', 31,
        '<', 255,
        '>', 259,
        '@', 210,
        '[', 72,
        '\\', 163,
        '_', 343,
        'b', 140,
        'c', 112,
        'd', 102,
        'f', 120,
        'i', 132,
        'l', 141,
        'p', 145,
        's', 114,
        'v', 136,
        '{', 289,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(433);
      END_STATE();
    case 17:
      if (lookahead == '$') ADVANCE(495);
      if (lookahead == '/') ADVANCE(232);
      if (lookahead == '\\') ADVANCE(47);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(481);
      if (lookahead != 0) ADVANCE(482);
      END_STATE();
    case 18:
      if (lookahead == '$') ADVANCE(497);
      if (lookahead == '/') ADVANCE(19);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(488);
      if (lookahead != 0) ADVANCE(490);
      END_STATE();
    case 19:
      if (lookahead == '$') ADVANCE(493);
      if (lookahead == '*') ADVANCE(489);
      if (lookahead == '/') ADVANCE(487);
      if (lookahead != 0) ADVANCE(490);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '%', 59,
        '&', 60,
        '(', 347,
        '*', 42,
        '+', 43,
        '-', 44,
        '.', 205,
        '/', 32,
        ';', 192,
        '<', 57,
        '=', 213,
        '>', 70,
        '?', 62,
        '@', 209,
        '[', 434,
        '^', 63,
        '_', 344,
        'b', 407,
        'c', 380,
        'd', 368,
        'f', 387,
        'i', 402,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '|', 64,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 21:
      if (lookahead == '\'') ADVANCE(454);
      END_STATE();
    case 22:
      if (lookahead == '\'') ADVANCE(456);
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(461);
      if (lookahead != 0) ADVANCE(460);
      END_STATE();
    case 23:
      if (lookahead == '\'') ADVANCE(179);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 24:
      if (lookahead == '\'') ADVANCE(448);
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      END_STATE();
    case 25:
      if (lookahead == '\'') ADVANCE(464);
      if (lookahead == '/') ADVANCE(459);
      if (lookahead == '\\') ADVANCE(163);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(462);
      if (lookahead != 0) ADVANCE(460);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '(', 347,
        ')', 348,
        ',', 288,
        '/', 31,
        ';', 192,
        '<', 255,
        '=', 213,
        '>', 259,
        '[', 72,
        '_', 431,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(430);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 27:
      if (lookahead == '(') ADVANCE(347);
      if (lookahead == '*') ADVANCE(207);
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '_') ADVANCE(431);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '(', 347,
        '/', 31,
        '>', 259,
        '@', 209,
        '_', 343,
        'b', 140,
        'c', 111,
        'd', 102,
        'f', 120,
        'i', 135,
        'l', 141,
        'p', 145,
        's', 114,
        'v', 136,
        '{', 289,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(28);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(433);
      END_STATE();
    case 29:
      if (lookahead == '(') ADVANCE(347);
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '_') ADVANCE(431);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(29);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        ')', 348,
        '-', 69,
        '/', 31,
        '_', 431,
        'b', 407,
        'c', 380,
        'd', 368,
        'f', 395,
        'i', 402,
        'l', 406,
        's', 383,
        'v', 403,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 31:
      if (lookahead == '*') ADVANCE(305);
      if (lookahead == '/') ADVANCE(304);
      END_STATE();
    case 32:
      if (lookahead == '*') ADVANCE(305);
      if (lookahead == '/') ADVANCE(304);
      if (lookahead == '=') ADVANCE(217);
      END_STATE();
    case 33:
      if (lookahead == '*') ADVANCE(55);
      if (lookahead == '/') ADVANCE(34);
      if (lookahead == '@') ADVANCE(341);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(41);
      if (lookahead != 0) ADVANCE(35);
      END_STATE();
    case 34:
      if (lookahead == '*') ADVANCE(306);
      if (lookahead == '/') ADVANCE(302);
      if (lookahead == '@') ADVANCE(71);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 35:
      if (lookahead == '*') ADVANCE(181);
      if (lookahead == '@') ADVANCE(71);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 36:
      if (lookahead == '*') ADVANCE(317);
      if (lookahead == '/') ADVANCE(31);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(316);
      END_STATE();
    case 37:
      if (lookahead == '*') ADVANCE(37);
      if (lookahead == '/') ADVANCE(310);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 38:
      if (lookahead == '*') ADVANCE(37);
      if (lookahead == '/') ADVANCE(40);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(38);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 39:
      if (lookahead == '*') ADVANCE(37);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 40:
      if (lookahead == '*') ADVANCE(309);
      if (lookahead == '/') ADVANCE(300);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 41:
      if (lookahead == '*') ADVANCE(56);
      if (lookahead == '/') ADVANCE(324);
      if (lookahead == '@') ADVANCE(340);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(41);
      if (lookahead != 0) ADVANCE(326);
      END_STATE();
    case 42:
      if (lookahead == '*') ADVANCE(68);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 43:
      if (lookahead == '+') ADVANCE(228);
      if (lookahead == '=') ADVANCE(219);
      END_STATE();
    case 44:
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 45:
      if (lookahead == '.') ADVANCE(254);
      END_STATE();
    case 46:
      if (lookahead == '/') ADVANCE(483);
      if (lookahead == 'u') ADVANCE(485);
      if (set_contains(sym_escape_sequence_character_set_1, 9, lookahead)) ADVANCE(494);
      if (lookahead != 0) ADVANCE(484);
      END_STATE();
    case 47:
      if (lookahead == '/') ADVANCE(483);
      if (lookahead != 0) ADVANCE(484);
      END_STATE();
    case 48:
      ADVANCE_MAP(
        '/', 31,
        '@', 118,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        's', 382,
        'v', 403,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(48);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 49:
      ADVANCE_MAP(
        '/', 31,
        '_', 344,
        'b', 407,
        'c', 380,
        'd', 368,
        'f', 387,
        'i', 402,
        'l', 406,
        's', 382,
        'v', 403,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(49);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 50:
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '_') ADVANCE(211);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(211);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(212);
      END_STATE();
    case 51:
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '_') ADVANCE(431);
      if (lookahead == 'f') ADVANCE(388);
      if (lookahead == 's') ADVANCE(422);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (lookahead == '$' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 52:
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '{') ADVANCE(289);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      END_STATE();
    case 53:
      if (lookahead == '/') ADVANCE(31);
      if (lookahead == '{') ADVANCE(289);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(498);
      END_STATE();
    case 54:
      if (lookahead == '/') ADVANCE(31);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      END_STATE();
    case 55:
      if (lookahead == '/') ADVANCE(327);
      if (lookahead != 0) ADVANCE(35);
      END_STATE();
    case 56:
      if (lookahead == '/') ADVANCE(327);
      if (lookahead != 0) ADVANCE(325);
      END_STATE();
    case 57:
      if (lookahead == '<') ADVANCE(65);
      END_STATE();
    case 58:
      if (lookahead == '=') ADVANCE(266);
      if (lookahead == 'i') ADVANCE(126);
      END_STATE();
    case 59:
      if (lookahead == '=') ADVANCE(218);
      END_STATE();
    case 60:
      if (lookahead == '=') ADVANCE(224);
      END_STATE();
    case 61:
      if (lookahead == '=') ADVANCE(265);
      if (lookahead == '~') ADVANCE(270);
      END_STATE();
    case 62:
      if (lookahead == '=') ADVANCE(227);
      END_STATE();
    case 63:
      if (lookahead == '=') ADVANCE(225);
      END_STATE();
    case 64:
      if (lookahead == '=') ADVANCE(226);
      END_STATE();
    case 65:
      if (lookahead == '=') ADVANCE(221);
      END_STATE();
    case 66:
      if (lookahead == '=') ADVANCE(222);
      if (lookahead == '>') ADVANCE(67);
      END_STATE();
    case 67:
      if (lookahead == '=') ADVANCE(223);
      END_STATE();
    case 68:
      if (lookahead == '=') ADVANCE(215);
      END_STATE();
    case 69:
      if (lookahead == '>') ADVANCE(290);
      END_STATE();
    case 70:
      if (lookahead == '>') ADVANCE(66);
      END_STATE();
    case 71:
      if (lookahead == '@') ADVANCE(325);
      END_STATE();
    case 72:
      if (lookahead == ']') ADVANCE(519);
      END_STATE();
    case 73:
      if (lookahead == 'a') ADVANCE(151);
      END_STATE();
    case 74:
      if (lookahead == 'a') ADVANCE(146);
      END_STATE();
    case 75:
      if (lookahead == 'a') ADVANCE(128);
      END_STATE();
    case 76:
      if (lookahead == 'a') ADVANCE(122);
      END_STATE();
    case 77:
      if (lookahead == 'a') ADVANCE(161);
      END_STATE();
    case 78:
      if (lookahead == 'a') ADVANCE(155);
      END_STATE();
    case 79:
      if (lookahead == 'a') ADVANCE(130);
      END_STATE();
    case 80:
      if (lookahead == 'a') ADVANCE(89);
      END_STATE();
    case 81:
      if (lookahead == 'a') ADVANCE(159);
      END_STATE();
    case 82:
      if (lookahead == 'a') ADVANCE(90);
      END_STATE();
    case 83:
      if (lookahead == 'b') ADVANCE(123);
      END_STATE();
    case 84:
      if (lookahead == 'b') ADVANCE(125);
      END_STATE();
    case 85:
      if (lookahead == 'c') ADVANCE(113);
      END_STATE();
    case 86:
      if (lookahead == 'c') ADVANCE(520);
      END_STATE();
    case 87:
      if (lookahead == 'c') ADVANCE(526);
      END_STATE();
    case 88:
      if (lookahead == 'c') ADVANCE(96);
      END_STATE();
    case 89:
      if (lookahead == 'c') ADVANCE(97);
      END_STATE();
    case 90:
      if (lookahead == 'c') ADVANCE(100);
      END_STATE();
    case 91:
      if (lookahead == 'c') ADVANCE(160);
      END_STATE();
    case 92:
      if (lookahead == 'd') ADVANCE(517);
      END_STATE();
    case 93:
      if (lookahead == 'd') ADVANCE(522);
      END_STATE();
    case 94:
      if (lookahead == 'd') ADVANCE(530);
      END_STATE();
    case 95:
      if (lookahead == 'e') ADVANCE(144);
      END_STATE();
    case 96:
      if (lookahead == 'e') ADVANCE(138);
      END_STATE();
    case 97:
      if (lookahead == 'e') ADVANCE(283);
      END_STATE();
    case 98:
      if (lookahead == 'e') ADVANCE(515);
      END_STATE();
    case 99:
      if (lookahead == 'e') ADVANCE(524);
      END_STATE();
    case 100:
      if (lookahead == 'e') ADVANCE(284);
      END_STATE();
    case 101:
      if (lookahead == 'e') ADVANCE(93);
      END_STATE();
    case 102:
      if (lookahead == 'e') ADVANCE(107);
      if (lookahead == 'o') ADVANCE(162);
      END_STATE();
    case 103:
      if (lookahead == 'e') ADVANCE(91);
      END_STATE();
    case 104:
      if (lookahead == 'e') ADVANCE(94);
      END_STATE();
    case 105:
      if (lookahead == 'e') ADVANCE(79);
      END_STATE();
    case 106:
      if (lookahead == 'f') ADVANCE(264);
      END_STATE();
    case 107:
      if (lookahead == 'f') ADVANCE(345);
      END_STATE();
    case 108:
      if (lookahead == 'f') ADVANCE(80);
      END_STATE();
    case 109:
      if (lookahead == 'f') ADVANCE(82);
      END_STATE();
    case 110:
      if (lookahead == 'g') ADVANCE(511);
      END_STATE();
    case 111:
      if (lookahead == 'h') ADVANCE(74);
      END_STATE();
    case 112:
      if (lookahead == 'h') ADVANCE(74);
      if (lookahead == 'l') ADVANCE(73);
      END_STATE();
    case 113:
      if (lookahead == 'h') ADVANCE(148);
      END_STATE();
    case 114:
      if (lookahead == 'h') ADVANCE(139);
      if (lookahead == 't') ADVANCE(77);
      if (lookahead == 'y') ADVANCE(131);
      END_STATE();
    case 115:
      if (lookahead == 'i') ADVANCE(164);
      if (lookahead == 'o') ADVANCE(158);
      END_STATE();
    case 116:
      if (lookahead == 'i') ADVANCE(92);
      END_STATE();
    case 117:
      if (lookahead == 'i') ADVANCE(165);
      END_STATE();
    case 118:
      if (lookahead == 'i') ADVANCE(127);
      END_STATE();
    case 119:
      if (lookahead == 'i') ADVANCE(86);
      END_STATE();
    case 120:
      if (lookahead == 'i') ADVANCE(134);
      if (lookahead == 'l') ADVANCE(142);
      END_STATE();
    case 121:
      if (lookahead == 'i') ADVANCE(87);
      END_STATE();
    case 122:
      if (lookahead == 'l') ADVANCE(528);
      END_STATE();
    case 123:
      if (lookahead == 'l') ADVANCE(119);
      END_STATE();
    case 124:
      if (lookahead == 'l') ADVANCE(105);
      END_STATE();
    case 125:
      if (lookahead == 'l') ADVANCE(98);
      END_STATE();
    case 126:
      if (lookahead == 'n') ADVANCE(263);
      END_STATE();
    case 127:
      if (lookahead == 'n') ADVANCE(152);
      END_STATE();
    case 128:
      if (lookahead == 'n') ADVANCE(88);
      END_STATE();
    case 129:
      if (lookahead == 'n') ADVANCE(110);
      END_STATE();
    case 130:
      if (lookahead == 'n') ADVANCE(505);
      END_STATE();
    case 131:
      if (lookahead == 'n') ADVANCE(85);
      END_STATE();
    case 132:
      if (lookahead == 'n') ADVANCE(154);
      END_STATE();
    case 133:
      if (lookahead == 'n') ADVANCE(117);
      END_STATE();
    case 134:
      if (lookahead == 'n') ADVANCE(76);
      END_STATE();
    case 135:
      if (lookahead == 'n') ADVANCE(157);
      END_STATE();
    case 136:
      if (lookahead == 'o') ADVANCE(116);
      END_STATE();
    case 137:
      if (lookahead == 'o') ADVANCE(124);
      END_STATE();
    case 138:
      if (lookahead == 'o') ADVANCE(106);
      END_STATE();
    case 139:
      if (lookahead == 'o') ADVANCE(147);
      END_STATE();
    case 140:
      if (lookahead == 'o') ADVANCE(137);
      END_STATE();
    case 141:
      if (lookahead == 'o') ADVANCE(129);
      END_STATE();
    case 142:
      if (lookahead == 'o') ADVANCE(78);
      END_STATE();
    case 143:
      if (lookahead == 'o') ADVANCE(133);
      END_STATE();
    case 144:
      if (lookahead == 'r') ADVANCE(108);
      END_STATE();
    case 145:
      if (lookahead == 'r') ADVANCE(115);
      if (lookahead == 'u') ADVANCE(83);
      END_STATE();
    case 146:
      if (lookahead == 'r') ADVANCE(507);
      END_STATE();
    case 147:
      if (lookahead == 'r') ADVANCE(156);
      END_STATE();
    case 148:
      if (lookahead == 'r') ADVANCE(143);
      END_STATE();
    case 149:
      if (lookahead == 'r') ADVANCE(109);
      END_STATE();
    case 150:
      if (lookahead == 's') ADVANCE(286);
      END_STATE();
    case 151:
      if (lookahead == 's') ADVANCE(150);
      END_STATE();
    case 152:
      if (lookahead == 't') ADVANCE(95);
      END_STATE();
    case 153:
      if (lookahead == 't') ADVANCE(75);
      END_STATE();
    case 154:
      if (lookahead == 't') ADVANCE(503);
      END_STATE();
    case 155:
      if (lookahead == 't') ADVANCE(513);
      END_STATE();
    case 156:
      if (lookahead == 't') ADVANCE(509);
      END_STATE();
    case 157:
      if (lookahead == 't') ADVANCE(501);
      END_STATE();
    case 158:
      if (lookahead == 't') ADVANCE(103);
      END_STATE();
    case 159:
      if (lookahead == 't') ADVANCE(99);
      END_STATE();
    case 160:
      if (lookahead == 't') ADVANCE(101);
      END_STATE();
    case 161:
      if (lookahead == 't') ADVANCE(121);
      END_STATE();
    case 162:
      if (lookahead == 'u') ADVANCE(84);
      END_STATE();
    case 163:
      if (lookahead == 'u') ADVANCE(173);
      if (set_contains(sym_escape_sequence_character_set_1, 9, lookahead)) ADVANCE(494);
      END_STATE();
    case 164:
      if (lookahead == 'v') ADVANCE(81);
      END_STATE();
    case 165:
      if (lookahead == 'z') ADVANCE(104);
      END_STATE();
    case 166:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(444);
      END_STATE();
    case 167:
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(440);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(438);
      END_STATE();
    case 168:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 169:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(446);
      END_STATE();
    case 170:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(447);
      END_STATE();
    case 171:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(494);
      END_STATE();
    case 172:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(442);
      END_STATE();
    case 173:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(175);
      END_STATE();
    case 174:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(171);
      END_STATE();
    case 175:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(174);
      END_STATE();
    case 176:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(498);
      END_STATE();
    case 177:
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '/' &&
          (lookahead < 'A' || 'Z' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead)) ADVANCE(490);
      END_STATE();
    case 178:
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 179:
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 180:
      if (lookahead != 0 &&
          lookahead != '$') ADVANCE(490);
      END_STATE();
    case 181:
      if (lookahead != 0 &&
          lookahead != '/') ADVANCE(325);
      END_STATE();
    case 182:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '#', 5,
        '$', 496,
        '%', 231,
        '&', 273,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 202,
        '+', 236,
        ',', 288,
        '-', 238,
        '.', 206,
        '/', 233,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 256,
        '=', 214,
        '>', 260,
        '?', 500,
        '@', 210,
        '[', 435,
        '\\', 46,
        ']', 436,
        '^', 275,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '|', 276,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(182);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 183:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '$', 349,
        '%', 231,
        '&', 273,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 198,
        '+', 236,
        ',', 288,
        '-', 238,
        '.', 206,
        '/', 233,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 256,
        '=', 214,
        '>', 260,
        '?', 500,
        '@', 210,
        '[', 434,
        '^', 275,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '|', 276,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(183);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 184:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '$', 349,
        '%', 230,
        '&', 272,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 201,
        '+', 235,
        ',', 288,
        '-', 240,
        '.', 206,
        '/', 232,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 257,
        '=', 61,
        '>', 261,
        '?', 499,
        '@', 210,
        '[', 434,
        '^', 274,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '|', 277,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 185:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 534,
        '"', 468,
        '$', 349,
        '%', 230,
        '&', 272,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 201,
        '+', 235,
        ',', 288,
        '-', 241,
        '.', 206,
        '/', 232,
        '0', 439,
        ':', 193,
        ';', 192,
        '<', 257,
        '=', 61,
        '>', 261,
        '?', 499,
        '@', 210,
        '[', 435,
        '^', 274,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '|', 277,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(185);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 186:
      if (eof) ADVANCE(187);
      ADVANCE_MAP(
        '!', 533,
        '"', 468,
        '#', 5,
        '$', 349,
        '\'', 449,
        '(', 347,
        ')', 348,
        '*', 203,
        '+', 235,
        ',', 288,
        '-', 240,
        '.', 205,
        '/', 232,
        '0', 439,
        ';', 192,
        '=', 213,
        '>', 259,
        '@', 210,
        '[', 434,
        '\\', 163,
        '_', 344,
        'b', 407,
        'c', 379,
        'd', 368,
        'f', 387,
        'i', 396,
        'l', 406,
        'p', 410,
        's', 382,
        'v', 403,
        '{', 289,
        '}', 291,
        '~', 532,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(186);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_POUND_BANG);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(aux_sym_shebang_token1);
      if (lookahead == '*') ADVANCE(308);
      if (lookahead == '/') ADVANCE(191);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(191);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(aux_sym_shebang_token1);
      if (lookahead == '/') ADVANCE(189);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(190);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(191);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(aux_sym_shebang_token1);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(191);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_DOT_AMP);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_DOT_AT);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_QMARK_DOT);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_STAR_DOT);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(282);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == ':') ADVANCE(204);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(282);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(281);
      if (lookahead == '.') ADVANCE(197);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(281);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == ':') ADVANCE(204);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == ':') ADVANCE(204);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == ':') ADVANCE(204);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_STAR_COLON);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '&') ADVANCE(194);
      if (lookahead == '.') ADVANCE(251);
      if (lookahead == '@') ADVANCE(195);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_STAR2);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_STAR2);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == ':') ADVANCE(204);
      if (lookahead == '=') ADVANCE(216);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == 'i') ADVANCE(127);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(aux_sym_annotation_token1);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(211);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(212);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(aux_sym_annotation_token1);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(212);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(265);
      if (lookahead == '~') ADVANCE(270);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT_EQ);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_QMARK_EQ);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(218);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(305);
      if (lookahead == '/') ADVANCE(304);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(305);
      if (lookahead == '/') ADVANCE(304);
      if (lookahead == '=') ADVANCE(217);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(228);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(228);
      if (lookahead == '=') ADVANCE(219);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '0') ADVANCE(439);
      if (lookahead == '=') ADVANCE(220);
      if (lookahead == '>') ADVANCE(290);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '0') ADVANCE(439);
      if (lookahead == '=') ADVANCE(220);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '0') ADVANCE(439);
      if (lookahead == '>') ADVANCE(290);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '0') ADVANCE(439);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '>') ADVANCE(290);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(290);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(221);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(222);
      if (lookahead == '>') ADVANCE(250);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(249);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      if (lookahead == '=') ADVANCE(223);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '<') ADVANCE(252);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_LT);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_LT_DOT_DOT_LT);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_LT_DOT_DOT);
      if (lookahead == '<') ADVANCE(253);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '.') ADVANCE(45);
      if (lookahead == '<') ADVANCE(246);
      if (lookahead == '=') ADVANCE(258);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '.') ADVANCE(45);
      if (lookahead == '<') ADVANCE(245);
      if (lookahead == '=') ADVANCE(258);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '>') ADVANCE(267);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(262);
      if (lookahead == '>') ADVANCE(247);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(262);
      if (lookahead == '>') ADVANCE(248);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(anon_sym_BANGin);
      if (lookahead == 's') ADVANCE(153);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(anon_sym_BANGinstanceof);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(268);
      if (lookahead == '~') ADVANCE(271);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(269);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(anon_sym_LT_EQ_GT);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_TILDE);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(278);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(278);
      if (lookahead == '=') ADVANCE(224);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(225);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(226);
      if (lookahead == '|') ADVANCE(279);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(279);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(anon_sym_QMARK_COLON);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(215);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(anon_sym_ATinterface);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(anon_sym_interface);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_interface);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(anon_sym_class);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(325);
      if (lookahead == '/') ADVANCE(304);
      if (lookahead != 0) ADVANCE(301);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(478);
      if (lookahead == '"') ADVANCE(294);
      if (lookahead == '$' ||
          lookahead == '\\') ADVANCE(304);
      if (lookahead != 0) ADVANCE(474);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(478);
      if (lookahead == '"' ||
          lookahead == '$' ||
          lookahead == '\\') ADVANCE(304);
      if (lookahead != 0) ADVANCE(474);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(490);
      if (lookahead == '$') ADVANCE(304);
      if (lookahead != 0) ADVANCE(487);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(490);
      if (lookahead == '$' ||
          lookahead == '/' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= '{')) ADVANCE(304);
      if (lookahead != 0) ADVANCE(487);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(460);
      if (lookahead == '\'') ADVANCE(298);
      if (lookahead == '\\') ADVANCE(304);
      if (lookahead != 0) ADVANCE(457);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(460);
      if (lookahead == '\'' ||
          lookahead == '\\') ADVANCE(304);
      if (lookahead != 0) ADVANCE(457);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(39);
      if (lookahead == '*') ADVANCE(299);
      if (lookahead == '/') ADVANCE(304);
      if (lookahead != 0) ADVANCE(300);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '\n') ADVANCE(39);
      if (lookahead == '*') ADVANCE(299);
      if (lookahead != 0) ADVANCE(300);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '*') ADVANCE(292);
      if (lookahead == '@') ADVANCE(303);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(304);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(302);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '*') ADVANCE(292);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(304);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(302);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '@') ADVANCE(301);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(304);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(304);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR);
      if (lookahead == '*') ADVANCE(311);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR);
      if (lookahead == '*') ADVANCE(312);
      if (lookahead != 0 &&
          lookahead != '/') ADVANCE(325);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR);
      if (lookahead == '*') ADVANCE(314);
      if (lookahead == '.') ADVANCE(318);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(321);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR);
      if (lookahead == '*') ADVANCE(315);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(191);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR);
      if (lookahead == '*') ADVANCE(313);
      if (lookahead == '/') ADVANCE(310);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(aux_sym_comment_token2);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR_STAR);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR_STAR);
      if (lookahead == '*') ADVANCE(181);
      if (lookahead == '@') ADVANCE(71);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR_STAR);
      if (lookahead == '*') ADVANCE(37);
      if (lookahead == '/') ADVANCE(310);
      if (lookahead != 0) ADVANCE(39);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR_STAR);
      if (lookahead == '.') ADVANCE(318);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(321);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(anon_sym_SLASH_STAR_STAR);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(191);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token1);
      if (lookahead == '*') ADVANCE(317);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(316);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == '*') ADVANCE(317);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      if (lookahead == '*') ADVANCE(307);
      if (lookahead == '.') ADVANCE(318);
      if (lookahead == '/') ADVANCE(322);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(321);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      if (lookahead == '.') ADVANCE(318);
      if (lookahead == '/') ADVANCE(319);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(320);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(321);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      if (lookahead == '.') ADVANCE(318);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(321);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      if (lookahead == '.') ADVANCE(323);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(322);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token2);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(304);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token3);
      if (lookahead == '*') ADVANCE(306);
      if (lookahead == '/') ADVANCE(302);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token3);
      if (lookahead == '*') ADVANCE(181);
      if (lookahead == '@') ADVANCE(71);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(aux_sym_groovy_doc_token3);
      if (lookahead == '*') ADVANCE(181);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ') ADVANCE(326);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(anon_sym_STAR_SLASH);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(anon_sym_ATparam);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(anon_sym_ATthrows);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'a') ADVANCE(336);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'a') ADVANCE(333);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'h') ADVANCE(335);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'm') ADVANCE(328);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'o') ADVANCE(338);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'r') ADVANCE(334);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'r') ADVANCE(331);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 's') ADVANCE(329);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (lookahead == 'w') ADVANCE(337);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym_groovy_doc_tag);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym_groovy_doc_at_text);
      if (lookahead == '@') ADVANCE(325);
      if (lookahead == 'p') ADVANCE(330);
      if (lookahead == 't') ADVANCE(332);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*') ADVANCE(342);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym_groovy_doc_at_text);
      if (lookahead == 'p') ADVANCE(330);
      if (lookahead == 't') ADVANCE(332);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(339);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym_groovy_doc_at_text);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != '*' &&
          lookahead != '@') ADVANCE(342);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(anon_sym__);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(431);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(anon_sym_def);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(anon_sym_def);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '/') ADVANCE(486);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(411);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(416);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(425);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(392);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(418);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(399);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(363);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(423);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'b') ADVANCE(393);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'b') ADVANCE(394);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(381);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(521);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(527);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 363:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(371);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 364:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(424);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 365:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(518);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 366:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(523);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(531);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(376);
      if (lookahead == 'o') ADVANCE(426);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(516);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 370:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(525);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(285);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(364);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(366);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(367);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(355);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 376:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(346);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(356);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 378:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'g') ADVANCE(512);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 379:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(350);
      if (lookahead == 'l') ADVANCE(351);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 380:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(350);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(414);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(405);
      if (lookahead == 't') ADVANCE(352);
      if (lookahead == 'y') ADVANCE(398);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(405);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(427);
      if (lookahead == 'o') ADVANCE(421);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(365);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(428);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(401);
      if (lookahead == 'l') ADVANCE(409);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(401);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(361);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(362);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(375);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 392:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(529);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 393:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(389);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(369);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(409);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 396:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(417);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(378);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(360);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(506);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(386);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(353);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(420);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(385);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 404:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(391);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 405:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(413);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(397);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(404);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 408:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(400);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 409:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(354);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 410:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(384);
      if (lookahead == 'u') ADVANCE(358);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 411:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(508);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(377);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(419);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(408);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 415:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(287);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(415);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 417:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(502);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(514);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 419:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(510);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(504);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(372);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(352);
      if (lookahead == 'y') ADVANCE(398);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(370);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(373);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(390);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'u') ADVANCE(359);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'v') ADVANCE(357);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'z') ADVANCE(374);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(429);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(433);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(430);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_') ADVANCE(431);
      if (('a' <= lookahead && lookahead <= 'z') ||
          (0xc0 <= lookahead && lookahead <= 0xd6) ||
          (0xd8 <= lookahead && lookahead <= 0xf6) ||
          lookahead == 0xf8 ||
          (0x100 <= lookahead && lookahead <= 0xfffe)) ADVANCE(432);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(sym_identifier);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(aux_sym__type_identifier_token1);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(433);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == ']') ADVANCE(519);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(aux_sym_number_literal_token1);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(aux_sym_number_literal_token1);
      if (lookahead == '.') ADVANCE(169);
      if (lookahead == '_') ADVANCE(167);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(440);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(438);
      if (set_contains(aux_sym_number_literal_token5_character_set_1, 9, lookahead)) ADVANCE(437);
      END_STATE();
    case 439:
      ACCEPT_TOKEN(aux_sym_number_literal_token1);
      if (lookahead == '.') ADVANCE(169);
      if (lookahead == '_') ADVANCE(168);
      if (lookahead == 'b') ADVANCE(166);
      if (lookahead == 'x') ADVANCE(172);
      if (lookahead == '8' ||
          lookahead == '9') ADVANCE(440);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(438);
      if (set_contains(aux_sym_number_literal_token5_character_set_1, 9, lookahead)) ADVANCE(437);
      END_STATE();
    case 440:
      ACCEPT_TOKEN(aux_sym_number_literal_token1);
      ADVANCE_MAP(
        '.', 169,
        '_', 168,
        'D', 437,
        'F', 437,
        'G', 437,
        'I', 437,
        'L', 437,
        'd', 437,
        'f', 437,
        'g', 437,
        'i', 437,
        'l', 437,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(440);
      END_STATE();
    case 441:
      ACCEPT_TOKEN(aux_sym_number_literal_token2);
      END_STATE();
    case 442:
      ACCEPT_TOKEN(aux_sym_number_literal_token2);
      ADVANCE_MAP(
        '_', 172,
        'D', 442,
        'F', 442,
        'd', 442,
        'f', 442,
        'G', 441,
        'I', 441,
        'L', 441,
        'g', 441,
        'i', 441,
        'l', 441,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'E') ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(442);
      END_STATE();
    case 443:
      ACCEPT_TOKEN(aux_sym_number_literal_token3);
      END_STATE();
    case 444:
      ACCEPT_TOKEN(aux_sym_number_literal_token3);
      ADVANCE_MAP(
        '_', 166,
        '0', 444,
        '1', 444,
        'D', 443,
        'F', 443,
        'G', 443,
        'I', 443,
        'L', 443,
        'd', 443,
        'f', 443,
        'g', 443,
        'i', 443,
        'l', 443,
      );
      END_STATE();
    case 445:
      ACCEPT_TOKEN(aux_sym_number_literal_token5);
      END_STATE();
    case 446:
      ACCEPT_TOKEN(aux_sym_number_literal_token5);
      ADVANCE_MAP(
        '_', 169,
        'E', 170,
        'e', 170,
        'D', 445,
        'F', 445,
        'G', 445,
        'I', 445,
        'L', 445,
        'd', 445,
        'f', 445,
        'g', 445,
        'i', 445,
        'l', 445,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(446);
      END_STATE();
    case 447:
      ACCEPT_TOKEN(aux_sym_number_literal_token5);
      ADVANCE_MAP(
        'D', 445,
        'F', 445,
        'G', 445,
        'I', 445,
        'L', 445,
        'd', 445,
        'f', 445,
        'g', 445,
        'i', 445,
        'l', 445,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(447);
      END_STATE();
    case 448:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 449:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      if (lookahead == '\'') ADVANCE(21);
      END_STATE();
    case 450:
      ACCEPT_TOKEN(aux_sym__plain_string_token1);
      if (lookahead == '*') ADVANCE(453);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(453);
      END_STATE();
    case 451:
      ACCEPT_TOKEN(aux_sym__plain_string_token1);
      if (lookahead == '*') ADVANCE(450);
      if (lookahead == '/') ADVANCE(453);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(453);
      END_STATE();
    case 452:
      ACCEPT_TOKEN(aux_sym__plain_string_token1);
      if (lookahead == '/') ADVANCE(451);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(452);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(453);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(aux_sym__plain_string_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(453);
      END_STATE();
    case 454:
      ACCEPT_TOKEN(anon_sym_SQUOTE_SQUOTE_SQUOTE);
      END_STATE();
    case 455:
      ACCEPT_TOKEN(aux_sym__plain_string_token2);
      if (lookahead == '\'') ADVANCE(454);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(aux_sym__plain_string_token2);
      if (lookahead == '\'') ADVANCE(455);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\n') ADVANCE(460);
      if (lookahead == '\'') ADVANCE(297);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(457);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\'') ADVANCE(23);
      if (lookahead == '*') ADVANCE(460);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\'') ADVANCE(23);
      if (lookahead == '*') ADVANCE(458);
      if (lookahead == '/') ADVANCE(457);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 460:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\'') ADVANCE(23);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 461:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\'') ADVANCE(465);
      if (lookahead == '/') ADVANCE(459);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(461);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 462:
      ACCEPT_TOKEN(aux_sym__plain_string_token3);
      if (lookahead == '\'') ADVANCE(464);
      if (lookahead == '/') ADVANCE(459);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(462);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 463:
      ACCEPT_TOKEN(aux_sym__plain_string_token4);
      if (lookahead == '\'') ADVANCE(454);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 464:
      ACCEPT_TOKEN(aux_sym__plain_string_token4);
      if (lookahead == '\'') ADVANCE(466);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 465:
      ACCEPT_TOKEN(aux_sym__plain_string_token4);
      if (lookahead == '\'') ADVANCE(463);
      if (lookahead != 0 &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 466:
      ACCEPT_TOKEN(aux_sym__plain_string_token4);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(460);
      END_STATE();
    case 467:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 468:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(12);
      END_STATE();
    case 469:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token1);
      if (lookahead == '*') ADVANCE(472);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(472);
      END_STATE();
    case 470:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token1);
      if (lookahead == '*') ADVANCE(469);
      if (lookahead == '/') ADVANCE(472);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(472);
      END_STATE();
    case 471:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token1);
      if (lookahead == '/') ADVANCE(470);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(471);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(472);
      END_STATE();
    case 472:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(472);
      END_STATE();
    case 473:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 474:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token2);
      if (lookahead == '\n') ADVANCE(478);
      if (lookahead == '"') ADVANCE(293);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(474);
      END_STATE();
    case 475:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token2);
      if (lookahead == '"') ADVANCE(480);
      if (lookahead == '/') ADVANCE(477);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(475);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 476:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token2);
      if (lookahead == '"') ADVANCE(14);
      if (lookahead == '*') ADVANCE(478);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 477:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token2);
      if (lookahead == '"') ADVANCE(14);
      if (lookahead == '*') ADVANCE(476);
      if (lookahead == '/') ADVANCE(474);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 478:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token2);
      if (lookahead == '"') ADVANCE(14);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 479:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token3);
      if (lookahead == '"') ADVANCE(473);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 480:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token3);
      if (lookahead == '"') ADVANCE(479);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '\\') ADVANCE(478);
      END_STATE();
    case 481:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token4);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(481);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '/' &&
          lookahead != '\\') ADVANCE(482);
      END_STATE();
    case 482:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token4);
      if (lookahead != 0 &&
          lookahead != '$' &&
          lookahead != '/' &&
          lookahead != '\\') ADVANCE(482);
      END_STATE();
    case 483:
      ACCEPT_TOKEN(anon_sym_BSLASH_SLASH);
      END_STATE();
    case 484:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token5);
      END_STATE();
    case 485:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token5);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(175);
      END_STATE();
    case 486:
      ACCEPT_TOKEN(anon_sym_DOLLAR_SLASH);
      END_STATE();
    case 487:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token6);
      if (lookahead == '\n') ADVANCE(490);
      if (lookahead == '$') ADVANCE(296);
      if (lookahead == '/') ADVANCE(295);
      if (lookahead != 0) ADVANCE(487);
      END_STATE();
    case 488:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token6);
      if (lookahead == '$') ADVANCE(497);
      if (lookahead == '/') ADVANCE(19);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(488);
      if (lookahead != 0) ADVANCE(490);
      END_STATE();
    case 489:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token6);
      if (lookahead == '$') ADVANCE(177);
      if (lookahead == '*') ADVANCE(490);
      if (lookahead == '/') ADVANCE(180);
      if (lookahead != 0) ADVANCE(490);
      END_STATE();
    case 490:
      ACCEPT_TOKEN(aux_sym__interpolate_string_token6);
      if (lookahead == '$') ADVANCE(177);
      if (lookahead == '/') ADVANCE(180);
      if (lookahead != 0) ADVANCE(490);
      END_STATE();
    case 491:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DOLLAR);
      END_STATE();
    case 492:
      ACCEPT_TOKEN(anon_sym_DOLLAR_DOLLAR);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 493:
      ACCEPT_TOKEN(anon_sym_SLASH_DOLLAR);
      END_STATE();
    case 494:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 495:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 496:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '$') ADVANCE(492);
      if (lookahead == '/') ADVANCE(486);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 497:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (lookahead == '$') ADVANCE(491);
      if (lookahead == '/') ADVANCE(486);
      if (lookahead != 0 &&
          (lookahead < 'A' || 'Z' < lookahead) &&
          (lookahead < 'a' || '{' < lookahead)) ADVANCE(490);
      END_STATE();
    case 498:
      ACCEPT_TOKEN(aux_sym_interpolation_token1);
      if (lookahead == '.') ADVANCE(176);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(498);
      END_STATE();
    case 499:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == ':') ADVANCE(280);
      END_STATE();
    case 500:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == ':') ADVANCE(280);
      if (lookahead == '=') ADVANCE(227);
      END_STATE();
    case 501:
      ACCEPT_TOKEN(anon_sym_int);
      END_STATE();
    case 502:
      ACCEPT_TOKEN(anon_sym_int);
      if (lookahead == 'e') ADVANCE(412);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 503:
      ACCEPT_TOKEN(anon_sym_int);
      if (lookahead == 'e') ADVANCE(149);
      END_STATE();
    case 504:
      ACCEPT_TOKEN(anon_sym_int);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 505:
      ACCEPT_TOKEN(anon_sym_boolean);
      END_STATE();
    case 506:
      ACCEPT_TOKEN(anon_sym_boolean);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 507:
      ACCEPT_TOKEN(anon_sym_char);
      END_STATE();
    case 508:
      ACCEPT_TOKEN(anon_sym_char);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 509:
      ACCEPT_TOKEN(anon_sym_short);
      END_STATE();
    case 510:
      ACCEPT_TOKEN(anon_sym_short);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 511:
      ACCEPT_TOKEN(anon_sym_long);
      END_STATE();
    case 512:
      ACCEPT_TOKEN(anon_sym_long);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 513:
      ACCEPT_TOKEN(anon_sym_float);
      END_STATE();
    case 514:
      ACCEPT_TOKEN(anon_sym_float);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 515:
      ACCEPT_TOKEN(anon_sym_double);
      END_STATE();
    case 516:
      ACCEPT_TOKEN(anon_sym_double);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 517:
      ACCEPT_TOKEN(anon_sym_void);
      END_STATE();
    case 518:
      ACCEPT_TOKEN(anon_sym_void);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 519:
      ACCEPT_TOKEN(anon_sym_LBRACK_RBRACK);
      END_STATE();
    case 520:
      ACCEPT_TOKEN(anon_sym_public);
      END_STATE();
    case 521:
      ACCEPT_TOKEN(anon_sym_public);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 522:
      ACCEPT_TOKEN(anon_sym_protected);
      END_STATE();
    case 523:
      ACCEPT_TOKEN(anon_sym_protected);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 524:
      ACCEPT_TOKEN(anon_sym_private);
      END_STATE();
    case 525:
      ACCEPT_TOKEN(anon_sym_private);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 526:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 527:
      ACCEPT_TOKEN(anon_sym_static);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 528:
      ACCEPT_TOKEN(anon_sym_final);
      END_STATE();
    case 529:
      ACCEPT_TOKEN(anon_sym_final);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 530:
      ACCEPT_TOKEN(anon_sym_synchronized);
      END_STATE();
    case 531:
      ACCEPT_TOKEN(anon_sym_synchronized);
      if (set_contains(aux_sym_annotation_token1_character_set_1, 9, lookahead)) ADVANCE(432);
      END_STATE();
    case 532:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 533:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 534:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(266);
      if (lookahead == 'i') ADVANCE(126);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'a', 1,
        'b', 2,
        'c', 3,
        'd', 4,
        'e', 5,
        'f', 6,
        'i', 7,
        'n', 8,
        'p', 9,
        'r', 10,
        's', 11,
        't', 12,
        'w', 13,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 's') ADVANCE(14);
      END_STATE();
    case 2:
      if (lookahead == 'r') ADVANCE(15);
      END_STATE();
    case 3:
      if (lookahead == 'a') ADVANCE(16);
      if (lookahead == 'o') ADVANCE(17);
      END_STATE();
    case 4:
      if (lookahead == 'e') ADVANCE(18);
      if (lookahead == 'o') ADVANCE(19);
      END_STATE();
    case 5:
      if (lookahead == 'l') ADVANCE(20);
      if (lookahead == 'x') ADVANCE(21);
      END_STATE();
    case 6:
      if (lookahead == 'a') ADVANCE(22);
      if (lookahead == 'i') ADVANCE(23);
      if (lookahead == 'o') ADVANCE(24);
      END_STATE();
    case 7:
      if (lookahead == 'f') ADVANCE(25);
      if (lookahead == 'm') ADVANCE(26);
      if (lookahead == 'n') ADVANCE(27);
      END_STATE();
    case 8:
      if (lookahead == 'e') ADVANCE(28);
      if (lookahead == 'u') ADVANCE(29);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(30);
      if (lookahead == 'i') ADVANCE(31);
      END_STATE();
    case 10:
      if (lookahead == 'e') ADVANCE(32);
      END_STATE();
    case 11:
      if (lookahead == 'w') ADVANCE(33);
      END_STATE();
    case 12:
      if (lookahead == 'h') ADVANCE(34);
      if (lookahead == 'r') ADVANCE(35);
      END_STATE();
    case 13:
      if (lookahead == 'h') ADVANCE(36);
      END_STATE();
    case 14:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 's') ADVANCE(37);
      END_STATE();
    case 15:
      if (lookahead == 'e') ADVANCE(38);
      END_STATE();
    case 16:
      if (lookahead == 's') ADVANCE(39);
      if (lookahead == 't') ADVANCE(40);
      END_STATE();
    case 17:
      if (lookahead == 'n') ADVANCE(41);
      END_STATE();
    case 18:
      if (lookahead == 'f') ADVANCE(42);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_do);
      END_STATE();
    case 20:
      if (lookahead == 's') ADVANCE(43);
      END_STATE();
    case 21:
      if (lookahead == 't') ADVANCE(44);
      END_STATE();
    case 22:
      if (lookahead == 'l') ADVANCE(45);
      END_STATE();
    case 23:
      if (lookahead == 'n') ADVANCE(46);
      END_STATE();
    case 24:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 25:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 26:
      if (lookahead == 'p') ADVANCE(48);
      END_STATE();
    case 27:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 's') ADVANCE(49);
      END_STATE();
    case 28:
      if (lookahead == 'w') ADVANCE(50);
      END_STATE();
    case 29:
      if (lookahead == 'l') ADVANCE(51);
      END_STATE();
    case 30:
      if (lookahead == 'c') ADVANCE(52);
      END_STATE();
    case 31:
      if (lookahead == 'p') ADVANCE(53);
      END_STATE();
    case 32:
      if (lookahead == 't') ADVANCE(54);
      END_STATE();
    case 33:
      if (lookahead == 'i') ADVANCE(55);
      END_STATE();
    case 34:
      if (lookahead == 'i') ADVANCE(56);
      END_STATE();
    case 35:
      if (lookahead == 'u') ADVANCE(57);
      if (lookahead == 'y') ADVANCE(58);
      END_STATE();
    case 36:
      if (lookahead == 'i') ADVANCE(59);
      END_STATE();
    case 37:
      if (lookahead == 'e') ADVANCE(60);
      END_STATE();
    case 38:
      if (lookahead == 'a') ADVANCE(61);
      END_STATE();
    case 39:
      if (lookahead == 'e') ADVANCE(62);
      END_STATE();
    case 40:
      if (lookahead == 'c') ADVANCE(63);
      END_STATE();
    case 41:
      if (lookahead == 't') ADVANCE(64);
      END_STATE();
    case 42:
      if (lookahead == 'a') ADVANCE(65);
      END_STATE();
    case 43:
      if (lookahead == 'e') ADVANCE(66);
      END_STATE();
    case 44:
      if (lookahead == 'e') ADVANCE(67);
      END_STATE();
    case 45:
      if (lookahead == 's') ADVANCE(68);
      END_STATE();
    case 46:
      if (lookahead == 'a') ADVANCE(69);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 48:
      if (lookahead == 'o') ADVANCE(70);
      END_STATE();
    case 49:
      if (lookahead == 't') ADVANCE(71);
      END_STATE();
    case 50:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 51:
      if (lookahead == 'l') ADVANCE(72);
      END_STATE();
    case 52:
      if (lookahead == 'k') ADVANCE(73);
      END_STATE();
    case 53:
      if (lookahead == 'e') ADVANCE(74);
      END_STATE();
    case 54:
      if (lookahead == 'u') ADVANCE(75);
      END_STATE();
    case 55:
      if (lookahead == 't') ADVANCE(76);
      END_STATE();
    case 56:
      if (lookahead == 's') ADVANCE(77);
      END_STATE();
    case 57:
      if (lookahead == 'e') ADVANCE(78);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 59:
      if (lookahead == 'l') ADVANCE(79);
      END_STATE();
    case 60:
      if (lookahead == 'r') ADVANCE(80);
      END_STATE();
    case 61:
      if (lookahead == 'k') ADVANCE(81);
      END_STATE();
    case 62:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 63:
      if (lookahead == 'h') ADVANCE(82);
      END_STATE();
    case 64:
      if (lookahead == 'i') ADVANCE(83);
      END_STATE();
    case 65:
      if (lookahead == 'u') ADVANCE(84);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 67:
      if (lookahead == 'n') ADVANCE(85);
      END_STATE();
    case 68:
      if (lookahead == 'e') ADVANCE(86);
      END_STATE();
    case 69:
      if (lookahead == 'l') ADVANCE(87);
      END_STATE();
    case 70:
      if (lookahead == 'r') ADVANCE(88);
      END_STATE();
    case 71:
      if (lookahead == 'a') ADVANCE(89);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_null);
      END_STATE();
    case 73:
      if (lookahead == 'a') ADVANCE(90);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(91);
      END_STATE();
    case 75:
      if (lookahead == 'r') ADVANCE(92);
      END_STATE();
    case 76:
      if (lookahead == 'c') ADVANCE(93);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(anon_sym_this);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 79:
      if (lookahead == 'e') ADVANCE(94);
      END_STATE();
    case 80:
      if (lookahead == 't') ADVANCE(95);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_catch);
      END_STATE();
    case 83:
      if (lookahead == 'n') ADVANCE(96);
      END_STATE();
    case 84:
      if (lookahead == 'l') ADVANCE(97);
      END_STATE();
    case 85:
      if (lookahead == 'd') ADVANCE(98);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 87:
      if (lookahead == 'l') ADVANCE(99);
      END_STATE();
    case 88:
      if (lookahead == 't') ADVANCE(100);
      END_STATE();
    case 89:
      if (lookahead == 'n') ADVANCE(101);
      END_STATE();
    case 90:
      if (lookahead == 'g') ADVANCE(102);
      END_STATE();
    case 91:
      if (lookahead == 'i') ADVANCE(103);
      END_STATE();
    case 92:
      if (lookahead == 'n') ADVANCE(104);
      END_STATE();
    case 93:
      if (lookahead == 'h') ADVANCE(105);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 96:
      if (lookahead == 'u') ADVANCE(106);
      END_STATE();
    case 97:
      if (lookahead == 't') ADVANCE(107);
      END_STATE();
    case 98:
      if (lookahead == 's') ADVANCE(108);
      END_STATE();
    case 99:
      if (lookahead == 'y') ADVANCE(109);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 101:
      if (lookahead == 'c') ADVANCE(110);
      END_STATE();
    case 102:
      if (lookahead == 'e') ADVANCE(111);
      END_STATE();
    case 103:
      if (lookahead == 'n') ADVANCE(112);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_switch);
      END_STATE();
    case 106:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_extends);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_finally);
      END_STATE();
    case 110:
      if (lookahead == 'e') ADVANCE(114);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_package);
      END_STATE();
    case 112:
      if (lookahead == 'e') ADVANCE(115);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 114:
      if (lookahead == 'o') ADVANCE(116);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(anon_sym_pipeline);
      END_STATE();
    case 116:
      if (lookahead == 'f') ADVANCE(117);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_instanceof);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0},
  [1] = {.lex_state = 186},
  [2] = {.lex_state = 183},
  [3] = {.lex_state = 183},
  [4] = {.lex_state = 183},
  [5] = {.lex_state = 183},
  [6] = {.lex_state = 183},
  [7] = {.lex_state = 183},
  [8] = {.lex_state = 183},
  [9] = {.lex_state = 183},
  [10] = {.lex_state = 183},
  [11] = {.lex_state = 183},
  [12] = {.lex_state = 183},
  [13] = {.lex_state = 183},
  [14] = {.lex_state = 183},
  [15] = {.lex_state = 183},
  [16] = {.lex_state = 183},
  [17] = {.lex_state = 183},
  [18] = {.lex_state = 183},
  [19] = {.lex_state = 183},
  [20] = {.lex_state = 183},
  [21] = {.lex_state = 183},
  [22] = {.lex_state = 183},
  [23] = {.lex_state = 183},
  [24] = {.lex_state = 183},
  [25] = {.lex_state = 183},
  [26] = {.lex_state = 183},
  [27] = {.lex_state = 183},
  [28] = {.lex_state = 183},
  [29] = {.lex_state = 183},
  [30] = {.lex_state = 183},
  [31] = {.lex_state = 183},
  [32] = {.lex_state = 183},
  [33] = {.lex_state = 183},
  [34] = {.lex_state = 183},
  [35] = {.lex_state = 183},
  [36] = {.lex_state = 183},
  [37] = {.lex_state = 183},
  [38] = {.lex_state = 183},
  [39] = {.lex_state = 183},
  [40] = {.lex_state = 183},
  [41] = {.lex_state = 183},
  [42] = {.lex_state = 183},
  [43] = {.lex_state = 186},
  [44] = {.lex_state = 186},
  [45] = {.lex_state = 186},
  [46] = {.lex_state = 186},
  [47] = {.lex_state = 186},
  [48] = {.lex_state = 186},
  [49] = {.lex_state = 186},
  [50] = {.lex_state = 186},
  [51] = {.lex_state = 186},
  [52] = {.lex_state = 186},
  [53] = {.lex_state = 186},
  [54] = {.lex_state = 186},
  [55] = {.lex_state = 186},
  [56] = {.lex_state = 186},
  [57] = {.lex_state = 186},
  [58] = {.lex_state = 186},
  [59] = {.lex_state = 186},
  [60] = {.lex_state = 186},
  [61] = {.lex_state = 186},
  [62] = {.lex_state = 186},
  [63] = {.lex_state = 186},
  [64] = {.lex_state = 186},
  [65] = {.lex_state = 186},
  [66] = {.lex_state = 186},
  [67] = {.lex_state = 186},
  [68] = {.lex_state = 186},
  [69] = {.lex_state = 186},
  [70] = {.lex_state = 186},
  [71] = {.lex_state = 186},
  [72] = {.lex_state = 186},
  [73] = {.lex_state = 186},
  [74] = {.lex_state = 186},
  [75] = {.lex_state = 186},
  [76] = {.lex_state = 186},
  [77] = {.lex_state = 186},
  [78] = {.lex_state = 186},
  [79] = {.lex_state = 186},
  [80] = {.lex_state = 186},
  [81] = {.lex_state = 186},
  [82] = {.lex_state = 186},
  [83] = {.lex_state = 186},
  [84] = {.lex_state = 186},
  [85] = {.lex_state = 186},
  [86] = {.lex_state = 186},
  [87] = {.lex_state = 186},
  [88] = {.lex_state = 186},
  [89] = {.lex_state = 186},
  [90] = {.lex_state = 186},
  [91] = {.lex_state = 186},
  [92] = {.lex_state = 186},
  [93] = {.lex_state = 186},
  [94] = {.lex_state = 186},
  [95] = {.lex_state = 186},
  [96] = {.lex_state = 186},
  [97] = {.lex_state = 186},
  [98] = {.lex_state = 186},
  [99] = {.lex_state = 186},
  [100] = {.lex_state = 186},
  [101] = {.lex_state = 186},
  [102] = {.lex_state = 186},
  [103] = {.lex_state = 186},
  [104] = {.lex_state = 186},
  [105] = {.lex_state = 186},
  [106] = {.lex_state = 186},
  [107] = {.lex_state = 186},
  [108] = {.lex_state = 186},
  [109] = {.lex_state = 186},
  [110] = {.lex_state = 186},
  [111] = {.lex_state = 186},
  [112] = {.lex_state = 186},
  [113] = {.lex_state = 186},
  [114] = {.lex_state = 186},
  [115] = {.lex_state = 186},
  [116] = {.lex_state = 186},
  [117] = {.lex_state = 186},
  [118] = {.lex_state = 186},
  [119] = {.lex_state = 186},
  [120] = {.lex_state = 186},
  [121] = {.lex_state = 186},
  [122] = {.lex_state = 186},
  [123] = {.lex_state = 186},
  [124] = {.lex_state = 186},
  [125] = {.lex_state = 186},
  [126] = {.lex_state = 186},
  [127] = {.lex_state = 186},
  [128] = {.lex_state = 186},
  [129] = {.lex_state = 186},
  [130] = {.lex_state = 186},
  [131] = {.lex_state = 186},
  [132] = {.lex_state = 186},
  [133] = {.lex_state = 186},
  [134] = {.lex_state = 186},
  [135] = {.lex_state = 186},
  [136] = {.lex_state = 186},
  [137] = {.lex_state = 186},
  [138] = {.lex_state = 186},
  [139] = {.lex_state = 186},
  [140] = {.lex_state = 186},
  [141] = {.lex_state = 186},
  [142] = {.lex_state = 186},
  [143] = {.lex_state = 186},
  [144] = {.lex_state = 186},
  [145] = {.lex_state = 186},
  [146] = {.lex_state = 186},
  [147] = {.lex_state = 186},
  [148] = {.lex_state = 186},
  [149] = {.lex_state = 186},
  [150] = {.lex_state = 186},
  [151] = {.lex_state = 186},
  [152] = {.lex_state = 186},
  [153] = {.lex_state = 186},
  [154] = {.lex_state = 186},
  [155] = {.lex_state = 186},
  [156] = {.lex_state = 186},
  [157] = {.lex_state = 186},
  [158] = {.lex_state = 186},
  [159] = {.lex_state = 186},
  [160] = {.lex_state = 186},
  [161] = {.lex_state = 186},
  [162] = {.lex_state = 186},
  [163] = {.lex_state = 186},
  [164] = {.lex_state = 186},
  [165] = {.lex_state = 186},
  [166] = {.lex_state = 186},
  [167] = {.lex_state = 186},
  [168] = {.lex_state = 186},
  [169] = {.lex_state = 186},
  [170] = {.lex_state = 186},
  [171] = {.lex_state = 186},
  [172] = {.lex_state = 186},
  [173] = {.lex_state = 186},
  [174] = {.lex_state = 186},
  [175] = {.lex_state = 186},
  [176] = {.lex_state = 186},
  [177] = {.lex_state = 186},
  [178] = {.lex_state = 186},
  [179] = {.lex_state = 186},
  [180] = {.lex_state = 186},
  [181] = {.lex_state = 186},
  [182] = {.lex_state = 186},
  [183] = {.lex_state = 186},
  [184] = {.lex_state = 186},
  [185] = {.lex_state = 186},
  [186] = {.lex_state = 186},
  [187] = {.lex_state = 186},
  [188] = {.lex_state = 186},
  [189] = {.lex_state = 186},
  [190] = {.lex_state = 186},
  [191] = {.lex_state = 186},
  [192] = {.lex_state = 186},
  [193] = {.lex_state = 186},
  [194] = {.lex_state = 186},
  [195] = {.lex_state = 186},
  [196] = {.lex_state = 186},
  [197] = {.lex_state = 186},
  [198] = {.lex_state = 186},
  [199] = {.lex_state = 186},
  [200] = {.lex_state = 186},
  [201] = {.lex_state = 186},
  [202] = {.lex_state = 186},
  [203] = {.lex_state = 186},
  [204] = {.lex_state = 186},
  [205] = {.lex_state = 186},
  [206] = {.lex_state = 186},
  [207] = {.lex_state = 186},
  [208] = {.lex_state = 186},
  [209] = {.lex_state = 186},
  [210] = {.lex_state = 186},
  [211] = {.lex_state = 186},
  [212] = {.lex_state = 186},
  [213] = {.lex_state = 186},
  [214] = {.lex_state = 186},
  [215] = {.lex_state = 186},
  [216] = {.lex_state = 186},
  [217] = {.lex_state = 186},
  [218] = {.lex_state = 186},
  [219] = {.lex_state = 186},
  [220] = {.lex_state = 186},
  [221] = {.lex_state = 186},
  [222] = {.lex_state = 186},
  [223] = {.lex_state = 186},
  [224] = {.lex_state = 186},
  [225] = {.lex_state = 4},
  [226] = {.lex_state = 186},
  [227] = {.lex_state = 186},
  [228] = {.lex_state = 186},
  [229] = {.lex_state = 186},
  [230] = {.lex_state = 186},
  [231] = {.lex_state = 186},
  [232] = {.lex_state = 4},
  [233] = {.lex_state = 186},
  [234] = {.lex_state = 186},
  [235] = {.lex_state = 186},
  [236] = {.lex_state = 186},
  [237] = {.lex_state = 186},
  [238] = {.lex_state = 186},
  [239] = {.lex_state = 186},
  [240] = {.lex_state = 186},
  [241] = {.lex_state = 186},
  [242] = {.lex_state = 186},
  [243] = {.lex_state = 186},
  [244] = {.lex_state = 186},
  [245] = {.lex_state = 186},
  [246] = {.lex_state = 186},
  [247] = {.lex_state = 4},
  [248] = {.lex_state = 185},
  [249] = {.lex_state = 184},
  [250] = {.lex_state = 184},
  [251] = {.lex_state = 184},
  [252] = {.lex_state = 184},
  [253] = {.lex_state = 184},
  [254] = {.lex_state = 184},
  [255] = {.lex_state = 185},
  [256] = {.lex_state = 184},
  [257] = {.lex_state = 184},
  [258] = {.lex_state = 184},
  [259] = {.lex_state = 184},
  [260] = {.lex_state = 184},
  [261] = {.lex_state = 184},
  [262] = {.lex_state = 184},
  [263] = {.lex_state = 184},
  [264] = {.lex_state = 184},
  [265] = {.lex_state = 184},
  [266] = {.lex_state = 184},
  [267] = {.lex_state = 184},
  [268] = {.lex_state = 184},
  [269] = {.lex_state = 184},
  [270] = {.lex_state = 184},
  [271] = {.lex_state = 184},
  [272] = {.lex_state = 184},
  [273] = {.lex_state = 184},
  [274] = {.lex_state = 184},
  [275] = {.lex_state = 184},
  [276] = {.lex_state = 184},
  [277] = {.lex_state = 184},
  [278] = {.lex_state = 184},
  [279] = {.lex_state = 184},
  [280] = {.lex_state = 184},
  [281] = {.lex_state = 184},
  [282] = {.lex_state = 184},
  [283] = {.lex_state = 4},
  [284] = {.lex_state = 184},
  [285] = {.lex_state = 184},
  [286] = {.lex_state = 184},
  [287] = {.lex_state = 184},
  [288] = {.lex_state = 184},
  [289] = {.lex_state = 184},
  [290] = {.lex_state = 184},
  [291] = {.lex_state = 184},
  [292] = {.lex_state = 184},
  [293] = {.lex_state = 184},
  [294] = {.lex_state = 184},
  [295] = {.lex_state = 184},
  [296] = {.lex_state = 185},
  [297] = {.lex_state = 184},
  [298] = {.lex_state = 184},
  [299] = {.lex_state = 184},
  [300] = {.lex_state = 184},
  [301] = {.lex_state = 184},
  [302] = {.lex_state = 184},
  [303] = {.lex_state = 184},
  [304] = {.lex_state = 184},
  [305] = {.lex_state = 184},
  [306] = {.lex_state = 184},
  [307] = {.lex_state = 184},
  [308] = {.lex_state = 184},
  [309] = {.lex_state = 184},
  [310] = {.lex_state = 184},
  [311] = {.lex_state = 184},
  [312] = {.lex_state = 184},
  [313] = {.lex_state = 184},
  [314] = {.lex_state = 184},
  [315] = {.lex_state = 184},
  [316] = {.lex_state = 184},
  [317] = {.lex_state = 184},
  [318] = {.lex_state = 184},
  [319] = {.lex_state = 184},
  [320] = {.lex_state = 184},
  [321] = {.lex_state = 184},
  [322] = {.lex_state = 184},
  [323] = {.lex_state = 184},
  [324] = {.lex_state = 184},
  [325] = {.lex_state = 184},
  [326] = {.lex_state = 184},
  [327] = {.lex_state = 184},
  [328] = {.lex_state = 184},
  [329] = {.lex_state = 184},
  [330] = {.lex_state = 184},
  [331] = {.lex_state = 184},
  [332] = {.lex_state = 185},
  [333] = {.lex_state = 184},
  [334] = {.lex_state = 184},
  [335] = {.lex_state = 184},
  [336] = {.lex_state = 184},
  [337] = {.lex_state = 184},
  [338] = {.lex_state = 184},
  [339] = {.lex_state = 184},
  [340] = {.lex_state = 184},
  [341] = {.lex_state = 184},
  [342] = {.lex_state = 184},
  [343] = {.lex_state = 184},
  [344] = {.lex_state = 184},
  [345] = {.lex_state = 184},
  [346] = {.lex_state = 184},
  [347] = {.lex_state = 184},
  [348] = {.lex_state = 184},
  [349] = {.lex_state = 184},
  [350] = {.lex_state = 184},
  [351] = {.lex_state = 184},
  [352] = {.lex_state = 184},
  [353] = {.lex_state = 184},
  [354] = {.lex_state = 184},
  [355] = {.lex_state = 184},
  [356] = {.lex_state = 184},
  [357] = {.lex_state = 184},
  [358] = {.lex_state = 184},
  [359] = {.lex_state = 184},
  [360] = {.lex_state = 184},
  [361] = {.lex_state = 184},
  [362] = {.lex_state = 184},
  [363] = {.lex_state = 185},
  [364] = {.lex_state = 184},
  [365] = {.lex_state = 184},
  [366] = {.lex_state = 184},
  [367] = {.lex_state = 184},
  [368] = {.lex_state = 184},
  [369] = {.lex_state = 184},
  [370] = {.lex_state = 184},
  [371] = {.lex_state = 184},
  [372] = {.lex_state = 184},
  [373] = {.lex_state = 184},
  [374] = {.lex_state = 184},
  [375] = {.lex_state = 184},
  [376] = {.lex_state = 184},
  [377] = {.lex_state = 184},
  [378] = {.lex_state = 184},
  [379] = {.lex_state = 184},
  [380] = {.lex_state = 184},
  [381] = {.lex_state = 184},
  [382] = {.lex_state = 184},
  [383] = {.lex_state = 184},
  [384] = {.lex_state = 184},
  [385] = {.lex_state = 184},
  [386] = {.lex_state = 184},
  [387] = {.lex_state = 184},
  [388] = {.lex_state = 184},
  [389] = {.lex_state = 184},
  [390] = {.lex_state = 184},
  [391] = {.lex_state = 184},
  [392] = {.lex_state = 184},
  [393] = {.lex_state = 184},
  [394] = {.lex_state = 184},
  [395] = {.lex_state = 184},
  [396] = {.lex_state = 184},
  [397] = {.lex_state = 184},
  [398] = {.lex_state = 184},
  [399] = {.lex_state = 184},
  [400] = {.lex_state = 184},
  [401] = {.lex_state = 184},
  [402] = {.lex_state = 184},
  [403] = {.lex_state = 184},
  [404] = {.lex_state = 184},
  [405] = {.lex_state = 184},
  [406] = {.lex_state = 184},
  [407] = {.lex_state = 184},
  [408] = {.lex_state = 184},
  [409] = {.lex_state = 184},
  [410] = {.lex_state = 184},
  [411] = {.lex_state = 184},
  [412] = {.lex_state = 184},
  [413] = {.lex_state = 184},
  [414] = {.lex_state = 184},
  [415] = {.lex_state = 184},
  [416] = {.lex_state = 184},
  [417] = {.lex_state = 184},
  [418] = {.lex_state = 184},
  [419] = {.lex_state = 184},
  [420] = {.lex_state = 184},
  [421] = {.lex_state = 185},
  [422] = {.lex_state = 184},
  [423] = {.lex_state = 184},
  [424] = {.lex_state = 184},
  [425] = {.lex_state = 184},
  [426] = {.lex_state = 184},
  [427] = {.lex_state = 184},
  [428] = {.lex_state = 184},
  [429] = {.lex_state = 184},
  [430] = {.lex_state = 184},
  [431] = {.lex_state = 184},
  [432] = {.lex_state = 184},
  [433] = {.lex_state = 184},
  [434] = {.lex_state = 184},
  [435] = {.lex_state = 184},
  [436] = {.lex_state = 184},
  [437] = {.lex_state = 184},
  [438] = {.lex_state = 184},
  [439] = {.lex_state = 184},
  [440] = {.lex_state = 184},
  [441] = {.lex_state = 184},
  [442] = {.lex_state = 184},
  [443] = {.lex_state = 184},
  [444] = {.lex_state = 184},
  [445] = {.lex_state = 184},
  [446] = {.lex_state = 184},
  [447] = {.lex_state = 184},
  [448] = {.lex_state = 184},
  [449] = {.lex_state = 184},
  [450] = {.lex_state = 184},
  [451] = {.lex_state = 184},
  [452] = {.lex_state = 184},
  [453] = {.lex_state = 184},
  [454] = {.lex_state = 184},
  [455] = {.lex_state = 184},
  [456] = {.lex_state = 184},
  [457] = {.lex_state = 184},
  [458] = {.lex_state = 184},
  [459] = {.lex_state = 184},
  [460] = {.lex_state = 184},
  [461] = {.lex_state = 184},
  [462] = {.lex_state = 184},
  [463] = {.lex_state = 184},
  [464] = {.lex_state = 184},
  [465] = {.lex_state = 184},
  [466] = {.lex_state = 184},
  [467] = {.lex_state = 185},
  [468] = {.lex_state = 184},
  [469] = {.lex_state = 184},
  [470] = {.lex_state = 184},
  [471] = {.lex_state = 184},
  [472] = {.lex_state = 184},
  [473] = {.lex_state = 184},
  [474] = {.lex_state = 184},
  [475] = {.lex_state = 184},
  [476] = {.lex_state = 184},
  [477] = {.lex_state = 184},
  [478] = {.lex_state = 184},
  [479] = {.lex_state = 184},
  [480] = {.lex_state = 184},
  [481] = {.lex_state = 184},
  [482] = {.lex_state = 184},
  [483] = {.lex_state = 184},
  [484] = {.lex_state = 184},
  [485] = {.lex_state = 184},
  [486] = {.lex_state = 184},
  [487] = {.lex_state = 184},
  [488] = {.lex_state = 184},
  [489] = {.lex_state = 184},
  [490] = {.lex_state = 184},
  [491] = {.lex_state = 184},
  [492] = {.lex_state = 184},
  [493] = {.lex_state = 184},
  [494] = {.lex_state = 184},
  [495] = {.lex_state = 184},
  [496] = {.lex_state = 184},
  [497] = {.lex_state = 185},
  [498] = {.lex_state = 184},
  [499] = {.lex_state = 184},
  [500] = {.lex_state = 184},
  [501] = {.lex_state = 184},
  [502] = {.lex_state = 184},
  [503] = {.lex_state = 184},
  [504] = {.lex_state = 184},
  [505] = {.lex_state = 184},
  [506] = {.lex_state = 184},
  [507] = {.lex_state = 184},
  [508] = {.lex_state = 184},
  [509] = {.lex_state = 184},
  [510] = {.lex_state = 184},
  [511] = {.lex_state = 184},
  [512] = {.lex_state = 184},
  [513] = {.lex_state = 184},
  [514] = {.lex_state = 184},
  [515] = {.lex_state = 184},
  [516] = {.lex_state = 184},
  [517] = {.lex_state = 184},
  [518] = {.lex_state = 184},
  [519] = {.lex_state = 184},
  [520] = {.lex_state = 184},
  [521] = {.lex_state = 184},
  [522] = {.lex_state = 184},
  [523] = {.lex_state = 184},
  [524] = {.lex_state = 184},
  [525] = {.lex_state = 184},
  [526] = {.lex_state = 184},
  [527] = {.lex_state = 184},
  [528] = {.lex_state = 184},
  [529] = {.lex_state = 184},
  [530] = {.lex_state = 184},
  [531] = {.lex_state = 184},
  [532] = {.lex_state = 184},
  [533] = {.lex_state = 184},
  [534] = {.lex_state = 184},
  [535] = {.lex_state = 184},
  [536] = {.lex_state = 184},
  [537] = {.lex_state = 184},
  [538] = {.lex_state = 184},
  [539] = {.lex_state = 184},
  [540] = {.lex_state = 184},
  [541] = {.lex_state = 184},
  [542] = {.lex_state = 184},
  [543] = {.lex_state = 184},
  [544] = {.lex_state = 184},
  [545] = {.lex_state = 184},
  [546] = {.lex_state = 184},
  [547] = {.lex_state = 184},
  [548] = {.lex_state = 184},
  [549] = {.lex_state = 184},
  [550] = {.lex_state = 184},
  [551] = {.lex_state = 184},
  [552] = {.lex_state = 184},
  [553] = {.lex_state = 184},
  [554] = {.lex_state = 184},
  [555] = {.lex_state = 184},
  [556] = {.lex_state = 184},
  [557] = {.lex_state = 184},
  [558] = {.lex_state = 184},
  [559] = {.lex_state = 184},
  [560] = {.lex_state = 184},
  [561] = {.lex_state = 184},
  [562] = {.lex_state = 184},
  [563] = {.lex_state = 184},
  [564] = {.lex_state = 184},
  [565] = {.lex_state = 184},
  [566] = {.lex_state = 184},
  [567] = {.lex_state = 184},
  [568] = {.lex_state = 184},
  [569] = {.lex_state = 184},
  [570] = {.lex_state = 184},
  [571] = {.lex_state = 184},
  [572] = {.lex_state = 184},
  [573] = {.lex_state = 184},
  [574] = {.lex_state = 184},
  [575] = {.lex_state = 184},
  [576] = {.lex_state = 184},
  [577] = {.lex_state = 184},
  [578] = {.lex_state = 184},
  [579] = {.lex_state = 184},
  [580] = {.lex_state = 184},
  [581] = {.lex_state = 184},
  [582] = {.lex_state = 184},
  [583] = {.lex_state = 184},
  [584] = {.lex_state = 184},
  [585] = {.lex_state = 184},
  [586] = {.lex_state = 184},
  [587] = {.lex_state = 184},
  [588] = {.lex_state = 184},
  [589] = {.lex_state = 184},
  [590] = {.lex_state = 184},
  [591] = {.lex_state = 184},
  [592] = {.lex_state = 184},
  [593] = {.lex_state = 184},
  [594] = {.lex_state = 184},
  [595] = {.lex_state = 184},
  [596] = {.lex_state = 184},
  [597] = {.lex_state = 184},
  [598] = {.lex_state = 184},
  [599] = {.lex_state = 184},
  [600] = {.lex_state = 184},
  [601] = {.lex_state = 184},
  [602] = {.lex_state = 184},
  [603] = {.lex_state = 184},
  [604] = {.lex_state = 184},
  [605] = {.lex_state = 184},
  [606] = {.lex_state = 184},
  [607] = {.lex_state = 184},
  [608] = {.lex_state = 184},
  [609] = {.lex_state = 184},
  [610] = {.lex_state = 184},
  [611] = {.lex_state = 184},
  [612] = {.lex_state = 184},
  [613] = {.lex_state = 184},
  [614] = {.lex_state = 184},
  [615] = {.lex_state = 184},
  [616] = {.lex_state = 184},
  [617] = {.lex_state = 184},
  [618] = {.lex_state = 184},
  [619] = {.lex_state = 184},
  [620] = {.lex_state = 184},
  [621] = {.lex_state = 184},
  [622] = {.lex_state = 184},
  [623] = {.lex_state = 184},
  [624] = {.lex_state = 184},
  [625] = {.lex_state = 184},
  [626] = {.lex_state = 184},
  [627] = {.lex_state = 184},
  [628] = {.lex_state = 184},
  [629] = {.lex_state = 184},
  [630] = {.lex_state = 184},
  [631] = {.lex_state = 184},
  [632] = {.lex_state = 184},
  [633] = {.lex_state = 184},
  [634] = {.lex_state = 184},
  [635] = {.lex_state = 184},
  [636] = {.lex_state = 184},
  [637] = {.lex_state = 184},
  [638] = {.lex_state = 184},
  [639] = {.lex_state = 184},
  [640] = {.lex_state = 184},
  [641] = {.lex_state = 184},
  [642] = {.lex_state = 184},
  [643] = {.lex_state = 184},
  [644] = {.lex_state = 184},
  [645] = {.lex_state = 184},
  [646] = {.lex_state = 184},
  [647] = {.lex_state = 184},
  [648] = {.lex_state = 184},
  [649] = {.lex_state = 184},
  [650] = {.lex_state = 184},
  [651] = {.lex_state = 184},
  [652] = {.lex_state = 184},
  [653] = {.lex_state = 184},
  [654] = {.lex_state = 184},
  [655] = {.lex_state = 184},
  [656] = {.lex_state = 184},
  [657] = {.lex_state = 186},
  [658] = {.lex_state = 186},
  [659] = {.lex_state = 186},
  [660] = {.lex_state = 4},
  [661] = {.lex_state = 4},
  [662] = {.lex_state = 186},
  [663] = {.lex_state = 4},
  [664] = {.lex_state = 4},
  [665] = {.lex_state = 4},
  [666] = {.lex_state = 4},
  [667] = {.lex_state = 4},
  [668] = {.lex_state = 4},
  [669] = {.lex_state = 7},
  [670] = {.lex_state = 7},
  [671] = {.lex_state = 7},
  [672] = {.lex_state = 7},
  [673] = {.lex_state = 7},
  [674] = {.lex_state = 7},
  [675] = {.lex_state = 186},
  [676] = {.lex_state = 186},
  [677] = {.lex_state = 186},
  [678] = {.lex_state = 186},
  [679] = {.lex_state = 186},
  [680] = {.lex_state = 186},
  [681] = {.lex_state = 186},
  [682] = {.lex_state = 186},
  [683] = {.lex_state = 186},
  [684] = {.lex_state = 186},
  [685] = {.lex_state = 186},
  [686] = {.lex_state = 186},
  [687] = {.lex_state = 186},
  [688] = {.lex_state = 186},
  [689] = {.lex_state = 186},
  [690] = {.lex_state = 186},
  [691] = {.lex_state = 186},
  [692] = {.lex_state = 186},
  [693] = {.lex_state = 186},
  [694] = {.lex_state = 186},
  [695] = {.lex_state = 186},
  [696] = {.lex_state = 186},
  [697] = {.lex_state = 186},
  [698] = {.lex_state = 186},
  [699] = {.lex_state = 186},
  [700] = {.lex_state = 186},
  [701] = {.lex_state = 186},
  [702] = {.lex_state = 186},
  [703] = {.lex_state = 186},
  [704] = {.lex_state = 186},
  [705] = {.lex_state = 186},
  [706] = {.lex_state = 186},
  [707] = {.lex_state = 186},
  [708] = {.lex_state = 186},
  [709] = {.lex_state = 186},
  [710] = {.lex_state = 186},
  [711] = {.lex_state = 186},
  [712] = {.lex_state = 186},
  [713] = {.lex_state = 186},
  [714] = {.lex_state = 186},
  [715] = {.lex_state = 186},
  [716] = {.lex_state = 186},
  [717] = {.lex_state = 186},
  [718] = {.lex_state = 186},
  [719] = {.lex_state = 186},
  [720] = {.lex_state = 186},
  [721] = {.lex_state = 186},
  [722] = {.lex_state = 186},
  [723] = {.lex_state = 186},
  [724] = {.lex_state = 186},
  [725] = {.lex_state = 186},
  [726] = {.lex_state = 186},
  [727] = {.lex_state = 186},
  [728] = {.lex_state = 186},
  [729] = {.lex_state = 186},
  [730] = {.lex_state = 186},
  [731] = {.lex_state = 186},
  [732] = {.lex_state = 186},
  [733] = {.lex_state = 186},
  [734] = {.lex_state = 186},
  [735] = {.lex_state = 186},
  [736] = {.lex_state = 186},
  [737] = {.lex_state = 186},
  [738] = {.lex_state = 186},
  [739] = {.lex_state = 186},
  [740] = {.lex_state = 186},
  [741] = {.lex_state = 186},
  [742] = {.lex_state = 186},
  [743] = {.lex_state = 186},
  [744] = {.lex_state = 186},
  [745] = {.lex_state = 186},
  [746] = {.lex_state = 186},
  [747] = {.lex_state = 186},
  [748] = {.lex_state = 186},
  [749] = {.lex_state = 186},
  [750] = {.lex_state = 186},
  [751] = {.lex_state = 186},
  [752] = {.lex_state = 186},
  [753] = {.lex_state = 186},
  [754] = {.lex_state = 186},
  [755] = {.lex_state = 186},
  [756] = {.lex_state = 186},
  [757] = {.lex_state = 186},
  [758] = {.lex_state = 186},
  [759] = {.lex_state = 186},
  [760] = {.lex_state = 186},
  [761] = {.lex_state = 186},
  [762] = {.lex_state = 186},
  [763] = {.lex_state = 186},
  [764] = {.lex_state = 186},
  [765] = {.lex_state = 186},
  [766] = {.lex_state = 186},
  [767] = {.lex_state = 186},
  [768] = {.lex_state = 186},
  [769] = {.lex_state = 186},
  [770] = {.lex_state = 186},
  [771] = {.lex_state = 186},
  [772] = {.lex_state = 186},
  [773] = {.lex_state = 186},
  [774] = {.lex_state = 186},
  [775] = {.lex_state = 186},
  [776] = {.lex_state = 186},
  [777] = {.lex_state = 186},
  [778] = {.lex_state = 186},
  [779] = {.lex_state = 186},
  [780] = {.lex_state = 186},
  [781] = {.lex_state = 186},
  [782] = {.lex_state = 186},
  [783] = {.lex_state = 186},
  [784] = {.lex_state = 186},
  [785] = {.lex_state = 186},
  [786] = {.lex_state = 186},
  [787] = {.lex_state = 186},
  [788] = {.lex_state = 186},
  [789] = {.lex_state = 186},
  [790] = {.lex_state = 186},
  [791] = {.lex_state = 186},
  [792] = {.lex_state = 186},
  [793] = {.lex_state = 186},
  [794] = {.lex_state = 186},
  [795] = {.lex_state = 186},
  [796] = {.lex_state = 186},
  [797] = {.lex_state = 186},
  [798] = {.lex_state = 186},
  [799] = {.lex_state = 186},
  [800] = {.lex_state = 186},
  [801] = {.lex_state = 186},
  [802] = {.lex_state = 186},
  [803] = {.lex_state = 186},
  [804] = {.lex_state = 186},
  [805] = {.lex_state = 186},
  [806] = {.lex_state = 186},
  [807] = {.lex_state = 186},
  [808] = {.lex_state = 186},
  [809] = {.lex_state = 186},
  [810] = {.lex_state = 186},
  [811] = {.lex_state = 186},
  [812] = {.lex_state = 186},
  [813] = {.lex_state = 186},
  [814] = {.lex_state = 186},
  [815] = {.lex_state = 186},
  [816] = {.lex_state = 186},
  [817] = {.lex_state = 186},
  [818] = {.lex_state = 186},
  [819] = {.lex_state = 186},
  [820] = {.lex_state = 186},
  [821] = {.lex_state = 186},
  [822] = {.lex_state = 186},
  [823] = {.lex_state = 186},
  [824] = {.lex_state = 186},
  [825] = {.lex_state = 186},
  [826] = {.lex_state = 186},
  [827] = {.lex_state = 186},
  [828] = {.lex_state = 186},
  [829] = {.lex_state = 186},
  [830] = {.lex_state = 186},
  [831] = {.lex_state = 186},
  [832] = {.lex_state = 186},
  [833] = {.lex_state = 186},
  [834] = {.lex_state = 186},
  [835] = {.lex_state = 186},
  [836] = {.lex_state = 186},
  [837] = {.lex_state = 186},
  [838] = {.lex_state = 186},
  [839] = {.lex_state = 186},
  [840] = {.lex_state = 186},
  [841] = {.lex_state = 186},
  [842] = {.lex_state = 186},
  [843] = {.lex_state = 186},
  [844] = {.lex_state = 186},
  [845] = {.lex_state = 186},
  [846] = {.lex_state = 186},
  [847] = {.lex_state = 186},
  [848] = {.lex_state = 186},
  [849] = {.lex_state = 186},
  [850] = {.lex_state = 186},
  [851] = {.lex_state = 186},
  [852] = {.lex_state = 186},
  [853] = {.lex_state = 186},
  [854] = {.lex_state = 186},
  [855] = {.lex_state = 186},
  [856] = {.lex_state = 186},
  [857] = {.lex_state = 186},
  [858] = {.lex_state = 186},
  [859] = {.lex_state = 186},
  [860] = {.lex_state = 186},
  [861] = {.lex_state = 186},
  [862] = {.lex_state = 186},
  [863] = {.lex_state = 186},
  [864] = {.lex_state = 186},
  [865] = {.lex_state = 186},
  [866] = {.lex_state = 186},
  [867] = {.lex_state = 186},
  [868] = {.lex_state = 186},
  [869] = {.lex_state = 186},
  [870] = {.lex_state = 186},
  [871] = {.lex_state = 186},
  [872] = {.lex_state = 186},
  [873] = {.lex_state = 186},
  [874] = {.lex_state = 186},
  [875] = {.lex_state = 186},
  [876] = {.lex_state = 186},
  [877] = {.lex_state = 186},
  [878] = {.lex_state = 186},
  [879] = {.lex_state = 186},
  [880] = {.lex_state = 186},
  [881] = {.lex_state = 186},
  [882] = {.lex_state = 186},
  [883] = {.lex_state = 186},
  [884] = {.lex_state = 186},
  [885] = {.lex_state = 186},
  [886] = {.lex_state = 186},
  [887] = {.lex_state = 186},
  [888] = {.lex_state = 186},
  [889] = {.lex_state = 186},
  [890] = {.lex_state = 186},
  [891] = {.lex_state = 186},
  [892] = {.lex_state = 186},
  [893] = {.lex_state = 186},
  [894] = {.lex_state = 186},
  [895] = {.lex_state = 186},
  [896] = {.lex_state = 186},
  [897] = {.lex_state = 186},
  [898] = {.lex_state = 186},
  [899] = {.lex_state = 186},
  [900] = {.lex_state = 186},
  [901] = {.lex_state = 186},
  [902] = {.lex_state = 186},
  [903] = {.lex_state = 186},
  [904] = {.lex_state = 186},
  [905] = {.lex_state = 186},
  [906] = {.lex_state = 186},
  [907] = {.lex_state = 186},
  [908] = {.lex_state = 186},
  [909] = {.lex_state = 186},
  [910] = {.lex_state = 186},
  [911] = {.lex_state = 186},
  [912] = {.lex_state = 186},
  [913] = {.lex_state = 186},
  [914] = {.lex_state = 186},
  [915] = {.lex_state = 186},
  [916] = {.lex_state = 186},
  [917] = {.lex_state = 186},
  [918] = {.lex_state = 186},
  [919] = {.lex_state = 186},
  [920] = {.lex_state = 186},
  [921] = {.lex_state = 186},
  [922] = {.lex_state = 186},
  [923] = {.lex_state = 186},
  [924] = {.lex_state = 186},
  [925] = {.lex_state = 186},
  [926] = {.lex_state = 186},
  [927] = {.lex_state = 186},
  [928] = {.lex_state = 186},
  [929] = {.lex_state = 186},
  [930] = {.lex_state = 186},
  [931] = {.lex_state = 186},
  [932] = {.lex_state = 186},
  [933] = {.lex_state = 186},
  [934] = {.lex_state = 186},
  [935] = {.lex_state = 186},
  [936] = {.lex_state = 186},
  [937] = {.lex_state = 186},
  [938] = {.lex_state = 186},
  [939] = {.lex_state = 186},
  [940] = {.lex_state = 186},
  [941] = {.lex_state = 186},
  [942] = {.lex_state = 186},
  [943] = {.lex_state = 186},
  [944] = {.lex_state = 186},
  [945] = {.lex_state = 186},
  [946] = {.lex_state = 186},
  [947] = {.lex_state = 186},
  [948] = {.lex_state = 186},
  [949] = {.lex_state = 186},
  [950] = {.lex_state = 186},
  [951] = {.lex_state = 186},
  [952] = {.lex_state = 8},
  [953] = {.lex_state = 186},
  [954] = {.lex_state = 186},
  [955] = {.lex_state = 186},
  [956] = {.lex_state = 186},
  [957] = {.lex_state = 186},
  [958] = {.lex_state = 186},
  [959] = {.lex_state = 186},
  [960] = {.lex_state = 186},
  [961] = {.lex_state = 186},
  [962] = {.lex_state = 186},
  [963] = {.lex_state = 186},
  [964] = {.lex_state = 186},
  [965] = {.lex_state = 186},
  [966] = {.lex_state = 186},
  [967] = {.lex_state = 186},
  [968] = {.lex_state = 186},
  [969] = {.lex_state = 186},
  [970] = {.lex_state = 186},
  [971] = {.lex_state = 186},
  [972] = {.lex_state = 186},
  [973] = {.lex_state = 186},
  [974] = {.lex_state = 186},
  [975] = {.lex_state = 186},
  [976] = {.lex_state = 186},
  [977] = {.lex_state = 186},
  [978] = {.lex_state = 186},
  [979] = {.lex_state = 186},
  [980] = {.lex_state = 186},
  [981] = {.lex_state = 186},
  [982] = {.lex_state = 186},
  [983] = {.lex_state = 186},
  [984] = {.lex_state = 186},
  [985] = {.lex_state = 186},
  [986] = {.lex_state = 186},
  [987] = {.lex_state = 186},
  [988] = {.lex_state = 186},
  [989] = {.lex_state = 186},
  [990] = {.lex_state = 186},
  [991] = {.lex_state = 186},
  [992] = {.lex_state = 186},
  [993] = {.lex_state = 186},
  [994] = {.lex_state = 186},
  [995] = {.lex_state = 186},
  [996] = {.lex_state = 186},
  [997] = {.lex_state = 186},
  [998] = {.lex_state = 186},
  [999] = {.lex_state = 186},
  [1000] = {.lex_state = 8},
  [1001] = {.lex_state = 186},
  [1002] = {.lex_state = 186},
  [1003] = {.lex_state = 186},
  [1004] = {.lex_state = 186},
  [1005] = {.lex_state = 186},
  [1006] = {.lex_state = 186},
  [1007] = {.lex_state = 186},
  [1008] = {.lex_state = 186},
  [1009] = {.lex_state = 186},
  [1010] = {.lex_state = 186},
  [1011] = {.lex_state = 186},
  [1012] = {.lex_state = 186},
  [1013] = {.lex_state = 186},
  [1014] = {.lex_state = 186},
  [1015] = {.lex_state = 186},
  [1016] = {.lex_state = 186},
  [1017] = {.lex_state = 186},
  [1018] = {.lex_state = 186},
  [1019] = {.lex_state = 186},
  [1020] = {.lex_state = 186},
  [1021] = {.lex_state = 186},
  [1022] = {.lex_state = 186},
  [1023] = {.lex_state = 186},
  [1024] = {.lex_state = 186},
  [1025] = {.lex_state = 6},
  [1026] = {.lex_state = 9},
  [1027] = {.lex_state = 9},
  [1028] = {.lex_state = 9},
  [1029] = {.lex_state = 9},
  [1030] = {.lex_state = 9},
  [1031] = {.lex_state = 9},
  [1032] = {.lex_state = 9},
  [1033] = {.lex_state = 9},
  [1034] = {.lex_state = 6},
  [1035] = {.lex_state = 6},
  [1036] = {.lex_state = 9},
  [1037] = {.lex_state = 9},
  [1038] = {.lex_state = 11},
  [1039] = {.lex_state = 11},
  [1040] = {.lex_state = 9},
  [1041] = {.lex_state = 6},
  [1042] = {.lex_state = 9},
  [1043] = {.lex_state = 11},
  [1044] = {.lex_state = 9},
  [1045] = {.lex_state = 9},
  [1046] = {.lex_state = 6},
  [1047] = {.lex_state = 11},
  [1048] = {.lex_state = 9},
  [1049] = {.lex_state = 9},
  [1050] = {.lex_state = 9},
  [1051] = {.lex_state = 6},
  [1052] = {.lex_state = 9},
  [1053] = {.lex_state = 9},
  [1054] = {.lex_state = 9},
  [1055] = {.lex_state = 9},
  [1056] = {.lex_state = 9},
  [1057] = {.lex_state = 9},
  [1058] = {.lex_state = 9},
  [1059] = {.lex_state = 6},
  [1060] = {.lex_state = 9},
  [1061] = {.lex_state = 9},
  [1062] = {.lex_state = 9},
  [1063] = {.lex_state = 11},
  [1064] = {.lex_state = 9},
  [1065] = {.lex_state = 9},
  [1066] = {.lex_state = 9},
  [1067] = {.lex_state = 9},
  [1068] = {.lex_state = 9},
  [1069] = {.lex_state = 9},
  [1070] = {.lex_state = 9},
  [1071] = {.lex_state = 9},
  [1072] = {.lex_state = 9},
  [1073] = {.lex_state = 9},
  [1074] = {.lex_state = 9},
  [1075] = {.lex_state = 10},
  [1076] = {.lex_state = 9},
  [1077] = {.lex_state = 11},
  [1078] = {.lex_state = 11},
  [1079] = {.lex_state = 11},
  [1080] = {.lex_state = 6},
  [1081] = {.lex_state = 10},
  [1082] = {.lex_state = 11},
  [1083] = {.lex_state = 6},
  [1084] = {.lex_state = 11},
  [1085] = {.lex_state = 6},
  [1086] = {.lex_state = 6},
  [1087] = {.lex_state = 11},
  [1088] = {.lex_state = 11},
  [1089] = {.lex_state = 6},
  [1090] = {.lex_state = 11},
  [1091] = {.lex_state = 11},
  [1092] = {.lex_state = 11},
  [1093] = {.lex_state = 11},
  [1094] = {.lex_state = 6},
  [1095] = {.lex_state = 11},
  [1096] = {.lex_state = 11},
  [1097] = {.lex_state = 6},
  [1098] = {.lex_state = 11},
  [1099] = {.lex_state = 11},
  [1100] = {.lex_state = 6},
  [1101] = {.lex_state = 11},
  [1102] = {.lex_state = 11},
  [1103] = {.lex_state = 11},
  [1104] = {.lex_state = 11},
  [1105] = {.lex_state = 6},
  [1106] = {.lex_state = 6},
  [1107] = {.lex_state = 11},
  [1108] = {.lex_state = 9},
  [1109] = {.lex_state = 11},
  [1110] = {.lex_state = 6},
  [1111] = {.lex_state = 6},
  [1112] = {.lex_state = 11},
  [1113] = {.lex_state = 11},
  [1114] = {.lex_state = 6},
  [1115] = {.lex_state = 11},
  [1116] = {.lex_state = 6},
  [1117] = {.lex_state = 11},
  [1118] = {.lex_state = 11},
  [1119] = {.lex_state = 6},
  [1120] = {.lex_state = 11},
  [1121] = {.lex_state = 6},
  [1122] = {.lex_state = 6},
  [1123] = {.lex_state = 6},
  [1124] = {.lex_state = 11},
  [1125] = {.lex_state = 11},
  [1126] = {.lex_state = 6},
  [1127] = {.lex_state = 11},
  [1128] = {.lex_state = 6},
  [1129] = {.lex_state = 6},
  [1130] = {.lex_state = 6},
  [1131] = {.lex_state = 6},
  [1132] = {.lex_state = 6},
  [1133] = {.lex_state = 11},
  [1134] = {.lex_state = 6},
  [1135] = {.lex_state = 6},
  [1136] = {.lex_state = 6},
  [1137] = {.lex_state = 6},
  [1138] = {.lex_state = 9},
  [1139] = {.lex_state = 6},
  [1140] = {.lex_state = 6},
  [1141] = {.lex_state = 6},
  [1142] = {.lex_state = 6},
  [1143] = {.lex_state = 6},
  [1144] = {.lex_state = 6},
  [1145] = {.lex_state = 6},
  [1146] = {.lex_state = 6},
  [1147] = {.lex_state = 6},
  [1148] = {.lex_state = 6},
  [1149] = {.lex_state = 6},
  [1150] = {.lex_state = 6},
  [1151] = {.lex_state = 11},
  [1152] = {.lex_state = 9},
  [1153] = {.lex_state = 11},
  [1154] = {.lex_state = 6},
  [1155] = {.lex_state = 11},
  [1156] = {.lex_state = 9},
  [1157] = {.lex_state = 9},
  [1158] = {.lex_state = 6},
  [1159] = {.lex_state = 9},
  [1160] = {.lex_state = 6},
  [1161] = {.lex_state = 6},
  [1162] = {.lex_state = 9},
  [1163] = {.lex_state = 9},
  [1164] = {.lex_state = 9},
  [1165] = {.lex_state = 9},
  [1166] = {.lex_state = 6},
  [1167] = {.lex_state = 9},
  [1168] = {.lex_state = 6},
  [1169] = {.lex_state = 9},
  [1170] = {.lex_state = 6},
  [1171] = {.lex_state = 6},
  [1172] = {.lex_state = 9},
  [1173] = {.lex_state = 9},
  [1174] = {.lex_state = 9},
  [1175] = {.lex_state = 6},
  [1176] = {.lex_state = 6},
  [1177] = {.lex_state = 6},
  [1178] = {.lex_state = 6},
  [1179] = {.lex_state = 6},
  [1180] = {.lex_state = 6},
  [1181] = {.lex_state = 6},
  [1182] = {.lex_state = 6},
  [1183] = {.lex_state = 6},
  [1184] = {.lex_state = 9},
  [1185] = {.lex_state = 6},
  [1186] = {.lex_state = 6},
  [1187] = {.lex_state = 6},
  [1188] = {.lex_state = 6},
  [1189] = {.lex_state = 11},
  [1190] = {.lex_state = 6},
  [1191] = {.lex_state = 6},
  [1192] = {.lex_state = 6},
  [1193] = {.lex_state = 6},
  [1194] = {.lex_state = 6},
  [1195] = {.lex_state = 6},
  [1196] = {.lex_state = 6},
  [1197] = {.lex_state = 6},
  [1198] = {.lex_state = 6},
  [1199] = {.lex_state = 6},
  [1200] = {.lex_state = 6},
  [1201] = {.lex_state = 6},
  [1202] = {.lex_state = 6},
  [1203] = {.lex_state = 6},
  [1204] = {.lex_state = 6},
  [1205] = {.lex_state = 6},
  [1206] = {.lex_state = 6},
  [1207] = {.lex_state = 6},
  [1208] = {.lex_state = 6},
  [1209] = {.lex_state = 6},
  [1210] = {.lex_state = 6},
  [1211] = {.lex_state = 6},
  [1212] = {.lex_state = 6},
  [1213] = {.lex_state = 6},
  [1214] = {.lex_state = 6},
  [1215] = {.lex_state = 6},
  [1216] = {.lex_state = 6},
  [1217] = {.lex_state = 6},
  [1218] = {.lex_state = 6},
  [1219] = {.lex_state = 6},
  [1220] = {.lex_state = 6},
  [1221] = {.lex_state = 6},
  [1222] = {.lex_state = 6},
  [1223] = {.lex_state = 6},
  [1224] = {.lex_state = 6},
  [1225] = {.lex_state = 6},
  [1226] = {.lex_state = 6},
  [1227] = {.lex_state = 6},
  [1228] = {.lex_state = 6},
  [1229] = {.lex_state = 6},
  [1230] = {.lex_state = 6},
  [1231] = {.lex_state = 6},
  [1232] = {.lex_state = 6},
  [1233] = {.lex_state = 6},
  [1234] = {.lex_state = 6},
  [1235] = {.lex_state = 6},
  [1236] = {.lex_state = 6},
  [1237] = {.lex_state = 6},
  [1238] = {.lex_state = 6},
  [1239] = {.lex_state = 6},
  [1240] = {.lex_state = 6},
  [1241] = {.lex_state = 6},
  [1242] = {.lex_state = 6},
  [1243] = {.lex_state = 6},
  [1244] = {.lex_state = 6},
  [1245] = {.lex_state = 6},
  [1246] = {.lex_state = 6},
  [1247] = {.lex_state = 6},
  [1248] = {.lex_state = 6},
  [1249] = {.lex_state = 6},
  [1250] = {.lex_state = 6},
  [1251] = {.lex_state = 6},
  [1252] = {.lex_state = 6},
  [1253] = {.lex_state = 6},
  [1254] = {.lex_state = 6},
  [1255] = {.lex_state = 6},
  [1256] = {.lex_state = 6},
  [1257] = {.lex_state = 6},
  [1258] = {.lex_state = 6},
  [1259] = {.lex_state = 6},
  [1260] = {.lex_state = 6},
  [1261] = {.lex_state = 6},
  [1262] = {.lex_state = 6},
  [1263] = {.lex_state = 6},
  [1264] = {.lex_state = 6},
  [1265] = {.lex_state = 6},
  [1266] = {.lex_state = 6},
  [1267] = {.lex_state = 6},
  [1268] = {.lex_state = 6},
  [1269] = {.lex_state = 6},
  [1270] = {.lex_state = 6},
  [1271] = {.lex_state = 6},
  [1272] = {.lex_state = 6},
  [1273] = {.lex_state = 6},
  [1274] = {.lex_state = 6},
  [1275] = {.lex_state = 6},
  [1276] = {.lex_state = 6},
  [1277] = {.lex_state = 6},
  [1278] = {.lex_state = 6},
  [1279] = {.lex_state = 6},
  [1280] = {.lex_state = 6},
  [1281] = {.lex_state = 6},
  [1282] = {.lex_state = 6},
  [1283] = {.lex_state = 11},
  [1284] = {.lex_state = 6},
  [1285] = {.lex_state = 6},
  [1286] = {.lex_state = 6},
  [1287] = {.lex_state = 6},
  [1288] = {.lex_state = 6},
  [1289] = {.lex_state = 6},
  [1290] = {.lex_state = 6},
  [1291] = {.lex_state = 6},
  [1292] = {.lex_state = 6},
  [1293] = {.lex_state = 6},
  [1294] = {.lex_state = 6},
  [1295] = {.lex_state = 6},
  [1296] = {.lex_state = 6},
  [1297] = {.lex_state = 6},
  [1298] = {.lex_state = 6},
  [1299] = {.lex_state = 6},
  [1300] = {.lex_state = 6},
  [1301] = {.lex_state = 6},
  [1302] = {.lex_state = 6},
  [1303] = {.lex_state = 9},
  [1304] = {.lex_state = 6},
  [1305] = {.lex_state = 11},
  [1306] = {.lex_state = 6},
  [1307] = {.lex_state = 11},
  [1308] = {.lex_state = 11},
  [1309] = {.lex_state = 6},
  [1310] = {.lex_state = 6},
  [1311] = {.lex_state = 6},
  [1312] = {.lex_state = 6},
  [1313] = {.lex_state = 6},
  [1314] = {.lex_state = 6},
  [1315] = {.lex_state = 6},
  [1316] = {.lex_state = 6},
  [1317] = {.lex_state = 6},
  [1318] = {.lex_state = 6},
  [1319] = {.lex_state = 9},
  [1320] = {.lex_state = 6},
  [1321] = {.lex_state = 6},
  [1322] = {.lex_state = 6},
  [1323] = {.lex_state = 6},
  [1324] = {.lex_state = 6},
  [1325] = {.lex_state = 6},
  [1326] = {.lex_state = 6},
  [1327] = {.lex_state = 6},
  [1328] = {.lex_state = 6},
  [1329] = {.lex_state = 11},
  [1330] = {.lex_state = 9},
  [1331] = {.lex_state = 9},
  [1332] = {.lex_state = 6},
  [1333] = {.lex_state = 6},
  [1334] = {.lex_state = 6},
  [1335] = {.lex_state = 6},
  [1336] = {.lex_state = 6},
  [1337] = {.lex_state = 6},
  [1338] = {.lex_state = 6},
  [1339] = {.lex_state = 6},
  [1340] = {.lex_state = 6},
  [1341] = {.lex_state = 6},
  [1342] = {.lex_state = 6},
  [1343] = {.lex_state = 6},
  [1344] = {.lex_state = 6},
  [1345] = {.lex_state = 6},
  [1346] = {.lex_state = 6},
  [1347] = {.lex_state = 6},
  [1348] = {.lex_state = 6},
  [1349] = {.lex_state = 6},
  [1350] = {.lex_state = 6},
  [1351] = {.lex_state = 6},
  [1352] = {.lex_state = 6},
  [1353] = {.lex_state = 6},
  [1354] = {.lex_state = 6},
  [1355] = {.lex_state = 6},
  [1356] = {.lex_state = 6},
  [1357] = {.lex_state = 11},
  [1358] = {.lex_state = 6},
  [1359] = {.lex_state = 6},
  [1360] = {.lex_state = 11},
  [1361] = {.lex_state = 6},
  [1362] = {.lex_state = 6},
  [1363] = {.lex_state = 11},
  [1364] = {.lex_state = 6},
  [1365] = {.lex_state = 6},
  [1366] = {.lex_state = 11},
  [1367] = {.lex_state = 6},
  [1368] = {.lex_state = 11},
  [1369] = {.lex_state = 6},
  [1370] = {.lex_state = 11},
  [1371] = {.lex_state = 6},
  [1372] = {.lex_state = 11},
  [1373] = {.lex_state = 11},
  [1374] = {.lex_state = 11},
  [1375] = {.lex_state = 11},
  [1376] = {.lex_state = 6},
  [1377] = {.lex_state = 11},
  [1378] = {.lex_state = 6},
  [1379] = {.lex_state = 11},
  [1380] = {.lex_state = 11},
  [1381] = {.lex_state = 6},
  [1382] = {.lex_state = 6},
  [1383] = {.lex_state = 9},
  [1384] = {.lex_state = 9},
  [1385] = {.lex_state = 9},
  [1386] = {.lex_state = 9},
  [1387] = {.lex_state = 9},
  [1388] = {.lex_state = 9},
  [1389] = {.lex_state = 9},
  [1390] = {.lex_state = 9},
  [1391] = {.lex_state = 9},
  [1392] = {.lex_state = 9},
  [1393] = {.lex_state = 9},
  [1394] = {.lex_state = 9},
  [1395] = {.lex_state = 10},
  [1396] = {.lex_state = 9},
  [1397] = {.lex_state = 11},
  [1398] = {.lex_state = 9},
  [1399] = {.lex_state = 11},
  [1400] = {.lex_state = 11},
  [1401] = {.lex_state = 9},
  [1402] = {.lex_state = 11},
  [1403] = {.lex_state = 11},
  [1404] = {.lex_state = 11},
  [1405] = {.lex_state = 11},
  [1406] = {.lex_state = 11},
  [1407] = {.lex_state = 11},
  [1408] = {.lex_state = 9},
  [1409] = {.lex_state = 11},
  [1410] = {.lex_state = 11},
  [1411] = {.lex_state = 11},
  [1412] = {.lex_state = 11},
  [1413] = {.lex_state = 11},
  [1414] = {.lex_state = 11},
  [1415] = {.lex_state = 9},
  [1416] = {.lex_state = 11},
  [1417] = {.lex_state = 11},
  [1418] = {.lex_state = 11},
  [1419] = {.lex_state = 11},
  [1420] = {.lex_state = 11},
  [1421] = {.lex_state = 11},
  [1422] = {.lex_state = 9},
  [1423] = {.lex_state = 11},
  [1424] = {.lex_state = 11},
  [1425] = {.lex_state = 11},
  [1426] = {.lex_state = 11},
  [1427] = {.lex_state = 9},
  [1428] = {.lex_state = 9},
  [1429] = {.lex_state = 9},
  [1430] = {.lex_state = 9},
  [1431] = {.lex_state = 9},
  [1432] = {.lex_state = 9},
  [1433] = {.lex_state = 9},
  [1434] = {.lex_state = 9},
  [1435] = {.lex_state = 11},
  [1436] = {.lex_state = 9},
  [1437] = {.lex_state = 9},
  [1438] = {.lex_state = 11},
  [1439] = {.lex_state = 9},
  [1440] = {.lex_state = 9},
  [1441] = {.lex_state = 11},
  [1442] = {.lex_state = 11},
  [1443] = {.lex_state = 11},
  [1444] = {.lex_state = 11},
  [1445] = {.lex_state = 11},
  [1446] = {.lex_state = 11},
  [1447] = {.lex_state = 11},
  [1448] = {.lex_state = 11},
  [1449] = {.lex_state = 11},
  [1450] = {.lex_state = 11},
  [1451] = {.lex_state = 11},
  [1452] = {.lex_state = 11},
  [1453] = {.lex_state = 11},
  [1454] = {.lex_state = 11},
  [1455] = {.lex_state = 11},
  [1456] = {.lex_state = 11},
  [1457] = {.lex_state = 11},
  [1458] = {.lex_state = 11},
  [1459] = {.lex_state = 11},
  [1460] = {.lex_state = 11},
  [1461] = {.lex_state = 11},
  [1462] = {.lex_state = 11},
  [1463] = {.lex_state = 11},
  [1464] = {.lex_state = 11},
  [1465] = {.lex_state = 11},
  [1466] = {.lex_state = 11},
  [1467] = {.lex_state = 11},
  [1468] = {.lex_state = 11},
  [1469] = {.lex_state = 11},
  [1470] = {.lex_state = 11},
  [1471] = {.lex_state = 11},
  [1472] = {.lex_state = 11},
  [1473] = {.lex_state = 11},
  [1474] = {.lex_state = 11},
  [1475] = {.lex_state = 11},
  [1476] = {.lex_state = 11},
  [1477] = {.lex_state = 11},
  [1478] = {.lex_state = 11},
  [1479] = {.lex_state = 11},
  [1480] = {.lex_state = 11},
  [1481] = {.lex_state = 11},
  [1482] = {.lex_state = 11},
  [1483] = {.lex_state = 11},
  [1484] = {.lex_state = 11},
  [1485] = {.lex_state = 11},
  [1486] = {.lex_state = 11},
  [1487] = {.lex_state = 11},
  [1488] = {.lex_state = 11},
  [1489] = {.lex_state = 11},
  [1490] = {.lex_state = 11},
  [1491] = {.lex_state = 11},
  [1492] = {.lex_state = 11},
  [1493] = {.lex_state = 11},
  [1494] = {.lex_state = 11},
  [1495] = {.lex_state = 11},
  [1496] = {.lex_state = 11},
  [1497] = {.lex_state = 11},
  [1498] = {.lex_state = 11},
  [1499] = {.lex_state = 11},
  [1500] = {.lex_state = 11},
  [1501] = {.lex_state = 11},
  [1502] = {.lex_state = 11},
  [1503] = {.lex_state = 11},
  [1504] = {.lex_state = 11},
  [1505] = {.lex_state = 11},
  [1506] = {.lex_state = 11},
  [1507] = {.lex_state = 11},
  [1508] = {.lex_state = 11},
  [1509] = {.lex_state = 11},
  [1510] = {.lex_state = 11},
  [1511] = {.lex_state = 11},
  [1512] = {.lex_state = 11},
  [1513] = {.lex_state = 11},
  [1514] = {.lex_state = 11},
  [1515] = {.lex_state = 11},
  [1516] = {.lex_state = 11},
  [1517] = {.lex_state = 11},
  [1518] = {.lex_state = 11},
  [1519] = {.lex_state = 11},
  [1520] = {.lex_state = 11},
  [1521] = {.lex_state = 11},
  [1522] = {.lex_state = 11},
  [1523] = {.lex_state = 11},
  [1524] = {.lex_state = 11},
  [1525] = {.lex_state = 11},
  [1526] = {.lex_state = 11},
  [1527] = {.lex_state = 11},
  [1528] = {.lex_state = 11},
  [1529] = {.lex_state = 11},
  [1530] = {.lex_state = 11},
  [1531] = {.lex_state = 11},
  [1532] = {.lex_state = 11},
  [1533] = {.lex_state = 11},
  [1534] = {.lex_state = 11},
  [1535] = {.lex_state = 11},
  [1536] = {.lex_state = 11},
  [1537] = {.lex_state = 11},
  [1538] = {.lex_state = 11},
  [1539] = {.lex_state = 11},
  [1540] = {.lex_state = 11},
  [1541] = {.lex_state = 11},
  [1542] = {.lex_state = 11},
  [1543] = {.lex_state = 11},
  [1544] = {.lex_state = 11},
  [1545] = {.lex_state = 11},
  [1546] = {.lex_state = 11},
  [1547] = {.lex_state = 11},
  [1548] = {.lex_state = 6},
  [1549] = {.lex_state = 6},
  [1550] = {.lex_state = 6},
  [1551] = {.lex_state = 6},
  [1552] = {.lex_state = 6},
  [1553] = {.lex_state = 6},
  [1554] = {.lex_state = 6},
  [1555] = {.lex_state = 6},
  [1556] = {.lex_state = 6},
  [1557] = {.lex_state = 6},
  [1558] = {.lex_state = 6},
  [1559] = {.lex_state = 6},
  [1560] = {.lex_state = 6},
  [1561] = {.lex_state = 6},
  [1562] = {.lex_state = 6},
  [1563] = {.lex_state = 6},
  [1564] = {.lex_state = 6},
  [1565] = {.lex_state = 6},
  [1566] = {.lex_state = 6},
  [1567] = {.lex_state = 6},
  [1568] = {.lex_state = 6},
  [1569] = {.lex_state = 6},
  [1570] = {.lex_state = 6},
  [1571] = {.lex_state = 6},
  [1572] = {.lex_state = 6},
  [1573] = {.lex_state = 6},
  [1574] = {.lex_state = 6},
  [1575] = {.lex_state = 6},
  [1576] = {.lex_state = 6},
  [1577] = {.lex_state = 6},
  [1578] = {.lex_state = 6},
  [1579] = {.lex_state = 6},
  [1580] = {.lex_state = 6},
  [1581] = {.lex_state = 6},
  [1582] = {.lex_state = 6},
  [1583] = {.lex_state = 6},
  [1584] = {.lex_state = 6},
  [1585] = {.lex_state = 6},
  [1586] = {.lex_state = 6},
  [1587] = {.lex_state = 6},
  [1588] = {.lex_state = 6},
  [1589] = {.lex_state = 6},
  [1590] = {.lex_state = 6},
  [1591] = {.lex_state = 6},
  [1592] = {.lex_state = 6},
  [1593] = {.lex_state = 6},
  [1594] = {.lex_state = 6},
  [1595] = {.lex_state = 6},
  [1596] = {.lex_state = 6},
  [1597] = {.lex_state = 6},
  [1598] = {.lex_state = 6},
  [1599] = {.lex_state = 6},
  [1600] = {.lex_state = 6},
  [1601] = {.lex_state = 6},
  [1602] = {.lex_state = 6},
  [1603] = {.lex_state = 20},
  [1604] = {.lex_state = 20},
  [1605] = {.lex_state = 16},
  [1606] = {.lex_state = 20},
  [1607] = {.lex_state = 20},
  [1608] = {.lex_state = 20},
  [1609] = {.lex_state = 16},
  [1610] = {.lex_state = 16},
  [1611] = {.lex_state = 16},
  [1612] = {.lex_state = 16},
  [1613] = {.lex_state = 16},
  [1614] = {.lex_state = 20},
  [1615] = {.lex_state = 20},
  [1616] = {.lex_state = 20},
  [1617] = {.lex_state = 20},
  [1618] = {.lex_state = 20},
  [1619] = {.lex_state = 20},
  [1620] = {.lex_state = 20},
  [1621] = {.lex_state = 20},
  [1622] = {.lex_state = 20},
  [1623] = {.lex_state = 20},
  [1624] = {.lex_state = 20},
  [1625] = {.lex_state = 6},
  [1626] = {.lex_state = 6},
  [1627] = {.lex_state = 6},
  [1628] = {.lex_state = 6},
  [1629] = {.lex_state = 6},
  [1630] = {.lex_state = 6},
  [1631] = {.lex_state = 6},
  [1632] = {.lex_state = 6},
  [1633] = {.lex_state = 6},
  [1634] = {.lex_state = 6},
  [1635] = {.lex_state = 6},
  [1636] = {.lex_state = 6},
  [1637] = {.lex_state = 6},
  [1638] = {.lex_state = 6},
  [1639] = {.lex_state = 6},
  [1640] = {.lex_state = 6},
  [1641] = {.lex_state = 6},
  [1642] = {.lex_state = 6},
  [1643] = {.lex_state = 6},
  [1644] = {.lex_state = 6},
  [1645] = {.lex_state = 6},
  [1646] = {.lex_state = 6},
  [1647] = {.lex_state = 6},
  [1648] = {.lex_state = 6},
  [1649] = {.lex_state = 6},
  [1650] = {.lex_state = 6},
  [1651] = {.lex_state = 6},
  [1652] = {.lex_state = 6},
  [1653] = {.lex_state = 6},
  [1654] = {.lex_state = 6},
  [1655] = {.lex_state = 28},
  [1656] = {.lex_state = 48},
  [1657] = {.lex_state = 48},
  [1658] = {.lex_state = 48},
  [1659] = {.lex_state = 48},
  [1660] = {.lex_state = 48},
  [1661] = {.lex_state = 48},
  [1662] = {.lex_state = 48},
  [1663] = {.lex_state = 48},
  [1664] = {.lex_state = 48},
  [1665] = {.lex_state = 48},
  [1666] = {.lex_state = 48},
  [1667] = {.lex_state = 48},
  [1668] = {.lex_state = 48},
  [1669] = {.lex_state = 48},
  [1670] = {.lex_state = 48},
  [1671] = {.lex_state = 48},
  [1672] = {.lex_state = 48},
  [1673] = {.lex_state = 48},
  [1674] = {.lex_state = 16},
  [1675] = {.lex_state = 16},
  [1676] = {.lex_state = 16},
  [1677] = {.lex_state = 16},
  [1678] = {.lex_state = 16},
  [1679] = {.lex_state = 16},
  [1680] = {.lex_state = 16},
  [1681] = {.lex_state = 16},
  [1682] = {.lex_state = 16},
  [1683] = {.lex_state = 16},
  [1684] = {.lex_state = 16},
  [1685] = {.lex_state = 16},
  [1686] = {.lex_state = 16},
  [1687] = {.lex_state = 16},
  [1688] = {.lex_state = 16},
  [1689] = {.lex_state = 16},
  [1690] = {.lex_state = 16},
  [1691] = {.lex_state = 16},
  [1692] = {.lex_state = 6},
  [1693] = {.lex_state = 6},
  [1694] = {.lex_state = 6},
  [1695] = {.lex_state = 49},
  [1696] = {.lex_state = 6},
  [1697] = {.lex_state = 6},
  [1698] = {.lex_state = 49},
  [1699] = {.lex_state = 49},
  [1700] = {.lex_state = 6},
  [1701] = {.lex_state = 6},
  [1702] = {.lex_state = 16},
  [1703] = {.lex_state = 16},
  [1704] = {.lex_state = 28},
  [1705] = {.lex_state = 28},
  [1706] = {.lex_state = 16},
  [1707] = {.lex_state = 16},
  [1708] = {.lex_state = 16},
  [1709] = {.lex_state = 16},
  [1710] = {.lex_state = 16},
  [1711] = {.lex_state = 16},
  [1712] = {.lex_state = 16},
  [1713] = {.lex_state = 16},
  [1714] = {.lex_state = 28},
  [1715] = {.lex_state = 28},
  [1716] = {.lex_state = 28},
  [1717] = {.lex_state = 20},
  [1718] = {.lex_state = 48},
  [1719] = {.lex_state = 28},
  [1720] = {.lex_state = 28},
  [1721] = {.lex_state = 28},
  [1722] = {.lex_state = 20},
  [1723] = {.lex_state = 20},
  [1724] = {.lex_state = 20},
  [1725] = {.lex_state = 30},
  [1726] = {.lex_state = 30},
  [1727] = {.lex_state = 20},
  [1728] = {.lex_state = 20},
  [1729] = {.lex_state = 30},
  [1730] = {.lex_state = 20},
  [1731] = {.lex_state = 30},
  [1732] = {.lex_state = 20},
  [1733] = {.lex_state = 28},
  [1734] = {.lex_state = 28},
  [1735] = {.lex_state = 28},
  [1736] = {.lex_state = 28},
  [1737] = {.lex_state = 30},
  [1738] = {.lex_state = 30},
  [1739] = {.lex_state = 28},
  [1740] = {.lex_state = 28},
  [1741] = {.lex_state = 30},
  [1742] = {.lex_state = 28},
  [1743] = {.lex_state = 30},
  [1744] = {.lex_state = 30},
  [1745] = {.lex_state = 28},
  [1746] = {.lex_state = 28},
  [1747] = {.lex_state = 28},
  [1748] = {.lex_state = 48},
  [1749] = {.lex_state = 48},
  [1750] = {.lex_state = 30},
  [1751] = {.lex_state = 49},
  [1752] = {.lex_state = 16},
  [1753] = {.lex_state = 6},
  [1754] = {.lex_state = 6},
  [1755] = {.lex_state = 49},
  [1756] = {.lex_state = 6},
  [1757] = {.lex_state = 6},
  [1758] = {.lex_state = 6},
  [1759] = {.lex_state = 6},
  [1760] = {.lex_state = 6},
  [1761] = {.lex_state = 49},
  [1762] = {.lex_state = 28},
  [1763] = {.lex_state = 6},
  [1764] = {.lex_state = 6},
  [1765] = {.lex_state = 28},
  [1766] = {.lex_state = 6},
  [1767] = {.lex_state = 6},
  [1768] = {.lex_state = 28},
  [1769] = {.lex_state = 6},
  [1770] = {.lex_state = 28},
  [1771] = {.lex_state = 6},
  [1772] = {.lex_state = 28},
  [1773] = {.lex_state = 28},
  [1774] = {.lex_state = 30},
  [1775] = {.lex_state = 30},
  [1776] = {.lex_state = 30},
  [1777] = {.lex_state = 30},
  [1778] = {.lex_state = 30},
  [1779] = {.lex_state = 30},
  [1780] = {.lex_state = 30},
  [1781] = {.lex_state = 30},
  [1782] = {.lex_state = 30},
  [1783] = {.lex_state = 30},
  [1784] = {.lex_state = 30},
  [1785] = {.lex_state = 26},
  [1786] = {.lex_state = 26},
  [1787] = {.lex_state = 28},
  [1788] = {.lex_state = 26},
  [1789] = {.lex_state = 26},
  [1790] = {.lex_state = 26},
  [1791] = {.lex_state = 26},
  [1792] = {.lex_state = 26},
  [1793] = {.lex_state = 28},
  [1794] = {.lex_state = 26},
  [1795] = {.lex_state = 8},
  [1796] = {.lex_state = 8},
  [1797] = {.lex_state = 8},
  [1798] = {.lex_state = 33},
  [1799] = {.lex_state = 28},
  [1800] = {.lex_state = 8},
  [1801] = {.lex_state = 8},
  [1802] = {.lex_state = 8},
  [1803] = {.lex_state = 33},
  [1804] = {.lex_state = 8},
  [1805] = {.lex_state = 8},
  [1806] = {.lex_state = 33},
  [1807] = {.lex_state = 26},
  [1808] = {.lex_state = 26},
  [1809] = {.lex_state = 51},
  [1810] = {.lex_state = 26},
  [1811] = {.lex_state = 26},
  [1812] = {.lex_state = 8},
  [1813] = {.lex_state = 26},
  [1814] = {.lex_state = 51},
  [1815] = {.lex_state = 26},
  [1816] = {.lex_state = 26},
  [1817] = {.lex_state = 51},
  [1818] = {.lex_state = 26},
  [1819] = {.lex_state = 26},
  [1820] = {.lex_state = 51},
  [1821] = {.lex_state = 8},
  [1822] = {.lex_state = 26},
  [1823] = {.lex_state = 51},
  [1824] = {.lex_state = 51},
  [1825] = {.lex_state = 26},
  [1826] = {.lex_state = 13},
  [1827] = {.lex_state = 17},
  [1828] = {.lex_state = 8},
  [1829] = {.lex_state = 8},
  [1830] = {.lex_state = 13},
  [1831] = {.lex_state = 8},
  [1832] = {.lex_state = 13},
  [1833] = {.lex_state = 18},
  [1834] = {.lex_state = 186},
  [1835] = {.lex_state = 18},
  [1836] = {.lex_state = 8},
  [1837] = {.lex_state = 8},
  [1838] = {.lex_state = 13},
  [1839] = {.lex_state = 13},
  [1840] = {.lex_state = 18},
  [1841] = {.lex_state = 8},
  [1842] = {.lex_state = 13},
  [1843] = {.lex_state = 18},
  [1844] = {.lex_state = 13},
  [1845] = {.lex_state = 18},
  [1846] = {.lex_state = 13},
  [1847] = {.lex_state = 18},
  [1848] = {.lex_state = 8},
  [1849] = {.lex_state = 13},
  [1850] = {.lex_state = 13},
  [1851] = {.lex_state = 13},
  [1852] = {.lex_state = 13},
  [1853] = {.lex_state = 18},
  [1854] = {.lex_state = 13},
  [1855] = {.lex_state = 18},
  [1856] = {.lex_state = 8},
  [1857] = {.lex_state = 18},
  [1858] = {.lex_state = 8},
  [1859] = {.lex_state = 17},
  [1860] = {.lex_state = 186},
  [1861] = {.lex_state = 17},
  [1862] = {.lex_state = 8},
  [1863] = {.lex_state = 18},
  [1864] = {.lex_state = 18},
  [1865] = {.lex_state = 17},
  [1866] = {.lex_state = 8},
  [1867] = {.lex_state = 18},
  [1868] = {.lex_state = 17},
  [1869] = {.lex_state = 186},
  [1870] = {.lex_state = 17},
  [1871] = {.lex_state = 17},
  [1872] = {.lex_state = 18},
  [1873] = {.lex_state = 1},
  [1874] = {.lex_state = 1},
  [1875] = {.lex_state = 1},
  [1876] = {.lex_state = 17},
  [1877] = {.lex_state = 1},
  [1878] = {.lex_state = 33},
  [1879] = {.lex_state = 22},
  [1880] = {.lex_state = 1},
  [1881] = {.lex_state = 22},
  [1882] = {.lex_state = 1},
  [1883] = {.lex_state = 186},
  [1884] = {.lex_state = 22},
  [1885] = {.lex_state = 26},
  [1886] = {.lex_state = 17},
  [1887] = {.lex_state = 22},
  [1888] = {.lex_state = 26},
  [1889] = {.lex_state = 17},
  [1890] = {.lex_state = 17},
  [1891] = {.lex_state = 22},
  [1892] = {.lex_state = 22},
  [1893] = {.lex_state = 1},
  [1894] = {.lex_state = 8},
  [1895] = {.lex_state = 33},
  [1896] = {.lex_state = 8},
  [1897] = {.lex_state = 22},
  [1898] = {.lex_state = 33},
  [1899] = {.lex_state = 1},
  [1900] = {.lex_state = 1},
  [1901] = {.lex_state = 17},
  [1902] = {.lex_state = 1},
  [1903] = {.lex_state = 22},
  [1904] = {.lex_state = 26},
  [1905] = {.lex_state = 8},
  [1906] = {.lex_state = 8},
  [1907] = {.lex_state = 17},
  [1908] = {.lex_state = 22},
  [1909] = {.lex_state = 22},
  [1910] = {.lex_state = 22},
  [1911] = {.lex_state = 1},
  [1912] = {.lex_state = 22},
  [1913] = {.lex_state = 1},
  [1914] = {.lex_state = 22},
  [1915] = {.lex_state = 1},
  [1916] = {.lex_state = 27},
  [1917] = {.lex_state = 26},
  [1918] = {.lex_state = 27},
  [1919] = {.lex_state = 8},
  [1920] = {.lex_state = 186},
  [1921] = {.lex_state = 8},
  [1922] = {.lex_state = 8},
  [1923] = {.lex_state = 186},
  [1924] = {.lex_state = 186},
  [1925] = {.lex_state = 8},
  [1926] = {.lex_state = 186},
  [1927] = {.lex_state = 17},
  [1928] = {.lex_state = 186},
  [1929] = {.lex_state = 8},
  [1930] = {.lex_state = 186},
  [1931] = {.lex_state = 186},
  [1932] = {.lex_state = 186},
  [1933] = {.lex_state = 186},
  [1934] = {.lex_state = 186},
  [1935] = {.lex_state = 26},
  [1936] = {.lex_state = 8},
  [1937] = {.lex_state = 8},
  [1938] = {.lex_state = 8},
  [1939] = {.lex_state = 16},
  [1940] = {.lex_state = 8},
  [1941] = {.lex_state = 186},
  [1942] = {.lex_state = 186},
  [1943] = {.lex_state = 27},
  [1944] = {.lex_state = 8},
  [1945] = {.lex_state = 186},
  [1946] = {.lex_state = 186},
  [1947] = {.lex_state = 8},
  [1948] = {.lex_state = 186},
  [1949] = {.lex_state = 18},
  [1950] = {.lex_state = 186},
  [1951] = {.lex_state = 186},
  [1952] = {.lex_state = 186},
  [1953] = {.lex_state = 8},
  [1954] = {.lex_state = 186},
  [1955] = {.lex_state = 18},
  [1956] = {.lex_state = 17},
  [1957] = {.lex_state = 186},
  [1958] = {.lex_state = 8},
  [1959] = {.lex_state = 186},
  [1960] = {.lex_state = 186},
  [1961] = {.lex_state = 186},
  [1962] = {.lex_state = 186},
  [1963] = {.lex_state = 186},
  [1964] = {.lex_state = 22},
  [1965] = {.lex_state = 27},
  [1966] = {.lex_state = 26},
  [1967] = {.lex_state = 8},
  [1968] = {.lex_state = 16},
  [1969] = {.lex_state = 27},
  [1970] = {.lex_state = 8},
  [1971] = {.lex_state = 186},
  [1972] = {.lex_state = 27},
  [1973] = {.lex_state = 18},
  [1974] = {.lex_state = 27},
  [1975] = {.lex_state = 26},
  [1976] = {.lex_state = 27},
  [1977] = {.lex_state = 186},
  [1978] = {.lex_state = 186},
  [1979] = {.lex_state = 13},
  [1980] = {.lex_state = 8},
  [1981] = {.lex_state = 8},
  [1982] = {.lex_state = 8},
  [1983] = {.lex_state = 27},
  [1984] = {.lex_state = 8},
  [1985] = {.lex_state = 27},
  [1986] = {.lex_state = 186},
  [1987] = {.lex_state = 26},
  [1988] = {.lex_state = 13},
  [1989] = {.lex_state = 8},
  [1990] = {.lex_state = 8},
  [1991] = {.lex_state = 186},
  [1992] = {.lex_state = 8},
  [1993] = {.lex_state = 8},
  [1994] = {.lex_state = 8},
  [1995] = {.lex_state = 186},
  [1996] = {.lex_state = 8},
  [1997] = {.lex_state = 13},
  [1998] = {.lex_state = 186},
  [1999] = {.lex_state = 8},
  [2000] = {.lex_state = 186},
  [2001] = {.lex_state = 186},
  [2002] = {.lex_state = 22},
  [2003] = {.lex_state = 186},
  [2004] = {.lex_state = 186},
  [2005] = {.lex_state = 186},
  [2006] = {.lex_state = 22},
  [2007] = {.lex_state = 26},
  [2008] = {.lex_state = 16},
  [2009] = {.lex_state = 8},
  [2010] = {.lex_state = 8},
  [2011] = {.lex_state = 186},
  [2012] = {.lex_state = 8},
  [2013] = {.lex_state = 27},
  [2014] = {.lex_state = 13},
  [2015] = {.lex_state = 8},
  [2016] = {.lex_state = 18},
  [2017] = {.lex_state = 8},
  [2018] = {.lex_state = 186},
  [2019] = {.lex_state = 8},
  [2020] = {.lex_state = 8},
  [2021] = {.lex_state = 27},
  [2022] = {.lex_state = 8},
  [2023] = {.lex_state = 186},
  [2024] = {.lex_state = 8},
  [2025] = {.lex_state = 186},
  [2026] = {.lex_state = 8},
  [2027] = {.lex_state = 186},
  [2028] = {.lex_state = 186},
  [2029] = {.lex_state = 186},
  [2030] = {.lex_state = 186},
  [2031] = {.lex_state = 186},
  [2032] = {.lex_state = 186},
  [2033] = {.lex_state = 17},
  [2034] = {.lex_state = 186},
  [2035] = {.lex_state = 186},
  [2036] = {.lex_state = 8},
  [2037] = {.lex_state = 186},
  [2038] = {.lex_state = 26},
  [2039] = {.lex_state = 8},
  [2040] = {.lex_state = 27},
  [2041] = {.lex_state = 186},
  [2042] = {.lex_state = 2},
  [2043] = {.lex_state = 10},
  [2044] = {.lex_state = 186},
  [2045] = {.lex_state = 10},
  [2046] = {.lex_state = 2},
  [2047] = {.lex_state = 10},
  [2048] = {.lex_state = 10},
  [2049] = {.lex_state = 10},
  [2050] = {.lex_state = 10},
  [2051] = {.lex_state = 27},
  [2052] = {.lex_state = 10},
  [2053] = {.lex_state = 2},
  [2054] = {.lex_state = 8},
  [2055] = {.lex_state = 186},
  [2056] = {.lex_state = 10},
  [2057] = {.lex_state = 10},
  [2058] = {.lex_state = 10},
  [2059] = {.lex_state = 10},
  [2060] = {.lex_state = 27},
  [2061] = {.lex_state = 27},
  [2062] = {.lex_state = 10},
  [2063] = {.lex_state = 2},
  [2064] = {.lex_state = 27},
  [2065] = {.lex_state = 2},
  [2066] = {.lex_state = 27},
  [2067] = {.lex_state = 10},
  [2068] = {.lex_state = 27},
  [2069] = {.lex_state = 27},
  [2070] = {.lex_state = 27},
  [2071] = {.lex_state = 10},
  [2072] = {.lex_state = 10},
  [2073] = {.lex_state = 2},
  [2074] = {.lex_state = 27},
  [2075] = {.lex_state = 186},
  [2076] = {.lex_state = 2},
  [2077] = {.lex_state = 2},
  [2078] = {.lex_state = 1},
  [2079] = {.lex_state = 10},
  [2080] = {.lex_state = 186},
  [2081] = {.lex_state = 0},
  [2082] = {.lex_state = 10},
  [2083] = {.lex_state = 27},
  [2084] = {.lex_state = 27},
  [2085] = {.lex_state = 10},
  [2086] = {.lex_state = 1},
  [2087] = {.lex_state = 27},
  [2088] = {.lex_state = 27},
  [2089] = {.lex_state = 2},
  [2090] = {.lex_state = 27},
  [2091] = {.lex_state = 10},
  [2092] = {.lex_state = 27},
  [2093] = {.lex_state = 10},
  [2094] = {.lex_state = 186},
  [2095] = {.lex_state = 10},
  [2096] = {.lex_state = 2},
  [2097] = {.lex_state = 186},
  [2098] = {.lex_state = 10},
  [2099] = {.lex_state = 1},
  [2100] = {.lex_state = 10},
  [2101] = {.lex_state = 2},
  [2102] = {.lex_state = 10},
  [2103] = {.lex_state = 10},
  [2104] = {.lex_state = 10},
  [2105] = {.lex_state = 27},
  [2106] = {.lex_state = 27},
  [2107] = {.lex_state = 27},
  [2108] = {.lex_state = 10},
  [2109] = {.lex_state = 16},
  [2110] = {.lex_state = 2},
  [2111] = {.lex_state = 27},
  [2112] = {.lex_state = 27},
  [2113] = {.lex_state = 27},
  [2114] = {.lex_state = 27},
  [2115] = {.lex_state = 2},
  [2116] = {.lex_state = 27},
  [2117] = {.lex_state = 10},
  [2118] = {.lex_state = 27},
  [2119] = {.lex_state = 10},
  [2120] = {.lex_state = 10},
  [2121] = {.lex_state = 8},
  [2122] = {.lex_state = 27},
  [2123] = {.lex_state = 0},
  [2124] = {.lex_state = 27},
  [2125] = {.lex_state = 26},
  [2126] = {.lex_state = 27},
  [2127] = {.lex_state = 8},
  [2128] = {.lex_state = 0},
  [2129] = {.lex_state = 27},
  [2130] = {.lex_state = 8},
  [2131] = {.lex_state = 27},
  [2132] = {.lex_state = 0},
  [2133] = {.lex_state = 27},
  [2134] = {.lex_state = 0},
  [2135] = {.lex_state = 27},
  [2136] = {.lex_state = 8},
  [2137] = {.lex_state = 27},
  [2138] = {.lex_state = 8},
  [2139] = {.lex_state = 0},
  [2140] = {.lex_state = 27},
  [2141] = {.lex_state = 27},
  [2142] = {.lex_state = 27},
  [2143] = {.lex_state = 8},
  [2144] = {.lex_state = 0},
  [2145] = {.lex_state = 27},
  [2146] = {.lex_state = 27},
  [2147] = {.lex_state = 8},
  [2148] = {.lex_state = 27},
  [2149] = {.lex_state = 8},
  [2150] = {.lex_state = 2},
  [2151] = {.lex_state = 27},
  [2152] = {.lex_state = 27},
  [2153] = {.lex_state = 27},
  [2154] = {.lex_state = 8},
  [2155] = {.lex_state = 27},
  [2156] = {.lex_state = 27},
  [2157] = {.lex_state = 27},
  [2158] = {.lex_state = 27},
  [2159] = {.lex_state = 27},
  [2160] = {.lex_state = 16},
  [2161] = {.lex_state = 27},
  [2162] = {.lex_state = 27},
  [2163] = {.lex_state = 0},
  [2164] = {.lex_state = 27},
  [2165] = {.lex_state = 8},
  [2166] = {.lex_state = 27},
  [2167] = {.lex_state = 27},
  [2168] = {.lex_state = 25},
  [2169] = {.lex_state = 8},
  [2170] = {.lex_state = 8},
  [2171] = {.lex_state = 8},
  [2172] = {.lex_state = 27},
  [2173] = {.lex_state = 8},
  [2174] = {.lex_state = 8},
  [2175] = {.lex_state = 0},
  [2176] = {.lex_state = 0},
  [2177] = {.lex_state = 0},
  [2178] = {.lex_state = 8},
  [2179] = {.lex_state = 8},
  [2180] = {.lex_state = 8},
  [2181] = {.lex_state = 0},
  [2182] = {.lex_state = 27},
  [2183] = {.lex_state = 0},
  [2184] = {.lex_state = 8},
  [2185] = {.lex_state = 8},
  [2186] = {.lex_state = 27},
  [2187] = {.lex_state = 0},
  [2188] = {.lex_state = 8},
  [2189] = {.lex_state = 8},
  [2190] = {.lex_state = 0},
  [2191] = {.lex_state = 0},
  [2192] = {.lex_state = 8},
  [2193] = {.lex_state = 27},
  [2194] = {.lex_state = 27},
  [2195] = {.lex_state = 8},
  [2196] = {.lex_state = 27},
  [2197] = {.lex_state = 27},
  [2198] = {.lex_state = 0},
  [2199] = {.lex_state = 27},
  [2200] = {.lex_state = 27},
  [2201] = {.lex_state = 8},
  [2202] = {.lex_state = 8},
  [2203] = {.lex_state = 27},
  [2204] = {.lex_state = 27},
  [2205] = {.lex_state = 8},
  [2206] = {.lex_state = 27},
  [2207] = {.lex_state = 8},
  [2208] = {.lex_state = 27},
  [2209] = {.lex_state = 27},
  [2210] = {.lex_state = 8},
  [2211] = {.lex_state = 27},
  [2212] = {.lex_state = 0},
  [2213] = {.lex_state = 0},
  [2214] = {.lex_state = 0},
  [2215] = {.lex_state = 0},
  [2216] = {.lex_state = 0},
  [2217] = {.lex_state = 0},
  [2218] = {.lex_state = 0},
  [2219] = {.lex_state = 0},
  [2220] = {.lex_state = 0},
  [2221] = {.lex_state = 0},
  [2222] = {.lex_state = 0},
  [2223] = {.lex_state = 0},
  [2224] = {.lex_state = 0},
  [2225] = {.lex_state = 0},
  [2226] = {.lex_state = 0},
  [2227] = {.lex_state = 0},
  [2228] = {.lex_state = 0},
  [2229] = {.lex_state = 0},
  [2230] = {.lex_state = 0},
  [2231] = {.lex_state = 0},
  [2232] = {.lex_state = 0},
  [2233] = {.lex_state = 0},
  [2234] = {.lex_state = 0},
  [2235] = {.lex_state = 0},
  [2236] = {.lex_state = 0},
  [2237] = {.lex_state = 0},
  [2238] = {.lex_state = 0},
  [2239] = {.lex_state = 0},
  [2240] = {.lex_state = 0},
  [2241] = {.lex_state = 0},
  [2242] = {.lex_state = 0},
  [2243] = {.lex_state = 0},
  [2244] = {.lex_state = 0},
  [2245] = {.lex_state = 0},
  [2246] = {.lex_state = 0},
  [2247] = {.lex_state = 0},
  [2248] = {.lex_state = 0},
  [2249] = {.lex_state = 0},
  [2250] = {.lex_state = 0},
  [2251] = {.lex_state = 0},
  [2252] = {.lex_state = 0},
  [2253] = {.lex_state = 0},
  [2254] = {.lex_state = 0},
  [2255] = {.lex_state = 0},
  [2256] = {.lex_state = 0},
  [2257] = {.lex_state = 0},
  [2258] = {.lex_state = 186},
  [2259] = {.lex_state = 0},
  [2260] = {.lex_state = 0},
  [2261] = {.lex_state = 0},
  [2262] = {.lex_state = 0},
  [2263] = {.lex_state = 0},
  [2264] = {.lex_state = 0},
  [2265] = {.lex_state = 0},
  [2266] = {.lex_state = 0},
  [2267] = {.lex_state = 0},
  [2268] = {.lex_state = 0},
  [2269] = {.lex_state = 0},
  [2270] = {.lex_state = 0},
  [2271] = {.lex_state = 53},
  [2272] = {.lex_state = 0},
  [2273] = {.lex_state = 53},
  [2274] = {.lex_state = 0},
  [2275] = {.lex_state = 0},
  [2276] = {.lex_state = 0},
  [2277] = {.lex_state = 0},
  [2278] = {.lex_state = 0},
  [2279] = {.lex_state = 53},
  [2280] = {.lex_state = 53},
  [2281] = {.lex_state = 0},
  [2282] = {.lex_state = 8},
  [2283] = {.lex_state = 186},
  [2284] = {.lex_state = 0},
  [2285] = {.lex_state = 0},
  [2286] = {.lex_state = 0},
  [2287] = {.lex_state = 0},
  [2288] = {.lex_state = 0},
  [2289] = {.lex_state = 0},
  [2290] = {.lex_state = 26},
  [2291] = {.lex_state = 0},
  [2292] = {.lex_state = 0},
  [2293] = {.lex_state = 0},
  [2294] = {.lex_state = 0},
  [2295] = {.lex_state = 0},
  [2296] = {.lex_state = 0},
  [2297] = {.lex_state = 0},
  [2298] = {.lex_state = 0},
  [2299] = {.lex_state = 0},
  [2300] = {.lex_state = 0},
  [2301] = {.lex_state = 0},
  [2302] = {.lex_state = 8},
  [2303] = {.lex_state = 0},
  [2304] = {.lex_state = 0},
  [2305] = {.lex_state = 0},
  [2306] = {.lex_state = 0},
  [2307] = {.lex_state = 0},
  [2308] = {.lex_state = 0},
  [2309] = {.lex_state = 0},
  [2310] = {.lex_state = 0},
  [2311] = {.lex_state = 0},
  [2312] = {.lex_state = 8},
  [2313] = {.lex_state = 0},
  [2314] = {.lex_state = 0},
  [2315] = {.lex_state = 0},
  [2316] = {.lex_state = 0},
  [2317] = {.lex_state = 0},
  [2318] = {.lex_state = 0},
  [2319] = {.lex_state = 0},
  [2320] = {.lex_state = 0},
  [2321] = {.lex_state = 0},
  [2322] = {.lex_state = 27},
  [2323] = {.lex_state = 0},
  [2324] = {.lex_state = 0},
  [2325] = {.lex_state = 0},
  [2326] = {.lex_state = 0},
  [2327] = {.lex_state = 0},
  [2328] = {.lex_state = 0},
  [2329] = {.lex_state = 0},
  [2330] = {.lex_state = 0},
  [2331] = {.lex_state = 0},
  [2332] = {.lex_state = 0},
  [2333] = {.lex_state = 8},
  [2334] = {.lex_state = 0},
  [2335] = {.lex_state = 0},
  [2336] = {.lex_state = 0},
  [2337] = {.lex_state = 0},
  [2338] = {.lex_state = 26},
  [2339] = {.lex_state = 0},
  [2340] = {.lex_state = 0},
  [2341] = {.lex_state = 8},
  [2342] = {.lex_state = 0},
  [2343] = {.lex_state = 0},
  [2344] = {.lex_state = 8},
  [2345] = {.lex_state = 0},
  [2346] = {.lex_state = 8},
  [2347] = {.lex_state = 0},
  [2348] = {.lex_state = 0},
  [2349] = {.lex_state = 0},
  [2350] = {.lex_state = 186},
  [2351] = {.lex_state = 8},
  [2352] = {.lex_state = 0},
  [2353] = {.lex_state = 0},
  [2354] = {.lex_state = 3},
  [2355] = {.lex_state = 0},
  [2356] = {.lex_state = 0},
  [2357] = {.lex_state = 0},
  [2358] = {.lex_state = 0},
  [2359] = {.lex_state = 0},
  [2360] = {.lex_state = 0},
  [2361] = {.lex_state = 0},
  [2362] = {.lex_state = 8},
  [2363] = {.lex_state = 190},
  [2364] = {.lex_state = 8},
  [2365] = {.lex_state = 8},
  [2366] = {.lex_state = 8},
  [2367] = {.lex_state = 0},
  [2368] = {.lex_state = 8},
  [2369] = {.lex_state = 8},
  [2370] = {.lex_state = 36},
  [2371] = {.lex_state = 0},
  [2372] = {.lex_state = 0},
  [2373] = {.lex_state = 8},
  [2374] = {.lex_state = 0},
  [2375] = {.lex_state = 8},
  [2376] = {.lex_state = 0},
  [2377] = {.lex_state = 8},
  [2378] = {.lex_state = 0},
  [2379] = {.lex_state = 0},
  [2380] = {.lex_state = 8},
  [2381] = {.lex_state = 0},
  [2382] = {.lex_state = 0},
  [2383] = {.lex_state = 8},
  [2384] = {.lex_state = 0},
  [2385] = {.lex_state = 0},
  [2386] = {.lex_state = 0},
  [2387] = {.lex_state = 0},
  [2388] = {.lex_state = 0},
  [2389] = {.lex_state = 0},
  [2390] = {.lex_state = 8},
  [2391] = {.lex_state = 8},
  [2392] = {.lex_state = 8},
  [2393] = {.lex_state = 0},
  [2394] = {.lex_state = 0},
  [2395] = {.lex_state = 0},
  [2396] = {.lex_state = 0},
  [2397] = {.lex_state = 0},
  [2398] = {.lex_state = 38},
  [2399] = {.lex_state = 186},
  [2400] = {.lex_state = 0},
  [2401] = {.lex_state = 0},
  [2402] = {.lex_state = 8},
  [2403] = {.lex_state = 8},
  [2404] = {.lex_state = 0},
  [2405] = {.lex_state = 0},
  [2406] = {.lex_state = 8},
  [2407] = {.lex_state = 0},
  [2408] = {.lex_state = 0},
  [2409] = {.lex_state = 8},
  [2410] = {.lex_state = 8},
  [2411] = {.lex_state = 0},
  [2412] = {.lex_state = 8},
  [2413] = {.lex_state = 8},
  [2414] = {.lex_state = 8},
  [2415] = {.lex_state = 0},
  [2416] = {.lex_state = 8},
  [2417] = {.lex_state = 0},
  [2418] = {.lex_state = 0},
  [2419] = {.lex_state = 8},
  [2420] = {.lex_state = 8},
  [2421] = {.lex_state = 8},
  [2422] = {.lex_state = 8},
  [2423] = {.lex_state = 8},
  [2424] = {.lex_state = 8},
  [2425] = {.lex_state = 8},
  [2426] = {.lex_state = 0},
  [2427] = {.lex_state = 8},
  [2428] = {.lex_state = 8},
  [2429] = {.lex_state = 50},
  [2430] = {.lex_state = 8},
  [2431] = {.lex_state = 8},
  [2432] = {.lex_state = 8},
  [2433] = {.lex_state = 8},
  [2434] = {.lex_state = 8},
  [2435] = {.lex_state = 8},
  [2436] = {.lex_state = 8},
  [2437] = {.lex_state = 8},
  [2438] = {.lex_state = 0},
  [2439] = {.lex_state = 8},
  [2440] = {.lex_state = 8},
  [2441] = {.lex_state = 8},
  [2442] = {.lex_state = 8},
  [2443] = {.lex_state = 0},
  [2444] = {.lex_state = 8},
  [2445] = {.lex_state = 0},
  [2446] = {.lex_state = 8},
  [2447] = {.lex_state = 0},
  [2448] = {.lex_state = 0},
  [2449] = {.lex_state = 0},
  [2450] = {.lex_state = 8},
  [2451] = {.lex_state = 8},
  [2452] = {.lex_state = 8},
  [2453] = {.lex_state = 0},
  [2454] = {.lex_state = 50},
  [2455] = {.lex_state = 0},
  [2456] = {.lex_state = 0},
  [2457] = {.lex_state = 8},
  [2458] = {.lex_state = 0},
  [2459] = {.lex_state = 0},
  [2460] = {.lex_state = 8},
  [2461] = {(TSStateId)(-1),},
  [2462] = {(TSStateId)(-1),},
  [2463] = {(TSStateId)(-1),},
  [2464] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_comment] = STATE(0),
    [sym_groovy_doc] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_POUND_BANG] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_DOT_AMP] = ACTIONS(1),
    [anon_sym_DOT_AT] = ACTIONS(1),
    [anon_sym_QMARK_DOT] = ACTIONS(1),
    [anon_sym_STAR_DOT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_STAR_COLON] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_STAR2] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_package] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_QMARK_EQ] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_GT_GT_GT] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_BANGin] = ACTIONS(1),
    [anon_sym_instanceof] = ACTIONS(1),
    [anon_sym_BANGinstanceof] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ_GT] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_EQ_TILDE] = ACTIONS(1),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_QMARK_COLON] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_ATinterface] = ACTIONS(1),
    [anon_sym_interface] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_extends] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_def] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_null] = ACTIONS(1),
    [anon_sym_this] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [aux_sym__type_identifier_token1] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [aux_sym_number_literal_token1] = ACTIONS(1),
    [aux_sym_number_literal_token2] = ACTIONS(1),
    [aux_sym_number_literal_token3] = ACTIONS(1),
    [aux_sym_number_literal_token4] = ACTIONS(1),
    [aux_sym_number_literal_token5] = ACTIONS(1),
    [anon_sym_pipeline] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [anon_sym_BSLASH_SLASH] = ACTIONS(1),
    [aux_sym__interpolate_string_token5] = ACTIONS(1),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1),
    [anon_sym_DOLLAR_DOLLAR] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_switch] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_catch] = ACTIONS(1),
    [anon_sym_finally] = ACTIONS(1),
    [anon_sym_int] = ACTIONS(1),
    [anon_sym_boolean] = ACTIONS(1),
    [anon_sym_char] = ACTIONS(1),
    [anon_sym_short] = ACTIONS(1),
    [anon_sym_long] = ACTIONS(1),
    [anon_sym_float] = ACTIONS(1),
    [anon_sym_double] = ACTIONS(1),
    [anon_sym_void] = ACTIONS(1),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1),
    [anon_sym_public] = ACTIONS(1),
    [anon_sym_protected] = ACTIONS(1),
    [anon_sym_private] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [anon_sym_final] = ACTIONS(1),
    [anon_sym_synchronized] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(2361),
    [sym_shebang] = STATE(57),
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(1),
    [sym_groovy_doc] = STATE(1),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(652),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_pipeline] = STATE(2353),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(58),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [ts_builtin_sym_end] = ACTIONS(9),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_POUND_BANG] = ACTIONS(13),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_pipeline] = ACTIONS(69),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(2)] = {
    [sym_access_op] = STATE(368),
    [sym_dotted_identifier] = STATE(263),
    [sym_increment_op] = STATE(368),
    [sym_binary_op] = STATE(368),
    [sym_boolean_literal] = STATE(264),
    [sym_closure] = STATE(368),
    [sym_comment] = STATE(2),
    [sym_groovy_doc] = STATE(2),
    [sym_parenthesized_expression] = STATE(1429),
    [sym__expression] = STATE(1545),
    [sym__primary_expression] = STATE(1437),
    [sym__callable_expression] = STATE(1430),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(261),
    [sym__type_identifier] = STATE(1408),
    [sym_index] = STATE(263),
    [sym__juxt_argument_list] = STATE(742),
    [sym_list] = STATE(264),
    [sym_map_item] = STATE(744),
    [sym_map] = STATE(264),
    [sym_number_literal] = STATE(250),
    [sym_string] = STATE(250),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_ternary_op] = STATE(368),
    [sym_unary_op] = STATE(368),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(99),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(104),
    [anon_sym_STAR_COLON] = ACTIONS(107),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(110),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(112),
    [anon_sym_STAR_EQ] = ACTIONS(112),
    [anon_sym_SLASH_EQ] = ACTIONS(112),
    [anon_sym_PERCENT_EQ] = ACTIONS(112),
    [anon_sym_PLUS_EQ] = ACTIONS(112),
    [anon_sym_DASH_EQ] = ACTIONS(112),
    [anon_sym_LT_LT_EQ] = ACTIONS(112),
    [anon_sym_GT_GT_EQ] = ACTIONS(112),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(112),
    [anon_sym_AMP_EQ] = ACTIONS(112),
    [anon_sym_CARET_EQ] = ACTIONS(112),
    [anon_sym_PIPE_EQ] = ACTIONS(112),
    [anon_sym_QMARK_EQ] = ACTIONS(112),
    [anon_sym_PLUS_PLUS] = ACTIONS(114),
    [anon_sym_DASH_DASH] = ACTIONS(114),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(117),
    [anon_sym_PLUS] = ACTIONS(120),
    [anon_sym_DASH] = ACTIONS(120),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(123),
    [anon_sym_false] = ACTIONS(123),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(126),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(132),
    [anon_sym_this] = ACTIONS(135),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(138),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(141),
    [aux_sym_number_literal_token1] = ACTIONS(144),
    [aux_sym_number_literal_token2] = ACTIONS(147),
    [aux_sym_number_literal_token3] = ACTIONS(147),
    [aux_sym_number_literal_token4] = ACTIONS(144),
    [aux_sym_number_literal_token5] = ACTIONS(147),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(150),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(156),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(159),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(162),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(168),
    [anon_sym_new] = ACTIONS(168),
  },
  [STATE(3)] = {
    [sym_access_op] = STATE(311),
    [sym_dotted_identifier] = STATE(272),
    [sym_increment_op] = STATE(311),
    [sym_binary_op] = STATE(311),
    [sym_boolean_literal] = STATE(276),
    [sym_closure] = STATE(311),
    [sym_comment] = STATE(3),
    [sym_groovy_doc] = STATE(3),
    [sym_parenthesized_expression] = STATE(1431),
    [sym__expression] = STATE(1543),
    [sym__primary_expression] = STATE(1440),
    [sym__callable_expression] = STATE(1427),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(268),
    [sym__type_identifier] = STATE(1422),
    [sym_index] = STATE(272),
    [sym__juxt_argument_list] = STATE(808),
    [sym_list] = STATE(276),
    [sym_map_item] = STATE(678),
    [sym_map] = STATE(276),
    [sym_number_literal] = STATE(252),
    [sym_string] = STATE(252),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_ternary_op] = STATE(311),
    [sym_unary_op] = STATE(311),
    [sym_identifier] = ACTIONS(171),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(174),
    [anon_sym_STAR_COLON] = ACTIONS(177),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(180),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_QMARK_EQ] = ACTIONS(182),
    [anon_sym_PLUS_PLUS] = ACTIONS(184),
    [anon_sym_DASH_DASH] = ACTIONS(184),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(187),
    [anon_sym_PLUS] = ACTIONS(190),
    [anon_sym_DASH] = ACTIONS(190),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(193),
    [anon_sym_false] = ACTIONS(193),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(196),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(199),
    [anon_sym_this] = ACTIONS(202),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(138),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(205),
    [aux_sym_number_literal_token1] = ACTIONS(208),
    [aux_sym_number_literal_token2] = ACTIONS(211),
    [aux_sym_number_literal_token3] = ACTIONS(211),
    [aux_sym_number_literal_token4] = ACTIONS(208),
    [aux_sym_number_literal_token5] = ACTIONS(211),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(214),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(217),
    [anon_sym_DQUOTE] = ACTIONS(220),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(223),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(226),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(229),
    [anon_sym_BANG] = ACTIONS(232),
    [anon_sym_new] = ACTIONS(232),
  },
  [STATE(4)] = {
    [sym_access_op] = STATE(548),
    [sym_dotted_identifier] = STATE(400),
    [sym_increment_op] = STATE(548),
    [sym_binary_op] = STATE(548),
    [sym_boolean_literal] = STATE(401),
    [sym_closure] = STATE(548),
    [sym_comment] = STATE(4),
    [sym_groovy_doc] = STATE(4),
    [sym_parenthesized_expression] = STATE(1428),
    [sym__expression] = STATE(1544),
    [sym__primary_expression] = STATE(1439),
    [sym__callable_expression] = STATE(1434),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(399),
    [sym__type_identifier] = STATE(1415),
    [sym_index] = STATE(400),
    [sym__juxt_argument_list] = STATE(971),
    [sym_list] = STATE(401),
    [sym_map_item] = STATE(809),
    [sym_map] = STATE(401),
    [sym_number_literal] = STATE(384),
    [sym_string] = STATE(384),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_ternary_op] = STATE(548),
    [sym_unary_op] = STATE(548),
    [sym_identifier] = ACTIONS(235),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(238),
    [anon_sym_STAR_COLON] = ACTIONS(241),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(244),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(246),
    [anon_sym_SLASH_EQ] = ACTIONS(246),
    [anon_sym_PERCENT_EQ] = ACTIONS(246),
    [anon_sym_PLUS_EQ] = ACTIONS(246),
    [anon_sym_DASH_EQ] = ACTIONS(246),
    [anon_sym_LT_LT_EQ] = ACTIONS(246),
    [anon_sym_GT_GT_EQ] = ACTIONS(246),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(246),
    [anon_sym_AMP_EQ] = ACTIONS(246),
    [anon_sym_CARET_EQ] = ACTIONS(246),
    [anon_sym_PIPE_EQ] = ACTIONS(246),
    [anon_sym_QMARK_EQ] = ACTIONS(246),
    [anon_sym_PLUS_PLUS] = ACTIONS(248),
    [anon_sym_DASH_DASH] = ACTIONS(248),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(251),
    [anon_sym_PLUS] = ACTIONS(254),
    [anon_sym_DASH] = ACTIONS(254),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(257),
    [anon_sym_false] = ACTIONS(257),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(260),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(263),
    [anon_sym_this] = ACTIONS(266),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(138),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(269),
    [aux_sym_number_literal_token1] = ACTIONS(272),
    [aux_sym_number_literal_token2] = ACTIONS(275),
    [aux_sym_number_literal_token3] = ACTIONS(275),
    [aux_sym_number_literal_token4] = ACTIONS(272),
    [aux_sym_number_literal_token5] = ACTIONS(275),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(278),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(281),
    [anon_sym_DQUOTE] = ACTIONS(284),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(287),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(290),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(293),
    [anon_sym_BANG] = ACTIONS(296),
    [anon_sym_new] = ACTIONS(296),
  },
  [STATE(5)] = {
    [sym_access_op] = STATE(558),
    [sym_dotted_identifier] = STATE(471),
    [sym_increment_op] = STATE(558),
    [sym_binary_op] = STATE(558),
    [sym_boolean_literal] = STATE(472),
    [sym_closure] = STATE(558),
    [sym_comment] = STATE(5),
    [sym_groovy_doc] = STATE(5),
    [sym_parenthesized_expression] = STATE(1433),
    [sym__expression] = STATE(1547),
    [sym__primary_expression] = STATE(1436),
    [sym__callable_expression] = STATE(1432),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(469),
    [sym__type_identifier] = STATE(1401),
    [sym_index] = STATE(471),
    [sym__juxt_argument_list] = STATE(895),
    [sym_list] = STATE(472),
    [sym_map_item] = STATE(898),
    [sym_map] = STATE(472),
    [sym_number_literal] = STATE(316),
    [sym_string] = STATE(316),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_ternary_op] = STATE(558),
    [sym_unary_op] = STATE(558),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(299),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(302),
    [anon_sym_STAR_COLON] = ACTIONS(305),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_QMARK_EQ] = ACTIONS(310),
    [anon_sym_PLUS_PLUS] = ACTIONS(312),
    [anon_sym_DASH_DASH] = ACTIONS(312),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(315),
    [anon_sym_PLUS] = ACTIONS(318),
    [anon_sym_DASH] = ACTIONS(318),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(321),
    [anon_sym_false] = ACTIONS(321),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(324),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(327),
    [anon_sym_this] = ACTIONS(330),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(138),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(333),
    [aux_sym_number_literal_token1] = ACTIONS(336),
    [aux_sym_number_literal_token2] = ACTIONS(339),
    [aux_sym_number_literal_token3] = ACTIONS(339),
    [aux_sym_number_literal_token4] = ACTIONS(336),
    [aux_sym_number_literal_token5] = ACTIONS(339),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(342),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(345),
    [anon_sym_DQUOTE] = ACTIONS(348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(351),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(354),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(357),
    [anon_sym_BANG] = ACTIONS(360),
    [anon_sym_new] = ACTIONS(360),
  },
  [STATE(6)] = {
    [sym_comment] = STATE(6),
    [sym_groovy_doc] = STATE(6),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(7)] = {
    [sym_comment] = STATE(7),
    [sym_groovy_doc] = STATE(7),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_COLON] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_RPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(8)] = {
    [sym_comment] = STATE(8),
    [sym_groovy_doc] = STATE(8),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(9)] = {
    [sym_comment] = STATE(9),
    [sym_groovy_doc] = STATE(9),
    [aux_sym_dotted_identifier_repeat1] = STATE(10),
    [ts_builtin_sym_end] = ACTIONS(373),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_EQ] = ACTIONS(375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(373),
    [anon_sym_STAR_EQ] = ACTIONS(373),
    [anon_sym_SLASH_EQ] = ACTIONS(373),
    [anon_sym_PERCENT_EQ] = ACTIONS(373),
    [anon_sym_PLUS_EQ] = ACTIONS(373),
    [anon_sym_DASH_EQ] = ACTIONS(373),
    [anon_sym_LT_LT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_AMP_EQ] = ACTIONS(373),
    [anon_sym_CARET_EQ] = ACTIONS(373),
    [anon_sym_PIPE_EQ] = ACTIONS(373),
    [anon_sym_QMARK_EQ] = ACTIONS(373),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(375),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(375),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(375),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(375),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_pipeline] = ACTIONS(375),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_catch] = ACTIONS(375),
    [anon_sym_finally] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(10)] = {
    [sym_comment] = STATE(10),
    [sym_groovy_doc] = STATE(10),
    [aux_sym_dotted_identifier_repeat1] = STATE(10),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(381),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(11)] = {
    [sym_comment] = STATE(11),
    [sym_groovy_doc] = STATE(11),
    [aux_sym_dotted_identifier_repeat1] = STATE(11),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(384),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(12)] = {
    [sym_comment] = STATE(12),
    [sym_groovy_doc] = STATE(12),
    [aux_sym_dotted_identifier_repeat1] = STATE(11),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_EQ] = ACTIONS(375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(373),
    [anon_sym_STAR_EQ] = ACTIONS(373),
    [anon_sym_SLASH_EQ] = ACTIONS(373),
    [anon_sym_PERCENT_EQ] = ACTIONS(373),
    [anon_sym_PLUS_EQ] = ACTIONS(373),
    [anon_sym_DASH_EQ] = ACTIONS(373),
    [anon_sym_LT_LT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_AMP_EQ] = ACTIONS(373),
    [anon_sym_CARET_EQ] = ACTIONS(373),
    [anon_sym_PIPE_EQ] = ACTIONS(373),
    [anon_sym_QMARK_EQ] = ACTIONS(373),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(375),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(375),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(375),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(375),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_case] = ACTIONS(375),
    [anon_sym_default] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_catch] = ACTIONS(375),
    [anon_sym_finally] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(13)] = {
    [sym_comment] = STATE(13),
    [sym_groovy_doc] = STATE(13),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(110),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(112),
    [anon_sym_STAR_EQ] = ACTIONS(112),
    [anon_sym_SLASH_EQ] = ACTIONS(112),
    [anon_sym_PERCENT_EQ] = ACTIONS(112),
    [anon_sym_PLUS_EQ] = ACTIONS(112),
    [anon_sym_DASH_EQ] = ACTIONS(112),
    [anon_sym_LT_LT_EQ] = ACTIONS(112),
    [anon_sym_GT_GT_EQ] = ACTIONS(112),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(112),
    [anon_sym_AMP_EQ] = ACTIONS(112),
    [anon_sym_CARET_EQ] = ACTIONS(112),
    [anon_sym_PIPE_EQ] = ACTIONS(112),
    [anon_sym_QMARK_EQ] = ACTIONS(112),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(14)] = {
    [sym_comment] = STATE(14),
    [sym_groovy_doc] = STATE(14),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_EQ] = ACTIONS(387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(389),
    [anon_sym_STAR_EQ] = ACTIONS(389),
    [anon_sym_SLASH_EQ] = ACTIONS(389),
    [anon_sym_PERCENT_EQ] = ACTIONS(389),
    [anon_sym_PLUS_EQ] = ACTIONS(389),
    [anon_sym_DASH_EQ] = ACTIONS(389),
    [anon_sym_LT_LT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_AMP_EQ] = ACTIONS(389),
    [anon_sym_CARET_EQ] = ACTIONS(389),
    [anon_sym_PIPE_EQ] = ACTIONS(389),
    [anon_sym_QMARK_EQ] = ACTIONS(389),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(387),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(387),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(15)] = {
    [sym_comment] = STATE(15),
    [sym_groovy_doc] = STATE(15),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(16)] = {
    [sym_comment] = STATE(16),
    [sym_groovy_doc] = STATE(16),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(17)] = {
    [sym_comment] = STATE(17),
    [sym_groovy_doc] = STATE(17),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(18)] = {
    [sym_comment] = STATE(18),
    [sym_groovy_doc] = STATE(18),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(19)] = {
    [sym_comment] = STATE(19),
    [sym_groovy_doc] = STATE(19),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(180),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(182),
    [anon_sym_STAR_EQ] = ACTIONS(182),
    [anon_sym_SLASH_EQ] = ACTIONS(182),
    [anon_sym_PERCENT_EQ] = ACTIONS(182),
    [anon_sym_PLUS_EQ] = ACTIONS(182),
    [anon_sym_DASH_EQ] = ACTIONS(182),
    [anon_sym_LT_LT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(182),
    [anon_sym_AMP_EQ] = ACTIONS(182),
    [anon_sym_CARET_EQ] = ACTIONS(182),
    [anon_sym_PIPE_EQ] = ACTIONS(182),
    [anon_sym_QMARK_EQ] = ACTIONS(182),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(20)] = {
    [sym_comment] = STATE(20),
    [sym_groovy_doc] = STATE(20),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(21)] = {
    [sym_comment] = STATE(21),
    [sym_groovy_doc] = STATE(21),
    [ts_builtin_sym_end] = ACTIONS(391),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_EQ] = ACTIONS(393),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(391),
    [anon_sym_STAR_EQ] = ACTIONS(391),
    [anon_sym_SLASH_EQ] = ACTIONS(391),
    [anon_sym_PERCENT_EQ] = ACTIONS(391),
    [anon_sym_PLUS_EQ] = ACTIONS(391),
    [anon_sym_DASH_EQ] = ACTIONS(391),
    [anon_sym_LT_LT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_AMP_EQ] = ACTIONS(391),
    [anon_sym_CARET_EQ] = ACTIONS(391),
    [anon_sym_PIPE_EQ] = ACTIONS(391),
    [anon_sym_QMARK_EQ] = ACTIONS(391),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(393),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(393),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(393),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(393),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(393),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_pipeline] = ACTIONS(393),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_catch] = ACTIONS(393),
    [anon_sym_finally] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(22)] = {
    [sym_comment] = STATE(22),
    [sym_groovy_doc] = STATE(22),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_EQ] = ACTIONS(393),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(391),
    [anon_sym_STAR_EQ] = ACTIONS(391),
    [anon_sym_SLASH_EQ] = ACTIONS(391),
    [anon_sym_PERCENT_EQ] = ACTIONS(391),
    [anon_sym_PLUS_EQ] = ACTIONS(391),
    [anon_sym_DASH_EQ] = ACTIONS(391),
    [anon_sym_LT_LT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_AMP_EQ] = ACTIONS(391),
    [anon_sym_CARET_EQ] = ACTIONS(391),
    [anon_sym_PIPE_EQ] = ACTIONS(391),
    [anon_sym_QMARK_EQ] = ACTIONS(391),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(393),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(393),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(393),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(393),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(393),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_case] = ACTIONS(393),
    [anon_sym_default] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_catch] = ACTIONS(393),
    [anon_sym_finally] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(23)] = {
    [sym_comment] = STATE(23),
    [sym_groovy_doc] = STATE(23),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_EQ] = ACTIONS(387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(389),
    [anon_sym_STAR_EQ] = ACTIONS(389),
    [anon_sym_SLASH_EQ] = ACTIONS(389),
    [anon_sym_PERCENT_EQ] = ACTIONS(389),
    [anon_sym_PLUS_EQ] = ACTIONS(389),
    [anon_sym_DASH_EQ] = ACTIONS(389),
    [anon_sym_LT_LT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_AMP_EQ] = ACTIONS(389),
    [anon_sym_CARET_EQ] = ACTIONS(389),
    [anon_sym_PIPE_EQ] = ACTIONS(389),
    [anon_sym_QMARK_EQ] = ACTIONS(389),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(387),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(387),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(24)] = {
    [sym_comment] = STATE(24),
    [sym_groovy_doc] = STATE(24),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(25)] = {
    [sym_comment] = STATE(25),
    [sym_groovy_doc] = STATE(25),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(26)] = {
    [sym_comment] = STATE(26),
    [sym_groovy_doc] = STATE(26),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(27)] = {
    [sym_comment] = STATE(27),
    [sym_groovy_doc] = STATE(27),
    [aux_sym_dotted_identifier_repeat1] = STATE(31),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_EQ] = ACTIONS(375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(373),
    [anon_sym_STAR_EQ] = ACTIONS(373),
    [anon_sym_SLASH_EQ] = ACTIONS(373),
    [anon_sym_PERCENT_EQ] = ACTIONS(373),
    [anon_sym_PLUS_EQ] = ACTIONS(373),
    [anon_sym_DASH_EQ] = ACTIONS(373),
    [anon_sym_LT_LT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_AMP_EQ] = ACTIONS(373),
    [anon_sym_CARET_EQ] = ACTIONS(373),
    [anon_sym_PIPE_EQ] = ACTIONS(373),
    [anon_sym_QMARK_EQ] = ACTIONS(373),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(375),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(375),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(375),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(375),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_case] = ACTIONS(375),
    [anon_sym_default] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(28)] = {
    [sym_comment] = STATE(28),
    [sym_groovy_doc] = STATE(28),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(29)] = {
    [sym_comment] = STATE(29),
    [sym_groovy_doc] = STATE(29),
    [aux_sym_dotted_identifier_repeat1] = STATE(29),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(395),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(30)] = {
    [sym_comment] = STATE(30),
    [sym_groovy_doc] = STATE(30),
    [ts_builtin_sym_end] = ACTIONS(391),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_EQ] = ACTIONS(393),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(391),
    [anon_sym_STAR_EQ] = ACTIONS(391),
    [anon_sym_SLASH_EQ] = ACTIONS(391),
    [anon_sym_PERCENT_EQ] = ACTIONS(391),
    [anon_sym_PLUS_EQ] = ACTIONS(391),
    [anon_sym_DASH_EQ] = ACTIONS(391),
    [anon_sym_LT_LT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_AMP_EQ] = ACTIONS(391),
    [anon_sym_CARET_EQ] = ACTIONS(391),
    [anon_sym_PIPE_EQ] = ACTIONS(391),
    [anon_sym_QMARK_EQ] = ACTIONS(391),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(393),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(393),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(393),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(393),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(393),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_pipeline] = ACTIONS(393),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(31)] = {
    [sym_comment] = STATE(31),
    [sym_groovy_doc] = STATE(31),
    [aux_sym_dotted_identifier_repeat1] = STATE(31),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(398),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(32)] = {
    [sym_comment] = STATE(32),
    [sym_groovy_doc] = STATE(32),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(33)] = {
    [sym_comment] = STATE(33),
    [sym_groovy_doc] = STATE(33),
    [aux_sym_dotted_identifier_repeat1] = STATE(29),
    [ts_builtin_sym_end] = ACTIONS(373),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_EQ] = ACTIONS(375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(373),
    [anon_sym_STAR_EQ] = ACTIONS(373),
    [anon_sym_SLASH_EQ] = ACTIONS(373),
    [anon_sym_PERCENT_EQ] = ACTIONS(373),
    [anon_sym_PLUS_EQ] = ACTIONS(373),
    [anon_sym_DASH_EQ] = ACTIONS(373),
    [anon_sym_LT_LT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_AMP_EQ] = ACTIONS(373),
    [anon_sym_CARET_EQ] = ACTIONS(373),
    [anon_sym_PIPE_EQ] = ACTIONS(373),
    [anon_sym_QMARK_EQ] = ACTIONS(373),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(375),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(375),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(375),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(375),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_pipeline] = ACTIONS(375),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(34)] = {
    [sym_comment] = STATE(34),
    [sym_groovy_doc] = STATE(34),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(35)] = {
    [sym_comment] = STATE(35),
    [sym_groovy_doc] = STATE(35),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(244),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(246),
    [anon_sym_SLASH_EQ] = ACTIONS(246),
    [anon_sym_PERCENT_EQ] = ACTIONS(246),
    [anon_sym_PLUS_EQ] = ACTIONS(246),
    [anon_sym_DASH_EQ] = ACTIONS(246),
    [anon_sym_LT_LT_EQ] = ACTIONS(246),
    [anon_sym_GT_GT_EQ] = ACTIONS(246),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(246),
    [anon_sym_AMP_EQ] = ACTIONS(246),
    [anon_sym_CARET_EQ] = ACTIONS(246),
    [anon_sym_PIPE_EQ] = ACTIONS(246),
    [anon_sym_QMARK_EQ] = ACTIONS(246),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(36)] = {
    [sym_comment] = STATE(36),
    [sym_groovy_doc] = STATE(36),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_EQ] = ACTIONS(393),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(391),
    [anon_sym_STAR_EQ] = ACTIONS(391),
    [anon_sym_SLASH_EQ] = ACTIONS(391),
    [anon_sym_PERCENT_EQ] = ACTIONS(391),
    [anon_sym_PLUS_EQ] = ACTIONS(391),
    [anon_sym_DASH_EQ] = ACTIONS(391),
    [anon_sym_LT_LT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_AMP_EQ] = ACTIONS(391),
    [anon_sym_CARET_EQ] = ACTIONS(391),
    [anon_sym_PIPE_EQ] = ACTIONS(391),
    [anon_sym_QMARK_EQ] = ACTIONS(391),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(393),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(393),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(393),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(393),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(393),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_case] = ACTIONS(393),
    [anon_sym_default] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(37)] = {
    [sym_comment] = STATE(37),
    [sym_groovy_doc] = STATE(37),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_EQ] = ACTIONS(387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(389),
    [anon_sym_STAR_EQ] = ACTIONS(389),
    [anon_sym_SLASH_EQ] = ACTIONS(389),
    [anon_sym_PERCENT_EQ] = ACTIONS(389),
    [anon_sym_PLUS_EQ] = ACTIONS(389),
    [anon_sym_DASH_EQ] = ACTIONS(389),
    [anon_sym_LT_LT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_AMP_EQ] = ACTIONS(389),
    [anon_sym_CARET_EQ] = ACTIONS(389),
    [anon_sym_PIPE_EQ] = ACTIONS(389),
    [anon_sym_QMARK_EQ] = ACTIONS(389),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(387),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(387),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(38)] = {
    [sym_comment] = STATE(38),
    [sym_groovy_doc] = STATE(38),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(39)] = {
    [sym_comment] = STATE(39),
    [sym_groovy_doc] = STATE(39),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(401),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_DASH_GT] = ACTIONS(401),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(40)] = {
    [sym_comment] = STATE(40),
    [sym_groovy_doc] = STATE(40),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(41)] = {
    [sym_comment] = STATE(41),
    [sym_groovy_doc] = STATE(41),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(308),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_QMARK_EQ] = ACTIONS(310),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(42)] = {
    [sym_comment] = STATE(42),
    [sym_groovy_doc] = STATE(42),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_EQ] = ACTIONS(387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(389),
    [anon_sym_STAR_EQ] = ACTIONS(389),
    [anon_sym_SLASH_EQ] = ACTIONS(389),
    [anon_sym_PERCENT_EQ] = ACTIONS(389),
    [anon_sym_PLUS_EQ] = ACTIONS(389),
    [anon_sym_DASH_EQ] = ACTIONS(389),
    [anon_sym_LT_LT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_AMP_EQ] = ACTIONS(389),
    [anon_sym_CARET_EQ] = ACTIONS(389),
    [anon_sym_PIPE_EQ] = ACTIONS(389),
    [anon_sym_QMARK_EQ] = ACTIONS(389),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(387),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(387),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(43)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(43),
    [sym_groovy_doc] = STATE(43),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2343),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(65),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(405),
    [anon_sym_RBRACE] = ACTIONS(407),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(44)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(44),
    [sym_groovy_doc] = STATE(44),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2348),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(71),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(411),
    [anon_sym_RBRACE] = ACTIONS(413),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(45)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(45),
    [sym_groovy_doc] = STATE(45),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2374),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(76),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(417),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(46)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(46),
    [sym_groovy_doc] = STATE(46),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2455),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(81),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(419),
    [anon_sym_RBRACE] = ACTIONS(421),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(47)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(47),
    [sym_groovy_doc] = STATE(47),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2359),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(87),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(423),
    [anon_sym_RBRACE] = ACTIONS(425),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(48)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(48),
    [sym_groovy_doc] = STATE(48),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2376),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(92),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(427),
    [anon_sym_RBRACE] = ACTIONS(429),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(49)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(49),
    [sym_groovy_doc] = STATE(49),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2381),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(97),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(431),
    [anon_sym_RBRACE] = ACTIONS(433),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(50)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(50),
    [sym_groovy_doc] = STATE(50),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2385),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(103),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(435),
    [anon_sym_RBRACE] = ACTIONS(437),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(51)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(51),
    [sym_groovy_doc] = STATE(51),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2389),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(109),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(439),
    [anon_sym_RBRACE] = ACTIONS(441),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(52)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(52),
    [sym_groovy_doc] = STATE(52),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2393),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(114),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(443),
    [anon_sym_RBRACE] = ACTIONS(445),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(53)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(53),
    [sym_groovy_doc] = STATE(53),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2397),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(64),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(447),
    [anon_sym_RBRACE] = ACTIONS(449),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(54)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(54),
    [sym_groovy_doc] = STATE(54),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(655),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym__param_list] = STATE(2447),
    [sym_parameter] = STATE(2214),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2072),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(101),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [aux_sym__param_list_repeat1] = STATE(1737),
    [sym_identifier] = ACTIONS(403),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(451),
    [anon_sym_RBRACE] = ACTIONS(453),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(409),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(55)] = {
    [sym__statement] = STATE(1004),
    [sym_label] = STATE(234),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1001),
    [sym_groovy_package] = STATE(1001),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1001),
    [sym_assignment] = STATE(1001),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1001),
    [sym_closure] = STATE(651),
    [sym_comment] = STATE(55),
    [sym_groovy_doc] = STATE(55),
    [sym_declaration] = STATE(1001),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(650),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1001),
    [sym_for_loop] = STATE(1001),
    [sym_for_in_loop] = STATE(1001),
    [sym_function_call] = STATE(608),
    [sym_function_declaration] = STATE(1001),
    [sym_function_definition] = STATE(1001),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1001),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1001),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1001),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1001),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1001),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1001),
    [aux_sym_source_file_repeat1] = STATE(55),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(455),
    [anon_sym_break] = ACTIONS(458),
    [anon_sym_continue] = ACTIONS(458),
    [anon_sym_STAR] = ACTIONS(461),
    [anon_sym_STAR_COLON] = ACTIONS(464),
    [anon_sym_import] = ACTIONS(467),
    [anon_sym_package] = ACTIONS(470),
    [anon_sym_AT] = ACTIONS(473),
    [anon_sym_assert] = ACTIONS(476),
    [anon_sym_PLUS_PLUS] = ACTIONS(479),
    [anon_sym_DASH_DASH] = ACTIONS(479),
    [anon_sym_SLASH] = ACTIONS(482),
    [anon_sym_PLUS] = ACTIONS(485),
    [anon_sym_DASH] = ACTIONS(485),
    [anon_sym_true] = ACTIONS(488),
    [anon_sym_false] = ACTIONS(488),
    [anon_sym_ATinterface] = ACTIONS(491),
    [anon_sym_interface] = ACTIONS(494),
    [anon_sym_class] = ACTIONS(494),
    [anon_sym_LBRACE] = ACTIONS(497),
    [anon_sym_RBRACE] = ACTIONS(500),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(502),
    [anon_sym_def] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(508),
    [anon_sym_null] = ACTIONS(511),
    [anon_sym_this] = ACTIONS(514),
    [anon_sym_do] = ACTIONS(517),
    [anon_sym_while] = ACTIONS(520),
    [anon_sym_for] = ACTIONS(523),
    [aux_sym__type_identifier_token1] = ACTIONS(526),
    [anon_sym_if] = ACTIONS(529),
    [anon_sym_LBRACK] = ACTIONS(532),
    [aux_sym_number_literal_token1] = ACTIONS(535),
    [aux_sym_number_literal_token2] = ACTIONS(538),
    [aux_sym_number_literal_token3] = ACTIONS(538),
    [aux_sym_number_literal_token4] = ACTIONS(535),
    [aux_sym_number_literal_token5] = ACTIONS(538),
    [anon_sym_return] = ACTIONS(541),
    [anon_sym_SQUOTE] = ACTIONS(544),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(547),
    [anon_sym_DQUOTE] = ACTIONS(550),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(553),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(556),
    [anon_sym_switch] = ACTIONS(559),
    [anon_sym_case] = ACTIONS(562),
    [anon_sym_default] = ACTIONS(562),
    [anon_sym_try] = ACTIONS(564),
    [anon_sym_int] = ACTIONS(567),
    [anon_sym_boolean] = ACTIONS(567),
    [anon_sym_char] = ACTIONS(567),
    [anon_sym_short] = ACTIONS(567),
    [anon_sym_long] = ACTIONS(567),
    [anon_sym_float] = ACTIONS(567),
    [anon_sym_double] = ACTIONS(567),
    [anon_sym_void] = ACTIONS(567),
    [anon_sym_public] = ACTIONS(570),
    [anon_sym_protected] = ACTIONS(570),
    [anon_sym_private] = ACTIONS(570),
    [anon_sym_static] = ACTIONS(573),
    [anon_sym_final] = ACTIONS(573),
    [anon_sym_synchronized] = ACTIONS(573),
    [anon_sym_TILDE] = ACTIONS(576),
    [anon_sym_BANG] = ACTIONS(576),
    [anon_sym_new] = ACTIONS(579),
  },
  [STATE(56)] = {
    [sym__statement] = STATE(1004),
    [sym_label] = STATE(234),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1001),
    [sym_groovy_package] = STATE(1001),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1001),
    [sym_assignment] = STATE(1001),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1001),
    [sym_closure] = STATE(651),
    [sym_comment] = STATE(56),
    [sym_groovy_doc] = STATE(56),
    [sym_declaration] = STATE(1001),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(650),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1001),
    [sym_for_loop] = STATE(1001),
    [sym_for_in_loop] = STATE(1001),
    [sym_function_call] = STATE(608),
    [sym_function_declaration] = STATE(1001),
    [sym_function_definition] = STATE(1001),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1001),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1001),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1001),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1001),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1001),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1001),
    [aux_sym_source_file_repeat1] = STATE(62),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(584),
    [anon_sym_continue] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(610),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_case] = ACTIONS(652),
    [anon_sym_default] = ACTIONS(652),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(57)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(57),
    [sym_groovy_doc] = STATE(57),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(652),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_pipeline] = STATE(2388),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(61),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [ts_builtin_sym_end] = ACTIONS(660),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_pipeline] = ACTIONS(69),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(58)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(58),
    [sym_groovy_doc] = STATE(58),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(652),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_pipeline] = STATE(2388),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [ts_builtin_sym_end] = ACTIONS(660),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_pipeline] = ACTIONS(69),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(59)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(59),
    [sym_groovy_doc] = STATE(59),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(652),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [ts_builtin_sym_end] = ACTIONS(500),
    [sym_identifier] = ACTIONS(662),
    [anon_sym_break] = ACTIONS(665),
    [anon_sym_continue] = ACTIONS(665),
    [anon_sym_STAR] = ACTIONS(668),
    [anon_sym_STAR_COLON] = ACTIONS(671),
    [anon_sym_import] = ACTIONS(674),
    [anon_sym_package] = ACTIONS(677),
    [anon_sym_AT] = ACTIONS(473),
    [anon_sym_assert] = ACTIONS(680),
    [anon_sym_PLUS_PLUS] = ACTIONS(683),
    [anon_sym_DASH_DASH] = ACTIONS(683),
    [anon_sym_SLASH] = ACTIONS(686),
    [anon_sym_PLUS] = ACTIONS(689),
    [anon_sym_DASH] = ACTIONS(689),
    [anon_sym_true] = ACTIONS(692),
    [anon_sym_false] = ACTIONS(692),
    [anon_sym_ATinterface] = ACTIONS(695),
    [anon_sym_interface] = ACTIONS(698),
    [anon_sym_class] = ACTIONS(698),
    [anon_sym_LBRACE] = ACTIONS(701),
    [anon_sym_RBRACE] = ACTIONS(500),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(704),
    [anon_sym_def] = ACTIONS(707),
    [anon_sym_LPAREN] = ACTIONS(710),
    [anon_sym_null] = ACTIONS(713),
    [anon_sym_this] = ACTIONS(716),
    [anon_sym_do] = ACTIONS(719),
    [anon_sym_while] = ACTIONS(722),
    [anon_sym_for] = ACTIONS(725),
    [aux_sym__type_identifier_token1] = ACTIONS(728),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_LBRACK] = ACTIONS(734),
    [aux_sym_number_literal_token1] = ACTIONS(737),
    [aux_sym_number_literal_token2] = ACTIONS(740),
    [aux_sym_number_literal_token3] = ACTIONS(740),
    [aux_sym_number_literal_token4] = ACTIONS(737),
    [aux_sym_number_literal_token5] = ACTIONS(740),
    [anon_sym_pipeline] = ACTIONS(562),
    [anon_sym_return] = ACTIONS(743),
    [anon_sym_SQUOTE] = ACTIONS(746),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(749),
    [anon_sym_DQUOTE] = ACTIONS(752),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(755),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(758),
    [anon_sym_switch] = ACTIONS(761),
    [anon_sym_try] = ACTIONS(764),
    [anon_sym_int] = ACTIONS(567),
    [anon_sym_boolean] = ACTIONS(567),
    [anon_sym_char] = ACTIONS(567),
    [anon_sym_short] = ACTIONS(567),
    [anon_sym_long] = ACTIONS(567),
    [anon_sym_float] = ACTIONS(567),
    [anon_sym_double] = ACTIONS(567),
    [anon_sym_void] = ACTIONS(567),
    [anon_sym_public] = ACTIONS(570),
    [anon_sym_protected] = ACTIONS(570),
    [anon_sym_private] = ACTIONS(570),
    [anon_sym_static] = ACTIONS(573),
    [anon_sym_final] = ACTIONS(573),
    [anon_sym_synchronized] = ACTIONS(573),
    [anon_sym_TILDE] = ACTIONS(767),
    [anon_sym_BANG] = ACTIONS(767),
    [anon_sym_new] = ACTIONS(770),
  },
  [STATE(60)] = {
    [sym__statement] = STATE(1004),
    [sym_label] = STATE(234),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1001),
    [sym_groovy_package] = STATE(1001),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1001),
    [sym_assignment] = STATE(1001),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1001),
    [sym_closure] = STATE(651),
    [sym_comment] = STATE(60),
    [sym_groovy_doc] = STATE(60),
    [sym_declaration] = STATE(1001),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(650),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1001),
    [sym_for_loop] = STATE(1001),
    [sym_for_in_loop] = STATE(1001),
    [sym_function_call] = STATE(608),
    [sym_function_declaration] = STATE(1001),
    [sym_function_definition] = STATE(1001),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1001),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1001),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1001),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1001),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1001),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1001),
    [aux_sym_source_file_repeat1] = STATE(55),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(584),
    [anon_sym_continue] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(773),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_case] = ACTIONS(775),
    [anon_sym_default] = ACTIONS(775),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(61)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(61),
    [sym_groovy_doc] = STATE(61),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(652),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_pipeline] = STATE(2387),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [ts_builtin_sym_end] = ACTIONS(777),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_pipeline] = ACTIONS(69),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(62)] = {
    [sym__statement] = STATE(1004),
    [sym_label] = STATE(234),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1001),
    [sym_groovy_package] = STATE(1001),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1001),
    [sym_assignment] = STATE(1001),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1001),
    [sym_closure] = STATE(651),
    [sym_comment] = STATE(62),
    [sym_groovy_doc] = STATE(62),
    [sym_declaration] = STATE(1001),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(650),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1001),
    [sym_for_loop] = STATE(1001),
    [sym_for_in_loop] = STATE(1001),
    [sym_function_call] = STATE(608),
    [sym_function_declaration] = STATE(1001),
    [sym_function_definition] = STATE(1001),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1001),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1001),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1001),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1001),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1001),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1001),
    [aux_sym_source_file_repeat1] = STATE(55),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(584),
    [anon_sym_continue] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(779),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_case] = ACTIONS(781),
    [anon_sym_default] = ACTIONS(781),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(63)] = {
    [sym__statement] = STATE(1004),
    [sym_label] = STATE(234),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1001),
    [sym_groovy_package] = STATE(1001),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1001),
    [sym_assignment] = STATE(1001),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1001),
    [sym_closure] = STATE(651),
    [sym_comment] = STATE(63),
    [sym_groovy_doc] = STATE(63),
    [sym_declaration] = STATE(1001),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(650),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1001),
    [sym_for_loop] = STATE(1001),
    [sym_for_in_loop] = STATE(1001),
    [sym_function_call] = STATE(608),
    [sym_function_declaration] = STATE(1001),
    [sym_function_definition] = STATE(1001),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1001),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1001),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1001),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1001),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1001),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1001),
    [aux_sym_source_file_repeat1] = STATE(60),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(584),
    [anon_sym_continue] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(783),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_case] = ACTIONS(785),
    [anon_sym_default] = ACTIONS(785),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(64)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(64),
    [sym_groovy_doc] = STATE(64),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(787),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(65)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(65),
    [sym_groovy_doc] = STATE(65),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(789),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(66)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(66),
    [sym_groovy_doc] = STATE(66),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(791),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(67)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(67),
    [sym_groovy_doc] = STATE(67),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(68),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(791),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(68)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(68),
    [sym_groovy_doc] = STATE(68),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(793),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(69)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(69),
    [sym_groovy_doc] = STATE(69),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(118),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(795),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(70)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(70),
    [sym_groovy_doc] = STATE(70),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(72),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(797),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(71)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(71),
    [sym_groovy_doc] = STATE(71),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(797),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(72)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(72),
    [sym_groovy_doc] = STATE(72),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(799),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(73)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(73),
    [sym_groovy_doc] = STATE(73),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(74),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(799),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(74)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(74),
    [sym_groovy_doc] = STATE(74),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(801),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(75)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(75),
    [sym_groovy_doc] = STATE(75),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(77),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(803),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(76)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(76),
    [sym_groovy_doc] = STATE(76),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(803),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(77)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(77),
    [sym_groovy_doc] = STATE(77),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(805),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(78)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(78),
    [sym_groovy_doc] = STATE(78),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(79),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(805),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(79)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(79),
    [sym_groovy_doc] = STATE(79),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(807),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(80)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(80),
    [sym_groovy_doc] = STATE(80),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(82),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(809),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(81)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(81),
    [sym_groovy_doc] = STATE(81),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(809),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(82)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(82),
    [sym_groovy_doc] = STATE(82),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(811),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(83)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(83),
    [sym_groovy_doc] = STATE(83),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(84),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(811),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(84)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(84),
    [sym_groovy_doc] = STATE(84),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(813),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(85)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(85),
    [sym_groovy_doc] = STATE(85),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(107),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(815),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(86)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(86),
    [sym_groovy_doc] = STATE(86),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(88),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(817),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(87)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(87),
    [sym_groovy_doc] = STATE(87),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(817),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(88)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(88),
    [sym_groovy_doc] = STATE(88),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(819),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(89)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(89),
    [sym_groovy_doc] = STATE(89),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(90),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(819),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(90)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(90),
    [sym_groovy_doc] = STATE(90),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(821),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(91)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(91),
    [sym_groovy_doc] = STATE(91),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(93),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(823),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(92)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(92),
    [sym_groovy_doc] = STATE(92),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(823),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(93)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(93),
    [sym_groovy_doc] = STATE(93),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(825),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(94)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(94),
    [sym_groovy_doc] = STATE(94),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(95),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(825),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(95)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(95),
    [sym_groovy_doc] = STATE(95),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(827),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(96)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(96),
    [sym_groovy_doc] = STATE(96),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(66),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(789),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(97)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(97),
    [sym_groovy_doc] = STATE(97),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(829),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(98)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(98),
    [sym_groovy_doc] = STATE(98),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(831),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(99)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(99),
    [sym_groovy_doc] = STATE(99),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(100),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(831),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(100)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(100),
    [sym_groovy_doc] = STATE(100),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(833),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(101)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(101),
    [sym_groovy_doc] = STATE(101),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(815),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(102)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(102),
    [sym_groovy_doc] = STATE(102),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(104),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(835),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(103)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(103),
    [sym_groovy_doc] = STATE(103),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(835),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(104)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(104),
    [sym_groovy_doc] = STATE(104),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(837),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(105)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(105),
    [sym_groovy_doc] = STATE(105),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(106),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(837),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(106)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(106),
    [sym_groovy_doc] = STATE(106),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(839),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(107)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(107),
    [sym_groovy_doc] = STATE(107),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(795),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(108)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(108),
    [sym_groovy_doc] = STATE(108),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(110),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(841),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(109)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(109),
    [sym_groovy_doc] = STATE(109),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(841),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(110)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(110),
    [sym_groovy_doc] = STATE(110),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(843),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(111)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(111),
    [sym_groovy_doc] = STATE(111),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(112),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(843),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(112)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(112),
    [sym_groovy_doc] = STATE(112),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(845),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(113)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(113),
    [sym_groovy_doc] = STATE(113),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(115),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(847),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(114)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(114),
    [sym_groovy_doc] = STATE(114),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(847),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(115)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(115),
    [sym_groovy_doc] = STATE(115),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(849),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(116)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(116),
    [sym_groovy_doc] = STATE(116),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(117),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(849),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(117)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(117),
    [sym_groovy_doc] = STATE(117),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(851),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(118)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(118),
    [sym_groovy_doc] = STATE(118),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(853),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(119)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(119),
    [sym_groovy_doc] = STATE(119),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(120),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(787),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(120)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(120),
    [sym_groovy_doc] = STATE(120),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(855),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(121)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(121),
    [sym_groovy_doc] = STATE(121),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(654),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(122),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(855),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(122)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(122),
    [sym_groovy_doc] = STATE(122),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(656),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(59),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(857),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(123)] = {
    [sym__statement] = STATE(1003),
    [sym_label] = STATE(235),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(647),
    [sym_comment] = STATE(123),
    [sym_groovy_doc] = STATE(123),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(614),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(918),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(918),
    [aux_sym_source_file_repeat1] = STATE(98),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(829),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(124)] = {
    [sym__statement] = STATE(2144),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(124),
    [sym_groovy_doc] = STATE(124),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(889),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(125)] = {
    [sym__statement] = STATE(2123),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(125),
    [sym_groovy_doc] = STATE(125),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(929),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(126)] = {
    [sym__statement] = STATE(2190),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(126),
    [sym_groovy_doc] = STATE(126),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(931),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(127)] = {
    [sym__statement] = STATE(2134),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(127),
    [sym_groovy_doc] = STATE(127),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(933),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(128)] = {
    [sym__statement] = STATE(2132),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(128),
    [sym_groovy_doc] = STATE(128),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(935),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(129)] = {
    [sym__statement] = STATE(2181),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(129),
    [sym_groovy_doc] = STATE(129),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_RPAREN] = ACTIONS(937),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(130)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(130),
    [sym_groovy_doc] = STATE(130),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(131)] = {
    [sym__statement] = STATE(2377),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(131),
    [sym_groovy_doc] = STATE(131),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(132)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(132),
    [sym_groovy_doc] = STATE(132),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(133)] = {
    [sym__statement] = STATE(976),
    [sym_label] = STATE(243),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(636),
    [sym_comment] = STATE(133),
    [sym_groovy_doc] = STATE(133),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(134)] = {
    [sym__statement] = STATE(978),
    [sym_label] = STATE(243),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(637),
    [sym_comment] = STATE(134),
    [sym_groovy_doc] = STATE(134),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(135)] = {
    [sym__statement] = STATE(992),
    [sym_label] = STATE(244),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(638),
    [sym_comment] = STATE(135),
    [sym_groovy_doc] = STATE(135),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(136)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(231),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(618),
    [sym_comment] = STATE(136),
    [sym_groovy_doc] = STATE(136),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(137)] = {
    [sym__statement] = STATE(788),
    [sym_label] = STATE(231),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(617),
    [sym_comment] = STATE(137),
    [sym_groovy_doc] = STATE(137),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(138)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(231),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(600),
    [sym_comment] = STATE(138),
    [sym_groovy_doc] = STATE(138),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(139)] = {
    [sym__statement] = STATE(941),
    [sym_label] = STATE(243),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(619),
    [sym_comment] = STATE(139),
    [sym_groovy_doc] = STATE(139),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(140)] = {
    [sym__statement] = STATE(979),
    [sym_label] = STATE(243),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(612),
    [sym_comment] = STATE(140),
    [sym_groovy_doc] = STATE(140),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(141)] = {
    [sym__statement] = STATE(813),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(483),
    [sym_comment] = STATE(141),
    [sym_groovy_doc] = STATE(141),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(142)] = {
    [sym__statement] = STATE(815),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(486),
    [sym_comment] = STATE(142),
    [sym_groovy_doc] = STATE(142),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(143)] = {
    [sym__statement] = STATE(792),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(501),
    [sym_comment] = STATE(143),
    [sym_groovy_doc] = STATE(143),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(144)] = {
    [sym__statement] = STATE(861),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(518),
    [sym_comment] = STATE(144),
    [sym_groovy_doc] = STATE(144),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(145)] = {
    [sym__statement] = STATE(877),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(533),
    [sym_comment] = STATE(145),
    [sym_groovy_doc] = STATE(145),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(146)] = {
    [sym__statement] = STATE(946),
    [sym_label] = STATE(246),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(601),
    [sym_comment] = STATE(146),
    [sym_groovy_doc] = STATE(146),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(147)] = {
    [sym__statement] = STATE(2314),
    [sym_label] = STATE(238),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1447),
    [sym_comment] = STATE(147),
    [sym_groovy_doc] = STATE(147),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(148)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(231),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(603),
    [sym_comment] = STATE(148),
    [sym_groovy_doc] = STATE(148),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(149)] = {
    [sym__statement] = STATE(868),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(476),
    [sym_comment] = STATE(149),
    [sym_groovy_doc] = STATE(149),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(150)] = {
    [sym__statement] = STATE(2406),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(150),
    [sym_groovy_doc] = STATE(150),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(151)] = {
    [sym__statement] = STATE(1831),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(151),
    [sym_groovy_doc] = STATE(151),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(152)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(230),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(152),
    [sym_groovy_doc] = STATE(152),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(153)] = {
    [sym__statement] = STATE(897),
    [sym_label] = STATE(246),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(601),
    [sym_comment] = STATE(153),
    [sym_groovy_doc] = STATE(153),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(154)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(230),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(154),
    [sym_groovy_doc] = STATE(154),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(155)] = {
    [sym__statement] = STATE(2207),
    [sym_label] = STATE(242),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(155),
    [sym_groovy_doc] = STATE(155),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(156)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(228),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(156),
    [sym_groovy_doc] = STATE(156),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(157)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(228),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(157),
    [sym_groovy_doc] = STATE(157),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(158)] = {
    [sym__statement] = STATE(897),
    [sym_label] = STATE(242),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(158),
    [sym_groovy_doc] = STATE(158),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(159)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(228),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(159),
    [sym_groovy_doc] = STATE(159),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(160)] = {
    [sym__statement] = STATE(718),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(523),
    [sym_comment] = STATE(160),
    [sym_groovy_doc] = STATE(160),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(161)] = {
    [sym__statement] = STATE(718),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(161),
    [sym_groovy_doc] = STATE(161),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(162)] = {
    [sym__statement] = STATE(947),
    [sym_label] = STATE(244),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(638),
    [sym_comment] = STATE(162),
    [sym_groovy_doc] = STATE(162),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(163)] = {
    [sym__statement] = STATE(805),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(501),
    [sym_comment] = STATE(163),
    [sym_groovy_doc] = STATE(163),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(164)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(476),
    [sym_comment] = STATE(164),
    [sym_groovy_doc] = STATE(164),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(165)] = {
    [sym__statement] = STATE(788),
    [sym_label] = STATE(230),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1418),
    [sym_comment] = STATE(165),
    [sym_groovy_doc] = STATE(165),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(166)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(230),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(166),
    [sym_groovy_doc] = STATE(166),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(167)] = {
    [sym__statement] = STATE(2054),
    [sym_label] = STATE(245),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(167),
    [sym_groovy_doc] = STATE(167),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(168)] = {
    [sym__statement] = STATE(718),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(168),
    [sym_groovy_doc] = STATE(168),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(169)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(169),
    [sym_groovy_doc] = STATE(169),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(170)] = {
    [sym__statement] = STATE(897),
    [sym_label] = STATE(245),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(170),
    [sym_groovy_doc] = STATE(170),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(171)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(476),
    [sym_comment] = STATE(171),
    [sym_groovy_doc] = STATE(171),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(172)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(521),
    [sym_comment] = STATE(172),
    [sym_groovy_doc] = STATE(172),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(173)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(173),
    [sym_groovy_doc] = STATE(173),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(174)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(522),
    [sym_comment] = STATE(174),
    [sym_groovy_doc] = STATE(174),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(175)] = {
    [sym__statement] = STATE(776),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(523),
    [sym_comment] = STATE(175),
    [sym_groovy_doc] = STATE(175),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(176)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(176),
    [sym_groovy_doc] = STATE(176),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(177)] = {
    [sym__statement] = STATE(878),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(510),
    [sym_comment] = STATE(177),
    [sym_groovy_doc] = STATE(177),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(178)] = {
    [sym__statement] = STATE(806),
    [sym_label] = STATE(240),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(510),
    [sym_comment] = STATE(178),
    [sym_groovy_doc] = STATE(178),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(179)] = {
    [sym__statement] = STATE(733),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(546),
    [sym_comment] = STATE(179),
    [sym_groovy_doc] = STATE(179),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(180)] = {
    [sym__statement] = STATE(695),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(180),
    [sym_groovy_doc] = STATE(180),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(181)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(233),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(554),
    [sym_comment] = STATE(181),
    [sym_groovy_doc] = STATE(181),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(182)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(182),
    [sym_groovy_doc] = STATE(182),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(183)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(183),
    [sym_groovy_doc] = STATE(183),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(184)] = {
    [sym__statement] = STATE(1967),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(184),
    [sym_groovy_doc] = STATE(184),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(185)] = {
    [sym__statement] = STATE(733),
    [sym_label] = STATE(236),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1418),
    [sym_comment] = STATE(185),
    [sym_groovy_doc] = STATE(185),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(186)] = {
    [sym__statement] = STATE(868),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(476),
    [sym_comment] = STATE(186),
    [sym_groovy_doc] = STATE(186),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(187)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(246),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(618),
    [sym_comment] = STATE(187),
    [sym_groovy_doc] = STATE(187),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(188)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(246),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(600),
    [sym_comment] = STATE(188),
    [sym_groovy_doc] = STATE(188),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(189)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(246),
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(986),
    [sym_groovy_package] = STATE(986),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(986),
    [sym_assignment] = STATE(986),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(986),
    [sym_closure] = STATE(603),
    [sym_comment] = STATE(189),
    [sym_groovy_doc] = STATE(189),
    [sym_declaration] = STATE(986),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(602),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(986),
    [sym_for_loop] = STATE(986),
    [sym_for_in_loop] = STATE(986),
    [sym_function_call] = STATE(596),
    [sym_function_declaration] = STATE(986),
    [sym_function_definition] = STATE(986),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(986),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(986),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(986),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(986),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(986),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(986),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(11),
    [anon_sym_break] = ACTIONS(991),
    [anon_sym_continue] = ACTIONS(991),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(190)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(242),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(190),
    [sym_groovy_doc] = STATE(190),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(191)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(242),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(191),
    [sym_groovy_doc] = STATE(191),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(192)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(242),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(192),
    [sym_groovy_doc] = STATE(192),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(193)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(521),
    [sym_comment] = STATE(193),
    [sym_groovy_doc] = STATE(193),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(194)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(522),
    [sym_comment] = STATE(194),
    [sym_groovy_doc] = STATE(194),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(195)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(239),
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(554),
    [sym_comment] = STATE(195),
    [sym_groovy_doc] = STATE(195),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(494),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(435),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(708),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1083),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(196)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(196),
    [sym_groovy_doc] = STATE(196),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(197)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(197),
    [sym_groovy_doc] = STATE(197),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(198)] = {
    [sym__statement] = STATE(1896),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1416),
    [sym_comment] = STATE(198),
    [sym_groovy_doc] = STATE(198),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(199)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(199),
    [sym_groovy_doc] = STATE(199),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(200)] = {
    [sym__statement] = STATE(976),
    [sym_label] = STATE(244),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(636),
    [sym_comment] = STATE(200),
    [sym_groovy_doc] = STATE(200),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(201)] = {
    [sym__statement] = STATE(978),
    [sym_label] = STATE(244),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(637),
    [sym_comment] = STATE(201),
    [sym_groovy_doc] = STATE(201),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(202)] = {
    [sym__statement] = STATE(979),
    [sym_label] = STATE(244),
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(943),
    [sym_groovy_package] = STATE(943),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(943),
    [sym_assignment] = STATE(943),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(943),
    [sym_closure] = STATE(612),
    [sym_comment] = STATE(202),
    [sym_groovy_doc] = STATE(202),
    [sym_declaration] = STATE(943),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(624),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(943),
    [sym_for_loop] = STATE(943),
    [sym_for_in_loop] = STATE(943),
    [sym_function_call] = STATE(594),
    [sym_function_declaration] = STATE(943),
    [sym_function_definition] = STATE(943),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(943),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(943),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(943),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(943),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(943),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(943),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_break] = ACTIONS(981),
    [anon_sym_continue] = ACTIONS(981),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(203)] = {
    [sym__statement] = STATE(813),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(483),
    [sym_comment] = STATE(203),
    [sym_groovy_doc] = STATE(203),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(204)] = {
    [sym__statement] = STATE(815),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(486),
    [sym_comment] = STATE(204),
    [sym_groovy_doc] = STATE(204),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(205)] = {
    [sym__statement] = STATE(877),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(533),
    [sym_comment] = STATE(205),
    [sym_groovy_doc] = STATE(205),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(206)] = {
    [sym__statement] = STATE(1831),
    [sym_label] = STATE(241),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(206),
    [sym_groovy_doc] = STATE(206),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(207)] = {
    [sym__statement] = STATE(817),
    [sym_label] = STATE(245),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(207),
    [sym_groovy_doc] = STATE(207),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(208)] = {
    [sym__statement] = STATE(831),
    [sym_label] = STATE(245),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(208),
    [sym_groovy_doc] = STATE(208),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(209)] = {
    [sym__statement] = STATE(830),
    [sym_label] = STATE(245),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(209),
    [sym_groovy_doc] = STATE(209),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(210)] = {
    [sym__statement] = STATE(878),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(510),
    [sym_comment] = STATE(210),
    [sym_groovy_doc] = STATE(210),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(211)] = {
    [sym__statement] = STATE(806),
    [sym_label] = STATE(237),
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(890),
    [sym_groovy_package] = STATE(890),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(890),
    [sym_assignment] = STATE(890),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(890),
    [sym_closure] = STATE(510),
    [sym_comment] = STATE(211),
    [sym_groovy_doc] = STATE(211),
    [sym_declaration] = STATE(890),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(539),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(890),
    [sym_for_loop] = STATE(890),
    [sym_for_in_loop] = STATE(890),
    [sym_function_call] = STATE(426),
    [sym_function_declaration] = STATE(890),
    [sym_function_definition] = STATE(890),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(890),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(890),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(890),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(890),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(890),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(890),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(993),
    [anon_sym_break] = ACTIONS(995),
    [anon_sym_continue] = ACTIONS(995),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(212)] = {
    [sym__statement] = STATE(703),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1413),
    [sym_comment] = STATE(212),
    [sym_groovy_doc] = STATE(212),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(213)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(213),
    [sym_groovy_doc] = STATE(213),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(214)] = {
    [sym__statement] = STATE(733),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1418),
    [sym_comment] = STATE(214),
    [sym_groovy_doc] = STATE(214),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(215)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(215),
    [sym_groovy_doc] = STATE(215),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(216)] = {
    [sym__statement] = STATE(1831),
    [sym_label] = STATE(227),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(216),
    [sym_groovy_doc] = STATE(216),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(217)] = {
    [sym__statement] = STATE(2412),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(217),
    [sym_groovy_doc] = STATE(217),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(218)] = {
    [sym__statement] = STATE(707),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1414),
    [sym_comment] = STATE(218),
    [sym_groovy_doc] = STATE(218),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(219)] = {
    [sym__statement] = STATE(727),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1399),
    [sym_comment] = STATE(219),
    [sym_groovy_doc] = STATE(219),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(220)] = {
    [sym__statement] = STATE(1831),
    [sym_label] = STATE(226),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1405),
    [sym_comment] = STATE(220),
    [sym_groovy_doc] = STATE(220),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(221)] = {
    [sym__statement] = STATE(2344),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(221),
    [sym_groovy_doc] = STATE(221),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(222)] = {
    [sym__statement] = STATE(2420),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(222),
    [sym_groovy_doc] = STATE(222),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(223)] = {
    [sym__statement] = STATE(2362),
    [sym_label] = STATE(229),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(918),
    [sym_groovy_package] = STATE(918),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(918),
    [sym_assignment] = STATE(918),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(918),
    [sym_closure] = STATE(1461),
    [sym_comment] = STATE(223),
    [sym_groovy_doc] = STATE(223),
    [sym_declaration] = STATE(918),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1443),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(918),
    [sym_for_loop] = STATE(918),
    [sym_for_in_loop] = STATE(918),
    [sym_function_call] = STATE(1394),
    [sym_function_declaration] = STATE(918),
    [sym_function_definition] = STATE(918),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(918),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(918),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(918),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(918),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(918),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(918),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(15),
    [anon_sym_continue] = ACTIONS(15),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(224)] = {
    [sym__statement] = STATE(788),
    [sym_label] = STATE(228),
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(708),
    [sym_groovy_package] = STATE(708),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(708),
    [sym_assignment] = STATE(708),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(708),
    [sym_closure] = STATE(1418),
    [sym_comment] = STATE(224),
    [sym_groovy_doc] = STATE(224),
    [sym_declaration] = STATE(708),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1409),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(708),
    [sym_for_loop] = STATE(708),
    [sym_for_in_loop] = STATE(708),
    [sym_function_call] = STATE(1156),
    [sym_function_declaration] = STATE(708),
    [sym_function_definition] = STATE(708),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(708),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(708),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(708),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(708),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(708),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(708),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(859),
    [anon_sym_break] = ACTIONS(939),
    [anon_sym_continue] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(225)] = {
    [sym_access_op] = STATE(1410),
    [sym_dotted_identifier] = STATE(1172),
    [sym_increment_op] = STATE(1410),
    [sym_binary_op] = STATE(1410),
    [sym_boolean_literal] = STATE(1330),
    [sym_closure] = STATE(1410),
    [sym_comment] = STATE(225),
    [sym_groovy_doc] = STATE(225),
    [sym_parenthesized_expression] = STATE(1393),
    [sym__expression] = STATE(1546),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1163),
    [sym__type_identifier] = STATE(1385),
    [sym_index] = STATE(1172),
    [sym__juxt_argument_list] = STATE(742),
    [sym_list] = STATE(1330),
    [sym_map_item] = STATE(1836),
    [sym_map] = STATE(1330),
    [sym_number_literal] = STATE(1173),
    [sym_string] = STATE(1173),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1410),
    [sym_unary_op] = STATE(1410),
    [sym_identifier] = ACTIONS(1191),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1193),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(1196),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1198),
    [anon_sym_STAR_EQ] = ACTIONS(1198),
    [anon_sym_SLASH_EQ] = ACTIONS(1198),
    [anon_sym_PERCENT_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_EQ] = ACTIONS(1198),
    [anon_sym_DASH_EQ] = ACTIONS(1198),
    [anon_sym_LT_LT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_AMP_EQ] = ACTIONS(1198),
    [anon_sym_CARET_EQ] = ACTIONS(1198),
    [anon_sym_PIPE_EQ] = ACTIONS(1198),
    [anon_sym_QMARK_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(1203),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_RPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(1209),
    [anon_sym_this] = ACTIONS(1211),
    [aux_sym__type_identifier_token1] = ACTIONS(1213),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(1215),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(226)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(226),
    [sym_groovy_doc] = STATE(226),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(955),
    [anon_sym_for] = ACTIONS(957),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(959),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(963),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(227)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(227),
    [sym_groovy_doc] = STATE(227),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(225),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1183),
    [anon_sym_for] = ACTIONS(1185),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1187),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1189),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(228)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(228),
    [sym_groovy_doc] = STATE(228),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(229)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(932),
    [sym_groovy_package] = STATE(932),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(932),
    [sym_assignment] = STATE(932),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(932),
    [sym_closure] = STATE(1441),
    [sym_comment] = STATE(229),
    [sym_groovy_doc] = STATE(229),
    [sym_declaration] = STATE(932),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1442),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(932),
    [sym_for_loop] = STATE(932),
    [sym_for_in_loop] = STATE(932),
    [sym_function_call] = STATE(1391),
    [sym_function_declaration] = STATE(932),
    [sym_function_definition] = STATE(932),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(932),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(932),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(932),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(932),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(932),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(932),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(230)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(230),
    [sym_groovy_doc] = STATE(230),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(965),
    [anon_sym_for] = ACTIONS(967),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(971),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(231)] = {
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(964),
    [sym_groovy_package] = STATE(964),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(964),
    [sym_assignment] = STATE(964),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(964),
    [sym_closure] = STATE(627),
    [sym_comment] = STATE(231),
    [sym_groovy_doc] = STATE(231),
    [sym_declaration] = STATE(964),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(632),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(964),
    [sym_for_loop] = STATE(964),
    [sym_for_in_loop] = STATE(964),
    [sym_function_call] = STATE(588),
    [sym_function_declaration] = STATE(964),
    [sym_function_definition] = STATE(964),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(964),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(964),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(964),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(964),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(964),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(964),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(232)] = {
    [sym_access_op] = STATE(1407),
    [sym_dotted_identifier] = STATE(1167),
    [sym_increment_op] = STATE(1407),
    [sym_binary_op] = STATE(1407),
    [sym_boolean_literal] = STATE(1319),
    [sym_closure] = STATE(1407),
    [sym_comment] = STATE(232),
    [sym_groovy_doc] = STATE(232),
    [sym_parenthesized_expression] = STATE(1393),
    [sym__expression] = STATE(1546),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1162),
    [sym__type_identifier] = STATE(1385),
    [sym_index] = STATE(1167),
    [sym__juxt_argument_list] = STATE(742),
    [sym_list] = STATE(1319),
    [sym_map_item] = STATE(1829),
    [sym_map] = STATE(1319),
    [sym_number_literal] = STATE(1169),
    [sym_string] = STATE(1169),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1407),
    [sym_unary_op] = STATE(1407),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1193),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(1196),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1198),
    [anon_sym_STAR_EQ] = ACTIONS(1198),
    [anon_sym_SLASH_EQ] = ACTIONS(1198),
    [anon_sym_PERCENT_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_EQ] = ACTIONS(1198),
    [anon_sym_DASH_EQ] = ACTIONS(1198),
    [anon_sym_LT_LT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_AMP_EQ] = ACTIONS(1198),
    [anon_sym_CARET_EQ] = ACTIONS(1198),
    [anon_sym_PIPE_EQ] = ACTIONS(1198),
    [anon_sym_QMARK_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(1203),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(1230),
    [anon_sym_this] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(1213),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(1215),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(233)] = {
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(512),
    [sym_comment] = STATE(233),
    [sym_groovy_doc] = STATE(233),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(513),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(441),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(704),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1119),
    [anon_sym_for] = ACTIONS(1121),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1125),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1145),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(234)] = {
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(1002),
    [sym_groovy_package] = STATE(1002),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(1002),
    [sym_assignment] = STATE(1002),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(1002),
    [sym_closure] = STATE(645),
    [sym_comment] = STATE(234),
    [sym_groovy_doc] = STATE(234),
    [sym_declaration] = STATE(1002),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(646),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(1002),
    [sym_for_loop] = STATE(1002),
    [sym_for_in_loop] = STATE(1002),
    [sym_function_call] = STATE(610),
    [sym_function_declaration] = STATE(1002),
    [sym_function_definition] = STATE(1002),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(1002),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(1002),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(1002),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(1002),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(1002),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(1002),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(1236),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(235)] = {
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(932),
    [sym_groovy_package] = STATE(932),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(932),
    [sym_assignment] = STATE(932),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(932),
    [sym_closure] = STATE(649),
    [sym_comment] = STATE(235),
    [sym_groovy_doc] = STATE(235),
    [sym_declaration] = STATE(932),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(648),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(932),
    [sym_for_loop] = STATE(932),
    [sym_for_in_loop] = STATE(932),
    [sym_function_call] = STATE(615),
    [sym_function_declaration] = STATE(932),
    [sym_function_definition] = STATE(932),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(932),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(932),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(932),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(932),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(932),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(932),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(55),
    [anon_sym_for] = ACTIONS(57),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(61),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(85),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(236)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(236),
    [sym_groovy_doc] = STATE(236),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(973),
    [anon_sym_for] = ACTIONS(975),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(977),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(979),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(237)] = {
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(899),
    [sym_groovy_package] = STATE(899),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(899),
    [sym_assignment] = STATE(899),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(899),
    [sym_closure] = STATE(542),
    [sym_comment] = STATE(237),
    [sym_groovy_doc] = STATE(237),
    [sym_declaration] = STATE(899),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(545),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(899),
    [sym_for_loop] = STATE(899),
    [sym_for_in_loop] = STATE(899),
    [sym_function_call] = STATE(427),
    [sym_function_declaration] = STATE(899),
    [sym_function_definition] = STATE(899),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(899),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(899),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(899),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(899),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(899),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(899),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(1240),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1067),
    [anon_sym_for] = ACTIONS(1069),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1071),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1073),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(238)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(932),
    [sym_groovy_package] = STATE(932),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(932),
    [sym_assignment] = STATE(932),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(932),
    [sym_closure] = STATE(1441),
    [sym_comment] = STATE(238),
    [sym_groovy_doc] = STATE(238),
    [sym_declaration] = STATE(932),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1442),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(932),
    [sym_for_loop] = STATE(932),
    [sym_for_in_loop] = STATE(932),
    [sym_function_call] = STATE(1391),
    [sym_function_declaration] = STATE(932),
    [sym_function_definition] = STATE(932),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(932),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(932),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(932),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(932),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(932),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(932),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(897),
    [anon_sym_for] = ACTIONS(899),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(903),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(923),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(239)] = {
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(15),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(491),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(512),
    [sym_comment] = STATE(239),
    [sym_groovy_doc] = STATE(239),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(13),
    [sym__expression] = STATE(513),
    [sym__primary_expression] = STATE(331),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(2),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(441),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(296),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(15),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_return] = STATE(704),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(475),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2082),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1683),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(475),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1605),
    [aux_sym_class_definition_repeat2] = STATE(1661),
    [sym_identifier] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1089),
    [anon_sym_package] = ACTIONS(1091),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1093),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1103),
    [anon_sym_interface] = ACTIONS(1105),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(1109),
    [anon_sym_LPAREN] = ACTIONS(1111),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1117),
    [anon_sym_while] = ACTIONS(1159),
    [anon_sym_for] = ACTIONS(1161),
    [aux_sym__type_identifier_token1] = ACTIONS(1123),
    [anon_sym_if] = ACTIONS(1163),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1165),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(240)] = {
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(18),
    [sym_groovy_import] = STATE(899),
    [sym_groovy_package] = STATE(899),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(899),
    [sym_assignment] = STATE(899),
    [sym_increment_op] = STATE(537),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_class_definition] = STATE(899),
    [sym_closure] = STATE(542),
    [sym_comment] = STATE(240),
    [sym_groovy_doc] = STATE(240),
    [sym_declaration] = STATE(899),
    [sym_parenthesized_expression] = STATE(19),
    [sym__expression] = STATE(545),
    [sym__primary_expression] = STATE(318),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(3),
    [sym_do_while_loop] = STATE(899),
    [sym_for_loop] = STATE(899),
    [sym_for_in_loop] = STATE(899),
    [sym_function_call] = STATE(427),
    [sym_function_declaration] = STATE(899),
    [sym_function_definition] = STATE(899),
    [sym__type_identifier] = STATE(255),
    [sym_if_statement] = STATE(899),
    [sym_index] = STATE(18),
    [sym_juxt_function_call] = STATE(899),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_return] = STATE(899),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_switch_statement] = STATE(899),
    [sym_ternary_op] = STATE(387),
    [sym_try_statement] = STATE(899),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2117),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1677),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(387),
    [sym_while_loop] = STATE(899),
    [aux_sym_class_definition_repeat1] = STATE(1611),
    [aux_sym_class_definition_repeat2] = STATE(1662),
    [sym_identifier] = ACTIONS(1240),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1001),
    [anon_sym_package] = ACTIONS(1003),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(1005),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1015),
    [anon_sym_interface] = ACTIONS(1017),
    [anon_sym_class] = ACTIONS(1017),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1021),
    [anon_sym_def] = ACTIONS(1023),
    [anon_sym_LPAREN] = ACTIONS(1025),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1031),
    [anon_sym_while] = ACTIONS(1033),
    [anon_sym_for] = ACTIONS(1035),
    [aux_sym__type_identifier_token1] = ACTIONS(1037),
    [anon_sym_if] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1047),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1059),
    [anon_sym_try] = ACTIONS(1061),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(241)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(241),
    [sym_groovy_doc] = STATE(241),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(232),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2119),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1681),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1609),
    [aux_sym_class_definition_repeat2] = STATE(1669),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(941),
    [anon_sym_package] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(945),
    [anon_sym_interface] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(949),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(953),
    [anon_sym_while] = ACTIONS(1167),
    [anon_sym_for] = ACTIONS(1169),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(961),
    [anon_sym_try] = ACTIONS(1173),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(242)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(242),
    [sym_groovy_doc] = STATE(242),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(283),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1151),
    [anon_sym_for] = ACTIONS(1153),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1157),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(243)] = {
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(937),
    [sym_groovy_package] = STATE(937),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(937),
    [sym_assignment] = STATE(937),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(937),
    [sym_closure] = STATE(630),
    [sym_comment] = STATE(243),
    [sym_groovy_doc] = STATE(243),
    [sym_declaration] = STATE(937),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(631),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(937),
    [sym_for_loop] = STATE(937),
    [sym_for_in_loop] = STATE(937),
    [sym_function_call] = STATE(598),
    [sym_function_declaration] = STATE(937),
    [sym_function_definition] = STATE(937),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(937),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(937),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(937),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(937),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(937),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(937),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(1236),
    [anon_sym_break] = ACTIONS(1244),
    [anon_sym_continue] = ACTIONS(1244),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(624),
    [anon_sym_for] = ACTIONS(626),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(630),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(654),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(244)] = {
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(34),
    [sym_groovy_import] = STATE(937),
    [sym_groovy_package] = STATE(937),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(937),
    [sym_assignment] = STATE(937),
    [sym_increment_op] = STATE(625),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_class_definition] = STATE(937),
    [sym_closure] = STATE(630),
    [sym_comment] = STATE(244),
    [sym_groovy_doc] = STATE(244),
    [sym_declaration] = STATE(937),
    [sym_parenthesized_expression] = STATE(35),
    [sym__expression] = STATE(631),
    [sym__primary_expression] = STATE(559),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(4),
    [sym_do_while_loop] = STATE(937),
    [sym_for_loop] = STATE(937),
    [sym_for_in_loop] = STATE(937),
    [sym_function_call] = STATE(598),
    [sym_function_declaration] = STATE(937),
    [sym_function_definition] = STATE(937),
    [sym__type_identifier] = STATE(467),
    [sym_if_statement] = STATE(937),
    [sym_index] = STATE(34),
    [sym_juxt_function_call] = STATE(937),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_return] = STATE(937),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_switch_statement] = STATE(937),
    [sym_ternary_op] = STATE(597),
    [sym_try_statement] = STATE(937),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2079),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1679),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(597),
    [sym_while_loop] = STATE(937),
    [aux_sym_class_definition_repeat1] = STATE(1610),
    [aux_sym_class_definition_repeat2] = STATE(1668),
    [sym_identifier] = ACTIONS(1236),
    [anon_sym_break] = ACTIONS(1244),
    [anon_sym_continue] = ACTIONS(1244),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(590),
    [anon_sym_package] = ACTIONS(592),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(594),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(604),
    [anon_sym_interface] = ACTIONS(606),
    [anon_sym_class] = ACTIONS(606),
    [anon_sym_LBRACE] = ACTIONS(608),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(612),
    [anon_sym_def] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(616),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(622),
    [anon_sym_while] = ACTIONS(983),
    [anon_sym_for] = ACTIONS(985),
    [aux_sym__type_identifier_token1] = ACTIONS(628),
    [anon_sym_if] = ACTIONS(987),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(638),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(650),
    [anon_sym_try] = ACTIONS(989),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(245)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(665),
    [sym_groovy_import] = STATE(704),
    [sym_groovy_package] = STATE(704),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(704),
    [sym_assignment] = STATE(704),
    [sym_increment_op] = STATE(1421),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_class_definition] = STATE(704),
    [sym_closure] = STATE(1423),
    [sym_comment] = STATE(245),
    [sym_groovy_doc] = STATE(245),
    [sym_declaration] = STATE(704),
    [sym_parenthesized_expression] = STATE(1000),
    [sym__expression] = STATE(1424),
    [sym__primary_expression] = STATE(1174),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(247),
    [sym_do_while_loop] = STATE(704),
    [sym_for_loop] = STATE(704),
    [sym_for_in_loop] = STATE(704),
    [sym_function_call] = STATE(1165),
    [sym_function_declaration] = STATE(704),
    [sym_function_definition] = STATE(704),
    [sym__type_identifier] = STATE(1075),
    [sym_if_statement] = STATE(704),
    [sym_index] = STATE(665),
    [sym_juxt_function_call] = STATE(704),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_return] = STATE(704),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_switch_statement] = STATE(704),
    [sym_ternary_op] = STATE(1283),
    [sym_try_statement] = STATE(704),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2045),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1678),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [sym_while_loop] = STATE(704),
    [aux_sym_class_definition_repeat1] = STATE(1612),
    [aux_sym_class_definition_repeat2] = STATE(1657),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1220),
    [anon_sym_continue] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_import] = ACTIONS(865),
    [anon_sym_package] = ACTIONS(867),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(869),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_ATinterface] = ACTIONS(879),
    [anon_sym_interface] = ACTIONS(881),
    [anon_sym_class] = ACTIONS(881),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(885),
    [anon_sym_LPAREN] = ACTIONS(887),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [anon_sym_do] = ACTIONS(895),
    [anon_sym_while] = ACTIONS(1175),
    [anon_sym_for] = ACTIONS(1177),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_if] = ACTIONS(1179),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_return] = ACTIONS(911),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1181),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(246)] = {
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(26),
    [sym_groovy_import] = STATE(964),
    [sym_groovy_package] = STATE(964),
    [sym_annotation] = STATE(1711),
    [sym_assertion] = STATE(964),
    [sym_assignment] = STATE(964),
    [sym_increment_op] = STATE(604),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_class_definition] = STATE(964),
    [sym_closure] = STATE(627),
    [sym_comment] = STATE(246),
    [sym_groovy_doc] = STATE(246),
    [sym_declaration] = STATE(964),
    [sym_parenthesized_expression] = STATE(41),
    [sym__expression] = STATE(632),
    [sym__primary_expression] = STATE(566),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(5),
    [sym_do_while_loop] = STATE(964),
    [sym_for_loop] = STATE(964),
    [sym_for_in_loop] = STATE(964),
    [sym_function_call] = STATE(588),
    [sym_function_declaration] = STATE(964),
    [sym_function_definition] = STATE(964),
    [sym__type_identifier] = STATE(421),
    [sym_if_statement] = STATE(964),
    [sym_index] = STATE(26),
    [sym_juxt_function_call] = STATE(964),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_return] = STATE(964),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_switch_statement] = STATE(964),
    [sym_ternary_op] = STATE(480),
    [sym_try_statement] = STATE(964),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2059),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1685),
    [sym_modifier] = STATE(1749),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(480),
    [sym_while_loop] = STATE(964),
    [aux_sym_class_definition_repeat1] = STATE(1613),
    [aux_sym_class_definition_repeat2] = STATE(1656),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(21),
    [anon_sym_package] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [anon_sym_assert] = ACTIONS(27),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(37),
    [anon_sym_interface] = ACTIONS(39),
    [anon_sym_class] = ACTIONS(39),
    [anon_sym_LBRACE] = ACTIONS(41),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(53),
    [anon_sym_while] = ACTIONS(1075),
    [anon_sym_for] = ACTIONS(1077),
    [aux_sym__type_identifier_token1] = ACTIONS(59),
    [anon_sym_if] = ACTIONS(1079),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_return] = ACTIONS(71),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(83),
    [anon_sym_try] = ACTIONS(1081),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(89),
    [anon_sym_protected] = ACTIONS(89),
    [anon_sym_private] = ACTIONS(89),
    [anon_sym_static] = ACTIONS(91),
    [anon_sym_final] = ACTIONS(91),
    [anon_sym_synchronized] = ACTIONS(91),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(247)] = {
    [sym_access_op] = STATE(1438),
    [sym_dotted_identifier] = STATE(1387),
    [sym_increment_op] = STATE(1438),
    [sym_binary_op] = STATE(1438),
    [sym_boolean_literal] = STATE(1390),
    [sym_closure] = STATE(1438),
    [sym_comment] = STATE(247),
    [sym_groovy_doc] = STATE(247),
    [sym_parenthesized_expression] = STATE(1393),
    [sym__expression] = STATE(1546),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1386),
    [sym__type_identifier] = STATE(1385),
    [sym_index] = STATE(1387),
    [sym__juxt_argument_list] = STATE(895),
    [sym_list] = STATE(1390),
    [sym_map_item] = STATE(1981),
    [sym_map] = STATE(1390),
    [sym_number_literal] = STATE(1384),
    [sym_string] = STATE(1384),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1438),
    [sym_unary_op] = STATE(1438),
    [sym_identifier] = ACTIONS(1246),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1193),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(1196),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1198),
    [anon_sym_STAR_EQ] = ACTIONS(1198),
    [anon_sym_SLASH_EQ] = ACTIONS(1198),
    [anon_sym_PERCENT_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_EQ] = ACTIONS(1198),
    [anon_sym_DASH_EQ] = ACTIONS(1198),
    [anon_sym_LT_LT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_AMP_EQ] = ACTIONS(1198),
    [anon_sym_CARET_EQ] = ACTIONS(1198),
    [anon_sym_PIPE_EQ] = ACTIONS(1198),
    [anon_sym_QMARK_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(1203),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_RPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(1248),
    [anon_sym_this] = ACTIONS(1250),
    [aux_sym__type_identifier_token1] = ACTIONS(1213),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(1215),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(248)] = {
    [sym_comment] = STATE(248),
    [sym_groovy_doc] = STATE(248),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_COLON] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_extends] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_RPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(387),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(389),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(249)] = {
    [sym_comment] = STATE(249),
    [sym_groovy_doc] = STATE(249),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1256),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(250)] = {
    [sym_comment] = STATE(250),
    [sym_groovy_doc] = STATE(250),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1256),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(251)] = {
    [sym_comment] = STATE(251),
    [sym_groovy_doc] = STATE(251),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1264),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(252)] = {
    [sym_comment] = STATE(252),
    [sym_groovy_doc] = STATE(252),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1264),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(253)] = {
    [sym_comment] = STATE(253),
    [sym_groovy_doc] = STATE(253),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1256),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(254)] = {
    [sym_comment] = STATE(254),
    [sym_groovy_doc] = STATE(254),
    [aux_sym_dotted_identifier_repeat1] = STATE(286),
    [ts_builtin_sym_end] = ACTIONS(1278),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_DOT] = ACTIONS(1282),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_pipeline] = ACTIONS(1280),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_catch] = ACTIONS(1280),
    [anon_sym_finally] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(255)] = {
    [sym_comment] = STATE(255),
    [sym_groovy_doc] = STATE(255),
    [aux_sym_dotted_identifier_repeat1] = STATE(12),
    [sym_identifier] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1288),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(102),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1290),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(256)] = {
    [sym_comment] = STATE(256),
    [sym_groovy_doc] = STATE(256),
    [aux_sym_dotted_identifier_repeat1] = STATE(269),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_DOT] = ACTIONS(1292),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1294),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_case] = ACTIONS(1280),
    [anon_sym_default] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_catch] = ACTIONS(1280),
    [anon_sym_finally] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(257)] = {
    [sym_comment] = STATE(257),
    [sym_groovy_doc] = STATE(257),
    [sym_identifier] = ACTIONS(1296),
    [anon_sym_break] = ACTIONS(1296),
    [anon_sym_continue] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1298),
    [anon_sym_COLON] = ACTIONS(1298),
    [anon_sym_DOT_AMP] = ACTIONS(1298),
    [anon_sym_DOT_AT] = ACTIONS(1298),
    [anon_sym_QMARK_DOT] = ACTIONS(1298),
    [anon_sym_STAR_DOT] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_STAR_COLON] = ACTIONS(1298),
    [anon_sym_DOT] = ACTIONS(1296),
    [anon_sym_import] = ACTIONS(1296),
    [anon_sym_as] = ACTIONS(1296),
    [anon_sym_package] = ACTIONS(1296),
    [anon_sym_AT] = ACTIONS(1296),
    [anon_sym_assert] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1298),
    [anon_sym_PERCENT] = ACTIONS(1298),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_LT_LT] = ACTIONS(1298),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_GT_GT_GT] = ACTIONS(1298),
    [anon_sym_DOT_DOT] = ACTIONS(1296),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_GT_EQ] = ACTIONS(1298),
    [anon_sym_in] = ACTIONS(1296),
    [anon_sym_BANGin] = ACTIONS(1296),
    [anon_sym_instanceof] = ACTIONS(1296),
    [anon_sym_BANGinstanceof] = ACTIONS(1298),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT_EQ_GT] = ACTIONS(1298),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1298),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1298),
    [anon_sym_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1298),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_AMP_AMP] = ACTIONS(1298),
    [anon_sym_PIPE_PIPE] = ACTIONS(1298),
    [anon_sym_QMARK_COLON] = ACTIONS(1298),
    [anon_sym_STAR_STAR] = ACTIONS(1298),
    [anon_sym_true] = ACTIONS(1296),
    [anon_sym_false] = ACTIONS(1296),
    [anon_sym_ATinterface] = ACTIONS(1298),
    [anon_sym_interface] = ACTIONS(1296),
    [anon_sym_class] = ACTIONS(1296),
    [anon_sym_COMMA] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_RBRACE] = ACTIONS(1298),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1296),
    [anon_sym_def] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1298),
    [anon_sym_null] = ACTIONS(1296),
    [anon_sym_this] = ACTIONS(1296),
    [anon_sym_do] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(1296),
    [anon_sym_for] = ACTIONS(1296),
    [aux_sym__type_identifier_token1] = ACTIONS(1296),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_else] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1298),
    [aux_sym_number_literal_token1] = ACTIONS(1296),
    [aux_sym_number_literal_token2] = ACTIONS(1298),
    [aux_sym_number_literal_token3] = ACTIONS(1298),
    [aux_sym_number_literal_token4] = ACTIONS(1296),
    [aux_sym_number_literal_token5] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1298),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1298),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1296),
    [anon_sym_case] = ACTIONS(1296),
    [anon_sym_default] = ACTIONS(1296),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_try] = ACTIONS(1296),
    [anon_sym_catch] = ACTIONS(1296),
    [anon_sym_finally] = ACTIONS(1296),
    [anon_sym_int] = ACTIONS(1296),
    [anon_sym_boolean] = ACTIONS(1296),
    [anon_sym_char] = ACTIONS(1296),
    [anon_sym_short] = ACTIONS(1296),
    [anon_sym_long] = ACTIONS(1296),
    [anon_sym_float] = ACTIONS(1296),
    [anon_sym_double] = ACTIONS(1296),
    [anon_sym_void] = ACTIONS(1296),
    [anon_sym_public] = ACTIONS(1296),
    [anon_sym_protected] = ACTIONS(1296),
    [anon_sym_private] = ACTIONS(1296),
    [anon_sym_static] = ACTIONS(1296),
    [anon_sym_final] = ACTIONS(1296),
    [anon_sym_synchronized] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1298),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_new] = ACTIONS(1296),
  },
  [STATE(258)] = {
    [sym_comment] = STATE(258),
    [sym_groovy_doc] = STATE(258),
    [ts_builtin_sym_end] = ACTIONS(1298),
    [sym_identifier] = ACTIONS(1296),
    [anon_sym_break] = ACTIONS(1296),
    [anon_sym_continue] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1298),
    [anon_sym_COLON] = ACTIONS(1298),
    [anon_sym_DOT_AMP] = ACTIONS(1298),
    [anon_sym_DOT_AT] = ACTIONS(1298),
    [anon_sym_QMARK_DOT] = ACTIONS(1298),
    [anon_sym_STAR_DOT] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_STAR_COLON] = ACTIONS(1298),
    [anon_sym_DOT] = ACTIONS(1296),
    [anon_sym_import] = ACTIONS(1296),
    [anon_sym_as] = ACTIONS(1296),
    [anon_sym_package] = ACTIONS(1296),
    [anon_sym_AT] = ACTIONS(1296),
    [anon_sym_assert] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1298),
    [anon_sym_PERCENT] = ACTIONS(1298),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_LT_LT] = ACTIONS(1298),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_GT_GT_GT] = ACTIONS(1298),
    [anon_sym_DOT_DOT] = ACTIONS(1296),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_GT_EQ] = ACTIONS(1298),
    [anon_sym_in] = ACTIONS(1296),
    [anon_sym_BANGin] = ACTIONS(1296),
    [anon_sym_instanceof] = ACTIONS(1296),
    [anon_sym_BANGinstanceof] = ACTIONS(1298),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT_EQ_GT] = ACTIONS(1298),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1298),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1298),
    [anon_sym_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1298),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_AMP_AMP] = ACTIONS(1298),
    [anon_sym_PIPE_PIPE] = ACTIONS(1298),
    [anon_sym_QMARK_COLON] = ACTIONS(1298),
    [anon_sym_STAR_STAR] = ACTIONS(1298),
    [anon_sym_true] = ACTIONS(1296),
    [anon_sym_false] = ACTIONS(1296),
    [anon_sym_ATinterface] = ACTIONS(1298),
    [anon_sym_interface] = ACTIONS(1296),
    [anon_sym_class] = ACTIONS(1296),
    [anon_sym_COMMA] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_RBRACE] = ACTIONS(1298),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1296),
    [anon_sym_def] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1298),
    [anon_sym_null] = ACTIONS(1296),
    [anon_sym_this] = ACTIONS(1296),
    [anon_sym_do] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(1296),
    [anon_sym_for] = ACTIONS(1296),
    [aux_sym__type_identifier_token1] = ACTIONS(1296),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_else] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1298),
    [aux_sym_number_literal_token1] = ACTIONS(1296),
    [aux_sym_number_literal_token2] = ACTIONS(1298),
    [aux_sym_number_literal_token3] = ACTIONS(1298),
    [aux_sym_number_literal_token4] = ACTIONS(1296),
    [aux_sym_number_literal_token5] = ACTIONS(1298),
    [anon_sym_pipeline] = ACTIONS(1296),
    [anon_sym_return] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1298),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1298),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1296),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_try] = ACTIONS(1296),
    [anon_sym_catch] = ACTIONS(1296),
    [anon_sym_finally] = ACTIONS(1296),
    [anon_sym_int] = ACTIONS(1296),
    [anon_sym_boolean] = ACTIONS(1296),
    [anon_sym_char] = ACTIONS(1296),
    [anon_sym_short] = ACTIONS(1296),
    [anon_sym_long] = ACTIONS(1296),
    [anon_sym_float] = ACTIONS(1296),
    [anon_sym_double] = ACTIONS(1296),
    [anon_sym_void] = ACTIONS(1296),
    [anon_sym_public] = ACTIONS(1296),
    [anon_sym_protected] = ACTIONS(1296),
    [anon_sym_private] = ACTIONS(1296),
    [anon_sym_static] = ACTIONS(1296),
    [anon_sym_final] = ACTIONS(1296),
    [anon_sym_synchronized] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1298),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_new] = ACTIONS(1296),
  },
  [STATE(259)] = {
    [sym_comment] = STATE(259),
    [sym_groovy_doc] = STATE(259),
    [ts_builtin_sym_end] = ACTIONS(1300),
    [sym_identifier] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_DOT_AMP] = ACTIONS(1300),
    [anon_sym_DOT_AT] = ACTIONS(1300),
    [anon_sym_QMARK_DOT] = ACTIONS(1300),
    [anon_sym_STAR_DOT] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1302),
    [anon_sym_STAR_COLON] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1302),
    [anon_sym_import] = ACTIONS(1302),
    [anon_sym_as] = ACTIONS(1302),
    [anon_sym_package] = ACTIONS(1302),
    [anon_sym_AT] = ACTIONS(1302),
    [anon_sym_assert] = ACTIONS(1302),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1302),
    [anon_sym_GT_GT_GT] = ACTIONS(1300),
    [anon_sym_DOT_DOT] = ACTIONS(1302),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1302),
    [anon_sym_LT] = ACTIONS(1302),
    [anon_sym_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT] = ACTIONS(1302),
    [anon_sym_GT_EQ] = ACTIONS(1300),
    [anon_sym_in] = ACTIONS(1302),
    [anon_sym_BANGin] = ACTIONS(1302),
    [anon_sym_instanceof] = ACTIONS(1302),
    [anon_sym_BANGinstanceof] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1302),
    [anon_sym_BANG_EQ] = ACTIONS(1302),
    [anon_sym_LT_EQ_GT] = ACTIONS(1300),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1300),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1302),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1302),
    [anon_sym_AMP_AMP] = ACTIONS(1300),
    [anon_sym_PIPE_PIPE] = ACTIONS(1300),
    [anon_sym_QMARK_COLON] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_true] = ACTIONS(1302),
    [anon_sym_false] = ACTIONS(1302),
    [anon_sym_ATinterface] = ACTIONS(1300),
    [anon_sym_interface] = ACTIONS(1302),
    [anon_sym_class] = ACTIONS(1302),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1302),
    [anon_sym_def] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_null] = ACTIONS(1302),
    [anon_sym_this] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [aux_sym__type_identifier_token1] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_else] = ACTIONS(1302),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [aux_sym_number_literal_token1] = ACTIONS(1302),
    [aux_sym_number_literal_token2] = ACTIONS(1300),
    [aux_sym_number_literal_token3] = ACTIONS(1300),
    [aux_sym_number_literal_token4] = ACTIONS(1302),
    [aux_sym_number_literal_token5] = ACTIONS(1300),
    [anon_sym_pipeline] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1300),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_QMARK] = ACTIONS(1302),
    [anon_sym_try] = ACTIONS(1302),
    [anon_sym_catch] = ACTIONS(1302),
    [anon_sym_finally] = ACTIONS(1302),
    [anon_sym_int] = ACTIONS(1302),
    [anon_sym_boolean] = ACTIONS(1302),
    [anon_sym_char] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_float] = ACTIONS(1302),
    [anon_sym_double] = ACTIONS(1302),
    [anon_sym_void] = ACTIONS(1302),
    [anon_sym_public] = ACTIONS(1302),
    [anon_sym_protected] = ACTIONS(1302),
    [anon_sym_private] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_final] = ACTIONS(1302),
    [anon_sym_synchronized] = ACTIONS(1302),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1302),
    [anon_sym_new] = ACTIONS(1302),
  },
  [STATE(260)] = {
    [sym_comment] = STATE(260),
    [sym_groovy_doc] = STATE(260),
    [sym_identifier] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_DOT_AMP] = ACTIONS(1300),
    [anon_sym_DOT_AT] = ACTIONS(1300),
    [anon_sym_QMARK_DOT] = ACTIONS(1300),
    [anon_sym_STAR_DOT] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1302),
    [anon_sym_STAR_COLON] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1302),
    [anon_sym_import] = ACTIONS(1302),
    [anon_sym_as] = ACTIONS(1302),
    [anon_sym_package] = ACTIONS(1302),
    [anon_sym_AT] = ACTIONS(1302),
    [anon_sym_assert] = ACTIONS(1302),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1302),
    [anon_sym_GT_GT_GT] = ACTIONS(1300),
    [anon_sym_DOT_DOT] = ACTIONS(1302),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1302),
    [anon_sym_LT] = ACTIONS(1302),
    [anon_sym_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT] = ACTIONS(1302),
    [anon_sym_GT_EQ] = ACTIONS(1300),
    [anon_sym_in] = ACTIONS(1302),
    [anon_sym_BANGin] = ACTIONS(1302),
    [anon_sym_instanceof] = ACTIONS(1302),
    [anon_sym_BANGinstanceof] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1302),
    [anon_sym_BANG_EQ] = ACTIONS(1302),
    [anon_sym_LT_EQ_GT] = ACTIONS(1300),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1300),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1302),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1302),
    [anon_sym_AMP_AMP] = ACTIONS(1300),
    [anon_sym_PIPE_PIPE] = ACTIONS(1300),
    [anon_sym_QMARK_COLON] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_true] = ACTIONS(1302),
    [anon_sym_false] = ACTIONS(1302),
    [anon_sym_ATinterface] = ACTIONS(1300),
    [anon_sym_interface] = ACTIONS(1302),
    [anon_sym_class] = ACTIONS(1302),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1302),
    [anon_sym_def] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_null] = ACTIONS(1302),
    [anon_sym_this] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [aux_sym__type_identifier_token1] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_else] = ACTIONS(1302),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [aux_sym_number_literal_token1] = ACTIONS(1302),
    [aux_sym_number_literal_token2] = ACTIONS(1300),
    [aux_sym_number_literal_token3] = ACTIONS(1300),
    [aux_sym_number_literal_token4] = ACTIONS(1302),
    [aux_sym_number_literal_token5] = ACTIONS(1300),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1300),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_QMARK] = ACTIONS(1302),
    [anon_sym_try] = ACTIONS(1302),
    [anon_sym_catch] = ACTIONS(1302),
    [anon_sym_finally] = ACTIONS(1302),
    [anon_sym_int] = ACTIONS(1302),
    [anon_sym_boolean] = ACTIONS(1302),
    [anon_sym_char] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_float] = ACTIONS(1302),
    [anon_sym_double] = ACTIONS(1302),
    [anon_sym_void] = ACTIONS(1302),
    [anon_sym_public] = ACTIONS(1302),
    [anon_sym_protected] = ACTIONS(1302),
    [anon_sym_private] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_final] = ACTIONS(1302),
    [anon_sym_synchronized] = ACTIONS(1302),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1302),
    [anon_sym_new] = ACTIONS(1302),
  },
  [STATE(261)] = {
    [sym_comment] = STATE(261),
    [sym_groovy_doc] = STATE(261),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(262)] = {
    [sym_closure] = STATE(385),
    [sym_comment] = STATE(262),
    [sym_groovy_doc] = STATE(262),
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_pipeline] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_catch] = ACTIONS(1306),
    [anon_sym_finally] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(263)] = {
    [sym_comment] = STATE(263),
    [sym_groovy_doc] = STATE(263),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(264)] = {
    [sym_comment] = STATE(264),
    [sym_groovy_doc] = STATE(264),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(265)] = {
    [sym_comment] = STATE(265),
    [sym_groovy_doc] = STATE(265),
    [aux_sym_dotted_identifier_repeat1] = STATE(286),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1282),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(266)] = {
    [sym_comment] = STATE(266),
    [sym_groovy_doc] = STATE(266),
    [ts_builtin_sym_end] = ACTIONS(1314),
    [sym_identifier] = ACTIONS(1316),
    [anon_sym_break] = ACTIONS(1316),
    [anon_sym_continue] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1314),
    [anon_sym_COLON] = ACTIONS(1314),
    [anon_sym_DOT_AMP] = ACTIONS(1314),
    [anon_sym_DOT_AT] = ACTIONS(1314),
    [anon_sym_QMARK_DOT] = ACTIONS(1314),
    [anon_sym_STAR_DOT] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_STAR_COLON] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1316),
    [anon_sym_import] = ACTIONS(1316),
    [anon_sym_as] = ACTIONS(1316),
    [anon_sym_package] = ACTIONS(1316),
    [anon_sym_AT] = ACTIONS(1316),
    [anon_sym_assert] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1314),
    [anon_sym_PERCENT] = ACTIONS(1314),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_LT_LT] = ACTIONS(1314),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_GT_GT_GT] = ACTIONS(1314),
    [anon_sym_DOT_DOT] = ACTIONS(1316),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_GT_EQ] = ACTIONS(1314),
    [anon_sym_in] = ACTIONS(1316),
    [anon_sym_BANGin] = ACTIONS(1316),
    [anon_sym_instanceof] = ACTIONS(1316),
    [anon_sym_BANGinstanceof] = ACTIONS(1314),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT_EQ_GT] = ACTIONS(1314),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1314),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1314),
    [anon_sym_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1314),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_AMP_AMP] = ACTIONS(1314),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_QMARK_COLON] = ACTIONS(1314),
    [anon_sym_STAR_STAR] = ACTIONS(1314),
    [anon_sym_true] = ACTIONS(1316),
    [anon_sym_false] = ACTIONS(1316),
    [anon_sym_ATinterface] = ACTIONS(1314),
    [anon_sym_interface] = ACTIONS(1316),
    [anon_sym_class] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1314),
    [anon_sym_RBRACE] = ACTIONS(1314),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1316),
    [anon_sym_def] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1314),
    [anon_sym_null] = ACTIONS(1316),
    [anon_sym_this] = ACTIONS(1316),
    [anon_sym_do] = ACTIONS(1316),
    [anon_sym_while] = ACTIONS(1316),
    [anon_sym_for] = ACTIONS(1316),
    [aux_sym__type_identifier_token1] = ACTIONS(1316),
    [anon_sym_if] = ACTIONS(1316),
    [anon_sym_else] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [aux_sym_number_literal_token1] = ACTIONS(1316),
    [aux_sym_number_literal_token2] = ACTIONS(1314),
    [aux_sym_number_literal_token3] = ACTIONS(1314),
    [aux_sym_number_literal_token4] = ACTIONS(1316),
    [aux_sym_number_literal_token5] = ACTIONS(1314),
    [anon_sym_pipeline] = ACTIONS(1316),
    [anon_sym_return] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1314),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1314),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1316),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_try] = ACTIONS(1316),
    [anon_sym_catch] = ACTIONS(1316),
    [anon_sym_finally] = ACTIONS(1316),
    [anon_sym_int] = ACTIONS(1316),
    [anon_sym_boolean] = ACTIONS(1316),
    [anon_sym_char] = ACTIONS(1316),
    [anon_sym_short] = ACTIONS(1316),
    [anon_sym_long] = ACTIONS(1316),
    [anon_sym_float] = ACTIONS(1316),
    [anon_sym_double] = ACTIONS(1316),
    [anon_sym_void] = ACTIONS(1316),
    [anon_sym_public] = ACTIONS(1316),
    [anon_sym_protected] = ACTIONS(1316),
    [anon_sym_private] = ACTIONS(1316),
    [anon_sym_static] = ACTIONS(1316),
    [anon_sym_final] = ACTIONS(1316),
    [anon_sym_synchronized] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1314),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_new] = ACTIONS(1316),
  },
  [STATE(267)] = {
    [sym_comment] = STATE(267),
    [sym_groovy_doc] = STATE(267),
    [ts_builtin_sym_end] = ACTIONS(1318),
    [sym_identifier] = ACTIONS(1320),
    [anon_sym_break] = ACTIONS(1320),
    [anon_sym_continue] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_DOT_AMP] = ACTIONS(1318),
    [anon_sym_DOT_AT] = ACTIONS(1318),
    [anon_sym_QMARK_DOT] = ACTIONS(1318),
    [anon_sym_STAR_DOT] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_STAR_COLON] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1320),
    [anon_sym_import] = ACTIONS(1320),
    [anon_sym_as] = ACTIONS(1320),
    [anon_sym_package] = ACTIONS(1320),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_assert] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_LT_LT] = ACTIONS(1318),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_GT_GT_GT] = ACTIONS(1318),
    [anon_sym_DOT_DOT] = ACTIONS(1320),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_GT_EQ] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1320),
    [anon_sym_BANGin] = ACTIONS(1320),
    [anon_sym_instanceof] = ACTIONS(1320),
    [anon_sym_BANGinstanceof] = ACTIONS(1318),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT_EQ_GT] = ACTIONS(1318),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1318),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1318),
    [anon_sym_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_AMP_AMP] = ACTIONS(1318),
    [anon_sym_PIPE_PIPE] = ACTIONS(1318),
    [anon_sym_QMARK_COLON] = ACTIONS(1318),
    [anon_sym_STAR_STAR] = ACTIONS(1318),
    [anon_sym_true] = ACTIONS(1320),
    [anon_sym_false] = ACTIONS(1320),
    [anon_sym_ATinterface] = ACTIONS(1318),
    [anon_sym_interface] = ACTIONS(1320),
    [anon_sym_class] = ACTIONS(1320),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1320),
    [anon_sym_def] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_null] = ACTIONS(1320),
    [anon_sym_this] = ACTIONS(1320),
    [anon_sym_do] = ACTIONS(1320),
    [anon_sym_while] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1320),
    [aux_sym__type_identifier_token1] = ACTIONS(1320),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_else] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1318),
    [aux_sym_number_literal_token1] = ACTIONS(1320),
    [aux_sym_number_literal_token2] = ACTIONS(1318),
    [aux_sym_number_literal_token3] = ACTIONS(1318),
    [aux_sym_number_literal_token4] = ACTIONS(1320),
    [aux_sym_number_literal_token5] = ACTIONS(1318),
    [anon_sym_pipeline] = ACTIONS(1320),
    [anon_sym_return] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1318),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_try] = ACTIONS(1320),
    [anon_sym_catch] = ACTIONS(1320),
    [anon_sym_finally] = ACTIONS(1320),
    [anon_sym_int] = ACTIONS(1320),
    [anon_sym_boolean] = ACTIONS(1320),
    [anon_sym_char] = ACTIONS(1320),
    [anon_sym_short] = ACTIONS(1320),
    [anon_sym_long] = ACTIONS(1320),
    [anon_sym_float] = ACTIONS(1320),
    [anon_sym_double] = ACTIONS(1320),
    [anon_sym_void] = ACTIONS(1320),
    [anon_sym_public] = ACTIONS(1320),
    [anon_sym_protected] = ACTIONS(1320),
    [anon_sym_private] = ACTIONS(1320),
    [anon_sym_static] = ACTIONS(1320),
    [anon_sym_final] = ACTIONS(1320),
    [anon_sym_synchronized] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_new] = ACTIONS(1320),
  },
  [STATE(268)] = {
    [sym_comment] = STATE(268),
    [sym_groovy_doc] = STATE(268),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(269)] = {
    [sym_comment] = STATE(269),
    [sym_groovy_doc] = STATE(269),
    [aux_sym_dotted_identifier_repeat1] = STATE(291),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(373),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(373),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(373),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(373),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(373),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_COMMA] = ACTIONS(373),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_case] = ACTIONS(375),
    [anon_sym_default] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_catch] = ACTIONS(375),
    [anon_sym_finally] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(270)] = {
    [sym_closure] = STATE(320),
    [sym_comment] = STATE(270),
    [sym_groovy_doc] = STATE(270),
    [ts_builtin_sym_end] = ACTIONS(1322),
    [sym_identifier] = ACTIONS(1324),
    [anon_sym_break] = ACTIONS(1324),
    [anon_sym_continue] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1322),
    [anon_sym_DOT_AMP] = ACTIONS(1322),
    [anon_sym_DOT_AT] = ACTIONS(1322),
    [anon_sym_QMARK_DOT] = ACTIONS(1322),
    [anon_sym_STAR_DOT] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_STAR_COLON] = ACTIONS(1322),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_import] = ACTIONS(1324),
    [anon_sym_as] = ACTIONS(1324),
    [anon_sym_package] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_assert] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1322),
    [anon_sym_PERCENT] = ACTIONS(1322),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_LT_LT] = ACTIONS(1322),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_GT_GT_GT] = ACTIONS(1322),
    [anon_sym_DOT_DOT] = ACTIONS(1324),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_GT_EQ] = ACTIONS(1322),
    [anon_sym_in] = ACTIONS(1324),
    [anon_sym_BANGin] = ACTIONS(1324),
    [anon_sym_instanceof] = ACTIONS(1324),
    [anon_sym_BANGinstanceof] = ACTIONS(1322),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT_EQ_GT] = ACTIONS(1322),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1322),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1322),
    [anon_sym_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1322),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_AMP_AMP] = ACTIONS(1322),
    [anon_sym_PIPE_PIPE] = ACTIONS(1322),
    [anon_sym_QMARK_COLON] = ACTIONS(1322),
    [anon_sym_STAR_STAR] = ACTIONS(1322),
    [anon_sym_true] = ACTIONS(1324),
    [anon_sym_false] = ACTIONS(1324),
    [anon_sym_ATinterface] = ACTIONS(1322),
    [anon_sym_interface] = ACTIONS(1324),
    [anon_sym_class] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(1322),
    [anon_sym_RBRACE] = ACTIONS(1322),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_def] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1322),
    [anon_sym_RPAREN] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_this] = ACTIONS(1324),
    [anon_sym_do] = ACTIONS(1324),
    [anon_sym_while] = ACTIONS(1324),
    [anon_sym_for] = ACTIONS(1324),
    [aux_sym__type_identifier_token1] = ACTIONS(1324),
    [anon_sym_if] = ACTIONS(1324),
    [anon_sym_else] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1322),
    [aux_sym_number_literal_token1] = ACTIONS(1324),
    [aux_sym_number_literal_token2] = ACTIONS(1322),
    [aux_sym_number_literal_token3] = ACTIONS(1322),
    [aux_sym_number_literal_token4] = ACTIONS(1324),
    [aux_sym_number_literal_token5] = ACTIONS(1322),
    [anon_sym_pipeline] = ACTIONS(1324),
    [anon_sym_return] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1322),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1322),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_try] = ACTIONS(1324),
    [anon_sym_int] = ACTIONS(1324),
    [anon_sym_boolean] = ACTIONS(1324),
    [anon_sym_char] = ACTIONS(1324),
    [anon_sym_short] = ACTIONS(1324),
    [anon_sym_long] = ACTIONS(1324),
    [anon_sym_float] = ACTIONS(1324),
    [anon_sym_double] = ACTIONS(1324),
    [anon_sym_void] = ACTIONS(1324),
    [anon_sym_public] = ACTIONS(1324),
    [anon_sym_protected] = ACTIONS(1324),
    [anon_sym_private] = ACTIONS(1324),
    [anon_sym_static] = ACTIONS(1324),
    [anon_sym_final] = ACTIONS(1324),
    [anon_sym_synchronized] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1322),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_new] = ACTIONS(1324),
  },
  [STATE(271)] = {
    [sym_comment] = STATE(271),
    [sym_groovy_doc] = STATE(271),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1256),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(272)] = {
    [sym_comment] = STATE(272),
    [sym_groovy_doc] = STATE(272),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(273)] = {
    [sym_comment] = STATE(273),
    [sym_groovy_doc] = STATE(273),
    [aux_sym_dotted_identifier_repeat1] = STATE(269),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1292),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1326),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1294),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(274)] = {
    [sym_comment] = STATE(274),
    [sym_groovy_doc] = STATE(274),
    [sym_identifier] = ACTIONS(1328),
    [anon_sym_break] = ACTIONS(1328),
    [anon_sym_continue] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1330),
    [anon_sym_COLON] = ACTIONS(1330),
    [anon_sym_DOT_AMP] = ACTIONS(1330),
    [anon_sym_DOT_AT] = ACTIONS(1330),
    [anon_sym_QMARK_DOT] = ACTIONS(1330),
    [anon_sym_STAR_DOT] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_STAR_COLON] = ACTIONS(1330),
    [anon_sym_DOT] = ACTIONS(1328),
    [anon_sym_import] = ACTIONS(1328),
    [anon_sym_as] = ACTIONS(1328),
    [anon_sym_package] = ACTIONS(1328),
    [anon_sym_AT] = ACTIONS(1328),
    [anon_sym_assert] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1330),
    [anon_sym_PERCENT] = ACTIONS(1330),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_LT_LT] = ACTIONS(1330),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_GT_GT_GT] = ACTIONS(1330),
    [anon_sym_DOT_DOT] = ACTIONS(1328),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_GT_EQ] = ACTIONS(1330),
    [anon_sym_in] = ACTIONS(1328),
    [anon_sym_BANGin] = ACTIONS(1328),
    [anon_sym_instanceof] = ACTIONS(1328),
    [anon_sym_BANGinstanceof] = ACTIONS(1330),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT_EQ_GT] = ACTIONS(1330),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1330),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1330),
    [anon_sym_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1330),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_AMP_AMP] = ACTIONS(1330),
    [anon_sym_PIPE_PIPE] = ACTIONS(1330),
    [anon_sym_QMARK_COLON] = ACTIONS(1330),
    [anon_sym_STAR_STAR] = ACTIONS(1330),
    [anon_sym_true] = ACTIONS(1328),
    [anon_sym_false] = ACTIONS(1328),
    [anon_sym_ATinterface] = ACTIONS(1330),
    [anon_sym_interface] = ACTIONS(1328),
    [anon_sym_class] = ACTIONS(1328),
    [anon_sym_COMMA] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1330),
    [anon_sym_RBRACE] = ACTIONS(1330),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1328),
    [anon_sym_def] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1330),
    [anon_sym_null] = ACTIONS(1328),
    [anon_sym_this] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1328),
    [anon_sym_while] = ACTIONS(1328),
    [anon_sym_for] = ACTIONS(1328),
    [aux_sym__type_identifier_token1] = ACTIONS(1328),
    [anon_sym_if] = ACTIONS(1328),
    [anon_sym_else] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [aux_sym_number_literal_token1] = ACTIONS(1328),
    [aux_sym_number_literal_token2] = ACTIONS(1330),
    [aux_sym_number_literal_token3] = ACTIONS(1330),
    [aux_sym_number_literal_token4] = ACTIONS(1328),
    [aux_sym_number_literal_token5] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1330),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1330),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1328),
    [anon_sym_case] = ACTIONS(1328),
    [anon_sym_default] = ACTIONS(1328),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_try] = ACTIONS(1328),
    [anon_sym_catch] = ACTIONS(1328),
    [anon_sym_finally] = ACTIONS(1328),
    [anon_sym_int] = ACTIONS(1328),
    [anon_sym_boolean] = ACTIONS(1328),
    [anon_sym_char] = ACTIONS(1328),
    [anon_sym_short] = ACTIONS(1328),
    [anon_sym_long] = ACTIONS(1328),
    [anon_sym_float] = ACTIONS(1328),
    [anon_sym_double] = ACTIONS(1328),
    [anon_sym_void] = ACTIONS(1328),
    [anon_sym_public] = ACTIONS(1328),
    [anon_sym_protected] = ACTIONS(1328),
    [anon_sym_private] = ACTIONS(1328),
    [anon_sym_static] = ACTIONS(1328),
    [anon_sym_final] = ACTIONS(1328),
    [anon_sym_synchronized] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1330),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_new] = ACTIONS(1328),
  },
  [STATE(275)] = {
    [sym_comment] = STATE(275),
    [sym_groovy_doc] = STATE(275),
    [aux_sym_dotted_identifier_repeat1] = STATE(286),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1282),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(276)] = {
    [sym_comment] = STATE(276),
    [sym_groovy_doc] = STATE(276),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(277)] = {
    [sym_closure] = STATE(353),
    [sym_comment] = STATE(277),
    [sym_groovy_doc] = STATE(277),
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_pipeline] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(278)] = {
    [sym_comment] = STATE(278),
    [sym_groovy_doc] = STATE(278),
    [sym_identifier] = ACTIONS(1316),
    [anon_sym_break] = ACTIONS(1316),
    [anon_sym_continue] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1314),
    [anon_sym_COLON] = ACTIONS(1314),
    [anon_sym_DOT_AMP] = ACTIONS(1314),
    [anon_sym_DOT_AT] = ACTIONS(1314),
    [anon_sym_QMARK_DOT] = ACTIONS(1314),
    [anon_sym_STAR_DOT] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_STAR_COLON] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1316),
    [anon_sym_import] = ACTIONS(1316),
    [anon_sym_as] = ACTIONS(1316),
    [anon_sym_package] = ACTIONS(1316),
    [anon_sym_AT] = ACTIONS(1316),
    [anon_sym_assert] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1314),
    [anon_sym_PERCENT] = ACTIONS(1314),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_LT_LT] = ACTIONS(1314),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_GT_GT_GT] = ACTIONS(1314),
    [anon_sym_DOT_DOT] = ACTIONS(1316),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_GT_EQ] = ACTIONS(1314),
    [anon_sym_in] = ACTIONS(1316),
    [anon_sym_BANGin] = ACTIONS(1316),
    [anon_sym_instanceof] = ACTIONS(1316),
    [anon_sym_BANGinstanceof] = ACTIONS(1314),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT_EQ_GT] = ACTIONS(1314),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1314),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1314),
    [anon_sym_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1314),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_AMP_AMP] = ACTIONS(1314),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_QMARK_COLON] = ACTIONS(1314),
    [anon_sym_STAR_STAR] = ACTIONS(1314),
    [anon_sym_true] = ACTIONS(1316),
    [anon_sym_false] = ACTIONS(1316),
    [anon_sym_ATinterface] = ACTIONS(1314),
    [anon_sym_interface] = ACTIONS(1316),
    [anon_sym_class] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1314),
    [anon_sym_RBRACE] = ACTIONS(1314),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1316),
    [anon_sym_def] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1314),
    [anon_sym_null] = ACTIONS(1316),
    [anon_sym_this] = ACTIONS(1316),
    [anon_sym_do] = ACTIONS(1316),
    [anon_sym_while] = ACTIONS(1316),
    [anon_sym_for] = ACTIONS(1316),
    [aux_sym__type_identifier_token1] = ACTIONS(1316),
    [anon_sym_if] = ACTIONS(1316),
    [anon_sym_else] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [aux_sym_number_literal_token1] = ACTIONS(1316),
    [aux_sym_number_literal_token2] = ACTIONS(1314),
    [aux_sym_number_literal_token3] = ACTIONS(1314),
    [aux_sym_number_literal_token4] = ACTIONS(1316),
    [aux_sym_number_literal_token5] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1314),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1314),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1316),
    [anon_sym_case] = ACTIONS(1316),
    [anon_sym_default] = ACTIONS(1316),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_try] = ACTIONS(1316),
    [anon_sym_catch] = ACTIONS(1316),
    [anon_sym_finally] = ACTIONS(1316),
    [anon_sym_int] = ACTIONS(1316),
    [anon_sym_boolean] = ACTIONS(1316),
    [anon_sym_char] = ACTIONS(1316),
    [anon_sym_short] = ACTIONS(1316),
    [anon_sym_long] = ACTIONS(1316),
    [anon_sym_float] = ACTIONS(1316),
    [anon_sym_double] = ACTIONS(1316),
    [anon_sym_void] = ACTIONS(1316),
    [anon_sym_public] = ACTIONS(1316),
    [anon_sym_protected] = ACTIONS(1316),
    [anon_sym_private] = ACTIONS(1316),
    [anon_sym_static] = ACTIONS(1316),
    [anon_sym_final] = ACTIONS(1316),
    [anon_sym_synchronized] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1314),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_new] = ACTIONS(1316),
  },
  [STATE(279)] = {
    [sym_closure] = STATE(334),
    [sym_comment] = STATE(279),
    [sym_groovy_doc] = STATE(279),
    [ts_builtin_sym_end] = ACTIONS(1322),
    [sym_identifier] = ACTIONS(1324),
    [anon_sym_break] = ACTIONS(1324),
    [anon_sym_continue] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1322),
    [anon_sym_DOT_AMP] = ACTIONS(1322),
    [anon_sym_DOT_AT] = ACTIONS(1322),
    [anon_sym_QMARK_DOT] = ACTIONS(1322),
    [anon_sym_STAR_DOT] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_STAR_COLON] = ACTIONS(1322),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_import] = ACTIONS(1324),
    [anon_sym_as] = ACTIONS(1324),
    [anon_sym_package] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_assert] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1322),
    [anon_sym_PERCENT] = ACTIONS(1322),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_LT_LT] = ACTIONS(1322),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_GT_GT_GT] = ACTIONS(1322),
    [anon_sym_DOT_DOT] = ACTIONS(1324),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_GT_EQ] = ACTIONS(1322),
    [anon_sym_in] = ACTIONS(1324),
    [anon_sym_BANGin] = ACTIONS(1324),
    [anon_sym_instanceof] = ACTIONS(1324),
    [anon_sym_BANGinstanceof] = ACTIONS(1322),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT_EQ_GT] = ACTIONS(1322),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1322),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1322),
    [anon_sym_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1322),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_AMP_AMP] = ACTIONS(1322),
    [anon_sym_PIPE_PIPE] = ACTIONS(1322),
    [anon_sym_QMARK_COLON] = ACTIONS(1322),
    [anon_sym_STAR_STAR] = ACTIONS(1322),
    [anon_sym_true] = ACTIONS(1324),
    [anon_sym_false] = ACTIONS(1324),
    [anon_sym_ATinterface] = ACTIONS(1322),
    [anon_sym_interface] = ACTIONS(1324),
    [anon_sym_class] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_RBRACE] = ACTIONS(1322),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_def] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_this] = ACTIONS(1324),
    [anon_sym_do] = ACTIONS(1324),
    [anon_sym_while] = ACTIONS(1324),
    [anon_sym_for] = ACTIONS(1324),
    [aux_sym__type_identifier_token1] = ACTIONS(1324),
    [anon_sym_if] = ACTIONS(1324),
    [anon_sym_else] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1322),
    [aux_sym_number_literal_token1] = ACTIONS(1324),
    [aux_sym_number_literal_token2] = ACTIONS(1322),
    [aux_sym_number_literal_token3] = ACTIONS(1322),
    [aux_sym_number_literal_token4] = ACTIONS(1324),
    [aux_sym_number_literal_token5] = ACTIONS(1322),
    [anon_sym_pipeline] = ACTIONS(1324),
    [anon_sym_return] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1322),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1322),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_try] = ACTIONS(1324),
    [anon_sym_catch] = ACTIONS(1324),
    [anon_sym_finally] = ACTIONS(1324),
    [anon_sym_int] = ACTIONS(1324),
    [anon_sym_boolean] = ACTIONS(1324),
    [anon_sym_char] = ACTIONS(1324),
    [anon_sym_short] = ACTIONS(1324),
    [anon_sym_long] = ACTIONS(1324),
    [anon_sym_float] = ACTIONS(1324),
    [anon_sym_double] = ACTIONS(1324),
    [anon_sym_void] = ACTIONS(1324),
    [anon_sym_public] = ACTIONS(1324),
    [anon_sym_protected] = ACTIONS(1324),
    [anon_sym_private] = ACTIONS(1324),
    [anon_sym_static] = ACTIONS(1324),
    [anon_sym_final] = ACTIONS(1324),
    [anon_sym_synchronized] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1322),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_new] = ACTIONS(1324),
  },
  [STATE(280)] = {
    [sym_closure] = STATE(378),
    [sym_comment] = STATE(280),
    [sym_groovy_doc] = STATE(280),
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_pipeline] = ACTIONS(1338),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_catch] = ACTIONS(1338),
    [anon_sym_finally] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(281)] = {
    [sym_comment] = STATE(281),
    [sym_groovy_doc] = STATE(281),
    [sym_identifier] = ACTIONS(1320),
    [anon_sym_break] = ACTIONS(1320),
    [anon_sym_continue] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_DOT_AMP] = ACTIONS(1318),
    [anon_sym_DOT_AT] = ACTIONS(1318),
    [anon_sym_QMARK_DOT] = ACTIONS(1318),
    [anon_sym_STAR_DOT] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_STAR_COLON] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1320),
    [anon_sym_import] = ACTIONS(1320),
    [anon_sym_as] = ACTIONS(1320),
    [anon_sym_package] = ACTIONS(1320),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_assert] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_LT_LT] = ACTIONS(1318),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_GT_GT_GT] = ACTIONS(1318),
    [anon_sym_DOT_DOT] = ACTIONS(1320),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_GT_EQ] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1320),
    [anon_sym_BANGin] = ACTIONS(1320),
    [anon_sym_instanceof] = ACTIONS(1320),
    [anon_sym_BANGinstanceof] = ACTIONS(1318),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT_EQ_GT] = ACTIONS(1318),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1318),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1318),
    [anon_sym_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_AMP_AMP] = ACTIONS(1318),
    [anon_sym_PIPE_PIPE] = ACTIONS(1318),
    [anon_sym_QMARK_COLON] = ACTIONS(1318),
    [anon_sym_STAR_STAR] = ACTIONS(1318),
    [anon_sym_true] = ACTIONS(1320),
    [anon_sym_false] = ACTIONS(1320),
    [anon_sym_ATinterface] = ACTIONS(1318),
    [anon_sym_interface] = ACTIONS(1320),
    [anon_sym_class] = ACTIONS(1320),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1320),
    [anon_sym_def] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_null] = ACTIONS(1320),
    [anon_sym_this] = ACTIONS(1320),
    [anon_sym_do] = ACTIONS(1320),
    [anon_sym_while] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1320),
    [aux_sym__type_identifier_token1] = ACTIONS(1320),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_else] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1318),
    [aux_sym_number_literal_token1] = ACTIONS(1320),
    [aux_sym_number_literal_token2] = ACTIONS(1318),
    [aux_sym_number_literal_token3] = ACTIONS(1318),
    [aux_sym_number_literal_token4] = ACTIONS(1320),
    [aux_sym_number_literal_token5] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1318),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1320),
    [anon_sym_case] = ACTIONS(1320),
    [anon_sym_default] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_try] = ACTIONS(1320),
    [anon_sym_catch] = ACTIONS(1320),
    [anon_sym_finally] = ACTIONS(1320),
    [anon_sym_int] = ACTIONS(1320),
    [anon_sym_boolean] = ACTIONS(1320),
    [anon_sym_char] = ACTIONS(1320),
    [anon_sym_short] = ACTIONS(1320),
    [anon_sym_long] = ACTIONS(1320),
    [anon_sym_float] = ACTIONS(1320),
    [anon_sym_double] = ACTIONS(1320),
    [anon_sym_void] = ACTIONS(1320),
    [anon_sym_public] = ACTIONS(1320),
    [anon_sym_protected] = ACTIONS(1320),
    [anon_sym_private] = ACTIONS(1320),
    [anon_sym_static] = ACTIONS(1320),
    [anon_sym_final] = ACTIONS(1320),
    [anon_sym_synchronized] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_new] = ACTIONS(1320),
  },
  [STATE(282)] = {
    [sym_closure] = STATE(375),
    [sym_comment] = STATE(282),
    [sym_groovy_doc] = STATE(282),
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_RPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_pipeline] = ACTIONS(1338),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(283)] = {
    [sym_access_op] = STATE(1435),
    [sym_dotted_identifier] = STATE(1389),
    [sym_increment_op] = STATE(1435),
    [sym_binary_op] = STATE(1435),
    [sym_boolean_literal] = STATE(1396),
    [sym_closure] = STATE(1435),
    [sym_comment] = STATE(283),
    [sym_groovy_doc] = STATE(283),
    [sym_parenthesized_expression] = STATE(1393),
    [sym__expression] = STATE(1546),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1388),
    [sym__type_identifier] = STATE(1385),
    [sym_index] = STATE(1389),
    [sym__juxt_argument_list] = STATE(895),
    [sym_list] = STATE(1396),
    [sym_map_item] = STATE(2024),
    [sym_map] = STATE(1396),
    [sym_number_literal] = STATE(1383),
    [sym_string] = STATE(1383),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1435),
    [sym_unary_op] = STATE(1435),
    [sym_identifier] = ACTIONS(1340),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1193),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_EQ] = ACTIONS(1196),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1198),
    [anon_sym_STAR_EQ] = ACTIONS(1198),
    [anon_sym_SLASH_EQ] = ACTIONS(1198),
    [anon_sym_PERCENT_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_EQ] = ACTIONS(1198),
    [anon_sym_DASH_EQ] = ACTIONS(1198),
    [anon_sym_LT_LT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_AMP_EQ] = ACTIONS(1198),
    [anon_sym_CARET_EQ] = ACTIONS(1198),
    [anon_sym_PIPE_EQ] = ACTIONS(1198),
    [anon_sym_QMARK_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(1200),
    [anon_sym_DASH_DASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(1203),
    [anon_sym_PLUS] = ACTIONS(1206),
    [anon_sym_DASH] = ACTIONS(1206),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(129),
    [anon_sym_null] = ACTIONS(1342),
    [anon_sym_this] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(1213),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(1215),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(284)] = {
    [sym_comment] = STATE(284),
    [sym_groovy_doc] = STATE(284),
    [sym_argument_list] = STATE(304),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1346),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_catch] = ACTIONS(1262),
    [anon_sym_finally] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(285)] = {
    [sym_comment] = STATE(285),
    [sym_groovy_doc] = STATE(285),
    [sym_identifier] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_DOT_AMP] = ACTIONS(1350),
    [anon_sym_DOT_AT] = ACTIONS(1350),
    [anon_sym_QMARK_DOT] = ACTIONS(1350),
    [anon_sym_STAR_DOT] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_STAR_COLON] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_import] = ACTIONS(1348),
    [anon_sym_as] = ACTIONS(1348),
    [anon_sym_package] = ACTIONS(1348),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_assert] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_LT_LT] = ACTIONS(1350),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_GT_GT_GT] = ACTIONS(1350),
    [anon_sym_DOT_DOT] = ACTIONS(1348),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_GT_EQ] = ACTIONS(1350),
    [anon_sym_in] = ACTIONS(1348),
    [anon_sym_BANGin] = ACTIONS(1348),
    [anon_sym_instanceof] = ACTIONS(1348),
    [anon_sym_BANGinstanceof] = ACTIONS(1350),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT_EQ_GT] = ACTIONS(1350),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1350),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1350),
    [anon_sym_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_AMP_AMP] = ACTIONS(1350),
    [anon_sym_PIPE_PIPE] = ACTIONS(1350),
    [anon_sym_QMARK_COLON] = ACTIONS(1350),
    [anon_sym_STAR_STAR] = ACTIONS(1350),
    [anon_sym_true] = ACTIONS(1348),
    [anon_sym_false] = ACTIONS(1348),
    [anon_sym_ATinterface] = ACTIONS(1350),
    [anon_sym_interface] = ACTIONS(1348),
    [anon_sym_class] = ACTIONS(1348),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1348),
    [anon_sym_def] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_null] = ACTIONS(1348),
    [anon_sym_this] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [aux_sym__type_identifier_token1] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_else] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1350),
    [aux_sym_number_literal_token1] = ACTIONS(1348),
    [aux_sym_number_literal_token2] = ACTIONS(1350),
    [aux_sym_number_literal_token3] = ACTIONS(1350),
    [aux_sym_number_literal_token4] = ACTIONS(1348),
    [aux_sym_number_literal_token5] = ACTIONS(1350),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1348),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1350),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1350),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_try] = ACTIONS(1348),
    [anon_sym_catch] = ACTIONS(1348),
    [anon_sym_finally] = ACTIONS(1348),
    [anon_sym_int] = ACTIONS(1348),
    [anon_sym_boolean] = ACTIONS(1348),
    [anon_sym_char] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_float] = ACTIONS(1348),
    [anon_sym_double] = ACTIONS(1348),
    [anon_sym_void] = ACTIONS(1348),
    [anon_sym_public] = ACTIONS(1348),
    [anon_sym_protected] = ACTIONS(1348),
    [anon_sym_private] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_final] = ACTIONS(1348),
    [anon_sym_synchronized] = ACTIONS(1348),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_new] = ACTIONS(1348),
  },
  [STATE(286)] = {
    [sym_comment] = STATE(286),
    [sym_groovy_doc] = STATE(286),
    [aux_sym_dotted_identifier_repeat1] = STATE(287),
    [ts_builtin_sym_end] = ACTIONS(373),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(373),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(373),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(373),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(373),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(373),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_COMMA] = ACTIONS(373),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_pipeline] = ACTIONS(375),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_catch] = ACTIONS(375),
    [anon_sym_finally] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(287)] = {
    [sym_comment] = STATE(287),
    [sym_groovy_doc] = STATE(287),
    [aux_sym_dotted_identifier_repeat1] = STATE(287),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(1352),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(288)] = {
    [sym_comment] = STATE(288),
    [sym_groovy_doc] = STATE(288),
    [ts_builtin_sym_end] = ACTIONS(1330),
    [sym_identifier] = ACTIONS(1328),
    [anon_sym_break] = ACTIONS(1328),
    [anon_sym_continue] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1330),
    [anon_sym_COLON] = ACTIONS(1330),
    [anon_sym_DOT_AMP] = ACTIONS(1330),
    [anon_sym_DOT_AT] = ACTIONS(1330),
    [anon_sym_QMARK_DOT] = ACTIONS(1330),
    [anon_sym_STAR_DOT] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_STAR_COLON] = ACTIONS(1330),
    [anon_sym_DOT] = ACTIONS(1328),
    [anon_sym_import] = ACTIONS(1328),
    [anon_sym_as] = ACTIONS(1328),
    [anon_sym_package] = ACTIONS(1328),
    [anon_sym_AT] = ACTIONS(1328),
    [anon_sym_assert] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1330),
    [anon_sym_PERCENT] = ACTIONS(1330),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_LT_LT] = ACTIONS(1330),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_GT_GT_GT] = ACTIONS(1330),
    [anon_sym_DOT_DOT] = ACTIONS(1328),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_GT_EQ] = ACTIONS(1330),
    [anon_sym_in] = ACTIONS(1328),
    [anon_sym_BANGin] = ACTIONS(1328),
    [anon_sym_instanceof] = ACTIONS(1328),
    [anon_sym_BANGinstanceof] = ACTIONS(1330),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT_EQ_GT] = ACTIONS(1330),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1330),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1330),
    [anon_sym_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1330),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_AMP_AMP] = ACTIONS(1330),
    [anon_sym_PIPE_PIPE] = ACTIONS(1330),
    [anon_sym_QMARK_COLON] = ACTIONS(1330),
    [anon_sym_STAR_STAR] = ACTIONS(1330),
    [anon_sym_true] = ACTIONS(1328),
    [anon_sym_false] = ACTIONS(1328),
    [anon_sym_ATinterface] = ACTIONS(1330),
    [anon_sym_interface] = ACTIONS(1328),
    [anon_sym_class] = ACTIONS(1328),
    [anon_sym_COMMA] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1330),
    [anon_sym_RBRACE] = ACTIONS(1330),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1328),
    [anon_sym_def] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1330),
    [anon_sym_null] = ACTIONS(1328),
    [anon_sym_this] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1328),
    [anon_sym_while] = ACTIONS(1328),
    [anon_sym_for] = ACTIONS(1328),
    [aux_sym__type_identifier_token1] = ACTIONS(1328),
    [anon_sym_if] = ACTIONS(1328),
    [anon_sym_else] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [aux_sym_number_literal_token1] = ACTIONS(1328),
    [aux_sym_number_literal_token2] = ACTIONS(1330),
    [aux_sym_number_literal_token3] = ACTIONS(1330),
    [aux_sym_number_literal_token4] = ACTIONS(1328),
    [aux_sym_number_literal_token5] = ACTIONS(1330),
    [anon_sym_pipeline] = ACTIONS(1328),
    [anon_sym_return] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1330),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1330),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1328),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_try] = ACTIONS(1328),
    [anon_sym_catch] = ACTIONS(1328),
    [anon_sym_finally] = ACTIONS(1328),
    [anon_sym_int] = ACTIONS(1328),
    [anon_sym_boolean] = ACTIONS(1328),
    [anon_sym_char] = ACTIONS(1328),
    [anon_sym_short] = ACTIONS(1328),
    [anon_sym_long] = ACTIONS(1328),
    [anon_sym_float] = ACTIONS(1328),
    [anon_sym_double] = ACTIONS(1328),
    [anon_sym_void] = ACTIONS(1328),
    [anon_sym_public] = ACTIONS(1328),
    [anon_sym_protected] = ACTIONS(1328),
    [anon_sym_private] = ACTIONS(1328),
    [anon_sym_static] = ACTIONS(1328),
    [anon_sym_final] = ACTIONS(1328),
    [anon_sym_synchronized] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1330),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_new] = ACTIONS(1328),
  },
  [STATE(289)] = {
    [sym_comment] = STATE(289),
    [sym_groovy_doc] = STATE(289),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1264),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(290)] = {
    [sym_closure] = STATE(359),
    [sym_comment] = STATE(290),
    [sym_groovy_doc] = STATE(290),
    [sym_identifier] = ACTIONS(1324),
    [anon_sym_break] = ACTIONS(1324),
    [anon_sym_continue] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1322),
    [anon_sym_DOT_AMP] = ACTIONS(1322),
    [anon_sym_DOT_AT] = ACTIONS(1322),
    [anon_sym_QMARK_DOT] = ACTIONS(1322),
    [anon_sym_STAR_DOT] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_STAR_COLON] = ACTIONS(1322),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_import] = ACTIONS(1324),
    [anon_sym_as] = ACTIONS(1324),
    [anon_sym_package] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_assert] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1322),
    [anon_sym_PERCENT] = ACTIONS(1322),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_LT_LT] = ACTIONS(1322),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_GT_GT_GT] = ACTIONS(1322),
    [anon_sym_DOT_DOT] = ACTIONS(1324),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_GT_EQ] = ACTIONS(1322),
    [anon_sym_in] = ACTIONS(1324),
    [anon_sym_BANGin] = ACTIONS(1324),
    [anon_sym_instanceof] = ACTIONS(1324),
    [anon_sym_BANGinstanceof] = ACTIONS(1322),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT_EQ_GT] = ACTIONS(1322),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1322),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1322),
    [anon_sym_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1322),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_AMP_AMP] = ACTIONS(1322),
    [anon_sym_PIPE_PIPE] = ACTIONS(1322),
    [anon_sym_QMARK_COLON] = ACTIONS(1322),
    [anon_sym_STAR_STAR] = ACTIONS(1322),
    [anon_sym_true] = ACTIONS(1324),
    [anon_sym_false] = ACTIONS(1324),
    [anon_sym_ATinterface] = ACTIONS(1322),
    [anon_sym_interface] = ACTIONS(1324),
    [anon_sym_class] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [anon_sym_RBRACE] = ACTIONS(1322),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_def] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_this] = ACTIONS(1324),
    [anon_sym_do] = ACTIONS(1324),
    [anon_sym_while] = ACTIONS(1324),
    [anon_sym_for] = ACTIONS(1324),
    [aux_sym__type_identifier_token1] = ACTIONS(1324),
    [anon_sym_if] = ACTIONS(1324),
    [anon_sym_else] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1322),
    [aux_sym_number_literal_token1] = ACTIONS(1324),
    [aux_sym_number_literal_token2] = ACTIONS(1322),
    [aux_sym_number_literal_token3] = ACTIONS(1322),
    [aux_sym_number_literal_token4] = ACTIONS(1324),
    [aux_sym_number_literal_token5] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1322),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1322),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1324),
    [anon_sym_case] = ACTIONS(1324),
    [anon_sym_default] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_try] = ACTIONS(1324),
    [anon_sym_catch] = ACTIONS(1324),
    [anon_sym_finally] = ACTIONS(1324),
    [anon_sym_int] = ACTIONS(1324),
    [anon_sym_boolean] = ACTIONS(1324),
    [anon_sym_char] = ACTIONS(1324),
    [anon_sym_short] = ACTIONS(1324),
    [anon_sym_long] = ACTIONS(1324),
    [anon_sym_float] = ACTIONS(1324),
    [anon_sym_double] = ACTIONS(1324),
    [anon_sym_void] = ACTIONS(1324),
    [anon_sym_public] = ACTIONS(1324),
    [anon_sym_protected] = ACTIONS(1324),
    [anon_sym_private] = ACTIONS(1324),
    [anon_sym_static] = ACTIONS(1324),
    [anon_sym_final] = ACTIONS(1324),
    [anon_sym_synchronized] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1322),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_new] = ACTIONS(1324),
  },
  [STATE(291)] = {
    [sym_comment] = STATE(291),
    [sym_groovy_doc] = STATE(291),
    [aux_sym_dotted_identifier_repeat1] = STATE(291),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(1355),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(292)] = {
    [sym_comment] = STATE(292),
    [sym_groovy_doc] = STATE(292),
    [aux_sym_dotted_identifier_repeat1] = STATE(269),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1292),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(293)] = {
    [sym_closure] = STATE(365),
    [sym_comment] = STATE(293),
    [sym_groovy_doc] = STATE(293),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_case] = ACTIONS(1338),
    [anon_sym_default] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_catch] = ACTIONS(1338),
    [anon_sym_finally] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(294)] = {
    [sym_comment] = STATE(294),
    [sym_groovy_doc] = STATE(294),
    [sym_argument_list] = STATE(364),
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1358),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_pipeline] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_catch] = ACTIONS(1262),
    [anon_sym_finally] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(295)] = {
    [sym_closure] = STATE(366),
    [sym_comment] = STATE(295),
    [sym_groovy_doc] = STATE(295),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_catch] = ACTIONS(1306),
    [anon_sym_finally] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(296)] = {
    [sym_comment] = STATE(296),
    [sym_groovy_doc] = STATE(296),
    [aux_sym_dotted_identifier_repeat1] = STATE(9),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1360),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(102),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1290),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(297)] = {
    [sym_closure] = STATE(369),
    [sym_comment] = STATE(297),
    [sym_groovy_doc] = STATE(297),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_catch] = ACTIONS(1334),
    [anon_sym_finally] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(298)] = {
    [sym_comment] = STATE(298),
    [sym_groovy_doc] = STATE(298),
    [ts_builtin_sym_end] = ACTIONS(1350),
    [sym_identifier] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_DOT_AMP] = ACTIONS(1350),
    [anon_sym_DOT_AT] = ACTIONS(1350),
    [anon_sym_QMARK_DOT] = ACTIONS(1350),
    [anon_sym_STAR_DOT] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_STAR_COLON] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_import] = ACTIONS(1348),
    [anon_sym_as] = ACTIONS(1348),
    [anon_sym_package] = ACTIONS(1348),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_assert] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_LT_LT] = ACTIONS(1350),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_GT_GT_GT] = ACTIONS(1350),
    [anon_sym_DOT_DOT] = ACTIONS(1348),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_GT_EQ] = ACTIONS(1350),
    [anon_sym_in] = ACTIONS(1348),
    [anon_sym_BANGin] = ACTIONS(1348),
    [anon_sym_instanceof] = ACTIONS(1348),
    [anon_sym_BANGinstanceof] = ACTIONS(1350),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT_EQ_GT] = ACTIONS(1350),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1350),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1350),
    [anon_sym_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_AMP_AMP] = ACTIONS(1350),
    [anon_sym_PIPE_PIPE] = ACTIONS(1350),
    [anon_sym_QMARK_COLON] = ACTIONS(1350),
    [anon_sym_STAR_STAR] = ACTIONS(1350),
    [anon_sym_true] = ACTIONS(1348),
    [anon_sym_false] = ACTIONS(1348),
    [anon_sym_ATinterface] = ACTIONS(1350),
    [anon_sym_interface] = ACTIONS(1348),
    [anon_sym_class] = ACTIONS(1348),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1348),
    [anon_sym_def] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_null] = ACTIONS(1348),
    [anon_sym_this] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [aux_sym__type_identifier_token1] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_else] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1350),
    [aux_sym_number_literal_token1] = ACTIONS(1348),
    [aux_sym_number_literal_token2] = ACTIONS(1350),
    [aux_sym_number_literal_token3] = ACTIONS(1350),
    [aux_sym_number_literal_token4] = ACTIONS(1348),
    [aux_sym_number_literal_token5] = ACTIONS(1350),
    [anon_sym_pipeline] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1348),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1350),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1350),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_try] = ACTIONS(1348),
    [anon_sym_catch] = ACTIONS(1348),
    [anon_sym_finally] = ACTIONS(1348),
    [anon_sym_int] = ACTIONS(1348),
    [anon_sym_boolean] = ACTIONS(1348),
    [anon_sym_char] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_float] = ACTIONS(1348),
    [anon_sym_double] = ACTIONS(1348),
    [anon_sym_void] = ACTIONS(1348),
    [anon_sym_public] = ACTIONS(1348),
    [anon_sym_protected] = ACTIONS(1348),
    [anon_sym_private] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_final] = ACTIONS(1348),
    [anon_sym_synchronized] = ACTIONS(1348),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_new] = ACTIONS(1348),
  },
  [STATE(299)] = {
    [sym_closure] = STATE(377),
    [sym_comment] = STATE(299),
    [sym_groovy_doc] = STATE(299),
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_pipeline] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_catch] = ACTIONS(1334),
    [anon_sym_finally] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(300)] = {
    [sym_closure] = STATE(348),
    [sym_comment] = STATE(300),
    [sym_groovy_doc] = STATE(300),
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_DASH_GT] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_RPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_pipeline] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(301)] = {
    [sym_comment] = STATE(301),
    [sym_groovy_doc] = STATE(301),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1264),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(302)] = {
    [sym_comment] = STATE(302),
    [sym_groovy_doc] = STATE(302),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_DOT_AMP] = ACTIONS(1364),
    [anon_sym_DOT_AT] = ACTIONS(1364),
    [anon_sym_QMARK_DOT] = ACTIONS(1364),
    [anon_sym_STAR_DOT] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_as] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_PERCENT] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_LT_LT] = ACTIONS(1364),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_GT_GT_GT] = ACTIONS(1364),
    [anon_sym_DOT_DOT] = ACTIONS(1362),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_GT_EQ] = ACTIONS(1364),
    [anon_sym_in] = ACTIONS(1362),
    [anon_sym_BANGin] = ACTIONS(1362),
    [anon_sym_instanceof] = ACTIONS(1362),
    [anon_sym_BANGinstanceof] = ACTIONS(1364),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT_EQ_GT] = ACTIONS(1364),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1364),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1364),
    [anon_sym_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_AMP_AMP] = ACTIONS(1364),
    [anon_sym_PIPE_PIPE] = ACTIONS(1364),
    [anon_sym_QMARK_COLON] = ACTIONS(1364),
    [anon_sym_STAR_STAR] = ACTIONS(1364),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_case] = ACTIONS(1362),
    [anon_sym_default] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_catch] = ACTIONS(1362),
    [anon_sym_finally] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(303)] = {
    [sym_comment] = STATE(303),
    [sym_groovy_doc] = STATE(303),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_catch] = ACTIONS(1262),
    [anon_sym_finally] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(304)] = {
    [sym_comment] = STATE(304),
    [sym_groovy_doc] = STATE(304),
    [sym_identifier] = ACTIONS(1366),
    [anon_sym_break] = ACTIONS(1366),
    [anon_sym_continue] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1368),
    [anon_sym_DOT_AMP] = ACTIONS(1368),
    [anon_sym_DOT_AT] = ACTIONS(1368),
    [anon_sym_QMARK_DOT] = ACTIONS(1368),
    [anon_sym_STAR_DOT] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_STAR_COLON] = ACTIONS(1368),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_import] = ACTIONS(1366),
    [anon_sym_as] = ACTIONS(1366),
    [anon_sym_package] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_assert] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1368),
    [anon_sym_PERCENT] = ACTIONS(1368),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_LT_LT] = ACTIONS(1368),
    [anon_sym_GT_GT] = ACTIONS(1366),
    [anon_sym_GT_GT_GT] = ACTIONS(1368),
    [anon_sym_DOT_DOT] = ACTIONS(1366),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_LT_EQ] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_GT_EQ] = ACTIONS(1368),
    [anon_sym_in] = ACTIONS(1366),
    [anon_sym_BANGin] = ACTIONS(1366),
    [anon_sym_instanceof] = ACTIONS(1366),
    [anon_sym_BANGinstanceof] = ACTIONS(1368),
    [anon_sym_EQ_EQ] = ACTIONS(1366),
    [anon_sym_BANG_EQ] = ACTIONS(1366),
    [anon_sym_LT_EQ_GT] = ACTIONS(1368),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1368),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1368),
    [anon_sym_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1368),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_AMP_AMP] = ACTIONS(1368),
    [anon_sym_PIPE_PIPE] = ACTIONS(1368),
    [anon_sym_QMARK_COLON] = ACTIONS(1368),
    [anon_sym_STAR_STAR] = ACTIONS(1368),
    [anon_sym_true] = ACTIONS(1366),
    [anon_sym_false] = ACTIONS(1366),
    [anon_sym_ATinterface] = ACTIONS(1368),
    [anon_sym_interface] = ACTIONS(1366),
    [anon_sym_class] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_def] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1368),
    [anon_sym_null] = ACTIONS(1366),
    [anon_sym_this] = ACTIONS(1366),
    [anon_sym_do] = ACTIONS(1366),
    [anon_sym_while] = ACTIONS(1366),
    [anon_sym_for] = ACTIONS(1366),
    [aux_sym__type_identifier_token1] = ACTIONS(1366),
    [anon_sym_if] = ACTIONS(1366),
    [anon_sym_else] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [aux_sym_number_literal_token1] = ACTIONS(1366),
    [aux_sym_number_literal_token2] = ACTIONS(1368),
    [aux_sym_number_literal_token3] = ACTIONS(1368),
    [aux_sym_number_literal_token4] = ACTIONS(1366),
    [aux_sym_number_literal_token5] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1368),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1368),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1366),
    [anon_sym_case] = ACTIONS(1366),
    [anon_sym_default] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_catch] = ACTIONS(1366),
    [anon_sym_finally] = ACTIONS(1366),
    [anon_sym_int] = ACTIONS(1366),
    [anon_sym_boolean] = ACTIONS(1366),
    [anon_sym_char] = ACTIONS(1366),
    [anon_sym_short] = ACTIONS(1366),
    [anon_sym_long] = ACTIONS(1366),
    [anon_sym_float] = ACTIONS(1366),
    [anon_sym_double] = ACTIONS(1366),
    [anon_sym_void] = ACTIONS(1366),
    [anon_sym_public] = ACTIONS(1366),
    [anon_sym_protected] = ACTIONS(1366),
    [anon_sym_private] = ACTIONS(1366),
    [anon_sym_static] = ACTIONS(1366),
    [anon_sym_final] = ACTIONS(1366),
    [anon_sym_synchronized] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_new] = ACTIONS(1366),
  },
  [STATE(305)] = {
    [sym_comment] = STATE(305),
    [sym_groovy_doc] = STATE(305),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1370),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(306)] = {
    [sym_comment] = STATE(306),
    [sym_groovy_doc] = STATE(306),
    [ts_builtin_sym_end] = ACTIONS(1374),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_DOT_AMP] = ACTIONS(1374),
    [anon_sym_DOT_AT] = ACTIONS(1374),
    [anon_sym_QMARK_DOT] = ACTIONS(1374),
    [anon_sym_STAR_DOT] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1376),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_as] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_LT_LT] = ACTIONS(1374),
    [anon_sym_GT_GT] = ACTIONS(1376),
    [anon_sym_GT_GT_GT] = ACTIONS(1374),
    [anon_sym_DOT_DOT] = ACTIONS(1376),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1376),
    [anon_sym_LT] = ACTIONS(1376),
    [anon_sym_LT_EQ] = ACTIONS(1376),
    [anon_sym_GT] = ACTIONS(1376),
    [anon_sym_GT_EQ] = ACTIONS(1374),
    [anon_sym_in] = ACTIONS(1376),
    [anon_sym_BANGin] = ACTIONS(1376),
    [anon_sym_instanceof] = ACTIONS(1376),
    [anon_sym_BANGinstanceof] = ACTIONS(1374),
    [anon_sym_EQ_EQ] = ACTIONS(1376),
    [anon_sym_BANG_EQ] = ACTIONS(1376),
    [anon_sym_LT_EQ_GT] = ACTIONS(1374),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1374),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1374),
    [anon_sym_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1376),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1376),
    [anon_sym_AMP_AMP] = ACTIONS(1374),
    [anon_sym_PIPE_PIPE] = ACTIONS(1374),
    [anon_sym_QMARK_COLON] = ACTIONS(1374),
    [anon_sym_STAR_STAR] = ACTIONS(1374),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_pipeline] = ACTIONS(1376),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_QMARK] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_catch] = ACTIONS(1376),
    [anon_sym_finally] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(307)] = {
    [sym_comment] = STATE(307),
    [sym_groovy_doc] = STATE(307),
    [ts_builtin_sym_end] = ACTIONS(1378),
    [sym_identifier] = ACTIONS(1380),
    [anon_sym_break] = ACTIONS(1380),
    [anon_sym_continue] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1378),
    [anon_sym_DOT_AMP] = ACTIONS(1378),
    [anon_sym_DOT_AT] = ACTIONS(1378),
    [anon_sym_QMARK_DOT] = ACTIONS(1378),
    [anon_sym_STAR_DOT] = ACTIONS(1378),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_STAR_COLON] = ACTIONS(1378),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_import] = ACTIONS(1380),
    [anon_sym_as] = ACTIONS(1380),
    [anon_sym_package] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_assert] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1378),
    [anon_sym_DASH_DASH] = ACTIONS(1378),
    [anon_sym_PERCENT] = ACTIONS(1378),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_LT_LT] = ACTIONS(1378),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym_GT_GT_GT] = ACTIONS(1378),
    [anon_sym_DOT_DOT] = ACTIONS(1380),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_GT_EQ] = ACTIONS(1378),
    [anon_sym_in] = ACTIONS(1380),
    [anon_sym_BANGin] = ACTIONS(1380),
    [anon_sym_instanceof] = ACTIONS(1380),
    [anon_sym_BANGinstanceof] = ACTIONS(1378),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ_GT] = ACTIONS(1378),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1378),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1378),
    [anon_sym_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1378),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1378),
    [anon_sym_PIPE_PIPE] = ACTIONS(1378),
    [anon_sym_QMARK_COLON] = ACTIONS(1378),
    [anon_sym_STAR_STAR] = ACTIONS(1378),
    [anon_sym_true] = ACTIONS(1380),
    [anon_sym_false] = ACTIONS(1380),
    [anon_sym_ATinterface] = ACTIONS(1378),
    [anon_sym_interface] = ACTIONS(1380),
    [anon_sym_class] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1378),
    [anon_sym_LBRACE] = ACTIONS(1378),
    [anon_sym_RBRACE] = ACTIONS(1378),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_def] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1378),
    [anon_sym_null] = ACTIONS(1380),
    [anon_sym_this] = ACTIONS(1380),
    [anon_sym_do] = ACTIONS(1380),
    [anon_sym_while] = ACTIONS(1380),
    [anon_sym_for] = ACTIONS(1380),
    [aux_sym__type_identifier_token1] = ACTIONS(1380),
    [anon_sym_if] = ACTIONS(1380),
    [anon_sym_else] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [aux_sym_number_literal_token1] = ACTIONS(1380),
    [aux_sym_number_literal_token2] = ACTIONS(1378),
    [aux_sym_number_literal_token3] = ACTIONS(1378),
    [aux_sym_number_literal_token4] = ACTIONS(1380),
    [aux_sym_number_literal_token5] = ACTIONS(1378),
    [anon_sym_pipeline] = ACTIONS(1380),
    [anon_sym_return] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1378),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1378),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1378),
    [anon_sym_switch] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_try] = ACTIONS(1380),
    [anon_sym_catch] = ACTIONS(1380),
    [anon_sym_finally] = ACTIONS(1380),
    [anon_sym_int] = ACTIONS(1380),
    [anon_sym_boolean] = ACTIONS(1380),
    [anon_sym_char] = ACTIONS(1380),
    [anon_sym_short] = ACTIONS(1380),
    [anon_sym_long] = ACTIONS(1380),
    [anon_sym_float] = ACTIONS(1380),
    [anon_sym_double] = ACTIONS(1380),
    [anon_sym_void] = ACTIONS(1380),
    [anon_sym_public] = ACTIONS(1380),
    [anon_sym_protected] = ACTIONS(1380),
    [anon_sym_private] = ACTIONS(1380),
    [anon_sym_static] = ACTIONS(1380),
    [anon_sym_final] = ACTIONS(1380),
    [anon_sym_synchronized] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1378),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_new] = ACTIONS(1380),
  },
  [STATE(308)] = {
    [sym_comment] = STATE(308),
    [sym_groovy_doc] = STATE(308),
    [ts_builtin_sym_end] = ACTIONS(1382),
    [sym_identifier] = ACTIONS(1384),
    [anon_sym_break] = ACTIONS(1384),
    [anon_sym_continue] = ACTIONS(1384),
    [anon_sym_SEMI] = ACTIONS(1382),
    [anon_sym_DOT_AMP] = ACTIONS(1382),
    [anon_sym_DOT_AT] = ACTIONS(1382),
    [anon_sym_QMARK_DOT] = ACTIONS(1382),
    [anon_sym_STAR_DOT] = ACTIONS(1382),
    [anon_sym_STAR] = ACTIONS(1384),
    [anon_sym_STAR_COLON] = ACTIONS(1382),
    [anon_sym_DOT] = ACTIONS(1384),
    [anon_sym_import] = ACTIONS(1384),
    [anon_sym_as] = ACTIONS(1384),
    [anon_sym_package] = ACTIONS(1384),
    [anon_sym_AT] = ACTIONS(1384),
    [anon_sym_assert] = ACTIONS(1384),
    [anon_sym_PLUS_PLUS] = ACTIONS(1382),
    [anon_sym_DASH_DASH] = ACTIONS(1382),
    [anon_sym_PERCENT] = ACTIONS(1382),
    [anon_sym_SLASH] = ACTIONS(1384),
    [anon_sym_PLUS] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_LT_LT] = ACTIONS(1382),
    [anon_sym_GT_GT] = ACTIONS(1384),
    [anon_sym_GT_GT_GT] = ACTIONS(1382),
    [anon_sym_DOT_DOT] = ACTIONS(1384),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1384),
    [anon_sym_LT] = ACTIONS(1384),
    [anon_sym_LT_EQ] = ACTIONS(1384),
    [anon_sym_GT] = ACTIONS(1384),
    [anon_sym_GT_EQ] = ACTIONS(1382),
    [anon_sym_in] = ACTIONS(1384),
    [anon_sym_BANGin] = ACTIONS(1384),
    [anon_sym_instanceof] = ACTIONS(1384),
    [anon_sym_BANGinstanceof] = ACTIONS(1382),
    [anon_sym_EQ_EQ] = ACTIONS(1384),
    [anon_sym_BANG_EQ] = ACTIONS(1384),
    [anon_sym_LT_EQ_GT] = ACTIONS(1382),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1382),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1382),
    [anon_sym_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_AMP] = ACTIONS(1384),
    [anon_sym_CARET] = ACTIONS(1382),
    [anon_sym_PIPE] = ACTIONS(1384),
    [anon_sym_AMP_AMP] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1382),
    [anon_sym_QMARK_COLON] = ACTIONS(1382),
    [anon_sym_STAR_STAR] = ACTIONS(1382),
    [anon_sym_true] = ACTIONS(1384),
    [anon_sym_false] = ACTIONS(1384),
    [anon_sym_ATinterface] = ACTIONS(1382),
    [anon_sym_interface] = ACTIONS(1384),
    [anon_sym_class] = ACTIONS(1384),
    [anon_sym_COMMA] = ACTIONS(1382),
    [anon_sym_LBRACE] = ACTIONS(1382),
    [anon_sym_RBRACE] = ACTIONS(1382),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1384),
    [anon_sym_def] = ACTIONS(1384),
    [anon_sym_LPAREN] = ACTIONS(1382),
    [anon_sym_null] = ACTIONS(1384),
    [anon_sym_this] = ACTIONS(1384),
    [anon_sym_do] = ACTIONS(1384),
    [anon_sym_while] = ACTIONS(1384),
    [anon_sym_for] = ACTIONS(1384),
    [aux_sym__type_identifier_token1] = ACTIONS(1384),
    [anon_sym_if] = ACTIONS(1384),
    [anon_sym_else] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1382),
    [aux_sym_number_literal_token1] = ACTIONS(1384),
    [aux_sym_number_literal_token2] = ACTIONS(1382),
    [aux_sym_number_literal_token3] = ACTIONS(1382),
    [aux_sym_number_literal_token4] = ACTIONS(1384),
    [aux_sym_number_literal_token5] = ACTIONS(1382),
    [anon_sym_pipeline] = ACTIONS(1384),
    [anon_sym_return] = ACTIONS(1384),
    [anon_sym_SQUOTE] = ACTIONS(1384),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1382),
    [anon_sym_DQUOTE] = ACTIONS(1384),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1382),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1382),
    [anon_sym_switch] = ACTIONS(1384),
    [anon_sym_QMARK] = ACTIONS(1384),
    [anon_sym_try] = ACTIONS(1384),
    [anon_sym_catch] = ACTIONS(1384),
    [anon_sym_finally] = ACTIONS(1384),
    [anon_sym_int] = ACTIONS(1384),
    [anon_sym_boolean] = ACTIONS(1384),
    [anon_sym_char] = ACTIONS(1384),
    [anon_sym_short] = ACTIONS(1384),
    [anon_sym_long] = ACTIONS(1384),
    [anon_sym_float] = ACTIONS(1384),
    [anon_sym_double] = ACTIONS(1384),
    [anon_sym_void] = ACTIONS(1384),
    [anon_sym_public] = ACTIONS(1384),
    [anon_sym_protected] = ACTIONS(1384),
    [anon_sym_private] = ACTIONS(1384),
    [anon_sym_static] = ACTIONS(1384),
    [anon_sym_final] = ACTIONS(1384),
    [anon_sym_synchronized] = ACTIONS(1384),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_new] = ACTIONS(1384),
  },
  [STATE(309)] = {
    [sym_comment] = STATE(309),
    [sym_groovy_doc] = STATE(309),
    [ts_builtin_sym_end] = ACTIONS(1386),
    [sym_identifier] = ACTIONS(1388),
    [anon_sym_break] = ACTIONS(1388),
    [anon_sym_continue] = ACTIONS(1388),
    [anon_sym_SEMI] = ACTIONS(1386),
    [anon_sym_DOT_AMP] = ACTIONS(1386),
    [anon_sym_DOT_AT] = ACTIONS(1386),
    [anon_sym_QMARK_DOT] = ACTIONS(1386),
    [anon_sym_STAR_DOT] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_STAR_COLON] = ACTIONS(1386),
    [anon_sym_DOT] = ACTIONS(1388),
    [anon_sym_import] = ACTIONS(1388),
    [anon_sym_as] = ACTIONS(1388),
    [anon_sym_package] = ACTIONS(1388),
    [anon_sym_AT] = ACTIONS(1388),
    [anon_sym_assert] = ACTIONS(1388),
    [anon_sym_PLUS_PLUS] = ACTIONS(1386),
    [anon_sym_DASH_DASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1388),
    [anon_sym_PLUS] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1388),
    [anon_sym_GT_GT_GT] = ACTIONS(1386),
    [anon_sym_DOT_DOT] = ACTIONS(1388),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1388),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_LT_EQ] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_EQ] = ACTIONS(1386),
    [anon_sym_in] = ACTIONS(1388),
    [anon_sym_BANGin] = ACTIONS(1388),
    [anon_sym_instanceof] = ACTIONS(1388),
    [anon_sym_BANGinstanceof] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1388),
    [anon_sym_BANG_EQ] = ACTIONS(1388),
    [anon_sym_LT_EQ_GT] = ACTIONS(1386),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1386),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1386),
    [anon_sym_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_PIPE] = ACTIONS(1388),
    [anon_sym_AMP_AMP] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1386),
    [anon_sym_QMARK_COLON] = ACTIONS(1386),
    [anon_sym_STAR_STAR] = ACTIONS(1386),
    [anon_sym_true] = ACTIONS(1388),
    [anon_sym_false] = ACTIONS(1388),
    [anon_sym_ATinterface] = ACTIONS(1386),
    [anon_sym_interface] = ACTIONS(1388),
    [anon_sym_class] = ACTIONS(1388),
    [anon_sym_COMMA] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1386),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1388),
    [anon_sym_def] = ACTIONS(1388),
    [anon_sym_LPAREN] = ACTIONS(1386),
    [anon_sym_null] = ACTIONS(1388),
    [anon_sym_this] = ACTIONS(1388),
    [anon_sym_do] = ACTIONS(1388),
    [anon_sym_while] = ACTIONS(1388),
    [anon_sym_for] = ACTIONS(1388),
    [aux_sym__type_identifier_token1] = ACTIONS(1388),
    [anon_sym_if] = ACTIONS(1388),
    [anon_sym_else] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1386),
    [aux_sym_number_literal_token1] = ACTIONS(1388),
    [aux_sym_number_literal_token2] = ACTIONS(1386),
    [aux_sym_number_literal_token3] = ACTIONS(1386),
    [aux_sym_number_literal_token4] = ACTIONS(1388),
    [aux_sym_number_literal_token5] = ACTIONS(1386),
    [anon_sym_pipeline] = ACTIONS(1388),
    [anon_sym_return] = ACTIONS(1388),
    [anon_sym_SQUOTE] = ACTIONS(1388),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1386),
    [anon_sym_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1386),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1386),
    [anon_sym_switch] = ACTIONS(1388),
    [anon_sym_QMARK] = ACTIONS(1388),
    [anon_sym_try] = ACTIONS(1388),
    [anon_sym_catch] = ACTIONS(1388),
    [anon_sym_finally] = ACTIONS(1388),
    [anon_sym_int] = ACTIONS(1388),
    [anon_sym_boolean] = ACTIONS(1388),
    [anon_sym_char] = ACTIONS(1388),
    [anon_sym_short] = ACTIONS(1388),
    [anon_sym_long] = ACTIONS(1388),
    [anon_sym_float] = ACTIONS(1388),
    [anon_sym_double] = ACTIONS(1388),
    [anon_sym_void] = ACTIONS(1388),
    [anon_sym_public] = ACTIONS(1388),
    [anon_sym_protected] = ACTIONS(1388),
    [anon_sym_private] = ACTIONS(1388),
    [anon_sym_static] = ACTIONS(1388),
    [anon_sym_final] = ACTIONS(1388),
    [anon_sym_synchronized] = ACTIONS(1388),
    [anon_sym_TILDE] = ACTIONS(1386),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_new] = ACTIONS(1388),
  },
  [STATE(310)] = {
    [sym_comment] = STATE(310),
    [sym_groovy_doc] = STATE(310),
    [ts_builtin_sym_end] = ACTIONS(1390),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_DOT_AMP] = ACTIONS(1390),
    [anon_sym_DOT_AT] = ACTIONS(1390),
    [anon_sym_QMARK_DOT] = ACTIONS(1390),
    [anon_sym_STAR_DOT] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_DOT] = ACTIONS(1392),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_as] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_PERCENT] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT] = ACTIONS(1390),
    [anon_sym_GT_GT] = ACTIONS(1392),
    [anon_sym_GT_GT_GT] = ACTIONS(1390),
    [anon_sym_DOT_DOT] = ACTIONS(1392),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1392),
    [anon_sym_LT] = ACTIONS(1392),
    [anon_sym_LT_EQ] = ACTIONS(1392),
    [anon_sym_GT] = ACTIONS(1392),
    [anon_sym_GT_EQ] = ACTIONS(1390),
    [anon_sym_in] = ACTIONS(1392),
    [anon_sym_BANGin] = ACTIONS(1392),
    [anon_sym_instanceof] = ACTIONS(1392),
    [anon_sym_BANGinstanceof] = ACTIONS(1390),
    [anon_sym_EQ_EQ] = ACTIONS(1392),
    [anon_sym_BANG_EQ] = ACTIONS(1392),
    [anon_sym_LT_EQ_GT] = ACTIONS(1390),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1390),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1390),
    [anon_sym_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_AMP] = ACTIONS(1392),
    [anon_sym_CARET] = ACTIONS(1390),
    [anon_sym_PIPE] = ACTIONS(1392),
    [anon_sym_AMP_AMP] = ACTIONS(1390),
    [anon_sym_PIPE_PIPE] = ACTIONS(1390),
    [anon_sym_QMARK_COLON] = ACTIONS(1390),
    [anon_sym_STAR_STAR] = ACTIONS(1390),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_COMMA] = ACTIONS(1390),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_pipeline] = ACTIONS(1392),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_QMARK] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_catch] = ACTIONS(1392),
    [anon_sym_finally] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(311)] = {
    [sym_comment] = STATE(311),
    [sym_groovy_doc] = STATE(311),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(312)] = {
    [sym_comment] = STATE(312),
    [sym_groovy_doc] = STATE(312),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(313)] = {
    [sym_comment] = STATE(313),
    [sym_groovy_doc] = STATE(313),
    [ts_builtin_sym_end] = ACTIONS(1394),
    [sym_identifier] = ACTIONS(1396),
    [anon_sym_break] = ACTIONS(1396),
    [anon_sym_continue] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1394),
    [anon_sym_DOT_AMP] = ACTIONS(1394),
    [anon_sym_DOT_AT] = ACTIONS(1394),
    [anon_sym_QMARK_DOT] = ACTIONS(1394),
    [anon_sym_STAR_DOT] = ACTIONS(1394),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_STAR_COLON] = ACTIONS(1394),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_import] = ACTIONS(1396),
    [anon_sym_as] = ACTIONS(1396),
    [anon_sym_package] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_assert] = ACTIONS(1396),
    [anon_sym_PLUS_PLUS] = ACTIONS(1394),
    [anon_sym_DASH_DASH] = ACTIONS(1394),
    [anon_sym_PERCENT] = ACTIONS(1394),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_LT_LT] = ACTIONS(1394),
    [anon_sym_GT_GT] = ACTIONS(1396),
    [anon_sym_GT_GT_GT] = ACTIONS(1394),
    [anon_sym_DOT_DOT] = ACTIONS(1396),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_LT_EQ] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_GT_EQ] = ACTIONS(1394),
    [anon_sym_in] = ACTIONS(1396),
    [anon_sym_BANGin] = ACTIONS(1396),
    [anon_sym_instanceof] = ACTIONS(1396),
    [anon_sym_BANGinstanceof] = ACTIONS(1394),
    [anon_sym_EQ_EQ] = ACTIONS(1396),
    [anon_sym_BANG_EQ] = ACTIONS(1396),
    [anon_sym_LT_EQ_GT] = ACTIONS(1394),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1394),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1394),
    [anon_sym_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1394),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_AMP_AMP] = ACTIONS(1394),
    [anon_sym_PIPE_PIPE] = ACTIONS(1394),
    [anon_sym_QMARK_COLON] = ACTIONS(1394),
    [anon_sym_STAR_STAR] = ACTIONS(1394),
    [anon_sym_true] = ACTIONS(1396),
    [anon_sym_false] = ACTIONS(1396),
    [anon_sym_ATinterface] = ACTIONS(1394),
    [anon_sym_interface] = ACTIONS(1396),
    [anon_sym_class] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(1394),
    [anon_sym_RBRACE] = ACTIONS(1394),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_def] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1394),
    [anon_sym_null] = ACTIONS(1396),
    [anon_sym_this] = ACTIONS(1396),
    [anon_sym_do] = ACTIONS(1396),
    [anon_sym_while] = ACTIONS(1396),
    [anon_sym_for] = ACTIONS(1396),
    [aux_sym__type_identifier_token1] = ACTIONS(1396),
    [anon_sym_if] = ACTIONS(1396),
    [anon_sym_else] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [aux_sym_number_literal_token1] = ACTIONS(1396),
    [aux_sym_number_literal_token2] = ACTIONS(1394),
    [aux_sym_number_literal_token3] = ACTIONS(1394),
    [aux_sym_number_literal_token4] = ACTIONS(1396),
    [aux_sym_number_literal_token5] = ACTIONS(1394),
    [anon_sym_pipeline] = ACTIONS(1396),
    [anon_sym_return] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1394),
    [anon_sym_switch] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_try] = ACTIONS(1396),
    [anon_sym_catch] = ACTIONS(1396),
    [anon_sym_finally] = ACTIONS(1396),
    [anon_sym_int] = ACTIONS(1396),
    [anon_sym_boolean] = ACTIONS(1396),
    [anon_sym_char] = ACTIONS(1396),
    [anon_sym_short] = ACTIONS(1396),
    [anon_sym_long] = ACTIONS(1396),
    [anon_sym_float] = ACTIONS(1396),
    [anon_sym_double] = ACTIONS(1396),
    [anon_sym_void] = ACTIONS(1396),
    [anon_sym_public] = ACTIONS(1396),
    [anon_sym_protected] = ACTIONS(1396),
    [anon_sym_private] = ACTIONS(1396),
    [anon_sym_static] = ACTIONS(1396),
    [anon_sym_final] = ACTIONS(1396),
    [anon_sym_synchronized] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1394),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_new] = ACTIONS(1396),
  },
  [STATE(314)] = {
    [sym_comment] = STATE(314),
    [sym_groovy_doc] = STATE(314),
    [ts_builtin_sym_end] = ACTIONS(1398),
    [sym_identifier] = ACTIONS(1400),
    [anon_sym_break] = ACTIONS(1400),
    [anon_sym_continue] = ACTIONS(1400),
    [anon_sym_SEMI] = ACTIONS(1398),
    [anon_sym_DOT_AMP] = ACTIONS(1398),
    [anon_sym_DOT_AT] = ACTIONS(1398),
    [anon_sym_QMARK_DOT] = ACTIONS(1398),
    [anon_sym_STAR_DOT] = ACTIONS(1398),
    [anon_sym_STAR] = ACTIONS(1400),
    [anon_sym_STAR_COLON] = ACTIONS(1398),
    [anon_sym_DOT] = ACTIONS(1400),
    [anon_sym_import] = ACTIONS(1400),
    [anon_sym_as] = ACTIONS(1400),
    [anon_sym_package] = ACTIONS(1400),
    [anon_sym_AT] = ACTIONS(1400),
    [anon_sym_assert] = ACTIONS(1400),
    [anon_sym_PLUS_PLUS] = ACTIONS(1398),
    [anon_sym_DASH_DASH] = ACTIONS(1398),
    [anon_sym_PERCENT] = ACTIONS(1398),
    [anon_sym_SLASH] = ACTIONS(1400),
    [anon_sym_PLUS] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_LT_LT] = ACTIONS(1398),
    [anon_sym_GT_GT] = ACTIONS(1400),
    [anon_sym_GT_GT_GT] = ACTIONS(1398),
    [anon_sym_DOT_DOT] = ACTIONS(1400),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1400),
    [anon_sym_LT] = ACTIONS(1400),
    [anon_sym_LT_EQ] = ACTIONS(1400),
    [anon_sym_GT] = ACTIONS(1400),
    [anon_sym_GT_EQ] = ACTIONS(1398),
    [anon_sym_in] = ACTIONS(1400),
    [anon_sym_BANGin] = ACTIONS(1400),
    [anon_sym_instanceof] = ACTIONS(1400),
    [anon_sym_BANGinstanceof] = ACTIONS(1398),
    [anon_sym_EQ_EQ] = ACTIONS(1400),
    [anon_sym_BANG_EQ] = ACTIONS(1400),
    [anon_sym_LT_EQ_GT] = ACTIONS(1398),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1398),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1398),
    [anon_sym_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_AMP] = ACTIONS(1400),
    [anon_sym_CARET] = ACTIONS(1398),
    [anon_sym_PIPE] = ACTIONS(1400),
    [anon_sym_AMP_AMP] = ACTIONS(1398),
    [anon_sym_PIPE_PIPE] = ACTIONS(1398),
    [anon_sym_QMARK_COLON] = ACTIONS(1398),
    [anon_sym_STAR_STAR] = ACTIONS(1398),
    [anon_sym_true] = ACTIONS(1400),
    [anon_sym_false] = ACTIONS(1400),
    [anon_sym_ATinterface] = ACTIONS(1398),
    [anon_sym_interface] = ACTIONS(1400),
    [anon_sym_class] = ACTIONS(1400),
    [anon_sym_COMMA] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1398),
    [anon_sym_RBRACE] = ACTIONS(1398),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1400),
    [anon_sym_def] = ACTIONS(1400),
    [anon_sym_LPAREN] = ACTIONS(1398),
    [anon_sym_null] = ACTIONS(1400),
    [anon_sym_this] = ACTIONS(1400),
    [anon_sym_do] = ACTIONS(1400),
    [anon_sym_while] = ACTIONS(1400),
    [anon_sym_for] = ACTIONS(1400),
    [aux_sym__type_identifier_token1] = ACTIONS(1400),
    [anon_sym_if] = ACTIONS(1400),
    [anon_sym_else] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1398),
    [aux_sym_number_literal_token1] = ACTIONS(1400),
    [aux_sym_number_literal_token2] = ACTIONS(1398),
    [aux_sym_number_literal_token3] = ACTIONS(1398),
    [aux_sym_number_literal_token4] = ACTIONS(1400),
    [aux_sym_number_literal_token5] = ACTIONS(1398),
    [anon_sym_pipeline] = ACTIONS(1400),
    [anon_sym_return] = ACTIONS(1400),
    [anon_sym_SQUOTE] = ACTIONS(1400),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1398),
    [anon_sym_DQUOTE] = ACTIONS(1400),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1398),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1398),
    [anon_sym_switch] = ACTIONS(1400),
    [anon_sym_QMARK] = ACTIONS(1400),
    [anon_sym_try] = ACTIONS(1400),
    [anon_sym_catch] = ACTIONS(1400),
    [anon_sym_finally] = ACTIONS(1400),
    [anon_sym_int] = ACTIONS(1400),
    [anon_sym_boolean] = ACTIONS(1400),
    [anon_sym_char] = ACTIONS(1400),
    [anon_sym_short] = ACTIONS(1400),
    [anon_sym_long] = ACTIONS(1400),
    [anon_sym_float] = ACTIONS(1400),
    [anon_sym_double] = ACTIONS(1400),
    [anon_sym_void] = ACTIONS(1400),
    [anon_sym_public] = ACTIONS(1400),
    [anon_sym_protected] = ACTIONS(1400),
    [anon_sym_private] = ACTIONS(1400),
    [anon_sym_static] = ACTIONS(1400),
    [anon_sym_final] = ACTIONS(1400),
    [anon_sym_synchronized] = ACTIONS(1400),
    [anon_sym_TILDE] = ACTIONS(1398),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_new] = ACTIONS(1400),
  },
  [STATE(315)] = {
    [sym_comment] = STATE(315),
    [sym_groovy_doc] = STATE(315),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(316)] = {
    [sym_comment] = STATE(316),
    [sym_groovy_doc] = STATE(316),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1370),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(317)] = {
    [sym_comment] = STATE(317),
    [sym_groovy_doc] = STATE(317),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(318)] = {
    [sym_comment] = STATE(318),
    [sym_groovy_doc] = STATE(318),
    [aux_sym_dotted_identifier_repeat1] = STATE(12),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1288),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1326),
    [anon_sym_DASH_DASH] = ACTIONS(1326),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1408),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(319)] = {
    [sym_comment] = STATE(319),
    [sym_groovy_doc] = STATE(319),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(320)] = {
    [sym_comment] = STATE(320),
    [sym_groovy_doc] = STATE(320),
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_DASH_GT] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_RPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_pipeline] = ACTIONS(1338),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(321)] = {
    [sym_comment] = STATE(321),
    [sym_groovy_doc] = STATE(321),
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_DOT_AMP] = ACTIONS(1410),
    [anon_sym_DOT_AT] = ACTIONS(1410),
    [anon_sym_QMARK_DOT] = ACTIONS(1410),
    [anon_sym_STAR_DOT] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_DOT] = ACTIONS(1412),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_as] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PERCENT] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_LT_LT] = ACTIONS(1410),
    [anon_sym_GT_GT] = ACTIONS(1412),
    [anon_sym_GT_GT_GT] = ACTIONS(1410),
    [anon_sym_DOT_DOT] = ACTIONS(1412),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1412),
    [anon_sym_LT] = ACTIONS(1412),
    [anon_sym_LT_EQ] = ACTIONS(1412),
    [anon_sym_GT] = ACTIONS(1412),
    [anon_sym_GT_EQ] = ACTIONS(1410),
    [anon_sym_in] = ACTIONS(1412),
    [anon_sym_BANGin] = ACTIONS(1412),
    [anon_sym_instanceof] = ACTIONS(1412),
    [anon_sym_BANGinstanceof] = ACTIONS(1410),
    [anon_sym_EQ_EQ] = ACTIONS(1412),
    [anon_sym_BANG_EQ] = ACTIONS(1412),
    [anon_sym_LT_EQ_GT] = ACTIONS(1410),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1410),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1410),
    [anon_sym_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_AMP] = ACTIONS(1412),
    [anon_sym_CARET] = ACTIONS(1410),
    [anon_sym_PIPE] = ACTIONS(1412),
    [anon_sym_AMP_AMP] = ACTIONS(1410),
    [anon_sym_PIPE_PIPE] = ACTIONS(1410),
    [anon_sym_QMARK_COLON] = ACTIONS(1410),
    [anon_sym_STAR_STAR] = ACTIONS(1410),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_COMMA] = ACTIONS(1410),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_pipeline] = ACTIONS(1412),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_QMARK] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_catch] = ACTIONS(1412),
    [anon_sym_finally] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(322)] = {
    [sym_comment] = STATE(322),
    [sym_groovy_doc] = STATE(322),
    [ts_builtin_sym_end] = ACTIONS(1414),
    [sym_identifier] = ACTIONS(1416),
    [anon_sym_break] = ACTIONS(1416),
    [anon_sym_continue] = ACTIONS(1416),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_DOT_AMP] = ACTIONS(1414),
    [anon_sym_DOT_AT] = ACTIONS(1414),
    [anon_sym_QMARK_DOT] = ACTIONS(1414),
    [anon_sym_STAR_DOT] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1416),
    [anon_sym_STAR_COLON] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1416),
    [anon_sym_import] = ACTIONS(1416),
    [anon_sym_as] = ACTIONS(1416),
    [anon_sym_package] = ACTIONS(1416),
    [anon_sym_AT] = ACTIONS(1416),
    [anon_sym_assert] = ACTIONS(1416),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_LT_LT] = ACTIONS(1414),
    [anon_sym_GT_GT] = ACTIONS(1416),
    [anon_sym_GT_GT_GT] = ACTIONS(1414),
    [anon_sym_DOT_DOT] = ACTIONS(1416),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1416),
    [anon_sym_LT] = ACTIONS(1416),
    [anon_sym_LT_EQ] = ACTIONS(1416),
    [anon_sym_GT] = ACTIONS(1416),
    [anon_sym_GT_EQ] = ACTIONS(1414),
    [anon_sym_in] = ACTIONS(1416),
    [anon_sym_BANGin] = ACTIONS(1416),
    [anon_sym_instanceof] = ACTIONS(1416),
    [anon_sym_BANGinstanceof] = ACTIONS(1414),
    [anon_sym_EQ_EQ] = ACTIONS(1416),
    [anon_sym_BANG_EQ] = ACTIONS(1416),
    [anon_sym_LT_EQ_GT] = ACTIONS(1414),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1414),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1414),
    [anon_sym_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1416),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1416),
    [anon_sym_AMP_AMP] = ACTIONS(1414),
    [anon_sym_PIPE_PIPE] = ACTIONS(1414),
    [anon_sym_QMARK_COLON] = ACTIONS(1414),
    [anon_sym_STAR_STAR] = ACTIONS(1414),
    [anon_sym_true] = ACTIONS(1416),
    [anon_sym_false] = ACTIONS(1416),
    [anon_sym_ATinterface] = ACTIONS(1414),
    [anon_sym_interface] = ACTIONS(1416),
    [anon_sym_class] = ACTIONS(1416),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1416),
    [anon_sym_def] = ACTIONS(1416),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_null] = ACTIONS(1416),
    [anon_sym_this] = ACTIONS(1416),
    [anon_sym_do] = ACTIONS(1416),
    [anon_sym_while] = ACTIONS(1416),
    [anon_sym_for] = ACTIONS(1416),
    [aux_sym__type_identifier_token1] = ACTIONS(1416),
    [anon_sym_if] = ACTIONS(1416),
    [anon_sym_else] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1414),
    [aux_sym_number_literal_token1] = ACTIONS(1416),
    [aux_sym_number_literal_token2] = ACTIONS(1414),
    [aux_sym_number_literal_token3] = ACTIONS(1414),
    [aux_sym_number_literal_token4] = ACTIONS(1416),
    [aux_sym_number_literal_token5] = ACTIONS(1414),
    [anon_sym_pipeline] = ACTIONS(1416),
    [anon_sym_return] = ACTIONS(1416),
    [anon_sym_SQUOTE] = ACTIONS(1416),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1416),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1414),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1414),
    [anon_sym_switch] = ACTIONS(1416),
    [anon_sym_QMARK] = ACTIONS(1416),
    [anon_sym_try] = ACTIONS(1416),
    [anon_sym_catch] = ACTIONS(1416),
    [anon_sym_finally] = ACTIONS(1416),
    [anon_sym_int] = ACTIONS(1416),
    [anon_sym_boolean] = ACTIONS(1416),
    [anon_sym_char] = ACTIONS(1416),
    [anon_sym_short] = ACTIONS(1416),
    [anon_sym_long] = ACTIONS(1416),
    [anon_sym_float] = ACTIONS(1416),
    [anon_sym_double] = ACTIONS(1416),
    [anon_sym_void] = ACTIONS(1416),
    [anon_sym_public] = ACTIONS(1416),
    [anon_sym_protected] = ACTIONS(1416),
    [anon_sym_private] = ACTIONS(1416),
    [anon_sym_static] = ACTIONS(1416),
    [anon_sym_final] = ACTIONS(1416),
    [anon_sym_synchronized] = ACTIONS(1416),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_new] = ACTIONS(1416),
  },
  [STATE(323)] = {
    [sym_comment] = STATE(323),
    [sym_groovy_doc] = STATE(323),
    [ts_builtin_sym_end] = ACTIONS(1418),
    [sym_identifier] = ACTIONS(1420),
    [anon_sym_break] = ACTIONS(1420),
    [anon_sym_continue] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1418),
    [anon_sym_DOT_AMP] = ACTIONS(1418),
    [anon_sym_DOT_AT] = ACTIONS(1418),
    [anon_sym_QMARK_DOT] = ACTIONS(1418),
    [anon_sym_STAR_DOT] = ACTIONS(1418),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_STAR_COLON] = ACTIONS(1418),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_import] = ACTIONS(1420),
    [anon_sym_as] = ACTIONS(1420),
    [anon_sym_package] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_assert] = ACTIONS(1420),
    [anon_sym_PLUS_PLUS] = ACTIONS(1418),
    [anon_sym_DASH_DASH] = ACTIONS(1418),
    [anon_sym_PERCENT] = ACTIONS(1418),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_LT_LT] = ACTIONS(1418),
    [anon_sym_GT_GT] = ACTIONS(1420),
    [anon_sym_GT_GT_GT] = ACTIONS(1418),
    [anon_sym_DOT_DOT] = ACTIONS(1420),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_LT_EQ] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_GT_EQ] = ACTIONS(1418),
    [anon_sym_in] = ACTIONS(1420),
    [anon_sym_BANGin] = ACTIONS(1420),
    [anon_sym_instanceof] = ACTIONS(1420),
    [anon_sym_BANGinstanceof] = ACTIONS(1418),
    [anon_sym_EQ_EQ] = ACTIONS(1420),
    [anon_sym_BANG_EQ] = ACTIONS(1420),
    [anon_sym_LT_EQ_GT] = ACTIONS(1418),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1418),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1418),
    [anon_sym_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1418),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_AMP_AMP] = ACTIONS(1418),
    [anon_sym_PIPE_PIPE] = ACTIONS(1418),
    [anon_sym_QMARK_COLON] = ACTIONS(1418),
    [anon_sym_STAR_STAR] = ACTIONS(1418),
    [anon_sym_true] = ACTIONS(1420),
    [anon_sym_false] = ACTIONS(1420),
    [anon_sym_ATinterface] = ACTIONS(1418),
    [anon_sym_interface] = ACTIONS(1420),
    [anon_sym_class] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1418),
    [anon_sym_RBRACE] = ACTIONS(1418),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_def] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1418),
    [anon_sym_null] = ACTIONS(1420),
    [anon_sym_this] = ACTIONS(1420),
    [anon_sym_do] = ACTIONS(1420),
    [anon_sym_while] = ACTIONS(1420),
    [anon_sym_for] = ACTIONS(1420),
    [aux_sym__type_identifier_token1] = ACTIONS(1420),
    [anon_sym_if] = ACTIONS(1420),
    [anon_sym_else] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [aux_sym_number_literal_token1] = ACTIONS(1420),
    [aux_sym_number_literal_token2] = ACTIONS(1418),
    [aux_sym_number_literal_token3] = ACTIONS(1418),
    [aux_sym_number_literal_token4] = ACTIONS(1420),
    [aux_sym_number_literal_token5] = ACTIONS(1418),
    [anon_sym_pipeline] = ACTIONS(1420),
    [anon_sym_return] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1418),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1418),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1418),
    [anon_sym_switch] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_try] = ACTIONS(1420),
    [anon_sym_catch] = ACTIONS(1420),
    [anon_sym_finally] = ACTIONS(1420),
    [anon_sym_int] = ACTIONS(1420),
    [anon_sym_boolean] = ACTIONS(1420),
    [anon_sym_char] = ACTIONS(1420),
    [anon_sym_short] = ACTIONS(1420),
    [anon_sym_long] = ACTIONS(1420),
    [anon_sym_float] = ACTIONS(1420),
    [anon_sym_double] = ACTIONS(1420),
    [anon_sym_void] = ACTIONS(1420),
    [anon_sym_public] = ACTIONS(1420),
    [anon_sym_protected] = ACTIONS(1420),
    [anon_sym_private] = ACTIONS(1420),
    [anon_sym_static] = ACTIONS(1420),
    [anon_sym_final] = ACTIONS(1420),
    [anon_sym_synchronized] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1418),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_new] = ACTIONS(1420),
  },
  [STATE(324)] = {
    [sym_comment] = STATE(324),
    [sym_groovy_doc] = STATE(324),
    [ts_builtin_sym_end] = ACTIONS(1374),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_DOT_AMP] = ACTIONS(1374),
    [anon_sym_DOT_AT] = ACTIONS(1374),
    [anon_sym_QMARK_DOT] = ACTIONS(1374),
    [anon_sym_STAR_DOT] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1376),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_as] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_LT_LT] = ACTIONS(1374),
    [anon_sym_GT_GT] = ACTIONS(1376),
    [anon_sym_GT_GT_GT] = ACTIONS(1374),
    [anon_sym_DOT_DOT] = ACTIONS(1376),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1376),
    [anon_sym_LT] = ACTIONS(1376),
    [anon_sym_LT_EQ] = ACTIONS(1376),
    [anon_sym_GT] = ACTIONS(1376),
    [anon_sym_GT_EQ] = ACTIONS(1374),
    [anon_sym_in] = ACTIONS(1376),
    [anon_sym_BANGin] = ACTIONS(1376),
    [anon_sym_instanceof] = ACTIONS(1376),
    [anon_sym_BANGinstanceof] = ACTIONS(1374),
    [anon_sym_EQ_EQ] = ACTIONS(1376),
    [anon_sym_BANG_EQ] = ACTIONS(1376),
    [anon_sym_LT_EQ_GT] = ACTIONS(1374),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1374),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1374),
    [anon_sym_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1376),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1376),
    [anon_sym_AMP_AMP] = ACTIONS(1374),
    [anon_sym_PIPE_PIPE] = ACTIONS(1374),
    [anon_sym_QMARK_COLON] = ACTIONS(1374),
    [anon_sym_STAR_STAR] = ACTIONS(1374),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_DASH_GT] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_RPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_pipeline] = ACTIONS(1376),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_QMARK] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(325)] = {
    [sym_comment] = STATE(325),
    [sym_groovy_doc] = STATE(325),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(326)] = {
    [sym_comment] = STATE(326),
    [sym_groovy_doc] = STATE(326),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(369),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(369),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(369),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(369),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(327)] = {
    [sym_comment] = STATE(327),
    [sym_groovy_doc] = STATE(327),
    [ts_builtin_sym_end] = ACTIONS(1318),
    [sym_identifier] = ACTIONS(1320),
    [anon_sym_break] = ACTIONS(1320),
    [anon_sym_continue] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_DOT_AMP] = ACTIONS(1318),
    [anon_sym_DOT_AT] = ACTIONS(1318),
    [anon_sym_QMARK_DOT] = ACTIONS(1318),
    [anon_sym_STAR_DOT] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_STAR_COLON] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1320),
    [anon_sym_import] = ACTIONS(1320),
    [anon_sym_as] = ACTIONS(1320),
    [anon_sym_package] = ACTIONS(1320),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_assert] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_LT_LT] = ACTIONS(1318),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_GT_GT_GT] = ACTIONS(1318),
    [anon_sym_DOT_DOT] = ACTIONS(1320),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_GT_EQ] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1320),
    [anon_sym_BANGin] = ACTIONS(1320),
    [anon_sym_instanceof] = ACTIONS(1320),
    [anon_sym_BANGinstanceof] = ACTIONS(1318),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT_EQ_GT] = ACTIONS(1318),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1318),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1318),
    [anon_sym_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_AMP_AMP] = ACTIONS(1318),
    [anon_sym_PIPE_PIPE] = ACTIONS(1318),
    [anon_sym_QMARK_COLON] = ACTIONS(1318),
    [anon_sym_STAR_STAR] = ACTIONS(1318),
    [anon_sym_true] = ACTIONS(1320),
    [anon_sym_false] = ACTIONS(1320),
    [anon_sym_ATinterface] = ACTIONS(1318),
    [anon_sym_interface] = ACTIONS(1320),
    [anon_sym_class] = ACTIONS(1320),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_DASH_GT] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1320),
    [anon_sym_def] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_null] = ACTIONS(1320),
    [anon_sym_this] = ACTIONS(1320),
    [anon_sym_do] = ACTIONS(1320),
    [anon_sym_while] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1320),
    [aux_sym__type_identifier_token1] = ACTIONS(1320),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_else] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1318),
    [aux_sym_number_literal_token1] = ACTIONS(1320),
    [aux_sym_number_literal_token2] = ACTIONS(1318),
    [aux_sym_number_literal_token3] = ACTIONS(1318),
    [aux_sym_number_literal_token4] = ACTIONS(1320),
    [aux_sym_number_literal_token5] = ACTIONS(1318),
    [anon_sym_pipeline] = ACTIONS(1320),
    [anon_sym_return] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1318),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_try] = ACTIONS(1320),
    [anon_sym_int] = ACTIONS(1320),
    [anon_sym_boolean] = ACTIONS(1320),
    [anon_sym_char] = ACTIONS(1320),
    [anon_sym_short] = ACTIONS(1320),
    [anon_sym_long] = ACTIONS(1320),
    [anon_sym_float] = ACTIONS(1320),
    [anon_sym_double] = ACTIONS(1320),
    [anon_sym_void] = ACTIONS(1320),
    [anon_sym_public] = ACTIONS(1320),
    [anon_sym_protected] = ACTIONS(1320),
    [anon_sym_private] = ACTIONS(1320),
    [anon_sym_static] = ACTIONS(1320),
    [anon_sym_final] = ACTIONS(1320),
    [anon_sym_synchronized] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_new] = ACTIONS(1320),
  },
  [STATE(328)] = {
    [sym_comment] = STATE(328),
    [sym_groovy_doc] = STATE(328),
    [aux_sym_dotted_identifier_repeat1] = STATE(376),
    [ts_builtin_sym_end] = ACTIONS(373),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(373),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(373),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(373),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(373),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(373),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_COMMA] = ACTIONS(373),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_DASH_GT] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_pipeline] = ACTIONS(375),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(329)] = {
    [sym_comment] = STATE(329),
    [sym_groovy_doc] = STATE(329),
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_pipeline] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_catch] = ACTIONS(1262),
    [anon_sym_finally] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(330)] = {
    [sym_comment] = STATE(330),
    [sym_groovy_doc] = STATE(330),
    [aux_sym_dotted_identifier_repeat1] = STATE(328),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1422),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_DASH_GT] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(331)] = {
    [sym_comment] = STATE(331),
    [sym_groovy_doc] = STATE(331),
    [aux_sym_dotted_identifier_repeat1] = STATE(9),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1360),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1312),
    [anon_sym_DASH_DASH] = ACTIONS(1312),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1424),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(332)] = {
    [sym_comment] = STATE(332),
    [sym_groovy_doc] = STATE(332),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(387),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(389),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(333)] = {
    [sym_comment] = STATE(333),
    [sym_groovy_doc] = STATE(333),
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_DOT_AMP] = ACTIONS(1410),
    [anon_sym_DOT_AT] = ACTIONS(1410),
    [anon_sym_QMARK_DOT] = ACTIONS(1410),
    [anon_sym_STAR_DOT] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_DOT] = ACTIONS(1412),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_as] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PERCENT] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_LT_LT] = ACTIONS(1410),
    [anon_sym_GT_GT] = ACTIONS(1412),
    [anon_sym_GT_GT_GT] = ACTIONS(1410),
    [anon_sym_DOT_DOT] = ACTIONS(1412),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1412),
    [anon_sym_LT] = ACTIONS(1412),
    [anon_sym_LT_EQ] = ACTIONS(1412),
    [anon_sym_GT] = ACTIONS(1412),
    [anon_sym_GT_EQ] = ACTIONS(1410),
    [anon_sym_in] = ACTIONS(1412),
    [anon_sym_BANGin] = ACTIONS(1412),
    [anon_sym_instanceof] = ACTIONS(1412),
    [anon_sym_BANGinstanceof] = ACTIONS(1410),
    [anon_sym_EQ_EQ] = ACTIONS(1412),
    [anon_sym_BANG_EQ] = ACTIONS(1412),
    [anon_sym_LT_EQ_GT] = ACTIONS(1410),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1410),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1410),
    [anon_sym_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_AMP] = ACTIONS(1412),
    [anon_sym_CARET] = ACTIONS(1410),
    [anon_sym_PIPE] = ACTIONS(1412),
    [anon_sym_AMP_AMP] = ACTIONS(1410),
    [anon_sym_PIPE_PIPE] = ACTIONS(1410),
    [anon_sym_QMARK_COLON] = ACTIONS(1410),
    [anon_sym_STAR_STAR] = ACTIONS(1410),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_COMMA] = ACTIONS(1410),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_DASH_GT] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_RPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_pipeline] = ACTIONS(1412),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_QMARK] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(334)] = {
    [sym_comment] = STATE(334),
    [sym_groovy_doc] = STATE(334),
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_pipeline] = ACTIONS(1338),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_catch] = ACTIONS(1338),
    [anon_sym_finally] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(335)] = {
    [sym_comment] = STATE(335),
    [sym_groovy_doc] = STATE(335),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(336)] = {
    [sym_comment] = STATE(336),
    [sym_groovy_doc] = STATE(336),
    [sym_identifier] = ACTIONS(1380),
    [anon_sym_break] = ACTIONS(1380),
    [anon_sym_continue] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1378),
    [anon_sym_DOT_AMP] = ACTIONS(1378),
    [anon_sym_DOT_AT] = ACTIONS(1378),
    [anon_sym_QMARK_DOT] = ACTIONS(1378),
    [anon_sym_STAR_DOT] = ACTIONS(1378),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_STAR_COLON] = ACTIONS(1378),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_import] = ACTIONS(1380),
    [anon_sym_as] = ACTIONS(1380),
    [anon_sym_package] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_assert] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1378),
    [anon_sym_DASH_DASH] = ACTIONS(1378),
    [anon_sym_PERCENT] = ACTIONS(1378),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_LT_LT] = ACTIONS(1378),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym_GT_GT_GT] = ACTIONS(1378),
    [anon_sym_DOT_DOT] = ACTIONS(1380),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_GT_EQ] = ACTIONS(1378),
    [anon_sym_in] = ACTIONS(1380),
    [anon_sym_BANGin] = ACTIONS(1380),
    [anon_sym_instanceof] = ACTIONS(1380),
    [anon_sym_BANGinstanceof] = ACTIONS(1378),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ_GT] = ACTIONS(1378),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1378),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1378),
    [anon_sym_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1378),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1378),
    [anon_sym_PIPE_PIPE] = ACTIONS(1378),
    [anon_sym_QMARK_COLON] = ACTIONS(1378),
    [anon_sym_STAR_STAR] = ACTIONS(1378),
    [anon_sym_true] = ACTIONS(1380),
    [anon_sym_false] = ACTIONS(1380),
    [anon_sym_ATinterface] = ACTIONS(1378),
    [anon_sym_interface] = ACTIONS(1380),
    [anon_sym_class] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1378),
    [anon_sym_LBRACE] = ACTIONS(1378),
    [anon_sym_RBRACE] = ACTIONS(1378),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_def] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1378),
    [anon_sym_null] = ACTIONS(1380),
    [anon_sym_this] = ACTIONS(1380),
    [anon_sym_do] = ACTIONS(1380),
    [anon_sym_while] = ACTIONS(1380),
    [anon_sym_for] = ACTIONS(1380),
    [aux_sym__type_identifier_token1] = ACTIONS(1380),
    [anon_sym_if] = ACTIONS(1380),
    [anon_sym_else] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [aux_sym_number_literal_token1] = ACTIONS(1380),
    [aux_sym_number_literal_token2] = ACTIONS(1378),
    [aux_sym_number_literal_token3] = ACTIONS(1378),
    [aux_sym_number_literal_token4] = ACTIONS(1380),
    [aux_sym_number_literal_token5] = ACTIONS(1378),
    [anon_sym_return] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1378),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1378),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1378),
    [anon_sym_switch] = ACTIONS(1380),
    [anon_sym_case] = ACTIONS(1380),
    [anon_sym_default] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_try] = ACTIONS(1380),
    [anon_sym_catch] = ACTIONS(1380),
    [anon_sym_finally] = ACTIONS(1380),
    [anon_sym_int] = ACTIONS(1380),
    [anon_sym_boolean] = ACTIONS(1380),
    [anon_sym_char] = ACTIONS(1380),
    [anon_sym_short] = ACTIONS(1380),
    [anon_sym_long] = ACTIONS(1380),
    [anon_sym_float] = ACTIONS(1380),
    [anon_sym_double] = ACTIONS(1380),
    [anon_sym_void] = ACTIONS(1380),
    [anon_sym_public] = ACTIONS(1380),
    [anon_sym_protected] = ACTIONS(1380),
    [anon_sym_private] = ACTIONS(1380),
    [anon_sym_static] = ACTIONS(1380),
    [anon_sym_final] = ACTIONS(1380),
    [anon_sym_synchronized] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1378),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_new] = ACTIONS(1380),
  },
  [STATE(337)] = {
    [sym_comment] = STATE(337),
    [sym_groovy_doc] = STATE(337),
    [sym_identifier] = ACTIONS(1384),
    [anon_sym_break] = ACTIONS(1384),
    [anon_sym_continue] = ACTIONS(1384),
    [anon_sym_SEMI] = ACTIONS(1382),
    [anon_sym_DOT_AMP] = ACTIONS(1382),
    [anon_sym_DOT_AT] = ACTIONS(1382),
    [anon_sym_QMARK_DOT] = ACTIONS(1382),
    [anon_sym_STAR_DOT] = ACTIONS(1382),
    [anon_sym_STAR] = ACTIONS(1384),
    [anon_sym_STAR_COLON] = ACTIONS(1382),
    [anon_sym_DOT] = ACTIONS(1384),
    [anon_sym_import] = ACTIONS(1384),
    [anon_sym_as] = ACTIONS(1384),
    [anon_sym_package] = ACTIONS(1384),
    [anon_sym_AT] = ACTIONS(1384),
    [anon_sym_assert] = ACTIONS(1384),
    [anon_sym_PLUS_PLUS] = ACTIONS(1382),
    [anon_sym_DASH_DASH] = ACTIONS(1382),
    [anon_sym_PERCENT] = ACTIONS(1382),
    [anon_sym_SLASH] = ACTIONS(1384),
    [anon_sym_PLUS] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_LT_LT] = ACTIONS(1382),
    [anon_sym_GT_GT] = ACTIONS(1384),
    [anon_sym_GT_GT_GT] = ACTIONS(1382),
    [anon_sym_DOT_DOT] = ACTIONS(1384),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1384),
    [anon_sym_LT] = ACTIONS(1384),
    [anon_sym_LT_EQ] = ACTIONS(1384),
    [anon_sym_GT] = ACTIONS(1384),
    [anon_sym_GT_EQ] = ACTIONS(1382),
    [anon_sym_in] = ACTIONS(1384),
    [anon_sym_BANGin] = ACTIONS(1384),
    [anon_sym_instanceof] = ACTIONS(1384),
    [anon_sym_BANGinstanceof] = ACTIONS(1382),
    [anon_sym_EQ_EQ] = ACTIONS(1384),
    [anon_sym_BANG_EQ] = ACTIONS(1384),
    [anon_sym_LT_EQ_GT] = ACTIONS(1382),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1382),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1382),
    [anon_sym_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_AMP] = ACTIONS(1384),
    [anon_sym_CARET] = ACTIONS(1382),
    [anon_sym_PIPE] = ACTIONS(1384),
    [anon_sym_AMP_AMP] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1382),
    [anon_sym_QMARK_COLON] = ACTIONS(1382),
    [anon_sym_STAR_STAR] = ACTIONS(1382),
    [anon_sym_true] = ACTIONS(1384),
    [anon_sym_false] = ACTIONS(1384),
    [anon_sym_ATinterface] = ACTIONS(1382),
    [anon_sym_interface] = ACTIONS(1384),
    [anon_sym_class] = ACTIONS(1384),
    [anon_sym_COMMA] = ACTIONS(1382),
    [anon_sym_LBRACE] = ACTIONS(1382),
    [anon_sym_RBRACE] = ACTIONS(1382),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1384),
    [anon_sym_def] = ACTIONS(1384),
    [anon_sym_LPAREN] = ACTIONS(1382),
    [anon_sym_null] = ACTIONS(1384),
    [anon_sym_this] = ACTIONS(1384),
    [anon_sym_do] = ACTIONS(1384),
    [anon_sym_while] = ACTIONS(1384),
    [anon_sym_for] = ACTIONS(1384),
    [aux_sym__type_identifier_token1] = ACTIONS(1384),
    [anon_sym_if] = ACTIONS(1384),
    [anon_sym_else] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1382),
    [aux_sym_number_literal_token1] = ACTIONS(1384),
    [aux_sym_number_literal_token2] = ACTIONS(1382),
    [aux_sym_number_literal_token3] = ACTIONS(1382),
    [aux_sym_number_literal_token4] = ACTIONS(1384),
    [aux_sym_number_literal_token5] = ACTIONS(1382),
    [anon_sym_return] = ACTIONS(1384),
    [anon_sym_SQUOTE] = ACTIONS(1384),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1382),
    [anon_sym_DQUOTE] = ACTIONS(1384),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1382),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1382),
    [anon_sym_switch] = ACTIONS(1384),
    [anon_sym_case] = ACTIONS(1384),
    [anon_sym_default] = ACTIONS(1384),
    [anon_sym_QMARK] = ACTIONS(1384),
    [anon_sym_try] = ACTIONS(1384),
    [anon_sym_catch] = ACTIONS(1384),
    [anon_sym_finally] = ACTIONS(1384),
    [anon_sym_int] = ACTIONS(1384),
    [anon_sym_boolean] = ACTIONS(1384),
    [anon_sym_char] = ACTIONS(1384),
    [anon_sym_short] = ACTIONS(1384),
    [anon_sym_long] = ACTIONS(1384),
    [anon_sym_float] = ACTIONS(1384),
    [anon_sym_double] = ACTIONS(1384),
    [anon_sym_void] = ACTIONS(1384),
    [anon_sym_public] = ACTIONS(1384),
    [anon_sym_protected] = ACTIONS(1384),
    [anon_sym_private] = ACTIONS(1384),
    [anon_sym_static] = ACTIONS(1384),
    [anon_sym_final] = ACTIONS(1384),
    [anon_sym_synchronized] = ACTIONS(1384),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_new] = ACTIONS(1384),
  },
  [STATE(338)] = {
    [sym_comment] = STATE(338),
    [sym_groovy_doc] = STATE(338),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(339)] = {
    [sym_comment] = STATE(339),
    [sym_groovy_doc] = STATE(339),
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_DOT_AMP] = ACTIONS(1364),
    [anon_sym_DOT_AT] = ACTIONS(1364),
    [anon_sym_QMARK_DOT] = ACTIONS(1364),
    [anon_sym_STAR_DOT] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_as] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_PERCENT] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_LT_LT] = ACTIONS(1364),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_GT_GT_GT] = ACTIONS(1364),
    [anon_sym_DOT_DOT] = ACTIONS(1362),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_GT_EQ] = ACTIONS(1364),
    [anon_sym_in] = ACTIONS(1362),
    [anon_sym_BANGin] = ACTIONS(1362),
    [anon_sym_instanceof] = ACTIONS(1362),
    [anon_sym_BANGinstanceof] = ACTIONS(1364),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT_EQ_GT] = ACTIONS(1364),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1364),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1364),
    [anon_sym_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_AMP_AMP] = ACTIONS(1364),
    [anon_sym_PIPE_PIPE] = ACTIONS(1364),
    [anon_sym_QMARK_COLON] = ACTIONS(1364),
    [anon_sym_STAR_STAR] = ACTIONS(1364),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_DASH_GT] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_RPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_pipeline] = ACTIONS(1362),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(340)] = {
    [sym_comment] = STATE(340),
    [sym_groovy_doc] = STATE(340),
    [ts_builtin_sym_end] = ACTIONS(1330),
    [sym_identifier] = ACTIONS(1328),
    [anon_sym_break] = ACTIONS(1328),
    [anon_sym_continue] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1330),
    [anon_sym_COLON] = ACTIONS(1330),
    [anon_sym_DOT_AMP] = ACTIONS(1330),
    [anon_sym_DOT_AT] = ACTIONS(1330),
    [anon_sym_QMARK_DOT] = ACTIONS(1330),
    [anon_sym_STAR_DOT] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_STAR_COLON] = ACTIONS(1330),
    [anon_sym_DOT] = ACTIONS(1328),
    [anon_sym_import] = ACTIONS(1328),
    [anon_sym_as] = ACTIONS(1328),
    [anon_sym_package] = ACTIONS(1328),
    [anon_sym_AT] = ACTIONS(1328),
    [anon_sym_assert] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1330),
    [anon_sym_PERCENT] = ACTIONS(1330),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_LT_LT] = ACTIONS(1330),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_GT_GT_GT] = ACTIONS(1330),
    [anon_sym_DOT_DOT] = ACTIONS(1328),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_GT_EQ] = ACTIONS(1330),
    [anon_sym_in] = ACTIONS(1328),
    [anon_sym_BANGin] = ACTIONS(1328),
    [anon_sym_instanceof] = ACTIONS(1328),
    [anon_sym_BANGinstanceof] = ACTIONS(1330),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT_EQ_GT] = ACTIONS(1330),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1330),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1330),
    [anon_sym_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1330),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_AMP_AMP] = ACTIONS(1330),
    [anon_sym_PIPE_PIPE] = ACTIONS(1330),
    [anon_sym_QMARK_COLON] = ACTIONS(1330),
    [anon_sym_STAR_STAR] = ACTIONS(1330),
    [anon_sym_true] = ACTIONS(1328),
    [anon_sym_false] = ACTIONS(1328),
    [anon_sym_ATinterface] = ACTIONS(1330),
    [anon_sym_interface] = ACTIONS(1328),
    [anon_sym_class] = ACTIONS(1328),
    [anon_sym_COMMA] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1330),
    [anon_sym_DASH_GT] = ACTIONS(1330),
    [anon_sym_RBRACE] = ACTIONS(1330),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1328),
    [anon_sym_def] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1330),
    [anon_sym_null] = ACTIONS(1328),
    [anon_sym_this] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1328),
    [anon_sym_while] = ACTIONS(1328),
    [anon_sym_for] = ACTIONS(1328),
    [aux_sym__type_identifier_token1] = ACTIONS(1328),
    [anon_sym_if] = ACTIONS(1328),
    [anon_sym_else] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [aux_sym_number_literal_token1] = ACTIONS(1328),
    [aux_sym_number_literal_token2] = ACTIONS(1330),
    [aux_sym_number_literal_token3] = ACTIONS(1330),
    [aux_sym_number_literal_token4] = ACTIONS(1328),
    [aux_sym_number_literal_token5] = ACTIONS(1330),
    [anon_sym_pipeline] = ACTIONS(1328),
    [anon_sym_return] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1330),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1330),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1328),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_try] = ACTIONS(1328),
    [anon_sym_int] = ACTIONS(1328),
    [anon_sym_boolean] = ACTIONS(1328),
    [anon_sym_char] = ACTIONS(1328),
    [anon_sym_short] = ACTIONS(1328),
    [anon_sym_long] = ACTIONS(1328),
    [anon_sym_float] = ACTIONS(1328),
    [anon_sym_double] = ACTIONS(1328),
    [anon_sym_void] = ACTIONS(1328),
    [anon_sym_public] = ACTIONS(1328),
    [anon_sym_protected] = ACTIONS(1328),
    [anon_sym_private] = ACTIONS(1328),
    [anon_sym_static] = ACTIONS(1328),
    [anon_sym_final] = ACTIONS(1328),
    [anon_sym_synchronized] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1330),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_new] = ACTIONS(1328),
  },
  [STATE(341)] = {
    [sym_comment] = STATE(341),
    [sym_groovy_doc] = STATE(341),
    [sym_identifier] = ACTIONS(1396),
    [anon_sym_break] = ACTIONS(1396),
    [anon_sym_continue] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1394),
    [anon_sym_DOT_AMP] = ACTIONS(1394),
    [anon_sym_DOT_AT] = ACTIONS(1394),
    [anon_sym_QMARK_DOT] = ACTIONS(1394),
    [anon_sym_STAR_DOT] = ACTIONS(1394),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_STAR_COLON] = ACTIONS(1394),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_import] = ACTIONS(1396),
    [anon_sym_as] = ACTIONS(1396),
    [anon_sym_package] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_assert] = ACTIONS(1396),
    [anon_sym_PLUS_PLUS] = ACTIONS(1394),
    [anon_sym_DASH_DASH] = ACTIONS(1394),
    [anon_sym_PERCENT] = ACTIONS(1394),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_LT_LT] = ACTIONS(1394),
    [anon_sym_GT_GT] = ACTIONS(1396),
    [anon_sym_GT_GT_GT] = ACTIONS(1394),
    [anon_sym_DOT_DOT] = ACTIONS(1396),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_LT_EQ] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_GT_EQ] = ACTIONS(1394),
    [anon_sym_in] = ACTIONS(1396),
    [anon_sym_BANGin] = ACTIONS(1396),
    [anon_sym_instanceof] = ACTIONS(1396),
    [anon_sym_BANGinstanceof] = ACTIONS(1394),
    [anon_sym_EQ_EQ] = ACTIONS(1396),
    [anon_sym_BANG_EQ] = ACTIONS(1396),
    [anon_sym_LT_EQ_GT] = ACTIONS(1394),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1394),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1394),
    [anon_sym_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1394),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_AMP_AMP] = ACTIONS(1394),
    [anon_sym_PIPE_PIPE] = ACTIONS(1394),
    [anon_sym_QMARK_COLON] = ACTIONS(1394),
    [anon_sym_STAR_STAR] = ACTIONS(1394),
    [anon_sym_true] = ACTIONS(1396),
    [anon_sym_false] = ACTIONS(1396),
    [anon_sym_ATinterface] = ACTIONS(1394),
    [anon_sym_interface] = ACTIONS(1396),
    [anon_sym_class] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(1394),
    [anon_sym_RBRACE] = ACTIONS(1394),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_def] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1394),
    [anon_sym_null] = ACTIONS(1396),
    [anon_sym_this] = ACTIONS(1396),
    [anon_sym_do] = ACTIONS(1396),
    [anon_sym_while] = ACTIONS(1396),
    [anon_sym_for] = ACTIONS(1396),
    [aux_sym__type_identifier_token1] = ACTIONS(1396),
    [anon_sym_if] = ACTIONS(1396),
    [anon_sym_else] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [aux_sym_number_literal_token1] = ACTIONS(1396),
    [aux_sym_number_literal_token2] = ACTIONS(1394),
    [aux_sym_number_literal_token3] = ACTIONS(1394),
    [aux_sym_number_literal_token4] = ACTIONS(1396),
    [aux_sym_number_literal_token5] = ACTIONS(1394),
    [anon_sym_return] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1394),
    [anon_sym_switch] = ACTIONS(1396),
    [anon_sym_case] = ACTIONS(1396),
    [anon_sym_default] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_try] = ACTIONS(1396),
    [anon_sym_catch] = ACTIONS(1396),
    [anon_sym_finally] = ACTIONS(1396),
    [anon_sym_int] = ACTIONS(1396),
    [anon_sym_boolean] = ACTIONS(1396),
    [anon_sym_char] = ACTIONS(1396),
    [anon_sym_short] = ACTIONS(1396),
    [anon_sym_long] = ACTIONS(1396),
    [anon_sym_float] = ACTIONS(1396),
    [anon_sym_double] = ACTIONS(1396),
    [anon_sym_void] = ACTIONS(1396),
    [anon_sym_public] = ACTIONS(1396),
    [anon_sym_protected] = ACTIONS(1396),
    [anon_sym_private] = ACTIONS(1396),
    [anon_sym_static] = ACTIONS(1396),
    [anon_sym_final] = ACTIONS(1396),
    [anon_sym_synchronized] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1394),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_new] = ACTIONS(1396),
  },
  [STATE(342)] = {
    [sym_comment] = STATE(342),
    [sym_groovy_doc] = STATE(342),
    [sym_identifier] = ACTIONS(1400),
    [anon_sym_break] = ACTIONS(1400),
    [anon_sym_continue] = ACTIONS(1400),
    [anon_sym_SEMI] = ACTIONS(1398),
    [anon_sym_DOT_AMP] = ACTIONS(1398),
    [anon_sym_DOT_AT] = ACTIONS(1398),
    [anon_sym_QMARK_DOT] = ACTIONS(1398),
    [anon_sym_STAR_DOT] = ACTIONS(1398),
    [anon_sym_STAR] = ACTIONS(1400),
    [anon_sym_STAR_COLON] = ACTIONS(1398),
    [anon_sym_DOT] = ACTIONS(1400),
    [anon_sym_import] = ACTIONS(1400),
    [anon_sym_as] = ACTIONS(1400),
    [anon_sym_package] = ACTIONS(1400),
    [anon_sym_AT] = ACTIONS(1400),
    [anon_sym_assert] = ACTIONS(1400),
    [anon_sym_PLUS_PLUS] = ACTIONS(1398),
    [anon_sym_DASH_DASH] = ACTIONS(1398),
    [anon_sym_PERCENT] = ACTIONS(1398),
    [anon_sym_SLASH] = ACTIONS(1400),
    [anon_sym_PLUS] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_LT_LT] = ACTIONS(1398),
    [anon_sym_GT_GT] = ACTIONS(1400),
    [anon_sym_GT_GT_GT] = ACTIONS(1398),
    [anon_sym_DOT_DOT] = ACTIONS(1400),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1400),
    [anon_sym_LT] = ACTIONS(1400),
    [anon_sym_LT_EQ] = ACTIONS(1400),
    [anon_sym_GT] = ACTIONS(1400),
    [anon_sym_GT_EQ] = ACTIONS(1398),
    [anon_sym_in] = ACTIONS(1400),
    [anon_sym_BANGin] = ACTIONS(1400),
    [anon_sym_instanceof] = ACTIONS(1400),
    [anon_sym_BANGinstanceof] = ACTIONS(1398),
    [anon_sym_EQ_EQ] = ACTIONS(1400),
    [anon_sym_BANG_EQ] = ACTIONS(1400),
    [anon_sym_LT_EQ_GT] = ACTIONS(1398),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1398),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1398),
    [anon_sym_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_AMP] = ACTIONS(1400),
    [anon_sym_CARET] = ACTIONS(1398),
    [anon_sym_PIPE] = ACTIONS(1400),
    [anon_sym_AMP_AMP] = ACTIONS(1398),
    [anon_sym_PIPE_PIPE] = ACTIONS(1398),
    [anon_sym_QMARK_COLON] = ACTIONS(1398),
    [anon_sym_STAR_STAR] = ACTIONS(1398),
    [anon_sym_true] = ACTIONS(1400),
    [anon_sym_false] = ACTIONS(1400),
    [anon_sym_ATinterface] = ACTIONS(1398),
    [anon_sym_interface] = ACTIONS(1400),
    [anon_sym_class] = ACTIONS(1400),
    [anon_sym_COMMA] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1398),
    [anon_sym_RBRACE] = ACTIONS(1398),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1400),
    [anon_sym_def] = ACTIONS(1400),
    [anon_sym_LPAREN] = ACTIONS(1398),
    [anon_sym_null] = ACTIONS(1400),
    [anon_sym_this] = ACTIONS(1400),
    [anon_sym_do] = ACTIONS(1400),
    [anon_sym_while] = ACTIONS(1400),
    [anon_sym_for] = ACTIONS(1400),
    [aux_sym__type_identifier_token1] = ACTIONS(1400),
    [anon_sym_if] = ACTIONS(1400),
    [anon_sym_else] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1398),
    [aux_sym_number_literal_token1] = ACTIONS(1400),
    [aux_sym_number_literal_token2] = ACTIONS(1398),
    [aux_sym_number_literal_token3] = ACTIONS(1398),
    [aux_sym_number_literal_token4] = ACTIONS(1400),
    [aux_sym_number_literal_token5] = ACTIONS(1398),
    [anon_sym_return] = ACTIONS(1400),
    [anon_sym_SQUOTE] = ACTIONS(1400),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1398),
    [anon_sym_DQUOTE] = ACTIONS(1400),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1398),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1398),
    [anon_sym_switch] = ACTIONS(1400),
    [anon_sym_case] = ACTIONS(1400),
    [anon_sym_default] = ACTIONS(1400),
    [anon_sym_QMARK] = ACTIONS(1400),
    [anon_sym_try] = ACTIONS(1400),
    [anon_sym_catch] = ACTIONS(1400),
    [anon_sym_finally] = ACTIONS(1400),
    [anon_sym_int] = ACTIONS(1400),
    [anon_sym_boolean] = ACTIONS(1400),
    [anon_sym_char] = ACTIONS(1400),
    [anon_sym_short] = ACTIONS(1400),
    [anon_sym_long] = ACTIONS(1400),
    [anon_sym_float] = ACTIONS(1400),
    [anon_sym_double] = ACTIONS(1400),
    [anon_sym_void] = ACTIONS(1400),
    [anon_sym_public] = ACTIONS(1400),
    [anon_sym_protected] = ACTIONS(1400),
    [anon_sym_private] = ACTIONS(1400),
    [anon_sym_static] = ACTIONS(1400),
    [anon_sym_final] = ACTIONS(1400),
    [anon_sym_synchronized] = ACTIONS(1400),
    [anon_sym_TILDE] = ACTIONS(1398),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_new] = ACTIONS(1400),
  },
  [STATE(343)] = {
    [sym_comment] = STATE(343),
    [sym_groovy_doc] = STATE(343),
    [sym_identifier] = ACTIONS(1416),
    [anon_sym_break] = ACTIONS(1416),
    [anon_sym_continue] = ACTIONS(1416),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_DOT_AMP] = ACTIONS(1414),
    [anon_sym_DOT_AT] = ACTIONS(1414),
    [anon_sym_QMARK_DOT] = ACTIONS(1414),
    [anon_sym_STAR_DOT] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1416),
    [anon_sym_STAR_COLON] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1416),
    [anon_sym_import] = ACTIONS(1416),
    [anon_sym_as] = ACTIONS(1416),
    [anon_sym_package] = ACTIONS(1416),
    [anon_sym_AT] = ACTIONS(1416),
    [anon_sym_assert] = ACTIONS(1416),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_LT_LT] = ACTIONS(1414),
    [anon_sym_GT_GT] = ACTIONS(1416),
    [anon_sym_GT_GT_GT] = ACTIONS(1414),
    [anon_sym_DOT_DOT] = ACTIONS(1416),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1416),
    [anon_sym_LT] = ACTIONS(1416),
    [anon_sym_LT_EQ] = ACTIONS(1416),
    [anon_sym_GT] = ACTIONS(1416),
    [anon_sym_GT_EQ] = ACTIONS(1414),
    [anon_sym_in] = ACTIONS(1416),
    [anon_sym_BANGin] = ACTIONS(1416),
    [anon_sym_instanceof] = ACTIONS(1416),
    [anon_sym_BANGinstanceof] = ACTIONS(1414),
    [anon_sym_EQ_EQ] = ACTIONS(1416),
    [anon_sym_BANG_EQ] = ACTIONS(1416),
    [anon_sym_LT_EQ_GT] = ACTIONS(1414),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1414),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1414),
    [anon_sym_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1416),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1416),
    [anon_sym_AMP_AMP] = ACTIONS(1414),
    [anon_sym_PIPE_PIPE] = ACTIONS(1414),
    [anon_sym_QMARK_COLON] = ACTIONS(1414),
    [anon_sym_STAR_STAR] = ACTIONS(1414),
    [anon_sym_true] = ACTIONS(1416),
    [anon_sym_false] = ACTIONS(1416),
    [anon_sym_ATinterface] = ACTIONS(1414),
    [anon_sym_interface] = ACTIONS(1416),
    [anon_sym_class] = ACTIONS(1416),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1416),
    [anon_sym_def] = ACTIONS(1416),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_null] = ACTIONS(1416),
    [anon_sym_this] = ACTIONS(1416),
    [anon_sym_do] = ACTIONS(1416),
    [anon_sym_while] = ACTIONS(1416),
    [anon_sym_for] = ACTIONS(1416),
    [aux_sym__type_identifier_token1] = ACTIONS(1416),
    [anon_sym_if] = ACTIONS(1416),
    [anon_sym_else] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1414),
    [aux_sym_number_literal_token1] = ACTIONS(1416),
    [aux_sym_number_literal_token2] = ACTIONS(1414),
    [aux_sym_number_literal_token3] = ACTIONS(1414),
    [aux_sym_number_literal_token4] = ACTIONS(1416),
    [aux_sym_number_literal_token5] = ACTIONS(1414),
    [anon_sym_return] = ACTIONS(1416),
    [anon_sym_SQUOTE] = ACTIONS(1416),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1416),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1414),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1414),
    [anon_sym_switch] = ACTIONS(1416),
    [anon_sym_case] = ACTIONS(1416),
    [anon_sym_default] = ACTIONS(1416),
    [anon_sym_QMARK] = ACTIONS(1416),
    [anon_sym_try] = ACTIONS(1416),
    [anon_sym_catch] = ACTIONS(1416),
    [anon_sym_finally] = ACTIONS(1416),
    [anon_sym_int] = ACTIONS(1416),
    [anon_sym_boolean] = ACTIONS(1416),
    [anon_sym_char] = ACTIONS(1416),
    [anon_sym_short] = ACTIONS(1416),
    [anon_sym_long] = ACTIONS(1416),
    [anon_sym_float] = ACTIONS(1416),
    [anon_sym_double] = ACTIONS(1416),
    [anon_sym_void] = ACTIONS(1416),
    [anon_sym_public] = ACTIONS(1416),
    [anon_sym_protected] = ACTIONS(1416),
    [anon_sym_private] = ACTIONS(1416),
    [anon_sym_static] = ACTIONS(1416),
    [anon_sym_final] = ACTIONS(1416),
    [anon_sym_synchronized] = ACTIONS(1416),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_new] = ACTIONS(1416),
  },
  [STATE(344)] = {
    [sym_comment] = STATE(344),
    [sym_groovy_doc] = STATE(344),
    [aux_sym_dotted_identifier_repeat1] = STATE(328),
    [ts_builtin_sym_end] = ACTIONS(1278),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_DOT] = ACTIONS(1422),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_DASH_GT] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1426),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_pipeline] = ACTIONS(1280),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(345)] = {
    [sym_comment] = STATE(345),
    [sym_groovy_doc] = STATE(345),
    [sym_identifier] = ACTIONS(1420),
    [anon_sym_break] = ACTIONS(1420),
    [anon_sym_continue] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1418),
    [anon_sym_DOT_AMP] = ACTIONS(1418),
    [anon_sym_DOT_AT] = ACTIONS(1418),
    [anon_sym_QMARK_DOT] = ACTIONS(1418),
    [anon_sym_STAR_DOT] = ACTIONS(1418),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_STAR_COLON] = ACTIONS(1418),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_import] = ACTIONS(1420),
    [anon_sym_as] = ACTIONS(1420),
    [anon_sym_package] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_assert] = ACTIONS(1420),
    [anon_sym_PLUS_PLUS] = ACTIONS(1418),
    [anon_sym_DASH_DASH] = ACTIONS(1418),
    [anon_sym_PERCENT] = ACTIONS(1418),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_LT_LT] = ACTIONS(1418),
    [anon_sym_GT_GT] = ACTIONS(1420),
    [anon_sym_GT_GT_GT] = ACTIONS(1418),
    [anon_sym_DOT_DOT] = ACTIONS(1420),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_LT_EQ] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_GT_EQ] = ACTIONS(1418),
    [anon_sym_in] = ACTIONS(1420),
    [anon_sym_BANGin] = ACTIONS(1420),
    [anon_sym_instanceof] = ACTIONS(1420),
    [anon_sym_BANGinstanceof] = ACTIONS(1418),
    [anon_sym_EQ_EQ] = ACTIONS(1420),
    [anon_sym_BANG_EQ] = ACTIONS(1420),
    [anon_sym_LT_EQ_GT] = ACTIONS(1418),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1418),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1418),
    [anon_sym_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1418),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_AMP_AMP] = ACTIONS(1418),
    [anon_sym_PIPE_PIPE] = ACTIONS(1418),
    [anon_sym_QMARK_COLON] = ACTIONS(1418),
    [anon_sym_STAR_STAR] = ACTIONS(1418),
    [anon_sym_true] = ACTIONS(1420),
    [anon_sym_false] = ACTIONS(1420),
    [anon_sym_ATinterface] = ACTIONS(1418),
    [anon_sym_interface] = ACTIONS(1420),
    [anon_sym_class] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1418),
    [anon_sym_RBRACE] = ACTIONS(1418),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_def] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1418),
    [anon_sym_null] = ACTIONS(1420),
    [anon_sym_this] = ACTIONS(1420),
    [anon_sym_do] = ACTIONS(1420),
    [anon_sym_while] = ACTIONS(1420),
    [anon_sym_for] = ACTIONS(1420),
    [aux_sym__type_identifier_token1] = ACTIONS(1420),
    [anon_sym_if] = ACTIONS(1420),
    [anon_sym_else] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [aux_sym_number_literal_token1] = ACTIONS(1420),
    [aux_sym_number_literal_token2] = ACTIONS(1418),
    [aux_sym_number_literal_token3] = ACTIONS(1418),
    [aux_sym_number_literal_token4] = ACTIONS(1420),
    [aux_sym_number_literal_token5] = ACTIONS(1418),
    [anon_sym_return] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1418),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1418),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1418),
    [anon_sym_switch] = ACTIONS(1420),
    [anon_sym_case] = ACTIONS(1420),
    [anon_sym_default] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_try] = ACTIONS(1420),
    [anon_sym_catch] = ACTIONS(1420),
    [anon_sym_finally] = ACTIONS(1420),
    [anon_sym_int] = ACTIONS(1420),
    [anon_sym_boolean] = ACTIONS(1420),
    [anon_sym_char] = ACTIONS(1420),
    [anon_sym_short] = ACTIONS(1420),
    [anon_sym_long] = ACTIONS(1420),
    [anon_sym_float] = ACTIONS(1420),
    [anon_sym_double] = ACTIONS(1420),
    [anon_sym_void] = ACTIONS(1420),
    [anon_sym_public] = ACTIONS(1420),
    [anon_sym_protected] = ACTIONS(1420),
    [anon_sym_private] = ACTIONS(1420),
    [anon_sym_static] = ACTIONS(1420),
    [anon_sym_final] = ACTIONS(1420),
    [anon_sym_synchronized] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1418),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_new] = ACTIONS(1420),
  },
  [STATE(346)] = {
    [sym_comment] = STATE(346),
    [sym_groovy_doc] = STATE(346),
    [ts_builtin_sym_end] = ACTIONS(1350),
    [sym_identifier] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_DOT_AMP] = ACTIONS(1350),
    [anon_sym_DOT_AT] = ACTIONS(1350),
    [anon_sym_QMARK_DOT] = ACTIONS(1350),
    [anon_sym_STAR_DOT] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_STAR_COLON] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_import] = ACTIONS(1348),
    [anon_sym_as] = ACTIONS(1348),
    [anon_sym_package] = ACTIONS(1348),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_assert] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_LT_LT] = ACTIONS(1350),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_GT_GT_GT] = ACTIONS(1350),
    [anon_sym_DOT_DOT] = ACTIONS(1348),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_GT_EQ] = ACTIONS(1350),
    [anon_sym_in] = ACTIONS(1348),
    [anon_sym_BANGin] = ACTIONS(1348),
    [anon_sym_instanceof] = ACTIONS(1348),
    [anon_sym_BANGinstanceof] = ACTIONS(1350),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT_EQ_GT] = ACTIONS(1350),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1350),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1350),
    [anon_sym_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_AMP_AMP] = ACTIONS(1350),
    [anon_sym_PIPE_PIPE] = ACTIONS(1350),
    [anon_sym_QMARK_COLON] = ACTIONS(1350),
    [anon_sym_STAR_STAR] = ACTIONS(1350),
    [anon_sym_true] = ACTIONS(1348),
    [anon_sym_false] = ACTIONS(1348),
    [anon_sym_ATinterface] = ACTIONS(1350),
    [anon_sym_interface] = ACTIONS(1348),
    [anon_sym_class] = ACTIONS(1348),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_DASH_GT] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1348),
    [anon_sym_def] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_null] = ACTIONS(1348),
    [anon_sym_this] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [aux_sym__type_identifier_token1] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_else] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1350),
    [aux_sym_number_literal_token1] = ACTIONS(1348),
    [aux_sym_number_literal_token2] = ACTIONS(1350),
    [aux_sym_number_literal_token3] = ACTIONS(1350),
    [aux_sym_number_literal_token4] = ACTIONS(1348),
    [aux_sym_number_literal_token5] = ACTIONS(1350),
    [anon_sym_pipeline] = ACTIONS(1348),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1348),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1350),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1350),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_try] = ACTIONS(1348),
    [anon_sym_int] = ACTIONS(1348),
    [anon_sym_boolean] = ACTIONS(1348),
    [anon_sym_char] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_float] = ACTIONS(1348),
    [anon_sym_double] = ACTIONS(1348),
    [anon_sym_void] = ACTIONS(1348),
    [anon_sym_public] = ACTIONS(1348),
    [anon_sym_protected] = ACTIONS(1348),
    [anon_sym_private] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_final] = ACTIONS(1348),
    [anon_sym_synchronized] = ACTIONS(1348),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_new] = ACTIONS(1348),
  },
  [STATE(347)] = {
    [sym_comment] = STATE(347),
    [sym_groovy_doc] = STATE(347),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(348)] = {
    [sym_comment] = STATE(348),
    [sym_groovy_doc] = STATE(348),
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_DASH_GT] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_RPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_pipeline] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(349)] = {
    [sym_comment] = STATE(349),
    [sym_groovy_doc] = STATE(349),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(350)] = {
    [sym_comment] = STATE(350),
    [sym_groovy_doc] = STATE(350),
    [ts_builtin_sym_end] = ACTIONS(1390),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_DOT_AMP] = ACTIONS(1390),
    [anon_sym_DOT_AT] = ACTIONS(1390),
    [anon_sym_QMARK_DOT] = ACTIONS(1390),
    [anon_sym_STAR_DOT] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_DOT] = ACTIONS(1392),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_as] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_PERCENT] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT] = ACTIONS(1390),
    [anon_sym_GT_GT] = ACTIONS(1392),
    [anon_sym_GT_GT_GT] = ACTIONS(1390),
    [anon_sym_DOT_DOT] = ACTIONS(1392),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1392),
    [anon_sym_LT] = ACTIONS(1392),
    [anon_sym_LT_EQ] = ACTIONS(1392),
    [anon_sym_GT] = ACTIONS(1392),
    [anon_sym_GT_EQ] = ACTIONS(1390),
    [anon_sym_in] = ACTIONS(1392),
    [anon_sym_BANGin] = ACTIONS(1392),
    [anon_sym_instanceof] = ACTIONS(1392),
    [anon_sym_BANGinstanceof] = ACTIONS(1390),
    [anon_sym_EQ_EQ] = ACTIONS(1392),
    [anon_sym_BANG_EQ] = ACTIONS(1392),
    [anon_sym_LT_EQ_GT] = ACTIONS(1390),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1390),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1390),
    [anon_sym_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_AMP] = ACTIONS(1392),
    [anon_sym_CARET] = ACTIONS(1390),
    [anon_sym_PIPE] = ACTIONS(1392),
    [anon_sym_AMP_AMP] = ACTIONS(1390),
    [anon_sym_PIPE_PIPE] = ACTIONS(1390),
    [anon_sym_QMARK_COLON] = ACTIONS(1390),
    [anon_sym_STAR_STAR] = ACTIONS(1390),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_COMMA] = ACTIONS(1390),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_DASH_GT] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_RPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_pipeline] = ACTIONS(1392),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_QMARK] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(351)] = {
    [sym_comment] = STATE(351),
    [sym_groovy_doc] = STATE(351),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(352)] = {
    [sym_comment] = STATE(352),
    [sym_groovy_doc] = STATE(352),
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_DOT_AMP] = ACTIONS(1364),
    [anon_sym_DOT_AT] = ACTIONS(1364),
    [anon_sym_QMARK_DOT] = ACTIONS(1364),
    [anon_sym_STAR_DOT] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_as] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_PERCENT] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_LT_LT] = ACTIONS(1364),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_GT_GT_GT] = ACTIONS(1364),
    [anon_sym_DOT_DOT] = ACTIONS(1362),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_GT_EQ] = ACTIONS(1364),
    [anon_sym_in] = ACTIONS(1362),
    [anon_sym_BANGin] = ACTIONS(1362),
    [anon_sym_instanceof] = ACTIONS(1362),
    [anon_sym_BANGinstanceof] = ACTIONS(1364),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT_EQ_GT] = ACTIONS(1364),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1364),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1364),
    [anon_sym_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_AMP_AMP] = ACTIONS(1364),
    [anon_sym_PIPE_PIPE] = ACTIONS(1364),
    [anon_sym_QMARK_COLON] = ACTIONS(1364),
    [anon_sym_STAR_STAR] = ACTIONS(1364),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_pipeline] = ACTIONS(1362),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_catch] = ACTIONS(1362),
    [anon_sym_finally] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(353)] = {
    [sym_comment] = STATE(353),
    [sym_groovy_doc] = STATE(353),
    [ts_builtin_sym_end] = ACTIONS(1428),
    [sym_identifier] = ACTIONS(1430),
    [anon_sym_break] = ACTIONS(1430),
    [anon_sym_continue] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1428),
    [anon_sym_DOT_AMP] = ACTIONS(1428),
    [anon_sym_DOT_AT] = ACTIONS(1428),
    [anon_sym_QMARK_DOT] = ACTIONS(1428),
    [anon_sym_STAR_DOT] = ACTIONS(1428),
    [anon_sym_STAR] = ACTIONS(1430),
    [anon_sym_STAR_COLON] = ACTIONS(1428),
    [anon_sym_DOT] = ACTIONS(1430),
    [anon_sym_import] = ACTIONS(1430),
    [anon_sym_as] = ACTIONS(1430),
    [anon_sym_package] = ACTIONS(1430),
    [anon_sym_AT] = ACTIONS(1430),
    [anon_sym_assert] = ACTIONS(1430),
    [anon_sym_PLUS_PLUS] = ACTIONS(1428),
    [anon_sym_DASH_DASH] = ACTIONS(1428),
    [anon_sym_PERCENT] = ACTIONS(1428),
    [anon_sym_SLASH] = ACTIONS(1430),
    [anon_sym_PLUS] = ACTIONS(1430),
    [anon_sym_DASH] = ACTIONS(1430),
    [anon_sym_LT_LT] = ACTIONS(1428),
    [anon_sym_GT_GT] = ACTIONS(1430),
    [anon_sym_GT_GT_GT] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(1430),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1430),
    [anon_sym_LT] = ACTIONS(1430),
    [anon_sym_LT_EQ] = ACTIONS(1430),
    [anon_sym_GT] = ACTIONS(1430),
    [anon_sym_GT_EQ] = ACTIONS(1428),
    [anon_sym_in] = ACTIONS(1430),
    [anon_sym_BANGin] = ACTIONS(1430),
    [anon_sym_instanceof] = ACTIONS(1430),
    [anon_sym_BANGinstanceof] = ACTIONS(1428),
    [anon_sym_EQ_EQ] = ACTIONS(1430),
    [anon_sym_BANG_EQ] = ACTIONS(1430),
    [anon_sym_LT_EQ_GT] = ACTIONS(1428),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1428),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1428),
    [anon_sym_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_AMP] = ACTIONS(1430),
    [anon_sym_CARET] = ACTIONS(1428),
    [anon_sym_PIPE] = ACTIONS(1430),
    [anon_sym_AMP_AMP] = ACTIONS(1428),
    [anon_sym_PIPE_PIPE] = ACTIONS(1428),
    [anon_sym_QMARK_COLON] = ACTIONS(1428),
    [anon_sym_STAR_STAR] = ACTIONS(1428),
    [anon_sym_true] = ACTIONS(1430),
    [anon_sym_false] = ACTIONS(1430),
    [anon_sym_ATinterface] = ACTIONS(1428),
    [anon_sym_interface] = ACTIONS(1430),
    [anon_sym_class] = ACTIONS(1430),
    [anon_sym_COMMA] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_DASH_GT] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1430),
    [anon_sym_def] = ACTIONS(1430),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_RPAREN] = ACTIONS(1428),
    [anon_sym_null] = ACTIONS(1430),
    [anon_sym_this] = ACTIONS(1430),
    [anon_sym_do] = ACTIONS(1430),
    [anon_sym_while] = ACTIONS(1430),
    [anon_sym_for] = ACTIONS(1430),
    [aux_sym__type_identifier_token1] = ACTIONS(1430),
    [anon_sym_if] = ACTIONS(1430),
    [anon_sym_else] = ACTIONS(1430),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [aux_sym_number_literal_token1] = ACTIONS(1430),
    [aux_sym_number_literal_token2] = ACTIONS(1428),
    [aux_sym_number_literal_token3] = ACTIONS(1428),
    [aux_sym_number_literal_token4] = ACTIONS(1430),
    [aux_sym_number_literal_token5] = ACTIONS(1428),
    [anon_sym_pipeline] = ACTIONS(1430),
    [anon_sym_return] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1428),
    [anon_sym_DQUOTE] = ACTIONS(1430),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1428),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1428),
    [anon_sym_switch] = ACTIONS(1430),
    [anon_sym_QMARK] = ACTIONS(1430),
    [anon_sym_try] = ACTIONS(1430),
    [anon_sym_int] = ACTIONS(1430),
    [anon_sym_boolean] = ACTIONS(1430),
    [anon_sym_char] = ACTIONS(1430),
    [anon_sym_short] = ACTIONS(1430),
    [anon_sym_long] = ACTIONS(1430),
    [anon_sym_float] = ACTIONS(1430),
    [anon_sym_double] = ACTIONS(1430),
    [anon_sym_void] = ACTIONS(1430),
    [anon_sym_public] = ACTIONS(1430),
    [anon_sym_protected] = ACTIONS(1430),
    [anon_sym_private] = ACTIONS(1430),
    [anon_sym_static] = ACTIONS(1430),
    [anon_sym_final] = ACTIONS(1430),
    [anon_sym_synchronized] = ACTIONS(1430),
    [anon_sym_TILDE] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1430),
    [anon_sym_new] = ACTIONS(1430),
  },
  [STATE(354)] = {
    [sym_comment] = STATE(354),
    [sym_groovy_doc] = STATE(354),
    [ts_builtin_sym_end] = ACTIONS(391),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(391),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(391),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(391),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(391),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_pipeline] = ACTIONS(393),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_catch] = ACTIONS(393),
    [anon_sym_finally] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(355)] = {
    [sym_comment] = STATE(355),
    [sym_groovy_doc] = STATE(355),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(356)] = {
    [sym_comment] = STATE(356),
    [sym_groovy_doc] = STATE(356),
    [sym_identifier] = ACTIONS(1388),
    [anon_sym_break] = ACTIONS(1388),
    [anon_sym_continue] = ACTIONS(1388),
    [anon_sym_SEMI] = ACTIONS(1386),
    [anon_sym_DOT_AMP] = ACTIONS(1386),
    [anon_sym_DOT_AT] = ACTIONS(1386),
    [anon_sym_QMARK_DOT] = ACTIONS(1386),
    [anon_sym_STAR_DOT] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_STAR_COLON] = ACTIONS(1386),
    [anon_sym_DOT] = ACTIONS(1388),
    [anon_sym_import] = ACTIONS(1388),
    [anon_sym_as] = ACTIONS(1388),
    [anon_sym_package] = ACTIONS(1388),
    [anon_sym_AT] = ACTIONS(1388),
    [anon_sym_assert] = ACTIONS(1388),
    [anon_sym_PLUS_PLUS] = ACTIONS(1386),
    [anon_sym_DASH_DASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1388),
    [anon_sym_PLUS] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1388),
    [anon_sym_GT_GT_GT] = ACTIONS(1386),
    [anon_sym_DOT_DOT] = ACTIONS(1388),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1388),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_LT_EQ] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_EQ] = ACTIONS(1386),
    [anon_sym_in] = ACTIONS(1388),
    [anon_sym_BANGin] = ACTIONS(1388),
    [anon_sym_instanceof] = ACTIONS(1388),
    [anon_sym_BANGinstanceof] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1388),
    [anon_sym_BANG_EQ] = ACTIONS(1388),
    [anon_sym_LT_EQ_GT] = ACTIONS(1386),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1386),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1386),
    [anon_sym_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_PIPE] = ACTIONS(1388),
    [anon_sym_AMP_AMP] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1386),
    [anon_sym_QMARK_COLON] = ACTIONS(1386),
    [anon_sym_STAR_STAR] = ACTIONS(1386),
    [anon_sym_true] = ACTIONS(1388),
    [anon_sym_false] = ACTIONS(1388),
    [anon_sym_ATinterface] = ACTIONS(1386),
    [anon_sym_interface] = ACTIONS(1388),
    [anon_sym_class] = ACTIONS(1388),
    [anon_sym_COMMA] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1386),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1388),
    [anon_sym_def] = ACTIONS(1388),
    [anon_sym_LPAREN] = ACTIONS(1386),
    [anon_sym_null] = ACTIONS(1388),
    [anon_sym_this] = ACTIONS(1388),
    [anon_sym_do] = ACTIONS(1388),
    [anon_sym_while] = ACTIONS(1388),
    [anon_sym_for] = ACTIONS(1388),
    [aux_sym__type_identifier_token1] = ACTIONS(1388),
    [anon_sym_if] = ACTIONS(1388),
    [anon_sym_else] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1386),
    [aux_sym_number_literal_token1] = ACTIONS(1388),
    [aux_sym_number_literal_token2] = ACTIONS(1386),
    [aux_sym_number_literal_token3] = ACTIONS(1386),
    [aux_sym_number_literal_token4] = ACTIONS(1388),
    [aux_sym_number_literal_token5] = ACTIONS(1386),
    [anon_sym_return] = ACTIONS(1388),
    [anon_sym_SQUOTE] = ACTIONS(1388),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1386),
    [anon_sym_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1386),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1386),
    [anon_sym_switch] = ACTIONS(1388),
    [anon_sym_case] = ACTIONS(1388),
    [anon_sym_default] = ACTIONS(1388),
    [anon_sym_QMARK] = ACTIONS(1388),
    [anon_sym_try] = ACTIONS(1388),
    [anon_sym_catch] = ACTIONS(1388),
    [anon_sym_finally] = ACTIONS(1388),
    [anon_sym_int] = ACTIONS(1388),
    [anon_sym_boolean] = ACTIONS(1388),
    [anon_sym_char] = ACTIONS(1388),
    [anon_sym_short] = ACTIONS(1388),
    [anon_sym_long] = ACTIONS(1388),
    [anon_sym_float] = ACTIONS(1388),
    [anon_sym_double] = ACTIONS(1388),
    [anon_sym_void] = ACTIONS(1388),
    [anon_sym_public] = ACTIONS(1388),
    [anon_sym_protected] = ACTIONS(1388),
    [anon_sym_private] = ACTIONS(1388),
    [anon_sym_static] = ACTIONS(1388),
    [anon_sym_final] = ACTIONS(1388),
    [anon_sym_synchronized] = ACTIONS(1388),
    [anon_sym_TILDE] = ACTIONS(1386),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_new] = ACTIONS(1388),
  },
  [STATE(357)] = {
    [sym_comment] = STATE(357),
    [sym_groovy_doc] = STATE(357),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(369),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(369),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(369),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(369),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(358)] = {
    [sym_comment] = STATE(358),
    [sym_groovy_doc] = STATE(358),
    [sym_identifier] = ACTIONS(1432),
    [anon_sym_break] = ACTIONS(1432),
    [anon_sym_continue] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1434),
    [anon_sym_DOT_AMP] = ACTIONS(1434),
    [anon_sym_DOT_AT] = ACTIONS(1434),
    [anon_sym_QMARK_DOT] = ACTIONS(1434),
    [anon_sym_STAR_DOT] = ACTIONS(1434),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_STAR_COLON] = ACTIONS(1434),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_import] = ACTIONS(1432),
    [anon_sym_as] = ACTIONS(1432),
    [anon_sym_package] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_assert] = ACTIONS(1432),
    [anon_sym_PLUS_PLUS] = ACTIONS(1434),
    [anon_sym_DASH_DASH] = ACTIONS(1434),
    [anon_sym_PERCENT] = ACTIONS(1434),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_LT_LT] = ACTIONS(1434),
    [anon_sym_GT_GT] = ACTIONS(1432),
    [anon_sym_GT_GT_GT] = ACTIONS(1434),
    [anon_sym_DOT_DOT] = ACTIONS(1432),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_LT_EQ] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_GT_EQ] = ACTIONS(1434),
    [anon_sym_in] = ACTIONS(1432),
    [anon_sym_BANGin] = ACTIONS(1432),
    [anon_sym_instanceof] = ACTIONS(1432),
    [anon_sym_BANGinstanceof] = ACTIONS(1434),
    [anon_sym_EQ_EQ] = ACTIONS(1432),
    [anon_sym_BANG_EQ] = ACTIONS(1432),
    [anon_sym_LT_EQ_GT] = ACTIONS(1434),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1434),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1434),
    [anon_sym_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1434),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_AMP_AMP] = ACTIONS(1434),
    [anon_sym_PIPE_PIPE] = ACTIONS(1434),
    [anon_sym_QMARK_COLON] = ACTIONS(1434),
    [anon_sym_STAR_STAR] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1432),
    [anon_sym_false] = ACTIONS(1432),
    [anon_sym_ATinterface] = ACTIONS(1434),
    [anon_sym_interface] = ACTIONS(1432),
    [anon_sym_class] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(1434),
    [anon_sym_RBRACE] = ACTIONS(1434),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_def] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1434),
    [anon_sym_null] = ACTIONS(1432),
    [anon_sym_this] = ACTIONS(1432),
    [anon_sym_do] = ACTIONS(1432),
    [anon_sym_while] = ACTIONS(1432),
    [anon_sym_for] = ACTIONS(1432),
    [aux_sym__type_identifier_token1] = ACTIONS(1432),
    [anon_sym_if] = ACTIONS(1432),
    [anon_sym_else] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [aux_sym_number_literal_token1] = ACTIONS(1432),
    [aux_sym_number_literal_token2] = ACTIONS(1434),
    [aux_sym_number_literal_token3] = ACTIONS(1434),
    [aux_sym_number_literal_token4] = ACTIONS(1432),
    [aux_sym_number_literal_token5] = ACTIONS(1434),
    [anon_sym_return] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1434),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1434),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1434),
    [anon_sym_switch] = ACTIONS(1432),
    [anon_sym_case] = ACTIONS(1432),
    [anon_sym_default] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_try] = ACTIONS(1432),
    [anon_sym_catch] = ACTIONS(1432),
    [anon_sym_finally] = ACTIONS(1432),
    [anon_sym_int] = ACTIONS(1432),
    [anon_sym_boolean] = ACTIONS(1432),
    [anon_sym_char] = ACTIONS(1432),
    [anon_sym_short] = ACTIONS(1432),
    [anon_sym_long] = ACTIONS(1432),
    [anon_sym_float] = ACTIONS(1432),
    [anon_sym_double] = ACTIONS(1432),
    [anon_sym_void] = ACTIONS(1432),
    [anon_sym_public] = ACTIONS(1432),
    [anon_sym_protected] = ACTIONS(1432),
    [anon_sym_private] = ACTIONS(1432),
    [anon_sym_static] = ACTIONS(1432),
    [anon_sym_final] = ACTIONS(1432),
    [anon_sym_synchronized] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1434),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_new] = ACTIONS(1432),
  },
  [STATE(359)] = {
    [sym_comment] = STATE(359),
    [sym_groovy_doc] = STATE(359),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_case] = ACTIONS(1338),
    [anon_sym_default] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_catch] = ACTIONS(1338),
    [anon_sym_finally] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(360)] = {
    [sym_comment] = STATE(360),
    [sym_groovy_doc] = STATE(360),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(102),
    [anon_sym_finally] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(361)] = {
    [sym_comment] = STATE(361),
    [sym_groovy_doc] = STATE(361),
    [ts_builtin_sym_end] = ACTIONS(1434),
    [sym_identifier] = ACTIONS(1432),
    [anon_sym_break] = ACTIONS(1432),
    [anon_sym_continue] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1434),
    [anon_sym_DOT_AMP] = ACTIONS(1434),
    [anon_sym_DOT_AT] = ACTIONS(1434),
    [anon_sym_QMARK_DOT] = ACTIONS(1434),
    [anon_sym_STAR_DOT] = ACTIONS(1434),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_STAR_COLON] = ACTIONS(1434),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_import] = ACTIONS(1432),
    [anon_sym_as] = ACTIONS(1432),
    [anon_sym_package] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_assert] = ACTIONS(1432),
    [anon_sym_PLUS_PLUS] = ACTIONS(1434),
    [anon_sym_DASH_DASH] = ACTIONS(1434),
    [anon_sym_PERCENT] = ACTIONS(1434),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_LT_LT] = ACTIONS(1434),
    [anon_sym_GT_GT] = ACTIONS(1432),
    [anon_sym_GT_GT_GT] = ACTIONS(1434),
    [anon_sym_DOT_DOT] = ACTIONS(1432),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_LT_EQ] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_GT_EQ] = ACTIONS(1434),
    [anon_sym_in] = ACTIONS(1432),
    [anon_sym_BANGin] = ACTIONS(1432),
    [anon_sym_instanceof] = ACTIONS(1432),
    [anon_sym_BANGinstanceof] = ACTIONS(1434),
    [anon_sym_EQ_EQ] = ACTIONS(1432),
    [anon_sym_BANG_EQ] = ACTIONS(1432),
    [anon_sym_LT_EQ_GT] = ACTIONS(1434),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1434),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1434),
    [anon_sym_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1434),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_AMP_AMP] = ACTIONS(1434),
    [anon_sym_PIPE_PIPE] = ACTIONS(1434),
    [anon_sym_QMARK_COLON] = ACTIONS(1434),
    [anon_sym_STAR_STAR] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1432),
    [anon_sym_false] = ACTIONS(1432),
    [anon_sym_ATinterface] = ACTIONS(1434),
    [anon_sym_interface] = ACTIONS(1432),
    [anon_sym_class] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(1434),
    [anon_sym_RBRACE] = ACTIONS(1434),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_def] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1434),
    [anon_sym_null] = ACTIONS(1432),
    [anon_sym_this] = ACTIONS(1432),
    [anon_sym_do] = ACTIONS(1432),
    [anon_sym_while] = ACTIONS(1432),
    [anon_sym_for] = ACTIONS(1432),
    [aux_sym__type_identifier_token1] = ACTIONS(1432),
    [anon_sym_if] = ACTIONS(1432),
    [anon_sym_else] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [aux_sym_number_literal_token1] = ACTIONS(1432),
    [aux_sym_number_literal_token2] = ACTIONS(1434),
    [aux_sym_number_literal_token3] = ACTIONS(1434),
    [aux_sym_number_literal_token4] = ACTIONS(1432),
    [aux_sym_number_literal_token5] = ACTIONS(1434),
    [anon_sym_pipeline] = ACTIONS(1432),
    [anon_sym_return] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1434),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1434),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1434),
    [anon_sym_switch] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_try] = ACTIONS(1432),
    [anon_sym_catch] = ACTIONS(1432),
    [anon_sym_finally] = ACTIONS(1432),
    [anon_sym_int] = ACTIONS(1432),
    [anon_sym_boolean] = ACTIONS(1432),
    [anon_sym_char] = ACTIONS(1432),
    [anon_sym_short] = ACTIONS(1432),
    [anon_sym_long] = ACTIONS(1432),
    [anon_sym_float] = ACTIONS(1432),
    [anon_sym_double] = ACTIONS(1432),
    [anon_sym_void] = ACTIONS(1432),
    [anon_sym_public] = ACTIONS(1432),
    [anon_sym_protected] = ACTIONS(1432),
    [anon_sym_private] = ACTIONS(1432),
    [anon_sym_static] = ACTIONS(1432),
    [anon_sym_final] = ACTIONS(1432),
    [anon_sym_synchronized] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1434),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_new] = ACTIONS(1432),
  },
  [STATE(362)] = {
    [sym_comment] = STATE(362),
    [sym_groovy_doc] = STATE(362),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(363)] = {
    [sym_comment] = STATE(363),
    [sym_groovy_doc] = STATE(363),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(387),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(389),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(364)] = {
    [sym_comment] = STATE(364),
    [sym_groovy_doc] = STATE(364),
    [ts_builtin_sym_end] = ACTIONS(1368),
    [sym_identifier] = ACTIONS(1366),
    [anon_sym_break] = ACTIONS(1366),
    [anon_sym_continue] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1368),
    [anon_sym_DOT_AMP] = ACTIONS(1368),
    [anon_sym_DOT_AT] = ACTIONS(1368),
    [anon_sym_QMARK_DOT] = ACTIONS(1368),
    [anon_sym_STAR_DOT] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_STAR_COLON] = ACTIONS(1368),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_import] = ACTIONS(1366),
    [anon_sym_as] = ACTIONS(1366),
    [anon_sym_package] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_assert] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1368),
    [anon_sym_PERCENT] = ACTIONS(1368),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_LT_LT] = ACTIONS(1368),
    [anon_sym_GT_GT] = ACTIONS(1366),
    [anon_sym_GT_GT_GT] = ACTIONS(1368),
    [anon_sym_DOT_DOT] = ACTIONS(1366),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_LT_EQ] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_GT_EQ] = ACTIONS(1368),
    [anon_sym_in] = ACTIONS(1366),
    [anon_sym_BANGin] = ACTIONS(1366),
    [anon_sym_instanceof] = ACTIONS(1366),
    [anon_sym_BANGinstanceof] = ACTIONS(1368),
    [anon_sym_EQ_EQ] = ACTIONS(1366),
    [anon_sym_BANG_EQ] = ACTIONS(1366),
    [anon_sym_LT_EQ_GT] = ACTIONS(1368),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1368),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1368),
    [anon_sym_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1368),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_AMP_AMP] = ACTIONS(1368),
    [anon_sym_PIPE_PIPE] = ACTIONS(1368),
    [anon_sym_QMARK_COLON] = ACTIONS(1368),
    [anon_sym_STAR_STAR] = ACTIONS(1368),
    [anon_sym_true] = ACTIONS(1366),
    [anon_sym_false] = ACTIONS(1366),
    [anon_sym_ATinterface] = ACTIONS(1368),
    [anon_sym_interface] = ACTIONS(1366),
    [anon_sym_class] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_def] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1368),
    [anon_sym_null] = ACTIONS(1366),
    [anon_sym_this] = ACTIONS(1366),
    [anon_sym_do] = ACTIONS(1366),
    [anon_sym_while] = ACTIONS(1366),
    [anon_sym_for] = ACTIONS(1366),
    [aux_sym__type_identifier_token1] = ACTIONS(1366),
    [anon_sym_if] = ACTIONS(1366),
    [anon_sym_else] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [aux_sym_number_literal_token1] = ACTIONS(1366),
    [aux_sym_number_literal_token2] = ACTIONS(1368),
    [aux_sym_number_literal_token3] = ACTIONS(1368),
    [aux_sym_number_literal_token4] = ACTIONS(1366),
    [aux_sym_number_literal_token5] = ACTIONS(1368),
    [anon_sym_pipeline] = ACTIONS(1366),
    [anon_sym_return] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1368),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1368),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_catch] = ACTIONS(1366),
    [anon_sym_finally] = ACTIONS(1366),
    [anon_sym_int] = ACTIONS(1366),
    [anon_sym_boolean] = ACTIONS(1366),
    [anon_sym_char] = ACTIONS(1366),
    [anon_sym_short] = ACTIONS(1366),
    [anon_sym_long] = ACTIONS(1366),
    [anon_sym_float] = ACTIONS(1366),
    [anon_sym_double] = ACTIONS(1366),
    [anon_sym_void] = ACTIONS(1366),
    [anon_sym_public] = ACTIONS(1366),
    [anon_sym_protected] = ACTIONS(1366),
    [anon_sym_private] = ACTIONS(1366),
    [anon_sym_static] = ACTIONS(1366),
    [anon_sym_final] = ACTIONS(1366),
    [anon_sym_synchronized] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_new] = ACTIONS(1366),
  },
  [STATE(365)] = {
    [sym_comment] = STATE(365),
    [sym_groovy_doc] = STATE(365),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_catch] = ACTIONS(1306),
    [anon_sym_finally] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(366)] = {
    [sym_comment] = STATE(366),
    [sym_groovy_doc] = STATE(366),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_catch] = ACTIONS(1334),
    [anon_sym_finally] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(367)] = {
    [sym_comment] = STATE(367),
    [sym_groovy_doc] = STATE(367),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(368)] = {
    [sym_comment] = STATE(368),
    [sym_groovy_doc] = STATE(368),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(369)] = {
    [sym_comment] = STATE(369),
    [sym_groovy_doc] = STATE(369),
    [sym_identifier] = ACTIONS(1430),
    [anon_sym_break] = ACTIONS(1430),
    [anon_sym_continue] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1428),
    [anon_sym_DOT_AMP] = ACTIONS(1428),
    [anon_sym_DOT_AT] = ACTIONS(1428),
    [anon_sym_QMARK_DOT] = ACTIONS(1428),
    [anon_sym_STAR_DOT] = ACTIONS(1428),
    [anon_sym_STAR] = ACTIONS(1430),
    [anon_sym_STAR_COLON] = ACTIONS(1428),
    [anon_sym_DOT] = ACTIONS(1430),
    [anon_sym_import] = ACTIONS(1430),
    [anon_sym_as] = ACTIONS(1430),
    [anon_sym_package] = ACTIONS(1430),
    [anon_sym_AT] = ACTIONS(1430),
    [anon_sym_assert] = ACTIONS(1430),
    [anon_sym_PLUS_PLUS] = ACTIONS(1428),
    [anon_sym_DASH_DASH] = ACTIONS(1428),
    [anon_sym_PERCENT] = ACTIONS(1428),
    [anon_sym_SLASH] = ACTIONS(1430),
    [anon_sym_PLUS] = ACTIONS(1430),
    [anon_sym_DASH] = ACTIONS(1430),
    [anon_sym_LT_LT] = ACTIONS(1428),
    [anon_sym_GT_GT] = ACTIONS(1430),
    [anon_sym_GT_GT_GT] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(1430),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1430),
    [anon_sym_LT] = ACTIONS(1430),
    [anon_sym_LT_EQ] = ACTIONS(1430),
    [anon_sym_GT] = ACTIONS(1430),
    [anon_sym_GT_EQ] = ACTIONS(1428),
    [anon_sym_in] = ACTIONS(1430),
    [anon_sym_BANGin] = ACTIONS(1430),
    [anon_sym_instanceof] = ACTIONS(1430),
    [anon_sym_BANGinstanceof] = ACTIONS(1428),
    [anon_sym_EQ_EQ] = ACTIONS(1430),
    [anon_sym_BANG_EQ] = ACTIONS(1430),
    [anon_sym_LT_EQ_GT] = ACTIONS(1428),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1428),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1428),
    [anon_sym_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_AMP] = ACTIONS(1430),
    [anon_sym_CARET] = ACTIONS(1428),
    [anon_sym_PIPE] = ACTIONS(1430),
    [anon_sym_AMP_AMP] = ACTIONS(1428),
    [anon_sym_PIPE_PIPE] = ACTIONS(1428),
    [anon_sym_QMARK_COLON] = ACTIONS(1428),
    [anon_sym_STAR_STAR] = ACTIONS(1428),
    [anon_sym_true] = ACTIONS(1430),
    [anon_sym_false] = ACTIONS(1430),
    [anon_sym_ATinterface] = ACTIONS(1428),
    [anon_sym_interface] = ACTIONS(1430),
    [anon_sym_class] = ACTIONS(1430),
    [anon_sym_COMMA] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1430),
    [anon_sym_def] = ACTIONS(1430),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_null] = ACTIONS(1430),
    [anon_sym_this] = ACTIONS(1430),
    [anon_sym_do] = ACTIONS(1430),
    [anon_sym_while] = ACTIONS(1430),
    [anon_sym_for] = ACTIONS(1430),
    [aux_sym__type_identifier_token1] = ACTIONS(1430),
    [anon_sym_if] = ACTIONS(1430),
    [anon_sym_else] = ACTIONS(1430),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [aux_sym_number_literal_token1] = ACTIONS(1430),
    [aux_sym_number_literal_token2] = ACTIONS(1428),
    [aux_sym_number_literal_token3] = ACTIONS(1428),
    [aux_sym_number_literal_token4] = ACTIONS(1430),
    [aux_sym_number_literal_token5] = ACTIONS(1428),
    [anon_sym_return] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1428),
    [anon_sym_DQUOTE] = ACTIONS(1430),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1428),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1428),
    [anon_sym_switch] = ACTIONS(1430),
    [anon_sym_case] = ACTIONS(1430),
    [anon_sym_default] = ACTIONS(1430),
    [anon_sym_QMARK] = ACTIONS(1430),
    [anon_sym_try] = ACTIONS(1430),
    [anon_sym_catch] = ACTIONS(1430),
    [anon_sym_finally] = ACTIONS(1430),
    [anon_sym_int] = ACTIONS(1430),
    [anon_sym_boolean] = ACTIONS(1430),
    [anon_sym_char] = ACTIONS(1430),
    [anon_sym_short] = ACTIONS(1430),
    [anon_sym_long] = ACTIONS(1430),
    [anon_sym_float] = ACTIONS(1430),
    [anon_sym_double] = ACTIONS(1430),
    [anon_sym_void] = ACTIONS(1430),
    [anon_sym_public] = ACTIONS(1430),
    [anon_sym_protected] = ACTIONS(1430),
    [anon_sym_private] = ACTIONS(1430),
    [anon_sym_static] = ACTIONS(1430),
    [anon_sym_final] = ACTIONS(1430),
    [anon_sym_synchronized] = ACTIONS(1430),
    [anon_sym_TILDE] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1430),
    [anon_sym_new] = ACTIONS(1430),
  },
  [STATE(370)] = {
    [sym_comment] = STATE(370),
    [sym_groovy_doc] = STATE(370),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(391),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(391),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(391),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(391),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_case] = ACTIONS(393),
    [anon_sym_default] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_catch] = ACTIONS(393),
    [anon_sym_finally] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(371)] = {
    [sym_comment] = STATE(371),
    [sym_groovy_doc] = STATE(371),
    [ts_builtin_sym_end] = ACTIONS(1314),
    [sym_identifier] = ACTIONS(1316),
    [anon_sym_break] = ACTIONS(1316),
    [anon_sym_continue] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1314),
    [anon_sym_COLON] = ACTIONS(1314),
    [anon_sym_DOT_AMP] = ACTIONS(1314),
    [anon_sym_DOT_AT] = ACTIONS(1314),
    [anon_sym_QMARK_DOT] = ACTIONS(1314),
    [anon_sym_STAR_DOT] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_STAR_COLON] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1316),
    [anon_sym_import] = ACTIONS(1316),
    [anon_sym_as] = ACTIONS(1316),
    [anon_sym_package] = ACTIONS(1316),
    [anon_sym_AT] = ACTIONS(1316),
    [anon_sym_assert] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1314),
    [anon_sym_PERCENT] = ACTIONS(1314),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_LT_LT] = ACTIONS(1314),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_GT_GT_GT] = ACTIONS(1314),
    [anon_sym_DOT_DOT] = ACTIONS(1316),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_GT_EQ] = ACTIONS(1314),
    [anon_sym_in] = ACTIONS(1316),
    [anon_sym_BANGin] = ACTIONS(1316),
    [anon_sym_instanceof] = ACTIONS(1316),
    [anon_sym_BANGinstanceof] = ACTIONS(1314),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT_EQ_GT] = ACTIONS(1314),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1314),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1314),
    [anon_sym_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1314),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_AMP_AMP] = ACTIONS(1314),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_QMARK_COLON] = ACTIONS(1314),
    [anon_sym_STAR_STAR] = ACTIONS(1314),
    [anon_sym_true] = ACTIONS(1316),
    [anon_sym_false] = ACTIONS(1316),
    [anon_sym_ATinterface] = ACTIONS(1314),
    [anon_sym_interface] = ACTIONS(1316),
    [anon_sym_class] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1314),
    [anon_sym_DASH_GT] = ACTIONS(1314),
    [anon_sym_RBRACE] = ACTIONS(1314),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1316),
    [anon_sym_def] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1314),
    [anon_sym_null] = ACTIONS(1316),
    [anon_sym_this] = ACTIONS(1316),
    [anon_sym_do] = ACTIONS(1316),
    [anon_sym_while] = ACTIONS(1316),
    [anon_sym_for] = ACTIONS(1316),
    [aux_sym__type_identifier_token1] = ACTIONS(1316),
    [anon_sym_if] = ACTIONS(1316),
    [anon_sym_else] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [aux_sym_number_literal_token1] = ACTIONS(1316),
    [aux_sym_number_literal_token2] = ACTIONS(1314),
    [aux_sym_number_literal_token3] = ACTIONS(1314),
    [aux_sym_number_literal_token4] = ACTIONS(1316),
    [aux_sym_number_literal_token5] = ACTIONS(1314),
    [anon_sym_pipeline] = ACTIONS(1316),
    [anon_sym_return] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1314),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1314),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1316),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_try] = ACTIONS(1316),
    [anon_sym_int] = ACTIONS(1316),
    [anon_sym_boolean] = ACTIONS(1316),
    [anon_sym_char] = ACTIONS(1316),
    [anon_sym_short] = ACTIONS(1316),
    [anon_sym_long] = ACTIONS(1316),
    [anon_sym_float] = ACTIONS(1316),
    [anon_sym_double] = ACTIONS(1316),
    [anon_sym_void] = ACTIONS(1316),
    [anon_sym_public] = ACTIONS(1316),
    [anon_sym_protected] = ACTIONS(1316),
    [anon_sym_private] = ACTIONS(1316),
    [anon_sym_static] = ACTIONS(1316),
    [anon_sym_final] = ACTIONS(1316),
    [anon_sym_synchronized] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1314),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_new] = ACTIONS(1316),
  },
  [STATE(372)] = {
    [sym_comment] = STATE(372),
    [sym_groovy_doc] = STATE(372),
    [ts_builtin_sym_end] = ACTIONS(1298),
    [sym_identifier] = ACTIONS(1296),
    [anon_sym_break] = ACTIONS(1296),
    [anon_sym_continue] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1298),
    [anon_sym_COLON] = ACTIONS(1298),
    [anon_sym_DOT_AMP] = ACTIONS(1298),
    [anon_sym_DOT_AT] = ACTIONS(1298),
    [anon_sym_QMARK_DOT] = ACTIONS(1298),
    [anon_sym_STAR_DOT] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_STAR_COLON] = ACTIONS(1298),
    [anon_sym_DOT] = ACTIONS(1296),
    [anon_sym_import] = ACTIONS(1296),
    [anon_sym_as] = ACTIONS(1296),
    [anon_sym_package] = ACTIONS(1296),
    [anon_sym_AT] = ACTIONS(1296),
    [anon_sym_assert] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1298),
    [anon_sym_PERCENT] = ACTIONS(1298),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_LT_LT] = ACTIONS(1298),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_GT_GT_GT] = ACTIONS(1298),
    [anon_sym_DOT_DOT] = ACTIONS(1296),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_GT_EQ] = ACTIONS(1298),
    [anon_sym_in] = ACTIONS(1296),
    [anon_sym_BANGin] = ACTIONS(1296),
    [anon_sym_instanceof] = ACTIONS(1296),
    [anon_sym_BANGinstanceof] = ACTIONS(1298),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT_EQ_GT] = ACTIONS(1298),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1298),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1298),
    [anon_sym_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1298),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_AMP_AMP] = ACTIONS(1298),
    [anon_sym_PIPE_PIPE] = ACTIONS(1298),
    [anon_sym_QMARK_COLON] = ACTIONS(1298),
    [anon_sym_STAR_STAR] = ACTIONS(1298),
    [anon_sym_true] = ACTIONS(1296),
    [anon_sym_false] = ACTIONS(1296),
    [anon_sym_ATinterface] = ACTIONS(1298),
    [anon_sym_interface] = ACTIONS(1296),
    [anon_sym_class] = ACTIONS(1296),
    [anon_sym_COMMA] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_DASH_GT] = ACTIONS(1298),
    [anon_sym_RBRACE] = ACTIONS(1298),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1296),
    [anon_sym_def] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1298),
    [anon_sym_null] = ACTIONS(1296),
    [anon_sym_this] = ACTIONS(1296),
    [anon_sym_do] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(1296),
    [anon_sym_for] = ACTIONS(1296),
    [aux_sym__type_identifier_token1] = ACTIONS(1296),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_else] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1298),
    [aux_sym_number_literal_token1] = ACTIONS(1296),
    [aux_sym_number_literal_token2] = ACTIONS(1298),
    [aux_sym_number_literal_token3] = ACTIONS(1298),
    [aux_sym_number_literal_token4] = ACTIONS(1296),
    [aux_sym_number_literal_token5] = ACTIONS(1298),
    [anon_sym_pipeline] = ACTIONS(1296),
    [anon_sym_return] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1298),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1298),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1296),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_try] = ACTIONS(1296),
    [anon_sym_int] = ACTIONS(1296),
    [anon_sym_boolean] = ACTIONS(1296),
    [anon_sym_char] = ACTIONS(1296),
    [anon_sym_short] = ACTIONS(1296),
    [anon_sym_long] = ACTIONS(1296),
    [anon_sym_float] = ACTIONS(1296),
    [anon_sym_double] = ACTIONS(1296),
    [anon_sym_void] = ACTIONS(1296),
    [anon_sym_public] = ACTIONS(1296),
    [anon_sym_protected] = ACTIONS(1296),
    [anon_sym_private] = ACTIONS(1296),
    [anon_sym_static] = ACTIONS(1296),
    [anon_sym_final] = ACTIONS(1296),
    [anon_sym_synchronized] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1298),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_new] = ACTIONS(1296),
  },
  [STATE(373)] = {
    [sym_comment] = STATE(373),
    [sym_groovy_doc] = STATE(373),
    [ts_builtin_sym_end] = ACTIONS(1300),
    [sym_identifier] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_DOT_AMP] = ACTIONS(1300),
    [anon_sym_DOT_AT] = ACTIONS(1300),
    [anon_sym_QMARK_DOT] = ACTIONS(1300),
    [anon_sym_STAR_DOT] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1302),
    [anon_sym_STAR_COLON] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1302),
    [anon_sym_import] = ACTIONS(1302),
    [anon_sym_as] = ACTIONS(1302),
    [anon_sym_package] = ACTIONS(1302),
    [anon_sym_AT] = ACTIONS(1302),
    [anon_sym_assert] = ACTIONS(1302),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1302),
    [anon_sym_GT_GT_GT] = ACTIONS(1300),
    [anon_sym_DOT_DOT] = ACTIONS(1302),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1302),
    [anon_sym_LT] = ACTIONS(1302),
    [anon_sym_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT] = ACTIONS(1302),
    [anon_sym_GT_EQ] = ACTIONS(1300),
    [anon_sym_in] = ACTIONS(1302),
    [anon_sym_BANGin] = ACTIONS(1302),
    [anon_sym_instanceof] = ACTIONS(1302),
    [anon_sym_BANGinstanceof] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1302),
    [anon_sym_BANG_EQ] = ACTIONS(1302),
    [anon_sym_LT_EQ_GT] = ACTIONS(1300),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1300),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1302),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1302),
    [anon_sym_AMP_AMP] = ACTIONS(1300),
    [anon_sym_PIPE_PIPE] = ACTIONS(1300),
    [anon_sym_QMARK_COLON] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_true] = ACTIONS(1302),
    [anon_sym_false] = ACTIONS(1302),
    [anon_sym_ATinterface] = ACTIONS(1300),
    [anon_sym_interface] = ACTIONS(1302),
    [anon_sym_class] = ACTIONS(1302),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_DASH_GT] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1302),
    [anon_sym_def] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_null] = ACTIONS(1302),
    [anon_sym_this] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [aux_sym__type_identifier_token1] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_else] = ACTIONS(1302),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [aux_sym_number_literal_token1] = ACTIONS(1302),
    [aux_sym_number_literal_token2] = ACTIONS(1300),
    [aux_sym_number_literal_token3] = ACTIONS(1300),
    [aux_sym_number_literal_token4] = ACTIONS(1302),
    [aux_sym_number_literal_token5] = ACTIONS(1300),
    [anon_sym_pipeline] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1300),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_QMARK] = ACTIONS(1302),
    [anon_sym_try] = ACTIONS(1302),
    [anon_sym_int] = ACTIONS(1302),
    [anon_sym_boolean] = ACTIONS(1302),
    [anon_sym_char] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_float] = ACTIONS(1302),
    [anon_sym_double] = ACTIONS(1302),
    [anon_sym_void] = ACTIONS(1302),
    [anon_sym_public] = ACTIONS(1302),
    [anon_sym_protected] = ACTIONS(1302),
    [anon_sym_private] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_final] = ACTIONS(1302),
    [anon_sym_synchronized] = ACTIONS(1302),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1302),
    [anon_sym_new] = ACTIONS(1302),
  },
  [STATE(374)] = {
    [sym_comment] = STATE(374),
    [sym_groovy_doc] = STATE(374),
    [aux_sym_dotted_identifier_repeat1] = STATE(328),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1422),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1436),
    [anon_sym_DASH_DASH] = ACTIONS(1436),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_DASH_GT] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1426),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(375)] = {
    [sym_comment] = STATE(375),
    [sym_groovy_doc] = STATE(375),
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_DASH_GT] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_RPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_pipeline] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(376)] = {
    [sym_comment] = STATE(376),
    [sym_groovy_doc] = STATE(376),
    [aux_sym_dotted_identifier_repeat1] = STATE(376),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(1438),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_DASH_GT] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(377)] = {
    [sym_comment] = STATE(377),
    [sym_groovy_doc] = STATE(377),
    [ts_builtin_sym_end] = ACTIONS(1428),
    [sym_identifier] = ACTIONS(1430),
    [anon_sym_break] = ACTIONS(1430),
    [anon_sym_continue] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1428),
    [anon_sym_DOT_AMP] = ACTIONS(1428),
    [anon_sym_DOT_AT] = ACTIONS(1428),
    [anon_sym_QMARK_DOT] = ACTIONS(1428),
    [anon_sym_STAR_DOT] = ACTIONS(1428),
    [anon_sym_STAR] = ACTIONS(1430),
    [anon_sym_STAR_COLON] = ACTIONS(1428),
    [anon_sym_DOT] = ACTIONS(1430),
    [anon_sym_import] = ACTIONS(1430),
    [anon_sym_as] = ACTIONS(1430),
    [anon_sym_package] = ACTIONS(1430),
    [anon_sym_AT] = ACTIONS(1430),
    [anon_sym_assert] = ACTIONS(1430),
    [anon_sym_PLUS_PLUS] = ACTIONS(1428),
    [anon_sym_DASH_DASH] = ACTIONS(1428),
    [anon_sym_PERCENT] = ACTIONS(1428),
    [anon_sym_SLASH] = ACTIONS(1430),
    [anon_sym_PLUS] = ACTIONS(1430),
    [anon_sym_DASH] = ACTIONS(1430),
    [anon_sym_LT_LT] = ACTIONS(1428),
    [anon_sym_GT_GT] = ACTIONS(1430),
    [anon_sym_GT_GT_GT] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(1430),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1430),
    [anon_sym_LT] = ACTIONS(1430),
    [anon_sym_LT_EQ] = ACTIONS(1430),
    [anon_sym_GT] = ACTIONS(1430),
    [anon_sym_GT_EQ] = ACTIONS(1428),
    [anon_sym_in] = ACTIONS(1430),
    [anon_sym_BANGin] = ACTIONS(1430),
    [anon_sym_instanceof] = ACTIONS(1430),
    [anon_sym_BANGinstanceof] = ACTIONS(1428),
    [anon_sym_EQ_EQ] = ACTIONS(1430),
    [anon_sym_BANG_EQ] = ACTIONS(1430),
    [anon_sym_LT_EQ_GT] = ACTIONS(1428),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1428),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1428),
    [anon_sym_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_AMP] = ACTIONS(1430),
    [anon_sym_CARET] = ACTIONS(1428),
    [anon_sym_PIPE] = ACTIONS(1430),
    [anon_sym_AMP_AMP] = ACTIONS(1428),
    [anon_sym_PIPE_PIPE] = ACTIONS(1428),
    [anon_sym_QMARK_COLON] = ACTIONS(1428),
    [anon_sym_STAR_STAR] = ACTIONS(1428),
    [anon_sym_true] = ACTIONS(1430),
    [anon_sym_false] = ACTIONS(1430),
    [anon_sym_ATinterface] = ACTIONS(1428),
    [anon_sym_interface] = ACTIONS(1430),
    [anon_sym_class] = ACTIONS(1430),
    [anon_sym_COMMA] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1430),
    [anon_sym_def] = ACTIONS(1430),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_null] = ACTIONS(1430),
    [anon_sym_this] = ACTIONS(1430),
    [anon_sym_do] = ACTIONS(1430),
    [anon_sym_while] = ACTIONS(1430),
    [anon_sym_for] = ACTIONS(1430),
    [aux_sym__type_identifier_token1] = ACTIONS(1430),
    [anon_sym_if] = ACTIONS(1430),
    [anon_sym_else] = ACTIONS(1430),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [aux_sym_number_literal_token1] = ACTIONS(1430),
    [aux_sym_number_literal_token2] = ACTIONS(1428),
    [aux_sym_number_literal_token3] = ACTIONS(1428),
    [aux_sym_number_literal_token4] = ACTIONS(1430),
    [aux_sym_number_literal_token5] = ACTIONS(1428),
    [anon_sym_pipeline] = ACTIONS(1430),
    [anon_sym_return] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1428),
    [anon_sym_DQUOTE] = ACTIONS(1430),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1428),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1428),
    [anon_sym_switch] = ACTIONS(1430),
    [anon_sym_QMARK] = ACTIONS(1430),
    [anon_sym_try] = ACTIONS(1430),
    [anon_sym_catch] = ACTIONS(1430),
    [anon_sym_finally] = ACTIONS(1430),
    [anon_sym_int] = ACTIONS(1430),
    [anon_sym_boolean] = ACTIONS(1430),
    [anon_sym_char] = ACTIONS(1430),
    [anon_sym_short] = ACTIONS(1430),
    [anon_sym_long] = ACTIONS(1430),
    [anon_sym_float] = ACTIONS(1430),
    [anon_sym_double] = ACTIONS(1430),
    [anon_sym_void] = ACTIONS(1430),
    [anon_sym_public] = ACTIONS(1430),
    [anon_sym_protected] = ACTIONS(1430),
    [anon_sym_private] = ACTIONS(1430),
    [anon_sym_static] = ACTIONS(1430),
    [anon_sym_final] = ACTIONS(1430),
    [anon_sym_synchronized] = ACTIONS(1430),
    [anon_sym_TILDE] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1430),
    [anon_sym_new] = ACTIONS(1430),
  },
  [STATE(378)] = {
    [sym_comment] = STATE(378),
    [sym_groovy_doc] = STATE(378),
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_pipeline] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_catch] = ACTIONS(1306),
    [anon_sym_finally] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(379)] = {
    [sym_comment] = STATE(379),
    [sym_groovy_doc] = STATE(379),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_DOT_AMP] = ACTIONS(1374),
    [anon_sym_DOT_AT] = ACTIONS(1374),
    [anon_sym_QMARK_DOT] = ACTIONS(1374),
    [anon_sym_STAR_DOT] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1376),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_as] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_LT_LT] = ACTIONS(1374),
    [anon_sym_GT_GT] = ACTIONS(1376),
    [anon_sym_GT_GT_GT] = ACTIONS(1374),
    [anon_sym_DOT_DOT] = ACTIONS(1376),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1376),
    [anon_sym_LT] = ACTIONS(1376),
    [anon_sym_LT_EQ] = ACTIONS(1376),
    [anon_sym_GT] = ACTIONS(1376),
    [anon_sym_GT_EQ] = ACTIONS(1374),
    [anon_sym_in] = ACTIONS(1376),
    [anon_sym_BANGin] = ACTIONS(1376),
    [anon_sym_instanceof] = ACTIONS(1376),
    [anon_sym_BANGinstanceof] = ACTIONS(1374),
    [anon_sym_EQ_EQ] = ACTIONS(1376),
    [anon_sym_BANG_EQ] = ACTIONS(1376),
    [anon_sym_LT_EQ_GT] = ACTIONS(1374),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1374),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1374),
    [anon_sym_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1376),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1376),
    [anon_sym_AMP_AMP] = ACTIONS(1374),
    [anon_sym_PIPE_PIPE] = ACTIONS(1374),
    [anon_sym_QMARK_COLON] = ACTIONS(1374),
    [anon_sym_STAR_STAR] = ACTIONS(1374),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_case] = ACTIONS(1376),
    [anon_sym_default] = ACTIONS(1376),
    [anon_sym_QMARK] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_catch] = ACTIONS(1376),
    [anon_sym_finally] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(380)] = {
    [sym_comment] = STATE(380),
    [sym_groovy_doc] = STATE(380),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_DOT_AMP] = ACTIONS(1390),
    [anon_sym_DOT_AT] = ACTIONS(1390),
    [anon_sym_QMARK_DOT] = ACTIONS(1390),
    [anon_sym_STAR_DOT] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_DOT] = ACTIONS(1392),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_as] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_PERCENT] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT] = ACTIONS(1390),
    [anon_sym_GT_GT] = ACTIONS(1392),
    [anon_sym_GT_GT_GT] = ACTIONS(1390),
    [anon_sym_DOT_DOT] = ACTIONS(1392),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1392),
    [anon_sym_LT] = ACTIONS(1392),
    [anon_sym_LT_EQ] = ACTIONS(1392),
    [anon_sym_GT] = ACTIONS(1392),
    [anon_sym_GT_EQ] = ACTIONS(1390),
    [anon_sym_in] = ACTIONS(1392),
    [anon_sym_BANGin] = ACTIONS(1392),
    [anon_sym_instanceof] = ACTIONS(1392),
    [anon_sym_BANGinstanceof] = ACTIONS(1390),
    [anon_sym_EQ_EQ] = ACTIONS(1392),
    [anon_sym_BANG_EQ] = ACTIONS(1392),
    [anon_sym_LT_EQ_GT] = ACTIONS(1390),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1390),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1390),
    [anon_sym_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_AMP] = ACTIONS(1392),
    [anon_sym_CARET] = ACTIONS(1390),
    [anon_sym_PIPE] = ACTIONS(1392),
    [anon_sym_AMP_AMP] = ACTIONS(1390),
    [anon_sym_PIPE_PIPE] = ACTIONS(1390),
    [anon_sym_QMARK_COLON] = ACTIONS(1390),
    [anon_sym_STAR_STAR] = ACTIONS(1390),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_COMMA] = ACTIONS(1390),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_case] = ACTIONS(1392),
    [anon_sym_default] = ACTIONS(1392),
    [anon_sym_QMARK] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_catch] = ACTIONS(1392),
    [anon_sym_finally] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(381)] = {
    [sym_comment] = STATE(381),
    [sym_groovy_doc] = STATE(381),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_DOT_AMP] = ACTIONS(1410),
    [anon_sym_DOT_AT] = ACTIONS(1410),
    [anon_sym_QMARK_DOT] = ACTIONS(1410),
    [anon_sym_STAR_DOT] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_DOT] = ACTIONS(1412),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_as] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PERCENT] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_LT_LT] = ACTIONS(1410),
    [anon_sym_GT_GT] = ACTIONS(1412),
    [anon_sym_GT_GT_GT] = ACTIONS(1410),
    [anon_sym_DOT_DOT] = ACTIONS(1412),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1412),
    [anon_sym_LT] = ACTIONS(1412),
    [anon_sym_LT_EQ] = ACTIONS(1412),
    [anon_sym_GT] = ACTIONS(1412),
    [anon_sym_GT_EQ] = ACTIONS(1410),
    [anon_sym_in] = ACTIONS(1412),
    [anon_sym_BANGin] = ACTIONS(1412),
    [anon_sym_instanceof] = ACTIONS(1412),
    [anon_sym_BANGinstanceof] = ACTIONS(1410),
    [anon_sym_EQ_EQ] = ACTIONS(1412),
    [anon_sym_BANG_EQ] = ACTIONS(1412),
    [anon_sym_LT_EQ_GT] = ACTIONS(1410),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1410),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1410),
    [anon_sym_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_AMP] = ACTIONS(1412),
    [anon_sym_CARET] = ACTIONS(1410),
    [anon_sym_PIPE] = ACTIONS(1412),
    [anon_sym_AMP_AMP] = ACTIONS(1410),
    [anon_sym_PIPE_PIPE] = ACTIONS(1410),
    [anon_sym_QMARK_COLON] = ACTIONS(1410),
    [anon_sym_STAR_STAR] = ACTIONS(1410),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_COMMA] = ACTIONS(1410),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_case] = ACTIONS(1412),
    [anon_sym_default] = ACTIONS(1412),
    [anon_sym_QMARK] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_catch] = ACTIONS(1412),
    [anon_sym_finally] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(382)] = {
    [sym_comment] = STATE(382),
    [sym_groovy_doc] = STATE(382),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1441),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(383)] = {
    [sym_comment] = STATE(383),
    [sym_groovy_doc] = STATE(383),
    [sym_argument_list] = STATE(466),
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_DASH_GT] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1445),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_pipeline] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(384)] = {
    [sym_comment] = STATE(384),
    [sym_groovy_doc] = STATE(384),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_COLON] = ACTIONS(1441),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(385)] = {
    [sym_comment] = STATE(385),
    [sym_groovy_doc] = STATE(385),
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_pipeline] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_catch] = ACTIONS(1334),
    [anon_sym_finally] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(386)] = {
    [sym_comment] = STATE(386),
    [sym_groovy_doc] = STATE(386),
    [sym_identifier] = ACTIONS(1296),
    [anon_sym_break] = ACTIONS(1296),
    [anon_sym_continue] = ACTIONS(1296),
    [anon_sym_SEMI] = ACTIONS(1298),
    [anon_sym_COLON] = ACTIONS(1298),
    [anon_sym_DOT_AMP] = ACTIONS(1298),
    [anon_sym_DOT_AT] = ACTIONS(1298),
    [anon_sym_QMARK_DOT] = ACTIONS(1298),
    [anon_sym_STAR_DOT] = ACTIONS(1298),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_STAR_COLON] = ACTIONS(1298),
    [anon_sym_DOT] = ACTIONS(1296),
    [anon_sym_import] = ACTIONS(1296),
    [anon_sym_as] = ACTIONS(1296),
    [anon_sym_package] = ACTIONS(1296),
    [anon_sym_AT] = ACTIONS(1296),
    [anon_sym_assert] = ACTIONS(1296),
    [anon_sym_PLUS_PLUS] = ACTIONS(1298),
    [anon_sym_DASH_DASH] = ACTIONS(1298),
    [anon_sym_PERCENT] = ACTIONS(1298),
    [anon_sym_SLASH] = ACTIONS(1296),
    [anon_sym_PLUS] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_LT_LT] = ACTIONS(1298),
    [anon_sym_GT_GT] = ACTIONS(1296),
    [anon_sym_GT_GT_GT] = ACTIONS(1298),
    [anon_sym_DOT_DOT] = ACTIONS(1296),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1298),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1296),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_LT_EQ] = ACTIONS(1296),
    [anon_sym_GT] = ACTIONS(1296),
    [anon_sym_GT_EQ] = ACTIONS(1298),
    [anon_sym_in] = ACTIONS(1296),
    [anon_sym_BANGin] = ACTIONS(1296),
    [anon_sym_instanceof] = ACTIONS(1296),
    [anon_sym_BANGinstanceof] = ACTIONS(1298),
    [anon_sym_EQ_EQ] = ACTIONS(1296),
    [anon_sym_BANG_EQ] = ACTIONS(1296),
    [anon_sym_LT_EQ_GT] = ACTIONS(1298),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1298),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1298),
    [anon_sym_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1298),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_CARET] = ACTIONS(1298),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_AMP_AMP] = ACTIONS(1298),
    [anon_sym_PIPE_PIPE] = ACTIONS(1298),
    [anon_sym_QMARK_COLON] = ACTIONS(1298),
    [anon_sym_STAR_STAR] = ACTIONS(1298),
    [anon_sym_true] = ACTIONS(1296),
    [anon_sym_false] = ACTIONS(1296),
    [anon_sym_ATinterface] = ACTIONS(1298),
    [anon_sym_interface] = ACTIONS(1296),
    [anon_sym_class] = ACTIONS(1296),
    [anon_sym_COMMA] = ACTIONS(1298),
    [anon_sym_LBRACE] = ACTIONS(1298),
    [anon_sym_RBRACE] = ACTIONS(1298),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1296),
    [anon_sym_def] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1298),
    [anon_sym_null] = ACTIONS(1296),
    [anon_sym_this] = ACTIONS(1296),
    [anon_sym_do] = ACTIONS(1296),
    [anon_sym_while] = ACTIONS(1296),
    [anon_sym_for] = ACTIONS(1296),
    [aux_sym__type_identifier_token1] = ACTIONS(1296),
    [anon_sym_if] = ACTIONS(1296),
    [anon_sym_else] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1298),
    [aux_sym_number_literal_token1] = ACTIONS(1296),
    [aux_sym_number_literal_token2] = ACTIONS(1298),
    [aux_sym_number_literal_token3] = ACTIONS(1298),
    [aux_sym_number_literal_token4] = ACTIONS(1296),
    [aux_sym_number_literal_token5] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1296),
    [anon_sym_SQUOTE] = ACTIONS(1296),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1298),
    [anon_sym_DQUOTE] = ACTIONS(1296),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1298),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1298),
    [anon_sym_switch] = ACTIONS(1296),
    [anon_sym_case] = ACTIONS(1296),
    [anon_sym_default] = ACTIONS(1296),
    [anon_sym_QMARK] = ACTIONS(1296),
    [anon_sym_try] = ACTIONS(1296),
    [anon_sym_int] = ACTIONS(1296),
    [anon_sym_boolean] = ACTIONS(1296),
    [anon_sym_char] = ACTIONS(1296),
    [anon_sym_short] = ACTIONS(1296),
    [anon_sym_long] = ACTIONS(1296),
    [anon_sym_float] = ACTIONS(1296),
    [anon_sym_double] = ACTIONS(1296),
    [anon_sym_void] = ACTIONS(1296),
    [anon_sym_public] = ACTIONS(1296),
    [anon_sym_protected] = ACTIONS(1296),
    [anon_sym_private] = ACTIONS(1296),
    [anon_sym_static] = ACTIONS(1296),
    [anon_sym_final] = ACTIONS(1296),
    [anon_sym_synchronized] = ACTIONS(1296),
    [anon_sym_TILDE] = ACTIONS(1298),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_new] = ACTIONS(1296),
  },
  [STATE(387)] = {
    [sym_comment] = STATE(387),
    [sym_groovy_doc] = STATE(387),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(388)] = {
    [sym_comment] = STATE(388),
    [sym_groovy_doc] = STATE(388),
    [sym_identifier] = ACTIONS(1328),
    [anon_sym_break] = ACTIONS(1328),
    [anon_sym_continue] = ACTIONS(1328),
    [anon_sym_SEMI] = ACTIONS(1330),
    [anon_sym_COLON] = ACTIONS(1330),
    [anon_sym_DOT_AMP] = ACTIONS(1330),
    [anon_sym_DOT_AT] = ACTIONS(1330),
    [anon_sym_QMARK_DOT] = ACTIONS(1330),
    [anon_sym_STAR_DOT] = ACTIONS(1330),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_STAR_COLON] = ACTIONS(1330),
    [anon_sym_DOT] = ACTIONS(1328),
    [anon_sym_import] = ACTIONS(1328),
    [anon_sym_as] = ACTIONS(1328),
    [anon_sym_package] = ACTIONS(1328),
    [anon_sym_AT] = ACTIONS(1328),
    [anon_sym_assert] = ACTIONS(1328),
    [anon_sym_PLUS_PLUS] = ACTIONS(1330),
    [anon_sym_DASH_DASH] = ACTIONS(1330),
    [anon_sym_PERCENT] = ACTIONS(1330),
    [anon_sym_SLASH] = ACTIONS(1328),
    [anon_sym_PLUS] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_LT_LT] = ACTIONS(1330),
    [anon_sym_GT_GT] = ACTIONS(1328),
    [anon_sym_GT_GT_GT] = ACTIONS(1330),
    [anon_sym_DOT_DOT] = ACTIONS(1328),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1330),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1328),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_LT_EQ] = ACTIONS(1328),
    [anon_sym_GT] = ACTIONS(1328),
    [anon_sym_GT_EQ] = ACTIONS(1330),
    [anon_sym_in] = ACTIONS(1328),
    [anon_sym_BANGin] = ACTIONS(1328),
    [anon_sym_instanceof] = ACTIONS(1328),
    [anon_sym_BANGinstanceof] = ACTIONS(1330),
    [anon_sym_EQ_EQ] = ACTIONS(1328),
    [anon_sym_BANG_EQ] = ACTIONS(1328),
    [anon_sym_LT_EQ_GT] = ACTIONS(1330),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1330),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1330),
    [anon_sym_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1330),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_CARET] = ACTIONS(1330),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_AMP_AMP] = ACTIONS(1330),
    [anon_sym_PIPE_PIPE] = ACTIONS(1330),
    [anon_sym_QMARK_COLON] = ACTIONS(1330),
    [anon_sym_STAR_STAR] = ACTIONS(1330),
    [anon_sym_true] = ACTIONS(1328),
    [anon_sym_false] = ACTIONS(1328),
    [anon_sym_ATinterface] = ACTIONS(1330),
    [anon_sym_interface] = ACTIONS(1328),
    [anon_sym_class] = ACTIONS(1328),
    [anon_sym_COMMA] = ACTIONS(1330),
    [anon_sym_LBRACE] = ACTIONS(1330),
    [anon_sym_RBRACE] = ACTIONS(1330),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1328),
    [anon_sym_def] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1330),
    [anon_sym_null] = ACTIONS(1328),
    [anon_sym_this] = ACTIONS(1328),
    [anon_sym_do] = ACTIONS(1328),
    [anon_sym_while] = ACTIONS(1328),
    [anon_sym_for] = ACTIONS(1328),
    [aux_sym__type_identifier_token1] = ACTIONS(1328),
    [anon_sym_if] = ACTIONS(1328),
    [anon_sym_else] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1330),
    [aux_sym_number_literal_token1] = ACTIONS(1328),
    [aux_sym_number_literal_token2] = ACTIONS(1330),
    [aux_sym_number_literal_token3] = ACTIONS(1330),
    [aux_sym_number_literal_token4] = ACTIONS(1328),
    [aux_sym_number_literal_token5] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1328),
    [anon_sym_SQUOTE] = ACTIONS(1328),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1330),
    [anon_sym_DQUOTE] = ACTIONS(1328),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1330),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1330),
    [anon_sym_switch] = ACTIONS(1328),
    [anon_sym_case] = ACTIONS(1328),
    [anon_sym_default] = ACTIONS(1328),
    [anon_sym_QMARK] = ACTIONS(1328),
    [anon_sym_try] = ACTIONS(1328),
    [anon_sym_int] = ACTIONS(1328),
    [anon_sym_boolean] = ACTIONS(1328),
    [anon_sym_char] = ACTIONS(1328),
    [anon_sym_short] = ACTIONS(1328),
    [anon_sym_long] = ACTIONS(1328),
    [anon_sym_float] = ACTIONS(1328),
    [anon_sym_double] = ACTIONS(1328),
    [anon_sym_void] = ACTIONS(1328),
    [anon_sym_public] = ACTIONS(1328),
    [anon_sym_protected] = ACTIONS(1328),
    [anon_sym_private] = ACTIONS(1328),
    [anon_sym_static] = ACTIONS(1328),
    [anon_sym_final] = ACTIONS(1328),
    [anon_sym_synchronized] = ACTIONS(1328),
    [anon_sym_TILDE] = ACTIONS(1330),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_new] = ACTIONS(1328),
  },
  [STATE(389)] = {
    [sym_comment] = STATE(389),
    [sym_groovy_doc] = STATE(389),
    [sym_argument_list] = STATE(547),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1447),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(390)] = {
    [sym_comment] = STATE(390),
    [sym_groovy_doc] = STATE(390),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_DASH_GT] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(391)] = {
    [sym_comment] = STATE(391),
    [sym_groovy_doc] = STATE(391),
    [sym_identifier] = ACTIONS(1348),
    [anon_sym_break] = ACTIONS(1348),
    [anon_sym_continue] = ACTIONS(1348),
    [anon_sym_SEMI] = ACTIONS(1350),
    [anon_sym_COLON] = ACTIONS(1350),
    [anon_sym_DOT_AMP] = ACTIONS(1350),
    [anon_sym_DOT_AT] = ACTIONS(1350),
    [anon_sym_QMARK_DOT] = ACTIONS(1350),
    [anon_sym_STAR_DOT] = ACTIONS(1350),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_STAR_COLON] = ACTIONS(1350),
    [anon_sym_DOT] = ACTIONS(1348),
    [anon_sym_import] = ACTIONS(1348),
    [anon_sym_as] = ACTIONS(1348),
    [anon_sym_package] = ACTIONS(1348),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_assert] = ACTIONS(1348),
    [anon_sym_PLUS_PLUS] = ACTIONS(1350),
    [anon_sym_DASH_DASH] = ACTIONS(1350),
    [anon_sym_PERCENT] = ACTIONS(1350),
    [anon_sym_SLASH] = ACTIONS(1348),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_LT_LT] = ACTIONS(1350),
    [anon_sym_GT_GT] = ACTIONS(1348),
    [anon_sym_GT_GT_GT] = ACTIONS(1350),
    [anon_sym_DOT_DOT] = ACTIONS(1348),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1350),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1348),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_LT_EQ] = ACTIONS(1348),
    [anon_sym_GT] = ACTIONS(1348),
    [anon_sym_GT_EQ] = ACTIONS(1350),
    [anon_sym_in] = ACTIONS(1348),
    [anon_sym_BANGin] = ACTIONS(1348),
    [anon_sym_instanceof] = ACTIONS(1348),
    [anon_sym_BANGinstanceof] = ACTIONS(1350),
    [anon_sym_EQ_EQ] = ACTIONS(1348),
    [anon_sym_BANG_EQ] = ACTIONS(1348),
    [anon_sym_LT_EQ_GT] = ACTIONS(1350),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1350),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1350),
    [anon_sym_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1350),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_CARET] = ACTIONS(1350),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_AMP_AMP] = ACTIONS(1350),
    [anon_sym_PIPE_PIPE] = ACTIONS(1350),
    [anon_sym_QMARK_COLON] = ACTIONS(1350),
    [anon_sym_STAR_STAR] = ACTIONS(1350),
    [anon_sym_true] = ACTIONS(1348),
    [anon_sym_false] = ACTIONS(1348),
    [anon_sym_ATinterface] = ACTIONS(1350),
    [anon_sym_interface] = ACTIONS(1348),
    [anon_sym_class] = ACTIONS(1348),
    [anon_sym_COMMA] = ACTIONS(1350),
    [anon_sym_LBRACE] = ACTIONS(1350),
    [anon_sym_RBRACE] = ACTIONS(1350),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1348),
    [anon_sym_def] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1350),
    [anon_sym_null] = ACTIONS(1348),
    [anon_sym_this] = ACTIONS(1348),
    [anon_sym_do] = ACTIONS(1348),
    [anon_sym_while] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1348),
    [aux_sym__type_identifier_token1] = ACTIONS(1348),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_else] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1350),
    [aux_sym_number_literal_token1] = ACTIONS(1348),
    [aux_sym_number_literal_token2] = ACTIONS(1350),
    [aux_sym_number_literal_token3] = ACTIONS(1350),
    [aux_sym_number_literal_token4] = ACTIONS(1348),
    [aux_sym_number_literal_token5] = ACTIONS(1350),
    [anon_sym_return] = ACTIONS(1348),
    [anon_sym_SQUOTE] = ACTIONS(1348),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1350),
    [anon_sym_DQUOTE] = ACTIONS(1348),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1350),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1350),
    [anon_sym_switch] = ACTIONS(1348),
    [anon_sym_case] = ACTIONS(1348),
    [anon_sym_default] = ACTIONS(1348),
    [anon_sym_QMARK] = ACTIONS(1348),
    [anon_sym_try] = ACTIONS(1348),
    [anon_sym_int] = ACTIONS(1348),
    [anon_sym_boolean] = ACTIONS(1348),
    [anon_sym_char] = ACTIONS(1348),
    [anon_sym_short] = ACTIONS(1348),
    [anon_sym_long] = ACTIONS(1348),
    [anon_sym_float] = ACTIONS(1348),
    [anon_sym_double] = ACTIONS(1348),
    [anon_sym_void] = ACTIONS(1348),
    [anon_sym_public] = ACTIONS(1348),
    [anon_sym_protected] = ACTIONS(1348),
    [anon_sym_private] = ACTIONS(1348),
    [anon_sym_static] = ACTIONS(1348),
    [anon_sym_final] = ACTIONS(1348),
    [anon_sym_synchronized] = ACTIONS(1348),
    [anon_sym_TILDE] = ACTIONS(1350),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_new] = ACTIONS(1348),
  },
  [STATE(392)] = {
    [sym_comment] = STATE(392),
    [sym_groovy_doc] = STATE(392),
    [sym_identifier] = ACTIONS(1449),
    [anon_sym_break] = ACTIONS(1449),
    [anon_sym_continue] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_DOT_AMP] = ACTIONS(1451),
    [anon_sym_DOT_AT] = ACTIONS(1451),
    [anon_sym_QMARK_DOT] = ACTIONS(1451),
    [anon_sym_STAR_DOT] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1449),
    [anon_sym_STAR_COLON] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1449),
    [anon_sym_as] = ACTIONS(1449),
    [anon_sym_package] = ACTIONS(1449),
    [anon_sym_AT] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1449),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_DOT_DOT] = ACTIONS(1449),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1449),
    [anon_sym_LT] = ACTIONS(1449),
    [anon_sym_LT_EQ] = ACTIONS(1449),
    [anon_sym_GT] = ACTIONS(1449),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1449),
    [anon_sym_BANGin] = ACTIONS(1449),
    [anon_sym_instanceof] = ACTIONS(1449),
    [anon_sym_BANGinstanceof] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1449),
    [anon_sym_BANG_EQ] = ACTIONS(1449),
    [anon_sym_LT_EQ_GT] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1449),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1449),
    [anon_sym_AMP_AMP] = ACTIONS(1451),
    [anon_sym_PIPE_PIPE] = ACTIONS(1451),
    [anon_sym_QMARK_COLON] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_true] = ACTIONS(1449),
    [anon_sym_false] = ACTIONS(1449),
    [anon_sym_ATinterface] = ACTIONS(1451),
    [anon_sym_interface] = ACTIONS(1449),
    [anon_sym_class] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [anon_sym_RBRACE] = ACTIONS(1451),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1449),
    [anon_sym_def] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_null] = ACTIONS(1449),
    [anon_sym_this] = ACTIONS(1449),
    [anon_sym_do] = ACTIONS(1449),
    [anon_sym_while] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1449),
    [aux_sym__type_identifier_token1] = ACTIONS(1449),
    [anon_sym_if] = ACTIONS(1449),
    [anon_sym_else] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_number_literal_token1] = ACTIONS(1449),
    [aux_sym_number_literal_token2] = ACTIONS(1451),
    [aux_sym_number_literal_token3] = ACTIONS(1451),
    [aux_sym_number_literal_token4] = ACTIONS(1449),
    [aux_sym_number_literal_token5] = ACTIONS(1451),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_SQUOTE] = ACTIONS(1449),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1451),
    [anon_sym_DQUOTE] = ACTIONS(1449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1451),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1451),
    [anon_sym_switch] = ACTIONS(1449),
    [anon_sym_case] = ACTIONS(1449),
    [anon_sym_default] = ACTIONS(1449),
    [anon_sym_QMARK] = ACTIONS(1449),
    [anon_sym_try] = ACTIONS(1449),
    [anon_sym_catch] = ACTIONS(1449),
    [anon_sym_finally] = ACTIONS(1449),
    [anon_sym_int] = ACTIONS(1449),
    [anon_sym_boolean] = ACTIONS(1449),
    [anon_sym_char] = ACTIONS(1449),
    [anon_sym_short] = ACTIONS(1449),
    [anon_sym_long] = ACTIONS(1449),
    [anon_sym_float] = ACTIONS(1449),
    [anon_sym_double] = ACTIONS(1449),
    [anon_sym_void] = ACTIONS(1449),
    [anon_sym_public] = ACTIONS(1449),
    [anon_sym_protected] = ACTIONS(1449),
    [anon_sym_private] = ACTIONS(1449),
    [anon_sym_static] = ACTIONS(1449),
    [anon_sym_final] = ACTIONS(1449),
    [anon_sym_synchronized] = ACTIONS(1449),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1449),
    [anon_sym_new] = ACTIONS(1449),
  },
  [STATE(393)] = {
    [sym_comment] = STATE(393),
    [sym_groovy_doc] = STATE(393),
    [aux_sym_dotted_identifier_repeat1] = STATE(432),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1453),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(394)] = {
    [sym_comment] = STATE(394),
    [sym_groovy_doc] = STATE(394),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_case] = ACTIONS(1455),
    [anon_sym_default] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_catch] = ACTIONS(1455),
    [anon_sym_finally] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(395)] = {
    [sym_comment] = STATE(395),
    [sym_groovy_doc] = STATE(395),
    [ts_builtin_sym_end] = ACTIONS(363),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_DASH_GT] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_pipeline] = ACTIONS(365),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(396)] = {
    [sym_comment] = STATE(396),
    [sym_groovy_doc] = STATE(396),
    [sym_identifier] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_COLON] = ACTIONS(1300),
    [anon_sym_DOT_AMP] = ACTIONS(1300),
    [anon_sym_DOT_AT] = ACTIONS(1300),
    [anon_sym_QMARK_DOT] = ACTIONS(1300),
    [anon_sym_STAR_DOT] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1302),
    [anon_sym_STAR_COLON] = ACTIONS(1300),
    [anon_sym_DOT] = ACTIONS(1302),
    [anon_sym_import] = ACTIONS(1302),
    [anon_sym_as] = ACTIONS(1302),
    [anon_sym_package] = ACTIONS(1302),
    [anon_sym_AT] = ACTIONS(1302),
    [anon_sym_assert] = ACTIONS(1302),
    [anon_sym_PLUS_PLUS] = ACTIONS(1300),
    [anon_sym_DASH_DASH] = ACTIONS(1300),
    [anon_sym_PERCENT] = ACTIONS(1300),
    [anon_sym_SLASH] = ACTIONS(1302),
    [anon_sym_PLUS] = ACTIONS(1302),
    [anon_sym_DASH] = ACTIONS(1302),
    [anon_sym_LT_LT] = ACTIONS(1300),
    [anon_sym_GT_GT] = ACTIONS(1302),
    [anon_sym_GT_GT_GT] = ACTIONS(1300),
    [anon_sym_DOT_DOT] = ACTIONS(1302),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1300),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1302),
    [anon_sym_LT] = ACTIONS(1302),
    [anon_sym_LT_EQ] = ACTIONS(1302),
    [anon_sym_GT] = ACTIONS(1302),
    [anon_sym_GT_EQ] = ACTIONS(1300),
    [anon_sym_in] = ACTIONS(1302),
    [anon_sym_BANGin] = ACTIONS(1302),
    [anon_sym_instanceof] = ACTIONS(1302),
    [anon_sym_BANGinstanceof] = ACTIONS(1300),
    [anon_sym_EQ_EQ] = ACTIONS(1302),
    [anon_sym_BANG_EQ] = ACTIONS(1302),
    [anon_sym_LT_EQ_GT] = ACTIONS(1300),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1300),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1300),
    [anon_sym_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1302),
    [anon_sym_CARET] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1302),
    [anon_sym_AMP_AMP] = ACTIONS(1300),
    [anon_sym_PIPE_PIPE] = ACTIONS(1300),
    [anon_sym_QMARK_COLON] = ACTIONS(1300),
    [anon_sym_STAR_STAR] = ACTIONS(1300),
    [anon_sym_true] = ACTIONS(1302),
    [anon_sym_false] = ACTIONS(1302),
    [anon_sym_ATinterface] = ACTIONS(1300),
    [anon_sym_interface] = ACTIONS(1302),
    [anon_sym_class] = ACTIONS(1302),
    [anon_sym_COMMA] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1302),
    [anon_sym_def] = ACTIONS(1302),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_null] = ACTIONS(1302),
    [anon_sym_this] = ACTIONS(1302),
    [anon_sym_do] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [aux_sym__type_identifier_token1] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_else] = ACTIONS(1302),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [aux_sym_number_literal_token1] = ACTIONS(1302),
    [aux_sym_number_literal_token2] = ACTIONS(1300),
    [aux_sym_number_literal_token3] = ACTIONS(1300),
    [aux_sym_number_literal_token4] = ACTIONS(1302),
    [aux_sym_number_literal_token5] = ACTIONS(1300),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1300),
    [anon_sym_DQUOTE] = ACTIONS(1302),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1300),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1300),
    [anon_sym_switch] = ACTIONS(1302),
    [anon_sym_case] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_QMARK] = ACTIONS(1302),
    [anon_sym_try] = ACTIONS(1302),
    [anon_sym_int] = ACTIONS(1302),
    [anon_sym_boolean] = ACTIONS(1302),
    [anon_sym_char] = ACTIONS(1302),
    [anon_sym_short] = ACTIONS(1302),
    [anon_sym_long] = ACTIONS(1302),
    [anon_sym_float] = ACTIONS(1302),
    [anon_sym_double] = ACTIONS(1302),
    [anon_sym_void] = ACTIONS(1302),
    [anon_sym_public] = ACTIONS(1302),
    [anon_sym_protected] = ACTIONS(1302),
    [anon_sym_private] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_final] = ACTIONS(1302),
    [anon_sym_synchronized] = ACTIONS(1302),
    [anon_sym_TILDE] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1302),
    [anon_sym_new] = ACTIONS(1302),
  },
  [STATE(397)] = {
    [sym_comment] = STATE(397),
    [sym_groovy_doc] = STATE(397),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1457),
    [anon_sym_DOT_AT] = ACTIONS(1457),
    [anon_sym_QMARK_DOT] = ACTIONS(1457),
    [anon_sym_STAR_DOT] = ACTIONS(1457),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1457),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_case] = ACTIONS(1455),
    [anon_sym_default] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_catch] = ACTIONS(1455),
    [anon_sym_finally] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(398)] = {
    [sym_comment] = STATE(398),
    [sym_groovy_doc] = STATE(398),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1278),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_case] = ACTIONS(1280),
    [anon_sym_default] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_catch] = ACTIONS(1280),
    [anon_sym_finally] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(399)] = {
    [sym_comment] = STATE(399),
    [sym_groovy_doc] = STATE(399),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(400)] = {
    [sym_comment] = STATE(400),
    [sym_groovy_doc] = STATE(400),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(401)] = {
    [sym_comment] = STATE(401),
    [sym_groovy_doc] = STATE(401),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(402)] = {
    [sym_comment] = STATE(402),
    [sym_groovy_doc] = STATE(402),
    [sym_identifier] = ACTIONS(1316),
    [anon_sym_break] = ACTIONS(1316),
    [anon_sym_continue] = ACTIONS(1316),
    [anon_sym_SEMI] = ACTIONS(1314),
    [anon_sym_COLON] = ACTIONS(1314),
    [anon_sym_DOT_AMP] = ACTIONS(1314),
    [anon_sym_DOT_AT] = ACTIONS(1314),
    [anon_sym_QMARK_DOT] = ACTIONS(1314),
    [anon_sym_STAR_DOT] = ACTIONS(1314),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_STAR_COLON] = ACTIONS(1314),
    [anon_sym_DOT] = ACTIONS(1316),
    [anon_sym_import] = ACTIONS(1316),
    [anon_sym_as] = ACTIONS(1316),
    [anon_sym_package] = ACTIONS(1316),
    [anon_sym_AT] = ACTIONS(1316),
    [anon_sym_assert] = ACTIONS(1316),
    [anon_sym_PLUS_PLUS] = ACTIONS(1314),
    [anon_sym_DASH_DASH] = ACTIONS(1314),
    [anon_sym_PERCENT] = ACTIONS(1314),
    [anon_sym_SLASH] = ACTIONS(1316),
    [anon_sym_PLUS] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_LT_LT] = ACTIONS(1314),
    [anon_sym_GT_GT] = ACTIONS(1316),
    [anon_sym_GT_GT_GT] = ACTIONS(1314),
    [anon_sym_DOT_DOT] = ACTIONS(1316),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1314),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1316),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_LT_EQ] = ACTIONS(1316),
    [anon_sym_GT] = ACTIONS(1316),
    [anon_sym_GT_EQ] = ACTIONS(1314),
    [anon_sym_in] = ACTIONS(1316),
    [anon_sym_BANGin] = ACTIONS(1316),
    [anon_sym_instanceof] = ACTIONS(1316),
    [anon_sym_BANGinstanceof] = ACTIONS(1314),
    [anon_sym_EQ_EQ] = ACTIONS(1316),
    [anon_sym_BANG_EQ] = ACTIONS(1316),
    [anon_sym_LT_EQ_GT] = ACTIONS(1314),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1314),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1314),
    [anon_sym_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1314),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_CARET] = ACTIONS(1314),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_AMP_AMP] = ACTIONS(1314),
    [anon_sym_PIPE_PIPE] = ACTIONS(1314),
    [anon_sym_QMARK_COLON] = ACTIONS(1314),
    [anon_sym_STAR_STAR] = ACTIONS(1314),
    [anon_sym_true] = ACTIONS(1316),
    [anon_sym_false] = ACTIONS(1316),
    [anon_sym_ATinterface] = ACTIONS(1314),
    [anon_sym_interface] = ACTIONS(1316),
    [anon_sym_class] = ACTIONS(1316),
    [anon_sym_COMMA] = ACTIONS(1314),
    [anon_sym_LBRACE] = ACTIONS(1314),
    [anon_sym_RBRACE] = ACTIONS(1314),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1316),
    [anon_sym_def] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1314),
    [anon_sym_null] = ACTIONS(1316),
    [anon_sym_this] = ACTIONS(1316),
    [anon_sym_do] = ACTIONS(1316),
    [anon_sym_while] = ACTIONS(1316),
    [anon_sym_for] = ACTIONS(1316),
    [aux_sym__type_identifier_token1] = ACTIONS(1316),
    [anon_sym_if] = ACTIONS(1316),
    [anon_sym_else] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1314),
    [aux_sym_number_literal_token1] = ACTIONS(1316),
    [aux_sym_number_literal_token2] = ACTIONS(1314),
    [aux_sym_number_literal_token3] = ACTIONS(1314),
    [aux_sym_number_literal_token4] = ACTIONS(1316),
    [aux_sym_number_literal_token5] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1316),
    [anon_sym_SQUOTE] = ACTIONS(1316),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1314),
    [anon_sym_DQUOTE] = ACTIONS(1316),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1314),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1314),
    [anon_sym_switch] = ACTIONS(1316),
    [anon_sym_case] = ACTIONS(1316),
    [anon_sym_default] = ACTIONS(1316),
    [anon_sym_QMARK] = ACTIONS(1316),
    [anon_sym_try] = ACTIONS(1316),
    [anon_sym_int] = ACTIONS(1316),
    [anon_sym_boolean] = ACTIONS(1316),
    [anon_sym_char] = ACTIONS(1316),
    [anon_sym_short] = ACTIONS(1316),
    [anon_sym_long] = ACTIONS(1316),
    [anon_sym_float] = ACTIONS(1316),
    [anon_sym_double] = ACTIONS(1316),
    [anon_sym_void] = ACTIONS(1316),
    [anon_sym_public] = ACTIONS(1316),
    [anon_sym_protected] = ACTIONS(1316),
    [anon_sym_private] = ACTIONS(1316),
    [anon_sym_static] = ACTIONS(1316),
    [anon_sym_final] = ACTIONS(1316),
    [anon_sym_synchronized] = ACTIONS(1316),
    [anon_sym_TILDE] = ACTIONS(1314),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_new] = ACTIONS(1316),
  },
  [STATE(403)] = {
    [sym_comment] = STATE(403),
    [sym_groovy_doc] = STATE(403),
    [sym_identifier] = ACTIONS(1320),
    [anon_sym_break] = ACTIONS(1320),
    [anon_sym_continue] = ACTIONS(1320),
    [anon_sym_SEMI] = ACTIONS(1318),
    [anon_sym_COLON] = ACTIONS(1318),
    [anon_sym_DOT_AMP] = ACTIONS(1318),
    [anon_sym_DOT_AT] = ACTIONS(1318),
    [anon_sym_QMARK_DOT] = ACTIONS(1318),
    [anon_sym_STAR_DOT] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_STAR_COLON] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1320),
    [anon_sym_import] = ACTIONS(1320),
    [anon_sym_as] = ACTIONS(1320),
    [anon_sym_package] = ACTIONS(1320),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_assert] = ACTIONS(1320),
    [anon_sym_PLUS_PLUS] = ACTIONS(1318),
    [anon_sym_DASH_DASH] = ACTIONS(1318),
    [anon_sym_PERCENT] = ACTIONS(1318),
    [anon_sym_SLASH] = ACTIONS(1320),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_LT_LT] = ACTIONS(1318),
    [anon_sym_GT_GT] = ACTIONS(1320),
    [anon_sym_GT_GT_GT] = ACTIONS(1318),
    [anon_sym_DOT_DOT] = ACTIONS(1320),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1318),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1320),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_LT_EQ] = ACTIONS(1320),
    [anon_sym_GT] = ACTIONS(1320),
    [anon_sym_GT_EQ] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1320),
    [anon_sym_BANGin] = ACTIONS(1320),
    [anon_sym_instanceof] = ACTIONS(1320),
    [anon_sym_BANGinstanceof] = ACTIONS(1318),
    [anon_sym_EQ_EQ] = ACTIONS(1320),
    [anon_sym_BANG_EQ] = ACTIONS(1320),
    [anon_sym_LT_EQ_GT] = ACTIONS(1318),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1318),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1318),
    [anon_sym_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1318),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_CARET] = ACTIONS(1318),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_AMP_AMP] = ACTIONS(1318),
    [anon_sym_PIPE_PIPE] = ACTIONS(1318),
    [anon_sym_QMARK_COLON] = ACTIONS(1318),
    [anon_sym_STAR_STAR] = ACTIONS(1318),
    [anon_sym_true] = ACTIONS(1320),
    [anon_sym_false] = ACTIONS(1320),
    [anon_sym_ATinterface] = ACTIONS(1318),
    [anon_sym_interface] = ACTIONS(1320),
    [anon_sym_class] = ACTIONS(1320),
    [anon_sym_COMMA] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [anon_sym_RBRACE] = ACTIONS(1318),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1320),
    [anon_sym_def] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1318),
    [anon_sym_null] = ACTIONS(1320),
    [anon_sym_this] = ACTIONS(1320),
    [anon_sym_do] = ACTIONS(1320),
    [anon_sym_while] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1320),
    [aux_sym__type_identifier_token1] = ACTIONS(1320),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_else] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1318),
    [aux_sym_number_literal_token1] = ACTIONS(1320),
    [aux_sym_number_literal_token2] = ACTIONS(1318),
    [aux_sym_number_literal_token3] = ACTIONS(1318),
    [aux_sym_number_literal_token4] = ACTIONS(1320),
    [aux_sym_number_literal_token5] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1320),
    [anon_sym_SQUOTE] = ACTIONS(1320),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1318),
    [anon_sym_DQUOTE] = ACTIONS(1320),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1318),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1318),
    [anon_sym_switch] = ACTIONS(1320),
    [anon_sym_case] = ACTIONS(1320),
    [anon_sym_default] = ACTIONS(1320),
    [anon_sym_QMARK] = ACTIONS(1320),
    [anon_sym_try] = ACTIONS(1320),
    [anon_sym_int] = ACTIONS(1320),
    [anon_sym_boolean] = ACTIONS(1320),
    [anon_sym_char] = ACTIONS(1320),
    [anon_sym_short] = ACTIONS(1320),
    [anon_sym_long] = ACTIONS(1320),
    [anon_sym_float] = ACTIONS(1320),
    [anon_sym_double] = ACTIONS(1320),
    [anon_sym_void] = ACTIONS(1320),
    [anon_sym_public] = ACTIONS(1320),
    [anon_sym_protected] = ACTIONS(1320),
    [anon_sym_private] = ACTIONS(1320),
    [anon_sym_static] = ACTIONS(1320),
    [anon_sym_final] = ACTIONS(1320),
    [anon_sym_synchronized] = ACTIONS(1320),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_new] = ACTIONS(1320),
  },
  [STATE(404)] = {
    [sym_comment] = STATE(404),
    [sym_groovy_doc] = STATE(404),
    [sym_identifier] = ACTIONS(1463),
    [anon_sym_break] = ACTIONS(1463),
    [anon_sym_continue] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_DOT_AMP] = ACTIONS(1465),
    [anon_sym_DOT_AT] = ACTIONS(1465),
    [anon_sym_QMARK_DOT] = ACTIONS(1465),
    [anon_sym_STAR_DOT] = ACTIONS(1465),
    [anon_sym_STAR] = ACTIONS(1463),
    [anon_sym_STAR_COLON] = ACTIONS(1465),
    [anon_sym_import] = ACTIONS(1463),
    [anon_sym_as] = ACTIONS(1463),
    [anon_sym_package] = ACTIONS(1463),
    [anon_sym_AT] = ACTIONS(1463),
    [anon_sym_assert] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_PERCENT] = ACTIONS(1465),
    [anon_sym_SLASH] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [anon_sym_LT_LT] = ACTIONS(1465),
    [anon_sym_GT_GT] = ACTIONS(1463),
    [anon_sym_GT_GT_GT] = ACTIONS(1465),
    [anon_sym_DOT_DOT] = ACTIONS(1463),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1463),
    [anon_sym_LT] = ACTIONS(1463),
    [anon_sym_LT_EQ] = ACTIONS(1463),
    [anon_sym_GT] = ACTIONS(1463),
    [anon_sym_GT_EQ] = ACTIONS(1465),
    [anon_sym_in] = ACTIONS(1463),
    [anon_sym_BANGin] = ACTIONS(1463),
    [anon_sym_instanceof] = ACTIONS(1463),
    [anon_sym_BANGinstanceof] = ACTIONS(1465),
    [anon_sym_EQ_EQ] = ACTIONS(1463),
    [anon_sym_BANG_EQ] = ACTIONS(1463),
    [anon_sym_LT_EQ_GT] = ACTIONS(1465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1465),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1465),
    [anon_sym_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1463),
    [anon_sym_CARET] = ACTIONS(1465),
    [anon_sym_PIPE] = ACTIONS(1463),
    [anon_sym_AMP_AMP] = ACTIONS(1465),
    [anon_sym_PIPE_PIPE] = ACTIONS(1465),
    [anon_sym_QMARK_COLON] = ACTIONS(1465),
    [anon_sym_STAR_STAR] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(1463),
    [anon_sym_false] = ACTIONS(1463),
    [anon_sym_ATinterface] = ACTIONS(1465),
    [anon_sym_interface] = ACTIONS(1463),
    [anon_sym_class] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [anon_sym_RBRACE] = ACTIONS(1465),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1463),
    [anon_sym_def] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_null] = ACTIONS(1463),
    [anon_sym_this] = ACTIONS(1463),
    [anon_sym_do] = ACTIONS(1463),
    [anon_sym_while] = ACTIONS(1463),
    [anon_sym_for] = ACTIONS(1463),
    [aux_sym__type_identifier_token1] = ACTIONS(1463),
    [anon_sym_if] = ACTIONS(1463),
    [anon_sym_else] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_number_literal_token1] = ACTIONS(1463),
    [aux_sym_number_literal_token2] = ACTIONS(1465),
    [aux_sym_number_literal_token3] = ACTIONS(1465),
    [aux_sym_number_literal_token4] = ACTIONS(1463),
    [aux_sym_number_literal_token5] = ACTIONS(1465),
    [anon_sym_return] = ACTIONS(1463),
    [anon_sym_SQUOTE] = ACTIONS(1463),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(1463),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1465),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1465),
    [anon_sym_switch] = ACTIONS(1463),
    [anon_sym_case] = ACTIONS(1463),
    [anon_sym_default] = ACTIONS(1463),
    [anon_sym_QMARK] = ACTIONS(1463),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_catch] = ACTIONS(1463),
    [anon_sym_finally] = ACTIONS(1463),
    [anon_sym_int] = ACTIONS(1463),
    [anon_sym_boolean] = ACTIONS(1463),
    [anon_sym_char] = ACTIONS(1463),
    [anon_sym_short] = ACTIONS(1463),
    [anon_sym_long] = ACTIONS(1463),
    [anon_sym_float] = ACTIONS(1463),
    [anon_sym_double] = ACTIONS(1463),
    [anon_sym_void] = ACTIONS(1463),
    [anon_sym_public] = ACTIONS(1463),
    [anon_sym_protected] = ACTIONS(1463),
    [anon_sym_private] = ACTIONS(1463),
    [anon_sym_static] = ACTIONS(1463),
    [anon_sym_final] = ACTIONS(1463),
    [anon_sym_synchronized] = ACTIONS(1463),
    [anon_sym_TILDE] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1463),
    [anon_sym_new] = ACTIONS(1463),
  },
  [STATE(405)] = {
    [sym_comment] = STATE(405),
    [sym_groovy_doc] = STATE(405),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(406)] = {
    [sym_comment] = STATE(406),
    [sym_groovy_doc] = STATE(406),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(407)] = {
    [sym_comment] = STATE(407),
    [sym_groovy_doc] = STATE(407),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(408)] = {
    [sym_comment] = STATE(408),
    [sym_groovy_doc] = STATE(408),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(409)] = {
    [sym_comment] = STATE(409),
    [sym_groovy_doc] = STATE(409),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(410)] = {
    [sym_comment] = STATE(410),
    [sym_groovy_doc] = STATE(410),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(411)] = {
    [sym_comment] = STATE(411),
    [sym_groovy_doc] = STATE(411),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(412)] = {
    [sym_comment] = STATE(412),
    [sym_groovy_doc] = STATE(412),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(413)] = {
    [sym_comment] = STATE(413),
    [sym_groovy_doc] = STATE(413),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(414)] = {
    [sym_comment] = STATE(414),
    [sym_groovy_doc] = STATE(414),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(415)] = {
    [sym_comment] = STATE(415),
    [sym_groovy_doc] = STATE(415),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(416)] = {
    [sym_comment] = STATE(416),
    [sym_groovy_doc] = STATE(416),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(417)] = {
    [sym_comment] = STATE(417),
    [sym_groovy_doc] = STATE(417),
    [ts_builtin_sym_end] = ACTIONS(1378),
    [sym_identifier] = ACTIONS(1380),
    [anon_sym_break] = ACTIONS(1380),
    [anon_sym_continue] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1378),
    [anon_sym_DOT_AMP] = ACTIONS(1378),
    [anon_sym_DOT_AT] = ACTIONS(1378),
    [anon_sym_QMARK_DOT] = ACTIONS(1378),
    [anon_sym_STAR_DOT] = ACTIONS(1378),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_STAR_COLON] = ACTIONS(1378),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_import] = ACTIONS(1380),
    [anon_sym_as] = ACTIONS(1380),
    [anon_sym_package] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_assert] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1378),
    [anon_sym_DASH_DASH] = ACTIONS(1378),
    [anon_sym_PERCENT] = ACTIONS(1378),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_LT_LT] = ACTIONS(1378),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym_GT_GT_GT] = ACTIONS(1378),
    [anon_sym_DOT_DOT] = ACTIONS(1380),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_GT_EQ] = ACTIONS(1378),
    [anon_sym_in] = ACTIONS(1380),
    [anon_sym_BANGin] = ACTIONS(1380),
    [anon_sym_instanceof] = ACTIONS(1380),
    [anon_sym_BANGinstanceof] = ACTIONS(1378),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ_GT] = ACTIONS(1378),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1378),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1378),
    [anon_sym_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1378),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1378),
    [anon_sym_PIPE_PIPE] = ACTIONS(1378),
    [anon_sym_QMARK_COLON] = ACTIONS(1378),
    [anon_sym_STAR_STAR] = ACTIONS(1378),
    [anon_sym_true] = ACTIONS(1380),
    [anon_sym_false] = ACTIONS(1380),
    [anon_sym_ATinterface] = ACTIONS(1378),
    [anon_sym_interface] = ACTIONS(1380),
    [anon_sym_class] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1378),
    [anon_sym_LBRACE] = ACTIONS(1378),
    [anon_sym_DASH_GT] = ACTIONS(1378),
    [anon_sym_RBRACE] = ACTIONS(1378),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_def] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1378),
    [anon_sym_null] = ACTIONS(1380),
    [anon_sym_this] = ACTIONS(1380),
    [anon_sym_do] = ACTIONS(1380),
    [anon_sym_while] = ACTIONS(1380),
    [anon_sym_for] = ACTIONS(1380),
    [aux_sym__type_identifier_token1] = ACTIONS(1380),
    [anon_sym_if] = ACTIONS(1380),
    [anon_sym_else] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [aux_sym_number_literal_token1] = ACTIONS(1380),
    [aux_sym_number_literal_token2] = ACTIONS(1378),
    [aux_sym_number_literal_token3] = ACTIONS(1378),
    [aux_sym_number_literal_token4] = ACTIONS(1380),
    [aux_sym_number_literal_token5] = ACTIONS(1378),
    [anon_sym_pipeline] = ACTIONS(1380),
    [anon_sym_return] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1378),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1378),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1378),
    [anon_sym_switch] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_try] = ACTIONS(1380),
    [anon_sym_int] = ACTIONS(1380),
    [anon_sym_boolean] = ACTIONS(1380),
    [anon_sym_char] = ACTIONS(1380),
    [anon_sym_short] = ACTIONS(1380),
    [anon_sym_long] = ACTIONS(1380),
    [anon_sym_float] = ACTIONS(1380),
    [anon_sym_double] = ACTIONS(1380),
    [anon_sym_void] = ACTIONS(1380),
    [anon_sym_public] = ACTIONS(1380),
    [anon_sym_protected] = ACTIONS(1380),
    [anon_sym_private] = ACTIONS(1380),
    [anon_sym_static] = ACTIONS(1380),
    [anon_sym_final] = ACTIONS(1380),
    [anon_sym_synchronized] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1378),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_new] = ACTIONS(1380),
  },
  [STATE(418)] = {
    [sym_comment] = STATE(418),
    [sym_groovy_doc] = STATE(418),
    [sym_identifier] = ACTIONS(1499),
    [anon_sym_break] = ACTIONS(1499),
    [anon_sym_continue] = ACTIONS(1499),
    [anon_sym_SEMI] = ACTIONS(1501),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1501),
    [anon_sym_import] = ACTIONS(1499),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1499),
    [anon_sym_AT] = ACTIONS(1499),
    [anon_sym_assert] = ACTIONS(1499),
    [anon_sym_PLUS_PLUS] = ACTIONS(1501),
    [anon_sym_DASH_DASH] = ACTIONS(1501),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1499),
    [anon_sym_false] = ACTIONS(1499),
    [anon_sym_ATinterface] = ACTIONS(1501),
    [anon_sym_interface] = ACTIONS(1499),
    [anon_sym_class] = ACTIONS(1499),
    [anon_sym_COMMA] = ACTIONS(1501),
    [anon_sym_LBRACE] = ACTIONS(1501),
    [anon_sym_RBRACE] = ACTIONS(1501),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1499),
    [anon_sym_def] = ACTIONS(1499),
    [anon_sym_LPAREN] = ACTIONS(1501),
    [anon_sym_null] = ACTIONS(1499),
    [anon_sym_this] = ACTIONS(1499),
    [anon_sym_do] = ACTIONS(1499),
    [anon_sym_while] = ACTIONS(1499),
    [anon_sym_for] = ACTIONS(1499),
    [aux_sym__type_identifier_token1] = ACTIONS(1499),
    [anon_sym_if] = ACTIONS(1499),
    [anon_sym_else] = ACTIONS(1499),
    [anon_sym_LBRACK] = ACTIONS(1501),
    [aux_sym_number_literal_token1] = ACTIONS(1499),
    [aux_sym_number_literal_token2] = ACTIONS(1501),
    [aux_sym_number_literal_token3] = ACTIONS(1501),
    [aux_sym_number_literal_token4] = ACTIONS(1499),
    [aux_sym_number_literal_token5] = ACTIONS(1501),
    [anon_sym_return] = ACTIONS(1499),
    [anon_sym_SQUOTE] = ACTIONS(1499),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(1499),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1501),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1501),
    [anon_sym_switch] = ACTIONS(1499),
    [anon_sym_case] = ACTIONS(1499),
    [anon_sym_default] = ACTIONS(1499),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1499),
    [anon_sym_catch] = ACTIONS(1499),
    [anon_sym_finally] = ACTIONS(1499),
    [anon_sym_int] = ACTIONS(1499),
    [anon_sym_boolean] = ACTIONS(1499),
    [anon_sym_char] = ACTIONS(1499),
    [anon_sym_short] = ACTIONS(1499),
    [anon_sym_long] = ACTIONS(1499),
    [anon_sym_float] = ACTIONS(1499),
    [anon_sym_double] = ACTIONS(1499),
    [anon_sym_void] = ACTIONS(1499),
    [anon_sym_public] = ACTIONS(1499),
    [anon_sym_protected] = ACTIONS(1499),
    [anon_sym_private] = ACTIONS(1499),
    [anon_sym_static] = ACTIONS(1499),
    [anon_sym_final] = ACTIONS(1499),
    [anon_sym_synchronized] = ACTIONS(1499),
    [anon_sym_TILDE] = ACTIONS(1501),
    [anon_sym_BANG] = ACTIONS(1499),
    [anon_sym_new] = ACTIONS(1499),
  },
  [STATE(419)] = {
    [sym_comment] = STATE(419),
    [sym_groovy_doc] = STATE(419),
    [ts_builtin_sym_end] = ACTIONS(1382),
    [sym_identifier] = ACTIONS(1384),
    [anon_sym_break] = ACTIONS(1384),
    [anon_sym_continue] = ACTIONS(1384),
    [anon_sym_SEMI] = ACTIONS(1382),
    [anon_sym_DOT_AMP] = ACTIONS(1382),
    [anon_sym_DOT_AT] = ACTIONS(1382),
    [anon_sym_QMARK_DOT] = ACTIONS(1382),
    [anon_sym_STAR_DOT] = ACTIONS(1382),
    [anon_sym_STAR] = ACTIONS(1384),
    [anon_sym_STAR_COLON] = ACTIONS(1382),
    [anon_sym_DOT] = ACTIONS(1384),
    [anon_sym_import] = ACTIONS(1384),
    [anon_sym_as] = ACTIONS(1384),
    [anon_sym_package] = ACTIONS(1384),
    [anon_sym_AT] = ACTIONS(1384),
    [anon_sym_assert] = ACTIONS(1384),
    [anon_sym_PLUS_PLUS] = ACTIONS(1382),
    [anon_sym_DASH_DASH] = ACTIONS(1382),
    [anon_sym_PERCENT] = ACTIONS(1382),
    [anon_sym_SLASH] = ACTIONS(1384),
    [anon_sym_PLUS] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_LT_LT] = ACTIONS(1382),
    [anon_sym_GT_GT] = ACTIONS(1384),
    [anon_sym_GT_GT_GT] = ACTIONS(1382),
    [anon_sym_DOT_DOT] = ACTIONS(1384),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1384),
    [anon_sym_LT] = ACTIONS(1384),
    [anon_sym_LT_EQ] = ACTIONS(1384),
    [anon_sym_GT] = ACTIONS(1384),
    [anon_sym_GT_EQ] = ACTIONS(1382),
    [anon_sym_in] = ACTIONS(1384),
    [anon_sym_BANGin] = ACTIONS(1384),
    [anon_sym_instanceof] = ACTIONS(1384),
    [anon_sym_BANGinstanceof] = ACTIONS(1382),
    [anon_sym_EQ_EQ] = ACTIONS(1384),
    [anon_sym_BANG_EQ] = ACTIONS(1384),
    [anon_sym_LT_EQ_GT] = ACTIONS(1382),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1382),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1382),
    [anon_sym_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_AMP] = ACTIONS(1384),
    [anon_sym_CARET] = ACTIONS(1382),
    [anon_sym_PIPE] = ACTIONS(1384),
    [anon_sym_AMP_AMP] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1382),
    [anon_sym_QMARK_COLON] = ACTIONS(1382),
    [anon_sym_STAR_STAR] = ACTIONS(1382),
    [anon_sym_true] = ACTIONS(1384),
    [anon_sym_false] = ACTIONS(1384),
    [anon_sym_ATinterface] = ACTIONS(1382),
    [anon_sym_interface] = ACTIONS(1384),
    [anon_sym_class] = ACTIONS(1384),
    [anon_sym_COMMA] = ACTIONS(1382),
    [anon_sym_LBRACE] = ACTIONS(1382),
    [anon_sym_DASH_GT] = ACTIONS(1382),
    [anon_sym_RBRACE] = ACTIONS(1382),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1384),
    [anon_sym_def] = ACTIONS(1384),
    [anon_sym_LPAREN] = ACTIONS(1382),
    [anon_sym_null] = ACTIONS(1384),
    [anon_sym_this] = ACTIONS(1384),
    [anon_sym_do] = ACTIONS(1384),
    [anon_sym_while] = ACTIONS(1384),
    [anon_sym_for] = ACTIONS(1384),
    [aux_sym__type_identifier_token1] = ACTIONS(1384),
    [anon_sym_if] = ACTIONS(1384),
    [anon_sym_else] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1382),
    [aux_sym_number_literal_token1] = ACTIONS(1384),
    [aux_sym_number_literal_token2] = ACTIONS(1382),
    [aux_sym_number_literal_token3] = ACTIONS(1382),
    [aux_sym_number_literal_token4] = ACTIONS(1384),
    [aux_sym_number_literal_token5] = ACTIONS(1382),
    [anon_sym_pipeline] = ACTIONS(1384),
    [anon_sym_return] = ACTIONS(1384),
    [anon_sym_SQUOTE] = ACTIONS(1384),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1382),
    [anon_sym_DQUOTE] = ACTIONS(1384),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1382),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1382),
    [anon_sym_switch] = ACTIONS(1384),
    [anon_sym_QMARK] = ACTIONS(1384),
    [anon_sym_try] = ACTIONS(1384),
    [anon_sym_int] = ACTIONS(1384),
    [anon_sym_boolean] = ACTIONS(1384),
    [anon_sym_char] = ACTIONS(1384),
    [anon_sym_short] = ACTIONS(1384),
    [anon_sym_long] = ACTIONS(1384),
    [anon_sym_float] = ACTIONS(1384),
    [anon_sym_double] = ACTIONS(1384),
    [anon_sym_void] = ACTIONS(1384),
    [anon_sym_public] = ACTIONS(1384),
    [anon_sym_protected] = ACTIONS(1384),
    [anon_sym_private] = ACTIONS(1384),
    [anon_sym_static] = ACTIONS(1384),
    [anon_sym_final] = ACTIONS(1384),
    [anon_sym_synchronized] = ACTIONS(1384),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_new] = ACTIONS(1384),
  },
  [STATE(420)] = {
    [sym_comment] = STATE(420),
    [sym_groovy_doc] = STATE(420),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(421)] = {
    [sym_comment] = STATE(421),
    [sym_groovy_doc] = STATE(421),
    [aux_sym_dotted_identifier_repeat1] = STATE(33),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1507),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(102),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1290),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(422)] = {
    [sym_comment] = STATE(422),
    [sym_groovy_doc] = STATE(422),
    [sym_identifier] = ACTIONS(1509),
    [anon_sym_break] = ACTIONS(1509),
    [anon_sym_continue] = ACTIONS(1509),
    [anon_sym_SEMI] = ACTIONS(1511),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1511),
    [anon_sym_import] = ACTIONS(1509),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1509),
    [anon_sym_assert] = ACTIONS(1509),
    [anon_sym_PLUS_PLUS] = ACTIONS(1511),
    [anon_sym_DASH_DASH] = ACTIONS(1511),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1509),
    [anon_sym_false] = ACTIONS(1509),
    [anon_sym_ATinterface] = ACTIONS(1511),
    [anon_sym_interface] = ACTIONS(1509),
    [anon_sym_class] = ACTIONS(1509),
    [anon_sym_COMMA] = ACTIONS(1511),
    [anon_sym_LBRACE] = ACTIONS(1511),
    [anon_sym_RBRACE] = ACTIONS(1511),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1509),
    [anon_sym_def] = ACTIONS(1509),
    [anon_sym_LPAREN] = ACTIONS(1511),
    [anon_sym_null] = ACTIONS(1509),
    [anon_sym_this] = ACTIONS(1509),
    [anon_sym_do] = ACTIONS(1509),
    [anon_sym_while] = ACTIONS(1509),
    [anon_sym_for] = ACTIONS(1509),
    [aux_sym__type_identifier_token1] = ACTIONS(1509),
    [anon_sym_if] = ACTIONS(1509),
    [anon_sym_else] = ACTIONS(1509),
    [anon_sym_LBRACK] = ACTIONS(1511),
    [aux_sym_number_literal_token1] = ACTIONS(1509),
    [aux_sym_number_literal_token2] = ACTIONS(1511),
    [aux_sym_number_literal_token3] = ACTIONS(1511),
    [aux_sym_number_literal_token4] = ACTIONS(1509),
    [aux_sym_number_literal_token5] = ACTIONS(1511),
    [anon_sym_return] = ACTIONS(1509),
    [anon_sym_SQUOTE] = ACTIONS(1509),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1511),
    [anon_sym_DQUOTE] = ACTIONS(1509),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1511),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1511),
    [anon_sym_switch] = ACTIONS(1509),
    [anon_sym_case] = ACTIONS(1509),
    [anon_sym_default] = ACTIONS(1509),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1509),
    [anon_sym_catch] = ACTIONS(1509),
    [anon_sym_finally] = ACTIONS(1509),
    [anon_sym_int] = ACTIONS(1509),
    [anon_sym_boolean] = ACTIONS(1509),
    [anon_sym_char] = ACTIONS(1509),
    [anon_sym_short] = ACTIONS(1509),
    [anon_sym_long] = ACTIONS(1509),
    [anon_sym_float] = ACTIONS(1509),
    [anon_sym_double] = ACTIONS(1509),
    [anon_sym_void] = ACTIONS(1509),
    [anon_sym_public] = ACTIONS(1509),
    [anon_sym_protected] = ACTIONS(1509),
    [anon_sym_private] = ACTIONS(1509),
    [anon_sym_static] = ACTIONS(1509),
    [anon_sym_final] = ACTIONS(1509),
    [anon_sym_synchronized] = ACTIONS(1509),
    [anon_sym_TILDE] = ACTIONS(1511),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_new] = ACTIONS(1509),
  },
  [STATE(423)] = {
    [sym_comment] = STATE(423),
    [sym_groovy_doc] = STATE(423),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(369),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(369),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(369),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(369),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_DASH_GT] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(424)] = {
    [sym_comment] = STATE(424),
    [sym_groovy_doc] = STATE(424),
    [ts_builtin_sym_end] = ACTIONS(377),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_DASH_GT] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_pipeline] = ACTIONS(379),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(425)] = {
    [sym_comment] = STATE(425),
    [sym_groovy_doc] = STATE(425),
    [ts_builtin_sym_end] = ACTIONS(391),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(391),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(391),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(391),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(391),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_DASH_GT] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_pipeline] = ACTIONS(393),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(426)] = {
    [sym_comment] = STATE(426),
    [sym_groovy_doc] = STATE(426),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(427)] = {
    [sym_comment] = STATE(427),
    [sym_groovy_doc] = STATE(427),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(428)] = {
    [sym_closure] = STATE(557),
    [sym_comment] = STATE(428),
    [sym_groovy_doc] = STATE(428),
    [sym_identifier] = ACTIONS(1324),
    [anon_sym_break] = ACTIONS(1324),
    [anon_sym_continue] = ACTIONS(1324),
    [anon_sym_SEMI] = ACTIONS(1322),
    [anon_sym_DOT_AMP] = ACTIONS(1322),
    [anon_sym_DOT_AT] = ACTIONS(1322),
    [anon_sym_QMARK_DOT] = ACTIONS(1322),
    [anon_sym_STAR_DOT] = ACTIONS(1322),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_STAR_COLON] = ACTIONS(1322),
    [anon_sym_DOT] = ACTIONS(1324),
    [anon_sym_import] = ACTIONS(1324),
    [anon_sym_as] = ACTIONS(1324),
    [anon_sym_package] = ACTIONS(1324),
    [anon_sym_AT] = ACTIONS(1324),
    [anon_sym_assert] = ACTIONS(1324),
    [anon_sym_PLUS_PLUS] = ACTIONS(1322),
    [anon_sym_DASH_DASH] = ACTIONS(1322),
    [anon_sym_PERCENT] = ACTIONS(1322),
    [anon_sym_SLASH] = ACTIONS(1324),
    [anon_sym_PLUS] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_LT_LT] = ACTIONS(1322),
    [anon_sym_GT_GT] = ACTIONS(1324),
    [anon_sym_GT_GT_GT] = ACTIONS(1322),
    [anon_sym_DOT_DOT] = ACTIONS(1324),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1322),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1324),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_LT_EQ] = ACTIONS(1324),
    [anon_sym_GT] = ACTIONS(1324),
    [anon_sym_GT_EQ] = ACTIONS(1322),
    [anon_sym_in] = ACTIONS(1324),
    [anon_sym_BANGin] = ACTIONS(1324),
    [anon_sym_instanceof] = ACTIONS(1324),
    [anon_sym_BANGinstanceof] = ACTIONS(1322),
    [anon_sym_EQ_EQ] = ACTIONS(1324),
    [anon_sym_BANG_EQ] = ACTIONS(1324),
    [anon_sym_LT_EQ_GT] = ACTIONS(1322),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1322),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1322),
    [anon_sym_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1322),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_CARET] = ACTIONS(1322),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_AMP_AMP] = ACTIONS(1322),
    [anon_sym_PIPE_PIPE] = ACTIONS(1322),
    [anon_sym_QMARK_COLON] = ACTIONS(1322),
    [anon_sym_STAR_STAR] = ACTIONS(1322),
    [anon_sym_true] = ACTIONS(1324),
    [anon_sym_false] = ACTIONS(1324),
    [anon_sym_ATinterface] = ACTIONS(1322),
    [anon_sym_interface] = ACTIONS(1324),
    [anon_sym_class] = ACTIONS(1324),
    [anon_sym_COMMA] = ACTIONS(1322),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(1322),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1324),
    [anon_sym_def] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1322),
    [anon_sym_null] = ACTIONS(1324),
    [anon_sym_this] = ACTIONS(1324),
    [anon_sym_do] = ACTIONS(1324),
    [anon_sym_while] = ACTIONS(1324),
    [anon_sym_for] = ACTIONS(1324),
    [aux_sym__type_identifier_token1] = ACTIONS(1324),
    [anon_sym_if] = ACTIONS(1324),
    [anon_sym_else] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1322),
    [aux_sym_number_literal_token1] = ACTIONS(1324),
    [aux_sym_number_literal_token2] = ACTIONS(1322),
    [aux_sym_number_literal_token3] = ACTIONS(1322),
    [aux_sym_number_literal_token4] = ACTIONS(1324),
    [aux_sym_number_literal_token5] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1324),
    [anon_sym_SQUOTE] = ACTIONS(1324),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1322),
    [anon_sym_DQUOTE] = ACTIONS(1324),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1322),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1322),
    [anon_sym_switch] = ACTIONS(1324),
    [anon_sym_case] = ACTIONS(1324),
    [anon_sym_default] = ACTIONS(1324),
    [anon_sym_QMARK] = ACTIONS(1324),
    [anon_sym_try] = ACTIONS(1324),
    [anon_sym_int] = ACTIONS(1324),
    [anon_sym_boolean] = ACTIONS(1324),
    [anon_sym_char] = ACTIONS(1324),
    [anon_sym_short] = ACTIONS(1324),
    [anon_sym_long] = ACTIONS(1324),
    [anon_sym_float] = ACTIONS(1324),
    [anon_sym_double] = ACTIONS(1324),
    [anon_sym_void] = ACTIONS(1324),
    [anon_sym_public] = ACTIONS(1324),
    [anon_sym_protected] = ACTIONS(1324),
    [anon_sym_private] = ACTIONS(1324),
    [anon_sym_static] = ACTIONS(1324),
    [anon_sym_final] = ACTIONS(1324),
    [anon_sym_synchronized] = ACTIONS(1324),
    [anon_sym_TILDE] = ACTIONS(1322),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_new] = ACTIONS(1324),
  },
  [STATE(429)] = {
    [sym_closure] = STATE(561),
    [sym_comment] = STATE(429),
    [sym_groovy_doc] = STATE(429),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_case] = ACTIONS(1338),
    [anon_sym_default] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(430)] = {
    [sym_closure] = STATE(562),
    [sym_comment] = STATE(430),
    [sym_groovy_doc] = STATE(430),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(431)] = {
    [sym_closure] = STATE(564),
    [sym_comment] = STATE(431),
    [sym_groovy_doc] = STATE(431),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(432)] = {
    [sym_comment] = STATE(432),
    [sym_groovy_doc] = STATE(432),
    [aux_sym_dotted_identifier_repeat1] = STATE(433),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_break] = ACTIONS(375),
    [anon_sym_continue] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_package] = ACTIONS(375),
    [anon_sym_AT] = ACTIONS(375),
    [anon_sym_assert] = ACTIONS(375),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(373),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(373),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(373),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(373),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(373),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_ATinterface] = ACTIONS(373),
    [anon_sym_interface] = ACTIONS(375),
    [anon_sym_class] = ACTIONS(375),
    [anon_sym_COMMA] = ACTIONS(373),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(375),
    [anon_sym_def] = ACTIONS(375),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_do] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [anon_sym_for] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_if] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_return] = ACTIONS(375),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_switch] = ACTIONS(375),
    [anon_sym_case] = ACTIONS(375),
    [anon_sym_default] = ACTIONS(375),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(375),
    [anon_sym_int] = ACTIONS(375),
    [anon_sym_boolean] = ACTIONS(375),
    [anon_sym_char] = ACTIONS(375),
    [anon_sym_short] = ACTIONS(375),
    [anon_sym_long] = ACTIONS(375),
    [anon_sym_float] = ACTIONS(375),
    [anon_sym_double] = ACTIONS(375),
    [anon_sym_void] = ACTIONS(375),
    [anon_sym_public] = ACTIONS(375),
    [anon_sym_protected] = ACTIONS(375),
    [anon_sym_private] = ACTIONS(375),
    [anon_sym_static] = ACTIONS(375),
    [anon_sym_final] = ACTIONS(375),
    [anon_sym_synchronized] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(433)] = {
    [sym_comment] = STATE(433),
    [sym_groovy_doc] = STATE(433),
    [aux_sym_dotted_identifier_repeat1] = STATE(433),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(1521),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(434)] = {
    [sym_comment] = STATE(434),
    [sym_groovy_doc] = STATE(434),
    [ts_builtin_sym_end] = ACTIONS(1386),
    [sym_identifier] = ACTIONS(1388),
    [anon_sym_break] = ACTIONS(1388),
    [anon_sym_continue] = ACTIONS(1388),
    [anon_sym_SEMI] = ACTIONS(1386),
    [anon_sym_DOT_AMP] = ACTIONS(1386),
    [anon_sym_DOT_AT] = ACTIONS(1386),
    [anon_sym_QMARK_DOT] = ACTIONS(1386),
    [anon_sym_STAR_DOT] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_STAR_COLON] = ACTIONS(1386),
    [anon_sym_DOT] = ACTIONS(1388),
    [anon_sym_import] = ACTIONS(1388),
    [anon_sym_as] = ACTIONS(1388),
    [anon_sym_package] = ACTIONS(1388),
    [anon_sym_AT] = ACTIONS(1388),
    [anon_sym_assert] = ACTIONS(1388),
    [anon_sym_PLUS_PLUS] = ACTIONS(1386),
    [anon_sym_DASH_DASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1388),
    [anon_sym_PLUS] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1388),
    [anon_sym_GT_GT_GT] = ACTIONS(1386),
    [anon_sym_DOT_DOT] = ACTIONS(1388),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1388),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_LT_EQ] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_EQ] = ACTIONS(1386),
    [anon_sym_in] = ACTIONS(1388),
    [anon_sym_BANGin] = ACTIONS(1388),
    [anon_sym_instanceof] = ACTIONS(1388),
    [anon_sym_BANGinstanceof] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1388),
    [anon_sym_BANG_EQ] = ACTIONS(1388),
    [anon_sym_LT_EQ_GT] = ACTIONS(1386),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1386),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1386),
    [anon_sym_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_PIPE] = ACTIONS(1388),
    [anon_sym_AMP_AMP] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1386),
    [anon_sym_QMARK_COLON] = ACTIONS(1386),
    [anon_sym_STAR_STAR] = ACTIONS(1386),
    [anon_sym_true] = ACTIONS(1388),
    [anon_sym_false] = ACTIONS(1388),
    [anon_sym_ATinterface] = ACTIONS(1386),
    [anon_sym_interface] = ACTIONS(1388),
    [anon_sym_class] = ACTIONS(1388),
    [anon_sym_COMMA] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1386),
    [anon_sym_DASH_GT] = ACTIONS(1386),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1388),
    [anon_sym_def] = ACTIONS(1388),
    [anon_sym_LPAREN] = ACTIONS(1386),
    [anon_sym_null] = ACTIONS(1388),
    [anon_sym_this] = ACTIONS(1388),
    [anon_sym_do] = ACTIONS(1388),
    [anon_sym_while] = ACTIONS(1388),
    [anon_sym_for] = ACTIONS(1388),
    [aux_sym__type_identifier_token1] = ACTIONS(1388),
    [anon_sym_if] = ACTIONS(1388),
    [anon_sym_else] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1386),
    [aux_sym_number_literal_token1] = ACTIONS(1388),
    [aux_sym_number_literal_token2] = ACTIONS(1386),
    [aux_sym_number_literal_token3] = ACTIONS(1386),
    [aux_sym_number_literal_token4] = ACTIONS(1388),
    [aux_sym_number_literal_token5] = ACTIONS(1386),
    [anon_sym_pipeline] = ACTIONS(1388),
    [anon_sym_return] = ACTIONS(1388),
    [anon_sym_SQUOTE] = ACTIONS(1388),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1386),
    [anon_sym_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1386),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1386),
    [anon_sym_switch] = ACTIONS(1388),
    [anon_sym_QMARK] = ACTIONS(1388),
    [anon_sym_try] = ACTIONS(1388),
    [anon_sym_int] = ACTIONS(1388),
    [anon_sym_boolean] = ACTIONS(1388),
    [anon_sym_char] = ACTIONS(1388),
    [anon_sym_short] = ACTIONS(1388),
    [anon_sym_long] = ACTIONS(1388),
    [anon_sym_float] = ACTIONS(1388),
    [anon_sym_double] = ACTIONS(1388),
    [anon_sym_void] = ACTIONS(1388),
    [anon_sym_public] = ACTIONS(1388),
    [anon_sym_protected] = ACTIONS(1388),
    [anon_sym_private] = ACTIONS(1388),
    [anon_sym_static] = ACTIONS(1388),
    [anon_sym_final] = ACTIONS(1388),
    [anon_sym_synchronized] = ACTIONS(1388),
    [anon_sym_TILDE] = ACTIONS(1386),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_new] = ACTIONS(1388),
  },
  [STATE(435)] = {
    [sym_comment] = STATE(435),
    [sym_groovy_doc] = STATE(435),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(436)] = {
    [sym_comment] = STATE(436),
    [sym_groovy_doc] = STATE(436),
    [ts_builtin_sym_end] = ACTIONS(1451),
    [sym_identifier] = ACTIONS(1449),
    [anon_sym_break] = ACTIONS(1449),
    [anon_sym_continue] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_DOT_AMP] = ACTIONS(1451),
    [anon_sym_DOT_AT] = ACTIONS(1451),
    [anon_sym_QMARK_DOT] = ACTIONS(1451),
    [anon_sym_STAR_DOT] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1449),
    [anon_sym_STAR_COLON] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1449),
    [anon_sym_as] = ACTIONS(1449),
    [anon_sym_package] = ACTIONS(1449),
    [anon_sym_AT] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1449),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_DOT_DOT] = ACTIONS(1449),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1449),
    [anon_sym_LT] = ACTIONS(1449),
    [anon_sym_LT_EQ] = ACTIONS(1449),
    [anon_sym_GT] = ACTIONS(1449),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1449),
    [anon_sym_BANGin] = ACTIONS(1449),
    [anon_sym_instanceof] = ACTIONS(1449),
    [anon_sym_BANGinstanceof] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1449),
    [anon_sym_BANG_EQ] = ACTIONS(1449),
    [anon_sym_LT_EQ_GT] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1449),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1449),
    [anon_sym_AMP_AMP] = ACTIONS(1451),
    [anon_sym_PIPE_PIPE] = ACTIONS(1451),
    [anon_sym_QMARK_COLON] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_true] = ACTIONS(1449),
    [anon_sym_false] = ACTIONS(1449),
    [anon_sym_ATinterface] = ACTIONS(1451),
    [anon_sym_interface] = ACTIONS(1449),
    [anon_sym_class] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [anon_sym_RBRACE] = ACTIONS(1451),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1449),
    [anon_sym_def] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_null] = ACTIONS(1449),
    [anon_sym_this] = ACTIONS(1449),
    [anon_sym_do] = ACTIONS(1449),
    [anon_sym_while] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1449),
    [aux_sym__type_identifier_token1] = ACTIONS(1449),
    [anon_sym_if] = ACTIONS(1449),
    [anon_sym_else] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_number_literal_token1] = ACTIONS(1449),
    [aux_sym_number_literal_token2] = ACTIONS(1451),
    [aux_sym_number_literal_token3] = ACTIONS(1451),
    [aux_sym_number_literal_token4] = ACTIONS(1449),
    [aux_sym_number_literal_token5] = ACTIONS(1451),
    [anon_sym_pipeline] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_SQUOTE] = ACTIONS(1449),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1451),
    [anon_sym_DQUOTE] = ACTIONS(1449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1451),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1451),
    [anon_sym_switch] = ACTIONS(1449),
    [anon_sym_QMARK] = ACTIONS(1449),
    [anon_sym_try] = ACTIONS(1449),
    [anon_sym_catch] = ACTIONS(1449),
    [anon_sym_finally] = ACTIONS(1449),
    [anon_sym_int] = ACTIONS(1449),
    [anon_sym_boolean] = ACTIONS(1449),
    [anon_sym_char] = ACTIONS(1449),
    [anon_sym_short] = ACTIONS(1449),
    [anon_sym_long] = ACTIONS(1449),
    [anon_sym_float] = ACTIONS(1449),
    [anon_sym_double] = ACTIONS(1449),
    [anon_sym_void] = ACTIONS(1449),
    [anon_sym_public] = ACTIONS(1449),
    [anon_sym_protected] = ACTIONS(1449),
    [anon_sym_private] = ACTIONS(1449),
    [anon_sym_static] = ACTIONS(1449),
    [anon_sym_final] = ACTIONS(1449),
    [anon_sym_synchronized] = ACTIONS(1449),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1449),
    [anon_sym_new] = ACTIONS(1449),
  },
  [STATE(437)] = {
    [sym_comment] = STATE(437),
    [sym_groovy_doc] = STATE(437),
    [ts_builtin_sym_end] = ACTIONS(1434),
    [sym_identifier] = ACTIONS(1432),
    [anon_sym_break] = ACTIONS(1432),
    [anon_sym_continue] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1434),
    [anon_sym_DOT_AMP] = ACTIONS(1434),
    [anon_sym_DOT_AT] = ACTIONS(1434),
    [anon_sym_QMARK_DOT] = ACTIONS(1434),
    [anon_sym_STAR_DOT] = ACTIONS(1434),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_STAR_COLON] = ACTIONS(1434),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_import] = ACTIONS(1432),
    [anon_sym_as] = ACTIONS(1432),
    [anon_sym_package] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_assert] = ACTIONS(1432),
    [anon_sym_PLUS_PLUS] = ACTIONS(1434),
    [anon_sym_DASH_DASH] = ACTIONS(1434),
    [anon_sym_PERCENT] = ACTIONS(1434),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_LT_LT] = ACTIONS(1434),
    [anon_sym_GT_GT] = ACTIONS(1432),
    [anon_sym_GT_GT_GT] = ACTIONS(1434),
    [anon_sym_DOT_DOT] = ACTIONS(1432),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_LT_EQ] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_GT_EQ] = ACTIONS(1434),
    [anon_sym_in] = ACTIONS(1432),
    [anon_sym_BANGin] = ACTIONS(1432),
    [anon_sym_instanceof] = ACTIONS(1432),
    [anon_sym_BANGinstanceof] = ACTIONS(1434),
    [anon_sym_EQ_EQ] = ACTIONS(1432),
    [anon_sym_BANG_EQ] = ACTIONS(1432),
    [anon_sym_LT_EQ_GT] = ACTIONS(1434),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1434),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1434),
    [anon_sym_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1434),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_AMP_AMP] = ACTIONS(1434),
    [anon_sym_PIPE_PIPE] = ACTIONS(1434),
    [anon_sym_QMARK_COLON] = ACTIONS(1434),
    [anon_sym_STAR_STAR] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1432),
    [anon_sym_false] = ACTIONS(1432),
    [anon_sym_ATinterface] = ACTIONS(1434),
    [anon_sym_interface] = ACTIONS(1432),
    [anon_sym_class] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(1434),
    [anon_sym_DASH_GT] = ACTIONS(1434),
    [anon_sym_RBRACE] = ACTIONS(1434),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_def] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1434),
    [anon_sym_null] = ACTIONS(1432),
    [anon_sym_this] = ACTIONS(1432),
    [anon_sym_do] = ACTIONS(1432),
    [anon_sym_while] = ACTIONS(1432),
    [anon_sym_for] = ACTIONS(1432),
    [aux_sym__type_identifier_token1] = ACTIONS(1432),
    [anon_sym_if] = ACTIONS(1432),
    [anon_sym_else] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [aux_sym_number_literal_token1] = ACTIONS(1432),
    [aux_sym_number_literal_token2] = ACTIONS(1434),
    [aux_sym_number_literal_token3] = ACTIONS(1434),
    [aux_sym_number_literal_token4] = ACTIONS(1432),
    [aux_sym_number_literal_token5] = ACTIONS(1434),
    [anon_sym_pipeline] = ACTIONS(1432),
    [anon_sym_return] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1434),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1434),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1434),
    [anon_sym_switch] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_try] = ACTIONS(1432),
    [anon_sym_int] = ACTIONS(1432),
    [anon_sym_boolean] = ACTIONS(1432),
    [anon_sym_char] = ACTIONS(1432),
    [anon_sym_short] = ACTIONS(1432),
    [anon_sym_long] = ACTIONS(1432),
    [anon_sym_float] = ACTIONS(1432),
    [anon_sym_double] = ACTIONS(1432),
    [anon_sym_void] = ACTIONS(1432),
    [anon_sym_public] = ACTIONS(1432),
    [anon_sym_protected] = ACTIONS(1432),
    [anon_sym_private] = ACTIONS(1432),
    [anon_sym_static] = ACTIONS(1432),
    [anon_sym_final] = ACTIONS(1432),
    [anon_sym_synchronized] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1434),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_new] = ACTIONS(1432),
  },
  [STATE(438)] = {
    [sym_comment] = STATE(438),
    [sym_groovy_doc] = STATE(438),
    [ts_builtin_sym_end] = ACTIONS(1457),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_pipeline] = ACTIONS(1455),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_catch] = ACTIONS(1455),
    [anon_sym_finally] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(439)] = {
    [sym_comment] = STATE(439),
    [sym_groovy_doc] = STATE(439),
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_DASH_GT] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_pipeline] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(440)] = {
    [sym_comment] = STATE(440),
    [sym_groovy_doc] = STATE(440),
    [ts_builtin_sym_end] = ACTIONS(1457),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1457),
    [anon_sym_DOT_AT] = ACTIONS(1457),
    [anon_sym_QMARK_DOT] = ACTIONS(1457),
    [anon_sym_STAR_DOT] = ACTIONS(1457),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1457),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_pipeline] = ACTIONS(1455),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_catch] = ACTIONS(1455),
    [anon_sym_finally] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(441)] = {
    [sym_comment] = STATE(441),
    [sym_groovy_doc] = STATE(441),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(442)] = {
    [sym_comment] = STATE(442),
    [sym_groovy_doc] = STATE(442),
    [ts_builtin_sym_end] = ACTIONS(1278),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1278),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_pipeline] = ACTIONS(1280),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_catch] = ACTIONS(1280),
    [anon_sym_finally] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(443)] = {
    [sym_comment] = STATE(443),
    [sym_groovy_doc] = STATE(443),
    [aux_sym_dotted_identifier_repeat1] = STATE(432),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1453),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1528),
    [anon_sym_DASH_DASH] = ACTIONS(1528),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(444)] = {
    [sym_comment] = STATE(444),
    [sym_groovy_doc] = STATE(444),
    [ts_builtin_sym_end] = ACTIONS(1394),
    [sym_identifier] = ACTIONS(1396),
    [anon_sym_break] = ACTIONS(1396),
    [anon_sym_continue] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1394),
    [anon_sym_DOT_AMP] = ACTIONS(1394),
    [anon_sym_DOT_AT] = ACTIONS(1394),
    [anon_sym_QMARK_DOT] = ACTIONS(1394),
    [anon_sym_STAR_DOT] = ACTIONS(1394),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_STAR_COLON] = ACTIONS(1394),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_import] = ACTIONS(1396),
    [anon_sym_as] = ACTIONS(1396),
    [anon_sym_package] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_assert] = ACTIONS(1396),
    [anon_sym_PLUS_PLUS] = ACTIONS(1394),
    [anon_sym_DASH_DASH] = ACTIONS(1394),
    [anon_sym_PERCENT] = ACTIONS(1394),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_LT_LT] = ACTIONS(1394),
    [anon_sym_GT_GT] = ACTIONS(1396),
    [anon_sym_GT_GT_GT] = ACTIONS(1394),
    [anon_sym_DOT_DOT] = ACTIONS(1396),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_LT_EQ] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_GT_EQ] = ACTIONS(1394),
    [anon_sym_in] = ACTIONS(1396),
    [anon_sym_BANGin] = ACTIONS(1396),
    [anon_sym_instanceof] = ACTIONS(1396),
    [anon_sym_BANGinstanceof] = ACTIONS(1394),
    [anon_sym_EQ_EQ] = ACTIONS(1396),
    [anon_sym_BANG_EQ] = ACTIONS(1396),
    [anon_sym_LT_EQ_GT] = ACTIONS(1394),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1394),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1394),
    [anon_sym_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1394),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_AMP_AMP] = ACTIONS(1394),
    [anon_sym_PIPE_PIPE] = ACTIONS(1394),
    [anon_sym_QMARK_COLON] = ACTIONS(1394),
    [anon_sym_STAR_STAR] = ACTIONS(1394),
    [anon_sym_true] = ACTIONS(1396),
    [anon_sym_false] = ACTIONS(1396),
    [anon_sym_ATinterface] = ACTIONS(1394),
    [anon_sym_interface] = ACTIONS(1396),
    [anon_sym_class] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(1394),
    [anon_sym_DASH_GT] = ACTIONS(1394),
    [anon_sym_RBRACE] = ACTIONS(1394),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_def] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1394),
    [anon_sym_null] = ACTIONS(1396),
    [anon_sym_this] = ACTIONS(1396),
    [anon_sym_do] = ACTIONS(1396),
    [anon_sym_while] = ACTIONS(1396),
    [anon_sym_for] = ACTIONS(1396),
    [aux_sym__type_identifier_token1] = ACTIONS(1396),
    [anon_sym_if] = ACTIONS(1396),
    [anon_sym_else] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [aux_sym_number_literal_token1] = ACTIONS(1396),
    [aux_sym_number_literal_token2] = ACTIONS(1394),
    [aux_sym_number_literal_token3] = ACTIONS(1394),
    [aux_sym_number_literal_token4] = ACTIONS(1396),
    [aux_sym_number_literal_token5] = ACTIONS(1394),
    [anon_sym_pipeline] = ACTIONS(1396),
    [anon_sym_return] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1394),
    [anon_sym_switch] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_try] = ACTIONS(1396),
    [anon_sym_int] = ACTIONS(1396),
    [anon_sym_boolean] = ACTIONS(1396),
    [anon_sym_char] = ACTIONS(1396),
    [anon_sym_short] = ACTIONS(1396),
    [anon_sym_long] = ACTIONS(1396),
    [anon_sym_float] = ACTIONS(1396),
    [anon_sym_double] = ACTIONS(1396),
    [anon_sym_void] = ACTIONS(1396),
    [anon_sym_public] = ACTIONS(1396),
    [anon_sym_protected] = ACTIONS(1396),
    [anon_sym_private] = ACTIONS(1396),
    [anon_sym_static] = ACTIONS(1396),
    [anon_sym_final] = ACTIONS(1396),
    [anon_sym_synchronized] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1394),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_new] = ACTIONS(1396),
  },
  [STATE(445)] = {
    [sym_comment] = STATE(445),
    [sym_groovy_doc] = STATE(445),
    [ts_builtin_sym_end] = ACTIONS(1398),
    [sym_identifier] = ACTIONS(1400),
    [anon_sym_break] = ACTIONS(1400),
    [anon_sym_continue] = ACTIONS(1400),
    [anon_sym_SEMI] = ACTIONS(1398),
    [anon_sym_DOT_AMP] = ACTIONS(1398),
    [anon_sym_DOT_AT] = ACTIONS(1398),
    [anon_sym_QMARK_DOT] = ACTIONS(1398),
    [anon_sym_STAR_DOT] = ACTIONS(1398),
    [anon_sym_STAR] = ACTIONS(1400),
    [anon_sym_STAR_COLON] = ACTIONS(1398),
    [anon_sym_DOT] = ACTIONS(1400),
    [anon_sym_import] = ACTIONS(1400),
    [anon_sym_as] = ACTIONS(1400),
    [anon_sym_package] = ACTIONS(1400),
    [anon_sym_AT] = ACTIONS(1400),
    [anon_sym_assert] = ACTIONS(1400),
    [anon_sym_PLUS_PLUS] = ACTIONS(1398),
    [anon_sym_DASH_DASH] = ACTIONS(1398),
    [anon_sym_PERCENT] = ACTIONS(1398),
    [anon_sym_SLASH] = ACTIONS(1400),
    [anon_sym_PLUS] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_LT_LT] = ACTIONS(1398),
    [anon_sym_GT_GT] = ACTIONS(1400),
    [anon_sym_GT_GT_GT] = ACTIONS(1398),
    [anon_sym_DOT_DOT] = ACTIONS(1400),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1400),
    [anon_sym_LT] = ACTIONS(1400),
    [anon_sym_LT_EQ] = ACTIONS(1400),
    [anon_sym_GT] = ACTIONS(1400),
    [anon_sym_GT_EQ] = ACTIONS(1398),
    [anon_sym_in] = ACTIONS(1400),
    [anon_sym_BANGin] = ACTIONS(1400),
    [anon_sym_instanceof] = ACTIONS(1400),
    [anon_sym_BANGinstanceof] = ACTIONS(1398),
    [anon_sym_EQ_EQ] = ACTIONS(1400),
    [anon_sym_BANG_EQ] = ACTIONS(1400),
    [anon_sym_LT_EQ_GT] = ACTIONS(1398),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1398),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1398),
    [anon_sym_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_AMP] = ACTIONS(1400),
    [anon_sym_CARET] = ACTIONS(1398),
    [anon_sym_PIPE] = ACTIONS(1400),
    [anon_sym_AMP_AMP] = ACTIONS(1398),
    [anon_sym_PIPE_PIPE] = ACTIONS(1398),
    [anon_sym_QMARK_COLON] = ACTIONS(1398),
    [anon_sym_STAR_STAR] = ACTIONS(1398),
    [anon_sym_true] = ACTIONS(1400),
    [anon_sym_false] = ACTIONS(1400),
    [anon_sym_ATinterface] = ACTIONS(1398),
    [anon_sym_interface] = ACTIONS(1400),
    [anon_sym_class] = ACTIONS(1400),
    [anon_sym_COMMA] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1398),
    [anon_sym_DASH_GT] = ACTIONS(1398),
    [anon_sym_RBRACE] = ACTIONS(1398),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1400),
    [anon_sym_def] = ACTIONS(1400),
    [anon_sym_LPAREN] = ACTIONS(1398),
    [anon_sym_null] = ACTIONS(1400),
    [anon_sym_this] = ACTIONS(1400),
    [anon_sym_do] = ACTIONS(1400),
    [anon_sym_while] = ACTIONS(1400),
    [anon_sym_for] = ACTIONS(1400),
    [aux_sym__type_identifier_token1] = ACTIONS(1400),
    [anon_sym_if] = ACTIONS(1400),
    [anon_sym_else] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1398),
    [aux_sym_number_literal_token1] = ACTIONS(1400),
    [aux_sym_number_literal_token2] = ACTIONS(1398),
    [aux_sym_number_literal_token3] = ACTIONS(1398),
    [aux_sym_number_literal_token4] = ACTIONS(1400),
    [aux_sym_number_literal_token5] = ACTIONS(1398),
    [anon_sym_pipeline] = ACTIONS(1400),
    [anon_sym_return] = ACTIONS(1400),
    [anon_sym_SQUOTE] = ACTIONS(1400),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1398),
    [anon_sym_DQUOTE] = ACTIONS(1400),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1398),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1398),
    [anon_sym_switch] = ACTIONS(1400),
    [anon_sym_QMARK] = ACTIONS(1400),
    [anon_sym_try] = ACTIONS(1400),
    [anon_sym_int] = ACTIONS(1400),
    [anon_sym_boolean] = ACTIONS(1400),
    [anon_sym_char] = ACTIONS(1400),
    [anon_sym_short] = ACTIONS(1400),
    [anon_sym_long] = ACTIONS(1400),
    [anon_sym_float] = ACTIONS(1400),
    [anon_sym_double] = ACTIONS(1400),
    [anon_sym_void] = ACTIONS(1400),
    [anon_sym_public] = ACTIONS(1400),
    [anon_sym_protected] = ACTIONS(1400),
    [anon_sym_private] = ACTIONS(1400),
    [anon_sym_static] = ACTIONS(1400),
    [anon_sym_final] = ACTIONS(1400),
    [anon_sym_synchronized] = ACTIONS(1400),
    [anon_sym_TILDE] = ACTIONS(1398),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_new] = ACTIONS(1400),
  },
  [STATE(446)] = {
    [sym_comment] = STATE(446),
    [sym_groovy_doc] = STATE(446),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1370),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(447)] = {
    [sym_comment] = STATE(447),
    [sym_groovy_doc] = STATE(447),
    [ts_builtin_sym_end] = ACTIONS(1465),
    [sym_identifier] = ACTIONS(1463),
    [anon_sym_break] = ACTIONS(1463),
    [anon_sym_continue] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_DOT_AMP] = ACTIONS(1465),
    [anon_sym_DOT_AT] = ACTIONS(1465),
    [anon_sym_QMARK_DOT] = ACTIONS(1465),
    [anon_sym_STAR_DOT] = ACTIONS(1465),
    [anon_sym_STAR] = ACTIONS(1463),
    [anon_sym_STAR_COLON] = ACTIONS(1465),
    [anon_sym_import] = ACTIONS(1463),
    [anon_sym_as] = ACTIONS(1463),
    [anon_sym_package] = ACTIONS(1463),
    [anon_sym_AT] = ACTIONS(1463),
    [anon_sym_assert] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_PERCENT] = ACTIONS(1465),
    [anon_sym_SLASH] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [anon_sym_LT_LT] = ACTIONS(1465),
    [anon_sym_GT_GT] = ACTIONS(1463),
    [anon_sym_GT_GT_GT] = ACTIONS(1465),
    [anon_sym_DOT_DOT] = ACTIONS(1463),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1463),
    [anon_sym_LT] = ACTIONS(1463),
    [anon_sym_LT_EQ] = ACTIONS(1463),
    [anon_sym_GT] = ACTIONS(1463),
    [anon_sym_GT_EQ] = ACTIONS(1465),
    [anon_sym_in] = ACTIONS(1463),
    [anon_sym_BANGin] = ACTIONS(1463),
    [anon_sym_instanceof] = ACTIONS(1463),
    [anon_sym_BANGinstanceof] = ACTIONS(1465),
    [anon_sym_EQ_EQ] = ACTIONS(1463),
    [anon_sym_BANG_EQ] = ACTIONS(1463),
    [anon_sym_LT_EQ_GT] = ACTIONS(1465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1465),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1465),
    [anon_sym_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1463),
    [anon_sym_CARET] = ACTIONS(1465),
    [anon_sym_PIPE] = ACTIONS(1463),
    [anon_sym_AMP_AMP] = ACTIONS(1465),
    [anon_sym_PIPE_PIPE] = ACTIONS(1465),
    [anon_sym_QMARK_COLON] = ACTIONS(1465),
    [anon_sym_STAR_STAR] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(1463),
    [anon_sym_false] = ACTIONS(1463),
    [anon_sym_ATinterface] = ACTIONS(1465),
    [anon_sym_interface] = ACTIONS(1463),
    [anon_sym_class] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [anon_sym_RBRACE] = ACTIONS(1465),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1463),
    [anon_sym_def] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_null] = ACTIONS(1463),
    [anon_sym_this] = ACTIONS(1463),
    [anon_sym_do] = ACTIONS(1463),
    [anon_sym_while] = ACTIONS(1463),
    [anon_sym_for] = ACTIONS(1463),
    [aux_sym__type_identifier_token1] = ACTIONS(1463),
    [anon_sym_if] = ACTIONS(1463),
    [anon_sym_else] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_number_literal_token1] = ACTIONS(1463),
    [aux_sym_number_literal_token2] = ACTIONS(1465),
    [aux_sym_number_literal_token3] = ACTIONS(1465),
    [aux_sym_number_literal_token4] = ACTIONS(1463),
    [aux_sym_number_literal_token5] = ACTIONS(1465),
    [anon_sym_pipeline] = ACTIONS(1463),
    [anon_sym_return] = ACTIONS(1463),
    [anon_sym_SQUOTE] = ACTIONS(1463),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(1463),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1465),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1465),
    [anon_sym_switch] = ACTIONS(1463),
    [anon_sym_QMARK] = ACTIONS(1463),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_catch] = ACTIONS(1463),
    [anon_sym_finally] = ACTIONS(1463),
    [anon_sym_int] = ACTIONS(1463),
    [anon_sym_boolean] = ACTIONS(1463),
    [anon_sym_char] = ACTIONS(1463),
    [anon_sym_short] = ACTIONS(1463),
    [anon_sym_long] = ACTIONS(1463),
    [anon_sym_float] = ACTIONS(1463),
    [anon_sym_double] = ACTIONS(1463),
    [anon_sym_void] = ACTIONS(1463),
    [anon_sym_public] = ACTIONS(1463),
    [anon_sym_protected] = ACTIONS(1463),
    [anon_sym_private] = ACTIONS(1463),
    [anon_sym_static] = ACTIONS(1463),
    [anon_sym_final] = ACTIONS(1463),
    [anon_sym_synchronized] = ACTIONS(1463),
    [anon_sym_TILDE] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1463),
    [anon_sym_new] = ACTIONS(1463),
  },
  [STATE(448)] = {
    [sym_comment] = STATE(448),
    [sym_groovy_doc] = STATE(448),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(449)] = {
    [sym_comment] = STATE(449),
    [sym_groovy_doc] = STATE(449),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(450)] = {
    [sym_comment] = STATE(450),
    [sym_groovy_doc] = STATE(450),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(451)] = {
    [sym_comment] = STATE(451),
    [sym_groovy_doc] = STATE(451),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(452)] = {
    [sym_comment] = STATE(452),
    [sym_groovy_doc] = STATE(452),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(453)] = {
    [sym_comment] = STATE(453),
    [sym_groovy_doc] = STATE(453),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(454)] = {
    [sym_comment] = STATE(454),
    [sym_groovy_doc] = STATE(454),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(455)] = {
    [sym_comment] = STATE(455),
    [sym_groovy_doc] = STATE(455),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(456)] = {
    [sym_comment] = STATE(456),
    [sym_groovy_doc] = STATE(456),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(457)] = {
    [sym_comment] = STATE(457),
    [sym_groovy_doc] = STATE(457),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(458)] = {
    [sym_comment] = STATE(458),
    [sym_groovy_doc] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(459)] = {
    [sym_comment] = STATE(459),
    [sym_groovy_doc] = STATE(459),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_catch] = ACTIONS(1467),
    [anon_sym_finally] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(460)] = {
    [sym_comment] = STATE(460),
    [sym_groovy_doc] = STATE(460),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1370),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(461)] = {
    [sym_comment] = STATE(461),
    [sym_groovy_doc] = STATE(461),
    [ts_builtin_sym_end] = ACTIONS(1501),
    [sym_identifier] = ACTIONS(1499),
    [anon_sym_break] = ACTIONS(1499),
    [anon_sym_continue] = ACTIONS(1499),
    [anon_sym_SEMI] = ACTIONS(1501),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1501),
    [anon_sym_import] = ACTIONS(1499),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1499),
    [anon_sym_AT] = ACTIONS(1499),
    [anon_sym_assert] = ACTIONS(1499),
    [anon_sym_PLUS_PLUS] = ACTIONS(1501),
    [anon_sym_DASH_DASH] = ACTIONS(1501),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1499),
    [anon_sym_false] = ACTIONS(1499),
    [anon_sym_ATinterface] = ACTIONS(1501),
    [anon_sym_interface] = ACTIONS(1499),
    [anon_sym_class] = ACTIONS(1499),
    [anon_sym_COMMA] = ACTIONS(1501),
    [anon_sym_LBRACE] = ACTIONS(1501),
    [anon_sym_RBRACE] = ACTIONS(1501),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1499),
    [anon_sym_def] = ACTIONS(1499),
    [anon_sym_LPAREN] = ACTIONS(1501),
    [anon_sym_null] = ACTIONS(1499),
    [anon_sym_this] = ACTIONS(1499),
    [anon_sym_do] = ACTIONS(1499),
    [anon_sym_while] = ACTIONS(1499),
    [anon_sym_for] = ACTIONS(1499),
    [aux_sym__type_identifier_token1] = ACTIONS(1499),
    [anon_sym_if] = ACTIONS(1499),
    [anon_sym_else] = ACTIONS(1499),
    [anon_sym_LBRACK] = ACTIONS(1501),
    [aux_sym_number_literal_token1] = ACTIONS(1499),
    [aux_sym_number_literal_token2] = ACTIONS(1501),
    [aux_sym_number_literal_token3] = ACTIONS(1501),
    [aux_sym_number_literal_token4] = ACTIONS(1499),
    [aux_sym_number_literal_token5] = ACTIONS(1501),
    [anon_sym_pipeline] = ACTIONS(1499),
    [anon_sym_return] = ACTIONS(1499),
    [anon_sym_SQUOTE] = ACTIONS(1499),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(1499),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1501),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1501),
    [anon_sym_switch] = ACTIONS(1499),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1499),
    [anon_sym_catch] = ACTIONS(1499),
    [anon_sym_finally] = ACTIONS(1499),
    [anon_sym_int] = ACTIONS(1499),
    [anon_sym_boolean] = ACTIONS(1499),
    [anon_sym_char] = ACTIONS(1499),
    [anon_sym_short] = ACTIONS(1499),
    [anon_sym_long] = ACTIONS(1499),
    [anon_sym_float] = ACTIONS(1499),
    [anon_sym_double] = ACTIONS(1499),
    [anon_sym_void] = ACTIONS(1499),
    [anon_sym_public] = ACTIONS(1499),
    [anon_sym_protected] = ACTIONS(1499),
    [anon_sym_private] = ACTIONS(1499),
    [anon_sym_static] = ACTIONS(1499),
    [anon_sym_final] = ACTIONS(1499),
    [anon_sym_synchronized] = ACTIONS(1499),
    [anon_sym_TILDE] = ACTIONS(1501),
    [anon_sym_BANG] = ACTIONS(1499),
    [anon_sym_new] = ACTIONS(1499),
  },
  [STATE(462)] = {
    [sym_comment] = STATE(462),
    [sym_groovy_doc] = STATE(462),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(463)] = {
    [sym_comment] = STATE(463),
    [sym_groovy_doc] = STATE(463),
    [ts_builtin_sym_end] = ACTIONS(1414),
    [sym_identifier] = ACTIONS(1416),
    [anon_sym_break] = ACTIONS(1416),
    [anon_sym_continue] = ACTIONS(1416),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_DOT_AMP] = ACTIONS(1414),
    [anon_sym_DOT_AT] = ACTIONS(1414),
    [anon_sym_QMARK_DOT] = ACTIONS(1414),
    [anon_sym_STAR_DOT] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1416),
    [anon_sym_STAR_COLON] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1416),
    [anon_sym_import] = ACTIONS(1416),
    [anon_sym_as] = ACTIONS(1416),
    [anon_sym_package] = ACTIONS(1416),
    [anon_sym_AT] = ACTIONS(1416),
    [anon_sym_assert] = ACTIONS(1416),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_LT_LT] = ACTIONS(1414),
    [anon_sym_GT_GT] = ACTIONS(1416),
    [anon_sym_GT_GT_GT] = ACTIONS(1414),
    [anon_sym_DOT_DOT] = ACTIONS(1416),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1416),
    [anon_sym_LT] = ACTIONS(1416),
    [anon_sym_LT_EQ] = ACTIONS(1416),
    [anon_sym_GT] = ACTIONS(1416),
    [anon_sym_GT_EQ] = ACTIONS(1414),
    [anon_sym_in] = ACTIONS(1416),
    [anon_sym_BANGin] = ACTIONS(1416),
    [anon_sym_instanceof] = ACTIONS(1416),
    [anon_sym_BANGinstanceof] = ACTIONS(1414),
    [anon_sym_EQ_EQ] = ACTIONS(1416),
    [anon_sym_BANG_EQ] = ACTIONS(1416),
    [anon_sym_LT_EQ_GT] = ACTIONS(1414),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1414),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1414),
    [anon_sym_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1416),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1416),
    [anon_sym_AMP_AMP] = ACTIONS(1414),
    [anon_sym_PIPE_PIPE] = ACTIONS(1414),
    [anon_sym_QMARK_COLON] = ACTIONS(1414),
    [anon_sym_STAR_STAR] = ACTIONS(1414),
    [anon_sym_true] = ACTIONS(1416),
    [anon_sym_false] = ACTIONS(1416),
    [anon_sym_ATinterface] = ACTIONS(1414),
    [anon_sym_interface] = ACTIONS(1416),
    [anon_sym_class] = ACTIONS(1416),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_DASH_GT] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1416),
    [anon_sym_def] = ACTIONS(1416),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_null] = ACTIONS(1416),
    [anon_sym_this] = ACTIONS(1416),
    [anon_sym_do] = ACTIONS(1416),
    [anon_sym_while] = ACTIONS(1416),
    [anon_sym_for] = ACTIONS(1416),
    [aux_sym__type_identifier_token1] = ACTIONS(1416),
    [anon_sym_if] = ACTIONS(1416),
    [anon_sym_else] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1414),
    [aux_sym_number_literal_token1] = ACTIONS(1416),
    [aux_sym_number_literal_token2] = ACTIONS(1414),
    [aux_sym_number_literal_token3] = ACTIONS(1414),
    [aux_sym_number_literal_token4] = ACTIONS(1416),
    [aux_sym_number_literal_token5] = ACTIONS(1414),
    [anon_sym_pipeline] = ACTIONS(1416),
    [anon_sym_return] = ACTIONS(1416),
    [anon_sym_SQUOTE] = ACTIONS(1416),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1416),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1414),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1414),
    [anon_sym_switch] = ACTIONS(1416),
    [anon_sym_QMARK] = ACTIONS(1416),
    [anon_sym_try] = ACTIONS(1416),
    [anon_sym_int] = ACTIONS(1416),
    [anon_sym_boolean] = ACTIONS(1416),
    [anon_sym_char] = ACTIONS(1416),
    [anon_sym_short] = ACTIONS(1416),
    [anon_sym_long] = ACTIONS(1416),
    [anon_sym_float] = ACTIONS(1416),
    [anon_sym_double] = ACTIONS(1416),
    [anon_sym_void] = ACTIONS(1416),
    [anon_sym_public] = ACTIONS(1416),
    [anon_sym_protected] = ACTIONS(1416),
    [anon_sym_private] = ACTIONS(1416),
    [anon_sym_static] = ACTIONS(1416),
    [anon_sym_final] = ACTIONS(1416),
    [anon_sym_synchronized] = ACTIONS(1416),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_new] = ACTIONS(1416),
  },
  [STATE(464)] = {
    [sym_comment] = STATE(464),
    [sym_groovy_doc] = STATE(464),
    [ts_builtin_sym_end] = ACTIONS(1418),
    [sym_identifier] = ACTIONS(1420),
    [anon_sym_break] = ACTIONS(1420),
    [anon_sym_continue] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1418),
    [anon_sym_DOT_AMP] = ACTIONS(1418),
    [anon_sym_DOT_AT] = ACTIONS(1418),
    [anon_sym_QMARK_DOT] = ACTIONS(1418),
    [anon_sym_STAR_DOT] = ACTIONS(1418),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_STAR_COLON] = ACTIONS(1418),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_import] = ACTIONS(1420),
    [anon_sym_as] = ACTIONS(1420),
    [anon_sym_package] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_assert] = ACTIONS(1420),
    [anon_sym_PLUS_PLUS] = ACTIONS(1418),
    [anon_sym_DASH_DASH] = ACTIONS(1418),
    [anon_sym_PERCENT] = ACTIONS(1418),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_LT_LT] = ACTIONS(1418),
    [anon_sym_GT_GT] = ACTIONS(1420),
    [anon_sym_GT_GT_GT] = ACTIONS(1418),
    [anon_sym_DOT_DOT] = ACTIONS(1420),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_LT_EQ] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_GT_EQ] = ACTIONS(1418),
    [anon_sym_in] = ACTIONS(1420),
    [anon_sym_BANGin] = ACTIONS(1420),
    [anon_sym_instanceof] = ACTIONS(1420),
    [anon_sym_BANGinstanceof] = ACTIONS(1418),
    [anon_sym_EQ_EQ] = ACTIONS(1420),
    [anon_sym_BANG_EQ] = ACTIONS(1420),
    [anon_sym_LT_EQ_GT] = ACTIONS(1418),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1418),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1418),
    [anon_sym_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1418),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_AMP_AMP] = ACTIONS(1418),
    [anon_sym_PIPE_PIPE] = ACTIONS(1418),
    [anon_sym_QMARK_COLON] = ACTIONS(1418),
    [anon_sym_STAR_STAR] = ACTIONS(1418),
    [anon_sym_true] = ACTIONS(1420),
    [anon_sym_false] = ACTIONS(1420),
    [anon_sym_ATinterface] = ACTIONS(1418),
    [anon_sym_interface] = ACTIONS(1420),
    [anon_sym_class] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1418),
    [anon_sym_DASH_GT] = ACTIONS(1418),
    [anon_sym_RBRACE] = ACTIONS(1418),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_def] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1418),
    [anon_sym_null] = ACTIONS(1420),
    [anon_sym_this] = ACTIONS(1420),
    [anon_sym_do] = ACTIONS(1420),
    [anon_sym_while] = ACTIONS(1420),
    [anon_sym_for] = ACTIONS(1420),
    [aux_sym__type_identifier_token1] = ACTIONS(1420),
    [anon_sym_if] = ACTIONS(1420),
    [anon_sym_else] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [aux_sym_number_literal_token1] = ACTIONS(1420),
    [aux_sym_number_literal_token2] = ACTIONS(1418),
    [aux_sym_number_literal_token3] = ACTIONS(1418),
    [aux_sym_number_literal_token4] = ACTIONS(1420),
    [aux_sym_number_literal_token5] = ACTIONS(1418),
    [anon_sym_pipeline] = ACTIONS(1420),
    [anon_sym_return] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1418),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1418),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1418),
    [anon_sym_switch] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_try] = ACTIONS(1420),
    [anon_sym_int] = ACTIONS(1420),
    [anon_sym_boolean] = ACTIONS(1420),
    [anon_sym_char] = ACTIONS(1420),
    [anon_sym_short] = ACTIONS(1420),
    [anon_sym_long] = ACTIONS(1420),
    [anon_sym_float] = ACTIONS(1420),
    [anon_sym_double] = ACTIONS(1420),
    [anon_sym_void] = ACTIONS(1420),
    [anon_sym_public] = ACTIONS(1420),
    [anon_sym_protected] = ACTIONS(1420),
    [anon_sym_private] = ACTIONS(1420),
    [anon_sym_static] = ACTIONS(1420),
    [anon_sym_final] = ACTIONS(1420),
    [anon_sym_synchronized] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1418),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_new] = ACTIONS(1420),
  },
  [STATE(465)] = {
    [sym_comment] = STATE(465),
    [sym_groovy_doc] = STATE(465),
    [ts_builtin_sym_end] = ACTIONS(1511),
    [sym_identifier] = ACTIONS(1509),
    [anon_sym_break] = ACTIONS(1509),
    [anon_sym_continue] = ACTIONS(1509),
    [anon_sym_SEMI] = ACTIONS(1511),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1511),
    [anon_sym_import] = ACTIONS(1509),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1509),
    [anon_sym_assert] = ACTIONS(1509),
    [anon_sym_PLUS_PLUS] = ACTIONS(1511),
    [anon_sym_DASH_DASH] = ACTIONS(1511),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1509),
    [anon_sym_false] = ACTIONS(1509),
    [anon_sym_ATinterface] = ACTIONS(1511),
    [anon_sym_interface] = ACTIONS(1509),
    [anon_sym_class] = ACTIONS(1509),
    [anon_sym_COMMA] = ACTIONS(1511),
    [anon_sym_LBRACE] = ACTIONS(1511),
    [anon_sym_RBRACE] = ACTIONS(1511),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1509),
    [anon_sym_def] = ACTIONS(1509),
    [anon_sym_LPAREN] = ACTIONS(1511),
    [anon_sym_null] = ACTIONS(1509),
    [anon_sym_this] = ACTIONS(1509),
    [anon_sym_do] = ACTIONS(1509),
    [anon_sym_while] = ACTIONS(1509),
    [anon_sym_for] = ACTIONS(1509),
    [aux_sym__type_identifier_token1] = ACTIONS(1509),
    [anon_sym_if] = ACTIONS(1509),
    [anon_sym_else] = ACTIONS(1509),
    [anon_sym_LBRACK] = ACTIONS(1511),
    [aux_sym_number_literal_token1] = ACTIONS(1509),
    [aux_sym_number_literal_token2] = ACTIONS(1511),
    [aux_sym_number_literal_token3] = ACTIONS(1511),
    [aux_sym_number_literal_token4] = ACTIONS(1509),
    [aux_sym_number_literal_token5] = ACTIONS(1511),
    [anon_sym_pipeline] = ACTIONS(1509),
    [anon_sym_return] = ACTIONS(1509),
    [anon_sym_SQUOTE] = ACTIONS(1509),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1511),
    [anon_sym_DQUOTE] = ACTIONS(1509),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1511),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1511),
    [anon_sym_switch] = ACTIONS(1509),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1509),
    [anon_sym_catch] = ACTIONS(1509),
    [anon_sym_finally] = ACTIONS(1509),
    [anon_sym_int] = ACTIONS(1509),
    [anon_sym_boolean] = ACTIONS(1509),
    [anon_sym_char] = ACTIONS(1509),
    [anon_sym_short] = ACTIONS(1509),
    [anon_sym_long] = ACTIONS(1509),
    [anon_sym_float] = ACTIONS(1509),
    [anon_sym_double] = ACTIONS(1509),
    [anon_sym_void] = ACTIONS(1509),
    [anon_sym_public] = ACTIONS(1509),
    [anon_sym_protected] = ACTIONS(1509),
    [anon_sym_private] = ACTIONS(1509),
    [anon_sym_static] = ACTIONS(1509),
    [anon_sym_final] = ACTIONS(1509),
    [anon_sym_synchronized] = ACTIONS(1509),
    [anon_sym_TILDE] = ACTIONS(1511),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_new] = ACTIONS(1509),
  },
  [STATE(466)] = {
    [sym_comment] = STATE(466),
    [sym_groovy_doc] = STATE(466),
    [ts_builtin_sym_end] = ACTIONS(1368),
    [sym_identifier] = ACTIONS(1366),
    [anon_sym_break] = ACTIONS(1366),
    [anon_sym_continue] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1368),
    [anon_sym_DOT_AMP] = ACTIONS(1368),
    [anon_sym_DOT_AT] = ACTIONS(1368),
    [anon_sym_QMARK_DOT] = ACTIONS(1368),
    [anon_sym_STAR_DOT] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_STAR_COLON] = ACTIONS(1368),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_import] = ACTIONS(1366),
    [anon_sym_as] = ACTIONS(1366),
    [anon_sym_package] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_assert] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1368),
    [anon_sym_PERCENT] = ACTIONS(1368),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_LT_LT] = ACTIONS(1368),
    [anon_sym_GT_GT] = ACTIONS(1366),
    [anon_sym_GT_GT_GT] = ACTIONS(1368),
    [anon_sym_DOT_DOT] = ACTIONS(1366),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_LT_EQ] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_GT_EQ] = ACTIONS(1368),
    [anon_sym_in] = ACTIONS(1366),
    [anon_sym_BANGin] = ACTIONS(1366),
    [anon_sym_instanceof] = ACTIONS(1366),
    [anon_sym_BANGinstanceof] = ACTIONS(1368),
    [anon_sym_EQ_EQ] = ACTIONS(1366),
    [anon_sym_BANG_EQ] = ACTIONS(1366),
    [anon_sym_LT_EQ_GT] = ACTIONS(1368),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1368),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1368),
    [anon_sym_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1368),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_AMP_AMP] = ACTIONS(1368),
    [anon_sym_PIPE_PIPE] = ACTIONS(1368),
    [anon_sym_QMARK_COLON] = ACTIONS(1368),
    [anon_sym_STAR_STAR] = ACTIONS(1368),
    [anon_sym_true] = ACTIONS(1366),
    [anon_sym_false] = ACTIONS(1366),
    [anon_sym_ATinterface] = ACTIONS(1368),
    [anon_sym_interface] = ACTIONS(1366),
    [anon_sym_class] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_DASH_GT] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_def] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1368),
    [anon_sym_null] = ACTIONS(1366),
    [anon_sym_this] = ACTIONS(1366),
    [anon_sym_do] = ACTIONS(1366),
    [anon_sym_while] = ACTIONS(1366),
    [anon_sym_for] = ACTIONS(1366),
    [aux_sym__type_identifier_token1] = ACTIONS(1366),
    [anon_sym_if] = ACTIONS(1366),
    [anon_sym_else] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [aux_sym_number_literal_token1] = ACTIONS(1366),
    [aux_sym_number_literal_token2] = ACTIONS(1368),
    [aux_sym_number_literal_token3] = ACTIONS(1368),
    [aux_sym_number_literal_token4] = ACTIONS(1366),
    [aux_sym_number_literal_token5] = ACTIONS(1368),
    [anon_sym_pipeline] = ACTIONS(1366),
    [anon_sym_return] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1368),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1368),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_int] = ACTIONS(1366),
    [anon_sym_boolean] = ACTIONS(1366),
    [anon_sym_char] = ACTIONS(1366),
    [anon_sym_short] = ACTIONS(1366),
    [anon_sym_long] = ACTIONS(1366),
    [anon_sym_float] = ACTIONS(1366),
    [anon_sym_double] = ACTIONS(1366),
    [anon_sym_void] = ACTIONS(1366),
    [anon_sym_public] = ACTIONS(1366),
    [anon_sym_protected] = ACTIONS(1366),
    [anon_sym_private] = ACTIONS(1366),
    [anon_sym_static] = ACTIONS(1366),
    [anon_sym_final] = ACTIONS(1366),
    [anon_sym_synchronized] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_new] = ACTIONS(1366),
  },
  [STATE(467)] = {
    [sym_comment] = STATE(467),
    [sym_groovy_doc] = STATE(467),
    [aux_sym_dotted_identifier_repeat1] = STATE(27),
    [sym_identifier] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(1286),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(102),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(1290),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(468)] = {
    [sym_comment] = STATE(468),
    [sym_groovy_doc] = STATE(468),
    [aux_sym_dotted_identifier_repeat1] = STATE(432),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_DOT] = ACTIONS(1453),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_case] = ACTIONS(1280),
    [anon_sym_default] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(469)] = {
    [sym_comment] = STATE(469),
    [sym_groovy_doc] = STATE(469),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(470)] = {
    [sym_comment] = STATE(470),
    [sym_groovy_doc] = STATE(470),
    [ts_builtin_sym_end] = ACTIONS(97),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_DASH_GT] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_pipeline] = ACTIONS(102),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(471)] = {
    [sym_comment] = STATE(471),
    [sym_groovy_doc] = STATE(471),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(472)] = {
    [sym_comment] = STATE(472),
    [sym_groovy_doc] = STATE(472),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(473)] = {
    [sym_comment] = STATE(473),
    [sym_groovy_doc] = STATE(473),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1441),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(474)] = {
    [sym_comment] = STATE(474),
    [sym_groovy_doc] = STATE(474),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_COLON] = ACTIONS(1441),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(475)] = {
    [sym_comment] = STATE(475),
    [sym_groovy_doc] = STATE(475),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(476)] = {
    [sym_comment] = STATE(476),
    [sym_groovy_doc] = STATE(476),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(477)] = {
    [sym_comment] = STATE(477),
    [sym_groovy_doc] = STATE(477),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_DOT_AMP] = ACTIONS(1374),
    [anon_sym_DOT_AT] = ACTIONS(1374),
    [anon_sym_QMARK_DOT] = ACTIONS(1374),
    [anon_sym_STAR_DOT] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_DOT] = ACTIONS(1376),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_as] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_PERCENT] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_LT_LT] = ACTIONS(1374),
    [anon_sym_GT_GT] = ACTIONS(1376),
    [anon_sym_GT_GT_GT] = ACTIONS(1374),
    [anon_sym_DOT_DOT] = ACTIONS(1376),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1374),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1376),
    [anon_sym_LT] = ACTIONS(1376),
    [anon_sym_LT_EQ] = ACTIONS(1376),
    [anon_sym_GT] = ACTIONS(1376),
    [anon_sym_GT_EQ] = ACTIONS(1374),
    [anon_sym_in] = ACTIONS(1376),
    [anon_sym_BANGin] = ACTIONS(1376),
    [anon_sym_instanceof] = ACTIONS(1376),
    [anon_sym_BANGinstanceof] = ACTIONS(1374),
    [anon_sym_EQ_EQ] = ACTIONS(1376),
    [anon_sym_BANG_EQ] = ACTIONS(1376),
    [anon_sym_LT_EQ_GT] = ACTIONS(1374),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1374),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1374),
    [anon_sym_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1374),
    [anon_sym_AMP] = ACTIONS(1376),
    [anon_sym_CARET] = ACTIONS(1374),
    [anon_sym_PIPE] = ACTIONS(1376),
    [anon_sym_AMP_AMP] = ACTIONS(1374),
    [anon_sym_PIPE_PIPE] = ACTIONS(1374),
    [anon_sym_QMARK_COLON] = ACTIONS(1374),
    [anon_sym_STAR_STAR] = ACTIONS(1374),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_COMMA] = ACTIONS(1374),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_case] = ACTIONS(1376),
    [anon_sym_default] = ACTIONS(1376),
    [anon_sym_QMARK] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(478)] = {
    [sym_comment] = STATE(478),
    [sym_groovy_doc] = STATE(478),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(479)] = {
    [sym_comment] = STATE(479),
    [sym_groovy_doc] = STATE(479),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_DOT_AMP] = ACTIONS(1390),
    [anon_sym_DOT_AT] = ACTIONS(1390),
    [anon_sym_QMARK_DOT] = ACTIONS(1390),
    [anon_sym_STAR_DOT] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_DOT] = ACTIONS(1392),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_as] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_PERCENT] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_LT_LT] = ACTIONS(1390),
    [anon_sym_GT_GT] = ACTIONS(1392),
    [anon_sym_GT_GT_GT] = ACTIONS(1390),
    [anon_sym_DOT_DOT] = ACTIONS(1392),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1390),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1392),
    [anon_sym_LT] = ACTIONS(1392),
    [anon_sym_LT_EQ] = ACTIONS(1392),
    [anon_sym_GT] = ACTIONS(1392),
    [anon_sym_GT_EQ] = ACTIONS(1390),
    [anon_sym_in] = ACTIONS(1392),
    [anon_sym_BANGin] = ACTIONS(1392),
    [anon_sym_instanceof] = ACTIONS(1392),
    [anon_sym_BANGinstanceof] = ACTIONS(1390),
    [anon_sym_EQ_EQ] = ACTIONS(1392),
    [anon_sym_BANG_EQ] = ACTIONS(1392),
    [anon_sym_LT_EQ_GT] = ACTIONS(1390),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1390),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1390),
    [anon_sym_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1390),
    [anon_sym_AMP] = ACTIONS(1392),
    [anon_sym_CARET] = ACTIONS(1390),
    [anon_sym_PIPE] = ACTIONS(1392),
    [anon_sym_AMP_AMP] = ACTIONS(1390),
    [anon_sym_PIPE_PIPE] = ACTIONS(1390),
    [anon_sym_QMARK_COLON] = ACTIONS(1390),
    [anon_sym_STAR_STAR] = ACTIONS(1390),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_COMMA] = ACTIONS(1390),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_case] = ACTIONS(1392),
    [anon_sym_default] = ACTIONS(1392),
    [anon_sym_QMARK] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(480)] = {
    [sym_comment] = STATE(480),
    [sym_groovy_doc] = STATE(480),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_DASH_GT] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(481)] = {
    [sym_comment] = STATE(481),
    [sym_groovy_doc] = STATE(481),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_break] = ACTIONS(365),
    [anon_sym_continue] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(365),
    [anon_sym_AT] = ACTIONS(365),
    [anon_sym_assert] = ACTIONS(365),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_ATinterface] = ACTIONS(363),
    [anon_sym_interface] = ACTIONS(365),
    [anon_sym_class] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [anon_sym_RBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(365),
    [anon_sym_def] = ACTIONS(365),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_do] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [anon_sym_for] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_return] = ACTIONS(365),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_switch] = ACTIONS(365),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(365),
    [anon_sym_int] = ACTIONS(365),
    [anon_sym_boolean] = ACTIONS(365),
    [anon_sym_char] = ACTIONS(365),
    [anon_sym_short] = ACTIONS(365),
    [anon_sym_long] = ACTIONS(365),
    [anon_sym_float] = ACTIONS(365),
    [anon_sym_double] = ACTIONS(365),
    [anon_sym_void] = ACTIONS(365),
    [anon_sym_public] = ACTIONS(365),
    [anon_sym_protected] = ACTIONS(365),
    [anon_sym_private] = ACTIONS(365),
    [anon_sym_static] = ACTIONS(365),
    [anon_sym_final] = ACTIONS(365),
    [anon_sym_synchronized] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(482)] = {
    [sym_comment] = STATE(482),
    [sym_groovy_doc] = STATE(482),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_break] = ACTIONS(379),
    [anon_sym_continue] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_import] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_package] = ACTIONS(379),
    [anon_sym_AT] = ACTIONS(379),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(377),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(377),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(377),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(377),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(377),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_ATinterface] = ACTIONS(377),
    [anon_sym_interface] = ACTIONS(379),
    [anon_sym_class] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [anon_sym_RBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(379),
    [anon_sym_def] = ACTIONS(379),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_do] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [anon_sym_for] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_if] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_return] = ACTIONS(379),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_switch] = ACTIONS(379),
    [anon_sym_case] = ACTIONS(379),
    [anon_sym_default] = ACTIONS(379),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_try] = ACTIONS(379),
    [anon_sym_int] = ACTIONS(379),
    [anon_sym_boolean] = ACTIONS(379),
    [anon_sym_char] = ACTIONS(379),
    [anon_sym_short] = ACTIONS(379),
    [anon_sym_long] = ACTIONS(379),
    [anon_sym_float] = ACTIONS(379),
    [anon_sym_double] = ACTIONS(379),
    [anon_sym_void] = ACTIONS(379),
    [anon_sym_public] = ACTIONS(379),
    [anon_sym_protected] = ACTIONS(379),
    [anon_sym_private] = ACTIONS(379),
    [anon_sym_static] = ACTIONS(379),
    [anon_sym_final] = ACTIONS(379),
    [anon_sym_synchronized] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(483)] = {
    [sym_comment] = STATE(483),
    [sym_groovy_doc] = STATE(483),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(484)] = {
    [sym_comment] = STATE(484),
    [sym_groovy_doc] = STATE(484),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_break] = ACTIONS(393),
    [anon_sym_continue] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_import] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_package] = ACTIONS(393),
    [anon_sym_AT] = ACTIONS(393),
    [anon_sym_assert] = ACTIONS(393),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(391),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(391),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(391),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(391),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(391),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_ATinterface] = ACTIONS(391),
    [anon_sym_interface] = ACTIONS(393),
    [anon_sym_class] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [anon_sym_RBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(393),
    [anon_sym_def] = ACTIONS(393),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_do] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [anon_sym_for] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_if] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_return] = ACTIONS(393),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_switch] = ACTIONS(393),
    [anon_sym_case] = ACTIONS(393),
    [anon_sym_default] = ACTIONS(393),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_try] = ACTIONS(393),
    [anon_sym_int] = ACTIONS(393),
    [anon_sym_boolean] = ACTIONS(393),
    [anon_sym_char] = ACTIONS(393),
    [anon_sym_short] = ACTIONS(393),
    [anon_sym_long] = ACTIONS(393),
    [anon_sym_float] = ACTIONS(393),
    [anon_sym_double] = ACTIONS(393),
    [anon_sym_void] = ACTIONS(393),
    [anon_sym_public] = ACTIONS(393),
    [anon_sym_protected] = ACTIONS(393),
    [anon_sym_private] = ACTIONS(393),
    [anon_sym_static] = ACTIONS(393),
    [anon_sym_final] = ACTIONS(393),
    [anon_sym_synchronized] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(485)] = {
    [sym_comment] = STATE(485),
    [sym_groovy_doc] = STATE(485),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(486)] = {
    [sym_comment] = STATE(486),
    [sym_groovy_doc] = STATE(486),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(487)] = {
    [sym_comment] = STATE(487),
    [sym_groovy_doc] = STATE(487),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(488)] = {
    [sym_comment] = STATE(488),
    [sym_groovy_doc] = STATE(488),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(489)] = {
    [sym_comment] = STATE(489),
    [sym_groovy_doc] = STATE(489),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(490)] = {
    [sym_comment] = STATE(490),
    [sym_groovy_doc] = STATE(490),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(491)] = {
    [sym_comment] = STATE(491),
    [sym_groovy_doc] = STATE(491),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(492)] = {
    [sym_comment] = STATE(492),
    [sym_groovy_doc] = STATE(492),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(493)] = {
    [sym_comment] = STATE(493),
    [sym_groovy_doc] = STATE(493),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(494)] = {
    [sym_comment] = STATE(494),
    [sym_groovy_doc] = STATE(494),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(495)] = {
    [sym_comment] = STATE(495),
    [sym_groovy_doc] = STATE(495),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(496)] = {
    [sym_comment] = STATE(496),
    [sym_groovy_doc] = STATE(496),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(369),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(369),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(369),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(369),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(497)] = {
    [sym_comment] = STATE(497),
    [sym_groovy_doc] = STATE(497),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(387),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_LBRACK_RBRACK] = ACTIONS(389),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(498)] = {
    [sym_comment] = STATE(498),
    [sym_groovy_doc] = STATE(498),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(499)] = {
    [sym_comment] = STATE(499),
    [sym_groovy_doc] = STATE(499),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(500)] = {
    [sym_comment] = STATE(500),
    [sym_groovy_doc] = STATE(500),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(389),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(389),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(389),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(389),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(501)] = {
    [sym_comment] = STATE(501),
    [sym_groovy_doc] = STATE(501),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(502)] = {
    [sym_comment] = STATE(502),
    [sym_groovy_doc] = STATE(502),
    [ts_builtin_sym_end] = ACTIONS(1469),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_DASH_GT] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_pipeline] = ACTIONS(1467),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(503)] = {
    [sym_comment] = STATE(503),
    [sym_groovy_doc] = STATE(503),
    [sym_identifier] = ACTIONS(1598),
    [anon_sym_break] = ACTIONS(1598),
    [anon_sym_continue] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1600),
    [anon_sym_import] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1598),
    [anon_sym_AT] = ACTIONS(1598),
    [anon_sym_assert] = ACTIONS(1598),
    [anon_sym_PLUS_PLUS] = ACTIONS(1600),
    [anon_sym_DASH_DASH] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_ATinterface] = ACTIONS(1600),
    [anon_sym_interface] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1598),
    [anon_sym_def] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_null] = ACTIONS(1598),
    [anon_sym_this] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_while] = ACTIONS(1598),
    [anon_sym_for] = ACTIONS(1598),
    [aux_sym__type_identifier_token1] = ACTIONS(1598),
    [anon_sym_if] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1600),
    [aux_sym_number_literal_token1] = ACTIONS(1598),
    [aux_sym_number_literal_token2] = ACTIONS(1600),
    [aux_sym_number_literal_token3] = ACTIONS(1600),
    [aux_sym_number_literal_token4] = ACTIONS(1598),
    [aux_sym_number_literal_token5] = ACTIONS(1600),
    [anon_sym_return] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1598),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1598),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1600),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1600),
    [anon_sym_switch] = ACTIONS(1598),
    [anon_sym_case] = ACTIONS(1598),
    [anon_sym_default] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1598),
    [anon_sym_catch] = ACTIONS(1598),
    [anon_sym_finally] = ACTIONS(1598),
    [anon_sym_int] = ACTIONS(1598),
    [anon_sym_boolean] = ACTIONS(1598),
    [anon_sym_char] = ACTIONS(1598),
    [anon_sym_short] = ACTIONS(1598),
    [anon_sym_long] = ACTIONS(1598),
    [anon_sym_float] = ACTIONS(1598),
    [anon_sym_double] = ACTIONS(1598),
    [anon_sym_void] = ACTIONS(1598),
    [anon_sym_public] = ACTIONS(1598),
    [anon_sym_protected] = ACTIONS(1598),
    [anon_sym_private] = ACTIONS(1598),
    [anon_sym_static] = ACTIONS(1598),
    [anon_sym_final] = ACTIONS(1598),
    [anon_sym_synchronized] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
  },
  [STATE(504)] = {
    [sym_comment] = STATE(504),
    [sym_groovy_doc] = STATE(504),
    [ts_builtin_sym_end] = ACTIONS(1602),
    [sym_identifier] = ACTIONS(1604),
    [anon_sym_break] = ACTIONS(1604),
    [anon_sym_continue] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1604),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1604),
    [anon_sym_AT] = ACTIONS(1604),
    [anon_sym_assert] = ACTIONS(1604),
    [anon_sym_PLUS_PLUS] = ACTIONS(1602),
    [anon_sym_DASH_DASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1604),
    [anon_sym_false] = ACTIONS(1604),
    [anon_sym_ATinterface] = ACTIONS(1602),
    [anon_sym_interface] = ACTIONS(1604),
    [anon_sym_class] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1604),
    [anon_sym_def] = ACTIONS(1604),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_null] = ACTIONS(1604),
    [anon_sym_this] = ACTIONS(1604),
    [anon_sym_do] = ACTIONS(1604),
    [anon_sym_while] = ACTIONS(1604),
    [anon_sym_for] = ACTIONS(1604),
    [aux_sym__type_identifier_token1] = ACTIONS(1604),
    [anon_sym_if] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [aux_sym_number_literal_token1] = ACTIONS(1604),
    [aux_sym_number_literal_token2] = ACTIONS(1602),
    [aux_sym_number_literal_token3] = ACTIONS(1602),
    [aux_sym_number_literal_token4] = ACTIONS(1604),
    [aux_sym_number_literal_token5] = ACTIONS(1602),
    [anon_sym_pipeline] = ACTIONS(1604),
    [anon_sym_return] = ACTIONS(1604),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1602),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1602),
    [anon_sym_switch] = ACTIONS(1604),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1604),
    [anon_sym_catch] = ACTIONS(1604),
    [anon_sym_finally] = ACTIONS(1604),
    [anon_sym_int] = ACTIONS(1604),
    [anon_sym_boolean] = ACTIONS(1604),
    [anon_sym_char] = ACTIONS(1604),
    [anon_sym_short] = ACTIONS(1604),
    [anon_sym_long] = ACTIONS(1604),
    [anon_sym_float] = ACTIONS(1604),
    [anon_sym_double] = ACTIONS(1604),
    [anon_sym_void] = ACTIONS(1604),
    [anon_sym_public] = ACTIONS(1604),
    [anon_sym_protected] = ACTIONS(1604),
    [anon_sym_private] = ACTIONS(1604),
    [anon_sym_static] = ACTIONS(1604),
    [anon_sym_final] = ACTIONS(1604),
    [anon_sym_synchronized] = ACTIONS(1604),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1604),
    [anon_sym_new] = ACTIONS(1604),
  },
  [STATE(505)] = {
    [sym_comment] = STATE(505),
    [sym_groovy_doc] = STATE(505),
    [sym_identifier] = ACTIONS(1380),
    [anon_sym_break] = ACTIONS(1380),
    [anon_sym_continue] = ACTIONS(1380),
    [anon_sym_SEMI] = ACTIONS(1378),
    [anon_sym_DOT_AMP] = ACTIONS(1378),
    [anon_sym_DOT_AT] = ACTIONS(1378),
    [anon_sym_QMARK_DOT] = ACTIONS(1378),
    [anon_sym_STAR_DOT] = ACTIONS(1378),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_STAR_COLON] = ACTIONS(1378),
    [anon_sym_DOT] = ACTIONS(1380),
    [anon_sym_import] = ACTIONS(1380),
    [anon_sym_as] = ACTIONS(1380),
    [anon_sym_package] = ACTIONS(1380),
    [anon_sym_AT] = ACTIONS(1380),
    [anon_sym_assert] = ACTIONS(1380),
    [anon_sym_PLUS_PLUS] = ACTIONS(1378),
    [anon_sym_DASH_DASH] = ACTIONS(1378),
    [anon_sym_PERCENT] = ACTIONS(1378),
    [anon_sym_SLASH] = ACTIONS(1380),
    [anon_sym_PLUS] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_LT_LT] = ACTIONS(1378),
    [anon_sym_GT_GT] = ACTIONS(1380),
    [anon_sym_GT_GT_GT] = ACTIONS(1378),
    [anon_sym_DOT_DOT] = ACTIONS(1380),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1378),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1380),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_LT_EQ] = ACTIONS(1380),
    [anon_sym_GT] = ACTIONS(1380),
    [anon_sym_GT_EQ] = ACTIONS(1378),
    [anon_sym_in] = ACTIONS(1380),
    [anon_sym_BANGin] = ACTIONS(1380),
    [anon_sym_instanceof] = ACTIONS(1380),
    [anon_sym_BANGinstanceof] = ACTIONS(1378),
    [anon_sym_EQ_EQ] = ACTIONS(1380),
    [anon_sym_BANG_EQ] = ACTIONS(1380),
    [anon_sym_LT_EQ_GT] = ACTIONS(1378),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1378),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1378),
    [anon_sym_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1378),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_CARET] = ACTIONS(1378),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_AMP_AMP] = ACTIONS(1378),
    [anon_sym_PIPE_PIPE] = ACTIONS(1378),
    [anon_sym_QMARK_COLON] = ACTIONS(1378),
    [anon_sym_STAR_STAR] = ACTIONS(1378),
    [anon_sym_true] = ACTIONS(1380),
    [anon_sym_false] = ACTIONS(1380),
    [anon_sym_ATinterface] = ACTIONS(1378),
    [anon_sym_interface] = ACTIONS(1380),
    [anon_sym_class] = ACTIONS(1380),
    [anon_sym_COMMA] = ACTIONS(1378),
    [anon_sym_LBRACE] = ACTIONS(1378),
    [anon_sym_RBRACE] = ACTIONS(1378),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1380),
    [anon_sym_def] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1378),
    [anon_sym_null] = ACTIONS(1380),
    [anon_sym_this] = ACTIONS(1380),
    [anon_sym_do] = ACTIONS(1380),
    [anon_sym_while] = ACTIONS(1380),
    [anon_sym_for] = ACTIONS(1380),
    [aux_sym__type_identifier_token1] = ACTIONS(1380),
    [anon_sym_if] = ACTIONS(1380),
    [anon_sym_else] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1378),
    [aux_sym_number_literal_token1] = ACTIONS(1380),
    [aux_sym_number_literal_token2] = ACTIONS(1378),
    [aux_sym_number_literal_token3] = ACTIONS(1378),
    [aux_sym_number_literal_token4] = ACTIONS(1380),
    [aux_sym_number_literal_token5] = ACTIONS(1378),
    [anon_sym_return] = ACTIONS(1380),
    [anon_sym_SQUOTE] = ACTIONS(1380),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1378),
    [anon_sym_DQUOTE] = ACTIONS(1380),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1378),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1378),
    [anon_sym_switch] = ACTIONS(1380),
    [anon_sym_case] = ACTIONS(1380),
    [anon_sym_default] = ACTIONS(1380),
    [anon_sym_QMARK] = ACTIONS(1380),
    [anon_sym_try] = ACTIONS(1380),
    [anon_sym_int] = ACTIONS(1380),
    [anon_sym_boolean] = ACTIONS(1380),
    [anon_sym_char] = ACTIONS(1380),
    [anon_sym_short] = ACTIONS(1380),
    [anon_sym_long] = ACTIONS(1380),
    [anon_sym_float] = ACTIONS(1380),
    [anon_sym_double] = ACTIONS(1380),
    [anon_sym_void] = ACTIONS(1380),
    [anon_sym_public] = ACTIONS(1380),
    [anon_sym_protected] = ACTIONS(1380),
    [anon_sym_private] = ACTIONS(1380),
    [anon_sym_static] = ACTIONS(1380),
    [anon_sym_final] = ACTIONS(1380),
    [anon_sym_synchronized] = ACTIONS(1380),
    [anon_sym_TILDE] = ACTIONS(1378),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_new] = ACTIONS(1380),
  },
  [STATE(506)] = {
    [sym_comment] = STATE(506),
    [sym_groovy_doc] = STATE(506),
    [sym_identifier] = ACTIONS(1396),
    [anon_sym_break] = ACTIONS(1396),
    [anon_sym_continue] = ACTIONS(1396),
    [anon_sym_SEMI] = ACTIONS(1394),
    [anon_sym_DOT_AMP] = ACTIONS(1394),
    [anon_sym_DOT_AT] = ACTIONS(1394),
    [anon_sym_QMARK_DOT] = ACTIONS(1394),
    [anon_sym_STAR_DOT] = ACTIONS(1394),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_STAR_COLON] = ACTIONS(1394),
    [anon_sym_DOT] = ACTIONS(1396),
    [anon_sym_import] = ACTIONS(1396),
    [anon_sym_as] = ACTIONS(1396),
    [anon_sym_package] = ACTIONS(1396),
    [anon_sym_AT] = ACTIONS(1396),
    [anon_sym_assert] = ACTIONS(1396),
    [anon_sym_PLUS_PLUS] = ACTIONS(1394),
    [anon_sym_DASH_DASH] = ACTIONS(1394),
    [anon_sym_PERCENT] = ACTIONS(1394),
    [anon_sym_SLASH] = ACTIONS(1396),
    [anon_sym_PLUS] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_LT_LT] = ACTIONS(1394),
    [anon_sym_GT_GT] = ACTIONS(1396),
    [anon_sym_GT_GT_GT] = ACTIONS(1394),
    [anon_sym_DOT_DOT] = ACTIONS(1396),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1394),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1396),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_LT_EQ] = ACTIONS(1396),
    [anon_sym_GT] = ACTIONS(1396),
    [anon_sym_GT_EQ] = ACTIONS(1394),
    [anon_sym_in] = ACTIONS(1396),
    [anon_sym_BANGin] = ACTIONS(1396),
    [anon_sym_instanceof] = ACTIONS(1396),
    [anon_sym_BANGinstanceof] = ACTIONS(1394),
    [anon_sym_EQ_EQ] = ACTIONS(1396),
    [anon_sym_BANG_EQ] = ACTIONS(1396),
    [anon_sym_LT_EQ_GT] = ACTIONS(1394),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1394),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1394),
    [anon_sym_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1394),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_CARET] = ACTIONS(1394),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_AMP_AMP] = ACTIONS(1394),
    [anon_sym_PIPE_PIPE] = ACTIONS(1394),
    [anon_sym_QMARK_COLON] = ACTIONS(1394),
    [anon_sym_STAR_STAR] = ACTIONS(1394),
    [anon_sym_true] = ACTIONS(1396),
    [anon_sym_false] = ACTIONS(1396),
    [anon_sym_ATinterface] = ACTIONS(1394),
    [anon_sym_interface] = ACTIONS(1396),
    [anon_sym_class] = ACTIONS(1396),
    [anon_sym_COMMA] = ACTIONS(1394),
    [anon_sym_LBRACE] = ACTIONS(1394),
    [anon_sym_RBRACE] = ACTIONS(1394),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1396),
    [anon_sym_def] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1394),
    [anon_sym_null] = ACTIONS(1396),
    [anon_sym_this] = ACTIONS(1396),
    [anon_sym_do] = ACTIONS(1396),
    [anon_sym_while] = ACTIONS(1396),
    [anon_sym_for] = ACTIONS(1396),
    [aux_sym__type_identifier_token1] = ACTIONS(1396),
    [anon_sym_if] = ACTIONS(1396),
    [anon_sym_else] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1394),
    [aux_sym_number_literal_token1] = ACTIONS(1396),
    [aux_sym_number_literal_token2] = ACTIONS(1394),
    [aux_sym_number_literal_token3] = ACTIONS(1394),
    [aux_sym_number_literal_token4] = ACTIONS(1396),
    [aux_sym_number_literal_token5] = ACTIONS(1394),
    [anon_sym_return] = ACTIONS(1396),
    [anon_sym_SQUOTE] = ACTIONS(1396),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1396),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1394),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1394),
    [anon_sym_switch] = ACTIONS(1396),
    [anon_sym_case] = ACTIONS(1396),
    [anon_sym_default] = ACTIONS(1396),
    [anon_sym_QMARK] = ACTIONS(1396),
    [anon_sym_try] = ACTIONS(1396),
    [anon_sym_int] = ACTIONS(1396),
    [anon_sym_boolean] = ACTIONS(1396),
    [anon_sym_char] = ACTIONS(1396),
    [anon_sym_short] = ACTIONS(1396),
    [anon_sym_long] = ACTIONS(1396),
    [anon_sym_float] = ACTIONS(1396),
    [anon_sym_double] = ACTIONS(1396),
    [anon_sym_void] = ACTIONS(1396),
    [anon_sym_public] = ACTIONS(1396),
    [anon_sym_protected] = ACTIONS(1396),
    [anon_sym_private] = ACTIONS(1396),
    [anon_sym_static] = ACTIONS(1396),
    [anon_sym_final] = ACTIONS(1396),
    [anon_sym_synchronized] = ACTIONS(1396),
    [anon_sym_TILDE] = ACTIONS(1394),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_new] = ACTIONS(1396),
  },
  [STATE(507)] = {
    [sym_comment] = STATE(507),
    [sym_groovy_doc] = STATE(507),
    [sym_identifier] = ACTIONS(1400),
    [anon_sym_break] = ACTIONS(1400),
    [anon_sym_continue] = ACTIONS(1400),
    [anon_sym_SEMI] = ACTIONS(1398),
    [anon_sym_DOT_AMP] = ACTIONS(1398),
    [anon_sym_DOT_AT] = ACTIONS(1398),
    [anon_sym_QMARK_DOT] = ACTIONS(1398),
    [anon_sym_STAR_DOT] = ACTIONS(1398),
    [anon_sym_STAR] = ACTIONS(1400),
    [anon_sym_STAR_COLON] = ACTIONS(1398),
    [anon_sym_DOT] = ACTIONS(1400),
    [anon_sym_import] = ACTIONS(1400),
    [anon_sym_as] = ACTIONS(1400),
    [anon_sym_package] = ACTIONS(1400),
    [anon_sym_AT] = ACTIONS(1400),
    [anon_sym_assert] = ACTIONS(1400),
    [anon_sym_PLUS_PLUS] = ACTIONS(1398),
    [anon_sym_DASH_DASH] = ACTIONS(1398),
    [anon_sym_PERCENT] = ACTIONS(1398),
    [anon_sym_SLASH] = ACTIONS(1400),
    [anon_sym_PLUS] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_LT_LT] = ACTIONS(1398),
    [anon_sym_GT_GT] = ACTIONS(1400),
    [anon_sym_GT_GT_GT] = ACTIONS(1398),
    [anon_sym_DOT_DOT] = ACTIONS(1400),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1398),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1400),
    [anon_sym_LT] = ACTIONS(1400),
    [anon_sym_LT_EQ] = ACTIONS(1400),
    [anon_sym_GT] = ACTIONS(1400),
    [anon_sym_GT_EQ] = ACTIONS(1398),
    [anon_sym_in] = ACTIONS(1400),
    [anon_sym_BANGin] = ACTIONS(1400),
    [anon_sym_instanceof] = ACTIONS(1400),
    [anon_sym_BANGinstanceof] = ACTIONS(1398),
    [anon_sym_EQ_EQ] = ACTIONS(1400),
    [anon_sym_BANG_EQ] = ACTIONS(1400),
    [anon_sym_LT_EQ_GT] = ACTIONS(1398),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1398),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1398),
    [anon_sym_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1398),
    [anon_sym_AMP] = ACTIONS(1400),
    [anon_sym_CARET] = ACTIONS(1398),
    [anon_sym_PIPE] = ACTIONS(1400),
    [anon_sym_AMP_AMP] = ACTIONS(1398),
    [anon_sym_PIPE_PIPE] = ACTIONS(1398),
    [anon_sym_QMARK_COLON] = ACTIONS(1398),
    [anon_sym_STAR_STAR] = ACTIONS(1398),
    [anon_sym_true] = ACTIONS(1400),
    [anon_sym_false] = ACTIONS(1400),
    [anon_sym_ATinterface] = ACTIONS(1398),
    [anon_sym_interface] = ACTIONS(1400),
    [anon_sym_class] = ACTIONS(1400),
    [anon_sym_COMMA] = ACTIONS(1398),
    [anon_sym_LBRACE] = ACTIONS(1398),
    [anon_sym_RBRACE] = ACTIONS(1398),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1400),
    [anon_sym_def] = ACTIONS(1400),
    [anon_sym_LPAREN] = ACTIONS(1398),
    [anon_sym_null] = ACTIONS(1400),
    [anon_sym_this] = ACTIONS(1400),
    [anon_sym_do] = ACTIONS(1400),
    [anon_sym_while] = ACTIONS(1400),
    [anon_sym_for] = ACTIONS(1400),
    [aux_sym__type_identifier_token1] = ACTIONS(1400),
    [anon_sym_if] = ACTIONS(1400),
    [anon_sym_else] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1398),
    [aux_sym_number_literal_token1] = ACTIONS(1400),
    [aux_sym_number_literal_token2] = ACTIONS(1398),
    [aux_sym_number_literal_token3] = ACTIONS(1398),
    [aux_sym_number_literal_token4] = ACTIONS(1400),
    [aux_sym_number_literal_token5] = ACTIONS(1398),
    [anon_sym_return] = ACTIONS(1400),
    [anon_sym_SQUOTE] = ACTIONS(1400),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1398),
    [anon_sym_DQUOTE] = ACTIONS(1400),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1398),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1398),
    [anon_sym_switch] = ACTIONS(1400),
    [anon_sym_case] = ACTIONS(1400),
    [anon_sym_default] = ACTIONS(1400),
    [anon_sym_QMARK] = ACTIONS(1400),
    [anon_sym_try] = ACTIONS(1400),
    [anon_sym_int] = ACTIONS(1400),
    [anon_sym_boolean] = ACTIONS(1400),
    [anon_sym_char] = ACTIONS(1400),
    [anon_sym_short] = ACTIONS(1400),
    [anon_sym_long] = ACTIONS(1400),
    [anon_sym_float] = ACTIONS(1400),
    [anon_sym_double] = ACTIONS(1400),
    [anon_sym_void] = ACTIONS(1400),
    [anon_sym_public] = ACTIONS(1400),
    [anon_sym_protected] = ACTIONS(1400),
    [anon_sym_private] = ACTIONS(1400),
    [anon_sym_static] = ACTIONS(1400),
    [anon_sym_final] = ACTIONS(1400),
    [anon_sym_synchronized] = ACTIONS(1400),
    [anon_sym_TILDE] = ACTIONS(1398),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_new] = ACTIONS(1400),
  },
  [STATE(508)] = {
    [sym_comment] = STATE(508),
    [sym_groovy_doc] = STATE(508),
    [ts_builtin_sym_end] = ACTIONS(1451),
    [sym_identifier] = ACTIONS(1449),
    [anon_sym_break] = ACTIONS(1449),
    [anon_sym_continue] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_DOT_AMP] = ACTIONS(1451),
    [anon_sym_DOT_AT] = ACTIONS(1451),
    [anon_sym_QMARK_DOT] = ACTIONS(1451),
    [anon_sym_STAR_DOT] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1449),
    [anon_sym_STAR_COLON] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1449),
    [anon_sym_as] = ACTIONS(1449),
    [anon_sym_package] = ACTIONS(1449),
    [anon_sym_AT] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1449),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_DOT_DOT] = ACTIONS(1449),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1449),
    [anon_sym_LT] = ACTIONS(1449),
    [anon_sym_LT_EQ] = ACTIONS(1449),
    [anon_sym_GT] = ACTIONS(1449),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1449),
    [anon_sym_BANGin] = ACTIONS(1449),
    [anon_sym_instanceof] = ACTIONS(1449),
    [anon_sym_BANGinstanceof] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1449),
    [anon_sym_BANG_EQ] = ACTIONS(1449),
    [anon_sym_LT_EQ_GT] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1449),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1449),
    [anon_sym_AMP_AMP] = ACTIONS(1451),
    [anon_sym_PIPE_PIPE] = ACTIONS(1451),
    [anon_sym_QMARK_COLON] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_true] = ACTIONS(1449),
    [anon_sym_false] = ACTIONS(1449),
    [anon_sym_ATinterface] = ACTIONS(1451),
    [anon_sym_interface] = ACTIONS(1449),
    [anon_sym_class] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [anon_sym_DASH_GT] = ACTIONS(1451),
    [anon_sym_RBRACE] = ACTIONS(1451),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1449),
    [anon_sym_def] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_null] = ACTIONS(1449),
    [anon_sym_this] = ACTIONS(1449),
    [anon_sym_do] = ACTIONS(1449),
    [anon_sym_while] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1449),
    [aux_sym__type_identifier_token1] = ACTIONS(1449),
    [anon_sym_if] = ACTIONS(1449),
    [anon_sym_else] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_number_literal_token1] = ACTIONS(1449),
    [aux_sym_number_literal_token2] = ACTIONS(1451),
    [aux_sym_number_literal_token3] = ACTIONS(1451),
    [aux_sym_number_literal_token4] = ACTIONS(1449),
    [aux_sym_number_literal_token5] = ACTIONS(1451),
    [anon_sym_pipeline] = ACTIONS(1449),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_SQUOTE] = ACTIONS(1449),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1451),
    [anon_sym_DQUOTE] = ACTIONS(1449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1451),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1451),
    [anon_sym_switch] = ACTIONS(1449),
    [anon_sym_QMARK] = ACTIONS(1449),
    [anon_sym_try] = ACTIONS(1449),
    [anon_sym_int] = ACTIONS(1449),
    [anon_sym_boolean] = ACTIONS(1449),
    [anon_sym_char] = ACTIONS(1449),
    [anon_sym_short] = ACTIONS(1449),
    [anon_sym_long] = ACTIONS(1449),
    [anon_sym_float] = ACTIONS(1449),
    [anon_sym_double] = ACTIONS(1449),
    [anon_sym_void] = ACTIONS(1449),
    [anon_sym_public] = ACTIONS(1449),
    [anon_sym_protected] = ACTIONS(1449),
    [anon_sym_private] = ACTIONS(1449),
    [anon_sym_static] = ACTIONS(1449),
    [anon_sym_final] = ACTIONS(1449),
    [anon_sym_synchronized] = ACTIONS(1449),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1449),
    [anon_sym_new] = ACTIONS(1449),
  },
  [STATE(509)] = {
    [sym_comment] = STATE(509),
    [sym_groovy_doc] = STATE(509),
    [ts_builtin_sym_end] = ACTIONS(1606),
    [sym_identifier] = ACTIONS(1608),
    [anon_sym_break] = ACTIONS(1608),
    [anon_sym_continue] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1606),
    [anon_sym_import] = ACTIONS(1608),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_assert] = ACTIONS(1608),
    [anon_sym_PLUS_PLUS] = ACTIONS(1606),
    [anon_sym_DASH_DASH] = ACTIONS(1606),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1608),
    [anon_sym_false] = ACTIONS(1608),
    [anon_sym_ATinterface] = ACTIONS(1606),
    [anon_sym_interface] = ACTIONS(1608),
    [anon_sym_class] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_def] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1606),
    [anon_sym_null] = ACTIONS(1608),
    [anon_sym_this] = ACTIONS(1608),
    [anon_sym_do] = ACTIONS(1608),
    [anon_sym_while] = ACTIONS(1608),
    [anon_sym_for] = ACTIONS(1608),
    [aux_sym__type_identifier_token1] = ACTIONS(1608),
    [anon_sym_if] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [aux_sym_number_literal_token1] = ACTIONS(1608),
    [aux_sym_number_literal_token2] = ACTIONS(1606),
    [aux_sym_number_literal_token3] = ACTIONS(1606),
    [aux_sym_number_literal_token4] = ACTIONS(1608),
    [aux_sym_number_literal_token5] = ACTIONS(1606),
    [anon_sym_pipeline] = ACTIONS(1608),
    [anon_sym_return] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1606),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1606),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1606),
    [anon_sym_switch] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1608),
    [anon_sym_catch] = ACTIONS(1608),
    [anon_sym_finally] = ACTIONS(1608),
    [anon_sym_int] = ACTIONS(1608),
    [anon_sym_boolean] = ACTIONS(1608),
    [anon_sym_char] = ACTIONS(1608),
    [anon_sym_short] = ACTIONS(1608),
    [anon_sym_long] = ACTIONS(1608),
    [anon_sym_float] = ACTIONS(1608),
    [anon_sym_double] = ACTIONS(1608),
    [anon_sym_void] = ACTIONS(1608),
    [anon_sym_public] = ACTIONS(1608),
    [anon_sym_protected] = ACTIONS(1608),
    [anon_sym_private] = ACTIONS(1608),
    [anon_sym_static] = ACTIONS(1608),
    [anon_sym_final] = ACTIONS(1608),
    [anon_sym_synchronized] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_new] = ACTIONS(1608),
  },
  [STATE(510)] = {
    [sym_comment] = STATE(510),
    [sym_groovy_doc] = STATE(510),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(511)] = {
    [sym_comment] = STATE(511),
    [sym_groovy_doc] = STATE(511),
    [sym_identifier] = ACTIONS(1384),
    [anon_sym_break] = ACTIONS(1384),
    [anon_sym_continue] = ACTIONS(1384),
    [anon_sym_SEMI] = ACTIONS(1382),
    [anon_sym_DOT_AMP] = ACTIONS(1382),
    [anon_sym_DOT_AT] = ACTIONS(1382),
    [anon_sym_QMARK_DOT] = ACTIONS(1382),
    [anon_sym_STAR_DOT] = ACTIONS(1382),
    [anon_sym_STAR] = ACTIONS(1384),
    [anon_sym_STAR_COLON] = ACTIONS(1382),
    [anon_sym_DOT] = ACTIONS(1384),
    [anon_sym_import] = ACTIONS(1384),
    [anon_sym_as] = ACTIONS(1384),
    [anon_sym_package] = ACTIONS(1384),
    [anon_sym_AT] = ACTIONS(1384),
    [anon_sym_assert] = ACTIONS(1384),
    [anon_sym_PLUS_PLUS] = ACTIONS(1382),
    [anon_sym_DASH_DASH] = ACTIONS(1382),
    [anon_sym_PERCENT] = ACTIONS(1382),
    [anon_sym_SLASH] = ACTIONS(1384),
    [anon_sym_PLUS] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_LT_LT] = ACTIONS(1382),
    [anon_sym_GT_GT] = ACTIONS(1384),
    [anon_sym_GT_GT_GT] = ACTIONS(1382),
    [anon_sym_DOT_DOT] = ACTIONS(1384),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1382),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1384),
    [anon_sym_LT] = ACTIONS(1384),
    [anon_sym_LT_EQ] = ACTIONS(1384),
    [anon_sym_GT] = ACTIONS(1384),
    [anon_sym_GT_EQ] = ACTIONS(1382),
    [anon_sym_in] = ACTIONS(1384),
    [anon_sym_BANGin] = ACTIONS(1384),
    [anon_sym_instanceof] = ACTIONS(1384),
    [anon_sym_BANGinstanceof] = ACTIONS(1382),
    [anon_sym_EQ_EQ] = ACTIONS(1384),
    [anon_sym_BANG_EQ] = ACTIONS(1384),
    [anon_sym_LT_EQ_GT] = ACTIONS(1382),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1382),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1382),
    [anon_sym_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1382),
    [anon_sym_AMP] = ACTIONS(1384),
    [anon_sym_CARET] = ACTIONS(1382),
    [anon_sym_PIPE] = ACTIONS(1384),
    [anon_sym_AMP_AMP] = ACTIONS(1382),
    [anon_sym_PIPE_PIPE] = ACTIONS(1382),
    [anon_sym_QMARK_COLON] = ACTIONS(1382),
    [anon_sym_STAR_STAR] = ACTIONS(1382),
    [anon_sym_true] = ACTIONS(1384),
    [anon_sym_false] = ACTIONS(1384),
    [anon_sym_ATinterface] = ACTIONS(1382),
    [anon_sym_interface] = ACTIONS(1384),
    [anon_sym_class] = ACTIONS(1384),
    [anon_sym_COMMA] = ACTIONS(1382),
    [anon_sym_LBRACE] = ACTIONS(1382),
    [anon_sym_RBRACE] = ACTIONS(1382),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1384),
    [anon_sym_def] = ACTIONS(1384),
    [anon_sym_LPAREN] = ACTIONS(1382),
    [anon_sym_null] = ACTIONS(1384),
    [anon_sym_this] = ACTIONS(1384),
    [anon_sym_do] = ACTIONS(1384),
    [anon_sym_while] = ACTIONS(1384),
    [anon_sym_for] = ACTIONS(1384),
    [aux_sym__type_identifier_token1] = ACTIONS(1384),
    [anon_sym_if] = ACTIONS(1384),
    [anon_sym_else] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1382),
    [aux_sym_number_literal_token1] = ACTIONS(1384),
    [aux_sym_number_literal_token2] = ACTIONS(1382),
    [aux_sym_number_literal_token3] = ACTIONS(1382),
    [aux_sym_number_literal_token4] = ACTIONS(1384),
    [aux_sym_number_literal_token5] = ACTIONS(1382),
    [anon_sym_return] = ACTIONS(1384),
    [anon_sym_SQUOTE] = ACTIONS(1384),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1382),
    [anon_sym_DQUOTE] = ACTIONS(1384),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1382),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1382),
    [anon_sym_switch] = ACTIONS(1384),
    [anon_sym_case] = ACTIONS(1384),
    [anon_sym_default] = ACTIONS(1384),
    [anon_sym_QMARK] = ACTIONS(1384),
    [anon_sym_try] = ACTIONS(1384),
    [anon_sym_int] = ACTIONS(1384),
    [anon_sym_boolean] = ACTIONS(1384),
    [anon_sym_char] = ACTIONS(1384),
    [anon_sym_short] = ACTIONS(1384),
    [anon_sym_long] = ACTIONS(1384),
    [anon_sym_float] = ACTIONS(1384),
    [anon_sym_double] = ACTIONS(1384),
    [anon_sym_void] = ACTIONS(1384),
    [anon_sym_public] = ACTIONS(1384),
    [anon_sym_protected] = ACTIONS(1384),
    [anon_sym_private] = ACTIONS(1384),
    [anon_sym_static] = ACTIONS(1384),
    [anon_sym_final] = ACTIONS(1384),
    [anon_sym_synchronized] = ACTIONS(1384),
    [anon_sym_TILDE] = ACTIONS(1382),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_new] = ACTIONS(1384),
  },
  [STATE(512)] = {
    [sym_comment] = STATE(512),
    [sym_groovy_doc] = STATE(512),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(513)] = {
    [sym_comment] = STATE(513),
    [sym_groovy_doc] = STATE(513),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(514)] = {
    [sym_comment] = STATE(514),
    [sym_groovy_doc] = STATE(514),
    [sym_identifier] = ACTIONS(1432),
    [anon_sym_break] = ACTIONS(1432),
    [anon_sym_continue] = ACTIONS(1432),
    [anon_sym_SEMI] = ACTIONS(1434),
    [anon_sym_DOT_AMP] = ACTIONS(1434),
    [anon_sym_DOT_AT] = ACTIONS(1434),
    [anon_sym_QMARK_DOT] = ACTIONS(1434),
    [anon_sym_STAR_DOT] = ACTIONS(1434),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_STAR_COLON] = ACTIONS(1434),
    [anon_sym_DOT] = ACTIONS(1432),
    [anon_sym_import] = ACTIONS(1432),
    [anon_sym_as] = ACTIONS(1432),
    [anon_sym_package] = ACTIONS(1432),
    [anon_sym_AT] = ACTIONS(1432),
    [anon_sym_assert] = ACTIONS(1432),
    [anon_sym_PLUS_PLUS] = ACTIONS(1434),
    [anon_sym_DASH_DASH] = ACTIONS(1434),
    [anon_sym_PERCENT] = ACTIONS(1434),
    [anon_sym_SLASH] = ACTIONS(1432),
    [anon_sym_PLUS] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_LT_LT] = ACTIONS(1434),
    [anon_sym_GT_GT] = ACTIONS(1432),
    [anon_sym_GT_GT_GT] = ACTIONS(1434),
    [anon_sym_DOT_DOT] = ACTIONS(1432),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1434),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1432),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_LT_EQ] = ACTIONS(1432),
    [anon_sym_GT] = ACTIONS(1432),
    [anon_sym_GT_EQ] = ACTIONS(1434),
    [anon_sym_in] = ACTIONS(1432),
    [anon_sym_BANGin] = ACTIONS(1432),
    [anon_sym_instanceof] = ACTIONS(1432),
    [anon_sym_BANGinstanceof] = ACTIONS(1434),
    [anon_sym_EQ_EQ] = ACTIONS(1432),
    [anon_sym_BANG_EQ] = ACTIONS(1432),
    [anon_sym_LT_EQ_GT] = ACTIONS(1434),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1434),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1434),
    [anon_sym_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1434),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_CARET] = ACTIONS(1434),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_AMP_AMP] = ACTIONS(1434),
    [anon_sym_PIPE_PIPE] = ACTIONS(1434),
    [anon_sym_QMARK_COLON] = ACTIONS(1434),
    [anon_sym_STAR_STAR] = ACTIONS(1434),
    [anon_sym_true] = ACTIONS(1432),
    [anon_sym_false] = ACTIONS(1432),
    [anon_sym_ATinterface] = ACTIONS(1434),
    [anon_sym_interface] = ACTIONS(1432),
    [anon_sym_class] = ACTIONS(1432),
    [anon_sym_COMMA] = ACTIONS(1434),
    [anon_sym_LBRACE] = ACTIONS(1434),
    [anon_sym_RBRACE] = ACTIONS(1434),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1432),
    [anon_sym_def] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1434),
    [anon_sym_null] = ACTIONS(1432),
    [anon_sym_this] = ACTIONS(1432),
    [anon_sym_do] = ACTIONS(1432),
    [anon_sym_while] = ACTIONS(1432),
    [anon_sym_for] = ACTIONS(1432),
    [aux_sym__type_identifier_token1] = ACTIONS(1432),
    [anon_sym_if] = ACTIONS(1432),
    [anon_sym_else] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1434),
    [aux_sym_number_literal_token1] = ACTIONS(1432),
    [aux_sym_number_literal_token2] = ACTIONS(1434),
    [aux_sym_number_literal_token3] = ACTIONS(1434),
    [aux_sym_number_literal_token4] = ACTIONS(1432),
    [aux_sym_number_literal_token5] = ACTIONS(1434),
    [anon_sym_return] = ACTIONS(1432),
    [anon_sym_SQUOTE] = ACTIONS(1432),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1434),
    [anon_sym_DQUOTE] = ACTIONS(1432),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1434),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1434),
    [anon_sym_switch] = ACTIONS(1432),
    [anon_sym_case] = ACTIONS(1432),
    [anon_sym_default] = ACTIONS(1432),
    [anon_sym_QMARK] = ACTIONS(1432),
    [anon_sym_try] = ACTIONS(1432),
    [anon_sym_int] = ACTIONS(1432),
    [anon_sym_boolean] = ACTIONS(1432),
    [anon_sym_char] = ACTIONS(1432),
    [anon_sym_short] = ACTIONS(1432),
    [anon_sym_long] = ACTIONS(1432),
    [anon_sym_float] = ACTIONS(1432),
    [anon_sym_double] = ACTIONS(1432),
    [anon_sym_void] = ACTIONS(1432),
    [anon_sym_public] = ACTIONS(1432),
    [anon_sym_protected] = ACTIONS(1432),
    [anon_sym_private] = ACTIONS(1432),
    [anon_sym_static] = ACTIONS(1432),
    [anon_sym_final] = ACTIONS(1432),
    [anon_sym_synchronized] = ACTIONS(1432),
    [anon_sym_TILDE] = ACTIONS(1434),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_new] = ACTIONS(1432),
  },
  [STATE(515)] = {
    [sym_comment] = STATE(515),
    [sym_groovy_doc] = STATE(515),
    [sym_identifier] = ACTIONS(1610),
    [anon_sym_break] = ACTIONS(1610),
    [anon_sym_continue] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1612),
    [anon_sym_import] = ACTIONS(1610),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_assert] = ACTIONS(1610),
    [anon_sym_PLUS_PLUS] = ACTIONS(1612),
    [anon_sym_DASH_DASH] = ACTIONS(1612),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_ATinterface] = ACTIONS(1612),
    [anon_sym_interface] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_def] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_null] = ACTIONS(1610),
    [anon_sym_this] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1610),
    [anon_sym_for] = ACTIONS(1610),
    [aux_sym__type_identifier_token1] = ACTIONS(1610),
    [anon_sym_if] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1612),
    [aux_sym_number_literal_token1] = ACTIONS(1610),
    [aux_sym_number_literal_token2] = ACTIONS(1612),
    [aux_sym_number_literal_token3] = ACTIONS(1612),
    [aux_sym_number_literal_token4] = ACTIONS(1610),
    [aux_sym_number_literal_token5] = ACTIONS(1612),
    [anon_sym_return] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1612),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1612),
    [anon_sym_switch] = ACTIONS(1610),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1610),
    [anon_sym_catch] = ACTIONS(1610),
    [anon_sym_finally] = ACTIONS(1610),
    [anon_sym_int] = ACTIONS(1610),
    [anon_sym_boolean] = ACTIONS(1610),
    [anon_sym_char] = ACTIONS(1610),
    [anon_sym_short] = ACTIONS(1610),
    [anon_sym_long] = ACTIONS(1610),
    [anon_sym_float] = ACTIONS(1610),
    [anon_sym_double] = ACTIONS(1610),
    [anon_sym_void] = ACTIONS(1610),
    [anon_sym_public] = ACTIONS(1610),
    [anon_sym_protected] = ACTIONS(1610),
    [anon_sym_private] = ACTIONS(1610),
    [anon_sym_static] = ACTIONS(1610),
    [anon_sym_final] = ACTIONS(1610),
    [anon_sym_synchronized] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
  },
  [STATE(516)] = {
    [sym_comment] = STATE(516),
    [sym_groovy_doc] = STATE(516),
    [ts_builtin_sym_end] = ACTIONS(1457),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1457),
    [anon_sym_DOT_AT] = ACTIONS(1457),
    [anon_sym_QMARK_DOT] = ACTIONS(1457),
    [anon_sym_STAR_DOT] = ACTIONS(1457),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1457),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_DASH_GT] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_pipeline] = ACTIONS(1455),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(517)] = {
    [sym_comment] = STATE(517),
    [sym_groovy_doc] = STATE(517),
    [sym_identifier] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1616),
    [anon_sym_import] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1614),
    [anon_sym_AT] = ACTIONS(1614),
    [anon_sym_assert] = ACTIONS(1614),
    [anon_sym_PLUS_PLUS] = ACTIONS(1616),
    [anon_sym_DASH_DASH] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_ATinterface] = ACTIONS(1616),
    [anon_sym_interface] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1614),
    [anon_sym_def] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_null] = ACTIONS(1614),
    [anon_sym_this] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [aux_sym__type_identifier_token1] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1616),
    [aux_sym_number_literal_token1] = ACTIONS(1614),
    [aux_sym_number_literal_token2] = ACTIONS(1616),
    [aux_sym_number_literal_token3] = ACTIONS(1616),
    [aux_sym_number_literal_token4] = ACTIONS(1614),
    [aux_sym_number_literal_token5] = ACTIONS(1616),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1614),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1616),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1616),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_case] = ACTIONS(1614),
    [anon_sym_default] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1614),
    [anon_sym_catch] = ACTIONS(1614),
    [anon_sym_finally] = ACTIONS(1614),
    [anon_sym_int] = ACTIONS(1614),
    [anon_sym_boolean] = ACTIONS(1614),
    [anon_sym_char] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_float] = ACTIONS(1614),
    [anon_sym_double] = ACTIONS(1614),
    [anon_sym_void] = ACTIONS(1614),
    [anon_sym_public] = ACTIONS(1614),
    [anon_sym_protected] = ACTIONS(1614),
    [anon_sym_private] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_final] = ACTIONS(1614),
    [anon_sym_synchronized] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
  },
  [STATE(518)] = {
    [sym_comment] = STATE(518),
    [sym_groovy_doc] = STATE(518),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(519)] = {
    [sym_comment] = STATE(519),
    [sym_groovy_doc] = STATE(519),
    [sym_identifier] = ACTIONS(1416),
    [anon_sym_break] = ACTIONS(1416),
    [anon_sym_continue] = ACTIONS(1416),
    [anon_sym_SEMI] = ACTIONS(1414),
    [anon_sym_DOT_AMP] = ACTIONS(1414),
    [anon_sym_DOT_AT] = ACTIONS(1414),
    [anon_sym_QMARK_DOT] = ACTIONS(1414),
    [anon_sym_STAR_DOT] = ACTIONS(1414),
    [anon_sym_STAR] = ACTIONS(1416),
    [anon_sym_STAR_COLON] = ACTIONS(1414),
    [anon_sym_DOT] = ACTIONS(1416),
    [anon_sym_import] = ACTIONS(1416),
    [anon_sym_as] = ACTIONS(1416),
    [anon_sym_package] = ACTIONS(1416),
    [anon_sym_AT] = ACTIONS(1416),
    [anon_sym_assert] = ACTIONS(1416),
    [anon_sym_PLUS_PLUS] = ACTIONS(1414),
    [anon_sym_DASH_DASH] = ACTIONS(1414),
    [anon_sym_PERCENT] = ACTIONS(1414),
    [anon_sym_SLASH] = ACTIONS(1416),
    [anon_sym_PLUS] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_LT_LT] = ACTIONS(1414),
    [anon_sym_GT_GT] = ACTIONS(1416),
    [anon_sym_GT_GT_GT] = ACTIONS(1414),
    [anon_sym_DOT_DOT] = ACTIONS(1416),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1414),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1416),
    [anon_sym_LT] = ACTIONS(1416),
    [anon_sym_LT_EQ] = ACTIONS(1416),
    [anon_sym_GT] = ACTIONS(1416),
    [anon_sym_GT_EQ] = ACTIONS(1414),
    [anon_sym_in] = ACTIONS(1416),
    [anon_sym_BANGin] = ACTIONS(1416),
    [anon_sym_instanceof] = ACTIONS(1416),
    [anon_sym_BANGinstanceof] = ACTIONS(1414),
    [anon_sym_EQ_EQ] = ACTIONS(1416),
    [anon_sym_BANG_EQ] = ACTIONS(1416),
    [anon_sym_LT_EQ_GT] = ACTIONS(1414),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1414),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1414),
    [anon_sym_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1414),
    [anon_sym_AMP] = ACTIONS(1416),
    [anon_sym_CARET] = ACTIONS(1414),
    [anon_sym_PIPE] = ACTIONS(1416),
    [anon_sym_AMP_AMP] = ACTIONS(1414),
    [anon_sym_PIPE_PIPE] = ACTIONS(1414),
    [anon_sym_QMARK_COLON] = ACTIONS(1414),
    [anon_sym_STAR_STAR] = ACTIONS(1414),
    [anon_sym_true] = ACTIONS(1416),
    [anon_sym_false] = ACTIONS(1416),
    [anon_sym_ATinterface] = ACTIONS(1414),
    [anon_sym_interface] = ACTIONS(1416),
    [anon_sym_class] = ACTIONS(1416),
    [anon_sym_COMMA] = ACTIONS(1414),
    [anon_sym_LBRACE] = ACTIONS(1414),
    [anon_sym_RBRACE] = ACTIONS(1414),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1416),
    [anon_sym_def] = ACTIONS(1416),
    [anon_sym_LPAREN] = ACTIONS(1414),
    [anon_sym_null] = ACTIONS(1416),
    [anon_sym_this] = ACTIONS(1416),
    [anon_sym_do] = ACTIONS(1416),
    [anon_sym_while] = ACTIONS(1416),
    [anon_sym_for] = ACTIONS(1416),
    [aux_sym__type_identifier_token1] = ACTIONS(1416),
    [anon_sym_if] = ACTIONS(1416),
    [anon_sym_else] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1414),
    [aux_sym_number_literal_token1] = ACTIONS(1416),
    [aux_sym_number_literal_token2] = ACTIONS(1414),
    [aux_sym_number_literal_token3] = ACTIONS(1414),
    [aux_sym_number_literal_token4] = ACTIONS(1416),
    [aux_sym_number_literal_token5] = ACTIONS(1414),
    [anon_sym_return] = ACTIONS(1416),
    [anon_sym_SQUOTE] = ACTIONS(1416),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1414),
    [anon_sym_DQUOTE] = ACTIONS(1416),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1414),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1414),
    [anon_sym_switch] = ACTIONS(1416),
    [anon_sym_case] = ACTIONS(1416),
    [anon_sym_default] = ACTIONS(1416),
    [anon_sym_QMARK] = ACTIONS(1416),
    [anon_sym_try] = ACTIONS(1416),
    [anon_sym_int] = ACTIONS(1416),
    [anon_sym_boolean] = ACTIONS(1416),
    [anon_sym_char] = ACTIONS(1416),
    [anon_sym_short] = ACTIONS(1416),
    [anon_sym_long] = ACTIONS(1416),
    [anon_sym_float] = ACTIONS(1416),
    [anon_sym_double] = ACTIONS(1416),
    [anon_sym_void] = ACTIONS(1416),
    [anon_sym_public] = ACTIONS(1416),
    [anon_sym_protected] = ACTIONS(1416),
    [anon_sym_private] = ACTIONS(1416),
    [anon_sym_static] = ACTIONS(1416),
    [anon_sym_final] = ACTIONS(1416),
    [anon_sym_synchronized] = ACTIONS(1416),
    [anon_sym_TILDE] = ACTIONS(1414),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_new] = ACTIONS(1416),
  },
  [STATE(520)] = {
    [sym_comment] = STATE(520),
    [sym_groovy_doc] = STATE(520),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(521)] = {
    [sym_comment] = STATE(521),
    [sym_groovy_doc] = STATE(521),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(522)] = {
    [sym_comment] = STATE(522),
    [sym_groovy_doc] = STATE(522),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(523)] = {
    [sym_comment] = STATE(523),
    [sym_groovy_doc] = STATE(523),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(524)] = {
    [sym_comment] = STATE(524),
    [sym_groovy_doc] = STATE(524),
    [sym_identifier] = ACTIONS(1420),
    [anon_sym_break] = ACTIONS(1420),
    [anon_sym_continue] = ACTIONS(1420),
    [anon_sym_SEMI] = ACTIONS(1418),
    [anon_sym_DOT_AMP] = ACTIONS(1418),
    [anon_sym_DOT_AT] = ACTIONS(1418),
    [anon_sym_QMARK_DOT] = ACTIONS(1418),
    [anon_sym_STAR_DOT] = ACTIONS(1418),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_STAR_COLON] = ACTIONS(1418),
    [anon_sym_DOT] = ACTIONS(1420),
    [anon_sym_import] = ACTIONS(1420),
    [anon_sym_as] = ACTIONS(1420),
    [anon_sym_package] = ACTIONS(1420),
    [anon_sym_AT] = ACTIONS(1420),
    [anon_sym_assert] = ACTIONS(1420),
    [anon_sym_PLUS_PLUS] = ACTIONS(1418),
    [anon_sym_DASH_DASH] = ACTIONS(1418),
    [anon_sym_PERCENT] = ACTIONS(1418),
    [anon_sym_SLASH] = ACTIONS(1420),
    [anon_sym_PLUS] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_LT_LT] = ACTIONS(1418),
    [anon_sym_GT_GT] = ACTIONS(1420),
    [anon_sym_GT_GT_GT] = ACTIONS(1418),
    [anon_sym_DOT_DOT] = ACTIONS(1420),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1418),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1420),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_LT_EQ] = ACTIONS(1420),
    [anon_sym_GT] = ACTIONS(1420),
    [anon_sym_GT_EQ] = ACTIONS(1418),
    [anon_sym_in] = ACTIONS(1420),
    [anon_sym_BANGin] = ACTIONS(1420),
    [anon_sym_instanceof] = ACTIONS(1420),
    [anon_sym_BANGinstanceof] = ACTIONS(1418),
    [anon_sym_EQ_EQ] = ACTIONS(1420),
    [anon_sym_BANG_EQ] = ACTIONS(1420),
    [anon_sym_LT_EQ_GT] = ACTIONS(1418),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1418),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1418),
    [anon_sym_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1418),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_CARET] = ACTIONS(1418),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_AMP_AMP] = ACTIONS(1418),
    [anon_sym_PIPE_PIPE] = ACTIONS(1418),
    [anon_sym_QMARK_COLON] = ACTIONS(1418),
    [anon_sym_STAR_STAR] = ACTIONS(1418),
    [anon_sym_true] = ACTIONS(1420),
    [anon_sym_false] = ACTIONS(1420),
    [anon_sym_ATinterface] = ACTIONS(1418),
    [anon_sym_interface] = ACTIONS(1420),
    [anon_sym_class] = ACTIONS(1420),
    [anon_sym_COMMA] = ACTIONS(1418),
    [anon_sym_LBRACE] = ACTIONS(1418),
    [anon_sym_RBRACE] = ACTIONS(1418),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1420),
    [anon_sym_def] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1418),
    [anon_sym_null] = ACTIONS(1420),
    [anon_sym_this] = ACTIONS(1420),
    [anon_sym_do] = ACTIONS(1420),
    [anon_sym_while] = ACTIONS(1420),
    [anon_sym_for] = ACTIONS(1420),
    [aux_sym__type_identifier_token1] = ACTIONS(1420),
    [anon_sym_if] = ACTIONS(1420),
    [anon_sym_else] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1418),
    [aux_sym_number_literal_token1] = ACTIONS(1420),
    [aux_sym_number_literal_token2] = ACTIONS(1418),
    [aux_sym_number_literal_token3] = ACTIONS(1418),
    [aux_sym_number_literal_token4] = ACTIONS(1420),
    [aux_sym_number_literal_token5] = ACTIONS(1418),
    [anon_sym_return] = ACTIONS(1420),
    [anon_sym_SQUOTE] = ACTIONS(1420),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1418),
    [anon_sym_DQUOTE] = ACTIONS(1420),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1418),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1418),
    [anon_sym_switch] = ACTIONS(1420),
    [anon_sym_case] = ACTIONS(1420),
    [anon_sym_default] = ACTIONS(1420),
    [anon_sym_QMARK] = ACTIONS(1420),
    [anon_sym_try] = ACTIONS(1420),
    [anon_sym_int] = ACTIONS(1420),
    [anon_sym_boolean] = ACTIONS(1420),
    [anon_sym_char] = ACTIONS(1420),
    [anon_sym_short] = ACTIONS(1420),
    [anon_sym_long] = ACTIONS(1420),
    [anon_sym_float] = ACTIONS(1420),
    [anon_sym_double] = ACTIONS(1420),
    [anon_sym_void] = ACTIONS(1420),
    [anon_sym_public] = ACTIONS(1420),
    [anon_sym_protected] = ACTIONS(1420),
    [anon_sym_private] = ACTIONS(1420),
    [anon_sym_static] = ACTIONS(1420),
    [anon_sym_final] = ACTIONS(1420),
    [anon_sym_synchronized] = ACTIONS(1420),
    [anon_sym_TILDE] = ACTIONS(1418),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_new] = ACTIONS(1420),
  },
  [STATE(525)] = {
    [sym_comment] = STATE(525),
    [sym_groovy_doc] = STATE(525),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(526)] = {
    [sym_comment] = STATE(526),
    [sym_groovy_doc] = STATE(526),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(527)] = {
    [sym_comment] = STATE(527),
    [sym_groovy_doc] = STATE(527),
    [ts_builtin_sym_end] = ACTIONS(1612),
    [sym_identifier] = ACTIONS(1610),
    [anon_sym_break] = ACTIONS(1610),
    [anon_sym_continue] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1612),
    [anon_sym_import] = ACTIONS(1610),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_assert] = ACTIONS(1610),
    [anon_sym_PLUS_PLUS] = ACTIONS(1612),
    [anon_sym_DASH_DASH] = ACTIONS(1612),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_ATinterface] = ACTIONS(1612),
    [anon_sym_interface] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_def] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_null] = ACTIONS(1610),
    [anon_sym_this] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1610),
    [anon_sym_for] = ACTIONS(1610),
    [aux_sym__type_identifier_token1] = ACTIONS(1610),
    [anon_sym_if] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1612),
    [aux_sym_number_literal_token1] = ACTIONS(1610),
    [aux_sym_number_literal_token2] = ACTIONS(1612),
    [aux_sym_number_literal_token3] = ACTIONS(1612),
    [aux_sym_number_literal_token4] = ACTIONS(1610),
    [aux_sym_number_literal_token5] = ACTIONS(1612),
    [anon_sym_pipeline] = ACTIONS(1610),
    [anon_sym_return] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1612),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1612),
    [anon_sym_switch] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1610),
    [anon_sym_catch] = ACTIONS(1610),
    [anon_sym_finally] = ACTIONS(1610),
    [anon_sym_int] = ACTIONS(1610),
    [anon_sym_boolean] = ACTIONS(1610),
    [anon_sym_char] = ACTIONS(1610),
    [anon_sym_short] = ACTIONS(1610),
    [anon_sym_long] = ACTIONS(1610),
    [anon_sym_float] = ACTIONS(1610),
    [anon_sym_double] = ACTIONS(1610),
    [anon_sym_void] = ACTIONS(1610),
    [anon_sym_public] = ACTIONS(1610),
    [anon_sym_protected] = ACTIONS(1610),
    [anon_sym_private] = ACTIONS(1610),
    [anon_sym_static] = ACTIONS(1610),
    [anon_sym_final] = ACTIONS(1610),
    [anon_sym_synchronized] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
  },
  [STATE(528)] = {
    [sym_comment] = STATE(528),
    [sym_groovy_doc] = STATE(528),
    [sym_identifier] = ACTIONS(1608),
    [anon_sym_break] = ACTIONS(1608),
    [anon_sym_continue] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1606),
    [anon_sym_import] = ACTIONS(1608),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_assert] = ACTIONS(1608),
    [anon_sym_PLUS_PLUS] = ACTIONS(1606),
    [anon_sym_DASH_DASH] = ACTIONS(1606),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1608),
    [anon_sym_false] = ACTIONS(1608),
    [anon_sym_ATinterface] = ACTIONS(1606),
    [anon_sym_interface] = ACTIONS(1608),
    [anon_sym_class] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_def] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1606),
    [anon_sym_null] = ACTIONS(1608),
    [anon_sym_this] = ACTIONS(1608),
    [anon_sym_do] = ACTIONS(1608),
    [anon_sym_while] = ACTIONS(1608),
    [anon_sym_for] = ACTIONS(1608),
    [aux_sym__type_identifier_token1] = ACTIONS(1608),
    [anon_sym_if] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [aux_sym_number_literal_token1] = ACTIONS(1608),
    [aux_sym_number_literal_token2] = ACTIONS(1606),
    [aux_sym_number_literal_token3] = ACTIONS(1606),
    [aux_sym_number_literal_token4] = ACTIONS(1608),
    [aux_sym_number_literal_token5] = ACTIONS(1606),
    [anon_sym_return] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1606),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1606),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1606),
    [anon_sym_switch] = ACTIONS(1608),
    [anon_sym_case] = ACTIONS(1608),
    [anon_sym_default] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1608),
    [anon_sym_catch] = ACTIONS(1608),
    [anon_sym_finally] = ACTIONS(1608),
    [anon_sym_int] = ACTIONS(1608),
    [anon_sym_boolean] = ACTIONS(1608),
    [anon_sym_char] = ACTIONS(1608),
    [anon_sym_short] = ACTIONS(1608),
    [anon_sym_long] = ACTIONS(1608),
    [anon_sym_float] = ACTIONS(1608),
    [anon_sym_double] = ACTIONS(1608),
    [anon_sym_void] = ACTIONS(1608),
    [anon_sym_public] = ACTIONS(1608),
    [anon_sym_protected] = ACTIONS(1608),
    [anon_sym_private] = ACTIONS(1608),
    [anon_sym_static] = ACTIONS(1608),
    [anon_sym_final] = ACTIONS(1608),
    [anon_sym_synchronized] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_new] = ACTIONS(1608),
  },
  [STATE(529)] = {
    [sym_comment] = STATE(529),
    [sym_groovy_doc] = STATE(529),
    [sym_identifier] = ACTIONS(1618),
    [anon_sym_break] = ACTIONS(1618),
    [anon_sym_continue] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1620),
    [anon_sym_import] = ACTIONS(1618),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_assert] = ACTIONS(1618),
    [anon_sym_PLUS_PLUS] = ACTIONS(1620),
    [anon_sym_DASH_DASH] = ACTIONS(1620),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_ATinterface] = ACTIONS(1620),
    [anon_sym_interface] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_def] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_null] = ACTIONS(1618),
    [anon_sym_this] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_while] = ACTIONS(1618),
    [anon_sym_for] = ACTIONS(1618),
    [aux_sym__type_identifier_token1] = ACTIONS(1618),
    [anon_sym_if] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1620),
    [aux_sym_number_literal_token1] = ACTIONS(1618),
    [aux_sym_number_literal_token2] = ACTIONS(1620),
    [aux_sym_number_literal_token3] = ACTIONS(1620),
    [aux_sym_number_literal_token4] = ACTIONS(1618),
    [aux_sym_number_literal_token5] = ACTIONS(1620),
    [anon_sym_return] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1620),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1620),
    [anon_sym_switch] = ACTIONS(1618),
    [anon_sym_case] = ACTIONS(1618),
    [anon_sym_default] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1618),
    [anon_sym_catch] = ACTIONS(1618),
    [anon_sym_finally] = ACTIONS(1618),
    [anon_sym_int] = ACTIONS(1618),
    [anon_sym_boolean] = ACTIONS(1618),
    [anon_sym_char] = ACTIONS(1618),
    [anon_sym_short] = ACTIONS(1618),
    [anon_sym_long] = ACTIONS(1618),
    [anon_sym_float] = ACTIONS(1618),
    [anon_sym_double] = ACTIONS(1618),
    [anon_sym_void] = ACTIONS(1618),
    [anon_sym_public] = ACTIONS(1618),
    [anon_sym_protected] = ACTIONS(1618),
    [anon_sym_private] = ACTIONS(1618),
    [anon_sym_static] = ACTIONS(1618),
    [anon_sym_final] = ACTIONS(1618),
    [anon_sym_synchronized] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1620),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
  },
  [STATE(530)] = {
    [sym_comment] = STATE(530),
    [sym_groovy_doc] = STATE(530),
    [sym_identifier] = ACTIONS(1622),
    [anon_sym_break] = ACTIONS(1622),
    [anon_sym_continue] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1624),
    [anon_sym_import] = ACTIONS(1622),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_assert] = ACTIONS(1622),
    [anon_sym_PLUS_PLUS] = ACTIONS(1624),
    [anon_sym_DASH_DASH] = ACTIONS(1624),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_ATinterface] = ACTIONS(1624),
    [anon_sym_interface] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_def] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_null] = ACTIONS(1622),
    [anon_sym_this] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_while] = ACTIONS(1622),
    [anon_sym_for] = ACTIONS(1622),
    [aux_sym__type_identifier_token1] = ACTIONS(1622),
    [anon_sym_if] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1624),
    [aux_sym_number_literal_token1] = ACTIONS(1622),
    [aux_sym_number_literal_token2] = ACTIONS(1624),
    [aux_sym_number_literal_token3] = ACTIONS(1624),
    [aux_sym_number_literal_token4] = ACTIONS(1622),
    [aux_sym_number_literal_token5] = ACTIONS(1624),
    [anon_sym_return] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1624),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1624),
    [anon_sym_switch] = ACTIONS(1622),
    [anon_sym_case] = ACTIONS(1622),
    [anon_sym_default] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1622),
    [anon_sym_catch] = ACTIONS(1622),
    [anon_sym_finally] = ACTIONS(1622),
    [anon_sym_int] = ACTIONS(1622),
    [anon_sym_boolean] = ACTIONS(1622),
    [anon_sym_char] = ACTIONS(1622),
    [anon_sym_short] = ACTIONS(1622),
    [anon_sym_long] = ACTIONS(1622),
    [anon_sym_float] = ACTIONS(1622),
    [anon_sym_double] = ACTIONS(1622),
    [anon_sym_void] = ACTIONS(1622),
    [anon_sym_public] = ACTIONS(1622),
    [anon_sym_protected] = ACTIONS(1622),
    [anon_sym_private] = ACTIONS(1622),
    [anon_sym_static] = ACTIONS(1622),
    [anon_sym_final] = ACTIONS(1622),
    [anon_sym_synchronized] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1624),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
  },
  [STATE(531)] = {
    [sym_comment] = STATE(531),
    [sym_groovy_doc] = STATE(531),
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_break] = ACTIONS(1626),
    [anon_sym_continue] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1628),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_PLUS_PLUS] = ACTIONS(1628),
    [anon_sym_DASH_DASH] = ACTIONS(1628),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_ATinterface] = ACTIONS(1628),
    [anon_sym_interface] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_null] = ACTIONS(1626),
    [anon_sym_this] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [aux_sym__type_identifier_token1] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1628),
    [aux_sym_number_literal_token1] = ACTIONS(1626),
    [aux_sym_number_literal_token2] = ACTIONS(1628),
    [aux_sym_number_literal_token3] = ACTIONS(1628),
    [aux_sym_number_literal_token4] = ACTIONS(1626),
    [aux_sym_number_literal_token5] = ACTIONS(1628),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1628),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1628),
    [anon_sym_switch] = ACTIONS(1626),
    [anon_sym_case] = ACTIONS(1626),
    [anon_sym_default] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_catch] = ACTIONS(1626),
    [anon_sym_finally] = ACTIONS(1626),
    [anon_sym_int] = ACTIONS(1626),
    [anon_sym_boolean] = ACTIONS(1626),
    [anon_sym_char] = ACTIONS(1626),
    [anon_sym_short] = ACTIONS(1626),
    [anon_sym_long] = ACTIONS(1626),
    [anon_sym_float] = ACTIONS(1626),
    [anon_sym_double] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_public] = ACTIONS(1626),
    [anon_sym_protected] = ACTIONS(1626),
    [anon_sym_private] = ACTIONS(1626),
    [anon_sym_static] = ACTIONS(1626),
    [anon_sym_final] = ACTIONS(1626),
    [anon_sym_synchronized] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1628),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
  },
  [STATE(532)] = {
    [sym_comment] = STATE(532),
    [sym_groovy_doc] = STATE(532),
    [sym_identifier] = ACTIONS(1630),
    [anon_sym_break] = ACTIONS(1630),
    [anon_sym_continue] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1632),
    [anon_sym_import] = ACTIONS(1630),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_assert] = ACTIONS(1630),
    [anon_sym_PLUS_PLUS] = ACTIONS(1632),
    [anon_sym_DASH_DASH] = ACTIONS(1632),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_ATinterface] = ACTIONS(1632),
    [anon_sym_interface] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_def] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_null] = ACTIONS(1630),
    [anon_sym_this] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_while] = ACTIONS(1630),
    [anon_sym_for] = ACTIONS(1630),
    [aux_sym__type_identifier_token1] = ACTIONS(1630),
    [anon_sym_if] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1632),
    [aux_sym_number_literal_token1] = ACTIONS(1630),
    [aux_sym_number_literal_token2] = ACTIONS(1632),
    [aux_sym_number_literal_token3] = ACTIONS(1632),
    [aux_sym_number_literal_token4] = ACTIONS(1630),
    [aux_sym_number_literal_token5] = ACTIONS(1632),
    [anon_sym_return] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1632),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1632),
    [anon_sym_switch] = ACTIONS(1630),
    [anon_sym_case] = ACTIONS(1630),
    [anon_sym_default] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1630),
    [anon_sym_catch] = ACTIONS(1630),
    [anon_sym_finally] = ACTIONS(1630),
    [anon_sym_int] = ACTIONS(1630),
    [anon_sym_boolean] = ACTIONS(1630),
    [anon_sym_char] = ACTIONS(1630),
    [anon_sym_short] = ACTIONS(1630),
    [anon_sym_long] = ACTIONS(1630),
    [anon_sym_float] = ACTIONS(1630),
    [anon_sym_double] = ACTIONS(1630),
    [anon_sym_void] = ACTIONS(1630),
    [anon_sym_public] = ACTIONS(1630),
    [anon_sym_protected] = ACTIONS(1630),
    [anon_sym_private] = ACTIONS(1630),
    [anon_sym_static] = ACTIONS(1630),
    [anon_sym_final] = ACTIONS(1630),
    [anon_sym_synchronized] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1632),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
  },
  [STATE(533)] = {
    [sym_comment] = STATE(533),
    [sym_groovy_doc] = STATE(533),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(534)] = {
    [sym_comment] = STATE(534),
    [sym_groovy_doc] = STATE(534),
    [sym_identifier] = ACTIONS(1634),
    [anon_sym_break] = ACTIONS(1634),
    [anon_sym_continue] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1636),
    [anon_sym_import] = ACTIONS(1634),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1634),
    [anon_sym_AT] = ACTIONS(1634),
    [anon_sym_assert] = ACTIONS(1634),
    [anon_sym_PLUS_PLUS] = ACTIONS(1636),
    [anon_sym_DASH_DASH] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_ATinterface] = ACTIONS(1636),
    [anon_sym_interface] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1634),
    [anon_sym_def] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_null] = ACTIONS(1634),
    [anon_sym_this] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_while] = ACTIONS(1634),
    [anon_sym_for] = ACTIONS(1634),
    [aux_sym__type_identifier_token1] = ACTIONS(1634),
    [anon_sym_if] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1636),
    [aux_sym_number_literal_token1] = ACTIONS(1634),
    [aux_sym_number_literal_token2] = ACTIONS(1636),
    [aux_sym_number_literal_token3] = ACTIONS(1636),
    [aux_sym_number_literal_token4] = ACTIONS(1634),
    [aux_sym_number_literal_token5] = ACTIONS(1636),
    [anon_sym_return] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1634),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1634),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1636),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1636),
    [anon_sym_switch] = ACTIONS(1634),
    [anon_sym_case] = ACTIONS(1634),
    [anon_sym_default] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1634),
    [anon_sym_catch] = ACTIONS(1634),
    [anon_sym_finally] = ACTIONS(1634),
    [anon_sym_int] = ACTIONS(1634),
    [anon_sym_boolean] = ACTIONS(1634),
    [anon_sym_char] = ACTIONS(1634),
    [anon_sym_short] = ACTIONS(1634),
    [anon_sym_long] = ACTIONS(1634),
    [anon_sym_float] = ACTIONS(1634),
    [anon_sym_double] = ACTIONS(1634),
    [anon_sym_void] = ACTIONS(1634),
    [anon_sym_public] = ACTIONS(1634),
    [anon_sym_protected] = ACTIONS(1634),
    [anon_sym_private] = ACTIONS(1634),
    [anon_sym_static] = ACTIONS(1634),
    [anon_sym_final] = ACTIONS(1634),
    [anon_sym_synchronized] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
  },
  [STATE(535)] = {
    [sym_comment] = STATE(535),
    [sym_groovy_doc] = STATE(535),
    [sym_identifier] = ACTIONS(1638),
    [anon_sym_break] = ACTIONS(1638),
    [anon_sym_continue] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1640),
    [anon_sym_import] = ACTIONS(1638),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1638),
    [anon_sym_AT] = ACTIONS(1638),
    [anon_sym_assert] = ACTIONS(1638),
    [anon_sym_PLUS_PLUS] = ACTIONS(1640),
    [anon_sym_DASH_DASH] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_ATinterface] = ACTIONS(1640),
    [anon_sym_interface] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1638),
    [anon_sym_def] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_null] = ACTIONS(1638),
    [anon_sym_this] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_while] = ACTIONS(1638),
    [anon_sym_for] = ACTIONS(1638),
    [aux_sym__type_identifier_token1] = ACTIONS(1638),
    [anon_sym_if] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1640),
    [aux_sym_number_literal_token1] = ACTIONS(1638),
    [aux_sym_number_literal_token2] = ACTIONS(1640),
    [aux_sym_number_literal_token3] = ACTIONS(1640),
    [aux_sym_number_literal_token4] = ACTIONS(1638),
    [aux_sym_number_literal_token5] = ACTIONS(1640),
    [anon_sym_return] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1638),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1638),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1640),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1640),
    [anon_sym_switch] = ACTIONS(1638),
    [anon_sym_case] = ACTIONS(1638),
    [anon_sym_default] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1638),
    [anon_sym_catch] = ACTIONS(1638),
    [anon_sym_finally] = ACTIONS(1638),
    [anon_sym_int] = ACTIONS(1638),
    [anon_sym_boolean] = ACTIONS(1638),
    [anon_sym_char] = ACTIONS(1638),
    [anon_sym_short] = ACTIONS(1638),
    [anon_sym_long] = ACTIONS(1638),
    [anon_sym_float] = ACTIONS(1638),
    [anon_sym_double] = ACTIONS(1638),
    [anon_sym_void] = ACTIONS(1638),
    [anon_sym_public] = ACTIONS(1638),
    [anon_sym_protected] = ACTIONS(1638),
    [anon_sym_private] = ACTIONS(1638),
    [anon_sym_static] = ACTIONS(1638),
    [anon_sym_final] = ACTIONS(1638),
    [anon_sym_synchronized] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
  },
  [STATE(536)] = {
    [sym_comment] = STATE(536),
    [sym_groovy_doc] = STATE(536),
    [sym_identifier] = ACTIONS(1388),
    [anon_sym_break] = ACTIONS(1388),
    [anon_sym_continue] = ACTIONS(1388),
    [anon_sym_SEMI] = ACTIONS(1386),
    [anon_sym_DOT_AMP] = ACTIONS(1386),
    [anon_sym_DOT_AT] = ACTIONS(1386),
    [anon_sym_QMARK_DOT] = ACTIONS(1386),
    [anon_sym_STAR_DOT] = ACTIONS(1386),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_STAR_COLON] = ACTIONS(1386),
    [anon_sym_DOT] = ACTIONS(1388),
    [anon_sym_import] = ACTIONS(1388),
    [anon_sym_as] = ACTIONS(1388),
    [anon_sym_package] = ACTIONS(1388),
    [anon_sym_AT] = ACTIONS(1388),
    [anon_sym_assert] = ACTIONS(1388),
    [anon_sym_PLUS_PLUS] = ACTIONS(1386),
    [anon_sym_DASH_DASH] = ACTIONS(1386),
    [anon_sym_PERCENT] = ACTIONS(1386),
    [anon_sym_SLASH] = ACTIONS(1388),
    [anon_sym_PLUS] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_LT_LT] = ACTIONS(1386),
    [anon_sym_GT_GT] = ACTIONS(1388),
    [anon_sym_GT_GT_GT] = ACTIONS(1386),
    [anon_sym_DOT_DOT] = ACTIONS(1388),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1386),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1388),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_LT_EQ] = ACTIONS(1388),
    [anon_sym_GT] = ACTIONS(1388),
    [anon_sym_GT_EQ] = ACTIONS(1386),
    [anon_sym_in] = ACTIONS(1388),
    [anon_sym_BANGin] = ACTIONS(1388),
    [anon_sym_instanceof] = ACTIONS(1388),
    [anon_sym_BANGinstanceof] = ACTIONS(1386),
    [anon_sym_EQ_EQ] = ACTIONS(1388),
    [anon_sym_BANG_EQ] = ACTIONS(1388),
    [anon_sym_LT_EQ_GT] = ACTIONS(1386),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1386),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1386),
    [anon_sym_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1386),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_CARET] = ACTIONS(1386),
    [anon_sym_PIPE] = ACTIONS(1388),
    [anon_sym_AMP_AMP] = ACTIONS(1386),
    [anon_sym_PIPE_PIPE] = ACTIONS(1386),
    [anon_sym_QMARK_COLON] = ACTIONS(1386),
    [anon_sym_STAR_STAR] = ACTIONS(1386),
    [anon_sym_true] = ACTIONS(1388),
    [anon_sym_false] = ACTIONS(1388),
    [anon_sym_ATinterface] = ACTIONS(1386),
    [anon_sym_interface] = ACTIONS(1388),
    [anon_sym_class] = ACTIONS(1388),
    [anon_sym_COMMA] = ACTIONS(1386),
    [anon_sym_LBRACE] = ACTIONS(1386),
    [anon_sym_RBRACE] = ACTIONS(1386),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1388),
    [anon_sym_def] = ACTIONS(1388),
    [anon_sym_LPAREN] = ACTIONS(1386),
    [anon_sym_null] = ACTIONS(1388),
    [anon_sym_this] = ACTIONS(1388),
    [anon_sym_do] = ACTIONS(1388),
    [anon_sym_while] = ACTIONS(1388),
    [anon_sym_for] = ACTIONS(1388),
    [aux_sym__type_identifier_token1] = ACTIONS(1388),
    [anon_sym_if] = ACTIONS(1388),
    [anon_sym_else] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1386),
    [aux_sym_number_literal_token1] = ACTIONS(1388),
    [aux_sym_number_literal_token2] = ACTIONS(1386),
    [aux_sym_number_literal_token3] = ACTIONS(1386),
    [aux_sym_number_literal_token4] = ACTIONS(1388),
    [aux_sym_number_literal_token5] = ACTIONS(1386),
    [anon_sym_return] = ACTIONS(1388),
    [anon_sym_SQUOTE] = ACTIONS(1388),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1386),
    [anon_sym_DQUOTE] = ACTIONS(1388),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1386),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1386),
    [anon_sym_switch] = ACTIONS(1388),
    [anon_sym_case] = ACTIONS(1388),
    [anon_sym_default] = ACTIONS(1388),
    [anon_sym_QMARK] = ACTIONS(1388),
    [anon_sym_try] = ACTIONS(1388),
    [anon_sym_int] = ACTIONS(1388),
    [anon_sym_boolean] = ACTIONS(1388),
    [anon_sym_char] = ACTIONS(1388),
    [anon_sym_short] = ACTIONS(1388),
    [anon_sym_long] = ACTIONS(1388),
    [anon_sym_float] = ACTIONS(1388),
    [anon_sym_double] = ACTIONS(1388),
    [anon_sym_void] = ACTIONS(1388),
    [anon_sym_public] = ACTIONS(1388),
    [anon_sym_protected] = ACTIONS(1388),
    [anon_sym_private] = ACTIONS(1388),
    [anon_sym_static] = ACTIONS(1388),
    [anon_sym_final] = ACTIONS(1388),
    [anon_sym_synchronized] = ACTIONS(1388),
    [anon_sym_TILDE] = ACTIONS(1386),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_new] = ACTIONS(1388),
  },
  [STATE(537)] = {
    [sym_comment] = STATE(537),
    [sym_groovy_doc] = STATE(537),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_catch] = ACTIONS(1310),
    [anon_sym_finally] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(538)] = {
    [sym_comment] = STATE(538),
    [sym_groovy_doc] = STATE(538),
    [ts_builtin_sym_end] = ACTIONS(1465),
    [sym_identifier] = ACTIONS(1463),
    [anon_sym_break] = ACTIONS(1463),
    [anon_sym_continue] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_DOT_AMP] = ACTIONS(1465),
    [anon_sym_DOT_AT] = ACTIONS(1465),
    [anon_sym_QMARK_DOT] = ACTIONS(1465),
    [anon_sym_STAR_DOT] = ACTIONS(1465),
    [anon_sym_STAR] = ACTIONS(1463),
    [anon_sym_STAR_COLON] = ACTIONS(1465),
    [anon_sym_import] = ACTIONS(1463),
    [anon_sym_as] = ACTIONS(1463),
    [anon_sym_package] = ACTIONS(1463),
    [anon_sym_AT] = ACTIONS(1463),
    [anon_sym_assert] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_PERCENT] = ACTIONS(1465),
    [anon_sym_SLASH] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [anon_sym_LT_LT] = ACTIONS(1465),
    [anon_sym_GT_GT] = ACTIONS(1463),
    [anon_sym_GT_GT_GT] = ACTIONS(1465),
    [anon_sym_DOT_DOT] = ACTIONS(1463),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1463),
    [anon_sym_LT] = ACTIONS(1463),
    [anon_sym_LT_EQ] = ACTIONS(1463),
    [anon_sym_GT] = ACTIONS(1463),
    [anon_sym_GT_EQ] = ACTIONS(1465),
    [anon_sym_in] = ACTIONS(1463),
    [anon_sym_BANGin] = ACTIONS(1463),
    [anon_sym_instanceof] = ACTIONS(1463),
    [anon_sym_BANGinstanceof] = ACTIONS(1465),
    [anon_sym_EQ_EQ] = ACTIONS(1463),
    [anon_sym_BANG_EQ] = ACTIONS(1463),
    [anon_sym_LT_EQ_GT] = ACTIONS(1465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1465),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1465),
    [anon_sym_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1463),
    [anon_sym_CARET] = ACTIONS(1465),
    [anon_sym_PIPE] = ACTIONS(1463),
    [anon_sym_AMP_AMP] = ACTIONS(1465),
    [anon_sym_PIPE_PIPE] = ACTIONS(1465),
    [anon_sym_QMARK_COLON] = ACTIONS(1465),
    [anon_sym_STAR_STAR] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(1463),
    [anon_sym_false] = ACTIONS(1463),
    [anon_sym_ATinterface] = ACTIONS(1465),
    [anon_sym_interface] = ACTIONS(1463),
    [anon_sym_class] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [anon_sym_DASH_GT] = ACTIONS(1465),
    [anon_sym_RBRACE] = ACTIONS(1465),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1463),
    [anon_sym_def] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_null] = ACTIONS(1463),
    [anon_sym_this] = ACTIONS(1463),
    [anon_sym_do] = ACTIONS(1463),
    [anon_sym_while] = ACTIONS(1463),
    [anon_sym_for] = ACTIONS(1463),
    [aux_sym__type_identifier_token1] = ACTIONS(1463),
    [anon_sym_if] = ACTIONS(1463),
    [anon_sym_else] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_number_literal_token1] = ACTIONS(1463),
    [aux_sym_number_literal_token2] = ACTIONS(1465),
    [aux_sym_number_literal_token3] = ACTIONS(1465),
    [aux_sym_number_literal_token4] = ACTIONS(1463),
    [aux_sym_number_literal_token5] = ACTIONS(1465),
    [anon_sym_pipeline] = ACTIONS(1463),
    [anon_sym_return] = ACTIONS(1463),
    [anon_sym_SQUOTE] = ACTIONS(1463),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(1463),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1465),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1465),
    [anon_sym_switch] = ACTIONS(1463),
    [anon_sym_QMARK] = ACTIONS(1463),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_int] = ACTIONS(1463),
    [anon_sym_boolean] = ACTIONS(1463),
    [anon_sym_char] = ACTIONS(1463),
    [anon_sym_short] = ACTIONS(1463),
    [anon_sym_long] = ACTIONS(1463),
    [anon_sym_float] = ACTIONS(1463),
    [anon_sym_double] = ACTIONS(1463),
    [anon_sym_void] = ACTIONS(1463),
    [anon_sym_public] = ACTIONS(1463),
    [anon_sym_protected] = ACTIONS(1463),
    [anon_sym_private] = ACTIONS(1463),
    [anon_sym_static] = ACTIONS(1463),
    [anon_sym_final] = ACTIONS(1463),
    [anon_sym_synchronized] = ACTIONS(1463),
    [anon_sym_TILDE] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1463),
    [anon_sym_new] = ACTIONS(1463),
  },
  [STATE(539)] = {
    [sym_comment] = STATE(539),
    [sym_groovy_doc] = STATE(539),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(540)] = {
    [sym_comment] = STATE(540),
    [sym_groovy_doc] = STATE(540),
    [ts_builtin_sym_end] = ACTIONS(1600),
    [sym_identifier] = ACTIONS(1598),
    [anon_sym_break] = ACTIONS(1598),
    [anon_sym_continue] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1600),
    [anon_sym_import] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1598),
    [anon_sym_AT] = ACTIONS(1598),
    [anon_sym_assert] = ACTIONS(1598),
    [anon_sym_PLUS_PLUS] = ACTIONS(1600),
    [anon_sym_DASH_DASH] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_ATinterface] = ACTIONS(1600),
    [anon_sym_interface] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1598),
    [anon_sym_def] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_null] = ACTIONS(1598),
    [anon_sym_this] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_while] = ACTIONS(1598),
    [anon_sym_for] = ACTIONS(1598),
    [aux_sym__type_identifier_token1] = ACTIONS(1598),
    [anon_sym_if] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1600),
    [aux_sym_number_literal_token1] = ACTIONS(1598),
    [aux_sym_number_literal_token2] = ACTIONS(1600),
    [aux_sym_number_literal_token3] = ACTIONS(1600),
    [aux_sym_number_literal_token4] = ACTIONS(1598),
    [aux_sym_number_literal_token5] = ACTIONS(1600),
    [anon_sym_pipeline] = ACTIONS(1598),
    [anon_sym_return] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1598),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1598),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1600),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1600),
    [anon_sym_switch] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1598),
    [anon_sym_catch] = ACTIONS(1598),
    [anon_sym_finally] = ACTIONS(1598),
    [anon_sym_int] = ACTIONS(1598),
    [anon_sym_boolean] = ACTIONS(1598),
    [anon_sym_char] = ACTIONS(1598),
    [anon_sym_short] = ACTIONS(1598),
    [anon_sym_long] = ACTIONS(1598),
    [anon_sym_float] = ACTIONS(1598),
    [anon_sym_double] = ACTIONS(1598),
    [anon_sym_void] = ACTIONS(1598),
    [anon_sym_public] = ACTIONS(1598),
    [anon_sym_protected] = ACTIONS(1598),
    [anon_sym_private] = ACTIONS(1598),
    [anon_sym_static] = ACTIONS(1598),
    [anon_sym_final] = ACTIONS(1598),
    [anon_sym_synchronized] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
  },
  [STATE(541)] = {
    [sym_comment] = STATE(541),
    [sym_groovy_doc] = STATE(541),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1262),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1262),
    [anon_sym_AT] = ACTIONS(1262),
    [anon_sym_assert] = ACTIONS(1262),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [anon_sym_ATinterface] = ACTIONS(1260),
    [anon_sym_interface] = ACTIONS(1262),
    [anon_sym_class] = ACTIONS(1262),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1262),
    [anon_sym_def] = ACTIONS(1262),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_null] = ACTIONS(1262),
    [anon_sym_this] = ACTIONS(1262),
    [anon_sym_do] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [aux_sym__type_identifier_token1] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_else] = ACTIONS(1262),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1262),
    [aux_sym_number_literal_token2] = ACTIONS(1260),
    [aux_sym_number_literal_token3] = ACTIONS(1260),
    [aux_sym_number_literal_token4] = ACTIONS(1262),
    [aux_sym_number_literal_token5] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1260),
    [anon_sym_DQUOTE] = ACTIONS(1262),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1260),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1260),
    [anon_sym_switch] = ACTIONS(1262),
    [anon_sym_case] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1262),
    [anon_sym_int] = ACTIONS(1262),
    [anon_sym_boolean] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_short] = ACTIONS(1262),
    [anon_sym_long] = ACTIONS(1262),
    [anon_sym_float] = ACTIONS(1262),
    [anon_sym_double] = ACTIONS(1262),
    [anon_sym_void] = ACTIONS(1262),
    [anon_sym_public] = ACTIONS(1262),
    [anon_sym_protected] = ACTIONS(1262),
    [anon_sym_private] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_final] = ACTIONS(1262),
    [anon_sym_synchronized] = ACTIONS(1262),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1262),
    [anon_sym_new] = ACTIONS(1262),
  },
  [STATE(542)] = {
    [sym_comment] = STATE(542),
    [sym_groovy_doc] = STATE(542),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(543)] = {
    [sym_comment] = STATE(543),
    [sym_groovy_doc] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1278),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_DASH_GT] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1278),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_pipeline] = ACTIONS(1280),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(544)] = {
    [sym_comment] = STATE(544),
    [sym_groovy_doc] = STATE(544),
    [ts_builtin_sym_end] = ACTIONS(1616),
    [sym_identifier] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1616),
    [anon_sym_import] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1614),
    [anon_sym_AT] = ACTIONS(1614),
    [anon_sym_assert] = ACTIONS(1614),
    [anon_sym_PLUS_PLUS] = ACTIONS(1616),
    [anon_sym_DASH_DASH] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_ATinterface] = ACTIONS(1616),
    [anon_sym_interface] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1614),
    [anon_sym_def] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_null] = ACTIONS(1614),
    [anon_sym_this] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [aux_sym__type_identifier_token1] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1616),
    [aux_sym_number_literal_token1] = ACTIONS(1614),
    [aux_sym_number_literal_token2] = ACTIONS(1616),
    [aux_sym_number_literal_token3] = ACTIONS(1616),
    [aux_sym_number_literal_token4] = ACTIONS(1614),
    [aux_sym_number_literal_token5] = ACTIONS(1616),
    [anon_sym_pipeline] = ACTIONS(1614),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1614),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1616),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1616),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1614),
    [anon_sym_catch] = ACTIONS(1614),
    [anon_sym_finally] = ACTIONS(1614),
    [anon_sym_int] = ACTIONS(1614),
    [anon_sym_boolean] = ACTIONS(1614),
    [anon_sym_char] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_float] = ACTIONS(1614),
    [anon_sym_double] = ACTIONS(1614),
    [anon_sym_void] = ACTIONS(1614),
    [anon_sym_public] = ACTIONS(1614),
    [anon_sym_protected] = ACTIONS(1614),
    [anon_sym_private] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_final] = ACTIONS(1614),
    [anon_sym_synchronized] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
  },
  [STATE(545)] = {
    [sym_comment] = STATE(545),
    [sym_groovy_doc] = STATE(545),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(546)] = {
    [sym_comment] = STATE(546),
    [sym_groovy_doc] = STATE(546),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(547)] = {
    [sym_comment] = STATE(547),
    [sym_groovy_doc] = STATE(547),
    [sym_identifier] = ACTIONS(1366),
    [anon_sym_break] = ACTIONS(1366),
    [anon_sym_continue] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1368),
    [anon_sym_DOT_AMP] = ACTIONS(1368),
    [anon_sym_DOT_AT] = ACTIONS(1368),
    [anon_sym_QMARK_DOT] = ACTIONS(1368),
    [anon_sym_STAR_DOT] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1366),
    [anon_sym_STAR_COLON] = ACTIONS(1368),
    [anon_sym_DOT] = ACTIONS(1366),
    [anon_sym_import] = ACTIONS(1366),
    [anon_sym_as] = ACTIONS(1366),
    [anon_sym_package] = ACTIONS(1366),
    [anon_sym_AT] = ACTIONS(1366),
    [anon_sym_assert] = ACTIONS(1366),
    [anon_sym_PLUS_PLUS] = ACTIONS(1368),
    [anon_sym_DASH_DASH] = ACTIONS(1368),
    [anon_sym_PERCENT] = ACTIONS(1368),
    [anon_sym_SLASH] = ACTIONS(1366),
    [anon_sym_PLUS] = ACTIONS(1366),
    [anon_sym_DASH] = ACTIONS(1366),
    [anon_sym_LT_LT] = ACTIONS(1368),
    [anon_sym_GT_GT] = ACTIONS(1366),
    [anon_sym_GT_GT_GT] = ACTIONS(1368),
    [anon_sym_DOT_DOT] = ACTIONS(1366),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1368),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1366),
    [anon_sym_LT_EQ] = ACTIONS(1366),
    [anon_sym_GT] = ACTIONS(1366),
    [anon_sym_GT_EQ] = ACTIONS(1368),
    [anon_sym_in] = ACTIONS(1366),
    [anon_sym_BANGin] = ACTIONS(1366),
    [anon_sym_instanceof] = ACTIONS(1366),
    [anon_sym_BANGinstanceof] = ACTIONS(1368),
    [anon_sym_EQ_EQ] = ACTIONS(1366),
    [anon_sym_BANG_EQ] = ACTIONS(1366),
    [anon_sym_LT_EQ_GT] = ACTIONS(1368),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1368),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1368),
    [anon_sym_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1368),
    [anon_sym_AMP] = ACTIONS(1366),
    [anon_sym_CARET] = ACTIONS(1368),
    [anon_sym_PIPE] = ACTIONS(1366),
    [anon_sym_AMP_AMP] = ACTIONS(1368),
    [anon_sym_PIPE_PIPE] = ACTIONS(1368),
    [anon_sym_QMARK_COLON] = ACTIONS(1368),
    [anon_sym_STAR_STAR] = ACTIONS(1368),
    [anon_sym_true] = ACTIONS(1366),
    [anon_sym_false] = ACTIONS(1366),
    [anon_sym_ATinterface] = ACTIONS(1368),
    [anon_sym_interface] = ACTIONS(1366),
    [anon_sym_class] = ACTIONS(1366),
    [anon_sym_COMMA] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1366),
    [anon_sym_def] = ACTIONS(1366),
    [anon_sym_LPAREN] = ACTIONS(1368),
    [anon_sym_null] = ACTIONS(1366),
    [anon_sym_this] = ACTIONS(1366),
    [anon_sym_do] = ACTIONS(1366),
    [anon_sym_while] = ACTIONS(1366),
    [anon_sym_for] = ACTIONS(1366),
    [aux_sym__type_identifier_token1] = ACTIONS(1366),
    [anon_sym_if] = ACTIONS(1366),
    [anon_sym_else] = ACTIONS(1366),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [aux_sym_number_literal_token1] = ACTIONS(1366),
    [aux_sym_number_literal_token2] = ACTIONS(1368),
    [aux_sym_number_literal_token3] = ACTIONS(1368),
    [aux_sym_number_literal_token4] = ACTIONS(1366),
    [aux_sym_number_literal_token5] = ACTIONS(1368),
    [anon_sym_return] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1368),
    [anon_sym_DQUOTE] = ACTIONS(1366),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1368),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1368),
    [anon_sym_switch] = ACTIONS(1366),
    [anon_sym_case] = ACTIONS(1366),
    [anon_sym_default] = ACTIONS(1366),
    [anon_sym_QMARK] = ACTIONS(1366),
    [anon_sym_try] = ACTIONS(1366),
    [anon_sym_int] = ACTIONS(1366),
    [anon_sym_boolean] = ACTIONS(1366),
    [anon_sym_char] = ACTIONS(1366),
    [anon_sym_short] = ACTIONS(1366),
    [anon_sym_long] = ACTIONS(1366),
    [anon_sym_float] = ACTIONS(1366),
    [anon_sym_double] = ACTIONS(1366),
    [anon_sym_void] = ACTIONS(1366),
    [anon_sym_public] = ACTIONS(1366),
    [anon_sym_protected] = ACTIONS(1366),
    [anon_sym_private] = ACTIONS(1366),
    [anon_sym_static] = ACTIONS(1366),
    [anon_sym_final] = ACTIONS(1366),
    [anon_sym_synchronized] = ACTIONS(1366),
    [anon_sym_TILDE] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1366),
    [anon_sym_new] = ACTIONS(1366),
  },
  [STATE(548)] = {
    [sym_comment] = STATE(548),
    [sym_groovy_doc] = STATE(548),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(549)] = {
    [sym_comment] = STATE(549),
    [sym_groovy_doc] = STATE(549),
    [ts_builtin_sym_end] = ACTIONS(1620),
    [sym_identifier] = ACTIONS(1618),
    [anon_sym_break] = ACTIONS(1618),
    [anon_sym_continue] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1620),
    [anon_sym_import] = ACTIONS(1618),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_assert] = ACTIONS(1618),
    [anon_sym_PLUS_PLUS] = ACTIONS(1620),
    [anon_sym_DASH_DASH] = ACTIONS(1620),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_ATinterface] = ACTIONS(1620),
    [anon_sym_interface] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_def] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_null] = ACTIONS(1618),
    [anon_sym_this] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_while] = ACTIONS(1618),
    [anon_sym_for] = ACTIONS(1618),
    [aux_sym__type_identifier_token1] = ACTIONS(1618),
    [anon_sym_if] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1620),
    [aux_sym_number_literal_token1] = ACTIONS(1618),
    [aux_sym_number_literal_token2] = ACTIONS(1620),
    [aux_sym_number_literal_token3] = ACTIONS(1620),
    [aux_sym_number_literal_token4] = ACTIONS(1618),
    [aux_sym_number_literal_token5] = ACTIONS(1620),
    [anon_sym_pipeline] = ACTIONS(1618),
    [anon_sym_return] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1620),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1620),
    [anon_sym_switch] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1618),
    [anon_sym_catch] = ACTIONS(1618),
    [anon_sym_finally] = ACTIONS(1618),
    [anon_sym_int] = ACTIONS(1618),
    [anon_sym_boolean] = ACTIONS(1618),
    [anon_sym_char] = ACTIONS(1618),
    [anon_sym_short] = ACTIONS(1618),
    [anon_sym_long] = ACTIONS(1618),
    [anon_sym_float] = ACTIONS(1618),
    [anon_sym_double] = ACTIONS(1618),
    [anon_sym_void] = ACTIONS(1618),
    [anon_sym_public] = ACTIONS(1618),
    [anon_sym_protected] = ACTIONS(1618),
    [anon_sym_private] = ACTIONS(1618),
    [anon_sym_static] = ACTIONS(1618),
    [anon_sym_final] = ACTIONS(1618),
    [anon_sym_synchronized] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1620),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
  },
  [STATE(550)] = {
    [sym_comment] = STATE(550),
    [sym_groovy_doc] = STATE(550),
    [ts_builtin_sym_end] = ACTIONS(1624),
    [sym_identifier] = ACTIONS(1622),
    [anon_sym_break] = ACTIONS(1622),
    [anon_sym_continue] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1624),
    [anon_sym_import] = ACTIONS(1622),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_assert] = ACTIONS(1622),
    [anon_sym_PLUS_PLUS] = ACTIONS(1624),
    [anon_sym_DASH_DASH] = ACTIONS(1624),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_ATinterface] = ACTIONS(1624),
    [anon_sym_interface] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_def] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_null] = ACTIONS(1622),
    [anon_sym_this] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_while] = ACTIONS(1622),
    [anon_sym_for] = ACTIONS(1622),
    [aux_sym__type_identifier_token1] = ACTIONS(1622),
    [anon_sym_if] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1624),
    [aux_sym_number_literal_token1] = ACTIONS(1622),
    [aux_sym_number_literal_token2] = ACTIONS(1624),
    [aux_sym_number_literal_token3] = ACTIONS(1624),
    [aux_sym_number_literal_token4] = ACTIONS(1622),
    [aux_sym_number_literal_token5] = ACTIONS(1624),
    [anon_sym_pipeline] = ACTIONS(1622),
    [anon_sym_return] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1624),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1624),
    [anon_sym_switch] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1622),
    [anon_sym_catch] = ACTIONS(1622),
    [anon_sym_finally] = ACTIONS(1622),
    [anon_sym_int] = ACTIONS(1622),
    [anon_sym_boolean] = ACTIONS(1622),
    [anon_sym_char] = ACTIONS(1622),
    [anon_sym_short] = ACTIONS(1622),
    [anon_sym_long] = ACTIONS(1622),
    [anon_sym_float] = ACTIONS(1622),
    [anon_sym_double] = ACTIONS(1622),
    [anon_sym_void] = ACTIONS(1622),
    [anon_sym_public] = ACTIONS(1622),
    [anon_sym_protected] = ACTIONS(1622),
    [anon_sym_private] = ACTIONS(1622),
    [anon_sym_static] = ACTIONS(1622),
    [anon_sym_final] = ACTIONS(1622),
    [anon_sym_synchronized] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1624),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
  },
  [STATE(551)] = {
    [sym_comment] = STATE(551),
    [sym_groovy_doc] = STATE(551),
    [ts_builtin_sym_end] = ACTIONS(1511),
    [sym_identifier] = ACTIONS(1509),
    [anon_sym_break] = ACTIONS(1509),
    [anon_sym_continue] = ACTIONS(1509),
    [anon_sym_SEMI] = ACTIONS(1511),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1511),
    [anon_sym_import] = ACTIONS(1509),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1509),
    [anon_sym_assert] = ACTIONS(1509),
    [anon_sym_PLUS_PLUS] = ACTIONS(1511),
    [anon_sym_DASH_DASH] = ACTIONS(1511),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1509),
    [anon_sym_false] = ACTIONS(1509),
    [anon_sym_ATinterface] = ACTIONS(1511),
    [anon_sym_interface] = ACTIONS(1509),
    [anon_sym_class] = ACTIONS(1509),
    [anon_sym_COMMA] = ACTIONS(1511),
    [anon_sym_LBRACE] = ACTIONS(1511),
    [anon_sym_DASH_GT] = ACTIONS(1511),
    [anon_sym_RBRACE] = ACTIONS(1511),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1509),
    [anon_sym_def] = ACTIONS(1509),
    [anon_sym_LPAREN] = ACTIONS(1511),
    [anon_sym_null] = ACTIONS(1509),
    [anon_sym_this] = ACTIONS(1509),
    [anon_sym_do] = ACTIONS(1509),
    [anon_sym_while] = ACTIONS(1509),
    [anon_sym_for] = ACTIONS(1509),
    [aux_sym__type_identifier_token1] = ACTIONS(1509),
    [anon_sym_if] = ACTIONS(1509),
    [anon_sym_else] = ACTIONS(1509),
    [anon_sym_LBRACK] = ACTIONS(1511),
    [aux_sym_number_literal_token1] = ACTIONS(1509),
    [aux_sym_number_literal_token2] = ACTIONS(1511),
    [aux_sym_number_literal_token3] = ACTIONS(1511),
    [aux_sym_number_literal_token4] = ACTIONS(1509),
    [aux_sym_number_literal_token5] = ACTIONS(1511),
    [anon_sym_pipeline] = ACTIONS(1509),
    [anon_sym_return] = ACTIONS(1509),
    [anon_sym_SQUOTE] = ACTIONS(1509),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1511),
    [anon_sym_DQUOTE] = ACTIONS(1509),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1511),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1511),
    [anon_sym_switch] = ACTIONS(1509),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1509),
    [anon_sym_int] = ACTIONS(1509),
    [anon_sym_boolean] = ACTIONS(1509),
    [anon_sym_char] = ACTIONS(1509),
    [anon_sym_short] = ACTIONS(1509),
    [anon_sym_long] = ACTIONS(1509),
    [anon_sym_float] = ACTIONS(1509),
    [anon_sym_double] = ACTIONS(1509),
    [anon_sym_void] = ACTIONS(1509),
    [anon_sym_public] = ACTIONS(1509),
    [anon_sym_protected] = ACTIONS(1509),
    [anon_sym_private] = ACTIONS(1509),
    [anon_sym_static] = ACTIONS(1509),
    [anon_sym_final] = ACTIONS(1509),
    [anon_sym_synchronized] = ACTIONS(1509),
    [anon_sym_TILDE] = ACTIONS(1511),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_new] = ACTIONS(1509),
  },
  [STATE(552)] = {
    [sym_comment] = STATE(552),
    [sym_groovy_doc] = STATE(552),
    [ts_builtin_sym_end] = ACTIONS(1628),
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_break] = ACTIONS(1626),
    [anon_sym_continue] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1628),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_PLUS_PLUS] = ACTIONS(1628),
    [anon_sym_DASH_DASH] = ACTIONS(1628),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_ATinterface] = ACTIONS(1628),
    [anon_sym_interface] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_null] = ACTIONS(1626),
    [anon_sym_this] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [aux_sym__type_identifier_token1] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1628),
    [aux_sym_number_literal_token1] = ACTIONS(1626),
    [aux_sym_number_literal_token2] = ACTIONS(1628),
    [aux_sym_number_literal_token3] = ACTIONS(1628),
    [aux_sym_number_literal_token4] = ACTIONS(1626),
    [aux_sym_number_literal_token5] = ACTIONS(1628),
    [anon_sym_pipeline] = ACTIONS(1626),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1628),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1628),
    [anon_sym_switch] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_catch] = ACTIONS(1626),
    [anon_sym_finally] = ACTIONS(1626),
    [anon_sym_int] = ACTIONS(1626),
    [anon_sym_boolean] = ACTIONS(1626),
    [anon_sym_char] = ACTIONS(1626),
    [anon_sym_short] = ACTIONS(1626),
    [anon_sym_long] = ACTIONS(1626),
    [anon_sym_float] = ACTIONS(1626),
    [anon_sym_double] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_public] = ACTIONS(1626),
    [anon_sym_protected] = ACTIONS(1626),
    [anon_sym_private] = ACTIONS(1626),
    [anon_sym_static] = ACTIONS(1626),
    [anon_sym_final] = ACTIONS(1626),
    [anon_sym_synchronized] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1628),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
  },
  [STATE(553)] = {
    [sym_comment] = STATE(553),
    [sym_groovy_doc] = STATE(553),
    [ts_builtin_sym_end] = ACTIONS(1632),
    [sym_identifier] = ACTIONS(1630),
    [anon_sym_break] = ACTIONS(1630),
    [anon_sym_continue] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1632),
    [anon_sym_import] = ACTIONS(1630),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_assert] = ACTIONS(1630),
    [anon_sym_PLUS_PLUS] = ACTIONS(1632),
    [anon_sym_DASH_DASH] = ACTIONS(1632),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_ATinterface] = ACTIONS(1632),
    [anon_sym_interface] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_def] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_null] = ACTIONS(1630),
    [anon_sym_this] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_while] = ACTIONS(1630),
    [anon_sym_for] = ACTIONS(1630),
    [aux_sym__type_identifier_token1] = ACTIONS(1630),
    [anon_sym_if] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1632),
    [aux_sym_number_literal_token1] = ACTIONS(1630),
    [aux_sym_number_literal_token2] = ACTIONS(1632),
    [aux_sym_number_literal_token3] = ACTIONS(1632),
    [aux_sym_number_literal_token4] = ACTIONS(1630),
    [aux_sym_number_literal_token5] = ACTIONS(1632),
    [anon_sym_pipeline] = ACTIONS(1630),
    [anon_sym_return] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1632),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1632),
    [anon_sym_switch] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1630),
    [anon_sym_catch] = ACTIONS(1630),
    [anon_sym_finally] = ACTIONS(1630),
    [anon_sym_int] = ACTIONS(1630),
    [anon_sym_boolean] = ACTIONS(1630),
    [anon_sym_char] = ACTIONS(1630),
    [anon_sym_short] = ACTIONS(1630),
    [anon_sym_long] = ACTIONS(1630),
    [anon_sym_float] = ACTIONS(1630),
    [anon_sym_double] = ACTIONS(1630),
    [anon_sym_void] = ACTIONS(1630),
    [anon_sym_public] = ACTIONS(1630),
    [anon_sym_protected] = ACTIONS(1630),
    [anon_sym_private] = ACTIONS(1630),
    [anon_sym_static] = ACTIONS(1630),
    [anon_sym_final] = ACTIONS(1630),
    [anon_sym_synchronized] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1632),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
  },
  [STATE(554)] = {
    [sym_comment] = STATE(554),
    [sym_groovy_doc] = STATE(554),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(555)] = {
    [sym_comment] = STATE(555),
    [sym_groovy_doc] = STATE(555),
    [ts_builtin_sym_end] = ACTIONS(1636),
    [sym_identifier] = ACTIONS(1634),
    [anon_sym_break] = ACTIONS(1634),
    [anon_sym_continue] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1636),
    [anon_sym_import] = ACTIONS(1634),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1634),
    [anon_sym_AT] = ACTIONS(1634),
    [anon_sym_assert] = ACTIONS(1634),
    [anon_sym_PLUS_PLUS] = ACTIONS(1636),
    [anon_sym_DASH_DASH] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_ATinterface] = ACTIONS(1636),
    [anon_sym_interface] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1634),
    [anon_sym_def] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_null] = ACTIONS(1634),
    [anon_sym_this] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_while] = ACTIONS(1634),
    [anon_sym_for] = ACTIONS(1634),
    [aux_sym__type_identifier_token1] = ACTIONS(1634),
    [anon_sym_if] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1636),
    [aux_sym_number_literal_token1] = ACTIONS(1634),
    [aux_sym_number_literal_token2] = ACTIONS(1636),
    [aux_sym_number_literal_token3] = ACTIONS(1636),
    [aux_sym_number_literal_token4] = ACTIONS(1634),
    [aux_sym_number_literal_token5] = ACTIONS(1636),
    [anon_sym_pipeline] = ACTIONS(1634),
    [anon_sym_return] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1634),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1634),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1636),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1636),
    [anon_sym_switch] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1634),
    [anon_sym_catch] = ACTIONS(1634),
    [anon_sym_finally] = ACTIONS(1634),
    [anon_sym_int] = ACTIONS(1634),
    [anon_sym_boolean] = ACTIONS(1634),
    [anon_sym_char] = ACTIONS(1634),
    [anon_sym_short] = ACTIONS(1634),
    [anon_sym_long] = ACTIONS(1634),
    [anon_sym_float] = ACTIONS(1634),
    [anon_sym_double] = ACTIONS(1634),
    [anon_sym_void] = ACTIONS(1634),
    [anon_sym_public] = ACTIONS(1634),
    [anon_sym_protected] = ACTIONS(1634),
    [anon_sym_private] = ACTIONS(1634),
    [anon_sym_static] = ACTIONS(1634),
    [anon_sym_final] = ACTIONS(1634),
    [anon_sym_synchronized] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
  },
  [STATE(556)] = {
    [sym_comment] = STATE(556),
    [sym_groovy_doc] = STATE(556),
    [ts_builtin_sym_end] = ACTIONS(1640),
    [sym_identifier] = ACTIONS(1638),
    [anon_sym_break] = ACTIONS(1638),
    [anon_sym_continue] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_DOT_AMP] = ACTIONS(1524),
    [anon_sym_DOT_AT] = ACTIONS(1524),
    [anon_sym_QMARK_DOT] = ACTIONS(1524),
    [anon_sym_STAR_DOT] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_COLON] = ACTIONS(1640),
    [anon_sym_import] = ACTIONS(1638),
    [anon_sym_as] = ACTIONS(1542),
    [anon_sym_package] = ACTIONS(1638),
    [anon_sym_AT] = ACTIONS(1638),
    [anon_sym_assert] = ACTIONS(1638),
    [anon_sym_PLUS_PLUS] = ACTIONS(1640),
    [anon_sym_DASH_DASH] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1534),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1538),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1538),
    [anon_sym_DOT_DOT] = ACTIONS(1540),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1538),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1542),
    [anon_sym_LT_EQ] = ACTIONS(1542),
    [anon_sym_GT] = ACTIONS(1542),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1542),
    [anon_sym_BANGin] = ACTIONS(1542),
    [anon_sym_instanceof] = ACTIONS(1542),
    [anon_sym_BANGinstanceof] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1546),
    [anon_sym_BANG_EQ] = ACTIONS(1546),
    [anon_sym_LT_EQ_GT] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_CARET] = ACTIONS(1552),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_AMP_AMP] = ACTIONS(1556),
    [anon_sym_PIPE_PIPE] = ACTIONS(1558),
    [anon_sym_QMARK_COLON] = ACTIONS(1560),
    [anon_sym_STAR_STAR] = ACTIONS(1526),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_ATinterface] = ACTIONS(1640),
    [anon_sym_interface] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1638),
    [anon_sym_def] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_null] = ACTIONS(1638),
    [anon_sym_this] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_while] = ACTIONS(1638),
    [anon_sym_for] = ACTIONS(1638),
    [aux_sym__type_identifier_token1] = ACTIONS(1638),
    [anon_sym_if] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1640),
    [aux_sym_number_literal_token1] = ACTIONS(1638),
    [aux_sym_number_literal_token2] = ACTIONS(1640),
    [aux_sym_number_literal_token3] = ACTIONS(1640),
    [aux_sym_number_literal_token4] = ACTIONS(1638),
    [aux_sym_number_literal_token5] = ACTIONS(1640),
    [anon_sym_pipeline] = ACTIONS(1638),
    [anon_sym_return] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1638),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1638),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1640),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1640),
    [anon_sym_switch] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1638),
    [anon_sym_catch] = ACTIONS(1638),
    [anon_sym_finally] = ACTIONS(1638),
    [anon_sym_int] = ACTIONS(1638),
    [anon_sym_boolean] = ACTIONS(1638),
    [anon_sym_char] = ACTIONS(1638),
    [anon_sym_short] = ACTIONS(1638),
    [anon_sym_long] = ACTIONS(1638),
    [anon_sym_float] = ACTIONS(1638),
    [anon_sym_double] = ACTIONS(1638),
    [anon_sym_void] = ACTIONS(1638),
    [anon_sym_public] = ACTIONS(1638),
    [anon_sym_protected] = ACTIONS(1638),
    [anon_sym_private] = ACTIONS(1638),
    [anon_sym_static] = ACTIONS(1638),
    [anon_sym_final] = ACTIONS(1638),
    [anon_sym_synchronized] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
  },
  [STATE(557)] = {
    [sym_comment] = STATE(557),
    [sym_groovy_doc] = STATE(557),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_DOT_AMP] = ACTIONS(1336),
    [anon_sym_DOT_AT] = ACTIONS(1336),
    [anon_sym_QMARK_DOT] = ACTIONS(1336),
    [anon_sym_STAR_DOT] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1338),
    [anon_sym_STAR_COLON] = ACTIONS(1336),
    [anon_sym_DOT] = ACTIONS(1338),
    [anon_sym_import] = ACTIONS(1338),
    [anon_sym_as] = ACTIONS(1338),
    [anon_sym_package] = ACTIONS(1338),
    [anon_sym_AT] = ACTIONS(1338),
    [anon_sym_assert] = ACTIONS(1338),
    [anon_sym_PLUS_PLUS] = ACTIONS(1336),
    [anon_sym_DASH_DASH] = ACTIONS(1336),
    [anon_sym_PERCENT] = ACTIONS(1336),
    [anon_sym_SLASH] = ACTIONS(1338),
    [anon_sym_PLUS] = ACTIONS(1338),
    [anon_sym_DASH] = ACTIONS(1338),
    [anon_sym_LT_LT] = ACTIONS(1336),
    [anon_sym_GT_GT] = ACTIONS(1338),
    [anon_sym_GT_GT_GT] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1338),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1336),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1338),
    [anon_sym_LT_EQ] = ACTIONS(1338),
    [anon_sym_GT] = ACTIONS(1338),
    [anon_sym_GT_EQ] = ACTIONS(1336),
    [anon_sym_in] = ACTIONS(1338),
    [anon_sym_BANGin] = ACTIONS(1338),
    [anon_sym_instanceof] = ACTIONS(1338),
    [anon_sym_BANGinstanceof] = ACTIONS(1336),
    [anon_sym_EQ_EQ] = ACTIONS(1338),
    [anon_sym_BANG_EQ] = ACTIONS(1338),
    [anon_sym_LT_EQ_GT] = ACTIONS(1336),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1336),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1336),
    [anon_sym_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1338),
    [anon_sym_CARET] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1338),
    [anon_sym_AMP_AMP] = ACTIONS(1336),
    [anon_sym_PIPE_PIPE] = ACTIONS(1336),
    [anon_sym_QMARK_COLON] = ACTIONS(1336),
    [anon_sym_STAR_STAR] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [anon_sym_ATinterface] = ACTIONS(1336),
    [anon_sym_interface] = ACTIONS(1338),
    [anon_sym_class] = ACTIONS(1338),
    [anon_sym_COMMA] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1338),
    [anon_sym_def] = ACTIONS(1338),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_null] = ACTIONS(1338),
    [anon_sym_this] = ACTIONS(1338),
    [anon_sym_do] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [aux_sym__type_identifier_token1] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_else] = ACTIONS(1338),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [aux_sym_number_literal_token1] = ACTIONS(1338),
    [aux_sym_number_literal_token2] = ACTIONS(1336),
    [aux_sym_number_literal_token3] = ACTIONS(1336),
    [aux_sym_number_literal_token4] = ACTIONS(1338),
    [aux_sym_number_literal_token5] = ACTIONS(1336),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1336),
    [anon_sym_DQUOTE] = ACTIONS(1338),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1336),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1336),
    [anon_sym_switch] = ACTIONS(1338),
    [anon_sym_case] = ACTIONS(1338),
    [anon_sym_default] = ACTIONS(1338),
    [anon_sym_QMARK] = ACTIONS(1338),
    [anon_sym_try] = ACTIONS(1338),
    [anon_sym_int] = ACTIONS(1338),
    [anon_sym_boolean] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_short] = ACTIONS(1338),
    [anon_sym_long] = ACTIONS(1338),
    [anon_sym_float] = ACTIONS(1338),
    [anon_sym_double] = ACTIONS(1338),
    [anon_sym_void] = ACTIONS(1338),
    [anon_sym_public] = ACTIONS(1338),
    [anon_sym_protected] = ACTIONS(1338),
    [anon_sym_private] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_final] = ACTIONS(1338),
    [anon_sym_synchronized] = ACTIONS(1338),
    [anon_sym_TILDE] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1338),
    [anon_sym_new] = ACTIONS(1338),
  },
  [STATE(558)] = {
    [sym_comment] = STATE(558),
    [sym_groovy_doc] = STATE(558),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1254),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(559)] = {
    [sym_comment] = STATE(559),
    [sym_groovy_doc] = STATE(559),
    [aux_sym_dotted_identifier_repeat1] = STATE(27),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1564),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1528),
    [anon_sym_DASH_DASH] = ACTIONS(1528),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1646),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(560)] = {
    [sym_comment] = STATE(560),
    [sym_groovy_doc] = STATE(560),
    [sym_identifier] = ACTIONS(1604),
    [anon_sym_break] = ACTIONS(1604),
    [anon_sym_continue] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_DOT_AMP] = ACTIONS(1459),
    [anon_sym_DOT_AT] = ACTIONS(1459),
    [anon_sym_QMARK_DOT] = ACTIONS(1459),
    [anon_sym_STAR_DOT] = ACTIONS(1459),
    [anon_sym_STAR] = ACTIONS(1471),
    [anon_sym_STAR_COLON] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1604),
    [anon_sym_as] = ACTIONS(1481),
    [anon_sym_package] = ACTIONS(1604),
    [anon_sym_AT] = ACTIONS(1604),
    [anon_sym_assert] = ACTIONS(1604),
    [anon_sym_PLUS_PLUS] = ACTIONS(1602),
    [anon_sym_DASH_DASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1473),
    [anon_sym_SLASH] = ACTIONS(1471),
    [anon_sym_PLUS] = ACTIONS(1475),
    [anon_sym_DASH] = ACTIONS(1475),
    [anon_sym_LT_LT] = ACTIONS(1477),
    [anon_sym_GT_GT] = ACTIONS(1479),
    [anon_sym_GT_GT_GT] = ACTIONS(1477),
    [anon_sym_DOT_DOT] = ACTIONS(1479),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1477),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1479),
    [anon_sym_LT] = ACTIONS(1481),
    [anon_sym_LT_EQ] = ACTIONS(1481),
    [anon_sym_GT] = ACTIONS(1481),
    [anon_sym_GT_EQ] = ACTIONS(1483),
    [anon_sym_in] = ACTIONS(1481),
    [anon_sym_BANGin] = ACTIONS(1481),
    [anon_sym_instanceof] = ACTIONS(1481),
    [anon_sym_BANGinstanceof] = ACTIONS(1483),
    [anon_sym_EQ_EQ] = ACTIONS(1485),
    [anon_sym_BANG_EQ] = ACTIONS(1485),
    [anon_sym_LT_EQ_GT] = ACTIONS(1487),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1487),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1487),
    [anon_sym_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1487),
    [anon_sym_AMP] = ACTIONS(1489),
    [anon_sym_CARET] = ACTIONS(1491),
    [anon_sym_PIPE] = ACTIONS(1493),
    [anon_sym_AMP_AMP] = ACTIONS(1495),
    [anon_sym_PIPE_PIPE] = ACTIONS(1497),
    [anon_sym_QMARK_COLON] = ACTIONS(1503),
    [anon_sym_STAR_STAR] = ACTIONS(1461),
    [anon_sym_true] = ACTIONS(1604),
    [anon_sym_false] = ACTIONS(1604),
    [anon_sym_ATinterface] = ACTIONS(1602),
    [anon_sym_interface] = ACTIONS(1604),
    [anon_sym_class] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1604),
    [anon_sym_def] = ACTIONS(1604),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_null] = ACTIONS(1604),
    [anon_sym_this] = ACTIONS(1604),
    [anon_sym_do] = ACTIONS(1604),
    [anon_sym_while] = ACTIONS(1604),
    [anon_sym_for] = ACTIONS(1604),
    [aux_sym__type_identifier_token1] = ACTIONS(1604),
    [anon_sym_if] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [aux_sym_number_literal_token1] = ACTIONS(1604),
    [aux_sym_number_literal_token2] = ACTIONS(1602),
    [aux_sym_number_literal_token3] = ACTIONS(1602),
    [aux_sym_number_literal_token4] = ACTIONS(1604),
    [aux_sym_number_literal_token5] = ACTIONS(1602),
    [anon_sym_return] = ACTIONS(1604),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1602),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1602),
    [anon_sym_switch] = ACTIONS(1604),
    [anon_sym_case] = ACTIONS(1604),
    [anon_sym_default] = ACTIONS(1604),
    [anon_sym_QMARK] = ACTIONS(1505),
    [anon_sym_try] = ACTIONS(1604),
    [anon_sym_catch] = ACTIONS(1604),
    [anon_sym_finally] = ACTIONS(1604),
    [anon_sym_int] = ACTIONS(1604),
    [anon_sym_boolean] = ACTIONS(1604),
    [anon_sym_char] = ACTIONS(1604),
    [anon_sym_short] = ACTIONS(1604),
    [anon_sym_long] = ACTIONS(1604),
    [anon_sym_float] = ACTIONS(1604),
    [anon_sym_double] = ACTIONS(1604),
    [anon_sym_void] = ACTIONS(1604),
    [anon_sym_public] = ACTIONS(1604),
    [anon_sym_protected] = ACTIONS(1604),
    [anon_sym_private] = ACTIONS(1604),
    [anon_sym_static] = ACTIONS(1604),
    [anon_sym_final] = ACTIONS(1604),
    [anon_sym_synchronized] = ACTIONS(1604),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1604),
    [anon_sym_new] = ACTIONS(1604),
  },
  [STATE(561)] = {
    [sym_comment] = STATE(561),
    [sym_groovy_doc] = STATE(561),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_DOT_AMP] = ACTIONS(1304),
    [anon_sym_DOT_AT] = ACTIONS(1304),
    [anon_sym_QMARK_DOT] = ACTIONS(1304),
    [anon_sym_STAR_DOT] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1306),
    [anon_sym_STAR_COLON] = ACTIONS(1304),
    [anon_sym_DOT] = ACTIONS(1306),
    [anon_sym_import] = ACTIONS(1306),
    [anon_sym_as] = ACTIONS(1306),
    [anon_sym_package] = ACTIONS(1306),
    [anon_sym_AT] = ACTIONS(1306),
    [anon_sym_assert] = ACTIONS(1306),
    [anon_sym_PLUS_PLUS] = ACTIONS(1304),
    [anon_sym_DASH_DASH] = ACTIONS(1304),
    [anon_sym_PERCENT] = ACTIONS(1304),
    [anon_sym_SLASH] = ACTIONS(1306),
    [anon_sym_PLUS] = ACTIONS(1306),
    [anon_sym_DASH] = ACTIONS(1306),
    [anon_sym_LT_LT] = ACTIONS(1304),
    [anon_sym_GT_GT] = ACTIONS(1306),
    [anon_sym_GT_GT_GT] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1306),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1304),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1306),
    [anon_sym_LT_EQ] = ACTIONS(1306),
    [anon_sym_GT] = ACTIONS(1306),
    [anon_sym_GT_EQ] = ACTIONS(1304),
    [anon_sym_in] = ACTIONS(1306),
    [anon_sym_BANGin] = ACTIONS(1306),
    [anon_sym_instanceof] = ACTIONS(1306),
    [anon_sym_BANGinstanceof] = ACTIONS(1304),
    [anon_sym_EQ_EQ] = ACTIONS(1306),
    [anon_sym_BANG_EQ] = ACTIONS(1306),
    [anon_sym_LT_EQ_GT] = ACTIONS(1304),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1304),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1304),
    [anon_sym_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1306),
    [anon_sym_CARET] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1306),
    [anon_sym_AMP_AMP] = ACTIONS(1304),
    [anon_sym_PIPE_PIPE] = ACTIONS(1304),
    [anon_sym_QMARK_COLON] = ACTIONS(1304),
    [anon_sym_STAR_STAR] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [anon_sym_ATinterface] = ACTIONS(1304),
    [anon_sym_interface] = ACTIONS(1306),
    [anon_sym_class] = ACTIONS(1306),
    [anon_sym_COMMA] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1306),
    [anon_sym_def] = ACTIONS(1306),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_null] = ACTIONS(1306),
    [anon_sym_this] = ACTIONS(1306),
    [anon_sym_do] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [aux_sym__type_identifier_token1] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_else] = ACTIONS(1306),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [aux_sym_number_literal_token1] = ACTIONS(1306),
    [aux_sym_number_literal_token2] = ACTIONS(1304),
    [aux_sym_number_literal_token3] = ACTIONS(1304),
    [aux_sym_number_literal_token4] = ACTIONS(1306),
    [aux_sym_number_literal_token5] = ACTIONS(1304),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1304),
    [anon_sym_DQUOTE] = ACTIONS(1306),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1304),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1304),
    [anon_sym_switch] = ACTIONS(1306),
    [anon_sym_case] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_QMARK] = ACTIONS(1306),
    [anon_sym_try] = ACTIONS(1306),
    [anon_sym_int] = ACTIONS(1306),
    [anon_sym_boolean] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_short] = ACTIONS(1306),
    [anon_sym_long] = ACTIONS(1306),
    [anon_sym_float] = ACTIONS(1306),
    [anon_sym_double] = ACTIONS(1306),
    [anon_sym_void] = ACTIONS(1306),
    [anon_sym_public] = ACTIONS(1306),
    [anon_sym_protected] = ACTIONS(1306),
    [anon_sym_private] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_final] = ACTIONS(1306),
    [anon_sym_synchronized] = ACTIONS(1306),
    [anon_sym_TILDE] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1306),
    [anon_sym_new] = ACTIONS(1306),
  },
  [STATE(562)] = {
    [sym_comment] = STATE(562),
    [sym_groovy_doc] = STATE(562),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_DOT_AMP] = ACTIONS(1332),
    [anon_sym_DOT_AT] = ACTIONS(1332),
    [anon_sym_QMARK_DOT] = ACTIONS(1332),
    [anon_sym_STAR_DOT] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1334),
    [anon_sym_STAR_COLON] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1334),
    [anon_sym_import] = ACTIONS(1334),
    [anon_sym_as] = ACTIONS(1334),
    [anon_sym_package] = ACTIONS(1334),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_assert] = ACTIONS(1334),
    [anon_sym_PLUS_PLUS] = ACTIONS(1332),
    [anon_sym_DASH_DASH] = ACTIONS(1332),
    [anon_sym_PERCENT] = ACTIONS(1332),
    [anon_sym_SLASH] = ACTIONS(1334),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_LT_LT] = ACTIONS(1332),
    [anon_sym_GT_GT] = ACTIONS(1334),
    [anon_sym_GT_GT_GT] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1334),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1332),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1334),
    [anon_sym_LT_EQ] = ACTIONS(1334),
    [anon_sym_GT] = ACTIONS(1334),
    [anon_sym_GT_EQ] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1334),
    [anon_sym_BANGin] = ACTIONS(1334),
    [anon_sym_instanceof] = ACTIONS(1334),
    [anon_sym_BANGinstanceof] = ACTIONS(1332),
    [anon_sym_EQ_EQ] = ACTIONS(1334),
    [anon_sym_BANG_EQ] = ACTIONS(1334),
    [anon_sym_LT_EQ_GT] = ACTIONS(1332),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1332),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1332),
    [anon_sym_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1334),
    [anon_sym_CARET] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1334),
    [anon_sym_AMP_AMP] = ACTIONS(1332),
    [anon_sym_PIPE_PIPE] = ACTIONS(1332),
    [anon_sym_QMARK_COLON] = ACTIONS(1332),
    [anon_sym_STAR_STAR] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [anon_sym_ATinterface] = ACTIONS(1332),
    [anon_sym_interface] = ACTIONS(1334),
    [anon_sym_class] = ACTIONS(1334),
    [anon_sym_COMMA] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1334),
    [anon_sym_def] = ACTIONS(1334),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_null] = ACTIONS(1334),
    [anon_sym_this] = ACTIONS(1334),
    [anon_sym_do] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [aux_sym__type_identifier_token1] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_else] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [aux_sym_number_literal_token1] = ACTIONS(1334),
    [aux_sym_number_literal_token2] = ACTIONS(1332),
    [aux_sym_number_literal_token3] = ACTIONS(1332),
    [aux_sym_number_literal_token4] = ACTIONS(1334),
    [aux_sym_number_literal_token5] = ACTIONS(1332),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1332),
    [anon_sym_DQUOTE] = ACTIONS(1334),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1332),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1332),
    [anon_sym_switch] = ACTIONS(1334),
    [anon_sym_case] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_QMARK] = ACTIONS(1334),
    [anon_sym_try] = ACTIONS(1334),
    [anon_sym_int] = ACTIONS(1334),
    [anon_sym_boolean] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_short] = ACTIONS(1334),
    [anon_sym_long] = ACTIONS(1334),
    [anon_sym_float] = ACTIONS(1334),
    [anon_sym_double] = ACTIONS(1334),
    [anon_sym_void] = ACTIONS(1334),
    [anon_sym_public] = ACTIONS(1334),
    [anon_sym_protected] = ACTIONS(1334),
    [anon_sym_private] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_final] = ACTIONS(1334),
    [anon_sym_synchronized] = ACTIONS(1334),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1334),
    [anon_sym_new] = ACTIONS(1334),
  },
  [STATE(563)] = {
    [sym_comment] = STATE(563),
    [sym_groovy_doc] = STATE(563),
    [sym_identifier] = ACTIONS(102),
    [anon_sym_break] = ACTIONS(102),
    [anon_sym_continue] = ACTIONS(102),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_STAR_COLON] = ACTIONS(97),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(102),
    [anon_sym_AT] = ACTIONS(102),
    [anon_sym_assert] = ACTIONS(102),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(102),
    [anon_sym_false] = ACTIONS(102),
    [anon_sym_ATinterface] = ACTIONS(97),
    [anon_sym_interface] = ACTIONS(102),
    [anon_sym_class] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [anon_sym_LBRACE] = ACTIONS(97),
    [anon_sym_RBRACE] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(102),
    [anon_sym_def] = ACTIONS(102),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_null] = ACTIONS(102),
    [anon_sym_this] = ACTIONS(102),
    [anon_sym_do] = ACTIONS(102),
    [anon_sym_while] = ACTIONS(102),
    [anon_sym_for] = ACTIONS(102),
    [aux_sym__type_identifier_token1] = ACTIONS(102),
    [anon_sym_if] = ACTIONS(102),
    [anon_sym_else] = ACTIONS(102),
    [anon_sym_LBRACK] = ACTIONS(97),
    [aux_sym_number_literal_token1] = ACTIONS(102),
    [aux_sym_number_literal_token2] = ACTIONS(97),
    [aux_sym_number_literal_token3] = ACTIONS(97),
    [aux_sym_number_literal_token4] = ACTIONS(102),
    [aux_sym_number_literal_token5] = ACTIONS(97),
    [anon_sym_return] = ACTIONS(102),
    [anon_sym_SQUOTE] = ACTIONS(102),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(97),
    [anon_sym_DQUOTE] = ACTIONS(102),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(97),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(97),
    [anon_sym_switch] = ACTIONS(102),
    [anon_sym_case] = ACTIONS(102),
    [anon_sym_default] = ACTIONS(102),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(102),
    [anon_sym_int] = ACTIONS(102),
    [anon_sym_boolean] = ACTIONS(102),
    [anon_sym_char] = ACTIONS(102),
    [anon_sym_short] = ACTIONS(102),
    [anon_sym_long] = ACTIONS(102),
    [anon_sym_float] = ACTIONS(102),
    [anon_sym_double] = ACTIONS(102),
    [anon_sym_void] = ACTIONS(102),
    [anon_sym_public] = ACTIONS(102),
    [anon_sym_protected] = ACTIONS(102),
    [anon_sym_private] = ACTIONS(102),
    [anon_sym_static] = ACTIONS(102),
    [anon_sym_final] = ACTIONS(102),
    [anon_sym_synchronized] = ACTIONS(102),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_BANG] = ACTIONS(102),
    [anon_sym_new] = ACTIONS(102),
  },
  [STATE(564)] = {
    [sym_comment] = STATE(564),
    [sym_groovy_doc] = STATE(564),
    [sym_identifier] = ACTIONS(1430),
    [anon_sym_break] = ACTIONS(1430),
    [anon_sym_continue] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1428),
    [anon_sym_DOT_AMP] = ACTIONS(1428),
    [anon_sym_DOT_AT] = ACTIONS(1428),
    [anon_sym_QMARK_DOT] = ACTIONS(1428),
    [anon_sym_STAR_DOT] = ACTIONS(1428),
    [anon_sym_STAR] = ACTIONS(1430),
    [anon_sym_STAR_COLON] = ACTIONS(1428),
    [anon_sym_DOT] = ACTIONS(1430),
    [anon_sym_import] = ACTIONS(1430),
    [anon_sym_as] = ACTIONS(1430),
    [anon_sym_package] = ACTIONS(1430),
    [anon_sym_AT] = ACTIONS(1430),
    [anon_sym_assert] = ACTIONS(1430),
    [anon_sym_PLUS_PLUS] = ACTIONS(1428),
    [anon_sym_DASH_DASH] = ACTIONS(1428),
    [anon_sym_PERCENT] = ACTIONS(1428),
    [anon_sym_SLASH] = ACTIONS(1430),
    [anon_sym_PLUS] = ACTIONS(1430),
    [anon_sym_DASH] = ACTIONS(1430),
    [anon_sym_LT_LT] = ACTIONS(1428),
    [anon_sym_GT_GT] = ACTIONS(1430),
    [anon_sym_GT_GT_GT] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(1430),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1428),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1430),
    [anon_sym_LT] = ACTIONS(1430),
    [anon_sym_LT_EQ] = ACTIONS(1430),
    [anon_sym_GT] = ACTIONS(1430),
    [anon_sym_GT_EQ] = ACTIONS(1428),
    [anon_sym_in] = ACTIONS(1430),
    [anon_sym_BANGin] = ACTIONS(1430),
    [anon_sym_instanceof] = ACTIONS(1430),
    [anon_sym_BANGinstanceof] = ACTIONS(1428),
    [anon_sym_EQ_EQ] = ACTIONS(1430),
    [anon_sym_BANG_EQ] = ACTIONS(1430),
    [anon_sym_LT_EQ_GT] = ACTIONS(1428),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1428),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1428),
    [anon_sym_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1428),
    [anon_sym_AMP] = ACTIONS(1430),
    [anon_sym_CARET] = ACTIONS(1428),
    [anon_sym_PIPE] = ACTIONS(1430),
    [anon_sym_AMP_AMP] = ACTIONS(1428),
    [anon_sym_PIPE_PIPE] = ACTIONS(1428),
    [anon_sym_QMARK_COLON] = ACTIONS(1428),
    [anon_sym_STAR_STAR] = ACTIONS(1428),
    [anon_sym_true] = ACTIONS(1430),
    [anon_sym_false] = ACTIONS(1430),
    [anon_sym_ATinterface] = ACTIONS(1428),
    [anon_sym_interface] = ACTIONS(1430),
    [anon_sym_class] = ACTIONS(1430),
    [anon_sym_COMMA] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1430),
    [anon_sym_def] = ACTIONS(1430),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_null] = ACTIONS(1430),
    [anon_sym_this] = ACTIONS(1430),
    [anon_sym_do] = ACTIONS(1430),
    [anon_sym_while] = ACTIONS(1430),
    [anon_sym_for] = ACTIONS(1430),
    [aux_sym__type_identifier_token1] = ACTIONS(1430),
    [anon_sym_if] = ACTIONS(1430),
    [anon_sym_else] = ACTIONS(1430),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [aux_sym_number_literal_token1] = ACTIONS(1430),
    [aux_sym_number_literal_token2] = ACTIONS(1428),
    [aux_sym_number_literal_token3] = ACTIONS(1428),
    [aux_sym_number_literal_token4] = ACTIONS(1430),
    [aux_sym_number_literal_token5] = ACTIONS(1428),
    [anon_sym_return] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1428),
    [anon_sym_DQUOTE] = ACTIONS(1430),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1428),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1428),
    [anon_sym_switch] = ACTIONS(1430),
    [anon_sym_case] = ACTIONS(1430),
    [anon_sym_default] = ACTIONS(1430),
    [anon_sym_QMARK] = ACTIONS(1430),
    [anon_sym_try] = ACTIONS(1430),
    [anon_sym_int] = ACTIONS(1430),
    [anon_sym_boolean] = ACTIONS(1430),
    [anon_sym_char] = ACTIONS(1430),
    [anon_sym_short] = ACTIONS(1430),
    [anon_sym_long] = ACTIONS(1430),
    [anon_sym_float] = ACTIONS(1430),
    [anon_sym_double] = ACTIONS(1430),
    [anon_sym_void] = ACTIONS(1430),
    [anon_sym_public] = ACTIONS(1430),
    [anon_sym_protected] = ACTIONS(1430),
    [anon_sym_private] = ACTIONS(1430),
    [anon_sym_static] = ACTIONS(1430),
    [anon_sym_final] = ACTIONS(1430),
    [anon_sym_synchronized] = ACTIONS(1430),
    [anon_sym_TILDE] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1430),
    [anon_sym_new] = ACTIONS(1430),
  },
  [STATE(565)] = {
    [sym_comment] = STATE(565),
    [sym_groovy_doc] = STATE(565),
    [ts_builtin_sym_end] = ACTIONS(1457),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_DASH_GT] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_pipeline] = ACTIONS(1455),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(566)] = {
    [sym_comment] = STATE(566),
    [sym_groovy_doc] = STATE(566),
    [aux_sym_dotted_identifier_repeat1] = STATE(33),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_DOT] = ACTIONS(1507),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1436),
    [anon_sym_DASH_DASH] = ACTIONS(1436),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1648),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(567)] = {
    [sym_comment] = STATE(567),
    [sym_groovy_doc] = STATE(567),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1402),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1405),
    [anon_sym_DASH_DASH] = ACTIONS(1405),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1402),
    [anon_sym_PLUS] = ACTIONS(1402),
    [anon_sym_DASH] = ACTIONS(1402),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1405),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1405),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(568)] = {
    [sym_comment] = STATE(568),
    [sym_groovy_doc] = STATE(568),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1275),
    [anon_sym_DASH_DASH] = ACTIONS(1275),
    [anon_sym_PERCENT] = ACTIONS(363),
    [anon_sym_SLASH] = ACTIONS(1272),
    [anon_sym_PLUS] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_LT_LT] = ACTIONS(363),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(363),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(363),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(363),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1275),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1275),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(569)] = {
    [sym_comment] = STATE(569),
    [sym_groovy_doc] = STATE(569),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1260),
    [anon_sym_DOT_AT] = ACTIONS(1260),
    [anon_sym_QMARK_DOT] = ACTIONS(1260),
    [anon_sym_STAR_DOT] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1262),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_DOT] = ACTIONS(1262),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1262),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1260),
    [anon_sym_DASH_DASH] = ACTIONS(1260),
    [anon_sym_PERCENT] = ACTIONS(1260),
    [anon_sym_SLASH] = ACTIONS(1262),
    [anon_sym_PLUS] = ACTIONS(1262),
    [anon_sym_DASH] = ACTIONS(1262),
    [anon_sym_LT_LT] = ACTIONS(1260),
    [anon_sym_GT_GT] = ACTIONS(1262),
    [anon_sym_GT_GT_GT] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1262),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1260),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1262),
    [anon_sym_LT_EQ] = ACTIONS(1262),
    [anon_sym_GT] = ACTIONS(1262),
    [anon_sym_GT_EQ] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1262),
    [anon_sym_BANGin] = ACTIONS(1262),
    [anon_sym_instanceof] = ACTIONS(1262),
    [anon_sym_BANGinstanceof] = ACTIONS(1260),
    [anon_sym_EQ_EQ] = ACTIONS(1262),
    [anon_sym_BANG_EQ] = ACTIONS(1262),
    [anon_sym_LT_EQ_GT] = ACTIONS(1260),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1260),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1260),
    [anon_sym_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1262),
    [anon_sym_CARET] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1262),
    [anon_sym_AMP_AMP] = ACTIONS(1260),
    [anon_sym_PIPE_PIPE] = ACTIONS(1260),
    [anon_sym_QMARK_COLON] = ACTIONS(1260),
    [anon_sym_STAR_STAR] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1262),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(570)] = {
    [sym_comment] = STATE(570),
    [sym_groovy_doc] = STATE(570),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_DOT_AMP] = ACTIONS(1364),
    [anon_sym_DOT_AT] = ACTIONS(1364),
    [anon_sym_QMARK_DOT] = ACTIONS(1364),
    [anon_sym_STAR_DOT] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_DOT] = ACTIONS(1362),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_as] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_PERCENT] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_LT_LT] = ACTIONS(1364),
    [anon_sym_GT_GT] = ACTIONS(1362),
    [anon_sym_GT_GT_GT] = ACTIONS(1364),
    [anon_sym_DOT_DOT] = ACTIONS(1362),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1364),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1362),
    [anon_sym_LT_EQ] = ACTIONS(1362),
    [anon_sym_GT] = ACTIONS(1362),
    [anon_sym_GT_EQ] = ACTIONS(1364),
    [anon_sym_in] = ACTIONS(1362),
    [anon_sym_BANGin] = ACTIONS(1362),
    [anon_sym_instanceof] = ACTIONS(1362),
    [anon_sym_BANGinstanceof] = ACTIONS(1364),
    [anon_sym_EQ_EQ] = ACTIONS(1362),
    [anon_sym_BANG_EQ] = ACTIONS(1362),
    [anon_sym_LT_EQ_GT] = ACTIONS(1364),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1364),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1364),
    [anon_sym_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1364),
    [anon_sym_AMP] = ACTIONS(1362),
    [anon_sym_CARET] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1362),
    [anon_sym_AMP_AMP] = ACTIONS(1364),
    [anon_sym_PIPE_PIPE] = ACTIONS(1364),
    [anon_sym_QMARK_COLON] = ACTIONS(1364),
    [anon_sym_STAR_STAR] = ACTIONS(1364),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_COMMA] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_case] = ACTIONS(1362),
    [anon_sym_default] = ACTIONS(1362),
    [anon_sym_QMARK] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1362),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(571)] = {
    [sym_comment] = STATE(571),
    [sym_groovy_doc] = STATE(571),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_DOT_AMP] = ACTIONS(1410),
    [anon_sym_DOT_AT] = ACTIONS(1410),
    [anon_sym_QMARK_DOT] = ACTIONS(1410),
    [anon_sym_STAR_DOT] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_DOT] = ACTIONS(1412),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_as] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_PERCENT] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_LT_LT] = ACTIONS(1410),
    [anon_sym_GT_GT] = ACTIONS(1412),
    [anon_sym_GT_GT_GT] = ACTIONS(1410),
    [anon_sym_DOT_DOT] = ACTIONS(1412),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1410),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1412),
    [anon_sym_LT] = ACTIONS(1412),
    [anon_sym_LT_EQ] = ACTIONS(1412),
    [anon_sym_GT] = ACTIONS(1412),
    [anon_sym_GT_EQ] = ACTIONS(1410),
    [anon_sym_in] = ACTIONS(1412),
    [anon_sym_BANGin] = ACTIONS(1412),
    [anon_sym_instanceof] = ACTIONS(1412),
    [anon_sym_BANGinstanceof] = ACTIONS(1410),
    [anon_sym_EQ_EQ] = ACTIONS(1412),
    [anon_sym_BANG_EQ] = ACTIONS(1412),
    [anon_sym_LT_EQ_GT] = ACTIONS(1410),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1410),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1410),
    [anon_sym_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1410),
    [anon_sym_AMP] = ACTIONS(1412),
    [anon_sym_CARET] = ACTIONS(1410),
    [anon_sym_PIPE] = ACTIONS(1412),
    [anon_sym_AMP_AMP] = ACTIONS(1410),
    [anon_sym_PIPE_PIPE] = ACTIONS(1410),
    [anon_sym_QMARK_COLON] = ACTIONS(1410),
    [anon_sym_STAR_STAR] = ACTIONS(1410),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_COMMA] = ACTIONS(1410),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_case] = ACTIONS(1412),
    [anon_sym_default] = ACTIONS(1412),
    [anon_sym_QMARK] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(572)] = {
    [sym_comment] = STATE(572),
    [sym_groovy_doc] = STATE(572),
    [ts_builtin_sym_end] = ACTIONS(1501),
    [sym_identifier] = ACTIONS(1499),
    [anon_sym_break] = ACTIONS(1499),
    [anon_sym_continue] = ACTIONS(1499),
    [anon_sym_SEMI] = ACTIONS(1501),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1501),
    [anon_sym_import] = ACTIONS(1499),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1499),
    [anon_sym_AT] = ACTIONS(1499),
    [anon_sym_assert] = ACTIONS(1499),
    [anon_sym_PLUS_PLUS] = ACTIONS(1501),
    [anon_sym_DASH_DASH] = ACTIONS(1501),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1499),
    [anon_sym_false] = ACTIONS(1499),
    [anon_sym_ATinterface] = ACTIONS(1501),
    [anon_sym_interface] = ACTIONS(1499),
    [anon_sym_class] = ACTIONS(1499),
    [anon_sym_COMMA] = ACTIONS(1501),
    [anon_sym_LBRACE] = ACTIONS(1501),
    [anon_sym_RBRACE] = ACTIONS(1501),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1499),
    [anon_sym_def] = ACTIONS(1499),
    [anon_sym_LPAREN] = ACTIONS(1501),
    [anon_sym_null] = ACTIONS(1499),
    [anon_sym_this] = ACTIONS(1499),
    [anon_sym_do] = ACTIONS(1499),
    [anon_sym_while] = ACTIONS(1499),
    [anon_sym_for] = ACTIONS(1499),
    [aux_sym__type_identifier_token1] = ACTIONS(1499),
    [anon_sym_if] = ACTIONS(1499),
    [anon_sym_else] = ACTIONS(1499),
    [anon_sym_LBRACK] = ACTIONS(1501),
    [aux_sym_number_literal_token1] = ACTIONS(1499),
    [aux_sym_number_literal_token2] = ACTIONS(1501),
    [aux_sym_number_literal_token3] = ACTIONS(1501),
    [aux_sym_number_literal_token4] = ACTIONS(1499),
    [aux_sym_number_literal_token5] = ACTIONS(1501),
    [anon_sym_pipeline] = ACTIONS(1499),
    [anon_sym_return] = ACTIONS(1499),
    [anon_sym_SQUOTE] = ACTIONS(1499),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(1499),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1501),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1501),
    [anon_sym_switch] = ACTIONS(1499),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1499),
    [anon_sym_int] = ACTIONS(1499),
    [anon_sym_boolean] = ACTIONS(1499),
    [anon_sym_char] = ACTIONS(1499),
    [anon_sym_short] = ACTIONS(1499),
    [anon_sym_long] = ACTIONS(1499),
    [anon_sym_float] = ACTIONS(1499),
    [anon_sym_double] = ACTIONS(1499),
    [anon_sym_void] = ACTIONS(1499),
    [anon_sym_public] = ACTIONS(1499),
    [anon_sym_protected] = ACTIONS(1499),
    [anon_sym_private] = ACTIONS(1499),
    [anon_sym_static] = ACTIONS(1499),
    [anon_sym_final] = ACTIONS(1499),
    [anon_sym_synchronized] = ACTIONS(1499),
    [anon_sym_TILDE] = ACTIONS(1501),
    [anon_sym_BANG] = ACTIONS(1499),
    [anon_sym_new] = ACTIONS(1499),
  },
  [STATE(573)] = {
    [sym_comment] = STATE(573),
    [sym_groovy_doc] = STATE(573),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1457),
    [anon_sym_DOT_AT] = ACTIONS(1457),
    [anon_sym_QMARK_DOT] = ACTIONS(1457),
    [anon_sym_STAR_DOT] = ACTIONS(1457),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1457),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_case] = ACTIONS(1455),
    [anon_sym_default] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(574)] = {
    [sym_comment] = STATE(574),
    [sym_groovy_doc] = STATE(574),
    [sym_identifier] = ACTIONS(1280),
    [anon_sym_break] = ACTIONS(1280),
    [anon_sym_continue] = ACTIONS(1280),
    [anon_sym_SEMI] = ACTIONS(1278),
    [anon_sym_DOT_AMP] = ACTIONS(1278),
    [anon_sym_DOT_AT] = ACTIONS(1278),
    [anon_sym_QMARK_DOT] = ACTIONS(1278),
    [anon_sym_STAR_DOT] = ACTIONS(1278),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_STAR_COLON] = ACTIONS(1278),
    [anon_sym_import] = ACTIONS(1280),
    [anon_sym_as] = ACTIONS(1280),
    [anon_sym_package] = ACTIONS(1280),
    [anon_sym_AT] = ACTIONS(1280),
    [anon_sym_assert] = ACTIONS(1280),
    [anon_sym_PLUS_PLUS] = ACTIONS(1278),
    [anon_sym_DASH_DASH] = ACTIONS(1278),
    [anon_sym_PERCENT] = ACTIONS(1278),
    [anon_sym_SLASH] = ACTIONS(1280),
    [anon_sym_PLUS] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_LT_LT] = ACTIONS(1278),
    [anon_sym_GT_GT] = ACTIONS(1280),
    [anon_sym_GT_GT_GT] = ACTIONS(1278),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1278),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1280),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_LT_EQ] = ACTIONS(1280),
    [anon_sym_GT] = ACTIONS(1280),
    [anon_sym_GT_EQ] = ACTIONS(1278),
    [anon_sym_in] = ACTIONS(1280),
    [anon_sym_BANGin] = ACTIONS(1280),
    [anon_sym_instanceof] = ACTIONS(1280),
    [anon_sym_BANGinstanceof] = ACTIONS(1278),
    [anon_sym_EQ_EQ] = ACTIONS(1280),
    [anon_sym_BANG_EQ] = ACTIONS(1280),
    [anon_sym_LT_EQ_GT] = ACTIONS(1278),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1278),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1278),
    [anon_sym_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1278),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_CARET] = ACTIONS(1278),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_AMP_AMP] = ACTIONS(1278),
    [anon_sym_PIPE_PIPE] = ACTIONS(1278),
    [anon_sym_QMARK_COLON] = ACTIONS(1278),
    [anon_sym_STAR_STAR] = ACTIONS(1278),
    [anon_sym_true] = ACTIONS(1280),
    [anon_sym_false] = ACTIONS(1280),
    [anon_sym_ATinterface] = ACTIONS(1278),
    [anon_sym_interface] = ACTIONS(1280),
    [anon_sym_class] = ACTIONS(1280),
    [anon_sym_COMMA] = ACTIONS(1278),
    [anon_sym_LBRACE] = ACTIONS(1278),
    [anon_sym_RBRACE] = ACTIONS(1278),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1280),
    [anon_sym_def] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1278),
    [anon_sym_null] = ACTIONS(1280),
    [anon_sym_this] = ACTIONS(1280),
    [anon_sym_do] = ACTIONS(1280),
    [anon_sym_while] = ACTIONS(1280),
    [anon_sym_for] = ACTIONS(1280),
    [aux_sym__type_identifier_token1] = ACTIONS(1280),
    [anon_sym_if] = ACTIONS(1280),
    [anon_sym_else] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1278),
    [aux_sym_number_literal_token1] = ACTIONS(1280),
    [aux_sym_number_literal_token2] = ACTIONS(1278),
    [aux_sym_number_literal_token3] = ACTIONS(1278),
    [aux_sym_number_literal_token4] = ACTIONS(1280),
    [aux_sym_number_literal_token5] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1280),
    [anon_sym_SQUOTE] = ACTIONS(1280),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1278),
    [anon_sym_DQUOTE] = ACTIONS(1280),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1278),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1278),
    [anon_sym_switch] = ACTIONS(1280),
    [anon_sym_case] = ACTIONS(1280),
    [anon_sym_default] = ACTIONS(1280),
    [anon_sym_QMARK] = ACTIONS(1280),
    [anon_sym_try] = ACTIONS(1280),
    [anon_sym_int] = ACTIONS(1280),
    [anon_sym_boolean] = ACTIONS(1280),
    [anon_sym_char] = ACTIONS(1280),
    [anon_sym_short] = ACTIONS(1280),
    [anon_sym_long] = ACTIONS(1280),
    [anon_sym_float] = ACTIONS(1280),
    [anon_sym_double] = ACTIONS(1280),
    [anon_sym_void] = ACTIONS(1280),
    [anon_sym_public] = ACTIONS(1280),
    [anon_sym_protected] = ACTIONS(1280),
    [anon_sym_private] = ACTIONS(1280),
    [anon_sym_static] = ACTIONS(1280),
    [anon_sym_final] = ACTIONS(1280),
    [anon_sym_synchronized] = ACTIONS(1280),
    [anon_sym_TILDE] = ACTIONS(1278),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_new] = ACTIONS(1280),
  },
  [STATE(575)] = {
    [sym_comment] = STATE(575),
    [sym_groovy_doc] = STATE(575),
    [sym_identifier] = ACTIONS(1463),
    [anon_sym_break] = ACTIONS(1463),
    [anon_sym_continue] = ACTIONS(1463),
    [anon_sym_SEMI] = ACTIONS(1465),
    [anon_sym_DOT_AMP] = ACTIONS(1465),
    [anon_sym_DOT_AT] = ACTIONS(1465),
    [anon_sym_QMARK_DOT] = ACTIONS(1465),
    [anon_sym_STAR_DOT] = ACTIONS(1465),
    [anon_sym_STAR] = ACTIONS(1463),
    [anon_sym_STAR_COLON] = ACTIONS(1465),
    [anon_sym_import] = ACTIONS(1463),
    [anon_sym_as] = ACTIONS(1463),
    [anon_sym_package] = ACTIONS(1463),
    [anon_sym_AT] = ACTIONS(1463),
    [anon_sym_assert] = ACTIONS(1463),
    [anon_sym_PLUS_PLUS] = ACTIONS(1465),
    [anon_sym_DASH_DASH] = ACTIONS(1465),
    [anon_sym_PERCENT] = ACTIONS(1465),
    [anon_sym_SLASH] = ACTIONS(1463),
    [anon_sym_PLUS] = ACTIONS(1463),
    [anon_sym_DASH] = ACTIONS(1463),
    [anon_sym_LT_LT] = ACTIONS(1465),
    [anon_sym_GT_GT] = ACTIONS(1463),
    [anon_sym_GT_GT_GT] = ACTIONS(1465),
    [anon_sym_DOT_DOT] = ACTIONS(1463),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1465),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1463),
    [anon_sym_LT] = ACTIONS(1463),
    [anon_sym_LT_EQ] = ACTIONS(1463),
    [anon_sym_GT] = ACTIONS(1463),
    [anon_sym_GT_EQ] = ACTIONS(1465),
    [anon_sym_in] = ACTIONS(1463),
    [anon_sym_BANGin] = ACTIONS(1463),
    [anon_sym_instanceof] = ACTIONS(1463),
    [anon_sym_BANGinstanceof] = ACTIONS(1465),
    [anon_sym_EQ_EQ] = ACTIONS(1463),
    [anon_sym_BANG_EQ] = ACTIONS(1463),
    [anon_sym_LT_EQ_GT] = ACTIONS(1465),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1465),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1465),
    [anon_sym_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1465),
    [anon_sym_AMP] = ACTIONS(1463),
    [anon_sym_CARET] = ACTIONS(1465),
    [anon_sym_PIPE] = ACTIONS(1463),
    [anon_sym_AMP_AMP] = ACTIONS(1465),
    [anon_sym_PIPE_PIPE] = ACTIONS(1465),
    [anon_sym_QMARK_COLON] = ACTIONS(1465),
    [anon_sym_STAR_STAR] = ACTIONS(1465),
    [anon_sym_true] = ACTIONS(1463),
    [anon_sym_false] = ACTIONS(1463),
    [anon_sym_ATinterface] = ACTIONS(1465),
    [anon_sym_interface] = ACTIONS(1463),
    [anon_sym_class] = ACTIONS(1463),
    [anon_sym_COMMA] = ACTIONS(1465),
    [anon_sym_LBRACE] = ACTIONS(1465),
    [anon_sym_RBRACE] = ACTIONS(1465),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1463),
    [anon_sym_def] = ACTIONS(1463),
    [anon_sym_LPAREN] = ACTIONS(1465),
    [anon_sym_null] = ACTIONS(1463),
    [anon_sym_this] = ACTIONS(1463),
    [anon_sym_do] = ACTIONS(1463),
    [anon_sym_while] = ACTIONS(1463),
    [anon_sym_for] = ACTIONS(1463),
    [aux_sym__type_identifier_token1] = ACTIONS(1463),
    [anon_sym_if] = ACTIONS(1463),
    [anon_sym_else] = ACTIONS(1463),
    [anon_sym_LBRACK] = ACTIONS(1465),
    [aux_sym_number_literal_token1] = ACTIONS(1463),
    [aux_sym_number_literal_token2] = ACTIONS(1465),
    [aux_sym_number_literal_token3] = ACTIONS(1465),
    [aux_sym_number_literal_token4] = ACTIONS(1463),
    [aux_sym_number_literal_token5] = ACTIONS(1465),
    [anon_sym_return] = ACTIONS(1463),
    [anon_sym_SQUOTE] = ACTIONS(1463),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1465),
    [anon_sym_DQUOTE] = ACTIONS(1463),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1465),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1465),
    [anon_sym_switch] = ACTIONS(1463),
    [anon_sym_case] = ACTIONS(1463),
    [anon_sym_default] = ACTIONS(1463),
    [anon_sym_QMARK] = ACTIONS(1463),
    [anon_sym_try] = ACTIONS(1463),
    [anon_sym_int] = ACTIONS(1463),
    [anon_sym_boolean] = ACTIONS(1463),
    [anon_sym_char] = ACTIONS(1463),
    [anon_sym_short] = ACTIONS(1463),
    [anon_sym_long] = ACTIONS(1463),
    [anon_sym_float] = ACTIONS(1463),
    [anon_sym_double] = ACTIONS(1463),
    [anon_sym_void] = ACTIONS(1463),
    [anon_sym_public] = ACTIONS(1463),
    [anon_sym_protected] = ACTIONS(1463),
    [anon_sym_private] = ACTIONS(1463),
    [anon_sym_static] = ACTIONS(1463),
    [anon_sym_final] = ACTIONS(1463),
    [anon_sym_synchronized] = ACTIONS(1463),
    [anon_sym_TILDE] = ACTIONS(1465),
    [anon_sym_BANG] = ACTIONS(1463),
    [anon_sym_new] = ACTIONS(1463),
  },
  [STATE(576)] = {
    [sym_comment] = STATE(576),
    [sym_groovy_doc] = STATE(576),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(577)] = {
    [sym_comment] = STATE(577),
    [sym_groovy_doc] = STATE(577),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(578)] = {
    [sym_comment] = STATE(578),
    [sym_groovy_doc] = STATE(578),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(579)] = {
    [sym_comment] = STATE(579),
    [sym_groovy_doc] = STATE(579),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(580)] = {
    [sym_comment] = STATE(580),
    [sym_groovy_doc] = STATE(580),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(581)] = {
    [sym_comment] = STATE(581),
    [sym_groovy_doc] = STATE(581),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(582)] = {
    [sym_comment] = STATE(582),
    [sym_groovy_doc] = STATE(582),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(583)] = {
    [sym_comment] = STATE(583),
    [sym_groovy_doc] = STATE(583),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(584)] = {
    [sym_comment] = STATE(584),
    [sym_groovy_doc] = STATE(584),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(585)] = {
    [sym_comment] = STATE(585),
    [sym_groovy_doc] = STATE(585),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(586)] = {
    [sym_comment] = STATE(586),
    [sym_groovy_doc] = STATE(586),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(587)] = {
    [sym_comment] = STATE(587),
    [sym_groovy_doc] = STATE(587),
    [sym_identifier] = ACTIONS(1467),
    [anon_sym_break] = ACTIONS(1467),
    [anon_sym_continue] = ACTIONS(1467),
    [anon_sym_SEMI] = ACTIONS(1469),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1467),
    [anon_sym_STAR_COLON] = ACTIONS(1469),
    [anon_sym_import] = ACTIONS(1467),
    [anon_sym_as] = ACTIONS(1467),
    [anon_sym_package] = ACTIONS(1467),
    [anon_sym_AT] = ACTIONS(1467),
    [anon_sym_assert] = ACTIONS(1467),
    [anon_sym_PLUS_PLUS] = ACTIONS(1469),
    [anon_sym_DASH_DASH] = ACTIONS(1469),
    [anon_sym_PERCENT] = ACTIONS(1469),
    [anon_sym_SLASH] = ACTIONS(1467),
    [anon_sym_PLUS] = ACTIONS(1467),
    [anon_sym_DASH] = ACTIONS(1467),
    [anon_sym_LT_LT] = ACTIONS(1469),
    [anon_sym_GT_GT] = ACTIONS(1467),
    [anon_sym_GT_GT_GT] = ACTIONS(1469),
    [anon_sym_DOT_DOT] = ACTIONS(1467),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1469),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1467),
    [anon_sym_LT] = ACTIONS(1467),
    [anon_sym_LT_EQ] = ACTIONS(1467),
    [anon_sym_GT] = ACTIONS(1467),
    [anon_sym_GT_EQ] = ACTIONS(1469),
    [anon_sym_in] = ACTIONS(1467),
    [anon_sym_BANGin] = ACTIONS(1467),
    [anon_sym_instanceof] = ACTIONS(1467),
    [anon_sym_BANGinstanceof] = ACTIONS(1469),
    [anon_sym_EQ_EQ] = ACTIONS(1467),
    [anon_sym_BANG_EQ] = ACTIONS(1467),
    [anon_sym_LT_EQ_GT] = ACTIONS(1469),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1469),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1469),
    [anon_sym_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1469),
    [anon_sym_AMP] = ACTIONS(1467),
    [anon_sym_CARET] = ACTIONS(1469),
    [anon_sym_PIPE] = ACTIONS(1467),
    [anon_sym_AMP_AMP] = ACTIONS(1469),
    [anon_sym_PIPE_PIPE] = ACTIONS(1469),
    [anon_sym_QMARK_COLON] = ACTIONS(1469),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1467),
    [anon_sym_false] = ACTIONS(1467),
    [anon_sym_ATinterface] = ACTIONS(1469),
    [anon_sym_interface] = ACTIONS(1467),
    [anon_sym_class] = ACTIONS(1467),
    [anon_sym_COMMA] = ACTIONS(1469),
    [anon_sym_LBRACE] = ACTIONS(1469),
    [anon_sym_RBRACE] = ACTIONS(1469),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1467),
    [anon_sym_def] = ACTIONS(1467),
    [anon_sym_LPAREN] = ACTIONS(1469),
    [anon_sym_null] = ACTIONS(1467),
    [anon_sym_this] = ACTIONS(1467),
    [anon_sym_do] = ACTIONS(1467),
    [anon_sym_while] = ACTIONS(1467),
    [anon_sym_for] = ACTIONS(1467),
    [aux_sym__type_identifier_token1] = ACTIONS(1467),
    [anon_sym_if] = ACTIONS(1467),
    [anon_sym_else] = ACTIONS(1467),
    [anon_sym_LBRACK] = ACTIONS(1469),
    [aux_sym_number_literal_token1] = ACTIONS(1467),
    [aux_sym_number_literal_token2] = ACTIONS(1469),
    [aux_sym_number_literal_token3] = ACTIONS(1469),
    [aux_sym_number_literal_token4] = ACTIONS(1467),
    [aux_sym_number_literal_token5] = ACTIONS(1469),
    [anon_sym_return] = ACTIONS(1467),
    [anon_sym_SQUOTE] = ACTIONS(1467),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1469),
    [anon_sym_DQUOTE] = ACTIONS(1467),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1469),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1469),
    [anon_sym_switch] = ACTIONS(1467),
    [anon_sym_case] = ACTIONS(1467),
    [anon_sym_default] = ACTIONS(1467),
    [anon_sym_QMARK] = ACTIONS(1467),
    [anon_sym_try] = ACTIONS(1467),
    [anon_sym_int] = ACTIONS(1467),
    [anon_sym_boolean] = ACTIONS(1467),
    [anon_sym_char] = ACTIONS(1467),
    [anon_sym_short] = ACTIONS(1467),
    [anon_sym_long] = ACTIONS(1467),
    [anon_sym_float] = ACTIONS(1467),
    [anon_sym_double] = ACTIONS(1467),
    [anon_sym_void] = ACTIONS(1467),
    [anon_sym_public] = ACTIONS(1467),
    [anon_sym_protected] = ACTIONS(1467),
    [anon_sym_private] = ACTIONS(1467),
    [anon_sym_static] = ACTIONS(1467),
    [anon_sym_final] = ACTIONS(1467),
    [anon_sym_synchronized] = ACTIONS(1467),
    [anon_sym_TILDE] = ACTIONS(1469),
    [anon_sym_BANG] = ACTIONS(1467),
    [anon_sym_new] = ACTIONS(1467),
  },
  [STATE(588)] = {
    [sym_comment] = STATE(588),
    [sym_groovy_doc] = STATE(588),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(589)] = {
    [sym_comment] = STATE(589),
    [sym_groovy_doc] = STATE(589),
    [sym_identifier] = ACTIONS(1509),
    [anon_sym_break] = ACTIONS(1509),
    [anon_sym_continue] = ACTIONS(1509),
    [anon_sym_SEMI] = ACTIONS(1511),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1511),
    [anon_sym_import] = ACTIONS(1509),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1509),
    [anon_sym_AT] = ACTIONS(1509),
    [anon_sym_assert] = ACTIONS(1509),
    [anon_sym_PLUS_PLUS] = ACTIONS(1511),
    [anon_sym_DASH_DASH] = ACTIONS(1511),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1509),
    [anon_sym_false] = ACTIONS(1509),
    [anon_sym_ATinterface] = ACTIONS(1511),
    [anon_sym_interface] = ACTIONS(1509),
    [anon_sym_class] = ACTIONS(1509),
    [anon_sym_COMMA] = ACTIONS(1511),
    [anon_sym_LBRACE] = ACTIONS(1511),
    [anon_sym_RBRACE] = ACTIONS(1511),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1509),
    [anon_sym_def] = ACTIONS(1509),
    [anon_sym_LPAREN] = ACTIONS(1511),
    [anon_sym_null] = ACTIONS(1509),
    [anon_sym_this] = ACTIONS(1509),
    [anon_sym_do] = ACTIONS(1509),
    [anon_sym_while] = ACTIONS(1509),
    [anon_sym_for] = ACTIONS(1509),
    [aux_sym__type_identifier_token1] = ACTIONS(1509),
    [anon_sym_if] = ACTIONS(1509),
    [anon_sym_else] = ACTIONS(1509),
    [anon_sym_LBRACK] = ACTIONS(1511),
    [aux_sym_number_literal_token1] = ACTIONS(1509),
    [aux_sym_number_literal_token2] = ACTIONS(1511),
    [aux_sym_number_literal_token3] = ACTIONS(1511),
    [aux_sym_number_literal_token4] = ACTIONS(1509),
    [aux_sym_number_literal_token5] = ACTIONS(1511),
    [anon_sym_return] = ACTIONS(1509),
    [anon_sym_SQUOTE] = ACTIONS(1509),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1511),
    [anon_sym_DQUOTE] = ACTIONS(1509),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1511),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1511),
    [anon_sym_switch] = ACTIONS(1509),
    [anon_sym_case] = ACTIONS(1509),
    [anon_sym_default] = ACTIONS(1509),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1509),
    [anon_sym_int] = ACTIONS(1509),
    [anon_sym_boolean] = ACTIONS(1509),
    [anon_sym_char] = ACTIONS(1509),
    [anon_sym_short] = ACTIONS(1509),
    [anon_sym_long] = ACTIONS(1509),
    [anon_sym_float] = ACTIONS(1509),
    [anon_sym_double] = ACTIONS(1509),
    [anon_sym_void] = ACTIONS(1509),
    [anon_sym_public] = ACTIONS(1509),
    [anon_sym_protected] = ACTIONS(1509),
    [anon_sym_private] = ACTIONS(1509),
    [anon_sym_static] = ACTIONS(1509),
    [anon_sym_final] = ACTIONS(1509),
    [anon_sym_synchronized] = ACTIONS(1509),
    [anon_sym_TILDE] = ACTIONS(1511),
    [anon_sym_BANG] = ACTIONS(1509),
    [anon_sym_new] = ACTIONS(1509),
  },
  [STATE(590)] = {
    [sym_comment] = STATE(590),
    [sym_groovy_doc] = STATE(590),
    [sym_identifier] = ACTIONS(1499),
    [anon_sym_break] = ACTIONS(1499),
    [anon_sym_continue] = ACTIONS(1499),
    [anon_sym_SEMI] = ACTIONS(1501),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1501),
    [anon_sym_import] = ACTIONS(1499),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1499),
    [anon_sym_AT] = ACTIONS(1499),
    [anon_sym_assert] = ACTIONS(1499),
    [anon_sym_PLUS_PLUS] = ACTIONS(1501),
    [anon_sym_DASH_DASH] = ACTIONS(1501),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1499),
    [anon_sym_false] = ACTIONS(1499),
    [anon_sym_ATinterface] = ACTIONS(1501),
    [anon_sym_interface] = ACTIONS(1499),
    [anon_sym_class] = ACTIONS(1499),
    [anon_sym_COMMA] = ACTIONS(1501),
    [anon_sym_LBRACE] = ACTIONS(1501),
    [anon_sym_RBRACE] = ACTIONS(1501),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1499),
    [anon_sym_def] = ACTIONS(1499),
    [anon_sym_LPAREN] = ACTIONS(1501),
    [anon_sym_null] = ACTIONS(1499),
    [anon_sym_this] = ACTIONS(1499),
    [anon_sym_do] = ACTIONS(1499),
    [anon_sym_while] = ACTIONS(1499),
    [anon_sym_for] = ACTIONS(1499),
    [aux_sym__type_identifier_token1] = ACTIONS(1499),
    [anon_sym_if] = ACTIONS(1499),
    [anon_sym_else] = ACTIONS(1499),
    [anon_sym_LBRACK] = ACTIONS(1501),
    [aux_sym_number_literal_token1] = ACTIONS(1499),
    [aux_sym_number_literal_token2] = ACTIONS(1501),
    [aux_sym_number_literal_token3] = ACTIONS(1501),
    [aux_sym_number_literal_token4] = ACTIONS(1499),
    [aux_sym_number_literal_token5] = ACTIONS(1501),
    [anon_sym_return] = ACTIONS(1499),
    [anon_sym_SQUOTE] = ACTIONS(1499),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1501),
    [anon_sym_DQUOTE] = ACTIONS(1499),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1501),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1501),
    [anon_sym_switch] = ACTIONS(1499),
    [anon_sym_case] = ACTIONS(1499),
    [anon_sym_default] = ACTIONS(1499),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1499),
    [anon_sym_int] = ACTIONS(1499),
    [anon_sym_boolean] = ACTIONS(1499),
    [anon_sym_char] = ACTIONS(1499),
    [anon_sym_short] = ACTIONS(1499),
    [anon_sym_long] = ACTIONS(1499),
    [anon_sym_float] = ACTIONS(1499),
    [anon_sym_double] = ACTIONS(1499),
    [anon_sym_void] = ACTIONS(1499),
    [anon_sym_public] = ACTIONS(1499),
    [anon_sym_protected] = ACTIONS(1499),
    [anon_sym_private] = ACTIONS(1499),
    [anon_sym_static] = ACTIONS(1499),
    [anon_sym_final] = ACTIONS(1499),
    [anon_sym_synchronized] = ACTIONS(1499),
    [anon_sym_TILDE] = ACTIONS(1501),
    [anon_sym_BANG] = ACTIONS(1499),
    [anon_sym_new] = ACTIONS(1499),
  },
  [STATE(591)] = {
    [sym_comment] = STATE(591),
    [sym_groovy_doc] = STATE(591),
    [sym_identifier] = ACTIONS(1449),
    [anon_sym_break] = ACTIONS(1449),
    [anon_sym_continue] = ACTIONS(1449),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_DOT_AMP] = ACTIONS(1451),
    [anon_sym_DOT_AT] = ACTIONS(1451),
    [anon_sym_QMARK_DOT] = ACTIONS(1451),
    [anon_sym_STAR_DOT] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1449),
    [anon_sym_STAR_COLON] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1449),
    [anon_sym_as] = ACTIONS(1449),
    [anon_sym_package] = ACTIONS(1449),
    [anon_sym_AT] = ACTIONS(1449),
    [anon_sym_assert] = ACTIONS(1449),
    [anon_sym_PLUS_PLUS] = ACTIONS(1451),
    [anon_sym_DASH_DASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1449),
    [anon_sym_PLUS] = ACTIONS(1449),
    [anon_sym_DASH] = ACTIONS(1449),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1449),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_DOT_DOT] = ACTIONS(1449),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1451),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1449),
    [anon_sym_LT] = ACTIONS(1449),
    [anon_sym_LT_EQ] = ACTIONS(1449),
    [anon_sym_GT] = ACTIONS(1449),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1449),
    [anon_sym_BANGin] = ACTIONS(1449),
    [anon_sym_instanceof] = ACTIONS(1449),
    [anon_sym_BANGinstanceof] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1449),
    [anon_sym_BANG_EQ] = ACTIONS(1449),
    [anon_sym_LT_EQ_GT] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1449),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1449),
    [anon_sym_AMP_AMP] = ACTIONS(1451),
    [anon_sym_PIPE_PIPE] = ACTIONS(1451),
    [anon_sym_QMARK_COLON] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_true] = ACTIONS(1449),
    [anon_sym_false] = ACTIONS(1449),
    [anon_sym_ATinterface] = ACTIONS(1451),
    [anon_sym_interface] = ACTIONS(1449),
    [anon_sym_class] = ACTIONS(1449),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [anon_sym_RBRACE] = ACTIONS(1451),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1449),
    [anon_sym_def] = ACTIONS(1449),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_null] = ACTIONS(1449),
    [anon_sym_this] = ACTIONS(1449),
    [anon_sym_do] = ACTIONS(1449),
    [anon_sym_while] = ACTIONS(1449),
    [anon_sym_for] = ACTIONS(1449),
    [aux_sym__type_identifier_token1] = ACTIONS(1449),
    [anon_sym_if] = ACTIONS(1449),
    [anon_sym_else] = ACTIONS(1449),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [aux_sym_number_literal_token1] = ACTIONS(1449),
    [aux_sym_number_literal_token2] = ACTIONS(1451),
    [aux_sym_number_literal_token3] = ACTIONS(1451),
    [aux_sym_number_literal_token4] = ACTIONS(1449),
    [aux_sym_number_literal_token5] = ACTIONS(1451),
    [anon_sym_return] = ACTIONS(1449),
    [anon_sym_SQUOTE] = ACTIONS(1449),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1451),
    [anon_sym_DQUOTE] = ACTIONS(1449),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1451),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1451),
    [anon_sym_switch] = ACTIONS(1449),
    [anon_sym_case] = ACTIONS(1449),
    [anon_sym_default] = ACTIONS(1449),
    [anon_sym_QMARK] = ACTIONS(1449),
    [anon_sym_try] = ACTIONS(1449),
    [anon_sym_int] = ACTIONS(1449),
    [anon_sym_boolean] = ACTIONS(1449),
    [anon_sym_char] = ACTIONS(1449),
    [anon_sym_short] = ACTIONS(1449),
    [anon_sym_long] = ACTIONS(1449),
    [anon_sym_float] = ACTIONS(1449),
    [anon_sym_double] = ACTIONS(1449),
    [anon_sym_void] = ACTIONS(1449),
    [anon_sym_public] = ACTIONS(1449),
    [anon_sym_protected] = ACTIONS(1449),
    [anon_sym_private] = ACTIONS(1449),
    [anon_sym_static] = ACTIONS(1449),
    [anon_sym_final] = ACTIONS(1449),
    [anon_sym_synchronized] = ACTIONS(1449),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_BANG] = ACTIONS(1449),
    [anon_sym_new] = ACTIONS(1449),
  },
  [STATE(592)] = {
    [sym_comment] = STATE(592),
    [sym_groovy_doc] = STATE(592),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(593)] = {
    [sym_comment] = STATE(593),
    [sym_groovy_doc] = STATE(593),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1270),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(594)] = {
    [sym_comment] = STATE(594),
    [sym_groovy_doc] = STATE(594),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(595)] = {
    [sym_comment] = STATE(595),
    [sym_groovy_doc] = STATE(595),
    [sym_identifier] = ACTIONS(1455),
    [anon_sym_break] = ACTIONS(1455),
    [anon_sym_continue] = ACTIONS(1455),
    [anon_sym_SEMI] = ACTIONS(1457),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1455),
    [anon_sym_STAR_COLON] = ACTIONS(1457),
    [anon_sym_import] = ACTIONS(1455),
    [anon_sym_as] = ACTIONS(1455),
    [anon_sym_package] = ACTIONS(1455),
    [anon_sym_AT] = ACTIONS(1455),
    [anon_sym_assert] = ACTIONS(1455),
    [anon_sym_PLUS_PLUS] = ACTIONS(1457),
    [anon_sym_DASH_DASH] = ACTIONS(1457),
    [anon_sym_PERCENT] = ACTIONS(1457),
    [anon_sym_SLASH] = ACTIONS(1455),
    [anon_sym_PLUS] = ACTIONS(1455),
    [anon_sym_DASH] = ACTIONS(1455),
    [anon_sym_LT_LT] = ACTIONS(1457),
    [anon_sym_GT_GT] = ACTIONS(1455),
    [anon_sym_GT_GT_GT] = ACTIONS(1457),
    [anon_sym_DOT_DOT] = ACTIONS(1455),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1457),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1455),
    [anon_sym_LT] = ACTIONS(1455),
    [anon_sym_LT_EQ] = ACTIONS(1455),
    [anon_sym_GT] = ACTIONS(1455),
    [anon_sym_GT_EQ] = ACTIONS(1457),
    [anon_sym_in] = ACTIONS(1455),
    [anon_sym_BANGin] = ACTIONS(1455),
    [anon_sym_instanceof] = ACTIONS(1455),
    [anon_sym_BANGinstanceof] = ACTIONS(1457),
    [anon_sym_EQ_EQ] = ACTIONS(1455),
    [anon_sym_BANG_EQ] = ACTIONS(1455),
    [anon_sym_LT_EQ_GT] = ACTIONS(1457),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1457),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1457),
    [anon_sym_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1457),
    [anon_sym_AMP] = ACTIONS(1455),
    [anon_sym_CARET] = ACTIONS(1457),
    [anon_sym_PIPE] = ACTIONS(1455),
    [anon_sym_AMP_AMP] = ACTIONS(1457),
    [anon_sym_PIPE_PIPE] = ACTIONS(1457),
    [anon_sym_QMARK_COLON] = ACTIONS(1457),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1455),
    [anon_sym_false] = ACTIONS(1455),
    [anon_sym_ATinterface] = ACTIONS(1457),
    [anon_sym_interface] = ACTIONS(1455),
    [anon_sym_class] = ACTIONS(1455),
    [anon_sym_COMMA] = ACTIONS(1457),
    [anon_sym_LBRACE] = ACTIONS(1457),
    [anon_sym_RBRACE] = ACTIONS(1457),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1455),
    [anon_sym_def] = ACTIONS(1455),
    [anon_sym_LPAREN] = ACTIONS(1457),
    [anon_sym_null] = ACTIONS(1455),
    [anon_sym_this] = ACTIONS(1455),
    [anon_sym_do] = ACTIONS(1455),
    [anon_sym_while] = ACTIONS(1455),
    [anon_sym_for] = ACTIONS(1455),
    [aux_sym__type_identifier_token1] = ACTIONS(1455),
    [anon_sym_if] = ACTIONS(1455),
    [anon_sym_else] = ACTIONS(1455),
    [anon_sym_LBRACK] = ACTIONS(1457),
    [aux_sym_number_literal_token1] = ACTIONS(1455),
    [aux_sym_number_literal_token2] = ACTIONS(1457),
    [aux_sym_number_literal_token3] = ACTIONS(1457),
    [aux_sym_number_literal_token4] = ACTIONS(1455),
    [aux_sym_number_literal_token5] = ACTIONS(1457),
    [anon_sym_return] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1455),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1455),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1457),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1457),
    [anon_sym_switch] = ACTIONS(1455),
    [anon_sym_case] = ACTIONS(1455),
    [anon_sym_default] = ACTIONS(1455),
    [anon_sym_QMARK] = ACTIONS(1455),
    [anon_sym_try] = ACTIONS(1455),
    [anon_sym_int] = ACTIONS(1455),
    [anon_sym_boolean] = ACTIONS(1455),
    [anon_sym_char] = ACTIONS(1455),
    [anon_sym_short] = ACTIONS(1455),
    [anon_sym_long] = ACTIONS(1455),
    [anon_sym_float] = ACTIONS(1455),
    [anon_sym_double] = ACTIONS(1455),
    [anon_sym_void] = ACTIONS(1455),
    [anon_sym_public] = ACTIONS(1455),
    [anon_sym_protected] = ACTIONS(1455),
    [anon_sym_private] = ACTIONS(1455),
    [anon_sym_static] = ACTIONS(1455),
    [anon_sym_final] = ACTIONS(1455),
    [anon_sym_synchronized] = ACTIONS(1455),
    [anon_sym_TILDE] = ACTIONS(1457),
    [anon_sym_BANG] = ACTIONS(1455),
    [anon_sym_new] = ACTIONS(1455),
  },
  [STATE(596)] = {
    [sym_comment] = STATE(596),
    [sym_groovy_doc] = STATE(596),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(597)] = {
    [sym_comment] = STATE(597),
    [sym_groovy_doc] = STATE(597),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_COMMA] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(598)] = {
    [sym_comment] = STATE(598),
    [sym_groovy_doc] = STATE(598),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(599)] = {
    [sym_comment] = STATE(599),
    [sym_groovy_doc] = STATE(599),
    [sym_identifier] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1616),
    [anon_sym_import] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1614),
    [anon_sym_AT] = ACTIONS(1614),
    [anon_sym_assert] = ACTIONS(1614),
    [anon_sym_PLUS_PLUS] = ACTIONS(1616),
    [anon_sym_DASH_DASH] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_ATinterface] = ACTIONS(1616),
    [anon_sym_interface] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1614),
    [anon_sym_def] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_null] = ACTIONS(1614),
    [anon_sym_this] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [aux_sym__type_identifier_token1] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1616),
    [aux_sym_number_literal_token1] = ACTIONS(1614),
    [aux_sym_number_literal_token2] = ACTIONS(1616),
    [aux_sym_number_literal_token3] = ACTIONS(1616),
    [aux_sym_number_literal_token4] = ACTIONS(1614),
    [aux_sym_number_literal_token5] = ACTIONS(1616),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1614),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1616),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1616),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_case] = ACTIONS(1614),
    [anon_sym_default] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1614),
    [anon_sym_int] = ACTIONS(1614),
    [anon_sym_boolean] = ACTIONS(1614),
    [anon_sym_char] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_float] = ACTIONS(1614),
    [anon_sym_double] = ACTIONS(1614),
    [anon_sym_void] = ACTIONS(1614),
    [anon_sym_public] = ACTIONS(1614),
    [anon_sym_protected] = ACTIONS(1614),
    [anon_sym_private] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_final] = ACTIONS(1614),
    [anon_sym_synchronized] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
  },
  [STATE(600)] = {
    [sym_comment] = STATE(600),
    [sym_groovy_doc] = STATE(600),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(601)] = {
    [sym_comment] = STATE(601),
    [sym_groovy_doc] = STATE(601),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(602)] = {
    [sym_comment] = STATE(602),
    [sym_groovy_doc] = STATE(602),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(603)] = {
    [sym_comment] = STATE(603),
    [sym_groovy_doc] = STATE(603),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(604)] = {
    [sym_comment] = STATE(604),
    [sym_groovy_doc] = STATE(604),
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_pipeline] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(605)] = {
    [sym_comment] = STATE(605),
    [sym_groovy_doc] = STATE(605),
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_break] = ACTIONS(1626),
    [anon_sym_continue] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1628),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_PLUS_PLUS] = ACTIONS(1628),
    [anon_sym_DASH_DASH] = ACTIONS(1628),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_ATinterface] = ACTIONS(1628),
    [anon_sym_interface] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_null] = ACTIONS(1626),
    [anon_sym_this] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [aux_sym__type_identifier_token1] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1628),
    [aux_sym_number_literal_token1] = ACTIONS(1626),
    [aux_sym_number_literal_token2] = ACTIONS(1628),
    [aux_sym_number_literal_token3] = ACTIONS(1628),
    [aux_sym_number_literal_token4] = ACTIONS(1626),
    [aux_sym_number_literal_token5] = ACTIONS(1628),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1628),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1628),
    [anon_sym_switch] = ACTIONS(1626),
    [anon_sym_case] = ACTIONS(1626),
    [anon_sym_default] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_int] = ACTIONS(1626),
    [anon_sym_boolean] = ACTIONS(1626),
    [anon_sym_char] = ACTIONS(1626),
    [anon_sym_short] = ACTIONS(1626),
    [anon_sym_long] = ACTIONS(1626),
    [anon_sym_float] = ACTIONS(1626),
    [anon_sym_double] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_public] = ACTIONS(1626),
    [anon_sym_protected] = ACTIONS(1626),
    [anon_sym_private] = ACTIONS(1626),
    [anon_sym_static] = ACTIONS(1626),
    [anon_sym_final] = ACTIONS(1626),
    [anon_sym_synchronized] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1628),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
  },
  [STATE(606)] = {
    [sym_comment] = STATE(606),
    [sym_groovy_doc] = STATE(606),
    [ts_builtin_sym_end] = ACTIONS(1606),
    [sym_identifier] = ACTIONS(1608),
    [anon_sym_break] = ACTIONS(1608),
    [anon_sym_continue] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1606),
    [anon_sym_import] = ACTIONS(1608),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_assert] = ACTIONS(1608),
    [anon_sym_PLUS_PLUS] = ACTIONS(1606),
    [anon_sym_DASH_DASH] = ACTIONS(1606),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1608),
    [anon_sym_false] = ACTIONS(1608),
    [anon_sym_ATinterface] = ACTIONS(1606),
    [anon_sym_interface] = ACTIONS(1608),
    [anon_sym_class] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_def] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1606),
    [anon_sym_null] = ACTIONS(1608),
    [anon_sym_this] = ACTIONS(1608),
    [anon_sym_do] = ACTIONS(1608),
    [anon_sym_while] = ACTIONS(1608),
    [anon_sym_for] = ACTIONS(1608),
    [aux_sym__type_identifier_token1] = ACTIONS(1608),
    [anon_sym_if] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [aux_sym_number_literal_token1] = ACTIONS(1608),
    [aux_sym_number_literal_token2] = ACTIONS(1606),
    [aux_sym_number_literal_token3] = ACTIONS(1606),
    [aux_sym_number_literal_token4] = ACTIONS(1608),
    [aux_sym_number_literal_token5] = ACTIONS(1606),
    [anon_sym_pipeline] = ACTIONS(1608),
    [anon_sym_return] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1606),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1606),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1606),
    [anon_sym_switch] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1608),
    [anon_sym_int] = ACTIONS(1608),
    [anon_sym_boolean] = ACTIONS(1608),
    [anon_sym_char] = ACTIONS(1608),
    [anon_sym_short] = ACTIONS(1608),
    [anon_sym_long] = ACTIONS(1608),
    [anon_sym_float] = ACTIONS(1608),
    [anon_sym_double] = ACTIONS(1608),
    [anon_sym_void] = ACTIONS(1608),
    [anon_sym_public] = ACTIONS(1608),
    [anon_sym_protected] = ACTIONS(1608),
    [anon_sym_private] = ACTIONS(1608),
    [anon_sym_static] = ACTIONS(1608),
    [anon_sym_final] = ACTIONS(1608),
    [anon_sym_synchronized] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_new] = ACTIONS(1608),
  },
  [STATE(607)] = {
    [sym_comment] = STATE(607),
    [sym_groovy_doc] = STATE(607),
    [ts_builtin_sym_end] = ACTIONS(1612),
    [sym_identifier] = ACTIONS(1610),
    [anon_sym_break] = ACTIONS(1610),
    [anon_sym_continue] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1612),
    [anon_sym_import] = ACTIONS(1610),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_assert] = ACTIONS(1610),
    [anon_sym_PLUS_PLUS] = ACTIONS(1612),
    [anon_sym_DASH_DASH] = ACTIONS(1612),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_ATinterface] = ACTIONS(1612),
    [anon_sym_interface] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_def] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_null] = ACTIONS(1610),
    [anon_sym_this] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1610),
    [anon_sym_for] = ACTIONS(1610),
    [aux_sym__type_identifier_token1] = ACTIONS(1610),
    [anon_sym_if] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1612),
    [aux_sym_number_literal_token1] = ACTIONS(1610),
    [aux_sym_number_literal_token2] = ACTIONS(1612),
    [aux_sym_number_literal_token3] = ACTIONS(1612),
    [aux_sym_number_literal_token4] = ACTIONS(1610),
    [aux_sym_number_literal_token5] = ACTIONS(1612),
    [anon_sym_pipeline] = ACTIONS(1610),
    [anon_sym_return] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1612),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1612),
    [anon_sym_switch] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1610),
    [anon_sym_int] = ACTIONS(1610),
    [anon_sym_boolean] = ACTIONS(1610),
    [anon_sym_char] = ACTIONS(1610),
    [anon_sym_short] = ACTIONS(1610),
    [anon_sym_long] = ACTIONS(1610),
    [anon_sym_float] = ACTIONS(1610),
    [anon_sym_double] = ACTIONS(1610),
    [anon_sym_void] = ACTIONS(1610),
    [anon_sym_public] = ACTIONS(1610),
    [anon_sym_protected] = ACTIONS(1610),
    [anon_sym_private] = ACTIONS(1610),
    [anon_sym_static] = ACTIONS(1610),
    [anon_sym_final] = ACTIONS(1610),
    [anon_sym_synchronized] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
  },
  [STATE(608)] = {
    [sym_comment] = STATE(608),
    [sym_groovy_doc] = STATE(608),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1686),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(609)] = {
    [sym_comment] = STATE(609),
    [sym_groovy_doc] = STATE(609),
    [ts_builtin_sym_end] = ACTIONS(1600),
    [sym_identifier] = ACTIONS(1598),
    [anon_sym_break] = ACTIONS(1598),
    [anon_sym_continue] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1600),
    [anon_sym_import] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1598),
    [anon_sym_AT] = ACTIONS(1598),
    [anon_sym_assert] = ACTIONS(1598),
    [anon_sym_PLUS_PLUS] = ACTIONS(1600),
    [anon_sym_DASH_DASH] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_ATinterface] = ACTIONS(1600),
    [anon_sym_interface] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1598),
    [anon_sym_def] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_null] = ACTIONS(1598),
    [anon_sym_this] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_while] = ACTIONS(1598),
    [anon_sym_for] = ACTIONS(1598),
    [aux_sym__type_identifier_token1] = ACTIONS(1598),
    [anon_sym_if] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1600),
    [aux_sym_number_literal_token1] = ACTIONS(1598),
    [aux_sym_number_literal_token2] = ACTIONS(1600),
    [aux_sym_number_literal_token3] = ACTIONS(1600),
    [aux_sym_number_literal_token4] = ACTIONS(1598),
    [aux_sym_number_literal_token5] = ACTIONS(1600),
    [anon_sym_pipeline] = ACTIONS(1598),
    [anon_sym_return] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1598),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1598),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1600),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1600),
    [anon_sym_switch] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1598),
    [anon_sym_int] = ACTIONS(1598),
    [anon_sym_boolean] = ACTIONS(1598),
    [anon_sym_char] = ACTIONS(1598),
    [anon_sym_short] = ACTIONS(1598),
    [anon_sym_long] = ACTIONS(1598),
    [anon_sym_float] = ACTIONS(1598),
    [anon_sym_double] = ACTIONS(1598),
    [anon_sym_void] = ACTIONS(1598),
    [anon_sym_public] = ACTIONS(1598),
    [anon_sym_protected] = ACTIONS(1598),
    [anon_sym_private] = ACTIONS(1598),
    [anon_sym_static] = ACTIONS(1598),
    [anon_sym_final] = ACTIONS(1598),
    [anon_sym_synchronized] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
  },
  [STATE(610)] = {
    [sym_comment] = STATE(610),
    [sym_groovy_doc] = STATE(610),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(611)] = {
    [sym_comment] = STATE(611),
    [sym_groovy_doc] = STATE(611),
    [sym_identifier] = ACTIONS(1630),
    [anon_sym_break] = ACTIONS(1630),
    [anon_sym_continue] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1632),
    [anon_sym_import] = ACTIONS(1630),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_assert] = ACTIONS(1630),
    [anon_sym_PLUS_PLUS] = ACTIONS(1632),
    [anon_sym_DASH_DASH] = ACTIONS(1632),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_ATinterface] = ACTIONS(1632),
    [anon_sym_interface] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_def] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_null] = ACTIONS(1630),
    [anon_sym_this] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_while] = ACTIONS(1630),
    [anon_sym_for] = ACTIONS(1630),
    [aux_sym__type_identifier_token1] = ACTIONS(1630),
    [anon_sym_if] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1632),
    [aux_sym_number_literal_token1] = ACTIONS(1630),
    [aux_sym_number_literal_token2] = ACTIONS(1632),
    [aux_sym_number_literal_token3] = ACTIONS(1632),
    [aux_sym_number_literal_token4] = ACTIONS(1630),
    [aux_sym_number_literal_token5] = ACTIONS(1632),
    [anon_sym_return] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1632),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1632),
    [anon_sym_switch] = ACTIONS(1630),
    [anon_sym_case] = ACTIONS(1630),
    [anon_sym_default] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1630),
    [anon_sym_int] = ACTIONS(1630),
    [anon_sym_boolean] = ACTIONS(1630),
    [anon_sym_char] = ACTIONS(1630),
    [anon_sym_short] = ACTIONS(1630),
    [anon_sym_long] = ACTIONS(1630),
    [anon_sym_float] = ACTIONS(1630),
    [anon_sym_double] = ACTIONS(1630),
    [anon_sym_void] = ACTIONS(1630),
    [anon_sym_public] = ACTIONS(1630),
    [anon_sym_protected] = ACTIONS(1630),
    [anon_sym_private] = ACTIONS(1630),
    [anon_sym_static] = ACTIONS(1630),
    [anon_sym_final] = ACTIONS(1630),
    [anon_sym_synchronized] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1632),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
  },
  [STATE(612)] = {
    [sym_comment] = STATE(612),
    [sym_groovy_doc] = STATE(612),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(613)] = {
    [sym_comment] = STATE(613),
    [sym_groovy_doc] = STATE(613),
    [ts_builtin_sym_end] = ACTIONS(1616),
    [sym_identifier] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1616),
    [anon_sym_import] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1614),
    [anon_sym_AT] = ACTIONS(1614),
    [anon_sym_assert] = ACTIONS(1614),
    [anon_sym_PLUS_PLUS] = ACTIONS(1616),
    [anon_sym_DASH_DASH] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_ATinterface] = ACTIONS(1616),
    [anon_sym_interface] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1614),
    [anon_sym_def] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_null] = ACTIONS(1614),
    [anon_sym_this] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [aux_sym__type_identifier_token1] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_else] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1616),
    [aux_sym_number_literal_token1] = ACTIONS(1614),
    [aux_sym_number_literal_token2] = ACTIONS(1616),
    [aux_sym_number_literal_token3] = ACTIONS(1616),
    [aux_sym_number_literal_token4] = ACTIONS(1614),
    [aux_sym_number_literal_token5] = ACTIONS(1616),
    [anon_sym_pipeline] = ACTIONS(1614),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1614),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1616),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1616),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1614),
    [anon_sym_int] = ACTIONS(1614),
    [anon_sym_boolean] = ACTIONS(1614),
    [anon_sym_char] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_float] = ACTIONS(1614),
    [anon_sym_double] = ACTIONS(1614),
    [anon_sym_void] = ACTIONS(1614),
    [anon_sym_public] = ACTIONS(1614),
    [anon_sym_protected] = ACTIONS(1614),
    [anon_sym_private] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_final] = ACTIONS(1614),
    [anon_sym_synchronized] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
  },
  [STATE(614)] = {
    [sym_comment] = STATE(614),
    [sym_groovy_doc] = STATE(614),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(615)] = {
    [sym_comment] = STATE(615),
    [sym_groovy_doc] = STATE(615),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1692),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(102),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(97),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(97),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(102),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(97),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(616)] = {
    [sym_comment] = STATE(616),
    [sym_groovy_doc] = STATE(616),
    [sym_identifier] = ACTIONS(1634),
    [anon_sym_break] = ACTIONS(1634),
    [anon_sym_continue] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1636),
    [anon_sym_import] = ACTIONS(1634),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1634),
    [anon_sym_AT] = ACTIONS(1634),
    [anon_sym_assert] = ACTIONS(1634),
    [anon_sym_PLUS_PLUS] = ACTIONS(1636),
    [anon_sym_DASH_DASH] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_ATinterface] = ACTIONS(1636),
    [anon_sym_interface] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1634),
    [anon_sym_def] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_null] = ACTIONS(1634),
    [anon_sym_this] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_while] = ACTIONS(1634),
    [anon_sym_for] = ACTIONS(1634),
    [aux_sym__type_identifier_token1] = ACTIONS(1634),
    [anon_sym_if] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1636),
    [aux_sym_number_literal_token1] = ACTIONS(1634),
    [aux_sym_number_literal_token2] = ACTIONS(1636),
    [aux_sym_number_literal_token3] = ACTIONS(1636),
    [aux_sym_number_literal_token4] = ACTIONS(1634),
    [aux_sym_number_literal_token5] = ACTIONS(1636),
    [anon_sym_return] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1634),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1634),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1636),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1636),
    [anon_sym_switch] = ACTIONS(1634),
    [anon_sym_case] = ACTIONS(1634),
    [anon_sym_default] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1634),
    [anon_sym_int] = ACTIONS(1634),
    [anon_sym_boolean] = ACTIONS(1634),
    [anon_sym_char] = ACTIONS(1634),
    [anon_sym_short] = ACTIONS(1634),
    [anon_sym_long] = ACTIONS(1634),
    [anon_sym_float] = ACTIONS(1634),
    [anon_sym_double] = ACTIONS(1634),
    [anon_sym_void] = ACTIONS(1634),
    [anon_sym_public] = ACTIONS(1634),
    [anon_sym_protected] = ACTIONS(1634),
    [anon_sym_private] = ACTIONS(1634),
    [anon_sym_static] = ACTIONS(1634),
    [anon_sym_final] = ACTIONS(1634),
    [anon_sym_synchronized] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
  },
  [STATE(617)] = {
    [sym_comment] = STATE(617),
    [sym_groovy_doc] = STATE(617),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(618)] = {
    [sym_comment] = STATE(618),
    [sym_groovy_doc] = STATE(618),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(619)] = {
    [sym_comment] = STATE(619),
    [sym_groovy_doc] = STATE(619),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(620)] = {
    [sym_comment] = STATE(620),
    [sym_groovy_doc] = STATE(620),
    [sym_identifier] = ACTIONS(1638),
    [anon_sym_break] = ACTIONS(1638),
    [anon_sym_continue] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1640),
    [anon_sym_import] = ACTIONS(1638),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1638),
    [anon_sym_AT] = ACTIONS(1638),
    [anon_sym_assert] = ACTIONS(1638),
    [anon_sym_PLUS_PLUS] = ACTIONS(1640),
    [anon_sym_DASH_DASH] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_ATinterface] = ACTIONS(1640),
    [anon_sym_interface] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1638),
    [anon_sym_def] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_null] = ACTIONS(1638),
    [anon_sym_this] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_while] = ACTIONS(1638),
    [anon_sym_for] = ACTIONS(1638),
    [aux_sym__type_identifier_token1] = ACTIONS(1638),
    [anon_sym_if] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1640),
    [aux_sym_number_literal_token1] = ACTIONS(1638),
    [aux_sym_number_literal_token2] = ACTIONS(1640),
    [aux_sym_number_literal_token3] = ACTIONS(1640),
    [aux_sym_number_literal_token4] = ACTIONS(1638),
    [aux_sym_number_literal_token5] = ACTIONS(1640),
    [anon_sym_return] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1638),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1638),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1640),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1640),
    [anon_sym_switch] = ACTIONS(1638),
    [anon_sym_case] = ACTIONS(1638),
    [anon_sym_default] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1638),
    [anon_sym_int] = ACTIONS(1638),
    [anon_sym_boolean] = ACTIONS(1638),
    [anon_sym_char] = ACTIONS(1638),
    [anon_sym_short] = ACTIONS(1638),
    [anon_sym_long] = ACTIONS(1638),
    [anon_sym_float] = ACTIONS(1638),
    [anon_sym_double] = ACTIONS(1638),
    [anon_sym_void] = ACTIONS(1638),
    [anon_sym_public] = ACTIONS(1638),
    [anon_sym_protected] = ACTIONS(1638),
    [anon_sym_private] = ACTIONS(1638),
    [anon_sym_static] = ACTIONS(1638),
    [anon_sym_final] = ACTIONS(1638),
    [anon_sym_synchronized] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
  },
  [STATE(621)] = {
    [sym_comment] = STATE(621),
    [sym_groovy_doc] = STATE(621),
    [ts_builtin_sym_end] = ACTIONS(1628),
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_break] = ACTIONS(1626),
    [anon_sym_continue] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1628),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1628),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_PLUS_PLUS] = ACTIONS(1628),
    [anon_sym_DASH_DASH] = ACTIONS(1628),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1626),
    [anon_sym_false] = ACTIONS(1626),
    [anon_sym_ATinterface] = ACTIONS(1628),
    [anon_sym_interface] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1628),
    [anon_sym_RBRACE] = ACTIONS(1628),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1628),
    [anon_sym_null] = ACTIONS(1626),
    [anon_sym_this] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [aux_sym__type_identifier_token1] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_else] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1628),
    [aux_sym_number_literal_token1] = ACTIONS(1626),
    [aux_sym_number_literal_token2] = ACTIONS(1628),
    [aux_sym_number_literal_token3] = ACTIONS(1628),
    [aux_sym_number_literal_token4] = ACTIONS(1626),
    [aux_sym_number_literal_token5] = ACTIONS(1628),
    [anon_sym_pipeline] = ACTIONS(1626),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_SQUOTE] = ACTIONS(1626),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1628),
    [anon_sym_DQUOTE] = ACTIONS(1626),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1628),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1628),
    [anon_sym_switch] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_int] = ACTIONS(1626),
    [anon_sym_boolean] = ACTIONS(1626),
    [anon_sym_char] = ACTIONS(1626),
    [anon_sym_short] = ACTIONS(1626),
    [anon_sym_long] = ACTIONS(1626),
    [anon_sym_float] = ACTIONS(1626),
    [anon_sym_double] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_public] = ACTIONS(1626),
    [anon_sym_protected] = ACTIONS(1626),
    [anon_sym_private] = ACTIONS(1626),
    [anon_sym_static] = ACTIONS(1626),
    [anon_sym_final] = ACTIONS(1626),
    [anon_sym_synchronized] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1628),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
  },
  [STATE(622)] = {
    [sym_comment] = STATE(622),
    [sym_groovy_doc] = STATE(622),
    [ts_builtin_sym_end] = ACTIONS(1632),
    [sym_identifier] = ACTIONS(1630),
    [anon_sym_break] = ACTIONS(1630),
    [anon_sym_continue] = ACTIONS(1630),
    [anon_sym_SEMI] = ACTIONS(1632),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1632),
    [anon_sym_import] = ACTIONS(1630),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1630),
    [anon_sym_AT] = ACTIONS(1630),
    [anon_sym_assert] = ACTIONS(1630),
    [anon_sym_PLUS_PLUS] = ACTIONS(1632),
    [anon_sym_DASH_DASH] = ACTIONS(1632),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1630),
    [anon_sym_false] = ACTIONS(1630),
    [anon_sym_ATinterface] = ACTIONS(1632),
    [anon_sym_interface] = ACTIONS(1630),
    [anon_sym_class] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1632),
    [anon_sym_RBRACE] = ACTIONS(1632),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1630),
    [anon_sym_def] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1632),
    [anon_sym_null] = ACTIONS(1630),
    [anon_sym_this] = ACTIONS(1630),
    [anon_sym_do] = ACTIONS(1630),
    [anon_sym_while] = ACTIONS(1630),
    [anon_sym_for] = ACTIONS(1630),
    [aux_sym__type_identifier_token1] = ACTIONS(1630),
    [anon_sym_if] = ACTIONS(1630),
    [anon_sym_else] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1632),
    [aux_sym_number_literal_token1] = ACTIONS(1630),
    [aux_sym_number_literal_token2] = ACTIONS(1632),
    [aux_sym_number_literal_token3] = ACTIONS(1632),
    [aux_sym_number_literal_token4] = ACTIONS(1630),
    [aux_sym_number_literal_token5] = ACTIONS(1632),
    [anon_sym_pipeline] = ACTIONS(1630),
    [anon_sym_return] = ACTIONS(1630),
    [anon_sym_SQUOTE] = ACTIONS(1630),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1632),
    [anon_sym_DQUOTE] = ACTIONS(1630),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1632),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1632),
    [anon_sym_switch] = ACTIONS(1630),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1630),
    [anon_sym_int] = ACTIONS(1630),
    [anon_sym_boolean] = ACTIONS(1630),
    [anon_sym_char] = ACTIONS(1630),
    [anon_sym_short] = ACTIONS(1630),
    [anon_sym_long] = ACTIONS(1630),
    [anon_sym_float] = ACTIONS(1630),
    [anon_sym_double] = ACTIONS(1630),
    [anon_sym_void] = ACTIONS(1630),
    [anon_sym_public] = ACTIONS(1630),
    [anon_sym_protected] = ACTIONS(1630),
    [anon_sym_private] = ACTIONS(1630),
    [anon_sym_static] = ACTIONS(1630),
    [anon_sym_final] = ACTIONS(1630),
    [anon_sym_synchronized] = ACTIONS(1630),
    [anon_sym_TILDE] = ACTIONS(1632),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_new] = ACTIONS(1630),
  },
  [STATE(623)] = {
    [sym_comment] = STATE(623),
    [sym_groovy_doc] = STATE(623),
    [sym_identifier] = ACTIONS(1618),
    [anon_sym_break] = ACTIONS(1618),
    [anon_sym_continue] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1620),
    [anon_sym_import] = ACTIONS(1618),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_assert] = ACTIONS(1618),
    [anon_sym_PLUS_PLUS] = ACTIONS(1620),
    [anon_sym_DASH_DASH] = ACTIONS(1620),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_ATinterface] = ACTIONS(1620),
    [anon_sym_interface] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_def] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_null] = ACTIONS(1618),
    [anon_sym_this] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_while] = ACTIONS(1618),
    [anon_sym_for] = ACTIONS(1618),
    [aux_sym__type_identifier_token1] = ACTIONS(1618),
    [anon_sym_if] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1620),
    [aux_sym_number_literal_token1] = ACTIONS(1618),
    [aux_sym_number_literal_token2] = ACTIONS(1620),
    [aux_sym_number_literal_token3] = ACTIONS(1620),
    [aux_sym_number_literal_token4] = ACTIONS(1618),
    [aux_sym_number_literal_token5] = ACTIONS(1620),
    [anon_sym_return] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1620),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1620),
    [anon_sym_switch] = ACTIONS(1618),
    [anon_sym_case] = ACTIONS(1618),
    [anon_sym_default] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1618),
    [anon_sym_int] = ACTIONS(1618),
    [anon_sym_boolean] = ACTIONS(1618),
    [anon_sym_char] = ACTIONS(1618),
    [anon_sym_short] = ACTIONS(1618),
    [anon_sym_long] = ACTIONS(1618),
    [anon_sym_float] = ACTIONS(1618),
    [anon_sym_double] = ACTIONS(1618),
    [anon_sym_void] = ACTIONS(1618),
    [anon_sym_public] = ACTIONS(1618),
    [anon_sym_protected] = ACTIONS(1618),
    [anon_sym_private] = ACTIONS(1618),
    [anon_sym_static] = ACTIONS(1618),
    [anon_sym_final] = ACTIONS(1618),
    [anon_sym_synchronized] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1620),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
  },
  [STATE(624)] = {
    [sym_comment] = STATE(624),
    [sym_groovy_doc] = STATE(624),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(625)] = {
    [sym_comment] = STATE(625),
    [sym_groovy_doc] = STATE(625),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1310),
    [anon_sym_STAR_COLON] = ACTIONS(1308),
    [anon_sym_import] = ACTIONS(1310),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1310),
    [anon_sym_AT] = ACTIONS(1310),
    [anon_sym_assert] = ACTIONS(1310),
    [anon_sym_PLUS_PLUS] = ACTIONS(1308),
    [anon_sym_DASH_DASH] = ACTIONS(1308),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1310),
    [anon_sym_PLUS] = ACTIONS(1310),
    [anon_sym_DASH] = ACTIONS(1310),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [anon_sym_ATinterface] = ACTIONS(1308),
    [anon_sym_interface] = ACTIONS(1310),
    [anon_sym_class] = ACTIONS(1310),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1310),
    [anon_sym_def] = ACTIONS(1310),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_null] = ACTIONS(1310),
    [anon_sym_this] = ACTIONS(1310),
    [anon_sym_do] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [aux_sym__type_identifier_token1] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_else] = ACTIONS(1310),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [aux_sym_number_literal_token1] = ACTIONS(1310),
    [aux_sym_number_literal_token2] = ACTIONS(1308),
    [aux_sym_number_literal_token3] = ACTIONS(1308),
    [aux_sym_number_literal_token4] = ACTIONS(1310),
    [aux_sym_number_literal_token5] = ACTIONS(1308),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1308),
    [anon_sym_DQUOTE] = ACTIONS(1310),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1308),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1308),
    [anon_sym_switch] = ACTIONS(1310),
    [anon_sym_case] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1310),
    [anon_sym_int] = ACTIONS(1310),
    [anon_sym_boolean] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_short] = ACTIONS(1310),
    [anon_sym_long] = ACTIONS(1310),
    [anon_sym_float] = ACTIONS(1310),
    [anon_sym_double] = ACTIONS(1310),
    [anon_sym_void] = ACTIONS(1310),
    [anon_sym_public] = ACTIONS(1310),
    [anon_sym_protected] = ACTIONS(1310),
    [anon_sym_private] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_final] = ACTIONS(1310),
    [anon_sym_synchronized] = ACTIONS(1310),
    [anon_sym_TILDE] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1310),
    [anon_sym_new] = ACTIONS(1310),
  },
  [STATE(626)] = {
    [sym_comment] = STATE(626),
    [sym_groovy_doc] = STATE(626),
    [ts_builtin_sym_end] = ACTIONS(1620),
    [sym_identifier] = ACTIONS(1618),
    [anon_sym_break] = ACTIONS(1618),
    [anon_sym_continue] = ACTIONS(1618),
    [anon_sym_SEMI] = ACTIONS(1620),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1620),
    [anon_sym_import] = ACTIONS(1618),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1618),
    [anon_sym_AT] = ACTIONS(1618),
    [anon_sym_assert] = ACTIONS(1618),
    [anon_sym_PLUS_PLUS] = ACTIONS(1620),
    [anon_sym_DASH_DASH] = ACTIONS(1620),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1618),
    [anon_sym_false] = ACTIONS(1618),
    [anon_sym_ATinterface] = ACTIONS(1620),
    [anon_sym_interface] = ACTIONS(1618),
    [anon_sym_class] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1620),
    [anon_sym_RBRACE] = ACTIONS(1620),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1618),
    [anon_sym_def] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1620),
    [anon_sym_null] = ACTIONS(1618),
    [anon_sym_this] = ACTIONS(1618),
    [anon_sym_do] = ACTIONS(1618),
    [anon_sym_while] = ACTIONS(1618),
    [anon_sym_for] = ACTIONS(1618),
    [aux_sym__type_identifier_token1] = ACTIONS(1618),
    [anon_sym_if] = ACTIONS(1618),
    [anon_sym_else] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1620),
    [aux_sym_number_literal_token1] = ACTIONS(1618),
    [aux_sym_number_literal_token2] = ACTIONS(1620),
    [aux_sym_number_literal_token3] = ACTIONS(1620),
    [aux_sym_number_literal_token4] = ACTIONS(1618),
    [aux_sym_number_literal_token5] = ACTIONS(1620),
    [anon_sym_pipeline] = ACTIONS(1618),
    [anon_sym_return] = ACTIONS(1618),
    [anon_sym_SQUOTE] = ACTIONS(1618),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1620),
    [anon_sym_DQUOTE] = ACTIONS(1618),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1620),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1620),
    [anon_sym_switch] = ACTIONS(1618),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1618),
    [anon_sym_int] = ACTIONS(1618),
    [anon_sym_boolean] = ACTIONS(1618),
    [anon_sym_char] = ACTIONS(1618),
    [anon_sym_short] = ACTIONS(1618),
    [anon_sym_long] = ACTIONS(1618),
    [anon_sym_float] = ACTIONS(1618),
    [anon_sym_double] = ACTIONS(1618),
    [anon_sym_void] = ACTIONS(1618),
    [anon_sym_public] = ACTIONS(1618),
    [anon_sym_protected] = ACTIONS(1618),
    [anon_sym_private] = ACTIONS(1618),
    [anon_sym_static] = ACTIONS(1618),
    [anon_sym_final] = ACTIONS(1618),
    [anon_sym_synchronized] = ACTIONS(1618),
    [anon_sym_TILDE] = ACTIONS(1620),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_new] = ACTIONS(1618),
  },
  [STATE(627)] = {
    [sym_comment] = STATE(627),
    [sym_groovy_doc] = STATE(627),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(628)] = {
    [sym_comment] = STATE(628),
    [sym_groovy_doc] = STATE(628),
    [ts_builtin_sym_end] = ACTIONS(1624),
    [sym_identifier] = ACTIONS(1622),
    [anon_sym_break] = ACTIONS(1622),
    [anon_sym_continue] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1624),
    [anon_sym_import] = ACTIONS(1622),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_assert] = ACTIONS(1622),
    [anon_sym_PLUS_PLUS] = ACTIONS(1624),
    [anon_sym_DASH_DASH] = ACTIONS(1624),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_ATinterface] = ACTIONS(1624),
    [anon_sym_interface] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_def] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_null] = ACTIONS(1622),
    [anon_sym_this] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_while] = ACTIONS(1622),
    [anon_sym_for] = ACTIONS(1622),
    [aux_sym__type_identifier_token1] = ACTIONS(1622),
    [anon_sym_if] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1624),
    [aux_sym_number_literal_token1] = ACTIONS(1622),
    [aux_sym_number_literal_token2] = ACTIONS(1624),
    [aux_sym_number_literal_token3] = ACTIONS(1624),
    [aux_sym_number_literal_token4] = ACTIONS(1622),
    [aux_sym_number_literal_token5] = ACTIONS(1624),
    [anon_sym_pipeline] = ACTIONS(1622),
    [anon_sym_return] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1624),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1624),
    [anon_sym_switch] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1622),
    [anon_sym_int] = ACTIONS(1622),
    [anon_sym_boolean] = ACTIONS(1622),
    [anon_sym_char] = ACTIONS(1622),
    [anon_sym_short] = ACTIONS(1622),
    [anon_sym_long] = ACTIONS(1622),
    [anon_sym_float] = ACTIONS(1622),
    [anon_sym_double] = ACTIONS(1622),
    [anon_sym_void] = ACTIONS(1622),
    [anon_sym_public] = ACTIONS(1622),
    [anon_sym_protected] = ACTIONS(1622),
    [anon_sym_private] = ACTIONS(1622),
    [anon_sym_static] = ACTIONS(1622),
    [anon_sym_final] = ACTIONS(1622),
    [anon_sym_synchronized] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1624),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
  },
  [STATE(629)] = {
    [sym_comment] = STATE(629),
    [sym_groovy_doc] = STATE(629),
    [sym_identifier] = ACTIONS(1604),
    [anon_sym_break] = ACTIONS(1604),
    [anon_sym_continue] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1604),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1604),
    [anon_sym_AT] = ACTIONS(1604),
    [anon_sym_assert] = ACTIONS(1604),
    [anon_sym_PLUS_PLUS] = ACTIONS(1602),
    [anon_sym_DASH_DASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1604),
    [anon_sym_false] = ACTIONS(1604),
    [anon_sym_ATinterface] = ACTIONS(1602),
    [anon_sym_interface] = ACTIONS(1604),
    [anon_sym_class] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1604),
    [anon_sym_def] = ACTIONS(1604),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_null] = ACTIONS(1604),
    [anon_sym_this] = ACTIONS(1604),
    [anon_sym_do] = ACTIONS(1604),
    [anon_sym_while] = ACTIONS(1604),
    [anon_sym_for] = ACTIONS(1604),
    [aux_sym__type_identifier_token1] = ACTIONS(1604),
    [anon_sym_if] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [aux_sym_number_literal_token1] = ACTIONS(1604),
    [aux_sym_number_literal_token2] = ACTIONS(1602),
    [aux_sym_number_literal_token3] = ACTIONS(1602),
    [aux_sym_number_literal_token4] = ACTIONS(1604),
    [aux_sym_number_literal_token5] = ACTIONS(1602),
    [anon_sym_return] = ACTIONS(1604),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1602),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1602),
    [anon_sym_switch] = ACTIONS(1604),
    [anon_sym_case] = ACTIONS(1604),
    [anon_sym_default] = ACTIONS(1604),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1604),
    [anon_sym_int] = ACTIONS(1604),
    [anon_sym_boolean] = ACTIONS(1604),
    [anon_sym_char] = ACTIONS(1604),
    [anon_sym_short] = ACTIONS(1604),
    [anon_sym_long] = ACTIONS(1604),
    [anon_sym_float] = ACTIONS(1604),
    [anon_sym_double] = ACTIONS(1604),
    [anon_sym_void] = ACTIONS(1604),
    [anon_sym_public] = ACTIONS(1604),
    [anon_sym_protected] = ACTIONS(1604),
    [anon_sym_private] = ACTIONS(1604),
    [anon_sym_static] = ACTIONS(1604),
    [anon_sym_final] = ACTIONS(1604),
    [anon_sym_synchronized] = ACTIONS(1604),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1604),
    [anon_sym_new] = ACTIONS(1604),
  },
  [STATE(630)] = {
    [sym_comment] = STATE(630),
    [sym_groovy_doc] = STATE(630),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(631)] = {
    [sym_comment] = STATE(631),
    [sym_groovy_doc] = STATE(631),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(632)] = {
    [sym_comment] = STATE(632),
    [sym_groovy_doc] = STATE(632),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(633)] = {
    [sym_comment] = STATE(633),
    [sym_groovy_doc] = STATE(633),
    [ts_builtin_sym_end] = ACTIONS(1636),
    [sym_identifier] = ACTIONS(1634),
    [anon_sym_break] = ACTIONS(1634),
    [anon_sym_continue] = ACTIONS(1634),
    [anon_sym_SEMI] = ACTIONS(1636),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1636),
    [anon_sym_import] = ACTIONS(1634),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1634),
    [anon_sym_AT] = ACTIONS(1634),
    [anon_sym_assert] = ACTIONS(1634),
    [anon_sym_PLUS_PLUS] = ACTIONS(1636),
    [anon_sym_DASH_DASH] = ACTIONS(1636),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1634),
    [anon_sym_false] = ACTIONS(1634),
    [anon_sym_ATinterface] = ACTIONS(1636),
    [anon_sym_interface] = ACTIONS(1634),
    [anon_sym_class] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1636),
    [anon_sym_RBRACE] = ACTIONS(1636),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1634),
    [anon_sym_def] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1636),
    [anon_sym_null] = ACTIONS(1634),
    [anon_sym_this] = ACTIONS(1634),
    [anon_sym_do] = ACTIONS(1634),
    [anon_sym_while] = ACTIONS(1634),
    [anon_sym_for] = ACTIONS(1634),
    [aux_sym__type_identifier_token1] = ACTIONS(1634),
    [anon_sym_if] = ACTIONS(1634),
    [anon_sym_else] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1636),
    [aux_sym_number_literal_token1] = ACTIONS(1634),
    [aux_sym_number_literal_token2] = ACTIONS(1636),
    [aux_sym_number_literal_token3] = ACTIONS(1636),
    [aux_sym_number_literal_token4] = ACTIONS(1634),
    [aux_sym_number_literal_token5] = ACTIONS(1636),
    [anon_sym_pipeline] = ACTIONS(1634),
    [anon_sym_return] = ACTIONS(1634),
    [anon_sym_SQUOTE] = ACTIONS(1634),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1636),
    [anon_sym_DQUOTE] = ACTIONS(1634),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1636),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1636),
    [anon_sym_switch] = ACTIONS(1634),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1634),
    [anon_sym_int] = ACTIONS(1634),
    [anon_sym_boolean] = ACTIONS(1634),
    [anon_sym_char] = ACTIONS(1634),
    [anon_sym_short] = ACTIONS(1634),
    [anon_sym_long] = ACTIONS(1634),
    [anon_sym_float] = ACTIONS(1634),
    [anon_sym_double] = ACTIONS(1634),
    [anon_sym_void] = ACTIONS(1634),
    [anon_sym_public] = ACTIONS(1634),
    [anon_sym_protected] = ACTIONS(1634),
    [anon_sym_private] = ACTIONS(1634),
    [anon_sym_static] = ACTIONS(1634),
    [anon_sym_final] = ACTIONS(1634),
    [anon_sym_synchronized] = ACTIONS(1634),
    [anon_sym_TILDE] = ACTIONS(1636),
    [anon_sym_BANG] = ACTIONS(1634),
    [anon_sym_new] = ACTIONS(1634),
  },
  [STATE(634)] = {
    [sym_comment] = STATE(634),
    [sym_groovy_doc] = STATE(634),
    [sym_identifier] = ACTIONS(1608),
    [anon_sym_break] = ACTIONS(1608),
    [anon_sym_continue] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1606),
    [anon_sym_import] = ACTIONS(1608),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1608),
    [anon_sym_AT] = ACTIONS(1608),
    [anon_sym_assert] = ACTIONS(1608),
    [anon_sym_PLUS_PLUS] = ACTIONS(1606),
    [anon_sym_DASH_DASH] = ACTIONS(1606),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1608),
    [anon_sym_false] = ACTIONS(1608),
    [anon_sym_ATinterface] = ACTIONS(1606),
    [anon_sym_interface] = ACTIONS(1608),
    [anon_sym_class] = ACTIONS(1608),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1608),
    [anon_sym_def] = ACTIONS(1608),
    [anon_sym_LPAREN] = ACTIONS(1606),
    [anon_sym_null] = ACTIONS(1608),
    [anon_sym_this] = ACTIONS(1608),
    [anon_sym_do] = ACTIONS(1608),
    [anon_sym_while] = ACTIONS(1608),
    [anon_sym_for] = ACTIONS(1608),
    [aux_sym__type_identifier_token1] = ACTIONS(1608),
    [anon_sym_if] = ACTIONS(1608),
    [anon_sym_else] = ACTIONS(1608),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [aux_sym_number_literal_token1] = ACTIONS(1608),
    [aux_sym_number_literal_token2] = ACTIONS(1606),
    [aux_sym_number_literal_token3] = ACTIONS(1606),
    [aux_sym_number_literal_token4] = ACTIONS(1608),
    [aux_sym_number_literal_token5] = ACTIONS(1606),
    [anon_sym_return] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1606),
    [anon_sym_DQUOTE] = ACTIONS(1608),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1606),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1606),
    [anon_sym_switch] = ACTIONS(1608),
    [anon_sym_case] = ACTIONS(1608),
    [anon_sym_default] = ACTIONS(1608),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1608),
    [anon_sym_int] = ACTIONS(1608),
    [anon_sym_boolean] = ACTIONS(1608),
    [anon_sym_char] = ACTIONS(1608),
    [anon_sym_short] = ACTIONS(1608),
    [anon_sym_long] = ACTIONS(1608),
    [anon_sym_float] = ACTIONS(1608),
    [anon_sym_double] = ACTIONS(1608),
    [anon_sym_void] = ACTIONS(1608),
    [anon_sym_public] = ACTIONS(1608),
    [anon_sym_protected] = ACTIONS(1608),
    [anon_sym_private] = ACTIONS(1608),
    [anon_sym_static] = ACTIONS(1608),
    [anon_sym_final] = ACTIONS(1608),
    [anon_sym_synchronized] = ACTIONS(1608),
    [anon_sym_TILDE] = ACTIONS(1606),
    [anon_sym_BANG] = ACTIONS(1608),
    [anon_sym_new] = ACTIONS(1608),
  },
  [STATE(635)] = {
    [sym_comment] = STATE(635),
    [sym_groovy_doc] = STATE(635),
    [sym_identifier] = ACTIONS(1622),
    [anon_sym_break] = ACTIONS(1622),
    [anon_sym_continue] = ACTIONS(1622),
    [anon_sym_SEMI] = ACTIONS(1624),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1624),
    [anon_sym_import] = ACTIONS(1622),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1622),
    [anon_sym_AT] = ACTIONS(1622),
    [anon_sym_assert] = ACTIONS(1622),
    [anon_sym_PLUS_PLUS] = ACTIONS(1624),
    [anon_sym_DASH_DASH] = ACTIONS(1624),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1622),
    [anon_sym_false] = ACTIONS(1622),
    [anon_sym_ATinterface] = ACTIONS(1624),
    [anon_sym_interface] = ACTIONS(1622),
    [anon_sym_class] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1624),
    [anon_sym_RBRACE] = ACTIONS(1624),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1622),
    [anon_sym_def] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1624),
    [anon_sym_null] = ACTIONS(1622),
    [anon_sym_this] = ACTIONS(1622),
    [anon_sym_do] = ACTIONS(1622),
    [anon_sym_while] = ACTIONS(1622),
    [anon_sym_for] = ACTIONS(1622),
    [aux_sym__type_identifier_token1] = ACTIONS(1622),
    [anon_sym_if] = ACTIONS(1622),
    [anon_sym_else] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1624),
    [aux_sym_number_literal_token1] = ACTIONS(1622),
    [aux_sym_number_literal_token2] = ACTIONS(1624),
    [aux_sym_number_literal_token3] = ACTIONS(1624),
    [aux_sym_number_literal_token4] = ACTIONS(1622),
    [aux_sym_number_literal_token5] = ACTIONS(1624),
    [anon_sym_return] = ACTIONS(1622),
    [anon_sym_SQUOTE] = ACTIONS(1622),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1624),
    [anon_sym_DQUOTE] = ACTIONS(1622),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1624),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1624),
    [anon_sym_switch] = ACTIONS(1622),
    [anon_sym_case] = ACTIONS(1622),
    [anon_sym_default] = ACTIONS(1622),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1622),
    [anon_sym_int] = ACTIONS(1622),
    [anon_sym_boolean] = ACTIONS(1622),
    [anon_sym_char] = ACTIONS(1622),
    [anon_sym_short] = ACTIONS(1622),
    [anon_sym_long] = ACTIONS(1622),
    [anon_sym_float] = ACTIONS(1622),
    [anon_sym_double] = ACTIONS(1622),
    [anon_sym_void] = ACTIONS(1622),
    [anon_sym_public] = ACTIONS(1622),
    [anon_sym_protected] = ACTIONS(1622),
    [anon_sym_private] = ACTIONS(1622),
    [anon_sym_static] = ACTIONS(1622),
    [anon_sym_final] = ACTIONS(1622),
    [anon_sym_synchronized] = ACTIONS(1622),
    [anon_sym_TILDE] = ACTIONS(1624),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_new] = ACTIONS(1622),
  },
  [STATE(636)] = {
    [sym_comment] = STATE(636),
    [sym_groovy_doc] = STATE(636),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(637)] = {
    [sym_comment] = STATE(637),
    [sym_groovy_doc] = STATE(637),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(638)] = {
    [sym_comment] = STATE(638),
    [sym_groovy_doc] = STATE(638),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(639)] = {
    [sym_comment] = STATE(639),
    [sym_groovy_doc] = STATE(639),
    [sym_identifier] = ACTIONS(1610),
    [anon_sym_break] = ACTIONS(1610),
    [anon_sym_continue] = ACTIONS(1610),
    [anon_sym_SEMI] = ACTIONS(1612),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1612),
    [anon_sym_import] = ACTIONS(1610),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1610),
    [anon_sym_AT] = ACTIONS(1610),
    [anon_sym_assert] = ACTIONS(1610),
    [anon_sym_PLUS_PLUS] = ACTIONS(1612),
    [anon_sym_DASH_DASH] = ACTIONS(1612),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1610),
    [anon_sym_false] = ACTIONS(1610),
    [anon_sym_ATinterface] = ACTIONS(1612),
    [anon_sym_interface] = ACTIONS(1610),
    [anon_sym_class] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1612),
    [anon_sym_RBRACE] = ACTIONS(1612),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1610),
    [anon_sym_def] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1612),
    [anon_sym_null] = ACTIONS(1610),
    [anon_sym_this] = ACTIONS(1610),
    [anon_sym_do] = ACTIONS(1610),
    [anon_sym_while] = ACTIONS(1610),
    [anon_sym_for] = ACTIONS(1610),
    [aux_sym__type_identifier_token1] = ACTIONS(1610),
    [anon_sym_if] = ACTIONS(1610),
    [anon_sym_else] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1612),
    [aux_sym_number_literal_token1] = ACTIONS(1610),
    [aux_sym_number_literal_token2] = ACTIONS(1612),
    [aux_sym_number_literal_token3] = ACTIONS(1612),
    [aux_sym_number_literal_token4] = ACTIONS(1610),
    [aux_sym_number_literal_token5] = ACTIONS(1612),
    [anon_sym_return] = ACTIONS(1610),
    [anon_sym_SQUOTE] = ACTIONS(1610),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1612),
    [anon_sym_DQUOTE] = ACTIONS(1610),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1612),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1612),
    [anon_sym_switch] = ACTIONS(1610),
    [anon_sym_case] = ACTIONS(1610),
    [anon_sym_default] = ACTIONS(1610),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1610),
    [anon_sym_int] = ACTIONS(1610),
    [anon_sym_boolean] = ACTIONS(1610),
    [anon_sym_char] = ACTIONS(1610),
    [anon_sym_short] = ACTIONS(1610),
    [anon_sym_long] = ACTIONS(1610),
    [anon_sym_float] = ACTIONS(1610),
    [anon_sym_double] = ACTIONS(1610),
    [anon_sym_void] = ACTIONS(1610),
    [anon_sym_public] = ACTIONS(1610),
    [anon_sym_protected] = ACTIONS(1610),
    [anon_sym_private] = ACTIONS(1610),
    [anon_sym_static] = ACTIONS(1610),
    [anon_sym_final] = ACTIONS(1610),
    [anon_sym_synchronized] = ACTIONS(1610),
    [anon_sym_TILDE] = ACTIONS(1612),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_new] = ACTIONS(1610),
  },
  [STATE(640)] = {
    [sym_comment] = STATE(640),
    [sym_groovy_doc] = STATE(640),
    [sym_identifier] = ACTIONS(1598),
    [anon_sym_break] = ACTIONS(1598),
    [anon_sym_continue] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1654),
    [anon_sym_STAR_COLON] = ACTIONS(1600),
    [anon_sym_import] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1598),
    [anon_sym_AT] = ACTIONS(1598),
    [anon_sym_assert] = ACTIONS(1598),
    [anon_sym_PLUS_PLUS] = ACTIONS(1600),
    [anon_sym_DASH_DASH] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1654),
    [anon_sym_PLUS] = ACTIONS(1658),
    [anon_sym_DASH] = ACTIONS(1658),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_ATinterface] = ACTIONS(1600),
    [anon_sym_interface] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1598),
    [anon_sym_def] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_null] = ACTIONS(1598),
    [anon_sym_this] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_while] = ACTIONS(1598),
    [anon_sym_for] = ACTIONS(1598),
    [aux_sym__type_identifier_token1] = ACTIONS(1598),
    [anon_sym_if] = ACTIONS(1598),
    [anon_sym_else] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1600),
    [aux_sym_number_literal_token1] = ACTIONS(1598),
    [aux_sym_number_literal_token2] = ACTIONS(1600),
    [aux_sym_number_literal_token3] = ACTIONS(1600),
    [aux_sym_number_literal_token4] = ACTIONS(1598),
    [aux_sym_number_literal_token5] = ACTIONS(1600),
    [anon_sym_return] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1598),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1598),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1600),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1600),
    [anon_sym_switch] = ACTIONS(1598),
    [anon_sym_case] = ACTIONS(1598),
    [anon_sym_default] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1598),
    [anon_sym_int] = ACTIONS(1598),
    [anon_sym_boolean] = ACTIONS(1598),
    [anon_sym_char] = ACTIONS(1598),
    [anon_sym_short] = ACTIONS(1598),
    [anon_sym_long] = ACTIONS(1598),
    [anon_sym_float] = ACTIONS(1598),
    [anon_sym_double] = ACTIONS(1598),
    [anon_sym_void] = ACTIONS(1598),
    [anon_sym_public] = ACTIONS(1598),
    [anon_sym_protected] = ACTIONS(1598),
    [anon_sym_private] = ACTIONS(1598),
    [anon_sym_static] = ACTIONS(1598),
    [anon_sym_final] = ACTIONS(1598),
    [anon_sym_synchronized] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
  },
  [STATE(641)] = {
    [sym_comment] = STATE(641),
    [sym_groovy_doc] = STATE(641),
    [ts_builtin_sym_end] = ACTIONS(1640),
    [sym_identifier] = ACTIONS(1638),
    [anon_sym_break] = ACTIONS(1638),
    [anon_sym_continue] = ACTIONS(1638),
    [anon_sym_SEMI] = ACTIONS(1640),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1640),
    [anon_sym_import] = ACTIONS(1638),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1638),
    [anon_sym_AT] = ACTIONS(1638),
    [anon_sym_assert] = ACTIONS(1638),
    [anon_sym_PLUS_PLUS] = ACTIONS(1640),
    [anon_sym_DASH_DASH] = ACTIONS(1640),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1638),
    [anon_sym_false] = ACTIONS(1638),
    [anon_sym_ATinterface] = ACTIONS(1640),
    [anon_sym_interface] = ACTIONS(1638),
    [anon_sym_class] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1640),
    [anon_sym_RBRACE] = ACTIONS(1640),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1638),
    [anon_sym_def] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1640),
    [anon_sym_null] = ACTIONS(1638),
    [anon_sym_this] = ACTIONS(1638),
    [anon_sym_do] = ACTIONS(1638),
    [anon_sym_while] = ACTIONS(1638),
    [anon_sym_for] = ACTIONS(1638),
    [aux_sym__type_identifier_token1] = ACTIONS(1638),
    [anon_sym_if] = ACTIONS(1638),
    [anon_sym_else] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1640),
    [aux_sym_number_literal_token1] = ACTIONS(1638),
    [aux_sym_number_literal_token2] = ACTIONS(1640),
    [aux_sym_number_literal_token3] = ACTIONS(1640),
    [aux_sym_number_literal_token4] = ACTIONS(1638),
    [aux_sym_number_literal_token5] = ACTIONS(1640),
    [anon_sym_pipeline] = ACTIONS(1638),
    [anon_sym_return] = ACTIONS(1638),
    [anon_sym_SQUOTE] = ACTIONS(1638),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1640),
    [anon_sym_DQUOTE] = ACTIONS(1638),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1640),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1640),
    [anon_sym_switch] = ACTIONS(1638),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1638),
    [anon_sym_int] = ACTIONS(1638),
    [anon_sym_boolean] = ACTIONS(1638),
    [anon_sym_char] = ACTIONS(1638),
    [anon_sym_short] = ACTIONS(1638),
    [anon_sym_long] = ACTIONS(1638),
    [anon_sym_float] = ACTIONS(1638),
    [anon_sym_double] = ACTIONS(1638),
    [anon_sym_void] = ACTIONS(1638),
    [anon_sym_public] = ACTIONS(1638),
    [anon_sym_protected] = ACTIONS(1638),
    [anon_sym_private] = ACTIONS(1638),
    [anon_sym_static] = ACTIONS(1638),
    [anon_sym_final] = ACTIONS(1638),
    [anon_sym_synchronized] = ACTIONS(1638),
    [anon_sym_TILDE] = ACTIONS(1640),
    [anon_sym_BANG] = ACTIONS(1638),
    [anon_sym_new] = ACTIONS(1638),
  },
  [STATE(642)] = {
    [sym_comment] = STATE(642),
    [sym_groovy_doc] = STATE(642),
    [ts_builtin_sym_end] = ACTIONS(1602),
    [sym_identifier] = ACTIONS(1604),
    [anon_sym_break] = ACTIONS(1604),
    [anon_sym_continue] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1604),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1604),
    [anon_sym_AT] = ACTIONS(1604),
    [anon_sym_assert] = ACTIONS(1604),
    [anon_sym_PLUS_PLUS] = ACTIONS(1602),
    [anon_sym_DASH_DASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1604),
    [anon_sym_false] = ACTIONS(1604),
    [anon_sym_ATinterface] = ACTIONS(1602),
    [anon_sym_interface] = ACTIONS(1604),
    [anon_sym_class] = ACTIONS(1604),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1604),
    [anon_sym_def] = ACTIONS(1604),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_null] = ACTIONS(1604),
    [anon_sym_this] = ACTIONS(1604),
    [anon_sym_do] = ACTIONS(1604),
    [anon_sym_while] = ACTIONS(1604),
    [anon_sym_for] = ACTIONS(1604),
    [aux_sym__type_identifier_token1] = ACTIONS(1604),
    [anon_sym_if] = ACTIONS(1604),
    [anon_sym_else] = ACTIONS(1604),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [aux_sym_number_literal_token1] = ACTIONS(1604),
    [aux_sym_number_literal_token2] = ACTIONS(1602),
    [aux_sym_number_literal_token3] = ACTIONS(1602),
    [aux_sym_number_literal_token4] = ACTIONS(1604),
    [aux_sym_number_literal_token5] = ACTIONS(1602),
    [anon_sym_pipeline] = ACTIONS(1604),
    [anon_sym_return] = ACTIONS(1604),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1602),
    [anon_sym_DQUOTE] = ACTIONS(1604),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1602),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1602),
    [anon_sym_switch] = ACTIONS(1604),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1604),
    [anon_sym_int] = ACTIONS(1604),
    [anon_sym_boolean] = ACTIONS(1604),
    [anon_sym_char] = ACTIONS(1604),
    [anon_sym_short] = ACTIONS(1604),
    [anon_sym_long] = ACTIONS(1604),
    [anon_sym_float] = ACTIONS(1604),
    [anon_sym_double] = ACTIONS(1604),
    [anon_sym_void] = ACTIONS(1604),
    [anon_sym_public] = ACTIONS(1604),
    [anon_sym_protected] = ACTIONS(1604),
    [anon_sym_private] = ACTIONS(1604),
    [anon_sym_static] = ACTIONS(1604),
    [anon_sym_final] = ACTIONS(1604),
    [anon_sym_synchronized] = ACTIONS(1604),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1604),
    [anon_sym_new] = ACTIONS(1604),
  },
  [STATE(643)] = {
    [sym_comment] = STATE(643),
    [sym_groovy_doc] = STATE(643),
    [sym_identifier] = ACTIONS(1598),
    [anon_sym_break] = ACTIONS(1598),
    [anon_sym_continue] = ACTIONS(1598),
    [anon_sym_SEMI] = ACTIONS(1600),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1600),
    [anon_sym_import] = ACTIONS(1598),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1598),
    [anon_sym_AT] = ACTIONS(1598),
    [anon_sym_assert] = ACTIONS(1598),
    [anon_sym_PLUS_PLUS] = ACTIONS(1600),
    [anon_sym_DASH_DASH] = ACTIONS(1600),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1598),
    [anon_sym_false] = ACTIONS(1598),
    [anon_sym_ATinterface] = ACTIONS(1600),
    [anon_sym_interface] = ACTIONS(1598),
    [anon_sym_class] = ACTIONS(1598),
    [anon_sym_COMMA] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1600),
    [anon_sym_DASH_GT] = ACTIONS(1694),
    [anon_sym_RBRACE] = ACTIONS(1600),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1598),
    [anon_sym_def] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1600),
    [anon_sym_null] = ACTIONS(1598),
    [anon_sym_this] = ACTIONS(1598),
    [anon_sym_do] = ACTIONS(1598),
    [anon_sym_while] = ACTIONS(1598),
    [anon_sym_for] = ACTIONS(1598),
    [aux_sym__type_identifier_token1] = ACTIONS(1598),
    [anon_sym_if] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1600),
    [aux_sym_number_literal_token1] = ACTIONS(1598),
    [aux_sym_number_literal_token2] = ACTIONS(1600),
    [aux_sym_number_literal_token3] = ACTIONS(1600),
    [aux_sym_number_literal_token4] = ACTIONS(1598),
    [aux_sym_number_literal_token5] = ACTIONS(1600),
    [anon_sym_return] = ACTIONS(1598),
    [anon_sym_SQUOTE] = ACTIONS(1598),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1600),
    [anon_sym_DQUOTE] = ACTIONS(1598),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1600),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1600),
    [anon_sym_switch] = ACTIONS(1598),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1598),
    [anon_sym_int] = ACTIONS(1598),
    [anon_sym_boolean] = ACTIONS(1598),
    [anon_sym_char] = ACTIONS(1598),
    [anon_sym_short] = ACTIONS(1598),
    [anon_sym_long] = ACTIONS(1598),
    [anon_sym_float] = ACTIONS(1598),
    [anon_sym_double] = ACTIONS(1598),
    [anon_sym_void] = ACTIONS(1598),
    [anon_sym_public] = ACTIONS(1598),
    [anon_sym_protected] = ACTIONS(1598),
    [anon_sym_private] = ACTIONS(1598),
    [anon_sym_static] = ACTIONS(1598),
    [anon_sym_final] = ACTIONS(1598),
    [anon_sym_synchronized] = ACTIONS(1598),
    [anon_sym_TILDE] = ACTIONS(1600),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_new] = ACTIONS(1598),
  },
  [STATE(644)] = {
    [sym_comment] = STATE(644),
    [sym_groovy_doc] = STATE(644),
    [sym_identifier] = ACTIONS(1614),
    [anon_sym_break] = ACTIONS(1614),
    [anon_sym_continue] = ACTIONS(1614),
    [anon_sym_SEMI] = ACTIONS(1616),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_STAR_COLON] = ACTIONS(1616),
    [anon_sym_import] = ACTIONS(1614),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1614),
    [anon_sym_AT] = ACTIONS(1614),
    [anon_sym_assert] = ACTIONS(1614),
    [anon_sym_PLUS_PLUS] = ACTIONS(1616),
    [anon_sym_DASH_DASH] = ACTIONS(1616),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1570),
    [anon_sym_PLUS] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1614),
    [anon_sym_false] = ACTIONS(1614),
    [anon_sym_ATinterface] = ACTIONS(1616),
    [anon_sym_interface] = ACTIONS(1614),
    [anon_sym_class] = ACTIONS(1614),
    [anon_sym_COMMA] = ACTIONS(1694),
    [anon_sym_LBRACE] = ACTIONS(1616),
    [anon_sym_DASH_GT] = ACTIONS(1694),
    [anon_sym_RBRACE] = ACTIONS(1616),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1614),
    [anon_sym_def] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1616),
    [anon_sym_null] = ACTIONS(1614),
    [anon_sym_this] = ACTIONS(1614),
    [anon_sym_do] = ACTIONS(1614),
    [anon_sym_while] = ACTIONS(1614),
    [anon_sym_for] = ACTIONS(1614),
    [aux_sym__type_identifier_token1] = ACTIONS(1614),
    [anon_sym_if] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1616),
    [aux_sym_number_literal_token1] = ACTIONS(1614),
    [aux_sym_number_literal_token2] = ACTIONS(1616),
    [aux_sym_number_literal_token3] = ACTIONS(1616),
    [aux_sym_number_literal_token4] = ACTIONS(1614),
    [aux_sym_number_literal_token5] = ACTIONS(1616),
    [anon_sym_return] = ACTIONS(1614),
    [anon_sym_SQUOTE] = ACTIONS(1614),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1616),
    [anon_sym_DQUOTE] = ACTIONS(1614),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1616),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1616),
    [anon_sym_switch] = ACTIONS(1614),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1614),
    [anon_sym_int] = ACTIONS(1614),
    [anon_sym_boolean] = ACTIONS(1614),
    [anon_sym_char] = ACTIONS(1614),
    [anon_sym_short] = ACTIONS(1614),
    [anon_sym_long] = ACTIONS(1614),
    [anon_sym_float] = ACTIONS(1614),
    [anon_sym_double] = ACTIONS(1614),
    [anon_sym_void] = ACTIONS(1614),
    [anon_sym_public] = ACTIONS(1614),
    [anon_sym_protected] = ACTIONS(1614),
    [anon_sym_private] = ACTIONS(1614),
    [anon_sym_static] = ACTIONS(1614),
    [anon_sym_final] = ACTIONS(1614),
    [anon_sym_synchronized] = ACTIONS(1614),
    [anon_sym_TILDE] = ACTIONS(1616),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_new] = ACTIONS(1614),
  },
  [STATE(645)] = {
    [sym_comment] = STATE(645),
    [sym_groovy_doc] = STATE(645),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(646)] = {
    [sym_comment] = STATE(646),
    [sym_groovy_doc] = STATE(646),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(647)] = {
    [sym_comment] = STATE(647),
    [sym_groovy_doc] = STATE(647),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(648)] = {
    [sym_comment] = STATE(648),
    [sym_groovy_doc] = STATE(648),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1692),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(649)] = {
    [sym_comment] = STATE(649),
    [sym_groovy_doc] = STATE(649),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1692),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1517),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(650)] = {
    [sym_comment] = STATE(650),
    [sym_groovy_doc] = STATE(650),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1686),
    [anon_sym_DOT_AMP] = ACTIONS(1650),
    [anon_sym_DOT_AT] = ACTIONS(1650),
    [anon_sym_QMARK_DOT] = ACTIONS(1650),
    [anon_sym_STAR_DOT] = ACTIONS(1650),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1664),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1656),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1660),
    [anon_sym_GT_GT] = ACTIONS(1662),
    [anon_sym_GT_GT_GT] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1662),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1660),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_LT_EQ] = ACTIONS(1664),
    [anon_sym_GT] = ACTIONS(1664),
    [anon_sym_GT_EQ] = ACTIONS(1666),
    [anon_sym_in] = ACTIONS(1664),
    [anon_sym_BANGin] = ACTIONS(1664),
    [anon_sym_instanceof] = ACTIONS(1664),
    [anon_sym_BANGinstanceof] = ACTIONS(1666),
    [anon_sym_EQ_EQ] = ACTIONS(1668),
    [anon_sym_BANG_EQ] = ACTIONS(1668),
    [anon_sym_LT_EQ_GT] = ACTIONS(1670),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1670),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1670),
    [anon_sym_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1670),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_CARET] = ACTIONS(1674),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_AMP_AMP] = ACTIONS(1678),
    [anon_sym_PIPE_PIPE] = ACTIONS(1680),
    [anon_sym_QMARK_COLON] = ACTIONS(1682),
    [anon_sym_STAR_STAR] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1684),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(651)] = {
    [sym_comment] = STATE(651),
    [sym_groovy_doc] = STATE(651),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1686),
    [anon_sym_DOT_AMP] = ACTIONS(1308),
    [anon_sym_DOT_AT] = ACTIONS(1308),
    [anon_sym_QMARK_DOT] = ACTIONS(1308),
    [anon_sym_STAR_DOT] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1310),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1308),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1308),
    [anon_sym_GT_GT] = ACTIONS(1310),
    [anon_sym_GT_GT_GT] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1310),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1308),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1310),
    [anon_sym_LT_EQ] = ACTIONS(1310),
    [anon_sym_GT] = ACTIONS(1310),
    [anon_sym_GT_EQ] = ACTIONS(1308),
    [anon_sym_in] = ACTIONS(1310),
    [anon_sym_BANGin] = ACTIONS(1310),
    [anon_sym_instanceof] = ACTIONS(1310),
    [anon_sym_BANGinstanceof] = ACTIONS(1308),
    [anon_sym_EQ_EQ] = ACTIONS(1310),
    [anon_sym_BANG_EQ] = ACTIONS(1310),
    [anon_sym_LT_EQ_GT] = ACTIONS(1308),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1308),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1308),
    [anon_sym_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1310),
    [anon_sym_CARET] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1310),
    [anon_sym_AMP_AMP] = ACTIONS(1308),
    [anon_sym_PIPE_PIPE] = ACTIONS(1308),
    [anon_sym_QMARK_COLON] = ACTIONS(1308),
    [anon_sym_STAR_STAR] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1310),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(652)] = {
    [sym_comment] = STATE(652),
    [sym_groovy_doc] = STATE(652),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(653)] = {
    [sym_comment] = STATE(653),
    [sym_groovy_doc] = STATE(653),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(654)] = {
    [sym_comment] = STATE(654),
    [sym_groovy_doc] = STATE(654),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(655)] = {
    [sym_comment] = STATE(655),
    [sym_groovy_doc] = STATE(655),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(656)] = {
    [sym_comment] = STATE(656),
    [sym_groovy_doc] = STATE(656),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_DOT_AMP] = ACTIONS(1566),
    [anon_sym_DOT_AT] = ACTIONS(1566),
    [anon_sym_QMARK_DOT] = ACTIONS(1566),
    [anon_sym_STAR_DOT] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_as] = ACTIONS(1580),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_PERCENT] = ACTIONS(1572),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_LT_LT] = ACTIONS(1576),
    [anon_sym_GT_GT] = ACTIONS(1578),
    [anon_sym_GT_GT_GT] = ACTIONS(1576),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(1576),
    [anon_sym_LT_DOT_DOT] = ACTIONS(1578),
    [anon_sym_LT] = ACTIONS(1580),
    [anon_sym_LT_EQ] = ACTIONS(1580),
    [anon_sym_GT] = ACTIONS(1580),
    [anon_sym_GT_EQ] = ACTIONS(1582),
    [anon_sym_in] = ACTIONS(1580),
    [anon_sym_BANGin] = ACTIONS(1580),
    [anon_sym_instanceof] = ACTIONS(1580),
    [anon_sym_BANGinstanceof] = ACTIONS(1582),
    [anon_sym_EQ_EQ] = ACTIONS(1584),
    [anon_sym_BANG_EQ] = ACTIONS(1584),
    [anon_sym_LT_EQ_GT] = ACTIONS(1586),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1586),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1586),
    [anon_sym_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1588),
    [anon_sym_CARET] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_AMP_AMP] = ACTIONS(1594),
    [anon_sym_PIPE_PIPE] = ACTIONS(1596),
    [anon_sym_QMARK_COLON] = ACTIONS(1642),
    [anon_sym_STAR_STAR] = ACTIONS(1568),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_QMARK] = ACTIONS(1644),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1513),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(657)] = {
    [sym_access_op] = STATE(475),
    [sym_dotted_identifier] = STATE(349),
    [sym_increment_op] = STATE(475),
    [sym_binary_op] = STATE(475),
    [sym_boolean_literal] = STATE(329),
    [sym_closure] = STATE(475),
    [sym_comment] = STATE(657),
    [sym_groovy_doc] = STATE(657),
    [sym_parenthesized_expression] = STATE(312),
    [sym__expression] = STATE(509),
    [sym__primary_expression] = STATE(265),
    [sym__callable_expression] = STATE(294),
    [sym__juxtable_expression] = STATE(312),
    [sym_function_call] = STATE(312),
    [sym__type_identifier] = STATE(275),
    [sym_index] = STATE(349),
    [sym_list] = STATE(329),
    [sym_map] = STATE(329),
    [sym_number_literal] = STATE(329),
    [sym_string] = STATE(329),
    [sym__plain_string] = STATE(298),
    [sym__interpolate_string] = STATE(298),
    [sym_ternary_op] = STATE(475),
    [sym_unary_op] = STATE(475),
    [ts_builtin_sym_end] = ACTIONS(1696),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_break] = ACTIONS(1700),
    [anon_sym_continue] = ACTIONS(1700),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(1085),
    [anon_sym_STAR_COLON] = ACTIONS(1087),
    [anon_sym_import] = ACTIONS(1700),
    [anon_sym_package] = ACTIONS(1700),
    [anon_sym_AT] = ACTIONS(1700),
    [anon_sym_assert] = ACTIONS(1700),
    [anon_sym_PLUS_PLUS] = ACTIONS(1095),
    [anon_sym_DASH_DASH] = ACTIONS(1095),
    [anon_sym_SLASH] = ACTIONS(1097),
    [anon_sym_PLUS] = ACTIONS(1099),
    [anon_sym_DASH] = ACTIONS(1099),
    [anon_sym_true] = ACTIONS(1101),
    [anon_sym_false] = ACTIONS(1101),
    [anon_sym_ATinterface] = ACTIONS(1696),
    [anon_sym_interface] = ACTIONS(1700),
    [anon_sym_class] = ACTIONS(1700),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1700),
    [anon_sym_def] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1702),
    [anon_sym_null] = ACTIONS(1113),
    [anon_sym_this] = ACTIONS(1115),
    [anon_sym_do] = ACTIONS(1700),
    [anon_sym_while] = ACTIONS(1700),
    [anon_sym_for] = ACTIONS(1700),
    [aux_sym__type_identifier_token1] = ACTIONS(1704),
    [anon_sym_if] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1700),
    [anon_sym_LBRACK] = ACTIONS(1127),
    [aux_sym_number_literal_token1] = ACTIONS(1129),
    [aux_sym_number_literal_token2] = ACTIONS(1131),
    [aux_sym_number_literal_token3] = ACTIONS(1131),
    [aux_sym_number_literal_token4] = ACTIONS(1129),
    [aux_sym_number_literal_token5] = ACTIONS(1131),
    [anon_sym_pipeline] = ACTIONS(1700),
    [anon_sym_return] = ACTIONS(1700),
    [anon_sym_SQUOTE] = ACTIONS(1135),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1137),
    [anon_sym_DQUOTE] = ACTIONS(1139),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1141),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1143),
    [anon_sym_switch] = ACTIONS(1700),
    [anon_sym_try] = ACTIONS(1700),
    [anon_sym_catch] = ACTIONS(1700),
    [anon_sym_finally] = ACTIONS(1700),
    [anon_sym_int] = ACTIONS(1700),
    [anon_sym_boolean] = ACTIONS(1700),
    [anon_sym_char] = ACTIONS(1700),
    [anon_sym_short] = ACTIONS(1700),
    [anon_sym_long] = ACTIONS(1700),
    [anon_sym_float] = ACTIONS(1700),
    [anon_sym_double] = ACTIONS(1700),
    [anon_sym_void] = ACTIONS(1700),
    [anon_sym_public] = ACTIONS(1700),
    [anon_sym_protected] = ACTIONS(1700),
    [anon_sym_private] = ACTIONS(1700),
    [anon_sym_static] = ACTIONS(1700),
    [anon_sym_final] = ACTIONS(1700),
    [anon_sym_synchronized] = ACTIONS(1700),
    [anon_sym_TILDE] = ACTIONS(1147),
    [anon_sym_BANG] = ACTIONS(1147),
    [anon_sym_new] = ACTIONS(1149),
  },
  [STATE(658)] = {
    [sym_access_op] = STATE(387),
    [sym_dotted_identifier] = STATE(367),
    [sym_increment_op] = STATE(387),
    [sym_binary_op] = STATE(387),
    [sym_boolean_literal] = STATE(303),
    [sym_closure] = STATE(387),
    [sym_comment] = STATE(658),
    [sym_groovy_doc] = STATE(658),
    [sym_parenthesized_expression] = STATE(360),
    [sym__expression] = STATE(528),
    [sym__primary_expression] = STATE(273),
    [sym__callable_expression] = STATE(284),
    [sym__juxtable_expression] = STATE(360),
    [sym_function_call] = STATE(360),
    [sym__type_identifier] = STATE(292),
    [sym_index] = STATE(367),
    [sym_list] = STATE(303),
    [sym_map] = STATE(303),
    [sym_number_literal] = STATE(303),
    [sym_string] = STATE(303),
    [sym__plain_string] = STATE(285),
    [sym__interpolate_string] = STATE(285),
    [sym_ternary_op] = STATE(387),
    [sym_unary_op] = STATE(387),
    [sym_identifier] = ACTIONS(1706),
    [anon_sym_break] = ACTIONS(1700),
    [anon_sym_continue] = ACTIONS(1700),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(997),
    [anon_sym_STAR_COLON] = ACTIONS(999),
    [anon_sym_import] = ACTIONS(1700),
    [anon_sym_package] = ACTIONS(1700),
    [anon_sym_AT] = ACTIONS(1700),
    [anon_sym_assert] = ACTIONS(1700),
    [anon_sym_PLUS_PLUS] = ACTIONS(1007),
    [anon_sym_DASH_DASH] = ACTIONS(1007),
    [anon_sym_SLASH] = ACTIONS(1009),
    [anon_sym_PLUS] = ACTIONS(1011),
    [anon_sym_DASH] = ACTIONS(1011),
    [anon_sym_true] = ACTIONS(1013),
    [anon_sym_false] = ACTIONS(1013),
    [anon_sym_ATinterface] = ACTIONS(1696),
    [anon_sym_interface] = ACTIONS(1700),
    [anon_sym_class] = ACTIONS(1700),
    [anon_sym_LBRACE] = ACTIONS(1019),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1700),
    [anon_sym_def] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1708),
    [anon_sym_null] = ACTIONS(1027),
    [anon_sym_this] = ACTIONS(1029),
    [anon_sym_do] = ACTIONS(1700),
    [anon_sym_while] = ACTIONS(1700),
    [anon_sym_for] = ACTIONS(1700),
    [aux_sym__type_identifier_token1] = ACTIONS(1710),
    [anon_sym_if] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1700),
    [anon_sym_LBRACK] = ACTIONS(1041),
    [aux_sym_number_literal_token1] = ACTIONS(1043),
    [aux_sym_number_literal_token2] = ACTIONS(1045),
    [aux_sym_number_literal_token3] = ACTIONS(1045),
    [aux_sym_number_literal_token4] = ACTIONS(1043),
    [aux_sym_number_literal_token5] = ACTIONS(1045),
    [anon_sym_return] = ACTIONS(1700),
    [anon_sym_SQUOTE] = ACTIONS(1049),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1051),
    [anon_sym_DQUOTE] = ACTIONS(1053),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1055),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1057),
    [anon_sym_switch] = ACTIONS(1700),
    [anon_sym_case] = ACTIONS(1700),
    [anon_sym_default] = ACTIONS(1700),
    [anon_sym_try] = ACTIONS(1700),
    [anon_sym_catch] = ACTIONS(1700),
    [anon_sym_finally] = ACTIONS(1700),
    [anon_sym_int] = ACTIONS(1700),
    [anon_sym_boolean] = ACTIONS(1700),
    [anon_sym_char] = ACTIONS(1700),
    [anon_sym_short] = ACTIONS(1700),
    [anon_sym_long] = ACTIONS(1700),
    [anon_sym_float] = ACTIONS(1700),
    [anon_sym_double] = ACTIONS(1700),
    [anon_sym_void] = ACTIONS(1700),
    [anon_sym_public] = ACTIONS(1700),
    [anon_sym_protected] = ACTIONS(1700),
    [anon_sym_private] = ACTIONS(1700),
    [anon_sym_static] = ACTIONS(1700),
    [anon_sym_final] = ACTIONS(1700),
    [anon_sym_synchronized] = ACTIONS(1700),
    [anon_sym_TILDE] = ACTIONS(1063),
    [anon_sym_BANG] = ACTIONS(1063),
    [anon_sym_new] = ACTIONS(1065),
  },
  [STATE(659)] = {
    [sym_access_op] = STATE(480),
    [sym_dotted_identifier] = STATE(395),
    [sym_increment_op] = STATE(480),
    [sym_binary_op] = STATE(480),
    [sym_boolean_literal] = STATE(439),
    [sym_closure] = STATE(480),
    [sym_comment] = STATE(659),
    [sym_groovy_doc] = STATE(659),
    [sym_parenthesized_expression] = STATE(470),
    [sym__expression] = STATE(606),
    [sym__primary_expression] = STATE(374),
    [sym__callable_expression] = STATE(383),
    [sym__juxtable_expression] = STATE(470),
    [sym_function_call] = STATE(470),
    [sym__type_identifier] = STATE(330),
    [sym_index] = STATE(395),
    [sym_list] = STATE(439),
    [sym_map] = STATE(439),
    [sym_number_literal] = STATE(439),
    [sym_string] = STATE(439),
    [sym__plain_string] = STATE(346),
    [sym__interpolate_string] = STATE(346),
    [sym_ternary_op] = STATE(480),
    [sym_unary_op] = STATE(480),
    [ts_builtin_sym_end] = ACTIONS(1696),
    [sym_identifier] = ACTIONS(1712),
    [anon_sym_break] = ACTIONS(1700),
    [anon_sym_continue] = ACTIONS(1700),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(17),
    [anon_sym_STAR_COLON] = ACTIONS(19),
    [anon_sym_import] = ACTIONS(1700),
    [anon_sym_package] = ACTIONS(1700),
    [anon_sym_AT] = ACTIONS(1700),
    [anon_sym_assert] = ACTIONS(1700),
    [anon_sym_PLUS_PLUS] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(29),
    [anon_sym_SLASH] = ACTIONS(31),
    [anon_sym_PLUS] = ACTIONS(33),
    [anon_sym_DASH] = ACTIONS(33),
    [anon_sym_true] = ACTIONS(35),
    [anon_sym_false] = ACTIONS(35),
    [anon_sym_ATinterface] = ACTIONS(1696),
    [anon_sym_interface] = ACTIONS(1700),
    [anon_sym_class] = ACTIONS(1700),
    [anon_sym_LBRACE] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1700),
    [anon_sym_def] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1714),
    [anon_sym_null] = ACTIONS(49),
    [anon_sym_this] = ACTIONS(51),
    [anon_sym_do] = ACTIONS(1700),
    [anon_sym_while] = ACTIONS(1700),
    [anon_sym_for] = ACTIONS(1700),
    [aux_sym__type_identifier_token1] = ACTIONS(1716),
    [anon_sym_if] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1700),
    [anon_sym_LBRACK] = ACTIONS(63),
    [aux_sym_number_literal_token1] = ACTIONS(65),
    [aux_sym_number_literal_token2] = ACTIONS(67),
    [aux_sym_number_literal_token3] = ACTIONS(67),
    [aux_sym_number_literal_token4] = ACTIONS(65),
    [aux_sym_number_literal_token5] = ACTIONS(67),
    [anon_sym_pipeline] = ACTIONS(1700),
    [anon_sym_return] = ACTIONS(1700),
    [anon_sym_SQUOTE] = ACTIONS(73),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(75),
    [anon_sym_DQUOTE] = ACTIONS(77),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(79),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(81),
    [anon_sym_switch] = ACTIONS(1700),
    [anon_sym_try] = ACTIONS(1700),
    [anon_sym_int] = ACTIONS(1700),
    [anon_sym_boolean] = ACTIONS(1700),
    [anon_sym_char] = ACTIONS(1700),
    [anon_sym_short] = ACTIONS(1700),
    [anon_sym_long] = ACTIONS(1700),
    [anon_sym_float] = ACTIONS(1700),
    [anon_sym_double] = ACTIONS(1700),
    [anon_sym_void] = ACTIONS(1700),
    [anon_sym_public] = ACTIONS(1700),
    [anon_sym_protected] = ACTIONS(1700),
    [anon_sym_private] = ACTIONS(1700),
    [anon_sym_static] = ACTIONS(1700),
    [anon_sym_final] = ACTIONS(1700),
    [anon_sym_synchronized] = ACTIONS(1700),
    [anon_sym_TILDE] = ACTIONS(93),
    [anon_sym_BANG] = ACTIONS(93),
    [anon_sym_new] = ACTIONS(95),
  },
  [STATE(660)] = {
    [sym_comment] = STATE(660),
    [sym_groovy_doc] = STATE(660),
    [aux_sym_dotted_identifier_repeat1] = STATE(661),
    [sym_identifier] = ACTIONS(375),
    [anon_sym_SEMI] = ACTIONS(373),
    [anon_sym_DOT_AMP] = ACTIONS(373),
    [anon_sym_DOT_AT] = ACTIONS(373),
    [anon_sym_QMARK_DOT] = ACTIONS(373),
    [anon_sym_STAR_DOT] = ACTIONS(373),
    [anon_sym_STAR] = ACTIONS(375),
    [anon_sym_STAR_COLON] = ACTIONS(373),
    [anon_sym_DOT] = ACTIONS(375),
    [anon_sym_as] = ACTIONS(375),
    [anon_sym_EQ] = ACTIONS(375),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(373),
    [anon_sym_STAR_EQ] = ACTIONS(373),
    [anon_sym_SLASH_EQ] = ACTIONS(373),
    [anon_sym_PERCENT_EQ] = ACTIONS(373),
    [anon_sym_PLUS_EQ] = ACTIONS(373),
    [anon_sym_DASH_EQ] = ACTIONS(373),
    [anon_sym_LT_LT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(373),
    [anon_sym_AMP_EQ] = ACTIONS(373),
    [anon_sym_CARET_EQ] = ACTIONS(373),
    [anon_sym_PIPE_EQ] = ACTIONS(373),
    [anon_sym_QMARK_EQ] = ACTIONS(373),
    [anon_sym_PLUS_PLUS] = ACTIONS(373),
    [anon_sym_DASH_DASH] = ACTIONS(373),
    [anon_sym_PERCENT] = ACTIONS(375),
    [anon_sym_SLASH] = ACTIONS(375),
    [anon_sym_PLUS] = ACTIONS(375),
    [anon_sym_DASH] = ACTIONS(375),
    [anon_sym_LT_LT] = ACTIONS(375),
    [anon_sym_GT_GT] = ACTIONS(375),
    [anon_sym_GT_GT_GT] = ACTIONS(375),
    [anon_sym_DOT_DOT] = ACTIONS(375),
    [anon_sym_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(373),
    [anon_sym_LT_DOT_DOT] = ACTIONS(375),
    [anon_sym_LT] = ACTIONS(375),
    [anon_sym_LT_EQ] = ACTIONS(375),
    [anon_sym_GT] = ACTIONS(375),
    [anon_sym_GT_EQ] = ACTIONS(373),
    [anon_sym_in] = ACTIONS(375),
    [anon_sym_BANGin] = ACTIONS(375),
    [anon_sym_instanceof] = ACTIONS(375),
    [anon_sym_BANGinstanceof] = ACTIONS(373),
    [anon_sym_EQ_EQ] = ACTIONS(375),
    [anon_sym_BANG_EQ] = ACTIONS(375),
    [anon_sym_LT_EQ_GT] = ACTIONS(373),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(373),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(373),
    [anon_sym_EQ_TILDE] = ACTIONS(373),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(373),
    [anon_sym_AMP] = ACTIONS(375),
    [anon_sym_CARET] = ACTIONS(375),
    [anon_sym_PIPE] = ACTIONS(375),
    [anon_sym_AMP_AMP] = ACTIONS(373),
    [anon_sym_PIPE_PIPE] = ACTIONS(373),
    [anon_sym_QMARK_COLON] = ACTIONS(373),
    [anon_sym_STAR_STAR] = ACTIONS(375),
    [anon_sym_true] = ACTIONS(375),
    [anon_sym_false] = ACTIONS(375),
    [anon_sym_COMMA] = ACTIONS(373),
    [anon_sym_LBRACE] = ACTIONS(373),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(373),
    [anon_sym_RPAREN] = ACTIONS(373),
    [anon_sym_null] = ACTIONS(375),
    [anon_sym_this] = ACTIONS(375),
    [anon_sym_while] = ACTIONS(375),
    [aux_sym__type_identifier_token1] = ACTIONS(375),
    [anon_sym_else] = ACTIONS(375),
    [anon_sym_LBRACK] = ACTIONS(373),
    [aux_sym_number_literal_token1] = ACTIONS(375),
    [aux_sym_number_literal_token2] = ACTIONS(373),
    [aux_sym_number_literal_token3] = ACTIONS(373),
    [aux_sym_number_literal_token4] = ACTIONS(375),
    [aux_sym_number_literal_token5] = ACTIONS(373),
    [anon_sym_SQUOTE] = ACTIONS(375),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(373),
    [anon_sym_DQUOTE] = ACTIONS(375),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(373),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(373),
    [anon_sym_QMARK] = ACTIONS(375),
    [anon_sym_catch] = ACTIONS(375),
    [anon_sym_finally] = ACTIONS(375),
    [anon_sym_TILDE] = ACTIONS(373),
    [anon_sym_BANG] = ACTIONS(375),
    [anon_sym_new] = ACTIONS(375),
  },
  [STATE(661)] = {
    [sym_comment] = STATE(661),
    [sym_groovy_doc] = STATE(661),
    [aux_sym_dotted_identifier_repeat1] = STATE(661),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(1718),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_RPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(662)] = {
    [sym_access_op] = STATE(597),
    [sym_dotted_identifier] = STATE(481),
    [sym_increment_op] = STATE(597),
    [sym_binary_op] = STATE(597),
    [sym_boolean_literal] = STATE(541),
    [sym_closure] = STATE(597),
    [sym_comment] = STATE(662),
    [sym_groovy_doc] = STATE(662),
    [sym_parenthesized_expression] = STATE(563),
    [sym__expression] = STATE(634),
    [sym__primary_expression] = STATE(443),
    [sym__callable_expression] = STATE(389),
    [sym__juxtable_expression] = STATE(563),
    [sym_function_call] = STATE(563),
    [sym__type_identifier] = STATE(393),
    [sym_index] = STATE(481),
    [sym_list] = STATE(541),
    [sym_map] = STATE(541),
    [sym_number_literal] = STATE(541),
    [sym_string] = STATE(541),
    [sym__plain_string] = STATE(391),
    [sym__interpolate_string] = STATE(391),
    [sym_ternary_op] = STATE(597),
    [sym_unary_op] = STATE(597),
    [sym_identifier] = ACTIONS(1721),
    [anon_sym_break] = ACTIONS(1700),
    [anon_sym_continue] = ACTIONS(1700),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(586),
    [anon_sym_STAR_COLON] = ACTIONS(588),
    [anon_sym_import] = ACTIONS(1700),
    [anon_sym_package] = ACTIONS(1700),
    [anon_sym_AT] = ACTIONS(1700),
    [anon_sym_assert] = ACTIONS(1700),
    [anon_sym_PLUS_PLUS] = ACTIONS(596),
    [anon_sym_DASH_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PLUS] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [anon_sym_ATinterface] = ACTIONS(1696),
    [anon_sym_interface] = ACTIONS(1700),
    [anon_sym_class] = ACTIONS(1700),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1700),
    [anon_sym_def] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1723),
    [anon_sym_null] = ACTIONS(618),
    [anon_sym_this] = ACTIONS(620),
    [anon_sym_do] = ACTIONS(1700),
    [anon_sym_while] = ACTIONS(1700),
    [anon_sym_for] = ACTIONS(1700),
    [aux_sym__type_identifier_token1] = ACTIONS(1725),
    [anon_sym_if] = ACTIONS(1700),
    [anon_sym_else] = ACTIONS(1700),
    [anon_sym_LBRACK] = ACTIONS(632),
    [aux_sym_number_literal_token1] = ACTIONS(634),
    [aux_sym_number_literal_token2] = ACTIONS(636),
    [aux_sym_number_literal_token3] = ACTIONS(636),
    [aux_sym_number_literal_token4] = ACTIONS(634),
    [aux_sym_number_literal_token5] = ACTIONS(636),
    [anon_sym_return] = ACTIONS(1700),
    [anon_sym_SQUOTE] = ACTIONS(640),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(642),
    [anon_sym_DQUOTE] = ACTIONS(644),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(646),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(648),
    [anon_sym_switch] = ACTIONS(1700),
    [anon_sym_case] = ACTIONS(1700),
    [anon_sym_default] = ACTIONS(1700),
    [anon_sym_try] = ACTIONS(1700),
    [anon_sym_int] = ACTIONS(1700),
    [anon_sym_boolean] = ACTIONS(1700),
    [anon_sym_char] = ACTIONS(1700),
    [anon_sym_short] = ACTIONS(1700),
    [anon_sym_long] = ACTIONS(1700),
    [anon_sym_float] = ACTIONS(1700),
    [anon_sym_double] = ACTIONS(1700),
    [anon_sym_void] = ACTIONS(1700),
    [anon_sym_public] = ACTIONS(1700),
    [anon_sym_protected] = ACTIONS(1700),
    [anon_sym_private] = ACTIONS(1700),
    [anon_sym_static] = ACTIONS(1700),
    [anon_sym_final] = ACTIONS(1700),
    [anon_sym_synchronized] = ACTIONS(1700),
    [anon_sym_TILDE] = ACTIONS(656),
    [anon_sym_BANG] = ACTIONS(656),
    [anon_sym_new] = ACTIONS(658),
  },
  [STATE(663)] = {
    [sym_comment] = STATE(663),
    [sym_groovy_doc] = STATE(663),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_COLON] = ACTIONS(367),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_RPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(664)] = {
    [sym_comment] = STATE(664),
    [sym_groovy_doc] = STATE(664),
    [sym_identifier] = ACTIONS(393),
    [anon_sym_SEMI] = ACTIONS(391),
    [anon_sym_DOT_AMP] = ACTIONS(391),
    [anon_sym_DOT_AT] = ACTIONS(391),
    [anon_sym_QMARK_DOT] = ACTIONS(391),
    [anon_sym_STAR_DOT] = ACTIONS(391),
    [anon_sym_STAR] = ACTIONS(393),
    [anon_sym_STAR_COLON] = ACTIONS(391),
    [anon_sym_DOT] = ACTIONS(393),
    [anon_sym_as] = ACTIONS(393),
    [anon_sym_EQ] = ACTIONS(393),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(391),
    [anon_sym_STAR_EQ] = ACTIONS(391),
    [anon_sym_SLASH_EQ] = ACTIONS(391),
    [anon_sym_PERCENT_EQ] = ACTIONS(391),
    [anon_sym_PLUS_EQ] = ACTIONS(391),
    [anon_sym_DASH_EQ] = ACTIONS(391),
    [anon_sym_LT_LT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(391),
    [anon_sym_AMP_EQ] = ACTIONS(391),
    [anon_sym_CARET_EQ] = ACTIONS(391),
    [anon_sym_PIPE_EQ] = ACTIONS(391),
    [anon_sym_QMARK_EQ] = ACTIONS(391),
    [anon_sym_PLUS_PLUS] = ACTIONS(391),
    [anon_sym_DASH_DASH] = ACTIONS(391),
    [anon_sym_PERCENT] = ACTIONS(393),
    [anon_sym_SLASH] = ACTIONS(393),
    [anon_sym_PLUS] = ACTIONS(393),
    [anon_sym_DASH] = ACTIONS(393),
    [anon_sym_LT_LT] = ACTIONS(393),
    [anon_sym_GT_GT] = ACTIONS(393),
    [anon_sym_GT_GT_GT] = ACTIONS(393),
    [anon_sym_DOT_DOT] = ACTIONS(393),
    [anon_sym_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(391),
    [anon_sym_LT_DOT_DOT] = ACTIONS(393),
    [anon_sym_LT] = ACTIONS(393),
    [anon_sym_LT_EQ] = ACTIONS(393),
    [anon_sym_GT] = ACTIONS(393),
    [anon_sym_GT_EQ] = ACTIONS(391),
    [anon_sym_in] = ACTIONS(393),
    [anon_sym_BANGin] = ACTIONS(393),
    [anon_sym_instanceof] = ACTIONS(393),
    [anon_sym_BANGinstanceof] = ACTIONS(391),
    [anon_sym_EQ_EQ] = ACTIONS(393),
    [anon_sym_BANG_EQ] = ACTIONS(393),
    [anon_sym_LT_EQ_GT] = ACTIONS(391),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(391),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(391),
    [anon_sym_EQ_TILDE] = ACTIONS(391),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(391),
    [anon_sym_AMP] = ACTIONS(393),
    [anon_sym_CARET] = ACTIONS(393),
    [anon_sym_PIPE] = ACTIONS(393),
    [anon_sym_AMP_AMP] = ACTIONS(391),
    [anon_sym_PIPE_PIPE] = ACTIONS(391),
    [anon_sym_QMARK_COLON] = ACTIONS(391),
    [anon_sym_STAR_STAR] = ACTIONS(393),
    [anon_sym_true] = ACTIONS(393),
    [anon_sym_false] = ACTIONS(393),
    [anon_sym_COMMA] = ACTIONS(391),
    [anon_sym_LBRACE] = ACTIONS(391),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(391),
    [anon_sym_RPAREN] = ACTIONS(391),
    [anon_sym_null] = ACTIONS(393),
    [anon_sym_this] = ACTIONS(393),
    [anon_sym_while] = ACTIONS(393),
    [aux_sym__type_identifier_token1] = ACTIONS(393),
    [anon_sym_else] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(391),
    [aux_sym_number_literal_token1] = ACTIONS(393),
    [aux_sym_number_literal_token2] = ACTIONS(391),
    [aux_sym_number_literal_token3] = ACTIONS(391),
    [aux_sym_number_literal_token4] = ACTIONS(393),
    [aux_sym_number_literal_token5] = ACTIONS(391),
    [anon_sym_SQUOTE] = ACTIONS(393),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(391),
    [anon_sym_DQUOTE] = ACTIONS(393),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(391),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(391),
    [anon_sym_QMARK] = ACTIONS(393),
    [anon_sym_catch] = ACTIONS(393),
    [anon_sym_finally] = ACTIONS(393),
    [anon_sym_TILDE] = ACTIONS(391),
    [anon_sym_BANG] = ACTIONS(393),
    [anon_sym_new] = ACTIONS(393),
  },
  [STATE(665)] = {
    [sym_comment] = STATE(665),
    [sym_groovy_doc] = STATE(665),
    [sym_identifier] = ACTIONS(365),
    [anon_sym_SEMI] = ACTIONS(363),
    [anon_sym_DOT_AMP] = ACTIONS(363),
    [anon_sym_DOT_AT] = ACTIONS(363),
    [anon_sym_QMARK_DOT] = ACTIONS(363),
    [anon_sym_STAR_DOT] = ACTIONS(363),
    [anon_sym_STAR] = ACTIONS(365),
    [anon_sym_STAR_COLON] = ACTIONS(363),
    [anon_sym_DOT] = ACTIONS(365),
    [anon_sym_as] = ACTIONS(365),
    [anon_sym_EQ] = ACTIONS(365),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(363),
    [anon_sym_STAR_EQ] = ACTIONS(363),
    [anon_sym_SLASH_EQ] = ACTIONS(363),
    [anon_sym_PERCENT_EQ] = ACTIONS(363),
    [anon_sym_PLUS_EQ] = ACTIONS(363),
    [anon_sym_DASH_EQ] = ACTIONS(363),
    [anon_sym_LT_LT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(363),
    [anon_sym_AMP_EQ] = ACTIONS(363),
    [anon_sym_CARET_EQ] = ACTIONS(363),
    [anon_sym_PIPE_EQ] = ACTIONS(363),
    [anon_sym_QMARK_EQ] = ACTIONS(363),
    [anon_sym_PLUS_PLUS] = ACTIONS(363),
    [anon_sym_DASH_DASH] = ACTIONS(363),
    [anon_sym_PERCENT] = ACTIONS(365),
    [anon_sym_SLASH] = ACTIONS(365),
    [anon_sym_PLUS] = ACTIONS(365),
    [anon_sym_DASH] = ACTIONS(365),
    [anon_sym_LT_LT] = ACTIONS(365),
    [anon_sym_GT_GT] = ACTIONS(365),
    [anon_sym_GT_GT_GT] = ACTIONS(365),
    [anon_sym_DOT_DOT] = ACTIONS(365),
    [anon_sym_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(363),
    [anon_sym_LT_DOT_DOT] = ACTIONS(365),
    [anon_sym_LT] = ACTIONS(365),
    [anon_sym_LT_EQ] = ACTIONS(365),
    [anon_sym_GT] = ACTIONS(365),
    [anon_sym_GT_EQ] = ACTIONS(363),
    [anon_sym_in] = ACTIONS(365),
    [anon_sym_BANGin] = ACTIONS(365),
    [anon_sym_instanceof] = ACTIONS(365),
    [anon_sym_BANGinstanceof] = ACTIONS(363),
    [anon_sym_EQ_EQ] = ACTIONS(365),
    [anon_sym_BANG_EQ] = ACTIONS(365),
    [anon_sym_LT_EQ_GT] = ACTIONS(363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(363),
    [anon_sym_EQ_TILDE] = ACTIONS(363),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(363),
    [anon_sym_AMP] = ACTIONS(365),
    [anon_sym_CARET] = ACTIONS(365),
    [anon_sym_PIPE] = ACTIONS(365),
    [anon_sym_AMP_AMP] = ACTIONS(363),
    [anon_sym_PIPE_PIPE] = ACTIONS(363),
    [anon_sym_QMARK_COLON] = ACTIONS(363),
    [anon_sym_STAR_STAR] = ACTIONS(365),
    [anon_sym_true] = ACTIONS(365),
    [anon_sym_false] = ACTIONS(365),
    [anon_sym_COMMA] = ACTIONS(363),
    [anon_sym_LBRACE] = ACTIONS(363),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(363),
    [anon_sym_RPAREN] = ACTIONS(363),
    [anon_sym_null] = ACTIONS(365),
    [anon_sym_this] = ACTIONS(365),
    [anon_sym_while] = ACTIONS(365),
    [aux_sym__type_identifier_token1] = ACTIONS(365),
    [anon_sym_else] = ACTIONS(365),
    [anon_sym_LBRACK] = ACTIONS(363),
    [aux_sym_number_literal_token1] = ACTIONS(365),
    [aux_sym_number_literal_token2] = ACTIONS(363),
    [aux_sym_number_literal_token3] = ACTIONS(363),
    [aux_sym_number_literal_token4] = ACTIONS(365),
    [aux_sym_number_literal_token5] = ACTIONS(363),
    [anon_sym_SQUOTE] = ACTIONS(365),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(363),
    [anon_sym_DQUOTE] = ACTIONS(365),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(363),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(363),
    [anon_sym_QMARK] = ACTIONS(365),
    [anon_sym_catch] = ACTIONS(365),
    [anon_sym_finally] = ACTIONS(365),
    [anon_sym_TILDE] = ACTIONS(363),
    [anon_sym_BANG] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(365),
  },
  [STATE(666)] = {
    [sym_comment] = STATE(666),
    [sym_groovy_doc] = STATE(666),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_DOT_AMP] = ACTIONS(389),
    [anon_sym_DOT_AT] = ACTIONS(389),
    [anon_sym_QMARK_DOT] = ACTIONS(389),
    [anon_sym_STAR_DOT] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_EQ] = ACTIONS(387),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(389),
    [anon_sym_STAR_EQ] = ACTIONS(389),
    [anon_sym_SLASH_EQ] = ACTIONS(389),
    [anon_sym_PERCENT_EQ] = ACTIONS(389),
    [anon_sym_PLUS_EQ] = ACTIONS(389),
    [anon_sym_DASH_EQ] = ACTIONS(389),
    [anon_sym_LT_LT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(389),
    [anon_sym_AMP_EQ] = ACTIONS(389),
    [anon_sym_CARET_EQ] = ACTIONS(389),
    [anon_sym_PIPE_EQ] = ACTIONS(389),
    [anon_sym_QMARK_EQ] = ACTIONS(389),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_PERCENT] = ACTIONS(387),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_LT_LT] = ACTIONS(387),
    [anon_sym_GT_GT] = ACTIONS(387),
    [anon_sym_GT_GT_GT] = ACTIONS(387),
    [anon_sym_DOT_DOT] = ACTIONS(387),
    [anon_sym_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(389),
    [anon_sym_LT_DOT_DOT] = ACTIONS(387),
    [anon_sym_LT] = ACTIONS(387),
    [anon_sym_LT_EQ] = ACTIONS(387),
    [anon_sym_GT] = ACTIONS(387),
    [anon_sym_GT_EQ] = ACTIONS(389),
    [anon_sym_in] = ACTIONS(387),
    [anon_sym_BANGin] = ACTIONS(387),
    [anon_sym_instanceof] = ACTIONS(387),
    [anon_sym_BANGinstanceof] = ACTIONS(389),
    [anon_sym_EQ_EQ] = ACTIONS(387),
    [anon_sym_BANG_EQ] = ACTIONS(387),
    [anon_sym_LT_EQ_GT] = ACTIONS(389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(389),
    [anon_sym_EQ_TILDE] = ACTIONS(389),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(389),
    [anon_sym_AMP] = ACTIONS(387),
    [anon_sym_CARET] = ACTIONS(387),
    [anon_sym_PIPE] = ACTIONS(387),
    [anon_sym_AMP_AMP] = ACTIONS(389),
    [anon_sym_PIPE_PIPE] = ACTIONS(389),
    [anon_sym_QMARK_COLON] = ACTIONS(389),
    [anon_sym_STAR_STAR] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_COMMA] = ACTIONS(389),
    [anon_sym_LBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_RPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_QMARK] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(387),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(667)] = {
    [sym_comment] = STATE(667),
    [sym_groovy_doc] = STATE(667),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(371),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(371),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [anon_sym_LBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_RPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(371),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(668)] = {
    [sym_comment] = STATE(668),
    [sym_groovy_doc] = STATE(668),
    [sym_identifier] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(377),
    [anon_sym_DOT_AMP] = ACTIONS(377),
    [anon_sym_DOT_AT] = ACTIONS(377),
    [anon_sym_QMARK_DOT] = ACTIONS(377),
    [anon_sym_STAR_DOT] = ACTIONS(377),
    [anon_sym_STAR] = ACTIONS(379),
    [anon_sym_STAR_COLON] = ACTIONS(377),
    [anon_sym_DOT] = ACTIONS(379),
    [anon_sym_as] = ACTIONS(379),
    [anon_sym_EQ] = ACTIONS(379),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(377),
    [anon_sym_STAR_EQ] = ACTIONS(377),
    [anon_sym_SLASH_EQ] = ACTIONS(377),
    [anon_sym_PERCENT_EQ] = ACTIONS(377),
    [anon_sym_PLUS_EQ] = ACTIONS(377),
    [anon_sym_DASH_EQ] = ACTIONS(377),
    [anon_sym_LT_LT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(377),
    [anon_sym_AMP_EQ] = ACTIONS(377),
    [anon_sym_CARET_EQ] = ACTIONS(377),
    [anon_sym_PIPE_EQ] = ACTIONS(377),
    [anon_sym_QMARK_EQ] = ACTIONS(377),
    [anon_sym_PLUS_PLUS] = ACTIONS(377),
    [anon_sym_DASH_DASH] = ACTIONS(377),
    [anon_sym_PERCENT] = ACTIONS(379),
    [anon_sym_SLASH] = ACTIONS(379),
    [anon_sym_PLUS] = ACTIONS(379),
    [anon_sym_DASH] = ACTIONS(379),
    [anon_sym_LT_LT] = ACTIONS(379),
    [anon_sym_GT_GT] = ACTIONS(379),
    [anon_sym_GT_GT_GT] = ACTIONS(379),
    [anon_sym_DOT_DOT] = ACTIONS(379),
    [anon_sym_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(377),
    [anon_sym_LT_DOT_DOT] = ACTIONS(379),
    [anon_sym_LT] = ACTIONS(379),
    [anon_sym_LT_EQ] = ACTIONS(379),
    [anon_sym_GT] = ACTIONS(379),
    [anon_sym_GT_EQ] = ACTIONS(377),
    [anon_sym_in] = ACTIONS(379),
    [anon_sym_BANGin] = ACTIONS(379),
    [anon_sym_instanceof] = ACTIONS(379),
    [anon_sym_BANGinstanceof] = ACTIONS(377),
    [anon_sym_EQ_EQ] = ACTIONS(379),
    [anon_sym_BANG_EQ] = ACTIONS(379),
    [anon_sym_LT_EQ_GT] = ACTIONS(377),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(377),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(377),
    [anon_sym_EQ_TILDE] = ACTIONS(377),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(377),
    [anon_sym_AMP] = ACTIONS(379),
    [anon_sym_CARET] = ACTIONS(379),
    [anon_sym_PIPE] = ACTIONS(379),
    [anon_sym_AMP_AMP] = ACTIONS(377),
    [anon_sym_PIPE_PIPE] = ACTIONS(377),
    [anon_sym_QMARK_COLON] = ACTIONS(377),
    [anon_sym_STAR_STAR] = ACTIONS(379),
    [anon_sym_true] = ACTIONS(379),
    [anon_sym_false] = ACTIONS(379),
    [anon_sym_COMMA] = ACTIONS(377),
    [anon_sym_LBRACE] = ACTIONS(377),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(377),
    [anon_sym_RPAREN] = ACTIONS(377),
    [anon_sym_null] = ACTIONS(379),
    [anon_sym_this] = ACTIONS(379),
    [anon_sym_while] = ACTIONS(379),
    [aux_sym__type_identifier_token1] = ACTIONS(379),
    [anon_sym_else] = ACTIONS(379),
    [anon_sym_LBRACK] = ACTIONS(377),
    [aux_sym_number_literal_token1] = ACTIONS(379),
    [aux_sym_number_literal_token2] = ACTIONS(377),
    [aux_sym_number_literal_token3] = ACTIONS(377),
    [aux_sym_number_literal_token4] = ACTIONS(379),
    [aux_sym_number_literal_token5] = ACTIONS(377),
    [anon_sym_SQUOTE] = ACTIONS(379),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(377),
    [anon_sym_DQUOTE] = ACTIONS(379),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(377),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(377),
    [anon_sym_QMARK] = ACTIONS(379),
    [anon_sym_catch] = ACTIONS(379),
    [anon_sym_finally] = ACTIONS(379),
    [anon_sym_TILDE] = ACTIONS(377),
    [anon_sym_BANG] = ACTIONS(379),
    [anon_sym_new] = ACTIONS(379),
  },
  [STATE(669)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(669),
    [sym_groovy_doc] = STATE(669),
    [sym_declaration] = STATE(2349),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1507),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(670)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(670),
    [sym_groovy_doc] = STATE(670),
    [sym_declaration] = STATE(2379),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1518),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(671)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(671),
    [sym_groovy_doc] = STATE(671),
    [sym_declaration] = STATE(2417),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1536),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(672)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(672),
    [sym_groovy_doc] = STATE(672),
    [sym_declaration] = STATE(2394),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1501),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(673)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(673),
    [sym_groovy_doc] = STATE(673),
    [sym_declaration] = STATE(2438),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1491),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(674)] = {
    [sym_access_op] = STATE(1283),
    [sym_dotted_identifier] = STATE(1048),
    [sym_annotation] = STATE(1740),
    [sym_increment_op] = STATE(1283),
    [sym_binary_op] = STATE(1283),
    [sym_boolean_literal] = STATE(1072),
    [sym_closure] = STATE(1283),
    [sym_comment] = STATE(674),
    [sym_groovy_doc] = STATE(674),
    [sym_declaration] = STATE(2407),
    [sym_parenthesized_expression] = STATE(1037),
    [sym__expression] = STATE(1498),
    [sym__primary_expression] = STATE(1045),
    [sym__callable_expression] = STATE(1030),
    [sym__juxtable_expression] = STATE(1037),
    [sym_function_call] = STATE(1037),
    [sym__type_identifier] = STATE(1395),
    [sym_index] = STATE(1048),
    [sym_list] = STATE(1072),
    [sym_map] = STATE(1072),
    [sym_number_literal] = STATE(1072),
    [sym_string] = STATE(1072),
    [sym__plain_string] = STATE(1076),
    [sym__interpolate_string] = STATE(1076),
    [sym_ternary_op] = STATE(1283),
    [sym_builtintype] = STATE(1935),
    [sym__type] = STATE(2103),
    [sym_array_type] = STATE(1935),
    [sym_access_modifier] = STATE(1704),
    [sym_modifier] = STATE(1755),
    [sym_type_with_generics] = STATE(1935),
    [sym_unary_op] = STATE(1283),
    [aux_sym_class_definition_repeat1] = STATE(1655),
    [aux_sym_class_definition_repeat2] = STATE(1695),
    [sym_identifier] = ACTIONS(1727),
    [anon_sym_STAR] = ACTIONS(861),
    [anon_sym_STAR_COLON] = ACTIONS(863),
    [anon_sym_AT] = ACTIONS(1729),
    [anon_sym_PLUS_PLUS] = ACTIONS(871),
    [anon_sym_DASH_DASH] = ACTIONS(871),
    [anon_sym_SLASH] = ACTIONS(873),
    [anon_sym_PLUS] = ACTIONS(875),
    [anon_sym_DASH] = ACTIONS(875),
    [anon_sym_true] = ACTIONS(877),
    [anon_sym_false] = ACTIONS(877),
    [anon_sym_LBRACE] = ACTIONS(883),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(43),
    [anon_sym_def] = ACTIONS(1731),
    [anon_sym_LPAREN] = ACTIONS(1733),
    [anon_sym_null] = ACTIONS(891),
    [anon_sym_this] = ACTIONS(893),
    [aux_sym__type_identifier_token1] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(905),
    [aux_sym_number_literal_token1] = ACTIONS(907),
    [aux_sym_number_literal_token2] = ACTIONS(909),
    [aux_sym_number_literal_token3] = ACTIONS(909),
    [aux_sym_number_literal_token4] = ACTIONS(907),
    [aux_sym_number_literal_token5] = ACTIONS(909),
    [anon_sym_SQUOTE] = ACTIONS(913),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(915),
    [anon_sym_DQUOTE] = ACTIONS(917),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(919),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(921),
    [anon_sym_int] = ACTIONS(87),
    [anon_sym_boolean] = ACTIONS(87),
    [anon_sym_char] = ACTIONS(87),
    [anon_sym_short] = ACTIONS(87),
    [anon_sym_long] = ACTIONS(87),
    [anon_sym_float] = ACTIONS(87),
    [anon_sym_double] = ACTIONS(87),
    [anon_sym_void] = ACTIONS(87),
    [anon_sym_public] = ACTIONS(1735),
    [anon_sym_protected] = ACTIONS(1735),
    [anon_sym_private] = ACTIONS(1735),
    [anon_sym_static] = ACTIONS(1737),
    [anon_sym_final] = ACTIONS(1737),
    [anon_sym_synchronized] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(925),
    [anon_sym_BANG] = ACTIONS(925),
    [anon_sym_new] = ACTIONS(927),
  },
  [STATE(675)] = {
    [sym_comment] = STATE(675),
    [sym_groovy_doc] = STATE(675),
    [sym_parameter_list] = STATE(769),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(1743),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_case] = ACTIONS(1739),
    [anon_sym_default] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_catch] = ACTIONS(1739),
    [anon_sym_finally] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(676)] = {
    [sym_comment] = STATE(676),
    [sym_groovy_doc] = STATE(676),
    [ts_builtin_sym_end] = ACTIONS(1747),
    [sym_identifier] = ACTIONS(1749),
    [anon_sym_break] = ACTIONS(1749),
    [anon_sym_continue] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_STAR] = ACTIONS(1749),
    [anon_sym_STAR_COLON] = ACTIONS(1747),
    [anon_sym_import] = ACTIONS(1749),
    [anon_sym_package] = ACTIONS(1749),
    [anon_sym_AT] = ACTIONS(1749),
    [anon_sym_assert] = ACTIONS(1749),
    [anon_sym_PLUS_PLUS] = ACTIONS(1747),
    [anon_sym_DASH_DASH] = ACTIONS(1747),
    [anon_sym_SLASH] = ACTIONS(1749),
    [anon_sym_PLUS] = ACTIONS(1749),
    [anon_sym_DASH] = ACTIONS(1749),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_ATinterface] = ACTIONS(1747),
    [anon_sym_interface] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_LBRACE] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1749),
    [anon_sym_def] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_null] = ACTIONS(1749),
    [anon_sym_this] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_while] = ACTIONS(1749),
    [anon_sym_for] = ACTIONS(1749),
    [aux_sym__type_identifier_token1] = ACTIONS(1749),
    [anon_sym_if] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1747),
    [aux_sym_number_literal_token1] = ACTIONS(1749),
    [aux_sym_number_literal_token2] = ACTIONS(1747),
    [aux_sym_number_literal_token3] = ACTIONS(1747),
    [aux_sym_number_literal_token4] = ACTIONS(1749),
    [aux_sym_number_literal_token5] = ACTIONS(1747),
    [anon_sym_pipeline] = ACTIONS(1749),
    [anon_sym_return] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1749),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1749),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1747),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1747),
    [anon_sym_switch] = ACTIONS(1749),
    [anon_sym_try] = ACTIONS(1749),
    [anon_sym_catch] = ACTIONS(1749),
    [anon_sym_finally] = ACTIONS(1749),
    [anon_sym_int] = ACTIONS(1749),
    [anon_sym_boolean] = ACTIONS(1749),
    [anon_sym_char] = ACTIONS(1749),
    [anon_sym_short] = ACTIONS(1749),
    [anon_sym_long] = ACTIONS(1749),
    [anon_sym_float] = ACTIONS(1749),
    [anon_sym_double] = ACTIONS(1749),
    [anon_sym_void] = ACTIONS(1749),
    [anon_sym_public] = ACTIONS(1749),
    [anon_sym_protected] = ACTIONS(1749),
    [anon_sym_private] = ACTIONS(1749),
    [anon_sym_static] = ACTIONS(1749),
    [anon_sym_final] = ACTIONS(1749),
    [anon_sym_synchronized] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1747),
    [anon_sym_BANG] = ACTIONS(1747),
    [anon_sym_new] = ACTIONS(1749),
  },
  [STATE(677)] = {
    [sym_comment] = STATE(677),
    [sym_groovy_doc] = STATE(677),
    [sym_parameter_list] = STATE(761),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_break] = ACTIONS(1751),
    [anon_sym_continue] = ACTIONS(1751),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_STAR] = ACTIONS(1751),
    [anon_sym_STAR_COLON] = ACTIONS(1753),
    [anon_sym_import] = ACTIONS(1751),
    [anon_sym_package] = ACTIONS(1751),
    [anon_sym_AT] = ACTIONS(1751),
    [anon_sym_assert] = ACTIONS(1751),
    [anon_sym_EQ] = ACTIONS(1755),
    [anon_sym_PLUS_PLUS] = ACTIONS(1753),
    [anon_sym_DASH_DASH] = ACTIONS(1753),
    [anon_sym_SLASH] = ACTIONS(1751),
    [anon_sym_PLUS] = ACTIONS(1751),
    [anon_sym_DASH] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1751),
    [anon_sym_false] = ACTIONS(1751),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_interface] = ACTIONS(1751),
    [anon_sym_class] = ACTIONS(1751),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_RBRACE] = ACTIONS(1753),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1751),
    [anon_sym_def] = ACTIONS(1751),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1751),
    [anon_sym_this] = ACTIONS(1751),
    [anon_sym_do] = ACTIONS(1751),
    [anon_sym_while] = ACTIONS(1751),
    [anon_sym_for] = ACTIONS(1751),
    [aux_sym__type_identifier_token1] = ACTIONS(1751),
    [anon_sym_if] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [aux_sym_number_literal_token1] = ACTIONS(1751),
    [aux_sym_number_literal_token2] = ACTIONS(1753),
    [aux_sym_number_literal_token3] = ACTIONS(1753),
    [aux_sym_number_literal_token4] = ACTIONS(1751),
    [aux_sym_number_literal_token5] = ACTIONS(1753),
    [anon_sym_return] = ACTIONS(1751),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1753),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1753),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1753),
    [anon_sym_switch] = ACTIONS(1751),
    [anon_sym_case] = ACTIONS(1751),
    [anon_sym_default] = ACTIONS(1751),
    [anon_sym_try] = ACTIONS(1751),
    [anon_sym_catch] = ACTIONS(1751),
    [anon_sym_finally] = ACTIONS(1751),
    [anon_sym_int] = ACTIONS(1751),
    [anon_sym_boolean] = ACTIONS(1751),
    [anon_sym_char] = ACTIONS(1751),
    [anon_sym_short] = ACTIONS(1751),
    [anon_sym_long] = ACTIONS(1751),
    [anon_sym_float] = ACTIONS(1751),
    [anon_sym_double] = ACTIONS(1751),
    [anon_sym_void] = ACTIONS(1751),
    [anon_sym_public] = ACTIONS(1751),
    [anon_sym_protected] = ACTIONS(1751),
    [anon_sym_private] = ACTIONS(1751),
    [anon_sym_static] = ACTIONS(1751),
    [anon_sym_final] = ACTIONS(1751),
    [anon_sym_synchronized] = ACTIONS(1751),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_BANG] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1751),
  },
  [STATE(678)] = {
    [sym_comment] = STATE(678),
    [sym_groovy_doc] = STATE(678),
    [aux_sym__juxt_argument_list_repeat1] = STATE(716),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(679)] = {
    [sym_comment] = STATE(679),
    [sym_groovy_doc] = STATE(679),
    [ts_builtin_sym_end] = ACTIONS(1757),
    [sym_identifier] = ACTIONS(1759),
    [anon_sym_break] = ACTIONS(1759),
    [anon_sym_continue] = ACTIONS(1759),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_STAR] = ACTIONS(1759),
    [anon_sym_STAR_COLON] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1757),
    [anon_sym_import] = ACTIONS(1759),
    [anon_sym_as] = ACTIONS(1759),
    [anon_sym_package] = ACTIONS(1759),
    [anon_sym_AT] = ACTIONS(1759),
    [anon_sym_assert] = ACTIONS(1759),
    [anon_sym_PLUS_PLUS] = ACTIONS(1757),
    [anon_sym_DASH_DASH] = ACTIONS(1757),
    [anon_sym_SLASH] = ACTIONS(1759),
    [anon_sym_PLUS] = ACTIONS(1759),
    [anon_sym_DASH] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1759),
    [anon_sym_false] = ACTIONS(1759),
    [anon_sym_ATinterface] = ACTIONS(1757),
    [anon_sym_interface] = ACTIONS(1759),
    [anon_sym_class] = ACTIONS(1759),
    [anon_sym_COMMA] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_RBRACE] = ACTIONS(1757),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1759),
    [anon_sym_def] = ACTIONS(1759),
    [anon_sym_LPAREN] = ACTIONS(1757),
    [anon_sym_RPAREN] = ACTIONS(1757),
    [anon_sym_null] = ACTIONS(1759),
    [anon_sym_this] = ACTIONS(1759),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_while] = ACTIONS(1759),
    [anon_sym_for] = ACTIONS(1759),
    [aux_sym__type_identifier_token1] = ACTIONS(1759),
    [anon_sym_if] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [aux_sym_number_literal_token1] = ACTIONS(1759),
    [aux_sym_number_literal_token2] = ACTIONS(1757),
    [aux_sym_number_literal_token3] = ACTIONS(1757),
    [aux_sym_number_literal_token4] = ACTIONS(1759),
    [aux_sym_number_literal_token5] = ACTIONS(1757),
    [anon_sym_pipeline] = ACTIONS(1759),
    [anon_sym_return] = ACTIONS(1759),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1757),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1757),
    [anon_sym_switch] = ACTIONS(1759),
    [anon_sym_try] = ACTIONS(1759),
    [anon_sym_int] = ACTIONS(1759),
    [anon_sym_boolean] = ACTIONS(1759),
    [anon_sym_char] = ACTIONS(1759),
    [anon_sym_short] = ACTIONS(1759),
    [anon_sym_long] = ACTIONS(1759),
    [anon_sym_float] = ACTIONS(1759),
    [anon_sym_double] = ACTIONS(1759),
    [anon_sym_void] = ACTIONS(1759),
    [anon_sym_public] = ACTIONS(1759),
    [anon_sym_protected] = ACTIONS(1759),
    [anon_sym_private] = ACTIONS(1759),
    [anon_sym_static] = ACTIONS(1759),
    [anon_sym_final] = ACTIONS(1759),
    [anon_sym_synchronized] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_BANG] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1759),
  },
  [STATE(680)] = {
    [sym_comment] = STATE(680),
    [sym_groovy_doc] = STATE(680),
    [sym_parameter_list] = STATE(761),
    [sym_identifier] = ACTIONS(1761),
    [anon_sym_break] = ACTIONS(1761),
    [anon_sym_continue] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_STAR] = ACTIONS(1761),
    [anon_sym_STAR_COLON] = ACTIONS(1763),
    [anon_sym_import] = ACTIONS(1761),
    [anon_sym_package] = ACTIONS(1761),
    [anon_sym_AT] = ACTIONS(1761),
    [anon_sym_assert] = ACTIONS(1761),
    [anon_sym_EQ] = ACTIONS(1765),
    [anon_sym_PLUS_PLUS] = ACTIONS(1763),
    [anon_sym_DASH_DASH] = ACTIONS(1763),
    [anon_sym_SLASH] = ACTIONS(1761),
    [anon_sym_PLUS] = ACTIONS(1761),
    [anon_sym_DASH] = ACTIONS(1761),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_ATinterface] = ACTIONS(1763),
    [anon_sym_interface] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1761),
    [anon_sym_def] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1761),
    [anon_sym_this] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_while] = ACTIONS(1761),
    [anon_sym_for] = ACTIONS(1761),
    [aux_sym__type_identifier_token1] = ACTIONS(1761),
    [anon_sym_if] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1763),
    [aux_sym_number_literal_token1] = ACTIONS(1761),
    [aux_sym_number_literal_token2] = ACTIONS(1763),
    [aux_sym_number_literal_token3] = ACTIONS(1763),
    [aux_sym_number_literal_token4] = ACTIONS(1761),
    [aux_sym_number_literal_token5] = ACTIONS(1763),
    [anon_sym_return] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1761),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1761),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1763),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1763),
    [anon_sym_switch] = ACTIONS(1761),
    [anon_sym_case] = ACTIONS(1761),
    [anon_sym_default] = ACTIONS(1761),
    [anon_sym_try] = ACTIONS(1761),
    [anon_sym_catch] = ACTIONS(1761),
    [anon_sym_finally] = ACTIONS(1761),
    [anon_sym_int] = ACTIONS(1761),
    [anon_sym_boolean] = ACTIONS(1761),
    [anon_sym_char] = ACTIONS(1761),
    [anon_sym_short] = ACTIONS(1761),
    [anon_sym_long] = ACTIONS(1761),
    [anon_sym_float] = ACTIONS(1761),
    [anon_sym_double] = ACTIONS(1761),
    [anon_sym_void] = ACTIONS(1761),
    [anon_sym_public] = ACTIONS(1761),
    [anon_sym_protected] = ACTIONS(1761),
    [anon_sym_private] = ACTIONS(1761),
    [anon_sym_static] = ACTIONS(1761),
    [anon_sym_final] = ACTIONS(1761),
    [anon_sym_synchronized] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1763),
    [anon_sym_BANG] = ACTIONS(1763),
    [anon_sym_new] = ACTIONS(1761),
  },
  [STATE(681)] = {
    [sym_comment] = STATE(681),
    [sym_groovy_doc] = STATE(681),
    [ts_builtin_sym_end] = ACTIONS(1767),
    [sym_identifier] = ACTIONS(1769),
    [anon_sym_break] = ACTIONS(1769),
    [anon_sym_continue] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_STAR] = ACTIONS(1769),
    [anon_sym_STAR_COLON] = ACTIONS(1767),
    [anon_sym_import] = ACTIONS(1769),
    [anon_sym_package] = ACTIONS(1769),
    [anon_sym_AT] = ACTIONS(1769),
    [anon_sym_assert] = ACTIONS(1769),
    [anon_sym_PLUS_PLUS] = ACTIONS(1767),
    [anon_sym_DASH_DASH] = ACTIONS(1767),
    [anon_sym_SLASH] = ACTIONS(1769),
    [anon_sym_PLUS] = ACTIONS(1769),
    [anon_sym_DASH] = ACTIONS(1769),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_ATinterface] = ACTIONS(1767),
    [anon_sym_interface] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1769),
    [anon_sym_def] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_RPAREN] = ACTIONS(1767),
    [anon_sym_null] = ACTIONS(1769),
    [anon_sym_this] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_while] = ACTIONS(1769),
    [anon_sym_for] = ACTIONS(1769),
    [aux_sym__type_identifier_token1] = ACTIONS(1769),
    [anon_sym_if] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [aux_sym_number_literal_token1] = ACTIONS(1769),
    [aux_sym_number_literal_token2] = ACTIONS(1767),
    [aux_sym_number_literal_token3] = ACTIONS(1767),
    [aux_sym_number_literal_token4] = ACTIONS(1769),
    [aux_sym_number_literal_token5] = ACTIONS(1767),
    [anon_sym_pipeline] = ACTIONS(1769),
    [anon_sym_return] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1769),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1769),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1767),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1767),
    [anon_sym_switch] = ACTIONS(1769),
    [anon_sym_try] = ACTIONS(1769),
    [anon_sym_catch] = ACTIONS(1769),
    [anon_sym_finally] = ACTIONS(1769),
    [anon_sym_int] = ACTIONS(1769),
    [anon_sym_boolean] = ACTIONS(1769),
    [anon_sym_char] = ACTIONS(1769),
    [anon_sym_short] = ACTIONS(1769),
    [anon_sym_long] = ACTIONS(1769),
    [anon_sym_float] = ACTIONS(1769),
    [anon_sym_double] = ACTIONS(1769),
    [anon_sym_void] = ACTIONS(1769),
    [anon_sym_public] = ACTIONS(1769),
    [anon_sym_protected] = ACTIONS(1769),
    [anon_sym_private] = ACTIONS(1769),
    [anon_sym_static] = ACTIONS(1769),
    [anon_sym_final] = ACTIONS(1769),
    [anon_sym_synchronized] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1767),
    [anon_sym_BANG] = ACTIONS(1767),
    [anon_sym_new] = ACTIONS(1769),
  },
  [STATE(682)] = {
    [sym_comment] = STATE(682),
    [sym_groovy_doc] = STATE(682),
    [sym_parameter_list] = STATE(768),
    [ts_builtin_sym_end] = ACTIONS(1771),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(1775),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_pipeline] = ACTIONS(1773),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_catch] = ACTIONS(1773),
    [anon_sym_finally] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(683)] = {
    [sym_comment] = STATE(683),
    [sym_groovy_doc] = STATE(683),
    [ts_builtin_sym_end] = ACTIONS(1779),
    [sym_identifier] = ACTIONS(1781),
    [anon_sym_break] = ACTIONS(1781),
    [anon_sym_continue] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_STAR] = ACTIONS(1781),
    [anon_sym_STAR_COLON] = ACTIONS(1779),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_import] = ACTIONS(1781),
    [anon_sym_as] = ACTIONS(1781),
    [anon_sym_package] = ACTIONS(1781),
    [anon_sym_AT] = ACTIONS(1781),
    [anon_sym_assert] = ACTIONS(1781),
    [anon_sym_PLUS_PLUS] = ACTIONS(1779),
    [anon_sym_DASH_DASH] = ACTIONS(1779),
    [anon_sym_SLASH] = ACTIONS(1781),
    [anon_sym_PLUS] = ACTIONS(1781),
    [anon_sym_DASH] = ACTIONS(1781),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_ATinterface] = ACTIONS(1779),
    [anon_sym_interface] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_COMMA] = ACTIONS(1779),
    [anon_sym_LBRACE] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1781),
    [anon_sym_def] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_RPAREN] = ACTIONS(1779),
    [anon_sym_null] = ACTIONS(1781),
    [anon_sym_this] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_while] = ACTIONS(1781),
    [anon_sym_for] = ACTIONS(1781),
    [aux_sym__type_identifier_token1] = ACTIONS(1781),
    [anon_sym_if] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1779),
    [aux_sym_number_literal_token1] = ACTIONS(1781),
    [aux_sym_number_literal_token2] = ACTIONS(1779),
    [aux_sym_number_literal_token3] = ACTIONS(1779),
    [aux_sym_number_literal_token4] = ACTIONS(1781),
    [aux_sym_number_literal_token5] = ACTIONS(1779),
    [anon_sym_pipeline] = ACTIONS(1781),
    [anon_sym_return] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1781),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1781),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1779),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1779),
    [anon_sym_switch] = ACTIONS(1781),
    [anon_sym_try] = ACTIONS(1781),
    [anon_sym_int] = ACTIONS(1781),
    [anon_sym_boolean] = ACTIONS(1781),
    [anon_sym_char] = ACTIONS(1781),
    [anon_sym_short] = ACTIONS(1781),
    [anon_sym_long] = ACTIONS(1781),
    [anon_sym_float] = ACTIONS(1781),
    [anon_sym_double] = ACTIONS(1781),
    [anon_sym_void] = ACTIONS(1781),
    [anon_sym_public] = ACTIONS(1781),
    [anon_sym_protected] = ACTIONS(1781),
    [anon_sym_private] = ACTIONS(1781),
    [anon_sym_static] = ACTIONS(1781),
    [anon_sym_final] = ACTIONS(1781),
    [anon_sym_synchronized] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1779),
    [anon_sym_BANG] = ACTIONS(1779),
    [anon_sym_new] = ACTIONS(1781),
  },
  [STATE(684)] = {
    [sym_comment] = STATE(684),
    [sym_groovy_doc] = STATE(684),
    [ts_builtin_sym_end] = ACTIONS(1783),
    [sym_identifier] = ACTIONS(1785),
    [anon_sym_break] = ACTIONS(1785),
    [anon_sym_continue] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_STAR] = ACTIONS(1785),
    [anon_sym_STAR_COLON] = ACTIONS(1783),
    [anon_sym_import] = ACTIONS(1785),
    [anon_sym_package] = ACTIONS(1785),
    [anon_sym_AT] = ACTIONS(1785),
    [anon_sym_assert] = ACTIONS(1785),
    [anon_sym_PLUS_PLUS] = ACTIONS(1783),
    [anon_sym_DASH_DASH] = ACTIONS(1783),
    [anon_sym_SLASH] = ACTIONS(1785),
    [anon_sym_PLUS] = ACTIONS(1785),
    [anon_sym_DASH] = ACTIONS(1785),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_ATinterface] = ACTIONS(1783),
    [anon_sym_interface] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_LBRACE] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1785),
    [anon_sym_def] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_RPAREN] = ACTIONS(1783),
    [anon_sym_null] = ACTIONS(1785),
    [anon_sym_this] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1785),
    [anon_sym_for] = ACTIONS(1785),
    [aux_sym__type_identifier_token1] = ACTIONS(1785),
    [anon_sym_if] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1783),
    [aux_sym_number_literal_token1] = ACTIONS(1785),
    [aux_sym_number_literal_token2] = ACTIONS(1783),
    [aux_sym_number_literal_token3] = ACTIONS(1783),
    [aux_sym_number_literal_token4] = ACTIONS(1785),
    [aux_sym_number_literal_token5] = ACTIONS(1783),
    [anon_sym_pipeline] = ACTIONS(1785),
    [anon_sym_return] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1785),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1785),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1783),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1783),
    [anon_sym_switch] = ACTIONS(1785),
    [anon_sym_try] = ACTIONS(1785),
    [anon_sym_catch] = ACTIONS(1785),
    [anon_sym_finally] = ACTIONS(1785),
    [anon_sym_int] = ACTIONS(1785),
    [anon_sym_boolean] = ACTIONS(1785),
    [anon_sym_char] = ACTIONS(1785),
    [anon_sym_short] = ACTIONS(1785),
    [anon_sym_long] = ACTIONS(1785),
    [anon_sym_float] = ACTIONS(1785),
    [anon_sym_double] = ACTIONS(1785),
    [anon_sym_void] = ACTIONS(1785),
    [anon_sym_public] = ACTIONS(1785),
    [anon_sym_protected] = ACTIONS(1785),
    [anon_sym_private] = ACTIONS(1785),
    [anon_sym_static] = ACTIONS(1785),
    [anon_sym_final] = ACTIONS(1785),
    [anon_sym_synchronized] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1783),
    [anon_sym_BANG] = ACTIONS(1783),
    [anon_sym_new] = ACTIONS(1785),
  },
  [STATE(685)] = {
    [sym_comment] = STATE(685),
    [sym_groovy_doc] = STATE(685),
    [aux_sym__juxt_argument_list_repeat1] = STATE(685),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1787),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(686)] = {
    [sym_comment] = STATE(686),
    [sym_groovy_doc] = STATE(686),
    [sym_parameter_list] = STATE(768),
    [ts_builtin_sym_end] = ACTIONS(1790),
    [sym_identifier] = ACTIONS(1792),
    [anon_sym_break] = ACTIONS(1792),
    [anon_sym_continue] = ACTIONS(1792),
    [anon_sym_SEMI] = ACTIONS(1790),
    [anon_sym_STAR] = ACTIONS(1792),
    [anon_sym_STAR_COLON] = ACTIONS(1790),
    [anon_sym_import] = ACTIONS(1792),
    [anon_sym_package] = ACTIONS(1792),
    [anon_sym_AT] = ACTIONS(1792),
    [anon_sym_assert] = ACTIONS(1792),
    [anon_sym_EQ] = ACTIONS(1794),
    [anon_sym_PLUS_PLUS] = ACTIONS(1790),
    [anon_sym_DASH_DASH] = ACTIONS(1790),
    [anon_sym_SLASH] = ACTIONS(1792),
    [anon_sym_PLUS] = ACTIONS(1792),
    [anon_sym_DASH] = ACTIONS(1792),
    [anon_sym_true] = ACTIONS(1792),
    [anon_sym_false] = ACTIONS(1792),
    [anon_sym_ATinterface] = ACTIONS(1790),
    [anon_sym_interface] = ACTIONS(1792),
    [anon_sym_class] = ACTIONS(1792),
    [anon_sym_LBRACE] = ACTIONS(1790),
    [anon_sym_RBRACE] = ACTIONS(1790),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1792),
    [anon_sym_def] = ACTIONS(1792),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1792),
    [anon_sym_this] = ACTIONS(1792),
    [anon_sym_do] = ACTIONS(1792),
    [anon_sym_while] = ACTIONS(1792),
    [anon_sym_for] = ACTIONS(1792),
    [aux_sym__type_identifier_token1] = ACTIONS(1792),
    [anon_sym_if] = ACTIONS(1792),
    [anon_sym_else] = ACTIONS(1792),
    [anon_sym_LBRACK] = ACTIONS(1790),
    [aux_sym_number_literal_token1] = ACTIONS(1792),
    [aux_sym_number_literal_token2] = ACTIONS(1790),
    [aux_sym_number_literal_token3] = ACTIONS(1790),
    [aux_sym_number_literal_token4] = ACTIONS(1792),
    [aux_sym_number_literal_token5] = ACTIONS(1790),
    [anon_sym_pipeline] = ACTIONS(1792),
    [anon_sym_return] = ACTIONS(1792),
    [anon_sym_SQUOTE] = ACTIONS(1792),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1790),
    [anon_sym_DQUOTE] = ACTIONS(1792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1790),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1790),
    [anon_sym_switch] = ACTIONS(1792),
    [anon_sym_try] = ACTIONS(1792),
    [anon_sym_catch] = ACTIONS(1792),
    [anon_sym_finally] = ACTIONS(1792),
    [anon_sym_int] = ACTIONS(1792),
    [anon_sym_boolean] = ACTIONS(1792),
    [anon_sym_char] = ACTIONS(1792),
    [anon_sym_short] = ACTIONS(1792),
    [anon_sym_long] = ACTIONS(1792),
    [anon_sym_float] = ACTIONS(1792),
    [anon_sym_double] = ACTIONS(1792),
    [anon_sym_void] = ACTIONS(1792),
    [anon_sym_public] = ACTIONS(1792),
    [anon_sym_protected] = ACTIONS(1792),
    [anon_sym_private] = ACTIONS(1792),
    [anon_sym_static] = ACTIONS(1792),
    [anon_sym_final] = ACTIONS(1792),
    [anon_sym_synchronized] = ACTIONS(1792),
    [anon_sym_TILDE] = ACTIONS(1790),
    [anon_sym_BANG] = ACTIONS(1790),
    [anon_sym_new] = ACTIONS(1792),
  },
  [STATE(687)] = {
    [sym_comment] = STATE(687),
    [sym_groovy_doc] = STATE(687),
    [ts_builtin_sym_end] = ACTIONS(1796),
    [sym_identifier] = ACTIONS(1798),
    [anon_sym_break] = ACTIONS(1798),
    [anon_sym_continue] = ACTIONS(1798),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym_STAR] = ACTIONS(1798),
    [anon_sym_STAR_COLON] = ACTIONS(1796),
    [anon_sym_import] = ACTIONS(1798),
    [anon_sym_package] = ACTIONS(1798),
    [anon_sym_AT] = ACTIONS(1798),
    [anon_sym_assert] = ACTIONS(1798),
    [anon_sym_PLUS_PLUS] = ACTIONS(1796),
    [anon_sym_DASH_DASH] = ACTIONS(1796),
    [anon_sym_SLASH] = ACTIONS(1798),
    [anon_sym_PLUS] = ACTIONS(1798),
    [anon_sym_DASH] = ACTIONS(1798),
    [anon_sym_true] = ACTIONS(1798),
    [anon_sym_false] = ACTIONS(1798),
    [anon_sym_ATinterface] = ACTIONS(1796),
    [anon_sym_interface] = ACTIONS(1798),
    [anon_sym_class] = ACTIONS(1798),
    [anon_sym_COMMA] = ACTIONS(1796),
    [anon_sym_LBRACE] = ACTIONS(1796),
    [anon_sym_RBRACE] = ACTIONS(1796),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1798),
    [anon_sym_def] = ACTIONS(1798),
    [anon_sym_LPAREN] = ACTIONS(1796),
    [anon_sym_RPAREN] = ACTIONS(1796),
    [anon_sym_null] = ACTIONS(1798),
    [anon_sym_this] = ACTIONS(1798),
    [anon_sym_do] = ACTIONS(1798),
    [anon_sym_while] = ACTIONS(1798),
    [anon_sym_for] = ACTIONS(1798),
    [aux_sym__type_identifier_token1] = ACTIONS(1798),
    [anon_sym_if] = ACTIONS(1798),
    [anon_sym_else] = ACTIONS(1798),
    [anon_sym_LBRACK] = ACTIONS(1796),
    [aux_sym_number_literal_token1] = ACTIONS(1798),
    [aux_sym_number_literal_token2] = ACTIONS(1796),
    [aux_sym_number_literal_token3] = ACTIONS(1796),
    [aux_sym_number_literal_token4] = ACTIONS(1798),
    [aux_sym_number_literal_token5] = ACTIONS(1796),
    [anon_sym_pipeline] = ACTIONS(1798),
    [anon_sym_return] = ACTIONS(1798),
    [anon_sym_SQUOTE] = ACTIONS(1798),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1796),
    [anon_sym_DQUOTE] = ACTIONS(1798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1796),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1796),
    [anon_sym_switch] = ACTIONS(1798),
    [anon_sym_try] = ACTIONS(1798),
    [anon_sym_catch] = ACTIONS(1798),
    [anon_sym_finally] = ACTIONS(1798),
    [anon_sym_int] = ACTIONS(1798),
    [anon_sym_boolean] = ACTIONS(1798),
    [anon_sym_char] = ACTIONS(1798),
    [anon_sym_short] = ACTIONS(1798),
    [anon_sym_long] = ACTIONS(1798),
    [anon_sym_float] = ACTIONS(1798),
    [anon_sym_double] = ACTIONS(1798),
    [anon_sym_void] = ACTIONS(1798),
    [anon_sym_public] = ACTIONS(1798),
    [anon_sym_protected] = ACTIONS(1798),
    [anon_sym_private] = ACTIONS(1798),
    [anon_sym_static] = ACTIONS(1798),
    [anon_sym_final] = ACTIONS(1798),
    [anon_sym_synchronized] = ACTIONS(1798),
    [anon_sym_TILDE] = ACTIONS(1796),
    [anon_sym_BANG] = ACTIONS(1796),
    [anon_sym_new] = ACTIONS(1798),
  },
  [STATE(688)] = {
    [sym_comment] = STATE(688),
    [sym_groovy_doc] = STATE(688),
    [ts_builtin_sym_end] = ACTIONS(1800),
    [sym_identifier] = ACTIONS(1802),
    [anon_sym_break] = ACTIONS(1802),
    [anon_sym_continue] = ACTIONS(1802),
    [anon_sym_SEMI] = ACTIONS(1800),
    [anon_sym_STAR] = ACTIONS(1802),
    [anon_sym_STAR_COLON] = ACTIONS(1800),
    [anon_sym_import] = ACTIONS(1802),
    [anon_sym_package] = ACTIONS(1802),
    [anon_sym_AT] = ACTIONS(1802),
    [anon_sym_assert] = ACTIONS(1802),
    [anon_sym_PLUS_PLUS] = ACTIONS(1800),
    [anon_sym_DASH_DASH] = ACTIONS(1800),
    [anon_sym_SLASH] = ACTIONS(1802),
    [anon_sym_PLUS] = ACTIONS(1802),
    [anon_sym_DASH] = ACTIONS(1802),
    [anon_sym_true] = ACTIONS(1802),
    [anon_sym_false] = ACTIONS(1802),
    [anon_sym_ATinterface] = ACTIONS(1800),
    [anon_sym_interface] = ACTIONS(1802),
    [anon_sym_class] = ACTIONS(1802),
    [anon_sym_COMMA] = ACTIONS(1800),
    [anon_sym_LBRACE] = ACTIONS(1800),
    [anon_sym_RBRACE] = ACTIONS(1800),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1802),
    [anon_sym_def] = ACTIONS(1802),
    [anon_sym_LPAREN] = ACTIONS(1800),
    [anon_sym_RPAREN] = ACTIONS(1800),
    [anon_sym_null] = ACTIONS(1802),
    [anon_sym_this] = ACTIONS(1802),
    [anon_sym_do] = ACTIONS(1802),
    [anon_sym_while] = ACTIONS(1802),
    [anon_sym_for] = ACTIONS(1802),
    [aux_sym__type_identifier_token1] = ACTIONS(1802),
    [anon_sym_if] = ACTIONS(1802),
    [anon_sym_else] = ACTIONS(1802),
    [anon_sym_LBRACK] = ACTIONS(1800),
    [aux_sym_number_literal_token1] = ACTIONS(1802),
    [aux_sym_number_literal_token2] = ACTIONS(1800),
    [aux_sym_number_literal_token3] = ACTIONS(1800),
    [aux_sym_number_literal_token4] = ACTIONS(1802),
    [aux_sym_number_literal_token5] = ACTIONS(1800),
    [anon_sym_pipeline] = ACTIONS(1802),
    [anon_sym_return] = ACTIONS(1802),
    [anon_sym_SQUOTE] = ACTIONS(1802),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1800),
    [anon_sym_DQUOTE] = ACTIONS(1802),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1800),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1800),
    [anon_sym_switch] = ACTIONS(1802),
    [anon_sym_try] = ACTIONS(1802),
    [anon_sym_catch] = ACTIONS(1802),
    [anon_sym_finally] = ACTIONS(1802),
    [anon_sym_int] = ACTIONS(1802),
    [anon_sym_boolean] = ACTIONS(1802),
    [anon_sym_char] = ACTIONS(1802),
    [anon_sym_short] = ACTIONS(1802),
    [anon_sym_long] = ACTIONS(1802),
    [anon_sym_float] = ACTIONS(1802),
    [anon_sym_double] = ACTIONS(1802),
    [anon_sym_void] = ACTIONS(1802),
    [anon_sym_public] = ACTIONS(1802),
    [anon_sym_protected] = ACTIONS(1802),
    [anon_sym_private] = ACTIONS(1802),
    [anon_sym_static] = ACTIONS(1802),
    [anon_sym_final] = ACTIONS(1802),
    [anon_sym_synchronized] = ACTIONS(1802),
    [anon_sym_TILDE] = ACTIONS(1800),
    [anon_sym_BANG] = ACTIONS(1800),
    [anon_sym_new] = ACTIONS(1802),
  },
  [STATE(689)] = {
    [sym_comment] = STATE(689),
    [sym_groovy_doc] = STATE(689),
    [ts_builtin_sym_end] = ACTIONS(1804),
    [sym_identifier] = ACTIONS(1806),
    [anon_sym_break] = ACTIONS(1806),
    [anon_sym_continue] = ACTIONS(1806),
    [anon_sym_SEMI] = ACTIONS(1804),
    [anon_sym_STAR] = ACTIONS(1806),
    [anon_sym_STAR_COLON] = ACTIONS(1804),
    [anon_sym_import] = ACTIONS(1806),
    [anon_sym_package] = ACTIONS(1806),
    [anon_sym_AT] = ACTIONS(1806),
    [anon_sym_assert] = ACTIONS(1806),
    [anon_sym_PLUS_PLUS] = ACTIONS(1804),
    [anon_sym_DASH_DASH] = ACTIONS(1804),
    [anon_sym_SLASH] = ACTIONS(1806),
    [anon_sym_PLUS] = ACTIONS(1806),
    [anon_sym_DASH] = ACTIONS(1806),
    [anon_sym_true] = ACTIONS(1806),
    [anon_sym_false] = ACTIONS(1806),
    [anon_sym_ATinterface] = ACTIONS(1804),
    [anon_sym_interface] = ACTIONS(1806),
    [anon_sym_class] = ACTIONS(1806),
    [anon_sym_COMMA] = ACTIONS(1804),
    [anon_sym_LBRACE] = ACTIONS(1804),
    [anon_sym_RBRACE] = ACTIONS(1804),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1806),
    [anon_sym_def] = ACTIONS(1806),
    [anon_sym_LPAREN] = ACTIONS(1804),
    [anon_sym_RPAREN] = ACTIONS(1804),
    [anon_sym_null] = ACTIONS(1806),
    [anon_sym_this] = ACTIONS(1806),
    [anon_sym_do] = ACTIONS(1806),
    [anon_sym_while] = ACTIONS(1806),
    [anon_sym_for] = ACTIONS(1806),
    [aux_sym__type_identifier_token1] = ACTIONS(1806),
    [anon_sym_if] = ACTIONS(1806),
    [anon_sym_else] = ACTIONS(1806),
    [anon_sym_LBRACK] = ACTIONS(1804),
    [aux_sym_number_literal_token1] = ACTIONS(1806),
    [aux_sym_number_literal_token2] = ACTIONS(1804),
    [aux_sym_number_literal_token3] = ACTIONS(1804),
    [aux_sym_number_literal_token4] = ACTIONS(1806),
    [aux_sym_number_literal_token5] = ACTIONS(1804),
    [anon_sym_pipeline] = ACTIONS(1806),
    [anon_sym_return] = ACTIONS(1806),
    [anon_sym_SQUOTE] = ACTIONS(1806),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1804),
    [anon_sym_DQUOTE] = ACTIONS(1806),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1804),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1804),
    [anon_sym_switch] = ACTIONS(1806),
    [anon_sym_try] = ACTIONS(1806),
    [anon_sym_catch] = ACTIONS(1806),
    [anon_sym_finally] = ACTIONS(1806),
    [anon_sym_int] = ACTIONS(1806),
    [anon_sym_boolean] = ACTIONS(1806),
    [anon_sym_char] = ACTIONS(1806),
    [anon_sym_short] = ACTIONS(1806),
    [anon_sym_long] = ACTIONS(1806),
    [anon_sym_float] = ACTIONS(1806),
    [anon_sym_double] = ACTIONS(1806),
    [anon_sym_void] = ACTIONS(1806),
    [anon_sym_public] = ACTIONS(1806),
    [anon_sym_protected] = ACTIONS(1806),
    [anon_sym_private] = ACTIONS(1806),
    [anon_sym_static] = ACTIONS(1806),
    [anon_sym_final] = ACTIONS(1806),
    [anon_sym_synchronized] = ACTIONS(1806),
    [anon_sym_TILDE] = ACTIONS(1804),
    [anon_sym_BANG] = ACTIONS(1804),
    [anon_sym_new] = ACTIONS(1806),
  },
  [STATE(690)] = {
    [sym_comment] = STATE(690),
    [sym_groovy_doc] = STATE(690),
    [sym_parameter_list] = STATE(765),
    [ts_builtin_sym_end] = ACTIONS(1763),
    [sym_identifier] = ACTIONS(1761),
    [anon_sym_break] = ACTIONS(1761),
    [anon_sym_continue] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_STAR] = ACTIONS(1761),
    [anon_sym_STAR_COLON] = ACTIONS(1763),
    [anon_sym_import] = ACTIONS(1761),
    [anon_sym_package] = ACTIONS(1761),
    [anon_sym_AT] = ACTIONS(1761),
    [anon_sym_assert] = ACTIONS(1761),
    [anon_sym_EQ] = ACTIONS(1808),
    [anon_sym_PLUS_PLUS] = ACTIONS(1763),
    [anon_sym_DASH_DASH] = ACTIONS(1763),
    [anon_sym_SLASH] = ACTIONS(1761),
    [anon_sym_PLUS] = ACTIONS(1761),
    [anon_sym_DASH] = ACTIONS(1761),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_ATinterface] = ACTIONS(1763),
    [anon_sym_interface] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1761),
    [anon_sym_def] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1761),
    [anon_sym_this] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_while] = ACTIONS(1761),
    [anon_sym_for] = ACTIONS(1761),
    [aux_sym__type_identifier_token1] = ACTIONS(1761),
    [anon_sym_if] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1763),
    [aux_sym_number_literal_token1] = ACTIONS(1761),
    [aux_sym_number_literal_token2] = ACTIONS(1763),
    [aux_sym_number_literal_token3] = ACTIONS(1763),
    [aux_sym_number_literal_token4] = ACTIONS(1761),
    [aux_sym_number_literal_token5] = ACTIONS(1763),
    [anon_sym_pipeline] = ACTIONS(1761),
    [anon_sym_return] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1761),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1761),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1763),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1763),
    [anon_sym_switch] = ACTIONS(1761),
    [anon_sym_try] = ACTIONS(1761),
    [anon_sym_catch] = ACTIONS(1761),
    [anon_sym_finally] = ACTIONS(1761),
    [anon_sym_int] = ACTIONS(1761),
    [anon_sym_boolean] = ACTIONS(1761),
    [anon_sym_char] = ACTIONS(1761),
    [anon_sym_short] = ACTIONS(1761),
    [anon_sym_long] = ACTIONS(1761),
    [anon_sym_float] = ACTIONS(1761),
    [anon_sym_double] = ACTIONS(1761),
    [anon_sym_void] = ACTIONS(1761),
    [anon_sym_public] = ACTIONS(1761),
    [anon_sym_protected] = ACTIONS(1761),
    [anon_sym_private] = ACTIONS(1761),
    [anon_sym_static] = ACTIONS(1761),
    [anon_sym_final] = ACTIONS(1761),
    [anon_sym_synchronized] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1763),
    [anon_sym_BANG] = ACTIONS(1763),
    [anon_sym_new] = ACTIONS(1761),
  },
  [STATE(691)] = {
    [sym_comment] = STATE(691),
    [sym_groovy_doc] = STATE(691),
    [sym_identifier] = ACTIONS(1810),
    [anon_sym_break] = ACTIONS(1810),
    [anon_sym_continue] = ACTIONS(1810),
    [anon_sym_SEMI] = ACTIONS(1812),
    [anon_sym_STAR] = ACTIONS(1810),
    [anon_sym_STAR_COLON] = ACTIONS(1812),
    [anon_sym_DOT] = ACTIONS(1814),
    [anon_sym_import] = ACTIONS(1810),
    [anon_sym_as] = ACTIONS(1816),
    [anon_sym_package] = ACTIONS(1810),
    [anon_sym_AT] = ACTIONS(1810),
    [anon_sym_assert] = ACTIONS(1810),
    [anon_sym_PLUS_PLUS] = ACTIONS(1812),
    [anon_sym_DASH_DASH] = ACTIONS(1812),
    [anon_sym_SLASH] = ACTIONS(1810),
    [anon_sym_PLUS] = ACTIONS(1810),
    [anon_sym_DASH] = ACTIONS(1810),
    [anon_sym_true] = ACTIONS(1810),
    [anon_sym_false] = ACTIONS(1810),
    [anon_sym_ATinterface] = ACTIONS(1812),
    [anon_sym_interface] = ACTIONS(1810),
    [anon_sym_class] = ACTIONS(1810),
    [anon_sym_LBRACE] = ACTIONS(1812),
    [anon_sym_RBRACE] = ACTIONS(1812),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1810),
    [anon_sym_def] = ACTIONS(1810),
    [anon_sym_LPAREN] = ACTIONS(1812),
    [anon_sym_null] = ACTIONS(1810),
    [anon_sym_this] = ACTIONS(1810),
    [anon_sym_do] = ACTIONS(1810),
    [anon_sym_while] = ACTIONS(1810),
    [anon_sym_for] = ACTIONS(1810),
    [aux_sym__type_identifier_token1] = ACTIONS(1810),
    [anon_sym_if] = ACTIONS(1810),
    [anon_sym_else] = ACTIONS(1810),
    [anon_sym_LBRACK] = ACTIONS(1812),
    [aux_sym_number_literal_token1] = ACTIONS(1810),
    [aux_sym_number_literal_token2] = ACTIONS(1812),
    [aux_sym_number_literal_token3] = ACTIONS(1812),
    [aux_sym_number_literal_token4] = ACTIONS(1810),
    [aux_sym_number_literal_token5] = ACTIONS(1812),
    [anon_sym_return] = ACTIONS(1810),
    [anon_sym_SQUOTE] = ACTIONS(1810),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1812),
    [anon_sym_DQUOTE] = ACTIONS(1810),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1812),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1812),
    [anon_sym_switch] = ACTIONS(1810),
    [anon_sym_case] = ACTIONS(1810),
    [anon_sym_default] = ACTIONS(1810),
    [anon_sym_try] = ACTIONS(1810),
    [anon_sym_catch] = ACTIONS(1810),
    [anon_sym_finally] = ACTIONS(1810),
    [anon_sym_int] = ACTIONS(1810),
    [anon_sym_boolean] = ACTIONS(1810),
    [anon_sym_char] = ACTIONS(1810),
    [anon_sym_short] = ACTIONS(1810),
    [anon_sym_long] = ACTIONS(1810),
    [anon_sym_float] = ACTIONS(1810),
    [anon_sym_double] = ACTIONS(1810),
    [anon_sym_void] = ACTIONS(1810),
    [anon_sym_public] = ACTIONS(1810),
    [anon_sym_protected] = ACTIONS(1810),
    [anon_sym_private] = ACTIONS(1810),
    [anon_sym_static] = ACTIONS(1810),
    [anon_sym_final] = ACTIONS(1810),
    [anon_sym_synchronized] = ACTIONS(1810),
    [anon_sym_TILDE] = ACTIONS(1812),
    [anon_sym_BANG] = ACTIONS(1812),
    [anon_sym_new] = ACTIONS(1810),
  },
  [STATE(692)] = {
    [sym_comment] = STATE(692),
    [sym_groovy_doc] = STATE(692),
    [ts_builtin_sym_end] = ACTIONS(1818),
    [sym_identifier] = ACTIONS(1820),
    [anon_sym_break] = ACTIONS(1820),
    [anon_sym_continue] = ACTIONS(1820),
    [anon_sym_SEMI] = ACTIONS(1818),
    [anon_sym_STAR] = ACTIONS(1820),
    [anon_sym_STAR_COLON] = ACTIONS(1818),
    [anon_sym_import] = ACTIONS(1820),
    [anon_sym_package] = ACTIONS(1820),
    [anon_sym_AT] = ACTIONS(1820),
    [anon_sym_assert] = ACTIONS(1820),
    [anon_sym_PLUS_PLUS] = ACTIONS(1818),
    [anon_sym_DASH_DASH] = ACTIONS(1818),
    [anon_sym_SLASH] = ACTIONS(1820),
    [anon_sym_PLUS] = ACTIONS(1820),
    [anon_sym_DASH] = ACTIONS(1820),
    [anon_sym_true] = ACTIONS(1820),
    [anon_sym_false] = ACTIONS(1820),
    [anon_sym_ATinterface] = ACTIONS(1818),
    [anon_sym_interface] = ACTIONS(1820),
    [anon_sym_class] = ACTIONS(1820),
    [anon_sym_COMMA] = ACTIONS(1818),
    [anon_sym_LBRACE] = ACTIONS(1818),
    [anon_sym_RBRACE] = ACTIONS(1818),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1820),
    [anon_sym_def] = ACTIONS(1820),
    [anon_sym_LPAREN] = ACTIONS(1818),
    [anon_sym_RPAREN] = ACTIONS(1818),
    [anon_sym_null] = ACTIONS(1820),
    [anon_sym_this] = ACTIONS(1820),
    [anon_sym_do] = ACTIONS(1820),
    [anon_sym_while] = ACTIONS(1820),
    [anon_sym_for] = ACTIONS(1820),
    [aux_sym__type_identifier_token1] = ACTIONS(1820),
    [anon_sym_if] = ACTIONS(1820),
    [anon_sym_else] = ACTIONS(1820),
    [anon_sym_LBRACK] = ACTIONS(1818),
    [aux_sym_number_literal_token1] = ACTIONS(1820),
    [aux_sym_number_literal_token2] = ACTIONS(1818),
    [aux_sym_number_literal_token3] = ACTIONS(1818),
    [aux_sym_number_literal_token4] = ACTIONS(1820),
    [aux_sym_number_literal_token5] = ACTIONS(1818),
    [anon_sym_pipeline] = ACTIONS(1820),
    [anon_sym_return] = ACTIONS(1820),
    [anon_sym_SQUOTE] = ACTIONS(1820),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1818),
    [anon_sym_DQUOTE] = ACTIONS(1820),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1818),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1818),
    [anon_sym_switch] = ACTIONS(1820),
    [anon_sym_try] = ACTIONS(1820),
    [anon_sym_catch] = ACTIONS(1820),
    [anon_sym_finally] = ACTIONS(1820),
    [anon_sym_int] = ACTIONS(1820),
    [anon_sym_boolean] = ACTIONS(1820),
    [anon_sym_char] = ACTIONS(1820),
    [anon_sym_short] = ACTIONS(1820),
    [anon_sym_long] = ACTIONS(1820),
    [anon_sym_float] = ACTIONS(1820),
    [anon_sym_double] = ACTIONS(1820),
    [anon_sym_void] = ACTIONS(1820),
    [anon_sym_public] = ACTIONS(1820),
    [anon_sym_protected] = ACTIONS(1820),
    [anon_sym_private] = ACTIONS(1820),
    [anon_sym_static] = ACTIONS(1820),
    [anon_sym_final] = ACTIONS(1820),
    [anon_sym_synchronized] = ACTIONS(1820),
    [anon_sym_TILDE] = ACTIONS(1818),
    [anon_sym_BANG] = ACTIONS(1818),
    [anon_sym_new] = ACTIONS(1820),
  },
  [STATE(693)] = {
    [sym_comment] = STATE(693),
    [sym_groovy_doc] = STATE(693),
    [ts_builtin_sym_end] = ACTIONS(1822),
    [sym_identifier] = ACTIONS(1824),
    [anon_sym_break] = ACTIONS(1824),
    [anon_sym_continue] = ACTIONS(1824),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_STAR] = ACTIONS(1824),
    [anon_sym_STAR_COLON] = ACTIONS(1822),
    [anon_sym_import] = ACTIONS(1824),
    [anon_sym_package] = ACTIONS(1824),
    [anon_sym_AT] = ACTIONS(1824),
    [anon_sym_assert] = ACTIONS(1824),
    [anon_sym_PLUS_PLUS] = ACTIONS(1822),
    [anon_sym_DASH_DASH] = ACTIONS(1822),
    [anon_sym_SLASH] = ACTIONS(1824),
    [anon_sym_PLUS] = ACTIONS(1824),
    [anon_sym_DASH] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1824),
    [anon_sym_false] = ACTIONS(1824),
    [anon_sym_ATinterface] = ACTIONS(1822),
    [anon_sym_interface] = ACTIONS(1824),
    [anon_sym_class] = ACTIONS(1824),
    [anon_sym_COMMA] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_RBRACE] = ACTIONS(1822),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1824),
    [anon_sym_def] = ACTIONS(1824),
    [anon_sym_LPAREN] = ACTIONS(1822),
    [anon_sym_RPAREN] = ACTIONS(1822),
    [anon_sym_null] = ACTIONS(1824),
    [anon_sym_this] = ACTIONS(1824),
    [anon_sym_do] = ACTIONS(1824),
    [anon_sym_while] = ACTIONS(1824),
    [anon_sym_for] = ACTIONS(1824),
    [aux_sym__type_identifier_token1] = ACTIONS(1824),
    [anon_sym_if] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [aux_sym_number_literal_token1] = ACTIONS(1824),
    [aux_sym_number_literal_token2] = ACTIONS(1822),
    [aux_sym_number_literal_token3] = ACTIONS(1822),
    [aux_sym_number_literal_token4] = ACTIONS(1824),
    [aux_sym_number_literal_token5] = ACTIONS(1822),
    [anon_sym_pipeline] = ACTIONS(1824),
    [anon_sym_return] = ACTIONS(1824),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1822),
    [anon_sym_switch] = ACTIONS(1824),
    [anon_sym_try] = ACTIONS(1824),
    [anon_sym_catch] = ACTIONS(1824),
    [anon_sym_finally] = ACTIONS(1824),
    [anon_sym_int] = ACTIONS(1824),
    [anon_sym_boolean] = ACTIONS(1824),
    [anon_sym_char] = ACTIONS(1824),
    [anon_sym_short] = ACTIONS(1824),
    [anon_sym_long] = ACTIONS(1824),
    [anon_sym_float] = ACTIONS(1824),
    [anon_sym_double] = ACTIONS(1824),
    [anon_sym_void] = ACTIONS(1824),
    [anon_sym_public] = ACTIONS(1824),
    [anon_sym_protected] = ACTIONS(1824),
    [anon_sym_private] = ACTIONS(1824),
    [anon_sym_static] = ACTIONS(1824),
    [anon_sym_final] = ACTIONS(1824),
    [anon_sym_synchronized] = ACTIONS(1824),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_BANG] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1824),
  },
  [STATE(694)] = {
    [sym_comment] = STATE(694),
    [sym_groovy_doc] = STATE(694),
    [ts_builtin_sym_end] = ACTIONS(1826),
    [sym_identifier] = ACTIONS(1828),
    [anon_sym_break] = ACTIONS(1828),
    [anon_sym_continue] = ACTIONS(1828),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_STAR] = ACTIONS(1828),
    [anon_sym_STAR_COLON] = ACTIONS(1826),
    [anon_sym_import] = ACTIONS(1828),
    [anon_sym_package] = ACTIONS(1828),
    [anon_sym_AT] = ACTIONS(1828),
    [anon_sym_assert] = ACTIONS(1828),
    [anon_sym_PLUS_PLUS] = ACTIONS(1826),
    [anon_sym_DASH_DASH] = ACTIONS(1826),
    [anon_sym_SLASH] = ACTIONS(1828),
    [anon_sym_PLUS] = ACTIONS(1828),
    [anon_sym_DASH] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1828),
    [anon_sym_false] = ACTIONS(1828),
    [anon_sym_ATinterface] = ACTIONS(1826),
    [anon_sym_interface] = ACTIONS(1828),
    [anon_sym_class] = ACTIONS(1828),
    [anon_sym_COMMA] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_RBRACE] = ACTIONS(1826),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1828),
    [anon_sym_def] = ACTIONS(1828),
    [anon_sym_LPAREN] = ACTIONS(1826),
    [anon_sym_RPAREN] = ACTIONS(1826),
    [anon_sym_null] = ACTIONS(1828),
    [anon_sym_this] = ACTIONS(1828),
    [anon_sym_do] = ACTIONS(1828),
    [anon_sym_while] = ACTIONS(1828),
    [anon_sym_for] = ACTIONS(1828),
    [aux_sym__type_identifier_token1] = ACTIONS(1828),
    [anon_sym_if] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [aux_sym_number_literal_token1] = ACTIONS(1828),
    [aux_sym_number_literal_token2] = ACTIONS(1826),
    [aux_sym_number_literal_token3] = ACTIONS(1826),
    [aux_sym_number_literal_token4] = ACTIONS(1828),
    [aux_sym_number_literal_token5] = ACTIONS(1826),
    [anon_sym_pipeline] = ACTIONS(1828),
    [anon_sym_return] = ACTIONS(1828),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1826),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1826),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1826),
    [anon_sym_switch] = ACTIONS(1828),
    [anon_sym_try] = ACTIONS(1828),
    [anon_sym_catch] = ACTIONS(1828),
    [anon_sym_finally] = ACTIONS(1828),
    [anon_sym_int] = ACTIONS(1828),
    [anon_sym_boolean] = ACTIONS(1828),
    [anon_sym_char] = ACTIONS(1828),
    [anon_sym_short] = ACTIONS(1828),
    [anon_sym_long] = ACTIONS(1828),
    [anon_sym_float] = ACTIONS(1828),
    [anon_sym_double] = ACTIONS(1828),
    [anon_sym_void] = ACTIONS(1828),
    [anon_sym_public] = ACTIONS(1828),
    [anon_sym_protected] = ACTIONS(1828),
    [anon_sym_private] = ACTIONS(1828),
    [anon_sym_static] = ACTIONS(1828),
    [anon_sym_final] = ACTIONS(1828),
    [anon_sym_synchronized] = ACTIONS(1828),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_BANG] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1828),
  },
  [STATE(695)] = {
    [sym_comment] = STATE(695),
    [sym_groovy_doc] = STATE(695),
    [ts_builtin_sym_end] = ACTIONS(1830),
    [sym_identifier] = ACTIONS(1832),
    [anon_sym_break] = ACTIONS(1832),
    [anon_sym_continue] = ACTIONS(1832),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_STAR] = ACTIONS(1832),
    [anon_sym_STAR_COLON] = ACTIONS(1830),
    [anon_sym_import] = ACTIONS(1832),
    [anon_sym_package] = ACTIONS(1832),
    [anon_sym_AT] = ACTIONS(1832),
    [anon_sym_assert] = ACTIONS(1832),
    [anon_sym_PLUS_PLUS] = ACTIONS(1830),
    [anon_sym_DASH_DASH] = ACTIONS(1830),
    [anon_sym_SLASH] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(1832),
    [anon_sym_DASH] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1832),
    [anon_sym_false] = ACTIONS(1832),
    [anon_sym_ATinterface] = ACTIONS(1830),
    [anon_sym_interface] = ACTIONS(1832),
    [anon_sym_class] = ACTIONS(1832),
    [anon_sym_COMMA] = ACTIONS(1830),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_RBRACE] = ACTIONS(1830),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1832),
    [anon_sym_def] = ACTIONS(1832),
    [anon_sym_LPAREN] = ACTIONS(1830),
    [anon_sym_RPAREN] = ACTIONS(1830),
    [anon_sym_null] = ACTIONS(1832),
    [anon_sym_this] = ACTIONS(1832),
    [anon_sym_do] = ACTIONS(1832),
    [anon_sym_while] = ACTIONS(1832),
    [anon_sym_for] = ACTIONS(1832),
    [aux_sym__type_identifier_token1] = ACTIONS(1832),
    [anon_sym_if] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [aux_sym_number_literal_token1] = ACTIONS(1832),
    [aux_sym_number_literal_token2] = ACTIONS(1830),
    [aux_sym_number_literal_token3] = ACTIONS(1830),
    [aux_sym_number_literal_token4] = ACTIONS(1832),
    [aux_sym_number_literal_token5] = ACTIONS(1830),
    [anon_sym_pipeline] = ACTIONS(1832),
    [anon_sym_return] = ACTIONS(1832),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1830),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1830),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1830),
    [anon_sym_switch] = ACTIONS(1832),
    [anon_sym_try] = ACTIONS(1832),
    [anon_sym_catch] = ACTIONS(1832),
    [anon_sym_finally] = ACTIONS(1832),
    [anon_sym_int] = ACTIONS(1832),
    [anon_sym_boolean] = ACTIONS(1832),
    [anon_sym_char] = ACTIONS(1832),
    [anon_sym_short] = ACTIONS(1832),
    [anon_sym_long] = ACTIONS(1832),
    [anon_sym_float] = ACTIONS(1832),
    [anon_sym_double] = ACTIONS(1832),
    [anon_sym_void] = ACTIONS(1832),
    [anon_sym_public] = ACTIONS(1832),
    [anon_sym_protected] = ACTIONS(1832),
    [anon_sym_private] = ACTIONS(1832),
    [anon_sym_static] = ACTIONS(1832),
    [anon_sym_final] = ACTIONS(1832),
    [anon_sym_synchronized] = ACTIONS(1832),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_BANG] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1832),
  },
  [STATE(696)] = {
    [sym_comment] = STATE(696),
    [sym_groovy_doc] = STATE(696),
    [sym_identifier] = ACTIONS(1759),
    [anon_sym_break] = ACTIONS(1759),
    [anon_sym_continue] = ACTIONS(1759),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_STAR] = ACTIONS(1759),
    [anon_sym_STAR_COLON] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1757),
    [anon_sym_import] = ACTIONS(1759),
    [anon_sym_as] = ACTIONS(1759),
    [anon_sym_package] = ACTIONS(1759),
    [anon_sym_AT] = ACTIONS(1759),
    [anon_sym_assert] = ACTIONS(1759),
    [anon_sym_PLUS_PLUS] = ACTIONS(1757),
    [anon_sym_DASH_DASH] = ACTIONS(1757),
    [anon_sym_SLASH] = ACTIONS(1759),
    [anon_sym_PLUS] = ACTIONS(1759),
    [anon_sym_DASH] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1759),
    [anon_sym_false] = ACTIONS(1759),
    [anon_sym_ATinterface] = ACTIONS(1757),
    [anon_sym_interface] = ACTIONS(1759),
    [anon_sym_class] = ACTIONS(1759),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_RBRACE] = ACTIONS(1757),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1759),
    [anon_sym_def] = ACTIONS(1759),
    [anon_sym_LPAREN] = ACTIONS(1757),
    [anon_sym_null] = ACTIONS(1759),
    [anon_sym_this] = ACTIONS(1759),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_while] = ACTIONS(1759),
    [anon_sym_for] = ACTIONS(1759),
    [aux_sym__type_identifier_token1] = ACTIONS(1759),
    [anon_sym_if] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [aux_sym_number_literal_token1] = ACTIONS(1759),
    [aux_sym_number_literal_token2] = ACTIONS(1757),
    [aux_sym_number_literal_token3] = ACTIONS(1757),
    [aux_sym_number_literal_token4] = ACTIONS(1759),
    [aux_sym_number_literal_token5] = ACTIONS(1757),
    [anon_sym_return] = ACTIONS(1759),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1757),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1757),
    [anon_sym_switch] = ACTIONS(1759),
    [anon_sym_case] = ACTIONS(1759),
    [anon_sym_default] = ACTIONS(1759),
    [anon_sym_try] = ACTIONS(1759),
    [anon_sym_catch] = ACTIONS(1759),
    [anon_sym_finally] = ACTIONS(1759),
    [anon_sym_int] = ACTIONS(1759),
    [anon_sym_boolean] = ACTIONS(1759),
    [anon_sym_char] = ACTIONS(1759),
    [anon_sym_short] = ACTIONS(1759),
    [anon_sym_long] = ACTIONS(1759),
    [anon_sym_float] = ACTIONS(1759),
    [anon_sym_double] = ACTIONS(1759),
    [anon_sym_void] = ACTIONS(1759),
    [anon_sym_public] = ACTIONS(1759),
    [anon_sym_protected] = ACTIONS(1759),
    [anon_sym_private] = ACTIONS(1759),
    [anon_sym_static] = ACTIONS(1759),
    [anon_sym_final] = ACTIONS(1759),
    [anon_sym_synchronized] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_BANG] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1759),
  },
  [STATE(697)] = {
    [sym_comment] = STATE(697),
    [sym_groovy_doc] = STATE(697),
    [sym_parameter_list] = STATE(765),
    [ts_builtin_sym_end] = ACTIONS(1753),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_break] = ACTIONS(1751),
    [anon_sym_continue] = ACTIONS(1751),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_STAR] = ACTIONS(1751),
    [anon_sym_STAR_COLON] = ACTIONS(1753),
    [anon_sym_import] = ACTIONS(1751),
    [anon_sym_package] = ACTIONS(1751),
    [anon_sym_AT] = ACTIONS(1751),
    [anon_sym_assert] = ACTIONS(1751),
    [anon_sym_EQ] = ACTIONS(1834),
    [anon_sym_PLUS_PLUS] = ACTIONS(1753),
    [anon_sym_DASH_DASH] = ACTIONS(1753),
    [anon_sym_SLASH] = ACTIONS(1751),
    [anon_sym_PLUS] = ACTIONS(1751),
    [anon_sym_DASH] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1751),
    [anon_sym_false] = ACTIONS(1751),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_interface] = ACTIONS(1751),
    [anon_sym_class] = ACTIONS(1751),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_RBRACE] = ACTIONS(1753),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1751),
    [anon_sym_def] = ACTIONS(1751),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1751),
    [anon_sym_this] = ACTIONS(1751),
    [anon_sym_do] = ACTIONS(1751),
    [anon_sym_while] = ACTIONS(1751),
    [anon_sym_for] = ACTIONS(1751),
    [aux_sym__type_identifier_token1] = ACTIONS(1751),
    [anon_sym_if] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [aux_sym_number_literal_token1] = ACTIONS(1751),
    [aux_sym_number_literal_token2] = ACTIONS(1753),
    [aux_sym_number_literal_token3] = ACTIONS(1753),
    [aux_sym_number_literal_token4] = ACTIONS(1751),
    [aux_sym_number_literal_token5] = ACTIONS(1753),
    [anon_sym_pipeline] = ACTIONS(1751),
    [anon_sym_return] = ACTIONS(1751),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1753),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1753),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1753),
    [anon_sym_switch] = ACTIONS(1751),
    [anon_sym_try] = ACTIONS(1751),
    [anon_sym_catch] = ACTIONS(1751),
    [anon_sym_finally] = ACTIONS(1751),
    [anon_sym_int] = ACTIONS(1751),
    [anon_sym_boolean] = ACTIONS(1751),
    [anon_sym_char] = ACTIONS(1751),
    [anon_sym_short] = ACTIONS(1751),
    [anon_sym_long] = ACTIONS(1751),
    [anon_sym_float] = ACTIONS(1751),
    [anon_sym_double] = ACTIONS(1751),
    [anon_sym_void] = ACTIONS(1751),
    [anon_sym_public] = ACTIONS(1751),
    [anon_sym_protected] = ACTIONS(1751),
    [anon_sym_private] = ACTIONS(1751),
    [anon_sym_static] = ACTIONS(1751),
    [anon_sym_final] = ACTIONS(1751),
    [anon_sym_synchronized] = ACTIONS(1751),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_BANG] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1751),
  },
  [STATE(698)] = {
    [sym_comment] = STATE(698),
    [sym_groovy_doc] = STATE(698),
    [ts_builtin_sym_end] = ACTIONS(1836),
    [sym_identifier] = ACTIONS(1838),
    [anon_sym_break] = ACTIONS(1838),
    [anon_sym_continue] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_STAR] = ACTIONS(1838),
    [anon_sym_STAR_COLON] = ACTIONS(1836),
    [anon_sym_import] = ACTIONS(1838),
    [anon_sym_package] = ACTIONS(1838),
    [anon_sym_AT] = ACTIONS(1838),
    [anon_sym_assert] = ACTIONS(1838),
    [anon_sym_PLUS_PLUS] = ACTIONS(1836),
    [anon_sym_DASH_DASH] = ACTIONS(1836),
    [anon_sym_SLASH] = ACTIONS(1838),
    [anon_sym_PLUS] = ACTIONS(1838),
    [anon_sym_DASH] = ACTIONS(1838),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_ATinterface] = ACTIONS(1836),
    [anon_sym_interface] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_LBRACE] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1838),
    [anon_sym_def] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_RPAREN] = ACTIONS(1836),
    [anon_sym_null] = ACTIONS(1838),
    [anon_sym_this] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_while] = ACTIONS(1838),
    [anon_sym_for] = ACTIONS(1838),
    [aux_sym__type_identifier_token1] = ACTIONS(1838),
    [anon_sym_if] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1836),
    [aux_sym_number_literal_token1] = ACTIONS(1838),
    [aux_sym_number_literal_token2] = ACTIONS(1836),
    [aux_sym_number_literal_token3] = ACTIONS(1836),
    [aux_sym_number_literal_token4] = ACTIONS(1838),
    [aux_sym_number_literal_token5] = ACTIONS(1836),
    [anon_sym_pipeline] = ACTIONS(1838),
    [anon_sym_return] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1838),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1836),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1836),
    [anon_sym_switch] = ACTIONS(1838),
    [anon_sym_try] = ACTIONS(1838),
    [anon_sym_catch] = ACTIONS(1838),
    [anon_sym_finally] = ACTIONS(1838),
    [anon_sym_int] = ACTIONS(1838),
    [anon_sym_boolean] = ACTIONS(1838),
    [anon_sym_char] = ACTIONS(1838),
    [anon_sym_short] = ACTIONS(1838),
    [anon_sym_long] = ACTIONS(1838),
    [anon_sym_float] = ACTIONS(1838),
    [anon_sym_double] = ACTIONS(1838),
    [anon_sym_void] = ACTIONS(1838),
    [anon_sym_public] = ACTIONS(1838),
    [anon_sym_protected] = ACTIONS(1838),
    [anon_sym_private] = ACTIONS(1838),
    [anon_sym_static] = ACTIONS(1838),
    [anon_sym_final] = ACTIONS(1838),
    [anon_sym_synchronized] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1836),
    [anon_sym_BANG] = ACTIONS(1836),
    [anon_sym_new] = ACTIONS(1838),
  },
  [STATE(699)] = {
    [sym_comment] = STATE(699),
    [sym_groovy_doc] = STATE(699),
    [ts_builtin_sym_end] = ACTIONS(1840),
    [sym_identifier] = ACTIONS(1842),
    [anon_sym_break] = ACTIONS(1842),
    [anon_sym_continue] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_STAR] = ACTIONS(1842),
    [anon_sym_STAR_COLON] = ACTIONS(1840),
    [anon_sym_import] = ACTIONS(1842),
    [anon_sym_package] = ACTIONS(1842),
    [anon_sym_AT] = ACTIONS(1842),
    [anon_sym_assert] = ACTIONS(1842),
    [anon_sym_PLUS_PLUS] = ACTIONS(1840),
    [anon_sym_DASH_DASH] = ACTIONS(1840),
    [anon_sym_SLASH] = ACTIONS(1842),
    [anon_sym_PLUS] = ACTIONS(1842),
    [anon_sym_DASH] = ACTIONS(1842),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_ATinterface] = ACTIONS(1840),
    [anon_sym_interface] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_LBRACE] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1842),
    [anon_sym_def] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_RPAREN] = ACTIONS(1840),
    [anon_sym_null] = ACTIONS(1842),
    [anon_sym_this] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_while] = ACTIONS(1842),
    [anon_sym_for] = ACTIONS(1842),
    [aux_sym__type_identifier_token1] = ACTIONS(1842),
    [anon_sym_if] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1840),
    [aux_sym_number_literal_token1] = ACTIONS(1842),
    [aux_sym_number_literal_token2] = ACTIONS(1840),
    [aux_sym_number_literal_token3] = ACTIONS(1840),
    [aux_sym_number_literal_token4] = ACTIONS(1842),
    [aux_sym_number_literal_token5] = ACTIONS(1840),
    [anon_sym_pipeline] = ACTIONS(1842),
    [anon_sym_return] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1842),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1842),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1840),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1840),
    [anon_sym_switch] = ACTIONS(1842),
    [anon_sym_try] = ACTIONS(1842),
    [anon_sym_catch] = ACTIONS(1842),
    [anon_sym_finally] = ACTIONS(1842),
    [anon_sym_int] = ACTIONS(1842),
    [anon_sym_boolean] = ACTIONS(1842),
    [anon_sym_char] = ACTIONS(1842),
    [anon_sym_short] = ACTIONS(1842),
    [anon_sym_long] = ACTIONS(1842),
    [anon_sym_float] = ACTIONS(1842),
    [anon_sym_double] = ACTIONS(1842),
    [anon_sym_void] = ACTIONS(1842),
    [anon_sym_public] = ACTIONS(1842),
    [anon_sym_protected] = ACTIONS(1842),
    [anon_sym_private] = ACTIONS(1842),
    [anon_sym_static] = ACTIONS(1842),
    [anon_sym_final] = ACTIONS(1842),
    [anon_sym_synchronized] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1840),
    [anon_sym_BANG] = ACTIONS(1840),
    [anon_sym_new] = ACTIONS(1842),
  },
  [STATE(700)] = {
    [sym_comment] = STATE(700),
    [sym_groovy_doc] = STATE(700),
    [ts_builtin_sym_end] = ACTIONS(1844),
    [sym_identifier] = ACTIONS(1846),
    [anon_sym_break] = ACTIONS(1846),
    [anon_sym_continue] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1846),
    [anon_sym_STAR_COLON] = ACTIONS(1844),
    [anon_sym_import] = ACTIONS(1846),
    [anon_sym_package] = ACTIONS(1846),
    [anon_sym_AT] = ACTIONS(1846),
    [anon_sym_assert] = ACTIONS(1846),
    [anon_sym_PLUS_PLUS] = ACTIONS(1844),
    [anon_sym_DASH_DASH] = ACTIONS(1844),
    [anon_sym_SLASH] = ACTIONS(1846),
    [anon_sym_PLUS] = ACTIONS(1846),
    [anon_sym_DASH] = ACTIONS(1846),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_ATinterface] = ACTIONS(1844),
    [anon_sym_interface] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1846),
    [anon_sym_def] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_RPAREN] = ACTIONS(1844),
    [anon_sym_null] = ACTIONS(1846),
    [anon_sym_this] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_while] = ACTIONS(1846),
    [anon_sym_for] = ACTIONS(1846),
    [aux_sym__type_identifier_token1] = ACTIONS(1846),
    [anon_sym_if] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1844),
    [aux_sym_number_literal_token1] = ACTIONS(1846),
    [aux_sym_number_literal_token2] = ACTIONS(1844),
    [aux_sym_number_literal_token3] = ACTIONS(1844),
    [aux_sym_number_literal_token4] = ACTIONS(1846),
    [aux_sym_number_literal_token5] = ACTIONS(1844),
    [anon_sym_pipeline] = ACTIONS(1846),
    [anon_sym_return] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1846),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1846),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1844),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1844),
    [anon_sym_switch] = ACTIONS(1846),
    [anon_sym_try] = ACTIONS(1846),
    [anon_sym_catch] = ACTIONS(1846),
    [anon_sym_finally] = ACTIONS(1846),
    [anon_sym_int] = ACTIONS(1846),
    [anon_sym_boolean] = ACTIONS(1846),
    [anon_sym_char] = ACTIONS(1846),
    [anon_sym_short] = ACTIONS(1846),
    [anon_sym_long] = ACTIONS(1846),
    [anon_sym_float] = ACTIONS(1846),
    [anon_sym_double] = ACTIONS(1846),
    [anon_sym_void] = ACTIONS(1846),
    [anon_sym_public] = ACTIONS(1846),
    [anon_sym_protected] = ACTIONS(1846),
    [anon_sym_private] = ACTIONS(1846),
    [anon_sym_static] = ACTIONS(1846),
    [anon_sym_final] = ACTIONS(1846),
    [anon_sym_synchronized] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1844),
    [anon_sym_new] = ACTIONS(1846),
  },
  [STATE(701)] = {
    [sym_comment] = STATE(701),
    [sym_groovy_doc] = STATE(701),
    [ts_builtin_sym_end] = ACTIONS(1848),
    [sym_identifier] = ACTIONS(1850),
    [anon_sym_break] = ACTIONS(1850),
    [anon_sym_continue] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_STAR] = ACTIONS(1850),
    [anon_sym_STAR_COLON] = ACTIONS(1848),
    [anon_sym_import] = ACTIONS(1850),
    [anon_sym_package] = ACTIONS(1850),
    [anon_sym_AT] = ACTIONS(1850),
    [anon_sym_assert] = ACTIONS(1850),
    [anon_sym_PLUS_PLUS] = ACTIONS(1848),
    [anon_sym_DASH_DASH] = ACTIONS(1848),
    [anon_sym_SLASH] = ACTIONS(1850),
    [anon_sym_PLUS] = ACTIONS(1850),
    [anon_sym_DASH] = ACTIONS(1850),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_ATinterface] = ACTIONS(1848),
    [anon_sym_interface] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_LBRACE] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1850),
    [anon_sym_def] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_RPAREN] = ACTIONS(1848),
    [anon_sym_null] = ACTIONS(1850),
    [anon_sym_this] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_while] = ACTIONS(1850),
    [anon_sym_for] = ACTIONS(1850),
    [aux_sym__type_identifier_token1] = ACTIONS(1850),
    [anon_sym_if] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1848),
    [aux_sym_number_literal_token1] = ACTIONS(1850),
    [aux_sym_number_literal_token2] = ACTIONS(1848),
    [aux_sym_number_literal_token3] = ACTIONS(1848),
    [aux_sym_number_literal_token4] = ACTIONS(1850),
    [aux_sym_number_literal_token5] = ACTIONS(1848),
    [anon_sym_pipeline] = ACTIONS(1850),
    [anon_sym_return] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1850),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1850),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1848),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1848),
    [anon_sym_switch] = ACTIONS(1850),
    [anon_sym_try] = ACTIONS(1850),
    [anon_sym_catch] = ACTIONS(1850),
    [anon_sym_finally] = ACTIONS(1850),
    [anon_sym_int] = ACTIONS(1850),
    [anon_sym_boolean] = ACTIONS(1850),
    [anon_sym_char] = ACTIONS(1850),
    [anon_sym_short] = ACTIONS(1850),
    [anon_sym_long] = ACTIONS(1850),
    [anon_sym_float] = ACTIONS(1850),
    [anon_sym_double] = ACTIONS(1850),
    [anon_sym_void] = ACTIONS(1850),
    [anon_sym_public] = ACTIONS(1850),
    [anon_sym_protected] = ACTIONS(1850),
    [anon_sym_private] = ACTIONS(1850),
    [anon_sym_static] = ACTIONS(1850),
    [anon_sym_final] = ACTIONS(1850),
    [anon_sym_synchronized] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1848),
    [anon_sym_BANG] = ACTIONS(1848),
    [anon_sym_new] = ACTIONS(1850),
  },
  [STATE(702)] = {
    [sym_comment] = STATE(702),
    [sym_groovy_doc] = STATE(702),
    [ts_builtin_sym_end] = ACTIONS(1852),
    [sym_identifier] = ACTIONS(1854),
    [anon_sym_break] = ACTIONS(1854),
    [anon_sym_continue] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_STAR] = ACTIONS(1854),
    [anon_sym_STAR_COLON] = ACTIONS(1852),
    [anon_sym_import] = ACTIONS(1854),
    [anon_sym_package] = ACTIONS(1854),
    [anon_sym_AT] = ACTIONS(1854),
    [anon_sym_assert] = ACTIONS(1854),
    [anon_sym_PLUS_PLUS] = ACTIONS(1852),
    [anon_sym_DASH_DASH] = ACTIONS(1852),
    [anon_sym_SLASH] = ACTIONS(1854),
    [anon_sym_PLUS] = ACTIONS(1854),
    [anon_sym_DASH] = ACTIONS(1854),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_ATinterface] = ACTIONS(1852),
    [anon_sym_interface] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_LBRACE] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1854),
    [anon_sym_def] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_RPAREN] = ACTIONS(1852),
    [anon_sym_null] = ACTIONS(1854),
    [anon_sym_this] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_while] = ACTIONS(1854),
    [anon_sym_for] = ACTIONS(1854),
    [aux_sym__type_identifier_token1] = ACTIONS(1854),
    [anon_sym_if] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1852),
    [aux_sym_number_literal_token1] = ACTIONS(1854),
    [aux_sym_number_literal_token2] = ACTIONS(1852),
    [aux_sym_number_literal_token3] = ACTIONS(1852),
    [aux_sym_number_literal_token4] = ACTIONS(1854),
    [aux_sym_number_literal_token5] = ACTIONS(1852),
    [anon_sym_pipeline] = ACTIONS(1854),
    [anon_sym_return] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1854),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1854),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1852),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1852),
    [anon_sym_switch] = ACTIONS(1854),
    [anon_sym_try] = ACTIONS(1854),
    [anon_sym_catch] = ACTIONS(1854),
    [anon_sym_finally] = ACTIONS(1854),
    [anon_sym_int] = ACTIONS(1854),
    [anon_sym_boolean] = ACTIONS(1854),
    [anon_sym_char] = ACTIONS(1854),
    [anon_sym_short] = ACTIONS(1854),
    [anon_sym_long] = ACTIONS(1854),
    [anon_sym_float] = ACTIONS(1854),
    [anon_sym_double] = ACTIONS(1854),
    [anon_sym_void] = ACTIONS(1854),
    [anon_sym_public] = ACTIONS(1854),
    [anon_sym_protected] = ACTIONS(1854),
    [anon_sym_private] = ACTIONS(1854),
    [anon_sym_static] = ACTIONS(1854),
    [anon_sym_final] = ACTIONS(1854),
    [anon_sym_synchronized] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1852),
    [anon_sym_BANG] = ACTIONS(1852),
    [anon_sym_new] = ACTIONS(1854),
  },
  [STATE(703)] = {
    [sym_comment] = STATE(703),
    [sym_groovy_doc] = STATE(703),
    [ts_builtin_sym_end] = ACTIONS(1856),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_break] = ACTIONS(1858),
    [anon_sym_continue] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_STAR] = ACTIONS(1858),
    [anon_sym_STAR_COLON] = ACTIONS(1856),
    [anon_sym_import] = ACTIONS(1858),
    [anon_sym_package] = ACTIONS(1858),
    [anon_sym_AT] = ACTIONS(1858),
    [anon_sym_assert] = ACTIONS(1858),
    [anon_sym_PLUS_PLUS] = ACTIONS(1856),
    [anon_sym_DASH_DASH] = ACTIONS(1856),
    [anon_sym_SLASH] = ACTIONS(1858),
    [anon_sym_PLUS] = ACTIONS(1858),
    [anon_sym_DASH] = ACTIONS(1858),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_ATinterface] = ACTIONS(1856),
    [anon_sym_interface] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_LBRACE] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1858),
    [anon_sym_def] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_RPAREN] = ACTIONS(1856),
    [anon_sym_null] = ACTIONS(1858),
    [anon_sym_this] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_while] = ACTIONS(1858),
    [anon_sym_for] = ACTIONS(1858),
    [aux_sym__type_identifier_token1] = ACTIONS(1858),
    [anon_sym_if] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1856),
    [aux_sym_number_literal_token1] = ACTIONS(1858),
    [aux_sym_number_literal_token2] = ACTIONS(1856),
    [aux_sym_number_literal_token3] = ACTIONS(1856),
    [aux_sym_number_literal_token4] = ACTIONS(1858),
    [aux_sym_number_literal_token5] = ACTIONS(1856),
    [anon_sym_pipeline] = ACTIONS(1858),
    [anon_sym_return] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1858),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1858),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1856),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1856),
    [anon_sym_switch] = ACTIONS(1858),
    [anon_sym_try] = ACTIONS(1858),
    [anon_sym_catch] = ACTIONS(1858),
    [anon_sym_finally] = ACTIONS(1858),
    [anon_sym_int] = ACTIONS(1858),
    [anon_sym_boolean] = ACTIONS(1858),
    [anon_sym_char] = ACTIONS(1858),
    [anon_sym_short] = ACTIONS(1858),
    [anon_sym_long] = ACTIONS(1858),
    [anon_sym_float] = ACTIONS(1858),
    [anon_sym_double] = ACTIONS(1858),
    [anon_sym_void] = ACTIONS(1858),
    [anon_sym_public] = ACTIONS(1858),
    [anon_sym_protected] = ACTIONS(1858),
    [anon_sym_private] = ACTIONS(1858),
    [anon_sym_static] = ACTIONS(1858),
    [anon_sym_final] = ACTIONS(1858),
    [anon_sym_synchronized] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1856),
    [anon_sym_BANG] = ACTIONS(1856),
    [anon_sym_new] = ACTIONS(1858),
  },
  [STATE(704)] = {
    [sym_comment] = STATE(704),
    [sym_groovy_doc] = STATE(704),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_COMMA] = ACTIONS(1519),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_RPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(705)] = {
    [sym_comment] = STATE(705),
    [sym_groovy_doc] = STATE(705),
    [ts_builtin_sym_end] = ACTIONS(1860),
    [sym_identifier] = ACTIONS(1862),
    [anon_sym_break] = ACTIONS(1862),
    [anon_sym_continue] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_STAR] = ACTIONS(1862),
    [anon_sym_STAR_COLON] = ACTIONS(1860),
    [anon_sym_import] = ACTIONS(1862),
    [anon_sym_package] = ACTIONS(1862),
    [anon_sym_AT] = ACTIONS(1862),
    [anon_sym_assert] = ACTIONS(1862),
    [anon_sym_PLUS_PLUS] = ACTIONS(1860),
    [anon_sym_DASH_DASH] = ACTIONS(1860),
    [anon_sym_SLASH] = ACTIONS(1862),
    [anon_sym_PLUS] = ACTIONS(1862),
    [anon_sym_DASH] = ACTIONS(1862),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_ATinterface] = ACTIONS(1860),
    [anon_sym_interface] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_LBRACE] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1862),
    [anon_sym_def] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_RPAREN] = ACTIONS(1860),
    [anon_sym_null] = ACTIONS(1862),
    [anon_sym_this] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_while] = ACTIONS(1862),
    [anon_sym_for] = ACTIONS(1862),
    [aux_sym__type_identifier_token1] = ACTIONS(1862),
    [anon_sym_if] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1860),
    [aux_sym_number_literal_token1] = ACTIONS(1862),
    [aux_sym_number_literal_token2] = ACTIONS(1860),
    [aux_sym_number_literal_token3] = ACTIONS(1860),
    [aux_sym_number_literal_token4] = ACTIONS(1862),
    [aux_sym_number_literal_token5] = ACTIONS(1860),
    [anon_sym_pipeline] = ACTIONS(1862),
    [anon_sym_return] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1862),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1862),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1860),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1860),
    [anon_sym_switch] = ACTIONS(1862),
    [anon_sym_try] = ACTIONS(1862),
    [anon_sym_catch] = ACTIONS(1862),
    [anon_sym_finally] = ACTIONS(1862),
    [anon_sym_int] = ACTIONS(1862),
    [anon_sym_boolean] = ACTIONS(1862),
    [anon_sym_char] = ACTIONS(1862),
    [anon_sym_short] = ACTIONS(1862),
    [anon_sym_long] = ACTIONS(1862),
    [anon_sym_float] = ACTIONS(1862),
    [anon_sym_double] = ACTIONS(1862),
    [anon_sym_void] = ACTIONS(1862),
    [anon_sym_public] = ACTIONS(1862),
    [anon_sym_protected] = ACTIONS(1862),
    [anon_sym_private] = ACTIONS(1862),
    [anon_sym_static] = ACTIONS(1862),
    [anon_sym_final] = ACTIONS(1862),
    [anon_sym_synchronized] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1860),
    [anon_sym_BANG] = ACTIONS(1860),
    [anon_sym_new] = ACTIONS(1862),
  },
  [STATE(706)] = {
    [sym_comment] = STATE(706),
    [sym_groovy_doc] = STATE(706),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(389),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(389),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(707)] = {
    [sym_comment] = STATE(707),
    [sym_groovy_doc] = STATE(707),
    [ts_builtin_sym_end] = ACTIONS(1864),
    [sym_identifier] = ACTIONS(1866),
    [anon_sym_break] = ACTIONS(1866),
    [anon_sym_continue] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_STAR] = ACTIONS(1866),
    [anon_sym_STAR_COLON] = ACTIONS(1864),
    [anon_sym_import] = ACTIONS(1866),
    [anon_sym_package] = ACTIONS(1866),
    [anon_sym_AT] = ACTIONS(1866),
    [anon_sym_assert] = ACTIONS(1866),
    [anon_sym_PLUS_PLUS] = ACTIONS(1864),
    [anon_sym_DASH_DASH] = ACTIONS(1864),
    [anon_sym_SLASH] = ACTIONS(1866),
    [anon_sym_PLUS] = ACTIONS(1866),
    [anon_sym_DASH] = ACTIONS(1866),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_ATinterface] = ACTIONS(1864),
    [anon_sym_interface] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_LBRACE] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1866),
    [anon_sym_def] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_RPAREN] = ACTIONS(1864),
    [anon_sym_null] = ACTIONS(1866),
    [anon_sym_this] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_while] = ACTIONS(1866),
    [anon_sym_for] = ACTIONS(1866),
    [aux_sym__type_identifier_token1] = ACTIONS(1866),
    [anon_sym_if] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1864),
    [aux_sym_number_literal_token1] = ACTIONS(1866),
    [aux_sym_number_literal_token2] = ACTIONS(1864),
    [aux_sym_number_literal_token3] = ACTIONS(1864),
    [aux_sym_number_literal_token4] = ACTIONS(1866),
    [aux_sym_number_literal_token5] = ACTIONS(1864),
    [anon_sym_pipeline] = ACTIONS(1866),
    [anon_sym_return] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1866),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1866),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1864),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1864),
    [anon_sym_switch] = ACTIONS(1866),
    [anon_sym_try] = ACTIONS(1866),
    [anon_sym_catch] = ACTIONS(1866),
    [anon_sym_finally] = ACTIONS(1866),
    [anon_sym_int] = ACTIONS(1866),
    [anon_sym_boolean] = ACTIONS(1866),
    [anon_sym_char] = ACTIONS(1866),
    [anon_sym_short] = ACTIONS(1866),
    [anon_sym_long] = ACTIONS(1866),
    [anon_sym_float] = ACTIONS(1866),
    [anon_sym_double] = ACTIONS(1866),
    [anon_sym_void] = ACTIONS(1866),
    [anon_sym_public] = ACTIONS(1866),
    [anon_sym_protected] = ACTIONS(1866),
    [anon_sym_private] = ACTIONS(1866),
    [anon_sym_static] = ACTIONS(1866),
    [anon_sym_final] = ACTIONS(1866),
    [anon_sym_synchronized] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1864),
    [anon_sym_BANG] = ACTIONS(1864),
    [anon_sym_new] = ACTIONS(1866),
  },
  [STATE(708)] = {
    [sym_comment] = STATE(708),
    [sym_groovy_doc] = STATE(708),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_COMMA] = ACTIONS(1515),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_RPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(709)] = {
    [sym_comment] = STATE(709),
    [sym_groovy_doc] = STATE(709),
    [ts_builtin_sym_end] = ACTIONS(1868),
    [sym_identifier] = ACTIONS(1870),
    [anon_sym_break] = ACTIONS(1870),
    [anon_sym_continue] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_STAR] = ACTIONS(1870),
    [anon_sym_STAR_COLON] = ACTIONS(1868),
    [anon_sym_DOT] = ACTIONS(1872),
    [anon_sym_import] = ACTIONS(1870),
    [anon_sym_as] = ACTIONS(1874),
    [anon_sym_package] = ACTIONS(1870),
    [anon_sym_AT] = ACTIONS(1870),
    [anon_sym_assert] = ACTIONS(1870),
    [anon_sym_PLUS_PLUS] = ACTIONS(1868),
    [anon_sym_DASH_DASH] = ACTIONS(1868),
    [anon_sym_SLASH] = ACTIONS(1870),
    [anon_sym_PLUS] = ACTIONS(1870),
    [anon_sym_DASH] = ACTIONS(1870),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_ATinterface] = ACTIONS(1868),
    [anon_sym_interface] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1870),
    [anon_sym_def] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_null] = ACTIONS(1870),
    [anon_sym_this] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_while] = ACTIONS(1870),
    [anon_sym_for] = ACTIONS(1870),
    [aux_sym__type_identifier_token1] = ACTIONS(1870),
    [anon_sym_if] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1868),
    [aux_sym_number_literal_token1] = ACTIONS(1870),
    [aux_sym_number_literal_token2] = ACTIONS(1868),
    [aux_sym_number_literal_token3] = ACTIONS(1868),
    [aux_sym_number_literal_token4] = ACTIONS(1870),
    [aux_sym_number_literal_token5] = ACTIONS(1868),
    [anon_sym_pipeline] = ACTIONS(1870),
    [anon_sym_return] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1870),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1870),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1868),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1868),
    [anon_sym_switch] = ACTIONS(1870),
    [anon_sym_try] = ACTIONS(1870),
    [anon_sym_catch] = ACTIONS(1870),
    [anon_sym_finally] = ACTIONS(1870),
    [anon_sym_int] = ACTIONS(1870),
    [anon_sym_boolean] = ACTIONS(1870),
    [anon_sym_char] = ACTIONS(1870),
    [anon_sym_short] = ACTIONS(1870),
    [anon_sym_long] = ACTIONS(1870),
    [anon_sym_float] = ACTIONS(1870),
    [anon_sym_double] = ACTIONS(1870),
    [anon_sym_void] = ACTIONS(1870),
    [anon_sym_public] = ACTIONS(1870),
    [anon_sym_protected] = ACTIONS(1870),
    [anon_sym_private] = ACTIONS(1870),
    [anon_sym_static] = ACTIONS(1870),
    [anon_sym_final] = ACTIONS(1870),
    [anon_sym_synchronized] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1868),
    [anon_sym_BANG] = ACTIONS(1868),
    [anon_sym_new] = ACTIONS(1870),
  },
  [STATE(710)] = {
    [sym_comment] = STATE(710),
    [sym_groovy_doc] = STATE(710),
    [ts_builtin_sym_end] = ACTIONS(1779),
    [sym_identifier] = ACTIONS(1781),
    [anon_sym_break] = ACTIONS(1781),
    [anon_sym_continue] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_STAR] = ACTIONS(1781),
    [anon_sym_STAR_COLON] = ACTIONS(1779),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_import] = ACTIONS(1781),
    [anon_sym_as] = ACTIONS(1781),
    [anon_sym_package] = ACTIONS(1781),
    [anon_sym_AT] = ACTIONS(1781),
    [anon_sym_assert] = ACTIONS(1781),
    [anon_sym_PLUS_PLUS] = ACTIONS(1779),
    [anon_sym_DASH_DASH] = ACTIONS(1779),
    [anon_sym_SLASH] = ACTIONS(1781),
    [anon_sym_PLUS] = ACTIONS(1781),
    [anon_sym_DASH] = ACTIONS(1781),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_ATinterface] = ACTIONS(1779),
    [anon_sym_interface] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1781),
    [anon_sym_def] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_null] = ACTIONS(1781),
    [anon_sym_this] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_while] = ACTIONS(1781),
    [anon_sym_for] = ACTIONS(1781),
    [aux_sym__type_identifier_token1] = ACTIONS(1781),
    [anon_sym_if] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1779),
    [aux_sym_number_literal_token1] = ACTIONS(1781),
    [aux_sym_number_literal_token2] = ACTIONS(1779),
    [aux_sym_number_literal_token3] = ACTIONS(1779),
    [aux_sym_number_literal_token4] = ACTIONS(1781),
    [aux_sym_number_literal_token5] = ACTIONS(1779),
    [anon_sym_pipeline] = ACTIONS(1781),
    [anon_sym_return] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1781),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1781),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1779),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1779),
    [anon_sym_switch] = ACTIONS(1781),
    [anon_sym_try] = ACTIONS(1781),
    [anon_sym_catch] = ACTIONS(1781),
    [anon_sym_finally] = ACTIONS(1781),
    [anon_sym_int] = ACTIONS(1781),
    [anon_sym_boolean] = ACTIONS(1781),
    [anon_sym_char] = ACTIONS(1781),
    [anon_sym_short] = ACTIONS(1781),
    [anon_sym_long] = ACTIONS(1781),
    [anon_sym_float] = ACTIONS(1781),
    [anon_sym_double] = ACTIONS(1781),
    [anon_sym_void] = ACTIONS(1781),
    [anon_sym_public] = ACTIONS(1781),
    [anon_sym_protected] = ACTIONS(1781),
    [anon_sym_private] = ACTIONS(1781),
    [anon_sym_static] = ACTIONS(1781),
    [anon_sym_final] = ACTIONS(1781),
    [anon_sym_synchronized] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1779),
    [anon_sym_BANG] = ACTIONS(1779),
    [anon_sym_new] = ACTIONS(1781),
  },
  [STATE(711)] = {
    [sym_comment] = STATE(711),
    [sym_groovy_doc] = STATE(711),
    [ts_builtin_sym_end] = ACTIONS(1876),
    [sym_identifier] = ACTIONS(1878),
    [anon_sym_break] = ACTIONS(1878),
    [anon_sym_continue] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_STAR] = ACTIONS(1878),
    [anon_sym_STAR_COLON] = ACTIONS(1876),
    [anon_sym_import] = ACTIONS(1878),
    [anon_sym_package] = ACTIONS(1878),
    [anon_sym_AT] = ACTIONS(1878),
    [anon_sym_assert] = ACTIONS(1878),
    [anon_sym_PLUS_PLUS] = ACTIONS(1876),
    [anon_sym_DASH_DASH] = ACTIONS(1876),
    [anon_sym_SLASH] = ACTIONS(1878),
    [anon_sym_PLUS] = ACTIONS(1878),
    [anon_sym_DASH] = ACTIONS(1878),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_ATinterface] = ACTIONS(1876),
    [anon_sym_interface] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_LBRACE] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1878),
    [anon_sym_def] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_RPAREN] = ACTIONS(1876),
    [anon_sym_null] = ACTIONS(1878),
    [anon_sym_this] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_while] = ACTIONS(1878),
    [anon_sym_for] = ACTIONS(1878),
    [aux_sym__type_identifier_token1] = ACTIONS(1878),
    [anon_sym_if] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1876),
    [aux_sym_number_literal_token1] = ACTIONS(1878),
    [aux_sym_number_literal_token2] = ACTIONS(1876),
    [aux_sym_number_literal_token3] = ACTIONS(1876),
    [aux_sym_number_literal_token4] = ACTIONS(1878),
    [aux_sym_number_literal_token5] = ACTIONS(1876),
    [anon_sym_pipeline] = ACTIONS(1878),
    [anon_sym_return] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1878),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1878),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1876),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1876),
    [anon_sym_switch] = ACTIONS(1878),
    [anon_sym_try] = ACTIONS(1878),
    [anon_sym_catch] = ACTIONS(1878),
    [anon_sym_finally] = ACTIONS(1878),
    [anon_sym_int] = ACTIONS(1878),
    [anon_sym_boolean] = ACTIONS(1878),
    [anon_sym_char] = ACTIONS(1878),
    [anon_sym_short] = ACTIONS(1878),
    [anon_sym_long] = ACTIONS(1878),
    [anon_sym_float] = ACTIONS(1878),
    [anon_sym_double] = ACTIONS(1878),
    [anon_sym_void] = ACTIONS(1878),
    [anon_sym_public] = ACTIONS(1878),
    [anon_sym_protected] = ACTIONS(1878),
    [anon_sym_private] = ACTIONS(1878),
    [anon_sym_static] = ACTIONS(1878),
    [anon_sym_final] = ACTIONS(1878),
    [anon_sym_synchronized] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1876),
    [anon_sym_BANG] = ACTIONS(1876),
    [anon_sym_new] = ACTIONS(1878),
  },
  [STATE(712)] = {
    [sym_comment] = STATE(712),
    [sym_groovy_doc] = STATE(712),
    [ts_builtin_sym_end] = ACTIONS(1880),
    [sym_identifier] = ACTIONS(1882),
    [anon_sym_break] = ACTIONS(1882),
    [anon_sym_continue] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_STAR] = ACTIONS(1882),
    [anon_sym_STAR_COLON] = ACTIONS(1880),
    [anon_sym_import] = ACTIONS(1882),
    [anon_sym_package] = ACTIONS(1882),
    [anon_sym_AT] = ACTIONS(1882),
    [anon_sym_assert] = ACTIONS(1882),
    [anon_sym_PLUS_PLUS] = ACTIONS(1880),
    [anon_sym_DASH_DASH] = ACTIONS(1880),
    [anon_sym_SLASH] = ACTIONS(1882),
    [anon_sym_PLUS] = ACTIONS(1882),
    [anon_sym_DASH] = ACTIONS(1882),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_ATinterface] = ACTIONS(1880),
    [anon_sym_interface] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_LBRACE] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1882),
    [anon_sym_def] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_RPAREN] = ACTIONS(1880),
    [anon_sym_null] = ACTIONS(1882),
    [anon_sym_this] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_while] = ACTIONS(1882),
    [anon_sym_for] = ACTIONS(1882),
    [aux_sym__type_identifier_token1] = ACTIONS(1882),
    [anon_sym_if] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1880),
    [aux_sym_number_literal_token1] = ACTIONS(1882),
    [aux_sym_number_literal_token2] = ACTIONS(1880),
    [aux_sym_number_literal_token3] = ACTIONS(1880),
    [aux_sym_number_literal_token4] = ACTIONS(1882),
    [aux_sym_number_literal_token5] = ACTIONS(1880),
    [anon_sym_pipeline] = ACTIONS(1882),
    [anon_sym_return] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1882),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1882),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1880),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1880),
    [anon_sym_switch] = ACTIONS(1882),
    [anon_sym_try] = ACTIONS(1882),
    [anon_sym_catch] = ACTIONS(1882),
    [anon_sym_finally] = ACTIONS(1882),
    [anon_sym_int] = ACTIONS(1882),
    [anon_sym_boolean] = ACTIONS(1882),
    [anon_sym_char] = ACTIONS(1882),
    [anon_sym_short] = ACTIONS(1882),
    [anon_sym_long] = ACTIONS(1882),
    [anon_sym_float] = ACTIONS(1882),
    [anon_sym_double] = ACTIONS(1882),
    [anon_sym_void] = ACTIONS(1882),
    [anon_sym_public] = ACTIONS(1882),
    [anon_sym_protected] = ACTIONS(1882),
    [anon_sym_private] = ACTIONS(1882),
    [anon_sym_static] = ACTIONS(1882),
    [anon_sym_final] = ACTIONS(1882),
    [anon_sym_synchronized] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1880),
    [anon_sym_BANG] = ACTIONS(1880),
    [anon_sym_new] = ACTIONS(1882),
  },
  [STATE(713)] = {
    [sym_comment] = STATE(713),
    [sym_groovy_doc] = STATE(713),
    [ts_builtin_sym_end] = ACTIONS(1812),
    [sym_identifier] = ACTIONS(1810),
    [anon_sym_break] = ACTIONS(1810),
    [anon_sym_continue] = ACTIONS(1810),
    [anon_sym_SEMI] = ACTIONS(1812),
    [anon_sym_STAR] = ACTIONS(1810),
    [anon_sym_STAR_COLON] = ACTIONS(1812),
    [anon_sym_DOT] = ACTIONS(1884),
    [anon_sym_import] = ACTIONS(1810),
    [anon_sym_as] = ACTIONS(1886),
    [anon_sym_package] = ACTIONS(1810),
    [anon_sym_AT] = ACTIONS(1810),
    [anon_sym_assert] = ACTIONS(1810),
    [anon_sym_PLUS_PLUS] = ACTIONS(1812),
    [anon_sym_DASH_DASH] = ACTIONS(1812),
    [anon_sym_SLASH] = ACTIONS(1810),
    [anon_sym_PLUS] = ACTIONS(1810),
    [anon_sym_DASH] = ACTIONS(1810),
    [anon_sym_true] = ACTIONS(1810),
    [anon_sym_false] = ACTIONS(1810),
    [anon_sym_ATinterface] = ACTIONS(1812),
    [anon_sym_interface] = ACTIONS(1810),
    [anon_sym_class] = ACTIONS(1810),
    [anon_sym_LBRACE] = ACTIONS(1812),
    [anon_sym_RBRACE] = ACTIONS(1812),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1810),
    [anon_sym_def] = ACTIONS(1810),
    [anon_sym_LPAREN] = ACTIONS(1812),
    [anon_sym_null] = ACTIONS(1810),
    [anon_sym_this] = ACTIONS(1810),
    [anon_sym_do] = ACTIONS(1810),
    [anon_sym_while] = ACTIONS(1810),
    [anon_sym_for] = ACTIONS(1810),
    [aux_sym__type_identifier_token1] = ACTIONS(1810),
    [anon_sym_if] = ACTIONS(1810),
    [anon_sym_else] = ACTIONS(1810),
    [anon_sym_LBRACK] = ACTIONS(1812),
    [aux_sym_number_literal_token1] = ACTIONS(1810),
    [aux_sym_number_literal_token2] = ACTIONS(1812),
    [aux_sym_number_literal_token3] = ACTIONS(1812),
    [aux_sym_number_literal_token4] = ACTIONS(1810),
    [aux_sym_number_literal_token5] = ACTIONS(1812),
    [anon_sym_pipeline] = ACTIONS(1810),
    [anon_sym_return] = ACTIONS(1810),
    [anon_sym_SQUOTE] = ACTIONS(1810),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1812),
    [anon_sym_DQUOTE] = ACTIONS(1810),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1812),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1812),
    [anon_sym_switch] = ACTIONS(1810),
    [anon_sym_try] = ACTIONS(1810),
    [anon_sym_catch] = ACTIONS(1810),
    [anon_sym_finally] = ACTIONS(1810),
    [anon_sym_int] = ACTIONS(1810),
    [anon_sym_boolean] = ACTIONS(1810),
    [anon_sym_char] = ACTIONS(1810),
    [anon_sym_short] = ACTIONS(1810),
    [anon_sym_long] = ACTIONS(1810),
    [anon_sym_float] = ACTIONS(1810),
    [anon_sym_double] = ACTIONS(1810),
    [anon_sym_void] = ACTIONS(1810),
    [anon_sym_public] = ACTIONS(1810),
    [anon_sym_protected] = ACTIONS(1810),
    [anon_sym_private] = ACTIONS(1810),
    [anon_sym_static] = ACTIONS(1810),
    [anon_sym_final] = ACTIONS(1810),
    [anon_sym_synchronized] = ACTIONS(1810),
    [anon_sym_TILDE] = ACTIONS(1812),
    [anon_sym_BANG] = ACTIONS(1812),
    [anon_sym_new] = ACTIONS(1810),
  },
  [STATE(714)] = {
    [sym_comment] = STATE(714),
    [sym_groovy_doc] = STATE(714),
    [sym_parameter_list] = STATE(773),
    [ts_builtin_sym_end] = ACTIONS(1741),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(1888),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_pipeline] = ACTIONS(1739),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_catch] = ACTIONS(1739),
    [anon_sym_finally] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(715)] = {
    [sym_comment] = STATE(715),
    [sym_groovy_doc] = STATE(715),
    [sym_parameter_list] = STATE(769),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1890),
    [anon_sym_STAR_COLON] = ACTIONS(1892),
    [anon_sym_import] = ACTIONS(1890),
    [anon_sym_package] = ACTIONS(1890),
    [anon_sym_AT] = ACTIONS(1890),
    [anon_sym_assert] = ACTIONS(1890),
    [anon_sym_EQ] = ACTIONS(1894),
    [anon_sym_PLUS_PLUS] = ACTIONS(1892),
    [anon_sym_DASH_DASH] = ACTIONS(1892),
    [anon_sym_SLASH] = ACTIONS(1890),
    [anon_sym_PLUS] = ACTIONS(1890),
    [anon_sym_DASH] = ACTIONS(1890),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_ATinterface] = ACTIONS(1892),
    [anon_sym_interface] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1890),
    [anon_sym_def] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1890),
    [anon_sym_this] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [aux_sym__type_identifier_token1] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [aux_sym_number_literal_token1] = ACTIONS(1890),
    [aux_sym_number_literal_token2] = ACTIONS(1892),
    [aux_sym_number_literal_token3] = ACTIONS(1892),
    [aux_sym_number_literal_token4] = ACTIONS(1890),
    [aux_sym_number_literal_token5] = ACTIONS(1892),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1892),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1892),
    [anon_sym_switch] = ACTIONS(1890),
    [anon_sym_case] = ACTIONS(1890),
    [anon_sym_default] = ACTIONS(1890),
    [anon_sym_try] = ACTIONS(1890),
    [anon_sym_catch] = ACTIONS(1890),
    [anon_sym_finally] = ACTIONS(1890),
    [anon_sym_int] = ACTIONS(1890),
    [anon_sym_boolean] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_short] = ACTIONS(1890),
    [anon_sym_long] = ACTIONS(1890),
    [anon_sym_float] = ACTIONS(1890),
    [anon_sym_double] = ACTIONS(1890),
    [anon_sym_void] = ACTIONS(1890),
    [anon_sym_public] = ACTIONS(1890),
    [anon_sym_protected] = ACTIONS(1890),
    [anon_sym_private] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_final] = ACTIONS(1890),
    [anon_sym_synchronized] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_new] = ACTIONS(1890),
  },
  [STATE(716)] = {
    [sym_comment] = STATE(716),
    [sym_groovy_doc] = STATE(716),
    [aux_sym__juxt_argument_list_repeat1] = STATE(740),
    [sym_identifier] = ACTIONS(1896),
    [anon_sym_break] = ACTIONS(1896),
    [anon_sym_continue] = ACTIONS(1896),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_STAR] = ACTIONS(1896),
    [anon_sym_STAR_COLON] = ACTIONS(1898),
    [anon_sym_import] = ACTIONS(1896),
    [anon_sym_package] = ACTIONS(1896),
    [anon_sym_AT] = ACTIONS(1896),
    [anon_sym_assert] = ACTIONS(1896),
    [anon_sym_PLUS_PLUS] = ACTIONS(1898),
    [anon_sym_DASH_DASH] = ACTIONS(1898),
    [anon_sym_SLASH] = ACTIONS(1896),
    [anon_sym_PLUS] = ACTIONS(1896),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1896),
    [anon_sym_false] = ACTIONS(1896),
    [anon_sym_ATinterface] = ACTIONS(1898),
    [anon_sym_interface] = ACTIONS(1896),
    [anon_sym_class] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_RBRACE] = ACTIONS(1898),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1896),
    [anon_sym_def] = ACTIONS(1896),
    [anon_sym_LPAREN] = ACTIONS(1898),
    [anon_sym_null] = ACTIONS(1896),
    [anon_sym_this] = ACTIONS(1896),
    [anon_sym_do] = ACTIONS(1896),
    [anon_sym_while] = ACTIONS(1896),
    [anon_sym_for] = ACTIONS(1896),
    [aux_sym__type_identifier_token1] = ACTIONS(1896),
    [anon_sym_if] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1896),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [aux_sym_number_literal_token1] = ACTIONS(1896),
    [aux_sym_number_literal_token2] = ACTIONS(1898),
    [aux_sym_number_literal_token3] = ACTIONS(1898),
    [aux_sym_number_literal_token4] = ACTIONS(1896),
    [aux_sym_number_literal_token5] = ACTIONS(1898),
    [anon_sym_return] = ACTIONS(1896),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1898),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1898),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1898),
    [anon_sym_switch] = ACTIONS(1896),
    [anon_sym_case] = ACTIONS(1896),
    [anon_sym_default] = ACTIONS(1896),
    [anon_sym_try] = ACTIONS(1896),
    [anon_sym_catch] = ACTIONS(1896),
    [anon_sym_finally] = ACTIONS(1896),
    [anon_sym_int] = ACTIONS(1896),
    [anon_sym_boolean] = ACTIONS(1896),
    [anon_sym_char] = ACTIONS(1896),
    [anon_sym_short] = ACTIONS(1896),
    [anon_sym_long] = ACTIONS(1896),
    [anon_sym_float] = ACTIONS(1896),
    [anon_sym_double] = ACTIONS(1896),
    [anon_sym_void] = ACTIONS(1896),
    [anon_sym_public] = ACTIONS(1896),
    [anon_sym_protected] = ACTIONS(1896),
    [anon_sym_private] = ACTIONS(1896),
    [anon_sym_static] = ACTIONS(1896),
    [anon_sym_final] = ACTIONS(1896),
    [anon_sym_synchronized] = ACTIONS(1896),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_BANG] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1896),
  },
  [STATE(717)] = {
    [sym_comment] = STATE(717),
    [sym_groovy_doc] = STATE(717),
    [ts_builtin_sym_end] = ACTIONS(1900),
    [sym_identifier] = ACTIONS(1902),
    [anon_sym_break] = ACTIONS(1902),
    [anon_sym_continue] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_STAR] = ACTIONS(1902),
    [anon_sym_STAR_COLON] = ACTIONS(1900),
    [anon_sym_import] = ACTIONS(1902),
    [anon_sym_package] = ACTIONS(1902),
    [anon_sym_AT] = ACTIONS(1902),
    [anon_sym_assert] = ACTIONS(1902),
    [anon_sym_PLUS_PLUS] = ACTIONS(1900),
    [anon_sym_DASH_DASH] = ACTIONS(1900),
    [anon_sym_SLASH] = ACTIONS(1902),
    [anon_sym_PLUS] = ACTIONS(1902),
    [anon_sym_DASH] = ACTIONS(1902),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_ATinterface] = ACTIONS(1900),
    [anon_sym_interface] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_LBRACE] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1902),
    [anon_sym_def] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_RPAREN] = ACTIONS(1900),
    [anon_sym_null] = ACTIONS(1902),
    [anon_sym_this] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_while] = ACTIONS(1902),
    [anon_sym_for] = ACTIONS(1902),
    [aux_sym__type_identifier_token1] = ACTIONS(1902),
    [anon_sym_if] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1900),
    [aux_sym_number_literal_token1] = ACTIONS(1902),
    [aux_sym_number_literal_token2] = ACTIONS(1900),
    [aux_sym_number_literal_token3] = ACTIONS(1900),
    [aux_sym_number_literal_token4] = ACTIONS(1902),
    [aux_sym_number_literal_token5] = ACTIONS(1900),
    [anon_sym_pipeline] = ACTIONS(1902),
    [anon_sym_return] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1902),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1902),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1900),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1900),
    [anon_sym_switch] = ACTIONS(1902),
    [anon_sym_try] = ACTIONS(1902),
    [anon_sym_catch] = ACTIONS(1902),
    [anon_sym_finally] = ACTIONS(1902),
    [anon_sym_int] = ACTIONS(1902),
    [anon_sym_boolean] = ACTIONS(1902),
    [anon_sym_char] = ACTIONS(1902),
    [anon_sym_short] = ACTIONS(1902),
    [anon_sym_long] = ACTIONS(1902),
    [anon_sym_float] = ACTIONS(1902),
    [anon_sym_double] = ACTIONS(1902),
    [anon_sym_void] = ACTIONS(1902),
    [anon_sym_public] = ACTIONS(1902),
    [anon_sym_protected] = ACTIONS(1902),
    [anon_sym_private] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1902),
    [anon_sym_final] = ACTIONS(1902),
    [anon_sym_synchronized] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1900),
    [anon_sym_BANG] = ACTIONS(1900),
    [anon_sym_new] = ACTIONS(1902),
  },
  [STATE(718)] = {
    [sym_comment] = STATE(718),
    [sym_groovy_doc] = STATE(718),
    [ts_builtin_sym_end] = ACTIONS(1904),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_RPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_pipeline] = ACTIONS(1906),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_catch] = ACTIONS(1906),
    [anon_sym_finally] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(719)] = {
    [sym_comment] = STATE(719),
    [sym_groovy_doc] = STATE(719),
    [sym_identifier] = ACTIONS(1781),
    [anon_sym_break] = ACTIONS(1781),
    [anon_sym_continue] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_STAR] = ACTIONS(1781),
    [anon_sym_STAR_COLON] = ACTIONS(1779),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_import] = ACTIONS(1781),
    [anon_sym_as] = ACTIONS(1781),
    [anon_sym_package] = ACTIONS(1781),
    [anon_sym_AT] = ACTIONS(1781),
    [anon_sym_assert] = ACTIONS(1781),
    [anon_sym_PLUS_PLUS] = ACTIONS(1779),
    [anon_sym_DASH_DASH] = ACTIONS(1779),
    [anon_sym_SLASH] = ACTIONS(1781),
    [anon_sym_PLUS] = ACTIONS(1781),
    [anon_sym_DASH] = ACTIONS(1781),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_ATinterface] = ACTIONS(1779),
    [anon_sym_interface] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1781),
    [anon_sym_def] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_null] = ACTIONS(1781),
    [anon_sym_this] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_while] = ACTIONS(1781),
    [anon_sym_for] = ACTIONS(1781),
    [aux_sym__type_identifier_token1] = ACTIONS(1781),
    [anon_sym_if] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1779),
    [aux_sym_number_literal_token1] = ACTIONS(1781),
    [aux_sym_number_literal_token2] = ACTIONS(1779),
    [aux_sym_number_literal_token3] = ACTIONS(1779),
    [aux_sym_number_literal_token4] = ACTIONS(1781),
    [aux_sym_number_literal_token5] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1781),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1781),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1779),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1779),
    [anon_sym_switch] = ACTIONS(1781),
    [anon_sym_case] = ACTIONS(1781),
    [anon_sym_default] = ACTIONS(1781),
    [anon_sym_try] = ACTIONS(1781),
    [anon_sym_catch] = ACTIONS(1781),
    [anon_sym_finally] = ACTIONS(1781),
    [anon_sym_int] = ACTIONS(1781),
    [anon_sym_boolean] = ACTIONS(1781),
    [anon_sym_char] = ACTIONS(1781),
    [anon_sym_short] = ACTIONS(1781),
    [anon_sym_long] = ACTIONS(1781),
    [anon_sym_float] = ACTIONS(1781),
    [anon_sym_double] = ACTIONS(1781),
    [anon_sym_void] = ACTIONS(1781),
    [anon_sym_public] = ACTIONS(1781),
    [anon_sym_protected] = ACTIONS(1781),
    [anon_sym_private] = ACTIONS(1781),
    [anon_sym_static] = ACTIONS(1781),
    [anon_sym_final] = ACTIONS(1781),
    [anon_sym_synchronized] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1779),
    [anon_sym_BANG] = ACTIONS(1779),
    [anon_sym_new] = ACTIONS(1781),
  },
  [STATE(720)] = {
    [sym_comment] = STATE(720),
    [sym_groovy_doc] = STATE(720),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_RPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(721)] = {
    [sym_comment] = STATE(721),
    [sym_groovy_doc] = STATE(721),
    [ts_builtin_sym_end] = ACTIONS(1908),
    [sym_identifier] = ACTIONS(1910),
    [anon_sym_break] = ACTIONS(1910),
    [anon_sym_continue] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_STAR] = ACTIONS(1910),
    [anon_sym_STAR_COLON] = ACTIONS(1908),
    [anon_sym_import] = ACTIONS(1910),
    [anon_sym_package] = ACTIONS(1910),
    [anon_sym_AT] = ACTIONS(1910),
    [anon_sym_assert] = ACTIONS(1910),
    [anon_sym_PLUS_PLUS] = ACTIONS(1908),
    [anon_sym_DASH_DASH] = ACTIONS(1908),
    [anon_sym_SLASH] = ACTIONS(1910),
    [anon_sym_PLUS] = ACTIONS(1910),
    [anon_sym_DASH] = ACTIONS(1910),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_ATinterface] = ACTIONS(1908),
    [anon_sym_interface] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_LBRACE] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1910),
    [anon_sym_def] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_RPAREN] = ACTIONS(1908),
    [anon_sym_null] = ACTIONS(1910),
    [anon_sym_this] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_while] = ACTIONS(1910),
    [anon_sym_for] = ACTIONS(1910),
    [aux_sym__type_identifier_token1] = ACTIONS(1910),
    [anon_sym_if] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1908),
    [aux_sym_number_literal_token1] = ACTIONS(1910),
    [aux_sym_number_literal_token2] = ACTIONS(1908),
    [aux_sym_number_literal_token3] = ACTIONS(1908),
    [aux_sym_number_literal_token4] = ACTIONS(1910),
    [aux_sym_number_literal_token5] = ACTIONS(1908),
    [anon_sym_pipeline] = ACTIONS(1910),
    [anon_sym_return] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1910),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1910),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1908),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1908),
    [anon_sym_switch] = ACTIONS(1910),
    [anon_sym_try] = ACTIONS(1910),
    [anon_sym_catch] = ACTIONS(1910),
    [anon_sym_finally] = ACTIONS(1910),
    [anon_sym_int] = ACTIONS(1910),
    [anon_sym_boolean] = ACTIONS(1910),
    [anon_sym_char] = ACTIONS(1910),
    [anon_sym_short] = ACTIONS(1910),
    [anon_sym_long] = ACTIONS(1910),
    [anon_sym_float] = ACTIONS(1910),
    [anon_sym_double] = ACTIONS(1910),
    [anon_sym_void] = ACTIONS(1910),
    [anon_sym_public] = ACTIONS(1910),
    [anon_sym_protected] = ACTIONS(1910),
    [anon_sym_private] = ACTIONS(1910),
    [anon_sym_static] = ACTIONS(1910),
    [anon_sym_final] = ACTIONS(1910),
    [anon_sym_synchronized] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1908),
    [anon_sym_BANG] = ACTIONS(1908),
    [anon_sym_new] = ACTIONS(1910),
  },
  [STATE(722)] = {
    [sym_comment] = STATE(722),
    [sym_groovy_doc] = STATE(722),
    [ts_builtin_sym_end] = ACTIONS(1912),
    [sym_identifier] = ACTIONS(1914),
    [anon_sym_break] = ACTIONS(1914),
    [anon_sym_continue] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_STAR] = ACTIONS(1914),
    [anon_sym_STAR_COLON] = ACTIONS(1912),
    [anon_sym_import] = ACTIONS(1914),
    [anon_sym_package] = ACTIONS(1914),
    [anon_sym_AT] = ACTIONS(1914),
    [anon_sym_assert] = ACTIONS(1914),
    [anon_sym_PLUS_PLUS] = ACTIONS(1912),
    [anon_sym_DASH_DASH] = ACTIONS(1912),
    [anon_sym_SLASH] = ACTIONS(1914),
    [anon_sym_PLUS] = ACTIONS(1914),
    [anon_sym_DASH] = ACTIONS(1914),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_ATinterface] = ACTIONS(1912),
    [anon_sym_interface] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_LBRACE] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1914),
    [anon_sym_def] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_RPAREN] = ACTIONS(1912),
    [anon_sym_null] = ACTIONS(1914),
    [anon_sym_this] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_while] = ACTIONS(1914),
    [anon_sym_for] = ACTIONS(1914),
    [aux_sym__type_identifier_token1] = ACTIONS(1914),
    [anon_sym_if] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1912),
    [aux_sym_number_literal_token1] = ACTIONS(1914),
    [aux_sym_number_literal_token2] = ACTIONS(1912),
    [aux_sym_number_literal_token3] = ACTIONS(1912),
    [aux_sym_number_literal_token4] = ACTIONS(1914),
    [aux_sym_number_literal_token5] = ACTIONS(1912),
    [anon_sym_pipeline] = ACTIONS(1914),
    [anon_sym_return] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1914),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1914),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1912),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1912),
    [anon_sym_switch] = ACTIONS(1914),
    [anon_sym_try] = ACTIONS(1914),
    [anon_sym_catch] = ACTIONS(1914),
    [anon_sym_finally] = ACTIONS(1914),
    [anon_sym_int] = ACTIONS(1914),
    [anon_sym_boolean] = ACTIONS(1914),
    [anon_sym_char] = ACTIONS(1914),
    [anon_sym_short] = ACTIONS(1914),
    [anon_sym_long] = ACTIONS(1914),
    [anon_sym_float] = ACTIONS(1914),
    [anon_sym_double] = ACTIONS(1914),
    [anon_sym_void] = ACTIONS(1914),
    [anon_sym_public] = ACTIONS(1914),
    [anon_sym_protected] = ACTIONS(1914),
    [anon_sym_private] = ACTIONS(1914),
    [anon_sym_static] = ACTIONS(1914),
    [anon_sym_final] = ACTIONS(1914),
    [anon_sym_synchronized] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1912),
    [anon_sym_BANG] = ACTIONS(1912),
    [anon_sym_new] = ACTIONS(1914),
  },
  [STATE(723)] = {
    [sym_comment] = STATE(723),
    [sym_groovy_doc] = STATE(723),
    [sym_identifier] = ACTIONS(1870),
    [anon_sym_break] = ACTIONS(1870),
    [anon_sym_continue] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_STAR] = ACTIONS(1870),
    [anon_sym_STAR_COLON] = ACTIONS(1868),
    [anon_sym_DOT] = ACTIONS(1916),
    [anon_sym_import] = ACTIONS(1870),
    [anon_sym_as] = ACTIONS(1918),
    [anon_sym_package] = ACTIONS(1870),
    [anon_sym_AT] = ACTIONS(1870),
    [anon_sym_assert] = ACTIONS(1870),
    [anon_sym_PLUS_PLUS] = ACTIONS(1868),
    [anon_sym_DASH_DASH] = ACTIONS(1868),
    [anon_sym_SLASH] = ACTIONS(1870),
    [anon_sym_PLUS] = ACTIONS(1870),
    [anon_sym_DASH] = ACTIONS(1870),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_ATinterface] = ACTIONS(1868),
    [anon_sym_interface] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1870),
    [anon_sym_def] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_null] = ACTIONS(1870),
    [anon_sym_this] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_while] = ACTIONS(1870),
    [anon_sym_for] = ACTIONS(1870),
    [aux_sym__type_identifier_token1] = ACTIONS(1870),
    [anon_sym_if] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1868),
    [aux_sym_number_literal_token1] = ACTIONS(1870),
    [aux_sym_number_literal_token2] = ACTIONS(1868),
    [aux_sym_number_literal_token3] = ACTIONS(1868),
    [aux_sym_number_literal_token4] = ACTIONS(1870),
    [aux_sym_number_literal_token5] = ACTIONS(1868),
    [anon_sym_return] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1870),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1870),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1868),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1868),
    [anon_sym_switch] = ACTIONS(1870),
    [anon_sym_case] = ACTIONS(1870),
    [anon_sym_default] = ACTIONS(1870),
    [anon_sym_try] = ACTIONS(1870),
    [anon_sym_catch] = ACTIONS(1870),
    [anon_sym_finally] = ACTIONS(1870),
    [anon_sym_int] = ACTIONS(1870),
    [anon_sym_boolean] = ACTIONS(1870),
    [anon_sym_char] = ACTIONS(1870),
    [anon_sym_short] = ACTIONS(1870),
    [anon_sym_long] = ACTIONS(1870),
    [anon_sym_float] = ACTIONS(1870),
    [anon_sym_double] = ACTIONS(1870),
    [anon_sym_void] = ACTIONS(1870),
    [anon_sym_public] = ACTIONS(1870),
    [anon_sym_protected] = ACTIONS(1870),
    [anon_sym_private] = ACTIONS(1870),
    [anon_sym_static] = ACTIONS(1870),
    [anon_sym_final] = ACTIONS(1870),
    [anon_sym_synchronized] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1868),
    [anon_sym_BANG] = ACTIONS(1868),
    [anon_sym_new] = ACTIONS(1870),
  },
  [STATE(724)] = {
    [sym_comment] = STATE(724),
    [sym_groovy_doc] = STATE(724),
    [ts_builtin_sym_end] = ACTIONS(1920),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_break] = ACTIONS(1922),
    [anon_sym_continue] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_STAR] = ACTIONS(1922),
    [anon_sym_STAR_COLON] = ACTIONS(1920),
    [anon_sym_import] = ACTIONS(1922),
    [anon_sym_package] = ACTIONS(1922),
    [anon_sym_AT] = ACTIONS(1922),
    [anon_sym_assert] = ACTIONS(1922),
    [anon_sym_PLUS_PLUS] = ACTIONS(1920),
    [anon_sym_DASH_DASH] = ACTIONS(1920),
    [anon_sym_SLASH] = ACTIONS(1922),
    [anon_sym_PLUS] = ACTIONS(1922),
    [anon_sym_DASH] = ACTIONS(1922),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_ATinterface] = ACTIONS(1920),
    [anon_sym_interface] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_LBRACE] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1922),
    [anon_sym_def] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_RPAREN] = ACTIONS(1920),
    [anon_sym_null] = ACTIONS(1922),
    [anon_sym_this] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_while] = ACTIONS(1922),
    [anon_sym_for] = ACTIONS(1922),
    [aux_sym__type_identifier_token1] = ACTIONS(1922),
    [anon_sym_if] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1920),
    [aux_sym_number_literal_token1] = ACTIONS(1922),
    [aux_sym_number_literal_token2] = ACTIONS(1920),
    [aux_sym_number_literal_token3] = ACTIONS(1920),
    [aux_sym_number_literal_token4] = ACTIONS(1922),
    [aux_sym_number_literal_token5] = ACTIONS(1920),
    [anon_sym_pipeline] = ACTIONS(1922),
    [anon_sym_return] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1922),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1922),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1920),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1920),
    [anon_sym_switch] = ACTIONS(1922),
    [anon_sym_try] = ACTIONS(1922),
    [anon_sym_catch] = ACTIONS(1922),
    [anon_sym_finally] = ACTIONS(1922),
    [anon_sym_int] = ACTIONS(1922),
    [anon_sym_boolean] = ACTIONS(1922),
    [anon_sym_char] = ACTIONS(1922),
    [anon_sym_short] = ACTIONS(1922),
    [anon_sym_long] = ACTIONS(1922),
    [anon_sym_float] = ACTIONS(1922),
    [anon_sym_double] = ACTIONS(1922),
    [anon_sym_void] = ACTIONS(1922),
    [anon_sym_public] = ACTIONS(1922),
    [anon_sym_protected] = ACTIONS(1922),
    [anon_sym_private] = ACTIONS(1922),
    [anon_sym_static] = ACTIONS(1922),
    [anon_sym_final] = ACTIONS(1922),
    [anon_sym_synchronized] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1920),
    [anon_sym_BANG] = ACTIONS(1920),
    [anon_sym_new] = ACTIONS(1922),
  },
  [STATE(725)] = {
    [sym_comment] = STATE(725),
    [sym_groovy_doc] = STATE(725),
    [ts_builtin_sym_end] = ACTIONS(1924),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_break] = ACTIONS(1926),
    [anon_sym_continue] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_STAR] = ACTIONS(1926),
    [anon_sym_STAR_COLON] = ACTIONS(1924),
    [anon_sym_import] = ACTIONS(1926),
    [anon_sym_package] = ACTIONS(1926),
    [anon_sym_AT] = ACTIONS(1926),
    [anon_sym_assert] = ACTIONS(1926),
    [anon_sym_PLUS_PLUS] = ACTIONS(1924),
    [anon_sym_DASH_DASH] = ACTIONS(1924),
    [anon_sym_SLASH] = ACTIONS(1926),
    [anon_sym_PLUS] = ACTIONS(1926),
    [anon_sym_DASH] = ACTIONS(1926),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_ATinterface] = ACTIONS(1924),
    [anon_sym_interface] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_LBRACE] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1926),
    [anon_sym_def] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_RPAREN] = ACTIONS(1924),
    [anon_sym_null] = ACTIONS(1926),
    [anon_sym_this] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_while] = ACTIONS(1926),
    [anon_sym_for] = ACTIONS(1926),
    [aux_sym__type_identifier_token1] = ACTIONS(1926),
    [anon_sym_if] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1924),
    [aux_sym_number_literal_token1] = ACTIONS(1926),
    [aux_sym_number_literal_token2] = ACTIONS(1924),
    [aux_sym_number_literal_token3] = ACTIONS(1924),
    [aux_sym_number_literal_token4] = ACTIONS(1926),
    [aux_sym_number_literal_token5] = ACTIONS(1924),
    [anon_sym_pipeline] = ACTIONS(1926),
    [anon_sym_return] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1926),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1924),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1924),
    [anon_sym_switch] = ACTIONS(1926),
    [anon_sym_try] = ACTIONS(1926),
    [anon_sym_catch] = ACTIONS(1926),
    [anon_sym_finally] = ACTIONS(1926),
    [anon_sym_int] = ACTIONS(1926),
    [anon_sym_boolean] = ACTIONS(1926),
    [anon_sym_char] = ACTIONS(1926),
    [anon_sym_short] = ACTIONS(1926),
    [anon_sym_long] = ACTIONS(1926),
    [anon_sym_float] = ACTIONS(1926),
    [anon_sym_double] = ACTIONS(1926),
    [anon_sym_void] = ACTIONS(1926),
    [anon_sym_public] = ACTIONS(1926),
    [anon_sym_protected] = ACTIONS(1926),
    [anon_sym_private] = ACTIONS(1926),
    [anon_sym_static] = ACTIONS(1926),
    [anon_sym_final] = ACTIONS(1926),
    [anon_sym_synchronized] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1924),
    [anon_sym_BANG] = ACTIONS(1924),
    [anon_sym_new] = ACTIONS(1926),
  },
  [STATE(726)] = {
    [sym_comment] = STATE(726),
    [sym_groovy_doc] = STATE(726),
    [ts_builtin_sym_end] = ACTIONS(1928),
    [sym_identifier] = ACTIONS(1930),
    [anon_sym_break] = ACTIONS(1930),
    [anon_sym_continue] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_STAR] = ACTIONS(1930),
    [anon_sym_STAR_COLON] = ACTIONS(1928),
    [anon_sym_import] = ACTIONS(1930),
    [anon_sym_package] = ACTIONS(1930),
    [anon_sym_AT] = ACTIONS(1930),
    [anon_sym_assert] = ACTIONS(1930),
    [anon_sym_PLUS_PLUS] = ACTIONS(1928),
    [anon_sym_DASH_DASH] = ACTIONS(1928),
    [anon_sym_SLASH] = ACTIONS(1930),
    [anon_sym_PLUS] = ACTIONS(1930),
    [anon_sym_DASH] = ACTIONS(1930),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_ATinterface] = ACTIONS(1928),
    [anon_sym_interface] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_LBRACE] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1930),
    [anon_sym_def] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_RPAREN] = ACTIONS(1928),
    [anon_sym_null] = ACTIONS(1930),
    [anon_sym_this] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_while] = ACTIONS(1930),
    [anon_sym_for] = ACTIONS(1930),
    [aux_sym__type_identifier_token1] = ACTIONS(1930),
    [anon_sym_if] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1928),
    [aux_sym_number_literal_token1] = ACTIONS(1930),
    [aux_sym_number_literal_token2] = ACTIONS(1928),
    [aux_sym_number_literal_token3] = ACTIONS(1928),
    [aux_sym_number_literal_token4] = ACTIONS(1930),
    [aux_sym_number_literal_token5] = ACTIONS(1928),
    [anon_sym_pipeline] = ACTIONS(1930),
    [anon_sym_return] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1930),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1928),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1928),
    [anon_sym_switch] = ACTIONS(1930),
    [anon_sym_try] = ACTIONS(1930),
    [anon_sym_catch] = ACTIONS(1930),
    [anon_sym_finally] = ACTIONS(1930),
    [anon_sym_int] = ACTIONS(1930),
    [anon_sym_boolean] = ACTIONS(1930),
    [anon_sym_char] = ACTIONS(1930),
    [anon_sym_short] = ACTIONS(1930),
    [anon_sym_long] = ACTIONS(1930),
    [anon_sym_float] = ACTIONS(1930),
    [anon_sym_double] = ACTIONS(1930),
    [anon_sym_void] = ACTIONS(1930),
    [anon_sym_public] = ACTIONS(1930),
    [anon_sym_protected] = ACTIONS(1930),
    [anon_sym_private] = ACTIONS(1930),
    [anon_sym_static] = ACTIONS(1930),
    [anon_sym_final] = ACTIONS(1930),
    [anon_sym_synchronized] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1928),
    [anon_sym_BANG] = ACTIONS(1928),
    [anon_sym_new] = ACTIONS(1930),
  },
  [STATE(727)] = {
    [sym_comment] = STATE(727),
    [sym_groovy_doc] = STATE(727),
    [ts_builtin_sym_end] = ACTIONS(1932),
    [sym_identifier] = ACTIONS(1934),
    [anon_sym_break] = ACTIONS(1934),
    [anon_sym_continue] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_STAR] = ACTIONS(1934),
    [anon_sym_STAR_COLON] = ACTIONS(1932),
    [anon_sym_import] = ACTIONS(1934),
    [anon_sym_package] = ACTIONS(1934),
    [anon_sym_AT] = ACTIONS(1934),
    [anon_sym_assert] = ACTIONS(1934),
    [anon_sym_PLUS_PLUS] = ACTIONS(1932),
    [anon_sym_DASH_DASH] = ACTIONS(1932),
    [anon_sym_SLASH] = ACTIONS(1934),
    [anon_sym_PLUS] = ACTIONS(1934),
    [anon_sym_DASH] = ACTIONS(1934),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_ATinterface] = ACTIONS(1932),
    [anon_sym_interface] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_LBRACE] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1934),
    [anon_sym_def] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_RPAREN] = ACTIONS(1932),
    [anon_sym_null] = ACTIONS(1934),
    [anon_sym_this] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_while] = ACTIONS(1934),
    [anon_sym_for] = ACTIONS(1934),
    [aux_sym__type_identifier_token1] = ACTIONS(1934),
    [anon_sym_if] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1932),
    [aux_sym_number_literal_token1] = ACTIONS(1934),
    [aux_sym_number_literal_token2] = ACTIONS(1932),
    [aux_sym_number_literal_token3] = ACTIONS(1932),
    [aux_sym_number_literal_token4] = ACTIONS(1934),
    [aux_sym_number_literal_token5] = ACTIONS(1932),
    [anon_sym_pipeline] = ACTIONS(1934),
    [anon_sym_return] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1934),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1932),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1932),
    [anon_sym_switch] = ACTIONS(1934),
    [anon_sym_try] = ACTIONS(1934),
    [anon_sym_catch] = ACTIONS(1934),
    [anon_sym_finally] = ACTIONS(1934),
    [anon_sym_int] = ACTIONS(1934),
    [anon_sym_boolean] = ACTIONS(1934),
    [anon_sym_char] = ACTIONS(1934),
    [anon_sym_short] = ACTIONS(1934),
    [anon_sym_long] = ACTIONS(1934),
    [anon_sym_float] = ACTIONS(1934),
    [anon_sym_double] = ACTIONS(1934),
    [anon_sym_void] = ACTIONS(1934),
    [anon_sym_public] = ACTIONS(1934),
    [anon_sym_protected] = ACTIONS(1934),
    [anon_sym_private] = ACTIONS(1934),
    [anon_sym_static] = ACTIONS(1934),
    [anon_sym_final] = ACTIONS(1934),
    [anon_sym_synchronized] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1932),
    [anon_sym_BANG] = ACTIONS(1932),
    [anon_sym_new] = ACTIONS(1934),
  },
  [STATE(728)] = {
    [sym_comment] = STATE(728),
    [sym_groovy_doc] = STATE(728),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(389),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_catch] = ACTIONS(387),
    [anon_sym_finally] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(389),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(729)] = {
    [sym_comment] = STATE(729),
    [sym_groovy_doc] = STATE(729),
    [ts_builtin_sym_end] = ACTIONS(1936),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_break] = ACTIONS(1938),
    [anon_sym_continue] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_STAR] = ACTIONS(1938),
    [anon_sym_STAR_COLON] = ACTIONS(1936),
    [anon_sym_import] = ACTIONS(1938),
    [anon_sym_package] = ACTIONS(1938),
    [anon_sym_AT] = ACTIONS(1938),
    [anon_sym_assert] = ACTIONS(1938),
    [anon_sym_PLUS_PLUS] = ACTIONS(1936),
    [anon_sym_DASH_DASH] = ACTIONS(1936),
    [anon_sym_SLASH] = ACTIONS(1938),
    [anon_sym_PLUS] = ACTIONS(1938),
    [anon_sym_DASH] = ACTIONS(1938),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_ATinterface] = ACTIONS(1936),
    [anon_sym_interface] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_COMMA] = ACTIONS(1936),
    [anon_sym_LBRACE] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1938),
    [anon_sym_def] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_RPAREN] = ACTIONS(1936),
    [anon_sym_null] = ACTIONS(1938),
    [anon_sym_this] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_while] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1938),
    [aux_sym__type_identifier_token1] = ACTIONS(1938),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1936),
    [aux_sym_number_literal_token1] = ACTIONS(1938),
    [aux_sym_number_literal_token2] = ACTIONS(1936),
    [aux_sym_number_literal_token3] = ACTIONS(1936),
    [aux_sym_number_literal_token4] = ACTIONS(1938),
    [aux_sym_number_literal_token5] = ACTIONS(1936),
    [anon_sym_pipeline] = ACTIONS(1938),
    [anon_sym_return] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1938),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1936),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1936),
    [anon_sym_switch] = ACTIONS(1938),
    [anon_sym_try] = ACTIONS(1938),
    [anon_sym_catch] = ACTIONS(1938),
    [anon_sym_finally] = ACTIONS(1938),
    [anon_sym_int] = ACTIONS(1938),
    [anon_sym_boolean] = ACTIONS(1938),
    [anon_sym_char] = ACTIONS(1938),
    [anon_sym_short] = ACTIONS(1938),
    [anon_sym_long] = ACTIONS(1938),
    [anon_sym_float] = ACTIONS(1938),
    [anon_sym_double] = ACTIONS(1938),
    [anon_sym_void] = ACTIONS(1938),
    [anon_sym_public] = ACTIONS(1938),
    [anon_sym_protected] = ACTIONS(1938),
    [anon_sym_private] = ACTIONS(1938),
    [anon_sym_static] = ACTIONS(1938),
    [anon_sym_final] = ACTIONS(1938),
    [anon_sym_synchronized] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1936),
    [anon_sym_BANG] = ACTIONS(1936),
    [anon_sym_new] = ACTIONS(1938),
  },
  [STATE(730)] = {
    [sym_comment] = STATE(730),
    [sym_groovy_doc] = STATE(730),
    [ts_builtin_sym_end] = ACTIONS(1940),
    [sym_identifier] = ACTIONS(1942),
    [anon_sym_break] = ACTIONS(1942),
    [anon_sym_continue] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1942),
    [anon_sym_STAR_COLON] = ACTIONS(1940),
    [anon_sym_import] = ACTIONS(1942),
    [anon_sym_package] = ACTIONS(1942),
    [anon_sym_AT] = ACTIONS(1942),
    [anon_sym_assert] = ACTIONS(1942),
    [anon_sym_PLUS_PLUS] = ACTIONS(1940),
    [anon_sym_DASH_DASH] = ACTIONS(1940),
    [anon_sym_SLASH] = ACTIONS(1942),
    [anon_sym_PLUS] = ACTIONS(1942),
    [anon_sym_DASH] = ACTIONS(1942),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_ATinterface] = ACTIONS(1940),
    [anon_sym_interface] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_LBRACE] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1942),
    [anon_sym_def] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_RPAREN] = ACTIONS(1940),
    [anon_sym_null] = ACTIONS(1942),
    [anon_sym_this] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_for] = ACTIONS(1942),
    [aux_sym__type_identifier_token1] = ACTIONS(1942),
    [anon_sym_if] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1940),
    [aux_sym_number_literal_token1] = ACTIONS(1942),
    [aux_sym_number_literal_token2] = ACTIONS(1940),
    [aux_sym_number_literal_token3] = ACTIONS(1940),
    [aux_sym_number_literal_token4] = ACTIONS(1942),
    [aux_sym_number_literal_token5] = ACTIONS(1940),
    [anon_sym_pipeline] = ACTIONS(1942),
    [anon_sym_return] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1942),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1940),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1940),
    [anon_sym_switch] = ACTIONS(1942),
    [anon_sym_try] = ACTIONS(1942),
    [anon_sym_catch] = ACTIONS(1942),
    [anon_sym_finally] = ACTIONS(1942),
    [anon_sym_int] = ACTIONS(1942),
    [anon_sym_boolean] = ACTIONS(1942),
    [anon_sym_char] = ACTIONS(1942),
    [anon_sym_short] = ACTIONS(1942),
    [anon_sym_long] = ACTIONS(1942),
    [anon_sym_float] = ACTIONS(1942),
    [anon_sym_double] = ACTIONS(1942),
    [anon_sym_void] = ACTIONS(1942),
    [anon_sym_public] = ACTIONS(1942),
    [anon_sym_protected] = ACTIONS(1942),
    [anon_sym_private] = ACTIONS(1942),
    [anon_sym_static] = ACTIONS(1942),
    [anon_sym_final] = ACTIONS(1942),
    [anon_sym_synchronized] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1940),
    [anon_sym_BANG] = ACTIONS(1940),
    [anon_sym_new] = ACTIONS(1942),
  },
  [STATE(731)] = {
    [sym_comment] = STATE(731),
    [sym_groovy_doc] = STATE(731),
    [ts_builtin_sym_end] = ACTIONS(1944),
    [sym_identifier] = ACTIONS(1946),
    [anon_sym_break] = ACTIONS(1946),
    [anon_sym_continue] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_STAR] = ACTIONS(1946),
    [anon_sym_STAR_COLON] = ACTIONS(1944),
    [anon_sym_import] = ACTIONS(1946),
    [anon_sym_package] = ACTIONS(1946),
    [anon_sym_AT] = ACTIONS(1946),
    [anon_sym_assert] = ACTIONS(1946),
    [anon_sym_PLUS_PLUS] = ACTIONS(1944),
    [anon_sym_DASH_DASH] = ACTIONS(1944),
    [anon_sym_SLASH] = ACTIONS(1946),
    [anon_sym_PLUS] = ACTIONS(1946),
    [anon_sym_DASH] = ACTIONS(1946),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_ATinterface] = ACTIONS(1944),
    [anon_sym_interface] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_LBRACE] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1946),
    [anon_sym_def] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_RPAREN] = ACTIONS(1944),
    [anon_sym_null] = ACTIONS(1946),
    [anon_sym_this] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_while] = ACTIONS(1946),
    [anon_sym_for] = ACTIONS(1946),
    [aux_sym__type_identifier_token1] = ACTIONS(1946),
    [anon_sym_if] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1944),
    [aux_sym_number_literal_token1] = ACTIONS(1946),
    [aux_sym_number_literal_token2] = ACTIONS(1944),
    [aux_sym_number_literal_token3] = ACTIONS(1944),
    [aux_sym_number_literal_token4] = ACTIONS(1946),
    [aux_sym_number_literal_token5] = ACTIONS(1944),
    [anon_sym_pipeline] = ACTIONS(1946),
    [anon_sym_return] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1946),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1944),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1944),
    [anon_sym_switch] = ACTIONS(1946),
    [anon_sym_try] = ACTIONS(1946),
    [anon_sym_catch] = ACTIONS(1946),
    [anon_sym_finally] = ACTIONS(1946),
    [anon_sym_int] = ACTIONS(1946),
    [anon_sym_boolean] = ACTIONS(1946),
    [anon_sym_char] = ACTIONS(1946),
    [anon_sym_short] = ACTIONS(1946),
    [anon_sym_long] = ACTIONS(1946),
    [anon_sym_float] = ACTIONS(1946),
    [anon_sym_double] = ACTIONS(1946),
    [anon_sym_void] = ACTIONS(1946),
    [anon_sym_public] = ACTIONS(1946),
    [anon_sym_protected] = ACTIONS(1946),
    [anon_sym_private] = ACTIONS(1946),
    [anon_sym_static] = ACTIONS(1946),
    [anon_sym_final] = ACTIONS(1946),
    [anon_sym_synchronized] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1944),
    [anon_sym_BANG] = ACTIONS(1944),
    [anon_sym_new] = ACTIONS(1946),
  },
  [STATE(732)] = {
    [sym_comment] = STATE(732),
    [sym_groovy_doc] = STATE(732),
    [ts_builtin_sym_end] = ACTIONS(1948),
    [sym_identifier] = ACTIONS(1950),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_STAR] = ACTIONS(1950),
    [anon_sym_STAR_COLON] = ACTIONS(1948),
    [anon_sym_import] = ACTIONS(1950),
    [anon_sym_package] = ACTIONS(1950),
    [anon_sym_AT] = ACTIONS(1950),
    [anon_sym_assert] = ACTIONS(1950),
    [anon_sym_PLUS_PLUS] = ACTIONS(1948),
    [anon_sym_DASH_DASH] = ACTIONS(1948),
    [anon_sym_SLASH] = ACTIONS(1950),
    [anon_sym_PLUS] = ACTIONS(1950),
    [anon_sym_DASH] = ACTIONS(1950),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_ATinterface] = ACTIONS(1948),
    [anon_sym_interface] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_LBRACE] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1950),
    [anon_sym_def] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_RPAREN] = ACTIONS(1948),
    [anon_sym_null] = ACTIONS(1950),
    [anon_sym_this] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_while] = ACTIONS(1950),
    [anon_sym_for] = ACTIONS(1950),
    [aux_sym__type_identifier_token1] = ACTIONS(1950),
    [anon_sym_if] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1948),
    [aux_sym_number_literal_token1] = ACTIONS(1950),
    [aux_sym_number_literal_token2] = ACTIONS(1948),
    [aux_sym_number_literal_token3] = ACTIONS(1948),
    [aux_sym_number_literal_token4] = ACTIONS(1950),
    [aux_sym_number_literal_token5] = ACTIONS(1948),
    [anon_sym_pipeline] = ACTIONS(1950),
    [anon_sym_return] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1950),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1948),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1948),
    [anon_sym_switch] = ACTIONS(1950),
    [anon_sym_try] = ACTIONS(1950),
    [anon_sym_catch] = ACTIONS(1950),
    [anon_sym_finally] = ACTIONS(1950),
    [anon_sym_int] = ACTIONS(1950),
    [anon_sym_boolean] = ACTIONS(1950),
    [anon_sym_char] = ACTIONS(1950),
    [anon_sym_short] = ACTIONS(1950),
    [anon_sym_long] = ACTIONS(1950),
    [anon_sym_float] = ACTIONS(1950),
    [anon_sym_double] = ACTIONS(1950),
    [anon_sym_void] = ACTIONS(1950),
    [anon_sym_public] = ACTIONS(1950),
    [anon_sym_protected] = ACTIONS(1950),
    [anon_sym_private] = ACTIONS(1950),
    [anon_sym_static] = ACTIONS(1950),
    [anon_sym_final] = ACTIONS(1950),
    [anon_sym_synchronized] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1948),
    [anon_sym_BANG] = ACTIONS(1948),
    [anon_sym_new] = ACTIONS(1950),
  },
  [STATE(733)] = {
    [sym_comment] = STATE(733),
    [sym_groovy_doc] = STATE(733),
    [ts_builtin_sym_end] = ACTIONS(1952),
    [sym_identifier] = ACTIONS(1954),
    [anon_sym_break] = ACTIONS(1954),
    [anon_sym_continue] = ACTIONS(1954),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_STAR] = ACTIONS(1954),
    [anon_sym_STAR_COLON] = ACTIONS(1952),
    [anon_sym_import] = ACTIONS(1954),
    [anon_sym_package] = ACTIONS(1954),
    [anon_sym_AT] = ACTIONS(1954),
    [anon_sym_assert] = ACTIONS(1954),
    [anon_sym_PLUS_PLUS] = ACTIONS(1952),
    [anon_sym_DASH_DASH] = ACTIONS(1952),
    [anon_sym_SLASH] = ACTIONS(1954),
    [anon_sym_PLUS] = ACTIONS(1954),
    [anon_sym_DASH] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(1954),
    [anon_sym_false] = ACTIONS(1954),
    [anon_sym_ATinterface] = ACTIONS(1952),
    [anon_sym_interface] = ACTIONS(1954),
    [anon_sym_class] = ACTIONS(1954),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_LBRACE] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1954),
    [anon_sym_def] = ACTIONS(1954),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_RPAREN] = ACTIONS(1952),
    [anon_sym_null] = ACTIONS(1954),
    [anon_sym_this] = ACTIONS(1954),
    [anon_sym_do] = ACTIONS(1954),
    [anon_sym_while] = ACTIONS(1954),
    [anon_sym_for] = ACTIONS(1954),
    [aux_sym__type_identifier_token1] = ACTIONS(1954),
    [anon_sym_if] = ACTIONS(1954),
    [anon_sym_else] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(1952),
    [aux_sym_number_literal_token1] = ACTIONS(1954),
    [aux_sym_number_literal_token2] = ACTIONS(1952),
    [aux_sym_number_literal_token3] = ACTIONS(1952),
    [aux_sym_number_literal_token4] = ACTIONS(1954),
    [aux_sym_number_literal_token5] = ACTIONS(1952),
    [anon_sym_pipeline] = ACTIONS(1954),
    [anon_sym_return] = ACTIONS(1954),
    [anon_sym_SQUOTE] = ACTIONS(1954),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1954),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1952),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1952),
    [anon_sym_switch] = ACTIONS(1954),
    [anon_sym_try] = ACTIONS(1954),
    [anon_sym_catch] = ACTIONS(1954),
    [anon_sym_finally] = ACTIONS(1954),
    [anon_sym_int] = ACTIONS(1954),
    [anon_sym_boolean] = ACTIONS(1954),
    [anon_sym_char] = ACTIONS(1954),
    [anon_sym_short] = ACTIONS(1954),
    [anon_sym_long] = ACTIONS(1954),
    [anon_sym_float] = ACTIONS(1954),
    [anon_sym_double] = ACTIONS(1954),
    [anon_sym_void] = ACTIONS(1954),
    [anon_sym_public] = ACTIONS(1954),
    [anon_sym_protected] = ACTIONS(1954),
    [anon_sym_private] = ACTIONS(1954),
    [anon_sym_static] = ACTIONS(1954),
    [anon_sym_final] = ACTIONS(1954),
    [anon_sym_synchronized] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(1952),
    [anon_sym_BANG] = ACTIONS(1952),
    [anon_sym_new] = ACTIONS(1954),
  },
  [STATE(734)] = {
    [sym_comment] = STATE(734),
    [sym_groovy_doc] = STATE(734),
    [ts_builtin_sym_end] = ACTIONS(1956),
    [sym_identifier] = ACTIONS(1958),
    [anon_sym_break] = ACTIONS(1958),
    [anon_sym_continue] = ACTIONS(1958),
    [anon_sym_SEMI] = ACTIONS(1956),
    [anon_sym_STAR] = ACTIONS(1958),
    [anon_sym_STAR_COLON] = ACTIONS(1956),
    [anon_sym_import] = ACTIONS(1958),
    [anon_sym_package] = ACTIONS(1958),
    [anon_sym_AT] = ACTIONS(1958),
    [anon_sym_assert] = ACTIONS(1958),
    [anon_sym_PLUS_PLUS] = ACTIONS(1956),
    [anon_sym_DASH_DASH] = ACTIONS(1956),
    [anon_sym_SLASH] = ACTIONS(1958),
    [anon_sym_PLUS] = ACTIONS(1958),
    [anon_sym_DASH] = ACTIONS(1958),
    [anon_sym_true] = ACTIONS(1958),
    [anon_sym_false] = ACTIONS(1958),
    [anon_sym_ATinterface] = ACTIONS(1956),
    [anon_sym_interface] = ACTIONS(1958),
    [anon_sym_class] = ACTIONS(1958),
    [anon_sym_COMMA] = ACTIONS(1956),
    [anon_sym_LBRACE] = ACTIONS(1956),
    [anon_sym_RBRACE] = ACTIONS(1956),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1958),
    [anon_sym_def] = ACTIONS(1958),
    [anon_sym_LPAREN] = ACTIONS(1956),
    [anon_sym_RPAREN] = ACTIONS(1956),
    [anon_sym_null] = ACTIONS(1958),
    [anon_sym_this] = ACTIONS(1958),
    [anon_sym_do] = ACTIONS(1958),
    [anon_sym_while] = ACTIONS(1958),
    [anon_sym_for] = ACTIONS(1958),
    [aux_sym__type_identifier_token1] = ACTIONS(1958),
    [anon_sym_if] = ACTIONS(1958),
    [anon_sym_else] = ACTIONS(1958),
    [anon_sym_LBRACK] = ACTIONS(1956),
    [aux_sym_number_literal_token1] = ACTIONS(1958),
    [aux_sym_number_literal_token2] = ACTIONS(1956),
    [aux_sym_number_literal_token3] = ACTIONS(1956),
    [aux_sym_number_literal_token4] = ACTIONS(1958),
    [aux_sym_number_literal_token5] = ACTIONS(1956),
    [anon_sym_pipeline] = ACTIONS(1958),
    [anon_sym_return] = ACTIONS(1958),
    [anon_sym_SQUOTE] = ACTIONS(1958),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1956),
    [anon_sym_DQUOTE] = ACTIONS(1958),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1956),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1956),
    [anon_sym_switch] = ACTIONS(1958),
    [anon_sym_try] = ACTIONS(1958),
    [anon_sym_catch] = ACTIONS(1958),
    [anon_sym_finally] = ACTIONS(1958),
    [anon_sym_int] = ACTIONS(1958),
    [anon_sym_boolean] = ACTIONS(1958),
    [anon_sym_char] = ACTIONS(1958),
    [anon_sym_short] = ACTIONS(1958),
    [anon_sym_long] = ACTIONS(1958),
    [anon_sym_float] = ACTIONS(1958),
    [anon_sym_double] = ACTIONS(1958),
    [anon_sym_void] = ACTIONS(1958),
    [anon_sym_public] = ACTIONS(1958),
    [anon_sym_protected] = ACTIONS(1958),
    [anon_sym_private] = ACTIONS(1958),
    [anon_sym_static] = ACTIONS(1958),
    [anon_sym_final] = ACTIONS(1958),
    [anon_sym_synchronized] = ACTIONS(1958),
    [anon_sym_TILDE] = ACTIONS(1956),
    [anon_sym_BANG] = ACTIONS(1956),
    [anon_sym_new] = ACTIONS(1958),
  },
  [STATE(735)] = {
    [sym_comment] = STATE(735),
    [sym_groovy_doc] = STATE(735),
    [ts_builtin_sym_end] = ACTIONS(1960),
    [sym_identifier] = ACTIONS(1962),
    [anon_sym_break] = ACTIONS(1962),
    [anon_sym_continue] = ACTIONS(1962),
    [anon_sym_SEMI] = ACTIONS(1960),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_STAR_COLON] = ACTIONS(1960),
    [anon_sym_import] = ACTIONS(1962),
    [anon_sym_package] = ACTIONS(1962),
    [anon_sym_AT] = ACTIONS(1962),
    [anon_sym_assert] = ACTIONS(1962),
    [anon_sym_PLUS_PLUS] = ACTIONS(1960),
    [anon_sym_DASH_DASH] = ACTIONS(1960),
    [anon_sym_SLASH] = ACTIONS(1962),
    [anon_sym_PLUS] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_true] = ACTIONS(1962),
    [anon_sym_false] = ACTIONS(1962),
    [anon_sym_ATinterface] = ACTIONS(1960),
    [anon_sym_interface] = ACTIONS(1962),
    [anon_sym_class] = ACTIONS(1962),
    [anon_sym_COMMA] = ACTIONS(1960),
    [anon_sym_LBRACE] = ACTIONS(1960),
    [anon_sym_RBRACE] = ACTIONS(1960),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1962),
    [anon_sym_def] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1960),
    [anon_sym_RPAREN] = ACTIONS(1960),
    [anon_sym_null] = ACTIONS(1962),
    [anon_sym_this] = ACTIONS(1962),
    [anon_sym_do] = ACTIONS(1962),
    [anon_sym_while] = ACTIONS(1962),
    [anon_sym_for] = ACTIONS(1962),
    [aux_sym__type_identifier_token1] = ACTIONS(1962),
    [anon_sym_if] = ACTIONS(1962),
    [anon_sym_else] = ACTIONS(1962),
    [anon_sym_LBRACK] = ACTIONS(1960),
    [aux_sym_number_literal_token1] = ACTIONS(1962),
    [aux_sym_number_literal_token2] = ACTIONS(1960),
    [aux_sym_number_literal_token3] = ACTIONS(1960),
    [aux_sym_number_literal_token4] = ACTIONS(1962),
    [aux_sym_number_literal_token5] = ACTIONS(1960),
    [anon_sym_pipeline] = ACTIONS(1962),
    [anon_sym_return] = ACTIONS(1962),
    [anon_sym_SQUOTE] = ACTIONS(1962),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1960),
    [anon_sym_DQUOTE] = ACTIONS(1962),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1960),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1960),
    [anon_sym_switch] = ACTIONS(1962),
    [anon_sym_try] = ACTIONS(1962),
    [anon_sym_catch] = ACTIONS(1962),
    [anon_sym_finally] = ACTIONS(1962),
    [anon_sym_int] = ACTIONS(1962),
    [anon_sym_boolean] = ACTIONS(1962),
    [anon_sym_char] = ACTIONS(1962),
    [anon_sym_short] = ACTIONS(1962),
    [anon_sym_long] = ACTIONS(1962),
    [anon_sym_float] = ACTIONS(1962),
    [anon_sym_double] = ACTIONS(1962),
    [anon_sym_void] = ACTIONS(1962),
    [anon_sym_public] = ACTIONS(1962),
    [anon_sym_protected] = ACTIONS(1962),
    [anon_sym_private] = ACTIONS(1962),
    [anon_sym_static] = ACTIONS(1962),
    [anon_sym_final] = ACTIONS(1962),
    [anon_sym_synchronized] = ACTIONS(1962),
    [anon_sym_TILDE] = ACTIONS(1960),
    [anon_sym_BANG] = ACTIONS(1960),
    [anon_sym_new] = ACTIONS(1962),
  },
  [STATE(736)] = {
    [sym_comment] = STATE(736),
    [sym_groovy_doc] = STATE(736),
    [sym_parameter_list] = STATE(766),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(1964),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_case] = ACTIONS(1773),
    [anon_sym_default] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_catch] = ACTIONS(1773),
    [anon_sym_finally] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(737)] = {
    [sym_comment] = STATE(737),
    [sym_groovy_doc] = STATE(737),
    [sym_parameter_list] = STATE(766),
    [sym_identifier] = ACTIONS(1792),
    [anon_sym_break] = ACTIONS(1792),
    [anon_sym_continue] = ACTIONS(1792),
    [anon_sym_SEMI] = ACTIONS(1790),
    [anon_sym_STAR] = ACTIONS(1792),
    [anon_sym_STAR_COLON] = ACTIONS(1790),
    [anon_sym_import] = ACTIONS(1792),
    [anon_sym_package] = ACTIONS(1792),
    [anon_sym_AT] = ACTIONS(1792),
    [anon_sym_assert] = ACTIONS(1792),
    [anon_sym_EQ] = ACTIONS(1966),
    [anon_sym_PLUS_PLUS] = ACTIONS(1790),
    [anon_sym_DASH_DASH] = ACTIONS(1790),
    [anon_sym_SLASH] = ACTIONS(1792),
    [anon_sym_PLUS] = ACTIONS(1792),
    [anon_sym_DASH] = ACTIONS(1792),
    [anon_sym_true] = ACTIONS(1792),
    [anon_sym_false] = ACTIONS(1792),
    [anon_sym_ATinterface] = ACTIONS(1790),
    [anon_sym_interface] = ACTIONS(1792),
    [anon_sym_class] = ACTIONS(1792),
    [anon_sym_LBRACE] = ACTIONS(1790),
    [anon_sym_RBRACE] = ACTIONS(1790),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1792),
    [anon_sym_def] = ACTIONS(1792),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1792),
    [anon_sym_this] = ACTIONS(1792),
    [anon_sym_do] = ACTIONS(1792),
    [anon_sym_while] = ACTIONS(1792),
    [anon_sym_for] = ACTIONS(1792),
    [aux_sym__type_identifier_token1] = ACTIONS(1792),
    [anon_sym_if] = ACTIONS(1792),
    [anon_sym_else] = ACTIONS(1792),
    [anon_sym_LBRACK] = ACTIONS(1790),
    [aux_sym_number_literal_token1] = ACTIONS(1792),
    [aux_sym_number_literal_token2] = ACTIONS(1790),
    [aux_sym_number_literal_token3] = ACTIONS(1790),
    [aux_sym_number_literal_token4] = ACTIONS(1792),
    [aux_sym_number_literal_token5] = ACTIONS(1790),
    [anon_sym_return] = ACTIONS(1792),
    [anon_sym_SQUOTE] = ACTIONS(1792),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1790),
    [anon_sym_DQUOTE] = ACTIONS(1792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1790),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1790),
    [anon_sym_switch] = ACTIONS(1792),
    [anon_sym_case] = ACTIONS(1792),
    [anon_sym_default] = ACTIONS(1792),
    [anon_sym_try] = ACTIONS(1792),
    [anon_sym_catch] = ACTIONS(1792),
    [anon_sym_finally] = ACTIONS(1792),
    [anon_sym_int] = ACTIONS(1792),
    [anon_sym_boolean] = ACTIONS(1792),
    [anon_sym_char] = ACTIONS(1792),
    [anon_sym_short] = ACTIONS(1792),
    [anon_sym_long] = ACTIONS(1792),
    [anon_sym_float] = ACTIONS(1792),
    [anon_sym_double] = ACTIONS(1792),
    [anon_sym_void] = ACTIONS(1792),
    [anon_sym_public] = ACTIONS(1792),
    [anon_sym_protected] = ACTIONS(1792),
    [anon_sym_private] = ACTIONS(1792),
    [anon_sym_static] = ACTIONS(1792),
    [anon_sym_final] = ACTIONS(1792),
    [anon_sym_synchronized] = ACTIONS(1792),
    [anon_sym_TILDE] = ACTIONS(1790),
    [anon_sym_BANG] = ACTIONS(1790),
    [anon_sym_new] = ACTIONS(1792),
  },
  [STATE(738)] = {
    [sym_comment] = STATE(738),
    [sym_groovy_doc] = STATE(738),
    [ts_builtin_sym_end] = ACTIONS(1968),
    [sym_identifier] = ACTIONS(1970),
    [anon_sym_break] = ACTIONS(1970),
    [anon_sym_continue] = ACTIONS(1970),
    [anon_sym_SEMI] = ACTIONS(1968),
    [anon_sym_STAR] = ACTIONS(1970),
    [anon_sym_STAR_COLON] = ACTIONS(1968),
    [anon_sym_import] = ACTIONS(1970),
    [anon_sym_package] = ACTIONS(1970),
    [anon_sym_AT] = ACTIONS(1970),
    [anon_sym_assert] = ACTIONS(1970),
    [anon_sym_PLUS_PLUS] = ACTIONS(1968),
    [anon_sym_DASH_DASH] = ACTIONS(1968),
    [anon_sym_SLASH] = ACTIONS(1970),
    [anon_sym_PLUS] = ACTIONS(1970),
    [anon_sym_DASH] = ACTIONS(1970),
    [anon_sym_true] = ACTIONS(1970),
    [anon_sym_false] = ACTIONS(1970),
    [anon_sym_ATinterface] = ACTIONS(1968),
    [anon_sym_interface] = ACTIONS(1970),
    [anon_sym_class] = ACTIONS(1970),
    [anon_sym_COMMA] = ACTIONS(1968),
    [anon_sym_LBRACE] = ACTIONS(1968),
    [anon_sym_RBRACE] = ACTIONS(1968),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1970),
    [anon_sym_def] = ACTIONS(1970),
    [anon_sym_LPAREN] = ACTIONS(1968),
    [anon_sym_RPAREN] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_this] = ACTIONS(1970),
    [anon_sym_do] = ACTIONS(1970),
    [anon_sym_while] = ACTIONS(1970),
    [anon_sym_for] = ACTIONS(1970),
    [aux_sym__type_identifier_token1] = ACTIONS(1970),
    [anon_sym_if] = ACTIONS(1970),
    [anon_sym_else] = ACTIONS(1970),
    [anon_sym_LBRACK] = ACTIONS(1968),
    [aux_sym_number_literal_token1] = ACTIONS(1970),
    [aux_sym_number_literal_token2] = ACTIONS(1968),
    [aux_sym_number_literal_token3] = ACTIONS(1968),
    [aux_sym_number_literal_token4] = ACTIONS(1970),
    [aux_sym_number_literal_token5] = ACTIONS(1968),
    [anon_sym_pipeline] = ACTIONS(1970),
    [anon_sym_return] = ACTIONS(1970),
    [anon_sym_SQUOTE] = ACTIONS(1970),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1968),
    [anon_sym_DQUOTE] = ACTIONS(1970),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1968),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1968),
    [anon_sym_switch] = ACTIONS(1970),
    [anon_sym_try] = ACTIONS(1970),
    [anon_sym_catch] = ACTIONS(1970),
    [anon_sym_finally] = ACTIONS(1970),
    [anon_sym_int] = ACTIONS(1970),
    [anon_sym_boolean] = ACTIONS(1970),
    [anon_sym_char] = ACTIONS(1970),
    [anon_sym_short] = ACTIONS(1970),
    [anon_sym_long] = ACTIONS(1970),
    [anon_sym_float] = ACTIONS(1970),
    [anon_sym_double] = ACTIONS(1970),
    [anon_sym_void] = ACTIONS(1970),
    [anon_sym_public] = ACTIONS(1970),
    [anon_sym_protected] = ACTIONS(1970),
    [anon_sym_private] = ACTIONS(1970),
    [anon_sym_static] = ACTIONS(1970),
    [anon_sym_final] = ACTIONS(1970),
    [anon_sym_synchronized] = ACTIONS(1970),
    [anon_sym_TILDE] = ACTIONS(1968),
    [anon_sym_BANG] = ACTIONS(1968),
    [anon_sym_new] = ACTIONS(1970),
  },
  [STATE(739)] = {
    [sym_comment] = STATE(739),
    [sym_groovy_doc] = STATE(739),
    [ts_builtin_sym_end] = ACTIONS(1972),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_break] = ACTIONS(1974),
    [anon_sym_continue] = ACTIONS(1974),
    [anon_sym_SEMI] = ACTIONS(1972),
    [anon_sym_STAR] = ACTIONS(1974),
    [anon_sym_STAR_COLON] = ACTIONS(1972),
    [anon_sym_import] = ACTIONS(1974),
    [anon_sym_package] = ACTIONS(1974),
    [anon_sym_AT] = ACTIONS(1974),
    [anon_sym_assert] = ACTIONS(1974),
    [anon_sym_PLUS_PLUS] = ACTIONS(1972),
    [anon_sym_DASH_DASH] = ACTIONS(1972),
    [anon_sym_SLASH] = ACTIONS(1974),
    [anon_sym_PLUS] = ACTIONS(1974),
    [anon_sym_DASH] = ACTIONS(1974),
    [anon_sym_true] = ACTIONS(1974),
    [anon_sym_false] = ACTIONS(1974),
    [anon_sym_ATinterface] = ACTIONS(1972),
    [anon_sym_interface] = ACTIONS(1974),
    [anon_sym_class] = ACTIONS(1974),
    [anon_sym_COMMA] = ACTIONS(1972),
    [anon_sym_LBRACE] = ACTIONS(1972),
    [anon_sym_RBRACE] = ACTIONS(1972),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1974),
    [anon_sym_def] = ACTIONS(1974),
    [anon_sym_LPAREN] = ACTIONS(1972),
    [anon_sym_RPAREN] = ACTIONS(1972),
    [anon_sym_null] = ACTIONS(1974),
    [anon_sym_this] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1974),
    [anon_sym_while] = ACTIONS(1974),
    [anon_sym_for] = ACTIONS(1974),
    [aux_sym__type_identifier_token1] = ACTIONS(1974),
    [anon_sym_if] = ACTIONS(1974),
    [anon_sym_else] = ACTIONS(1974),
    [anon_sym_LBRACK] = ACTIONS(1972),
    [aux_sym_number_literal_token1] = ACTIONS(1974),
    [aux_sym_number_literal_token2] = ACTIONS(1972),
    [aux_sym_number_literal_token3] = ACTIONS(1972),
    [aux_sym_number_literal_token4] = ACTIONS(1974),
    [aux_sym_number_literal_token5] = ACTIONS(1972),
    [anon_sym_pipeline] = ACTIONS(1974),
    [anon_sym_return] = ACTIONS(1974),
    [anon_sym_SQUOTE] = ACTIONS(1974),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1972),
    [anon_sym_DQUOTE] = ACTIONS(1974),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1972),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1972),
    [anon_sym_switch] = ACTIONS(1974),
    [anon_sym_try] = ACTIONS(1974),
    [anon_sym_catch] = ACTIONS(1974),
    [anon_sym_finally] = ACTIONS(1974),
    [anon_sym_int] = ACTIONS(1974),
    [anon_sym_boolean] = ACTIONS(1974),
    [anon_sym_char] = ACTIONS(1974),
    [anon_sym_short] = ACTIONS(1974),
    [anon_sym_long] = ACTIONS(1974),
    [anon_sym_float] = ACTIONS(1974),
    [anon_sym_double] = ACTIONS(1974),
    [anon_sym_void] = ACTIONS(1974),
    [anon_sym_public] = ACTIONS(1974),
    [anon_sym_protected] = ACTIONS(1974),
    [anon_sym_private] = ACTIONS(1974),
    [anon_sym_static] = ACTIONS(1974),
    [anon_sym_final] = ACTIONS(1974),
    [anon_sym_synchronized] = ACTIONS(1974),
    [anon_sym_TILDE] = ACTIONS(1972),
    [anon_sym_BANG] = ACTIONS(1972),
    [anon_sym_new] = ACTIONS(1974),
  },
  [STATE(740)] = {
    [sym_comment] = STATE(740),
    [sym_groovy_doc] = STATE(740),
    [aux_sym__juxt_argument_list_repeat1] = STATE(740),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1976),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(741)] = {
    [sym_comment] = STATE(741),
    [sym_groovy_doc] = STATE(741),
    [sym_parameter_list] = STATE(774),
    [ts_builtin_sym_end] = ACTIONS(1979),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(1983),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_pipeline] = ACTIONS(1981),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_catch] = ACTIONS(1981),
    [anon_sym_finally] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(742)] = {
    [sym_comment] = STATE(742),
    [sym_groovy_doc] = STATE(742),
    [ts_builtin_sym_end] = ACTIONS(1985),
    [sym_identifier] = ACTIONS(1987),
    [anon_sym_break] = ACTIONS(1987),
    [anon_sym_continue] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_STAR] = ACTIONS(1987),
    [anon_sym_STAR_COLON] = ACTIONS(1985),
    [anon_sym_import] = ACTIONS(1987),
    [anon_sym_package] = ACTIONS(1987),
    [anon_sym_AT] = ACTIONS(1987),
    [anon_sym_assert] = ACTIONS(1987),
    [anon_sym_PLUS_PLUS] = ACTIONS(1985),
    [anon_sym_DASH_DASH] = ACTIONS(1985),
    [anon_sym_SLASH] = ACTIONS(1987),
    [anon_sym_PLUS] = ACTIONS(1987),
    [anon_sym_DASH] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_ATinterface] = ACTIONS(1985),
    [anon_sym_interface] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_LBRACE] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1987),
    [anon_sym_def] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_RPAREN] = ACTIONS(1985),
    [anon_sym_null] = ACTIONS(1987),
    [anon_sym_this] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_while] = ACTIONS(1987),
    [anon_sym_for] = ACTIONS(1987),
    [aux_sym__type_identifier_token1] = ACTIONS(1987),
    [anon_sym_if] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1985),
    [aux_sym_number_literal_token1] = ACTIONS(1987),
    [aux_sym_number_literal_token2] = ACTIONS(1985),
    [aux_sym_number_literal_token3] = ACTIONS(1985),
    [aux_sym_number_literal_token4] = ACTIONS(1987),
    [aux_sym_number_literal_token5] = ACTIONS(1985),
    [anon_sym_pipeline] = ACTIONS(1987),
    [anon_sym_return] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1987),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1987),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1985),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1985),
    [anon_sym_switch] = ACTIONS(1987),
    [anon_sym_try] = ACTIONS(1987),
    [anon_sym_catch] = ACTIONS(1987),
    [anon_sym_finally] = ACTIONS(1987),
    [anon_sym_int] = ACTIONS(1987),
    [anon_sym_boolean] = ACTIONS(1987),
    [anon_sym_char] = ACTIONS(1987),
    [anon_sym_short] = ACTIONS(1987),
    [anon_sym_long] = ACTIONS(1987),
    [anon_sym_float] = ACTIONS(1987),
    [anon_sym_double] = ACTIONS(1987),
    [anon_sym_void] = ACTIONS(1987),
    [anon_sym_public] = ACTIONS(1987),
    [anon_sym_protected] = ACTIONS(1987),
    [anon_sym_private] = ACTIONS(1987),
    [anon_sym_static] = ACTIONS(1987),
    [anon_sym_final] = ACTIONS(1987),
    [anon_sym_synchronized] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1985),
    [anon_sym_BANG] = ACTIONS(1985),
    [anon_sym_new] = ACTIONS(1987),
  },
  [STATE(743)] = {
    [sym_comment] = STATE(743),
    [sym_groovy_doc] = STATE(743),
    [ts_builtin_sym_end] = ACTIONS(1989),
    [sym_identifier] = ACTIONS(1991),
    [anon_sym_break] = ACTIONS(1991),
    [anon_sym_continue] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_STAR] = ACTIONS(1991),
    [anon_sym_STAR_COLON] = ACTIONS(1989),
    [anon_sym_import] = ACTIONS(1991),
    [anon_sym_package] = ACTIONS(1991),
    [anon_sym_AT] = ACTIONS(1991),
    [anon_sym_assert] = ACTIONS(1991),
    [anon_sym_PLUS_PLUS] = ACTIONS(1989),
    [anon_sym_DASH_DASH] = ACTIONS(1989),
    [anon_sym_SLASH] = ACTIONS(1991),
    [anon_sym_PLUS] = ACTIONS(1991),
    [anon_sym_DASH] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_ATinterface] = ACTIONS(1989),
    [anon_sym_interface] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1991),
    [anon_sym_def] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_RPAREN] = ACTIONS(1989),
    [anon_sym_null] = ACTIONS(1991),
    [anon_sym_this] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_while] = ACTIONS(1991),
    [anon_sym_for] = ACTIONS(1991),
    [aux_sym__type_identifier_token1] = ACTIONS(1991),
    [anon_sym_if] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1989),
    [aux_sym_number_literal_token1] = ACTIONS(1991),
    [aux_sym_number_literal_token2] = ACTIONS(1989),
    [aux_sym_number_literal_token3] = ACTIONS(1989),
    [aux_sym_number_literal_token4] = ACTIONS(1991),
    [aux_sym_number_literal_token5] = ACTIONS(1989),
    [anon_sym_pipeline] = ACTIONS(1991),
    [anon_sym_return] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1991),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1991),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1989),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1989),
    [anon_sym_switch] = ACTIONS(1991),
    [anon_sym_try] = ACTIONS(1991),
    [anon_sym_catch] = ACTIONS(1991),
    [anon_sym_finally] = ACTIONS(1991),
    [anon_sym_int] = ACTIONS(1991),
    [anon_sym_boolean] = ACTIONS(1991),
    [anon_sym_char] = ACTIONS(1991),
    [anon_sym_short] = ACTIONS(1991),
    [anon_sym_long] = ACTIONS(1991),
    [anon_sym_float] = ACTIONS(1991),
    [anon_sym_double] = ACTIONS(1991),
    [anon_sym_void] = ACTIONS(1991),
    [anon_sym_public] = ACTIONS(1991),
    [anon_sym_protected] = ACTIONS(1991),
    [anon_sym_private] = ACTIONS(1991),
    [anon_sym_static] = ACTIONS(1991),
    [anon_sym_final] = ACTIONS(1991),
    [anon_sym_synchronized] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1989),
    [anon_sym_BANG] = ACTIONS(1989),
    [anon_sym_new] = ACTIONS(1991),
  },
  [STATE(744)] = {
    [sym_comment] = STATE(744),
    [sym_groovy_doc] = STATE(744),
    [aux_sym__juxt_argument_list_repeat1] = STATE(752),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_catch] = ACTIONS(1254),
    [anon_sym_finally] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(745)] = {
    [sym_comment] = STATE(745),
    [sym_groovy_doc] = STATE(745),
    [ts_builtin_sym_end] = ACTIONS(1993),
    [sym_identifier] = ACTIONS(1995),
    [anon_sym_break] = ACTIONS(1995),
    [anon_sym_continue] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_STAR] = ACTIONS(1995),
    [anon_sym_STAR_COLON] = ACTIONS(1993),
    [anon_sym_import] = ACTIONS(1995),
    [anon_sym_package] = ACTIONS(1995),
    [anon_sym_AT] = ACTIONS(1995),
    [anon_sym_assert] = ACTIONS(1995),
    [anon_sym_PLUS_PLUS] = ACTIONS(1993),
    [anon_sym_DASH_DASH] = ACTIONS(1993),
    [anon_sym_SLASH] = ACTIONS(1995),
    [anon_sym_PLUS] = ACTIONS(1995),
    [anon_sym_DASH] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_ATinterface] = ACTIONS(1993),
    [anon_sym_interface] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_LBRACE] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1995),
    [anon_sym_def] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_RPAREN] = ACTIONS(1993),
    [anon_sym_null] = ACTIONS(1995),
    [anon_sym_this] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_while] = ACTIONS(1995),
    [anon_sym_for] = ACTIONS(1995),
    [aux_sym__type_identifier_token1] = ACTIONS(1995),
    [anon_sym_if] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1993),
    [aux_sym_number_literal_token1] = ACTIONS(1995),
    [aux_sym_number_literal_token2] = ACTIONS(1993),
    [aux_sym_number_literal_token3] = ACTIONS(1993),
    [aux_sym_number_literal_token4] = ACTIONS(1995),
    [aux_sym_number_literal_token5] = ACTIONS(1993),
    [anon_sym_pipeline] = ACTIONS(1995),
    [anon_sym_return] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1995),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1995),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1993),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1993),
    [anon_sym_switch] = ACTIONS(1995),
    [anon_sym_try] = ACTIONS(1995),
    [anon_sym_catch] = ACTIONS(1995),
    [anon_sym_finally] = ACTIONS(1995),
    [anon_sym_int] = ACTIONS(1995),
    [anon_sym_boolean] = ACTIONS(1995),
    [anon_sym_char] = ACTIONS(1995),
    [anon_sym_short] = ACTIONS(1995),
    [anon_sym_long] = ACTIONS(1995),
    [anon_sym_float] = ACTIONS(1995),
    [anon_sym_double] = ACTIONS(1995),
    [anon_sym_void] = ACTIONS(1995),
    [anon_sym_public] = ACTIONS(1995),
    [anon_sym_protected] = ACTIONS(1995),
    [anon_sym_private] = ACTIONS(1995),
    [anon_sym_static] = ACTIONS(1995),
    [anon_sym_final] = ACTIONS(1995),
    [anon_sym_synchronized] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1993),
    [anon_sym_BANG] = ACTIONS(1993),
    [anon_sym_new] = ACTIONS(1995),
  },
  [STATE(746)] = {
    [sym_comment] = STATE(746),
    [sym_groovy_doc] = STATE(746),
    [sym_parameter_list] = STATE(773),
    [ts_builtin_sym_end] = ACTIONS(1892),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1890),
    [anon_sym_STAR_COLON] = ACTIONS(1892),
    [anon_sym_import] = ACTIONS(1890),
    [anon_sym_package] = ACTIONS(1890),
    [anon_sym_AT] = ACTIONS(1890),
    [anon_sym_assert] = ACTIONS(1890),
    [anon_sym_EQ] = ACTIONS(1997),
    [anon_sym_PLUS_PLUS] = ACTIONS(1892),
    [anon_sym_DASH_DASH] = ACTIONS(1892),
    [anon_sym_SLASH] = ACTIONS(1890),
    [anon_sym_PLUS] = ACTIONS(1890),
    [anon_sym_DASH] = ACTIONS(1890),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_ATinterface] = ACTIONS(1892),
    [anon_sym_interface] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1890),
    [anon_sym_def] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(1890),
    [anon_sym_this] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [aux_sym__type_identifier_token1] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [aux_sym_number_literal_token1] = ACTIONS(1890),
    [aux_sym_number_literal_token2] = ACTIONS(1892),
    [aux_sym_number_literal_token3] = ACTIONS(1892),
    [aux_sym_number_literal_token4] = ACTIONS(1890),
    [aux_sym_number_literal_token5] = ACTIONS(1892),
    [anon_sym_pipeline] = ACTIONS(1890),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1892),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1892),
    [anon_sym_switch] = ACTIONS(1890),
    [anon_sym_try] = ACTIONS(1890),
    [anon_sym_catch] = ACTIONS(1890),
    [anon_sym_finally] = ACTIONS(1890),
    [anon_sym_int] = ACTIONS(1890),
    [anon_sym_boolean] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_short] = ACTIONS(1890),
    [anon_sym_long] = ACTIONS(1890),
    [anon_sym_float] = ACTIONS(1890),
    [anon_sym_double] = ACTIONS(1890),
    [anon_sym_void] = ACTIONS(1890),
    [anon_sym_public] = ACTIONS(1890),
    [anon_sym_protected] = ACTIONS(1890),
    [anon_sym_private] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_final] = ACTIONS(1890),
    [anon_sym_synchronized] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_new] = ACTIONS(1890),
  },
  [STATE(747)] = {
    [sym_comment] = STATE(747),
    [sym_groovy_doc] = STATE(747),
    [ts_builtin_sym_end] = ACTIONS(1999),
    [sym_identifier] = ACTIONS(2001),
    [anon_sym_break] = ACTIONS(2001),
    [anon_sym_continue] = ACTIONS(2001),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_STAR] = ACTIONS(2001),
    [anon_sym_STAR_COLON] = ACTIONS(1999),
    [anon_sym_import] = ACTIONS(2001),
    [anon_sym_package] = ACTIONS(2001),
    [anon_sym_AT] = ACTIONS(2001),
    [anon_sym_assert] = ACTIONS(2001),
    [anon_sym_PLUS_PLUS] = ACTIONS(1999),
    [anon_sym_DASH_DASH] = ACTIONS(1999),
    [anon_sym_SLASH] = ACTIONS(2001),
    [anon_sym_PLUS] = ACTIONS(2001),
    [anon_sym_DASH] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(2001),
    [anon_sym_false] = ACTIONS(2001),
    [anon_sym_ATinterface] = ACTIONS(1999),
    [anon_sym_interface] = ACTIONS(2001),
    [anon_sym_class] = ACTIONS(2001),
    [anon_sym_COMMA] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_RBRACE] = ACTIONS(1999),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2001),
    [anon_sym_def] = ACTIONS(2001),
    [anon_sym_LPAREN] = ACTIONS(1999),
    [anon_sym_RPAREN] = ACTIONS(1999),
    [anon_sym_null] = ACTIONS(2001),
    [anon_sym_this] = ACTIONS(2001),
    [anon_sym_do] = ACTIONS(2001),
    [anon_sym_while] = ACTIONS(2001),
    [anon_sym_for] = ACTIONS(2001),
    [aux_sym__type_identifier_token1] = ACTIONS(2001),
    [anon_sym_if] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [aux_sym_number_literal_token1] = ACTIONS(2001),
    [aux_sym_number_literal_token2] = ACTIONS(1999),
    [aux_sym_number_literal_token3] = ACTIONS(1999),
    [aux_sym_number_literal_token4] = ACTIONS(2001),
    [aux_sym_number_literal_token5] = ACTIONS(1999),
    [anon_sym_pipeline] = ACTIONS(2001),
    [anon_sym_return] = ACTIONS(2001),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1999),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1999),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1999),
    [anon_sym_switch] = ACTIONS(2001),
    [anon_sym_try] = ACTIONS(2001),
    [anon_sym_catch] = ACTIONS(2001),
    [anon_sym_finally] = ACTIONS(2001),
    [anon_sym_int] = ACTIONS(2001),
    [anon_sym_boolean] = ACTIONS(2001),
    [anon_sym_char] = ACTIONS(2001),
    [anon_sym_short] = ACTIONS(2001),
    [anon_sym_long] = ACTIONS(2001),
    [anon_sym_float] = ACTIONS(2001),
    [anon_sym_double] = ACTIONS(2001),
    [anon_sym_void] = ACTIONS(2001),
    [anon_sym_public] = ACTIONS(2001),
    [anon_sym_protected] = ACTIONS(2001),
    [anon_sym_private] = ACTIONS(2001),
    [anon_sym_static] = ACTIONS(2001),
    [anon_sym_final] = ACTIONS(2001),
    [anon_sym_synchronized] = ACTIONS(2001),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_BANG] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(2001),
  },
  [STATE(748)] = {
    [sym_comment] = STATE(748),
    [sym_groovy_doc] = STATE(748),
    [ts_builtin_sym_end] = ACTIONS(2003),
    [sym_identifier] = ACTIONS(2005),
    [anon_sym_break] = ACTIONS(2005),
    [anon_sym_continue] = ACTIONS(2005),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_STAR] = ACTIONS(2005),
    [anon_sym_STAR_COLON] = ACTIONS(2003),
    [anon_sym_import] = ACTIONS(2005),
    [anon_sym_package] = ACTIONS(2005),
    [anon_sym_AT] = ACTIONS(2005),
    [anon_sym_assert] = ACTIONS(2005),
    [anon_sym_PLUS_PLUS] = ACTIONS(2003),
    [anon_sym_DASH_DASH] = ACTIONS(2003),
    [anon_sym_SLASH] = ACTIONS(2005),
    [anon_sym_PLUS] = ACTIONS(2005),
    [anon_sym_DASH] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2005),
    [anon_sym_false] = ACTIONS(2005),
    [anon_sym_ATinterface] = ACTIONS(2003),
    [anon_sym_interface] = ACTIONS(2005),
    [anon_sym_class] = ACTIONS(2005),
    [anon_sym_COMMA] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_RBRACE] = ACTIONS(2003),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2005),
    [anon_sym_def] = ACTIONS(2005),
    [anon_sym_LPAREN] = ACTIONS(2003),
    [anon_sym_RPAREN] = ACTIONS(2003),
    [anon_sym_null] = ACTIONS(2005),
    [anon_sym_this] = ACTIONS(2005),
    [anon_sym_do] = ACTIONS(2005),
    [anon_sym_while] = ACTIONS(2005),
    [anon_sym_for] = ACTIONS(2005),
    [aux_sym__type_identifier_token1] = ACTIONS(2005),
    [anon_sym_if] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [aux_sym_number_literal_token1] = ACTIONS(2005),
    [aux_sym_number_literal_token2] = ACTIONS(2003),
    [aux_sym_number_literal_token3] = ACTIONS(2003),
    [aux_sym_number_literal_token4] = ACTIONS(2005),
    [aux_sym_number_literal_token5] = ACTIONS(2003),
    [anon_sym_pipeline] = ACTIONS(2005),
    [anon_sym_return] = ACTIONS(2005),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2003),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2003),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2003),
    [anon_sym_switch] = ACTIONS(2005),
    [anon_sym_try] = ACTIONS(2005),
    [anon_sym_catch] = ACTIONS(2005),
    [anon_sym_finally] = ACTIONS(2005),
    [anon_sym_int] = ACTIONS(2005),
    [anon_sym_boolean] = ACTIONS(2005),
    [anon_sym_char] = ACTIONS(2005),
    [anon_sym_short] = ACTIONS(2005),
    [anon_sym_long] = ACTIONS(2005),
    [anon_sym_float] = ACTIONS(2005),
    [anon_sym_double] = ACTIONS(2005),
    [anon_sym_void] = ACTIONS(2005),
    [anon_sym_public] = ACTIONS(2005),
    [anon_sym_protected] = ACTIONS(2005),
    [anon_sym_private] = ACTIONS(2005),
    [anon_sym_static] = ACTIONS(2005),
    [anon_sym_final] = ACTIONS(2005),
    [anon_sym_synchronized] = ACTIONS(2005),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_BANG] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2005),
  },
  [STATE(749)] = {
    [sym_comment] = STATE(749),
    [sym_groovy_doc] = STATE(749),
    [sym_parameter_list] = STATE(774),
    [ts_builtin_sym_end] = ACTIONS(2007),
    [sym_identifier] = ACTIONS(2009),
    [anon_sym_break] = ACTIONS(2009),
    [anon_sym_continue] = ACTIONS(2009),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_STAR] = ACTIONS(2009),
    [anon_sym_STAR_COLON] = ACTIONS(2007),
    [anon_sym_import] = ACTIONS(2009),
    [anon_sym_package] = ACTIONS(2009),
    [anon_sym_AT] = ACTIONS(2009),
    [anon_sym_assert] = ACTIONS(2009),
    [anon_sym_EQ] = ACTIONS(2011),
    [anon_sym_PLUS_PLUS] = ACTIONS(2007),
    [anon_sym_DASH_DASH] = ACTIONS(2007),
    [anon_sym_SLASH] = ACTIONS(2009),
    [anon_sym_PLUS] = ACTIONS(2009),
    [anon_sym_DASH] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2009),
    [anon_sym_false] = ACTIONS(2009),
    [anon_sym_ATinterface] = ACTIONS(2007),
    [anon_sym_interface] = ACTIONS(2009),
    [anon_sym_class] = ACTIONS(2009),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_RBRACE] = ACTIONS(2007),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2009),
    [anon_sym_def] = ACTIONS(2009),
    [anon_sym_LPAREN] = ACTIONS(1777),
    [anon_sym_null] = ACTIONS(2009),
    [anon_sym_this] = ACTIONS(2009),
    [anon_sym_do] = ACTIONS(2009),
    [anon_sym_while] = ACTIONS(2009),
    [anon_sym_for] = ACTIONS(2009),
    [aux_sym__type_identifier_token1] = ACTIONS(2009),
    [anon_sym_if] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [aux_sym_number_literal_token1] = ACTIONS(2009),
    [aux_sym_number_literal_token2] = ACTIONS(2007),
    [aux_sym_number_literal_token3] = ACTIONS(2007),
    [aux_sym_number_literal_token4] = ACTIONS(2009),
    [aux_sym_number_literal_token5] = ACTIONS(2007),
    [anon_sym_pipeline] = ACTIONS(2009),
    [anon_sym_return] = ACTIONS(2009),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2007),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2007),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2007),
    [anon_sym_switch] = ACTIONS(2009),
    [anon_sym_try] = ACTIONS(2009),
    [anon_sym_catch] = ACTIONS(2009),
    [anon_sym_finally] = ACTIONS(2009),
    [anon_sym_int] = ACTIONS(2009),
    [anon_sym_boolean] = ACTIONS(2009),
    [anon_sym_char] = ACTIONS(2009),
    [anon_sym_short] = ACTIONS(2009),
    [anon_sym_long] = ACTIONS(2009),
    [anon_sym_float] = ACTIONS(2009),
    [anon_sym_double] = ACTIONS(2009),
    [anon_sym_void] = ACTIONS(2009),
    [anon_sym_public] = ACTIONS(2009),
    [anon_sym_protected] = ACTIONS(2009),
    [anon_sym_private] = ACTIONS(2009),
    [anon_sym_static] = ACTIONS(2009),
    [anon_sym_final] = ACTIONS(2009),
    [anon_sym_synchronized] = ACTIONS(2009),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_BANG] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2009),
  },
  [STATE(750)] = {
    [sym_comment] = STATE(750),
    [sym_groovy_doc] = STATE(750),
    [ts_builtin_sym_end] = ACTIONS(2013),
    [sym_identifier] = ACTIONS(2015),
    [anon_sym_break] = ACTIONS(2015),
    [anon_sym_continue] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_STAR] = ACTIONS(2015),
    [anon_sym_STAR_COLON] = ACTIONS(2013),
    [anon_sym_import] = ACTIONS(2015),
    [anon_sym_package] = ACTIONS(2015),
    [anon_sym_AT] = ACTIONS(2015),
    [anon_sym_assert] = ACTIONS(2015),
    [anon_sym_PLUS_PLUS] = ACTIONS(2013),
    [anon_sym_DASH_DASH] = ACTIONS(2013),
    [anon_sym_SLASH] = ACTIONS(2015),
    [anon_sym_PLUS] = ACTIONS(2015),
    [anon_sym_DASH] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_ATinterface] = ACTIONS(2013),
    [anon_sym_interface] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_LBRACE] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2015),
    [anon_sym_def] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_RPAREN] = ACTIONS(2013),
    [anon_sym_null] = ACTIONS(2015),
    [anon_sym_this] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_while] = ACTIONS(2015),
    [anon_sym_for] = ACTIONS(2015),
    [aux_sym__type_identifier_token1] = ACTIONS(2015),
    [anon_sym_if] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2013),
    [aux_sym_number_literal_token1] = ACTIONS(2015),
    [aux_sym_number_literal_token2] = ACTIONS(2013),
    [aux_sym_number_literal_token3] = ACTIONS(2013),
    [aux_sym_number_literal_token4] = ACTIONS(2015),
    [aux_sym_number_literal_token5] = ACTIONS(2013),
    [anon_sym_pipeline] = ACTIONS(2015),
    [anon_sym_return] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2015),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2015),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2013),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2013),
    [anon_sym_switch] = ACTIONS(2015),
    [anon_sym_try] = ACTIONS(2015),
    [anon_sym_catch] = ACTIONS(2015),
    [anon_sym_finally] = ACTIONS(2015),
    [anon_sym_int] = ACTIONS(2015),
    [anon_sym_boolean] = ACTIONS(2015),
    [anon_sym_char] = ACTIONS(2015),
    [anon_sym_short] = ACTIONS(2015),
    [anon_sym_long] = ACTIONS(2015),
    [anon_sym_float] = ACTIONS(2015),
    [anon_sym_double] = ACTIONS(2015),
    [anon_sym_void] = ACTIONS(2015),
    [anon_sym_public] = ACTIONS(2015),
    [anon_sym_protected] = ACTIONS(2015),
    [anon_sym_private] = ACTIONS(2015),
    [anon_sym_static] = ACTIONS(2015),
    [anon_sym_final] = ACTIONS(2015),
    [anon_sym_synchronized] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2013),
    [anon_sym_BANG] = ACTIONS(2013),
    [anon_sym_new] = ACTIONS(2015),
  },
  [STATE(751)] = {
    [sym_comment] = STATE(751),
    [sym_groovy_doc] = STATE(751),
    [sym_parameter_list] = STATE(770),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2017),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_case] = ACTIONS(1981),
    [anon_sym_default] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_catch] = ACTIONS(1981),
    [anon_sym_finally] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(752)] = {
    [sym_comment] = STATE(752),
    [sym_groovy_doc] = STATE(752),
    [aux_sym__juxt_argument_list_repeat1] = STATE(685),
    [ts_builtin_sym_end] = ACTIONS(1898),
    [sym_identifier] = ACTIONS(1896),
    [anon_sym_break] = ACTIONS(1896),
    [anon_sym_continue] = ACTIONS(1896),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_STAR] = ACTIONS(1896),
    [anon_sym_STAR_COLON] = ACTIONS(1898),
    [anon_sym_import] = ACTIONS(1896),
    [anon_sym_package] = ACTIONS(1896),
    [anon_sym_AT] = ACTIONS(1896),
    [anon_sym_assert] = ACTIONS(1896),
    [anon_sym_PLUS_PLUS] = ACTIONS(1898),
    [anon_sym_DASH_DASH] = ACTIONS(1898),
    [anon_sym_SLASH] = ACTIONS(1896),
    [anon_sym_PLUS] = ACTIONS(1896),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1896),
    [anon_sym_false] = ACTIONS(1896),
    [anon_sym_ATinterface] = ACTIONS(1898),
    [anon_sym_interface] = ACTIONS(1896),
    [anon_sym_class] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1258),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_RBRACE] = ACTIONS(1898),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1896),
    [anon_sym_def] = ACTIONS(1896),
    [anon_sym_LPAREN] = ACTIONS(1898),
    [anon_sym_null] = ACTIONS(1896),
    [anon_sym_this] = ACTIONS(1896),
    [anon_sym_do] = ACTIONS(1896),
    [anon_sym_while] = ACTIONS(1896),
    [anon_sym_for] = ACTIONS(1896),
    [aux_sym__type_identifier_token1] = ACTIONS(1896),
    [anon_sym_if] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1896),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [aux_sym_number_literal_token1] = ACTIONS(1896),
    [aux_sym_number_literal_token2] = ACTIONS(1898),
    [aux_sym_number_literal_token3] = ACTIONS(1898),
    [aux_sym_number_literal_token4] = ACTIONS(1896),
    [aux_sym_number_literal_token5] = ACTIONS(1898),
    [anon_sym_pipeline] = ACTIONS(1896),
    [anon_sym_return] = ACTIONS(1896),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1898),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1898),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1898),
    [anon_sym_switch] = ACTIONS(1896),
    [anon_sym_try] = ACTIONS(1896),
    [anon_sym_catch] = ACTIONS(1896),
    [anon_sym_finally] = ACTIONS(1896),
    [anon_sym_int] = ACTIONS(1896),
    [anon_sym_boolean] = ACTIONS(1896),
    [anon_sym_char] = ACTIONS(1896),
    [anon_sym_short] = ACTIONS(1896),
    [anon_sym_long] = ACTIONS(1896),
    [anon_sym_float] = ACTIONS(1896),
    [anon_sym_double] = ACTIONS(1896),
    [anon_sym_void] = ACTIONS(1896),
    [anon_sym_public] = ACTIONS(1896),
    [anon_sym_protected] = ACTIONS(1896),
    [anon_sym_private] = ACTIONS(1896),
    [anon_sym_static] = ACTIONS(1896),
    [anon_sym_final] = ACTIONS(1896),
    [anon_sym_synchronized] = ACTIONS(1896),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_BANG] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1896),
  },
  [STATE(753)] = {
    [sym_comment] = STATE(753),
    [sym_groovy_doc] = STATE(753),
    [sym_parameter_list] = STATE(770),
    [sym_identifier] = ACTIONS(2009),
    [anon_sym_break] = ACTIONS(2009),
    [anon_sym_continue] = ACTIONS(2009),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_STAR] = ACTIONS(2009),
    [anon_sym_STAR_COLON] = ACTIONS(2007),
    [anon_sym_import] = ACTIONS(2009),
    [anon_sym_package] = ACTIONS(2009),
    [anon_sym_AT] = ACTIONS(2009),
    [anon_sym_assert] = ACTIONS(2009),
    [anon_sym_EQ] = ACTIONS(2019),
    [anon_sym_PLUS_PLUS] = ACTIONS(2007),
    [anon_sym_DASH_DASH] = ACTIONS(2007),
    [anon_sym_SLASH] = ACTIONS(2009),
    [anon_sym_PLUS] = ACTIONS(2009),
    [anon_sym_DASH] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2009),
    [anon_sym_false] = ACTIONS(2009),
    [anon_sym_ATinterface] = ACTIONS(2007),
    [anon_sym_interface] = ACTIONS(2009),
    [anon_sym_class] = ACTIONS(2009),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_RBRACE] = ACTIONS(2007),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2009),
    [anon_sym_def] = ACTIONS(2009),
    [anon_sym_LPAREN] = ACTIONS(1745),
    [anon_sym_null] = ACTIONS(2009),
    [anon_sym_this] = ACTIONS(2009),
    [anon_sym_do] = ACTIONS(2009),
    [anon_sym_while] = ACTIONS(2009),
    [anon_sym_for] = ACTIONS(2009),
    [aux_sym__type_identifier_token1] = ACTIONS(2009),
    [anon_sym_if] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [aux_sym_number_literal_token1] = ACTIONS(2009),
    [aux_sym_number_literal_token2] = ACTIONS(2007),
    [aux_sym_number_literal_token3] = ACTIONS(2007),
    [aux_sym_number_literal_token4] = ACTIONS(2009),
    [aux_sym_number_literal_token5] = ACTIONS(2007),
    [anon_sym_return] = ACTIONS(2009),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2007),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2007),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2007),
    [anon_sym_switch] = ACTIONS(2009),
    [anon_sym_case] = ACTIONS(2009),
    [anon_sym_default] = ACTIONS(2009),
    [anon_sym_try] = ACTIONS(2009),
    [anon_sym_catch] = ACTIONS(2009),
    [anon_sym_finally] = ACTIONS(2009),
    [anon_sym_int] = ACTIONS(2009),
    [anon_sym_boolean] = ACTIONS(2009),
    [anon_sym_char] = ACTIONS(2009),
    [anon_sym_short] = ACTIONS(2009),
    [anon_sym_long] = ACTIONS(2009),
    [anon_sym_float] = ACTIONS(2009),
    [anon_sym_double] = ACTIONS(2009),
    [anon_sym_void] = ACTIONS(2009),
    [anon_sym_public] = ACTIONS(2009),
    [anon_sym_protected] = ACTIONS(2009),
    [anon_sym_private] = ACTIONS(2009),
    [anon_sym_static] = ACTIONS(2009),
    [anon_sym_final] = ACTIONS(2009),
    [anon_sym_synchronized] = ACTIONS(2009),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_BANG] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2009),
  },
  [STATE(754)] = {
    [sym_comment] = STATE(754),
    [sym_groovy_doc] = STATE(754),
    [ts_builtin_sym_end] = ACTIONS(1757),
    [sym_identifier] = ACTIONS(1759),
    [anon_sym_break] = ACTIONS(1759),
    [anon_sym_continue] = ACTIONS(1759),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_STAR] = ACTIONS(1759),
    [anon_sym_STAR_COLON] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1757),
    [anon_sym_import] = ACTIONS(1759),
    [anon_sym_as] = ACTIONS(1759),
    [anon_sym_package] = ACTIONS(1759),
    [anon_sym_AT] = ACTIONS(1759),
    [anon_sym_assert] = ACTIONS(1759),
    [anon_sym_PLUS_PLUS] = ACTIONS(1757),
    [anon_sym_DASH_DASH] = ACTIONS(1757),
    [anon_sym_SLASH] = ACTIONS(1759),
    [anon_sym_PLUS] = ACTIONS(1759),
    [anon_sym_DASH] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1759),
    [anon_sym_false] = ACTIONS(1759),
    [anon_sym_ATinterface] = ACTIONS(1757),
    [anon_sym_interface] = ACTIONS(1759),
    [anon_sym_class] = ACTIONS(1759),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_RBRACE] = ACTIONS(1757),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1759),
    [anon_sym_def] = ACTIONS(1759),
    [anon_sym_LPAREN] = ACTIONS(1757),
    [anon_sym_null] = ACTIONS(1759),
    [anon_sym_this] = ACTIONS(1759),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_while] = ACTIONS(1759),
    [anon_sym_for] = ACTIONS(1759),
    [aux_sym__type_identifier_token1] = ACTIONS(1759),
    [anon_sym_if] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [aux_sym_number_literal_token1] = ACTIONS(1759),
    [aux_sym_number_literal_token2] = ACTIONS(1757),
    [aux_sym_number_literal_token3] = ACTIONS(1757),
    [aux_sym_number_literal_token4] = ACTIONS(1759),
    [aux_sym_number_literal_token5] = ACTIONS(1757),
    [anon_sym_pipeline] = ACTIONS(1759),
    [anon_sym_return] = ACTIONS(1759),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1757),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1757),
    [anon_sym_switch] = ACTIONS(1759),
    [anon_sym_try] = ACTIONS(1759),
    [anon_sym_catch] = ACTIONS(1759),
    [anon_sym_finally] = ACTIONS(1759),
    [anon_sym_int] = ACTIONS(1759),
    [anon_sym_boolean] = ACTIONS(1759),
    [anon_sym_char] = ACTIONS(1759),
    [anon_sym_short] = ACTIONS(1759),
    [anon_sym_long] = ACTIONS(1759),
    [anon_sym_float] = ACTIONS(1759),
    [anon_sym_double] = ACTIONS(1759),
    [anon_sym_void] = ACTIONS(1759),
    [anon_sym_public] = ACTIONS(1759),
    [anon_sym_protected] = ACTIONS(1759),
    [anon_sym_private] = ACTIONS(1759),
    [anon_sym_static] = ACTIONS(1759),
    [anon_sym_final] = ACTIONS(1759),
    [anon_sym_synchronized] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_BANG] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1759),
  },
  [STATE(755)] = {
    [sym_comment] = STATE(755),
    [sym_groovy_doc] = STATE(755),
    [ts_builtin_sym_end] = ACTIONS(2021),
    [sym_identifier] = ACTIONS(2023),
    [anon_sym_break] = ACTIONS(2023),
    [anon_sym_continue] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_STAR] = ACTIONS(2023),
    [anon_sym_STAR_COLON] = ACTIONS(2021),
    [anon_sym_import] = ACTIONS(2023),
    [anon_sym_package] = ACTIONS(2023),
    [anon_sym_AT] = ACTIONS(2023),
    [anon_sym_assert] = ACTIONS(2023),
    [anon_sym_PLUS_PLUS] = ACTIONS(2021),
    [anon_sym_DASH_DASH] = ACTIONS(2021),
    [anon_sym_SLASH] = ACTIONS(2023),
    [anon_sym_PLUS] = ACTIONS(2023),
    [anon_sym_DASH] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_ATinterface] = ACTIONS(2021),
    [anon_sym_interface] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2023),
    [anon_sym_def] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_RPAREN] = ACTIONS(2021),
    [anon_sym_null] = ACTIONS(2023),
    [anon_sym_this] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_while] = ACTIONS(2023),
    [anon_sym_for] = ACTIONS(2023),
    [aux_sym__type_identifier_token1] = ACTIONS(2023),
    [anon_sym_if] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [aux_sym_number_literal_token1] = ACTIONS(2023),
    [aux_sym_number_literal_token2] = ACTIONS(2021),
    [aux_sym_number_literal_token3] = ACTIONS(2021),
    [aux_sym_number_literal_token4] = ACTIONS(2023),
    [aux_sym_number_literal_token5] = ACTIONS(2021),
    [anon_sym_pipeline] = ACTIONS(2023),
    [anon_sym_return] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2021),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2021),
    [anon_sym_switch] = ACTIONS(2023),
    [anon_sym_try] = ACTIONS(2023),
    [anon_sym_catch] = ACTIONS(2023),
    [anon_sym_finally] = ACTIONS(2023),
    [anon_sym_int] = ACTIONS(2023),
    [anon_sym_boolean] = ACTIONS(2023),
    [anon_sym_char] = ACTIONS(2023),
    [anon_sym_short] = ACTIONS(2023),
    [anon_sym_long] = ACTIONS(2023),
    [anon_sym_float] = ACTIONS(2023),
    [anon_sym_double] = ACTIONS(2023),
    [anon_sym_void] = ACTIONS(2023),
    [anon_sym_public] = ACTIONS(2023),
    [anon_sym_protected] = ACTIONS(2023),
    [anon_sym_private] = ACTIONS(2023),
    [anon_sym_static] = ACTIONS(2023),
    [anon_sym_final] = ACTIONS(2023),
    [anon_sym_synchronized] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_BANG] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2023),
  },
  [STATE(756)] = {
    [sym_comment] = STATE(756),
    [sym_groovy_doc] = STATE(756),
    [ts_builtin_sym_end] = ACTIONS(2025),
    [sym_identifier] = ACTIONS(2027),
    [anon_sym_break] = ACTIONS(2027),
    [anon_sym_continue] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_STAR] = ACTIONS(2027),
    [anon_sym_STAR_COLON] = ACTIONS(2025),
    [anon_sym_import] = ACTIONS(2027),
    [anon_sym_package] = ACTIONS(2027),
    [anon_sym_AT] = ACTIONS(2027),
    [anon_sym_assert] = ACTIONS(2027),
    [anon_sym_PLUS_PLUS] = ACTIONS(2025),
    [anon_sym_DASH_DASH] = ACTIONS(2025),
    [anon_sym_SLASH] = ACTIONS(2027),
    [anon_sym_PLUS] = ACTIONS(2027),
    [anon_sym_DASH] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_ATinterface] = ACTIONS(2025),
    [anon_sym_interface] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2027),
    [anon_sym_def] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_RPAREN] = ACTIONS(2025),
    [anon_sym_null] = ACTIONS(2027),
    [anon_sym_this] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_while] = ACTIONS(2027),
    [anon_sym_for] = ACTIONS(2027),
    [aux_sym__type_identifier_token1] = ACTIONS(2027),
    [anon_sym_if] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [aux_sym_number_literal_token1] = ACTIONS(2027),
    [aux_sym_number_literal_token2] = ACTIONS(2025),
    [aux_sym_number_literal_token3] = ACTIONS(2025),
    [aux_sym_number_literal_token4] = ACTIONS(2027),
    [aux_sym_number_literal_token5] = ACTIONS(2025),
    [anon_sym_pipeline] = ACTIONS(2027),
    [anon_sym_return] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2025),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2025),
    [anon_sym_switch] = ACTIONS(2027),
    [anon_sym_try] = ACTIONS(2027),
    [anon_sym_catch] = ACTIONS(2027),
    [anon_sym_finally] = ACTIONS(2027),
    [anon_sym_int] = ACTIONS(2027),
    [anon_sym_boolean] = ACTIONS(2027),
    [anon_sym_char] = ACTIONS(2027),
    [anon_sym_short] = ACTIONS(2027),
    [anon_sym_long] = ACTIONS(2027),
    [anon_sym_float] = ACTIONS(2027),
    [anon_sym_double] = ACTIONS(2027),
    [anon_sym_void] = ACTIONS(2027),
    [anon_sym_public] = ACTIONS(2027),
    [anon_sym_protected] = ACTIONS(2027),
    [anon_sym_private] = ACTIONS(2027),
    [anon_sym_static] = ACTIONS(2027),
    [anon_sym_final] = ACTIONS(2027),
    [anon_sym_synchronized] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_BANG] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2027),
  },
  [STATE(757)] = {
    [sym_comment] = STATE(757),
    [sym_groovy_doc] = STATE(757),
    [ts_builtin_sym_end] = ACTIONS(2029),
    [sym_identifier] = ACTIONS(2031),
    [anon_sym_break] = ACTIONS(2031),
    [anon_sym_continue] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_STAR] = ACTIONS(2031),
    [anon_sym_STAR_COLON] = ACTIONS(2029),
    [anon_sym_import] = ACTIONS(2031),
    [anon_sym_package] = ACTIONS(2031),
    [anon_sym_AT] = ACTIONS(2031),
    [anon_sym_assert] = ACTIONS(2031),
    [anon_sym_PLUS_PLUS] = ACTIONS(2029),
    [anon_sym_DASH_DASH] = ACTIONS(2029),
    [anon_sym_SLASH] = ACTIONS(2031),
    [anon_sym_PLUS] = ACTIONS(2031),
    [anon_sym_DASH] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_ATinterface] = ACTIONS(2029),
    [anon_sym_interface] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_LBRACE] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2031),
    [anon_sym_def] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_RPAREN] = ACTIONS(2029),
    [anon_sym_null] = ACTIONS(2031),
    [anon_sym_this] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_while] = ACTIONS(2031),
    [anon_sym_for] = ACTIONS(2031),
    [aux_sym__type_identifier_token1] = ACTIONS(2031),
    [anon_sym_if] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2029),
    [aux_sym_number_literal_token1] = ACTIONS(2031),
    [aux_sym_number_literal_token2] = ACTIONS(2029),
    [aux_sym_number_literal_token3] = ACTIONS(2029),
    [aux_sym_number_literal_token4] = ACTIONS(2031),
    [aux_sym_number_literal_token5] = ACTIONS(2029),
    [anon_sym_pipeline] = ACTIONS(2031),
    [anon_sym_return] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2031),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2031),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2029),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2029),
    [anon_sym_switch] = ACTIONS(2031),
    [anon_sym_try] = ACTIONS(2031),
    [anon_sym_catch] = ACTIONS(2031),
    [anon_sym_finally] = ACTIONS(2031),
    [anon_sym_int] = ACTIONS(2031),
    [anon_sym_boolean] = ACTIONS(2031),
    [anon_sym_char] = ACTIONS(2031),
    [anon_sym_short] = ACTIONS(2031),
    [anon_sym_long] = ACTIONS(2031),
    [anon_sym_float] = ACTIONS(2031),
    [anon_sym_double] = ACTIONS(2031),
    [anon_sym_void] = ACTIONS(2031),
    [anon_sym_public] = ACTIONS(2031),
    [anon_sym_protected] = ACTIONS(2031),
    [anon_sym_private] = ACTIONS(2031),
    [anon_sym_static] = ACTIONS(2031),
    [anon_sym_final] = ACTIONS(2031),
    [anon_sym_synchronized] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2029),
    [anon_sym_BANG] = ACTIONS(2029),
    [anon_sym_new] = ACTIONS(2031),
  },
  [STATE(758)] = {
    [sym_comment] = STATE(758),
    [sym_groovy_doc] = STATE(758),
    [sym_identifier] = ACTIONS(2033),
    [anon_sym_break] = ACTIONS(2033),
    [anon_sym_continue] = ACTIONS(2033),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_STAR] = ACTIONS(2033),
    [anon_sym_STAR_COLON] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2037),
    [anon_sym_import] = ACTIONS(2033),
    [anon_sym_package] = ACTIONS(2033),
    [anon_sym_AT] = ACTIONS(2033),
    [anon_sym_assert] = ACTIONS(2033),
    [anon_sym_PLUS_PLUS] = ACTIONS(2035),
    [anon_sym_DASH_DASH] = ACTIONS(2035),
    [anon_sym_SLASH] = ACTIONS(2033),
    [anon_sym_PLUS] = ACTIONS(2033),
    [anon_sym_DASH] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2033),
    [anon_sym_false] = ACTIONS(2033),
    [anon_sym_ATinterface] = ACTIONS(2035),
    [anon_sym_interface] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_RBRACE] = ACTIONS(2035),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2033),
    [anon_sym_def] = ACTIONS(2033),
    [anon_sym_LPAREN] = ACTIONS(2035),
    [anon_sym_null] = ACTIONS(2033),
    [anon_sym_this] = ACTIONS(2033),
    [anon_sym_do] = ACTIONS(2033),
    [anon_sym_while] = ACTIONS(2033),
    [anon_sym_for] = ACTIONS(2033),
    [aux_sym__type_identifier_token1] = ACTIONS(2033),
    [anon_sym_if] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [aux_sym_number_literal_token1] = ACTIONS(2033),
    [aux_sym_number_literal_token2] = ACTIONS(2035),
    [aux_sym_number_literal_token3] = ACTIONS(2035),
    [aux_sym_number_literal_token4] = ACTIONS(2033),
    [aux_sym_number_literal_token5] = ACTIONS(2035),
    [anon_sym_return] = ACTIONS(2033),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2035),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2035),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2035),
    [anon_sym_switch] = ACTIONS(2033),
    [anon_sym_case] = ACTIONS(2033),
    [anon_sym_default] = ACTIONS(2033),
    [anon_sym_try] = ACTIONS(2033),
    [anon_sym_catch] = ACTIONS(2033),
    [anon_sym_finally] = ACTIONS(2033),
    [anon_sym_int] = ACTIONS(2033),
    [anon_sym_boolean] = ACTIONS(2033),
    [anon_sym_char] = ACTIONS(2033),
    [anon_sym_short] = ACTIONS(2033),
    [anon_sym_long] = ACTIONS(2033),
    [anon_sym_float] = ACTIONS(2033),
    [anon_sym_double] = ACTIONS(2033),
    [anon_sym_void] = ACTIONS(2033),
    [anon_sym_public] = ACTIONS(2033),
    [anon_sym_protected] = ACTIONS(2033),
    [anon_sym_private] = ACTIONS(2033),
    [anon_sym_static] = ACTIONS(2033),
    [anon_sym_final] = ACTIONS(2033),
    [anon_sym_synchronized] = ACTIONS(2033),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_BANG] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2033),
  },
  [STATE(759)] = {
    [sym_comment] = STATE(759),
    [sym_groovy_doc] = STATE(759),
    [ts_builtin_sym_end] = ACTIONS(1771),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(1775),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_pipeline] = ACTIONS(1773),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_catch] = ACTIONS(1773),
    [anon_sym_finally] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(760)] = {
    [sym_comment] = STATE(760),
    [sym_groovy_doc] = STATE(760),
    [ts_builtin_sym_end] = ACTIONS(1741),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(1888),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1741),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_pipeline] = ACTIONS(1739),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_catch] = ACTIONS(1739),
    [anon_sym_finally] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(761)] = {
    [sym_closure] = STATE(885),
    [sym_comment] = STATE(761),
    [sym_groovy_doc] = STATE(761),
    [sym_identifier] = ACTIONS(2039),
    [anon_sym_break] = ACTIONS(2039),
    [anon_sym_continue] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_STAR] = ACTIONS(2039),
    [anon_sym_STAR_COLON] = ACTIONS(2041),
    [anon_sym_import] = ACTIONS(2039),
    [anon_sym_package] = ACTIONS(2039),
    [anon_sym_AT] = ACTIONS(2039),
    [anon_sym_assert] = ACTIONS(2039),
    [anon_sym_PLUS_PLUS] = ACTIONS(2041),
    [anon_sym_DASH_DASH] = ACTIONS(2041),
    [anon_sym_SLASH] = ACTIONS(2039),
    [anon_sym_PLUS] = ACTIONS(2039),
    [anon_sym_DASH] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_ATinterface] = ACTIONS(2041),
    [anon_sym_interface] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2039),
    [anon_sym_def] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_null] = ACTIONS(2039),
    [anon_sym_this] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_while] = ACTIONS(2039),
    [anon_sym_for] = ACTIONS(2039),
    [aux_sym__type_identifier_token1] = ACTIONS(2039),
    [anon_sym_if] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2041),
    [aux_sym_number_literal_token1] = ACTIONS(2039),
    [aux_sym_number_literal_token2] = ACTIONS(2041),
    [aux_sym_number_literal_token3] = ACTIONS(2041),
    [aux_sym_number_literal_token4] = ACTIONS(2039),
    [aux_sym_number_literal_token5] = ACTIONS(2041),
    [anon_sym_return] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2039),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2039),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2041),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2041),
    [anon_sym_switch] = ACTIONS(2039),
    [anon_sym_case] = ACTIONS(2039),
    [anon_sym_default] = ACTIONS(2039),
    [anon_sym_try] = ACTIONS(2039),
    [anon_sym_catch] = ACTIONS(2039),
    [anon_sym_finally] = ACTIONS(2039),
    [anon_sym_int] = ACTIONS(2039),
    [anon_sym_boolean] = ACTIONS(2039),
    [anon_sym_char] = ACTIONS(2039),
    [anon_sym_short] = ACTIONS(2039),
    [anon_sym_long] = ACTIONS(2039),
    [anon_sym_float] = ACTIONS(2039),
    [anon_sym_double] = ACTIONS(2039),
    [anon_sym_void] = ACTIONS(2039),
    [anon_sym_public] = ACTIONS(2039),
    [anon_sym_protected] = ACTIONS(2039),
    [anon_sym_private] = ACTIONS(2039),
    [anon_sym_static] = ACTIONS(2039),
    [anon_sym_final] = ACTIONS(2039),
    [anon_sym_synchronized] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2041),
    [anon_sym_BANG] = ACTIONS(2041),
    [anon_sym_new] = ACTIONS(2039),
  },
  [STATE(762)] = {
    [sym_comment] = STATE(762),
    [sym_groovy_doc] = STATE(762),
    [ts_builtin_sym_end] = ACTIONS(1979),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(1983),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1979),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_pipeline] = ACTIONS(1981),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_catch] = ACTIONS(1981),
    [anon_sym_finally] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(763)] = {
    [sym_comment] = STATE(763),
    [sym_groovy_doc] = STATE(763),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_catch] = ACTIONS(1270),
    [anon_sym_finally] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(764)] = {
    [sym_comment] = STATE(764),
    [sym_groovy_doc] = STATE(764),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(1743),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1741),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_case] = ACTIONS(1739),
    [anon_sym_default] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_catch] = ACTIONS(1739),
    [anon_sym_finally] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(765)] = {
    [sym_closure] = STATE(738),
    [sym_comment] = STATE(765),
    [sym_groovy_doc] = STATE(765),
    [ts_builtin_sym_end] = ACTIONS(2041),
    [sym_identifier] = ACTIONS(2039),
    [anon_sym_break] = ACTIONS(2039),
    [anon_sym_continue] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_STAR] = ACTIONS(2039),
    [anon_sym_STAR_COLON] = ACTIONS(2041),
    [anon_sym_import] = ACTIONS(2039),
    [anon_sym_package] = ACTIONS(2039),
    [anon_sym_AT] = ACTIONS(2039),
    [anon_sym_assert] = ACTIONS(2039),
    [anon_sym_PLUS_PLUS] = ACTIONS(2041),
    [anon_sym_DASH_DASH] = ACTIONS(2041),
    [anon_sym_SLASH] = ACTIONS(2039),
    [anon_sym_PLUS] = ACTIONS(2039),
    [anon_sym_DASH] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_ATinterface] = ACTIONS(2041),
    [anon_sym_interface] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2039),
    [anon_sym_def] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_null] = ACTIONS(2039),
    [anon_sym_this] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_while] = ACTIONS(2039),
    [anon_sym_for] = ACTIONS(2039),
    [aux_sym__type_identifier_token1] = ACTIONS(2039),
    [anon_sym_if] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2041),
    [aux_sym_number_literal_token1] = ACTIONS(2039),
    [aux_sym_number_literal_token2] = ACTIONS(2041),
    [aux_sym_number_literal_token3] = ACTIONS(2041),
    [aux_sym_number_literal_token4] = ACTIONS(2039),
    [aux_sym_number_literal_token5] = ACTIONS(2041),
    [anon_sym_pipeline] = ACTIONS(2039),
    [anon_sym_return] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2039),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2039),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2041),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2041),
    [anon_sym_switch] = ACTIONS(2039),
    [anon_sym_try] = ACTIONS(2039),
    [anon_sym_catch] = ACTIONS(2039),
    [anon_sym_finally] = ACTIONS(2039),
    [anon_sym_int] = ACTIONS(2039),
    [anon_sym_boolean] = ACTIONS(2039),
    [anon_sym_char] = ACTIONS(2039),
    [anon_sym_short] = ACTIONS(2039),
    [anon_sym_long] = ACTIONS(2039),
    [anon_sym_float] = ACTIONS(2039),
    [anon_sym_double] = ACTIONS(2039),
    [anon_sym_void] = ACTIONS(2039),
    [anon_sym_public] = ACTIONS(2039),
    [anon_sym_protected] = ACTIONS(2039),
    [anon_sym_private] = ACTIONS(2039),
    [anon_sym_static] = ACTIONS(2039),
    [anon_sym_final] = ACTIONS(2039),
    [anon_sym_synchronized] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2041),
    [anon_sym_BANG] = ACTIONS(2041),
    [anon_sym_new] = ACTIONS(2039),
  },
  [STATE(766)] = {
    [sym_closure] = STATE(865),
    [sym_comment] = STATE(766),
    [sym_groovy_doc] = STATE(766),
    [sym_identifier] = ACTIONS(2047),
    [anon_sym_break] = ACTIONS(2047),
    [anon_sym_continue] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_STAR] = ACTIONS(2047),
    [anon_sym_STAR_COLON] = ACTIONS(2049),
    [anon_sym_import] = ACTIONS(2047),
    [anon_sym_package] = ACTIONS(2047),
    [anon_sym_AT] = ACTIONS(2047),
    [anon_sym_assert] = ACTIONS(2047),
    [anon_sym_PLUS_PLUS] = ACTIONS(2049),
    [anon_sym_DASH_DASH] = ACTIONS(2049),
    [anon_sym_SLASH] = ACTIONS(2047),
    [anon_sym_PLUS] = ACTIONS(2047),
    [anon_sym_DASH] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_ATinterface] = ACTIONS(2049),
    [anon_sym_interface] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2047),
    [anon_sym_def] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_null] = ACTIONS(2047),
    [anon_sym_this] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_while] = ACTIONS(2047),
    [anon_sym_for] = ACTIONS(2047),
    [aux_sym__type_identifier_token1] = ACTIONS(2047),
    [anon_sym_if] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2049),
    [aux_sym_number_literal_token1] = ACTIONS(2047),
    [aux_sym_number_literal_token2] = ACTIONS(2049),
    [aux_sym_number_literal_token3] = ACTIONS(2049),
    [aux_sym_number_literal_token4] = ACTIONS(2047),
    [aux_sym_number_literal_token5] = ACTIONS(2049),
    [anon_sym_return] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2047),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2047),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2049),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2049),
    [anon_sym_switch] = ACTIONS(2047),
    [anon_sym_case] = ACTIONS(2047),
    [anon_sym_default] = ACTIONS(2047),
    [anon_sym_try] = ACTIONS(2047),
    [anon_sym_catch] = ACTIONS(2047),
    [anon_sym_finally] = ACTIONS(2047),
    [anon_sym_int] = ACTIONS(2047),
    [anon_sym_boolean] = ACTIONS(2047),
    [anon_sym_char] = ACTIONS(2047),
    [anon_sym_short] = ACTIONS(2047),
    [anon_sym_long] = ACTIONS(2047),
    [anon_sym_float] = ACTIONS(2047),
    [anon_sym_double] = ACTIONS(2047),
    [anon_sym_void] = ACTIONS(2047),
    [anon_sym_public] = ACTIONS(2047),
    [anon_sym_protected] = ACTIONS(2047),
    [anon_sym_private] = ACTIONS(2047),
    [anon_sym_static] = ACTIONS(2047),
    [anon_sym_final] = ACTIONS(2047),
    [anon_sym_synchronized] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2049),
    [anon_sym_BANG] = ACTIONS(2049),
    [anon_sym_new] = ACTIONS(2047),
  },
  [STATE(767)] = {
    [sym_comment] = STATE(767),
    [sym_groovy_doc] = STATE(767),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(1964),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_case] = ACTIONS(1773),
    [anon_sym_default] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_catch] = ACTIONS(1773),
    [anon_sym_finally] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(768)] = {
    [sym_closure] = STATE(687),
    [sym_comment] = STATE(768),
    [sym_groovy_doc] = STATE(768),
    [ts_builtin_sym_end] = ACTIONS(2049),
    [sym_identifier] = ACTIONS(2047),
    [anon_sym_break] = ACTIONS(2047),
    [anon_sym_continue] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_STAR] = ACTIONS(2047),
    [anon_sym_STAR_COLON] = ACTIONS(2049),
    [anon_sym_import] = ACTIONS(2047),
    [anon_sym_package] = ACTIONS(2047),
    [anon_sym_AT] = ACTIONS(2047),
    [anon_sym_assert] = ACTIONS(2047),
    [anon_sym_PLUS_PLUS] = ACTIONS(2049),
    [anon_sym_DASH_DASH] = ACTIONS(2049),
    [anon_sym_SLASH] = ACTIONS(2047),
    [anon_sym_PLUS] = ACTIONS(2047),
    [anon_sym_DASH] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_ATinterface] = ACTIONS(2049),
    [anon_sym_interface] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2047),
    [anon_sym_def] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_null] = ACTIONS(2047),
    [anon_sym_this] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_while] = ACTIONS(2047),
    [anon_sym_for] = ACTIONS(2047),
    [aux_sym__type_identifier_token1] = ACTIONS(2047),
    [anon_sym_if] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2049),
    [aux_sym_number_literal_token1] = ACTIONS(2047),
    [aux_sym_number_literal_token2] = ACTIONS(2049),
    [aux_sym_number_literal_token3] = ACTIONS(2049),
    [aux_sym_number_literal_token4] = ACTIONS(2047),
    [aux_sym_number_literal_token5] = ACTIONS(2049),
    [anon_sym_pipeline] = ACTIONS(2047),
    [anon_sym_return] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2047),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2047),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2049),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2049),
    [anon_sym_switch] = ACTIONS(2047),
    [anon_sym_try] = ACTIONS(2047),
    [anon_sym_catch] = ACTIONS(2047),
    [anon_sym_finally] = ACTIONS(2047),
    [anon_sym_int] = ACTIONS(2047),
    [anon_sym_boolean] = ACTIONS(2047),
    [anon_sym_char] = ACTIONS(2047),
    [anon_sym_short] = ACTIONS(2047),
    [anon_sym_long] = ACTIONS(2047),
    [anon_sym_float] = ACTIONS(2047),
    [anon_sym_double] = ACTIONS(2047),
    [anon_sym_void] = ACTIONS(2047),
    [anon_sym_public] = ACTIONS(2047),
    [anon_sym_protected] = ACTIONS(2047),
    [anon_sym_private] = ACTIONS(2047),
    [anon_sym_static] = ACTIONS(2047),
    [anon_sym_final] = ACTIONS(2047),
    [anon_sym_synchronized] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2049),
    [anon_sym_BANG] = ACTIONS(2049),
    [anon_sym_new] = ACTIONS(2047),
  },
  [STATE(769)] = {
    [sym_closure] = STATE(843),
    [sym_comment] = STATE(769),
    [sym_groovy_doc] = STATE(769),
    [sym_identifier] = ACTIONS(2051),
    [anon_sym_break] = ACTIONS(2051),
    [anon_sym_continue] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_STAR] = ACTIONS(2051),
    [anon_sym_STAR_COLON] = ACTIONS(2053),
    [anon_sym_import] = ACTIONS(2051),
    [anon_sym_package] = ACTIONS(2051),
    [anon_sym_AT] = ACTIONS(2051),
    [anon_sym_assert] = ACTIONS(2051),
    [anon_sym_PLUS_PLUS] = ACTIONS(2053),
    [anon_sym_DASH_DASH] = ACTIONS(2053),
    [anon_sym_SLASH] = ACTIONS(2051),
    [anon_sym_PLUS] = ACTIONS(2051),
    [anon_sym_DASH] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_ATinterface] = ACTIONS(2053),
    [anon_sym_interface] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2051),
    [anon_sym_def] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_null] = ACTIONS(2051),
    [anon_sym_this] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_while] = ACTIONS(2051),
    [anon_sym_for] = ACTIONS(2051),
    [aux_sym__type_identifier_token1] = ACTIONS(2051),
    [anon_sym_if] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2053),
    [aux_sym_number_literal_token1] = ACTIONS(2051),
    [aux_sym_number_literal_token2] = ACTIONS(2053),
    [aux_sym_number_literal_token3] = ACTIONS(2053),
    [aux_sym_number_literal_token4] = ACTIONS(2051),
    [aux_sym_number_literal_token5] = ACTIONS(2053),
    [anon_sym_return] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2051),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2051),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2053),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2053),
    [anon_sym_switch] = ACTIONS(2051),
    [anon_sym_case] = ACTIONS(2051),
    [anon_sym_default] = ACTIONS(2051),
    [anon_sym_try] = ACTIONS(2051),
    [anon_sym_catch] = ACTIONS(2051),
    [anon_sym_finally] = ACTIONS(2051),
    [anon_sym_int] = ACTIONS(2051),
    [anon_sym_boolean] = ACTIONS(2051),
    [anon_sym_char] = ACTIONS(2051),
    [anon_sym_short] = ACTIONS(2051),
    [anon_sym_long] = ACTIONS(2051),
    [anon_sym_float] = ACTIONS(2051),
    [anon_sym_double] = ACTIONS(2051),
    [anon_sym_void] = ACTIONS(2051),
    [anon_sym_public] = ACTIONS(2051),
    [anon_sym_protected] = ACTIONS(2051),
    [anon_sym_private] = ACTIONS(2051),
    [anon_sym_static] = ACTIONS(2051),
    [anon_sym_final] = ACTIONS(2051),
    [anon_sym_synchronized] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2053),
    [anon_sym_BANG] = ACTIONS(2053),
    [anon_sym_new] = ACTIONS(2051),
  },
  [STATE(770)] = {
    [sym_closure] = STATE(874),
    [sym_comment] = STATE(770),
    [sym_groovy_doc] = STATE(770),
    [sym_identifier] = ACTIONS(2055),
    [anon_sym_break] = ACTIONS(2055),
    [anon_sym_continue] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_STAR] = ACTIONS(2055),
    [anon_sym_STAR_COLON] = ACTIONS(2057),
    [anon_sym_import] = ACTIONS(2055),
    [anon_sym_package] = ACTIONS(2055),
    [anon_sym_AT] = ACTIONS(2055),
    [anon_sym_assert] = ACTIONS(2055),
    [anon_sym_PLUS_PLUS] = ACTIONS(2057),
    [anon_sym_DASH_DASH] = ACTIONS(2057),
    [anon_sym_SLASH] = ACTIONS(2055),
    [anon_sym_PLUS] = ACTIONS(2055),
    [anon_sym_DASH] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_ATinterface] = ACTIONS(2057),
    [anon_sym_interface] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2043),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2055),
    [anon_sym_def] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_null] = ACTIONS(2055),
    [anon_sym_this] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_while] = ACTIONS(2055),
    [anon_sym_for] = ACTIONS(2055),
    [aux_sym__type_identifier_token1] = ACTIONS(2055),
    [anon_sym_if] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2057),
    [aux_sym_number_literal_token1] = ACTIONS(2055),
    [aux_sym_number_literal_token2] = ACTIONS(2057),
    [aux_sym_number_literal_token3] = ACTIONS(2057),
    [aux_sym_number_literal_token4] = ACTIONS(2055),
    [aux_sym_number_literal_token5] = ACTIONS(2057),
    [anon_sym_return] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2055),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2057),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2057),
    [anon_sym_switch] = ACTIONS(2055),
    [anon_sym_case] = ACTIONS(2055),
    [anon_sym_default] = ACTIONS(2055),
    [anon_sym_try] = ACTIONS(2055),
    [anon_sym_catch] = ACTIONS(2055),
    [anon_sym_finally] = ACTIONS(2055),
    [anon_sym_int] = ACTIONS(2055),
    [anon_sym_boolean] = ACTIONS(2055),
    [anon_sym_char] = ACTIONS(2055),
    [anon_sym_short] = ACTIONS(2055),
    [anon_sym_long] = ACTIONS(2055),
    [anon_sym_float] = ACTIONS(2055),
    [anon_sym_double] = ACTIONS(2055),
    [anon_sym_void] = ACTIONS(2055),
    [anon_sym_public] = ACTIONS(2055),
    [anon_sym_protected] = ACTIONS(2055),
    [anon_sym_private] = ACTIONS(2055),
    [anon_sym_static] = ACTIONS(2055),
    [anon_sym_final] = ACTIONS(2055),
    [anon_sym_synchronized] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2057),
    [anon_sym_BANG] = ACTIONS(2057),
    [anon_sym_new] = ACTIONS(2055),
  },
  [STATE(771)] = {
    [sym_comment] = STATE(771),
    [sym_groovy_doc] = STATE(771),
    [ts_builtin_sym_end] = ACTIONS(2035),
    [sym_identifier] = ACTIONS(2033),
    [anon_sym_break] = ACTIONS(2033),
    [anon_sym_continue] = ACTIONS(2033),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_STAR] = ACTIONS(2033),
    [anon_sym_STAR_COLON] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2059),
    [anon_sym_import] = ACTIONS(2033),
    [anon_sym_package] = ACTIONS(2033),
    [anon_sym_AT] = ACTIONS(2033),
    [anon_sym_assert] = ACTIONS(2033),
    [anon_sym_PLUS_PLUS] = ACTIONS(2035),
    [anon_sym_DASH_DASH] = ACTIONS(2035),
    [anon_sym_SLASH] = ACTIONS(2033),
    [anon_sym_PLUS] = ACTIONS(2033),
    [anon_sym_DASH] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2033),
    [anon_sym_false] = ACTIONS(2033),
    [anon_sym_ATinterface] = ACTIONS(2035),
    [anon_sym_interface] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_RBRACE] = ACTIONS(2035),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2033),
    [anon_sym_def] = ACTIONS(2033),
    [anon_sym_LPAREN] = ACTIONS(2035),
    [anon_sym_null] = ACTIONS(2033),
    [anon_sym_this] = ACTIONS(2033),
    [anon_sym_do] = ACTIONS(2033),
    [anon_sym_while] = ACTIONS(2033),
    [anon_sym_for] = ACTIONS(2033),
    [aux_sym__type_identifier_token1] = ACTIONS(2033),
    [anon_sym_if] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [aux_sym_number_literal_token1] = ACTIONS(2033),
    [aux_sym_number_literal_token2] = ACTIONS(2035),
    [aux_sym_number_literal_token3] = ACTIONS(2035),
    [aux_sym_number_literal_token4] = ACTIONS(2033),
    [aux_sym_number_literal_token5] = ACTIONS(2035),
    [anon_sym_pipeline] = ACTIONS(2033),
    [anon_sym_return] = ACTIONS(2033),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2035),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2035),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2035),
    [anon_sym_switch] = ACTIONS(2033),
    [anon_sym_try] = ACTIONS(2033),
    [anon_sym_catch] = ACTIONS(2033),
    [anon_sym_finally] = ACTIONS(2033),
    [anon_sym_int] = ACTIONS(2033),
    [anon_sym_boolean] = ACTIONS(2033),
    [anon_sym_char] = ACTIONS(2033),
    [anon_sym_short] = ACTIONS(2033),
    [anon_sym_long] = ACTIONS(2033),
    [anon_sym_float] = ACTIONS(2033),
    [anon_sym_double] = ACTIONS(2033),
    [anon_sym_void] = ACTIONS(2033),
    [anon_sym_public] = ACTIONS(2033),
    [anon_sym_protected] = ACTIONS(2033),
    [anon_sym_private] = ACTIONS(2033),
    [anon_sym_static] = ACTIONS(2033),
    [anon_sym_final] = ACTIONS(2033),
    [anon_sym_synchronized] = ACTIONS(2033),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_BANG] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2033),
  },
  [STATE(772)] = {
    [sym_comment] = STATE(772),
    [sym_groovy_doc] = STATE(772),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2017),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1979),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_case] = ACTIONS(1981),
    [anon_sym_default] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_catch] = ACTIONS(1981),
    [anon_sym_finally] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(773)] = {
    [sym_closure] = STATE(702),
    [sym_comment] = STATE(773),
    [sym_groovy_doc] = STATE(773),
    [ts_builtin_sym_end] = ACTIONS(2053),
    [sym_identifier] = ACTIONS(2051),
    [anon_sym_break] = ACTIONS(2051),
    [anon_sym_continue] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_STAR] = ACTIONS(2051),
    [anon_sym_STAR_COLON] = ACTIONS(2053),
    [anon_sym_import] = ACTIONS(2051),
    [anon_sym_package] = ACTIONS(2051),
    [anon_sym_AT] = ACTIONS(2051),
    [anon_sym_assert] = ACTIONS(2051),
    [anon_sym_PLUS_PLUS] = ACTIONS(2053),
    [anon_sym_DASH_DASH] = ACTIONS(2053),
    [anon_sym_SLASH] = ACTIONS(2051),
    [anon_sym_PLUS] = ACTIONS(2051),
    [anon_sym_DASH] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_ATinterface] = ACTIONS(2053),
    [anon_sym_interface] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2051),
    [anon_sym_def] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_null] = ACTIONS(2051),
    [anon_sym_this] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_while] = ACTIONS(2051),
    [anon_sym_for] = ACTIONS(2051),
    [aux_sym__type_identifier_token1] = ACTIONS(2051),
    [anon_sym_if] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2053),
    [aux_sym_number_literal_token1] = ACTIONS(2051),
    [aux_sym_number_literal_token2] = ACTIONS(2053),
    [aux_sym_number_literal_token3] = ACTIONS(2053),
    [aux_sym_number_literal_token4] = ACTIONS(2051),
    [aux_sym_number_literal_token5] = ACTIONS(2053),
    [anon_sym_pipeline] = ACTIONS(2051),
    [anon_sym_return] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2051),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2051),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2053),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2053),
    [anon_sym_switch] = ACTIONS(2051),
    [anon_sym_try] = ACTIONS(2051),
    [anon_sym_catch] = ACTIONS(2051),
    [anon_sym_finally] = ACTIONS(2051),
    [anon_sym_int] = ACTIONS(2051),
    [anon_sym_boolean] = ACTIONS(2051),
    [anon_sym_char] = ACTIONS(2051),
    [anon_sym_short] = ACTIONS(2051),
    [anon_sym_long] = ACTIONS(2051),
    [anon_sym_float] = ACTIONS(2051),
    [anon_sym_double] = ACTIONS(2051),
    [anon_sym_void] = ACTIONS(2051),
    [anon_sym_public] = ACTIONS(2051),
    [anon_sym_protected] = ACTIONS(2051),
    [anon_sym_private] = ACTIONS(2051),
    [anon_sym_static] = ACTIONS(2051),
    [anon_sym_final] = ACTIONS(2051),
    [anon_sym_synchronized] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2053),
    [anon_sym_BANG] = ACTIONS(2053),
    [anon_sym_new] = ACTIONS(2051),
  },
  [STATE(774)] = {
    [sym_closure] = STATE(724),
    [sym_comment] = STATE(774),
    [sym_groovy_doc] = STATE(774),
    [ts_builtin_sym_end] = ACTIONS(2057),
    [sym_identifier] = ACTIONS(2055),
    [anon_sym_break] = ACTIONS(2055),
    [anon_sym_continue] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_STAR] = ACTIONS(2055),
    [anon_sym_STAR_COLON] = ACTIONS(2057),
    [anon_sym_import] = ACTIONS(2055),
    [anon_sym_package] = ACTIONS(2055),
    [anon_sym_AT] = ACTIONS(2055),
    [anon_sym_assert] = ACTIONS(2055),
    [anon_sym_PLUS_PLUS] = ACTIONS(2057),
    [anon_sym_DASH_DASH] = ACTIONS(2057),
    [anon_sym_SLASH] = ACTIONS(2055),
    [anon_sym_PLUS] = ACTIONS(2055),
    [anon_sym_DASH] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_ATinterface] = ACTIONS(2057),
    [anon_sym_interface] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2045),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2055),
    [anon_sym_def] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_null] = ACTIONS(2055),
    [anon_sym_this] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_while] = ACTIONS(2055),
    [anon_sym_for] = ACTIONS(2055),
    [aux_sym__type_identifier_token1] = ACTIONS(2055),
    [anon_sym_if] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2057),
    [aux_sym_number_literal_token1] = ACTIONS(2055),
    [aux_sym_number_literal_token2] = ACTIONS(2057),
    [aux_sym_number_literal_token3] = ACTIONS(2057),
    [aux_sym_number_literal_token4] = ACTIONS(2055),
    [aux_sym_number_literal_token5] = ACTIONS(2057),
    [anon_sym_pipeline] = ACTIONS(2055),
    [anon_sym_return] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2055),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2057),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2057),
    [anon_sym_switch] = ACTIONS(2055),
    [anon_sym_try] = ACTIONS(2055),
    [anon_sym_catch] = ACTIONS(2055),
    [anon_sym_finally] = ACTIONS(2055),
    [anon_sym_int] = ACTIONS(2055),
    [anon_sym_boolean] = ACTIONS(2055),
    [anon_sym_char] = ACTIONS(2055),
    [anon_sym_short] = ACTIONS(2055),
    [anon_sym_long] = ACTIONS(2055),
    [anon_sym_float] = ACTIONS(2055),
    [anon_sym_double] = ACTIONS(2055),
    [anon_sym_void] = ACTIONS(2055),
    [anon_sym_public] = ACTIONS(2055),
    [anon_sym_protected] = ACTIONS(2055),
    [anon_sym_private] = ACTIONS(2055),
    [anon_sym_static] = ACTIONS(2055),
    [anon_sym_final] = ACTIONS(2055),
    [anon_sym_synchronized] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2057),
    [anon_sym_BANG] = ACTIONS(2057),
    [anon_sym_new] = ACTIONS(2055),
  },
  [STATE(775)] = {
    [sym_comment] = STATE(775),
    [sym_groovy_doc] = STATE(775),
    [sym_identifier] = ACTIONS(2023),
    [anon_sym_break] = ACTIONS(2023),
    [anon_sym_continue] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_STAR] = ACTIONS(2023),
    [anon_sym_STAR_COLON] = ACTIONS(2021),
    [anon_sym_import] = ACTIONS(2023),
    [anon_sym_package] = ACTIONS(2023),
    [anon_sym_AT] = ACTIONS(2023),
    [anon_sym_assert] = ACTIONS(2023),
    [anon_sym_PLUS_PLUS] = ACTIONS(2021),
    [anon_sym_DASH_DASH] = ACTIONS(2021),
    [anon_sym_SLASH] = ACTIONS(2023),
    [anon_sym_PLUS] = ACTIONS(2023),
    [anon_sym_DASH] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_ATinterface] = ACTIONS(2021),
    [anon_sym_interface] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2023),
    [anon_sym_def] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_null] = ACTIONS(2023),
    [anon_sym_this] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_while] = ACTIONS(2023),
    [anon_sym_for] = ACTIONS(2023),
    [aux_sym__type_identifier_token1] = ACTIONS(2023),
    [anon_sym_if] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [aux_sym_number_literal_token1] = ACTIONS(2023),
    [aux_sym_number_literal_token2] = ACTIONS(2021),
    [aux_sym_number_literal_token3] = ACTIONS(2021),
    [aux_sym_number_literal_token4] = ACTIONS(2023),
    [aux_sym_number_literal_token5] = ACTIONS(2021),
    [anon_sym_return] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2021),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2021),
    [anon_sym_switch] = ACTIONS(2023),
    [anon_sym_case] = ACTIONS(2023),
    [anon_sym_default] = ACTIONS(2023),
    [anon_sym_try] = ACTIONS(2023),
    [anon_sym_catch] = ACTIONS(2023),
    [anon_sym_finally] = ACTIONS(2023),
    [anon_sym_int] = ACTIONS(2023),
    [anon_sym_boolean] = ACTIONS(2023),
    [anon_sym_char] = ACTIONS(2023),
    [anon_sym_short] = ACTIONS(2023),
    [anon_sym_long] = ACTIONS(2023),
    [anon_sym_float] = ACTIONS(2023),
    [anon_sym_double] = ACTIONS(2023),
    [anon_sym_void] = ACTIONS(2023),
    [anon_sym_public] = ACTIONS(2023),
    [anon_sym_protected] = ACTIONS(2023),
    [anon_sym_private] = ACTIONS(2023),
    [anon_sym_static] = ACTIONS(2023),
    [anon_sym_final] = ACTIONS(2023),
    [anon_sym_synchronized] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_BANG] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2023),
  },
  [STATE(776)] = {
    [sym_comment] = STATE(776),
    [sym_groovy_doc] = STATE(776),
    [ts_builtin_sym_end] = ACTIONS(1904),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(2061),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_pipeline] = ACTIONS(1906),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_catch] = ACTIONS(1906),
    [anon_sym_finally] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(777)] = {
    [sym_comment] = STATE(777),
    [sym_groovy_doc] = STATE(777),
    [ts_builtin_sym_end] = ACTIONS(1848),
    [sym_identifier] = ACTIONS(1850),
    [anon_sym_break] = ACTIONS(1850),
    [anon_sym_continue] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_STAR] = ACTIONS(1850),
    [anon_sym_STAR_COLON] = ACTIONS(1848),
    [anon_sym_import] = ACTIONS(1850),
    [anon_sym_package] = ACTIONS(1850),
    [anon_sym_AT] = ACTIONS(1850),
    [anon_sym_assert] = ACTIONS(1850),
    [anon_sym_PLUS_PLUS] = ACTIONS(1848),
    [anon_sym_DASH_DASH] = ACTIONS(1848),
    [anon_sym_SLASH] = ACTIONS(1850),
    [anon_sym_PLUS] = ACTIONS(1850),
    [anon_sym_DASH] = ACTIONS(1850),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_ATinterface] = ACTIONS(1848),
    [anon_sym_interface] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_COMMA] = ACTIONS(1848),
    [anon_sym_LBRACE] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1850),
    [anon_sym_def] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_RPAREN] = ACTIONS(1848),
    [anon_sym_null] = ACTIONS(1850),
    [anon_sym_this] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_while] = ACTIONS(1850),
    [anon_sym_for] = ACTIONS(1850),
    [aux_sym__type_identifier_token1] = ACTIONS(1850),
    [anon_sym_if] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1848),
    [aux_sym_number_literal_token1] = ACTIONS(1850),
    [aux_sym_number_literal_token2] = ACTIONS(1848),
    [aux_sym_number_literal_token3] = ACTIONS(1848),
    [aux_sym_number_literal_token4] = ACTIONS(1850),
    [aux_sym_number_literal_token5] = ACTIONS(1848),
    [anon_sym_pipeline] = ACTIONS(1850),
    [anon_sym_return] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1850),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1850),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1848),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1848),
    [anon_sym_switch] = ACTIONS(1850),
    [anon_sym_try] = ACTIONS(1850),
    [anon_sym_int] = ACTIONS(1850),
    [anon_sym_boolean] = ACTIONS(1850),
    [anon_sym_char] = ACTIONS(1850),
    [anon_sym_short] = ACTIONS(1850),
    [anon_sym_long] = ACTIONS(1850),
    [anon_sym_float] = ACTIONS(1850),
    [anon_sym_double] = ACTIONS(1850),
    [anon_sym_void] = ACTIONS(1850),
    [anon_sym_public] = ACTIONS(1850),
    [anon_sym_protected] = ACTIONS(1850),
    [anon_sym_private] = ACTIONS(1850),
    [anon_sym_static] = ACTIONS(1850),
    [anon_sym_final] = ACTIONS(1850),
    [anon_sym_synchronized] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1848),
    [anon_sym_BANG] = ACTIONS(1848),
    [anon_sym_new] = ACTIONS(1850),
  },
  [STATE(778)] = {
    [sym_comment] = STATE(778),
    [sym_groovy_doc] = STATE(778),
    [ts_builtin_sym_end] = ACTIONS(1860),
    [sym_identifier] = ACTIONS(1862),
    [anon_sym_break] = ACTIONS(1862),
    [anon_sym_continue] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_STAR] = ACTIONS(1862),
    [anon_sym_STAR_COLON] = ACTIONS(1860),
    [anon_sym_import] = ACTIONS(1862),
    [anon_sym_package] = ACTIONS(1862),
    [anon_sym_AT] = ACTIONS(1862),
    [anon_sym_assert] = ACTIONS(1862),
    [anon_sym_PLUS_PLUS] = ACTIONS(1860),
    [anon_sym_DASH_DASH] = ACTIONS(1860),
    [anon_sym_SLASH] = ACTIONS(1862),
    [anon_sym_PLUS] = ACTIONS(1862),
    [anon_sym_DASH] = ACTIONS(1862),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_ATinterface] = ACTIONS(1860),
    [anon_sym_interface] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_COMMA] = ACTIONS(1860),
    [anon_sym_LBRACE] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1862),
    [anon_sym_def] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_RPAREN] = ACTIONS(1860),
    [anon_sym_null] = ACTIONS(1862),
    [anon_sym_this] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_while] = ACTIONS(1862),
    [anon_sym_for] = ACTIONS(1862),
    [aux_sym__type_identifier_token1] = ACTIONS(1862),
    [anon_sym_if] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1860),
    [aux_sym_number_literal_token1] = ACTIONS(1862),
    [aux_sym_number_literal_token2] = ACTIONS(1860),
    [aux_sym_number_literal_token3] = ACTIONS(1860),
    [aux_sym_number_literal_token4] = ACTIONS(1862),
    [aux_sym_number_literal_token5] = ACTIONS(1860),
    [anon_sym_pipeline] = ACTIONS(1862),
    [anon_sym_return] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1862),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1862),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1860),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1860),
    [anon_sym_switch] = ACTIONS(1862),
    [anon_sym_try] = ACTIONS(1862),
    [anon_sym_int] = ACTIONS(1862),
    [anon_sym_boolean] = ACTIONS(1862),
    [anon_sym_char] = ACTIONS(1862),
    [anon_sym_short] = ACTIONS(1862),
    [anon_sym_long] = ACTIONS(1862),
    [anon_sym_float] = ACTIONS(1862),
    [anon_sym_double] = ACTIONS(1862),
    [anon_sym_void] = ACTIONS(1862),
    [anon_sym_public] = ACTIONS(1862),
    [anon_sym_protected] = ACTIONS(1862),
    [anon_sym_private] = ACTIONS(1862),
    [anon_sym_static] = ACTIONS(1862),
    [anon_sym_final] = ACTIONS(1862),
    [anon_sym_synchronized] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1860),
    [anon_sym_BANG] = ACTIONS(1860),
    [anon_sym_new] = ACTIONS(1862),
  },
  [STATE(779)] = {
    [sym_comment] = STATE(779),
    [sym_groovy_doc] = STATE(779),
    [aux_sym__juxt_argument_list_repeat1] = STATE(852),
    [sym_identifier] = ACTIONS(1896),
    [anon_sym_break] = ACTIONS(1896),
    [anon_sym_continue] = ACTIONS(1896),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_STAR] = ACTIONS(1896),
    [anon_sym_STAR_COLON] = ACTIONS(1898),
    [anon_sym_import] = ACTIONS(1896),
    [anon_sym_package] = ACTIONS(1896),
    [anon_sym_AT] = ACTIONS(1896),
    [anon_sym_assert] = ACTIONS(1896),
    [anon_sym_PLUS_PLUS] = ACTIONS(1898),
    [anon_sym_DASH_DASH] = ACTIONS(1898),
    [anon_sym_SLASH] = ACTIONS(1896),
    [anon_sym_PLUS] = ACTIONS(1896),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1896),
    [anon_sym_false] = ACTIONS(1896),
    [anon_sym_ATinterface] = ACTIONS(1898),
    [anon_sym_interface] = ACTIONS(1896),
    [anon_sym_class] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_RBRACE] = ACTIONS(1898),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1896),
    [anon_sym_def] = ACTIONS(1896),
    [anon_sym_LPAREN] = ACTIONS(1898),
    [anon_sym_null] = ACTIONS(1896),
    [anon_sym_this] = ACTIONS(1896),
    [anon_sym_do] = ACTIONS(1896),
    [anon_sym_while] = ACTIONS(1896),
    [anon_sym_for] = ACTIONS(1896),
    [aux_sym__type_identifier_token1] = ACTIONS(1896),
    [anon_sym_if] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1896),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [aux_sym_number_literal_token1] = ACTIONS(1896),
    [aux_sym_number_literal_token2] = ACTIONS(1898),
    [aux_sym_number_literal_token3] = ACTIONS(1898),
    [aux_sym_number_literal_token4] = ACTIONS(1896),
    [aux_sym_number_literal_token5] = ACTIONS(1898),
    [anon_sym_return] = ACTIONS(1896),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1898),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1898),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1898),
    [anon_sym_switch] = ACTIONS(1896),
    [anon_sym_case] = ACTIONS(1896),
    [anon_sym_default] = ACTIONS(1896),
    [anon_sym_try] = ACTIONS(1896),
    [anon_sym_int] = ACTIONS(1896),
    [anon_sym_boolean] = ACTIONS(1896),
    [anon_sym_char] = ACTIONS(1896),
    [anon_sym_short] = ACTIONS(1896),
    [anon_sym_long] = ACTIONS(1896),
    [anon_sym_float] = ACTIONS(1896),
    [anon_sym_double] = ACTIONS(1896),
    [anon_sym_void] = ACTIONS(1896),
    [anon_sym_public] = ACTIONS(1896),
    [anon_sym_protected] = ACTIONS(1896),
    [anon_sym_private] = ACTIONS(1896),
    [anon_sym_static] = ACTIONS(1896),
    [anon_sym_final] = ACTIONS(1896),
    [anon_sym_synchronized] = ACTIONS(1896),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_BANG] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1896),
  },
  [STATE(780)] = {
    [sym_comment] = STATE(780),
    [sym_groovy_doc] = STATE(780),
    [ts_builtin_sym_end] = ACTIONS(1390),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_pipeline] = ACTIONS(1392),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_catch] = ACTIONS(1392),
    [anon_sym_finally] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1390),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(781)] = {
    [sym_comment] = STATE(781),
    [sym_groovy_doc] = STATE(781),
    [ts_builtin_sym_end] = ACTIONS(1868),
    [sym_identifier] = ACTIONS(1870),
    [anon_sym_break] = ACTIONS(1870),
    [anon_sym_continue] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_STAR] = ACTIONS(1870),
    [anon_sym_STAR_COLON] = ACTIONS(1868),
    [anon_sym_DOT] = ACTIONS(2063),
    [anon_sym_import] = ACTIONS(1870),
    [anon_sym_as] = ACTIONS(2065),
    [anon_sym_package] = ACTIONS(1870),
    [anon_sym_AT] = ACTIONS(1870),
    [anon_sym_assert] = ACTIONS(1870),
    [anon_sym_PLUS_PLUS] = ACTIONS(1868),
    [anon_sym_DASH_DASH] = ACTIONS(1868),
    [anon_sym_SLASH] = ACTIONS(1870),
    [anon_sym_PLUS] = ACTIONS(1870),
    [anon_sym_DASH] = ACTIONS(1870),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_ATinterface] = ACTIONS(1868),
    [anon_sym_interface] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1870),
    [anon_sym_def] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_null] = ACTIONS(1870),
    [anon_sym_this] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_while] = ACTIONS(1870),
    [anon_sym_for] = ACTIONS(1870),
    [aux_sym__type_identifier_token1] = ACTIONS(1870),
    [anon_sym_if] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1868),
    [aux_sym_number_literal_token1] = ACTIONS(1870),
    [aux_sym_number_literal_token2] = ACTIONS(1868),
    [aux_sym_number_literal_token3] = ACTIONS(1868),
    [aux_sym_number_literal_token4] = ACTIONS(1870),
    [aux_sym_number_literal_token5] = ACTIONS(1868),
    [anon_sym_pipeline] = ACTIONS(1870),
    [anon_sym_return] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1870),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1870),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1868),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1868),
    [anon_sym_switch] = ACTIONS(1870),
    [anon_sym_try] = ACTIONS(1870),
    [anon_sym_int] = ACTIONS(1870),
    [anon_sym_boolean] = ACTIONS(1870),
    [anon_sym_char] = ACTIONS(1870),
    [anon_sym_short] = ACTIONS(1870),
    [anon_sym_long] = ACTIONS(1870),
    [anon_sym_float] = ACTIONS(1870),
    [anon_sym_double] = ACTIONS(1870),
    [anon_sym_void] = ACTIONS(1870),
    [anon_sym_public] = ACTIONS(1870),
    [anon_sym_protected] = ACTIONS(1870),
    [anon_sym_private] = ACTIONS(1870),
    [anon_sym_static] = ACTIONS(1870),
    [anon_sym_final] = ACTIONS(1870),
    [anon_sym_synchronized] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1868),
    [anon_sym_BANG] = ACTIONS(1868),
    [anon_sym_new] = ACTIONS(1870),
  },
  [STATE(782)] = {
    [sym_comment] = STATE(782),
    [sym_groovy_doc] = STATE(782),
    [ts_builtin_sym_end] = ACTIONS(1783),
    [sym_identifier] = ACTIONS(1785),
    [anon_sym_break] = ACTIONS(1785),
    [anon_sym_continue] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_STAR] = ACTIONS(1785),
    [anon_sym_STAR_COLON] = ACTIONS(1783),
    [anon_sym_import] = ACTIONS(1785),
    [anon_sym_package] = ACTIONS(1785),
    [anon_sym_AT] = ACTIONS(1785),
    [anon_sym_assert] = ACTIONS(1785),
    [anon_sym_PLUS_PLUS] = ACTIONS(1783),
    [anon_sym_DASH_DASH] = ACTIONS(1783),
    [anon_sym_SLASH] = ACTIONS(1785),
    [anon_sym_PLUS] = ACTIONS(1785),
    [anon_sym_DASH] = ACTIONS(1785),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_ATinterface] = ACTIONS(1783),
    [anon_sym_interface] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_LBRACE] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1785),
    [anon_sym_def] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_RPAREN] = ACTIONS(1783),
    [anon_sym_null] = ACTIONS(1785),
    [anon_sym_this] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1785),
    [anon_sym_for] = ACTIONS(1785),
    [aux_sym__type_identifier_token1] = ACTIONS(1785),
    [anon_sym_if] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1783),
    [aux_sym_number_literal_token1] = ACTIONS(1785),
    [aux_sym_number_literal_token2] = ACTIONS(1783),
    [aux_sym_number_literal_token3] = ACTIONS(1783),
    [aux_sym_number_literal_token4] = ACTIONS(1785),
    [aux_sym_number_literal_token5] = ACTIONS(1783),
    [anon_sym_pipeline] = ACTIONS(1785),
    [anon_sym_return] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1785),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1785),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1783),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1783),
    [anon_sym_switch] = ACTIONS(1785),
    [anon_sym_try] = ACTIONS(1785),
    [anon_sym_int] = ACTIONS(1785),
    [anon_sym_boolean] = ACTIONS(1785),
    [anon_sym_char] = ACTIONS(1785),
    [anon_sym_short] = ACTIONS(1785),
    [anon_sym_long] = ACTIONS(1785),
    [anon_sym_float] = ACTIONS(1785),
    [anon_sym_double] = ACTIONS(1785),
    [anon_sym_void] = ACTIONS(1785),
    [anon_sym_public] = ACTIONS(1785),
    [anon_sym_protected] = ACTIONS(1785),
    [anon_sym_private] = ACTIONS(1785),
    [anon_sym_static] = ACTIONS(1785),
    [anon_sym_final] = ACTIONS(1785),
    [anon_sym_synchronized] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1783),
    [anon_sym_BANG] = ACTIONS(1783),
    [anon_sym_new] = ACTIONS(1785),
  },
  [STATE(783)] = {
    [sym_comment] = STATE(783),
    [sym_groovy_doc] = STATE(783),
    [ts_builtin_sym_end] = ACTIONS(1944),
    [sym_identifier] = ACTIONS(1946),
    [anon_sym_break] = ACTIONS(1946),
    [anon_sym_continue] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_STAR] = ACTIONS(1946),
    [anon_sym_STAR_COLON] = ACTIONS(1944),
    [anon_sym_import] = ACTIONS(1946),
    [anon_sym_package] = ACTIONS(1946),
    [anon_sym_AT] = ACTIONS(1946),
    [anon_sym_assert] = ACTIONS(1946),
    [anon_sym_PLUS_PLUS] = ACTIONS(1944),
    [anon_sym_DASH_DASH] = ACTIONS(1944),
    [anon_sym_SLASH] = ACTIONS(1946),
    [anon_sym_PLUS] = ACTIONS(1946),
    [anon_sym_DASH] = ACTIONS(1946),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_ATinterface] = ACTIONS(1944),
    [anon_sym_interface] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_COMMA] = ACTIONS(1944),
    [anon_sym_LBRACE] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1946),
    [anon_sym_def] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_RPAREN] = ACTIONS(1944),
    [anon_sym_null] = ACTIONS(1946),
    [anon_sym_this] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_while] = ACTIONS(1946),
    [anon_sym_for] = ACTIONS(1946),
    [aux_sym__type_identifier_token1] = ACTIONS(1946),
    [anon_sym_if] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1944),
    [aux_sym_number_literal_token1] = ACTIONS(1946),
    [aux_sym_number_literal_token2] = ACTIONS(1944),
    [aux_sym_number_literal_token3] = ACTIONS(1944),
    [aux_sym_number_literal_token4] = ACTIONS(1946),
    [aux_sym_number_literal_token5] = ACTIONS(1944),
    [anon_sym_pipeline] = ACTIONS(1946),
    [anon_sym_return] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1946),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1944),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1944),
    [anon_sym_switch] = ACTIONS(1946),
    [anon_sym_try] = ACTIONS(1946),
    [anon_sym_int] = ACTIONS(1946),
    [anon_sym_boolean] = ACTIONS(1946),
    [anon_sym_char] = ACTIONS(1946),
    [anon_sym_short] = ACTIONS(1946),
    [anon_sym_long] = ACTIONS(1946),
    [anon_sym_float] = ACTIONS(1946),
    [anon_sym_double] = ACTIONS(1946),
    [anon_sym_void] = ACTIONS(1946),
    [anon_sym_public] = ACTIONS(1946),
    [anon_sym_protected] = ACTIONS(1946),
    [anon_sym_private] = ACTIONS(1946),
    [anon_sym_static] = ACTIONS(1946),
    [anon_sym_final] = ACTIONS(1946),
    [anon_sym_synchronized] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1944),
    [anon_sym_BANG] = ACTIONS(1944),
    [anon_sym_new] = ACTIONS(1946),
  },
  [STATE(784)] = {
    [sym_comment] = STATE(784),
    [sym_groovy_doc] = STATE(784),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_COMMA] = ACTIONS(1519),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_RPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(785)] = {
    [sym_comment] = STATE(785),
    [sym_groovy_doc] = STATE(785),
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_pipeline] = ACTIONS(1412),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_catch] = ACTIONS(1412),
    [anon_sym_finally] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(786)] = {
    [sym_comment] = STATE(786),
    [sym_groovy_doc] = STATE(786),
    [sym_parameter_list] = STATE(923),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2067),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_case] = ACTIONS(1981),
    [anon_sym_default] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(787)] = {
    [sym_comment] = STATE(787),
    [sym_groovy_doc] = STATE(787),
    [sym_parameter_list] = STATE(923),
    [sym_identifier] = ACTIONS(2009),
    [anon_sym_break] = ACTIONS(2009),
    [anon_sym_continue] = ACTIONS(2009),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_STAR] = ACTIONS(2009),
    [anon_sym_STAR_COLON] = ACTIONS(2007),
    [anon_sym_import] = ACTIONS(2009),
    [anon_sym_package] = ACTIONS(2009),
    [anon_sym_AT] = ACTIONS(2009),
    [anon_sym_assert] = ACTIONS(2009),
    [anon_sym_EQ] = ACTIONS(2071),
    [anon_sym_PLUS_PLUS] = ACTIONS(2007),
    [anon_sym_DASH_DASH] = ACTIONS(2007),
    [anon_sym_SLASH] = ACTIONS(2009),
    [anon_sym_PLUS] = ACTIONS(2009),
    [anon_sym_DASH] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2009),
    [anon_sym_false] = ACTIONS(2009),
    [anon_sym_ATinterface] = ACTIONS(2007),
    [anon_sym_interface] = ACTIONS(2009),
    [anon_sym_class] = ACTIONS(2009),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_RBRACE] = ACTIONS(2007),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2009),
    [anon_sym_def] = ACTIONS(2009),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(2009),
    [anon_sym_this] = ACTIONS(2009),
    [anon_sym_do] = ACTIONS(2009),
    [anon_sym_while] = ACTIONS(2009),
    [anon_sym_for] = ACTIONS(2009),
    [aux_sym__type_identifier_token1] = ACTIONS(2009),
    [anon_sym_if] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [aux_sym_number_literal_token1] = ACTIONS(2009),
    [aux_sym_number_literal_token2] = ACTIONS(2007),
    [aux_sym_number_literal_token3] = ACTIONS(2007),
    [aux_sym_number_literal_token4] = ACTIONS(2009),
    [aux_sym_number_literal_token5] = ACTIONS(2007),
    [anon_sym_return] = ACTIONS(2009),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2007),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2007),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2007),
    [anon_sym_switch] = ACTIONS(2009),
    [anon_sym_case] = ACTIONS(2009),
    [anon_sym_default] = ACTIONS(2009),
    [anon_sym_try] = ACTIONS(2009),
    [anon_sym_int] = ACTIONS(2009),
    [anon_sym_boolean] = ACTIONS(2009),
    [anon_sym_char] = ACTIONS(2009),
    [anon_sym_short] = ACTIONS(2009),
    [anon_sym_long] = ACTIONS(2009),
    [anon_sym_float] = ACTIONS(2009),
    [anon_sym_double] = ACTIONS(2009),
    [anon_sym_void] = ACTIONS(2009),
    [anon_sym_public] = ACTIONS(2009),
    [anon_sym_protected] = ACTIONS(2009),
    [anon_sym_private] = ACTIONS(2009),
    [anon_sym_static] = ACTIONS(2009),
    [anon_sym_final] = ACTIONS(2009),
    [anon_sym_synchronized] = ACTIONS(2009),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_BANG] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2009),
  },
  [STATE(788)] = {
    [sym_comment] = STATE(788),
    [sym_groovy_doc] = STATE(788),
    [ts_builtin_sym_end] = ACTIONS(1952),
    [sym_identifier] = ACTIONS(1954),
    [anon_sym_break] = ACTIONS(1954),
    [anon_sym_continue] = ACTIONS(1954),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_STAR] = ACTIONS(1954),
    [anon_sym_STAR_COLON] = ACTIONS(1952),
    [anon_sym_import] = ACTIONS(1954),
    [anon_sym_package] = ACTIONS(1954),
    [anon_sym_AT] = ACTIONS(1954),
    [anon_sym_assert] = ACTIONS(1954),
    [anon_sym_PLUS_PLUS] = ACTIONS(1952),
    [anon_sym_DASH_DASH] = ACTIONS(1952),
    [anon_sym_SLASH] = ACTIONS(1954),
    [anon_sym_PLUS] = ACTIONS(1954),
    [anon_sym_DASH] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(1954),
    [anon_sym_false] = ACTIONS(1954),
    [anon_sym_ATinterface] = ACTIONS(1952),
    [anon_sym_interface] = ACTIONS(1954),
    [anon_sym_class] = ACTIONS(1954),
    [anon_sym_COMMA] = ACTIONS(1952),
    [anon_sym_LBRACE] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1954),
    [anon_sym_def] = ACTIONS(1954),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_RPAREN] = ACTIONS(1952),
    [anon_sym_null] = ACTIONS(1954),
    [anon_sym_this] = ACTIONS(1954),
    [anon_sym_do] = ACTIONS(1954),
    [anon_sym_while] = ACTIONS(1954),
    [anon_sym_for] = ACTIONS(1954),
    [aux_sym__type_identifier_token1] = ACTIONS(1954),
    [anon_sym_if] = ACTIONS(1954),
    [anon_sym_else] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(1952),
    [aux_sym_number_literal_token1] = ACTIONS(1954),
    [aux_sym_number_literal_token2] = ACTIONS(1952),
    [aux_sym_number_literal_token3] = ACTIONS(1952),
    [aux_sym_number_literal_token4] = ACTIONS(1954),
    [aux_sym_number_literal_token5] = ACTIONS(1952),
    [anon_sym_pipeline] = ACTIONS(1954),
    [anon_sym_return] = ACTIONS(1954),
    [anon_sym_SQUOTE] = ACTIONS(1954),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1954),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1952),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1952),
    [anon_sym_switch] = ACTIONS(1954),
    [anon_sym_try] = ACTIONS(1954),
    [anon_sym_int] = ACTIONS(1954),
    [anon_sym_boolean] = ACTIONS(1954),
    [anon_sym_char] = ACTIONS(1954),
    [anon_sym_short] = ACTIONS(1954),
    [anon_sym_long] = ACTIONS(1954),
    [anon_sym_float] = ACTIONS(1954),
    [anon_sym_double] = ACTIONS(1954),
    [anon_sym_void] = ACTIONS(1954),
    [anon_sym_public] = ACTIONS(1954),
    [anon_sym_protected] = ACTIONS(1954),
    [anon_sym_private] = ACTIONS(1954),
    [anon_sym_static] = ACTIONS(1954),
    [anon_sym_final] = ACTIONS(1954),
    [anon_sym_synchronized] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(1952),
    [anon_sym_BANG] = ACTIONS(1952),
    [anon_sym_new] = ACTIONS(1954),
  },
  [STATE(789)] = {
    [sym_comment] = STATE(789),
    [sym_groovy_doc] = STATE(789),
    [sym_parameter_list] = STATE(929),
    [sym_identifier] = ACTIONS(1761),
    [anon_sym_break] = ACTIONS(1761),
    [anon_sym_continue] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_STAR] = ACTIONS(1761),
    [anon_sym_STAR_COLON] = ACTIONS(1763),
    [anon_sym_import] = ACTIONS(1761),
    [anon_sym_package] = ACTIONS(1761),
    [anon_sym_AT] = ACTIONS(1761),
    [anon_sym_assert] = ACTIONS(1761),
    [anon_sym_EQ] = ACTIONS(2073),
    [anon_sym_PLUS_PLUS] = ACTIONS(1763),
    [anon_sym_DASH_DASH] = ACTIONS(1763),
    [anon_sym_SLASH] = ACTIONS(1761),
    [anon_sym_PLUS] = ACTIONS(1761),
    [anon_sym_DASH] = ACTIONS(1761),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_ATinterface] = ACTIONS(1763),
    [anon_sym_interface] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1761),
    [anon_sym_def] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1761),
    [anon_sym_this] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_while] = ACTIONS(1761),
    [anon_sym_for] = ACTIONS(1761),
    [aux_sym__type_identifier_token1] = ACTIONS(1761),
    [anon_sym_if] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1763),
    [aux_sym_number_literal_token1] = ACTIONS(1761),
    [aux_sym_number_literal_token2] = ACTIONS(1763),
    [aux_sym_number_literal_token3] = ACTIONS(1763),
    [aux_sym_number_literal_token4] = ACTIONS(1761),
    [aux_sym_number_literal_token5] = ACTIONS(1763),
    [anon_sym_return] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1761),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1761),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1763),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1763),
    [anon_sym_switch] = ACTIONS(1761),
    [anon_sym_case] = ACTIONS(1761),
    [anon_sym_default] = ACTIONS(1761),
    [anon_sym_try] = ACTIONS(1761),
    [anon_sym_int] = ACTIONS(1761),
    [anon_sym_boolean] = ACTIONS(1761),
    [anon_sym_char] = ACTIONS(1761),
    [anon_sym_short] = ACTIONS(1761),
    [anon_sym_long] = ACTIONS(1761),
    [anon_sym_float] = ACTIONS(1761),
    [anon_sym_double] = ACTIONS(1761),
    [anon_sym_void] = ACTIONS(1761),
    [anon_sym_public] = ACTIONS(1761),
    [anon_sym_protected] = ACTIONS(1761),
    [anon_sym_private] = ACTIONS(1761),
    [anon_sym_static] = ACTIONS(1761),
    [anon_sym_final] = ACTIONS(1761),
    [anon_sym_synchronized] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1763),
    [anon_sym_BANG] = ACTIONS(1763),
    [anon_sym_new] = ACTIONS(1761),
  },
  [STATE(790)] = {
    [sym_comment] = STATE(790),
    [sym_groovy_doc] = STATE(790),
    [sym_parameter_list] = STATE(929),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_break] = ACTIONS(1751),
    [anon_sym_continue] = ACTIONS(1751),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_STAR] = ACTIONS(1751),
    [anon_sym_STAR_COLON] = ACTIONS(1753),
    [anon_sym_import] = ACTIONS(1751),
    [anon_sym_package] = ACTIONS(1751),
    [anon_sym_AT] = ACTIONS(1751),
    [anon_sym_assert] = ACTIONS(1751),
    [anon_sym_EQ] = ACTIONS(2075),
    [anon_sym_PLUS_PLUS] = ACTIONS(1753),
    [anon_sym_DASH_DASH] = ACTIONS(1753),
    [anon_sym_SLASH] = ACTIONS(1751),
    [anon_sym_PLUS] = ACTIONS(1751),
    [anon_sym_DASH] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1751),
    [anon_sym_false] = ACTIONS(1751),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_interface] = ACTIONS(1751),
    [anon_sym_class] = ACTIONS(1751),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_RBRACE] = ACTIONS(1753),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1751),
    [anon_sym_def] = ACTIONS(1751),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1751),
    [anon_sym_this] = ACTIONS(1751),
    [anon_sym_do] = ACTIONS(1751),
    [anon_sym_while] = ACTIONS(1751),
    [anon_sym_for] = ACTIONS(1751),
    [aux_sym__type_identifier_token1] = ACTIONS(1751),
    [anon_sym_if] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [aux_sym_number_literal_token1] = ACTIONS(1751),
    [aux_sym_number_literal_token2] = ACTIONS(1753),
    [aux_sym_number_literal_token3] = ACTIONS(1753),
    [aux_sym_number_literal_token4] = ACTIONS(1751),
    [aux_sym_number_literal_token5] = ACTIONS(1753),
    [anon_sym_return] = ACTIONS(1751),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1753),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1753),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1753),
    [anon_sym_switch] = ACTIONS(1751),
    [anon_sym_case] = ACTIONS(1751),
    [anon_sym_default] = ACTIONS(1751),
    [anon_sym_try] = ACTIONS(1751),
    [anon_sym_int] = ACTIONS(1751),
    [anon_sym_boolean] = ACTIONS(1751),
    [anon_sym_char] = ACTIONS(1751),
    [anon_sym_short] = ACTIONS(1751),
    [anon_sym_long] = ACTIONS(1751),
    [anon_sym_float] = ACTIONS(1751),
    [anon_sym_double] = ACTIONS(1751),
    [anon_sym_void] = ACTIONS(1751),
    [anon_sym_public] = ACTIONS(1751),
    [anon_sym_protected] = ACTIONS(1751),
    [anon_sym_private] = ACTIONS(1751),
    [anon_sym_static] = ACTIONS(1751),
    [anon_sym_final] = ACTIONS(1751),
    [anon_sym_synchronized] = ACTIONS(1751),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_BANG] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1751),
  },
  [STATE(791)] = {
    [sym_comment] = STATE(791),
    [sym_groovy_doc] = STATE(791),
    [ts_builtin_sym_end] = ACTIONS(2013),
    [sym_identifier] = ACTIONS(2015),
    [anon_sym_break] = ACTIONS(2015),
    [anon_sym_continue] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_STAR] = ACTIONS(2015),
    [anon_sym_STAR_COLON] = ACTIONS(2013),
    [anon_sym_import] = ACTIONS(2015),
    [anon_sym_package] = ACTIONS(2015),
    [anon_sym_AT] = ACTIONS(2015),
    [anon_sym_assert] = ACTIONS(2015),
    [anon_sym_PLUS_PLUS] = ACTIONS(2013),
    [anon_sym_DASH_DASH] = ACTIONS(2013),
    [anon_sym_SLASH] = ACTIONS(2015),
    [anon_sym_PLUS] = ACTIONS(2015),
    [anon_sym_DASH] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_ATinterface] = ACTIONS(2013),
    [anon_sym_interface] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_COMMA] = ACTIONS(2013),
    [anon_sym_LBRACE] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2015),
    [anon_sym_def] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_RPAREN] = ACTIONS(2013),
    [anon_sym_null] = ACTIONS(2015),
    [anon_sym_this] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_while] = ACTIONS(2015),
    [anon_sym_for] = ACTIONS(2015),
    [aux_sym__type_identifier_token1] = ACTIONS(2015),
    [anon_sym_if] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2013),
    [aux_sym_number_literal_token1] = ACTIONS(2015),
    [aux_sym_number_literal_token2] = ACTIONS(2013),
    [aux_sym_number_literal_token3] = ACTIONS(2013),
    [aux_sym_number_literal_token4] = ACTIONS(2015),
    [aux_sym_number_literal_token5] = ACTIONS(2013),
    [anon_sym_pipeline] = ACTIONS(2015),
    [anon_sym_return] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2015),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2015),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2013),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2013),
    [anon_sym_switch] = ACTIONS(2015),
    [anon_sym_try] = ACTIONS(2015),
    [anon_sym_int] = ACTIONS(2015),
    [anon_sym_boolean] = ACTIONS(2015),
    [anon_sym_char] = ACTIONS(2015),
    [anon_sym_short] = ACTIONS(2015),
    [anon_sym_long] = ACTIONS(2015),
    [anon_sym_float] = ACTIONS(2015),
    [anon_sym_double] = ACTIONS(2015),
    [anon_sym_void] = ACTIONS(2015),
    [anon_sym_public] = ACTIONS(2015),
    [anon_sym_protected] = ACTIONS(2015),
    [anon_sym_private] = ACTIONS(2015),
    [anon_sym_static] = ACTIONS(2015),
    [anon_sym_final] = ACTIONS(2015),
    [anon_sym_synchronized] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2013),
    [anon_sym_BANG] = ACTIONS(2013),
    [anon_sym_new] = ACTIONS(2015),
  },
  [STATE(792)] = {
    [sym_comment] = STATE(792),
    [sym_groovy_doc] = STATE(792),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(2077),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_case] = ACTIONS(1906),
    [anon_sym_default] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_catch] = ACTIONS(1906),
    [anon_sym_finally] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(793)] = {
    [sym_comment] = STATE(793),
    [sym_groovy_doc] = STATE(793),
    [sym_identifier] = ACTIONS(1769),
    [anon_sym_break] = ACTIONS(1769),
    [anon_sym_continue] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_STAR] = ACTIONS(1769),
    [anon_sym_STAR_COLON] = ACTIONS(1767),
    [anon_sym_import] = ACTIONS(1769),
    [anon_sym_package] = ACTIONS(1769),
    [anon_sym_AT] = ACTIONS(1769),
    [anon_sym_assert] = ACTIONS(1769),
    [anon_sym_PLUS_PLUS] = ACTIONS(1767),
    [anon_sym_DASH_DASH] = ACTIONS(1767),
    [anon_sym_SLASH] = ACTIONS(1769),
    [anon_sym_PLUS] = ACTIONS(1769),
    [anon_sym_DASH] = ACTIONS(1769),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_ATinterface] = ACTIONS(1767),
    [anon_sym_interface] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1769),
    [anon_sym_def] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_null] = ACTIONS(1769),
    [anon_sym_this] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_while] = ACTIONS(1769),
    [anon_sym_for] = ACTIONS(1769),
    [aux_sym__type_identifier_token1] = ACTIONS(1769),
    [anon_sym_if] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [aux_sym_number_literal_token1] = ACTIONS(1769),
    [aux_sym_number_literal_token2] = ACTIONS(1767),
    [aux_sym_number_literal_token3] = ACTIONS(1767),
    [aux_sym_number_literal_token4] = ACTIONS(1769),
    [aux_sym_number_literal_token5] = ACTIONS(1767),
    [anon_sym_return] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1769),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1769),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1767),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1767),
    [anon_sym_switch] = ACTIONS(1769),
    [anon_sym_case] = ACTIONS(1769),
    [anon_sym_default] = ACTIONS(1769),
    [anon_sym_try] = ACTIONS(1769),
    [anon_sym_catch] = ACTIONS(1769),
    [anon_sym_finally] = ACTIONS(1769),
    [anon_sym_int] = ACTIONS(1769),
    [anon_sym_boolean] = ACTIONS(1769),
    [anon_sym_char] = ACTIONS(1769),
    [anon_sym_short] = ACTIONS(1769),
    [anon_sym_long] = ACTIONS(1769),
    [anon_sym_float] = ACTIONS(1769),
    [anon_sym_double] = ACTIONS(1769),
    [anon_sym_void] = ACTIONS(1769),
    [anon_sym_public] = ACTIONS(1769),
    [anon_sym_protected] = ACTIONS(1769),
    [anon_sym_private] = ACTIONS(1769),
    [anon_sym_static] = ACTIONS(1769),
    [anon_sym_final] = ACTIONS(1769),
    [anon_sym_synchronized] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1767),
    [anon_sym_BANG] = ACTIONS(1767),
    [anon_sym_new] = ACTIONS(1769),
  },
  [STATE(794)] = {
    [sym_comment] = STATE(794),
    [sym_groovy_doc] = STATE(794),
    [ts_builtin_sym_end] = ACTIONS(2021),
    [sym_identifier] = ACTIONS(2023),
    [anon_sym_break] = ACTIONS(2023),
    [anon_sym_continue] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_STAR] = ACTIONS(2023),
    [anon_sym_STAR_COLON] = ACTIONS(2021),
    [anon_sym_import] = ACTIONS(2023),
    [anon_sym_package] = ACTIONS(2023),
    [anon_sym_AT] = ACTIONS(2023),
    [anon_sym_assert] = ACTIONS(2023),
    [anon_sym_PLUS_PLUS] = ACTIONS(2021),
    [anon_sym_DASH_DASH] = ACTIONS(2021),
    [anon_sym_SLASH] = ACTIONS(2023),
    [anon_sym_PLUS] = ACTIONS(2023),
    [anon_sym_DASH] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_ATinterface] = ACTIONS(2021),
    [anon_sym_interface] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_COMMA] = ACTIONS(2021),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2023),
    [anon_sym_def] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_RPAREN] = ACTIONS(2021),
    [anon_sym_null] = ACTIONS(2023),
    [anon_sym_this] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_while] = ACTIONS(2023),
    [anon_sym_for] = ACTIONS(2023),
    [aux_sym__type_identifier_token1] = ACTIONS(2023),
    [anon_sym_if] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [aux_sym_number_literal_token1] = ACTIONS(2023),
    [aux_sym_number_literal_token2] = ACTIONS(2021),
    [aux_sym_number_literal_token3] = ACTIONS(2021),
    [aux_sym_number_literal_token4] = ACTIONS(2023),
    [aux_sym_number_literal_token5] = ACTIONS(2021),
    [anon_sym_pipeline] = ACTIONS(2023),
    [anon_sym_return] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2021),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2021),
    [anon_sym_switch] = ACTIONS(2023),
    [anon_sym_try] = ACTIONS(2023),
    [anon_sym_int] = ACTIONS(2023),
    [anon_sym_boolean] = ACTIONS(2023),
    [anon_sym_char] = ACTIONS(2023),
    [anon_sym_short] = ACTIONS(2023),
    [anon_sym_long] = ACTIONS(2023),
    [anon_sym_float] = ACTIONS(2023),
    [anon_sym_double] = ACTIONS(2023),
    [anon_sym_void] = ACTIONS(2023),
    [anon_sym_public] = ACTIONS(2023),
    [anon_sym_protected] = ACTIONS(2023),
    [anon_sym_private] = ACTIONS(2023),
    [anon_sym_static] = ACTIONS(2023),
    [anon_sym_final] = ACTIONS(2023),
    [anon_sym_synchronized] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_BANG] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2023),
  },
  [STATE(795)] = {
    [sym_comment] = STATE(795),
    [sym_groovy_doc] = STATE(795),
    [aux_sym__juxt_argument_list_repeat1] = STATE(903),
    [ts_builtin_sym_end] = ACTIONS(1898),
    [sym_identifier] = ACTIONS(1896),
    [anon_sym_break] = ACTIONS(1896),
    [anon_sym_continue] = ACTIONS(1896),
    [anon_sym_SEMI] = ACTIONS(1898),
    [anon_sym_STAR] = ACTIONS(1896),
    [anon_sym_STAR_COLON] = ACTIONS(1898),
    [anon_sym_import] = ACTIONS(1896),
    [anon_sym_package] = ACTIONS(1896),
    [anon_sym_AT] = ACTIONS(1896),
    [anon_sym_assert] = ACTIONS(1896),
    [anon_sym_PLUS_PLUS] = ACTIONS(1898),
    [anon_sym_DASH_DASH] = ACTIONS(1898),
    [anon_sym_SLASH] = ACTIONS(1896),
    [anon_sym_PLUS] = ACTIONS(1896),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1896),
    [anon_sym_false] = ACTIONS(1896),
    [anon_sym_ATinterface] = ACTIONS(1898),
    [anon_sym_interface] = ACTIONS(1896),
    [anon_sym_class] = ACTIONS(1896),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1898),
    [anon_sym_RBRACE] = ACTIONS(1898),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1896),
    [anon_sym_def] = ACTIONS(1896),
    [anon_sym_LPAREN] = ACTIONS(1898),
    [anon_sym_null] = ACTIONS(1896),
    [anon_sym_this] = ACTIONS(1896),
    [anon_sym_do] = ACTIONS(1896),
    [anon_sym_while] = ACTIONS(1896),
    [anon_sym_for] = ACTIONS(1896),
    [aux_sym__type_identifier_token1] = ACTIONS(1896),
    [anon_sym_if] = ACTIONS(1896),
    [anon_sym_else] = ACTIONS(1896),
    [anon_sym_LBRACK] = ACTIONS(1898),
    [aux_sym_number_literal_token1] = ACTIONS(1896),
    [aux_sym_number_literal_token2] = ACTIONS(1898),
    [aux_sym_number_literal_token3] = ACTIONS(1898),
    [aux_sym_number_literal_token4] = ACTIONS(1896),
    [aux_sym_number_literal_token5] = ACTIONS(1898),
    [anon_sym_pipeline] = ACTIONS(1896),
    [anon_sym_return] = ACTIONS(1896),
    [anon_sym_SQUOTE] = ACTIONS(1896),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1898),
    [anon_sym_DQUOTE] = ACTIONS(1896),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1898),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1898),
    [anon_sym_switch] = ACTIONS(1896),
    [anon_sym_try] = ACTIONS(1896),
    [anon_sym_int] = ACTIONS(1896),
    [anon_sym_boolean] = ACTIONS(1896),
    [anon_sym_char] = ACTIONS(1896),
    [anon_sym_short] = ACTIONS(1896),
    [anon_sym_long] = ACTIONS(1896),
    [anon_sym_float] = ACTIONS(1896),
    [anon_sym_double] = ACTIONS(1896),
    [anon_sym_void] = ACTIONS(1896),
    [anon_sym_public] = ACTIONS(1896),
    [anon_sym_protected] = ACTIONS(1896),
    [anon_sym_private] = ACTIONS(1896),
    [anon_sym_static] = ACTIONS(1896),
    [anon_sym_final] = ACTIONS(1896),
    [anon_sym_synchronized] = ACTIONS(1896),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_BANG] = ACTIONS(1898),
    [anon_sym_new] = ACTIONS(1896),
  },
  [STATE(796)] = {
    [sym_comment] = STATE(796),
    [sym_groovy_doc] = STATE(796),
    [ts_builtin_sym_end] = ACTIONS(1796),
    [sym_identifier] = ACTIONS(1798),
    [anon_sym_break] = ACTIONS(1798),
    [anon_sym_continue] = ACTIONS(1798),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym_STAR] = ACTIONS(1798),
    [anon_sym_STAR_COLON] = ACTIONS(1796),
    [anon_sym_import] = ACTIONS(1798),
    [anon_sym_package] = ACTIONS(1798),
    [anon_sym_AT] = ACTIONS(1798),
    [anon_sym_assert] = ACTIONS(1798),
    [anon_sym_PLUS_PLUS] = ACTIONS(1796),
    [anon_sym_DASH_DASH] = ACTIONS(1796),
    [anon_sym_SLASH] = ACTIONS(1798),
    [anon_sym_PLUS] = ACTIONS(1798),
    [anon_sym_DASH] = ACTIONS(1798),
    [anon_sym_true] = ACTIONS(1798),
    [anon_sym_false] = ACTIONS(1798),
    [anon_sym_ATinterface] = ACTIONS(1796),
    [anon_sym_interface] = ACTIONS(1798),
    [anon_sym_class] = ACTIONS(1798),
    [anon_sym_COMMA] = ACTIONS(1796),
    [anon_sym_LBRACE] = ACTIONS(1796),
    [anon_sym_RBRACE] = ACTIONS(1796),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1798),
    [anon_sym_def] = ACTIONS(1798),
    [anon_sym_LPAREN] = ACTIONS(1796),
    [anon_sym_RPAREN] = ACTIONS(1796),
    [anon_sym_null] = ACTIONS(1798),
    [anon_sym_this] = ACTIONS(1798),
    [anon_sym_do] = ACTIONS(1798),
    [anon_sym_while] = ACTIONS(1798),
    [anon_sym_for] = ACTIONS(1798),
    [aux_sym__type_identifier_token1] = ACTIONS(1798),
    [anon_sym_if] = ACTIONS(1798),
    [anon_sym_else] = ACTIONS(1798),
    [anon_sym_LBRACK] = ACTIONS(1796),
    [aux_sym_number_literal_token1] = ACTIONS(1798),
    [aux_sym_number_literal_token2] = ACTIONS(1796),
    [aux_sym_number_literal_token3] = ACTIONS(1796),
    [aux_sym_number_literal_token4] = ACTIONS(1798),
    [aux_sym_number_literal_token5] = ACTIONS(1796),
    [anon_sym_pipeline] = ACTIONS(1798),
    [anon_sym_return] = ACTIONS(1798),
    [anon_sym_SQUOTE] = ACTIONS(1798),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1796),
    [anon_sym_DQUOTE] = ACTIONS(1798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1796),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1796),
    [anon_sym_switch] = ACTIONS(1798),
    [anon_sym_try] = ACTIONS(1798),
    [anon_sym_int] = ACTIONS(1798),
    [anon_sym_boolean] = ACTIONS(1798),
    [anon_sym_char] = ACTIONS(1798),
    [anon_sym_short] = ACTIONS(1798),
    [anon_sym_long] = ACTIONS(1798),
    [anon_sym_float] = ACTIONS(1798),
    [anon_sym_double] = ACTIONS(1798),
    [anon_sym_void] = ACTIONS(1798),
    [anon_sym_public] = ACTIONS(1798),
    [anon_sym_protected] = ACTIONS(1798),
    [anon_sym_private] = ACTIONS(1798),
    [anon_sym_static] = ACTIONS(1798),
    [anon_sym_final] = ACTIONS(1798),
    [anon_sym_synchronized] = ACTIONS(1798),
    [anon_sym_TILDE] = ACTIONS(1796),
    [anon_sym_BANG] = ACTIONS(1796),
    [anon_sym_new] = ACTIONS(1798),
  },
  [STATE(797)] = {
    [sym_comment] = STATE(797),
    [sym_groovy_doc] = STATE(797),
    [sym_parameter_list] = STATE(934),
    [ts_builtin_sym_end] = ACTIONS(1771),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(2079),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_pipeline] = ACTIONS(1773),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(798)] = {
    [sym_comment] = STATE(798),
    [sym_groovy_doc] = STATE(798),
    [ts_builtin_sym_end] = ACTIONS(1800),
    [sym_identifier] = ACTIONS(1802),
    [anon_sym_break] = ACTIONS(1802),
    [anon_sym_continue] = ACTIONS(1802),
    [anon_sym_SEMI] = ACTIONS(1800),
    [anon_sym_STAR] = ACTIONS(1802),
    [anon_sym_STAR_COLON] = ACTIONS(1800),
    [anon_sym_import] = ACTIONS(1802),
    [anon_sym_package] = ACTIONS(1802),
    [anon_sym_AT] = ACTIONS(1802),
    [anon_sym_assert] = ACTIONS(1802),
    [anon_sym_PLUS_PLUS] = ACTIONS(1800),
    [anon_sym_DASH_DASH] = ACTIONS(1800),
    [anon_sym_SLASH] = ACTIONS(1802),
    [anon_sym_PLUS] = ACTIONS(1802),
    [anon_sym_DASH] = ACTIONS(1802),
    [anon_sym_true] = ACTIONS(1802),
    [anon_sym_false] = ACTIONS(1802),
    [anon_sym_ATinterface] = ACTIONS(1800),
    [anon_sym_interface] = ACTIONS(1802),
    [anon_sym_class] = ACTIONS(1802),
    [anon_sym_COMMA] = ACTIONS(1800),
    [anon_sym_LBRACE] = ACTIONS(1800),
    [anon_sym_RBRACE] = ACTIONS(1800),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1802),
    [anon_sym_def] = ACTIONS(1802),
    [anon_sym_LPAREN] = ACTIONS(1800),
    [anon_sym_RPAREN] = ACTIONS(1800),
    [anon_sym_null] = ACTIONS(1802),
    [anon_sym_this] = ACTIONS(1802),
    [anon_sym_do] = ACTIONS(1802),
    [anon_sym_while] = ACTIONS(1802),
    [anon_sym_for] = ACTIONS(1802),
    [aux_sym__type_identifier_token1] = ACTIONS(1802),
    [anon_sym_if] = ACTIONS(1802),
    [anon_sym_else] = ACTIONS(1802),
    [anon_sym_LBRACK] = ACTIONS(1800),
    [aux_sym_number_literal_token1] = ACTIONS(1802),
    [aux_sym_number_literal_token2] = ACTIONS(1800),
    [aux_sym_number_literal_token3] = ACTIONS(1800),
    [aux_sym_number_literal_token4] = ACTIONS(1802),
    [aux_sym_number_literal_token5] = ACTIONS(1800),
    [anon_sym_pipeline] = ACTIONS(1802),
    [anon_sym_return] = ACTIONS(1802),
    [anon_sym_SQUOTE] = ACTIONS(1802),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1800),
    [anon_sym_DQUOTE] = ACTIONS(1802),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1800),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1800),
    [anon_sym_switch] = ACTIONS(1802),
    [anon_sym_try] = ACTIONS(1802),
    [anon_sym_int] = ACTIONS(1802),
    [anon_sym_boolean] = ACTIONS(1802),
    [anon_sym_char] = ACTIONS(1802),
    [anon_sym_short] = ACTIONS(1802),
    [anon_sym_long] = ACTIONS(1802),
    [anon_sym_float] = ACTIONS(1802),
    [anon_sym_double] = ACTIONS(1802),
    [anon_sym_void] = ACTIONS(1802),
    [anon_sym_public] = ACTIONS(1802),
    [anon_sym_protected] = ACTIONS(1802),
    [anon_sym_private] = ACTIONS(1802),
    [anon_sym_static] = ACTIONS(1802),
    [anon_sym_final] = ACTIONS(1802),
    [anon_sym_synchronized] = ACTIONS(1802),
    [anon_sym_TILDE] = ACTIONS(1800),
    [anon_sym_BANG] = ACTIONS(1800),
    [anon_sym_new] = ACTIONS(1802),
  },
  [STATE(799)] = {
    [sym_comment] = STATE(799),
    [sym_groovy_doc] = STATE(799),
    [sym_identifier] = ACTIONS(1781),
    [anon_sym_break] = ACTIONS(1781),
    [anon_sym_continue] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_STAR] = ACTIONS(1781),
    [anon_sym_STAR_COLON] = ACTIONS(1779),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_import] = ACTIONS(1781),
    [anon_sym_as] = ACTIONS(1781),
    [anon_sym_package] = ACTIONS(1781),
    [anon_sym_AT] = ACTIONS(1781),
    [anon_sym_assert] = ACTIONS(1781),
    [anon_sym_PLUS_PLUS] = ACTIONS(1779),
    [anon_sym_DASH_DASH] = ACTIONS(1779),
    [anon_sym_SLASH] = ACTIONS(1781),
    [anon_sym_PLUS] = ACTIONS(1781),
    [anon_sym_DASH] = ACTIONS(1781),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_ATinterface] = ACTIONS(1779),
    [anon_sym_interface] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1781),
    [anon_sym_def] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_null] = ACTIONS(1781),
    [anon_sym_this] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_while] = ACTIONS(1781),
    [anon_sym_for] = ACTIONS(1781),
    [aux_sym__type_identifier_token1] = ACTIONS(1781),
    [anon_sym_if] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1779),
    [aux_sym_number_literal_token1] = ACTIONS(1781),
    [aux_sym_number_literal_token2] = ACTIONS(1779),
    [aux_sym_number_literal_token3] = ACTIONS(1779),
    [aux_sym_number_literal_token4] = ACTIONS(1781),
    [aux_sym_number_literal_token5] = ACTIONS(1779),
    [anon_sym_return] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1781),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1781),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1779),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1779),
    [anon_sym_switch] = ACTIONS(1781),
    [anon_sym_case] = ACTIONS(1781),
    [anon_sym_default] = ACTIONS(1781),
    [anon_sym_try] = ACTIONS(1781),
    [anon_sym_int] = ACTIONS(1781),
    [anon_sym_boolean] = ACTIONS(1781),
    [anon_sym_char] = ACTIONS(1781),
    [anon_sym_short] = ACTIONS(1781),
    [anon_sym_long] = ACTIONS(1781),
    [anon_sym_float] = ACTIONS(1781),
    [anon_sym_double] = ACTIONS(1781),
    [anon_sym_void] = ACTIONS(1781),
    [anon_sym_public] = ACTIONS(1781),
    [anon_sym_protected] = ACTIONS(1781),
    [anon_sym_private] = ACTIONS(1781),
    [anon_sym_static] = ACTIONS(1781),
    [anon_sym_final] = ACTIONS(1781),
    [anon_sym_synchronized] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1779),
    [anon_sym_BANG] = ACTIONS(1779),
    [anon_sym_new] = ACTIONS(1781),
  },
  [STATE(800)] = {
    [sym_comment] = STATE(800),
    [sym_groovy_doc] = STATE(800),
    [sym_parameter_list] = STATE(927),
    [ts_builtin_sym_end] = ACTIONS(1763),
    [sym_identifier] = ACTIONS(1761),
    [anon_sym_break] = ACTIONS(1761),
    [anon_sym_continue] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_STAR] = ACTIONS(1761),
    [anon_sym_STAR_COLON] = ACTIONS(1763),
    [anon_sym_import] = ACTIONS(1761),
    [anon_sym_package] = ACTIONS(1761),
    [anon_sym_AT] = ACTIONS(1761),
    [anon_sym_assert] = ACTIONS(1761),
    [anon_sym_EQ] = ACTIONS(2083),
    [anon_sym_PLUS_PLUS] = ACTIONS(1763),
    [anon_sym_DASH_DASH] = ACTIONS(1763),
    [anon_sym_SLASH] = ACTIONS(1761),
    [anon_sym_PLUS] = ACTIONS(1761),
    [anon_sym_DASH] = ACTIONS(1761),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_ATinterface] = ACTIONS(1763),
    [anon_sym_interface] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1761),
    [anon_sym_def] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1761),
    [anon_sym_this] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_while] = ACTIONS(1761),
    [anon_sym_for] = ACTIONS(1761),
    [aux_sym__type_identifier_token1] = ACTIONS(1761),
    [anon_sym_if] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1763),
    [aux_sym_number_literal_token1] = ACTIONS(1761),
    [aux_sym_number_literal_token2] = ACTIONS(1763),
    [aux_sym_number_literal_token3] = ACTIONS(1763),
    [aux_sym_number_literal_token4] = ACTIONS(1761),
    [aux_sym_number_literal_token5] = ACTIONS(1763),
    [anon_sym_pipeline] = ACTIONS(1761),
    [anon_sym_return] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1761),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1761),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1763),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1763),
    [anon_sym_switch] = ACTIONS(1761),
    [anon_sym_try] = ACTIONS(1761),
    [anon_sym_int] = ACTIONS(1761),
    [anon_sym_boolean] = ACTIONS(1761),
    [anon_sym_char] = ACTIONS(1761),
    [anon_sym_short] = ACTIONS(1761),
    [anon_sym_long] = ACTIONS(1761),
    [anon_sym_float] = ACTIONS(1761),
    [anon_sym_double] = ACTIONS(1761),
    [anon_sym_void] = ACTIONS(1761),
    [anon_sym_public] = ACTIONS(1761),
    [anon_sym_protected] = ACTIONS(1761),
    [anon_sym_private] = ACTIONS(1761),
    [anon_sym_static] = ACTIONS(1761),
    [anon_sym_final] = ACTIONS(1761),
    [anon_sym_synchronized] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1763),
    [anon_sym_BANG] = ACTIONS(1763),
    [anon_sym_new] = ACTIONS(1761),
  },
  [STATE(801)] = {
    [sym_comment] = STATE(801),
    [sym_groovy_doc] = STATE(801),
    [sym_parameter_list] = STATE(927),
    [ts_builtin_sym_end] = ACTIONS(1753),
    [sym_identifier] = ACTIONS(1751),
    [anon_sym_break] = ACTIONS(1751),
    [anon_sym_continue] = ACTIONS(1751),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_STAR] = ACTIONS(1751),
    [anon_sym_STAR_COLON] = ACTIONS(1753),
    [anon_sym_import] = ACTIONS(1751),
    [anon_sym_package] = ACTIONS(1751),
    [anon_sym_AT] = ACTIONS(1751),
    [anon_sym_assert] = ACTIONS(1751),
    [anon_sym_EQ] = ACTIONS(2085),
    [anon_sym_PLUS_PLUS] = ACTIONS(1753),
    [anon_sym_DASH_DASH] = ACTIONS(1753),
    [anon_sym_SLASH] = ACTIONS(1751),
    [anon_sym_PLUS] = ACTIONS(1751),
    [anon_sym_DASH] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1751),
    [anon_sym_false] = ACTIONS(1751),
    [anon_sym_ATinterface] = ACTIONS(1753),
    [anon_sym_interface] = ACTIONS(1751),
    [anon_sym_class] = ACTIONS(1751),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_RBRACE] = ACTIONS(1753),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1751),
    [anon_sym_def] = ACTIONS(1751),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1751),
    [anon_sym_this] = ACTIONS(1751),
    [anon_sym_do] = ACTIONS(1751),
    [anon_sym_while] = ACTIONS(1751),
    [anon_sym_for] = ACTIONS(1751),
    [aux_sym__type_identifier_token1] = ACTIONS(1751),
    [anon_sym_if] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [aux_sym_number_literal_token1] = ACTIONS(1751),
    [aux_sym_number_literal_token2] = ACTIONS(1753),
    [aux_sym_number_literal_token3] = ACTIONS(1753),
    [aux_sym_number_literal_token4] = ACTIONS(1751),
    [aux_sym_number_literal_token5] = ACTIONS(1753),
    [anon_sym_pipeline] = ACTIONS(1751),
    [anon_sym_return] = ACTIONS(1751),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1753),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1753),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1753),
    [anon_sym_switch] = ACTIONS(1751),
    [anon_sym_try] = ACTIONS(1751),
    [anon_sym_int] = ACTIONS(1751),
    [anon_sym_boolean] = ACTIONS(1751),
    [anon_sym_char] = ACTIONS(1751),
    [anon_sym_short] = ACTIONS(1751),
    [anon_sym_long] = ACTIONS(1751),
    [anon_sym_float] = ACTIONS(1751),
    [anon_sym_double] = ACTIONS(1751),
    [anon_sym_void] = ACTIONS(1751),
    [anon_sym_public] = ACTIONS(1751),
    [anon_sym_protected] = ACTIONS(1751),
    [anon_sym_private] = ACTIONS(1751),
    [anon_sym_static] = ACTIONS(1751),
    [anon_sym_final] = ACTIONS(1751),
    [anon_sym_synchronized] = ACTIONS(1751),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_BANG] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1751),
  },
  [STATE(802)] = {
    [sym_comment] = STATE(802),
    [sym_groovy_doc] = STATE(802),
    [ts_builtin_sym_end] = ACTIONS(1836),
    [sym_identifier] = ACTIONS(1838),
    [anon_sym_break] = ACTIONS(1838),
    [anon_sym_continue] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_STAR] = ACTIONS(1838),
    [anon_sym_STAR_COLON] = ACTIONS(1836),
    [anon_sym_import] = ACTIONS(1838),
    [anon_sym_package] = ACTIONS(1838),
    [anon_sym_AT] = ACTIONS(1838),
    [anon_sym_assert] = ACTIONS(1838),
    [anon_sym_PLUS_PLUS] = ACTIONS(1836),
    [anon_sym_DASH_DASH] = ACTIONS(1836),
    [anon_sym_SLASH] = ACTIONS(1838),
    [anon_sym_PLUS] = ACTIONS(1838),
    [anon_sym_DASH] = ACTIONS(1838),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_ATinterface] = ACTIONS(1836),
    [anon_sym_interface] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_COMMA] = ACTIONS(1836),
    [anon_sym_LBRACE] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1838),
    [anon_sym_def] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_RPAREN] = ACTIONS(1836),
    [anon_sym_null] = ACTIONS(1838),
    [anon_sym_this] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_while] = ACTIONS(1838),
    [anon_sym_for] = ACTIONS(1838),
    [aux_sym__type_identifier_token1] = ACTIONS(1838),
    [anon_sym_if] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1836),
    [aux_sym_number_literal_token1] = ACTIONS(1838),
    [aux_sym_number_literal_token2] = ACTIONS(1836),
    [aux_sym_number_literal_token3] = ACTIONS(1836),
    [aux_sym_number_literal_token4] = ACTIONS(1838),
    [aux_sym_number_literal_token5] = ACTIONS(1836),
    [anon_sym_pipeline] = ACTIONS(1838),
    [anon_sym_return] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1838),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1836),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1836),
    [anon_sym_switch] = ACTIONS(1838),
    [anon_sym_try] = ACTIONS(1838),
    [anon_sym_int] = ACTIONS(1838),
    [anon_sym_boolean] = ACTIONS(1838),
    [anon_sym_char] = ACTIONS(1838),
    [anon_sym_short] = ACTIONS(1838),
    [anon_sym_long] = ACTIONS(1838),
    [anon_sym_float] = ACTIONS(1838),
    [anon_sym_double] = ACTIONS(1838),
    [anon_sym_void] = ACTIONS(1838),
    [anon_sym_public] = ACTIONS(1838),
    [anon_sym_protected] = ACTIONS(1838),
    [anon_sym_private] = ACTIONS(1838),
    [anon_sym_static] = ACTIONS(1838),
    [anon_sym_final] = ACTIONS(1838),
    [anon_sym_synchronized] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1836),
    [anon_sym_BANG] = ACTIONS(1836),
    [anon_sym_new] = ACTIONS(1838),
  },
  [STATE(803)] = {
    [sym_comment] = STATE(803),
    [sym_groovy_doc] = STATE(803),
    [ts_builtin_sym_end] = ACTIONS(1876),
    [sym_identifier] = ACTIONS(1878),
    [anon_sym_break] = ACTIONS(1878),
    [anon_sym_continue] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_STAR] = ACTIONS(1878),
    [anon_sym_STAR_COLON] = ACTIONS(1876),
    [anon_sym_import] = ACTIONS(1878),
    [anon_sym_package] = ACTIONS(1878),
    [anon_sym_AT] = ACTIONS(1878),
    [anon_sym_assert] = ACTIONS(1878),
    [anon_sym_PLUS_PLUS] = ACTIONS(1876),
    [anon_sym_DASH_DASH] = ACTIONS(1876),
    [anon_sym_SLASH] = ACTIONS(1878),
    [anon_sym_PLUS] = ACTIONS(1878),
    [anon_sym_DASH] = ACTIONS(1878),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_ATinterface] = ACTIONS(1876),
    [anon_sym_interface] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_COMMA] = ACTIONS(1876),
    [anon_sym_LBRACE] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1878),
    [anon_sym_def] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_RPAREN] = ACTIONS(1876),
    [anon_sym_null] = ACTIONS(1878),
    [anon_sym_this] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_while] = ACTIONS(1878),
    [anon_sym_for] = ACTIONS(1878),
    [aux_sym__type_identifier_token1] = ACTIONS(1878),
    [anon_sym_if] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1876),
    [aux_sym_number_literal_token1] = ACTIONS(1878),
    [aux_sym_number_literal_token2] = ACTIONS(1876),
    [aux_sym_number_literal_token3] = ACTIONS(1876),
    [aux_sym_number_literal_token4] = ACTIONS(1878),
    [aux_sym_number_literal_token5] = ACTIONS(1876),
    [anon_sym_pipeline] = ACTIONS(1878),
    [anon_sym_return] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1878),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1878),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1876),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1876),
    [anon_sym_switch] = ACTIONS(1878),
    [anon_sym_try] = ACTIONS(1878),
    [anon_sym_int] = ACTIONS(1878),
    [anon_sym_boolean] = ACTIONS(1878),
    [anon_sym_char] = ACTIONS(1878),
    [anon_sym_short] = ACTIONS(1878),
    [anon_sym_long] = ACTIONS(1878),
    [anon_sym_float] = ACTIONS(1878),
    [anon_sym_double] = ACTIONS(1878),
    [anon_sym_void] = ACTIONS(1878),
    [anon_sym_public] = ACTIONS(1878),
    [anon_sym_protected] = ACTIONS(1878),
    [anon_sym_private] = ACTIONS(1878),
    [anon_sym_static] = ACTIONS(1878),
    [anon_sym_final] = ACTIONS(1878),
    [anon_sym_synchronized] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1876),
    [anon_sym_BANG] = ACTIONS(1876),
    [anon_sym_new] = ACTIONS(1878),
  },
  [STATE(804)] = {
    [sym_comment] = STATE(804),
    [sym_groovy_doc] = STATE(804),
    [sym_parameter_list] = STATE(934),
    [ts_builtin_sym_end] = ACTIONS(1790),
    [sym_identifier] = ACTIONS(1792),
    [anon_sym_break] = ACTIONS(1792),
    [anon_sym_continue] = ACTIONS(1792),
    [anon_sym_SEMI] = ACTIONS(1790),
    [anon_sym_STAR] = ACTIONS(1792),
    [anon_sym_STAR_COLON] = ACTIONS(1790),
    [anon_sym_import] = ACTIONS(1792),
    [anon_sym_package] = ACTIONS(1792),
    [anon_sym_AT] = ACTIONS(1792),
    [anon_sym_assert] = ACTIONS(1792),
    [anon_sym_EQ] = ACTIONS(2087),
    [anon_sym_PLUS_PLUS] = ACTIONS(1790),
    [anon_sym_DASH_DASH] = ACTIONS(1790),
    [anon_sym_SLASH] = ACTIONS(1792),
    [anon_sym_PLUS] = ACTIONS(1792),
    [anon_sym_DASH] = ACTIONS(1792),
    [anon_sym_true] = ACTIONS(1792),
    [anon_sym_false] = ACTIONS(1792),
    [anon_sym_ATinterface] = ACTIONS(1790),
    [anon_sym_interface] = ACTIONS(1792),
    [anon_sym_class] = ACTIONS(1792),
    [anon_sym_LBRACE] = ACTIONS(1790),
    [anon_sym_RBRACE] = ACTIONS(1790),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1792),
    [anon_sym_def] = ACTIONS(1792),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1792),
    [anon_sym_this] = ACTIONS(1792),
    [anon_sym_do] = ACTIONS(1792),
    [anon_sym_while] = ACTIONS(1792),
    [anon_sym_for] = ACTIONS(1792),
    [aux_sym__type_identifier_token1] = ACTIONS(1792),
    [anon_sym_if] = ACTIONS(1792),
    [anon_sym_else] = ACTIONS(1792),
    [anon_sym_LBRACK] = ACTIONS(1790),
    [aux_sym_number_literal_token1] = ACTIONS(1792),
    [aux_sym_number_literal_token2] = ACTIONS(1790),
    [aux_sym_number_literal_token3] = ACTIONS(1790),
    [aux_sym_number_literal_token4] = ACTIONS(1792),
    [aux_sym_number_literal_token5] = ACTIONS(1790),
    [anon_sym_pipeline] = ACTIONS(1792),
    [anon_sym_return] = ACTIONS(1792),
    [anon_sym_SQUOTE] = ACTIONS(1792),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1790),
    [anon_sym_DQUOTE] = ACTIONS(1792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1790),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1790),
    [anon_sym_switch] = ACTIONS(1792),
    [anon_sym_try] = ACTIONS(1792),
    [anon_sym_int] = ACTIONS(1792),
    [anon_sym_boolean] = ACTIONS(1792),
    [anon_sym_char] = ACTIONS(1792),
    [anon_sym_short] = ACTIONS(1792),
    [anon_sym_long] = ACTIONS(1792),
    [anon_sym_float] = ACTIONS(1792),
    [anon_sym_double] = ACTIONS(1792),
    [anon_sym_void] = ACTIONS(1792),
    [anon_sym_public] = ACTIONS(1792),
    [anon_sym_protected] = ACTIONS(1792),
    [anon_sym_private] = ACTIONS(1792),
    [anon_sym_static] = ACTIONS(1792),
    [anon_sym_final] = ACTIONS(1792),
    [anon_sym_synchronized] = ACTIONS(1792),
    [anon_sym_TILDE] = ACTIONS(1790),
    [anon_sym_BANG] = ACTIONS(1790),
    [anon_sym_new] = ACTIONS(1792),
  },
  [STATE(805)] = {
    [sym_comment] = STATE(805),
    [sym_groovy_doc] = STATE(805),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_case] = ACTIONS(1906),
    [anon_sym_default] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_catch] = ACTIONS(1906),
    [anon_sym_finally] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(806)] = {
    [sym_comment] = STATE(806),
    [sym_groovy_doc] = STATE(806),
    [sym_identifier] = ACTIONS(1832),
    [anon_sym_break] = ACTIONS(1832),
    [anon_sym_continue] = ACTIONS(1832),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_STAR] = ACTIONS(1832),
    [anon_sym_STAR_COLON] = ACTIONS(1830),
    [anon_sym_import] = ACTIONS(1832),
    [anon_sym_package] = ACTIONS(1832),
    [anon_sym_AT] = ACTIONS(1832),
    [anon_sym_assert] = ACTIONS(1832),
    [anon_sym_PLUS_PLUS] = ACTIONS(1830),
    [anon_sym_DASH_DASH] = ACTIONS(1830),
    [anon_sym_SLASH] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(1832),
    [anon_sym_DASH] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1832),
    [anon_sym_false] = ACTIONS(1832),
    [anon_sym_ATinterface] = ACTIONS(1830),
    [anon_sym_interface] = ACTIONS(1832),
    [anon_sym_class] = ACTIONS(1832),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_RBRACE] = ACTIONS(1830),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1832),
    [anon_sym_def] = ACTIONS(1832),
    [anon_sym_LPAREN] = ACTIONS(1830),
    [anon_sym_null] = ACTIONS(1832),
    [anon_sym_this] = ACTIONS(1832),
    [anon_sym_do] = ACTIONS(1832),
    [anon_sym_while] = ACTIONS(1832),
    [anon_sym_for] = ACTIONS(1832),
    [aux_sym__type_identifier_token1] = ACTIONS(1832),
    [anon_sym_if] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [aux_sym_number_literal_token1] = ACTIONS(1832),
    [aux_sym_number_literal_token2] = ACTIONS(1830),
    [aux_sym_number_literal_token3] = ACTIONS(1830),
    [aux_sym_number_literal_token4] = ACTIONS(1832),
    [aux_sym_number_literal_token5] = ACTIONS(1830),
    [anon_sym_return] = ACTIONS(1832),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1830),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1830),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1830),
    [anon_sym_switch] = ACTIONS(1832),
    [anon_sym_case] = ACTIONS(1832),
    [anon_sym_default] = ACTIONS(1832),
    [anon_sym_try] = ACTIONS(1832),
    [anon_sym_catch] = ACTIONS(1832),
    [anon_sym_finally] = ACTIONS(1832),
    [anon_sym_int] = ACTIONS(1832),
    [anon_sym_boolean] = ACTIONS(1832),
    [anon_sym_char] = ACTIONS(1832),
    [anon_sym_short] = ACTIONS(1832),
    [anon_sym_long] = ACTIONS(1832),
    [anon_sym_float] = ACTIONS(1832),
    [anon_sym_double] = ACTIONS(1832),
    [anon_sym_void] = ACTIONS(1832),
    [anon_sym_public] = ACTIONS(1832),
    [anon_sym_protected] = ACTIONS(1832),
    [anon_sym_private] = ACTIONS(1832),
    [anon_sym_static] = ACTIONS(1832),
    [anon_sym_final] = ACTIONS(1832),
    [anon_sym_synchronized] = ACTIONS(1832),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_BANG] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1832),
  },
  [STATE(807)] = {
    [sym_comment] = STATE(807),
    [sym_groovy_doc] = STATE(807),
    [ts_builtin_sym_end] = ACTIONS(1804),
    [sym_identifier] = ACTIONS(1806),
    [anon_sym_break] = ACTIONS(1806),
    [anon_sym_continue] = ACTIONS(1806),
    [anon_sym_SEMI] = ACTIONS(1804),
    [anon_sym_STAR] = ACTIONS(1806),
    [anon_sym_STAR_COLON] = ACTIONS(1804),
    [anon_sym_import] = ACTIONS(1806),
    [anon_sym_package] = ACTIONS(1806),
    [anon_sym_AT] = ACTIONS(1806),
    [anon_sym_assert] = ACTIONS(1806),
    [anon_sym_PLUS_PLUS] = ACTIONS(1804),
    [anon_sym_DASH_DASH] = ACTIONS(1804),
    [anon_sym_SLASH] = ACTIONS(1806),
    [anon_sym_PLUS] = ACTIONS(1806),
    [anon_sym_DASH] = ACTIONS(1806),
    [anon_sym_true] = ACTIONS(1806),
    [anon_sym_false] = ACTIONS(1806),
    [anon_sym_ATinterface] = ACTIONS(1804),
    [anon_sym_interface] = ACTIONS(1806),
    [anon_sym_class] = ACTIONS(1806),
    [anon_sym_COMMA] = ACTIONS(1804),
    [anon_sym_LBRACE] = ACTIONS(1804),
    [anon_sym_RBRACE] = ACTIONS(1804),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1806),
    [anon_sym_def] = ACTIONS(1806),
    [anon_sym_LPAREN] = ACTIONS(1804),
    [anon_sym_RPAREN] = ACTIONS(1804),
    [anon_sym_null] = ACTIONS(1806),
    [anon_sym_this] = ACTIONS(1806),
    [anon_sym_do] = ACTIONS(1806),
    [anon_sym_while] = ACTIONS(1806),
    [anon_sym_for] = ACTIONS(1806),
    [aux_sym__type_identifier_token1] = ACTIONS(1806),
    [anon_sym_if] = ACTIONS(1806),
    [anon_sym_else] = ACTIONS(1806),
    [anon_sym_LBRACK] = ACTIONS(1804),
    [aux_sym_number_literal_token1] = ACTIONS(1806),
    [aux_sym_number_literal_token2] = ACTIONS(1804),
    [aux_sym_number_literal_token3] = ACTIONS(1804),
    [aux_sym_number_literal_token4] = ACTIONS(1806),
    [aux_sym_number_literal_token5] = ACTIONS(1804),
    [anon_sym_pipeline] = ACTIONS(1806),
    [anon_sym_return] = ACTIONS(1806),
    [anon_sym_SQUOTE] = ACTIONS(1806),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1804),
    [anon_sym_DQUOTE] = ACTIONS(1806),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1804),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1804),
    [anon_sym_switch] = ACTIONS(1806),
    [anon_sym_try] = ACTIONS(1806),
    [anon_sym_int] = ACTIONS(1806),
    [anon_sym_boolean] = ACTIONS(1806),
    [anon_sym_char] = ACTIONS(1806),
    [anon_sym_short] = ACTIONS(1806),
    [anon_sym_long] = ACTIONS(1806),
    [anon_sym_float] = ACTIONS(1806),
    [anon_sym_double] = ACTIONS(1806),
    [anon_sym_void] = ACTIONS(1806),
    [anon_sym_public] = ACTIONS(1806),
    [anon_sym_protected] = ACTIONS(1806),
    [anon_sym_private] = ACTIONS(1806),
    [anon_sym_static] = ACTIONS(1806),
    [anon_sym_final] = ACTIONS(1806),
    [anon_sym_synchronized] = ACTIONS(1806),
    [anon_sym_TILDE] = ACTIONS(1804),
    [anon_sym_BANG] = ACTIONS(1804),
    [anon_sym_new] = ACTIONS(1806),
  },
  [STATE(808)] = {
    [sym_comment] = STATE(808),
    [sym_groovy_doc] = STATE(808),
    [sym_identifier] = ACTIONS(1987),
    [anon_sym_break] = ACTIONS(1987),
    [anon_sym_continue] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_STAR] = ACTIONS(1987),
    [anon_sym_STAR_COLON] = ACTIONS(1985),
    [anon_sym_import] = ACTIONS(1987),
    [anon_sym_package] = ACTIONS(1987),
    [anon_sym_AT] = ACTIONS(1987),
    [anon_sym_assert] = ACTIONS(1987),
    [anon_sym_PLUS_PLUS] = ACTIONS(1985),
    [anon_sym_DASH_DASH] = ACTIONS(1985),
    [anon_sym_SLASH] = ACTIONS(1987),
    [anon_sym_PLUS] = ACTIONS(1987),
    [anon_sym_DASH] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_ATinterface] = ACTIONS(1985),
    [anon_sym_interface] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1987),
    [anon_sym_def] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_null] = ACTIONS(1987),
    [anon_sym_this] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_while] = ACTIONS(1987),
    [anon_sym_for] = ACTIONS(1987),
    [aux_sym__type_identifier_token1] = ACTIONS(1987),
    [anon_sym_if] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1985),
    [aux_sym_number_literal_token1] = ACTIONS(1987),
    [aux_sym_number_literal_token2] = ACTIONS(1985),
    [aux_sym_number_literal_token3] = ACTIONS(1985),
    [aux_sym_number_literal_token4] = ACTIONS(1987),
    [aux_sym_number_literal_token5] = ACTIONS(1985),
    [anon_sym_return] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1987),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1987),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1985),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1985),
    [anon_sym_switch] = ACTIONS(1987),
    [anon_sym_case] = ACTIONS(1987),
    [anon_sym_default] = ACTIONS(1987),
    [anon_sym_try] = ACTIONS(1987),
    [anon_sym_catch] = ACTIONS(1987),
    [anon_sym_finally] = ACTIONS(1987),
    [anon_sym_int] = ACTIONS(1987),
    [anon_sym_boolean] = ACTIONS(1987),
    [anon_sym_char] = ACTIONS(1987),
    [anon_sym_short] = ACTIONS(1987),
    [anon_sym_long] = ACTIONS(1987),
    [anon_sym_float] = ACTIONS(1987),
    [anon_sym_double] = ACTIONS(1987),
    [anon_sym_void] = ACTIONS(1987),
    [anon_sym_public] = ACTIONS(1987),
    [anon_sym_protected] = ACTIONS(1987),
    [anon_sym_private] = ACTIONS(1987),
    [anon_sym_static] = ACTIONS(1987),
    [anon_sym_final] = ACTIONS(1987),
    [anon_sym_synchronized] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1985),
    [anon_sym_BANG] = ACTIONS(1985),
    [anon_sym_new] = ACTIONS(1987),
  },
  [STATE(809)] = {
    [sym_comment] = STATE(809),
    [sym_groovy_doc] = STATE(809),
    [aux_sym__juxt_argument_list_repeat1] = STATE(779),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_case] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(810)] = {
    [sym_comment] = STATE(810),
    [sym_groovy_doc] = STATE(810),
    [sym_identifier] = ACTIONS(2001),
    [anon_sym_break] = ACTIONS(2001),
    [anon_sym_continue] = ACTIONS(2001),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_STAR] = ACTIONS(2001),
    [anon_sym_STAR_COLON] = ACTIONS(1999),
    [anon_sym_import] = ACTIONS(2001),
    [anon_sym_package] = ACTIONS(2001),
    [anon_sym_AT] = ACTIONS(2001),
    [anon_sym_assert] = ACTIONS(2001),
    [anon_sym_PLUS_PLUS] = ACTIONS(1999),
    [anon_sym_DASH_DASH] = ACTIONS(1999),
    [anon_sym_SLASH] = ACTIONS(2001),
    [anon_sym_PLUS] = ACTIONS(2001),
    [anon_sym_DASH] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(2001),
    [anon_sym_false] = ACTIONS(2001),
    [anon_sym_ATinterface] = ACTIONS(1999),
    [anon_sym_interface] = ACTIONS(2001),
    [anon_sym_class] = ACTIONS(2001),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_RBRACE] = ACTIONS(1999),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2001),
    [anon_sym_def] = ACTIONS(2001),
    [anon_sym_LPAREN] = ACTIONS(1999),
    [anon_sym_null] = ACTIONS(2001),
    [anon_sym_this] = ACTIONS(2001),
    [anon_sym_do] = ACTIONS(2001),
    [anon_sym_while] = ACTIONS(2001),
    [anon_sym_for] = ACTIONS(2001),
    [aux_sym__type_identifier_token1] = ACTIONS(2001),
    [anon_sym_if] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [aux_sym_number_literal_token1] = ACTIONS(2001),
    [aux_sym_number_literal_token2] = ACTIONS(1999),
    [aux_sym_number_literal_token3] = ACTIONS(1999),
    [aux_sym_number_literal_token4] = ACTIONS(2001),
    [aux_sym_number_literal_token5] = ACTIONS(1999),
    [anon_sym_return] = ACTIONS(2001),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1999),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1999),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1999),
    [anon_sym_switch] = ACTIONS(2001),
    [anon_sym_case] = ACTIONS(2001),
    [anon_sym_default] = ACTIONS(2001),
    [anon_sym_try] = ACTIONS(2001),
    [anon_sym_catch] = ACTIONS(2001),
    [anon_sym_finally] = ACTIONS(2001),
    [anon_sym_int] = ACTIONS(2001),
    [anon_sym_boolean] = ACTIONS(2001),
    [anon_sym_char] = ACTIONS(2001),
    [anon_sym_short] = ACTIONS(2001),
    [anon_sym_long] = ACTIONS(2001),
    [anon_sym_float] = ACTIONS(2001),
    [anon_sym_double] = ACTIONS(2001),
    [anon_sym_void] = ACTIONS(2001),
    [anon_sym_public] = ACTIONS(2001),
    [anon_sym_protected] = ACTIONS(2001),
    [anon_sym_private] = ACTIONS(2001),
    [anon_sym_static] = ACTIONS(2001),
    [anon_sym_final] = ACTIONS(2001),
    [anon_sym_synchronized] = ACTIONS(2001),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_BANG] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(2001),
  },
  [STATE(811)] = {
    [sym_comment] = STATE(811),
    [sym_groovy_doc] = STATE(811),
    [ts_builtin_sym_end] = ACTIONS(1822),
    [sym_identifier] = ACTIONS(1824),
    [anon_sym_break] = ACTIONS(1824),
    [anon_sym_continue] = ACTIONS(1824),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_STAR] = ACTIONS(1824),
    [anon_sym_STAR_COLON] = ACTIONS(1822),
    [anon_sym_import] = ACTIONS(1824),
    [anon_sym_package] = ACTIONS(1824),
    [anon_sym_AT] = ACTIONS(1824),
    [anon_sym_assert] = ACTIONS(1824),
    [anon_sym_PLUS_PLUS] = ACTIONS(1822),
    [anon_sym_DASH_DASH] = ACTIONS(1822),
    [anon_sym_SLASH] = ACTIONS(1824),
    [anon_sym_PLUS] = ACTIONS(1824),
    [anon_sym_DASH] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1824),
    [anon_sym_false] = ACTIONS(1824),
    [anon_sym_ATinterface] = ACTIONS(1822),
    [anon_sym_interface] = ACTIONS(1824),
    [anon_sym_class] = ACTIONS(1824),
    [anon_sym_COMMA] = ACTIONS(1822),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_RBRACE] = ACTIONS(1822),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1824),
    [anon_sym_def] = ACTIONS(1824),
    [anon_sym_LPAREN] = ACTIONS(1822),
    [anon_sym_RPAREN] = ACTIONS(1822),
    [anon_sym_null] = ACTIONS(1824),
    [anon_sym_this] = ACTIONS(1824),
    [anon_sym_do] = ACTIONS(1824),
    [anon_sym_while] = ACTIONS(1824),
    [anon_sym_for] = ACTIONS(1824),
    [aux_sym__type_identifier_token1] = ACTIONS(1824),
    [anon_sym_if] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [aux_sym_number_literal_token1] = ACTIONS(1824),
    [aux_sym_number_literal_token2] = ACTIONS(1822),
    [aux_sym_number_literal_token3] = ACTIONS(1822),
    [aux_sym_number_literal_token4] = ACTIONS(1824),
    [aux_sym_number_literal_token5] = ACTIONS(1822),
    [anon_sym_pipeline] = ACTIONS(1824),
    [anon_sym_return] = ACTIONS(1824),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1822),
    [anon_sym_switch] = ACTIONS(1824),
    [anon_sym_try] = ACTIONS(1824),
    [anon_sym_int] = ACTIONS(1824),
    [anon_sym_boolean] = ACTIONS(1824),
    [anon_sym_char] = ACTIONS(1824),
    [anon_sym_short] = ACTIONS(1824),
    [anon_sym_long] = ACTIONS(1824),
    [anon_sym_float] = ACTIONS(1824),
    [anon_sym_double] = ACTIONS(1824),
    [anon_sym_void] = ACTIONS(1824),
    [anon_sym_public] = ACTIONS(1824),
    [anon_sym_protected] = ACTIONS(1824),
    [anon_sym_private] = ACTIONS(1824),
    [anon_sym_static] = ACTIONS(1824),
    [anon_sym_final] = ACTIONS(1824),
    [anon_sym_synchronized] = ACTIONS(1824),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_BANG] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1824),
  },
  [STATE(812)] = {
    [sym_comment] = STATE(812),
    [sym_groovy_doc] = STATE(812),
    [sym_identifier] = ACTIONS(1785),
    [anon_sym_break] = ACTIONS(1785),
    [anon_sym_continue] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_STAR] = ACTIONS(1785),
    [anon_sym_STAR_COLON] = ACTIONS(1783),
    [anon_sym_import] = ACTIONS(1785),
    [anon_sym_package] = ACTIONS(1785),
    [anon_sym_AT] = ACTIONS(1785),
    [anon_sym_assert] = ACTIONS(1785),
    [anon_sym_PLUS_PLUS] = ACTIONS(1783),
    [anon_sym_DASH_DASH] = ACTIONS(1783),
    [anon_sym_SLASH] = ACTIONS(1785),
    [anon_sym_PLUS] = ACTIONS(1785),
    [anon_sym_DASH] = ACTIONS(1785),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_ATinterface] = ACTIONS(1783),
    [anon_sym_interface] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1785),
    [anon_sym_def] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_null] = ACTIONS(1785),
    [anon_sym_this] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1785),
    [anon_sym_for] = ACTIONS(1785),
    [aux_sym__type_identifier_token1] = ACTIONS(1785),
    [anon_sym_if] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1783),
    [aux_sym_number_literal_token1] = ACTIONS(1785),
    [aux_sym_number_literal_token2] = ACTIONS(1783),
    [aux_sym_number_literal_token3] = ACTIONS(1783),
    [aux_sym_number_literal_token4] = ACTIONS(1785),
    [aux_sym_number_literal_token5] = ACTIONS(1783),
    [anon_sym_return] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1785),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1785),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1783),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1783),
    [anon_sym_switch] = ACTIONS(1785),
    [anon_sym_case] = ACTIONS(1785),
    [anon_sym_default] = ACTIONS(1785),
    [anon_sym_try] = ACTIONS(1785),
    [anon_sym_catch] = ACTIONS(1785),
    [anon_sym_finally] = ACTIONS(1785),
    [anon_sym_int] = ACTIONS(1785),
    [anon_sym_boolean] = ACTIONS(1785),
    [anon_sym_char] = ACTIONS(1785),
    [anon_sym_short] = ACTIONS(1785),
    [anon_sym_long] = ACTIONS(1785),
    [anon_sym_float] = ACTIONS(1785),
    [anon_sym_double] = ACTIONS(1785),
    [anon_sym_void] = ACTIONS(1785),
    [anon_sym_public] = ACTIONS(1785),
    [anon_sym_protected] = ACTIONS(1785),
    [anon_sym_private] = ACTIONS(1785),
    [anon_sym_static] = ACTIONS(1785),
    [anon_sym_final] = ACTIONS(1785),
    [anon_sym_synchronized] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1783),
    [anon_sym_BANG] = ACTIONS(1783),
    [anon_sym_new] = ACTIONS(1785),
  },
  [STATE(813)] = {
    [sym_comment] = STATE(813),
    [sym_groovy_doc] = STATE(813),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_break] = ACTIONS(1858),
    [anon_sym_continue] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_STAR] = ACTIONS(1858),
    [anon_sym_STAR_COLON] = ACTIONS(1856),
    [anon_sym_import] = ACTIONS(1858),
    [anon_sym_package] = ACTIONS(1858),
    [anon_sym_AT] = ACTIONS(1858),
    [anon_sym_assert] = ACTIONS(1858),
    [anon_sym_PLUS_PLUS] = ACTIONS(1856),
    [anon_sym_DASH_DASH] = ACTIONS(1856),
    [anon_sym_SLASH] = ACTIONS(1858),
    [anon_sym_PLUS] = ACTIONS(1858),
    [anon_sym_DASH] = ACTIONS(1858),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_ATinterface] = ACTIONS(1856),
    [anon_sym_interface] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1858),
    [anon_sym_def] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_null] = ACTIONS(1858),
    [anon_sym_this] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_while] = ACTIONS(1858),
    [anon_sym_for] = ACTIONS(1858),
    [aux_sym__type_identifier_token1] = ACTIONS(1858),
    [anon_sym_if] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1856),
    [aux_sym_number_literal_token1] = ACTIONS(1858),
    [aux_sym_number_literal_token2] = ACTIONS(1856),
    [aux_sym_number_literal_token3] = ACTIONS(1856),
    [aux_sym_number_literal_token4] = ACTIONS(1858),
    [aux_sym_number_literal_token5] = ACTIONS(1856),
    [anon_sym_return] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1858),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1858),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1856),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1856),
    [anon_sym_switch] = ACTIONS(1858),
    [anon_sym_case] = ACTIONS(1858),
    [anon_sym_default] = ACTIONS(1858),
    [anon_sym_try] = ACTIONS(1858),
    [anon_sym_catch] = ACTIONS(1858),
    [anon_sym_finally] = ACTIONS(1858),
    [anon_sym_int] = ACTIONS(1858),
    [anon_sym_boolean] = ACTIONS(1858),
    [anon_sym_char] = ACTIONS(1858),
    [anon_sym_short] = ACTIONS(1858),
    [anon_sym_long] = ACTIONS(1858),
    [anon_sym_float] = ACTIONS(1858),
    [anon_sym_double] = ACTIONS(1858),
    [anon_sym_void] = ACTIONS(1858),
    [anon_sym_public] = ACTIONS(1858),
    [anon_sym_protected] = ACTIONS(1858),
    [anon_sym_private] = ACTIONS(1858),
    [anon_sym_static] = ACTIONS(1858),
    [anon_sym_final] = ACTIONS(1858),
    [anon_sym_synchronized] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1856),
    [anon_sym_BANG] = ACTIONS(1856),
    [anon_sym_new] = ACTIONS(1858),
  },
  [STATE(814)] = {
    [sym_comment] = STATE(814),
    [sym_groovy_doc] = STATE(814),
    [ts_builtin_sym_end] = ACTIONS(1826),
    [sym_identifier] = ACTIONS(1828),
    [anon_sym_break] = ACTIONS(1828),
    [anon_sym_continue] = ACTIONS(1828),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_STAR] = ACTIONS(1828),
    [anon_sym_STAR_COLON] = ACTIONS(1826),
    [anon_sym_import] = ACTIONS(1828),
    [anon_sym_package] = ACTIONS(1828),
    [anon_sym_AT] = ACTIONS(1828),
    [anon_sym_assert] = ACTIONS(1828),
    [anon_sym_PLUS_PLUS] = ACTIONS(1826),
    [anon_sym_DASH_DASH] = ACTIONS(1826),
    [anon_sym_SLASH] = ACTIONS(1828),
    [anon_sym_PLUS] = ACTIONS(1828),
    [anon_sym_DASH] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1828),
    [anon_sym_false] = ACTIONS(1828),
    [anon_sym_ATinterface] = ACTIONS(1826),
    [anon_sym_interface] = ACTIONS(1828),
    [anon_sym_class] = ACTIONS(1828),
    [anon_sym_COMMA] = ACTIONS(1826),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_RBRACE] = ACTIONS(1826),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1828),
    [anon_sym_def] = ACTIONS(1828),
    [anon_sym_LPAREN] = ACTIONS(1826),
    [anon_sym_RPAREN] = ACTIONS(1826),
    [anon_sym_null] = ACTIONS(1828),
    [anon_sym_this] = ACTIONS(1828),
    [anon_sym_do] = ACTIONS(1828),
    [anon_sym_while] = ACTIONS(1828),
    [anon_sym_for] = ACTIONS(1828),
    [aux_sym__type_identifier_token1] = ACTIONS(1828),
    [anon_sym_if] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [aux_sym_number_literal_token1] = ACTIONS(1828),
    [aux_sym_number_literal_token2] = ACTIONS(1826),
    [aux_sym_number_literal_token3] = ACTIONS(1826),
    [aux_sym_number_literal_token4] = ACTIONS(1828),
    [aux_sym_number_literal_token5] = ACTIONS(1826),
    [anon_sym_pipeline] = ACTIONS(1828),
    [anon_sym_return] = ACTIONS(1828),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1826),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1826),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1826),
    [anon_sym_switch] = ACTIONS(1828),
    [anon_sym_try] = ACTIONS(1828),
    [anon_sym_int] = ACTIONS(1828),
    [anon_sym_boolean] = ACTIONS(1828),
    [anon_sym_char] = ACTIONS(1828),
    [anon_sym_short] = ACTIONS(1828),
    [anon_sym_long] = ACTIONS(1828),
    [anon_sym_float] = ACTIONS(1828),
    [anon_sym_double] = ACTIONS(1828),
    [anon_sym_void] = ACTIONS(1828),
    [anon_sym_public] = ACTIONS(1828),
    [anon_sym_protected] = ACTIONS(1828),
    [anon_sym_private] = ACTIONS(1828),
    [anon_sym_static] = ACTIONS(1828),
    [anon_sym_final] = ACTIONS(1828),
    [anon_sym_synchronized] = ACTIONS(1828),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_BANG] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1828),
  },
  [STATE(815)] = {
    [sym_comment] = STATE(815),
    [sym_groovy_doc] = STATE(815),
    [sym_identifier] = ACTIONS(1866),
    [anon_sym_break] = ACTIONS(1866),
    [anon_sym_continue] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_STAR] = ACTIONS(1866),
    [anon_sym_STAR_COLON] = ACTIONS(1864),
    [anon_sym_import] = ACTIONS(1866),
    [anon_sym_package] = ACTIONS(1866),
    [anon_sym_AT] = ACTIONS(1866),
    [anon_sym_assert] = ACTIONS(1866),
    [anon_sym_PLUS_PLUS] = ACTIONS(1864),
    [anon_sym_DASH_DASH] = ACTIONS(1864),
    [anon_sym_SLASH] = ACTIONS(1866),
    [anon_sym_PLUS] = ACTIONS(1866),
    [anon_sym_DASH] = ACTIONS(1866),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_ATinterface] = ACTIONS(1864),
    [anon_sym_interface] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1866),
    [anon_sym_def] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_null] = ACTIONS(1866),
    [anon_sym_this] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_while] = ACTIONS(1866),
    [anon_sym_for] = ACTIONS(1866),
    [aux_sym__type_identifier_token1] = ACTIONS(1866),
    [anon_sym_if] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1864),
    [aux_sym_number_literal_token1] = ACTIONS(1866),
    [aux_sym_number_literal_token2] = ACTIONS(1864),
    [aux_sym_number_literal_token3] = ACTIONS(1864),
    [aux_sym_number_literal_token4] = ACTIONS(1866),
    [aux_sym_number_literal_token5] = ACTIONS(1864),
    [anon_sym_return] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1866),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1866),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1864),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1864),
    [anon_sym_switch] = ACTIONS(1866),
    [anon_sym_case] = ACTIONS(1866),
    [anon_sym_default] = ACTIONS(1866),
    [anon_sym_try] = ACTIONS(1866),
    [anon_sym_catch] = ACTIONS(1866),
    [anon_sym_finally] = ACTIONS(1866),
    [anon_sym_int] = ACTIONS(1866),
    [anon_sym_boolean] = ACTIONS(1866),
    [anon_sym_char] = ACTIONS(1866),
    [anon_sym_short] = ACTIONS(1866),
    [anon_sym_long] = ACTIONS(1866),
    [anon_sym_float] = ACTIONS(1866),
    [anon_sym_double] = ACTIONS(1866),
    [anon_sym_void] = ACTIONS(1866),
    [anon_sym_public] = ACTIONS(1866),
    [anon_sym_protected] = ACTIONS(1866),
    [anon_sym_private] = ACTIONS(1866),
    [anon_sym_static] = ACTIONS(1866),
    [anon_sym_final] = ACTIONS(1866),
    [anon_sym_synchronized] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1864),
    [anon_sym_BANG] = ACTIONS(1864),
    [anon_sym_new] = ACTIONS(1866),
  },
  [STATE(816)] = {
    [sym_comment] = STATE(816),
    [sym_groovy_doc] = STATE(816),
    [ts_builtin_sym_end] = ACTIONS(1840),
    [sym_identifier] = ACTIONS(1842),
    [anon_sym_break] = ACTIONS(1842),
    [anon_sym_continue] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_STAR] = ACTIONS(1842),
    [anon_sym_STAR_COLON] = ACTIONS(1840),
    [anon_sym_import] = ACTIONS(1842),
    [anon_sym_package] = ACTIONS(1842),
    [anon_sym_AT] = ACTIONS(1842),
    [anon_sym_assert] = ACTIONS(1842),
    [anon_sym_PLUS_PLUS] = ACTIONS(1840),
    [anon_sym_DASH_DASH] = ACTIONS(1840),
    [anon_sym_SLASH] = ACTIONS(1842),
    [anon_sym_PLUS] = ACTIONS(1842),
    [anon_sym_DASH] = ACTIONS(1842),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_ATinterface] = ACTIONS(1840),
    [anon_sym_interface] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_COMMA] = ACTIONS(1840),
    [anon_sym_LBRACE] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1842),
    [anon_sym_def] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_RPAREN] = ACTIONS(1840),
    [anon_sym_null] = ACTIONS(1842),
    [anon_sym_this] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_while] = ACTIONS(1842),
    [anon_sym_for] = ACTIONS(1842),
    [aux_sym__type_identifier_token1] = ACTIONS(1842),
    [anon_sym_if] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1840),
    [aux_sym_number_literal_token1] = ACTIONS(1842),
    [aux_sym_number_literal_token2] = ACTIONS(1840),
    [aux_sym_number_literal_token3] = ACTIONS(1840),
    [aux_sym_number_literal_token4] = ACTIONS(1842),
    [aux_sym_number_literal_token5] = ACTIONS(1840),
    [anon_sym_pipeline] = ACTIONS(1842),
    [anon_sym_return] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1842),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1842),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1840),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1840),
    [anon_sym_switch] = ACTIONS(1842),
    [anon_sym_try] = ACTIONS(1842),
    [anon_sym_int] = ACTIONS(1842),
    [anon_sym_boolean] = ACTIONS(1842),
    [anon_sym_char] = ACTIONS(1842),
    [anon_sym_short] = ACTIONS(1842),
    [anon_sym_long] = ACTIONS(1842),
    [anon_sym_float] = ACTIONS(1842),
    [anon_sym_double] = ACTIONS(1842),
    [anon_sym_void] = ACTIONS(1842),
    [anon_sym_public] = ACTIONS(1842),
    [anon_sym_protected] = ACTIONS(1842),
    [anon_sym_private] = ACTIONS(1842),
    [anon_sym_static] = ACTIONS(1842),
    [anon_sym_final] = ACTIONS(1842),
    [anon_sym_synchronized] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1840),
    [anon_sym_BANG] = ACTIONS(1840),
    [anon_sym_new] = ACTIONS(1842),
  },
  [STATE(817)] = {
    [sym_comment] = STATE(817),
    [sym_groovy_doc] = STATE(817),
    [ts_builtin_sym_end] = ACTIONS(1856),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_break] = ACTIONS(1858),
    [anon_sym_continue] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_STAR] = ACTIONS(1858),
    [anon_sym_STAR_COLON] = ACTIONS(1856),
    [anon_sym_import] = ACTIONS(1858),
    [anon_sym_package] = ACTIONS(1858),
    [anon_sym_AT] = ACTIONS(1858),
    [anon_sym_assert] = ACTIONS(1858),
    [anon_sym_PLUS_PLUS] = ACTIONS(1856),
    [anon_sym_DASH_DASH] = ACTIONS(1856),
    [anon_sym_SLASH] = ACTIONS(1858),
    [anon_sym_PLUS] = ACTIONS(1858),
    [anon_sym_DASH] = ACTIONS(1858),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_ATinterface] = ACTIONS(1856),
    [anon_sym_interface] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_COMMA] = ACTIONS(1856),
    [anon_sym_LBRACE] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1858),
    [anon_sym_def] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_RPAREN] = ACTIONS(1856),
    [anon_sym_null] = ACTIONS(1858),
    [anon_sym_this] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_while] = ACTIONS(1858),
    [anon_sym_for] = ACTIONS(1858),
    [aux_sym__type_identifier_token1] = ACTIONS(1858),
    [anon_sym_if] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1856),
    [aux_sym_number_literal_token1] = ACTIONS(1858),
    [aux_sym_number_literal_token2] = ACTIONS(1856),
    [aux_sym_number_literal_token3] = ACTIONS(1856),
    [aux_sym_number_literal_token4] = ACTIONS(1858),
    [aux_sym_number_literal_token5] = ACTIONS(1856),
    [anon_sym_pipeline] = ACTIONS(1858),
    [anon_sym_return] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1858),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1858),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1856),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1856),
    [anon_sym_switch] = ACTIONS(1858),
    [anon_sym_try] = ACTIONS(1858),
    [anon_sym_int] = ACTIONS(1858),
    [anon_sym_boolean] = ACTIONS(1858),
    [anon_sym_char] = ACTIONS(1858),
    [anon_sym_short] = ACTIONS(1858),
    [anon_sym_long] = ACTIONS(1858),
    [anon_sym_float] = ACTIONS(1858),
    [anon_sym_double] = ACTIONS(1858),
    [anon_sym_void] = ACTIONS(1858),
    [anon_sym_public] = ACTIONS(1858),
    [anon_sym_protected] = ACTIONS(1858),
    [anon_sym_private] = ACTIONS(1858),
    [anon_sym_static] = ACTIONS(1858),
    [anon_sym_final] = ACTIONS(1858),
    [anon_sym_synchronized] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1856),
    [anon_sym_BANG] = ACTIONS(1856),
    [anon_sym_new] = ACTIONS(1858),
  },
  [STATE(818)] = {
    [sym_comment] = STATE(818),
    [sym_groovy_doc] = STATE(818),
    [ts_builtin_sym_end] = ACTIONS(1908),
    [sym_identifier] = ACTIONS(1910),
    [anon_sym_break] = ACTIONS(1910),
    [anon_sym_continue] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_STAR] = ACTIONS(1910),
    [anon_sym_STAR_COLON] = ACTIONS(1908),
    [anon_sym_import] = ACTIONS(1910),
    [anon_sym_package] = ACTIONS(1910),
    [anon_sym_AT] = ACTIONS(1910),
    [anon_sym_assert] = ACTIONS(1910),
    [anon_sym_PLUS_PLUS] = ACTIONS(1908),
    [anon_sym_DASH_DASH] = ACTIONS(1908),
    [anon_sym_SLASH] = ACTIONS(1910),
    [anon_sym_PLUS] = ACTIONS(1910),
    [anon_sym_DASH] = ACTIONS(1910),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_ATinterface] = ACTIONS(1908),
    [anon_sym_interface] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_COMMA] = ACTIONS(1908),
    [anon_sym_LBRACE] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1910),
    [anon_sym_def] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_RPAREN] = ACTIONS(1908),
    [anon_sym_null] = ACTIONS(1910),
    [anon_sym_this] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_while] = ACTIONS(1910),
    [anon_sym_for] = ACTIONS(1910),
    [aux_sym__type_identifier_token1] = ACTIONS(1910),
    [anon_sym_if] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1908),
    [aux_sym_number_literal_token1] = ACTIONS(1910),
    [aux_sym_number_literal_token2] = ACTIONS(1908),
    [aux_sym_number_literal_token3] = ACTIONS(1908),
    [aux_sym_number_literal_token4] = ACTIONS(1910),
    [aux_sym_number_literal_token5] = ACTIONS(1908),
    [anon_sym_pipeline] = ACTIONS(1910),
    [anon_sym_return] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1910),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1910),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1908),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1908),
    [anon_sym_switch] = ACTIONS(1910),
    [anon_sym_try] = ACTIONS(1910),
    [anon_sym_int] = ACTIONS(1910),
    [anon_sym_boolean] = ACTIONS(1910),
    [anon_sym_char] = ACTIONS(1910),
    [anon_sym_short] = ACTIONS(1910),
    [anon_sym_long] = ACTIONS(1910),
    [anon_sym_float] = ACTIONS(1910),
    [anon_sym_double] = ACTIONS(1910),
    [anon_sym_void] = ACTIONS(1910),
    [anon_sym_public] = ACTIONS(1910),
    [anon_sym_protected] = ACTIONS(1910),
    [anon_sym_private] = ACTIONS(1910),
    [anon_sym_static] = ACTIONS(1910),
    [anon_sym_final] = ACTIONS(1910),
    [anon_sym_synchronized] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1908),
    [anon_sym_BANG] = ACTIONS(1908),
    [anon_sym_new] = ACTIONS(1910),
  },
  [STATE(819)] = {
    [sym_comment] = STATE(819),
    [sym_groovy_doc] = STATE(819),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(369),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(820)] = {
    [sym_comment] = STATE(820),
    [sym_groovy_doc] = STATE(820),
    [sym_identifier] = ACTIONS(1749),
    [anon_sym_break] = ACTIONS(1749),
    [anon_sym_continue] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_STAR] = ACTIONS(1749),
    [anon_sym_STAR_COLON] = ACTIONS(1747),
    [anon_sym_import] = ACTIONS(1749),
    [anon_sym_package] = ACTIONS(1749),
    [anon_sym_AT] = ACTIONS(1749),
    [anon_sym_assert] = ACTIONS(1749),
    [anon_sym_PLUS_PLUS] = ACTIONS(1747),
    [anon_sym_DASH_DASH] = ACTIONS(1747),
    [anon_sym_SLASH] = ACTIONS(1749),
    [anon_sym_PLUS] = ACTIONS(1749),
    [anon_sym_DASH] = ACTIONS(1749),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_ATinterface] = ACTIONS(1747),
    [anon_sym_interface] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1749),
    [anon_sym_def] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_null] = ACTIONS(1749),
    [anon_sym_this] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_while] = ACTIONS(1749),
    [anon_sym_for] = ACTIONS(1749),
    [aux_sym__type_identifier_token1] = ACTIONS(1749),
    [anon_sym_if] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1747),
    [aux_sym_number_literal_token1] = ACTIONS(1749),
    [aux_sym_number_literal_token2] = ACTIONS(1747),
    [aux_sym_number_literal_token3] = ACTIONS(1747),
    [aux_sym_number_literal_token4] = ACTIONS(1749),
    [aux_sym_number_literal_token5] = ACTIONS(1747),
    [anon_sym_return] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1749),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1749),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1747),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1747),
    [anon_sym_switch] = ACTIONS(1749),
    [anon_sym_case] = ACTIONS(1749),
    [anon_sym_default] = ACTIONS(1749),
    [anon_sym_try] = ACTIONS(1749),
    [anon_sym_catch] = ACTIONS(1749),
    [anon_sym_finally] = ACTIONS(1749),
    [anon_sym_int] = ACTIONS(1749),
    [anon_sym_boolean] = ACTIONS(1749),
    [anon_sym_char] = ACTIONS(1749),
    [anon_sym_short] = ACTIONS(1749),
    [anon_sym_long] = ACTIONS(1749),
    [anon_sym_float] = ACTIONS(1749),
    [anon_sym_double] = ACTIONS(1749),
    [anon_sym_void] = ACTIONS(1749),
    [anon_sym_public] = ACTIONS(1749),
    [anon_sym_protected] = ACTIONS(1749),
    [anon_sym_private] = ACTIONS(1749),
    [anon_sym_static] = ACTIONS(1749),
    [anon_sym_final] = ACTIONS(1749),
    [anon_sym_synchronized] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1747),
    [anon_sym_BANG] = ACTIONS(1747),
    [anon_sym_new] = ACTIONS(1749),
  },
  [STATE(821)] = {
    [sym_comment] = STATE(821),
    [sym_groovy_doc] = STATE(821),
    [ts_builtin_sym_end] = ACTIONS(1912),
    [sym_identifier] = ACTIONS(1914),
    [anon_sym_break] = ACTIONS(1914),
    [anon_sym_continue] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_STAR] = ACTIONS(1914),
    [anon_sym_STAR_COLON] = ACTIONS(1912),
    [anon_sym_import] = ACTIONS(1914),
    [anon_sym_package] = ACTIONS(1914),
    [anon_sym_AT] = ACTIONS(1914),
    [anon_sym_assert] = ACTIONS(1914),
    [anon_sym_PLUS_PLUS] = ACTIONS(1912),
    [anon_sym_DASH_DASH] = ACTIONS(1912),
    [anon_sym_SLASH] = ACTIONS(1914),
    [anon_sym_PLUS] = ACTIONS(1914),
    [anon_sym_DASH] = ACTIONS(1914),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_ATinterface] = ACTIONS(1912),
    [anon_sym_interface] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_COMMA] = ACTIONS(1912),
    [anon_sym_LBRACE] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1914),
    [anon_sym_def] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_RPAREN] = ACTIONS(1912),
    [anon_sym_null] = ACTIONS(1914),
    [anon_sym_this] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_while] = ACTIONS(1914),
    [anon_sym_for] = ACTIONS(1914),
    [aux_sym__type_identifier_token1] = ACTIONS(1914),
    [anon_sym_if] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1912),
    [aux_sym_number_literal_token1] = ACTIONS(1914),
    [aux_sym_number_literal_token2] = ACTIONS(1912),
    [aux_sym_number_literal_token3] = ACTIONS(1912),
    [aux_sym_number_literal_token4] = ACTIONS(1914),
    [aux_sym_number_literal_token5] = ACTIONS(1912),
    [anon_sym_pipeline] = ACTIONS(1914),
    [anon_sym_return] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1914),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1914),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1912),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1912),
    [anon_sym_switch] = ACTIONS(1914),
    [anon_sym_try] = ACTIONS(1914),
    [anon_sym_int] = ACTIONS(1914),
    [anon_sym_boolean] = ACTIONS(1914),
    [anon_sym_char] = ACTIONS(1914),
    [anon_sym_short] = ACTIONS(1914),
    [anon_sym_long] = ACTIONS(1914),
    [anon_sym_float] = ACTIONS(1914),
    [anon_sym_double] = ACTIONS(1914),
    [anon_sym_void] = ACTIONS(1914),
    [anon_sym_public] = ACTIONS(1914),
    [anon_sym_protected] = ACTIONS(1914),
    [anon_sym_private] = ACTIONS(1914),
    [anon_sym_static] = ACTIONS(1914),
    [anon_sym_final] = ACTIONS(1914),
    [anon_sym_synchronized] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1912),
    [anon_sym_BANG] = ACTIONS(1912),
    [anon_sym_new] = ACTIONS(1914),
  },
  [STATE(822)] = {
    [sym_comment] = STATE(822),
    [sym_groovy_doc] = STATE(822),
    [ts_builtin_sym_end] = ACTIONS(1767),
    [sym_identifier] = ACTIONS(1769),
    [anon_sym_break] = ACTIONS(1769),
    [anon_sym_continue] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_STAR] = ACTIONS(1769),
    [anon_sym_STAR_COLON] = ACTIONS(1767),
    [anon_sym_import] = ACTIONS(1769),
    [anon_sym_package] = ACTIONS(1769),
    [anon_sym_AT] = ACTIONS(1769),
    [anon_sym_assert] = ACTIONS(1769),
    [anon_sym_PLUS_PLUS] = ACTIONS(1767),
    [anon_sym_DASH_DASH] = ACTIONS(1767),
    [anon_sym_SLASH] = ACTIONS(1769),
    [anon_sym_PLUS] = ACTIONS(1769),
    [anon_sym_DASH] = ACTIONS(1769),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_ATinterface] = ACTIONS(1767),
    [anon_sym_interface] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_LBRACE] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1769),
    [anon_sym_def] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_RPAREN] = ACTIONS(1767),
    [anon_sym_null] = ACTIONS(1769),
    [anon_sym_this] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_while] = ACTIONS(1769),
    [anon_sym_for] = ACTIONS(1769),
    [aux_sym__type_identifier_token1] = ACTIONS(1769),
    [anon_sym_if] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [aux_sym_number_literal_token1] = ACTIONS(1769),
    [aux_sym_number_literal_token2] = ACTIONS(1767),
    [aux_sym_number_literal_token3] = ACTIONS(1767),
    [aux_sym_number_literal_token4] = ACTIONS(1769),
    [aux_sym_number_literal_token5] = ACTIONS(1767),
    [anon_sym_pipeline] = ACTIONS(1769),
    [anon_sym_return] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1769),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1769),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1767),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1767),
    [anon_sym_switch] = ACTIONS(1769),
    [anon_sym_try] = ACTIONS(1769),
    [anon_sym_int] = ACTIONS(1769),
    [anon_sym_boolean] = ACTIONS(1769),
    [anon_sym_char] = ACTIONS(1769),
    [anon_sym_short] = ACTIONS(1769),
    [anon_sym_long] = ACTIONS(1769),
    [anon_sym_float] = ACTIONS(1769),
    [anon_sym_double] = ACTIONS(1769),
    [anon_sym_void] = ACTIONS(1769),
    [anon_sym_public] = ACTIONS(1769),
    [anon_sym_protected] = ACTIONS(1769),
    [anon_sym_private] = ACTIONS(1769),
    [anon_sym_static] = ACTIONS(1769),
    [anon_sym_final] = ACTIONS(1769),
    [anon_sym_synchronized] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1767),
    [anon_sym_BANG] = ACTIONS(1767),
    [anon_sym_new] = ACTIONS(1769),
  },
  [STATE(823)] = {
    [sym_comment] = STATE(823),
    [sym_groovy_doc] = STATE(823),
    [ts_builtin_sym_end] = ACTIONS(1920),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_break] = ACTIONS(1922),
    [anon_sym_continue] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_STAR] = ACTIONS(1922),
    [anon_sym_STAR_COLON] = ACTIONS(1920),
    [anon_sym_import] = ACTIONS(1922),
    [anon_sym_package] = ACTIONS(1922),
    [anon_sym_AT] = ACTIONS(1922),
    [anon_sym_assert] = ACTIONS(1922),
    [anon_sym_PLUS_PLUS] = ACTIONS(1920),
    [anon_sym_DASH_DASH] = ACTIONS(1920),
    [anon_sym_SLASH] = ACTIONS(1922),
    [anon_sym_PLUS] = ACTIONS(1922),
    [anon_sym_DASH] = ACTIONS(1922),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_ATinterface] = ACTIONS(1920),
    [anon_sym_interface] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_COMMA] = ACTIONS(1920),
    [anon_sym_LBRACE] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1922),
    [anon_sym_def] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_RPAREN] = ACTIONS(1920),
    [anon_sym_null] = ACTIONS(1922),
    [anon_sym_this] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_while] = ACTIONS(1922),
    [anon_sym_for] = ACTIONS(1922),
    [aux_sym__type_identifier_token1] = ACTIONS(1922),
    [anon_sym_if] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1920),
    [aux_sym_number_literal_token1] = ACTIONS(1922),
    [aux_sym_number_literal_token2] = ACTIONS(1920),
    [aux_sym_number_literal_token3] = ACTIONS(1920),
    [aux_sym_number_literal_token4] = ACTIONS(1922),
    [aux_sym_number_literal_token5] = ACTIONS(1920),
    [anon_sym_pipeline] = ACTIONS(1922),
    [anon_sym_return] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1922),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1922),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1920),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1920),
    [anon_sym_switch] = ACTIONS(1922),
    [anon_sym_try] = ACTIONS(1922),
    [anon_sym_int] = ACTIONS(1922),
    [anon_sym_boolean] = ACTIONS(1922),
    [anon_sym_char] = ACTIONS(1922),
    [anon_sym_short] = ACTIONS(1922),
    [anon_sym_long] = ACTIONS(1922),
    [anon_sym_float] = ACTIONS(1922),
    [anon_sym_double] = ACTIONS(1922),
    [anon_sym_void] = ACTIONS(1922),
    [anon_sym_public] = ACTIONS(1922),
    [anon_sym_protected] = ACTIONS(1922),
    [anon_sym_private] = ACTIONS(1922),
    [anon_sym_static] = ACTIONS(1922),
    [anon_sym_final] = ACTIONS(1922),
    [anon_sym_synchronized] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1920),
    [anon_sym_BANG] = ACTIONS(1920),
    [anon_sym_new] = ACTIONS(1922),
  },
  [STATE(824)] = {
    [sym_comment] = STATE(824),
    [sym_groovy_doc] = STATE(824),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_case] = ACTIONS(1362),
    [anon_sym_default] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_catch] = ACTIONS(1362),
    [anon_sym_finally] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(825)] = {
    [sym_comment] = STATE(825),
    [sym_groovy_doc] = STATE(825),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_case] = ACTIONS(1376),
    [anon_sym_default] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_catch] = ACTIONS(1376),
    [anon_sym_finally] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(826)] = {
    [sym_comment] = STATE(826),
    [sym_groovy_doc] = STATE(826),
    [ts_builtin_sym_end] = ACTIONS(1924),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_break] = ACTIONS(1926),
    [anon_sym_continue] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_STAR] = ACTIONS(1926),
    [anon_sym_STAR_COLON] = ACTIONS(1924),
    [anon_sym_import] = ACTIONS(1926),
    [anon_sym_package] = ACTIONS(1926),
    [anon_sym_AT] = ACTIONS(1926),
    [anon_sym_assert] = ACTIONS(1926),
    [anon_sym_PLUS_PLUS] = ACTIONS(1924),
    [anon_sym_DASH_DASH] = ACTIONS(1924),
    [anon_sym_SLASH] = ACTIONS(1926),
    [anon_sym_PLUS] = ACTIONS(1926),
    [anon_sym_DASH] = ACTIONS(1926),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_ATinterface] = ACTIONS(1924),
    [anon_sym_interface] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_COMMA] = ACTIONS(1924),
    [anon_sym_LBRACE] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1926),
    [anon_sym_def] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_RPAREN] = ACTIONS(1924),
    [anon_sym_null] = ACTIONS(1926),
    [anon_sym_this] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_while] = ACTIONS(1926),
    [anon_sym_for] = ACTIONS(1926),
    [aux_sym__type_identifier_token1] = ACTIONS(1926),
    [anon_sym_if] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1924),
    [aux_sym_number_literal_token1] = ACTIONS(1926),
    [aux_sym_number_literal_token2] = ACTIONS(1924),
    [aux_sym_number_literal_token3] = ACTIONS(1924),
    [aux_sym_number_literal_token4] = ACTIONS(1926),
    [aux_sym_number_literal_token5] = ACTIONS(1924),
    [anon_sym_pipeline] = ACTIONS(1926),
    [anon_sym_return] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1926),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1924),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1924),
    [anon_sym_switch] = ACTIONS(1926),
    [anon_sym_try] = ACTIONS(1926),
    [anon_sym_int] = ACTIONS(1926),
    [anon_sym_boolean] = ACTIONS(1926),
    [anon_sym_char] = ACTIONS(1926),
    [anon_sym_short] = ACTIONS(1926),
    [anon_sym_long] = ACTIONS(1926),
    [anon_sym_float] = ACTIONS(1926),
    [anon_sym_double] = ACTIONS(1926),
    [anon_sym_void] = ACTIONS(1926),
    [anon_sym_public] = ACTIONS(1926),
    [anon_sym_protected] = ACTIONS(1926),
    [anon_sym_private] = ACTIONS(1926),
    [anon_sym_static] = ACTIONS(1926),
    [anon_sym_final] = ACTIONS(1926),
    [anon_sym_synchronized] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1924),
    [anon_sym_BANG] = ACTIONS(1924),
    [anon_sym_new] = ACTIONS(1926),
  },
  [STATE(827)] = {
    [sym_comment] = STATE(827),
    [sym_groovy_doc] = STATE(827),
    [ts_builtin_sym_end] = ACTIONS(1928),
    [sym_identifier] = ACTIONS(1930),
    [anon_sym_break] = ACTIONS(1930),
    [anon_sym_continue] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_STAR] = ACTIONS(1930),
    [anon_sym_STAR_COLON] = ACTIONS(1928),
    [anon_sym_import] = ACTIONS(1930),
    [anon_sym_package] = ACTIONS(1930),
    [anon_sym_AT] = ACTIONS(1930),
    [anon_sym_assert] = ACTIONS(1930),
    [anon_sym_PLUS_PLUS] = ACTIONS(1928),
    [anon_sym_DASH_DASH] = ACTIONS(1928),
    [anon_sym_SLASH] = ACTIONS(1930),
    [anon_sym_PLUS] = ACTIONS(1930),
    [anon_sym_DASH] = ACTIONS(1930),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_ATinterface] = ACTIONS(1928),
    [anon_sym_interface] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_COMMA] = ACTIONS(1928),
    [anon_sym_LBRACE] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1930),
    [anon_sym_def] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_RPAREN] = ACTIONS(1928),
    [anon_sym_null] = ACTIONS(1930),
    [anon_sym_this] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_while] = ACTIONS(1930),
    [anon_sym_for] = ACTIONS(1930),
    [aux_sym__type_identifier_token1] = ACTIONS(1930),
    [anon_sym_if] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1928),
    [aux_sym_number_literal_token1] = ACTIONS(1930),
    [aux_sym_number_literal_token2] = ACTIONS(1928),
    [aux_sym_number_literal_token3] = ACTIONS(1928),
    [aux_sym_number_literal_token4] = ACTIONS(1930),
    [aux_sym_number_literal_token5] = ACTIONS(1928),
    [anon_sym_pipeline] = ACTIONS(1930),
    [anon_sym_return] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1930),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1928),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1928),
    [anon_sym_switch] = ACTIONS(1930),
    [anon_sym_try] = ACTIONS(1930),
    [anon_sym_int] = ACTIONS(1930),
    [anon_sym_boolean] = ACTIONS(1930),
    [anon_sym_char] = ACTIONS(1930),
    [anon_sym_short] = ACTIONS(1930),
    [anon_sym_long] = ACTIONS(1930),
    [anon_sym_float] = ACTIONS(1930),
    [anon_sym_double] = ACTIONS(1930),
    [anon_sym_void] = ACTIONS(1930),
    [anon_sym_public] = ACTIONS(1930),
    [anon_sym_protected] = ACTIONS(1930),
    [anon_sym_private] = ACTIONS(1930),
    [anon_sym_static] = ACTIONS(1930),
    [anon_sym_final] = ACTIONS(1930),
    [anon_sym_synchronized] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1928),
    [anon_sym_BANG] = ACTIONS(1928),
    [anon_sym_new] = ACTIONS(1930),
  },
  [STATE(828)] = {
    [sym_comment] = STATE(828),
    [sym_groovy_doc] = STATE(828),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_case] = ACTIONS(1392),
    [anon_sym_default] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_catch] = ACTIONS(1392),
    [anon_sym_finally] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1390),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(829)] = {
    [sym_comment] = STATE(829),
    [sym_groovy_doc] = STATE(829),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_case] = ACTIONS(1412),
    [anon_sym_default] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_catch] = ACTIONS(1412),
    [anon_sym_finally] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(830)] = {
    [sym_comment] = STATE(830),
    [sym_groovy_doc] = STATE(830),
    [ts_builtin_sym_end] = ACTIONS(1932),
    [sym_identifier] = ACTIONS(1934),
    [anon_sym_break] = ACTIONS(1934),
    [anon_sym_continue] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_STAR] = ACTIONS(1934),
    [anon_sym_STAR_COLON] = ACTIONS(1932),
    [anon_sym_import] = ACTIONS(1934),
    [anon_sym_package] = ACTIONS(1934),
    [anon_sym_AT] = ACTIONS(1934),
    [anon_sym_assert] = ACTIONS(1934),
    [anon_sym_PLUS_PLUS] = ACTIONS(1932),
    [anon_sym_DASH_DASH] = ACTIONS(1932),
    [anon_sym_SLASH] = ACTIONS(1934),
    [anon_sym_PLUS] = ACTIONS(1934),
    [anon_sym_DASH] = ACTIONS(1934),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_ATinterface] = ACTIONS(1932),
    [anon_sym_interface] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_COMMA] = ACTIONS(1932),
    [anon_sym_LBRACE] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1934),
    [anon_sym_def] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_RPAREN] = ACTIONS(1932),
    [anon_sym_null] = ACTIONS(1934),
    [anon_sym_this] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_while] = ACTIONS(1934),
    [anon_sym_for] = ACTIONS(1934),
    [aux_sym__type_identifier_token1] = ACTIONS(1934),
    [anon_sym_if] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1932),
    [aux_sym_number_literal_token1] = ACTIONS(1934),
    [aux_sym_number_literal_token2] = ACTIONS(1932),
    [aux_sym_number_literal_token3] = ACTIONS(1932),
    [aux_sym_number_literal_token4] = ACTIONS(1934),
    [aux_sym_number_literal_token5] = ACTIONS(1932),
    [anon_sym_pipeline] = ACTIONS(1934),
    [anon_sym_return] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1934),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1932),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1932),
    [anon_sym_switch] = ACTIONS(1934),
    [anon_sym_try] = ACTIONS(1934),
    [anon_sym_int] = ACTIONS(1934),
    [anon_sym_boolean] = ACTIONS(1934),
    [anon_sym_char] = ACTIONS(1934),
    [anon_sym_short] = ACTIONS(1934),
    [anon_sym_long] = ACTIONS(1934),
    [anon_sym_float] = ACTIONS(1934),
    [anon_sym_double] = ACTIONS(1934),
    [anon_sym_void] = ACTIONS(1934),
    [anon_sym_public] = ACTIONS(1934),
    [anon_sym_protected] = ACTIONS(1934),
    [anon_sym_private] = ACTIONS(1934),
    [anon_sym_static] = ACTIONS(1934),
    [anon_sym_final] = ACTIONS(1934),
    [anon_sym_synchronized] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1932),
    [anon_sym_BANG] = ACTIONS(1932),
    [anon_sym_new] = ACTIONS(1934),
  },
  [STATE(831)] = {
    [sym_comment] = STATE(831),
    [sym_groovy_doc] = STATE(831),
    [ts_builtin_sym_end] = ACTIONS(1864),
    [sym_identifier] = ACTIONS(1866),
    [anon_sym_break] = ACTIONS(1866),
    [anon_sym_continue] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_STAR] = ACTIONS(1866),
    [anon_sym_STAR_COLON] = ACTIONS(1864),
    [anon_sym_import] = ACTIONS(1866),
    [anon_sym_package] = ACTIONS(1866),
    [anon_sym_AT] = ACTIONS(1866),
    [anon_sym_assert] = ACTIONS(1866),
    [anon_sym_PLUS_PLUS] = ACTIONS(1864),
    [anon_sym_DASH_DASH] = ACTIONS(1864),
    [anon_sym_SLASH] = ACTIONS(1866),
    [anon_sym_PLUS] = ACTIONS(1866),
    [anon_sym_DASH] = ACTIONS(1866),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_ATinterface] = ACTIONS(1864),
    [anon_sym_interface] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_COMMA] = ACTIONS(1864),
    [anon_sym_LBRACE] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1866),
    [anon_sym_def] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_RPAREN] = ACTIONS(1864),
    [anon_sym_null] = ACTIONS(1866),
    [anon_sym_this] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_while] = ACTIONS(1866),
    [anon_sym_for] = ACTIONS(1866),
    [aux_sym__type_identifier_token1] = ACTIONS(1866),
    [anon_sym_if] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1864),
    [aux_sym_number_literal_token1] = ACTIONS(1866),
    [aux_sym_number_literal_token2] = ACTIONS(1864),
    [aux_sym_number_literal_token3] = ACTIONS(1864),
    [aux_sym_number_literal_token4] = ACTIONS(1866),
    [aux_sym_number_literal_token5] = ACTIONS(1864),
    [anon_sym_pipeline] = ACTIONS(1866),
    [anon_sym_return] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1866),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1866),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1864),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1864),
    [anon_sym_switch] = ACTIONS(1866),
    [anon_sym_try] = ACTIONS(1866),
    [anon_sym_int] = ACTIONS(1866),
    [anon_sym_boolean] = ACTIONS(1866),
    [anon_sym_char] = ACTIONS(1866),
    [anon_sym_short] = ACTIONS(1866),
    [anon_sym_long] = ACTIONS(1866),
    [anon_sym_float] = ACTIONS(1866),
    [anon_sym_double] = ACTIONS(1866),
    [anon_sym_void] = ACTIONS(1866),
    [anon_sym_public] = ACTIONS(1866),
    [anon_sym_protected] = ACTIONS(1866),
    [anon_sym_private] = ACTIONS(1866),
    [anon_sym_static] = ACTIONS(1866),
    [anon_sym_final] = ACTIONS(1866),
    [anon_sym_synchronized] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1864),
    [anon_sym_BANG] = ACTIONS(1864),
    [anon_sym_new] = ACTIONS(1866),
  },
  [STATE(832)] = {
    [sym_comment] = STATE(832),
    [sym_groovy_doc] = STATE(832),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(371),
    [anon_sym_finally] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(369),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(833)] = {
    [sym_comment] = STATE(833),
    [sym_groovy_doc] = STATE(833),
    [ts_builtin_sym_end] = ACTIONS(1844),
    [sym_identifier] = ACTIONS(1846),
    [anon_sym_break] = ACTIONS(1846),
    [anon_sym_continue] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1846),
    [anon_sym_STAR_COLON] = ACTIONS(1844),
    [anon_sym_import] = ACTIONS(1846),
    [anon_sym_package] = ACTIONS(1846),
    [anon_sym_AT] = ACTIONS(1846),
    [anon_sym_assert] = ACTIONS(1846),
    [anon_sym_PLUS_PLUS] = ACTIONS(1844),
    [anon_sym_DASH_DASH] = ACTIONS(1844),
    [anon_sym_SLASH] = ACTIONS(1846),
    [anon_sym_PLUS] = ACTIONS(1846),
    [anon_sym_DASH] = ACTIONS(1846),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_ATinterface] = ACTIONS(1844),
    [anon_sym_interface] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_COMMA] = ACTIONS(1844),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1846),
    [anon_sym_def] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_RPAREN] = ACTIONS(1844),
    [anon_sym_null] = ACTIONS(1846),
    [anon_sym_this] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_while] = ACTIONS(1846),
    [anon_sym_for] = ACTIONS(1846),
    [aux_sym__type_identifier_token1] = ACTIONS(1846),
    [anon_sym_if] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1844),
    [aux_sym_number_literal_token1] = ACTIONS(1846),
    [aux_sym_number_literal_token2] = ACTIONS(1844),
    [aux_sym_number_literal_token3] = ACTIONS(1844),
    [aux_sym_number_literal_token4] = ACTIONS(1846),
    [aux_sym_number_literal_token5] = ACTIONS(1844),
    [anon_sym_pipeline] = ACTIONS(1846),
    [anon_sym_return] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1846),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1846),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1844),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1844),
    [anon_sym_switch] = ACTIONS(1846),
    [anon_sym_try] = ACTIONS(1846),
    [anon_sym_int] = ACTIONS(1846),
    [anon_sym_boolean] = ACTIONS(1846),
    [anon_sym_char] = ACTIONS(1846),
    [anon_sym_short] = ACTIONS(1846),
    [anon_sym_long] = ACTIONS(1846),
    [anon_sym_float] = ACTIONS(1846),
    [anon_sym_double] = ACTIONS(1846),
    [anon_sym_void] = ACTIONS(1846),
    [anon_sym_public] = ACTIONS(1846),
    [anon_sym_protected] = ACTIONS(1846),
    [anon_sym_private] = ACTIONS(1846),
    [anon_sym_static] = ACTIONS(1846),
    [anon_sym_final] = ACTIONS(1846),
    [anon_sym_synchronized] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1844),
    [anon_sym_new] = ACTIONS(1846),
  },
  [STATE(834)] = {
    [sym_comment] = STATE(834),
    [sym_groovy_doc] = STATE(834),
    [ts_builtin_sym_end] = ACTIONS(1852),
    [sym_identifier] = ACTIONS(1854),
    [anon_sym_break] = ACTIONS(1854),
    [anon_sym_continue] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_STAR] = ACTIONS(1854),
    [anon_sym_STAR_COLON] = ACTIONS(1852),
    [anon_sym_import] = ACTIONS(1854),
    [anon_sym_package] = ACTIONS(1854),
    [anon_sym_AT] = ACTIONS(1854),
    [anon_sym_assert] = ACTIONS(1854),
    [anon_sym_PLUS_PLUS] = ACTIONS(1852),
    [anon_sym_DASH_DASH] = ACTIONS(1852),
    [anon_sym_SLASH] = ACTIONS(1854),
    [anon_sym_PLUS] = ACTIONS(1854),
    [anon_sym_DASH] = ACTIONS(1854),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_ATinterface] = ACTIONS(1852),
    [anon_sym_interface] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_COMMA] = ACTIONS(1852),
    [anon_sym_LBRACE] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1854),
    [anon_sym_def] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_RPAREN] = ACTIONS(1852),
    [anon_sym_null] = ACTIONS(1854),
    [anon_sym_this] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_while] = ACTIONS(1854),
    [anon_sym_for] = ACTIONS(1854),
    [aux_sym__type_identifier_token1] = ACTIONS(1854),
    [anon_sym_if] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1852),
    [aux_sym_number_literal_token1] = ACTIONS(1854),
    [aux_sym_number_literal_token2] = ACTIONS(1852),
    [aux_sym_number_literal_token3] = ACTIONS(1852),
    [aux_sym_number_literal_token4] = ACTIONS(1854),
    [aux_sym_number_literal_token5] = ACTIONS(1852),
    [anon_sym_pipeline] = ACTIONS(1854),
    [anon_sym_return] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1854),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1854),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1852),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1852),
    [anon_sym_switch] = ACTIONS(1854),
    [anon_sym_try] = ACTIONS(1854),
    [anon_sym_int] = ACTIONS(1854),
    [anon_sym_boolean] = ACTIONS(1854),
    [anon_sym_char] = ACTIONS(1854),
    [anon_sym_short] = ACTIONS(1854),
    [anon_sym_long] = ACTIONS(1854),
    [anon_sym_float] = ACTIONS(1854),
    [anon_sym_double] = ACTIONS(1854),
    [anon_sym_void] = ACTIONS(1854),
    [anon_sym_public] = ACTIONS(1854),
    [anon_sym_protected] = ACTIONS(1854),
    [anon_sym_private] = ACTIONS(1854),
    [anon_sym_static] = ACTIONS(1854),
    [anon_sym_final] = ACTIONS(1854),
    [anon_sym_synchronized] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1852),
    [anon_sym_BANG] = ACTIONS(1852),
    [anon_sym_new] = ACTIONS(1854),
  },
  [STATE(835)] = {
    [sym_comment] = STATE(835),
    [sym_groovy_doc] = STATE(835),
    [sym_identifier] = ACTIONS(1806),
    [anon_sym_break] = ACTIONS(1806),
    [anon_sym_continue] = ACTIONS(1806),
    [anon_sym_SEMI] = ACTIONS(1804),
    [anon_sym_STAR] = ACTIONS(1806),
    [anon_sym_STAR_COLON] = ACTIONS(1804),
    [anon_sym_import] = ACTIONS(1806),
    [anon_sym_package] = ACTIONS(1806),
    [anon_sym_AT] = ACTIONS(1806),
    [anon_sym_assert] = ACTIONS(1806),
    [anon_sym_PLUS_PLUS] = ACTIONS(1804),
    [anon_sym_DASH_DASH] = ACTIONS(1804),
    [anon_sym_SLASH] = ACTIONS(1806),
    [anon_sym_PLUS] = ACTIONS(1806),
    [anon_sym_DASH] = ACTIONS(1806),
    [anon_sym_true] = ACTIONS(1806),
    [anon_sym_false] = ACTIONS(1806),
    [anon_sym_ATinterface] = ACTIONS(1804),
    [anon_sym_interface] = ACTIONS(1806),
    [anon_sym_class] = ACTIONS(1806),
    [anon_sym_LBRACE] = ACTIONS(1804),
    [anon_sym_RBRACE] = ACTIONS(1804),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1806),
    [anon_sym_def] = ACTIONS(1806),
    [anon_sym_LPAREN] = ACTIONS(1804),
    [anon_sym_null] = ACTIONS(1806),
    [anon_sym_this] = ACTIONS(1806),
    [anon_sym_do] = ACTIONS(1806),
    [anon_sym_while] = ACTIONS(1806),
    [anon_sym_for] = ACTIONS(1806),
    [aux_sym__type_identifier_token1] = ACTIONS(1806),
    [anon_sym_if] = ACTIONS(1806),
    [anon_sym_else] = ACTIONS(1806),
    [anon_sym_LBRACK] = ACTIONS(1804),
    [aux_sym_number_literal_token1] = ACTIONS(1806),
    [aux_sym_number_literal_token2] = ACTIONS(1804),
    [aux_sym_number_literal_token3] = ACTIONS(1804),
    [aux_sym_number_literal_token4] = ACTIONS(1806),
    [aux_sym_number_literal_token5] = ACTIONS(1804),
    [anon_sym_return] = ACTIONS(1806),
    [anon_sym_SQUOTE] = ACTIONS(1806),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1804),
    [anon_sym_DQUOTE] = ACTIONS(1806),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1804),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1804),
    [anon_sym_switch] = ACTIONS(1806),
    [anon_sym_case] = ACTIONS(1806),
    [anon_sym_default] = ACTIONS(1806),
    [anon_sym_try] = ACTIONS(1806),
    [anon_sym_catch] = ACTIONS(1806),
    [anon_sym_finally] = ACTIONS(1806),
    [anon_sym_int] = ACTIONS(1806),
    [anon_sym_boolean] = ACTIONS(1806),
    [anon_sym_char] = ACTIONS(1806),
    [anon_sym_short] = ACTIONS(1806),
    [anon_sym_long] = ACTIONS(1806),
    [anon_sym_float] = ACTIONS(1806),
    [anon_sym_double] = ACTIONS(1806),
    [anon_sym_void] = ACTIONS(1806),
    [anon_sym_public] = ACTIONS(1806),
    [anon_sym_protected] = ACTIONS(1806),
    [anon_sym_private] = ACTIONS(1806),
    [anon_sym_static] = ACTIONS(1806),
    [anon_sym_final] = ACTIONS(1806),
    [anon_sym_synchronized] = ACTIONS(1806),
    [anon_sym_TILDE] = ACTIONS(1804),
    [anon_sym_BANG] = ACTIONS(1804),
    [anon_sym_new] = ACTIONS(1806),
  },
  [STATE(836)] = {
    [sym_comment] = STATE(836),
    [sym_groovy_doc] = STATE(836),
    [ts_builtin_sym_end] = ACTIONS(1880),
    [sym_identifier] = ACTIONS(1882),
    [anon_sym_break] = ACTIONS(1882),
    [anon_sym_continue] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_STAR] = ACTIONS(1882),
    [anon_sym_STAR_COLON] = ACTIONS(1880),
    [anon_sym_import] = ACTIONS(1882),
    [anon_sym_package] = ACTIONS(1882),
    [anon_sym_AT] = ACTIONS(1882),
    [anon_sym_assert] = ACTIONS(1882),
    [anon_sym_PLUS_PLUS] = ACTIONS(1880),
    [anon_sym_DASH_DASH] = ACTIONS(1880),
    [anon_sym_SLASH] = ACTIONS(1882),
    [anon_sym_PLUS] = ACTIONS(1882),
    [anon_sym_DASH] = ACTIONS(1882),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_ATinterface] = ACTIONS(1880),
    [anon_sym_interface] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_COMMA] = ACTIONS(1880),
    [anon_sym_LBRACE] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1882),
    [anon_sym_def] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_RPAREN] = ACTIONS(1880),
    [anon_sym_null] = ACTIONS(1882),
    [anon_sym_this] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_while] = ACTIONS(1882),
    [anon_sym_for] = ACTIONS(1882),
    [aux_sym__type_identifier_token1] = ACTIONS(1882),
    [anon_sym_if] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1880),
    [aux_sym_number_literal_token1] = ACTIONS(1882),
    [aux_sym_number_literal_token2] = ACTIONS(1880),
    [aux_sym_number_literal_token3] = ACTIONS(1880),
    [aux_sym_number_literal_token4] = ACTIONS(1882),
    [aux_sym_number_literal_token5] = ACTIONS(1880),
    [anon_sym_pipeline] = ACTIONS(1882),
    [anon_sym_return] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1882),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1882),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1880),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1880),
    [anon_sym_switch] = ACTIONS(1882),
    [anon_sym_try] = ACTIONS(1882),
    [anon_sym_int] = ACTIONS(1882),
    [anon_sym_boolean] = ACTIONS(1882),
    [anon_sym_char] = ACTIONS(1882),
    [anon_sym_short] = ACTIONS(1882),
    [anon_sym_long] = ACTIONS(1882),
    [anon_sym_float] = ACTIONS(1882),
    [anon_sym_double] = ACTIONS(1882),
    [anon_sym_void] = ACTIONS(1882),
    [anon_sym_public] = ACTIONS(1882),
    [anon_sym_protected] = ACTIONS(1882),
    [anon_sym_private] = ACTIONS(1882),
    [anon_sym_static] = ACTIONS(1882),
    [anon_sym_final] = ACTIONS(1882),
    [anon_sym_synchronized] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1880),
    [anon_sym_BANG] = ACTIONS(1880),
    [anon_sym_new] = ACTIONS(1882),
  },
  [STATE(837)] = {
    [sym_comment] = STATE(837),
    [sym_groovy_doc] = STATE(837),
    [sym_identifier] = ACTIONS(1824),
    [anon_sym_break] = ACTIONS(1824),
    [anon_sym_continue] = ACTIONS(1824),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_STAR] = ACTIONS(1824),
    [anon_sym_STAR_COLON] = ACTIONS(1822),
    [anon_sym_import] = ACTIONS(1824),
    [anon_sym_package] = ACTIONS(1824),
    [anon_sym_AT] = ACTIONS(1824),
    [anon_sym_assert] = ACTIONS(1824),
    [anon_sym_PLUS_PLUS] = ACTIONS(1822),
    [anon_sym_DASH_DASH] = ACTIONS(1822),
    [anon_sym_SLASH] = ACTIONS(1824),
    [anon_sym_PLUS] = ACTIONS(1824),
    [anon_sym_DASH] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1824),
    [anon_sym_false] = ACTIONS(1824),
    [anon_sym_ATinterface] = ACTIONS(1822),
    [anon_sym_interface] = ACTIONS(1824),
    [anon_sym_class] = ACTIONS(1824),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_RBRACE] = ACTIONS(1822),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1824),
    [anon_sym_def] = ACTIONS(1824),
    [anon_sym_LPAREN] = ACTIONS(1822),
    [anon_sym_null] = ACTIONS(1824),
    [anon_sym_this] = ACTIONS(1824),
    [anon_sym_do] = ACTIONS(1824),
    [anon_sym_while] = ACTIONS(1824),
    [anon_sym_for] = ACTIONS(1824),
    [aux_sym__type_identifier_token1] = ACTIONS(1824),
    [anon_sym_if] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [aux_sym_number_literal_token1] = ACTIONS(1824),
    [aux_sym_number_literal_token2] = ACTIONS(1822),
    [aux_sym_number_literal_token3] = ACTIONS(1822),
    [aux_sym_number_literal_token4] = ACTIONS(1824),
    [aux_sym_number_literal_token5] = ACTIONS(1822),
    [anon_sym_return] = ACTIONS(1824),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1822),
    [anon_sym_switch] = ACTIONS(1824),
    [anon_sym_case] = ACTIONS(1824),
    [anon_sym_default] = ACTIONS(1824),
    [anon_sym_try] = ACTIONS(1824),
    [anon_sym_catch] = ACTIONS(1824),
    [anon_sym_finally] = ACTIONS(1824),
    [anon_sym_int] = ACTIONS(1824),
    [anon_sym_boolean] = ACTIONS(1824),
    [anon_sym_char] = ACTIONS(1824),
    [anon_sym_short] = ACTIONS(1824),
    [anon_sym_long] = ACTIONS(1824),
    [anon_sym_float] = ACTIONS(1824),
    [anon_sym_double] = ACTIONS(1824),
    [anon_sym_void] = ACTIONS(1824),
    [anon_sym_public] = ACTIONS(1824),
    [anon_sym_protected] = ACTIONS(1824),
    [anon_sym_private] = ACTIONS(1824),
    [anon_sym_static] = ACTIONS(1824),
    [anon_sym_final] = ACTIONS(1824),
    [anon_sym_synchronized] = ACTIONS(1824),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_BANG] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1824),
  },
  [STATE(838)] = {
    [sym_comment] = STATE(838),
    [sym_groovy_doc] = STATE(838),
    [sym_identifier] = ACTIONS(1759),
    [anon_sym_break] = ACTIONS(1759),
    [anon_sym_continue] = ACTIONS(1759),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_STAR] = ACTIONS(1759),
    [anon_sym_STAR_COLON] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1757),
    [anon_sym_import] = ACTIONS(1759),
    [anon_sym_as] = ACTIONS(1759),
    [anon_sym_package] = ACTIONS(1759),
    [anon_sym_AT] = ACTIONS(1759),
    [anon_sym_assert] = ACTIONS(1759),
    [anon_sym_PLUS_PLUS] = ACTIONS(1757),
    [anon_sym_DASH_DASH] = ACTIONS(1757),
    [anon_sym_SLASH] = ACTIONS(1759),
    [anon_sym_PLUS] = ACTIONS(1759),
    [anon_sym_DASH] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1759),
    [anon_sym_false] = ACTIONS(1759),
    [anon_sym_ATinterface] = ACTIONS(1757),
    [anon_sym_interface] = ACTIONS(1759),
    [anon_sym_class] = ACTIONS(1759),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_RBRACE] = ACTIONS(1757),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1759),
    [anon_sym_def] = ACTIONS(1759),
    [anon_sym_LPAREN] = ACTIONS(1757),
    [anon_sym_null] = ACTIONS(1759),
    [anon_sym_this] = ACTIONS(1759),
    [anon_sym_do] = ACTIONS(1759),
    [anon_sym_while] = ACTIONS(1759),
    [anon_sym_for] = ACTIONS(1759),
    [aux_sym__type_identifier_token1] = ACTIONS(1759),
    [anon_sym_if] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [aux_sym_number_literal_token1] = ACTIONS(1759),
    [aux_sym_number_literal_token2] = ACTIONS(1757),
    [aux_sym_number_literal_token3] = ACTIONS(1757),
    [aux_sym_number_literal_token4] = ACTIONS(1759),
    [aux_sym_number_literal_token5] = ACTIONS(1757),
    [anon_sym_return] = ACTIONS(1759),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1757),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1757),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1757),
    [anon_sym_switch] = ACTIONS(1759),
    [anon_sym_case] = ACTIONS(1759),
    [anon_sym_default] = ACTIONS(1759),
    [anon_sym_try] = ACTIONS(1759),
    [anon_sym_int] = ACTIONS(1759),
    [anon_sym_boolean] = ACTIONS(1759),
    [anon_sym_char] = ACTIONS(1759),
    [anon_sym_short] = ACTIONS(1759),
    [anon_sym_long] = ACTIONS(1759),
    [anon_sym_float] = ACTIONS(1759),
    [anon_sym_double] = ACTIONS(1759),
    [anon_sym_void] = ACTIONS(1759),
    [anon_sym_public] = ACTIONS(1759),
    [anon_sym_protected] = ACTIONS(1759),
    [anon_sym_private] = ACTIONS(1759),
    [anon_sym_static] = ACTIONS(1759),
    [anon_sym_final] = ACTIONS(1759),
    [anon_sym_synchronized] = ACTIONS(1759),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_BANG] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1759),
  },
  [STATE(839)] = {
    [sym_comment] = STATE(839),
    [sym_groovy_doc] = STATE(839),
    [sym_identifier] = ACTIONS(1828),
    [anon_sym_break] = ACTIONS(1828),
    [anon_sym_continue] = ACTIONS(1828),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_STAR] = ACTIONS(1828),
    [anon_sym_STAR_COLON] = ACTIONS(1826),
    [anon_sym_import] = ACTIONS(1828),
    [anon_sym_package] = ACTIONS(1828),
    [anon_sym_AT] = ACTIONS(1828),
    [anon_sym_assert] = ACTIONS(1828),
    [anon_sym_PLUS_PLUS] = ACTIONS(1826),
    [anon_sym_DASH_DASH] = ACTIONS(1826),
    [anon_sym_SLASH] = ACTIONS(1828),
    [anon_sym_PLUS] = ACTIONS(1828),
    [anon_sym_DASH] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1828),
    [anon_sym_false] = ACTIONS(1828),
    [anon_sym_ATinterface] = ACTIONS(1826),
    [anon_sym_interface] = ACTIONS(1828),
    [anon_sym_class] = ACTIONS(1828),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_RBRACE] = ACTIONS(1826),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1828),
    [anon_sym_def] = ACTIONS(1828),
    [anon_sym_LPAREN] = ACTIONS(1826),
    [anon_sym_null] = ACTIONS(1828),
    [anon_sym_this] = ACTIONS(1828),
    [anon_sym_do] = ACTIONS(1828),
    [anon_sym_while] = ACTIONS(1828),
    [anon_sym_for] = ACTIONS(1828),
    [aux_sym__type_identifier_token1] = ACTIONS(1828),
    [anon_sym_if] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [aux_sym_number_literal_token1] = ACTIONS(1828),
    [aux_sym_number_literal_token2] = ACTIONS(1826),
    [aux_sym_number_literal_token3] = ACTIONS(1826),
    [aux_sym_number_literal_token4] = ACTIONS(1828),
    [aux_sym_number_literal_token5] = ACTIONS(1826),
    [anon_sym_return] = ACTIONS(1828),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1826),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1826),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1826),
    [anon_sym_switch] = ACTIONS(1828),
    [anon_sym_case] = ACTIONS(1828),
    [anon_sym_default] = ACTIONS(1828),
    [anon_sym_try] = ACTIONS(1828),
    [anon_sym_catch] = ACTIONS(1828),
    [anon_sym_finally] = ACTIONS(1828),
    [anon_sym_int] = ACTIONS(1828),
    [anon_sym_boolean] = ACTIONS(1828),
    [anon_sym_char] = ACTIONS(1828),
    [anon_sym_short] = ACTIONS(1828),
    [anon_sym_long] = ACTIONS(1828),
    [anon_sym_float] = ACTIONS(1828),
    [anon_sym_double] = ACTIONS(1828),
    [anon_sym_void] = ACTIONS(1828),
    [anon_sym_public] = ACTIONS(1828),
    [anon_sym_protected] = ACTIONS(1828),
    [anon_sym_private] = ACTIONS(1828),
    [anon_sym_static] = ACTIONS(1828),
    [anon_sym_final] = ACTIONS(1828),
    [anon_sym_synchronized] = ACTIONS(1828),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_BANG] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1828),
  },
  [STATE(840)] = {
    [sym_comment] = STATE(840),
    [sym_groovy_doc] = STATE(840),
    [sym_identifier] = ACTIONS(1842),
    [anon_sym_break] = ACTIONS(1842),
    [anon_sym_continue] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_STAR] = ACTIONS(1842),
    [anon_sym_STAR_COLON] = ACTIONS(1840),
    [anon_sym_import] = ACTIONS(1842),
    [anon_sym_package] = ACTIONS(1842),
    [anon_sym_AT] = ACTIONS(1842),
    [anon_sym_assert] = ACTIONS(1842),
    [anon_sym_PLUS_PLUS] = ACTIONS(1840),
    [anon_sym_DASH_DASH] = ACTIONS(1840),
    [anon_sym_SLASH] = ACTIONS(1842),
    [anon_sym_PLUS] = ACTIONS(1842),
    [anon_sym_DASH] = ACTIONS(1842),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_ATinterface] = ACTIONS(1840),
    [anon_sym_interface] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1842),
    [anon_sym_def] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_null] = ACTIONS(1842),
    [anon_sym_this] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_while] = ACTIONS(1842),
    [anon_sym_for] = ACTIONS(1842),
    [aux_sym__type_identifier_token1] = ACTIONS(1842),
    [anon_sym_if] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1840),
    [aux_sym_number_literal_token1] = ACTIONS(1842),
    [aux_sym_number_literal_token2] = ACTIONS(1840),
    [aux_sym_number_literal_token3] = ACTIONS(1840),
    [aux_sym_number_literal_token4] = ACTIONS(1842),
    [aux_sym_number_literal_token5] = ACTIONS(1840),
    [anon_sym_return] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1842),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1842),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1840),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1840),
    [anon_sym_switch] = ACTIONS(1842),
    [anon_sym_case] = ACTIONS(1842),
    [anon_sym_default] = ACTIONS(1842),
    [anon_sym_try] = ACTIONS(1842),
    [anon_sym_catch] = ACTIONS(1842),
    [anon_sym_finally] = ACTIONS(1842),
    [anon_sym_int] = ACTIONS(1842),
    [anon_sym_boolean] = ACTIONS(1842),
    [anon_sym_char] = ACTIONS(1842),
    [anon_sym_short] = ACTIONS(1842),
    [anon_sym_long] = ACTIONS(1842),
    [anon_sym_float] = ACTIONS(1842),
    [anon_sym_double] = ACTIONS(1842),
    [anon_sym_void] = ACTIONS(1842),
    [anon_sym_public] = ACTIONS(1842),
    [anon_sym_protected] = ACTIONS(1842),
    [anon_sym_private] = ACTIONS(1842),
    [anon_sym_static] = ACTIONS(1842),
    [anon_sym_final] = ACTIONS(1842),
    [anon_sym_synchronized] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1840),
    [anon_sym_BANG] = ACTIONS(1840),
    [anon_sym_new] = ACTIONS(1842),
  },
  [STATE(841)] = {
    [sym_comment] = STATE(841),
    [sym_groovy_doc] = STATE(841),
    [ts_builtin_sym_end] = ACTIONS(389),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(389),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_pipeline] = ACTIONS(387),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(389),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(842)] = {
    [sym_comment] = STATE(842),
    [sym_groovy_doc] = STATE(842),
    [sym_identifier] = ACTIONS(1846),
    [anon_sym_break] = ACTIONS(1846),
    [anon_sym_continue] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1846),
    [anon_sym_STAR_COLON] = ACTIONS(1844),
    [anon_sym_import] = ACTIONS(1846),
    [anon_sym_package] = ACTIONS(1846),
    [anon_sym_AT] = ACTIONS(1846),
    [anon_sym_assert] = ACTIONS(1846),
    [anon_sym_PLUS_PLUS] = ACTIONS(1844),
    [anon_sym_DASH_DASH] = ACTIONS(1844),
    [anon_sym_SLASH] = ACTIONS(1846),
    [anon_sym_PLUS] = ACTIONS(1846),
    [anon_sym_DASH] = ACTIONS(1846),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_ATinterface] = ACTIONS(1844),
    [anon_sym_interface] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1846),
    [anon_sym_def] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_null] = ACTIONS(1846),
    [anon_sym_this] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_while] = ACTIONS(1846),
    [anon_sym_for] = ACTIONS(1846),
    [aux_sym__type_identifier_token1] = ACTIONS(1846),
    [anon_sym_if] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1844),
    [aux_sym_number_literal_token1] = ACTIONS(1846),
    [aux_sym_number_literal_token2] = ACTIONS(1844),
    [aux_sym_number_literal_token3] = ACTIONS(1844),
    [aux_sym_number_literal_token4] = ACTIONS(1846),
    [aux_sym_number_literal_token5] = ACTIONS(1844),
    [anon_sym_return] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1846),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1846),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1844),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1844),
    [anon_sym_switch] = ACTIONS(1846),
    [anon_sym_case] = ACTIONS(1846),
    [anon_sym_default] = ACTIONS(1846),
    [anon_sym_try] = ACTIONS(1846),
    [anon_sym_catch] = ACTIONS(1846),
    [anon_sym_finally] = ACTIONS(1846),
    [anon_sym_int] = ACTIONS(1846),
    [anon_sym_boolean] = ACTIONS(1846),
    [anon_sym_char] = ACTIONS(1846),
    [anon_sym_short] = ACTIONS(1846),
    [anon_sym_long] = ACTIONS(1846),
    [anon_sym_float] = ACTIONS(1846),
    [anon_sym_double] = ACTIONS(1846),
    [anon_sym_void] = ACTIONS(1846),
    [anon_sym_public] = ACTIONS(1846),
    [anon_sym_protected] = ACTIONS(1846),
    [anon_sym_private] = ACTIONS(1846),
    [anon_sym_static] = ACTIONS(1846),
    [anon_sym_final] = ACTIONS(1846),
    [anon_sym_synchronized] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1844),
    [anon_sym_new] = ACTIONS(1846),
  },
  [STATE(843)] = {
    [sym_comment] = STATE(843),
    [sym_groovy_doc] = STATE(843),
    [sym_identifier] = ACTIONS(1854),
    [anon_sym_break] = ACTIONS(1854),
    [anon_sym_continue] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_STAR] = ACTIONS(1854),
    [anon_sym_STAR_COLON] = ACTIONS(1852),
    [anon_sym_import] = ACTIONS(1854),
    [anon_sym_package] = ACTIONS(1854),
    [anon_sym_AT] = ACTIONS(1854),
    [anon_sym_assert] = ACTIONS(1854),
    [anon_sym_PLUS_PLUS] = ACTIONS(1852),
    [anon_sym_DASH_DASH] = ACTIONS(1852),
    [anon_sym_SLASH] = ACTIONS(1854),
    [anon_sym_PLUS] = ACTIONS(1854),
    [anon_sym_DASH] = ACTIONS(1854),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_ATinterface] = ACTIONS(1852),
    [anon_sym_interface] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1854),
    [anon_sym_def] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_null] = ACTIONS(1854),
    [anon_sym_this] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_while] = ACTIONS(1854),
    [anon_sym_for] = ACTIONS(1854),
    [aux_sym__type_identifier_token1] = ACTIONS(1854),
    [anon_sym_if] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1852),
    [aux_sym_number_literal_token1] = ACTIONS(1854),
    [aux_sym_number_literal_token2] = ACTIONS(1852),
    [aux_sym_number_literal_token3] = ACTIONS(1852),
    [aux_sym_number_literal_token4] = ACTIONS(1854),
    [aux_sym_number_literal_token5] = ACTIONS(1852),
    [anon_sym_return] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1854),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1854),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1852),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1852),
    [anon_sym_switch] = ACTIONS(1854),
    [anon_sym_case] = ACTIONS(1854),
    [anon_sym_default] = ACTIONS(1854),
    [anon_sym_try] = ACTIONS(1854),
    [anon_sym_catch] = ACTIONS(1854),
    [anon_sym_finally] = ACTIONS(1854),
    [anon_sym_int] = ACTIONS(1854),
    [anon_sym_boolean] = ACTIONS(1854),
    [anon_sym_char] = ACTIONS(1854),
    [anon_sym_short] = ACTIONS(1854),
    [anon_sym_long] = ACTIONS(1854),
    [anon_sym_float] = ACTIONS(1854),
    [anon_sym_double] = ACTIONS(1854),
    [anon_sym_void] = ACTIONS(1854),
    [anon_sym_public] = ACTIONS(1854),
    [anon_sym_protected] = ACTIONS(1854),
    [anon_sym_private] = ACTIONS(1854),
    [anon_sym_static] = ACTIONS(1854),
    [anon_sym_final] = ACTIONS(1854),
    [anon_sym_synchronized] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1852),
    [anon_sym_BANG] = ACTIONS(1852),
    [anon_sym_new] = ACTIONS(1854),
  },
  [STATE(844)] = {
    [sym_comment] = STATE(844),
    [sym_groovy_doc] = STATE(844),
    [sym_identifier] = ACTIONS(1882),
    [anon_sym_break] = ACTIONS(1882),
    [anon_sym_continue] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_STAR] = ACTIONS(1882),
    [anon_sym_STAR_COLON] = ACTIONS(1880),
    [anon_sym_import] = ACTIONS(1882),
    [anon_sym_package] = ACTIONS(1882),
    [anon_sym_AT] = ACTIONS(1882),
    [anon_sym_assert] = ACTIONS(1882),
    [anon_sym_PLUS_PLUS] = ACTIONS(1880),
    [anon_sym_DASH_DASH] = ACTIONS(1880),
    [anon_sym_SLASH] = ACTIONS(1882),
    [anon_sym_PLUS] = ACTIONS(1882),
    [anon_sym_DASH] = ACTIONS(1882),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_ATinterface] = ACTIONS(1880),
    [anon_sym_interface] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1882),
    [anon_sym_def] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_null] = ACTIONS(1882),
    [anon_sym_this] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_while] = ACTIONS(1882),
    [anon_sym_for] = ACTIONS(1882),
    [aux_sym__type_identifier_token1] = ACTIONS(1882),
    [anon_sym_if] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1880),
    [aux_sym_number_literal_token1] = ACTIONS(1882),
    [aux_sym_number_literal_token2] = ACTIONS(1880),
    [aux_sym_number_literal_token3] = ACTIONS(1880),
    [aux_sym_number_literal_token4] = ACTIONS(1882),
    [aux_sym_number_literal_token5] = ACTIONS(1880),
    [anon_sym_return] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1882),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1882),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1880),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1880),
    [anon_sym_switch] = ACTIONS(1882),
    [anon_sym_case] = ACTIONS(1882),
    [anon_sym_default] = ACTIONS(1882),
    [anon_sym_try] = ACTIONS(1882),
    [anon_sym_catch] = ACTIONS(1882),
    [anon_sym_finally] = ACTIONS(1882),
    [anon_sym_int] = ACTIONS(1882),
    [anon_sym_boolean] = ACTIONS(1882),
    [anon_sym_char] = ACTIONS(1882),
    [anon_sym_short] = ACTIONS(1882),
    [anon_sym_long] = ACTIONS(1882),
    [anon_sym_float] = ACTIONS(1882),
    [anon_sym_double] = ACTIONS(1882),
    [anon_sym_void] = ACTIONS(1882),
    [anon_sym_public] = ACTIONS(1882),
    [anon_sym_protected] = ACTIONS(1882),
    [anon_sym_private] = ACTIONS(1882),
    [anon_sym_static] = ACTIONS(1882),
    [anon_sym_final] = ACTIONS(1882),
    [anon_sym_synchronized] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1880),
    [anon_sym_BANG] = ACTIONS(1880),
    [anon_sym_new] = ACTIONS(1882),
  },
  [STATE(845)] = {
    [sym_comment] = STATE(845),
    [sym_groovy_doc] = STATE(845),
    [ts_builtin_sym_end] = ACTIONS(1940),
    [sym_identifier] = ACTIONS(1942),
    [anon_sym_break] = ACTIONS(1942),
    [anon_sym_continue] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1942),
    [anon_sym_STAR_COLON] = ACTIONS(1940),
    [anon_sym_import] = ACTIONS(1942),
    [anon_sym_package] = ACTIONS(1942),
    [anon_sym_AT] = ACTIONS(1942),
    [anon_sym_assert] = ACTIONS(1942),
    [anon_sym_PLUS_PLUS] = ACTIONS(1940),
    [anon_sym_DASH_DASH] = ACTIONS(1940),
    [anon_sym_SLASH] = ACTIONS(1942),
    [anon_sym_PLUS] = ACTIONS(1942),
    [anon_sym_DASH] = ACTIONS(1942),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_ATinterface] = ACTIONS(1940),
    [anon_sym_interface] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_COMMA] = ACTIONS(1940),
    [anon_sym_LBRACE] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1942),
    [anon_sym_def] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_RPAREN] = ACTIONS(1940),
    [anon_sym_null] = ACTIONS(1942),
    [anon_sym_this] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_for] = ACTIONS(1942),
    [aux_sym__type_identifier_token1] = ACTIONS(1942),
    [anon_sym_if] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1940),
    [aux_sym_number_literal_token1] = ACTIONS(1942),
    [aux_sym_number_literal_token2] = ACTIONS(1940),
    [aux_sym_number_literal_token3] = ACTIONS(1940),
    [aux_sym_number_literal_token4] = ACTIONS(1942),
    [aux_sym_number_literal_token5] = ACTIONS(1940),
    [anon_sym_pipeline] = ACTIONS(1942),
    [anon_sym_return] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1942),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1940),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1940),
    [anon_sym_switch] = ACTIONS(1942),
    [anon_sym_try] = ACTIONS(1942),
    [anon_sym_int] = ACTIONS(1942),
    [anon_sym_boolean] = ACTIONS(1942),
    [anon_sym_char] = ACTIONS(1942),
    [anon_sym_short] = ACTIONS(1942),
    [anon_sym_long] = ACTIONS(1942),
    [anon_sym_float] = ACTIONS(1942),
    [anon_sym_double] = ACTIONS(1942),
    [anon_sym_void] = ACTIONS(1942),
    [anon_sym_public] = ACTIONS(1942),
    [anon_sym_protected] = ACTIONS(1942),
    [anon_sym_private] = ACTIONS(1942),
    [anon_sym_static] = ACTIONS(1942),
    [anon_sym_final] = ACTIONS(1942),
    [anon_sym_synchronized] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1940),
    [anon_sym_BANG] = ACTIONS(1940),
    [anon_sym_new] = ACTIONS(1942),
  },
  [STATE(846)] = {
    [sym_comment] = STATE(846),
    [sym_groovy_doc] = STATE(846),
    [ts_builtin_sym_end] = ACTIONS(1948),
    [sym_identifier] = ACTIONS(1950),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_STAR] = ACTIONS(1950),
    [anon_sym_STAR_COLON] = ACTIONS(1948),
    [anon_sym_import] = ACTIONS(1950),
    [anon_sym_package] = ACTIONS(1950),
    [anon_sym_AT] = ACTIONS(1950),
    [anon_sym_assert] = ACTIONS(1950),
    [anon_sym_PLUS_PLUS] = ACTIONS(1948),
    [anon_sym_DASH_DASH] = ACTIONS(1948),
    [anon_sym_SLASH] = ACTIONS(1950),
    [anon_sym_PLUS] = ACTIONS(1950),
    [anon_sym_DASH] = ACTIONS(1950),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_ATinterface] = ACTIONS(1948),
    [anon_sym_interface] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_COMMA] = ACTIONS(1948),
    [anon_sym_LBRACE] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1950),
    [anon_sym_def] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_RPAREN] = ACTIONS(1948),
    [anon_sym_null] = ACTIONS(1950),
    [anon_sym_this] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_while] = ACTIONS(1950),
    [anon_sym_for] = ACTIONS(1950),
    [aux_sym__type_identifier_token1] = ACTIONS(1950),
    [anon_sym_if] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1948),
    [aux_sym_number_literal_token1] = ACTIONS(1950),
    [aux_sym_number_literal_token2] = ACTIONS(1948),
    [aux_sym_number_literal_token3] = ACTIONS(1948),
    [aux_sym_number_literal_token4] = ACTIONS(1950),
    [aux_sym_number_literal_token5] = ACTIONS(1948),
    [anon_sym_pipeline] = ACTIONS(1950),
    [anon_sym_return] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1950),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1948),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1948),
    [anon_sym_switch] = ACTIONS(1950),
    [anon_sym_try] = ACTIONS(1950),
    [anon_sym_int] = ACTIONS(1950),
    [anon_sym_boolean] = ACTIONS(1950),
    [anon_sym_char] = ACTIONS(1950),
    [anon_sym_short] = ACTIONS(1950),
    [anon_sym_long] = ACTIONS(1950),
    [anon_sym_float] = ACTIONS(1950),
    [anon_sym_double] = ACTIONS(1950),
    [anon_sym_void] = ACTIONS(1950),
    [anon_sym_public] = ACTIONS(1950),
    [anon_sym_protected] = ACTIONS(1950),
    [anon_sym_private] = ACTIONS(1950),
    [anon_sym_static] = ACTIONS(1950),
    [anon_sym_final] = ACTIONS(1950),
    [anon_sym_synchronized] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1948),
    [anon_sym_BANG] = ACTIONS(1948),
    [anon_sym_new] = ACTIONS(1950),
  },
  [STATE(847)] = {
    [sym_comment] = STATE(847),
    [sym_groovy_doc] = STATE(847),
    [sym_identifier] = ACTIONS(2031),
    [anon_sym_break] = ACTIONS(2031),
    [anon_sym_continue] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_STAR] = ACTIONS(2031),
    [anon_sym_STAR_COLON] = ACTIONS(2029),
    [anon_sym_import] = ACTIONS(2031),
    [anon_sym_package] = ACTIONS(2031),
    [anon_sym_AT] = ACTIONS(2031),
    [anon_sym_assert] = ACTIONS(2031),
    [anon_sym_PLUS_PLUS] = ACTIONS(2029),
    [anon_sym_DASH_DASH] = ACTIONS(2029),
    [anon_sym_SLASH] = ACTIONS(2031),
    [anon_sym_PLUS] = ACTIONS(2031),
    [anon_sym_DASH] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_ATinterface] = ACTIONS(2029),
    [anon_sym_interface] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2031),
    [anon_sym_def] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_null] = ACTIONS(2031),
    [anon_sym_this] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_while] = ACTIONS(2031),
    [anon_sym_for] = ACTIONS(2031),
    [aux_sym__type_identifier_token1] = ACTIONS(2031),
    [anon_sym_if] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2029),
    [aux_sym_number_literal_token1] = ACTIONS(2031),
    [aux_sym_number_literal_token2] = ACTIONS(2029),
    [aux_sym_number_literal_token3] = ACTIONS(2029),
    [aux_sym_number_literal_token4] = ACTIONS(2031),
    [aux_sym_number_literal_token5] = ACTIONS(2029),
    [anon_sym_return] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2031),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2031),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2029),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2029),
    [anon_sym_switch] = ACTIONS(2031),
    [anon_sym_case] = ACTIONS(2031),
    [anon_sym_default] = ACTIONS(2031),
    [anon_sym_try] = ACTIONS(2031),
    [anon_sym_catch] = ACTIONS(2031),
    [anon_sym_finally] = ACTIONS(2031),
    [anon_sym_int] = ACTIONS(2031),
    [anon_sym_boolean] = ACTIONS(2031),
    [anon_sym_char] = ACTIONS(2031),
    [anon_sym_short] = ACTIONS(2031),
    [anon_sym_long] = ACTIONS(2031),
    [anon_sym_float] = ACTIONS(2031),
    [anon_sym_double] = ACTIONS(2031),
    [anon_sym_void] = ACTIONS(2031),
    [anon_sym_public] = ACTIONS(2031),
    [anon_sym_protected] = ACTIONS(2031),
    [anon_sym_private] = ACTIONS(2031),
    [anon_sym_static] = ACTIONS(2031),
    [anon_sym_final] = ACTIONS(2031),
    [anon_sym_synchronized] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2029),
    [anon_sym_BANG] = ACTIONS(2029),
    [anon_sym_new] = ACTIONS(2031),
  },
  [STATE(848)] = {
    [sym_comment] = STATE(848),
    [sym_groovy_doc] = STATE(848),
    [sym_identifier] = ACTIONS(1902),
    [anon_sym_break] = ACTIONS(1902),
    [anon_sym_continue] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_STAR] = ACTIONS(1902),
    [anon_sym_STAR_COLON] = ACTIONS(1900),
    [anon_sym_import] = ACTIONS(1902),
    [anon_sym_package] = ACTIONS(1902),
    [anon_sym_AT] = ACTIONS(1902),
    [anon_sym_assert] = ACTIONS(1902),
    [anon_sym_PLUS_PLUS] = ACTIONS(1900),
    [anon_sym_DASH_DASH] = ACTIONS(1900),
    [anon_sym_SLASH] = ACTIONS(1902),
    [anon_sym_PLUS] = ACTIONS(1902),
    [anon_sym_DASH] = ACTIONS(1902),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_ATinterface] = ACTIONS(1900),
    [anon_sym_interface] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1902),
    [anon_sym_def] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_null] = ACTIONS(1902),
    [anon_sym_this] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_while] = ACTIONS(1902),
    [anon_sym_for] = ACTIONS(1902),
    [aux_sym__type_identifier_token1] = ACTIONS(1902),
    [anon_sym_if] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1900),
    [aux_sym_number_literal_token1] = ACTIONS(1902),
    [aux_sym_number_literal_token2] = ACTIONS(1900),
    [aux_sym_number_literal_token3] = ACTIONS(1900),
    [aux_sym_number_literal_token4] = ACTIONS(1902),
    [aux_sym_number_literal_token5] = ACTIONS(1900),
    [anon_sym_return] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1902),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1902),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1900),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1900),
    [anon_sym_switch] = ACTIONS(1902),
    [anon_sym_case] = ACTIONS(1902),
    [anon_sym_default] = ACTIONS(1902),
    [anon_sym_try] = ACTIONS(1902),
    [anon_sym_catch] = ACTIONS(1902),
    [anon_sym_finally] = ACTIONS(1902),
    [anon_sym_int] = ACTIONS(1902),
    [anon_sym_boolean] = ACTIONS(1902),
    [anon_sym_char] = ACTIONS(1902),
    [anon_sym_short] = ACTIONS(1902),
    [anon_sym_long] = ACTIONS(1902),
    [anon_sym_float] = ACTIONS(1902),
    [anon_sym_double] = ACTIONS(1902),
    [anon_sym_void] = ACTIONS(1902),
    [anon_sym_public] = ACTIONS(1902),
    [anon_sym_protected] = ACTIONS(1902),
    [anon_sym_private] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1902),
    [anon_sym_final] = ACTIONS(1902),
    [anon_sym_synchronized] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1900),
    [anon_sym_BANG] = ACTIONS(1900),
    [anon_sym_new] = ACTIONS(1902),
  },
  [STATE(849)] = {
    [sym_comment] = STATE(849),
    [sym_groovy_doc] = STATE(849),
    [ts_builtin_sym_end] = ACTIONS(1960),
    [sym_identifier] = ACTIONS(1962),
    [anon_sym_break] = ACTIONS(1962),
    [anon_sym_continue] = ACTIONS(1962),
    [anon_sym_SEMI] = ACTIONS(1960),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_STAR_COLON] = ACTIONS(1960),
    [anon_sym_import] = ACTIONS(1962),
    [anon_sym_package] = ACTIONS(1962),
    [anon_sym_AT] = ACTIONS(1962),
    [anon_sym_assert] = ACTIONS(1962),
    [anon_sym_PLUS_PLUS] = ACTIONS(1960),
    [anon_sym_DASH_DASH] = ACTIONS(1960),
    [anon_sym_SLASH] = ACTIONS(1962),
    [anon_sym_PLUS] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_true] = ACTIONS(1962),
    [anon_sym_false] = ACTIONS(1962),
    [anon_sym_ATinterface] = ACTIONS(1960),
    [anon_sym_interface] = ACTIONS(1962),
    [anon_sym_class] = ACTIONS(1962),
    [anon_sym_COMMA] = ACTIONS(1960),
    [anon_sym_LBRACE] = ACTIONS(1960),
    [anon_sym_RBRACE] = ACTIONS(1960),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1962),
    [anon_sym_def] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1960),
    [anon_sym_RPAREN] = ACTIONS(1960),
    [anon_sym_null] = ACTIONS(1962),
    [anon_sym_this] = ACTIONS(1962),
    [anon_sym_do] = ACTIONS(1962),
    [anon_sym_while] = ACTIONS(1962),
    [anon_sym_for] = ACTIONS(1962),
    [aux_sym__type_identifier_token1] = ACTIONS(1962),
    [anon_sym_if] = ACTIONS(1962),
    [anon_sym_else] = ACTIONS(1962),
    [anon_sym_LBRACK] = ACTIONS(1960),
    [aux_sym_number_literal_token1] = ACTIONS(1962),
    [aux_sym_number_literal_token2] = ACTIONS(1960),
    [aux_sym_number_literal_token3] = ACTIONS(1960),
    [aux_sym_number_literal_token4] = ACTIONS(1962),
    [aux_sym_number_literal_token5] = ACTIONS(1960),
    [anon_sym_pipeline] = ACTIONS(1962),
    [anon_sym_return] = ACTIONS(1962),
    [anon_sym_SQUOTE] = ACTIONS(1962),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1960),
    [anon_sym_DQUOTE] = ACTIONS(1962),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1960),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1960),
    [anon_sym_switch] = ACTIONS(1962),
    [anon_sym_try] = ACTIONS(1962),
    [anon_sym_int] = ACTIONS(1962),
    [anon_sym_boolean] = ACTIONS(1962),
    [anon_sym_char] = ACTIONS(1962),
    [anon_sym_short] = ACTIONS(1962),
    [anon_sym_long] = ACTIONS(1962),
    [anon_sym_float] = ACTIONS(1962),
    [anon_sym_double] = ACTIONS(1962),
    [anon_sym_void] = ACTIONS(1962),
    [anon_sym_public] = ACTIONS(1962),
    [anon_sym_protected] = ACTIONS(1962),
    [anon_sym_private] = ACTIONS(1962),
    [anon_sym_static] = ACTIONS(1962),
    [anon_sym_final] = ACTIONS(1962),
    [anon_sym_synchronized] = ACTIONS(1962),
    [anon_sym_TILDE] = ACTIONS(1960),
    [anon_sym_BANG] = ACTIONS(1960),
    [anon_sym_new] = ACTIONS(1962),
  },
  [STATE(850)] = {
    [sym_comment] = STATE(850),
    [sym_groovy_doc] = STATE(850),
    [ts_builtin_sym_end] = ACTIONS(1812),
    [sym_identifier] = ACTIONS(1810),
    [anon_sym_break] = ACTIONS(1810),
    [anon_sym_continue] = ACTIONS(1810),
    [anon_sym_SEMI] = ACTIONS(1812),
    [anon_sym_STAR] = ACTIONS(1810),
    [anon_sym_STAR_COLON] = ACTIONS(1812),
    [anon_sym_DOT] = ACTIONS(2089),
    [anon_sym_import] = ACTIONS(1810),
    [anon_sym_as] = ACTIONS(2091),
    [anon_sym_package] = ACTIONS(1810),
    [anon_sym_AT] = ACTIONS(1810),
    [anon_sym_assert] = ACTIONS(1810),
    [anon_sym_PLUS_PLUS] = ACTIONS(1812),
    [anon_sym_DASH_DASH] = ACTIONS(1812),
    [anon_sym_SLASH] = ACTIONS(1810),
    [anon_sym_PLUS] = ACTIONS(1810),
    [anon_sym_DASH] = ACTIONS(1810),
    [anon_sym_true] = ACTIONS(1810),
    [anon_sym_false] = ACTIONS(1810),
    [anon_sym_ATinterface] = ACTIONS(1812),
    [anon_sym_interface] = ACTIONS(1810),
    [anon_sym_class] = ACTIONS(1810),
    [anon_sym_LBRACE] = ACTIONS(1812),
    [anon_sym_RBRACE] = ACTIONS(1812),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1810),
    [anon_sym_def] = ACTIONS(1810),
    [anon_sym_LPAREN] = ACTIONS(1812),
    [anon_sym_null] = ACTIONS(1810),
    [anon_sym_this] = ACTIONS(1810),
    [anon_sym_do] = ACTIONS(1810),
    [anon_sym_while] = ACTIONS(1810),
    [anon_sym_for] = ACTIONS(1810),
    [aux_sym__type_identifier_token1] = ACTIONS(1810),
    [anon_sym_if] = ACTIONS(1810),
    [anon_sym_else] = ACTIONS(1810),
    [anon_sym_LBRACK] = ACTIONS(1812),
    [aux_sym_number_literal_token1] = ACTIONS(1810),
    [aux_sym_number_literal_token2] = ACTIONS(1812),
    [aux_sym_number_literal_token3] = ACTIONS(1812),
    [aux_sym_number_literal_token4] = ACTIONS(1810),
    [aux_sym_number_literal_token5] = ACTIONS(1812),
    [anon_sym_pipeline] = ACTIONS(1810),
    [anon_sym_return] = ACTIONS(1810),
    [anon_sym_SQUOTE] = ACTIONS(1810),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1812),
    [anon_sym_DQUOTE] = ACTIONS(1810),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1812),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1812),
    [anon_sym_switch] = ACTIONS(1810),
    [anon_sym_try] = ACTIONS(1810),
    [anon_sym_int] = ACTIONS(1810),
    [anon_sym_boolean] = ACTIONS(1810),
    [anon_sym_char] = ACTIONS(1810),
    [anon_sym_short] = ACTIONS(1810),
    [anon_sym_long] = ACTIONS(1810),
    [anon_sym_float] = ACTIONS(1810),
    [anon_sym_double] = ACTIONS(1810),
    [anon_sym_void] = ACTIONS(1810),
    [anon_sym_public] = ACTIONS(1810),
    [anon_sym_protected] = ACTIONS(1810),
    [anon_sym_private] = ACTIONS(1810),
    [anon_sym_static] = ACTIONS(1810),
    [anon_sym_final] = ACTIONS(1810),
    [anon_sym_synchronized] = ACTIONS(1810),
    [anon_sym_TILDE] = ACTIONS(1812),
    [anon_sym_BANG] = ACTIONS(1812),
    [anon_sym_new] = ACTIONS(1810),
  },
  [STATE(851)] = {
    [sym_comment] = STATE(851),
    [sym_groovy_doc] = STATE(851),
    [ts_builtin_sym_end] = ACTIONS(1968),
    [sym_identifier] = ACTIONS(1970),
    [anon_sym_break] = ACTIONS(1970),
    [anon_sym_continue] = ACTIONS(1970),
    [anon_sym_SEMI] = ACTIONS(1968),
    [anon_sym_STAR] = ACTIONS(1970),
    [anon_sym_STAR_COLON] = ACTIONS(1968),
    [anon_sym_import] = ACTIONS(1970),
    [anon_sym_package] = ACTIONS(1970),
    [anon_sym_AT] = ACTIONS(1970),
    [anon_sym_assert] = ACTIONS(1970),
    [anon_sym_PLUS_PLUS] = ACTIONS(1968),
    [anon_sym_DASH_DASH] = ACTIONS(1968),
    [anon_sym_SLASH] = ACTIONS(1970),
    [anon_sym_PLUS] = ACTIONS(1970),
    [anon_sym_DASH] = ACTIONS(1970),
    [anon_sym_true] = ACTIONS(1970),
    [anon_sym_false] = ACTIONS(1970),
    [anon_sym_ATinterface] = ACTIONS(1968),
    [anon_sym_interface] = ACTIONS(1970),
    [anon_sym_class] = ACTIONS(1970),
    [anon_sym_COMMA] = ACTIONS(1968),
    [anon_sym_LBRACE] = ACTIONS(1968),
    [anon_sym_RBRACE] = ACTIONS(1968),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1970),
    [anon_sym_def] = ACTIONS(1970),
    [anon_sym_LPAREN] = ACTIONS(1968),
    [anon_sym_RPAREN] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_this] = ACTIONS(1970),
    [anon_sym_do] = ACTIONS(1970),
    [anon_sym_while] = ACTIONS(1970),
    [anon_sym_for] = ACTIONS(1970),
    [aux_sym__type_identifier_token1] = ACTIONS(1970),
    [anon_sym_if] = ACTIONS(1970),
    [anon_sym_else] = ACTIONS(1970),
    [anon_sym_LBRACK] = ACTIONS(1968),
    [aux_sym_number_literal_token1] = ACTIONS(1970),
    [aux_sym_number_literal_token2] = ACTIONS(1968),
    [aux_sym_number_literal_token3] = ACTIONS(1968),
    [aux_sym_number_literal_token4] = ACTIONS(1970),
    [aux_sym_number_literal_token5] = ACTIONS(1968),
    [anon_sym_pipeline] = ACTIONS(1970),
    [anon_sym_return] = ACTIONS(1970),
    [anon_sym_SQUOTE] = ACTIONS(1970),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1968),
    [anon_sym_DQUOTE] = ACTIONS(1970),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1968),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1968),
    [anon_sym_switch] = ACTIONS(1970),
    [anon_sym_try] = ACTIONS(1970),
    [anon_sym_int] = ACTIONS(1970),
    [anon_sym_boolean] = ACTIONS(1970),
    [anon_sym_char] = ACTIONS(1970),
    [anon_sym_short] = ACTIONS(1970),
    [anon_sym_long] = ACTIONS(1970),
    [anon_sym_float] = ACTIONS(1970),
    [anon_sym_double] = ACTIONS(1970),
    [anon_sym_void] = ACTIONS(1970),
    [anon_sym_public] = ACTIONS(1970),
    [anon_sym_protected] = ACTIONS(1970),
    [anon_sym_private] = ACTIONS(1970),
    [anon_sym_static] = ACTIONS(1970),
    [anon_sym_final] = ACTIONS(1970),
    [anon_sym_synchronized] = ACTIONS(1970),
    [anon_sym_TILDE] = ACTIONS(1968),
    [anon_sym_BANG] = ACTIONS(1968),
    [anon_sym_new] = ACTIONS(1970),
  },
  [STATE(852)] = {
    [sym_comment] = STATE(852),
    [sym_groovy_doc] = STATE(852),
    [aux_sym__juxt_argument_list_repeat1] = STATE(852),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(2093),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(853)] = {
    [sym_comment] = STATE(853),
    [sym_groovy_doc] = STATE(853),
    [ts_builtin_sym_end] = ACTIONS(1972),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_break] = ACTIONS(1974),
    [anon_sym_continue] = ACTIONS(1974),
    [anon_sym_SEMI] = ACTIONS(1972),
    [anon_sym_STAR] = ACTIONS(1974),
    [anon_sym_STAR_COLON] = ACTIONS(1972),
    [anon_sym_import] = ACTIONS(1974),
    [anon_sym_package] = ACTIONS(1974),
    [anon_sym_AT] = ACTIONS(1974),
    [anon_sym_assert] = ACTIONS(1974),
    [anon_sym_PLUS_PLUS] = ACTIONS(1972),
    [anon_sym_DASH_DASH] = ACTIONS(1972),
    [anon_sym_SLASH] = ACTIONS(1974),
    [anon_sym_PLUS] = ACTIONS(1974),
    [anon_sym_DASH] = ACTIONS(1974),
    [anon_sym_true] = ACTIONS(1974),
    [anon_sym_false] = ACTIONS(1974),
    [anon_sym_ATinterface] = ACTIONS(1972),
    [anon_sym_interface] = ACTIONS(1974),
    [anon_sym_class] = ACTIONS(1974),
    [anon_sym_COMMA] = ACTIONS(1972),
    [anon_sym_LBRACE] = ACTIONS(1972),
    [anon_sym_RBRACE] = ACTIONS(1972),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1974),
    [anon_sym_def] = ACTIONS(1974),
    [anon_sym_LPAREN] = ACTIONS(1972),
    [anon_sym_RPAREN] = ACTIONS(1972),
    [anon_sym_null] = ACTIONS(1974),
    [anon_sym_this] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1974),
    [anon_sym_while] = ACTIONS(1974),
    [anon_sym_for] = ACTIONS(1974),
    [aux_sym__type_identifier_token1] = ACTIONS(1974),
    [anon_sym_if] = ACTIONS(1974),
    [anon_sym_else] = ACTIONS(1974),
    [anon_sym_LBRACK] = ACTIONS(1972),
    [aux_sym_number_literal_token1] = ACTIONS(1974),
    [aux_sym_number_literal_token2] = ACTIONS(1972),
    [aux_sym_number_literal_token3] = ACTIONS(1972),
    [aux_sym_number_literal_token4] = ACTIONS(1974),
    [aux_sym_number_literal_token5] = ACTIONS(1972),
    [anon_sym_pipeline] = ACTIONS(1974),
    [anon_sym_return] = ACTIONS(1974),
    [anon_sym_SQUOTE] = ACTIONS(1974),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1972),
    [anon_sym_DQUOTE] = ACTIONS(1974),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1972),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1972),
    [anon_sym_switch] = ACTIONS(1974),
    [anon_sym_try] = ACTIONS(1974),
    [anon_sym_int] = ACTIONS(1974),
    [anon_sym_boolean] = ACTIONS(1974),
    [anon_sym_char] = ACTIONS(1974),
    [anon_sym_short] = ACTIONS(1974),
    [anon_sym_long] = ACTIONS(1974),
    [anon_sym_float] = ACTIONS(1974),
    [anon_sym_double] = ACTIONS(1974),
    [anon_sym_void] = ACTIONS(1974),
    [anon_sym_public] = ACTIONS(1974),
    [anon_sym_protected] = ACTIONS(1974),
    [anon_sym_private] = ACTIONS(1974),
    [anon_sym_static] = ACTIONS(1974),
    [anon_sym_final] = ACTIONS(1974),
    [anon_sym_synchronized] = ACTIONS(1974),
    [anon_sym_TILDE] = ACTIONS(1972),
    [anon_sym_BANG] = ACTIONS(1972),
    [anon_sym_new] = ACTIONS(1974),
  },
  [STATE(854)] = {
    [sym_comment] = STATE(854),
    [sym_groovy_doc] = STATE(854),
    [sym_identifier] = ACTIONS(1958),
    [anon_sym_break] = ACTIONS(1958),
    [anon_sym_continue] = ACTIONS(1958),
    [anon_sym_SEMI] = ACTIONS(1956),
    [anon_sym_STAR] = ACTIONS(1958),
    [anon_sym_STAR_COLON] = ACTIONS(1956),
    [anon_sym_import] = ACTIONS(1958),
    [anon_sym_package] = ACTIONS(1958),
    [anon_sym_AT] = ACTIONS(1958),
    [anon_sym_assert] = ACTIONS(1958),
    [anon_sym_PLUS_PLUS] = ACTIONS(1956),
    [anon_sym_DASH_DASH] = ACTIONS(1956),
    [anon_sym_SLASH] = ACTIONS(1958),
    [anon_sym_PLUS] = ACTIONS(1958),
    [anon_sym_DASH] = ACTIONS(1958),
    [anon_sym_true] = ACTIONS(1958),
    [anon_sym_false] = ACTIONS(1958),
    [anon_sym_ATinterface] = ACTIONS(1956),
    [anon_sym_interface] = ACTIONS(1958),
    [anon_sym_class] = ACTIONS(1958),
    [anon_sym_LBRACE] = ACTIONS(1956),
    [anon_sym_RBRACE] = ACTIONS(1956),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1958),
    [anon_sym_def] = ACTIONS(1958),
    [anon_sym_LPAREN] = ACTIONS(1956),
    [anon_sym_null] = ACTIONS(1958),
    [anon_sym_this] = ACTIONS(1958),
    [anon_sym_do] = ACTIONS(1958),
    [anon_sym_while] = ACTIONS(1958),
    [anon_sym_for] = ACTIONS(1958),
    [aux_sym__type_identifier_token1] = ACTIONS(1958),
    [anon_sym_if] = ACTIONS(1958),
    [anon_sym_else] = ACTIONS(1958),
    [anon_sym_LBRACK] = ACTIONS(1956),
    [aux_sym_number_literal_token1] = ACTIONS(1958),
    [aux_sym_number_literal_token2] = ACTIONS(1956),
    [aux_sym_number_literal_token3] = ACTIONS(1956),
    [aux_sym_number_literal_token4] = ACTIONS(1958),
    [aux_sym_number_literal_token5] = ACTIONS(1956),
    [anon_sym_return] = ACTIONS(1958),
    [anon_sym_SQUOTE] = ACTIONS(1958),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1956),
    [anon_sym_DQUOTE] = ACTIONS(1958),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1956),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1956),
    [anon_sym_switch] = ACTIONS(1958),
    [anon_sym_case] = ACTIONS(1958),
    [anon_sym_default] = ACTIONS(1958),
    [anon_sym_try] = ACTIONS(1958),
    [anon_sym_catch] = ACTIONS(1958),
    [anon_sym_finally] = ACTIONS(1958),
    [anon_sym_int] = ACTIONS(1958),
    [anon_sym_boolean] = ACTIONS(1958),
    [anon_sym_char] = ACTIONS(1958),
    [anon_sym_short] = ACTIONS(1958),
    [anon_sym_long] = ACTIONS(1958),
    [anon_sym_float] = ACTIONS(1958),
    [anon_sym_double] = ACTIONS(1958),
    [anon_sym_void] = ACTIONS(1958),
    [anon_sym_public] = ACTIONS(1958),
    [anon_sym_protected] = ACTIONS(1958),
    [anon_sym_private] = ACTIONS(1958),
    [anon_sym_static] = ACTIONS(1958),
    [anon_sym_final] = ACTIONS(1958),
    [anon_sym_synchronized] = ACTIONS(1958),
    [anon_sym_TILDE] = ACTIONS(1956),
    [anon_sym_BANG] = ACTIONS(1956),
    [anon_sym_new] = ACTIONS(1958),
  },
  [STATE(855)] = {
    [sym_comment] = STATE(855),
    [sym_groovy_doc] = STATE(855),
    [ts_builtin_sym_end] = ACTIONS(1989),
    [sym_identifier] = ACTIONS(1991),
    [anon_sym_break] = ACTIONS(1991),
    [anon_sym_continue] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_STAR] = ACTIONS(1991),
    [anon_sym_STAR_COLON] = ACTIONS(1989),
    [anon_sym_import] = ACTIONS(1991),
    [anon_sym_package] = ACTIONS(1991),
    [anon_sym_AT] = ACTIONS(1991),
    [anon_sym_assert] = ACTIONS(1991),
    [anon_sym_PLUS_PLUS] = ACTIONS(1989),
    [anon_sym_DASH_DASH] = ACTIONS(1989),
    [anon_sym_SLASH] = ACTIONS(1991),
    [anon_sym_PLUS] = ACTIONS(1991),
    [anon_sym_DASH] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_ATinterface] = ACTIONS(1989),
    [anon_sym_interface] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_COMMA] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1991),
    [anon_sym_def] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_RPAREN] = ACTIONS(1989),
    [anon_sym_null] = ACTIONS(1991),
    [anon_sym_this] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_while] = ACTIONS(1991),
    [anon_sym_for] = ACTIONS(1991),
    [aux_sym__type_identifier_token1] = ACTIONS(1991),
    [anon_sym_if] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1989),
    [aux_sym_number_literal_token1] = ACTIONS(1991),
    [aux_sym_number_literal_token2] = ACTIONS(1989),
    [aux_sym_number_literal_token3] = ACTIONS(1989),
    [aux_sym_number_literal_token4] = ACTIONS(1991),
    [aux_sym_number_literal_token5] = ACTIONS(1989),
    [anon_sym_pipeline] = ACTIONS(1991),
    [anon_sym_return] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1991),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1991),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1989),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1989),
    [anon_sym_switch] = ACTIONS(1991),
    [anon_sym_try] = ACTIONS(1991),
    [anon_sym_int] = ACTIONS(1991),
    [anon_sym_boolean] = ACTIONS(1991),
    [anon_sym_char] = ACTIONS(1991),
    [anon_sym_short] = ACTIONS(1991),
    [anon_sym_long] = ACTIONS(1991),
    [anon_sym_float] = ACTIONS(1991),
    [anon_sym_double] = ACTIONS(1991),
    [anon_sym_void] = ACTIONS(1991),
    [anon_sym_public] = ACTIONS(1991),
    [anon_sym_protected] = ACTIONS(1991),
    [anon_sym_private] = ACTIONS(1991),
    [anon_sym_static] = ACTIONS(1991),
    [anon_sym_final] = ACTIONS(1991),
    [anon_sym_synchronized] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1989),
    [anon_sym_BANG] = ACTIONS(1989),
    [anon_sym_new] = ACTIONS(1991),
  },
  [STATE(856)] = {
    [sym_comment] = STATE(856),
    [sym_groovy_doc] = STATE(856),
    [sym_identifier] = ACTIONS(2027),
    [anon_sym_break] = ACTIONS(2027),
    [anon_sym_continue] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_STAR] = ACTIONS(2027),
    [anon_sym_STAR_COLON] = ACTIONS(2025),
    [anon_sym_import] = ACTIONS(2027),
    [anon_sym_package] = ACTIONS(2027),
    [anon_sym_AT] = ACTIONS(2027),
    [anon_sym_assert] = ACTIONS(2027),
    [anon_sym_PLUS_PLUS] = ACTIONS(2025),
    [anon_sym_DASH_DASH] = ACTIONS(2025),
    [anon_sym_SLASH] = ACTIONS(2027),
    [anon_sym_PLUS] = ACTIONS(2027),
    [anon_sym_DASH] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_ATinterface] = ACTIONS(2025),
    [anon_sym_interface] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2027),
    [anon_sym_def] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_null] = ACTIONS(2027),
    [anon_sym_this] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_while] = ACTIONS(2027),
    [anon_sym_for] = ACTIONS(2027),
    [aux_sym__type_identifier_token1] = ACTIONS(2027),
    [anon_sym_if] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [aux_sym_number_literal_token1] = ACTIONS(2027),
    [aux_sym_number_literal_token2] = ACTIONS(2025),
    [aux_sym_number_literal_token3] = ACTIONS(2025),
    [aux_sym_number_literal_token4] = ACTIONS(2027),
    [aux_sym_number_literal_token5] = ACTIONS(2025),
    [anon_sym_return] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2025),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2025),
    [anon_sym_switch] = ACTIONS(2027),
    [anon_sym_case] = ACTIONS(2027),
    [anon_sym_default] = ACTIONS(2027),
    [anon_sym_try] = ACTIONS(2027),
    [anon_sym_catch] = ACTIONS(2027),
    [anon_sym_finally] = ACTIONS(2027),
    [anon_sym_int] = ACTIONS(2027),
    [anon_sym_boolean] = ACTIONS(2027),
    [anon_sym_char] = ACTIONS(2027),
    [anon_sym_short] = ACTIONS(2027),
    [anon_sym_long] = ACTIONS(2027),
    [anon_sym_float] = ACTIONS(2027),
    [anon_sym_double] = ACTIONS(2027),
    [anon_sym_void] = ACTIONS(2027),
    [anon_sym_public] = ACTIONS(2027),
    [anon_sym_protected] = ACTIONS(2027),
    [anon_sym_private] = ACTIONS(2027),
    [anon_sym_static] = ACTIONS(2027),
    [anon_sym_final] = ACTIONS(2027),
    [anon_sym_synchronized] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_BANG] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2027),
  },
  [STATE(857)] = {
    [sym_comment] = STATE(857),
    [sym_groovy_doc] = STATE(857),
    [sym_identifier] = ACTIONS(1820),
    [anon_sym_break] = ACTIONS(1820),
    [anon_sym_continue] = ACTIONS(1820),
    [anon_sym_SEMI] = ACTIONS(1818),
    [anon_sym_STAR] = ACTIONS(1820),
    [anon_sym_STAR_COLON] = ACTIONS(1818),
    [anon_sym_import] = ACTIONS(1820),
    [anon_sym_package] = ACTIONS(1820),
    [anon_sym_AT] = ACTIONS(1820),
    [anon_sym_assert] = ACTIONS(1820),
    [anon_sym_PLUS_PLUS] = ACTIONS(1818),
    [anon_sym_DASH_DASH] = ACTIONS(1818),
    [anon_sym_SLASH] = ACTIONS(1820),
    [anon_sym_PLUS] = ACTIONS(1820),
    [anon_sym_DASH] = ACTIONS(1820),
    [anon_sym_true] = ACTIONS(1820),
    [anon_sym_false] = ACTIONS(1820),
    [anon_sym_ATinterface] = ACTIONS(1818),
    [anon_sym_interface] = ACTIONS(1820),
    [anon_sym_class] = ACTIONS(1820),
    [anon_sym_LBRACE] = ACTIONS(1818),
    [anon_sym_RBRACE] = ACTIONS(1818),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1820),
    [anon_sym_def] = ACTIONS(1820),
    [anon_sym_LPAREN] = ACTIONS(1818),
    [anon_sym_null] = ACTIONS(1820),
    [anon_sym_this] = ACTIONS(1820),
    [anon_sym_do] = ACTIONS(1820),
    [anon_sym_while] = ACTIONS(1820),
    [anon_sym_for] = ACTIONS(1820),
    [aux_sym__type_identifier_token1] = ACTIONS(1820),
    [anon_sym_if] = ACTIONS(1820),
    [anon_sym_else] = ACTIONS(1820),
    [anon_sym_LBRACK] = ACTIONS(1818),
    [aux_sym_number_literal_token1] = ACTIONS(1820),
    [aux_sym_number_literal_token2] = ACTIONS(1818),
    [aux_sym_number_literal_token3] = ACTIONS(1818),
    [aux_sym_number_literal_token4] = ACTIONS(1820),
    [aux_sym_number_literal_token5] = ACTIONS(1818),
    [anon_sym_return] = ACTIONS(1820),
    [anon_sym_SQUOTE] = ACTIONS(1820),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1818),
    [anon_sym_DQUOTE] = ACTIONS(1820),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1818),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1818),
    [anon_sym_switch] = ACTIONS(1820),
    [anon_sym_case] = ACTIONS(1820),
    [anon_sym_default] = ACTIONS(1820),
    [anon_sym_try] = ACTIONS(1820),
    [anon_sym_catch] = ACTIONS(1820),
    [anon_sym_finally] = ACTIONS(1820),
    [anon_sym_int] = ACTIONS(1820),
    [anon_sym_boolean] = ACTIONS(1820),
    [anon_sym_char] = ACTIONS(1820),
    [anon_sym_short] = ACTIONS(1820),
    [anon_sym_long] = ACTIONS(1820),
    [anon_sym_float] = ACTIONS(1820),
    [anon_sym_double] = ACTIONS(1820),
    [anon_sym_void] = ACTIONS(1820),
    [anon_sym_public] = ACTIONS(1820),
    [anon_sym_protected] = ACTIONS(1820),
    [anon_sym_private] = ACTIONS(1820),
    [anon_sym_static] = ACTIONS(1820),
    [anon_sym_final] = ACTIONS(1820),
    [anon_sym_synchronized] = ACTIONS(1820),
    [anon_sym_TILDE] = ACTIONS(1818),
    [anon_sym_BANG] = ACTIONS(1818),
    [anon_sym_new] = ACTIONS(1820),
  },
  [STATE(858)] = {
    [sym_comment] = STATE(858),
    [sym_groovy_doc] = STATE(858),
    [sym_identifier] = ACTIONS(1850),
    [anon_sym_break] = ACTIONS(1850),
    [anon_sym_continue] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_STAR] = ACTIONS(1850),
    [anon_sym_STAR_COLON] = ACTIONS(1848),
    [anon_sym_import] = ACTIONS(1850),
    [anon_sym_package] = ACTIONS(1850),
    [anon_sym_AT] = ACTIONS(1850),
    [anon_sym_assert] = ACTIONS(1850),
    [anon_sym_PLUS_PLUS] = ACTIONS(1848),
    [anon_sym_DASH_DASH] = ACTIONS(1848),
    [anon_sym_SLASH] = ACTIONS(1850),
    [anon_sym_PLUS] = ACTIONS(1850),
    [anon_sym_DASH] = ACTIONS(1850),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_ATinterface] = ACTIONS(1848),
    [anon_sym_interface] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1850),
    [anon_sym_def] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_null] = ACTIONS(1850),
    [anon_sym_this] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_while] = ACTIONS(1850),
    [anon_sym_for] = ACTIONS(1850),
    [aux_sym__type_identifier_token1] = ACTIONS(1850),
    [anon_sym_if] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1848),
    [aux_sym_number_literal_token1] = ACTIONS(1850),
    [aux_sym_number_literal_token2] = ACTIONS(1848),
    [aux_sym_number_literal_token3] = ACTIONS(1848),
    [aux_sym_number_literal_token4] = ACTIONS(1850),
    [aux_sym_number_literal_token5] = ACTIONS(1848),
    [anon_sym_return] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1850),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1850),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1848),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1848),
    [anon_sym_switch] = ACTIONS(1850),
    [anon_sym_case] = ACTIONS(1850),
    [anon_sym_default] = ACTIONS(1850),
    [anon_sym_try] = ACTIONS(1850),
    [anon_sym_catch] = ACTIONS(1850),
    [anon_sym_finally] = ACTIONS(1850),
    [anon_sym_int] = ACTIONS(1850),
    [anon_sym_boolean] = ACTIONS(1850),
    [anon_sym_char] = ACTIONS(1850),
    [anon_sym_short] = ACTIONS(1850),
    [anon_sym_long] = ACTIONS(1850),
    [anon_sym_float] = ACTIONS(1850),
    [anon_sym_double] = ACTIONS(1850),
    [anon_sym_void] = ACTIONS(1850),
    [anon_sym_public] = ACTIONS(1850),
    [anon_sym_protected] = ACTIONS(1850),
    [anon_sym_private] = ACTIONS(1850),
    [anon_sym_static] = ACTIONS(1850),
    [anon_sym_final] = ACTIONS(1850),
    [anon_sym_synchronized] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1848),
    [anon_sym_BANG] = ACTIONS(1848),
    [anon_sym_new] = ACTIONS(1850),
  },
  [STATE(859)] = {
    [sym_comment] = STATE(859),
    [sym_groovy_doc] = STATE(859),
    [sym_identifier] = ACTIONS(1862),
    [anon_sym_break] = ACTIONS(1862),
    [anon_sym_continue] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_STAR] = ACTIONS(1862),
    [anon_sym_STAR_COLON] = ACTIONS(1860),
    [anon_sym_import] = ACTIONS(1862),
    [anon_sym_package] = ACTIONS(1862),
    [anon_sym_AT] = ACTIONS(1862),
    [anon_sym_assert] = ACTIONS(1862),
    [anon_sym_PLUS_PLUS] = ACTIONS(1860),
    [anon_sym_DASH_DASH] = ACTIONS(1860),
    [anon_sym_SLASH] = ACTIONS(1862),
    [anon_sym_PLUS] = ACTIONS(1862),
    [anon_sym_DASH] = ACTIONS(1862),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_ATinterface] = ACTIONS(1860),
    [anon_sym_interface] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1862),
    [anon_sym_def] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_null] = ACTIONS(1862),
    [anon_sym_this] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_while] = ACTIONS(1862),
    [anon_sym_for] = ACTIONS(1862),
    [aux_sym__type_identifier_token1] = ACTIONS(1862),
    [anon_sym_if] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1860),
    [aux_sym_number_literal_token1] = ACTIONS(1862),
    [aux_sym_number_literal_token2] = ACTIONS(1860),
    [aux_sym_number_literal_token3] = ACTIONS(1860),
    [aux_sym_number_literal_token4] = ACTIONS(1862),
    [aux_sym_number_literal_token5] = ACTIONS(1860),
    [anon_sym_return] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1862),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1862),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1860),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1860),
    [anon_sym_switch] = ACTIONS(1862),
    [anon_sym_case] = ACTIONS(1862),
    [anon_sym_default] = ACTIONS(1862),
    [anon_sym_try] = ACTIONS(1862),
    [anon_sym_catch] = ACTIONS(1862),
    [anon_sym_finally] = ACTIONS(1862),
    [anon_sym_int] = ACTIONS(1862),
    [anon_sym_boolean] = ACTIONS(1862),
    [anon_sym_char] = ACTIONS(1862),
    [anon_sym_short] = ACTIONS(1862),
    [anon_sym_long] = ACTIONS(1862),
    [anon_sym_float] = ACTIONS(1862),
    [anon_sym_double] = ACTIONS(1862),
    [anon_sym_void] = ACTIONS(1862),
    [anon_sym_public] = ACTIONS(1862),
    [anon_sym_protected] = ACTIONS(1862),
    [anon_sym_private] = ACTIONS(1862),
    [anon_sym_static] = ACTIONS(1862),
    [anon_sym_final] = ACTIONS(1862),
    [anon_sym_synchronized] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1860),
    [anon_sym_BANG] = ACTIONS(1860),
    [anon_sym_new] = ACTIONS(1862),
  },
  [STATE(860)] = {
    [sym_comment] = STATE(860),
    [sym_groovy_doc] = STATE(860),
    [sym_identifier] = ACTIONS(1946),
    [anon_sym_break] = ACTIONS(1946),
    [anon_sym_continue] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_STAR] = ACTIONS(1946),
    [anon_sym_STAR_COLON] = ACTIONS(1944),
    [anon_sym_import] = ACTIONS(1946),
    [anon_sym_package] = ACTIONS(1946),
    [anon_sym_AT] = ACTIONS(1946),
    [anon_sym_assert] = ACTIONS(1946),
    [anon_sym_PLUS_PLUS] = ACTIONS(1944),
    [anon_sym_DASH_DASH] = ACTIONS(1944),
    [anon_sym_SLASH] = ACTIONS(1946),
    [anon_sym_PLUS] = ACTIONS(1946),
    [anon_sym_DASH] = ACTIONS(1946),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_ATinterface] = ACTIONS(1944),
    [anon_sym_interface] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1946),
    [anon_sym_def] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_null] = ACTIONS(1946),
    [anon_sym_this] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_while] = ACTIONS(1946),
    [anon_sym_for] = ACTIONS(1946),
    [aux_sym__type_identifier_token1] = ACTIONS(1946),
    [anon_sym_if] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1944),
    [aux_sym_number_literal_token1] = ACTIONS(1946),
    [aux_sym_number_literal_token2] = ACTIONS(1944),
    [aux_sym_number_literal_token3] = ACTIONS(1944),
    [aux_sym_number_literal_token4] = ACTIONS(1946),
    [aux_sym_number_literal_token5] = ACTIONS(1944),
    [anon_sym_return] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1946),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1944),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1944),
    [anon_sym_switch] = ACTIONS(1946),
    [anon_sym_case] = ACTIONS(1946),
    [anon_sym_default] = ACTIONS(1946),
    [anon_sym_try] = ACTIONS(1946),
    [anon_sym_catch] = ACTIONS(1946),
    [anon_sym_finally] = ACTIONS(1946),
    [anon_sym_int] = ACTIONS(1946),
    [anon_sym_boolean] = ACTIONS(1946),
    [anon_sym_char] = ACTIONS(1946),
    [anon_sym_short] = ACTIONS(1946),
    [anon_sym_long] = ACTIONS(1946),
    [anon_sym_float] = ACTIONS(1946),
    [anon_sym_double] = ACTIONS(1946),
    [anon_sym_void] = ACTIONS(1946),
    [anon_sym_public] = ACTIONS(1946),
    [anon_sym_protected] = ACTIONS(1946),
    [anon_sym_private] = ACTIONS(1946),
    [anon_sym_static] = ACTIONS(1946),
    [anon_sym_final] = ACTIONS(1946),
    [anon_sym_synchronized] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1944),
    [anon_sym_BANG] = ACTIONS(1944),
    [anon_sym_new] = ACTIONS(1946),
  },
  [STATE(861)] = {
    [sym_comment] = STATE(861),
    [sym_groovy_doc] = STATE(861),
    [sym_identifier] = ACTIONS(1954),
    [anon_sym_break] = ACTIONS(1954),
    [anon_sym_continue] = ACTIONS(1954),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_STAR] = ACTIONS(1954),
    [anon_sym_STAR_COLON] = ACTIONS(1952),
    [anon_sym_import] = ACTIONS(1954),
    [anon_sym_package] = ACTIONS(1954),
    [anon_sym_AT] = ACTIONS(1954),
    [anon_sym_assert] = ACTIONS(1954),
    [anon_sym_PLUS_PLUS] = ACTIONS(1952),
    [anon_sym_DASH_DASH] = ACTIONS(1952),
    [anon_sym_SLASH] = ACTIONS(1954),
    [anon_sym_PLUS] = ACTIONS(1954),
    [anon_sym_DASH] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(1954),
    [anon_sym_false] = ACTIONS(1954),
    [anon_sym_ATinterface] = ACTIONS(1952),
    [anon_sym_interface] = ACTIONS(1954),
    [anon_sym_class] = ACTIONS(1954),
    [anon_sym_LBRACE] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1954),
    [anon_sym_def] = ACTIONS(1954),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_null] = ACTIONS(1954),
    [anon_sym_this] = ACTIONS(1954),
    [anon_sym_do] = ACTIONS(1954),
    [anon_sym_while] = ACTIONS(1954),
    [anon_sym_for] = ACTIONS(1954),
    [aux_sym__type_identifier_token1] = ACTIONS(1954),
    [anon_sym_if] = ACTIONS(1954),
    [anon_sym_else] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(1952),
    [aux_sym_number_literal_token1] = ACTIONS(1954),
    [aux_sym_number_literal_token2] = ACTIONS(1952),
    [aux_sym_number_literal_token3] = ACTIONS(1952),
    [aux_sym_number_literal_token4] = ACTIONS(1954),
    [aux_sym_number_literal_token5] = ACTIONS(1952),
    [anon_sym_return] = ACTIONS(1954),
    [anon_sym_SQUOTE] = ACTIONS(1954),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1954),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1952),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1952),
    [anon_sym_switch] = ACTIONS(1954),
    [anon_sym_case] = ACTIONS(1954),
    [anon_sym_default] = ACTIONS(1954),
    [anon_sym_try] = ACTIONS(1954),
    [anon_sym_catch] = ACTIONS(1954),
    [anon_sym_finally] = ACTIONS(1954),
    [anon_sym_int] = ACTIONS(1954),
    [anon_sym_boolean] = ACTIONS(1954),
    [anon_sym_char] = ACTIONS(1954),
    [anon_sym_short] = ACTIONS(1954),
    [anon_sym_long] = ACTIONS(1954),
    [anon_sym_float] = ACTIONS(1954),
    [anon_sym_double] = ACTIONS(1954),
    [anon_sym_void] = ACTIONS(1954),
    [anon_sym_public] = ACTIONS(1954),
    [anon_sym_protected] = ACTIONS(1954),
    [anon_sym_private] = ACTIONS(1954),
    [anon_sym_static] = ACTIONS(1954),
    [anon_sym_final] = ACTIONS(1954),
    [anon_sym_synchronized] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(1952),
    [anon_sym_BANG] = ACTIONS(1952),
    [anon_sym_new] = ACTIONS(1954),
  },
  [STATE(862)] = {
    [sym_comment] = STATE(862),
    [sym_groovy_doc] = STATE(862),
    [sym_identifier] = ACTIONS(2015),
    [anon_sym_break] = ACTIONS(2015),
    [anon_sym_continue] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_STAR] = ACTIONS(2015),
    [anon_sym_STAR_COLON] = ACTIONS(2013),
    [anon_sym_import] = ACTIONS(2015),
    [anon_sym_package] = ACTIONS(2015),
    [anon_sym_AT] = ACTIONS(2015),
    [anon_sym_assert] = ACTIONS(2015),
    [anon_sym_PLUS_PLUS] = ACTIONS(2013),
    [anon_sym_DASH_DASH] = ACTIONS(2013),
    [anon_sym_SLASH] = ACTIONS(2015),
    [anon_sym_PLUS] = ACTIONS(2015),
    [anon_sym_DASH] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_ATinterface] = ACTIONS(2013),
    [anon_sym_interface] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2015),
    [anon_sym_def] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_null] = ACTIONS(2015),
    [anon_sym_this] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_while] = ACTIONS(2015),
    [anon_sym_for] = ACTIONS(2015),
    [aux_sym__type_identifier_token1] = ACTIONS(2015),
    [anon_sym_if] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2013),
    [aux_sym_number_literal_token1] = ACTIONS(2015),
    [aux_sym_number_literal_token2] = ACTIONS(2013),
    [aux_sym_number_literal_token3] = ACTIONS(2013),
    [aux_sym_number_literal_token4] = ACTIONS(2015),
    [aux_sym_number_literal_token5] = ACTIONS(2013),
    [anon_sym_return] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2015),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2015),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2013),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2013),
    [anon_sym_switch] = ACTIONS(2015),
    [anon_sym_case] = ACTIONS(2015),
    [anon_sym_default] = ACTIONS(2015),
    [anon_sym_try] = ACTIONS(2015),
    [anon_sym_catch] = ACTIONS(2015),
    [anon_sym_finally] = ACTIONS(2015),
    [anon_sym_int] = ACTIONS(2015),
    [anon_sym_boolean] = ACTIONS(2015),
    [anon_sym_char] = ACTIONS(2015),
    [anon_sym_short] = ACTIONS(2015),
    [anon_sym_long] = ACTIONS(2015),
    [anon_sym_float] = ACTIONS(2015),
    [anon_sym_double] = ACTIONS(2015),
    [anon_sym_void] = ACTIONS(2015),
    [anon_sym_public] = ACTIONS(2015),
    [anon_sym_protected] = ACTIONS(2015),
    [anon_sym_private] = ACTIONS(2015),
    [anon_sym_static] = ACTIONS(2015),
    [anon_sym_final] = ACTIONS(2015),
    [anon_sym_synchronized] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2013),
    [anon_sym_BANG] = ACTIONS(2013),
    [anon_sym_new] = ACTIONS(2015),
  },
  [STATE(863)] = {
    [sym_comment] = STATE(863),
    [sym_groovy_doc] = STATE(863),
    [ts_builtin_sym_end] = ACTIONS(1993),
    [sym_identifier] = ACTIONS(1995),
    [anon_sym_break] = ACTIONS(1995),
    [anon_sym_continue] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_STAR] = ACTIONS(1995),
    [anon_sym_STAR_COLON] = ACTIONS(1993),
    [anon_sym_import] = ACTIONS(1995),
    [anon_sym_package] = ACTIONS(1995),
    [anon_sym_AT] = ACTIONS(1995),
    [anon_sym_assert] = ACTIONS(1995),
    [anon_sym_PLUS_PLUS] = ACTIONS(1993),
    [anon_sym_DASH_DASH] = ACTIONS(1993),
    [anon_sym_SLASH] = ACTIONS(1995),
    [anon_sym_PLUS] = ACTIONS(1995),
    [anon_sym_DASH] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_ATinterface] = ACTIONS(1993),
    [anon_sym_interface] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_COMMA] = ACTIONS(1993),
    [anon_sym_LBRACE] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1995),
    [anon_sym_def] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_RPAREN] = ACTIONS(1993),
    [anon_sym_null] = ACTIONS(1995),
    [anon_sym_this] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_while] = ACTIONS(1995),
    [anon_sym_for] = ACTIONS(1995),
    [aux_sym__type_identifier_token1] = ACTIONS(1995),
    [anon_sym_if] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1993),
    [aux_sym_number_literal_token1] = ACTIONS(1995),
    [aux_sym_number_literal_token2] = ACTIONS(1993),
    [aux_sym_number_literal_token3] = ACTIONS(1993),
    [aux_sym_number_literal_token4] = ACTIONS(1995),
    [aux_sym_number_literal_token5] = ACTIONS(1993),
    [anon_sym_pipeline] = ACTIONS(1995),
    [anon_sym_return] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1995),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1995),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1993),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1993),
    [anon_sym_switch] = ACTIONS(1995),
    [anon_sym_try] = ACTIONS(1995),
    [anon_sym_int] = ACTIONS(1995),
    [anon_sym_boolean] = ACTIONS(1995),
    [anon_sym_char] = ACTIONS(1995),
    [anon_sym_short] = ACTIONS(1995),
    [anon_sym_long] = ACTIONS(1995),
    [anon_sym_float] = ACTIONS(1995),
    [anon_sym_double] = ACTIONS(1995),
    [anon_sym_void] = ACTIONS(1995),
    [anon_sym_public] = ACTIONS(1995),
    [anon_sym_protected] = ACTIONS(1995),
    [anon_sym_private] = ACTIONS(1995),
    [anon_sym_static] = ACTIONS(1995),
    [anon_sym_final] = ACTIONS(1995),
    [anon_sym_synchronized] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1993),
    [anon_sym_BANG] = ACTIONS(1993),
    [anon_sym_new] = ACTIONS(1995),
  },
  [STATE(864)] = {
    [sym_comment] = STATE(864),
    [sym_groovy_doc] = STATE(864),
    [ts_builtin_sym_end] = ACTIONS(2003),
    [sym_identifier] = ACTIONS(2005),
    [anon_sym_break] = ACTIONS(2005),
    [anon_sym_continue] = ACTIONS(2005),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_STAR] = ACTIONS(2005),
    [anon_sym_STAR_COLON] = ACTIONS(2003),
    [anon_sym_import] = ACTIONS(2005),
    [anon_sym_package] = ACTIONS(2005),
    [anon_sym_AT] = ACTIONS(2005),
    [anon_sym_assert] = ACTIONS(2005),
    [anon_sym_PLUS_PLUS] = ACTIONS(2003),
    [anon_sym_DASH_DASH] = ACTIONS(2003),
    [anon_sym_SLASH] = ACTIONS(2005),
    [anon_sym_PLUS] = ACTIONS(2005),
    [anon_sym_DASH] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2005),
    [anon_sym_false] = ACTIONS(2005),
    [anon_sym_ATinterface] = ACTIONS(2003),
    [anon_sym_interface] = ACTIONS(2005),
    [anon_sym_class] = ACTIONS(2005),
    [anon_sym_COMMA] = ACTIONS(2003),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_RBRACE] = ACTIONS(2003),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2005),
    [anon_sym_def] = ACTIONS(2005),
    [anon_sym_LPAREN] = ACTIONS(2003),
    [anon_sym_RPAREN] = ACTIONS(2003),
    [anon_sym_null] = ACTIONS(2005),
    [anon_sym_this] = ACTIONS(2005),
    [anon_sym_do] = ACTIONS(2005),
    [anon_sym_while] = ACTIONS(2005),
    [anon_sym_for] = ACTIONS(2005),
    [aux_sym__type_identifier_token1] = ACTIONS(2005),
    [anon_sym_if] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [aux_sym_number_literal_token1] = ACTIONS(2005),
    [aux_sym_number_literal_token2] = ACTIONS(2003),
    [aux_sym_number_literal_token3] = ACTIONS(2003),
    [aux_sym_number_literal_token4] = ACTIONS(2005),
    [aux_sym_number_literal_token5] = ACTIONS(2003),
    [anon_sym_pipeline] = ACTIONS(2005),
    [anon_sym_return] = ACTIONS(2005),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2003),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2003),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2003),
    [anon_sym_switch] = ACTIONS(2005),
    [anon_sym_try] = ACTIONS(2005),
    [anon_sym_int] = ACTIONS(2005),
    [anon_sym_boolean] = ACTIONS(2005),
    [anon_sym_char] = ACTIONS(2005),
    [anon_sym_short] = ACTIONS(2005),
    [anon_sym_long] = ACTIONS(2005),
    [anon_sym_float] = ACTIONS(2005),
    [anon_sym_double] = ACTIONS(2005),
    [anon_sym_void] = ACTIONS(2005),
    [anon_sym_public] = ACTIONS(2005),
    [anon_sym_protected] = ACTIONS(2005),
    [anon_sym_private] = ACTIONS(2005),
    [anon_sym_static] = ACTIONS(2005),
    [anon_sym_final] = ACTIONS(2005),
    [anon_sym_synchronized] = ACTIONS(2005),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_BANG] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2005),
  },
  [STATE(865)] = {
    [sym_comment] = STATE(865),
    [sym_groovy_doc] = STATE(865),
    [sym_identifier] = ACTIONS(1798),
    [anon_sym_break] = ACTIONS(1798),
    [anon_sym_continue] = ACTIONS(1798),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym_STAR] = ACTIONS(1798),
    [anon_sym_STAR_COLON] = ACTIONS(1796),
    [anon_sym_import] = ACTIONS(1798),
    [anon_sym_package] = ACTIONS(1798),
    [anon_sym_AT] = ACTIONS(1798),
    [anon_sym_assert] = ACTIONS(1798),
    [anon_sym_PLUS_PLUS] = ACTIONS(1796),
    [anon_sym_DASH_DASH] = ACTIONS(1796),
    [anon_sym_SLASH] = ACTIONS(1798),
    [anon_sym_PLUS] = ACTIONS(1798),
    [anon_sym_DASH] = ACTIONS(1798),
    [anon_sym_true] = ACTIONS(1798),
    [anon_sym_false] = ACTIONS(1798),
    [anon_sym_ATinterface] = ACTIONS(1796),
    [anon_sym_interface] = ACTIONS(1798),
    [anon_sym_class] = ACTIONS(1798),
    [anon_sym_LBRACE] = ACTIONS(1796),
    [anon_sym_RBRACE] = ACTIONS(1796),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1798),
    [anon_sym_def] = ACTIONS(1798),
    [anon_sym_LPAREN] = ACTIONS(1796),
    [anon_sym_null] = ACTIONS(1798),
    [anon_sym_this] = ACTIONS(1798),
    [anon_sym_do] = ACTIONS(1798),
    [anon_sym_while] = ACTIONS(1798),
    [anon_sym_for] = ACTIONS(1798),
    [aux_sym__type_identifier_token1] = ACTIONS(1798),
    [anon_sym_if] = ACTIONS(1798),
    [anon_sym_else] = ACTIONS(1798),
    [anon_sym_LBRACK] = ACTIONS(1796),
    [aux_sym_number_literal_token1] = ACTIONS(1798),
    [aux_sym_number_literal_token2] = ACTIONS(1796),
    [aux_sym_number_literal_token3] = ACTIONS(1796),
    [aux_sym_number_literal_token4] = ACTIONS(1798),
    [aux_sym_number_literal_token5] = ACTIONS(1796),
    [anon_sym_return] = ACTIONS(1798),
    [anon_sym_SQUOTE] = ACTIONS(1798),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1796),
    [anon_sym_DQUOTE] = ACTIONS(1798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1796),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1796),
    [anon_sym_switch] = ACTIONS(1798),
    [anon_sym_case] = ACTIONS(1798),
    [anon_sym_default] = ACTIONS(1798),
    [anon_sym_try] = ACTIONS(1798),
    [anon_sym_catch] = ACTIONS(1798),
    [anon_sym_finally] = ACTIONS(1798),
    [anon_sym_int] = ACTIONS(1798),
    [anon_sym_boolean] = ACTIONS(1798),
    [anon_sym_char] = ACTIONS(1798),
    [anon_sym_short] = ACTIONS(1798),
    [anon_sym_long] = ACTIONS(1798),
    [anon_sym_float] = ACTIONS(1798),
    [anon_sym_double] = ACTIONS(1798),
    [anon_sym_void] = ACTIONS(1798),
    [anon_sym_public] = ACTIONS(1798),
    [anon_sym_protected] = ACTIONS(1798),
    [anon_sym_private] = ACTIONS(1798),
    [anon_sym_static] = ACTIONS(1798),
    [anon_sym_final] = ACTIONS(1798),
    [anon_sym_synchronized] = ACTIONS(1798),
    [anon_sym_TILDE] = ACTIONS(1796),
    [anon_sym_BANG] = ACTIONS(1796),
    [anon_sym_new] = ACTIONS(1798),
  },
  [STATE(866)] = {
    [sym_comment] = STATE(866),
    [sym_groovy_doc] = STATE(866),
    [sym_identifier] = ACTIONS(1802),
    [anon_sym_break] = ACTIONS(1802),
    [anon_sym_continue] = ACTIONS(1802),
    [anon_sym_SEMI] = ACTIONS(1800),
    [anon_sym_STAR] = ACTIONS(1802),
    [anon_sym_STAR_COLON] = ACTIONS(1800),
    [anon_sym_import] = ACTIONS(1802),
    [anon_sym_package] = ACTIONS(1802),
    [anon_sym_AT] = ACTIONS(1802),
    [anon_sym_assert] = ACTIONS(1802),
    [anon_sym_PLUS_PLUS] = ACTIONS(1800),
    [anon_sym_DASH_DASH] = ACTIONS(1800),
    [anon_sym_SLASH] = ACTIONS(1802),
    [anon_sym_PLUS] = ACTIONS(1802),
    [anon_sym_DASH] = ACTIONS(1802),
    [anon_sym_true] = ACTIONS(1802),
    [anon_sym_false] = ACTIONS(1802),
    [anon_sym_ATinterface] = ACTIONS(1800),
    [anon_sym_interface] = ACTIONS(1802),
    [anon_sym_class] = ACTIONS(1802),
    [anon_sym_LBRACE] = ACTIONS(1800),
    [anon_sym_RBRACE] = ACTIONS(1800),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1802),
    [anon_sym_def] = ACTIONS(1802),
    [anon_sym_LPAREN] = ACTIONS(1800),
    [anon_sym_null] = ACTIONS(1802),
    [anon_sym_this] = ACTIONS(1802),
    [anon_sym_do] = ACTIONS(1802),
    [anon_sym_while] = ACTIONS(1802),
    [anon_sym_for] = ACTIONS(1802),
    [aux_sym__type_identifier_token1] = ACTIONS(1802),
    [anon_sym_if] = ACTIONS(1802),
    [anon_sym_else] = ACTIONS(1802),
    [anon_sym_LBRACK] = ACTIONS(1800),
    [aux_sym_number_literal_token1] = ACTIONS(1802),
    [aux_sym_number_literal_token2] = ACTIONS(1800),
    [aux_sym_number_literal_token3] = ACTIONS(1800),
    [aux_sym_number_literal_token4] = ACTIONS(1802),
    [aux_sym_number_literal_token5] = ACTIONS(1800),
    [anon_sym_return] = ACTIONS(1802),
    [anon_sym_SQUOTE] = ACTIONS(1802),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1800),
    [anon_sym_DQUOTE] = ACTIONS(1802),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1800),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1800),
    [anon_sym_switch] = ACTIONS(1802),
    [anon_sym_case] = ACTIONS(1802),
    [anon_sym_default] = ACTIONS(1802),
    [anon_sym_try] = ACTIONS(1802),
    [anon_sym_catch] = ACTIONS(1802),
    [anon_sym_finally] = ACTIONS(1802),
    [anon_sym_int] = ACTIONS(1802),
    [anon_sym_boolean] = ACTIONS(1802),
    [anon_sym_char] = ACTIONS(1802),
    [anon_sym_short] = ACTIONS(1802),
    [anon_sym_long] = ACTIONS(1802),
    [anon_sym_float] = ACTIONS(1802),
    [anon_sym_double] = ACTIONS(1802),
    [anon_sym_void] = ACTIONS(1802),
    [anon_sym_public] = ACTIONS(1802),
    [anon_sym_protected] = ACTIONS(1802),
    [anon_sym_private] = ACTIONS(1802),
    [anon_sym_static] = ACTIONS(1802),
    [anon_sym_final] = ACTIONS(1802),
    [anon_sym_synchronized] = ACTIONS(1802),
    [anon_sym_TILDE] = ACTIONS(1800),
    [anon_sym_BANG] = ACTIONS(1800),
    [anon_sym_new] = ACTIONS(1802),
  },
  [STATE(867)] = {
    [sym_comment] = STATE(867),
    [sym_groovy_doc] = STATE(867),
    [sym_parameter_list] = STATE(930),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(2096),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_case] = ACTIONS(1739),
    [anon_sym_default] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(868)] = {
    [sym_comment] = STATE(868),
    [sym_groovy_doc] = STATE(868),
    [ts_builtin_sym_end] = ACTIONS(1830),
    [sym_identifier] = ACTIONS(1832),
    [anon_sym_break] = ACTIONS(1832),
    [anon_sym_continue] = ACTIONS(1832),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_STAR] = ACTIONS(1832),
    [anon_sym_STAR_COLON] = ACTIONS(1830),
    [anon_sym_import] = ACTIONS(1832),
    [anon_sym_package] = ACTIONS(1832),
    [anon_sym_AT] = ACTIONS(1832),
    [anon_sym_assert] = ACTIONS(1832),
    [anon_sym_PLUS_PLUS] = ACTIONS(1830),
    [anon_sym_DASH_DASH] = ACTIONS(1830),
    [anon_sym_SLASH] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(1832),
    [anon_sym_DASH] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1832),
    [anon_sym_false] = ACTIONS(1832),
    [anon_sym_ATinterface] = ACTIONS(1830),
    [anon_sym_interface] = ACTIONS(1832),
    [anon_sym_class] = ACTIONS(1832),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_RBRACE] = ACTIONS(1830),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1832),
    [anon_sym_def] = ACTIONS(1832),
    [anon_sym_LPAREN] = ACTIONS(1830),
    [anon_sym_null] = ACTIONS(1832),
    [anon_sym_this] = ACTIONS(1832),
    [anon_sym_do] = ACTIONS(1832),
    [anon_sym_while] = ACTIONS(1832),
    [anon_sym_for] = ACTIONS(1832),
    [aux_sym__type_identifier_token1] = ACTIONS(1832),
    [anon_sym_if] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [aux_sym_number_literal_token1] = ACTIONS(1832),
    [aux_sym_number_literal_token2] = ACTIONS(1830),
    [aux_sym_number_literal_token3] = ACTIONS(1830),
    [aux_sym_number_literal_token4] = ACTIONS(1832),
    [aux_sym_number_literal_token5] = ACTIONS(1830),
    [anon_sym_pipeline] = ACTIONS(1832),
    [anon_sym_return] = ACTIONS(1832),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1830),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1830),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1830),
    [anon_sym_switch] = ACTIONS(1832),
    [anon_sym_try] = ACTIONS(1832),
    [anon_sym_catch] = ACTIONS(2098),
    [anon_sym_finally] = ACTIONS(2100),
    [anon_sym_int] = ACTIONS(1832),
    [anon_sym_boolean] = ACTIONS(1832),
    [anon_sym_char] = ACTIONS(1832),
    [anon_sym_short] = ACTIONS(1832),
    [anon_sym_long] = ACTIONS(1832),
    [anon_sym_float] = ACTIONS(1832),
    [anon_sym_double] = ACTIONS(1832),
    [anon_sym_void] = ACTIONS(1832),
    [anon_sym_public] = ACTIONS(1832),
    [anon_sym_protected] = ACTIONS(1832),
    [anon_sym_private] = ACTIONS(1832),
    [anon_sym_static] = ACTIONS(1832),
    [anon_sym_final] = ACTIONS(1832),
    [anon_sym_synchronized] = ACTIONS(1832),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_BANG] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1832),
  },
  [STATE(869)] = {
    [sym_comment] = STATE(869),
    [sym_groovy_doc] = STATE(869),
    [sym_identifier] = ACTIONS(1838),
    [anon_sym_break] = ACTIONS(1838),
    [anon_sym_continue] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_STAR] = ACTIONS(1838),
    [anon_sym_STAR_COLON] = ACTIONS(1836),
    [anon_sym_import] = ACTIONS(1838),
    [anon_sym_package] = ACTIONS(1838),
    [anon_sym_AT] = ACTIONS(1838),
    [anon_sym_assert] = ACTIONS(1838),
    [anon_sym_PLUS_PLUS] = ACTIONS(1836),
    [anon_sym_DASH_DASH] = ACTIONS(1836),
    [anon_sym_SLASH] = ACTIONS(1838),
    [anon_sym_PLUS] = ACTIONS(1838),
    [anon_sym_DASH] = ACTIONS(1838),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_ATinterface] = ACTIONS(1836),
    [anon_sym_interface] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1838),
    [anon_sym_def] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_null] = ACTIONS(1838),
    [anon_sym_this] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_while] = ACTIONS(1838),
    [anon_sym_for] = ACTIONS(1838),
    [aux_sym__type_identifier_token1] = ACTIONS(1838),
    [anon_sym_if] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1836),
    [aux_sym_number_literal_token1] = ACTIONS(1838),
    [aux_sym_number_literal_token2] = ACTIONS(1836),
    [aux_sym_number_literal_token3] = ACTIONS(1836),
    [aux_sym_number_literal_token4] = ACTIONS(1838),
    [aux_sym_number_literal_token5] = ACTIONS(1836),
    [anon_sym_return] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1838),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1836),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1836),
    [anon_sym_switch] = ACTIONS(1838),
    [anon_sym_case] = ACTIONS(1838),
    [anon_sym_default] = ACTIONS(1838),
    [anon_sym_try] = ACTIONS(1838),
    [anon_sym_catch] = ACTIONS(1838),
    [anon_sym_finally] = ACTIONS(1838),
    [anon_sym_int] = ACTIONS(1838),
    [anon_sym_boolean] = ACTIONS(1838),
    [anon_sym_char] = ACTIONS(1838),
    [anon_sym_short] = ACTIONS(1838),
    [anon_sym_long] = ACTIONS(1838),
    [anon_sym_float] = ACTIONS(1838),
    [anon_sym_double] = ACTIONS(1838),
    [anon_sym_void] = ACTIONS(1838),
    [anon_sym_public] = ACTIONS(1838),
    [anon_sym_protected] = ACTIONS(1838),
    [anon_sym_private] = ACTIONS(1838),
    [anon_sym_static] = ACTIONS(1838),
    [anon_sym_final] = ACTIONS(1838),
    [anon_sym_synchronized] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1836),
    [anon_sym_BANG] = ACTIONS(1836),
    [anon_sym_new] = ACTIONS(1838),
  },
  [STATE(870)] = {
    [sym_comment] = STATE(870),
    [sym_groovy_doc] = STATE(870),
    [sym_identifier] = ACTIONS(1878),
    [anon_sym_break] = ACTIONS(1878),
    [anon_sym_continue] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_STAR] = ACTIONS(1878),
    [anon_sym_STAR_COLON] = ACTIONS(1876),
    [anon_sym_import] = ACTIONS(1878),
    [anon_sym_package] = ACTIONS(1878),
    [anon_sym_AT] = ACTIONS(1878),
    [anon_sym_assert] = ACTIONS(1878),
    [anon_sym_PLUS_PLUS] = ACTIONS(1876),
    [anon_sym_DASH_DASH] = ACTIONS(1876),
    [anon_sym_SLASH] = ACTIONS(1878),
    [anon_sym_PLUS] = ACTIONS(1878),
    [anon_sym_DASH] = ACTIONS(1878),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_ATinterface] = ACTIONS(1876),
    [anon_sym_interface] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1878),
    [anon_sym_def] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_null] = ACTIONS(1878),
    [anon_sym_this] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_while] = ACTIONS(1878),
    [anon_sym_for] = ACTIONS(1878),
    [aux_sym__type_identifier_token1] = ACTIONS(1878),
    [anon_sym_if] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1876),
    [aux_sym_number_literal_token1] = ACTIONS(1878),
    [aux_sym_number_literal_token2] = ACTIONS(1876),
    [aux_sym_number_literal_token3] = ACTIONS(1876),
    [aux_sym_number_literal_token4] = ACTIONS(1878),
    [aux_sym_number_literal_token5] = ACTIONS(1876),
    [anon_sym_return] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1878),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1878),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1876),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1876),
    [anon_sym_switch] = ACTIONS(1878),
    [anon_sym_case] = ACTIONS(1878),
    [anon_sym_default] = ACTIONS(1878),
    [anon_sym_try] = ACTIONS(1878),
    [anon_sym_catch] = ACTIONS(1878),
    [anon_sym_finally] = ACTIONS(1878),
    [anon_sym_int] = ACTIONS(1878),
    [anon_sym_boolean] = ACTIONS(1878),
    [anon_sym_char] = ACTIONS(1878),
    [anon_sym_short] = ACTIONS(1878),
    [anon_sym_long] = ACTIONS(1878),
    [anon_sym_float] = ACTIONS(1878),
    [anon_sym_double] = ACTIONS(1878),
    [anon_sym_void] = ACTIONS(1878),
    [anon_sym_public] = ACTIONS(1878),
    [anon_sym_protected] = ACTIONS(1878),
    [anon_sym_private] = ACTIONS(1878),
    [anon_sym_static] = ACTIONS(1878),
    [anon_sym_final] = ACTIONS(1878),
    [anon_sym_synchronized] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1876),
    [anon_sym_BANG] = ACTIONS(1876),
    [anon_sym_new] = ACTIONS(1878),
  },
  [STATE(871)] = {
    [sym_comment] = STATE(871),
    [sym_groovy_doc] = STATE(871),
    [ts_builtin_sym_end] = ACTIONS(2029),
    [sym_identifier] = ACTIONS(2031),
    [anon_sym_break] = ACTIONS(2031),
    [anon_sym_continue] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_STAR] = ACTIONS(2031),
    [anon_sym_STAR_COLON] = ACTIONS(2029),
    [anon_sym_import] = ACTIONS(2031),
    [anon_sym_package] = ACTIONS(2031),
    [anon_sym_AT] = ACTIONS(2031),
    [anon_sym_assert] = ACTIONS(2031),
    [anon_sym_PLUS_PLUS] = ACTIONS(2029),
    [anon_sym_DASH_DASH] = ACTIONS(2029),
    [anon_sym_SLASH] = ACTIONS(2031),
    [anon_sym_PLUS] = ACTIONS(2031),
    [anon_sym_DASH] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_ATinterface] = ACTIONS(2029),
    [anon_sym_interface] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_COMMA] = ACTIONS(2029),
    [anon_sym_LBRACE] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2031),
    [anon_sym_def] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_RPAREN] = ACTIONS(2029),
    [anon_sym_null] = ACTIONS(2031),
    [anon_sym_this] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_while] = ACTIONS(2031),
    [anon_sym_for] = ACTIONS(2031),
    [aux_sym__type_identifier_token1] = ACTIONS(2031),
    [anon_sym_if] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2029),
    [aux_sym_number_literal_token1] = ACTIONS(2031),
    [aux_sym_number_literal_token2] = ACTIONS(2029),
    [aux_sym_number_literal_token3] = ACTIONS(2029),
    [aux_sym_number_literal_token4] = ACTIONS(2031),
    [aux_sym_number_literal_token5] = ACTIONS(2029),
    [anon_sym_pipeline] = ACTIONS(2031),
    [anon_sym_return] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2031),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2031),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2029),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2029),
    [anon_sym_switch] = ACTIONS(2031),
    [anon_sym_try] = ACTIONS(2031),
    [anon_sym_int] = ACTIONS(2031),
    [anon_sym_boolean] = ACTIONS(2031),
    [anon_sym_char] = ACTIONS(2031),
    [anon_sym_short] = ACTIONS(2031),
    [anon_sym_long] = ACTIONS(2031),
    [anon_sym_float] = ACTIONS(2031),
    [anon_sym_double] = ACTIONS(2031),
    [anon_sym_void] = ACTIONS(2031),
    [anon_sym_public] = ACTIONS(2031),
    [anon_sym_protected] = ACTIONS(2031),
    [anon_sym_private] = ACTIONS(2031),
    [anon_sym_static] = ACTIONS(2031),
    [anon_sym_final] = ACTIONS(2031),
    [anon_sym_synchronized] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2029),
    [anon_sym_BANG] = ACTIONS(2029),
    [anon_sym_new] = ACTIONS(2031),
  },
  [STATE(872)] = {
    [sym_comment] = STATE(872),
    [sym_groovy_doc] = STATE(872),
    [sym_identifier] = ACTIONS(1910),
    [anon_sym_break] = ACTIONS(1910),
    [anon_sym_continue] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_STAR] = ACTIONS(1910),
    [anon_sym_STAR_COLON] = ACTIONS(1908),
    [anon_sym_import] = ACTIONS(1910),
    [anon_sym_package] = ACTIONS(1910),
    [anon_sym_AT] = ACTIONS(1910),
    [anon_sym_assert] = ACTIONS(1910),
    [anon_sym_PLUS_PLUS] = ACTIONS(1908),
    [anon_sym_DASH_DASH] = ACTIONS(1908),
    [anon_sym_SLASH] = ACTIONS(1910),
    [anon_sym_PLUS] = ACTIONS(1910),
    [anon_sym_DASH] = ACTIONS(1910),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_ATinterface] = ACTIONS(1908),
    [anon_sym_interface] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1910),
    [anon_sym_def] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_null] = ACTIONS(1910),
    [anon_sym_this] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_while] = ACTIONS(1910),
    [anon_sym_for] = ACTIONS(1910),
    [aux_sym__type_identifier_token1] = ACTIONS(1910),
    [anon_sym_if] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1908),
    [aux_sym_number_literal_token1] = ACTIONS(1910),
    [aux_sym_number_literal_token2] = ACTIONS(1908),
    [aux_sym_number_literal_token3] = ACTIONS(1908),
    [aux_sym_number_literal_token4] = ACTIONS(1910),
    [aux_sym_number_literal_token5] = ACTIONS(1908),
    [anon_sym_return] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1910),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1910),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1908),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1908),
    [anon_sym_switch] = ACTIONS(1910),
    [anon_sym_case] = ACTIONS(1910),
    [anon_sym_default] = ACTIONS(1910),
    [anon_sym_try] = ACTIONS(1910),
    [anon_sym_catch] = ACTIONS(1910),
    [anon_sym_finally] = ACTIONS(1910),
    [anon_sym_int] = ACTIONS(1910),
    [anon_sym_boolean] = ACTIONS(1910),
    [anon_sym_char] = ACTIONS(1910),
    [anon_sym_short] = ACTIONS(1910),
    [anon_sym_long] = ACTIONS(1910),
    [anon_sym_float] = ACTIONS(1910),
    [anon_sym_double] = ACTIONS(1910),
    [anon_sym_void] = ACTIONS(1910),
    [anon_sym_public] = ACTIONS(1910),
    [anon_sym_protected] = ACTIONS(1910),
    [anon_sym_private] = ACTIONS(1910),
    [anon_sym_static] = ACTIONS(1910),
    [anon_sym_final] = ACTIONS(1910),
    [anon_sym_synchronized] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1908),
    [anon_sym_BANG] = ACTIONS(1908),
    [anon_sym_new] = ACTIONS(1910),
  },
  [STATE(873)] = {
    [sym_comment] = STATE(873),
    [sym_groovy_doc] = STATE(873),
    [sym_identifier] = ACTIONS(1914),
    [anon_sym_break] = ACTIONS(1914),
    [anon_sym_continue] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_STAR] = ACTIONS(1914),
    [anon_sym_STAR_COLON] = ACTIONS(1912),
    [anon_sym_import] = ACTIONS(1914),
    [anon_sym_package] = ACTIONS(1914),
    [anon_sym_AT] = ACTIONS(1914),
    [anon_sym_assert] = ACTIONS(1914),
    [anon_sym_PLUS_PLUS] = ACTIONS(1912),
    [anon_sym_DASH_DASH] = ACTIONS(1912),
    [anon_sym_SLASH] = ACTIONS(1914),
    [anon_sym_PLUS] = ACTIONS(1914),
    [anon_sym_DASH] = ACTIONS(1914),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_ATinterface] = ACTIONS(1912),
    [anon_sym_interface] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1914),
    [anon_sym_def] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_null] = ACTIONS(1914),
    [anon_sym_this] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_while] = ACTIONS(1914),
    [anon_sym_for] = ACTIONS(1914),
    [aux_sym__type_identifier_token1] = ACTIONS(1914),
    [anon_sym_if] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1912),
    [aux_sym_number_literal_token1] = ACTIONS(1914),
    [aux_sym_number_literal_token2] = ACTIONS(1912),
    [aux_sym_number_literal_token3] = ACTIONS(1912),
    [aux_sym_number_literal_token4] = ACTIONS(1914),
    [aux_sym_number_literal_token5] = ACTIONS(1912),
    [anon_sym_return] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1914),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1914),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1912),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1912),
    [anon_sym_switch] = ACTIONS(1914),
    [anon_sym_case] = ACTIONS(1914),
    [anon_sym_default] = ACTIONS(1914),
    [anon_sym_try] = ACTIONS(1914),
    [anon_sym_catch] = ACTIONS(1914),
    [anon_sym_finally] = ACTIONS(1914),
    [anon_sym_int] = ACTIONS(1914),
    [anon_sym_boolean] = ACTIONS(1914),
    [anon_sym_char] = ACTIONS(1914),
    [anon_sym_short] = ACTIONS(1914),
    [anon_sym_long] = ACTIONS(1914),
    [anon_sym_float] = ACTIONS(1914),
    [anon_sym_double] = ACTIONS(1914),
    [anon_sym_void] = ACTIONS(1914),
    [anon_sym_public] = ACTIONS(1914),
    [anon_sym_protected] = ACTIONS(1914),
    [anon_sym_private] = ACTIONS(1914),
    [anon_sym_static] = ACTIONS(1914),
    [anon_sym_final] = ACTIONS(1914),
    [anon_sym_synchronized] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1912),
    [anon_sym_BANG] = ACTIONS(1912),
    [anon_sym_new] = ACTIONS(1914),
  },
  [STATE(874)] = {
    [sym_comment] = STATE(874),
    [sym_groovy_doc] = STATE(874),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_break] = ACTIONS(1922),
    [anon_sym_continue] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_STAR] = ACTIONS(1922),
    [anon_sym_STAR_COLON] = ACTIONS(1920),
    [anon_sym_import] = ACTIONS(1922),
    [anon_sym_package] = ACTIONS(1922),
    [anon_sym_AT] = ACTIONS(1922),
    [anon_sym_assert] = ACTIONS(1922),
    [anon_sym_PLUS_PLUS] = ACTIONS(1920),
    [anon_sym_DASH_DASH] = ACTIONS(1920),
    [anon_sym_SLASH] = ACTIONS(1922),
    [anon_sym_PLUS] = ACTIONS(1922),
    [anon_sym_DASH] = ACTIONS(1922),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_ATinterface] = ACTIONS(1920),
    [anon_sym_interface] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1922),
    [anon_sym_def] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_null] = ACTIONS(1922),
    [anon_sym_this] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_while] = ACTIONS(1922),
    [anon_sym_for] = ACTIONS(1922),
    [aux_sym__type_identifier_token1] = ACTIONS(1922),
    [anon_sym_if] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1920),
    [aux_sym_number_literal_token1] = ACTIONS(1922),
    [aux_sym_number_literal_token2] = ACTIONS(1920),
    [aux_sym_number_literal_token3] = ACTIONS(1920),
    [aux_sym_number_literal_token4] = ACTIONS(1922),
    [aux_sym_number_literal_token5] = ACTIONS(1920),
    [anon_sym_return] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1922),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1922),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1920),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1920),
    [anon_sym_switch] = ACTIONS(1922),
    [anon_sym_case] = ACTIONS(1922),
    [anon_sym_default] = ACTIONS(1922),
    [anon_sym_try] = ACTIONS(1922),
    [anon_sym_catch] = ACTIONS(1922),
    [anon_sym_finally] = ACTIONS(1922),
    [anon_sym_int] = ACTIONS(1922),
    [anon_sym_boolean] = ACTIONS(1922),
    [anon_sym_char] = ACTIONS(1922),
    [anon_sym_short] = ACTIONS(1922),
    [anon_sym_long] = ACTIONS(1922),
    [anon_sym_float] = ACTIONS(1922),
    [anon_sym_double] = ACTIONS(1922),
    [anon_sym_void] = ACTIONS(1922),
    [anon_sym_public] = ACTIONS(1922),
    [anon_sym_protected] = ACTIONS(1922),
    [anon_sym_private] = ACTIONS(1922),
    [anon_sym_static] = ACTIONS(1922),
    [anon_sym_final] = ACTIONS(1922),
    [anon_sym_synchronized] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1920),
    [anon_sym_BANG] = ACTIONS(1920),
    [anon_sym_new] = ACTIONS(1922),
  },
  [STATE(875)] = {
    [sym_comment] = STATE(875),
    [sym_groovy_doc] = STATE(875),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_break] = ACTIONS(1926),
    [anon_sym_continue] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_STAR] = ACTIONS(1926),
    [anon_sym_STAR_COLON] = ACTIONS(1924),
    [anon_sym_import] = ACTIONS(1926),
    [anon_sym_package] = ACTIONS(1926),
    [anon_sym_AT] = ACTIONS(1926),
    [anon_sym_assert] = ACTIONS(1926),
    [anon_sym_PLUS_PLUS] = ACTIONS(1924),
    [anon_sym_DASH_DASH] = ACTIONS(1924),
    [anon_sym_SLASH] = ACTIONS(1926),
    [anon_sym_PLUS] = ACTIONS(1926),
    [anon_sym_DASH] = ACTIONS(1926),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_ATinterface] = ACTIONS(1924),
    [anon_sym_interface] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1926),
    [anon_sym_def] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_null] = ACTIONS(1926),
    [anon_sym_this] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_while] = ACTIONS(1926),
    [anon_sym_for] = ACTIONS(1926),
    [aux_sym__type_identifier_token1] = ACTIONS(1926),
    [anon_sym_if] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1924),
    [aux_sym_number_literal_token1] = ACTIONS(1926),
    [aux_sym_number_literal_token2] = ACTIONS(1924),
    [aux_sym_number_literal_token3] = ACTIONS(1924),
    [aux_sym_number_literal_token4] = ACTIONS(1926),
    [aux_sym_number_literal_token5] = ACTIONS(1924),
    [anon_sym_return] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1926),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1924),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1924),
    [anon_sym_switch] = ACTIONS(1926),
    [anon_sym_case] = ACTIONS(1926),
    [anon_sym_default] = ACTIONS(1926),
    [anon_sym_try] = ACTIONS(1926),
    [anon_sym_catch] = ACTIONS(1926),
    [anon_sym_finally] = ACTIONS(1926),
    [anon_sym_int] = ACTIONS(1926),
    [anon_sym_boolean] = ACTIONS(1926),
    [anon_sym_char] = ACTIONS(1926),
    [anon_sym_short] = ACTIONS(1926),
    [anon_sym_long] = ACTIONS(1926),
    [anon_sym_float] = ACTIONS(1926),
    [anon_sym_double] = ACTIONS(1926),
    [anon_sym_void] = ACTIONS(1926),
    [anon_sym_public] = ACTIONS(1926),
    [anon_sym_protected] = ACTIONS(1926),
    [anon_sym_private] = ACTIONS(1926),
    [anon_sym_static] = ACTIONS(1926),
    [anon_sym_final] = ACTIONS(1926),
    [anon_sym_synchronized] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1924),
    [anon_sym_BANG] = ACTIONS(1924),
    [anon_sym_new] = ACTIONS(1926),
  },
  [STATE(876)] = {
    [sym_comment] = STATE(876),
    [sym_groovy_doc] = STATE(876),
    [sym_identifier] = ACTIONS(1930),
    [anon_sym_break] = ACTIONS(1930),
    [anon_sym_continue] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_STAR] = ACTIONS(1930),
    [anon_sym_STAR_COLON] = ACTIONS(1928),
    [anon_sym_import] = ACTIONS(1930),
    [anon_sym_package] = ACTIONS(1930),
    [anon_sym_AT] = ACTIONS(1930),
    [anon_sym_assert] = ACTIONS(1930),
    [anon_sym_PLUS_PLUS] = ACTIONS(1928),
    [anon_sym_DASH_DASH] = ACTIONS(1928),
    [anon_sym_SLASH] = ACTIONS(1930),
    [anon_sym_PLUS] = ACTIONS(1930),
    [anon_sym_DASH] = ACTIONS(1930),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_ATinterface] = ACTIONS(1928),
    [anon_sym_interface] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1930),
    [anon_sym_def] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_null] = ACTIONS(1930),
    [anon_sym_this] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_while] = ACTIONS(1930),
    [anon_sym_for] = ACTIONS(1930),
    [aux_sym__type_identifier_token1] = ACTIONS(1930),
    [anon_sym_if] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1928),
    [aux_sym_number_literal_token1] = ACTIONS(1930),
    [aux_sym_number_literal_token2] = ACTIONS(1928),
    [aux_sym_number_literal_token3] = ACTIONS(1928),
    [aux_sym_number_literal_token4] = ACTIONS(1930),
    [aux_sym_number_literal_token5] = ACTIONS(1928),
    [anon_sym_return] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1930),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1928),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1928),
    [anon_sym_switch] = ACTIONS(1930),
    [anon_sym_case] = ACTIONS(1930),
    [anon_sym_default] = ACTIONS(1930),
    [anon_sym_try] = ACTIONS(1930),
    [anon_sym_catch] = ACTIONS(1930),
    [anon_sym_finally] = ACTIONS(1930),
    [anon_sym_int] = ACTIONS(1930),
    [anon_sym_boolean] = ACTIONS(1930),
    [anon_sym_char] = ACTIONS(1930),
    [anon_sym_short] = ACTIONS(1930),
    [anon_sym_long] = ACTIONS(1930),
    [anon_sym_float] = ACTIONS(1930),
    [anon_sym_double] = ACTIONS(1930),
    [anon_sym_void] = ACTIONS(1930),
    [anon_sym_public] = ACTIONS(1930),
    [anon_sym_protected] = ACTIONS(1930),
    [anon_sym_private] = ACTIONS(1930),
    [anon_sym_static] = ACTIONS(1930),
    [anon_sym_final] = ACTIONS(1930),
    [anon_sym_synchronized] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1928),
    [anon_sym_BANG] = ACTIONS(1928),
    [anon_sym_new] = ACTIONS(1930),
  },
  [STATE(877)] = {
    [sym_comment] = STATE(877),
    [sym_groovy_doc] = STATE(877),
    [sym_identifier] = ACTIONS(1934),
    [anon_sym_break] = ACTIONS(1934),
    [anon_sym_continue] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_STAR] = ACTIONS(1934),
    [anon_sym_STAR_COLON] = ACTIONS(1932),
    [anon_sym_import] = ACTIONS(1934),
    [anon_sym_package] = ACTIONS(1934),
    [anon_sym_AT] = ACTIONS(1934),
    [anon_sym_assert] = ACTIONS(1934),
    [anon_sym_PLUS_PLUS] = ACTIONS(1932),
    [anon_sym_DASH_DASH] = ACTIONS(1932),
    [anon_sym_SLASH] = ACTIONS(1934),
    [anon_sym_PLUS] = ACTIONS(1934),
    [anon_sym_DASH] = ACTIONS(1934),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_ATinterface] = ACTIONS(1932),
    [anon_sym_interface] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1934),
    [anon_sym_def] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_null] = ACTIONS(1934),
    [anon_sym_this] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_while] = ACTIONS(1934),
    [anon_sym_for] = ACTIONS(1934),
    [aux_sym__type_identifier_token1] = ACTIONS(1934),
    [anon_sym_if] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1932),
    [aux_sym_number_literal_token1] = ACTIONS(1934),
    [aux_sym_number_literal_token2] = ACTIONS(1932),
    [aux_sym_number_literal_token3] = ACTIONS(1932),
    [aux_sym_number_literal_token4] = ACTIONS(1934),
    [aux_sym_number_literal_token5] = ACTIONS(1932),
    [anon_sym_return] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1934),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1932),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1932),
    [anon_sym_switch] = ACTIONS(1934),
    [anon_sym_case] = ACTIONS(1934),
    [anon_sym_default] = ACTIONS(1934),
    [anon_sym_try] = ACTIONS(1934),
    [anon_sym_catch] = ACTIONS(1934),
    [anon_sym_finally] = ACTIONS(1934),
    [anon_sym_int] = ACTIONS(1934),
    [anon_sym_boolean] = ACTIONS(1934),
    [anon_sym_char] = ACTIONS(1934),
    [anon_sym_short] = ACTIONS(1934),
    [anon_sym_long] = ACTIONS(1934),
    [anon_sym_float] = ACTIONS(1934),
    [anon_sym_double] = ACTIONS(1934),
    [anon_sym_void] = ACTIONS(1934),
    [anon_sym_public] = ACTIONS(1934),
    [anon_sym_protected] = ACTIONS(1934),
    [anon_sym_private] = ACTIONS(1934),
    [anon_sym_static] = ACTIONS(1934),
    [anon_sym_final] = ACTIONS(1934),
    [anon_sym_synchronized] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1932),
    [anon_sym_BANG] = ACTIONS(1932),
    [anon_sym_new] = ACTIONS(1934),
  },
  [STATE(878)] = {
    [sym_comment] = STATE(878),
    [sym_groovy_doc] = STATE(878),
    [sym_identifier] = ACTIONS(1832),
    [anon_sym_break] = ACTIONS(1832),
    [anon_sym_continue] = ACTIONS(1832),
    [anon_sym_SEMI] = ACTIONS(1830),
    [anon_sym_STAR] = ACTIONS(1832),
    [anon_sym_STAR_COLON] = ACTIONS(1830),
    [anon_sym_import] = ACTIONS(1832),
    [anon_sym_package] = ACTIONS(1832),
    [anon_sym_AT] = ACTIONS(1832),
    [anon_sym_assert] = ACTIONS(1832),
    [anon_sym_PLUS_PLUS] = ACTIONS(1830),
    [anon_sym_DASH_DASH] = ACTIONS(1830),
    [anon_sym_SLASH] = ACTIONS(1832),
    [anon_sym_PLUS] = ACTIONS(1832),
    [anon_sym_DASH] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1832),
    [anon_sym_false] = ACTIONS(1832),
    [anon_sym_ATinterface] = ACTIONS(1830),
    [anon_sym_interface] = ACTIONS(1832),
    [anon_sym_class] = ACTIONS(1832),
    [anon_sym_LBRACE] = ACTIONS(1830),
    [anon_sym_RBRACE] = ACTIONS(1830),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1832),
    [anon_sym_def] = ACTIONS(1832),
    [anon_sym_LPAREN] = ACTIONS(1830),
    [anon_sym_null] = ACTIONS(1832),
    [anon_sym_this] = ACTIONS(1832),
    [anon_sym_do] = ACTIONS(1832),
    [anon_sym_while] = ACTIONS(1832),
    [anon_sym_for] = ACTIONS(1832),
    [aux_sym__type_identifier_token1] = ACTIONS(1832),
    [anon_sym_if] = ACTIONS(1832),
    [anon_sym_else] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1830),
    [aux_sym_number_literal_token1] = ACTIONS(1832),
    [aux_sym_number_literal_token2] = ACTIONS(1830),
    [aux_sym_number_literal_token3] = ACTIONS(1830),
    [aux_sym_number_literal_token4] = ACTIONS(1832),
    [aux_sym_number_literal_token5] = ACTIONS(1830),
    [anon_sym_return] = ACTIONS(1832),
    [anon_sym_SQUOTE] = ACTIONS(1832),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1830),
    [anon_sym_DQUOTE] = ACTIONS(1832),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1830),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1830),
    [anon_sym_switch] = ACTIONS(1832),
    [anon_sym_case] = ACTIONS(1832),
    [anon_sym_default] = ACTIONS(1832),
    [anon_sym_try] = ACTIONS(1832),
    [anon_sym_catch] = ACTIONS(2102),
    [anon_sym_finally] = ACTIONS(2104),
    [anon_sym_int] = ACTIONS(1832),
    [anon_sym_boolean] = ACTIONS(1832),
    [anon_sym_char] = ACTIONS(1832),
    [anon_sym_short] = ACTIONS(1832),
    [anon_sym_long] = ACTIONS(1832),
    [anon_sym_float] = ACTIONS(1832),
    [anon_sym_double] = ACTIONS(1832),
    [anon_sym_void] = ACTIONS(1832),
    [anon_sym_public] = ACTIONS(1832),
    [anon_sym_protected] = ACTIONS(1832),
    [anon_sym_private] = ACTIONS(1832),
    [anon_sym_static] = ACTIONS(1832),
    [anon_sym_final] = ACTIONS(1832),
    [anon_sym_synchronized] = ACTIONS(1832),
    [anon_sym_TILDE] = ACTIONS(1830),
    [anon_sym_BANG] = ACTIONS(1830),
    [anon_sym_new] = ACTIONS(1832),
  },
  [STATE(879)] = {
    [sym_comment] = STATE(879),
    [sym_groovy_doc] = STATE(879),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_break] = ACTIONS(1938),
    [anon_sym_continue] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_STAR] = ACTIONS(1938),
    [anon_sym_STAR_COLON] = ACTIONS(1936),
    [anon_sym_import] = ACTIONS(1938),
    [anon_sym_package] = ACTIONS(1938),
    [anon_sym_AT] = ACTIONS(1938),
    [anon_sym_assert] = ACTIONS(1938),
    [anon_sym_PLUS_PLUS] = ACTIONS(1936),
    [anon_sym_DASH_DASH] = ACTIONS(1936),
    [anon_sym_SLASH] = ACTIONS(1938),
    [anon_sym_PLUS] = ACTIONS(1938),
    [anon_sym_DASH] = ACTIONS(1938),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_ATinterface] = ACTIONS(1936),
    [anon_sym_interface] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1938),
    [anon_sym_def] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_null] = ACTIONS(1938),
    [anon_sym_this] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_while] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1938),
    [aux_sym__type_identifier_token1] = ACTIONS(1938),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1936),
    [aux_sym_number_literal_token1] = ACTIONS(1938),
    [aux_sym_number_literal_token2] = ACTIONS(1936),
    [aux_sym_number_literal_token3] = ACTIONS(1936),
    [aux_sym_number_literal_token4] = ACTIONS(1938),
    [aux_sym_number_literal_token5] = ACTIONS(1936),
    [anon_sym_return] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1938),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1936),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1936),
    [anon_sym_switch] = ACTIONS(1938),
    [anon_sym_case] = ACTIONS(1938),
    [anon_sym_default] = ACTIONS(1938),
    [anon_sym_try] = ACTIONS(1938),
    [anon_sym_catch] = ACTIONS(1938),
    [anon_sym_finally] = ACTIONS(1938),
    [anon_sym_int] = ACTIONS(1938),
    [anon_sym_boolean] = ACTIONS(1938),
    [anon_sym_char] = ACTIONS(1938),
    [anon_sym_short] = ACTIONS(1938),
    [anon_sym_long] = ACTIONS(1938),
    [anon_sym_float] = ACTIONS(1938),
    [anon_sym_double] = ACTIONS(1938),
    [anon_sym_void] = ACTIONS(1938),
    [anon_sym_public] = ACTIONS(1938),
    [anon_sym_protected] = ACTIONS(1938),
    [anon_sym_private] = ACTIONS(1938),
    [anon_sym_static] = ACTIONS(1938),
    [anon_sym_final] = ACTIONS(1938),
    [anon_sym_synchronized] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1936),
    [anon_sym_BANG] = ACTIONS(1936),
    [anon_sym_new] = ACTIONS(1938),
  },
  [STATE(880)] = {
    [sym_comment] = STATE(880),
    [sym_groovy_doc] = STATE(880),
    [sym_parameter_list] = STATE(931),
    [ts_builtin_sym_end] = ACTIONS(1741),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(2106),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_pipeline] = ACTIONS(1739),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(881)] = {
    [sym_comment] = STATE(881),
    [sym_groovy_doc] = STATE(881),
    [sym_identifier] = ACTIONS(1942),
    [anon_sym_break] = ACTIONS(1942),
    [anon_sym_continue] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1942),
    [anon_sym_STAR_COLON] = ACTIONS(1940),
    [anon_sym_import] = ACTIONS(1942),
    [anon_sym_package] = ACTIONS(1942),
    [anon_sym_AT] = ACTIONS(1942),
    [anon_sym_assert] = ACTIONS(1942),
    [anon_sym_PLUS_PLUS] = ACTIONS(1940),
    [anon_sym_DASH_DASH] = ACTIONS(1940),
    [anon_sym_SLASH] = ACTIONS(1942),
    [anon_sym_PLUS] = ACTIONS(1942),
    [anon_sym_DASH] = ACTIONS(1942),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_ATinterface] = ACTIONS(1940),
    [anon_sym_interface] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1942),
    [anon_sym_def] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_null] = ACTIONS(1942),
    [anon_sym_this] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_for] = ACTIONS(1942),
    [aux_sym__type_identifier_token1] = ACTIONS(1942),
    [anon_sym_if] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1940),
    [aux_sym_number_literal_token1] = ACTIONS(1942),
    [aux_sym_number_literal_token2] = ACTIONS(1940),
    [aux_sym_number_literal_token3] = ACTIONS(1940),
    [aux_sym_number_literal_token4] = ACTIONS(1942),
    [aux_sym_number_literal_token5] = ACTIONS(1940),
    [anon_sym_return] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1942),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1940),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1940),
    [anon_sym_switch] = ACTIONS(1942),
    [anon_sym_case] = ACTIONS(1942),
    [anon_sym_default] = ACTIONS(1942),
    [anon_sym_try] = ACTIONS(1942),
    [anon_sym_catch] = ACTIONS(1942),
    [anon_sym_finally] = ACTIONS(1942),
    [anon_sym_int] = ACTIONS(1942),
    [anon_sym_boolean] = ACTIONS(1942),
    [anon_sym_char] = ACTIONS(1942),
    [anon_sym_short] = ACTIONS(1942),
    [anon_sym_long] = ACTIONS(1942),
    [anon_sym_float] = ACTIONS(1942),
    [anon_sym_double] = ACTIONS(1942),
    [anon_sym_void] = ACTIONS(1942),
    [anon_sym_public] = ACTIONS(1942),
    [anon_sym_protected] = ACTIONS(1942),
    [anon_sym_private] = ACTIONS(1942),
    [anon_sym_static] = ACTIONS(1942),
    [anon_sym_final] = ACTIONS(1942),
    [anon_sym_synchronized] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1940),
    [anon_sym_BANG] = ACTIONS(1940),
    [anon_sym_new] = ACTIONS(1942),
  },
  [STATE(882)] = {
    [sym_comment] = STATE(882),
    [sym_groovy_doc] = STATE(882),
    [sym_identifier] = ACTIONS(1950),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_STAR] = ACTIONS(1950),
    [anon_sym_STAR_COLON] = ACTIONS(1948),
    [anon_sym_import] = ACTIONS(1950),
    [anon_sym_package] = ACTIONS(1950),
    [anon_sym_AT] = ACTIONS(1950),
    [anon_sym_assert] = ACTIONS(1950),
    [anon_sym_PLUS_PLUS] = ACTIONS(1948),
    [anon_sym_DASH_DASH] = ACTIONS(1948),
    [anon_sym_SLASH] = ACTIONS(1950),
    [anon_sym_PLUS] = ACTIONS(1950),
    [anon_sym_DASH] = ACTIONS(1950),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_ATinterface] = ACTIONS(1948),
    [anon_sym_interface] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1950),
    [anon_sym_def] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_null] = ACTIONS(1950),
    [anon_sym_this] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_while] = ACTIONS(1950),
    [anon_sym_for] = ACTIONS(1950),
    [aux_sym__type_identifier_token1] = ACTIONS(1950),
    [anon_sym_if] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1948),
    [aux_sym_number_literal_token1] = ACTIONS(1950),
    [aux_sym_number_literal_token2] = ACTIONS(1948),
    [aux_sym_number_literal_token3] = ACTIONS(1948),
    [aux_sym_number_literal_token4] = ACTIONS(1950),
    [aux_sym_number_literal_token5] = ACTIONS(1948),
    [anon_sym_return] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1950),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1948),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1948),
    [anon_sym_switch] = ACTIONS(1950),
    [anon_sym_case] = ACTIONS(1950),
    [anon_sym_default] = ACTIONS(1950),
    [anon_sym_try] = ACTIONS(1950),
    [anon_sym_catch] = ACTIONS(1950),
    [anon_sym_finally] = ACTIONS(1950),
    [anon_sym_int] = ACTIONS(1950),
    [anon_sym_boolean] = ACTIONS(1950),
    [anon_sym_char] = ACTIONS(1950),
    [anon_sym_short] = ACTIONS(1950),
    [anon_sym_long] = ACTIONS(1950),
    [anon_sym_float] = ACTIONS(1950),
    [anon_sym_double] = ACTIONS(1950),
    [anon_sym_void] = ACTIONS(1950),
    [anon_sym_public] = ACTIONS(1950),
    [anon_sym_protected] = ACTIONS(1950),
    [anon_sym_private] = ACTIONS(1950),
    [anon_sym_static] = ACTIONS(1950),
    [anon_sym_final] = ACTIONS(1950),
    [anon_sym_synchronized] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1948),
    [anon_sym_BANG] = ACTIONS(1948),
    [anon_sym_new] = ACTIONS(1950),
  },
  [STATE(883)] = {
    [sym_comment] = STATE(883),
    [sym_groovy_doc] = STATE(883),
    [sym_identifier] = ACTIONS(1962),
    [anon_sym_break] = ACTIONS(1962),
    [anon_sym_continue] = ACTIONS(1962),
    [anon_sym_SEMI] = ACTIONS(1960),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_STAR_COLON] = ACTIONS(1960),
    [anon_sym_import] = ACTIONS(1962),
    [anon_sym_package] = ACTIONS(1962),
    [anon_sym_AT] = ACTIONS(1962),
    [anon_sym_assert] = ACTIONS(1962),
    [anon_sym_PLUS_PLUS] = ACTIONS(1960),
    [anon_sym_DASH_DASH] = ACTIONS(1960),
    [anon_sym_SLASH] = ACTIONS(1962),
    [anon_sym_PLUS] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_true] = ACTIONS(1962),
    [anon_sym_false] = ACTIONS(1962),
    [anon_sym_ATinterface] = ACTIONS(1960),
    [anon_sym_interface] = ACTIONS(1962),
    [anon_sym_class] = ACTIONS(1962),
    [anon_sym_LBRACE] = ACTIONS(1960),
    [anon_sym_RBRACE] = ACTIONS(1960),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1962),
    [anon_sym_def] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1960),
    [anon_sym_null] = ACTIONS(1962),
    [anon_sym_this] = ACTIONS(1962),
    [anon_sym_do] = ACTIONS(1962),
    [anon_sym_while] = ACTIONS(1962),
    [anon_sym_for] = ACTIONS(1962),
    [aux_sym__type_identifier_token1] = ACTIONS(1962),
    [anon_sym_if] = ACTIONS(1962),
    [anon_sym_else] = ACTIONS(1962),
    [anon_sym_LBRACK] = ACTIONS(1960),
    [aux_sym_number_literal_token1] = ACTIONS(1962),
    [aux_sym_number_literal_token2] = ACTIONS(1960),
    [aux_sym_number_literal_token3] = ACTIONS(1960),
    [aux_sym_number_literal_token4] = ACTIONS(1962),
    [aux_sym_number_literal_token5] = ACTIONS(1960),
    [anon_sym_return] = ACTIONS(1962),
    [anon_sym_SQUOTE] = ACTIONS(1962),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1960),
    [anon_sym_DQUOTE] = ACTIONS(1962),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1960),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1960),
    [anon_sym_switch] = ACTIONS(1962),
    [anon_sym_case] = ACTIONS(1962),
    [anon_sym_default] = ACTIONS(1962),
    [anon_sym_try] = ACTIONS(1962),
    [anon_sym_catch] = ACTIONS(1962),
    [anon_sym_finally] = ACTIONS(1962),
    [anon_sym_int] = ACTIONS(1962),
    [anon_sym_boolean] = ACTIONS(1962),
    [anon_sym_char] = ACTIONS(1962),
    [anon_sym_short] = ACTIONS(1962),
    [anon_sym_long] = ACTIONS(1962),
    [anon_sym_float] = ACTIONS(1962),
    [anon_sym_double] = ACTIONS(1962),
    [anon_sym_void] = ACTIONS(1962),
    [anon_sym_public] = ACTIONS(1962),
    [anon_sym_protected] = ACTIONS(1962),
    [anon_sym_private] = ACTIONS(1962),
    [anon_sym_static] = ACTIONS(1962),
    [anon_sym_final] = ACTIONS(1962),
    [anon_sym_synchronized] = ACTIONS(1962),
    [anon_sym_TILDE] = ACTIONS(1960),
    [anon_sym_BANG] = ACTIONS(1960),
    [anon_sym_new] = ACTIONS(1962),
  },
  [STATE(884)] = {
    [sym_comment] = STATE(884),
    [sym_groovy_doc] = STATE(884),
    [ts_builtin_sym_end] = ACTIONS(1900),
    [sym_identifier] = ACTIONS(1902),
    [anon_sym_break] = ACTIONS(1902),
    [anon_sym_continue] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_STAR] = ACTIONS(1902),
    [anon_sym_STAR_COLON] = ACTIONS(1900),
    [anon_sym_import] = ACTIONS(1902),
    [anon_sym_package] = ACTIONS(1902),
    [anon_sym_AT] = ACTIONS(1902),
    [anon_sym_assert] = ACTIONS(1902),
    [anon_sym_PLUS_PLUS] = ACTIONS(1900),
    [anon_sym_DASH_DASH] = ACTIONS(1900),
    [anon_sym_SLASH] = ACTIONS(1902),
    [anon_sym_PLUS] = ACTIONS(1902),
    [anon_sym_DASH] = ACTIONS(1902),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_ATinterface] = ACTIONS(1900),
    [anon_sym_interface] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_COMMA] = ACTIONS(1900),
    [anon_sym_LBRACE] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1902),
    [anon_sym_def] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_RPAREN] = ACTIONS(1900),
    [anon_sym_null] = ACTIONS(1902),
    [anon_sym_this] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_while] = ACTIONS(1902),
    [anon_sym_for] = ACTIONS(1902),
    [aux_sym__type_identifier_token1] = ACTIONS(1902),
    [anon_sym_if] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1900),
    [aux_sym_number_literal_token1] = ACTIONS(1902),
    [aux_sym_number_literal_token2] = ACTIONS(1900),
    [aux_sym_number_literal_token3] = ACTIONS(1900),
    [aux_sym_number_literal_token4] = ACTIONS(1902),
    [aux_sym_number_literal_token5] = ACTIONS(1900),
    [anon_sym_pipeline] = ACTIONS(1902),
    [anon_sym_return] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1902),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1902),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1900),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1900),
    [anon_sym_switch] = ACTIONS(1902),
    [anon_sym_try] = ACTIONS(1902),
    [anon_sym_int] = ACTIONS(1902),
    [anon_sym_boolean] = ACTIONS(1902),
    [anon_sym_char] = ACTIONS(1902),
    [anon_sym_short] = ACTIONS(1902),
    [anon_sym_long] = ACTIONS(1902),
    [anon_sym_float] = ACTIONS(1902),
    [anon_sym_double] = ACTIONS(1902),
    [anon_sym_void] = ACTIONS(1902),
    [anon_sym_public] = ACTIONS(1902),
    [anon_sym_protected] = ACTIONS(1902),
    [anon_sym_private] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1902),
    [anon_sym_final] = ACTIONS(1902),
    [anon_sym_synchronized] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1900),
    [anon_sym_BANG] = ACTIONS(1900),
    [anon_sym_new] = ACTIONS(1902),
  },
  [STATE(885)] = {
    [sym_comment] = STATE(885),
    [sym_groovy_doc] = STATE(885),
    [sym_identifier] = ACTIONS(1970),
    [anon_sym_break] = ACTIONS(1970),
    [anon_sym_continue] = ACTIONS(1970),
    [anon_sym_SEMI] = ACTIONS(1968),
    [anon_sym_STAR] = ACTIONS(1970),
    [anon_sym_STAR_COLON] = ACTIONS(1968),
    [anon_sym_import] = ACTIONS(1970),
    [anon_sym_package] = ACTIONS(1970),
    [anon_sym_AT] = ACTIONS(1970),
    [anon_sym_assert] = ACTIONS(1970),
    [anon_sym_PLUS_PLUS] = ACTIONS(1968),
    [anon_sym_DASH_DASH] = ACTIONS(1968),
    [anon_sym_SLASH] = ACTIONS(1970),
    [anon_sym_PLUS] = ACTIONS(1970),
    [anon_sym_DASH] = ACTIONS(1970),
    [anon_sym_true] = ACTIONS(1970),
    [anon_sym_false] = ACTIONS(1970),
    [anon_sym_ATinterface] = ACTIONS(1968),
    [anon_sym_interface] = ACTIONS(1970),
    [anon_sym_class] = ACTIONS(1970),
    [anon_sym_LBRACE] = ACTIONS(1968),
    [anon_sym_RBRACE] = ACTIONS(1968),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1970),
    [anon_sym_def] = ACTIONS(1970),
    [anon_sym_LPAREN] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_this] = ACTIONS(1970),
    [anon_sym_do] = ACTIONS(1970),
    [anon_sym_while] = ACTIONS(1970),
    [anon_sym_for] = ACTIONS(1970),
    [aux_sym__type_identifier_token1] = ACTIONS(1970),
    [anon_sym_if] = ACTIONS(1970),
    [anon_sym_else] = ACTIONS(1970),
    [anon_sym_LBRACK] = ACTIONS(1968),
    [aux_sym_number_literal_token1] = ACTIONS(1970),
    [aux_sym_number_literal_token2] = ACTIONS(1968),
    [aux_sym_number_literal_token3] = ACTIONS(1968),
    [aux_sym_number_literal_token4] = ACTIONS(1970),
    [aux_sym_number_literal_token5] = ACTIONS(1968),
    [anon_sym_return] = ACTIONS(1970),
    [anon_sym_SQUOTE] = ACTIONS(1970),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1968),
    [anon_sym_DQUOTE] = ACTIONS(1970),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1968),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1968),
    [anon_sym_switch] = ACTIONS(1970),
    [anon_sym_case] = ACTIONS(1970),
    [anon_sym_default] = ACTIONS(1970),
    [anon_sym_try] = ACTIONS(1970),
    [anon_sym_catch] = ACTIONS(1970),
    [anon_sym_finally] = ACTIONS(1970),
    [anon_sym_int] = ACTIONS(1970),
    [anon_sym_boolean] = ACTIONS(1970),
    [anon_sym_char] = ACTIONS(1970),
    [anon_sym_short] = ACTIONS(1970),
    [anon_sym_long] = ACTIONS(1970),
    [anon_sym_float] = ACTIONS(1970),
    [anon_sym_double] = ACTIONS(1970),
    [anon_sym_void] = ACTIONS(1970),
    [anon_sym_public] = ACTIONS(1970),
    [anon_sym_protected] = ACTIONS(1970),
    [anon_sym_private] = ACTIONS(1970),
    [anon_sym_static] = ACTIONS(1970),
    [anon_sym_final] = ACTIONS(1970),
    [anon_sym_synchronized] = ACTIONS(1970),
    [anon_sym_TILDE] = ACTIONS(1968),
    [anon_sym_BANG] = ACTIONS(1968),
    [anon_sym_new] = ACTIONS(1970),
  },
  [STATE(886)] = {
    [sym_comment] = STATE(886),
    [sym_groovy_doc] = STATE(886),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_break] = ACTIONS(1974),
    [anon_sym_continue] = ACTIONS(1974),
    [anon_sym_SEMI] = ACTIONS(1972),
    [anon_sym_STAR] = ACTIONS(1974),
    [anon_sym_STAR_COLON] = ACTIONS(1972),
    [anon_sym_import] = ACTIONS(1974),
    [anon_sym_package] = ACTIONS(1974),
    [anon_sym_AT] = ACTIONS(1974),
    [anon_sym_assert] = ACTIONS(1974),
    [anon_sym_PLUS_PLUS] = ACTIONS(1972),
    [anon_sym_DASH_DASH] = ACTIONS(1972),
    [anon_sym_SLASH] = ACTIONS(1974),
    [anon_sym_PLUS] = ACTIONS(1974),
    [anon_sym_DASH] = ACTIONS(1974),
    [anon_sym_true] = ACTIONS(1974),
    [anon_sym_false] = ACTIONS(1974),
    [anon_sym_ATinterface] = ACTIONS(1972),
    [anon_sym_interface] = ACTIONS(1974),
    [anon_sym_class] = ACTIONS(1974),
    [anon_sym_LBRACE] = ACTIONS(1972),
    [anon_sym_RBRACE] = ACTIONS(1972),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1974),
    [anon_sym_def] = ACTIONS(1974),
    [anon_sym_LPAREN] = ACTIONS(1972),
    [anon_sym_null] = ACTIONS(1974),
    [anon_sym_this] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1974),
    [anon_sym_while] = ACTIONS(1974),
    [anon_sym_for] = ACTIONS(1974),
    [aux_sym__type_identifier_token1] = ACTIONS(1974),
    [anon_sym_if] = ACTIONS(1974),
    [anon_sym_else] = ACTIONS(1974),
    [anon_sym_LBRACK] = ACTIONS(1972),
    [aux_sym_number_literal_token1] = ACTIONS(1974),
    [aux_sym_number_literal_token2] = ACTIONS(1972),
    [aux_sym_number_literal_token3] = ACTIONS(1972),
    [aux_sym_number_literal_token4] = ACTIONS(1974),
    [aux_sym_number_literal_token5] = ACTIONS(1972),
    [anon_sym_return] = ACTIONS(1974),
    [anon_sym_SQUOTE] = ACTIONS(1974),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1972),
    [anon_sym_DQUOTE] = ACTIONS(1974),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1972),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1972),
    [anon_sym_switch] = ACTIONS(1974),
    [anon_sym_case] = ACTIONS(1974),
    [anon_sym_default] = ACTIONS(1974),
    [anon_sym_try] = ACTIONS(1974),
    [anon_sym_catch] = ACTIONS(1974),
    [anon_sym_finally] = ACTIONS(1974),
    [anon_sym_int] = ACTIONS(1974),
    [anon_sym_boolean] = ACTIONS(1974),
    [anon_sym_char] = ACTIONS(1974),
    [anon_sym_short] = ACTIONS(1974),
    [anon_sym_long] = ACTIONS(1974),
    [anon_sym_float] = ACTIONS(1974),
    [anon_sym_double] = ACTIONS(1974),
    [anon_sym_void] = ACTIONS(1974),
    [anon_sym_public] = ACTIONS(1974),
    [anon_sym_protected] = ACTIONS(1974),
    [anon_sym_private] = ACTIONS(1974),
    [anon_sym_static] = ACTIONS(1974),
    [anon_sym_final] = ACTIONS(1974),
    [anon_sym_synchronized] = ACTIONS(1974),
    [anon_sym_TILDE] = ACTIONS(1972),
    [anon_sym_BANG] = ACTIONS(1972),
    [anon_sym_new] = ACTIONS(1974),
  },
  [STATE(887)] = {
    [sym_comment] = STATE(887),
    [sym_groovy_doc] = STATE(887),
    [sym_identifier] = ACTIONS(1991),
    [anon_sym_break] = ACTIONS(1991),
    [anon_sym_continue] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_STAR] = ACTIONS(1991),
    [anon_sym_STAR_COLON] = ACTIONS(1989),
    [anon_sym_import] = ACTIONS(1991),
    [anon_sym_package] = ACTIONS(1991),
    [anon_sym_AT] = ACTIONS(1991),
    [anon_sym_assert] = ACTIONS(1991),
    [anon_sym_PLUS_PLUS] = ACTIONS(1989),
    [anon_sym_DASH_DASH] = ACTIONS(1989),
    [anon_sym_SLASH] = ACTIONS(1991),
    [anon_sym_PLUS] = ACTIONS(1991),
    [anon_sym_DASH] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_ATinterface] = ACTIONS(1989),
    [anon_sym_interface] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1991),
    [anon_sym_def] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_null] = ACTIONS(1991),
    [anon_sym_this] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_while] = ACTIONS(1991),
    [anon_sym_for] = ACTIONS(1991),
    [aux_sym__type_identifier_token1] = ACTIONS(1991),
    [anon_sym_if] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1989),
    [aux_sym_number_literal_token1] = ACTIONS(1991),
    [aux_sym_number_literal_token2] = ACTIONS(1989),
    [aux_sym_number_literal_token3] = ACTIONS(1989),
    [aux_sym_number_literal_token4] = ACTIONS(1991),
    [aux_sym_number_literal_token5] = ACTIONS(1989),
    [anon_sym_return] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1991),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1991),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1989),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1989),
    [anon_sym_switch] = ACTIONS(1991),
    [anon_sym_case] = ACTIONS(1991),
    [anon_sym_default] = ACTIONS(1991),
    [anon_sym_try] = ACTIONS(1991),
    [anon_sym_catch] = ACTIONS(1991),
    [anon_sym_finally] = ACTIONS(1991),
    [anon_sym_int] = ACTIONS(1991),
    [anon_sym_boolean] = ACTIONS(1991),
    [anon_sym_char] = ACTIONS(1991),
    [anon_sym_short] = ACTIONS(1991),
    [anon_sym_long] = ACTIONS(1991),
    [anon_sym_float] = ACTIONS(1991),
    [anon_sym_double] = ACTIONS(1991),
    [anon_sym_void] = ACTIONS(1991),
    [anon_sym_public] = ACTIONS(1991),
    [anon_sym_protected] = ACTIONS(1991),
    [anon_sym_private] = ACTIONS(1991),
    [anon_sym_static] = ACTIONS(1991),
    [anon_sym_final] = ACTIONS(1991),
    [anon_sym_synchronized] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1989),
    [anon_sym_BANG] = ACTIONS(1989),
    [anon_sym_new] = ACTIONS(1991),
  },
  [STATE(888)] = {
    [sym_comment] = STATE(888),
    [sym_groovy_doc] = STATE(888),
    [sym_identifier] = ACTIONS(1995),
    [anon_sym_break] = ACTIONS(1995),
    [anon_sym_continue] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_STAR] = ACTIONS(1995),
    [anon_sym_STAR_COLON] = ACTIONS(1993),
    [anon_sym_import] = ACTIONS(1995),
    [anon_sym_package] = ACTIONS(1995),
    [anon_sym_AT] = ACTIONS(1995),
    [anon_sym_assert] = ACTIONS(1995),
    [anon_sym_PLUS_PLUS] = ACTIONS(1993),
    [anon_sym_DASH_DASH] = ACTIONS(1993),
    [anon_sym_SLASH] = ACTIONS(1995),
    [anon_sym_PLUS] = ACTIONS(1995),
    [anon_sym_DASH] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_ATinterface] = ACTIONS(1993),
    [anon_sym_interface] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1995),
    [anon_sym_def] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_null] = ACTIONS(1995),
    [anon_sym_this] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_while] = ACTIONS(1995),
    [anon_sym_for] = ACTIONS(1995),
    [aux_sym__type_identifier_token1] = ACTIONS(1995),
    [anon_sym_if] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1993),
    [aux_sym_number_literal_token1] = ACTIONS(1995),
    [aux_sym_number_literal_token2] = ACTIONS(1993),
    [aux_sym_number_literal_token3] = ACTIONS(1993),
    [aux_sym_number_literal_token4] = ACTIONS(1995),
    [aux_sym_number_literal_token5] = ACTIONS(1993),
    [anon_sym_return] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1995),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1995),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1993),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1993),
    [anon_sym_switch] = ACTIONS(1995),
    [anon_sym_case] = ACTIONS(1995),
    [anon_sym_default] = ACTIONS(1995),
    [anon_sym_try] = ACTIONS(1995),
    [anon_sym_catch] = ACTIONS(1995),
    [anon_sym_finally] = ACTIONS(1995),
    [anon_sym_int] = ACTIONS(1995),
    [anon_sym_boolean] = ACTIONS(1995),
    [anon_sym_char] = ACTIONS(1995),
    [anon_sym_short] = ACTIONS(1995),
    [anon_sym_long] = ACTIONS(1995),
    [anon_sym_float] = ACTIONS(1995),
    [anon_sym_double] = ACTIONS(1995),
    [anon_sym_void] = ACTIONS(1995),
    [anon_sym_public] = ACTIONS(1995),
    [anon_sym_protected] = ACTIONS(1995),
    [anon_sym_private] = ACTIONS(1995),
    [anon_sym_static] = ACTIONS(1995),
    [anon_sym_final] = ACTIONS(1995),
    [anon_sym_synchronized] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1993),
    [anon_sym_BANG] = ACTIONS(1993),
    [anon_sym_new] = ACTIONS(1995),
  },
  [STATE(889)] = {
    [sym_comment] = STATE(889),
    [sym_groovy_doc] = STATE(889),
    [sym_identifier] = ACTIONS(2005),
    [anon_sym_break] = ACTIONS(2005),
    [anon_sym_continue] = ACTIONS(2005),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_STAR] = ACTIONS(2005),
    [anon_sym_STAR_COLON] = ACTIONS(2003),
    [anon_sym_import] = ACTIONS(2005),
    [anon_sym_package] = ACTIONS(2005),
    [anon_sym_AT] = ACTIONS(2005),
    [anon_sym_assert] = ACTIONS(2005),
    [anon_sym_PLUS_PLUS] = ACTIONS(2003),
    [anon_sym_DASH_DASH] = ACTIONS(2003),
    [anon_sym_SLASH] = ACTIONS(2005),
    [anon_sym_PLUS] = ACTIONS(2005),
    [anon_sym_DASH] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2005),
    [anon_sym_false] = ACTIONS(2005),
    [anon_sym_ATinterface] = ACTIONS(2003),
    [anon_sym_interface] = ACTIONS(2005),
    [anon_sym_class] = ACTIONS(2005),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_RBRACE] = ACTIONS(2003),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2005),
    [anon_sym_def] = ACTIONS(2005),
    [anon_sym_LPAREN] = ACTIONS(2003),
    [anon_sym_null] = ACTIONS(2005),
    [anon_sym_this] = ACTIONS(2005),
    [anon_sym_do] = ACTIONS(2005),
    [anon_sym_while] = ACTIONS(2005),
    [anon_sym_for] = ACTIONS(2005),
    [aux_sym__type_identifier_token1] = ACTIONS(2005),
    [anon_sym_if] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [aux_sym_number_literal_token1] = ACTIONS(2005),
    [aux_sym_number_literal_token2] = ACTIONS(2003),
    [aux_sym_number_literal_token3] = ACTIONS(2003),
    [aux_sym_number_literal_token4] = ACTIONS(2005),
    [aux_sym_number_literal_token5] = ACTIONS(2003),
    [anon_sym_return] = ACTIONS(2005),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2003),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2003),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2003),
    [anon_sym_switch] = ACTIONS(2005),
    [anon_sym_case] = ACTIONS(2005),
    [anon_sym_default] = ACTIONS(2005),
    [anon_sym_try] = ACTIONS(2005),
    [anon_sym_catch] = ACTIONS(2005),
    [anon_sym_finally] = ACTIONS(2005),
    [anon_sym_int] = ACTIONS(2005),
    [anon_sym_boolean] = ACTIONS(2005),
    [anon_sym_char] = ACTIONS(2005),
    [anon_sym_short] = ACTIONS(2005),
    [anon_sym_long] = ACTIONS(2005),
    [anon_sym_float] = ACTIONS(2005),
    [anon_sym_double] = ACTIONS(2005),
    [anon_sym_void] = ACTIONS(2005),
    [anon_sym_public] = ACTIONS(2005),
    [anon_sym_protected] = ACTIONS(2005),
    [anon_sym_private] = ACTIONS(2005),
    [anon_sym_static] = ACTIONS(2005),
    [anon_sym_final] = ACTIONS(2005),
    [anon_sym_synchronized] = ACTIONS(2005),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_BANG] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2005),
  },
  [STATE(890)] = {
    [sym_comment] = STATE(890),
    [sym_groovy_doc] = STATE(890),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_catch] = ACTIONS(1513),
    [anon_sym_finally] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(891)] = {
    [sym_comment] = STATE(891),
    [sym_groovy_doc] = STATE(891),
    [sym_parameter_list] = STATE(930),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1890),
    [anon_sym_STAR_COLON] = ACTIONS(1892),
    [anon_sym_import] = ACTIONS(1890),
    [anon_sym_package] = ACTIONS(1890),
    [anon_sym_AT] = ACTIONS(1890),
    [anon_sym_assert] = ACTIONS(1890),
    [anon_sym_EQ] = ACTIONS(2108),
    [anon_sym_PLUS_PLUS] = ACTIONS(1892),
    [anon_sym_DASH_DASH] = ACTIONS(1892),
    [anon_sym_SLASH] = ACTIONS(1890),
    [anon_sym_PLUS] = ACTIONS(1890),
    [anon_sym_DASH] = ACTIONS(1890),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_ATinterface] = ACTIONS(1892),
    [anon_sym_interface] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1890),
    [anon_sym_def] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1890),
    [anon_sym_this] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [aux_sym__type_identifier_token1] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [aux_sym_number_literal_token1] = ACTIONS(1890),
    [aux_sym_number_literal_token2] = ACTIONS(1892),
    [aux_sym_number_literal_token3] = ACTIONS(1892),
    [aux_sym_number_literal_token4] = ACTIONS(1890),
    [aux_sym_number_literal_token5] = ACTIONS(1892),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1892),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1892),
    [anon_sym_switch] = ACTIONS(1890),
    [anon_sym_case] = ACTIONS(1890),
    [anon_sym_default] = ACTIONS(1890),
    [anon_sym_try] = ACTIONS(1890),
    [anon_sym_int] = ACTIONS(1890),
    [anon_sym_boolean] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_short] = ACTIONS(1890),
    [anon_sym_long] = ACTIONS(1890),
    [anon_sym_float] = ACTIONS(1890),
    [anon_sym_double] = ACTIONS(1890),
    [anon_sym_void] = ACTIONS(1890),
    [anon_sym_public] = ACTIONS(1890),
    [anon_sym_protected] = ACTIONS(1890),
    [anon_sym_private] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_final] = ACTIONS(1890),
    [anon_sym_synchronized] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_new] = ACTIONS(1890),
  },
  [STATE(892)] = {
    [sym_comment] = STATE(892),
    [sym_groovy_doc] = STATE(892),
    [ts_builtin_sym_end] = ACTIONS(1747),
    [sym_identifier] = ACTIONS(1749),
    [anon_sym_break] = ACTIONS(1749),
    [anon_sym_continue] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_STAR] = ACTIONS(1749),
    [anon_sym_STAR_COLON] = ACTIONS(1747),
    [anon_sym_import] = ACTIONS(1749),
    [anon_sym_package] = ACTIONS(1749),
    [anon_sym_AT] = ACTIONS(1749),
    [anon_sym_assert] = ACTIONS(1749),
    [anon_sym_PLUS_PLUS] = ACTIONS(1747),
    [anon_sym_DASH_DASH] = ACTIONS(1747),
    [anon_sym_SLASH] = ACTIONS(1749),
    [anon_sym_PLUS] = ACTIONS(1749),
    [anon_sym_DASH] = ACTIONS(1749),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_ATinterface] = ACTIONS(1747),
    [anon_sym_interface] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_LBRACE] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1749),
    [anon_sym_def] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_null] = ACTIONS(1749),
    [anon_sym_this] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_while] = ACTIONS(1749),
    [anon_sym_for] = ACTIONS(1749),
    [aux_sym__type_identifier_token1] = ACTIONS(1749),
    [anon_sym_if] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1747),
    [aux_sym_number_literal_token1] = ACTIONS(1749),
    [aux_sym_number_literal_token2] = ACTIONS(1747),
    [aux_sym_number_literal_token3] = ACTIONS(1747),
    [aux_sym_number_literal_token4] = ACTIONS(1749),
    [aux_sym_number_literal_token5] = ACTIONS(1747),
    [anon_sym_pipeline] = ACTIONS(1749),
    [anon_sym_return] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1749),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1749),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1747),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1747),
    [anon_sym_switch] = ACTIONS(1749),
    [anon_sym_try] = ACTIONS(1749),
    [anon_sym_int] = ACTIONS(1749),
    [anon_sym_boolean] = ACTIONS(1749),
    [anon_sym_char] = ACTIONS(1749),
    [anon_sym_short] = ACTIONS(1749),
    [anon_sym_long] = ACTIONS(1749),
    [anon_sym_float] = ACTIONS(1749),
    [anon_sym_double] = ACTIONS(1749),
    [anon_sym_void] = ACTIONS(1749),
    [anon_sym_public] = ACTIONS(1749),
    [anon_sym_protected] = ACTIONS(1749),
    [anon_sym_private] = ACTIONS(1749),
    [anon_sym_static] = ACTIONS(1749),
    [anon_sym_final] = ACTIONS(1749),
    [anon_sym_synchronized] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1747),
    [anon_sym_BANG] = ACTIONS(1747),
    [anon_sym_new] = ACTIONS(1749),
  },
  [STATE(893)] = {
    [sym_comment] = STATE(893),
    [sym_groovy_doc] = STATE(893),
    [ts_builtin_sym_end] = ACTIONS(2110),
    [sym_identifier] = ACTIONS(2112),
    [anon_sym_break] = ACTIONS(2112),
    [anon_sym_continue] = ACTIONS(2112),
    [anon_sym_SEMI] = ACTIONS(2110),
    [anon_sym_STAR] = ACTIONS(2112),
    [anon_sym_STAR_COLON] = ACTIONS(2110),
    [anon_sym_import] = ACTIONS(2112),
    [anon_sym_package] = ACTIONS(2112),
    [anon_sym_AT] = ACTIONS(2112),
    [anon_sym_assert] = ACTIONS(2112),
    [anon_sym_PLUS_PLUS] = ACTIONS(2110),
    [anon_sym_DASH_DASH] = ACTIONS(2110),
    [anon_sym_SLASH] = ACTIONS(2112),
    [anon_sym_PLUS] = ACTIONS(2112),
    [anon_sym_DASH] = ACTIONS(2112),
    [anon_sym_true] = ACTIONS(2112),
    [anon_sym_false] = ACTIONS(2112),
    [anon_sym_ATinterface] = ACTIONS(2110),
    [anon_sym_interface] = ACTIONS(2112),
    [anon_sym_class] = ACTIONS(2112),
    [anon_sym_COMMA] = ACTIONS(2110),
    [anon_sym_LBRACE] = ACTIONS(2110),
    [anon_sym_RBRACE] = ACTIONS(2110),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2112),
    [anon_sym_def] = ACTIONS(2112),
    [anon_sym_LPAREN] = ACTIONS(2110),
    [anon_sym_RPAREN] = ACTIONS(2110),
    [anon_sym_null] = ACTIONS(2112),
    [anon_sym_this] = ACTIONS(2112),
    [anon_sym_do] = ACTIONS(2112),
    [anon_sym_while] = ACTIONS(2112),
    [anon_sym_for] = ACTIONS(2112),
    [aux_sym__type_identifier_token1] = ACTIONS(2112),
    [anon_sym_if] = ACTIONS(2112),
    [anon_sym_else] = ACTIONS(2112),
    [anon_sym_LBRACK] = ACTIONS(2110),
    [aux_sym_number_literal_token1] = ACTIONS(2112),
    [aux_sym_number_literal_token2] = ACTIONS(2110),
    [aux_sym_number_literal_token3] = ACTIONS(2110),
    [aux_sym_number_literal_token4] = ACTIONS(2112),
    [aux_sym_number_literal_token5] = ACTIONS(2110),
    [anon_sym_pipeline] = ACTIONS(2112),
    [anon_sym_return] = ACTIONS(2112),
    [anon_sym_SQUOTE] = ACTIONS(2112),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2110),
    [anon_sym_DQUOTE] = ACTIONS(2112),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2110),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2110),
    [anon_sym_switch] = ACTIONS(2112),
    [anon_sym_try] = ACTIONS(2112),
    [anon_sym_int] = ACTIONS(2112),
    [anon_sym_boolean] = ACTIONS(2112),
    [anon_sym_char] = ACTIONS(2112),
    [anon_sym_short] = ACTIONS(2112),
    [anon_sym_long] = ACTIONS(2112),
    [anon_sym_float] = ACTIONS(2112),
    [anon_sym_double] = ACTIONS(2112),
    [anon_sym_void] = ACTIONS(2112),
    [anon_sym_public] = ACTIONS(2112),
    [anon_sym_protected] = ACTIONS(2112),
    [anon_sym_private] = ACTIONS(2112),
    [anon_sym_static] = ACTIONS(2112),
    [anon_sym_final] = ACTIONS(2112),
    [anon_sym_synchronized] = ACTIONS(2112),
    [anon_sym_TILDE] = ACTIONS(2110),
    [anon_sym_BANG] = ACTIONS(2110),
    [anon_sym_new] = ACTIONS(2112),
  },
  [STATE(894)] = {
    [sym_comment] = STATE(894),
    [sym_groovy_doc] = STATE(894),
    [sym_identifier] = ACTIONS(1810),
    [anon_sym_break] = ACTIONS(1810),
    [anon_sym_continue] = ACTIONS(1810),
    [anon_sym_SEMI] = ACTIONS(1812),
    [anon_sym_STAR] = ACTIONS(1810),
    [anon_sym_STAR_COLON] = ACTIONS(1812),
    [anon_sym_DOT] = ACTIONS(2114),
    [anon_sym_import] = ACTIONS(1810),
    [anon_sym_as] = ACTIONS(2116),
    [anon_sym_package] = ACTIONS(1810),
    [anon_sym_AT] = ACTIONS(1810),
    [anon_sym_assert] = ACTIONS(1810),
    [anon_sym_PLUS_PLUS] = ACTIONS(1812),
    [anon_sym_DASH_DASH] = ACTIONS(1812),
    [anon_sym_SLASH] = ACTIONS(1810),
    [anon_sym_PLUS] = ACTIONS(1810),
    [anon_sym_DASH] = ACTIONS(1810),
    [anon_sym_true] = ACTIONS(1810),
    [anon_sym_false] = ACTIONS(1810),
    [anon_sym_ATinterface] = ACTIONS(1812),
    [anon_sym_interface] = ACTIONS(1810),
    [anon_sym_class] = ACTIONS(1810),
    [anon_sym_LBRACE] = ACTIONS(1812),
    [anon_sym_RBRACE] = ACTIONS(1812),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1810),
    [anon_sym_def] = ACTIONS(1810),
    [anon_sym_LPAREN] = ACTIONS(1812),
    [anon_sym_null] = ACTIONS(1810),
    [anon_sym_this] = ACTIONS(1810),
    [anon_sym_do] = ACTIONS(1810),
    [anon_sym_while] = ACTIONS(1810),
    [anon_sym_for] = ACTIONS(1810),
    [aux_sym__type_identifier_token1] = ACTIONS(1810),
    [anon_sym_if] = ACTIONS(1810),
    [anon_sym_else] = ACTIONS(1810),
    [anon_sym_LBRACK] = ACTIONS(1812),
    [aux_sym_number_literal_token1] = ACTIONS(1810),
    [aux_sym_number_literal_token2] = ACTIONS(1812),
    [aux_sym_number_literal_token3] = ACTIONS(1812),
    [aux_sym_number_literal_token4] = ACTIONS(1810),
    [aux_sym_number_literal_token5] = ACTIONS(1812),
    [anon_sym_return] = ACTIONS(1810),
    [anon_sym_SQUOTE] = ACTIONS(1810),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1812),
    [anon_sym_DQUOTE] = ACTIONS(1810),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1812),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1812),
    [anon_sym_switch] = ACTIONS(1810),
    [anon_sym_case] = ACTIONS(1810),
    [anon_sym_default] = ACTIONS(1810),
    [anon_sym_try] = ACTIONS(1810),
    [anon_sym_int] = ACTIONS(1810),
    [anon_sym_boolean] = ACTIONS(1810),
    [anon_sym_char] = ACTIONS(1810),
    [anon_sym_short] = ACTIONS(1810),
    [anon_sym_long] = ACTIONS(1810),
    [anon_sym_float] = ACTIONS(1810),
    [anon_sym_double] = ACTIONS(1810),
    [anon_sym_void] = ACTIONS(1810),
    [anon_sym_public] = ACTIONS(1810),
    [anon_sym_protected] = ACTIONS(1810),
    [anon_sym_private] = ACTIONS(1810),
    [anon_sym_static] = ACTIONS(1810),
    [anon_sym_final] = ACTIONS(1810),
    [anon_sym_synchronized] = ACTIONS(1810),
    [anon_sym_TILDE] = ACTIONS(1812),
    [anon_sym_BANG] = ACTIONS(1812),
    [anon_sym_new] = ACTIONS(1810),
  },
  [STATE(895)] = {
    [sym_comment] = STATE(895),
    [sym_groovy_doc] = STATE(895),
    [ts_builtin_sym_end] = ACTIONS(1985),
    [sym_identifier] = ACTIONS(1987),
    [anon_sym_break] = ACTIONS(1987),
    [anon_sym_continue] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_STAR] = ACTIONS(1987),
    [anon_sym_STAR_COLON] = ACTIONS(1985),
    [anon_sym_import] = ACTIONS(1987),
    [anon_sym_package] = ACTIONS(1987),
    [anon_sym_AT] = ACTIONS(1987),
    [anon_sym_assert] = ACTIONS(1987),
    [anon_sym_PLUS_PLUS] = ACTIONS(1985),
    [anon_sym_DASH_DASH] = ACTIONS(1985),
    [anon_sym_SLASH] = ACTIONS(1987),
    [anon_sym_PLUS] = ACTIONS(1987),
    [anon_sym_DASH] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_ATinterface] = ACTIONS(1985),
    [anon_sym_interface] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_COMMA] = ACTIONS(1985),
    [anon_sym_LBRACE] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1987),
    [anon_sym_def] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_RPAREN] = ACTIONS(1985),
    [anon_sym_null] = ACTIONS(1987),
    [anon_sym_this] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_while] = ACTIONS(1987),
    [anon_sym_for] = ACTIONS(1987),
    [aux_sym__type_identifier_token1] = ACTIONS(1987),
    [anon_sym_if] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1985),
    [aux_sym_number_literal_token1] = ACTIONS(1987),
    [aux_sym_number_literal_token2] = ACTIONS(1985),
    [aux_sym_number_literal_token3] = ACTIONS(1985),
    [aux_sym_number_literal_token4] = ACTIONS(1987),
    [aux_sym_number_literal_token5] = ACTIONS(1985),
    [anon_sym_pipeline] = ACTIONS(1987),
    [anon_sym_return] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1987),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1987),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1985),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1985),
    [anon_sym_switch] = ACTIONS(1987),
    [anon_sym_try] = ACTIONS(1987),
    [anon_sym_int] = ACTIONS(1987),
    [anon_sym_boolean] = ACTIONS(1987),
    [anon_sym_char] = ACTIONS(1987),
    [anon_sym_short] = ACTIONS(1987),
    [anon_sym_long] = ACTIONS(1987),
    [anon_sym_float] = ACTIONS(1987),
    [anon_sym_double] = ACTIONS(1987),
    [anon_sym_void] = ACTIONS(1987),
    [anon_sym_public] = ACTIONS(1987),
    [anon_sym_protected] = ACTIONS(1987),
    [anon_sym_private] = ACTIONS(1987),
    [anon_sym_static] = ACTIONS(1987),
    [anon_sym_final] = ACTIONS(1987),
    [anon_sym_synchronized] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1985),
    [anon_sym_BANG] = ACTIONS(1985),
    [anon_sym_new] = ACTIONS(1987),
  },
  [STATE(896)] = {
    [sym_comment] = STATE(896),
    [sym_groovy_doc] = STATE(896),
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_pipeline] = ACTIONS(1362),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_catch] = ACTIONS(1362),
    [anon_sym_finally] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(897)] = {
    [sym_comment] = STATE(897),
    [sym_groovy_doc] = STATE(897),
    [ts_builtin_sym_end] = ACTIONS(1904),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_COMMA] = ACTIONS(1904),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_RPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_pipeline] = ACTIONS(1906),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(898)] = {
    [sym_comment] = STATE(898),
    [sym_groovy_doc] = STATE(898),
    [aux_sym__juxt_argument_list_repeat1] = STATE(795),
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_STAR] = ACTIONS(1254),
    [anon_sym_STAR_COLON] = ACTIONS(1252),
    [anon_sym_import] = ACTIONS(1254),
    [anon_sym_package] = ACTIONS(1254),
    [anon_sym_AT] = ACTIONS(1254),
    [anon_sym_assert] = ACTIONS(1254),
    [anon_sym_PLUS_PLUS] = ACTIONS(1252),
    [anon_sym_DASH_DASH] = ACTIONS(1252),
    [anon_sym_SLASH] = ACTIONS(1254),
    [anon_sym_PLUS] = ACTIONS(1254),
    [anon_sym_DASH] = ACTIONS(1254),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [anon_sym_ATinterface] = ACTIONS(1252),
    [anon_sym_interface] = ACTIONS(1254),
    [anon_sym_class] = ACTIONS(1254),
    [anon_sym_COMMA] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1254),
    [anon_sym_def] = ACTIONS(1254),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_null] = ACTIONS(1254),
    [anon_sym_this] = ACTIONS(1254),
    [anon_sym_do] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [aux_sym__type_identifier_token1] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_else] = ACTIONS(1254),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [aux_sym_number_literal_token1] = ACTIONS(1254),
    [aux_sym_number_literal_token2] = ACTIONS(1252),
    [aux_sym_number_literal_token3] = ACTIONS(1252),
    [aux_sym_number_literal_token4] = ACTIONS(1254),
    [aux_sym_number_literal_token5] = ACTIONS(1252),
    [anon_sym_pipeline] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1252),
    [anon_sym_DQUOTE] = ACTIONS(1254),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1252),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1252),
    [anon_sym_switch] = ACTIONS(1254),
    [anon_sym_try] = ACTIONS(1254),
    [anon_sym_int] = ACTIONS(1254),
    [anon_sym_boolean] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_short] = ACTIONS(1254),
    [anon_sym_long] = ACTIONS(1254),
    [anon_sym_float] = ACTIONS(1254),
    [anon_sym_double] = ACTIONS(1254),
    [anon_sym_void] = ACTIONS(1254),
    [anon_sym_public] = ACTIONS(1254),
    [anon_sym_protected] = ACTIONS(1254),
    [anon_sym_private] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_final] = ACTIONS(1254),
    [anon_sym_synchronized] = ACTIONS(1254),
    [anon_sym_TILDE] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_new] = ACTIONS(1254),
  },
  [STATE(899)] = {
    [sym_comment] = STATE(899),
    [sym_groovy_doc] = STATE(899),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_catch] = ACTIONS(1517),
    [anon_sym_finally] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(900)] = {
    [sym_comment] = STATE(900),
    [sym_groovy_doc] = STATE(900),
    [sym_parameter_list] = STATE(931),
    [ts_builtin_sym_end] = ACTIONS(1892),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1890),
    [anon_sym_STAR_COLON] = ACTIONS(1892),
    [anon_sym_import] = ACTIONS(1890),
    [anon_sym_package] = ACTIONS(1890),
    [anon_sym_AT] = ACTIONS(1890),
    [anon_sym_assert] = ACTIONS(1890),
    [anon_sym_EQ] = ACTIONS(2118),
    [anon_sym_PLUS_PLUS] = ACTIONS(1892),
    [anon_sym_DASH_DASH] = ACTIONS(1892),
    [anon_sym_SLASH] = ACTIONS(1890),
    [anon_sym_PLUS] = ACTIONS(1890),
    [anon_sym_DASH] = ACTIONS(1890),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_ATinterface] = ACTIONS(1892),
    [anon_sym_interface] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1890),
    [anon_sym_def] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1890),
    [anon_sym_this] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [aux_sym__type_identifier_token1] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_else] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [aux_sym_number_literal_token1] = ACTIONS(1890),
    [aux_sym_number_literal_token2] = ACTIONS(1892),
    [aux_sym_number_literal_token3] = ACTIONS(1892),
    [aux_sym_number_literal_token4] = ACTIONS(1890),
    [aux_sym_number_literal_token5] = ACTIONS(1892),
    [anon_sym_pipeline] = ACTIONS(1890),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1892),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1892),
    [anon_sym_switch] = ACTIONS(1890),
    [anon_sym_try] = ACTIONS(1890),
    [anon_sym_int] = ACTIONS(1890),
    [anon_sym_boolean] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_short] = ACTIONS(1890),
    [anon_sym_long] = ACTIONS(1890),
    [anon_sym_float] = ACTIONS(1890),
    [anon_sym_double] = ACTIONS(1890),
    [anon_sym_void] = ACTIONS(1890),
    [anon_sym_public] = ACTIONS(1890),
    [anon_sym_protected] = ACTIONS(1890),
    [anon_sym_private] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_final] = ACTIONS(1890),
    [anon_sym_synchronized] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_new] = ACTIONS(1890),
  },
  [STATE(901)] = {
    [sym_comment] = STATE(901),
    [sym_groovy_doc] = STATE(901),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_RPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(902)] = {
    [sym_comment] = STATE(902),
    [sym_groovy_doc] = STATE(902),
    [ts_builtin_sym_end] = ACTIONS(1999),
    [sym_identifier] = ACTIONS(2001),
    [anon_sym_break] = ACTIONS(2001),
    [anon_sym_continue] = ACTIONS(2001),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_STAR] = ACTIONS(2001),
    [anon_sym_STAR_COLON] = ACTIONS(1999),
    [anon_sym_import] = ACTIONS(2001),
    [anon_sym_package] = ACTIONS(2001),
    [anon_sym_AT] = ACTIONS(2001),
    [anon_sym_assert] = ACTIONS(2001),
    [anon_sym_PLUS_PLUS] = ACTIONS(1999),
    [anon_sym_DASH_DASH] = ACTIONS(1999),
    [anon_sym_SLASH] = ACTIONS(2001),
    [anon_sym_PLUS] = ACTIONS(2001),
    [anon_sym_DASH] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(2001),
    [anon_sym_false] = ACTIONS(2001),
    [anon_sym_ATinterface] = ACTIONS(1999),
    [anon_sym_interface] = ACTIONS(2001),
    [anon_sym_class] = ACTIONS(2001),
    [anon_sym_COMMA] = ACTIONS(1999),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_RBRACE] = ACTIONS(1999),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2001),
    [anon_sym_def] = ACTIONS(2001),
    [anon_sym_LPAREN] = ACTIONS(1999),
    [anon_sym_RPAREN] = ACTIONS(1999),
    [anon_sym_null] = ACTIONS(2001),
    [anon_sym_this] = ACTIONS(2001),
    [anon_sym_do] = ACTIONS(2001),
    [anon_sym_while] = ACTIONS(2001),
    [anon_sym_for] = ACTIONS(2001),
    [aux_sym__type_identifier_token1] = ACTIONS(2001),
    [anon_sym_if] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [aux_sym_number_literal_token1] = ACTIONS(2001),
    [aux_sym_number_literal_token2] = ACTIONS(1999),
    [aux_sym_number_literal_token3] = ACTIONS(1999),
    [aux_sym_number_literal_token4] = ACTIONS(2001),
    [aux_sym_number_literal_token5] = ACTIONS(1999),
    [anon_sym_pipeline] = ACTIONS(2001),
    [anon_sym_return] = ACTIONS(2001),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1999),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1999),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1999),
    [anon_sym_switch] = ACTIONS(2001),
    [anon_sym_try] = ACTIONS(2001),
    [anon_sym_int] = ACTIONS(2001),
    [anon_sym_boolean] = ACTIONS(2001),
    [anon_sym_char] = ACTIONS(2001),
    [anon_sym_short] = ACTIONS(2001),
    [anon_sym_long] = ACTIONS(2001),
    [anon_sym_float] = ACTIONS(2001),
    [anon_sym_double] = ACTIONS(2001),
    [anon_sym_void] = ACTIONS(2001),
    [anon_sym_public] = ACTIONS(2001),
    [anon_sym_protected] = ACTIONS(2001),
    [anon_sym_private] = ACTIONS(2001),
    [anon_sym_static] = ACTIONS(2001),
    [anon_sym_final] = ACTIONS(2001),
    [anon_sym_synchronized] = ACTIONS(2001),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_BANG] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(2001),
  },
  [STATE(903)] = {
    [sym_comment] = STATE(903),
    [sym_groovy_doc] = STATE(903),
    [aux_sym__juxt_argument_list_repeat1] = STATE(903),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(2120),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_pipeline] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(904)] = {
    [sym_comment] = STATE(904),
    [sym_groovy_doc] = STATE(904),
    [ts_builtin_sym_end] = ACTIONS(1956),
    [sym_identifier] = ACTIONS(1958),
    [anon_sym_break] = ACTIONS(1958),
    [anon_sym_continue] = ACTIONS(1958),
    [anon_sym_SEMI] = ACTIONS(1956),
    [anon_sym_STAR] = ACTIONS(1958),
    [anon_sym_STAR_COLON] = ACTIONS(1956),
    [anon_sym_import] = ACTIONS(1958),
    [anon_sym_package] = ACTIONS(1958),
    [anon_sym_AT] = ACTIONS(1958),
    [anon_sym_assert] = ACTIONS(1958),
    [anon_sym_PLUS_PLUS] = ACTIONS(1956),
    [anon_sym_DASH_DASH] = ACTIONS(1956),
    [anon_sym_SLASH] = ACTIONS(1958),
    [anon_sym_PLUS] = ACTIONS(1958),
    [anon_sym_DASH] = ACTIONS(1958),
    [anon_sym_true] = ACTIONS(1958),
    [anon_sym_false] = ACTIONS(1958),
    [anon_sym_ATinterface] = ACTIONS(1956),
    [anon_sym_interface] = ACTIONS(1958),
    [anon_sym_class] = ACTIONS(1958),
    [anon_sym_COMMA] = ACTIONS(1956),
    [anon_sym_LBRACE] = ACTIONS(1956),
    [anon_sym_RBRACE] = ACTIONS(1956),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1958),
    [anon_sym_def] = ACTIONS(1958),
    [anon_sym_LPAREN] = ACTIONS(1956),
    [anon_sym_RPAREN] = ACTIONS(1956),
    [anon_sym_null] = ACTIONS(1958),
    [anon_sym_this] = ACTIONS(1958),
    [anon_sym_do] = ACTIONS(1958),
    [anon_sym_while] = ACTIONS(1958),
    [anon_sym_for] = ACTIONS(1958),
    [aux_sym__type_identifier_token1] = ACTIONS(1958),
    [anon_sym_if] = ACTIONS(1958),
    [anon_sym_else] = ACTIONS(1958),
    [anon_sym_LBRACK] = ACTIONS(1956),
    [aux_sym_number_literal_token1] = ACTIONS(1958),
    [aux_sym_number_literal_token2] = ACTIONS(1956),
    [aux_sym_number_literal_token3] = ACTIONS(1956),
    [aux_sym_number_literal_token4] = ACTIONS(1958),
    [aux_sym_number_literal_token5] = ACTIONS(1956),
    [anon_sym_pipeline] = ACTIONS(1958),
    [anon_sym_return] = ACTIONS(1958),
    [anon_sym_SQUOTE] = ACTIONS(1958),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1956),
    [anon_sym_DQUOTE] = ACTIONS(1958),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1956),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1956),
    [anon_sym_switch] = ACTIONS(1958),
    [anon_sym_try] = ACTIONS(1958),
    [anon_sym_int] = ACTIONS(1958),
    [anon_sym_boolean] = ACTIONS(1958),
    [anon_sym_char] = ACTIONS(1958),
    [anon_sym_short] = ACTIONS(1958),
    [anon_sym_long] = ACTIONS(1958),
    [anon_sym_float] = ACTIONS(1958),
    [anon_sym_double] = ACTIONS(1958),
    [anon_sym_void] = ACTIONS(1958),
    [anon_sym_public] = ACTIONS(1958),
    [anon_sym_protected] = ACTIONS(1958),
    [anon_sym_private] = ACTIONS(1958),
    [anon_sym_static] = ACTIONS(1958),
    [anon_sym_final] = ACTIONS(1958),
    [anon_sym_synchronized] = ACTIONS(1958),
    [anon_sym_TILDE] = ACTIONS(1956),
    [anon_sym_BANG] = ACTIONS(1956),
    [anon_sym_new] = ACTIONS(1958),
  },
  [STATE(905)] = {
    [sym_comment] = STATE(905),
    [sym_groovy_doc] = STATE(905),
    [sym_parameter_list] = STATE(916),
    [ts_builtin_sym_end] = ACTIONS(1979),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2123),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_pipeline] = ACTIONS(1981),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(906)] = {
    [sym_comment] = STATE(906),
    [sym_groovy_doc] = STATE(906),
    [sym_parameter_list] = STATE(916),
    [ts_builtin_sym_end] = ACTIONS(2007),
    [sym_identifier] = ACTIONS(2009),
    [anon_sym_break] = ACTIONS(2009),
    [anon_sym_continue] = ACTIONS(2009),
    [anon_sym_SEMI] = ACTIONS(2007),
    [anon_sym_STAR] = ACTIONS(2009),
    [anon_sym_STAR_COLON] = ACTIONS(2007),
    [anon_sym_import] = ACTIONS(2009),
    [anon_sym_package] = ACTIONS(2009),
    [anon_sym_AT] = ACTIONS(2009),
    [anon_sym_assert] = ACTIONS(2009),
    [anon_sym_EQ] = ACTIONS(2125),
    [anon_sym_PLUS_PLUS] = ACTIONS(2007),
    [anon_sym_DASH_DASH] = ACTIONS(2007),
    [anon_sym_SLASH] = ACTIONS(2009),
    [anon_sym_PLUS] = ACTIONS(2009),
    [anon_sym_DASH] = ACTIONS(2009),
    [anon_sym_true] = ACTIONS(2009),
    [anon_sym_false] = ACTIONS(2009),
    [anon_sym_ATinterface] = ACTIONS(2007),
    [anon_sym_interface] = ACTIONS(2009),
    [anon_sym_class] = ACTIONS(2009),
    [anon_sym_LBRACE] = ACTIONS(2007),
    [anon_sym_RBRACE] = ACTIONS(2007),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2009),
    [anon_sym_def] = ACTIONS(2009),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(2009),
    [anon_sym_this] = ACTIONS(2009),
    [anon_sym_do] = ACTIONS(2009),
    [anon_sym_while] = ACTIONS(2009),
    [anon_sym_for] = ACTIONS(2009),
    [aux_sym__type_identifier_token1] = ACTIONS(2009),
    [anon_sym_if] = ACTIONS(2009),
    [anon_sym_else] = ACTIONS(2009),
    [anon_sym_LBRACK] = ACTIONS(2007),
    [aux_sym_number_literal_token1] = ACTIONS(2009),
    [aux_sym_number_literal_token2] = ACTIONS(2007),
    [aux_sym_number_literal_token3] = ACTIONS(2007),
    [aux_sym_number_literal_token4] = ACTIONS(2009),
    [aux_sym_number_literal_token5] = ACTIONS(2007),
    [anon_sym_pipeline] = ACTIONS(2009),
    [anon_sym_return] = ACTIONS(2009),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2007),
    [anon_sym_DQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2007),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2007),
    [anon_sym_switch] = ACTIONS(2009),
    [anon_sym_try] = ACTIONS(2009),
    [anon_sym_int] = ACTIONS(2009),
    [anon_sym_boolean] = ACTIONS(2009),
    [anon_sym_char] = ACTIONS(2009),
    [anon_sym_short] = ACTIONS(2009),
    [anon_sym_long] = ACTIONS(2009),
    [anon_sym_float] = ACTIONS(2009),
    [anon_sym_double] = ACTIONS(2009),
    [anon_sym_void] = ACTIONS(2009),
    [anon_sym_public] = ACTIONS(2009),
    [anon_sym_protected] = ACTIONS(2009),
    [anon_sym_private] = ACTIONS(2009),
    [anon_sym_static] = ACTIONS(2009),
    [anon_sym_final] = ACTIONS(2009),
    [anon_sym_synchronized] = ACTIONS(2009),
    [anon_sym_TILDE] = ACTIONS(2007),
    [anon_sym_BANG] = ACTIONS(2007),
    [anon_sym_new] = ACTIONS(2009),
  },
  [STATE(907)] = {
    [sym_comment] = STATE(907),
    [sym_groovy_doc] = STATE(907),
    [sym_identifier] = ACTIONS(1870),
    [anon_sym_break] = ACTIONS(1870),
    [anon_sym_continue] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_STAR] = ACTIONS(1870),
    [anon_sym_STAR_COLON] = ACTIONS(1868),
    [anon_sym_DOT] = ACTIONS(2127),
    [anon_sym_import] = ACTIONS(1870),
    [anon_sym_as] = ACTIONS(2129),
    [anon_sym_package] = ACTIONS(1870),
    [anon_sym_AT] = ACTIONS(1870),
    [anon_sym_assert] = ACTIONS(1870),
    [anon_sym_PLUS_PLUS] = ACTIONS(1868),
    [anon_sym_DASH_DASH] = ACTIONS(1868),
    [anon_sym_SLASH] = ACTIONS(1870),
    [anon_sym_PLUS] = ACTIONS(1870),
    [anon_sym_DASH] = ACTIONS(1870),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [anon_sym_ATinterface] = ACTIONS(1868),
    [anon_sym_interface] = ACTIONS(1870),
    [anon_sym_class] = ACTIONS(1870),
    [anon_sym_LBRACE] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1870),
    [anon_sym_def] = ACTIONS(1870),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_null] = ACTIONS(1870),
    [anon_sym_this] = ACTIONS(1870),
    [anon_sym_do] = ACTIONS(1870),
    [anon_sym_while] = ACTIONS(1870),
    [anon_sym_for] = ACTIONS(1870),
    [aux_sym__type_identifier_token1] = ACTIONS(1870),
    [anon_sym_if] = ACTIONS(1870),
    [anon_sym_else] = ACTIONS(1870),
    [anon_sym_LBRACK] = ACTIONS(1868),
    [aux_sym_number_literal_token1] = ACTIONS(1870),
    [aux_sym_number_literal_token2] = ACTIONS(1868),
    [aux_sym_number_literal_token3] = ACTIONS(1868),
    [aux_sym_number_literal_token4] = ACTIONS(1870),
    [aux_sym_number_literal_token5] = ACTIONS(1868),
    [anon_sym_return] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1870),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1868),
    [anon_sym_DQUOTE] = ACTIONS(1870),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1868),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1868),
    [anon_sym_switch] = ACTIONS(1870),
    [anon_sym_case] = ACTIONS(1870),
    [anon_sym_default] = ACTIONS(1870),
    [anon_sym_try] = ACTIONS(1870),
    [anon_sym_int] = ACTIONS(1870),
    [anon_sym_boolean] = ACTIONS(1870),
    [anon_sym_char] = ACTIONS(1870),
    [anon_sym_short] = ACTIONS(1870),
    [anon_sym_long] = ACTIONS(1870),
    [anon_sym_float] = ACTIONS(1870),
    [anon_sym_double] = ACTIONS(1870),
    [anon_sym_void] = ACTIONS(1870),
    [anon_sym_public] = ACTIONS(1870),
    [anon_sym_protected] = ACTIONS(1870),
    [anon_sym_private] = ACTIONS(1870),
    [anon_sym_static] = ACTIONS(1870),
    [anon_sym_final] = ACTIONS(1870),
    [anon_sym_synchronized] = ACTIONS(1870),
    [anon_sym_TILDE] = ACTIONS(1868),
    [anon_sym_BANG] = ACTIONS(1868),
    [anon_sym_new] = ACTIONS(1870),
  },
  [STATE(908)] = {
    [sym_comment] = STATE(908),
    [sym_groovy_doc] = STATE(908),
    [ts_builtin_sym_end] = ACTIONS(2025),
    [sym_identifier] = ACTIONS(2027),
    [anon_sym_break] = ACTIONS(2027),
    [anon_sym_continue] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_STAR] = ACTIONS(2027),
    [anon_sym_STAR_COLON] = ACTIONS(2025),
    [anon_sym_import] = ACTIONS(2027),
    [anon_sym_package] = ACTIONS(2027),
    [anon_sym_AT] = ACTIONS(2027),
    [anon_sym_assert] = ACTIONS(2027),
    [anon_sym_PLUS_PLUS] = ACTIONS(2025),
    [anon_sym_DASH_DASH] = ACTIONS(2025),
    [anon_sym_SLASH] = ACTIONS(2027),
    [anon_sym_PLUS] = ACTIONS(2027),
    [anon_sym_DASH] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_ATinterface] = ACTIONS(2025),
    [anon_sym_interface] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_COMMA] = ACTIONS(2025),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2027),
    [anon_sym_def] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_RPAREN] = ACTIONS(2025),
    [anon_sym_null] = ACTIONS(2027),
    [anon_sym_this] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_while] = ACTIONS(2027),
    [anon_sym_for] = ACTIONS(2027),
    [aux_sym__type_identifier_token1] = ACTIONS(2027),
    [anon_sym_if] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [aux_sym_number_literal_token1] = ACTIONS(2027),
    [aux_sym_number_literal_token2] = ACTIONS(2025),
    [aux_sym_number_literal_token3] = ACTIONS(2025),
    [aux_sym_number_literal_token4] = ACTIONS(2027),
    [aux_sym_number_literal_token5] = ACTIONS(2025),
    [anon_sym_pipeline] = ACTIONS(2027),
    [anon_sym_return] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2025),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2025),
    [anon_sym_switch] = ACTIONS(2027),
    [anon_sym_try] = ACTIONS(2027),
    [anon_sym_int] = ACTIONS(2027),
    [anon_sym_boolean] = ACTIONS(2027),
    [anon_sym_char] = ACTIONS(2027),
    [anon_sym_short] = ACTIONS(2027),
    [anon_sym_long] = ACTIONS(2027),
    [anon_sym_float] = ACTIONS(2027),
    [anon_sym_double] = ACTIONS(2027),
    [anon_sym_void] = ACTIONS(2027),
    [anon_sym_public] = ACTIONS(2027),
    [anon_sym_protected] = ACTIONS(2027),
    [anon_sym_private] = ACTIONS(2027),
    [anon_sym_static] = ACTIONS(2027),
    [anon_sym_final] = ACTIONS(2027),
    [anon_sym_synchronized] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_BANG] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2027),
  },
  [STATE(909)] = {
    [sym_comment] = STATE(909),
    [sym_groovy_doc] = STATE(909),
    [ts_builtin_sym_end] = ACTIONS(1374),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_pipeline] = ACTIONS(1376),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_catch] = ACTIONS(1376),
    [anon_sym_finally] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(910)] = {
    [sym_comment] = STATE(910),
    [sym_groovy_doc] = STATE(910),
    [ts_builtin_sym_end] = ACTIONS(1818),
    [sym_identifier] = ACTIONS(1820),
    [anon_sym_break] = ACTIONS(1820),
    [anon_sym_continue] = ACTIONS(1820),
    [anon_sym_SEMI] = ACTIONS(1818),
    [anon_sym_STAR] = ACTIONS(1820),
    [anon_sym_STAR_COLON] = ACTIONS(1818),
    [anon_sym_import] = ACTIONS(1820),
    [anon_sym_package] = ACTIONS(1820),
    [anon_sym_AT] = ACTIONS(1820),
    [anon_sym_assert] = ACTIONS(1820),
    [anon_sym_PLUS_PLUS] = ACTIONS(1818),
    [anon_sym_DASH_DASH] = ACTIONS(1818),
    [anon_sym_SLASH] = ACTIONS(1820),
    [anon_sym_PLUS] = ACTIONS(1820),
    [anon_sym_DASH] = ACTIONS(1820),
    [anon_sym_true] = ACTIONS(1820),
    [anon_sym_false] = ACTIONS(1820),
    [anon_sym_ATinterface] = ACTIONS(1818),
    [anon_sym_interface] = ACTIONS(1820),
    [anon_sym_class] = ACTIONS(1820),
    [anon_sym_COMMA] = ACTIONS(1818),
    [anon_sym_LBRACE] = ACTIONS(1818),
    [anon_sym_RBRACE] = ACTIONS(1818),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1820),
    [anon_sym_def] = ACTIONS(1820),
    [anon_sym_LPAREN] = ACTIONS(1818),
    [anon_sym_RPAREN] = ACTIONS(1818),
    [anon_sym_null] = ACTIONS(1820),
    [anon_sym_this] = ACTIONS(1820),
    [anon_sym_do] = ACTIONS(1820),
    [anon_sym_while] = ACTIONS(1820),
    [anon_sym_for] = ACTIONS(1820),
    [aux_sym__type_identifier_token1] = ACTIONS(1820),
    [anon_sym_if] = ACTIONS(1820),
    [anon_sym_else] = ACTIONS(1820),
    [anon_sym_LBRACK] = ACTIONS(1818),
    [aux_sym_number_literal_token1] = ACTIONS(1820),
    [aux_sym_number_literal_token2] = ACTIONS(1818),
    [aux_sym_number_literal_token3] = ACTIONS(1818),
    [aux_sym_number_literal_token4] = ACTIONS(1820),
    [aux_sym_number_literal_token5] = ACTIONS(1818),
    [anon_sym_pipeline] = ACTIONS(1820),
    [anon_sym_return] = ACTIONS(1820),
    [anon_sym_SQUOTE] = ACTIONS(1820),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1818),
    [anon_sym_DQUOTE] = ACTIONS(1820),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1818),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1818),
    [anon_sym_switch] = ACTIONS(1820),
    [anon_sym_try] = ACTIONS(1820),
    [anon_sym_int] = ACTIONS(1820),
    [anon_sym_boolean] = ACTIONS(1820),
    [anon_sym_char] = ACTIONS(1820),
    [anon_sym_short] = ACTIONS(1820),
    [anon_sym_long] = ACTIONS(1820),
    [anon_sym_float] = ACTIONS(1820),
    [anon_sym_double] = ACTIONS(1820),
    [anon_sym_void] = ACTIONS(1820),
    [anon_sym_public] = ACTIONS(1820),
    [anon_sym_protected] = ACTIONS(1820),
    [anon_sym_private] = ACTIONS(1820),
    [anon_sym_static] = ACTIONS(1820),
    [anon_sym_final] = ACTIONS(1820),
    [anon_sym_synchronized] = ACTIONS(1820),
    [anon_sym_TILDE] = ACTIONS(1818),
    [anon_sym_BANG] = ACTIONS(1818),
    [anon_sym_new] = ACTIONS(1820),
  },
  [STATE(911)] = {
    [sym_comment] = STATE(911),
    [sym_groovy_doc] = STATE(911),
    [sym_parameter_list] = STATE(915),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(2131),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_case] = ACTIONS(1773),
    [anon_sym_default] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(912)] = {
    [sym_comment] = STATE(912),
    [sym_groovy_doc] = STATE(912),
    [sym_parameter_list] = STATE(915),
    [sym_identifier] = ACTIONS(1792),
    [anon_sym_break] = ACTIONS(1792),
    [anon_sym_continue] = ACTIONS(1792),
    [anon_sym_SEMI] = ACTIONS(1790),
    [anon_sym_STAR] = ACTIONS(1792),
    [anon_sym_STAR_COLON] = ACTIONS(1790),
    [anon_sym_import] = ACTIONS(1792),
    [anon_sym_package] = ACTIONS(1792),
    [anon_sym_AT] = ACTIONS(1792),
    [anon_sym_assert] = ACTIONS(1792),
    [anon_sym_EQ] = ACTIONS(2133),
    [anon_sym_PLUS_PLUS] = ACTIONS(1790),
    [anon_sym_DASH_DASH] = ACTIONS(1790),
    [anon_sym_SLASH] = ACTIONS(1792),
    [anon_sym_PLUS] = ACTIONS(1792),
    [anon_sym_DASH] = ACTIONS(1792),
    [anon_sym_true] = ACTIONS(1792),
    [anon_sym_false] = ACTIONS(1792),
    [anon_sym_ATinterface] = ACTIONS(1790),
    [anon_sym_interface] = ACTIONS(1792),
    [anon_sym_class] = ACTIONS(1792),
    [anon_sym_LBRACE] = ACTIONS(1790),
    [anon_sym_RBRACE] = ACTIONS(1790),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1792),
    [anon_sym_def] = ACTIONS(1792),
    [anon_sym_LPAREN] = ACTIONS(2069),
    [anon_sym_null] = ACTIONS(1792),
    [anon_sym_this] = ACTIONS(1792),
    [anon_sym_do] = ACTIONS(1792),
    [anon_sym_while] = ACTIONS(1792),
    [anon_sym_for] = ACTIONS(1792),
    [aux_sym__type_identifier_token1] = ACTIONS(1792),
    [anon_sym_if] = ACTIONS(1792),
    [anon_sym_else] = ACTIONS(1792),
    [anon_sym_LBRACK] = ACTIONS(1790),
    [aux_sym_number_literal_token1] = ACTIONS(1792),
    [aux_sym_number_literal_token2] = ACTIONS(1790),
    [aux_sym_number_literal_token3] = ACTIONS(1790),
    [aux_sym_number_literal_token4] = ACTIONS(1792),
    [aux_sym_number_literal_token5] = ACTIONS(1790),
    [anon_sym_return] = ACTIONS(1792),
    [anon_sym_SQUOTE] = ACTIONS(1792),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1790),
    [anon_sym_DQUOTE] = ACTIONS(1792),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1790),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1790),
    [anon_sym_switch] = ACTIONS(1792),
    [anon_sym_case] = ACTIONS(1792),
    [anon_sym_default] = ACTIONS(1792),
    [anon_sym_try] = ACTIONS(1792),
    [anon_sym_int] = ACTIONS(1792),
    [anon_sym_boolean] = ACTIONS(1792),
    [anon_sym_char] = ACTIONS(1792),
    [anon_sym_short] = ACTIONS(1792),
    [anon_sym_long] = ACTIONS(1792),
    [anon_sym_float] = ACTIONS(1792),
    [anon_sym_double] = ACTIONS(1792),
    [anon_sym_void] = ACTIONS(1792),
    [anon_sym_public] = ACTIONS(1792),
    [anon_sym_protected] = ACTIONS(1792),
    [anon_sym_private] = ACTIONS(1792),
    [anon_sym_static] = ACTIONS(1792),
    [anon_sym_final] = ACTIONS(1792),
    [anon_sym_synchronized] = ACTIONS(1792),
    [anon_sym_TILDE] = ACTIONS(1790),
    [anon_sym_BANG] = ACTIONS(1790),
    [anon_sym_new] = ACTIONS(1792),
  },
  [STATE(913)] = {
    [sym_comment] = STATE(913),
    [sym_groovy_doc] = STATE(913),
    [sym_identifier] = ACTIONS(387),
    [anon_sym_break] = ACTIONS(387),
    [anon_sym_continue] = ACTIONS(387),
    [anon_sym_SEMI] = ACTIONS(389),
    [anon_sym_STAR] = ACTIONS(387),
    [anon_sym_STAR_COLON] = ACTIONS(389),
    [anon_sym_DOT] = ACTIONS(389),
    [anon_sym_import] = ACTIONS(387),
    [anon_sym_as] = ACTIONS(387),
    [anon_sym_package] = ACTIONS(387),
    [anon_sym_AT] = ACTIONS(387),
    [anon_sym_assert] = ACTIONS(387),
    [anon_sym_PLUS_PLUS] = ACTIONS(389),
    [anon_sym_DASH_DASH] = ACTIONS(389),
    [anon_sym_SLASH] = ACTIONS(387),
    [anon_sym_PLUS] = ACTIONS(387),
    [anon_sym_DASH] = ACTIONS(387),
    [anon_sym_true] = ACTIONS(387),
    [anon_sym_false] = ACTIONS(387),
    [anon_sym_ATinterface] = ACTIONS(389),
    [anon_sym_interface] = ACTIONS(387),
    [anon_sym_class] = ACTIONS(387),
    [anon_sym_LBRACE] = ACTIONS(389),
    [anon_sym_RBRACE] = ACTIONS(389),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(387),
    [anon_sym_def] = ACTIONS(387),
    [anon_sym_LPAREN] = ACTIONS(389),
    [anon_sym_null] = ACTIONS(387),
    [anon_sym_this] = ACTIONS(387),
    [anon_sym_do] = ACTIONS(387),
    [anon_sym_while] = ACTIONS(387),
    [anon_sym_for] = ACTIONS(387),
    [aux_sym__type_identifier_token1] = ACTIONS(387),
    [anon_sym_if] = ACTIONS(387),
    [anon_sym_else] = ACTIONS(387),
    [anon_sym_LBRACK] = ACTIONS(389),
    [aux_sym_number_literal_token1] = ACTIONS(387),
    [aux_sym_number_literal_token2] = ACTIONS(389),
    [aux_sym_number_literal_token3] = ACTIONS(389),
    [aux_sym_number_literal_token4] = ACTIONS(387),
    [aux_sym_number_literal_token5] = ACTIONS(389),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_SQUOTE] = ACTIONS(387),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(389),
    [anon_sym_DQUOTE] = ACTIONS(387),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(389),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(389),
    [anon_sym_switch] = ACTIONS(387),
    [anon_sym_case] = ACTIONS(387),
    [anon_sym_default] = ACTIONS(387),
    [anon_sym_try] = ACTIONS(387),
    [anon_sym_int] = ACTIONS(387),
    [anon_sym_boolean] = ACTIONS(387),
    [anon_sym_char] = ACTIONS(387),
    [anon_sym_short] = ACTIONS(387),
    [anon_sym_long] = ACTIONS(387),
    [anon_sym_float] = ACTIONS(387),
    [anon_sym_double] = ACTIONS(387),
    [anon_sym_void] = ACTIONS(387),
    [anon_sym_public] = ACTIONS(387),
    [anon_sym_protected] = ACTIONS(387),
    [anon_sym_private] = ACTIONS(387),
    [anon_sym_static] = ACTIONS(387),
    [anon_sym_final] = ACTIONS(387),
    [anon_sym_synchronized] = ACTIONS(387),
    [anon_sym_TILDE] = ACTIONS(389),
    [anon_sym_BANG] = ACTIONS(389),
    [anon_sym_new] = ACTIONS(387),
  },
  [STATE(914)] = {
    [sym_comment] = STATE(914),
    [sym_groovy_doc] = STATE(914),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(2131),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_case] = ACTIONS(1773),
    [anon_sym_default] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(915)] = {
    [sym_closure] = STATE(953),
    [sym_comment] = STATE(915),
    [sym_groovy_doc] = STATE(915),
    [sym_identifier] = ACTIONS(2047),
    [anon_sym_break] = ACTIONS(2047),
    [anon_sym_continue] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_STAR] = ACTIONS(2047),
    [anon_sym_STAR_COLON] = ACTIONS(2049),
    [anon_sym_import] = ACTIONS(2047),
    [anon_sym_package] = ACTIONS(2047),
    [anon_sym_AT] = ACTIONS(2047),
    [anon_sym_assert] = ACTIONS(2047),
    [anon_sym_PLUS_PLUS] = ACTIONS(2049),
    [anon_sym_DASH_DASH] = ACTIONS(2049),
    [anon_sym_SLASH] = ACTIONS(2047),
    [anon_sym_PLUS] = ACTIONS(2047),
    [anon_sym_DASH] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_ATinterface] = ACTIONS(2049),
    [anon_sym_interface] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2047),
    [anon_sym_def] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_null] = ACTIONS(2047),
    [anon_sym_this] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_while] = ACTIONS(2047),
    [anon_sym_for] = ACTIONS(2047),
    [aux_sym__type_identifier_token1] = ACTIONS(2047),
    [anon_sym_if] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2049),
    [aux_sym_number_literal_token1] = ACTIONS(2047),
    [aux_sym_number_literal_token2] = ACTIONS(2049),
    [aux_sym_number_literal_token3] = ACTIONS(2049),
    [aux_sym_number_literal_token4] = ACTIONS(2047),
    [aux_sym_number_literal_token5] = ACTIONS(2049),
    [anon_sym_return] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2047),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2047),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2049),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2049),
    [anon_sym_switch] = ACTIONS(2047),
    [anon_sym_case] = ACTIONS(2047),
    [anon_sym_default] = ACTIONS(2047),
    [anon_sym_try] = ACTIONS(2047),
    [anon_sym_int] = ACTIONS(2047),
    [anon_sym_boolean] = ACTIONS(2047),
    [anon_sym_char] = ACTIONS(2047),
    [anon_sym_short] = ACTIONS(2047),
    [anon_sym_long] = ACTIONS(2047),
    [anon_sym_float] = ACTIONS(2047),
    [anon_sym_double] = ACTIONS(2047),
    [anon_sym_void] = ACTIONS(2047),
    [anon_sym_public] = ACTIONS(2047),
    [anon_sym_protected] = ACTIONS(2047),
    [anon_sym_private] = ACTIONS(2047),
    [anon_sym_static] = ACTIONS(2047),
    [anon_sym_final] = ACTIONS(2047),
    [anon_sym_synchronized] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2049),
    [anon_sym_BANG] = ACTIONS(2049),
    [anon_sym_new] = ACTIONS(2047),
  },
  [STATE(916)] = {
    [sym_closure] = STATE(823),
    [sym_comment] = STATE(916),
    [sym_groovy_doc] = STATE(916),
    [ts_builtin_sym_end] = ACTIONS(2057),
    [sym_identifier] = ACTIONS(2055),
    [anon_sym_break] = ACTIONS(2055),
    [anon_sym_continue] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_STAR] = ACTIONS(2055),
    [anon_sym_STAR_COLON] = ACTIONS(2057),
    [anon_sym_import] = ACTIONS(2055),
    [anon_sym_package] = ACTIONS(2055),
    [anon_sym_AT] = ACTIONS(2055),
    [anon_sym_assert] = ACTIONS(2055),
    [anon_sym_PLUS_PLUS] = ACTIONS(2057),
    [anon_sym_DASH_DASH] = ACTIONS(2057),
    [anon_sym_SLASH] = ACTIONS(2055),
    [anon_sym_PLUS] = ACTIONS(2055),
    [anon_sym_DASH] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_ATinterface] = ACTIONS(2057),
    [anon_sym_interface] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2055),
    [anon_sym_def] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_null] = ACTIONS(2055),
    [anon_sym_this] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_while] = ACTIONS(2055),
    [anon_sym_for] = ACTIONS(2055),
    [aux_sym__type_identifier_token1] = ACTIONS(2055),
    [anon_sym_if] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2057),
    [aux_sym_number_literal_token1] = ACTIONS(2055),
    [aux_sym_number_literal_token2] = ACTIONS(2057),
    [aux_sym_number_literal_token3] = ACTIONS(2057),
    [aux_sym_number_literal_token4] = ACTIONS(2055),
    [aux_sym_number_literal_token5] = ACTIONS(2057),
    [anon_sym_pipeline] = ACTIONS(2055),
    [anon_sym_return] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2055),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2057),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2057),
    [anon_sym_switch] = ACTIONS(2055),
    [anon_sym_try] = ACTIONS(2055),
    [anon_sym_int] = ACTIONS(2055),
    [anon_sym_boolean] = ACTIONS(2055),
    [anon_sym_char] = ACTIONS(2055),
    [anon_sym_short] = ACTIONS(2055),
    [anon_sym_long] = ACTIONS(2055),
    [anon_sym_float] = ACTIONS(2055),
    [anon_sym_double] = ACTIONS(2055),
    [anon_sym_void] = ACTIONS(2055),
    [anon_sym_public] = ACTIONS(2055),
    [anon_sym_protected] = ACTIONS(2055),
    [anon_sym_private] = ACTIONS(2055),
    [anon_sym_static] = ACTIONS(2055),
    [anon_sym_final] = ACTIONS(2055),
    [anon_sym_synchronized] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2057),
    [anon_sym_BANG] = ACTIONS(2057),
    [anon_sym_new] = ACTIONS(2055),
  },
  [STATE(917)] = {
    [sym_comment] = STATE(917),
    [sym_groovy_doc] = STATE(917),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2067),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1979),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_case] = ACTIONS(1981),
    [anon_sym_default] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(918)] = {
    [sym_comment] = STATE(918),
    [sym_groovy_doc] = STATE(918),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1690),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_COMMA] = ACTIONS(1515),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_RPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(919)] = {
    [sym_comment] = STATE(919),
    [sym_groovy_doc] = STATE(919),
    [sym_identifier] = ACTIONS(2033),
    [anon_sym_break] = ACTIONS(2033),
    [anon_sym_continue] = ACTIONS(2033),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_STAR] = ACTIONS(2033),
    [anon_sym_STAR_COLON] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2139),
    [anon_sym_import] = ACTIONS(2033),
    [anon_sym_package] = ACTIONS(2033),
    [anon_sym_AT] = ACTIONS(2033),
    [anon_sym_assert] = ACTIONS(2033),
    [anon_sym_PLUS_PLUS] = ACTIONS(2035),
    [anon_sym_DASH_DASH] = ACTIONS(2035),
    [anon_sym_SLASH] = ACTIONS(2033),
    [anon_sym_PLUS] = ACTIONS(2033),
    [anon_sym_DASH] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2033),
    [anon_sym_false] = ACTIONS(2033),
    [anon_sym_ATinterface] = ACTIONS(2035),
    [anon_sym_interface] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_RBRACE] = ACTIONS(2035),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2033),
    [anon_sym_def] = ACTIONS(2033),
    [anon_sym_LPAREN] = ACTIONS(2035),
    [anon_sym_null] = ACTIONS(2033),
    [anon_sym_this] = ACTIONS(2033),
    [anon_sym_do] = ACTIONS(2033),
    [anon_sym_while] = ACTIONS(2033),
    [anon_sym_for] = ACTIONS(2033),
    [aux_sym__type_identifier_token1] = ACTIONS(2033),
    [anon_sym_if] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [aux_sym_number_literal_token1] = ACTIONS(2033),
    [aux_sym_number_literal_token2] = ACTIONS(2035),
    [aux_sym_number_literal_token3] = ACTIONS(2035),
    [aux_sym_number_literal_token4] = ACTIONS(2033),
    [aux_sym_number_literal_token5] = ACTIONS(2035),
    [anon_sym_return] = ACTIONS(2033),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2035),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2035),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2035),
    [anon_sym_switch] = ACTIONS(2033),
    [anon_sym_case] = ACTIONS(2033),
    [anon_sym_default] = ACTIONS(2033),
    [anon_sym_try] = ACTIONS(2033),
    [anon_sym_int] = ACTIONS(2033),
    [anon_sym_boolean] = ACTIONS(2033),
    [anon_sym_char] = ACTIONS(2033),
    [anon_sym_short] = ACTIONS(2033),
    [anon_sym_long] = ACTIONS(2033),
    [anon_sym_float] = ACTIONS(2033),
    [anon_sym_double] = ACTIONS(2033),
    [anon_sym_void] = ACTIONS(2033),
    [anon_sym_public] = ACTIONS(2033),
    [anon_sym_protected] = ACTIONS(2033),
    [anon_sym_private] = ACTIONS(2033),
    [anon_sym_static] = ACTIONS(2033),
    [anon_sym_final] = ACTIONS(2033),
    [anon_sym_synchronized] = ACTIONS(2033),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_BANG] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2033),
  },
  [STATE(920)] = {
    [sym_comment] = STATE(920),
    [sym_groovy_doc] = STATE(920),
    [sym_parameter_list] = STATE(931),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(2141),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_DASH_GT] = ACTIONS(2143),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(921)] = {
    [sym_comment] = STATE(921),
    [sym_groovy_doc] = STATE(921),
    [ts_builtin_sym_end] = ACTIONS(1771),
    [sym_identifier] = ACTIONS(1773),
    [anon_sym_break] = ACTIONS(1773),
    [anon_sym_continue] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_STAR] = ACTIONS(1773),
    [anon_sym_STAR_COLON] = ACTIONS(1771),
    [anon_sym_import] = ACTIONS(1773),
    [anon_sym_package] = ACTIONS(1773),
    [anon_sym_AT] = ACTIONS(1773),
    [anon_sym_assert] = ACTIONS(1773),
    [anon_sym_EQ] = ACTIONS(2079),
    [anon_sym_PLUS_PLUS] = ACTIONS(1771),
    [anon_sym_DASH_DASH] = ACTIONS(1771),
    [anon_sym_SLASH] = ACTIONS(1773),
    [anon_sym_PLUS] = ACTIONS(1773),
    [anon_sym_DASH] = ACTIONS(1773),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_ATinterface] = ACTIONS(1771),
    [anon_sym_interface] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1773),
    [anon_sym_def] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_null] = ACTIONS(1773),
    [anon_sym_this] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_while] = ACTIONS(1773),
    [anon_sym_for] = ACTIONS(1773),
    [aux_sym__type_identifier_token1] = ACTIONS(1773),
    [anon_sym_if] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1771),
    [aux_sym_number_literal_token1] = ACTIONS(1773),
    [aux_sym_number_literal_token2] = ACTIONS(1771),
    [aux_sym_number_literal_token3] = ACTIONS(1771),
    [aux_sym_number_literal_token4] = ACTIONS(1773),
    [aux_sym_number_literal_token5] = ACTIONS(1771),
    [anon_sym_pipeline] = ACTIONS(1773),
    [anon_sym_return] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1773),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1773),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1771),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1771),
    [anon_sym_switch] = ACTIONS(1773),
    [anon_sym_try] = ACTIONS(1773),
    [anon_sym_int] = ACTIONS(1773),
    [anon_sym_boolean] = ACTIONS(1773),
    [anon_sym_char] = ACTIONS(1773),
    [anon_sym_short] = ACTIONS(1773),
    [anon_sym_long] = ACTIONS(1773),
    [anon_sym_float] = ACTIONS(1773),
    [anon_sym_double] = ACTIONS(1773),
    [anon_sym_void] = ACTIONS(1773),
    [anon_sym_public] = ACTIONS(1773),
    [anon_sym_protected] = ACTIONS(1773),
    [anon_sym_private] = ACTIONS(1773),
    [anon_sym_static] = ACTIONS(1773),
    [anon_sym_final] = ACTIONS(1773),
    [anon_sym_synchronized] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1771),
    [anon_sym_BANG] = ACTIONS(1771),
    [anon_sym_new] = ACTIONS(1773),
  },
  [STATE(922)] = {
    [sym_comment] = STATE(922),
    [sym_groovy_doc] = STATE(922),
    [ts_builtin_sym_end] = ACTIONS(2035),
    [sym_identifier] = ACTIONS(2033),
    [anon_sym_break] = ACTIONS(2033),
    [anon_sym_continue] = ACTIONS(2033),
    [anon_sym_SEMI] = ACTIONS(2035),
    [anon_sym_STAR] = ACTIONS(2033),
    [anon_sym_STAR_COLON] = ACTIONS(2035),
    [anon_sym_DOT] = ACTIONS(2145),
    [anon_sym_import] = ACTIONS(2033),
    [anon_sym_package] = ACTIONS(2033),
    [anon_sym_AT] = ACTIONS(2033),
    [anon_sym_assert] = ACTIONS(2033),
    [anon_sym_PLUS_PLUS] = ACTIONS(2035),
    [anon_sym_DASH_DASH] = ACTIONS(2035),
    [anon_sym_SLASH] = ACTIONS(2033),
    [anon_sym_PLUS] = ACTIONS(2033),
    [anon_sym_DASH] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(2033),
    [anon_sym_false] = ACTIONS(2033),
    [anon_sym_ATinterface] = ACTIONS(2035),
    [anon_sym_interface] = ACTIONS(2033),
    [anon_sym_class] = ACTIONS(2033),
    [anon_sym_LBRACE] = ACTIONS(2035),
    [anon_sym_RBRACE] = ACTIONS(2035),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2033),
    [anon_sym_def] = ACTIONS(2033),
    [anon_sym_LPAREN] = ACTIONS(2035),
    [anon_sym_null] = ACTIONS(2033),
    [anon_sym_this] = ACTIONS(2033),
    [anon_sym_do] = ACTIONS(2033),
    [anon_sym_while] = ACTIONS(2033),
    [anon_sym_for] = ACTIONS(2033),
    [aux_sym__type_identifier_token1] = ACTIONS(2033),
    [anon_sym_if] = ACTIONS(2033),
    [anon_sym_else] = ACTIONS(2033),
    [anon_sym_LBRACK] = ACTIONS(2035),
    [aux_sym_number_literal_token1] = ACTIONS(2033),
    [aux_sym_number_literal_token2] = ACTIONS(2035),
    [aux_sym_number_literal_token3] = ACTIONS(2035),
    [aux_sym_number_literal_token4] = ACTIONS(2033),
    [aux_sym_number_literal_token5] = ACTIONS(2035),
    [anon_sym_pipeline] = ACTIONS(2033),
    [anon_sym_return] = ACTIONS(2033),
    [anon_sym_SQUOTE] = ACTIONS(2033),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2035),
    [anon_sym_DQUOTE] = ACTIONS(2033),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2035),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2035),
    [anon_sym_switch] = ACTIONS(2033),
    [anon_sym_try] = ACTIONS(2033),
    [anon_sym_int] = ACTIONS(2033),
    [anon_sym_boolean] = ACTIONS(2033),
    [anon_sym_char] = ACTIONS(2033),
    [anon_sym_short] = ACTIONS(2033),
    [anon_sym_long] = ACTIONS(2033),
    [anon_sym_float] = ACTIONS(2033),
    [anon_sym_double] = ACTIONS(2033),
    [anon_sym_void] = ACTIONS(2033),
    [anon_sym_public] = ACTIONS(2033),
    [anon_sym_protected] = ACTIONS(2033),
    [anon_sym_private] = ACTIONS(2033),
    [anon_sym_static] = ACTIONS(2033),
    [anon_sym_final] = ACTIONS(2033),
    [anon_sym_synchronized] = ACTIONS(2033),
    [anon_sym_TILDE] = ACTIONS(2035),
    [anon_sym_BANG] = ACTIONS(2035),
    [anon_sym_new] = ACTIONS(2033),
  },
  [STATE(923)] = {
    [sym_closure] = STATE(968),
    [sym_comment] = STATE(923),
    [sym_groovy_doc] = STATE(923),
    [sym_identifier] = ACTIONS(2055),
    [anon_sym_break] = ACTIONS(2055),
    [anon_sym_continue] = ACTIONS(2055),
    [anon_sym_SEMI] = ACTIONS(2057),
    [anon_sym_STAR] = ACTIONS(2055),
    [anon_sym_STAR_COLON] = ACTIONS(2057),
    [anon_sym_import] = ACTIONS(2055),
    [anon_sym_package] = ACTIONS(2055),
    [anon_sym_AT] = ACTIONS(2055),
    [anon_sym_assert] = ACTIONS(2055),
    [anon_sym_PLUS_PLUS] = ACTIONS(2057),
    [anon_sym_DASH_DASH] = ACTIONS(2057),
    [anon_sym_SLASH] = ACTIONS(2055),
    [anon_sym_PLUS] = ACTIONS(2055),
    [anon_sym_DASH] = ACTIONS(2055),
    [anon_sym_true] = ACTIONS(2055),
    [anon_sym_false] = ACTIONS(2055),
    [anon_sym_ATinterface] = ACTIONS(2057),
    [anon_sym_interface] = ACTIONS(2055),
    [anon_sym_class] = ACTIONS(2055),
    [anon_sym_LBRACE] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2057),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2055),
    [anon_sym_def] = ACTIONS(2055),
    [anon_sym_LPAREN] = ACTIONS(2057),
    [anon_sym_null] = ACTIONS(2055),
    [anon_sym_this] = ACTIONS(2055),
    [anon_sym_do] = ACTIONS(2055),
    [anon_sym_while] = ACTIONS(2055),
    [anon_sym_for] = ACTIONS(2055),
    [aux_sym__type_identifier_token1] = ACTIONS(2055),
    [anon_sym_if] = ACTIONS(2055),
    [anon_sym_else] = ACTIONS(2055),
    [anon_sym_LBRACK] = ACTIONS(2057),
    [aux_sym_number_literal_token1] = ACTIONS(2055),
    [aux_sym_number_literal_token2] = ACTIONS(2057),
    [aux_sym_number_literal_token3] = ACTIONS(2057),
    [aux_sym_number_literal_token4] = ACTIONS(2055),
    [aux_sym_number_literal_token5] = ACTIONS(2057),
    [anon_sym_return] = ACTIONS(2055),
    [anon_sym_SQUOTE] = ACTIONS(2055),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2057),
    [anon_sym_DQUOTE] = ACTIONS(2055),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2057),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2057),
    [anon_sym_switch] = ACTIONS(2055),
    [anon_sym_case] = ACTIONS(2055),
    [anon_sym_default] = ACTIONS(2055),
    [anon_sym_try] = ACTIONS(2055),
    [anon_sym_int] = ACTIONS(2055),
    [anon_sym_boolean] = ACTIONS(2055),
    [anon_sym_char] = ACTIONS(2055),
    [anon_sym_short] = ACTIONS(2055),
    [anon_sym_long] = ACTIONS(2055),
    [anon_sym_float] = ACTIONS(2055),
    [anon_sym_double] = ACTIONS(2055),
    [anon_sym_void] = ACTIONS(2055),
    [anon_sym_public] = ACTIONS(2055),
    [anon_sym_protected] = ACTIONS(2055),
    [anon_sym_private] = ACTIONS(2055),
    [anon_sym_static] = ACTIONS(2055),
    [anon_sym_final] = ACTIONS(2055),
    [anon_sym_synchronized] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(2057),
    [anon_sym_BANG] = ACTIONS(2057),
    [anon_sym_new] = ACTIONS(2055),
  },
  [STATE(924)] = {
    [sym_comment] = STATE(924),
    [sym_groovy_doc] = STATE(924),
    [ts_builtin_sym_end] = ACTIONS(1741),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(2106),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1741),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_pipeline] = ACTIONS(1739),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(925)] = {
    [sym_comment] = STATE(925),
    [sym_groovy_doc] = STATE(925),
    [sym_parameter_list] = STATE(931),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1890),
    [anon_sym_STAR_COLON] = ACTIONS(1892),
    [anon_sym_import] = ACTIONS(1890),
    [anon_sym_package] = ACTIONS(1890),
    [anon_sym_AT] = ACTIONS(1890),
    [anon_sym_assert] = ACTIONS(1890),
    [anon_sym_EQ] = ACTIONS(2147),
    [anon_sym_PLUS_PLUS] = ACTIONS(1892),
    [anon_sym_DASH_DASH] = ACTIONS(1892),
    [anon_sym_SLASH] = ACTIONS(1890),
    [anon_sym_PLUS] = ACTIONS(1890),
    [anon_sym_DASH] = ACTIONS(1890),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [anon_sym_ATinterface] = ACTIONS(1892),
    [anon_sym_interface] = ACTIONS(1890),
    [anon_sym_class] = ACTIONS(1890),
    [anon_sym_COMMA] = ACTIONS(2143),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_DASH_GT] = ACTIONS(2143),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1890),
    [anon_sym_def] = ACTIONS(1890),
    [anon_sym_LPAREN] = ACTIONS(2081),
    [anon_sym_null] = ACTIONS(1890),
    [anon_sym_this] = ACTIONS(1890),
    [anon_sym_do] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [aux_sym__type_identifier_token1] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [aux_sym_number_literal_token1] = ACTIONS(1890),
    [aux_sym_number_literal_token2] = ACTIONS(1892),
    [aux_sym_number_literal_token3] = ACTIONS(1892),
    [aux_sym_number_literal_token4] = ACTIONS(1890),
    [aux_sym_number_literal_token5] = ACTIONS(1892),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1892),
    [anon_sym_DQUOTE] = ACTIONS(1890),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1892),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1892),
    [anon_sym_switch] = ACTIONS(1890),
    [anon_sym_try] = ACTIONS(1890),
    [anon_sym_int] = ACTIONS(1890),
    [anon_sym_boolean] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_short] = ACTIONS(1890),
    [anon_sym_long] = ACTIONS(1890),
    [anon_sym_float] = ACTIONS(1890),
    [anon_sym_double] = ACTIONS(1890),
    [anon_sym_void] = ACTIONS(1890),
    [anon_sym_public] = ACTIONS(1890),
    [anon_sym_protected] = ACTIONS(1890),
    [anon_sym_private] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_final] = ACTIONS(1890),
    [anon_sym_synchronized] = ACTIONS(1890),
    [anon_sym_TILDE] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_new] = ACTIONS(1890),
  },
  [STATE(926)] = {
    [sym_comment] = STATE(926),
    [sym_groovy_doc] = STATE(926),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1270),
    [anon_sym_STAR_COLON] = ACTIONS(1268),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_package] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_PLUS_PLUS] = ACTIONS(1268),
    [anon_sym_DASH_DASH] = ACTIONS(1268),
    [anon_sym_SLASH] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1270),
    [anon_sym_DASH] = ACTIONS(1270),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [anon_sym_ATinterface] = ACTIONS(1268),
    [anon_sym_interface] = ACTIONS(1270),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_null] = ACTIONS(1270),
    [anon_sym_this] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [aux_sym__type_identifier_token1] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_else] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [aux_sym_number_literal_token1] = ACTIONS(1270),
    [aux_sym_number_literal_token2] = ACTIONS(1268),
    [aux_sym_number_literal_token3] = ACTIONS(1268),
    [aux_sym_number_literal_token4] = ACTIONS(1270),
    [aux_sym_number_literal_token5] = ACTIONS(1268),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1268),
    [anon_sym_DQUOTE] = ACTIONS(1270),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1268),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1268),
    [anon_sym_switch] = ACTIONS(1270),
    [anon_sym_case] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_int] = ACTIONS(1270),
    [anon_sym_boolean] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_short] = ACTIONS(1270),
    [anon_sym_long] = ACTIONS(1270),
    [anon_sym_float] = ACTIONS(1270),
    [anon_sym_double] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_public] = ACTIONS(1270),
    [anon_sym_protected] = ACTIONS(1270),
    [anon_sym_private] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_final] = ACTIONS(1270),
    [anon_sym_synchronized] = ACTIONS(1270),
    [anon_sym_TILDE] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_new] = ACTIONS(1270),
  },
  [STATE(927)] = {
    [sym_closure] = STATE(851),
    [sym_comment] = STATE(927),
    [sym_groovy_doc] = STATE(927),
    [ts_builtin_sym_end] = ACTIONS(2041),
    [sym_identifier] = ACTIONS(2039),
    [anon_sym_break] = ACTIONS(2039),
    [anon_sym_continue] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_STAR] = ACTIONS(2039),
    [anon_sym_STAR_COLON] = ACTIONS(2041),
    [anon_sym_import] = ACTIONS(2039),
    [anon_sym_package] = ACTIONS(2039),
    [anon_sym_AT] = ACTIONS(2039),
    [anon_sym_assert] = ACTIONS(2039),
    [anon_sym_PLUS_PLUS] = ACTIONS(2041),
    [anon_sym_DASH_DASH] = ACTIONS(2041),
    [anon_sym_SLASH] = ACTIONS(2039),
    [anon_sym_PLUS] = ACTIONS(2039),
    [anon_sym_DASH] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_ATinterface] = ACTIONS(2041),
    [anon_sym_interface] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2039),
    [anon_sym_def] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_null] = ACTIONS(2039),
    [anon_sym_this] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_while] = ACTIONS(2039),
    [anon_sym_for] = ACTIONS(2039),
    [aux_sym__type_identifier_token1] = ACTIONS(2039),
    [anon_sym_if] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2041),
    [aux_sym_number_literal_token1] = ACTIONS(2039),
    [aux_sym_number_literal_token2] = ACTIONS(2041),
    [aux_sym_number_literal_token3] = ACTIONS(2041),
    [aux_sym_number_literal_token4] = ACTIONS(2039),
    [aux_sym_number_literal_token5] = ACTIONS(2041),
    [anon_sym_pipeline] = ACTIONS(2039),
    [anon_sym_return] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2039),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2039),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2041),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2041),
    [anon_sym_switch] = ACTIONS(2039),
    [anon_sym_try] = ACTIONS(2039),
    [anon_sym_int] = ACTIONS(2039),
    [anon_sym_boolean] = ACTIONS(2039),
    [anon_sym_char] = ACTIONS(2039),
    [anon_sym_short] = ACTIONS(2039),
    [anon_sym_long] = ACTIONS(2039),
    [anon_sym_float] = ACTIONS(2039),
    [anon_sym_double] = ACTIONS(2039),
    [anon_sym_void] = ACTIONS(2039),
    [anon_sym_public] = ACTIONS(2039),
    [anon_sym_protected] = ACTIONS(2039),
    [anon_sym_private] = ACTIONS(2039),
    [anon_sym_static] = ACTIONS(2039),
    [anon_sym_final] = ACTIONS(2039),
    [anon_sym_synchronized] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2041),
    [anon_sym_BANG] = ACTIONS(2041),
    [anon_sym_new] = ACTIONS(2039),
  },
  [STATE(928)] = {
    [sym_comment] = STATE(928),
    [sym_groovy_doc] = STATE(928),
    [sym_identifier] = ACTIONS(1739),
    [anon_sym_break] = ACTIONS(1739),
    [anon_sym_continue] = ACTIONS(1739),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_STAR] = ACTIONS(1739),
    [anon_sym_STAR_COLON] = ACTIONS(1741),
    [anon_sym_import] = ACTIONS(1739),
    [anon_sym_package] = ACTIONS(1739),
    [anon_sym_AT] = ACTIONS(1739),
    [anon_sym_assert] = ACTIONS(1739),
    [anon_sym_EQ] = ACTIONS(2096),
    [anon_sym_PLUS_PLUS] = ACTIONS(1741),
    [anon_sym_DASH_DASH] = ACTIONS(1741),
    [anon_sym_SLASH] = ACTIONS(1739),
    [anon_sym_PLUS] = ACTIONS(1739),
    [anon_sym_DASH] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1739),
    [anon_sym_false] = ACTIONS(1739),
    [anon_sym_ATinterface] = ACTIONS(1741),
    [anon_sym_interface] = ACTIONS(1739),
    [anon_sym_class] = ACTIONS(1739),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_RBRACE] = ACTIONS(1741),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1739),
    [anon_sym_def] = ACTIONS(1739),
    [anon_sym_LPAREN] = ACTIONS(1741),
    [anon_sym_null] = ACTIONS(1739),
    [anon_sym_this] = ACTIONS(1739),
    [anon_sym_do] = ACTIONS(1739),
    [anon_sym_while] = ACTIONS(1739),
    [anon_sym_for] = ACTIONS(1739),
    [aux_sym__type_identifier_token1] = ACTIONS(1739),
    [anon_sym_if] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [aux_sym_number_literal_token1] = ACTIONS(1739),
    [aux_sym_number_literal_token2] = ACTIONS(1741),
    [aux_sym_number_literal_token3] = ACTIONS(1741),
    [aux_sym_number_literal_token4] = ACTIONS(1739),
    [aux_sym_number_literal_token5] = ACTIONS(1741),
    [anon_sym_return] = ACTIONS(1739),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1741),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1741),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1741),
    [anon_sym_switch] = ACTIONS(1739),
    [anon_sym_case] = ACTIONS(1739),
    [anon_sym_default] = ACTIONS(1739),
    [anon_sym_try] = ACTIONS(1739),
    [anon_sym_int] = ACTIONS(1739),
    [anon_sym_boolean] = ACTIONS(1739),
    [anon_sym_char] = ACTIONS(1739),
    [anon_sym_short] = ACTIONS(1739),
    [anon_sym_long] = ACTIONS(1739),
    [anon_sym_float] = ACTIONS(1739),
    [anon_sym_double] = ACTIONS(1739),
    [anon_sym_void] = ACTIONS(1739),
    [anon_sym_public] = ACTIONS(1739),
    [anon_sym_protected] = ACTIONS(1739),
    [anon_sym_private] = ACTIONS(1739),
    [anon_sym_static] = ACTIONS(1739),
    [anon_sym_final] = ACTIONS(1739),
    [anon_sym_synchronized] = ACTIONS(1739),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_BANG] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1739),
  },
  [STATE(929)] = {
    [sym_closure] = STATE(995),
    [sym_comment] = STATE(929),
    [sym_groovy_doc] = STATE(929),
    [sym_identifier] = ACTIONS(2039),
    [anon_sym_break] = ACTIONS(2039),
    [anon_sym_continue] = ACTIONS(2039),
    [anon_sym_SEMI] = ACTIONS(2041),
    [anon_sym_STAR] = ACTIONS(2039),
    [anon_sym_STAR_COLON] = ACTIONS(2041),
    [anon_sym_import] = ACTIONS(2039),
    [anon_sym_package] = ACTIONS(2039),
    [anon_sym_AT] = ACTIONS(2039),
    [anon_sym_assert] = ACTIONS(2039),
    [anon_sym_PLUS_PLUS] = ACTIONS(2041),
    [anon_sym_DASH_DASH] = ACTIONS(2041),
    [anon_sym_SLASH] = ACTIONS(2039),
    [anon_sym_PLUS] = ACTIONS(2039),
    [anon_sym_DASH] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(2039),
    [anon_sym_false] = ACTIONS(2039),
    [anon_sym_ATinterface] = ACTIONS(2041),
    [anon_sym_interface] = ACTIONS(2039),
    [anon_sym_class] = ACTIONS(2039),
    [anon_sym_LBRACE] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2041),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2039),
    [anon_sym_def] = ACTIONS(2039),
    [anon_sym_LPAREN] = ACTIONS(2041),
    [anon_sym_null] = ACTIONS(2039),
    [anon_sym_this] = ACTIONS(2039),
    [anon_sym_do] = ACTIONS(2039),
    [anon_sym_while] = ACTIONS(2039),
    [anon_sym_for] = ACTIONS(2039),
    [aux_sym__type_identifier_token1] = ACTIONS(2039),
    [anon_sym_if] = ACTIONS(2039),
    [anon_sym_else] = ACTIONS(2039),
    [anon_sym_LBRACK] = ACTIONS(2041),
    [aux_sym_number_literal_token1] = ACTIONS(2039),
    [aux_sym_number_literal_token2] = ACTIONS(2041),
    [aux_sym_number_literal_token3] = ACTIONS(2041),
    [aux_sym_number_literal_token4] = ACTIONS(2039),
    [aux_sym_number_literal_token5] = ACTIONS(2041),
    [anon_sym_return] = ACTIONS(2039),
    [anon_sym_SQUOTE] = ACTIONS(2039),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2041),
    [anon_sym_DQUOTE] = ACTIONS(2039),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2041),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2041),
    [anon_sym_switch] = ACTIONS(2039),
    [anon_sym_case] = ACTIONS(2039),
    [anon_sym_default] = ACTIONS(2039),
    [anon_sym_try] = ACTIONS(2039),
    [anon_sym_int] = ACTIONS(2039),
    [anon_sym_boolean] = ACTIONS(2039),
    [anon_sym_char] = ACTIONS(2039),
    [anon_sym_short] = ACTIONS(2039),
    [anon_sym_long] = ACTIONS(2039),
    [anon_sym_float] = ACTIONS(2039),
    [anon_sym_double] = ACTIONS(2039),
    [anon_sym_void] = ACTIONS(2039),
    [anon_sym_public] = ACTIONS(2039),
    [anon_sym_protected] = ACTIONS(2039),
    [anon_sym_private] = ACTIONS(2039),
    [anon_sym_static] = ACTIONS(2039),
    [anon_sym_final] = ACTIONS(2039),
    [anon_sym_synchronized] = ACTIONS(2039),
    [anon_sym_TILDE] = ACTIONS(2041),
    [anon_sym_BANG] = ACTIONS(2041),
    [anon_sym_new] = ACTIONS(2039),
  },
  [STATE(930)] = {
    [sym_closure] = STATE(967),
    [sym_comment] = STATE(930),
    [sym_groovy_doc] = STATE(930),
    [sym_identifier] = ACTIONS(2051),
    [anon_sym_break] = ACTIONS(2051),
    [anon_sym_continue] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_STAR] = ACTIONS(2051),
    [anon_sym_STAR_COLON] = ACTIONS(2053),
    [anon_sym_import] = ACTIONS(2051),
    [anon_sym_package] = ACTIONS(2051),
    [anon_sym_AT] = ACTIONS(2051),
    [anon_sym_assert] = ACTIONS(2051),
    [anon_sym_PLUS_PLUS] = ACTIONS(2053),
    [anon_sym_DASH_DASH] = ACTIONS(2053),
    [anon_sym_SLASH] = ACTIONS(2051),
    [anon_sym_PLUS] = ACTIONS(2051),
    [anon_sym_DASH] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_ATinterface] = ACTIONS(2053),
    [anon_sym_interface] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2135),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2051),
    [anon_sym_def] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_null] = ACTIONS(2051),
    [anon_sym_this] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_while] = ACTIONS(2051),
    [anon_sym_for] = ACTIONS(2051),
    [aux_sym__type_identifier_token1] = ACTIONS(2051),
    [anon_sym_if] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2053),
    [aux_sym_number_literal_token1] = ACTIONS(2051),
    [aux_sym_number_literal_token2] = ACTIONS(2053),
    [aux_sym_number_literal_token3] = ACTIONS(2053),
    [aux_sym_number_literal_token4] = ACTIONS(2051),
    [aux_sym_number_literal_token5] = ACTIONS(2053),
    [anon_sym_return] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2051),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2051),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2053),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2053),
    [anon_sym_switch] = ACTIONS(2051),
    [anon_sym_case] = ACTIONS(2051),
    [anon_sym_default] = ACTIONS(2051),
    [anon_sym_try] = ACTIONS(2051),
    [anon_sym_int] = ACTIONS(2051),
    [anon_sym_boolean] = ACTIONS(2051),
    [anon_sym_char] = ACTIONS(2051),
    [anon_sym_short] = ACTIONS(2051),
    [anon_sym_long] = ACTIONS(2051),
    [anon_sym_float] = ACTIONS(2051),
    [anon_sym_double] = ACTIONS(2051),
    [anon_sym_void] = ACTIONS(2051),
    [anon_sym_public] = ACTIONS(2051),
    [anon_sym_protected] = ACTIONS(2051),
    [anon_sym_private] = ACTIONS(2051),
    [anon_sym_static] = ACTIONS(2051),
    [anon_sym_final] = ACTIONS(2051),
    [anon_sym_synchronized] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2053),
    [anon_sym_BANG] = ACTIONS(2053),
    [anon_sym_new] = ACTIONS(2051),
  },
  [STATE(931)] = {
    [sym_closure] = STATE(834),
    [sym_comment] = STATE(931),
    [sym_groovy_doc] = STATE(931),
    [ts_builtin_sym_end] = ACTIONS(2053),
    [sym_identifier] = ACTIONS(2051),
    [anon_sym_break] = ACTIONS(2051),
    [anon_sym_continue] = ACTIONS(2051),
    [anon_sym_SEMI] = ACTIONS(2053),
    [anon_sym_STAR] = ACTIONS(2051),
    [anon_sym_STAR_COLON] = ACTIONS(2053),
    [anon_sym_import] = ACTIONS(2051),
    [anon_sym_package] = ACTIONS(2051),
    [anon_sym_AT] = ACTIONS(2051),
    [anon_sym_assert] = ACTIONS(2051),
    [anon_sym_PLUS_PLUS] = ACTIONS(2053),
    [anon_sym_DASH_DASH] = ACTIONS(2053),
    [anon_sym_SLASH] = ACTIONS(2051),
    [anon_sym_PLUS] = ACTIONS(2051),
    [anon_sym_DASH] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(2051),
    [anon_sym_false] = ACTIONS(2051),
    [anon_sym_ATinterface] = ACTIONS(2053),
    [anon_sym_interface] = ACTIONS(2051),
    [anon_sym_class] = ACTIONS(2051),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_RBRACE] = ACTIONS(2053),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2051),
    [anon_sym_def] = ACTIONS(2051),
    [anon_sym_LPAREN] = ACTIONS(2053),
    [anon_sym_null] = ACTIONS(2051),
    [anon_sym_this] = ACTIONS(2051),
    [anon_sym_do] = ACTIONS(2051),
    [anon_sym_while] = ACTIONS(2051),
    [anon_sym_for] = ACTIONS(2051),
    [aux_sym__type_identifier_token1] = ACTIONS(2051),
    [anon_sym_if] = ACTIONS(2051),
    [anon_sym_else] = ACTIONS(2051),
    [anon_sym_LBRACK] = ACTIONS(2053),
    [aux_sym_number_literal_token1] = ACTIONS(2051),
    [aux_sym_number_literal_token2] = ACTIONS(2053),
    [aux_sym_number_literal_token3] = ACTIONS(2053),
    [aux_sym_number_literal_token4] = ACTIONS(2051),
    [aux_sym_number_literal_token5] = ACTIONS(2053),
    [anon_sym_pipeline] = ACTIONS(2051),
    [anon_sym_return] = ACTIONS(2051),
    [anon_sym_SQUOTE] = ACTIONS(2051),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2053),
    [anon_sym_DQUOTE] = ACTIONS(2051),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2053),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2053),
    [anon_sym_switch] = ACTIONS(2051),
    [anon_sym_try] = ACTIONS(2051),
    [anon_sym_int] = ACTIONS(2051),
    [anon_sym_boolean] = ACTIONS(2051),
    [anon_sym_char] = ACTIONS(2051),
    [anon_sym_short] = ACTIONS(2051),
    [anon_sym_long] = ACTIONS(2051),
    [anon_sym_float] = ACTIONS(2051),
    [anon_sym_double] = ACTIONS(2051),
    [anon_sym_void] = ACTIONS(2051),
    [anon_sym_public] = ACTIONS(2051),
    [anon_sym_protected] = ACTIONS(2051),
    [anon_sym_private] = ACTIONS(2051),
    [anon_sym_static] = ACTIONS(2051),
    [anon_sym_final] = ACTIONS(2051),
    [anon_sym_synchronized] = ACTIONS(2051),
    [anon_sym_TILDE] = ACTIONS(2053),
    [anon_sym_BANG] = ACTIONS(2053),
    [anon_sym_new] = ACTIONS(2051),
  },
  [STATE(932)] = {
    [sym_comment] = STATE(932),
    [sym_groovy_doc] = STATE(932),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1692),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_COMMA] = ACTIONS(1519),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_RPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(933)] = {
    [sym_comment] = STATE(933),
    [sym_groovy_doc] = STATE(933),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_break] = ACTIONS(1938),
    [anon_sym_continue] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_STAR] = ACTIONS(1938),
    [anon_sym_STAR_COLON] = ACTIONS(1936),
    [anon_sym_import] = ACTIONS(1938),
    [anon_sym_package] = ACTIONS(1938),
    [anon_sym_AT] = ACTIONS(1938),
    [anon_sym_assert] = ACTIONS(1938),
    [anon_sym_PLUS_PLUS] = ACTIONS(1936),
    [anon_sym_DASH_DASH] = ACTIONS(1936),
    [anon_sym_SLASH] = ACTIONS(1938),
    [anon_sym_PLUS] = ACTIONS(1938),
    [anon_sym_DASH] = ACTIONS(1938),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_ATinterface] = ACTIONS(1936),
    [anon_sym_interface] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1938),
    [anon_sym_def] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_null] = ACTIONS(1938),
    [anon_sym_this] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_while] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1938),
    [aux_sym__type_identifier_token1] = ACTIONS(1938),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1936),
    [aux_sym_number_literal_token1] = ACTIONS(1938),
    [aux_sym_number_literal_token2] = ACTIONS(1936),
    [aux_sym_number_literal_token3] = ACTIONS(1936),
    [aux_sym_number_literal_token4] = ACTIONS(1938),
    [aux_sym_number_literal_token5] = ACTIONS(1936),
    [anon_sym_return] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1938),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1936),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1936),
    [anon_sym_switch] = ACTIONS(1938),
    [anon_sym_case] = ACTIONS(1938),
    [anon_sym_default] = ACTIONS(1938),
    [anon_sym_try] = ACTIONS(1938),
    [anon_sym_finally] = ACTIONS(2149),
    [anon_sym_int] = ACTIONS(1938),
    [anon_sym_boolean] = ACTIONS(1938),
    [anon_sym_char] = ACTIONS(1938),
    [anon_sym_short] = ACTIONS(1938),
    [anon_sym_long] = ACTIONS(1938),
    [anon_sym_float] = ACTIONS(1938),
    [anon_sym_double] = ACTIONS(1938),
    [anon_sym_void] = ACTIONS(1938),
    [anon_sym_public] = ACTIONS(1938),
    [anon_sym_protected] = ACTIONS(1938),
    [anon_sym_private] = ACTIONS(1938),
    [anon_sym_static] = ACTIONS(1938),
    [anon_sym_final] = ACTIONS(1938),
    [anon_sym_synchronized] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1936),
    [anon_sym_BANG] = ACTIONS(1936),
    [anon_sym_new] = ACTIONS(1938),
  },
  [STATE(934)] = {
    [sym_closure] = STATE(796),
    [sym_comment] = STATE(934),
    [sym_groovy_doc] = STATE(934),
    [ts_builtin_sym_end] = ACTIONS(2049),
    [sym_identifier] = ACTIONS(2047),
    [anon_sym_break] = ACTIONS(2047),
    [anon_sym_continue] = ACTIONS(2047),
    [anon_sym_SEMI] = ACTIONS(2049),
    [anon_sym_STAR] = ACTIONS(2047),
    [anon_sym_STAR_COLON] = ACTIONS(2049),
    [anon_sym_import] = ACTIONS(2047),
    [anon_sym_package] = ACTIONS(2047),
    [anon_sym_AT] = ACTIONS(2047),
    [anon_sym_assert] = ACTIONS(2047),
    [anon_sym_PLUS_PLUS] = ACTIONS(2049),
    [anon_sym_DASH_DASH] = ACTIONS(2049),
    [anon_sym_SLASH] = ACTIONS(2047),
    [anon_sym_PLUS] = ACTIONS(2047),
    [anon_sym_DASH] = ACTIONS(2047),
    [anon_sym_true] = ACTIONS(2047),
    [anon_sym_false] = ACTIONS(2047),
    [anon_sym_ATinterface] = ACTIONS(2049),
    [anon_sym_interface] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_LBRACE] = ACTIONS(2137),
    [anon_sym_RBRACE] = ACTIONS(2049),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2047),
    [anon_sym_def] = ACTIONS(2047),
    [anon_sym_LPAREN] = ACTIONS(2049),
    [anon_sym_null] = ACTIONS(2047),
    [anon_sym_this] = ACTIONS(2047),
    [anon_sym_do] = ACTIONS(2047),
    [anon_sym_while] = ACTIONS(2047),
    [anon_sym_for] = ACTIONS(2047),
    [aux_sym__type_identifier_token1] = ACTIONS(2047),
    [anon_sym_if] = ACTIONS(2047),
    [anon_sym_else] = ACTIONS(2047),
    [anon_sym_LBRACK] = ACTIONS(2049),
    [aux_sym_number_literal_token1] = ACTIONS(2047),
    [aux_sym_number_literal_token2] = ACTIONS(2049),
    [aux_sym_number_literal_token3] = ACTIONS(2049),
    [aux_sym_number_literal_token4] = ACTIONS(2047),
    [aux_sym_number_literal_token5] = ACTIONS(2049),
    [anon_sym_pipeline] = ACTIONS(2047),
    [anon_sym_return] = ACTIONS(2047),
    [anon_sym_SQUOTE] = ACTIONS(2047),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2049),
    [anon_sym_DQUOTE] = ACTIONS(2047),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2049),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2049),
    [anon_sym_switch] = ACTIONS(2047),
    [anon_sym_try] = ACTIONS(2047),
    [anon_sym_int] = ACTIONS(2047),
    [anon_sym_boolean] = ACTIONS(2047),
    [anon_sym_char] = ACTIONS(2047),
    [anon_sym_short] = ACTIONS(2047),
    [anon_sym_long] = ACTIONS(2047),
    [anon_sym_float] = ACTIONS(2047),
    [anon_sym_double] = ACTIONS(2047),
    [anon_sym_void] = ACTIONS(2047),
    [anon_sym_public] = ACTIONS(2047),
    [anon_sym_protected] = ACTIONS(2047),
    [anon_sym_private] = ACTIONS(2047),
    [anon_sym_static] = ACTIONS(2047),
    [anon_sym_final] = ACTIONS(2047),
    [anon_sym_synchronized] = ACTIONS(2047),
    [anon_sym_TILDE] = ACTIONS(2049),
    [anon_sym_BANG] = ACTIONS(2049),
    [anon_sym_new] = ACTIONS(2047),
  },
  [STATE(935)] = {
    [sym_comment] = STATE(935),
    [sym_groovy_doc] = STATE(935),
    [ts_builtin_sym_end] = ACTIONS(1936),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_break] = ACTIONS(1938),
    [anon_sym_continue] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_STAR] = ACTIONS(1938),
    [anon_sym_STAR_COLON] = ACTIONS(1936),
    [anon_sym_import] = ACTIONS(1938),
    [anon_sym_package] = ACTIONS(1938),
    [anon_sym_AT] = ACTIONS(1938),
    [anon_sym_assert] = ACTIONS(1938),
    [anon_sym_PLUS_PLUS] = ACTIONS(1936),
    [anon_sym_DASH_DASH] = ACTIONS(1936),
    [anon_sym_SLASH] = ACTIONS(1938),
    [anon_sym_PLUS] = ACTIONS(1938),
    [anon_sym_DASH] = ACTIONS(1938),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [anon_sym_ATinterface] = ACTIONS(1936),
    [anon_sym_interface] = ACTIONS(1938),
    [anon_sym_class] = ACTIONS(1938),
    [anon_sym_LBRACE] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1938),
    [anon_sym_def] = ACTIONS(1938),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_null] = ACTIONS(1938),
    [anon_sym_this] = ACTIONS(1938),
    [anon_sym_do] = ACTIONS(1938),
    [anon_sym_while] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1938),
    [aux_sym__type_identifier_token1] = ACTIONS(1938),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_else] = ACTIONS(1938),
    [anon_sym_LBRACK] = ACTIONS(1936),
    [aux_sym_number_literal_token1] = ACTIONS(1938),
    [aux_sym_number_literal_token2] = ACTIONS(1936),
    [aux_sym_number_literal_token3] = ACTIONS(1936),
    [aux_sym_number_literal_token4] = ACTIONS(1938),
    [aux_sym_number_literal_token5] = ACTIONS(1936),
    [anon_sym_pipeline] = ACTIONS(1938),
    [anon_sym_return] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1938),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1936),
    [anon_sym_DQUOTE] = ACTIONS(1938),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1936),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1936),
    [anon_sym_switch] = ACTIONS(1938),
    [anon_sym_try] = ACTIONS(1938),
    [anon_sym_finally] = ACTIONS(2151),
    [anon_sym_int] = ACTIONS(1938),
    [anon_sym_boolean] = ACTIONS(1938),
    [anon_sym_char] = ACTIONS(1938),
    [anon_sym_short] = ACTIONS(1938),
    [anon_sym_long] = ACTIONS(1938),
    [anon_sym_float] = ACTIONS(1938),
    [anon_sym_double] = ACTIONS(1938),
    [anon_sym_void] = ACTIONS(1938),
    [anon_sym_public] = ACTIONS(1938),
    [anon_sym_protected] = ACTIONS(1938),
    [anon_sym_private] = ACTIONS(1938),
    [anon_sym_static] = ACTIONS(1938),
    [anon_sym_final] = ACTIONS(1938),
    [anon_sym_synchronized] = ACTIONS(1938),
    [anon_sym_TILDE] = ACTIONS(1936),
    [anon_sym_BANG] = ACTIONS(1936),
    [anon_sym_new] = ACTIONS(1938),
  },
  [STATE(936)] = {
    [sym_comment] = STATE(936),
    [sym_groovy_doc] = STATE(936),
    [ts_builtin_sym_end] = ACTIONS(1979),
    [sym_identifier] = ACTIONS(1981),
    [anon_sym_break] = ACTIONS(1981),
    [anon_sym_continue] = ACTIONS(1981),
    [anon_sym_SEMI] = ACTIONS(1979),
    [anon_sym_STAR] = ACTIONS(1981),
    [anon_sym_STAR_COLON] = ACTIONS(1979),
    [anon_sym_import] = ACTIONS(1981),
    [anon_sym_package] = ACTIONS(1981),
    [anon_sym_AT] = ACTIONS(1981),
    [anon_sym_assert] = ACTIONS(1981),
    [anon_sym_EQ] = ACTIONS(2123),
    [anon_sym_PLUS_PLUS] = ACTIONS(1979),
    [anon_sym_DASH_DASH] = ACTIONS(1979),
    [anon_sym_SLASH] = ACTIONS(1981),
    [anon_sym_PLUS] = ACTIONS(1981),
    [anon_sym_DASH] = ACTIONS(1981),
    [anon_sym_true] = ACTIONS(1981),
    [anon_sym_false] = ACTIONS(1981),
    [anon_sym_ATinterface] = ACTIONS(1979),
    [anon_sym_interface] = ACTIONS(1981),
    [anon_sym_class] = ACTIONS(1981),
    [anon_sym_LBRACE] = ACTIONS(1979),
    [anon_sym_RBRACE] = ACTIONS(1979),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1981),
    [anon_sym_def] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1979),
    [anon_sym_null] = ACTIONS(1981),
    [anon_sym_this] = ACTIONS(1981),
    [anon_sym_do] = ACTIONS(1981),
    [anon_sym_while] = ACTIONS(1981),
    [anon_sym_for] = ACTIONS(1981),
    [aux_sym__type_identifier_token1] = ACTIONS(1981),
    [anon_sym_if] = ACTIONS(1981),
    [anon_sym_else] = ACTIONS(1981),
    [anon_sym_LBRACK] = ACTIONS(1979),
    [aux_sym_number_literal_token1] = ACTIONS(1981),
    [aux_sym_number_literal_token2] = ACTIONS(1979),
    [aux_sym_number_literal_token3] = ACTIONS(1979),
    [aux_sym_number_literal_token4] = ACTIONS(1981),
    [aux_sym_number_literal_token5] = ACTIONS(1979),
    [anon_sym_pipeline] = ACTIONS(1981),
    [anon_sym_return] = ACTIONS(1981),
    [anon_sym_SQUOTE] = ACTIONS(1981),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1979),
    [anon_sym_DQUOTE] = ACTIONS(1981),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1979),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1979),
    [anon_sym_switch] = ACTIONS(1981),
    [anon_sym_try] = ACTIONS(1981),
    [anon_sym_int] = ACTIONS(1981),
    [anon_sym_boolean] = ACTIONS(1981),
    [anon_sym_char] = ACTIONS(1981),
    [anon_sym_short] = ACTIONS(1981),
    [anon_sym_long] = ACTIONS(1981),
    [anon_sym_float] = ACTIONS(1981),
    [anon_sym_double] = ACTIONS(1981),
    [anon_sym_void] = ACTIONS(1981),
    [anon_sym_public] = ACTIONS(1981),
    [anon_sym_protected] = ACTIONS(1981),
    [anon_sym_private] = ACTIONS(1981),
    [anon_sym_static] = ACTIONS(1981),
    [anon_sym_final] = ACTIONS(1981),
    [anon_sym_synchronized] = ACTIONS(1981),
    [anon_sym_TILDE] = ACTIONS(1979),
    [anon_sym_BANG] = ACTIONS(1979),
    [anon_sym_new] = ACTIONS(1981),
  },
  [STATE(937)] = {
    [sym_comment] = STATE(937),
    [sym_groovy_doc] = STATE(937),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(938)] = {
    [sym_comment] = STATE(938),
    [sym_groovy_doc] = STATE(938),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(939)] = {
    [sym_comment] = STATE(939),
    [sym_groovy_doc] = STATE(939),
    [sym_identifier] = ACTIONS(1862),
    [anon_sym_break] = ACTIONS(1862),
    [anon_sym_continue] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_STAR] = ACTIONS(1862),
    [anon_sym_STAR_COLON] = ACTIONS(1860),
    [anon_sym_import] = ACTIONS(1862),
    [anon_sym_package] = ACTIONS(1862),
    [anon_sym_AT] = ACTIONS(1862),
    [anon_sym_assert] = ACTIONS(1862),
    [anon_sym_PLUS_PLUS] = ACTIONS(1860),
    [anon_sym_DASH_DASH] = ACTIONS(1860),
    [anon_sym_SLASH] = ACTIONS(1862),
    [anon_sym_PLUS] = ACTIONS(1862),
    [anon_sym_DASH] = ACTIONS(1862),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [anon_sym_ATinterface] = ACTIONS(1860),
    [anon_sym_interface] = ACTIONS(1862),
    [anon_sym_class] = ACTIONS(1862),
    [anon_sym_LBRACE] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1862),
    [anon_sym_def] = ACTIONS(1862),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_null] = ACTIONS(1862),
    [anon_sym_this] = ACTIONS(1862),
    [anon_sym_do] = ACTIONS(1862),
    [anon_sym_while] = ACTIONS(1862),
    [anon_sym_for] = ACTIONS(1862),
    [aux_sym__type_identifier_token1] = ACTIONS(1862),
    [anon_sym_if] = ACTIONS(1862),
    [anon_sym_else] = ACTIONS(1862),
    [anon_sym_LBRACK] = ACTIONS(1860),
    [aux_sym_number_literal_token1] = ACTIONS(1862),
    [aux_sym_number_literal_token2] = ACTIONS(1860),
    [aux_sym_number_literal_token3] = ACTIONS(1860),
    [aux_sym_number_literal_token4] = ACTIONS(1862),
    [aux_sym_number_literal_token5] = ACTIONS(1860),
    [anon_sym_return] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1862),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1860),
    [anon_sym_DQUOTE] = ACTIONS(1862),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1860),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1860),
    [anon_sym_switch] = ACTIONS(1862),
    [anon_sym_case] = ACTIONS(1862),
    [anon_sym_default] = ACTIONS(1862),
    [anon_sym_try] = ACTIONS(1862),
    [anon_sym_int] = ACTIONS(1862),
    [anon_sym_boolean] = ACTIONS(1862),
    [anon_sym_char] = ACTIONS(1862),
    [anon_sym_short] = ACTIONS(1862),
    [anon_sym_long] = ACTIONS(1862),
    [anon_sym_float] = ACTIONS(1862),
    [anon_sym_double] = ACTIONS(1862),
    [anon_sym_void] = ACTIONS(1862),
    [anon_sym_public] = ACTIONS(1862),
    [anon_sym_protected] = ACTIONS(1862),
    [anon_sym_private] = ACTIONS(1862),
    [anon_sym_static] = ACTIONS(1862),
    [anon_sym_final] = ACTIONS(1862),
    [anon_sym_synchronized] = ACTIONS(1862),
    [anon_sym_TILDE] = ACTIONS(1860),
    [anon_sym_BANG] = ACTIONS(1860),
    [anon_sym_new] = ACTIONS(1862),
  },
  [STATE(940)] = {
    [sym_comment] = STATE(940),
    [sym_groovy_doc] = STATE(940),
    [sym_identifier] = ACTIONS(1946),
    [anon_sym_break] = ACTIONS(1946),
    [anon_sym_continue] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_STAR] = ACTIONS(1946),
    [anon_sym_STAR_COLON] = ACTIONS(1944),
    [anon_sym_import] = ACTIONS(1946),
    [anon_sym_package] = ACTIONS(1946),
    [anon_sym_AT] = ACTIONS(1946),
    [anon_sym_assert] = ACTIONS(1946),
    [anon_sym_PLUS_PLUS] = ACTIONS(1944),
    [anon_sym_DASH_DASH] = ACTIONS(1944),
    [anon_sym_SLASH] = ACTIONS(1946),
    [anon_sym_PLUS] = ACTIONS(1946),
    [anon_sym_DASH] = ACTIONS(1946),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [anon_sym_ATinterface] = ACTIONS(1944),
    [anon_sym_interface] = ACTIONS(1946),
    [anon_sym_class] = ACTIONS(1946),
    [anon_sym_LBRACE] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1946),
    [anon_sym_def] = ACTIONS(1946),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_null] = ACTIONS(1946),
    [anon_sym_this] = ACTIONS(1946),
    [anon_sym_do] = ACTIONS(1946),
    [anon_sym_while] = ACTIONS(1946),
    [anon_sym_for] = ACTIONS(1946),
    [aux_sym__type_identifier_token1] = ACTIONS(1946),
    [anon_sym_if] = ACTIONS(1946),
    [anon_sym_else] = ACTIONS(1946),
    [anon_sym_LBRACK] = ACTIONS(1944),
    [aux_sym_number_literal_token1] = ACTIONS(1946),
    [aux_sym_number_literal_token2] = ACTIONS(1944),
    [aux_sym_number_literal_token3] = ACTIONS(1944),
    [aux_sym_number_literal_token4] = ACTIONS(1946),
    [aux_sym_number_literal_token5] = ACTIONS(1944),
    [anon_sym_return] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1946),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1944),
    [anon_sym_DQUOTE] = ACTIONS(1946),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1944),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1944),
    [anon_sym_switch] = ACTIONS(1946),
    [anon_sym_case] = ACTIONS(1946),
    [anon_sym_default] = ACTIONS(1946),
    [anon_sym_try] = ACTIONS(1946),
    [anon_sym_int] = ACTIONS(1946),
    [anon_sym_boolean] = ACTIONS(1946),
    [anon_sym_char] = ACTIONS(1946),
    [anon_sym_short] = ACTIONS(1946),
    [anon_sym_long] = ACTIONS(1946),
    [anon_sym_float] = ACTIONS(1946),
    [anon_sym_double] = ACTIONS(1946),
    [anon_sym_void] = ACTIONS(1946),
    [anon_sym_public] = ACTIONS(1946),
    [anon_sym_protected] = ACTIONS(1946),
    [anon_sym_private] = ACTIONS(1946),
    [anon_sym_static] = ACTIONS(1946),
    [anon_sym_final] = ACTIONS(1946),
    [anon_sym_synchronized] = ACTIONS(1946),
    [anon_sym_TILDE] = ACTIONS(1944),
    [anon_sym_BANG] = ACTIONS(1944),
    [anon_sym_new] = ACTIONS(1946),
  },
  [STATE(941)] = {
    [sym_comment] = STATE(941),
    [sym_groovy_doc] = STATE(941),
    [sym_identifier] = ACTIONS(1954),
    [anon_sym_break] = ACTIONS(1954),
    [anon_sym_continue] = ACTIONS(1954),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_STAR] = ACTIONS(1954),
    [anon_sym_STAR_COLON] = ACTIONS(1952),
    [anon_sym_import] = ACTIONS(1954),
    [anon_sym_package] = ACTIONS(1954),
    [anon_sym_AT] = ACTIONS(1954),
    [anon_sym_assert] = ACTIONS(1954),
    [anon_sym_PLUS_PLUS] = ACTIONS(1952),
    [anon_sym_DASH_DASH] = ACTIONS(1952),
    [anon_sym_SLASH] = ACTIONS(1954),
    [anon_sym_PLUS] = ACTIONS(1954),
    [anon_sym_DASH] = ACTIONS(1954),
    [anon_sym_true] = ACTIONS(1954),
    [anon_sym_false] = ACTIONS(1954),
    [anon_sym_ATinterface] = ACTIONS(1952),
    [anon_sym_interface] = ACTIONS(1954),
    [anon_sym_class] = ACTIONS(1954),
    [anon_sym_LBRACE] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1954),
    [anon_sym_def] = ACTIONS(1954),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_null] = ACTIONS(1954),
    [anon_sym_this] = ACTIONS(1954),
    [anon_sym_do] = ACTIONS(1954),
    [anon_sym_while] = ACTIONS(1954),
    [anon_sym_for] = ACTIONS(1954),
    [aux_sym__type_identifier_token1] = ACTIONS(1954),
    [anon_sym_if] = ACTIONS(1954),
    [anon_sym_else] = ACTIONS(1954),
    [anon_sym_LBRACK] = ACTIONS(1952),
    [aux_sym_number_literal_token1] = ACTIONS(1954),
    [aux_sym_number_literal_token2] = ACTIONS(1952),
    [aux_sym_number_literal_token3] = ACTIONS(1952),
    [aux_sym_number_literal_token4] = ACTIONS(1954),
    [aux_sym_number_literal_token5] = ACTIONS(1952),
    [anon_sym_return] = ACTIONS(1954),
    [anon_sym_SQUOTE] = ACTIONS(1954),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1952),
    [anon_sym_DQUOTE] = ACTIONS(1954),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1952),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1952),
    [anon_sym_switch] = ACTIONS(1954),
    [anon_sym_case] = ACTIONS(1954),
    [anon_sym_default] = ACTIONS(1954),
    [anon_sym_try] = ACTIONS(1954),
    [anon_sym_int] = ACTIONS(1954),
    [anon_sym_boolean] = ACTIONS(1954),
    [anon_sym_char] = ACTIONS(1954),
    [anon_sym_short] = ACTIONS(1954),
    [anon_sym_long] = ACTIONS(1954),
    [anon_sym_float] = ACTIONS(1954),
    [anon_sym_double] = ACTIONS(1954),
    [anon_sym_void] = ACTIONS(1954),
    [anon_sym_public] = ACTIONS(1954),
    [anon_sym_protected] = ACTIONS(1954),
    [anon_sym_private] = ACTIONS(1954),
    [anon_sym_static] = ACTIONS(1954),
    [anon_sym_final] = ACTIONS(1954),
    [anon_sym_synchronized] = ACTIONS(1954),
    [anon_sym_TILDE] = ACTIONS(1952),
    [anon_sym_BANG] = ACTIONS(1952),
    [anon_sym_new] = ACTIONS(1954),
  },
  [STATE(942)] = {
    [sym_comment] = STATE(942),
    [sym_groovy_doc] = STATE(942),
    [ts_builtin_sym_end] = ACTIONS(1410),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_pipeline] = ACTIONS(1412),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(943)] = {
    [sym_comment] = STATE(943),
    [sym_groovy_doc] = STATE(943),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(944)] = {
    [sym_comment] = STATE(944),
    [sym_groovy_doc] = STATE(944),
    [sym_identifier] = ACTIONS(2031),
    [anon_sym_break] = ACTIONS(2031),
    [anon_sym_continue] = ACTIONS(2031),
    [anon_sym_SEMI] = ACTIONS(2029),
    [anon_sym_STAR] = ACTIONS(2031),
    [anon_sym_STAR_COLON] = ACTIONS(2029),
    [anon_sym_import] = ACTIONS(2031),
    [anon_sym_package] = ACTIONS(2031),
    [anon_sym_AT] = ACTIONS(2031),
    [anon_sym_assert] = ACTIONS(2031),
    [anon_sym_PLUS_PLUS] = ACTIONS(2029),
    [anon_sym_DASH_DASH] = ACTIONS(2029),
    [anon_sym_SLASH] = ACTIONS(2031),
    [anon_sym_PLUS] = ACTIONS(2031),
    [anon_sym_DASH] = ACTIONS(2031),
    [anon_sym_true] = ACTIONS(2031),
    [anon_sym_false] = ACTIONS(2031),
    [anon_sym_ATinterface] = ACTIONS(2029),
    [anon_sym_interface] = ACTIONS(2031),
    [anon_sym_class] = ACTIONS(2031),
    [anon_sym_LBRACE] = ACTIONS(2029),
    [anon_sym_RBRACE] = ACTIONS(2029),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2031),
    [anon_sym_def] = ACTIONS(2031),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_null] = ACTIONS(2031),
    [anon_sym_this] = ACTIONS(2031),
    [anon_sym_do] = ACTIONS(2031),
    [anon_sym_while] = ACTIONS(2031),
    [anon_sym_for] = ACTIONS(2031),
    [aux_sym__type_identifier_token1] = ACTIONS(2031),
    [anon_sym_if] = ACTIONS(2031),
    [anon_sym_else] = ACTIONS(2031),
    [anon_sym_LBRACK] = ACTIONS(2029),
    [aux_sym_number_literal_token1] = ACTIONS(2031),
    [aux_sym_number_literal_token2] = ACTIONS(2029),
    [aux_sym_number_literal_token3] = ACTIONS(2029),
    [aux_sym_number_literal_token4] = ACTIONS(2031),
    [aux_sym_number_literal_token5] = ACTIONS(2029),
    [anon_sym_return] = ACTIONS(2031),
    [anon_sym_SQUOTE] = ACTIONS(2031),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2029),
    [anon_sym_DQUOTE] = ACTIONS(2031),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2029),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2029),
    [anon_sym_switch] = ACTIONS(2031),
    [anon_sym_case] = ACTIONS(2031),
    [anon_sym_default] = ACTIONS(2031),
    [anon_sym_try] = ACTIONS(2031),
    [anon_sym_int] = ACTIONS(2031),
    [anon_sym_boolean] = ACTIONS(2031),
    [anon_sym_char] = ACTIONS(2031),
    [anon_sym_short] = ACTIONS(2031),
    [anon_sym_long] = ACTIONS(2031),
    [anon_sym_float] = ACTIONS(2031),
    [anon_sym_double] = ACTIONS(2031),
    [anon_sym_void] = ACTIONS(2031),
    [anon_sym_public] = ACTIONS(2031),
    [anon_sym_protected] = ACTIONS(2031),
    [anon_sym_private] = ACTIONS(2031),
    [anon_sym_static] = ACTIONS(2031),
    [anon_sym_final] = ACTIONS(2031),
    [anon_sym_synchronized] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(2029),
    [anon_sym_BANG] = ACTIONS(2029),
    [anon_sym_new] = ACTIONS(2031),
  },
  [STATE(945)] = {
    [sym_comment] = STATE(945),
    [sym_groovy_doc] = STATE(945),
    [sym_identifier] = ACTIONS(2015),
    [anon_sym_break] = ACTIONS(2015),
    [anon_sym_continue] = ACTIONS(2015),
    [anon_sym_SEMI] = ACTIONS(2013),
    [anon_sym_STAR] = ACTIONS(2015),
    [anon_sym_STAR_COLON] = ACTIONS(2013),
    [anon_sym_import] = ACTIONS(2015),
    [anon_sym_package] = ACTIONS(2015),
    [anon_sym_AT] = ACTIONS(2015),
    [anon_sym_assert] = ACTIONS(2015),
    [anon_sym_PLUS_PLUS] = ACTIONS(2013),
    [anon_sym_DASH_DASH] = ACTIONS(2013),
    [anon_sym_SLASH] = ACTIONS(2015),
    [anon_sym_PLUS] = ACTIONS(2015),
    [anon_sym_DASH] = ACTIONS(2015),
    [anon_sym_true] = ACTIONS(2015),
    [anon_sym_false] = ACTIONS(2015),
    [anon_sym_ATinterface] = ACTIONS(2013),
    [anon_sym_interface] = ACTIONS(2015),
    [anon_sym_class] = ACTIONS(2015),
    [anon_sym_LBRACE] = ACTIONS(2013),
    [anon_sym_RBRACE] = ACTIONS(2013),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2015),
    [anon_sym_def] = ACTIONS(2015),
    [anon_sym_LPAREN] = ACTIONS(2013),
    [anon_sym_null] = ACTIONS(2015),
    [anon_sym_this] = ACTIONS(2015),
    [anon_sym_do] = ACTIONS(2015),
    [anon_sym_while] = ACTIONS(2015),
    [anon_sym_for] = ACTIONS(2015),
    [aux_sym__type_identifier_token1] = ACTIONS(2015),
    [anon_sym_if] = ACTIONS(2015),
    [anon_sym_else] = ACTIONS(2015),
    [anon_sym_LBRACK] = ACTIONS(2013),
    [aux_sym_number_literal_token1] = ACTIONS(2015),
    [aux_sym_number_literal_token2] = ACTIONS(2013),
    [aux_sym_number_literal_token3] = ACTIONS(2013),
    [aux_sym_number_literal_token4] = ACTIONS(2015),
    [aux_sym_number_literal_token5] = ACTIONS(2013),
    [anon_sym_return] = ACTIONS(2015),
    [anon_sym_SQUOTE] = ACTIONS(2015),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2013),
    [anon_sym_DQUOTE] = ACTIONS(2015),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2013),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2013),
    [anon_sym_switch] = ACTIONS(2015),
    [anon_sym_case] = ACTIONS(2015),
    [anon_sym_default] = ACTIONS(2015),
    [anon_sym_try] = ACTIONS(2015),
    [anon_sym_int] = ACTIONS(2015),
    [anon_sym_boolean] = ACTIONS(2015),
    [anon_sym_char] = ACTIONS(2015),
    [anon_sym_short] = ACTIONS(2015),
    [anon_sym_long] = ACTIONS(2015),
    [anon_sym_float] = ACTIONS(2015),
    [anon_sym_double] = ACTIONS(2015),
    [anon_sym_void] = ACTIONS(2015),
    [anon_sym_public] = ACTIONS(2015),
    [anon_sym_protected] = ACTIONS(2015),
    [anon_sym_private] = ACTIONS(2015),
    [anon_sym_static] = ACTIONS(2015),
    [anon_sym_final] = ACTIONS(2015),
    [anon_sym_synchronized] = ACTIONS(2015),
    [anon_sym_TILDE] = ACTIONS(2013),
    [anon_sym_BANG] = ACTIONS(2013),
    [anon_sym_new] = ACTIONS(2015),
  },
  [STATE(946)] = {
    [sym_comment] = STATE(946),
    [sym_groovy_doc] = STATE(946),
    [ts_builtin_sym_end] = ACTIONS(1904),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(2153),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_pipeline] = ACTIONS(1906),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(947)] = {
    [sym_comment] = STATE(947),
    [sym_groovy_doc] = STATE(947),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(1906),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_case] = ACTIONS(1906),
    [anon_sym_default] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(948)] = {
    [sym_comment] = STATE(948),
    [sym_groovy_doc] = STATE(948),
    [sym_identifier] = ACTIONS(1902),
    [anon_sym_break] = ACTIONS(1902),
    [anon_sym_continue] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_STAR] = ACTIONS(1902),
    [anon_sym_STAR_COLON] = ACTIONS(1900),
    [anon_sym_import] = ACTIONS(1902),
    [anon_sym_package] = ACTIONS(1902),
    [anon_sym_AT] = ACTIONS(1902),
    [anon_sym_assert] = ACTIONS(1902),
    [anon_sym_PLUS_PLUS] = ACTIONS(1900),
    [anon_sym_DASH_DASH] = ACTIONS(1900),
    [anon_sym_SLASH] = ACTIONS(1902),
    [anon_sym_PLUS] = ACTIONS(1902),
    [anon_sym_DASH] = ACTIONS(1902),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [anon_sym_ATinterface] = ACTIONS(1900),
    [anon_sym_interface] = ACTIONS(1902),
    [anon_sym_class] = ACTIONS(1902),
    [anon_sym_LBRACE] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1902),
    [anon_sym_def] = ACTIONS(1902),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_null] = ACTIONS(1902),
    [anon_sym_this] = ACTIONS(1902),
    [anon_sym_do] = ACTIONS(1902),
    [anon_sym_while] = ACTIONS(1902),
    [anon_sym_for] = ACTIONS(1902),
    [aux_sym__type_identifier_token1] = ACTIONS(1902),
    [anon_sym_if] = ACTIONS(1902),
    [anon_sym_else] = ACTIONS(1902),
    [anon_sym_LBRACK] = ACTIONS(1900),
    [aux_sym_number_literal_token1] = ACTIONS(1902),
    [aux_sym_number_literal_token2] = ACTIONS(1900),
    [aux_sym_number_literal_token3] = ACTIONS(1900),
    [aux_sym_number_literal_token4] = ACTIONS(1902),
    [aux_sym_number_literal_token5] = ACTIONS(1900),
    [anon_sym_return] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1902),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1900),
    [anon_sym_DQUOTE] = ACTIONS(1902),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1900),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1900),
    [anon_sym_switch] = ACTIONS(1902),
    [anon_sym_case] = ACTIONS(1902),
    [anon_sym_default] = ACTIONS(1902),
    [anon_sym_try] = ACTIONS(1902),
    [anon_sym_int] = ACTIONS(1902),
    [anon_sym_boolean] = ACTIONS(1902),
    [anon_sym_char] = ACTIONS(1902),
    [anon_sym_short] = ACTIONS(1902),
    [anon_sym_long] = ACTIONS(1902),
    [anon_sym_float] = ACTIONS(1902),
    [anon_sym_double] = ACTIONS(1902),
    [anon_sym_void] = ACTIONS(1902),
    [anon_sym_public] = ACTIONS(1902),
    [anon_sym_protected] = ACTIONS(1902),
    [anon_sym_private] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1902),
    [anon_sym_final] = ACTIONS(1902),
    [anon_sym_synchronized] = ACTIONS(1902),
    [anon_sym_TILDE] = ACTIONS(1900),
    [anon_sym_BANG] = ACTIONS(1900),
    [anon_sym_new] = ACTIONS(1902),
  },
  [STATE(949)] = {
    [sym_comment] = STATE(949),
    [sym_groovy_doc] = STATE(949),
    [sym_identifier] = ACTIONS(2023),
    [anon_sym_break] = ACTIONS(2023),
    [anon_sym_continue] = ACTIONS(2023),
    [anon_sym_SEMI] = ACTIONS(2021),
    [anon_sym_STAR] = ACTIONS(2023),
    [anon_sym_STAR_COLON] = ACTIONS(2021),
    [anon_sym_import] = ACTIONS(2023),
    [anon_sym_package] = ACTIONS(2023),
    [anon_sym_AT] = ACTIONS(2023),
    [anon_sym_assert] = ACTIONS(2023),
    [anon_sym_PLUS_PLUS] = ACTIONS(2021),
    [anon_sym_DASH_DASH] = ACTIONS(2021),
    [anon_sym_SLASH] = ACTIONS(2023),
    [anon_sym_PLUS] = ACTIONS(2023),
    [anon_sym_DASH] = ACTIONS(2023),
    [anon_sym_true] = ACTIONS(2023),
    [anon_sym_false] = ACTIONS(2023),
    [anon_sym_ATinterface] = ACTIONS(2021),
    [anon_sym_interface] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_LBRACE] = ACTIONS(2021),
    [anon_sym_RBRACE] = ACTIONS(2021),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2023),
    [anon_sym_def] = ACTIONS(2023),
    [anon_sym_LPAREN] = ACTIONS(2021),
    [anon_sym_null] = ACTIONS(2023),
    [anon_sym_this] = ACTIONS(2023),
    [anon_sym_do] = ACTIONS(2023),
    [anon_sym_while] = ACTIONS(2023),
    [anon_sym_for] = ACTIONS(2023),
    [aux_sym__type_identifier_token1] = ACTIONS(2023),
    [anon_sym_if] = ACTIONS(2023),
    [anon_sym_else] = ACTIONS(2023),
    [anon_sym_LBRACK] = ACTIONS(2021),
    [aux_sym_number_literal_token1] = ACTIONS(2023),
    [aux_sym_number_literal_token2] = ACTIONS(2021),
    [aux_sym_number_literal_token3] = ACTIONS(2021),
    [aux_sym_number_literal_token4] = ACTIONS(2023),
    [aux_sym_number_literal_token5] = ACTIONS(2021),
    [anon_sym_return] = ACTIONS(2023),
    [anon_sym_SQUOTE] = ACTIONS(2023),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2021),
    [anon_sym_DQUOTE] = ACTIONS(2023),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2021),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2021),
    [anon_sym_switch] = ACTIONS(2023),
    [anon_sym_case] = ACTIONS(2023),
    [anon_sym_default] = ACTIONS(2023),
    [anon_sym_try] = ACTIONS(2023),
    [anon_sym_int] = ACTIONS(2023),
    [anon_sym_boolean] = ACTIONS(2023),
    [anon_sym_char] = ACTIONS(2023),
    [anon_sym_short] = ACTIONS(2023),
    [anon_sym_long] = ACTIONS(2023),
    [anon_sym_float] = ACTIONS(2023),
    [anon_sym_double] = ACTIONS(2023),
    [anon_sym_void] = ACTIONS(2023),
    [anon_sym_public] = ACTIONS(2023),
    [anon_sym_protected] = ACTIONS(2023),
    [anon_sym_private] = ACTIONS(2023),
    [anon_sym_static] = ACTIONS(2023),
    [anon_sym_final] = ACTIONS(2023),
    [anon_sym_synchronized] = ACTIONS(2023),
    [anon_sym_TILDE] = ACTIONS(2021),
    [anon_sym_BANG] = ACTIONS(2021),
    [anon_sym_new] = ACTIONS(2023),
  },
  [STATE(950)] = {
    [sym_comment] = STATE(950),
    [sym_groovy_doc] = STATE(950),
    [sym_identifier] = ACTIONS(1958),
    [anon_sym_break] = ACTIONS(1958),
    [anon_sym_continue] = ACTIONS(1958),
    [anon_sym_SEMI] = ACTIONS(1956),
    [anon_sym_STAR] = ACTIONS(1958),
    [anon_sym_STAR_COLON] = ACTIONS(1956),
    [anon_sym_import] = ACTIONS(1958),
    [anon_sym_package] = ACTIONS(1958),
    [anon_sym_AT] = ACTIONS(1958),
    [anon_sym_assert] = ACTIONS(1958),
    [anon_sym_PLUS_PLUS] = ACTIONS(1956),
    [anon_sym_DASH_DASH] = ACTIONS(1956),
    [anon_sym_SLASH] = ACTIONS(1958),
    [anon_sym_PLUS] = ACTIONS(1958),
    [anon_sym_DASH] = ACTIONS(1958),
    [anon_sym_true] = ACTIONS(1958),
    [anon_sym_false] = ACTIONS(1958),
    [anon_sym_ATinterface] = ACTIONS(1956),
    [anon_sym_interface] = ACTIONS(1958),
    [anon_sym_class] = ACTIONS(1958),
    [anon_sym_LBRACE] = ACTIONS(1956),
    [anon_sym_RBRACE] = ACTIONS(1956),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1958),
    [anon_sym_def] = ACTIONS(1958),
    [anon_sym_LPAREN] = ACTIONS(1956),
    [anon_sym_null] = ACTIONS(1958),
    [anon_sym_this] = ACTIONS(1958),
    [anon_sym_do] = ACTIONS(1958),
    [anon_sym_while] = ACTIONS(1958),
    [anon_sym_for] = ACTIONS(1958),
    [aux_sym__type_identifier_token1] = ACTIONS(1958),
    [anon_sym_if] = ACTIONS(1958),
    [anon_sym_else] = ACTIONS(1958),
    [anon_sym_LBRACK] = ACTIONS(1956),
    [aux_sym_number_literal_token1] = ACTIONS(1958),
    [aux_sym_number_literal_token2] = ACTIONS(1956),
    [aux_sym_number_literal_token3] = ACTIONS(1956),
    [aux_sym_number_literal_token4] = ACTIONS(1958),
    [aux_sym_number_literal_token5] = ACTIONS(1956),
    [anon_sym_return] = ACTIONS(1958),
    [anon_sym_SQUOTE] = ACTIONS(1958),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1956),
    [anon_sym_DQUOTE] = ACTIONS(1958),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1956),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1956),
    [anon_sym_switch] = ACTIONS(1958),
    [anon_sym_case] = ACTIONS(1958),
    [anon_sym_default] = ACTIONS(1958),
    [anon_sym_try] = ACTIONS(1958),
    [anon_sym_int] = ACTIONS(1958),
    [anon_sym_boolean] = ACTIONS(1958),
    [anon_sym_char] = ACTIONS(1958),
    [anon_sym_short] = ACTIONS(1958),
    [anon_sym_long] = ACTIONS(1958),
    [anon_sym_float] = ACTIONS(1958),
    [anon_sym_double] = ACTIONS(1958),
    [anon_sym_void] = ACTIONS(1958),
    [anon_sym_public] = ACTIONS(1958),
    [anon_sym_protected] = ACTIONS(1958),
    [anon_sym_private] = ACTIONS(1958),
    [anon_sym_static] = ACTIONS(1958),
    [anon_sym_final] = ACTIONS(1958),
    [anon_sym_synchronized] = ACTIONS(1958),
    [anon_sym_TILDE] = ACTIONS(1956),
    [anon_sym_BANG] = ACTIONS(1956),
    [anon_sym_new] = ACTIONS(1958),
  },
  [STATE(951)] = {
    [sym_comment] = STATE(951),
    [sym_groovy_doc] = STATE(951),
    [ts_builtin_sym_end] = ACTIONS(1374),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_pipeline] = ACTIONS(1376),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(952)] = {
    [sym_comment] = STATE(952),
    [sym_groovy_doc] = STATE(952),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_DOT_AMP] = ACTIONS(369),
    [anon_sym_DOT_AT] = ACTIONS(369),
    [anon_sym_QMARK_DOT] = ACTIONS(369),
    [anon_sym_STAR_DOT] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_DOT] = ACTIONS(371),
    [anon_sym_as] = ACTIONS(369),
    [anon_sym_EQ] = ACTIONS(371),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(369),
    [anon_sym_STAR_EQ] = ACTIONS(369),
    [anon_sym_SLASH_EQ] = ACTIONS(369),
    [anon_sym_PERCENT_EQ] = ACTIONS(369),
    [anon_sym_PLUS_EQ] = ACTIONS(369),
    [anon_sym_DASH_EQ] = ACTIONS(369),
    [anon_sym_LT_LT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(369),
    [anon_sym_AMP_EQ] = ACTIONS(369),
    [anon_sym_CARET_EQ] = ACTIONS(369),
    [anon_sym_PIPE_EQ] = ACTIONS(369),
    [anon_sym_QMARK_EQ] = ACTIONS(369),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_PERCENT] = ACTIONS(371),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_LT_LT] = ACTIONS(371),
    [anon_sym_GT_GT] = ACTIONS(371),
    [anon_sym_GT_GT_GT] = ACTIONS(371),
    [anon_sym_DOT_DOT] = ACTIONS(371),
    [anon_sym_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(369),
    [anon_sym_LT_DOT_DOT] = ACTIONS(371),
    [anon_sym_LT] = ACTIONS(371),
    [anon_sym_LT_EQ] = ACTIONS(371),
    [anon_sym_GT] = ACTIONS(371),
    [anon_sym_GT_EQ] = ACTIONS(369),
    [anon_sym_in] = ACTIONS(371),
    [anon_sym_BANGin] = ACTIONS(371),
    [anon_sym_instanceof] = ACTIONS(369),
    [anon_sym_BANGinstanceof] = ACTIONS(369),
    [anon_sym_EQ_EQ] = ACTIONS(371),
    [anon_sym_BANG_EQ] = ACTIONS(371),
    [anon_sym_LT_EQ_GT] = ACTIONS(369),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(369),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(369),
    [anon_sym_EQ_TILDE] = ACTIONS(369),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(369),
    [anon_sym_AMP] = ACTIONS(371),
    [anon_sym_CARET] = ACTIONS(371),
    [anon_sym_PIPE] = ACTIONS(371),
    [anon_sym_AMP_AMP] = ACTIONS(369),
    [anon_sym_PIPE_PIPE] = ACTIONS(369),
    [anon_sym_QMARK_COLON] = ACTIONS(369),
    [anon_sym_STAR_STAR] = ACTIONS(371),
    [anon_sym_COMMA] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_RPAREN] = ACTIONS(369),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_else] = ACTIONS(369),
    [anon_sym_LBRACK] = ACTIONS(369),
    [anon_sym_QMARK] = ACTIONS(371),
    [anon_sym_catch] = ACTIONS(369),
    [anon_sym_finally] = ACTIONS(369),
  },
  [STATE(953)] = {
    [sym_comment] = STATE(953),
    [sym_groovy_doc] = STATE(953),
    [sym_identifier] = ACTIONS(1798),
    [anon_sym_break] = ACTIONS(1798),
    [anon_sym_continue] = ACTIONS(1798),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym_STAR] = ACTIONS(1798),
    [anon_sym_STAR_COLON] = ACTIONS(1796),
    [anon_sym_import] = ACTIONS(1798),
    [anon_sym_package] = ACTIONS(1798),
    [anon_sym_AT] = ACTIONS(1798),
    [anon_sym_assert] = ACTIONS(1798),
    [anon_sym_PLUS_PLUS] = ACTIONS(1796),
    [anon_sym_DASH_DASH] = ACTIONS(1796),
    [anon_sym_SLASH] = ACTIONS(1798),
    [anon_sym_PLUS] = ACTIONS(1798),
    [anon_sym_DASH] = ACTIONS(1798),
    [anon_sym_true] = ACTIONS(1798),
    [anon_sym_false] = ACTIONS(1798),
    [anon_sym_ATinterface] = ACTIONS(1796),
    [anon_sym_interface] = ACTIONS(1798),
    [anon_sym_class] = ACTIONS(1798),
    [anon_sym_LBRACE] = ACTIONS(1796),
    [anon_sym_RBRACE] = ACTIONS(1796),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1798),
    [anon_sym_def] = ACTIONS(1798),
    [anon_sym_LPAREN] = ACTIONS(1796),
    [anon_sym_null] = ACTIONS(1798),
    [anon_sym_this] = ACTIONS(1798),
    [anon_sym_do] = ACTIONS(1798),
    [anon_sym_while] = ACTIONS(1798),
    [anon_sym_for] = ACTIONS(1798),
    [aux_sym__type_identifier_token1] = ACTIONS(1798),
    [anon_sym_if] = ACTIONS(1798),
    [anon_sym_else] = ACTIONS(1798),
    [anon_sym_LBRACK] = ACTIONS(1796),
    [aux_sym_number_literal_token1] = ACTIONS(1798),
    [aux_sym_number_literal_token2] = ACTIONS(1796),
    [aux_sym_number_literal_token3] = ACTIONS(1796),
    [aux_sym_number_literal_token4] = ACTIONS(1798),
    [aux_sym_number_literal_token5] = ACTIONS(1796),
    [anon_sym_return] = ACTIONS(1798),
    [anon_sym_SQUOTE] = ACTIONS(1798),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1796),
    [anon_sym_DQUOTE] = ACTIONS(1798),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1796),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1796),
    [anon_sym_switch] = ACTIONS(1798),
    [anon_sym_case] = ACTIONS(1798),
    [anon_sym_default] = ACTIONS(1798),
    [anon_sym_try] = ACTIONS(1798),
    [anon_sym_int] = ACTIONS(1798),
    [anon_sym_boolean] = ACTIONS(1798),
    [anon_sym_char] = ACTIONS(1798),
    [anon_sym_short] = ACTIONS(1798),
    [anon_sym_long] = ACTIONS(1798),
    [anon_sym_float] = ACTIONS(1798),
    [anon_sym_double] = ACTIONS(1798),
    [anon_sym_void] = ACTIONS(1798),
    [anon_sym_public] = ACTIONS(1798),
    [anon_sym_protected] = ACTIONS(1798),
    [anon_sym_private] = ACTIONS(1798),
    [anon_sym_static] = ACTIONS(1798),
    [anon_sym_final] = ACTIONS(1798),
    [anon_sym_synchronized] = ACTIONS(1798),
    [anon_sym_TILDE] = ACTIONS(1796),
    [anon_sym_BANG] = ACTIONS(1796),
    [anon_sym_new] = ACTIONS(1798),
  },
  [STATE(954)] = {
    [sym_comment] = STATE(954),
    [sym_groovy_doc] = STATE(954),
    [sym_identifier] = ACTIONS(1802),
    [anon_sym_break] = ACTIONS(1802),
    [anon_sym_continue] = ACTIONS(1802),
    [anon_sym_SEMI] = ACTIONS(1800),
    [anon_sym_STAR] = ACTIONS(1802),
    [anon_sym_STAR_COLON] = ACTIONS(1800),
    [anon_sym_import] = ACTIONS(1802),
    [anon_sym_package] = ACTIONS(1802),
    [anon_sym_AT] = ACTIONS(1802),
    [anon_sym_assert] = ACTIONS(1802),
    [anon_sym_PLUS_PLUS] = ACTIONS(1800),
    [anon_sym_DASH_DASH] = ACTIONS(1800),
    [anon_sym_SLASH] = ACTIONS(1802),
    [anon_sym_PLUS] = ACTIONS(1802),
    [anon_sym_DASH] = ACTIONS(1802),
    [anon_sym_true] = ACTIONS(1802),
    [anon_sym_false] = ACTIONS(1802),
    [anon_sym_ATinterface] = ACTIONS(1800),
    [anon_sym_interface] = ACTIONS(1802),
    [anon_sym_class] = ACTIONS(1802),
    [anon_sym_LBRACE] = ACTIONS(1800),
    [anon_sym_RBRACE] = ACTIONS(1800),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1802),
    [anon_sym_def] = ACTIONS(1802),
    [anon_sym_LPAREN] = ACTIONS(1800),
    [anon_sym_null] = ACTIONS(1802),
    [anon_sym_this] = ACTIONS(1802),
    [anon_sym_do] = ACTIONS(1802),
    [anon_sym_while] = ACTIONS(1802),
    [anon_sym_for] = ACTIONS(1802),
    [aux_sym__type_identifier_token1] = ACTIONS(1802),
    [anon_sym_if] = ACTIONS(1802),
    [anon_sym_else] = ACTIONS(1802),
    [anon_sym_LBRACK] = ACTIONS(1800),
    [aux_sym_number_literal_token1] = ACTIONS(1802),
    [aux_sym_number_literal_token2] = ACTIONS(1800),
    [aux_sym_number_literal_token3] = ACTIONS(1800),
    [aux_sym_number_literal_token4] = ACTIONS(1802),
    [aux_sym_number_literal_token5] = ACTIONS(1800),
    [anon_sym_return] = ACTIONS(1802),
    [anon_sym_SQUOTE] = ACTIONS(1802),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1800),
    [anon_sym_DQUOTE] = ACTIONS(1802),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1800),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1800),
    [anon_sym_switch] = ACTIONS(1802),
    [anon_sym_case] = ACTIONS(1802),
    [anon_sym_default] = ACTIONS(1802),
    [anon_sym_try] = ACTIONS(1802),
    [anon_sym_int] = ACTIONS(1802),
    [anon_sym_boolean] = ACTIONS(1802),
    [anon_sym_char] = ACTIONS(1802),
    [anon_sym_short] = ACTIONS(1802),
    [anon_sym_long] = ACTIONS(1802),
    [anon_sym_float] = ACTIONS(1802),
    [anon_sym_double] = ACTIONS(1802),
    [anon_sym_void] = ACTIONS(1802),
    [anon_sym_public] = ACTIONS(1802),
    [anon_sym_protected] = ACTIONS(1802),
    [anon_sym_private] = ACTIONS(1802),
    [anon_sym_static] = ACTIONS(1802),
    [anon_sym_final] = ACTIONS(1802),
    [anon_sym_synchronized] = ACTIONS(1802),
    [anon_sym_TILDE] = ACTIONS(1800),
    [anon_sym_BANG] = ACTIONS(1800),
    [anon_sym_new] = ACTIONS(1802),
  },
  [STATE(955)] = {
    [sym_comment] = STATE(955),
    [sym_groovy_doc] = STATE(955),
    [sym_identifier] = ACTIONS(1995),
    [anon_sym_break] = ACTIONS(1995),
    [anon_sym_continue] = ACTIONS(1995),
    [anon_sym_SEMI] = ACTIONS(1993),
    [anon_sym_STAR] = ACTIONS(1995),
    [anon_sym_STAR_COLON] = ACTIONS(1993),
    [anon_sym_import] = ACTIONS(1995),
    [anon_sym_package] = ACTIONS(1995),
    [anon_sym_AT] = ACTIONS(1995),
    [anon_sym_assert] = ACTIONS(1995),
    [anon_sym_PLUS_PLUS] = ACTIONS(1993),
    [anon_sym_DASH_DASH] = ACTIONS(1993),
    [anon_sym_SLASH] = ACTIONS(1995),
    [anon_sym_PLUS] = ACTIONS(1995),
    [anon_sym_DASH] = ACTIONS(1995),
    [anon_sym_true] = ACTIONS(1995),
    [anon_sym_false] = ACTIONS(1995),
    [anon_sym_ATinterface] = ACTIONS(1993),
    [anon_sym_interface] = ACTIONS(1995),
    [anon_sym_class] = ACTIONS(1995),
    [anon_sym_LBRACE] = ACTIONS(1993),
    [anon_sym_RBRACE] = ACTIONS(1993),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1995),
    [anon_sym_def] = ACTIONS(1995),
    [anon_sym_LPAREN] = ACTIONS(1993),
    [anon_sym_null] = ACTIONS(1995),
    [anon_sym_this] = ACTIONS(1995),
    [anon_sym_do] = ACTIONS(1995),
    [anon_sym_while] = ACTIONS(1995),
    [anon_sym_for] = ACTIONS(1995),
    [aux_sym__type_identifier_token1] = ACTIONS(1995),
    [anon_sym_if] = ACTIONS(1995),
    [anon_sym_else] = ACTIONS(1995),
    [anon_sym_LBRACK] = ACTIONS(1993),
    [aux_sym_number_literal_token1] = ACTIONS(1995),
    [aux_sym_number_literal_token2] = ACTIONS(1993),
    [aux_sym_number_literal_token3] = ACTIONS(1993),
    [aux_sym_number_literal_token4] = ACTIONS(1995),
    [aux_sym_number_literal_token5] = ACTIONS(1993),
    [anon_sym_return] = ACTIONS(1995),
    [anon_sym_SQUOTE] = ACTIONS(1995),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1993),
    [anon_sym_DQUOTE] = ACTIONS(1995),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1993),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1993),
    [anon_sym_switch] = ACTIONS(1995),
    [anon_sym_case] = ACTIONS(1995),
    [anon_sym_default] = ACTIONS(1995),
    [anon_sym_try] = ACTIONS(1995),
    [anon_sym_int] = ACTIONS(1995),
    [anon_sym_boolean] = ACTIONS(1995),
    [anon_sym_char] = ACTIONS(1995),
    [anon_sym_short] = ACTIONS(1995),
    [anon_sym_long] = ACTIONS(1995),
    [anon_sym_float] = ACTIONS(1995),
    [anon_sym_double] = ACTIONS(1995),
    [anon_sym_void] = ACTIONS(1995),
    [anon_sym_public] = ACTIONS(1995),
    [anon_sym_protected] = ACTIONS(1995),
    [anon_sym_private] = ACTIONS(1995),
    [anon_sym_static] = ACTIONS(1995),
    [anon_sym_final] = ACTIONS(1995),
    [anon_sym_synchronized] = ACTIONS(1995),
    [anon_sym_TILDE] = ACTIONS(1993),
    [anon_sym_BANG] = ACTIONS(1993),
    [anon_sym_new] = ACTIONS(1995),
  },
  [STATE(956)] = {
    [sym_comment] = STATE(956),
    [sym_groovy_doc] = STATE(956),
    [sym_identifier] = ACTIONS(1882),
    [anon_sym_break] = ACTIONS(1882),
    [anon_sym_continue] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_STAR] = ACTIONS(1882),
    [anon_sym_STAR_COLON] = ACTIONS(1880),
    [anon_sym_import] = ACTIONS(1882),
    [anon_sym_package] = ACTIONS(1882),
    [anon_sym_AT] = ACTIONS(1882),
    [anon_sym_assert] = ACTIONS(1882),
    [anon_sym_PLUS_PLUS] = ACTIONS(1880),
    [anon_sym_DASH_DASH] = ACTIONS(1880),
    [anon_sym_SLASH] = ACTIONS(1882),
    [anon_sym_PLUS] = ACTIONS(1882),
    [anon_sym_DASH] = ACTIONS(1882),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [anon_sym_ATinterface] = ACTIONS(1880),
    [anon_sym_interface] = ACTIONS(1882),
    [anon_sym_class] = ACTIONS(1882),
    [anon_sym_LBRACE] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1882),
    [anon_sym_def] = ACTIONS(1882),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_null] = ACTIONS(1882),
    [anon_sym_this] = ACTIONS(1882),
    [anon_sym_do] = ACTIONS(1882),
    [anon_sym_while] = ACTIONS(1882),
    [anon_sym_for] = ACTIONS(1882),
    [aux_sym__type_identifier_token1] = ACTIONS(1882),
    [anon_sym_if] = ACTIONS(1882),
    [anon_sym_else] = ACTIONS(1882),
    [anon_sym_LBRACK] = ACTIONS(1880),
    [aux_sym_number_literal_token1] = ACTIONS(1882),
    [aux_sym_number_literal_token2] = ACTIONS(1880),
    [aux_sym_number_literal_token3] = ACTIONS(1880),
    [aux_sym_number_literal_token4] = ACTIONS(1882),
    [aux_sym_number_literal_token5] = ACTIONS(1880),
    [anon_sym_return] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1882),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1880),
    [anon_sym_DQUOTE] = ACTIONS(1882),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1880),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1880),
    [anon_sym_switch] = ACTIONS(1882),
    [anon_sym_case] = ACTIONS(1882),
    [anon_sym_default] = ACTIONS(1882),
    [anon_sym_try] = ACTIONS(1882),
    [anon_sym_int] = ACTIONS(1882),
    [anon_sym_boolean] = ACTIONS(1882),
    [anon_sym_char] = ACTIONS(1882),
    [anon_sym_short] = ACTIONS(1882),
    [anon_sym_long] = ACTIONS(1882),
    [anon_sym_float] = ACTIONS(1882),
    [anon_sym_double] = ACTIONS(1882),
    [anon_sym_void] = ACTIONS(1882),
    [anon_sym_public] = ACTIONS(1882),
    [anon_sym_protected] = ACTIONS(1882),
    [anon_sym_private] = ACTIONS(1882),
    [anon_sym_static] = ACTIONS(1882),
    [anon_sym_final] = ACTIONS(1882),
    [anon_sym_synchronized] = ACTIONS(1882),
    [anon_sym_TILDE] = ACTIONS(1880),
    [anon_sym_BANG] = ACTIONS(1880),
    [anon_sym_new] = ACTIONS(1882),
  },
  [STATE(957)] = {
    [sym_comment] = STATE(957),
    [sym_groovy_doc] = STATE(957),
    [sym_identifier] = ACTIONS(1838),
    [anon_sym_break] = ACTIONS(1838),
    [anon_sym_continue] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_STAR] = ACTIONS(1838),
    [anon_sym_STAR_COLON] = ACTIONS(1836),
    [anon_sym_import] = ACTIONS(1838),
    [anon_sym_package] = ACTIONS(1838),
    [anon_sym_AT] = ACTIONS(1838),
    [anon_sym_assert] = ACTIONS(1838),
    [anon_sym_PLUS_PLUS] = ACTIONS(1836),
    [anon_sym_DASH_DASH] = ACTIONS(1836),
    [anon_sym_SLASH] = ACTIONS(1838),
    [anon_sym_PLUS] = ACTIONS(1838),
    [anon_sym_DASH] = ACTIONS(1838),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [anon_sym_ATinterface] = ACTIONS(1836),
    [anon_sym_interface] = ACTIONS(1838),
    [anon_sym_class] = ACTIONS(1838),
    [anon_sym_LBRACE] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1838),
    [anon_sym_def] = ACTIONS(1838),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_null] = ACTIONS(1838),
    [anon_sym_this] = ACTIONS(1838),
    [anon_sym_do] = ACTIONS(1838),
    [anon_sym_while] = ACTIONS(1838),
    [anon_sym_for] = ACTIONS(1838),
    [aux_sym__type_identifier_token1] = ACTIONS(1838),
    [anon_sym_if] = ACTIONS(1838),
    [anon_sym_else] = ACTIONS(1838),
    [anon_sym_LBRACK] = ACTIONS(1836),
    [aux_sym_number_literal_token1] = ACTIONS(1838),
    [aux_sym_number_literal_token2] = ACTIONS(1836),
    [aux_sym_number_literal_token3] = ACTIONS(1836),
    [aux_sym_number_literal_token4] = ACTIONS(1838),
    [aux_sym_number_literal_token5] = ACTIONS(1836),
    [anon_sym_return] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1838),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1836),
    [anon_sym_DQUOTE] = ACTIONS(1838),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1836),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1836),
    [anon_sym_switch] = ACTIONS(1838),
    [anon_sym_case] = ACTIONS(1838),
    [anon_sym_default] = ACTIONS(1838),
    [anon_sym_try] = ACTIONS(1838),
    [anon_sym_int] = ACTIONS(1838),
    [anon_sym_boolean] = ACTIONS(1838),
    [anon_sym_char] = ACTIONS(1838),
    [anon_sym_short] = ACTIONS(1838),
    [anon_sym_long] = ACTIONS(1838),
    [anon_sym_float] = ACTIONS(1838),
    [anon_sym_double] = ACTIONS(1838),
    [anon_sym_void] = ACTIONS(1838),
    [anon_sym_public] = ACTIONS(1838),
    [anon_sym_protected] = ACTIONS(1838),
    [anon_sym_private] = ACTIONS(1838),
    [anon_sym_static] = ACTIONS(1838),
    [anon_sym_final] = ACTIONS(1838),
    [anon_sym_synchronized] = ACTIONS(1838),
    [anon_sym_TILDE] = ACTIONS(1836),
    [anon_sym_BANG] = ACTIONS(1836),
    [anon_sym_new] = ACTIONS(1838),
  },
  [STATE(958)] = {
    [sym_comment] = STATE(958),
    [sym_groovy_doc] = STATE(958),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_case] = ACTIONS(1362),
    [anon_sym_default] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(959)] = {
    [sym_comment] = STATE(959),
    [sym_groovy_doc] = STATE(959),
    [sym_identifier] = ACTIONS(1376),
    [anon_sym_break] = ACTIONS(1376),
    [anon_sym_continue] = ACTIONS(1376),
    [anon_sym_SEMI] = ACTIONS(1374),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_STAR_COLON] = ACTIONS(1374),
    [anon_sym_import] = ACTIONS(1376),
    [anon_sym_package] = ACTIONS(1376),
    [anon_sym_AT] = ACTIONS(1376),
    [anon_sym_assert] = ACTIONS(1376),
    [anon_sym_PLUS_PLUS] = ACTIONS(1374),
    [anon_sym_DASH_DASH] = ACTIONS(1374),
    [anon_sym_SLASH] = ACTIONS(1376),
    [anon_sym_PLUS] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_true] = ACTIONS(1376),
    [anon_sym_false] = ACTIONS(1376),
    [anon_sym_ATinterface] = ACTIONS(1374),
    [anon_sym_interface] = ACTIONS(1376),
    [anon_sym_class] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1374),
    [anon_sym_RBRACE] = ACTIONS(1374),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1376),
    [anon_sym_def] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1374),
    [anon_sym_null] = ACTIONS(1376),
    [anon_sym_this] = ACTIONS(1376),
    [anon_sym_do] = ACTIONS(1376),
    [anon_sym_while] = ACTIONS(1376),
    [anon_sym_for] = ACTIONS(1376),
    [aux_sym__type_identifier_token1] = ACTIONS(1376),
    [anon_sym_if] = ACTIONS(1376),
    [anon_sym_else] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1374),
    [aux_sym_number_literal_token1] = ACTIONS(1376),
    [aux_sym_number_literal_token2] = ACTIONS(1374),
    [aux_sym_number_literal_token3] = ACTIONS(1374),
    [aux_sym_number_literal_token4] = ACTIONS(1376),
    [aux_sym_number_literal_token5] = ACTIONS(1374),
    [anon_sym_return] = ACTIONS(1376),
    [anon_sym_SQUOTE] = ACTIONS(1376),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1374),
    [anon_sym_DQUOTE] = ACTIONS(1376),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1374),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1374),
    [anon_sym_switch] = ACTIONS(1376),
    [anon_sym_case] = ACTIONS(1376),
    [anon_sym_default] = ACTIONS(1376),
    [anon_sym_try] = ACTIONS(1376),
    [anon_sym_int] = ACTIONS(1376),
    [anon_sym_boolean] = ACTIONS(1376),
    [anon_sym_char] = ACTIONS(1376),
    [anon_sym_short] = ACTIONS(1376),
    [anon_sym_long] = ACTIONS(1376),
    [anon_sym_float] = ACTIONS(1376),
    [anon_sym_double] = ACTIONS(1376),
    [anon_sym_void] = ACTIONS(1376),
    [anon_sym_public] = ACTIONS(1376),
    [anon_sym_protected] = ACTIONS(1376),
    [anon_sym_private] = ACTIONS(1376),
    [anon_sym_static] = ACTIONS(1376),
    [anon_sym_final] = ACTIONS(1376),
    [anon_sym_synchronized] = ACTIONS(1376),
    [anon_sym_TILDE] = ACTIONS(1374),
    [anon_sym_BANG] = ACTIONS(1374),
    [anon_sym_new] = ACTIONS(1376),
  },
  [STATE(960)] = {
    [sym_comment] = STATE(960),
    [sym_groovy_doc] = STATE(960),
    [sym_identifier] = ACTIONS(1878),
    [anon_sym_break] = ACTIONS(1878),
    [anon_sym_continue] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_STAR] = ACTIONS(1878),
    [anon_sym_STAR_COLON] = ACTIONS(1876),
    [anon_sym_import] = ACTIONS(1878),
    [anon_sym_package] = ACTIONS(1878),
    [anon_sym_AT] = ACTIONS(1878),
    [anon_sym_assert] = ACTIONS(1878),
    [anon_sym_PLUS_PLUS] = ACTIONS(1876),
    [anon_sym_DASH_DASH] = ACTIONS(1876),
    [anon_sym_SLASH] = ACTIONS(1878),
    [anon_sym_PLUS] = ACTIONS(1878),
    [anon_sym_DASH] = ACTIONS(1878),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [anon_sym_ATinterface] = ACTIONS(1876),
    [anon_sym_interface] = ACTIONS(1878),
    [anon_sym_class] = ACTIONS(1878),
    [anon_sym_LBRACE] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1878),
    [anon_sym_def] = ACTIONS(1878),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_null] = ACTIONS(1878),
    [anon_sym_this] = ACTIONS(1878),
    [anon_sym_do] = ACTIONS(1878),
    [anon_sym_while] = ACTIONS(1878),
    [anon_sym_for] = ACTIONS(1878),
    [aux_sym__type_identifier_token1] = ACTIONS(1878),
    [anon_sym_if] = ACTIONS(1878),
    [anon_sym_else] = ACTIONS(1878),
    [anon_sym_LBRACK] = ACTIONS(1876),
    [aux_sym_number_literal_token1] = ACTIONS(1878),
    [aux_sym_number_literal_token2] = ACTIONS(1876),
    [aux_sym_number_literal_token3] = ACTIONS(1876),
    [aux_sym_number_literal_token4] = ACTIONS(1878),
    [aux_sym_number_literal_token5] = ACTIONS(1876),
    [anon_sym_return] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1878),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1876),
    [anon_sym_DQUOTE] = ACTIONS(1878),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1876),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1876),
    [anon_sym_switch] = ACTIONS(1878),
    [anon_sym_case] = ACTIONS(1878),
    [anon_sym_default] = ACTIONS(1878),
    [anon_sym_try] = ACTIONS(1878),
    [anon_sym_int] = ACTIONS(1878),
    [anon_sym_boolean] = ACTIONS(1878),
    [anon_sym_char] = ACTIONS(1878),
    [anon_sym_short] = ACTIONS(1878),
    [anon_sym_long] = ACTIONS(1878),
    [anon_sym_float] = ACTIONS(1878),
    [anon_sym_double] = ACTIONS(1878),
    [anon_sym_void] = ACTIONS(1878),
    [anon_sym_public] = ACTIONS(1878),
    [anon_sym_protected] = ACTIONS(1878),
    [anon_sym_private] = ACTIONS(1878),
    [anon_sym_static] = ACTIONS(1878),
    [anon_sym_final] = ACTIONS(1878),
    [anon_sym_synchronized] = ACTIONS(1878),
    [anon_sym_TILDE] = ACTIONS(1876),
    [anon_sym_BANG] = ACTIONS(1876),
    [anon_sym_new] = ACTIONS(1878),
  },
  [STATE(961)] = {
    [sym_comment] = STATE(961),
    [sym_groovy_doc] = STATE(961),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_case] = ACTIONS(1392),
    [anon_sym_default] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1390),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(962)] = {
    [sym_comment] = STATE(962),
    [sym_groovy_doc] = STATE(962),
    [sym_identifier] = ACTIONS(1846),
    [anon_sym_break] = ACTIONS(1846),
    [anon_sym_continue] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1846),
    [anon_sym_STAR_COLON] = ACTIONS(1844),
    [anon_sym_import] = ACTIONS(1846),
    [anon_sym_package] = ACTIONS(1846),
    [anon_sym_AT] = ACTIONS(1846),
    [anon_sym_assert] = ACTIONS(1846),
    [anon_sym_PLUS_PLUS] = ACTIONS(1844),
    [anon_sym_DASH_DASH] = ACTIONS(1844),
    [anon_sym_SLASH] = ACTIONS(1846),
    [anon_sym_PLUS] = ACTIONS(1846),
    [anon_sym_DASH] = ACTIONS(1846),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [anon_sym_ATinterface] = ACTIONS(1844),
    [anon_sym_interface] = ACTIONS(1846),
    [anon_sym_class] = ACTIONS(1846),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1846),
    [anon_sym_def] = ACTIONS(1846),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_null] = ACTIONS(1846),
    [anon_sym_this] = ACTIONS(1846),
    [anon_sym_do] = ACTIONS(1846),
    [anon_sym_while] = ACTIONS(1846),
    [anon_sym_for] = ACTIONS(1846),
    [aux_sym__type_identifier_token1] = ACTIONS(1846),
    [anon_sym_if] = ACTIONS(1846),
    [anon_sym_else] = ACTIONS(1846),
    [anon_sym_LBRACK] = ACTIONS(1844),
    [aux_sym_number_literal_token1] = ACTIONS(1846),
    [aux_sym_number_literal_token2] = ACTIONS(1844),
    [aux_sym_number_literal_token3] = ACTIONS(1844),
    [aux_sym_number_literal_token4] = ACTIONS(1846),
    [aux_sym_number_literal_token5] = ACTIONS(1844),
    [anon_sym_return] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1846),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1844),
    [anon_sym_DQUOTE] = ACTIONS(1846),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1844),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1844),
    [anon_sym_switch] = ACTIONS(1846),
    [anon_sym_case] = ACTIONS(1846),
    [anon_sym_default] = ACTIONS(1846),
    [anon_sym_try] = ACTIONS(1846),
    [anon_sym_int] = ACTIONS(1846),
    [anon_sym_boolean] = ACTIONS(1846),
    [anon_sym_char] = ACTIONS(1846),
    [anon_sym_short] = ACTIONS(1846),
    [anon_sym_long] = ACTIONS(1846),
    [anon_sym_float] = ACTIONS(1846),
    [anon_sym_double] = ACTIONS(1846),
    [anon_sym_void] = ACTIONS(1846),
    [anon_sym_public] = ACTIONS(1846),
    [anon_sym_protected] = ACTIONS(1846),
    [anon_sym_private] = ACTIONS(1846),
    [anon_sym_static] = ACTIONS(1846),
    [anon_sym_final] = ACTIONS(1846),
    [anon_sym_synchronized] = ACTIONS(1846),
    [anon_sym_TILDE] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1844),
    [anon_sym_new] = ACTIONS(1846),
  },
  [STATE(963)] = {
    [sym_comment] = STATE(963),
    [sym_groovy_doc] = STATE(963),
    [sym_identifier] = ACTIONS(1412),
    [anon_sym_break] = ACTIONS(1412),
    [anon_sym_continue] = ACTIONS(1412),
    [anon_sym_SEMI] = ACTIONS(1410),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_STAR_COLON] = ACTIONS(1410),
    [anon_sym_import] = ACTIONS(1412),
    [anon_sym_package] = ACTIONS(1412),
    [anon_sym_AT] = ACTIONS(1412),
    [anon_sym_assert] = ACTIONS(1412),
    [anon_sym_PLUS_PLUS] = ACTIONS(1410),
    [anon_sym_DASH_DASH] = ACTIONS(1410),
    [anon_sym_SLASH] = ACTIONS(1412),
    [anon_sym_PLUS] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_true] = ACTIONS(1412),
    [anon_sym_false] = ACTIONS(1412),
    [anon_sym_ATinterface] = ACTIONS(1410),
    [anon_sym_interface] = ACTIONS(1412),
    [anon_sym_class] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1410),
    [anon_sym_RBRACE] = ACTIONS(1410),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1412),
    [anon_sym_def] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1410),
    [anon_sym_null] = ACTIONS(1412),
    [anon_sym_this] = ACTIONS(1412),
    [anon_sym_do] = ACTIONS(1412),
    [anon_sym_while] = ACTIONS(1412),
    [anon_sym_for] = ACTIONS(1412),
    [aux_sym__type_identifier_token1] = ACTIONS(1412),
    [anon_sym_if] = ACTIONS(1412),
    [anon_sym_else] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1410),
    [aux_sym_number_literal_token1] = ACTIONS(1412),
    [aux_sym_number_literal_token2] = ACTIONS(1410),
    [aux_sym_number_literal_token3] = ACTIONS(1410),
    [aux_sym_number_literal_token4] = ACTIONS(1412),
    [aux_sym_number_literal_token5] = ACTIONS(1410),
    [anon_sym_return] = ACTIONS(1412),
    [anon_sym_SQUOTE] = ACTIONS(1412),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1410),
    [anon_sym_DQUOTE] = ACTIONS(1412),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1410),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1410),
    [anon_sym_switch] = ACTIONS(1412),
    [anon_sym_case] = ACTIONS(1412),
    [anon_sym_default] = ACTIONS(1412),
    [anon_sym_try] = ACTIONS(1412),
    [anon_sym_int] = ACTIONS(1412),
    [anon_sym_boolean] = ACTIONS(1412),
    [anon_sym_char] = ACTIONS(1412),
    [anon_sym_short] = ACTIONS(1412),
    [anon_sym_long] = ACTIONS(1412),
    [anon_sym_float] = ACTIONS(1412),
    [anon_sym_double] = ACTIONS(1412),
    [anon_sym_void] = ACTIONS(1412),
    [anon_sym_public] = ACTIONS(1412),
    [anon_sym_protected] = ACTIONS(1412),
    [anon_sym_private] = ACTIONS(1412),
    [anon_sym_static] = ACTIONS(1412),
    [anon_sym_final] = ACTIONS(1412),
    [anon_sym_synchronized] = ACTIONS(1412),
    [anon_sym_TILDE] = ACTIONS(1410),
    [anon_sym_BANG] = ACTIONS(1410),
    [anon_sym_new] = ACTIONS(1412),
  },
  [STATE(964)] = {
    [sym_comment] = STATE(964),
    [sym_groovy_doc] = STATE(964),
    [ts_builtin_sym_end] = ACTIONS(1519),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1519),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_else] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_pipeline] = ACTIONS(1517),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(965)] = {
    [sym_comment] = STATE(965),
    [sym_groovy_doc] = STATE(965),
    [sym_identifier] = ACTIONS(1910),
    [anon_sym_break] = ACTIONS(1910),
    [anon_sym_continue] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_STAR] = ACTIONS(1910),
    [anon_sym_STAR_COLON] = ACTIONS(1908),
    [anon_sym_import] = ACTIONS(1910),
    [anon_sym_package] = ACTIONS(1910),
    [anon_sym_AT] = ACTIONS(1910),
    [anon_sym_assert] = ACTIONS(1910),
    [anon_sym_PLUS_PLUS] = ACTIONS(1908),
    [anon_sym_DASH_DASH] = ACTIONS(1908),
    [anon_sym_SLASH] = ACTIONS(1910),
    [anon_sym_PLUS] = ACTIONS(1910),
    [anon_sym_DASH] = ACTIONS(1910),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [anon_sym_ATinterface] = ACTIONS(1908),
    [anon_sym_interface] = ACTIONS(1910),
    [anon_sym_class] = ACTIONS(1910),
    [anon_sym_LBRACE] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1910),
    [anon_sym_def] = ACTIONS(1910),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_null] = ACTIONS(1910),
    [anon_sym_this] = ACTIONS(1910),
    [anon_sym_do] = ACTIONS(1910),
    [anon_sym_while] = ACTIONS(1910),
    [anon_sym_for] = ACTIONS(1910),
    [aux_sym__type_identifier_token1] = ACTIONS(1910),
    [anon_sym_if] = ACTIONS(1910),
    [anon_sym_else] = ACTIONS(1910),
    [anon_sym_LBRACK] = ACTIONS(1908),
    [aux_sym_number_literal_token1] = ACTIONS(1910),
    [aux_sym_number_literal_token2] = ACTIONS(1908),
    [aux_sym_number_literal_token3] = ACTIONS(1908),
    [aux_sym_number_literal_token4] = ACTIONS(1910),
    [aux_sym_number_literal_token5] = ACTIONS(1908),
    [anon_sym_return] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1910),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1908),
    [anon_sym_DQUOTE] = ACTIONS(1910),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1908),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1908),
    [anon_sym_switch] = ACTIONS(1910),
    [anon_sym_case] = ACTIONS(1910),
    [anon_sym_default] = ACTIONS(1910),
    [anon_sym_try] = ACTIONS(1910),
    [anon_sym_int] = ACTIONS(1910),
    [anon_sym_boolean] = ACTIONS(1910),
    [anon_sym_char] = ACTIONS(1910),
    [anon_sym_short] = ACTIONS(1910),
    [anon_sym_long] = ACTIONS(1910),
    [anon_sym_float] = ACTIONS(1910),
    [anon_sym_double] = ACTIONS(1910),
    [anon_sym_void] = ACTIONS(1910),
    [anon_sym_public] = ACTIONS(1910),
    [anon_sym_protected] = ACTIONS(1910),
    [anon_sym_private] = ACTIONS(1910),
    [anon_sym_static] = ACTIONS(1910),
    [anon_sym_final] = ACTIONS(1910),
    [anon_sym_synchronized] = ACTIONS(1910),
    [anon_sym_TILDE] = ACTIONS(1908),
    [anon_sym_BANG] = ACTIONS(1908),
    [anon_sym_new] = ACTIONS(1910),
  },
  [STATE(966)] = {
    [sym_comment] = STATE(966),
    [sym_groovy_doc] = STATE(966),
    [sym_identifier] = ACTIONS(1914),
    [anon_sym_break] = ACTIONS(1914),
    [anon_sym_continue] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_STAR] = ACTIONS(1914),
    [anon_sym_STAR_COLON] = ACTIONS(1912),
    [anon_sym_import] = ACTIONS(1914),
    [anon_sym_package] = ACTIONS(1914),
    [anon_sym_AT] = ACTIONS(1914),
    [anon_sym_assert] = ACTIONS(1914),
    [anon_sym_PLUS_PLUS] = ACTIONS(1912),
    [anon_sym_DASH_DASH] = ACTIONS(1912),
    [anon_sym_SLASH] = ACTIONS(1914),
    [anon_sym_PLUS] = ACTIONS(1914),
    [anon_sym_DASH] = ACTIONS(1914),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [anon_sym_ATinterface] = ACTIONS(1912),
    [anon_sym_interface] = ACTIONS(1914),
    [anon_sym_class] = ACTIONS(1914),
    [anon_sym_LBRACE] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1914),
    [anon_sym_def] = ACTIONS(1914),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_null] = ACTIONS(1914),
    [anon_sym_this] = ACTIONS(1914),
    [anon_sym_do] = ACTIONS(1914),
    [anon_sym_while] = ACTIONS(1914),
    [anon_sym_for] = ACTIONS(1914),
    [aux_sym__type_identifier_token1] = ACTIONS(1914),
    [anon_sym_if] = ACTIONS(1914),
    [anon_sym_else] = ACTIONS(1914),
    [anon_sym_LBRACK] = ACTIONS(1912),
    [aux_sym_number_literal_token1] = ACTIONS(1914),
    [aux_sym_number_literal_token2] = ACTIONS(1912),
    [aux_sym_number_literal_token3] = ACTIONS(1912),
    [aux_sym_number_literal_token4] = ACTIONS(1914),
    [aux_sym_number_literal_token5] = ACTIONS(1912),
    [anon_sym_return] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1914),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1912),
    [anon_sym_DQUOTE] = ACTIONS(1914),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1912),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1912),
    [anon_sym_switch] = ACTIONS(1914),
    [anon_sym_case] = ACTIONS(1914),
    [anon_sym_default] = ACTIONS(1914),
    [anon_sym_try] = ACTIONS(1914),
    [anon_sym_int] = ACTIONS(1914),
    [anon_sym_boolean] = ACTIONS(1914),
    [anon_sym_char] = ACTIONS(1914),
    [anon_sym_short] = ACTIONS(1914),
    [anon_sym_long] = ACTIONS(1914),
    [anon_sym_float] = ACTIONS(1914),
    [anon_sym_double] = ACTIONS(1914),
    [anon_sym_void] = ACTIONS(1914),
    [anon_sym_public] = ACTIONS(1914),
    [anon_sym_protected] = ACTIONS(1914),
    [anon_sym_private] = ACTIONS(1914),
    [anon_sym_static] = ACTIONS(1914),
    [anon_sym_final] = ACTIONS(1914),
    [anon_sym_synchronized] = ACTIONS(1914),
    [anon_sym_TILDE] = ACTIONS(1912),
    [anon_sym_BANG] = ACTIONS(1912),
    [anon_sym_new] = ACTIONS(1914),
  },
  [STATE(967)] = {
    [sym_comment] = STATE(967),
    [sym_groovy_doc] = STATE(967),
    [sym_identifier] = ACTIONS(1854),
    [anon_sym_break] = ACTIONS(1854),
    [anon_sym_continue] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_STAR] = ACTIONS(1854),
    [anon_sym_STAR_COLON] = ACTIONS(1852),
    [anon_sym_import] = ACTIONS(1854),
    [anon_sym_package] = ACTIONS(1854),
    [anon_sym_AT] = ACTIONS(1854),
    [anon_sym_assert] = ACTIONS(1854),
    [anon_sym_PLUS_PLUS] = ACTIONS(1852),
    [anon_sym_DASH_DASH] = ACTIONS(1852),
    [anon_sym_SLASH] = ACTIONS(1854),
    [anon_sym_PLUS] = ACTIONS(1854),
    [anon_sym_DASH] = ACTIONS(1854),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [anon_sym_ATinterface] = ACTIONS(1852),
    [anon_sym_interface] = ACTIONS(1854),
    [anon_sym_class] = ACTIONS(1854),
    [anon_sym_LBRACE] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1854),
    [anon_sym_def] = ACTIONS(1854),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_null] = ACTIONS(1854),
    [anon_sym_this] = ACTIONS(1854),
    [anon_sym_do] = ACTIONS(1854),
    [anon_sym_while] = ACTIONS(1854),
    [anon_sym_for] = ACTIONS(1854),
    [aux_sym__type_identifier_token1] = ACTIONS(1854),
    [anon_sym_if] = ACTIONS(1854),
    [anon_sym_else] = ACTIONS(1854),
    [anon_sym_LBRACK] = ACTIONS(1852),
    [aux_sym_number_literal_token1] = ACTIONS(1854),
    [aux_sym_number_literal_token2] = ACTIONS(1852),
    [aux_sym_number_literal_token3] = ACTIONS(1852),
    [aux_sym_number_literal_token4] = ACTIONS(1854),
    [aux_sym_number_literal_token5] = ACTIONS(1852),
    [anon_sym_return] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1854),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1852),
    [anon_sym_DQUOTE] = ACTIONS(1854),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1852),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1852),
    [anon_sym_switch] = ACTIONS(1854),
    [anon_sym_case] = ACTIONS(1854),
    [anon_sym_default] = ACTIONS(1854),
    [anon_sym_try] = ACTIONS(1854),
    [anon_sym_int] = ACTIONS(1854),
    [anon_sym_boolean] = ACTIONS(1854),
    [anon_sym_char] = ACTIONS(1854),
    [anon_sym_short] = ACTIONS(1854),
    [anon_sym_long] = ACTIONS(1854),
    [anon_sym_float] = ACTIONS(1854),
    [anon_sym_double] = ACTIONS(1854),
    [anon_sym_void] = ACTIONS(1854),
    [anon_sym_public] = ACTIONS(1854),
    [anon_sym_protected] = ACTIONS(1854),
    [anon_sym_private] = ACTIONS(1854),
    [anon_sym_static] = ACTIONS(1854),
    [anon_sym_final] = ACTIONS(1854),
    [anon_sym_synchronized] = ACTIONS(1854),
    [anon_sym_TILDE] = ACTIONS(1852),
    [anon_sym_BANG] = ACTIONS(1852),
    [anon_sym_new] = ACTIONS(1854),
  },
  [STATE(968)] = {
    [sym_comment] = STATE(968),
    [sym_groovy_doc] = STATE(968),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_break] = ACTIONS(1922),
    [anon_sym_continue] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_STAR] = ACTIONS(1922),
    [anon_sym_STAR_COLON] = ACTIONS(1920),
    [anon_sym_import] = ACTIONS(1922),
    [anon_sym_package] = ACTIONS(1922),
    [anon_sym_AT] = ACTIONS(1922),
    [anon_sym_assert] = ACTIONS(1922),
    [anon_sym_PLUS_PLUS] = ACTIONS(1920),
    [anon_sym_DASH_DASH] = ACTIONS(1920),
    [anon_sym_SLASH] = ACTIONS(1922),
    [anon_sym_PLUS] = ACTIONS(1922),
    [anon_sym_DASH] = ACTIONS(1922),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [anon_sym_ATinterface] = ACTIONS(1920),
    [anon_sym_interface] = ACTIONS(1922),
    [anon_sym_class] = ACTIONS(1922),
    [anon_sym_LBRACE] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1922),
    [anon_sym_def] = ACTIONS(1922),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_null] = ACTIONS(1922),
    [anon_sym_this] = ACTIONS(1922),
    [anon_sym_do] = ACTIONS(1922),
    [anon_sym_while] = ACTIONS(1922),
    [anon_sym_for] = ACTIONS(1922),
    [aux_sym__type_identifier_token1] = ACTIONS(1922),
    [anon_sym_if] = ACTIONS(1922),
    [anon_sym_else] = ACTIONS(1922),
    [anon_sym_LBRACK] = ACTIONS(1920),
    [aux_sym_number_literal_token1] = ACTIONS(1922),
    [aux_sym_number_literal_token2] = ACTIONS(1920),
    [aux_sym_number_literal_token3] = ACTIONS(1920),
    [aux_sym_number_literal_token4] = ACTIONS(1922),
    [aux_sym_number_literal_token5] = ACTIONS(1920),
    [anon_sym_return] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1922),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1920),
    [anon_sym_DQUOTE] = ACTIONS(1922),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1920),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1920),
    [anon_sym_switch] = ACTIONS(1922),
    [anon_sym_case] = ACTIONS(1922),
    [anon_sym_default] = ACTIONS(1922),
    [anon_sym_try] = ACTIONS(1922),
    [anon_sym_int] = ACTIONS(1922),
    [anon_sym_boolean] = ACTIONS(1922),
    [anon_sym_char] = ACTIONS(1922),
    [anon_sym_short] = ACTIONS(1922),
    [anon_sym_long] = ACTIONS(1922),
    [anon_sym_float] = ACTIONS(1922),
    [anon_sym_double] = ACTIONS(1922),
    [anon_sym_void] = ACTIONS(1922),
    [anon_sym_public] = ACTIONS(1922),
    [anon_sym_protected] = ACTIONS(1922),
    [anon_sym_private] = ACTIONS(1922),
    [anon_sym_static] = ACTIONS(1922),
    [anon_sym_final] = ACTIONS(1922),
    [anon_sym_synchronized] = ACTIONS(1922),
    [anon_sym_TILDE] = ACTIONS(1920),
    [anon_sym_BANG] = ACTIONS(1920),
    [anon_sym_new] = ACTIONS(1922),
  },
  [STATE(969)] = {
    [sym_comment] = STATE(969),
    [sym_groovy_doc] = STATE(969),
    [ts_builtin_sym_end] = ACTIONS(1390),
    [sym_identifier] = ACTIONS(1392),
    [anon_sym_break] = ACTIONS(1392),
    [anon_sym_continue] = ACTIONS(1392),
    [anon_sym_SEMI] = ACTIONS(1390),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_STAR_COLON] = ACTIONS(1390),
    [anon_sym_import] = ACTIONS(1392),
    [anon_sym_package] = ACTIONS(1392),
    [anon_sym_AT] = ACTIONS(1392),
    [anon_sym_assert] = ACTIONS(1392),
    [anon_sym_PLUS_PLUS] = ACTIONS(1390),
    [anon_sym_DASH_DASH] = ACTIONS(1390),
    [anon_sym_SLASH] = ACTIONS(1392),
    [anon_sym_PLUS] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_true] = ACTIONS(1392),
    [anon_sym_false] = ACTIONS(1392),
    [anon_sym_ATinterface] = ACTIONS(1390),
    [anon_sym_interface] = ACTIONS(1392),
    [anon_sym_class] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1390),
    [anon_sym_RBRACE] = ACTIONS(1390),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1392),
    [anon_sym_def] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1390),
    [anon_sym_null] = ACTIONS(1392),
    [anon_sym_this] = ACTIONS(1392),
    [anon_sym_do] = ACTIONS(1392),
    [anon_sym_while] = ACTIONS(1392),
    [anon_sym_for] = ACTIONS(1392),
    [aux_sym__type_identifier_token1] = ACTIONS(1392),
    [anon_sym_if] = ACTIONS(1392),
    [anon_sym_else] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1390),
    [aux_sym_number_literal_token1] = ACTIONS(1392),
    [aux_sym_number_literal_token2] = ACTIONS(1390),
    [aux_sym_number_literal_token3] = ACTIONS(1390),
    [aux_sym_number_literal_token4] = ACTIONS(1392),
    [aux_sym_number_literal_token5] = ACTIONS(1390),
    [anon_sym_pipeline] = ACTIONS(1392),
    [anon_sym_return] = ACTIONS(1392),
    [anon_sym_SQUOTE] = ACTIONS(1392),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1390),
    [anon_sym_DQUOTE] = ACTIONS(1392),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1390),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1390),
    [anon_sym_switch] = ACTIONS(1392),
    [anon_sym_try] = ACTIONS(1392),
    [anon_sym_int] = ACTIONS(1392),
    [anon_sym_boolean] = ACTIONS(1392),
    [anon_sym_char] = ACTIONS(1392),
    [anon_sym_short] = ACTIONS(1392),
    [anon_sym_long] = ACTIONS(1392),
    [anon_sym_float] = ACTIONS(1392),
    [anon_sym_double] = ACTIONS(1392),
    [anon_sym_void] = ACTIONS(1392),
    [anon_sym_public] = ACTIONS(1392),
    [anon_sym_protected] = ACTIONS(1392),
    [anon_sym_private] = ACTIONS(1392),
    [anon_sym_static] = ACTIONS(1392),
    [anon_sym_final] = ACTIONS(1392),
    [anon_sym_synchronized] = ACTIONS(1392),
    [anon_sym_TILDE] = ACTIONS(1390),
    [anon_sym_BANG] = ACTIONS(1390),
    [anon_sym_new] = ACTIONS(1392),
  },
  [STATE(970)] = {
    [sym_comment] = STATE(970),
    [sym_groovy_doc] = STATE(970),
    [sym_identifier] = ACTIONS(1991),
    [anon_sym_break] = ACTIONS(1991),
    [anon_sym_continue] = ACTIONS(1991),
    [anon_sym_SEMI] = ACTIONS(1989),
    [anon_sym_STAR] = ACTIONS(1991),
    [anon_sym_STAR_COLON] = ACTIONS(1989),
    [anon_sym_import] = ACTIONS(1991),
    [anon_sym_package] = ACTIONS(1991),
    [anon_sym_AT] = ACTIONS(1991),
    [anon_sym_assert] = ACTIONS(1991),
    [anon_sym_PLUS_PLUS] = ACTIONS(1989),
    [anon_sym_DASH_DASH] = ACTIONS(1989),
    [anon_sym_SLASH] = ACTIONS(1991),
    [anon_sym_PLUS] = ACTIONS(1991),
    [anon_sym_DASH] = ACTIONS(1991),
    [anon_sym_true] = ACTIONS(1991),
    [anon_sym_false] = ACTIONS(1991),
    [anon_sym_ATinterface] = ACTIONS(1989),
    [anon_sym_interface] = ACTIONS(1991),
    [anon_sym_class] = ACTIONS(1991),
    [anon_sym_LBRACE] = ACTIONS(1989),
    [anon_sym_RBRACE] = ACTIONS(1989),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1991),
    [anon_sym_def] = ACTIONS(1991),
    [anon_sym_LPAREN] = ACTIONS(1989),
    [anon_sym_null] = ACTIONS(1991),
    [anon_sym_this] = ACTIONS(1991),
    [anon_sym_do] = ACTIONS(1991),
    [anon_sym_while] = ACTIONS(1991),
    [anon_sym_for] = ACTIONS(1991),
    [aux_sym__type_identifier_token1] = ACTIONS(1991),
    [anon_sym_if] = ACTIONS(1991),
    [anon_sym_else] = ACTIONS(1991),
    [anon_sym_LBRACK] = ACTIONS(1989),
    [aux_sym_number_literal_token1] = ACTIONS(1991),
    [aux_sym_number_literal_token2] = ACTIONS(1989),
    [aux_sym_number_literal_token3] = ACTIONS(1989),
    [aux_sym_number_literal_token4] = ACTIONS(1991),
    [aux_sym_number_literal_token5] = ACTIONS(1989),
    [anon_sym_return] = ACTIONS(1991),
    [anon_sym_SQUOTE] = ACTIONS(1991),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1989),
    [anon_sym_DQUOTE] = ACTIONS(1991),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1989),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1989),
    [anon_sym_switch] = ACTIONS(1991),
    [anon_sym_case] = ACTIONS(1991),
    [anon_sym_default] = ACTIONS(1991),
    [anon_sym_try] = ACTIONS(1991),
    [anon_sym_int] = ACTIONS(1991),
    [anon_sym_boolean] = ACTIONS(1991),
    [anon_sym_char] = ACTIONS(1991),
    [anon_sym_short] = ACTIONS(1991),
    [anon_sym_long] = ACTIONS(1991),
    [anon_sym_float] = ACTIONS(1991),
    [anon_sym_double] = ACTIONS(1991),
    [anon_sym_void] = ACTIONS(1991),
    [anon_sym_public] = ACTIONS(1991),
    [anon_sym_protected] = ACTIONS(1991),
    [anon_sym_private] = ACTIONS(1991),
    [anon_sym_static] = ACTIONS(1991),
    [anon_sym_final] = ACTIONS(1991),
    [anon_sym_synchronized] = ACTIONS(1991),
    [anon_sym_TILDE] = ACTIONS(1989),
    [anon_sym_BANG] = ACTIONS(1989),
    [anon_sym_new] = ACTIONS(1991),
  },
  [STATE(971)] = {
    [sym_comment] = STATE(971),
    [sym_groovy_doc] = STATE(971),
    [sym_identifier] = ACTIONS(1987),
    [anon_sym_break] = ACTIONS(1987),
    [anon_sym_continue] = ACTIONS(1987),
    [anon_sym_SEMI] = ACTIONS(1985),
    [anon_sym_STAR] = ACTIONS(1987),
    [anon_sym_STAR_COLON] = ACTIONS(1985),
    [anon_sym_import] = ACTIONS(1987),
    [anon_sym_package] = ACTIONS(1987),
    [anon_sym_AT] = ACTIONS(1987),
    [anon_sym_assert] = ACTIONS(1987),
    [anon_sym_PLUS_PLUS] = ACTIONS(1985),
    [anon_sym_DASH_DASH] = ACTIONS(1985),
    [anon_sym_SLASH] = ACTIONS(1987),
    [anon_sym_PLUS] = ACTIONS(1987),
    [anon_sym_DASH] = ACTIONS(1987),
    [anon_sym_true] = ACTIONS(1987),
    [anon_sym_false] = ACTIONS(1987),
    [anon_sym_ATinterface] = ACTIONS(1985),
    [anon_sym_interface] = ACTIONS(1987),
    [anon_sym_class] = ACTIONS(1987),
    [anon_sym_LBRACE] = ACTIONS(1985),
    [anon_sym_RBRACE] = ACTIONS(1985),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1987),
    [anon_sym_def] = ACTIONS(1987),
    [anon_sym_LPAREN] = ACTIONS(1985),
    [anon_sym_null] = ACTIONS(1987),
    [anon_sym_this] = ACTIONS(1987),
    [anon_sym_do] = ACTIONS(1987),
    [anon_sym_while] = ACTIONS(1987),
    [anon_sym_for] = ACTIONS(1987),
    [aux_sym__type_identifier_token1] = ACTIONS(1987),
    [anon_sym_if] = ACTIONS(1987),
    [anon_sym_else] = ACTIONS(1987),
    [anon_sym_LBRACK] = ACTIONS(1985),
    [aux_sym_number_literal_token1] = ACTIONS(1987),
    [aux_sym_number_literal_token2] = ACTIONS(1985),
    [aux_sym_number_literal_token3] = ACTIONS(1985),
    [aux_sym_number_literal_token4] = ACTIONS(1987),
    [aux_sym_number_literal_token5] = ACTIONS(1985),
    [anon_sym_return] = ACTIONS(1987),
    [anon_sym_SQUOTE] = ACTIONS(1987),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1985),
    [anon_sym_DQUOTE] = ACTIONS(1987),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1985),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1985),
    [anon_sym_switch] = ACTIONS(1987),
    [anon_sym_case] = ACTIONS(1987),
    [anon_sym_default] = ACTIONS(1987),
    [anon_sym_try] = ACTIONS(1987),
    [anon_sym_int] = ACTIONS(1987),
    [anon_sym_boolean] = ACTIONS(1987),
    [anon_sym_char] = ACTIONS(1987),
    [anon_sym_short] = ACTIONS(1987),
    [anon_sym_long] = ACTIONS(1987),
    [anon_sym_float] = ACTIONS(1987),
    [anon_sym_double] = ACTIONS(1987),
    [anon_sym_void] = ACTIONS(1987),
    [anon_sym_public] = ACTIONS(1987),
    [anon_sym_protected] = ACTIONS(1987),
    [anon_sym_private] = ACTIONS(1987),
    [anon_sym_static] = ACTIONS(1987),
    [anon_sym_final] = ACTIONS(1987),
    [anon_sym_synchronized] = ACTIONS(1987),
    [anon_sym_TILDE] = ACTIONS(1985),
    [anon_sym_BANG] = ACTIONS(1985),
    [anon_sym_new] = ACTIONS(1987),
  },
  [STATE(972)] = {
    [sym_comment] = STATE(972),
    [sym_groovy_doc] = STATE(972),
    [sym_identifier] = ACTIONS(2027),
    [anon_sym_break] = ACTIONS(2027),
    [anon_sym_continue] = ACTIONS(2027),
    [anon_sym_SEMI] = ACTIONS(2025),
    [anon_sym_STAR] = ACTIONS(2027),
    [anon_sym_STAR_COLON] = ACTIONS(2025),
    [anon_sym_import] = ACTIONS(2027),
    [anon_sym_package] = ACTIONS(2027),
    [anon_sym_AT] = ACTIONS(2027),
    [anon_sym_assert] = ACTIONS(2027),
    [anon_sym_PLUS_PLUS] = ACTIONS(2025),
    [anon_sym_DASH_DASH] = ACTIONS(2025),
    [anon_sym_SLASH] = ACTIONS(2027),
    [anon_sym_PLUS] = ACTIONS(2027),
    [anon_sym_DASH] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(2027),
    [anon_sym_false] = ACTIONS(2027),
    [anon_sym_ATinterface] = ACTIONS(2025),
    [anon_sym_interface] = ACTIONS(2027),
    [anon_sym_class] = ACTIONS(2027),
    [anon_sym_LBRACE] = ACTIONS(2025),
    [anon_sym_RBRACE] = ACTIONS(2025),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2027),
    [anon_sym_def] = ACTIONS(2027),
    [anon_sym_LPAREN] = ACTIONS(2025),
    [anon_sym_null] = ACTIONS(2027),
    [anon_sym_this] = ACTIONS(2027),
    [anon_sym_do] = ACTIONS(2027),
    [anon_sym_while] = ACTIONS(2027),
    [anon_sym_for] = ACTIONS(2027),
    [aux_sym__type_identifier_token1] = ACTIONS(2027),
    [anon_sym_if] = ACTIONS(2027),
    [anon_sym_else] = ACTIONS(2027),
    [anon_sym_LBRACK] = ACTIONS(2025),
    [aux_sym_number_literal_token1] = ACTIONS(2027),
    [aux_sym_number_literal_token2] = ACTIONS(2025),
    [aux_sym_number_literal_token3] = ACTIONS(2025),
    [aux_sym_number_literal_token4] = ACTIONS(2027),
    [aux_sym_number_literal_token5] = ACTIONS(2025),
    [anon_sym_return] = ACTIONS(2027),
    [anon_sym_SQUOTE] = ACTIONS(2027),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2025),
    [anon_sym_DQUOTE] = ACTIONS(2027),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2025),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2025),
    [anon_sym_switch] = ACTIONS(2027),
    [anon_sym_case] = ACTIONS(2027),
    [anon_sym_default] = ACTIONS(2027),
    [anon_sym_try] = ACTIONS(2027),
    [anon_sym_int] = ACTIONS(2027),
    [anon_sym_boolean] = ACTIONS(2027),
    [anon_sym_char] = ACTIONS(2027),
    [anon_sym_short] = ACTIONS(2027),
    [anon_sym_long] = ACTIONS(2027),
    [anon_sym_float] = ACTIONS(2027),
    [anon_sym_double] = ACTIONS(2027),
    [anon_sym_void] = ACTIONS(2027),
    [anon_sym_public] = ACTIONS(2027),
    [anon_sym_protected] = ACTIONS(2027),
    [anon_sym_private] = ACTIONS(2027),
    [anon_sym_static] = ACTIONS(2027),
    [anon_sym_final] = ACTIONS(2027),
    [anon_sym_synchronized] = ACTIONS(2027),
    [anon_sym_TILDE] = ACTIONS(2025),
    [anon_sym_BANG] = ACTIONS(2025),
    [anon_sym_new] = ACTIONS(2027),
  },
  [STATE(973)] = {
    [sym_comment] = STATE(973),
    [sym_groovy_doc] = STATE(973),
    [sym_identifier] = ACTIONS(2001),
    [anon_sym_break] = ACTIONS(2001),
    [anon_sym_continue] = ACTIONS(2001),
    [anon_sym_SEMI] = ACTIONS(1999),
    [anon_sym_STAR] = ACTIONS(2001),
    [anon_sym_STAR_COLON] = ACTIONS(1999),
    [anon_sym_import] = ACTIONS(2001),
    [anon_sym_package] = ACTIONS(2001),
    [anon_sym_AT] = ACTIONS(2001),
    [anon_sym_assert] = ACTIONS(2001),
    [anon_sym_PLUS_PLUS] = ACTIONS(1999),
    [anon_sym_DASH_DASH] = ACTIONS(1999),
    [anon_sym_SLASH] = ACTIONS(2001),
    [anon_sym_PLUS] = ACTIONS(2001),
    [anon_sym_DASH] = ACTIONS(2001),
    [anon_sym_true] = ACTIONS(2001),
    [anon_sym_false] = ACTIONS(2001),
    [anon_sym_ATinterface] = ACTIONS(1999),
    [anon_sym_interface] = ACTIONS(2001),
    [anon_sym_class] = ACTIONS(2001),
    [anon_sym_LBRACE] = ACTIONS(1999),
    [anon_sym_RBRACE] = ACTIONS(1999),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2001),
    [anon_sym_def] = ACTIONS(2001),
    [anon_sym_LPAREN] = ACTIONS(1999),
    [anon_sym_null] = ACTIONS(2001),
    [anon_sym_this] = ACTIONS(2001),
    [anon_sym_do] = ACTIONS(2001),
    [anon_sym_while] = ACTIONS(2001),
    [anon_sym_for] = ACTIONS(2001),
    [aux_sym__type_identifier_token1] = ACTIONS(2001),
    [anon_sym_if] = ACTIONS(2001),
    [anon_sym_else] = ACTIONS(2001),
    [anon_sym_LBRACK] = ACTIONS(1999),
    [aux_sym_number_literal_token1] = ACTIONS(2001),
    [aux_sym_number_literal_token2] = ACTIONS(1999),
    [aux_sym_number_literal_token3] = ACTIONS(1999),
    [aux_sym_number_literal_token4] = ACTIONS(2001),
    [aux_sym_number_literal_token5] = ACTIONS(1999),
    [anon_sym_return] = ACTIONS(2001),
    [anon_sym_SQUOTE] = ACTIONS(2001),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1999),
    [anon_sym_DQUOTE] = ACTIONS(2001),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1999),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1999),
    [anon_sym_switch] = ACTIONS(2001),
    [anon_sym_case] = ACTIONS(2001),
    [anon_sym_default] = ACTIONS(2001),
    [anon_sym_try] = ACTIONS(2001),
    [anon_sym_int] = ACTIONS(2001),
    [anon_sym_boolean] = ACTIONS(2001),
    [anon_sym_char] = ACTIONS(2001),
    [anon_sym_short] = ACTIONS(2001),
    [anon_sym_long] = ACTIONS(2001),
    [anon_sym_float] = ACTIONS(2001),
    [anon_sym_double] = ACTIONS(2001),
    [anon_sym_void] = ACTIONS(2001),
    [anon_sym_public] = ACTIONS(2001),
    [anon_sym_protected] = ACTIONS(2001),
    [anon_sym_private] = ACTIONS(2001),
    [anon_sym_static] = ACTIONS(2001),
    [anon_sym_final] = ACTIONS(2001),
    [anon_sym_synchronized] = ACTIONS(2001),
    [anon_sym_TILDE] = ACTIONS(1999),
    [anon_sym_BANG] = ACTIONS(1999),
    [anon_sym_new] = ACTIONS(2001),
  },
  [STATE(974)] = {
    [sym_comment] = STATE(974),
    [sym_groovy_doc] = STATE(974),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_break] = ACTIONS(1926),
    [anon_sym_continue] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_STAR] = ACTIONS(1926),
    [anon_sym_STAR_COLON] = ACTIONS(1924),
    [anon_sym_import] = ACTIONS(1926),
    [anon_sym_package] = ACTIONS(1926),
    [anon_sym_AT] = ACTIONS(1926),
    [anon_sym_assert] = ACTIONS(1926),
    [anon_sym_PLUS_PLUS] = ACTIONS(1924),
    [anon_sym_DASH_DASH] = ACTIONS(1924),
    [anon_sym_SLASH] = ACTIONS(1926),
    [anon_sym_PLUS] = ACTIONS(1926),
    [anon_sym_DASH] = ACTIONS(1926),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [anon_sym_ATinterface] = ACTIONS(1924),
    [anon_sym_interface] = ACTIONS(1926),
    [anon_sym_class] = ACTIONS(1926),
    [anon_sym_LBRACE] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1926),
    [anon_sym_def] = ACTIONS(1926),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_null] = ACTIONS(1926),
    [anon_sym_this] = ACTIONS(1926),
    [anon_sym_do] = ACTIONS(1926),
    [anon_sym_while] = ACTIONS(1926),
    [anon_sym_for] = ACTIONS(1926),
    [aux_sym__type_identifier_token1] = ACTIONS(1926),
    [anon_sym_if] = ACTIONS(1926),
    [anon_sym_else] = ACTIONS(1926),
    [anon_sym_LBRACK] = ACTIONS(1924),
    [aux_sym_number_literal_token1] = ACTIONS(1926),
    [aux_sym_number_literal_token2] = ACTIONS(1924),
    [aux_sym_number_literal_token3] = ACTIONS(1924),
    [aux_sym_number_literal_token4] = ACTIONS(1926),
    [aux_sym_number_literal_token5] = ACTIONS(1924),
    [anon_sym_return] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1924),
    [anon_sym_DQUOTE] = ACTIONS(1926),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1924),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1924),
    [anon_sym_switch] = ACTIONS(1926),
    [anon_sym_case] = ACTIONS(1926),
    [anon_sym_default] = ACTIONS(1926),
    [anon_sym_try] = ACTIONS(1926),
    [anon_sym_int] = ACTIONS(1926),
    [anon_sym_boolean] = ACTIONS(1926),
    [anon_sym_char] = ACTIONS(1926),
    [anon_sym_short] = ACTIONS(1926),
    [anon_sym_long] = ACTIONS(1926),
    [anon_sym_float] = ACTIONS(1926),
    [anon_sym_double] = ACTIONS(1926),
    [anon_sym_void] = ACTIONS(1926),
    [anon_sym_public] = ACTIONS(1926),
    [anon_sym_protected] = ACTIONS(1926),
    [anon_sym_private] = ACTIONS(1926),
    [anon_sym_static] = ACTIONS(1926),
    [anon_sym_final] = ACTIONS(1926),
    [anon_sym_synchronized] = ACTIONS(1926),
    [anon_sym_TILDE] = ACTIONS(1924),
    [anon_sym_BANG] = ACTIONS(1924),
    [anon_sym_new] = ACTIONS(1926),
  },
  [STATE(975)] = {
    [sym_comment] = STATE(975),
    [sym_groovy_doc] = STATE(975),
    [sym_identifier] = ACTIONS(1785),
    [anon_sym_break] = ACTIONS(1785),
    [anon_sym_continue] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_STAR] = ACTIONS(1785),
    [anon_sym_STAR_COLON] = ACTIONS(1783),
    [anon_sym_import] = ACTIONS(1785),
    [anon_sym_package] = ACTIONS(1785),
    [anon_sym_AT] = ACTIONS(1785),
    [anon_sym_assert] = ACTIONS(1785),
    [anon_sym_PLUS_PLUS] = ACTIONS(1783),
    [anon_sym_DASH_DASH] = ACTIONS(1783),
    [anon_sym_SLASH] = ACTIONS(1785),
    [anon_sym_PLUS] = ACTIONS(1785),
    [anon_sym_DASH] = ACTIONS(1785),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_ATinterface] = ACTIONS(1783),
    [anon_sym_interface] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1785),
    [anon_sym_def] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_null] = ACTIONS(1785),
    [anon_sym_this] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_while] = ACTIONS(1785),
    [anon_sym_for] = ACTIONS(1785),
    [aux_sym__type_identifier_token1] = ACTIONS(1785),
    [anon_sym_if] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1783),
    [aux_sym_number_literal_token1] = ACTIONS(1785),
    [aux_sym_number_literal_token2] = ACTIONS(1783),
    [aux_sym_number_literal_token3] = ACTIONS(1783),
    [aux_sym_number_literal_token4] = ACTIONS(1785),
    [aux_sym_number_literal_token5] = ACTIONS(1783),
    [anon_sym_return] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1785),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1785),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1783),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1783),
    [anon_sym_switch] = ACTIONS(1785),
    [anon_sym_case] = ACTIONS(1785),
    [anon_sym_default] = ACTIONS(1785),
    [anon_sym_try] = ACTIONS(1785),
    [anon_sym_int] = ACTIONS(1785),
    [anon_sym_boolean] = ACTIONS(1785),
    [anon_sym_char] = ACTIONS(1785),
    [anon_sym_short] = ACTIONS(1785),
    [anon_sym_long] = ACTIONS(1785),
    [anon_sym_float] = ACTIONS(1785),
    [anon_sym_double] = ACTIONS(1785),
    [anon_sym_void] = ACTIONS(1785),
    [anon_sym_public] = ACTIONS(1785),
    [anon_sym_protected] = ACTIONS(1785),
    [anon_sym_private] = ACTIONS(1785),
    [anon_sym_static] = ACTIONS(1785),
    [anon_sym_final] = ACTIONS(1785),
    [anon_sym_synchronized] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1783),
    [anon_sym_BANG] = ACTIONS(1783),
    [anon_sym_new] = ACTIONS(1785),
  },
  [STATE(976)] = {
    [sym_comment] = STATE(976),
    [sym_groovy_doc] = STATE(976),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_break] = ACTIONS(1858),
    [anon_sym_continue] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_STAR] = ACTIONS(1858),
    [anon_sym_STAR_COLON] = ACTIONS(1856),
    [anon_sym_import] = ACTIONS(1858),
    [anon_sym_package] = ACTIONS(1858),
    [anon_sym_AT] = ACTIONS(1858),
    [anon_sym_assert] = ACTIONS(1858),
    [anon_sym_PLUS_PLUS] = ACTIONS(1856),
    [anon_sym_DASH_DASH] = ACTIONS(1856),
    [anon_sym_SLASH] = ACTIONS(1858),
    [anon_sym_PLUS] = ACTIONS(1858),
    [anon_sym_DASH] = ACTIONS(1858),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [anon_sym_ATinterface] = ACTIONS(1856),
    [anon_sym_interface] = ACTIONS(1858),
    [anon_sym_class] = ACTIONS(1858),
    [anon_sym_LBRACE] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1858),
    [anon_sym_def] = ACTIONS(1858),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_null] = ACTIONS(1858),
    [anon_sym_this] = ACTIONS(1858),
    [anon_sym_do] = ACTIONS(1858),
    [anon_sym_while] = ACTIONS(1858),
    [anon_sym_for] = ACTIONS(1858),
    [aux_sym__type_identifier_token1] = ACTIONS(1858),
    [anon_sym_if] = ACTIONS(1858),
    [anon_sym_else] = ACTIONS(1858),
    [anon_sym_LBRACK] = ACTIONS(1856),
    [aux_sym_number_literal_token1] = ACTIONS(1858),
    [aux_sym_number_literal_token2] = ACTIONS(1856),
    [aux_sym_number_literal_token3] = ACTIONS(1856),
    [aux_sym_number_literal_token4] = ACTIONS(1858),
    [aux_sym_number_literal_token5] = ACTIONS(1856),
    [anon_sym_return] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1858),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1856),
    [anon_sym_DQUOTE] = ACTIONS(1858),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1856),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1856),
    [anon_sym_switch] = ACTIONS(1858),
    [anon_sym_case] = ACTIONS(1858),
    [anon_sym_default] = ACTIONS(1858),
    [anon_sym_try] = ACTIONS(1858),
    [anon_sym_int] = ACTIONS(1858),
    [anon_sym_boolean] = ACTIONS(1858),
    [anon_sym_char] = ACTIONS(1858),
    [anon_sym_short] = ACTIONS(1858),
    [anon_sym_long] = ACTIONS(1858),
    [anon_sym_float] = ACTIONS(1858),
    [anon_sym_double] = ACTIONS(1858),
    [anon_sym_void] = ACTIONS(1858),
    [anon_sym_public] = ACTIONS(1858),
    [anon_sym_protected] = ACTIONS(1858),
    [anon_sym_private] = ACTIONS(1858),
    [anon_sym_static] = ACTIONS(1858),
    [anon_sym_final] = ACTIONS(1858),
    [anon_sym_synchronized] = ACTIONS(1858),
    [anon_sym_TILDE] = ACTIONS(1856),
    [anon_sym_BANG] = ACTIONS(1856),
    [anon_sym_new] = ACTIONS(1858),
  },
  [STATE(977)] = {
    [sym_comment] = STATE(977),
    [sym_groovy_doc] = STATE(977),
    [sym_identifier] = ACTIONS(1930),
    [anon_sym_break] = ACTIONS(1930),
    [anon_sym_continue] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_STAR] = ACTIONS(1930),
    [anon_sym_STAR_COLON] = ACTIONS(1928),
    [anon_sym_import] = ACTIONS(1930),
    [anon_sym_package] = ACTIONS(1930),
    [anon_sym_AT] = ACTIONS(1930),
    [anon_sym_assert] = ACTIONS(1930),
    [anon_sym_PLUS_PLUS] = ACTIONS(1928),
    [anon_sym_DASH_DASH] = ACTIONS(1928),
    [anon_sym_SLASH] = ACTIONS(1930),
    [anon_sym_PLUS] = ACTIONS(1930),
    [anon_sym_DASH] = ACTIONS(1930),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [anon_sym_ATinterface] = ACTIONS(1928),
    [anon_sym_interface] = ACTIONS(1930),
    [anon_sym_class] = ACTIONS(1930),
    [anon_sym_LBRACE] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1930),
    [anon_sym_def] = ACTIONS(1930),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_null] = ACTIONS(1930),
    [anon_sym_this] = ACTIONS(1930),
    [anon_sym_do] = ACTIONS(1930),
    [anon_sym_while] = ACTIONS(1930),
    [anon_sym_for] = ACTIONS(1930),
    [aux_sym__type_identifier_token1] = ACTIONS(1930),
    [anon_sym_if] = ACTIONS(1930),
    [anon_sym_else] = ACTIONS(1930),
    [anon_sym_LBRACK] = ACTIONS(1928),
    [aux_sym_number_literal_token1] = ACTIONS(1930),
    [aux_sym_number_literal_token2] = ACTIONS(1928),
    [aux_sym_number_literal_token3] = ACTIONS(1928),
    [aux_sym_number_literal_token4] = ACTIONS(1930),
    [aux_sym_number_literal_token5] = ACTIONS(1928),
    [anon_sym_return] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1930),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1928),
    [anon_sym_DQUOTE] = ACTIONS(1930),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1928),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1928),
    [anon_sym_switch] = ACTIONS(1930),
    [anon_sym_case] = ACTIONS(1930),
    [anon_sym_default] = ACTIONS(1930),
    [anon_sym_try] = ACTIONS(1930),
    [anon_sym_int] = ACTIONS(1930),
    [anon_sym_boolean] = ACTIONS(1930),
    [anon_sym_char] = ACTIONS(1930),
    [anon_sym_short] = ACTIONS(1930),
    [anon_sym_long] = ACTIONS(1930),
    [anon_sym_float] = ACTIONS(1930),
    [anon_sym_double] = ACTIONS(1930),
    [anon_sym_void] = ACTIONS(1930),
    [anon_sym_public] = ACTIONS(1930),
    [anon_sym_protected] = ACTIONS(1930),
    [anon_sym_private] = ACTIONS(1930),
    [anon_sym_static] = ACTIONS(1930),
    [anon_sym_final] = ACTIONS(1930),
    [anon_sym_synchronized] = ACTIONS(1930),
    [anon_sym_TILDE] = ACTIONS(1928),
    [anon_sym_BANG] = ACTIONS(1928),
    [anon_sym_new] = ACTIONS(1930),
  },
  [STATE(978)] = {
    [sym_comment] = STATE(978),
    [sym_groovy_doc] = STATE(978),
    [sym_identifier] = ACTIONS(1866),
    [anon_sym_break] = ACTIONS(1866),
    [anon_sym_continue] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_STAR] = ACTIONS(1866),
    [anon_sym_STAR_COLON] = ACTIONS(1864),
    [anon_sym_import] = ACTIONS(1866),
    [anon_sym_package] = ACTIONS(1866),
    [anon_sym_AT] = ACTIONS(1866),
    [anon_sym_assert] = ACTIONS(1866),
    [anon_sym_PLUS_PLUS] = ACTIONS(1864),
    [anon_sym_DASH_DASH] = ACTIONS(1864),
    [anon_sym_SLASH] = ACTIONS(1866),
    [anon_sym_PLUS] = ACTIONS(1866),
    [anon_sym_DASH] = ACTIONS(1866),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [anon_sym_ATinterface] = ACTIONS(1864),
    [anon_sym_interface] = ACTIONS(1866),
    [anon_sym_class] = ACTIONS(1866),
    [anon_sym_LBRACE] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1866),
    [anon_sym_def] = ACTIONS(1866),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_null] = ACTIONS(1866),
    [anon_sym_this] = ACTIONS(1866),
    [anon_sym_do] = ACTIONS(1866),
    [anon_sym_while] = ACTIONS(1866),
    [anon_sym_for] = ACTIONS(1866),
    [aux_sym__type_identifier_token1] = ACTIONS(1866),
    [anon_sym_if] = ACTIONS(1866),
    [anon_sym_else] = ACTIONS(1866),
    [anon_sym_LBRACK] = ACTIONS(1864),
    [aux_sym_number_literal_token1] = ACTIONS(1866),
    [aux_sym_number_literal_token2] = ACTIONS(1864),
    [aux_sym_number_literal_token3] = ACTIONS(1864),
    [aux_sym_number_literal_token4] = ACTIONS(1866),
    [aux_sym_number_literal_token5] = ACTIONS(1864),
    [anon_sym_return] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1866),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1864),
    [anon_sym_DQUOTE] = ACTIONS(1866),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1864),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1864),
    [anon_sym_switch] = ACTIONS(1866),
    [anon_sym_case] = ACTIONS(1866),
    [anon_sym_default] = ACTIONS(1866),
    [anon_sym_try] = ACTIONS(1866),
    [anon_sym_int] = ACTIONS(1866),
    [anon_sym_boolean] = ACTIONS(1866),
    [anon_sym_char] = ACTIONS(1866),
    [anon_sym_short] = ACTIONS(1866),
    [anon_sym_long] = ACTIONS(1866),
    [anon_sym_float] = ACTIONS(1866),
    [anon_sym_double] = ACTIONS(1866),
    [anon_sym_void] = ACTIONS(1866),
    [anon_sym_public] = ACTIONS(1866),
    [anon_sym_protected] = ACTIONS(1866),
    [anon_sym_private] = ACTIONS(1866),
    [anon_sym_static] = ACTIONS(1866),
    [anon_sym_final] = ACTIONS(1866),
    [anon_sym_synchronized] = ACTIONS(1866),
    [anon_sym_TILDE] = ACTIONS(1864),
    [anon_sym_BANG] = ACTIONS(1864),
    [anon_sym_new] = ACTIONS(1866),
  },
  [STATE(979)] = {
    [sym_comment] = STATE(979),
    [sym_groovy_doc] = STATE(979),
    [sym_identifier] = ACTIONS(1934),
    [anon_sym_break] = ACTIONS(1934),
    [anon_sym_continue] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_STAR] = ACTIONS(1934),
    [anon_sym_STAR_COLON] = ACTIONS(1932),
    [anon_sym_import] = ACTIONS(1934),
    [anon_sym_package] = ACTIONS(1934),
    [anon_sym_AT] = ACTIONS(1934),
    [anon_sym_assert] = ACTIONS(1934),
    [anon_sym_PLUS_PLUS] = ACTIONS(1932),
    [anon_sym_DASH_DASH] = ACTIONS(1932),
    [anon_sym_SLASH] = ACTIONS(1934),
    [anon_sym_PLUS] = ACTIONS(1934),
    [anon_sym_DASH] = ACTIONS(1934),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [anon_sym_ATinterface] = ACTIONS(1932),
    [anon_sym_interface] = ACTIONS(1934),
    [anon_sym_class] = ACTIONS(1934),
    [anon_sym_LBRACE] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1934),
    [anon_sym_def] = ACTIONS(1934),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_null] = ACTIONS(1934),
    [anon_sym_this] = ACTIONS(1934),
    [anon_sym_do] = ACTIONS(1934),
    [anon_sym_while] = ACTIONS(1934),
    [anon_sym_for] = ACTIONS(1934),
    [aux_sym__type_identifier_token1] = ACTIONS(1934),
    [anon_sym_if] = ACTIONS(1934),
    [anon_sym_else] = ACTIONS(1934),
    [anon_sym_LBRACK] = ACTIONS(1932),
    [aux_sym_number_literal_token1] = ACTIONS(1934),
    [aux_sym_number_literal_token2] = ACTIONS(1932),
    [aux_sym_number_literal_token3] = ACTIONS(1932),
    [aux_sym_number_literal_token4] = ACTIONS(1934),
    [aux_sym_number_literal_token5] = ACTIONS(1932),
    [anon_sym_return] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1934),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1932),
    [anon_sym_DQUOTE] = ACTIONS(1934),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1932),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1932),
    [anon_sym_switch] = ACTIONS(1934),
    [anon_sym_case] = ACTIONS(1934),
    [anon_sym_default] = ACTIONS(1934),
    [anon_sym_try] = ACTIONS(1934),
    [anon_sym_int] = ACTIONS(1934),
    [anon_sym_boolean] = ACTIONS(1934),
    [anon_sym_char] = ACTIONS(1934),
    [anon_sym_short] = ACTIONS(1934),
    [anon_sym_long] = ACTIONS(1934),
    [anon_sym_float] = ACTIONS(1934),
    [anon_sym_double] = ACTIONS(1934),
    [anon_sym_void] = ACTIONS(1934),
    [anon_sym_public] = ACTIONS(1934),
    [anon_sym_protected] = ACTIONS(1934),
    [anon_sym_private] = ACTIONS(1934),
    [anon_sym_static] = ACTIONS(1934),
    [anon_sym_final] = ACTIONS(1934),
    [anon_sym_synchronized] = ACTIONS(1934),
    [anon_sym_TILDE] = ACTIONS(1932),
    [anon_sym_BANG] = ACTIONS(1932),
    [anon_sym_new] = ACTIONS(1934),
  },
  [STATE(980)] = {
    [sym_comment] = STATE(980),
    [sym_groovy_doc] = STATE(980),
    [sym_identifier] = ACTIONS(1749),
    [anon_sym_break] = ACTIONS(1749),
    [anon_sym_continue] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_STAR] = ACTIONS(1749),
    [anon_sym_STAR_COLON] = ACTIONS(1747),
    [anon_sym_import] = ACTIONS(1749),
    [anon_sym_package] = ACTIONS(1749),
    [anon_sym_AT] = ACTIONS(1749),
    [anon_sym_assert] = ACTIONS(1749),
    [anon_sym_PLUS_PLUS] = ACTIONS(1747),
    [anon_sym_DASH_DASH] = ACTIONS(1747),
    [anon_sym_SLASH] = ACTIONS(1749),
    [anon_sym_PLUS] = ACTIONS(1749),
    [anon_sym_DASH] = ACTIONS(1749),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_ATinterface] = ACTIONS(1747),
    [anon_sym_interface] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1749),
    [anon_sym_def] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_null] = ACTIONS(1749),
    [anon_sym_this] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_while] = ACTIONS(1749),
    [anon_sym_for] = ACTIONS(1749),
    [aux_sym__type_identifier_token1] = ACTIONS(1749),
    [anon_sym_if] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1747),
    [aux_sym_number_literal_token1] = ACTIONS(1749),
    [aux_sym_number_literal_token2] = ACTIONS(1747),
    [aux_sym_number_literal_token3] = ACTIONS(1747),
    [aux_sym_number_literal_token4] = ACTIONS(1749),
    [aux_sym_number_literal_token5] = ACTIONS(1747),
    [anon_sym_return] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1749),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1749),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1747),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1747),
    [anon_sym_switch] = ACTIONS(1749),
    [anon_sym_case] = ACTIONS(1749),
    [anon_sym_default] = ACTIONS(1749),
    [anon_sym_try] = ACTIONS(1749),
    [anon_sym_int] = ACTIONS(1749),
    [anon_sym_boolean] = ACTIONS(1749),
    [anon_sym_char] = ACTIONS(1749),
    [anon_sym_short] = ACTIONS(1749),
    [anon_sym_long] = ACTIONS(1749),
    [anon_sym_float] = ACTIONS(1749),
    [anon_sym_double] = ACTIONS(1749),
    [anon_sym_void] = ACTIONS(1749),
    [anon_sym_public] = ACTIONS(1749),
    [anon_sym_protected] = ACTIONS(1749),
    [anon_sym_private] = ACTIONS(1749),
    [anon_sym_static] = ACTIONS(1749),
    [anon_sym_final] = ACTIONS(1749),
    [anon_sym_synchronized] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1747),
    [anon_sym_BANG] = ACTIONS(1747),
    [anon_sym_new] = ACTIONS(1749),
  },
  [STATE(981)] = {
    [sym_comment] = STATE(981),
    [sym_groovy_doc] = STATE(981),
    [sym_identifier] = ACTIONS(2112),
    [anon_sym_break] = ACTIONS(2112),
    [anon_sym_continue] = ACTIONS(2112),
    [anon_sym_SEMI] = ACTIONS(2110),
    [anon_sym_STAR] = ACTIONS(2112),
    [anon_sym_STAR_COLON] = ACTIONS(2110),
    [anon_sym_import] = ACTIONS(2112),
    [anon_sym_package] = ACTIONS(2112),
    [anon_sym_AT] = ACTIONS(2112),
    [anon_sym_assert] = ACTIONS(2112),
    [anon_sym_PLUS_PLUS] = ACTIONS(2110),
    [anon_sym_DASH_DASH] = ACTIONS(2110),
    [anon_sym_SLASH] = ACTIONS(2112),
    [anon_sym_PLUS] = ACTIONS(2112),
    [anon_sym_DASH] = ACTIONS(2112),
    [anon_sym_true] = ACTIONS(2112),
    [anon_sym_false] = ACTIONS(2112),
    [anon_sym_ATinterface] = ACTIONS(2110),
    [anon_sym_interface] = ACTIONS(2112),
    [anon_sym_class] = ACTIONS(2112),
    [anon_sym_LBRACE] = ACTIONS(2110),
    [anon_sym_RBRACE] = ACTIONS(2110),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2112),
    [anon_sym_def] = ACTIONS(2112),
    [anon_sym_LPAREN] = ACTIONS(2110),
    [anon_sym_null] = ACTIONS(2112),
    [anon_sym_this] = ACTIONS(2112),
    [anon_sym_do] = ACTIONS(2112),
    [anon_sym_while] = ACTIONS(2112),
    [anon_sym_for] = ACTIONS(2112),
    [aux_sym__type_identifier_token1] = ACTIONS(2112),
    [anon_sym_if] = ACTIONS(2112),
    [anon_sym_else] = ACTIONS(2112),
    [anon_sym_LBRACK] = ACTIONS(2110),
    [aux_sym_number_literal_token1] = ACTIONS(2112),
    [aux_sym_number_literal_token2] = ACTIONS(2110),
    [aux_sym_number_literal_token3] = ACTIONS(2110),
    [aux_sym_number_literal_token4] = ACTIONS(2112),
    [aux_sym_number_literal_token5] = ACTIONS(2110),
    [anon_sym_return] = ACTIONS(2112),
    [anon_sym_SQUOTE] = ACTIONS(2112),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2110),
    [anon_sym_DQUOTE] = ACTIONS(2112),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2110),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2110),
    [anon_sym_switch] = ACTIONS(2112),
    [anon_sym_case] = ACTIONS(2112),
    [anon_sym_default] = ACTIONS(2112),
    [anon_sym_try] = ACTIONS(2112),
    [anon_sym_int] = ACTIONS(2112),
    [anon_sym_boolean] = ACTIONS(2112),
    [anon_sym_char] = ACTIONS(2112),
    [anon_sym_short] = ACTIONS(2112),
    [anon_sym_long] = ACTIONS(2112),
    [anon_sym_float] = ACTIONS(2112),
    [anon_sym_double] = ACTIONS(2112),
    [anon_sym_void] = ACTIONS(2112),
    [anon_sym_public] = ACTIONS(2112),
    [anon_sym_protected] = ACTIONS(2112),
    [anon_sym_private] = ACTIONS(2112),
    [anon_sym_static] = ACTIONS(2112),
    [anon_sym_final] = ACTIONS(2112),
    [anon_sym_synchronized] = ACTIONS(2112),
    [anon_sym_TILDE] = ACTIONS(2110),
    [anon_sym_BANG] = ACTIONS(2110),
    [anon_sym_new] = ACTIONS(2112),
  },
  [STATE(982)] = {
    [sym_comment] = STATE(982),
    [sym_groovy_doc] = STATE(982),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_case] = ACTIONS(371),
    [anon_sym_default] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(369),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(983)] = {
    [sym_comment] = STATE(983),
    [sym_groovy_doc] = STATE(983),
    [sym_identifier] = ACTIONS(1842),
    [anon_sym_break] = ACTIONS(1842),
    [anon_sym_continue] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_STAR] = ACTIONS(1842),
    [anon_sym_STAR_COLON] = ACTIONS(1840),
    [anon_sym_import] = ACTIONS(1842),
    [anon_sym_package] = ACTIONS(1842),
    [anon_sym_AT] = ACTIONS(1842),
    [anon_sym_assert] = ACTIONS(1842),
    [anon_sym_PLUS_PLUS] = ACTIONS(1840),
    [anon_sym_DASH_DASH] = ACTIONS(1840),
    [anon_sym_SLASH] = ACTIONS(1842),
    [anon_sym_PLUS] = ACTIONS(1842),
    [anon_sym_DASH] = ACTIONS(1842),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [anon_sym_ATinterface] = ACTIONS(1840),
    [anon_sym_interface] = ACTIONS(1842),
    [anon_sym_class] = ACTIONS(1842),
    [anon_sym_LBRACE] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1842),
    [anon_sym_def] = ACTIONS(1842),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_null] = ACTIONS(1842),
    [anon_sym_this] = ACTIONS(1842),
    [anon_sym_do] = ACTIONS(1842),
    [anon_sym_while] = ACTIONS(1842),
    [anon_sym_for] = ACTIONS(1842),
    [aux_sym__type_identifier_token1] = ACTIONS(1842),
    [anon_sym_if] = ACTIONS(1842),
    [anon_sym_else] = ACTIONS(1842),
    [anon_sym_LBRACK] = ACTIONS(1840),
    [aux_sym_number_literal_token1] = ACTIONS(1842),
    [aux_sym_number_literal_token2] = ACTIONS(1840),
    [aux_sym_number_literal_token3] = ACTIONS(1840),
    [aux_sym_number_literal_token4] = ACTIONS(1842),
    [aux_sym_number_literal_token5] = ACTIONS(1840),
    [anon_sym_return] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1842),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1840),
    [anon_sym_DQUOTE] = ACTIONS(1842),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1840),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1840),
    [anon_sym_switch] = ACTIONS(1842),
    [anon_sym_case] = ACTIONS(1842),
    [anon_sym_default] = ACTIONS(1842),
    [anon_sym_try] = ACTIONS(1842),
    [anon_sym_int] = ACTIONS(1842),
    [anon_sym_boolean] = ACTIONS(1842),
    [anon_sym_char] = ACTIONS(1842),
    [anon_sym_short] = ACTIONS(1842),
    [anon_sym_long] = ACTIONS(1842),
    [anon_sym_float] = ACTIONS(1842),
    [anon_sym_double] = ACTIONS(1842),
    [anon_sym_void] = ACTIONS(1842),
    [anon_sym_public] = ACTIONS(1842),
    [anon_sym_protected] = ACTIONS(1842),
    [anon_sym_private] = ACTIONS(1842),
    [anon_sym_static] = ACTIONS(1842),
    [anon_sym_final] = ACTIONS(1842),
    [anon_sym_synchronized] = ACTIONS(1842),
    [anon_sym_TILDE] = ACTIONS(1840),
    [anon_sym_BANG] = ACTIONS(1840),
    [anon_sym_new] = ACTIONS(1842),
  },
  [STATE(984)] = {
    [sym_comment] = STATE(984),
    [sym_groovy_doc] = STATE(984),
    [sym_identifier] = ACTIONS(2005),
    [anon_sym_break] = ACTIONS(2005),
    [anon_sym_continue] = ACTIONS(2005),
    [anon_sym_SEMI] = ACTIONS(2003),
    [anon_sym_STAR] = ACTIONS(2005),
    [anon_sym_STAR_COLON] = ACTIONS(2003),
    [anon_sym_import] = ACTIONS(2005),
    [anon_sym_package] = ACTIONS(2005),
    [anon_sym_AT] = ACTIONS(2005),
    [anon_sym_assert] = ACTIONS(2005),
    [anon_sym_PLUS_PLUS] = ACTIONS(2003),
    [anon_sym_DASH_DASH] = ACTIONS(2003),
    [anon_sym_SLASH] = ACTIONS(2005),
    [anon_sym_PLUS] = ACTIONS(2005),
    [anon_sym_DASH] = ACTIONS(2005),
    [anon_sym_true] = ACTIONS(2005),
    [anon_sym_false] = ACTIONS(2005),
    [anon_sym_ATinterface] = ACTIONS(2003),
    [anon_sym_interface] = ACTIONS(2005),
    [anon_sym_class] = ACTIONS(2005),
    [anon_sym_LBRACE] = ACTIONS(2003),
    [anon_sym_RBRACE] = ACTIONS(2003),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2005),
    [anon_sym_def] = ACTIONS(2005),
    [anon_sym_LPAREN] = ACTIONS(2003),
    [anon_sym_null] = ACTIONS(2005),
    [anon_sym_this] = ACTIONS(2005),
    [anon_sym_do] = ACTIONS(2005),
    [anon_sym_while] = ACTIONS(2005),
    [anon_sym_for] = ACTIONS(2005),
    [aux_sym__type_identifier_token1] = ACTIONS(2005),
    [anon_sym_if] = ACTIONS(2005),
    [anon_sym_else] = ACTIONS(2005),
    [anon_sym_LBRACK] = ACTIONS(2003),
    [aux_sym_number_literal_token1] = ACTIONS(2005),
    [aux_sym_number_literal_token2] = ACTIONS(2003),
    [aux_sym_number_literal_token3] = ACTIONS(2003),
    [aux_sym_number_literal_token4] = ACTIONS(2005),
    [aux_sym_number_literal_token5] = ACTIONS(2003),
    [anon_sym_return] = ACTIONS(2005),
    [anon_sym_SQUOTE] = ACTIONS(2005),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2003),
    [anon_sym_DQUOTE] = ACTIONS(2005),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2003),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2003),
    [anon_sym_switch] = ACTIONS(2005),
    [anon_sym_case] = ACTIONS(2005),
    [anon_sym_default] = ACTIONS(2005),
    [anon_sym_try] = ACTIONS(2005),
    [anon_sym_int] = ACTIONS(2005),
    [anon_sym_boolean] = ACTIONS(2005),
    [anon_sym_char] = ACTIONS(2005),
    [anon_sym_short] = ACTIONS(2005),
    [anon_sym_long] = ACTIONS(2005),
    [anon_sym_float] = ACTIONS(2005),
    [anon_sym_double] = ACTIONS(2005),
    [anon_sym_void] = ACTIONS(2005),
    [anon_sym_public] = ACTIONS(2005),
    [anon_sym_protected] = ACTIONS(2005),
    [anon_sym_private] = ACTIONS(2005),
    [anon_sym_static] = ACTIONS(2005),
    [anon_sym_final] = ACTIONS(2005),
    [anon_sym_synchronized] = ACTIONS(2005),
    [anon_sym_TILDE] = ACTIONS(2003),
    [anon_sym_BANG] = ACTIONS(2003),
    [anon_sym_new] = ACTIONS(2005),
  },
  [STATE(985)] = {
    [sym_comment] = STATE(985),
    [sym_groovy_doc] = STATE(985),
    [sym_identifier] = ACTIONS(1820),
    [anon_sym_break] = ACTIONS(1820),
    [anon_sym_continue] = ACTIONS(1820),
    [anon_sym_SEMI] = ACTIONS(1818),
    [anon_sym_STAR] = ACTIONS(1820),
    [anon_sym_STAR_COLON] = ACTIONS(1818),
    [anon_sym_import] = ACTIONS(1820),
    [anon_sym_package] = ACTIONS(1820),
    [anon_sym_AT] = ACTIONS(1820),
    [anon_sym_assert] = ACTIONS(1820),
    [anon_sym_PLUS_PLUS] = ACTIONS(1818),
    [anon_sym_DASH_DASH] = ACTIONS(1818),
    [anon_sym_SLASH] = ACTIONS(1820),
    [anon_sym_PLUS] = ACTIONS(1820),
    [anon_sym_DASH] = ACTIONS(1820),
    [anon_sym_true] = ACTIONS(1820),
    [anon_sym_false] = ACTIONS(1820),
    [anon_sym_ATinterface] = ACTIONS(1818),
    [anon_sym_interface] = ACTIONS(1820),
    [anon_sym_class] = ACTIONS(1820),
    [anon_sym_LBRACE] = ACTIONS(1818),
    [anon_sym_RBRACE] = ACTIONS(1818),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1820),
    [anon_sym_def] = ACTIONS(1820),
    [anon_sym_LPAREN] = ACTIONS(1818),
    [anon_sym_null] = ACTIONS(1820),
    [anon_sym_this] = ACTIONS(1820),
    [anon_sym_do] = ACTIONS(1820),
    [anon_sym_while] = ACTIONS(1820),
    [anon_sym_for] = ACTIONS(1820),
    [aux_sym__type_identifier_token1] = ACTIONS(1820),
    [anon_sym_if] = ACTIONS(1820),
    [anon_sym_else] = ACTIONS(1820),
    [anon_sym_LBRACK] = ACTIONS(1818),
    [aux_sym_number_literal_token1] = ACTIONS(1820),
    [aux_sym_number_literal_token2] = ACTIONS(1818),
    [aux_sym_number_literal_token3] = ACTIONS(1818),
    [aux_sym_number_literal_token4] = ACTIONS(1820),
    [aux_sym_number_literal_token5] = ACTIONS(1818),
    [anon_sym_return] = ACTIONS(1820),
    [anon_sym_SQUOTE] = ACTIONS(1820),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1818),
    [anon_sym_DQUOTE] = ACTIONS(1820),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1818),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1818),
    [anon_sym_switch] = ACTIONS(1820),
    [anon_sym_case] = ACTIONS(1820),
    [anon_sym_default] = ACTIONS(1820),
    [anon_sym_try] = ACTIONS(1820),
    [anon_sym_int] = ACTIONS(1820),
    [anon_sym_boolean] = ACTIONS(1820),
    [anon_sym_char] = ACTIONS(1820),
    [anon_sym_short] = ACTIONS(1820),
    [anon_sym_long] = ACTIONS(1820),
    [anon_sym_float] = ACTIONS(1820),
    [anon_sym_double] = ACTIONS(1820),
    [anon_sym_void] = ACTIONS(1820),
    [anon_sym_public] = ACTIONS(1820),
    [anon_sym_protected] = ACTIONS(1820),
    [anon_sym_private] = ACTIONS(1820),
    [anon_sym_static] = ACTIONS(1820),
    [anon_sym_final] = ACTIONS(1820),
    [anon_sym_synchronized] = ACTIONS(1820),
    [anon_sym_TILDE] = ACTIONS(1818),
    [anon_sym_BANG] = ACTIONS(1818),
    [anon_sym_new] = ACTIONS(1820),
  },
  [STATE(986)] = {
    [sym_comment] = STATE(986),
    [sym_groovy_doc] = STATE(986),
    [ts_builtin_sym_end] = ACTIONS(1515),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1515),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_else] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_pipeline] = ACTIONS(1513),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(987)] = {
    [sym_comment] = STATE(987),
    [sym_groovy_doc] = STATE(987),
    [sym_identifier] = ACTIONS(1942),
    [anon_sym_break] = ACTIONS(1942),
    [anon_sym_continue] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1942),
    [anon_sym_STAR_COLON] = ACTIONS(1940),
    [anon_sym_import] = ACTIONS(1942),
    [anon_sym_package] = ACTIONS(1942),
    [anon_sym_AT] = ACTIONS(1942),
    [anon_sym_assert] = ACTIONS(1942),
    [anon_sym_PLUS_PLUS] = ACTIONS(1940),
    [anon_sym_DASH_DASH] = ACTIONS(1940),
    [anon_sym_SLASH] = ACTIONS(1942),
    [anon_sym_PLUS] = ACTIONS(1942),
    [anon_sym_DASH] = ACTIONS(1942),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [anon_sym_ATinterface] = ACTIONS(1940),
    [anon_sym_interface] = ACTIONS(1942),
    [anon_sym_class] = ACTIONS(1942),
    [anon_sym_LBRACE] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1942),
    [anon_sym_def] = ACTIONS(1942),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_null] = ACTIONS(1942),
    [anon_sym_this] = ACTIONS(1942),
    [anon_sym_do] = ACTIONS(1942),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_for] = ACTIONS(1942),
    [aux_sym__type_identifier_token1] = ACTIONS(1942),
    [anon_sym_if] = ACTIONS(1942),
    [anon_sym_else] = ACTIONS(1942),
    [anon_sym_LBRACK] = ACTIONS(1940),
    [aux_sym_number_literal_token1] = ACTIONS(1942),
    [aux_sym_number_literal_token2] = ACTIONS(1940),
    [aux_sym_number_literal_token3] = ACTIONS(1940),
    [aux_sym_number_literal_token4] = ACTIONS(1942),
    [aux_sym_number_literal_token5] = ACTIONS(1940),
    [anon_sym_return] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1942),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1942),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1940),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1940),
    [anon_sym_switch] = ACTIONS(1942),
    [anon_sym_case] = ACTIONS(1942),
    [anon_sym_default] = ACTIONS(1942),
    [anon_sym_try] = ACTIONS(1942),
    [anon_sym_int] = ACTIONS(1942),
    [anon_sym_boolean] = ACTIONS(1942),
    [anon_sym_char] = ACTIONS(1942),
    [anon_sym_short] = ACTIONS(1942),
    [anon_sym_long] = ACTIONS(1942),
    [anon_sym_float] = ACTIONS(1942),
    [anon_sym_double] = ACTIONS(1942),
    [anon_sym_void] = ACTIONS(1942),
    [anon_sym_public] = ACTIONS(1942),
    [anon_sym_protected] = ACTIONS(1942),
    [anon_sym_private] = ACTIONS(1942),
    [anon_sym_static] = ACTIONS(1942),
    [anon_sym_final] = ACTIONS(1942),
    [anon_sym_synchronized] = ACTIONS(1942),
    [anon_sym_TILDE] = ACTIONS(1940),
    [anon_sym_BANG] = ACTIONS(1940),
    [anon_sym_new] = ACTIONS(1942),
  },
  [STATE(988)] = {
    [sym_comment] = STATE(988),
    [sym_groovy_doc] = STATE(988),
    [sym_identifier] = ACTIONS(1769),
    [anon_sym_break] = ACTIONS(1769),
    [anon_sym_continue] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_STAR] = ACTIONS(1769),
    [anon_sym_STAR_COLON] = ACTIONS(1767),
    [anon_sym_import] = ACTIONS(1769),
    [anon_sym_package] = ACTIONS(1769),
    [anon_sym_AT] = ACTIONS(1769),
    [anon_sym_assert] = ACTIONS(1769),
    [anon_sym_PLUS_PLUS] = ACTIONS(1767),
    [anon_sym_DASH_DASH] = ACTIONS(1767),
    [anon_sym_SLASH] = ACTIONS(1769),
    [anon_sym_PLUS] = ACTIONS(1769),
    [anon_sym_DASH] = ACTIONS(1769),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_ATinterface] = ACTIONS(1767),
    [anon_sym_interface] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1769),
    [anon_sym_def] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_null] = ACTIONS(1769),
    [anon_sym_this] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_while] = ACTIONS(1769),
    [anon_sym_for] = ACTIONS(1769),
    [aux_sym__type_identifier_token1] = ACTIONS(1769),
    [anon_sym_if] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1767),
    [aux_sym_number_literal_token1] = ACTIONS(1769),
    [aux_sym_number_literal_token2] = ACTIONS(1767),
    [aux_sym_number_literal_token3] = ACTIONS(1767),
    [aux_sym_number_literal_token4] = ACTIONS(1769),
    [aux_sym_number_literal_token5] = ACTIONS(1767),
    [anon_sym_return] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1769),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1769),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1767),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1767),
    [anon_sym_switch] = ACTIONS(1769),
    [anon_sym_case] = ACTIONS(1769),
    [anon_sym_default] = ACTIONS(1769),
    [anon_sym_try] = ACTIONS(1769),
    [anon_sym_int] = ACTIONS(1769),
    [anon_sym_boolean] = ACTIONS(1769),
    [anon_sym_char] = ACTIONS(1769),
    [anon_sym_short] = ACTIONS(1769),
    [anon_sym_long] = ACTIONS(1769),
    [anon_sym_float] = ACTIONS(1769),
    [anon_sym_double] = ACTIONS(1769),
    [anon_sym_void] = ACTIONS(1769),
    [anon_sym_public] = ACTIONS(1769),
    [anon_sym_protected] = ACTIONS(1769),
    [anon_sym_private] = ACTIONS(1769),
    [anon_sym_static] = ACTIONS(1769),
    [anon_sym_final] = ACTIONS(1769),
    [anon_sym_synchronized] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1767),
    [anon_sym_BANG] = ACTIONS(1767),
    [anon_sym_new] = ACTIONS(1769),
  },
  [STATE(989)] = {
    [sym_comment] = STATE(989),
    [sym_groovy_doc] = STATE(989),
    [sym_identifier] = ACTIONS(1850),
    [anon_sym_break] = ACTIONS(1850),
    [anon_sym_continue] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_STAR] = ACTIONS(1850),
    [anon_sym_STAR_COLON] = ACTIONS(1848),
    [anon_sym_import] = ACTIONS(1850),
    [anon_sym_package] = ACTIONS(1850),
    [anon_sym_AT] = ACTIONS(1850),
    [anon_sym_assert] = ACTIONS(1850),
    [anon_sym_PLUS_PLUS] = ACTIONS(1848),
    [anon_sym_DASH_DASH] = ACTIONS(1848),
    [anon_sym_SLASH] = ACTIONS(1850),
    [anon_sym_PLUS] = ACTIONS(1850),
    [anon_sym_DASH] = ACTIONS(1850),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [anon_sym_ATinterface] = ACTIONS(1848),
    [anon_sym_interface] = ACTIONS(1850),
    [anon_sym_class] = ACTIONS(1850),
    [anon_sym_LBRACE] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1850),
    [anon_sym_def] = ACTIONS(1850),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_null] = ACTIONS(1850),
    [anon_sym_this] = ACTIONS(1850),
    [anon_sym_do] = ACTIONS(1850),
    [anon_sym_while] = ACTIONS(1850),
    [anon_sym_for] = ACTIONS(1850),
    [aux_sym__type_identifier_token1] = ACTIONS(1850),
    [anon_sym_if] = ACTIONS(1850),
    [anon_sym_else] = ACTIONS(1850),
    [anon_sym_LBRACK] = ACTIONS(1848),
    [aux_sym_number_literal_token1] = ACTIONS(1850),
    [aux_sym_number_literal_token2] = ACTIONS(1848),
    [aux_sym_number_literal_token3] = ACTIONS(1848),
    [aux_sym_number_literal_token4] = ACTIONS(1850),
    [aux_sym_number_literal_token5] = ACTIONS(1848),
    [anon_sym_return] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1850),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1848),
    [anon_sym_DQUOTE] = ACTIONS(1850),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1848),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1848),
    [anon_sym_switch] = ACTIONS(1850),
    [anon_sym_case] = ACTIONS(1850),
    [anon_sym_default] = ACTIONS(1850),
    [anon_sym_try] = ACTIONS(1850),
    [anon_sym_int] = ACTIONS(1850),
    [anon_sym_boolean] = ACTIONS(1850),
    [anon_sym_char] = ACTIONS(1850),
    [anon_sym_short] = ACTIONS(1850),
    [anon_sym_long] = ACTIONS(1850),
    [anon_sym_float] = ACTIONS(1850),
    [anon_sym_double] = ACTIONS(1850),
    [anon_sym_void] = ACTIONS(1850),
    [anon_sym_public] = ACTIONS(1850),
    [anon_sym_protected] = ACTIONS(1850),
    [anon_sym_private] = ACTIONS(1850),
    [anon_sym_static] = ACTIONS(1850),
    [anon_sym_final] = ACTIONS(1850),
    [anon_sym_synchronized] = ACTIONS(1850),
    [anon_sym_TILDE] = ACTIONS(1848),
    [anon_sym_BANG] = ACTIONS(1848),
    [anon_sym_new] = ACTIONS(1850),
  },
  [STATE(990)] = {
    [sym_comment] = STATE(990),
    [sym_groovy_doc] = STATE(990),
    [ts_builtin_sym_end] = ACTIONS(369),
    [sym_identifier] = ACTIONS(371),
    [anon_sym_break] = ACTIONS(371),
    [anon_sym_continue] = ACTIONS(371),
    [anon_sym_SEMI] = ACTIONS(369),
    [anon_sym_STAR] = ACTIONS(371),
    [anon_sym_STAR_COLON] = ACTIONS(369),
    [anon_sym_import] = ACTIONS(371),
    [anon_sym_package] = ACTIONS(371),
    [anon_sym_AT] = ACTIONS(371),
    [anon_sym_assert] = ACTIONS(371),
    [anon_sym_PLUS_PLUS] = ACTIONS(369),
    [anon_sym_DASH_DASH] = ACTIONS(369),
    [anon_sym_SLASH] = ACTIONS(371),
    [anon_sym_PLUS] = ACTIONS(371),
    [anon_sym_DASH] = ACTIONS(371),
    [anon_sym_true] = ACTIONS(371),
    [anon_sym_false] = ACTIONS(371),
    [anon_sym_ATinterface] = ACTIONS(369),
    [anon_sym_interface] = ACTIONS(371),
    [anon_sym_class] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(369),
    [anon_sym_RBRACE] = ACTIONS(369),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(371),
    [anon_sym_def] = ACTIONS(371),
    [anon_sym_LPAREN] = ACTIONS(369),
    [anon_sym_null] = ACTIONS(371),
    [anon_sym_this] = ACTIONS(371),
    [anon_sym_do] = ACTIONS(371),
    [anon_sym_while] = ACTIONS(371),
    [anon_sym_for] = ACTIONS(371),
    [aux_sym__type_identifier_token1] = ACTIONS(371),
    [anon_sym_if] = ACTIONS(371),
    [anon_sym_else] = ACTIONS(371),
    [anon_sym_LBRACK] = ACTIONS(369),
    [aux_sym_number_literal_token1] = ACTIONS(371),
    [aux_sym_number_literal_token2] = ACTIONS(369),
    [aux_sym_number_literal_token3] = ACTIONS(369),
    [aux_sym_number_literal_token4] = ACTIONS(371),
    [aux_sym_number_literal_token5] = ACTIONS(369),
    [anon_sym_pipeline] = ACTIONS(371),
    [anon_sym_return] = ACTIONS(371),
    [anon_sym_SQUOTE] = ACTIONS(371),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(369),
    [anon_sym_DQUOTE] = ACTIONS(371),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(369),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(369),
    [anon_sym_switch] = ACTIONS(371),
    [anon_sym_try] = ACTIONS(371),
    [anon_sym_int] = ACTIONS(371),
    [anon_sym_boolean] = ACTIONS(371),
    [anon_sym_char] = ACTIONS(371),
    [anon_sym_short] = ACTIONS(371),
    [anon_sym_long] = ACTIONS(371),
    [anon_sym_float] = ACTIONS(371),
    [anon_sym_double] = ACTIONS(371),
    [anon_sym_void] = ACTIONS(371),
    [anon_sym_public] = ACTIONS(371),
    [anon_sym_protected] = ACTIONS(371),
    [anon_sym_private] = ACTIONS(371),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_final] = ACTIONS(371),
    [anon_sym_synchronized] = ACTIONS(371),
    [anon_sym_TILDE] = ACTIONS(369),
    [anon_sym_BANG] = ACTIONS(369),
    [anon_sym_new] = ACTIONS(371),
  },
  [STATE(991)] = {
    [sym_comment] = STATE(991),
    [sym_groovy_doc] = STATE(991),
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1362),
    [anon_sym_STAR_COLON] = ACTIONS(1364),
    [anon_sym_import] = ACTIONS(1362),
    [anon_sym_package] = ACTIONS(1362),
    [anon_sym_AT] = ACTIONS(1362),
    [anon_sym_assert] = ACTIONS(1362),
    [anon_sym_PLUS_PLUS] = ACTIONS(1364),
    [anon_sym_DASH_DASH] = ACTIONS(1364),
    [anon_sym_SLASH] = ACTIONS(1362),
    [anon_sym_PLUS] = ACTIONS(1362),
    [anon_sym_DASH] = ACTIONS(1362),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [anon_sym_ATinterface] = ACTIONS(1364),
    [anon_sym_interface] = ACTIONS(1362),
    [anon_sym_class] = ACTIONS(1362),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1362),
    [anon_sym_def] = ACTIONS(1362),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_null] = ACTIONS(1362),
    [anon_sym_this] = ACTIONS(1362),
    [anon_sym_do] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [aux_sym__type_identifier_token1] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_else] = ACTIONS(1362),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [aux_sym_number_literal_token1] = ACTIONS(1362),
    [aux_sym_number_literal_token2] = ACTIONS(1364),
    [aux_sym_number_literal_token3] = ACTIONS(1364),
    [aux_sym_number_literal_token4] = ACTIONS(1362),
    [aux_sym_number_literal_token5] = ACTIONS(1364),
    [anon_sym_pipeline] = ACTIONS(1362),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1364),
    [anon_sym_DQUOTE] = ACTIONS(1362),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1364),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1364),
    [anon_sym_switch] = ACTIONS(1362),
    [anon_sym_try] = ACTIONS(1362),
    [anon_sym_int] = ACTIONS(1362),
    [anon_sym_boolean] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_short] = ACTIONS(1362),
    [anon_sym_long] = ACTIONS(1362),
    [anon_sym_float] = ACTIONS(1362),
    [anon_sym_double] = ACTIONS(1362),
    [anon_sym_void] = ACTIONS(1362),
    [anon_sym_public] = ACTIONS(1362),
    [anon_sym_protected] = ACTIONS(1362),
    [anon_sym_private] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_final] = ACTIONS(1362),
    [anon_sym_synchronized] = ACTIONS(1362),
    [anon_sym_TILDE] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_new] = ACTIONS(1362),
  },
  [STATE(992)] = {
    [sym_comment] = STATE(992),
    [sym_groovy_doc] = STATE(992),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1906),
    [anon_sym_STAR_COLON] = ACTIONS(1904),
    [anon_sym_import] = ACTIONS(1906),
    [anon_sym_package] = ACTIONS(1906),
    [anon_sym_AT] = ACTIONS(1906),
    [anon_sym_assert] = ACTIONS(1906),
    [anon_sym_PLUS_PLUS] = ACTIONS(1904),
    [anon_sym_DASH_DASH] = ACTIONS(1904),
    [anon_sym_SLASH] = ACTIONS(1906),
    [anon_sym_PLUS] = ACTIONS(1906),
    [anon_sym_DASH] = ACTIONS(1906),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [anon_sym_ATinterface] = ACTIONS(1904),
    [anon_sym_interface] = ACTIONS(1906),
    [anon_sym_class] = ACTIONS(1906),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1906),
    [anon_sym_def] = ACTIONS(1906),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_null] = ACTIONS(1906),
    [anon_sym_this] = ACTIONS(1906),
    [anon_sym_do] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [aux_sym__type_identifier_token1] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_else] = ACTIONS(2155),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [aux_sym_number_literal_token1] = ACTIONS(1906),
    [aux_sym_number_literal_token2] = ACTIONS(1904),
    [aux_sym_number_literal_token3] = ACTIONS(1904),
    [aux_sym_number_literal_token4] = ACTIONS(1906),
    [aux_sym_number_literal_token5] = ACTIONS(1904),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1904),
    [anon_sym_DQUOTE] = ACTIONS(1906),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1904),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1904),
    [anon_sym_switch] = ACTIONS(1906),
    [anon_sym_case] = ACTIONS(1906),
    [anon_sym_default] = ACTIONS(1906),
    [anon_sym_try] = ACTIONS(1906),
    [anon_sym_int] = ACTIONS(1906),
    [anon_sym_boolean] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_short] = ACTIONS(1906),
    [anon_sym_long] = ACTIONS(1906),
    [anon_sym_float] = ACTIONS(1906),
    [anon_sym_double] = ACTIONS(1906),
    [anon_sym_void] = ACTIONS(1906),
    [anon_sym_public] = ACTIONS(1906),
    [anon_sym_protected] = ACTIONS(1906),
    [anon_sym_private] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_final] = ACTIONS(1906),
    [anon_sym_synchronized] = ACTIONS(1906),
    [anon_sym_TILDE] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_new] = ACTIONS(1906),
  },
  [STATE(993)] = {
    [sym_comment] = STATE(993),
    [sym_groovy_doc] = STATE(993),
    [sym_identifier] = ACTIONS(1950),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_STAR] = ACTIONS(1950),
    [anon_sym_STAR_COLON] = ACTIONS(1948),
    [anon_sym_import] = ACTIONS(1950),
    [anon_sym_package] = ACTIONS(1950),
    [anon_sym_AT] = ACTIONS(1950),
    [anon_sym_assert] = ACTIONS(1950),
    [anon_sym_PLUS_PLUS] = ACTIONS(1948),
    [anon_sym_DASH_DASH] = ACTIONS(1948),
    [anon_sym_SLASH] = ACTIONS(1950),
    [anon_sym_PLUS] = ACTIONS(1950),
    [anon_sym_DASH] = ACTIONS(1950),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [anon_sym_ATinterface] = ACTIONS(1948),
    [anon_sym_interface] = ACTIONS(1950),
    [anon_sym_class] = ACTIONS(1950),
    [anon_sym_LBRACE] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1950),
    [anon_sym_def] = ACTIONS(1950),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_null] = ACTIONS(1950),
    [anon_sym_this] = ACTIONS(1950),
    [anon_sym_do] = ACTIONS(1950),
    [anon_sym_while] = ACTIONS(1950),
    [anon_sym_for] = ACTIONS(1950),
    [aux_sym__type_identifier_token1] = ACTIONS(1950),
    [anon_sym_if] = ACTIONS(1950),
    [anon_sym_else] = ACTIONS(1950),
    [anon_sym_LBRACK] = ACTIONS(1948),
    [aux_sym_number_literal_token1] = ACTIONS(1950),
    [aux_sym_number_literal_token2] = ACTIONS(1948),
    [aux_sym_number_literal_token3] = ACTIONS(1948),
    [aux_sym_number_literal_token4] = ACTIONS(1950),
    [aux_sym_number_literal_token5] = ACTIONS(1948),
    [anon_sym_return] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1950),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1948),
    [anon_sym_DQUOTE] = ACTIONS(1950),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1948),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1948),
    [anon_sym_switch] = ACTIONS(1950),
    [anon_sym_case] = ACTIONS(1950),
    [anon_sym_default] = ACTIONS(1950),
    [anon_sym_try] = ACTIONS(1950),
    [anon_sym_int] = ACTIONS(1950),
    [anon_sym_boolean] = ACTIONS(1950),
    [anon_sym_char] = ACTIONS(1950),
    [anon_sym_short] = ACTIONS(1950),
    [anon_sym_long] = ACTIONS(1950),
    [anon_sym_float] = ACTIONS(1950),
    [anon_sym_double] = ACTIONS(1950),
    [anon_sym_void] = ACTIONS(1950),
    [anon_sym_public] = ACTIONS(1950),
    [anon_sym_protected] = ACTIONS(1950),
    [anon_sym_private] = ACTIONS(1950),
    [anon_sym_static] = ACTIONS(1950),
    [anon_sym_final] = ACTIONS(1950),
    [anon_sym_synchronized] = ACTIONS(1950),
    [anon_sym_TILDE] = ACTIONS(1948),
    [anon_sym_BANG] = ACTIONS(1948),
    [anon_sym_new] = ACTIONS(1950),
  },
  [STATE(994)] = {
    [sym_comment] = STATE(994),
    [sym_groovy_doc] = STATE(994),
    [sym_identifier] = ACTIONS(1962),
    [anon_sym_break] = ACTIONS(1962),
    [anon_sym_continue] = ACTIONS(1962),
    [anon_sym_SEMI] = ACTIONS(1960),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_STAR_COLON] = ACTIONS(1960),
    [anon_sym_import] = ACTIONS(1962),
    [anon_sym_package] = ACTIONS(1962),
    [anon_sym_AT] = ACTIONS(1962),
    [anon_sym_assert] = ACTIONS(1962),
    [anon_sym_PLUS_PLUS] = ACTIONS(1960),
    [anon_sym_DASH_DASH] = ACTIONS(1960),
    [anon_sym_SLASH] = ACTIONS(1962),
    [anon_sym_PLUS] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_true] = ACTIONS(1962),
    [anon_sym_false] = ACTIONS(1962),
    [anon_sym_ATinterface] = ACTIONS(1960),
    [anon_sym_interface] = ACTIONS(1962),
    [anon_sym_class] = ACTIONS(1962),
    [anon_sym_LBRACE] = ACTIONS(1960),
    [anon_sym_RBRACE] = ACTIONS(1960),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1962),
    [anon_sym_def] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1960),
    [anon_sym_null] = ACTIONS(1962),
    [anon_sym_this] = ACTIONS(1962),
    [anon_sym_do] = ACTIONS(1962),
    [anon_sym_while] = ACTIONS(1962),
    [anon_sym_for] = ACTIONS(1962),
    [aux_sym__type_identifier_token1] = ACTIONS(1962),
    [anon_sym_if] = ACTIONS(1962),
    [anon_sym_else] = ACTIONS(1962),
    [anon_sym_LBRACK] = ACTIONS(1960),
    [aux_sym_number_literal_token1] = ACTIONS(1962),
    [aux_sym_number_literal_token2] = ACTIONS(1960),
    [aux_sym_number_literal_token3] = ACTIONS(1960),
    [aux_sym_number_literal_token4] = ACTIONS(1962),
    [aux_sym_number_literal_token5] = ACTIONS(1960),
    [anon_sym_return] = ACTIONS(1962),
    [anon_sym_SQUOTE] = ACTIONS(1962),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1960),
    [anon_sym_DQUOTE] = ACTIONS(1962),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1960),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1960),
    [anon_sym_switch] = ACTIONS(1962),
    [anon_sym_case] = ACTIONS(1962),
    [anon_sym_default] = ACTIONS(1962),
    [anon_sym_try] = ACTIONS(1962),
    [anon_sym_int] = ACTIONS(1962),
    [anon_sym_boolean] = ACTIONS(1962),
    [anon_sym_char] = ACTIONS(1962),
    [anon_sym_short] = ACTIONS(1962),
    [anon_sym_long] = ACTIONS(1962),
    [anon_sym_float] = ACTIONS(1962),
    [anon_sym_double] = ACTIONS(1962),
    [anon_sym_void] = ACTIONS(1962),
    [anon_sym_public] = ACTIONS(1962),
    [anon_sym_protected] = ACTIONS(1962),
    [anon_sym_private] = ACTIONS(1962),
    [anon_sym_static] = ACTIONS(1962),
    [anon_sym_final] = ACTIONS(1962),
    [anon_sym_synchronized] = ACTIONS(1962),
    [anon_sym_TILDE] = ACTIONS(1960),
    [anon_sym_BANG] = ACTIONS(1960),
    [anon_sym_new] = ACTIONS(1962),
  },
  [STATE(995)] = {
    [sym_comment] = STATE(995),
    [sym_groovy_doc] = STATE(995),
    [sym_identifier] = ACTIONS(1970),
    [anon_sym_break] = ACTIONS(1970),
    [anon_sym_continue] = ACTIONS(1970),
    [anon_sym_SEMI] = ACTIONS(1968),
    [anon_sym_STAR] = ACTIONS(1970),
    [anon_sym_STAR_COLON] = ACTIONS(1968),
    [anon_sym_import] = ACTIONS(1970),
    [anon_sym_package] = ACTIONS(1970),
    [anon_sym_AT] = ACTIONS(1970),
    [anon_sym_assert] = ACTIONS(1970),
    [anon_sym_PLUS_PLUS] = ACTIONS(1968),
    [anon_sym_DASH_DASH] = ACTIONS(1968),
    [anon_sym_SLASH] = ACTIONS(1970),
    [anon_sym_PLUS] = ACTIONS(1970),
    [anon_sym_DASH] = ACTIONS(1970),
    [anon_sym_true] = ACTIONS(1970),
    [anon_sym_false] = ACTIONS(1970),
    [anon_sym_ATinterface] = ACTIONS(1968),
    [anon_sym_interface] = ACTIONS(1970),
    [anon_sym_class] = ACTIONS(1970),
    [anon_sym_LBRACE] = ACTIONS(1968),
    [anon_sym_RBRACE] = ACTIONS(1968),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1970),
    [anon_sym_def] = ACTIONS(1970),
    [anon_sym_LPAREN] = ACTIONS(1968),
    [anon_sym_null] = ACTIONS(1970),
    [anon_sym_this] = ACTIONS(1970),
    [anon_sym_do] = ACTIONS(1970),
    [anon_sym_while] = ACTIONS(1970),
    [anon_sym_for] = ACTIONS(1970),
    [aux_sym__type_identifier_token1] = ACTIONS(1970),
    [anon_sym_if] = ACTIONS(1970),
    [anon_sym_else] = ACTIONS(1970),
    [anon_sym_LBRACK] = ACTIONS(1968),
    [aux_sym_number_literal_token1] = ACTIONS(1970),
    [aux_sym_number_literal_token2] = ACTIONS(1968),
    [aux_sym_number_literal_token3] = ACTIONS(1968),
    [aux_sym_number_literal_token4] = ACTIONS(1970),
    [aux_sym_number_literal_token5] = ACTIONS(1968),
    [anon_sym_return] = ACTIONS(1970),
    [anon_sym_SQUOTE] = ACTIONS(1970),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1968),
    [anon_sym_DQUOTE] = ACTIONS(1970),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1968),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1968),
    [anon_sym_switch] = ACTIONS(1970),
    [anon_sym_case] = ACTIONS(1970),
    [anon_sym_default] = ACTIONS(1970),
    [anon_sym_try] = ACTIONS(1970),
    [anon_sym_int] = ACTIONS(1970),
    [anon_sym_boolean] = ACTIONS(1970),
    [anon_sym_char] = ACTIONS(1970),
    [anon_sym_short] = ACTIONS(1970),
    [anon_sym_long] = ACTIONS(1970),
    [anon_sym_float] = ACTIONS(1970),
    [anon_sym_double] = ACTIONS(1970),
    [anon_sym_void] = ACTIONS(1970),
    [anon_sym_public] = ACTIONS(1970),
    [anon_sym_protected] = ACTIONS(1970),
    [anon_sym_private] = ACTIONS(1970),
    [anon_sym_static] = ACTIONS(1970),
    [anon_sym_final] = ACTIONS(1970),
    [anon_sym_synchronized] = ACTIONS(1970),
    [anon_sym_TILDE] = ACTIONS(1968),
    [anon_sym_BANG] = ACTIONS(1968),
    [anon_sym_new] = ACTIONS(1970),
  },
  [STATE(996)] = {
    [sym_comment] = STATE(996),
    [sym_groovy_doc] = STATE(996),
    [sym_identifier] = ACTIONS(1806),
    [anon_sym_break] = ACTIONS(1806),
    [anon_sym_continue] = ACTIONS(1806),
    [anon_sym_SEMI] = ACTIONS(1804),
    [anon_sym_STAR] = ACTIONS(1806),
    [anon_sym_STAR_COLON] = ACTIONS(1804),
    [anon_sym_import] = ACTIONS(1806),
    [anon_sym_package] = ACTIONS(1806),
    [anon_sym_AT] = ACTIONS(1806),
    [anon_sym_assert] = ACTIONS(1806),
    [anon_sym_PLUS_PLUS] = ACTIONS(1804),
    [anon_sym_DASH_DASH] = ACTIONS(1804),
    [anon_sym_SLASH] = ACTIONS(1806),
    [anon_sym_PLUS] = ACTIONS(1806),
    [anon_sym_DASH] = ACTIONS(1806),
    [anon_sym_true] = ACTIONS(1806),
    [anon_sym_false] = ACTIONS(1806),
    [anon_sym_ATinterface] = ACTIONS(1804),
    [anon_sym_interface] = ACTIONS(1806),
    [anon_sym_class] = ACTIONS(1806),
    [anon_sym_LBRACE] = ACTIONS(1804),
    [anon_sym_RBRACE] = ACTIONS(1804),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1806),
    [anon_sym_def] = ACTIONS(1806),
    [anon_sym_LPAREN] = ACTIONS(1804),
    [anon_sym_null] = ACTIONS(1806),
    [anon_sym_this] = ACTIONS(1806),
    [anon_sym_do] = ACTIONS(1806),
    [anon_sym_while] = ACTIONS(1806),
    [anon_sym_for] = ACTIONS(1806),
    [aux_sym__type_identifier_token1] = ACTIONS(1806),
    [anon_sym_if] = ACTIONS(1806),
    [anon_sym_else] = ACTIONS(1806),
    [anon_sym_LBRACK] = ACTIONS(1804),
    [aux_sym_number_literal_token1] = ACTIONS(1806),
    [aux_sym_number_literal_token2] = ACTIONS(1804),
    [aux_sym_number_literal_token3] = ACTIONS(1804),
    [aux_sym_number_literal_token4] = ACTIONS(1806),
    [aux_sym_number_literal_token5] = ACTIONS(1804),
    [anon_sym_return] = ACTIONS(1806),
    [anon_sym_SQUOTE] = ACTIONS(1806),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1804),
    [anon_sym_DQUOTE] = ACTIONS(1806),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1804),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1804),
    [anon_sym_switch] = ACTIONS(1806),
    [anon_sym_case] = ACTIONS(1806),
    [anon_sym_default] = ACTIONS(1806),
    [anon_sym_try] = ACTIONS(1806),
    [anon_sym_int] = ACTIONS(1806),
    [anon_sym_boolean] = ACTIONS(1806),
    [anon_sym_char] = ACTIONS(1806),
    [anon_sym_short] = ACTIONS(1806),
    [anon_sym_long] = ACTIONS(1806),
    [anon_sym_float] = ACTIONS(1806),
    [anon_sym_double] = ACTIONS(1806),
    [anon_sym_void] = ACTIONS(1806),
    [anon_sym_public] = ACTIONS(1806),
    [anon_sym_protected] = ACTIONS(1806),
    [anon_sym_private] = ACTIONS(1806),
    [anon_sym_static] = ACTIONS(1806),
    [anon_sym_final] = ACTIONS(1806),
    [anon_sym_synchronized] = ACTIONS(1806),
    [anon_sym_TILDE] = ACTIONS(1804),
    [anon_sym_BANG] = ACTIONS(1804),
    [anon_sym_new] = ACTIONS(1806),
  },
  [STATE(997)] = {
    [sym_comment] = STATE(997),
    [sym_groovy_doc] = STATE(997),
    [sym_identifier] = ACTIONS(1824),
    [anon_sym_break] = ACTIONS(1824),
    [anon_sym_continue] = ACTIONS(1824),
    [anon_sym_SEMI] = ACTIONS(1822),
    [anon_sym_STAR] = ACTIONS(1824),
    [anon_sym_STAR_COLON] = ACTIONS(1822),
    [anon_sym_import] = ACTIONS(1824),
    [anon_sym_package] = ACTIONS(1824),
    [anon_sym_AT] = ACTIONS(1824),
    [anon_sym_assert] = ACTIONS(1824),
    [anon_sym_PLUS_PLUS] = ACTIONS(1822),
    [anon_sym_DASH_DASH] = ACTIONS(1822),
    [anon_sym_SLASH] = ACTIONS(1824),
    [anon_sym_PLUS] = ACTIONS(1824),
    [anon_sym_DASH] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1824),
    [anon_sym_false] = ACTIONS(1824),
    [anon_sym_ATinterface] = ACTIONS(1822),
    [anon_sym_interface] = ACTIONS(1824),
    [anon_sym_class] = ACTIONS(1824),
    [anon_sym_LBRACE] = ACTIONS(1822),
    [anon_sym_RBRACE] = ACTIONS(1822),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1824),
    [anon_sym_def] = ACTIONS(1824),
    [anon_sym_LPAREN] = ACTIONS(1822),
    [anon_sym_null] = ACTIONS(1824),
    [anon_sym_this] = ACTIONS(1824),
    [anon_sym_do] = ACTIONS(1824),
    [anon_sym_while] = ACTIONS(1824),
    [anon_sym_for] = ACTIONS(1824),
    [aux_sym__type_identifier_token1] = ACTIONS(1824),
    [anon_sym_if] = ACTIONS(1824),
    [anon_sym_else] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1822),
    [aux_sym_number_literal_token1] = ACTIONS(1824),
    [aux_sym_number_literal_token2] = ACTIONS(1822),
    [aux_sym_number_literal_token3] = ACTIONS(1822),
    [aux_sym_number_literal_token4] = ACTIONS(1824),
    [aux_sym_number_literal_token5] = ACTIONS(1822),
    [anon_sym_return] = ACTIONS(1824),
    [anon_sym_SQUOTE] = ACTIONS(1824),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1822),
    [anon_sym_DQUOTE] = ACTIONS(1824),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1822),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1822),
    [anon_sym_switch] = ACTIONS(1824),
    [anon_sym_case] = ACTIONS(1824),
    [anon_sym_default] = ACTIONS(1824),
    [anon_sym_try] = ACTIONS(1824),
    [anon_sym_int] = ACTIONS(1824),
    [anon_sym_boolean] = ACTIONS(1824),
    [anon_sym_char] = ACTIONS(1824),
    [anon_sym_short] = ACTIONS(1824),
    [anon_sym_long] = ACTIONS(1824),
    [anon_sym_float] = ACTIONS(1824),
    [anon_sym_double] = ACTIONS(1824),
    [anon_sym_void] = ACTIONS(1824),
    [anon_sym_public] = ACTIONS(1824),
    [anon_sym_protected] = ACTIONS(1824),
    [anon_sym_private] = ACTIONS(1824),
    [anon_sym_static] = ACTIONS(1824),
    [anon_sym_final] = ACTIONS(1824),
    [anon_sym_synchronized] = ACTIONS(1824),
    [anon_sym_TILDE] = ACTIONS(1822),
    [anon_sym_BANG] = ACTIONS(1822),
    [anon_sym_new] = ACTIONS(1824),
  },
  [STATE(998)] = {
    [sym_comment] = STATE(998),
    [sym_groovy_doc] = STATE(998),
    [sym_identifier] = ACTIONS(1974),
    [anon_sym_break] = ACTIONS(1974),
    [anon_sym_continue] = ACTIONS(1974),
    [anon_sym_SEMI] = ACTIONS(1972),
    [anon_sym_STAR] = ACTIONS(1974),
    [anon_sym_STAR_COLON] = ACTIONS(1972),
    [anon_sym_import] = ACTIONS(1974),
    [anon_sym_package] = ACTIONS(1974),
    [anon_sym_AT] = ACTIONS(1974),
    [anon_sym_assert] = ACTIONS(1974),
    [anon_sym_PLUS_PLUS] = ACTIONS(1972),
    [anon_sym_DASH_DASH] = ACTIONS(1972),
    [anon_sym_SLASH] = ACTIONS(1974),
    [anon_sym_PLUS] = ACTIONS(1974),
    [anon_sym_DASH] = ACTIONS(1974),
    [anon_sym_true] = ACTIONS(1974),
    [anon_sym_false] = ACTIONS(1974),
    [anon_sym_ATinterface] = ACTIONS(1972),
    [anon_sym_interface] = ACTIONS(1974),
    [anon_sym_class] = ACTIONS(1974),
    [anon_sym_LBRACE] = ACTIONS(1972),
    [anon_sym_RBRACE] = ACTIONS(1972),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1974),
    [anon_sym_def] = ACTIONS(1974),
    [anon_sym_LPAREN] = ACTIONS(1972),
    [anon_sym_null] = ACTIONS(1974),
    [anon_sym_this] = ACTIONS(1974),
    [anon_sym_do] = ACTIONS(1974),
    [anon_sym_while] = ACTIONS(1974),
    [anon_sym_for] = ACTIONS(1974),
    [aux_sym__type_identifier_token1] = ACTIONS(1974),
    [anon_sym_if] = ACTIONS(1974),
    [anon_sym_else] = ACTIONS(1974),
    [anon_sym_LBRACK] = ACTIONS(1972),
    [aux_sym_number_literal_token1] = ACTIONS(1974),
    [aux_sym_number_literal_token2] = ACTIONS(1972),
    [aux_sym_number_literal_token3] = ACTIONS(1972),
    [aux_sym_number_literal_token4] = ACTIONS(1974),
    [aux_sym_number_literal_token5] = ACTIONS(1972),
    [anon_sym_return] = ACTIONS(1974),
    [anon_sym_SQUOTE] = ACTIONS(1974),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1972),
    [anon_sym_DQUOTE] = ACTIONS(1974),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1972),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1972),
    [anon_sym_switch] = ACTIONS(1974),
    [anon_sym_case] = ACTIONS(1974),
    [anon_sym_default] = ACTIONS(1974),
    [anon_sym_try] = ACTIONS(1974),
    [anon_sym_int] = ACTIONS(1974),
    [anon_sym_boolean] = ACTIONS(1974),
    [anon_sym_char] = ACTIONS(1974),
    [anon_sym_short] = ACTIONS(1974),
    [anon_sym_long] = ACTIONS(1974),
    [anon_sym_float] = ACTIONS(1974),
    [anon_sym_double] = ACTIONS(1974),
    [anon_sym_void] = ACTIONS(1974),
    [anon_sym_public] = ACTIONS(1974),
    [anon_sym_protected] = ACTIONS(1974),
    [anon_sym_private] = ACTIONS(1974),
    [anon_sym_static] = ACTIONS(1974),
    [anon_sym_final] = ACTIONS(1974),
    [anon_sym_synchronized] = ACTIONS(1974),
    [anon_sym_TILDE] = ACTIONS(1972),
    [anon_sym_BANG] = ACTIONS(1972),
    [anon_sym_new] = ACTIONS(1974),
  },
  [STATE(999)] = {
    [sym_comment] = STATE(999),
    [sym_groovy_doc] = STATE(999),
    [sym_identifier] = ACTIONS(1828),
    [anon_sym_break] = ACTIONS(1828),
    [anon_sym_continue] = ACTIONS(1828),
    [anon_sym_SEMI] = ACTIONS(1826),
    [anon_sym_STAR] = ACTIONS(1828),
    [anon_sym_STAR_COLON] = ACTIONS(1826),
    [anon_sym_import] = ACTIONS(1828),
    [anon_sym_package] = ACTIONS(1828),
    [anon_sym_AT] = ACTIONS(1828),
    [anon_sym_assert] = ACTIONS(1828),
    [anon_sym_PLUS_PLUS] = ACTIONS(1826),
    [anon_sym_DASH_DASH] = ACTIONS(1826),
    [anon_sym_SLASH] = ACTIONS(1828),
    [anon_sym_PLUS] = ACTIONS(1828),
    [anon_sym_DASH] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1828),
    [anon_sym_false] = ACTIONS(1828),
    [anon_sym_ATinterface] = ACTIONS(1826),
    [anon_sym_interface] = ACTIONS(1828),
    [anon_sym_class] = ACTIONS(1828),
    [anon_sym_LBRACE] = ACTIONS(1826),
    [anon_sym_RBRACE] = ACTIONS(1826),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1828),
    [anon_sym_def] = ACTIONS(1828),
    [anon_sym_LPAREN] = ACTIONS(1826),
    [anon_sym_null] = ACTIONS(1828),
    [anon_sym_this] = ACTIONS(1828),
    [anon_sym_do] = ACTIONS(1828),
    [anon_sym_while] = ACTIONS(1828),
    [anon_sym_for] = ACTIONS(1828),
    [aux_sym__type_identifier_token1] = ACTIONS(1828),
    [anon_sym_if] = ACTIONS(1828),
    [anon_sym_else] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1826),
    [aux_sym_number_literal_token1] = ACTIONS(1828),
    [aux_sym_number_literal_token2] = ACTIONS(1826),
    [aux_sym_number_literal_token3] = ACTIONS(1826),
    [aux_sym_number_literal_token4] = ACTIONS(1828),
    [aux_sym_number_literal_token5] = ACTIONS(1826),
    [anon_sym_return] = ACTIONS(1828),
    [anon_sym_SQUOTE] = ACTIONS(1828),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1826),
    [anon_sym_DQUOTE] = ACTIONS(1828),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1826),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1826),
    [anon_sym_switch] = ACTIONS(1828),
    [anon_sym_case] = ACTIONS(1828),
    [anon_sym_default] = ACTIONS(1828),
    [anon_sym_try] = ACTIONS(1828),
    [anon_sym_int] = ACTIONS(1828),
    [anon_sym_boolean] = ACTIONS(1828),
    [anon_sym_char] = ACTIONS(1828),
    [anon_sym_short] = ACTIONS(1828),
    [anon_sym_long] = ACTIONS(1828),
    [anon_sym_float] = ACTIONS(1828),
    [anon_sym_double] = ACTIONS(1828),
    [anon_sym_void] = ACTIONS(1828),
    [anon_sym_public] = ACTIONS(1828),
    [anon_sym_protected] = ACTIONS(1828),
    [anon_sym_private] = ACTIONS(1828),
    [anon_sym_static] = ACTIONS(1828),
    [anon_sym_final] = ACTIONS(1828),
    [anon_sym_synchronized] = ACTIONS(1828),
    [anon_sym_TILDE] = ACTIONS(1826),
    [anon_sym_BANG] = ACTIONS(1826),
    [anon_sym_new] = ACTIONS(1828),
  },
  [STATE(1000)] = {
    [sym_comment] = STATE(1000),
    [sym_groovy_doc] = STATE(1000),
    [anon_sym_SEMI] = ACTIONS(97),
    [anon_sym_DOT_AMP] = ACTIONS(97),
    [anon_sym_DOT_AT] = ACTIONS(97),
    [anon_sym_QMARK_DOT] = ACTIONS(97),
    [anon_sym_STAR_DOT] = ACTIONS(97),
    [anon_sym_STAR] = ACTIONS(102),
    [anon_sym_DOT] = ACTIONS(102),
    [anon_sym_as] = ACTIONS(97),
    [anon_sym_EQ] = ACTIONS(1196),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1198),
    [anon_sym_STAR_EQ] = ACTIONS(1198),
    [anon_sym_SLASH_EQ] = ACTIONS(1198),
    [anon_sym_PERCENT_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_EQ] = ACTIONS(1198),
    [anon_sym_DASH_EQ] = ACTIONS(1198),
    [anon_sym_LT_LT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1198),
    [anon_sym_AMP_EQ] = ACTIONS(1198),
    [anon_sym_CARET_EQ] = ACTIONS(1198),
    [anon_sym_PIPE_EQ] = ACTIONS(1198),
    [anon_sym_QMARK_EQ] = ACTIONS(1198),
    [anon_sym_PLUS_PLUS] = ACTIONS(97),
    [anon_sym_DASH_DASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(102),
    [anon_sym_SLASH] = ACTIONS(102),
    [anon_sym_PLUS] = ACTIONS(102),
    [anon_sym_DASH] = ACTIONS(102),
    [anon_sym_LT_LT] = ACTIONS(102),
    [anon_sym_GT_GT] = ACTIONS(102),
    [anon_sym_GT_GT_GT] = ACTIONS(102),
    [anon_sym_DOT_DOT] = ACTIONS(102),
    [anon_sym_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT_LT] = ACTIONS(97),
    [anon_sym_LT_DOT_DOT] = ACTIONS(102),
    [anon_sym_LT] = ACTIONS(102),
    [anon_sym_LT_EQ] = ACTIONS(102),
    [anon_sym_GT] = ACTIONS(102),
    [anon_sym_GT_EQ] = ACTIONS(97),
    [anon_sym_in] = ACTIONS(102),
    [anon_sym_BANGin] = ACTIONS(102),
    [anon_sym_instanceof] = ACTIONS(97),
    [anon_sym_BANGinstanceof] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(102),
    [anon_sym_BANG_EQ] = ACTIONS(102),
    [anon_sym_LT_EQ_GT] = ACTIONS(97),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(97),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(97),
    [anon_sym_EQ_TILDE] = ACTIONS(97),
    [anon_sym_EQ_EQ_TILDE] = ACTIONS(97),
    [anon_sym_AMP] = ACTIONS(102),
    [anon_sym_CARET] = ACTIONS(102),
    [anon_sym_PIPE] = ACTIONS(102),
    [anon_sym_AMP_AMP] = ACTIONS(97),
    [anon_sym_PIPE_PIPE] = ACTIONS(97),
    [anon_sym_QMARK_COLON] = ACTIONS(97),
    [anon_sym_STAR_STAR] = ACTIONS(102),
    [anon_sym_COMMA] = ACTIONS(97),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(97),
    [anon_sym_RPAREN] = ACTIONS(97),
    [anon_sym_while] = ACTIONS(97),
    [anon_sym_else] = ACTIONS(97),
    [anon_sym_LBRACK] = ACTIONS(97),
    [anon_sym_QMARK] = ACTIONS(102),
    [anon_sym_catch] = ACTIONS(97),
    [anon_sym_finally] = ACTIONS(97),
  },
  [STATE(1001)] = {
    [sym_comment] = STATE(1001),
    [sym_groovy_doc] = STATE(1001),
    [sym_identifier] = ACTIONS(1513),
    [anon_sym_break] = ACTIONS(1513),
    [anon_sym_continue] = ACTIONS(1513),
    [anon_sym_SEMI] = ACTIONS(1686),
    [anon_sym_STAR] = ACTIONS(1513),
    [anon_sym_STAR_COLON] = ACTIONS(1515),
    [anon_sym_import] = ACTIONS(1513),
    [anon_sym_package] = ACTIONS(1513),
    [anon_sym_AT] = ACTIONS(1513),
    [anon_sym_assert] = ACTIONS(1513),
    [anon_sym_PLUS_PLUS] = ACTIONS(1515),
    [anon_sym_DASH_DASH] = ACTIONS(1515),
    [anon_sym_SLASH] = ACTIONS(1513),
    [anon_sym_PLUS] = ACTIONS(1513),
    [anon_sym_DASH] = ACTIONS(1513),
    [anon_sym_true] = ACTIONS(1513),
    [anon_sym_false] = ACTIONS(1513),
    [anon_sym_ATinterface] = ACTIONS(1515),
    [anon_sym_interface] = ACTIONS(1513),
    [anon_sym_class] = ACTIONS(1513),
    [anon_sym_LBRACE] = ACTIONS(1515),
    [anon_sym_RBRACE] = ACTIONS(1515),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1513),
    [anon_sym_def] = ACTIONS(1513),
    [anon_sym_LPAREN] = ACTIONS(1515),
    [anon_sym_null] = ACTIONS(1513),
    [anon_sym_this] = ACTIONS(1513),
    [anon_sym_do] = ACTIONS(1513),
    [anon_sym_while] = ACTIONS(1513),
    [anon_sym_for] = ACTIONS(1513),
    [aux_sym__type_identifier_token1] = ACTIONS(1513),
    [anon_sym_if] = ACTIONS(1513),
    [anon_sym_LBRACK] = ACTIONS(1515),
    [aux_sym_number_literal_token1] = ACTIONS(1513),
    [aux_sym_number_literal_token2] = ACTIONS(1515),
    [aux_sym_number_literal_token3] = ACTIONS(1515),
    [aux_sym_number_literal_token4] = ACTIONS(1513),
    [aux_sym_number_literal_token5] = ACTIONS(1515),
    [anon_sym_return] = ACTIONS(1513),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1515),
    [anon_sym_DQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1515),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1515),
    [anon_sym_switch] = ACTIONS(1513),
    [anon_sym_case] = ACTIONS(1513),
    [anon_sym_default] = ACTIONS(1513),
    [anon_sym_try] = ACTIONS(1513),
    [anon_sym_int] = ACTIONS(1513),
    [anon_sym_boolean] = ACTIONS(1513),
    [anon_sym_char] = ACTIONS(1513),
    [anon_sym_short] = ACTIONS(1513),
    [anon_sym_long] = ACTIONS(1513),
    [anon_sym_float] = ACTIONS(1513),
    [anon_sym_double] = ACTIONS(1513),
    [anon_sym_void] = ACTIONS(1513),
    [anon_sym_public] = ACTIONS(1513),
    [anon_sym_protected] = ACTIONS(1513),
    [anon_sym_private] = ACTIONS(1513),
    [anon_sym_static] = ACTIONS(1513),
    [anon_sym_final] = ACTIONS(1513),
    [anon_sym_synchronized] = ACTIONS(1513),
    [anon_sym_TILDE] = ACTIONS(1515),
    [anon_sym_BANG] = ACTIONS(1515),
    [anon_sym_new] = ACTIONS(1513),
  },
  [STATE(1002)] = {
    [sym_comment] = STATE(1002),
    [sym_groovy_doc] = STATE(1002),
    [sym_identifier] = ACTIONS(1517),
    [anon_sym_break] = ACTIONS(1517),
    [anon_sym_continue] = ACTIONS(1517),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_STAR] = ACTIONS(1517),
    [anon_sym_STAR_COLON] = ACTIONS(1519),
    [anon_sym_import] = ACTIONS(1517),
    [anon_sym_package] = ACTIONS(1517),
    [anon_sym_AT] = ACTIONS(1517),
    [anon_sym_assert] = ACTIONS(1517),
    [anon_sym_PLUS_PLUS] = ACTIONS(1519),
    [anon_sym_DASH_DASH] = ACTIONS(1519),
    [anon_sym_SLASH] = ACTIONS(1517),
    [anon_sym_PLUS] = ACTIONS(1517),
    [anon_sym_DASH] = ACTIONS(1517),
    [anon_sym_true] = ACTIONS(1517),
    [anon_sym_false] = ACTIONS(1517),
    [anon_sym_ATinterface] = ACTIONS(1519),
    [anon_sym_interface] = ACTIONS(1517),
    [anon_sym_class] = ACTIONS(1517),
    [anon_sym_LBRACE] = ACTIONS(1519),
    [anon_sym_RBRACE] = ACTIONS(1519),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(1517),
    [anon_sym_def] = ACTIONS(1517),
    [anon_sym_LPAREN] = ACTIONS(1519),
    [anon_sym_null] = ACTIONS(1517),
    [anon_sym_this] = ACTIONS(1517),
    [anon_sym_do] = ACTIONS(1517),
    [anon_sym_while] = ACTIONS(1517),
    [anon_sym_for] = ACTIONS(1517),
    [aux_sym__type_identifier_token1] = ACTIONS(1517),
    [anon_sym_if] = ACTIONS(1517),
    [anon_sym_LBRACK] = ACTIONS(1519),
    [aux_sym_number_literal_token1] = ACTIONS(1517),
    [aux_sym_number_literal_token2] = ACTIONS(1519),
    [aux_sym_number_literal_token3] = ACTIONS(1519),
    [aux_sym_number_literal_token4] = ACTIONS(1517),
    [aux_sym_number_literal_token5] = ACTIONS(1519),
    [anon_sym_return] = ACTIONS(1517),
    [anon_sym_SQUOTE] = ACTIONS(1517),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1519),
    [anon_sym_DQUOTE] = ACTIONS(1517),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1519),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(1519),
    [anon_sym_switch] = ACTIONS(1517),
    [anon_sym_case] = ACTIONS(1517),
    [anon_sym_default] = ACTIONS(1517),
    [anon_sym_try] = ACTIONS(1517),
    [anon_sym_int] = ACTIONS(1517),
    [anon_sym_boolean] = ACTIONS(1517),
    [anon_sym_char] = ACTIONS(1517),
    [anon_sym_short] = ACTIONS(1517),
    [anon_sym_long] = ACTIONS(1517),
    [anon_sym_float] = ACTIONS(1517),
    [anon_sym_double] = ACTIONS(1517),
    [anon_sym_void] = ACTIONS(1517),
    [anon_sym_public] = ACTIONS(1517),
    [anon_sym_protected] = ACTIONS(1517),
    [anon_sym_private] = ACTIONS(1517),
    [anon_sym_static] = ACTIONS(1517),
    [anon_sym_final] = ACTIONS(1517),
    [anon_sym_synchronized] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1519),
    [anon_sym_BANG] = ACTIONS(1519),
    [anon_sym_new] = ACTIONS(1517),
  },
  [STATE(1003)] = {
    [sym_comment] = STATE(1003),
    [sym_groovy_doc] = STATE(1003),
    [ts_builtin_sym_end] = ACTIONS(2157),
    [sym_identifier] = ACTIONS(2159),
    [anon_sym_break] = ACTIONS(2159),
    [anon_sym_continue] = ACTIONS(2159),
    [anon_sym_STAR] = ACTIONS(2159),
    [anon_sym_STAR_COLON] = ACTIONS(2157),
    [anon_sym_import] = ACTIONS(2159),
    [anon_sym_package] = ACTIONS(2159),
    [anon_sym_AT] = ACTIONS(2159),
    [anon_sym_assert] = ACTIONS(2159),
    [anon_sym_PLUS_PLUS] = ACTIONS(2157),
    [anon_sym_DASH_DASH] = ACTIONS(2157),
    [anon_sym_SLASH] = ACTIONS(2159),
    [anon_sym_PLUS] = ACTIONS(2159),
    [anon_sym_DASH] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2159),
    [anon_sym_false] = ACTIONS(2159),
    [anon_sym_ATinterface] = ACTIONS(2157),
    [anon_sym_interface] = ACTIONS(2159),
    [anon_sym_class] = ACTIONS(2159),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_RBRACE] = ACTIONS(2157),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2159),
    [anon_sym_def] = ACTIONS(2159),
    [anon_sym_LPAREN] = ACTIONS(2157),
    [anon_sym_null] = ACTIONS(2159),
    [anon_sym_this] = ACTIONS(2159),
    [anon_sym_do] = ACTIONS(2159),
    [anon_sym_while] = ACTIONS(2159),
    [anon_sym_for] = ACTIONS(2159),
    [aux_sym__type_identifier_token1] = ACTIONS(2159),
    [anon_sym_if] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [aux_sym_number_literal_token1] = ACTIONS(2159),
    [aux_sym_number_literal_token2] = ACTIONS(2157),
    [aux_sym_number_literal_token3] = ACTIONS(2157),
    [aux_sym_number_literal_token4] = ACTIONS(2159),
    [aux_sym_number_literal_token5] = ACTIONS(2157),
    [anon_sym_pipeline] = ACTIONS(2159),
    [anon_sym_return] = ACTIONS(2159),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2157),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2157),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2157),
    [anon_sym_switch] = ACTIONS(2159),
    [anon_sym_try] = ACTIONS(2159),
    [anon_sym_int] = ACTIONS(2159),
    [anon_sym_boolean] = ACTIONS(2159),
    [anon_sym_char] = ACTIONS(2159),
    [anon_sym_short] = ACTIONS(2159),
    [anon_sym_long] = ACTIONS(2159),
    [anon_sym_float] = ACTIONS(2159),
    [anon_sym_double] = ACTIONS(2159),
    [anon_sym_void] = ACTIONS(2159),
    [anon_sym_public] = ACTIONS(2159),
    [anon_sym_protected] = ACTIONS(2159),
    [anon_sym_private] = ACTIONS(2159),
    [anon_sym_static] = ACTIONS(2159),
    [anon_sym_final] = ACTIONS(2159),
    [anon_sym_synchronized] = ACTIONS(2159),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_BANG] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2159),
  },
  [STATE(1004)] = {
    [sym_comment] = STATE(1004),
    [sym_groovy_doc] = STATE(1004),
    [sym_identifier] = ACTIONS(2159),
    [anon_sym_break] = ACTIONS(2159),
    [anon_sym_continue] = ACTIONS(2159),
    [anon_sym_STAR] = ACTIONS(2159),
    [anon_sym_STAR_COLON] = ACTIONS(2157),
    [anon_sym_import] = ACTIONS(2159),
    [anon_sym_package] = ACTIONS(2159),
    [anon_sym_AT] = ACTIONS(2159),
    [anon_sym_assert] = ACTIONS(2159),
    [anon_sym_PLUS_PLUS] = ACTIONS(2157),
    [anon_sym_DASH_DASH] = ACTIONS(2157),
    [anon_sym_SLASH] = ACTIONS(2159),
    [anon_sym_PLUS] = ACTIONS(2159),
    [anon_sym_DASH] = ACTIONS(2159),
    [anon_sym_true] = ACTIONS(2159),
    [anon_sym_false] = ACTIONS(2159),
    [anon_sym_ATinterface] = ACTIONS(2157),
    [anon_sym_interface] = ACTIONS(2159),
    [anon_sym_class] = ACTIONS(2159),
    [anon_sym_LBRACE] = ACTIONS(2157),
    [anon_sym_RBRACE] = ACTIONS(2157),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2159),
    [anon_sym_def] = ACTIONS(2159),
    [anon_sym_LPAREN] = ACTIONS(2157),
    [anon_sym_null] = ACTIONS(2159),
    [anon_sym_this] = ACTIONS(2159),
    [anon_sym_do] = ACTIONS(2159),
    [anon_sym_while] = ACTIONS(2159),
    [anon_sym_for] = ACTIONS(2159),
    [aux_sym__type_identifier_token1] = ACTIONS(2159),
    [anon_sym_if] = ACTIONS(2159),
    [anon_sym_LBRACK] = ACTIONS(2157),
    [aux_sym_number_literal_token1] = ACTIONS(2159),
    [aux_sym_number_literal_token2] = ACTIONS(2157),
    [aux_sym_number_literal_token3] = ACTIONS(2157),
    [aux_sym_number_literal_token4] = ACTIONS(2159),
    [aux_sym_number_literal_token5] = ACTIONS(2157),
    [anon_sym_return] = ACTIONS(2159),
    [anon_sym_SQUOTE] = ACTIONS(2159),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2157),
    [anon_sym_DQUOTE] = ACTIONS(2159),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2157),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2157),
    [anon_sym_switch] = ACTIONS(2159),
    [anon_sym_case] = ACTIONS(2159),
    [anon_sym_default] = ACTIONS(2159),
    [anon_sym_try] = ACTIONS(2159),
    [anon_sym_int] = ACTIONS(2159),
    [anon_sym_boolean] = ACTIONS(2159),
    [anon_sym_char] = ACTIONS(2159),
    [anon_sym_short] = ACTIONS(2159),
    [anon_sym_long] = ACTIONS(2159),
    [anon_sym_float] = ACTIONS(2159),
    [anon_sym_double] = ACTIONS(2159),
    [anon_sym_void] = ACTIONS(2159),
    [anon_sym_public] = ACTIONS(2159),
    [anon_sym_protected] = ACTIONS(2159),
    [anon_sym_private] = ACTIONS(2159),
    [anon_sym_static] = ACTIONS(2159),
    [anon_sym_final] = ACTIONS(2159),
    [anon_sym_synchronized] = ACTIONS(2159),
    [anon_sym_TILDE] = ACTIONS(2157),
    [anon_sym_BANG] = ACTIONS(2157),
    [anon_sym_new] = ACTIONS(2159),
  },
  [STATE(1005)] = {
    [sym_comment] = STATE(1005),
    [sym_groovy_doc] = STATE(1005),
    [ts_builtin_sym_end] = ACTIONS(2161),
    [sym_identifier] = ACTIONS(2163),
    [anon_sym_break] = ACTIONS(2163),
    [anon_sym_continue] = ACTIONS(2163),
    [anon_sym_STAR] = ACTIONS(2163),
    [anon_sym_STAR_COLON] = ACTIONS(2161),
    [anon_sym_import] = ACTIONS(2163),
    [anon_sym_package] = ACTIONS(2163),
    [anon_sym_AT] = ACTIONS(2163),
    [anon_sym_assert] = ACTIONS(2163),
    [anon_sym_PLUS_PLUS] = ACTIONS(2161),
    [anon_sym_DASH_DASH] = ACTIONS(2161),
    [anon_sym_SLASH] = ACTIONS(2163),
    [anon_sym_PLUS] = ACTIONS(2163),
    [anon_sym_DASH] = ACTIONS(2163),
    [anon_sym_true] = ACTIONS(2163),
    [anon_sym_false] = ACTIONS(2163),
    [anon_sym_ATinterface] = ACTIONS(2161),
    [anon_sym_interface] = ACTIONS(2163),
    [anon_sym_class] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2161),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2163),
    [anon_sym_def] = ACTIONS(2163),
    [anon_sym_LPAREN] = ACTIONS(2161),
    [anon_sym_null] = ACTIONS(2163),
    [anon_sym_this] = ACTIONS(2163),
    [anon_sym_do] = ACTIONS(2163),
    [anon_sym_while] = ACTIONS(2163),
    [anon_sym_for] = ACTIONS(2163),
    [aux_sym__type_identifier_token1] = ACTIONS(2163),
    [anon_sym_if] = ACTIONS(2163),
    [anon_sym_LBRACK] = ACTIONS(2161),
    [aux_sym_number_literal_token1] = ACTIONS(2163),
    [aux_sym_number_literal_token2] = ACTIONS(2161),
    [aux_sym_number_literal_token3] = ACTIONS(2161),
    [aux_sym_number_literal_token4] = ACTIONS(2163),
    [aux_sym_number_literal_token5] = ACTIONS(2161),
    [anon_sym_pipeline] = ACTIONS(2163),
    [anon_sym_return] = ACTIONS(2163),
    [anon_sym_SQUOTE] = ACTIONS(2163),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2161),
    [anon_sym_DQUOTE] = ACTIONS(2163),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2161),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2161),
    [anon_sym_switch] = ACTIONS(2163),
    [anon_sym_try] = ACTIONS(2163),
    [anon_sym_int] = ACTIONS(2163),
    [anon_sym_boolean] = ACTIONS(2163),
    [anon_sym_char] = ACTIONS(2163),
    [anon_sym_short] = ACTIONS(2163),
    [anon_sym_long] = ACTIONS(2163),
    [anon_sym_float] = ACTIONS(2163),
    [anon_sym_double] = ACTIONS(2163),
    [anon_sym_void] = ACTIONS(2163),
    [anon_sym_public] = ACTIONS(2163),
    [anon_sym_protected] = ACTIONS(2163),
    [anon_sym_private] = ACTIONS(2163),
    [anon_sym_static] = ACTIONS(2163),
    [anon_sym_final] = ACTIONS(2163),
    [anon_sym_synchronized] = ACTIONS(2163),
    [anon_sym_TILDE] = ACTIONS(2161),
    [anon_sym_BANG] = ACTIONS(2161),
    [anon_sym_new] = ACTIONS(2163),
  },
  [STATE(1006)] = {
    [sym_comment] = STATE(1006),
    [sym_groovy_doc] = STATE(1006),
    [sym_identifier] = ACTIONS(2165),
    [anon_sym_break] = ACTIONS(2165),
    [anon_sym_continue] = ACTIONS(2165),
    [anon_sym_STAR] = ACTIONS(2165),
    [anon_sym_STAR_COLON] = ACTIONS(2167),
    [anon_sym_import] = ACTIONS(2165),
    [anon_sym_package] = ACTIONS(2165),
    [anon_sym_AT] = ACTIONS(2165),
    [anon_sym_assert] = ACTIONS(2165),
    [anon_sym_PLUS_PLUS] = ACTIONS(2167),
    [anon_sym_DASH_DASH] = ACTIONS(2167),
    [anon_sym_SLASH] = ACTIONS(2165),
    [anon_sym_PLUS] = ACTIONS(2165),
    [anon_sym_DASH] = ACTIONS(2165),
    [anon_sym_true] = ACTIONS(2165),
    [anon_sym_false] = ACTIONS(2165),
    [anon_sym_ATinterface] = ACTIONS(2167),
    [anon_sym_interface] = ACTIONS(2165),
    [anon_sym_class] = ACTIONS(2165),
    [anon_sym_LBRACE] = ACTIONS(2167),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2165),
    [anon_sym_def] = ACTIONS(2165),
    [anon_sym_LPAREN] = ACTIONS(2167),
    [anon_sym_null] = ACTIONS(2165),
    [anon_sym_this] = ACTIONS(2165),
    [anon_sym_do] = ACTIONS(2165),
    [anon_sym_while] = ACTIONS(2165),
    [anon_sym_for] = ACTIONS(2165),
    [aux_sym__type_identifier_token1] = ACTIONS(2165),
    [anon_sym_if] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2167),
    [aux_sym_number_literal_token1] = ACTIONS(2165),
    [aux_sym_number_literal_token2] = ACTIONS(2167),
    [aux_sym_number_literal_token3] = ACTIONS(2167),
    [aux_sym_number_literal_token4] = ACTIONS(2165),
    [aux_sym_number_literal_token5] = ACTIONS(2167),
    [anon_sym_return] = ACTIONS(2165),
    [anon_sym_SQUOTE] = ACTIONS(2165),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2167),
    [anon_sym_DQUOTE] = ACTIONS(2165),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2167),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2167),
    [anon_sym_switch] = ACTIONS(2165),
    [anon_sym_try] = ACTIONS(2165),
    [anon_sym_int] = ACTIONS(2165),
    [anon_sym_boolean] = ACTIONS(2165),
    [anon_sym_char] = ACTIONS(2165),
    [anon_sym_short] = ACTIONS(2165),
    [anon_sym_long] = ACTIONS(2165),
    [anon_sym_float] = ACTIONS(2165),
    [anon_sym_double] = ACTIONS(2165),
    [anon_sym_void] = ACTIONS(2165),
    [anon_sym_public] = ACTIONS(2165),
    [anon_sym_protected] = ACTIONS(2165),
    [anon_sym_private] = ACTIONS(2165),
    [anon_sym_static] = ACTIONS(2165),
    [anon_sym_final] = ACTIONS(2165),
    [anon_sym_synchronized] = ACTIONS(2165),
    [anon_sym_TILDE] = ACTIONS(2167),
    [anon_sym_BANG] = ACTIONS(2167),
    [anon_sym_new] = ACTIONS(2165),
  },
  [STATE(1007)] = {
    [sym_comment] = STATE(1007),
    [sym_groovy_doc] = STATE(1007),
    [sym_identifier] = ACTIONS(2169),
    [anon_sym_break] = ACTIONS(2169),
    [anon_sym_continue] = ACTIONS(2169),
    [anon_sym_STAR] = ACTIONS(2169),
    [anon_sym_STAR_COLON] = ACTIONS(2171),
    [anon_sym_import] = ACTIONS(2169),
    [anon_sym_package] = ACTIONS(2169),
    [anon_sym_AT] = ACTIONS(2169),
    [anon_sym_assert] = ACTIONS(2169),
    [anon_sym_PLUS_PLUS] = ACTIONS(2171),
    [anon_sym_DASH_DASH] = ACTIONS(2171),
    [anon_sym_SLASH] = ACTIONS(2169),
    [anon_sym_PLUS] = ACTIONS(2169),
    [anon_sym_DASH] = ACTIONS(2169),
    [anon_sym_true] = ACTIONS(2169),
    [anon_sym_false] = ACTIONS(2169),
    [anon_sym_ATinterface] = ACTIONS(2171),
    [anon_sym_interface] = ACTIONS(2169),
    [anon_sym_class] = ACTIONS(2169),
    [anon_sym_LBRACE] = ACTIONS(2171),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2169),
    [anon_sym_def] = ACTIONS(2169),
    [anon_sym_LPAREN] = ACTIONS(2171),
    [anon_sym_null] = ACTIONS(2169),
    [anon_sym_this] = ACTIONS(2169),
    [anon_sym_do] = ACTIONS(2169),
    [anon_sym_while] = ACTIONS(2169),
    [anon_sym_for] = ACTIONS(2169),
    [aux_sym__type_identifier_token1] = ACTIONS(2169),
    [anon_sym_if] = ACTIONS(2169),
    [anon_sym_LBRACK] = ACTIONS(2171),
    [aux_sym_number_literal_token1] = ACTIONS(2169),
    [aux_sym_number_literal_token2] = ACTIONS(2171),
    [aux_sym_number_literal_token3] = ACTIONS(2171),
    [aux_sym_number_literal_token4] = ACTIONS(2169),
    [aux_sym_number_literal_token5] = ACTIONS(2171),
    [anon_sym_return] = ACTIONS(2169),
    [anon_sym_SQUOTE] = ACTIONS(2169),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2171),
    [anon_sym_DQUOTE] = ACTIONS(2169),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2171),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2171),
    [anon_sym_switch] = ACTIONS(2169),
    [anon_sym_try] = ACTIONS(2169),
    [anon_sym_int] = ACTIONS(2169),
    [anon_sym_boolean] = ACTIONS(2169),
    [anon_sym_char] = ACTIONS(2169),
    [anon_sym_short] = ACTIONS(2169),
    [anon_sym_long] = ACTIONS(2169),
    [anon_sym_float] = ACTIONS(2169),
    [anon_sym_double] = ACTIONS(2169),
    [anon_sym_void] = ACTIONS(2169),
    [anon_sym_public] = ACTIONS(2169),
    [anon_sym_protected] = ACTIONS(2169),
    [anon_sym_private] = ACTIONS(2169),
    [anon_sym_static] = ACTIONS(2169),
    [anon_sym_final] = ACTIONS(2169),
    [anon_sym_synchronized] = ACTIONS(2169),
    [anon_sym_TILDE] = ACTIONS(2171),
    [anon_sym_BANG] = ACTIONS(2171),
    [anon_sym_new] = ACTIONS(2169),
  },
  [STATE(1008)] = {
    [sym_comment] = STATE(1008),
    [sym_groovy_doc] = STATE(1008),
    [sym_identifier] = ACTIONS(2173),
    [anon_sym_break] = ACTIONS(2173),
    [anon_sym_continue] = ACTIONS(2173),
    [anon_sym_STAR] = ACTIONS(2173),
    [anon_sym_STAR_COLON] = ACTIONS(2175),
    [anon_sym_import] = ACTIONS(2173),
    [anon_sym_package] = ACTIONS(2173),
    [anon_sym_AT] = ACTIONS(2173),
    [anon_sym_assert] = ACTIONS(2173),
    [anon_sym_PLUS_PLUS] = ACTIONS(2175),
    [anon_sym_DASH_DASH] = ACTIONS(2175),
    [anon_sym_SLASH] = ACTIONS(2173),
    [anon_sym_PLUS] = ACTIONS(2173),
    [anon_sym_DASH] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2173),
    [anon_sym_false] = ACTIONS(2173),
    [anon_sym_ATinterface] = ACTIONS(2175),
    [anon_sym_interface] = ACTIONS(2173),
    [anon_sym_class] = ACTIONS(2173),
    [anon_sym_LBRACE] = ACTIONS(2175),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2173),
    [anon_sym_def] = ACTIONS(2173),
    [anon_sym_LPAREN] = ACTIONS(2175),
    [anon_sym_null] = ACTIONS(2173),
    [anon_sym_this] = ACTIONS(2173),
    [anon_sym_do] = ACTIONS(2173),
    [anon_sym_while] = ACTIONS(2173),
    [anon_sym_for] = ACTIONS(2173),
    [aux_sym__type_identifier_token1] = ACTIONS(2173),
    [anon_sym_if] = ACTIONS(2173),
    [anon_sym_LBRACK] = ACTIONS(2175),
    [aux_sym_number_literal_token1] = ACTIONS(2173),
    [aux_sym_number_literal_token2] = ACTIONS(2175),
    [aux_sym_number_literal_token3] = ACTIONS(2175),
    [aux_sym_number_literal_token4] = ACTIONS(2173),
    [aux_sym_number_literal_token5] = ACTIONS(2175),
    [anon_sym_return] = ACTIONS(2173),
    [anon_sym_SQUOTE] = ACTIONS(2173),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2175),
    [anon_sym_DQUOTE] = ACTIONS(2173),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2175),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2175),
    [anon_sym_switch] = ACTIONS(2173),
    [anon_sym_try] = ACTIONS(2173),
    [anon_sym_int] = ACTIONS(2173),
    [anon_sym_boolean] = ACTIONS(2173),
    [anon_sym_char] = ACTIONS(2173),
    [anon_sym_short] = ACTIONS(2173),
    [anon_sym_long] = ACTIONS(2173),
    [anon_sym_float] = ACTIONS(2173),
    [anon_sym_double] = ACTIONS(2173),
    [anon_sym_void] = ACTIONS(2173),
    [anon_sym_public] = ACTIONS(2173),
    [anon_sym_protected] = ACTIONS(2173),
    [anon_sym_private] = ACTIONS(2173),
    [anon_sym_static] = ACTIONS(2173),
    [anon_sym_final] = ACTIONS(2173),
    [anon_sym_synchronized] = ACTIONS(2173),
    [anon_sym_TILDE] = ACTIONS(2175),
    [anon_sym_BANG] = ACTIONS(2175),
    [anon_sym_new] = ACTIONS(2173),
  },
  [STATE(1009)] = {
    [sym_comment] = STATE(1009),
    [sym_groovy_doc] = STATE(1009),
    [sym_identifier] = ACTIONS(2177),
    [anon_sym_break] = ACTIONS(2177),
    [anon_sym_continue] = ACTIONS(2177),
    [anon_sym_STAR] = ACTIONS(2177),
    [anon_sym_STAR_COLON] = ACTIONS(2179),
    [anon_sym_import] = ACTIONS(2177),
    [anon_sym_package] = ACTIONS(2177),
    [anon_sym_AT] = ACTIONS(2177),
    [anon_sym_assert] = ACTIONS(2177),
    [anon_sym_PLUS_PLUS] = ACTIONS(2179),
    [anon_sym_DASH_DASH] = ACTIONS(2179),
    [anon_sym_SLASH] = ACTIONS(2177),
    [anon_sym_PLUS] = ACTIONS(2177),
    [anon_sym_DASH] = ACTIONS(2177),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [anon_sym_ATinterface] = ACTIONS(2179),
    [anon_sym_interface] = ACTIONS(2177),
    [anon_sym_class] = ACTIONS(2177),
    [anon_sym_LBRACE] = ACTIONS(2179),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2177),
    [anon_sym_def] = ACTIONS(2177),
    [anon_sym_LPAREN] = ACTIONS(2179),
    [anon_sym_null] = ACTIONS(2177),
    [anon_sym_this] = ACTIONS(2177),
    [anon_sym_do] = ACTIONS(2177),
    [anon_sym_while] = ACTIONS(2177),
    [anon_sym_for] = ACTIONS(2177),
    [aux_sym__type_identifier_token1] = ACTIONS(2177),
    [anon_sym_if] = ACTIONS(2177),
    [anon_sym_LBRACK] = ACTIONS(2179),
    [aux_sym_number_literal_token1] = ACTIONS(2177),
    [aux_sym_number_literal_token2] = ACTIONS(2179),
    [aux_sym_number_literal_token3] = ACTIONS(2179),
    [aux_sym_number_literal_token4] = ACTIONS(2177),
    [aux_sym_number_literal_token5] = ACTIONS(2179),
    [anon_sym_return] = ACTIONS(2177),
    [anon_sym_SQUOTE] = ACTIONS(2177),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2179),
    [anon_sym_DQUOTE] = ACTIONS(2177),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2179),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2179),
    [anon_sym_switch] = ACTIONS(2177),
    [anon_sym_try] = ACTIONS(2177),
    [anon_sym_int] = ACTIONS(2177),
    [anon_sym_boolean] = ACTIONS(2177),
    [anon_sym_char] = ACTIONS(2177),
    [anon_sym_short] = ACTIONS(2177),
    [anon_sym_long] = ACTIONS(2177),
    [anon_sym_float] = ACTIONS(2177),
    [anon_sym_double] = ACTIONS(2177),
    [anon_sym_void] = ACTIONS(2177),
    [anon_sym_public] = ACTIONS(2177),
    [anon_sym_protected] = ACTIONS(2177),
    [anon_sym_private] = ACTIONS(2177),
    [anon_sym_static] = ACTIONS(2177),
    [anon_sym_final] = ACTIONS(2177),
    [anon_sym_synchronized] = ACTIONS(2177),
    [anon_sym_TILDE] = ACTIONS(2179),
    [anon_sym_BANG] = ACTIONS(2179),
    [anon_sym_new] = ACTIONS(2177),
  },
  [STATE(1010)] = {
    [sym_comment] = STATE(1010),
    [sym_groovy_doc] = STATE(1010),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_STAR] = ACTIONS(2181),
    [anon_sym_STAR_COLON] = ACTIONS(2183),
    [anon_sym_import] = ACTIONS(2181),
    [anon_sym_package] = ACTIONS(2181),
    [anon_sym_AT] = ACTIONS(2181),
    [anon_sym_assert] = ACTIONS(2181),
    [anon_sym_PLUS_PLUS] = ACTIONS(2183),
    [anon_sym_DASH_DASH] = ACTIONS(2183),
    [anon_sym_SLASH] = ACTIONS(2181),
    [anon_sym_PLUS] = ACTIONS(2181),
    [anon_sym_DASH] = ACTIONS(2181),
    [anon_sym_true] = ACTIONS(2181),
    [anon_sym_false] = ACTIONS(2181),
    [anon_sym_ATinterface] = ACTIONS(2183),
    [anon_sym_interface] = ACTIONS(2181),
    [anon_sym_class] = ACTIONS(2181),
    [anon_sym_LBRACE] = ACTIONS(2183),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2181),
    [anon_sym_def] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_null] = ACTIONS(2181),
    [anon_sym_this] = ACTIONS(2181),
    [anon_sym_do] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [aux_sym__type_identifier_token1] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_LBRACK] = ACTIONS(2183),
    [aux_sym_number_literal_token1] = ACTIONS(2181),
    [aux_sym_number_literal_token2] = ACTIONS(2183),
    [aux_sym_number_literal_token3] = ACTIONS(2183),
    [aux_sym_number_literal_token4] = ACTIONS(2181),
    [aux_sym_number_literal_token5] = ACTIONS(2183),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2183),
    [anon_sym_DQUOTE] = ACTIONS(2181),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2183),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2183),
    [anon_sym_switch] = ACTIONS(2181),
    [anon_sym_try] = ACTIONS(2181),
    [anon_sym_int] = ACTIONS(2181),
    [anon_sym_boolean] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [anon_sym_short] = ACTIONS(2181),
    [anon_sym_long] = ACTIONS(2181),
    [anon_sym_float] = ACTIONS(2181),
    [anon_sym_double] = ACTIONS(2181),
    [anon_sym_void] = ACTIONS(2181),
    [anon_sym_public] = ACTIONS(2181),
    [anon_sym_protected] = ACTIONS(2181),
    [anon_sym_private] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_final] = ACTIONS(2181),
    [anon_sym_synchronized] = ACTIONS(2181),
    [anon_sym_TILDE] = ACTIONS(2183),
    [anon_sym_BANG] = ACTIONS(2183),
    [anon_sym_new] = ACTIONS(2181),
  },
  [STATE(1011)] = {
    [sym_comment] = STATE(1011),
    [sym_groovy_doc] = STATE(1011),
    [sym_identifier] = ACTIONS(2185),
    [anon_sym_break] = ACTIONS(2185),
    [anon_sym_continue] = ACTIONS(2185),
    [anon_sym_STAR] = ACTIONS(2185),
    [anon_sym_STAR_COLON] = ACTIONS(2187),
    [anon_sym_import] = ACTIONS(2185),
    [anon_sym_package] = ACTIONS(2185),
    [anon_sym_AT] = ACTIONS(2185),
    [anon_sym_assert] = ACTIONS(2185),
    [anon_sym_PLUS_PLUS] = ACTIONS(2187),
    [anon_sym_DASH_DASH] = ACTIONS(2187),
    [anon_sym_SLASH] = ACTIONS(2185),
    [anon_sym_PLUS] = ACTIONS(2185),
    [anon_sym_DASH] = ACTIONS(2185),
    [anon_sym_true] = ACTIONS(2185),
    [anon_sym_false] = ACTIONS(2185),
    [anon_sym_ATinterface] = ACTIONS(2187),
    [anon_sym_interface] = ACTIONS(2185),
    [anon_sym_class] = ACTIONS(2185),
    [anon_sym_LBRACE] = ACTIONS(2187),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2185),
    [anon_sym_def] = ACTIONS(2185),
    [anon_sym_LPAREN] = ACTIONS(2187),
    [anon_sym_null] = ACTIONS(2185),
    [anon_sym_this] = ACTIONS(2185),
    [anon_sym_do] = ACTIONS(2185),
    [anon_sym_while] = ACTIONS(2185),
    [anon_sym_for] = ACTIONS(2185),
    [aux_sym__type_identifier_token1] = ACTIONS(2185),
    [anon_sym_if] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2187),
    [aux_sym_number_literal_token1] = ACTIONS(2185),
    [aux_sym_number_literal_token2] = ACTIONS(2187),
    [aux_sym_number_literal_token3] = ACTIONS(2187),
    [aux_sym_number_literal_token4] = ACTIONS(2185),
    [aux_sym_number_literal_token5] = ACTIONS(2187),
    [anon_sym_return] = ACTIONS(2185),
    [anon_sym_SQUOTE] = ACTIONS(2185),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2187),
    [anon_sym_DQUOTE] = ACTIONS(2185),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2187),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2187),
    [anon_sym_switch] = ACTIONS(2185),
    [anon_sym_try] = ACTIONS(2185),
    [anon_sym_int] = ACTIONS(2185),
    [anon_sym_boolean] = ACTIONS(2185),
    [anon_sym_char] = ACTIONS(2185),
    [anon_sym_short] = ACTIONS(2185),
    [anon_sym_long] = ACTIONS(2185),
    [anon_sym_float] = ACTIONS(2185),
    [anon_sym_double] = ACTIONS(2185),
    [anon_sym_void] = ACTIONS(2185),
    [anon_sym_public] = ACTIONS(2185),
    [anon_sym_protected] = ACTIONS(2185),
    [anon_sym_private] = ACTIONS(2185),
    [anon_sym_static] = ACTIONS(2185),
    [anon_sym_final] = ACTIONS(2185),
    [anon_sym_synchronized] = ACTIONS(2185),
    [anon_sym_TILDE] = ACTIONS(2187),
    [anon_sym_BANG] = ACTIONS(2187),
    [anon_sym_new] = ACTIONS(2185),
  },
  [STATE(1012)] = {
    [sym_comment] = STATE(1012),
    [sym_groovy_doc] = STATE(1012),
    [sym_identifier] = ACTIONS(2189),
    [anon_sym_break] = ACTIONS(2189),
    [anon_sym_continue] = ACTIONS(2189),
    [anon_sym_STAR] = ACTIONS(2189),
    [anon_sym_STAR_COLON] = ACTIONS(2191),
    [anon_sym_import] = ACTIONS(2189),
    [anon_sym_package] = ACTIONS(2189),
    [anon_sym_AT] = ACTIONS(2189),
    [anon_sym_assert] = ACTIONS(2189),
    [anon_sym_PLUS_PLUS] = ACTIONS(2191),
    [anon_sym_DASH_DASH] = ACTIONS(2191),
    [anon_sym_SLASH] = ACTIONS(2189),
    [anon_sym_PLUS] = ACTIONS(2189),
    [anon_sym_DASH] = ACTIONS(2189),
    [anon_sym_true] = ACTIONS(2189),
    [anon_sym_false] = ACTIONS(2189),
    [anon_sym_ATinterface] = ACTIONS(2191),
    [anon_sym_interface] = ACTIONS(2189),
    [anon_sym_class] = ACTIONS(2189),
    [anon_sym_LBRACE] = ACTIONS(2191),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2189),
    [anon_sym_def] = ACTIONS(2189),
    [anon_sym_LPAREN] = ACTIONS(2191),
    [anon_sym_null] = ACTIONS(2189),
    [anon_sym_this] = ACTIONS(2189),
    [anon_sym_do] = ACTIONS(2189),
    [anon_sym_while] = ACTIONS(2189),
    [anon_sym_for] = ACTIONS(2189),
    [aux_sym__type_identifier_token1] = ACTIONS(2189),
    [anon_sym_if] = ACTIONS(2189),
    [anon_sym_LBRACK] = ACTIONS(2191),
    [aux_sym_number_literal_token1] = ACTIONS(2189),
    [aux_sym_number_literal_token2] = ACTIONS(2191),
    [aux_sym_number_literal_token3] = ACTIONS(2191),
    [aux_sym_number_literal_token4] = ACTIONS(2189),
    [aux_sym_number_literal_token5] = ACTIONS(2191),
    [anon_sym_return] = ACTIONS(2189),
    [anon_sym_SQUOTE] = ACTIONS(2189),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2191),
    [anon_sym_DQUOTE] = ACTIONS(2189),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2191),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2191),
    [anon_sym_switch] = ACTIONS(2189),
    [anon_sym_try] = ACTIONS(2189),
    [anon_sym_int] = ACTIONS(2189),
    [anon_sym_boolean] = ACTIONS(2189),
    [anon_sym_char] = ACTIONS(2189),
    [anon_sym_short] = ACTIONS(2189),
    [anon_sym_long] = ACTIONS(2189),
    [anon_sym_float] = ACTIONS(2189),
    [anon_sym_double] = ACTIONS(2189),
    [anon_sym_void] = ACTIONS(2189),
    [anon_sym_public] = ACTIONS(2189),
    [anon_sym_protected] = ACTIONS(2189),
    [anon_sym_private] = ACTIONS(2189),
    [anon_sym_static] = ACTIONS(2189),
    [anon_sym_final] = ACTIONS(2189),
    [anon_sym_synchronized] = ACTIONS(2189),
    [anon_sym_TILDE] = ACTIONS(2191),
    [anon_sym_BANG] = ACTIONS(2191),
    [anon_sym_new] = ACTIONS(2189),
  },
  [STATE(1013)] = {
    [sym_comment] = STATE(1013),
    [sym_groovy_doc] = STATE(1013),
    [sym_identifier] = ACTIONS(2193),
    [anon_sym_break] = ACTIONS(2193),
    [anon_sym_continue] = ACTIONS(2193),
    [anon_sym_STAR] = ACTIONS(2193),
    [anon_sym_STAR_COLON] = ACTIONS(2195),
    [anon_sym_import] = ACTIONS(2193),
    [anon_sym_package] = ACTIONS(2193),
    [anon_sym_AT] = ACTIONS(2193),
    [anon_sym_assert] = ACTIONS(2193),
    [anon_sym_PLUS_PLUS] = ACTIONS(2195),
    [anon_sym_DASH_DASH] = ACTIONS(2195),
    [anon_sym_SLASH] = ACTIONS(2193),
    [anon_sym_PLUS] = ACTIONS(2193),
    [anon_sym_DASH] = ACTIONS(2193),
    [anon_sym_true] = ACTIONS(2193),
    [anon_sym_false] = ACTIONS(2193),
    [anon_sym_ATinterface] = ACTIONS(2195),
    [anon_sym_interface] = ACTIONS(2193),
    [anon_sym_class] = ACTIONS(2193),
    [anon_sym_LBRACE] = ACTIONS(2195),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2193),
    [anon_sym_def] = ACTIONS(2193),
    [anon_sym_LPAREN] = ACTIONS(2195),
    [anon_sym_null] = ACTIONS(2193),
    [anon_sym_this] = ACTIONS(2193),
    [anon_sym_do] = ACTIONS(2193),
    [anon_sym_while] = ACTIONS(2193),
    [anon_sym_for] = ACTIONS(2193),
    [aux_sym__type_identifier_token1] = ACTIONS(2193),
    [anon_sym_if] = ACTIONS(2193),
    [anon_sym_LBRACK] = ACTIONS(2195),
    [aux_sym_number_literal_token1] = ACTIONS(2193),
    [aux_sym_number_literal_token2] = ACTIONS(2195),
    [aux_sym_number_literal_token3] = ACTIONS(2195),
    [aux_sym_number_literal_token4] = ACTIONS(2193),
    [aux_sym_number_literal_token5] = ACTIONS(2195),
    [anon_sym_return] = ACTIONS(2193),
    [anon_sym_SQUOTE] = ACTIONS(2193),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2195),
    [anon_sym_DQUOTE] = ACTIONS(2193),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2195),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2195),
    [anon_sym_switch] = ACTIONS(2193),
    [anon_sym_try] = ACTIONS(2193),
    [anon_sym_int] = ACTIONS(2193),
    [anon_sym_boolean] = ACTIONS(2193),
    [anon_sym_char] = ACTIONS(2193),
    [anon_sym_short] = ACTIONS(2193),
    [anon_sym_long] = ACTIONS(2193),
    [anon_sym_float] = ACTIONS(2193),
    [anon_sym_double] = ACTIONS(2193),
    [anon_sym_void] = ACTIONS(2193),
    [anon_sym_public] = ACTIONS(2193),
    [anon_sym_protected] = ACTIONS(2193),
    [anon_sym_private] = ACTIONS(2193),
    [anon_sym_static] = ACTIONS(2193),
    [anon_sym_final] = ACTIONS(2193),
    [anon_sym_synchronized] = ACTIONS(2193),
    [anon_sym_TILDE] = ACTIONS(2195),
    [anon_sym_BANG] = ACTIONS(2195),
    [anon_sym_new] = ACTIONS(2193),
  },
  [STATE(1014)] = {
    [sym_comment] = STATE(1014),
    [sym_groovy_doc] = STATE(1014),
    [sym_identifier] = ACTIONS(2197),
    [anon_sym_break] = ACTIONS(2197),
    [anon_sym_continue] = ACTIONS(2197),
    [anon_sym_STAR] = ACTIONS(2197),
    [anon_sym_STAR_COLON] = ACTIONS(2199),
    [anon_sym_import] = ACTIONS(2197),
    [anon_sym_package] = ACTIONS(2197),
    [anon_sym_AT] = ACTIONS(2197),
    [anon_sym_assert] = ACTIONS(2197),
    [anon_sym_PLUS_PLUS] = ACTIONS(2199),
    [anon_sym_DASH_DASH] = ACTIONS(2199),
    [anon_sym_SLASH] = ACTIONS(2197),
    [anon_sym_PLUS] = ACTIONS(2197),
    [anon_sym_DASH] = ACTIONS(2197),
    [anon_sym_true] = ACTIONS(2197),
    [anon_sym_false] = ACTIONS(2197),
    [anon_sym_ATinterface] = ACTIONS(2199),
    [anon_sym_interface] = ACTIONS(2197),
    [anon_sym_class] = ACTIONS(2197),
    [anon_sym_LBRACE] = ACTIONS(2199),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2197),
    [anon_sym_def] = ACTIONS(2197),
    [anon_sym_LPAREN] = ACTIONS(2199),
    [anon_sym_null] = ACTIONS(2197),
    [anon_sym_this] = ACTIONS(2197),
    [anon_sym_do] = ACTIONS(2197),
    [anon_sym_while] = ACTIONS(2197),
    [anon_sym_for] = ACTIONS(2197),
    [aux_sym__type_identifier_token1] = ACTIONS(2197),
    [anon_sym_if] = ACTIONS(2197),
    [anon_sym_LBRACK] = ACTIONS(2199),
    [aux_sym_number_literal_token1] = ACTIONS(2197),
    [aux_sym_number_literal_token2] = ACTIONS(2199),
    [aux_sym_number_literal_token3] = ACTIONS(2199),
    [aux_sym_number_literal_token4] = ACTIONS(2197),
    [aux_sym_number_literal_token5] = ACTIONS(2199),
    [anon_sym_return] = ACTIONS(2197),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2199),
    [anon_sym_DQUOTE] = ACTIONS(2197),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2199),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2199),
    [anon_sym_switch] = ACTIONS(2197),
    [anon_sym_try] = ACTIONS(2197),
    [anon_sym_int] = ACTIONS(2197),
    [anon_sym_boolean] = ACTIONS(2197),
    [anon_sym_char] = ACTIONS(2197),
    [anon_sym_short] = ACTIONS(2197),
    [anon_sym_long] = ACTIONS(2197),
    [anon_sym_float] = ACTIONS(2197),
    [anon_sym_double] = ACTIONS(2197),
    [anon_sym_void] = ACTIONS(2197),
    [anon_sym_public] = ACTIONS(2197),
    [anon_sym_protected] = ACTIONS(2197),
    [anon_sym_private] = ACTIONS(2197),
    [anon_sym_static] = ACTIONS(2197),
    [anon_sym_final] = ACTIONS(2197),
    [anon_sym_synchronized] = ACTIONS(2197),
    [anon_sym_TILDE] = ACTIONS(2199),
    [anon_sym_BANG] = ACTIONS(2199),
    [anon_sym_new] = ACTIONS(2197),
  },
  [STATE(1015)] = {
    [sym_comment] = STATE(1015),
    [sym_groovy_doc] = STATE(1015),
    [sym_identifier] = ACTIONS(2201),
    [anon_sym_break] = ACTIONS(2201),
    [anon_sym_continue] = ACTIONS(2201),
    [anon_sym_STAR] = ACTIONS(2201),
    [anon_sym_STAR_COLON] = ACTIONS(2203),
    [anon_sym_import] = ACTIONS(2201),
    [anon_sym_package] = ACTIONS(2201),
    [anon_sym_AT] = ACTIONS(2201),
    [anon_sym_assert] = ACTIONS(2201),
    [anon_sym_PLUS_PLUS] = ACTIONS(2203),
    [anon_sym_DASH_DASH] = ACTIONS(2203),
    [anon_sym_SLASH] = ACTIONS(2201),
    [anon_sym_PLUS] = ACTIONS(2201),
    [anon_sym_DASH] = ACTIONS(2201),
    [anon_sym_true] = ACTIONS(2201),
    [anon_sym_false] = ACTIONS(2201),
    [anon_sym_ATinterface] = ACTIONS(2203),
    [anon_sym_interface] = ACTIONS(2201),
    [anon_sym_class] = ACTIONS(2201),
    [anon_sym_LBRACE] = ACTIONS(2203),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2201),
    [anon_sym_def] = ACTIONS(2201),
    [anon_sym_LPAREN] = ACTIONS(2203),
    [anon_sym_null] = ACTIONS(2201),
    [anon_sym_this] = ACTIONS(2201),
    [anon_sym_do] = ACTIONS(2201),
    [anon_sym_while] = ACTIONS(2201),
    [anon_sym_for] = ACTIONS(2201),
    [aux_sym__type_identifier_token1] = ACTIONS(2201),
    [anon_sym_if] = ACTIONS(2201),
    [anon_sym_LBRACK] = ACTIONS(2203),
    [aux_sym_number_literal_token1] = ACTIONS(2201),
    [aux_sym_number_literal_token2] = ACTIONS(2203),
    [aux_sym_number_literal_token3] = ACTIONS(2203),
    [aux_sym_number_literal_token4] = ACTIONS(2201),
    [aux_sym_number_literal_token5] = ACTIONS(2203),
    [anon_sym_return] = ACTIONS(2201),
    [anon_sym_SQUOTE] = ACTIONS(2201),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2203),
    [anon_sym_DQUOTE] = ACTIONS(2201),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2203),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2203),
    [anon_sym_switch] = ACTIONS(2201),
    [anon_sym_try] = ACTIONS(2201),
    [anon_sym_int] = ACTIONS(2201),
    [anon_sym_boolean] = ACTIONS(2201),
    [anon_sym_char] = ACTIONS(2201),
    [anon_sym_short] = ACTIONS(2201),
    [anon_sym_long] = ACTIONS(2201),
    [anon_sym_float] = ACTIONS(2201),
    [anon_sym_double] = ACTIONS(2201),
    [anon_sym_void] = ACTIONS(2201),
    [anon_sym_public] = ACTIONS(2201),
    [anon_sym_protected] = ACTIONS(2201),
    [anon_sym_private] = ACTIONS(2201),
    [anon_sym_static] = ACTIONS(2201),
    [anon_sym_final] = ACTIONS(2201),
    [anon_sym_synchronized] = ACTIONS(2201),
    [anon_sym_TILDE] = ACTIONS(2203),
    [anon_sym_BANG] = ACTIONS(2203),
    [anon_sym_new] = ACTIONS(2201),
  },
  [STATE(1016)] = {
    [sym_comment] = STATE(1016),
    [sym_groovy_doc] = STATE(1016),
    [sym_identifier] = ACTIONS(2205),
    [anon_sym_break] = ACTIONS(2205),
    [anon_sym_continue] = ACTIONS(2205),
    [anon_sym_STAR] = ACTIONS(2205),
    [anon_sym_STAR_COLON] = ACTIONS(2207),
    [anon_sym_import] = ACTIONS(2205),
    [anon_sym_package] = ACTIONS(2205),
    [anon_sym_AT] = ACTIONS(2205),
    [anon_sym_assert] = ACTIONS(2205),
    [anon_sym_PLUS_PLUS] = ACTIONS(2207),
    [anon_sym_DASH_DASH] = ACTIONS(2207),
    [anon_sym_SLASH] = ACTIONS(2205),
    [anon_sym_PLUS] = ACTIONS(2205),
    [anon_sym_DASH] = ACTIONS(2205),
    [anon_sym_true] = ACTIONS(2205),
    [anon_sym_false] = ACTIONS(2205),
    [anon_sym_ATinterface] = ACTIONS(2207),
    [anon_sym_interface] = ACTIONS(2205),
    [anon_sym_class] = ACTIONS(2205),
    [anon_sym_LBRACE] = ACTIONS(2207),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2205),
    [anon_sym_def] = ACTIONS(2205),
    [anon_sym_LPAREN] = ACTIONS(2207),
    [anon_sym_null] = ACTIONS(2205),
    [anon_sym_this] = ACTIONS(2205),
    [anon_sym_do] = ACTIONS(2205),
    [anon_sym_while] = ACTIONS(2205),
    [anon_sym_for] = ACTIONS(2205),
    [aux_sym__type_identifier_token1] = ACTIONS(2205),
    [anon_sym_if] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2207),
    [aux_sym_number_literal_token1] = ACTIONS(2205),
    [aux_sym_number_literal_token2] = ACTIONS(2207),
    [aux_sym_number_literal_token3] = ACTIONS(2207),
    [aux_sym_number_literal_token4] = ACTIONS(2205),
    [aux_sym_number_literal_token5] = ACTIONS(2207),
    [anon_sym_return] = ACTIONS(2205),
    [anon_sym_SQUOTE] = ACTIONS(2205),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2207),
    [anon_sym_DQUOTE] = ACTIONS(2205),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2207),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2207),
    [anon_sym_switch] = ACTIONS(2205),
    [anon_sym_try] = ACTIONS(2205),
    [anon_sym_int] = ACTIONS(2205),
    [anon_sym_boolean] = ACTIONS(2205),
    [anon_sym_char] = ACTIONS(2205),
    [anon_sym_short] = ACTIONS(2205),
    [anon_sym_long] = ACTIONS(2205),
    [anon_sym_float] = ACTIONS(2205),
    [anon_sym_double] = ACTIONS(2205),
    [anon_sym_void] = ACTIONS(2205),
    [anon_sym_public] = ACTIONS(2205),
    [anon_sym_protected] = ACTIONS(2205),
    [anon_sym_private] = ACTIONS(2205),
    [anon_sym_static] = ACTIONS(2205),
    [anon_sym_final] = ACTIONS(2205),
    [anon_sym_synchronized] = ACTIONS(2205),
    [anon_sym_TILDE] = ACTIONS(2207),
    [anon_sym_BANG] = ACTIONS(2207),
    [anon_sym_new] = ACTIONS(2205),
  },
  [STATE(1017)] = {
    [sym_comment] = STATE(1017),
    [sym_groovy_doc] = STATE(1017),
    [sym_identifier] = ACTIONS(2209),
    [anon_sym_break] = ACTIONS(2209),
    [anon_sym_continue] = ACTIONS(2209),
    [anon_sym_STAR] = ACTIONS(2209),
    [anon_sym_STAR_COLON] = ACTIONS(2211),
    [anon_sym_import] = ACTIONS(2209),
    [anon_sym_package] = ACTIONS(2209),
    [anon_sym_AT] = ACTIONS(2209),
    [anon_sym_assert] = ACTIONS(2209),
    [anon_sym_PLUS_PLUS] = ACTIONS(2211),
    [anon_sym_DASH_DASH] = ACTIONS(2211),
    [anon_sym_SLASH] = ACTIONS(2209),
    [anon_sym_PLUS] = ACTIONS(2209),
    [anon_sym_DASH] = ACTIONS(2209),
    [anon_sym_true] = ACTIONS(2209),
    [anon_sym_false] = ACTIONS(2209),
    [anon_sym_ATinterface] = ACTIONS(2211),
    [anon_sym_interface] = ACTIONS(2209),
    [anon_sym_class] = ACTIONS(2209),
    [anon_sym_LBRACE] = ACTIONS(2211),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2209),
    [anon_sym_def] = ACTIONS(2209),
    [anon_sym_LPAREN] = ACTIONS(2211),
    [anon_sym_null] = ACTIONS(2209),
    [anon_sym_this] = ACTIONS(2209),
    [anon_sym_do] = ACTIONS(2209),
    [anon_sym_while] = ACTIONS(2209),
    [anon_sym_for] = ACTIONS(2209),
    [aux_sym__type_identifier_token1] = ACTIONS(2209),
    [anon_sym_if] = ACTIONS(2209),
    [anon_sym_LBRACK] = ACTIONS(2211),
    [aux_sym_number_literal_token1] = ACTIONS(2209),
    [aux_sym_number_literal_token2] = ACTIONS(2211),
    [aux_sym_number_literal_token3] = ACTIONS(2211),
    [aux_sym_number_literal_token4] = ACTIONS(2209),
    [aux_sym_number_literal_token5] = ACTIONS(2211),
    [anon_sym_return] = ACTIONS(2209),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2211),
    [anon_sym_DQUOTE] = ACTIONS(2209),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2211),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2211),
    [anon_sym_switch] = ACTIONS(2209),
    [anon_sym_try] = ACTIONS(2209),
    [anon_sym_int] = ACTIONS(2209),
    [anon_sym_boolean] = ACTIONS(2209),
    [anon_sym_char] = ACTIONS(2209),
    [anon_sym_short] = ACTIONS(2209),
    [anon_sym_long] = ACTIONS(2209),
    [anon_sym_float] = ACTIONS(2209),
    [anon_sym_double] = ACTIONS(2209),
    [anon_sym_void] = ACTIONS(2209),
    [anon_sym_public] = ACTIONS(2209),
    [anon_sym_protected] = ACTIONS(2209),
    [anon_sym_private] = ACTIONS(2209),
    [anon_sym_static] = ACTIONS(2209),
    [anon_sym_final] = ACTIONS(2209),
    [anon_sym_synchronized] = ACTIONS(2209),
    [anon_sym_TILDE] = ACTIONS(2211),
    [anon_sym_BANG] = ACTIONS(2211),
    [anon_sym_new] = ACTIONS(2209),
  },
  [STATE(1018)] = {
    [sym_comment] = STATE(1018),
    [sym_groovy_doc] = STATE(1018),
    [sym_identifier] = ACTIONS(2213),
    [anon_sym_break] = ACTIONS(2213),
    [anon_sym_continue] = ACTIONS(2213),
    [anon_sym_STAR] = ACTIONS(2213),
    [anon_sym_STAR_COLON] = ACTIONS(2215),
    [anon_sym_import] = ACTIONS(2213),
    [anon_sym_package] = ACTIONS(2213),
    [anon_sym_AT] = ACTIONS(2213),
    [anon_sym_assert] = ACTIONS(2213),
    [anon_sym_PLUS_PLUS] = ACTIONS(2215),
    [anon_sym_DASH_DASH] = ACTIONS(2215),
    [anon_sym_SLASH] = ACTIONS(2213),
    [anon_sym_PLUS] = ACTIONS(2213),
    [anon_sym_DASH] = ACTIONS(2213),
    [anon_sym_true] = ACTIONS(2213),
    [anon_sym_false] = ACTIONS(2213),
    [anon_sym_ATinterface] = ACTIONS(2215),
    [anon_sym_interface] = ACTIONS(2213),
    [anon_sym_class] = ACTIONS(2213),
    [anon_sym_LBRACE] = ACTIONS(2215),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2213),
    [anon_sym_def] = ACTIONS(2213),
    [anon_sym_LPAREN] = ACTIONS(2215),
    [anon_sym_null] = ACTIONS(2213),
    [anon_sym_this] = ACTIONS(2213),
    [anon_sym_do] = ACTIONS(2213),
    [anon_sym_while] = ACTIONS(2213),
    [anon_sym_for] = ACTIONS(2213),
    [aux_sym__type_identifier_token1] = ACTIONS(2213),
    [anon_sym_if] = ACTIONS(2213),
    [anon_sym_LBRACK] = ACTIONS(2215),
    [aux_sym_number_literal_token1] = ACTIONS(2213),
    [aux_sym_number_literal_token2] = ACTIONS(2215),
    [aux_sym_number_literal_token3] = ACTIONS(2215),
    [aux_sym_number_literal_token4] = ACTIONS(2213),
    [aux_sym_number_literal_token5] = ACTIONS(2215),
    [anon_sym_return] = ACTIONS(2213),
    [anon_sym_SQUOTE] = ACTIONS(2213),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2215),
    [anon_sym_DQUOTE] = ACTIONS(2213),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2215),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2215),
    [anon_sym_switch] = ACTIONS(2213),
    [anon_sym_try] = ACTIONS(2213),
    [anon_sym_int] = ACTIONS(2213),
    [anon_sym_boolean] = ACTIONS(2213),
    [anon_sym_char] = ACTIONS(2213),
    [anon_sym_short] = ACTIONS(2213),
    [anon_sym_long] = ACTIONS(2213),
    [anon_sym_float] = ACTIONS(2213),
    [anon_sym_double] = ACTIONS(2213),
    [anon_sym_void] = ACTIONS(2213),
    [anon_sym_public] = ACTIONS(2213),
    [anon_sym_protected] = ACTIONS(2213),
    [anon_sym_private] = ACTIONS(2213),
    [anon_sym_static] = ACTIONS(2213),
    [anon_sym_final] = ACTIONS(2213),
    [anon_sym_synchronized] = ACTIONS(2213),
    [anon_sym_TILDE] = ACTIONS(2215),
    [anon_sym_BANG] = ACTIONS(2215),
    [anon_sym_new] = ACTIONS(2213),
  },
  [STATE(1019)] = {
    [sym_comment] = STATE(1019),
    [sym_groovy_doc] = STATE(1019),
    [sym_identifier] = ACTIONS(2217),
    [anon_sym_break] = ACTIONS(2217),
    [anon_sym_continue] = ACTIONS(2217),
    [anon_sym_STAR] = ACTIONS(2217),
    [anon_sym_STAR_COLON] = ACTIONS(2219),
    [anon_sym_import] = ACTIONS(2217),
    [anon_sym_package] = ACTIONS(2217),
    [anon_sym_AT] = ACTIONS(2217),
    [anon_sym_assert] = ACTIONS(2217),
    [anon_sym_PLUS_PLUS] = ACTIONS(2219),
    [anon_sym_DASH_DASH] = ACTIONS(2219),
    [anon_sym_SLASH] = ACTIONS(2217),
    [anon_sym_PLUS] = ACTIONS(2217),
    [anon_sym_DASH] = ACTIONS(2217),
    [anon_sym_true] = ACTIONS(2217),
    [anon_sym_false] = ACTIONS(2217),
    [anon_sym_ATinterface] = ACTIONS(2219),
    [anon_sym_interface] = ACTIONS(2217),
    [anon_sym_class] = ACTIONS(2217),
    [anon_sym_LBRACE] = ACTIONS(2219),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2217),
    [anon_sym_def] = ACTIONS(2217),
    [anon_sym_LPAREN] = ACTIONS(2219),
    [anon_sym_null] = ACTIONS(2217),
    [anon_sym_this] = ACTIONS(2217),
    [anon_sym_do] = ACTIONS(2217),
    [anon_sym_while] = ACTIONS(2217),
    [anon_sym_for] = ACTIONS(2217),
    [aux_sym__type_identifier_token1] = ACTIONS(2217),
    [anon_sym_if] = ACTIONS(2217),
    [anon_sym_LBRACK] = ACTIONS(2219),
    [aux_sym_number_literal_token1] = ACTIONS(2217),
    [aux_sym_number_literal_token2] = ACTIONS(2219),
    [aux_sym_number_literal_token3] = ACTIONS(2219),
    [aux_sym_number_literal_token4] = ACTIONS(2217),
    [aux_sym_number_literal_token5] = ACTIONS(2219),
    [anon_sym_return] = ACTIONS(2217),
    [anon_sym_SQUOTE] = ACTIONS(2217),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2219),
    [anon_sym_DQUOTE] = ACTIONS(2217),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2219),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2219),
    [anon_sym_switch] = ACTIONS(2217),
    [anon_sym_try] = ACTIONS(2217),
    [anon_sym_int] = ACTIONS(2217),
    [anon_sym_boolean] = ACTIONS(2217),
    [anon_sym_char] = ACTIONS(2217),
    [anon_sym_short] = ACTIONS(2217),
    [anon_sym_long] = ACTIONS(2217),
    [anon_sym_float] = ACTIONS(2217),
    [anon_sym_double] = ACTIONS(2217),
    [anon_sym_void] = ACTIONS(2217),
    [anon_sym_public] = ACTIONS(2217),
    [anon_sym_protected] = ACTIONS(2217),
    [anon_sym_private] = ACTIONS(2217),
    [anon_sym_static] = ACTIONS(2217),
    [anon_sym_final] = ACTIONS(2217),
    [anon_sym_synchronized] = ACTIONS(2217),
    [anon_sym_TILDE] = ACTIONS(2219),
    [anon_sym_BANG] = ACTIONS(2219),
    [anon_sym_new] = ACTIONS(2217),
  },
  [STATE(1020)] = {
    [sym_comment] = STATE(1020),
    [sym_groovy_doc] = STATE(1020),
    [sym_identifier] = ACTIONS(2221),
    [anon_sym_break] = ACTIONS(2221),
    [anon_sym_continue] = ACTIONS(2221),
    [anon_sym_STAR] = ACTIONS(2221),
    [anon_sym_STAR_COLON] = ACTIONS(2223),
    [anon_sym_import] = ACTIONS(2221),
    [anon_sym_package] = ACTIONS(2221),
    [anon_sym_AT] = ACTIONS(2221),
    [anon_sym_assert] = ACTIONS(2221),
    [anon_sym_PLUS_PLUS] = ACTIONS(2223),
    [anon_sym_DASH_DASH] = ACTIONS(2223),
    [anon_sym_SLASH] = ACTIONS(2221),
    [anon_sym_PLUS] = ACTIONS(2221),
    [anon_sym_DASH] = ACTIONS(2221),
    [anon_sym_true] = ACTIONS(2221),
    [anon_sym_false] = ACTIONS(2221),
    [anon_sym_ATinterface] = ACTIONS(2223),
    [anon_sym_interface] = ACTIONS(2221),
    [anon_sym_class] = ACTIONS(2221),
    [anon_sym_LBRACE] = ACTIONS(2223),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2221),
    [anon_sym_def] = ACTIONS(2221),
    [anon_sym_LPAREN] = ACTIONS(2223),
    [anon_sym_null] = ACTIONS(2221),
    [anon_sym_this] = ACTIONS(2221),
    [anon_sym_do] = ACTIONS(2221),
    [anon_sym_while] = ACTIONS(2221),
    [anon_sym_for] = ACTIONS(2221),
    [aux_sym__type_identifier_token1] = ACTIONS(2221),
    [anon_sym_if] = ACTIONS(2221),
    [anon_sym_LBRACK] = ACTIONS(2223),
    [aux_sym_number_literal_token1] = ACTIONS(2221),
    [aux_sym_number_literal_token2] = ACTIONS(2223),
    [aux_sym_number_literal_token3] = ACTIONS(2223),
    [aux_sym_number_literal_token4] = ACTIONS(2221),
    [aux_sym_number_literal_token5] = ACTIONS(2223),
    [anon_sym_return] = ACTIONS(2221),
    [anon_sym_SQUOTE] = ACTIONS(2221),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2223),
    [anon_sym_DQUOTE] = ACTIONS(2221),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2223),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2223),
    [anon_sym_switch] = ACTIONS(2221),
    [anon_sym_try] = ACTIONS(2221),
    [anon_sym_int] = ACTIONS(2221),
    [anon_sym_boolean] = ACTIONS(2221),
    [anon_sym_char] = ACTIONS(2221),
    [anon_sym_short] = ACTIONS(2221),
    [anon_sym_long] = ACTIONS(2221),
    [anon_sym_float] = ACTIONS(2221),
    [anon_sym_double] = ACTIONS(2221),
    [anon_sym_void] = ACTIONS(2221),
    [anon_sym_public] = ACTIONS(2221),
    [anon_sym_protected] = ACTIONS(2221),
    [anon_sym_private] = ACTIONS(2221),
    [anon_sym_static] = ACTIONS(2221),
    [anon_sym_final] = ACTIONS(2221),
    [anon_sym_synchronized] = ACTIONS(2221),
    [anon_sym_TILDE] = ACTIONS(2223),
    [anon_sym_BANG] = ACTIONS(2223),
    [anon_sym_new] = ACTIONS(2221),
  },
  [STATE(1021)] = {
    [sym_comment] = STATE(1021),
    [sym_groovy_doc] = STATE(1021),
    [sym_identifier] = ACTIONS(2225),
    [anon_sym_break] = ACTIONS(2225),
    [anon_sym_continue] = ACTIONS(2225),
    [anon_sym_STAR] = ACTIONS(2225),
    [anon_sym_STAR_COLON] = ACTIONS(2227),
    [anon_sym_import] = ACTIONS(2225),
    [anon_sym_package] = ACTIONS(2225),
    [anon_sym_AT] = ACTIONS(2225),
    [anon_sym_assert] = ACTIONS(2225),
    [anon_sym_PLUS_PLUS] = ACTIONS(2227),
    [anon_sym_DASH_DASH] = ACTIONS(2227),
    [anon_sym_SLASH] = ACTIONS(2225),
    [anon_sym_PLUS] = ACTIONS(2225),
    [anon_sym_DASH] = ACTIONS(2225),
    [anon_sym_true] = ACTIONS(2225),
    [anon_sym_false] = ACTIONS(2225),
    [anon_sym_ATinterface] = ACTIONS(2227),
    [anon_sym_interface] = ACTIONS(2225),
    [anon_sym_class] = ACTIONS(2225),
    [anon_sym_LBRACE] = ACTIONS(2227),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2225),
    [anon_sym_def] = ACTIONS(2225),
    [anon_sym_LPAREN] = ACTIONS(2227),
    [anon_sym_null] = ACTIONS(2225),
    [anon_sym_this] = ACTIONS(2225),
    [anon_sym_do] = ACTIONS(2225),
    [anon_sym_while] = ACTIONS(2225),
    [anon_sym_for] = ACTIONS(2225),
    [aux_sym__type_identifier_token1] = ACTIONS(2225),
    [anon_sym_if] = ACTIONS(2225),
    [anon_sym_LBRACK] = ACTIONS(2227),
    [aux_sym_number_literal_token1] = ACTIONS(2225),
    [aux_sym_number_literal_token2] = ACTIONS(2227),
    [aux_sym_number_literal_token3] = ACTIONS(2227),
    [aux_sym_number_literal_token4] = ACTIONS(2225),
    [aux_sym_number_literal_token5] = ACTIONS(2227),
    [anon_sym_return] = ACTIONS(2225),
    [anon_sym_SQUOTE] = ACTIONS(2225),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2227),
    [anon_sym_DQUOTE] = ACTIONS(2225),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2227),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2227),
    [anon_sym_switch] = ACTIONS(2225),
    [anon_sym_try] = ACTIONS(2225),
    [anon_sym_int] = ACTIONS(2225),
    [anon_sym_boolean] = ACTIONS(2225),
    [anon_sym_char] = ACTIONS(2225),
    [anon_sym_short] = ACTIONS(2225),
    [anon_sym_long] = ACTIONS(2225),
    [anon_sym_float] = ACTIONS(2225),
    [anon_sym_double] = ACTIONS(2225),
    [anon_sym_void] = ACTIONS(2225),
    [anon_sym_public] = ACTIONS(2225),
    [anon_sym_protected] = ACTIONS(2225),
    [anon_sym_private] = ACTIONS(2225),
    [anon_sym_static] = ACTIONS(2225),
    [anon_sym_final] = ACTIONS(2225),
    [anon_sym_synchronized] = ACTIONS(2225),
    [anon_sym_TILDE] = ACTIONS(2227),
    [anon_sym_BANG] = ACTIONS(2227),
    [anon_sym_new] = ACTIONS(2225),
  },
  [STATE(1022)] = {
    [sym_comment] = STATE(1022),
    [sym_groovy_doc] = STATE(1022),
    [sym_identifier] = ACTIONS(2229),
    [anon_sym_break] = ACTIONS(2229),
    [anon_sym_continue] = ACTIONS(2229),
    [anon_sym_STAR] = ACTIONS(2229),
    [anon_sym_STAR_COLON] = ACTIONS(2231),
    [anon_sym_import] = ACTIONS(2229),
    [anon_sym_package] = ACTIONS(2229),
    [anon_sym_AT] = ACTIONS(2229),
    [anon_sym_assert] = ACTIONS(2229),
    [anon_sym_PLUS_PLUS] = ACTIONS(2231),
    [anon_sym_DASH_DASH] = ACTIONS(2231),
    [anon_sym_SLASH] = ACTIONS(2229),
    [anon_sym_PLUS] = ACTIONS(2229),
    [anon_sym_DASH] = ACTIONS(2229),
    [anon_sym_true] = ACTIONS(2229),
    [anon_sym_false] = ACTIONS(2229),
    [anon_sym_ATinterface] = ACTIONS(2231),
    [anon_sym_interface] = ACTIONS(2229),
    [anon_sym_class] = ACTIONS(2229),
    [anon_sym_LBRACE] = ACTIONS(2231),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2229),
    [anon_sym_def] = ACTIONS(2229),
    [anon_sym_LPAREN] = ACTIONS(2231),
    [anon_sym_null] = ACTIONS(2229),
    [anon_sym_this] = ACTIONS(2229),
    [anon_sym_do] = ACTIONS(2229),
    [anon_sym_while] = ACTIONS(2229),
    [anon_sym_for] = ACTIONS(2229),
    [aux_sym__type_identifier_token1] = ACTIONS(2229),
    [anon_sym_if] = ACTIONS(2229),
    [anon_sym_LBRACK] = ACTIONS(2231),
    [aux_sym_number_literal_token1] = ACTIONS(2229),
    [aux_sym_number_literal_token2] = ACTIONS(2231),
    [aux_sym_number_literal_token3] = ACTIONS(2231),
    [aux_sym_number_literal_token4] = ACTIONS(2229),
    [aux_sym_number_literal_token5] = ACTIONS(2231),
    [anon_sym_return] = ACTIONS(2229),
    [anon_sym_SQUOTE] = ACTIONS(2229),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2231),
    [anon_sym_DQUOTE] = ACTIONS(2229),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2231),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2231),
    [anon_sym_switch] = ACTIONS(2229),
    [anon_sym_try] = ACTIONS(2229),
    [anon_sym_int] = ACTIONS(2229),
    [anon_sym_boolean] = ACTIONS(2229),
    [anon_sym_char] = ACTIONS(2229),
    [anon_sym_short] = ACTIONS(2229),
    [anon_sym_long] = ACTIONS(2229),
    [anon_sym_float] = ACTIONS(2229),
    [anon_sym_double] = ACTIONS(2229),
    [anon_sym_void] = ACTIONS(2229),
    [anon_sym_public] = ACTIONS(2229),
    [anon_sym_protected] = ACTIONS(2229),
    [anon_sym_private] = ACTIONS(2229),
    [anon_sym_static] = ACTIONS(2229),
    [anon_sym_final] = ACTIONS(2229),
    [anon_sym_synchronized] = ACTIONS(2229),
    [anon_sym_TILDE] = ACTIONS(2231),
    [anon_sym_BANG] = ACTIONS(2231),
    [anon_sym_new] = ACTIONS(2229),
  },
  [STATE(1023)] = {
    [sym_comment] = STATE(1023),
    [sym_groovy_doc] = STATE(1023),
    [sym_identifier] = ACTIONS(2233),
    [anon_sym_break] = ACTIONS(2233),
    [anon_sym_continue] = ACTIONS(2233),
    [anon_sym_STAR] = ACTIONS(2233),
    [anon_sym_STAR_COLON] = ACTIONS(2235),
    [anon_sym_import] = ACTIONS(2233),
    [anon_sym_package] = ACTIONS(2233),
    [anon_sym_AT] = ACTIONS(2233),
    [anon_sym_assert] = ACTIONS(2233),
    [anon_sym_PLUS_PLUS] = ACTIONS(2235),
    [anon_sym_DASH_DASH] = ACTIONS(2235),
    [anon_sym_SLASH] = ACTIONS(2233),
    [anon_sym_PLUS] = ACTIONS(2233),
    [anon_sym_DASH] = ACTIONS(2233),
    [anon_sym_true] = ACTIONS(2233),
    [anon_sym_false] = ACTIONS(2233),
    [anon_sym_ATinterface] = ACTIONS(2235),
    [anon_sym_interface] = ACTIONS(2233),
    [anon_sym_class] = ACTIONS(2233),
    [anon_sym_LBRACE] = ACTIONS(2235),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2233),
    [anon_sym_def] = ACTIONS(2233),
    [anon_sym_LPAREN] = ACTIONS(2235),
    [anon_sym_null] = ACTIONS(2233),
    [anon_sym_this] = ACTIONS(2233),
    [anon_sym_do] = ACTIONS(2233),
    [anon_sym_while] = ACTIONS(2233),
    [anon_sym_for] = ACTIONS(2233),
    [aux_sym__type_identifier_token1] = ACTIONS(2233),
    [anon_sym_if] = ACTIONS(2233),
    [anon_sym_LBRACK] = ACTIONS(2235),
    [aux_sym_number_literal_token1] = ACTIONS(2233),
    [aux_sym_number_literal_token2] = ACTIONS(2235),
    [aux_sym_number_literal_token3] = ACTIONS(2235),
    [aux_sym_number_literal_token4] = ACTIONS(2233),
    [aux_sym_number_literal_token5] = ACTIONS(2235),
    [anon_sym_return] = ACTIONS(2233),
    [anon_sym_SQUOTE] = ACTIONS(2233),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2235),
    [anon_sym_DQUOTE] = ACTIONS(2233),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2235),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2235),
    [anon_sym_switch] = ACTIONS(2233),
    [anon_sym_try] = ACTIONS(2233),
    [anon_sym_int] = ACTIONS(2233),
    [anon_sym_boolean] = ACTIONS(2233),
    [anon_sym_char] = ACTIONS(2233),
    [anon_sym_short] = ACTIONS(2233),
    [anon_sym_long] = ACTIONS(2233),
    [anon_sym_float] = ACTIONS(2233),
    [anon_sym_double] = ACTIONS(2233),
    [anon_sym_void] = ACTIONS(2233),
    [anon_sym_public] = ACTIONS(2233),
    [anon_sym_protected] = ACTIONS(2233),
    [anon_sym_private] = ACTIONS(2233),
    [anon_sym_static] = ACTIONS(2233),
    [anon_sym_final] = ACTIONS(2233),
    [anon_sym_synchronized] = ACTIONS(2233),
    [anon_sym_TILDE] = ACTIONS(2235),
    [anon_sym_BANG] = ACTIONS(2235),
    [anon_sym_new] = ACTIONS(2233),
  },
  [STATE(1024)] = {
    [sym_comment] = STATE(1024),
    [sym_groovy_doc] = STATE(1024),
    [sym_identifier] = ACTIONS(2237),
    [anon_sym_break] = ACTIONS(2237),
    [anon_sym_continue] = ACTIONS(2237),
    [anon_sym_STAR] = ACTIONS(2237),
    [anon_sym_STAR_COLON] = ACTIONS(2239),
    [anon_sym_import] = ACTIONS(2237),
    [anon_sym_package] = ACTIONS(2237),
    [anon_sym_AT] = ACTIONS(2237),
    [anon_sym_assert] = ACTIONS(2237),
    [anon_sym_PLUS_PLUS] = ACTIONS(2239),
    [anon_sym_DASH_DASH] = ACTIONS(2239),
    [anon_sym_SLASH] = ACTIONS(2237),
    [anon_sym_PLUS] = ACTIONS(2237),
    [anon_sym_DASH] = ACTIONS(2237),
    [anon_sym_true] = ACTIONS(2237),
    [anon_sym_false] = ACTIONS(2237),
    [anon_sym_ATinterface] = ACTIONS(2239),
    [anon_sym_interface] = ACTIONS(2237),
    [anon_sym_class] = ACTIONS(2237),
    [anon_sym_LBRACE] = ACTIONS(2239),
    [aux_sym_comment_token1] = ACTIONS(3),
    [anon_sym_SLASH_STAR] = ACTIONS(5),
    [anon_sym_SLASH_STAR_STAR] = ACTIONS(7),
    [anon_sym__] = ACTIONS(2237),
    [anon_sym_def] = ACTIONS(2237),
    [anon_sym_LPAREN] = ACTIONS(2239),
    [anon_sym_null] = ACTIONS(2237),
    [anon_sym_this] = ACTIONS(2237),
    [anon_sym_do] = ACTIONS(2237),
    [anon_sym_while] = ACTIONS(2237),
    [anon_sym_for] = ACTIONS(2237),
    [aux_sym__type_identifier_token1] = ACTIONS(2237),
    [anon_sym_if] = ACTIONS(2237),
    [anon_sym_LBRACK] = ACTIONS(2239),
    [aux_sym_number_literal_token1] = ACTIONS(2237),
    [aux_sym_number_literal_token2] = ACTIONS(2239),
    [aux_sym_number_literal_token3] = ACTIONS(2239),
    [aux_sym_number_literal_token4] = ACTIONS(2237),
    [aux_sym_number_literal_token5] = ACTIONS(2239),
    [anon_sym_return] = ACTIONS(2237),
    [anon_sym_SQUOTE] = ACTIONS(2237),
    [anon_sym_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(2239),
    [anon_sym_DQUOTE] = ACTIONS(2237),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(2239),
    [anon_sym_DOLLAR_SLASH] = ACTIONS(2239),
    [anon_sym_switch] = ACTIONS(2237),
    [anon_sym_try] = ACTIONS(2237),
    [anon_sym_int] = ACTIONS(2237),
    [anon_sym_boolean] = ACTIONS(2237),
    [anon_sym_char] = ACTIONS(2237),
    [anon_sym_short] = ACTIONS(2237),
    [anon_sym_long] = ACTIONS(2237),
    [anon_sym_float] = ACTIONS(2237),
    [anon_sym_double] = ACTIONS(2237),
    [anon_sym_void] = ACTIONS(2237),
    [anon_sym_public] = ACTIONS(2237),
    [anon_sym_protected] = ACTIONS(2237),
    [anon_sym_private] = ACTIONS(2237),
    [anon_sym_static] = ACTIONS(2237),
    [anon_sym_final] = ACTIONS(2237),
    [anon_sym_synchronized] = ACTIONS(2237),
    [anon_sym_TILDE] = ACTIONS(2239),
    [anon_sym_BANG] = ACTIONS(2239),
    [anon_sym_new] = ACTIONS(2237),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1404), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1025), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    ACTIONS(1696), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    ACTIONS(1700), 4,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [138] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2241), 1,
      anon_sym_LBRACE,
    STATE(1040), 1,
      sym_closure,
    STATE(1026), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1338), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1336), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [218] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2241), 1,
      anon_sym_LBRACE,
    STATE(1042), 1,
      sym_closure,
    STATE(1027), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1334), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1332), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [298] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2241), 1,
      anon_sym_LBRACE,
    STATE(1054), 1,
      sym_closure,
    STATE(1028), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1306), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1304), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [378] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2241), 1,
      anon_sym_LBRACE,
    STATE(1050), 1,
      sym_closure,
    STATE(1029), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1324), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1322), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [458] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2243), 1,
      anon_sym_LPAREN,
    STATE(1053), 1,
      sym_argument_list,
    STATE(1030), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [537] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1033), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1031), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(375), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(373), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [614] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2245), 1,
      anon_sym_DOT,
    STATE(1031), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1032), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [693] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2247), 1,
      anon_sym_DOT,
    STATE(1033), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(379), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(377), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [770] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2252), 1,
      anon_sym_COLON,
    ACTIONS(2254), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1131), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1464), 1,
      sym__expression,
    STATE(1701), 1,
      aux_sym_map_repeat1,
    STATE(2277), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1034), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [916] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2256), 1,
      anon_sym_COLON,
    ACTIONS(2258), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1111), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1475), 1,
      sym__expression,
    STATE(1697), 1,
      aux_sym_map_repeat1,
    STATE(2267), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1035), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [1062] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1036), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1412), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1410), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1136] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1037), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1210] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(1103), 1,
      sym_closure,
    STATE(1038), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1338), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1336), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1288] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(1095), 1,
      sym_closure,
    STATE(1039), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1306), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1304), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1366] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1040), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1306), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1304), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1440] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2260), 1,
      anon_sym_COLON,
    ACTIONS(2262), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1119), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1463), 1,
      sym__expression,
    STATE(1693), 1,
      aux_sym_map_repeat1,
    STATE(2295), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1041), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [1586] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1042), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1430), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1428), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1660] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(1099), 1,
      sym_closure,
    STATE(1043), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1324), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1322), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1738] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1044), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(393), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(391), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1812] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2245), 1,
      anon_sym_DOT,
    ACTIONS(2266), 1,
      anon_sym_LBRACK,
    STATE(1031), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(2264), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1045), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 34,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [1894] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2268), 1,
      anon_sym_COLON,
    ACTIONS(2270), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1154), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1474), 1,
      sym__expression,
    STATE(1700), 1,
      aux_sym_map_repeat1,
    STATE(2262), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1046), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [2040] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(1098), 1,
      sym_closure,
    STATE(1047), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1334), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1332), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2118] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1048), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2192] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1049), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(371), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(369), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2266] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1050), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1338), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1336), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2340] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2272), 1,
      anon_sym_COLON,
    ACTIONS(2274), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1114), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1473), 1,
      sym__expression,
    STATE(1694), 1,
      aux_sym_map_repeat1,
    STATE(2291), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1051), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [2486] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1052), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1392), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1390), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2560] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1053), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1366), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1368), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2634] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1054), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1334), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1332), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2708] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1055), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1362), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1364), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2782] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1056), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(387), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(389), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2856] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1057), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(379), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(377), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [2930] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1058), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1376), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1374), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3004] = 42,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2276), 1,
      anon_sym_COLON,
    ACTIONS(2278), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1106), 1,
      aux_sym_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1462), 1,
      sym__expression,
    STATE(1696), 1,
      aux_sym_map_repeat1,
    STATE(2303), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1059), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [3150] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1060), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1400), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1398), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3223] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1061), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1416), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1414), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3296] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1062), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1420), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1418), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3369] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2280), 1,
      anon_sym_LPAREN,
    STATE(1090), 1,
      sym_argument_list,
    STATE(1063), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3446] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1064), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1384), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1382), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3519] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1065), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1432), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1434), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3592] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1066), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1320), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1318), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3665] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1067), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1296), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1298), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3738] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1068), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1302), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1300), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3811] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1069), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1328), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1330), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3884] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1070), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1396), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1394), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [3957] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1071), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1380), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1378), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4030] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1072), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4103] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1073), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1316), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1314), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4176] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1074), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1388), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1386), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4249] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1290), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(2282), 1,
      anon_sym_DOT,
    STATE(660), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1286), 2,
      anon_sym_LT,
      sym_identifier,
    STATE(1075), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 23,
      anon_sym_STAR,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_instanceof,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(97), 27,
      anon_sym_SEMI,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [4330] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1076), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1348), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1350), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4403] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1079), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1077), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(375), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(373), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4478] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2284), 1,
      anon_sym_DOT,
    STATE(1077), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1078), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4555] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2286), 1,
      anon_sym_DOT,
    STATE(1079), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(379), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(377), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4630] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2289), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1130), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1453), 1,
      sym__expression,
    STATE(2297), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1080), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [4770] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1081), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(387), 26,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_instanceof,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_while,
      sym_identifier,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(389), 28,
      anon_sym_SEMI,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK_RBRACK,
  [4842] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1082), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(365), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [4914] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2291), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1141), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1452), 1,
      sym__expression,
    STATE(2335), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1083), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [5054] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1084), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(379), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(377), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5126] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2293), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1137), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1456), 1,
      sym__expression,
    STATE(2332), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1085), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [5266] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2295), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1126), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1469), 1,
      sym__expression,
    STATE(2225), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1086), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [5406] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1087), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1392), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1390), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5478] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2284), 1,
      anon_sym_DOT,
    ACTIONS(2297), 1,
      anon_sym_LBRACK,
    STATE(1077), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1088), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1280), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1278), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5556] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2299), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1123), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1455), 1,
      sym__expression,
    STATE(2326), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1089), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [5696] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1090), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1366), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1368), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5768] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1091), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(387), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(389), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5840] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1092), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1362), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1364), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5912] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1093), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [5984] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2301), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1121), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1449), 1,
      sym__expression,
    STATE(2251), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1094), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [6124] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1095), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1334), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1332), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6196] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1096), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1376), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1374), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6268] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2303), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1139), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1476), 1,
      sym__expression,
    STATE(2305), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1097), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [6408] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1098), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1430), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1428), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6480] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1099), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1338), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1336), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6552] = 40,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    ACTIONS(2305), 1,
      anon_sym_RPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1135), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1467), 1,
      sym__expression,
    STATE(2212), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1100), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [6692] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1101), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1412), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1410), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6764] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1102), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(371), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(369), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6836] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1103), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1306), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1304), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6908] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1104), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(393), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(391), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [6980] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1515), 1,
      sym__expression,
    STATE(2357), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1105), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [7115] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2307), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1466), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1106), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [7248] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1107), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1316), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1314), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [7319] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1108), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [7396] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1109), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1348), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1350), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [7467] = 36,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2311), 1,
      sym_identifier,
    ACTIONS(2314), 1,
      anon_sym_STAR,
    ACTIONS(2317), 1,
      anon_sym_STAR_COLON,
    ACTIONS(2323), 1,
      anon_sym_SLASH,
    ACTIONS(2332), 1,
      anon_sym_LBRACE,
    ACTIONS(2335), 1,
      anon_sym_LPAREN,
    ACTIONS(2338), 1,
      anon_sym_null,
    ACTIONS(2341), 1,
      anon_sym_this,
    ACTIONS(2344), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2347), 1,
      anon_sym_LBRACK,
    ACTIONS(2350), 1,
      anon_sym_RBRACK,
    ACTIONS(2358), 1,
      anon_sym_SQUOTE,
    ACTIONS(2361), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(2364), 1,
      anon_sym_DQUOTE,
    ACTIONS(2367), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(2370), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(2376), 1,
      anon_sym_new,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1496), 1,
      sym__expression,
    ACTIONS(2320), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2326), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2329), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2352), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(2373), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    ACTIONS(2355), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1110), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_list_repeat1,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [7598] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2379), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1470), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1111), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [7731] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1112), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [7802] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1113), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1432), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1434), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [7873] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2381), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1450), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1114), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [8006] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1115), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1296), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1298), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [8077] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1480), 1,
      sym__expression,
    STATE(2347), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1116), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [8212] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1117), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1302), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1300), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [8283] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1118), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1388), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1386), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [8354] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2383), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1465), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1119), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [8487] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1120), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1380), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1378), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [8558] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1451), 1,
      sym__expression,
    STATE(2336), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1121), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [8695] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1520), 1,
      sym__expression,
    STATE(2448), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1122), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [8830] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1457), 1,
      sym__expression,
    STATE(2216), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1123), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [8967] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1124), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1396), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1394), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [9038] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1125), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1384), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1382), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [9109] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1471), 1,
      sym__expression,
    STATE(2310), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1126), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [9246] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1127), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1400), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1398), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [9317] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1512), 1,
      sym__expression,
    STATE(2418), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1128), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [9452] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1497), 1,
      sym__expression,
    STATE(2395), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1129), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [9587] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1454), 1,
      sym__expression,
    STATE(2308), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1130), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [9724] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2385), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1448), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1131), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [9857] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2387), 1,
      sym_identifier,
    ACTIONS(2390), 1,
      anon_sym_STAR,
    ACTIONS(2393), 1,
      anon_sym_STAR_COLON,
    ACTIONS(2399), 1,
      anon_sym_SLASH,
    ACTIONS(2408), 1,
      anon_sym_LBRACE,
    ACTIONS(2411), 1,
      anon_sym_LPAREN,
    ACTIONS(2414), 1,
      anon_sym_null,
    ACTIONS(2417), 1,
      anon_sym_this,
    ACTIONS(2420), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2423), 1,
      anon_sym_LBRACK,
    ACTIONS(2432), 1,
      anon_sym_SQUOTE,
    ACTIONS(2435), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(2438), 1,
      anon_sym_DQUOTE,
    ACTIONS(2441), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(2444), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(2450), 1,
      anon_sym_new,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1525), 1,
      sym__expression,
    STATE(2449), 1,
      sym_map_item,
    ACTIONS(2396), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2402), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2405), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2426), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(2447), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(2429), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1132), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_argument_list_repeat1,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [9992] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1133), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1328), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1330), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [10063] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1521), 1,
      sym__expression,
    STATE(2360), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1134), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [10198] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1468), 1,
      sym__expression,
    STATE(2327), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1135), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [10335] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1539), 1,
      sym__expression,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    STATE(2456), 1,
      sym__immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1136), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [10470] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1460), 1,
      sym__expression,
    STATE(2229), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1137), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [10607] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1138), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [10686] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1459), 1,
      sym__expression,
    STATE(2309), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1139), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [10823] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1499), 1,
      sym__expression,
    STATE(2411), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1140), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [10958] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2250), 1,
      sym_identifier,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1132), 1,
      aux_sym_argument_list_repeat1,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1458), 1,
      sym__expression,
    STATE(2213), 1,
      sym_map_item,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1037), 2,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1141), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1398), 2,
      sym_number_literal,
      sym_string,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1072), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [11095] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1492), 1,
      sym__expression,
    STATE(2443), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1142), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11230] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1503), 1,
      sym__expression,
    STATE(2378), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1143), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11365] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1519), 1,
      sym__expression,
    STATE(2372), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1144), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11500] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1529), 1,
      sym__expression,
    STATE(2345), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1145), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11635] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1534), 1,
      sym__expression,
    STATE(2384), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1146), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11770] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1538), 1,
      sym__expression,
    STATE(2400), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1147), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [11905] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1540), 1,
      sym__expression,
    STATE(2404), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1148), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [12040] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1541), 1,
      sym__expression,
    STATE(2408), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1149), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [12175] = 38,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1446), 1,
      sym_closure,
    STATE(1478), 1,
      sym__expression,
    STATE(2415), 1,
      sym__immediately_invoked_closure,
    STATE(2453), 1,
      sym___immediately_invoked_closure,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1150), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 5,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_ternary_op,
      sym_unary_op,
  [12310] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1151), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1416), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1414), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [12381] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1152), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [12456] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1153), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1420), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1418), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [12527] = 37,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2455), 1,
      anon_sym_RBRACK,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1110), 1,
      aux_sym_list_repeat1,
    STATE(1472), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1154), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [12660] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1155), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1320), 17,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1318), 36,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [12731] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1156), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [12803] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1157), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [12875] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2457), 1,
      sym_identifier,
    ACTIONS(2459), 1,
      anon_sym_null,
    ACTIONS(2461), 1,
      anon_sym_this,
    STATE(325), 1,
      sym_function_call,
    STATE(763), 1,
      sym_map_item,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1422), 1,
      sym__type_identifier,
    STATE(1427), 1,
      sym__callable_expression,
    STATE(1431), 1,
      sym_parenthesized_expression,
    STATE(1440), 1,
      sym__primary_expression,
    STATE(1543), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(289), 2,
      sym_number_literal,
      sym_string,
    STATE(335), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1158), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(338), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(420), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [13011] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1159), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [13083] = 36,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2463), 1,
      anon_sym_SEMI,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1514), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1160), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [13213] = 36,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2465), 1,
      anon_sym_SEMI,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1517), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1161), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [13343] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1162), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [13419] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1163), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [13493] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1164), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [13567] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1165), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1519), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [13639] = 36,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2467), 1,
      anon_sym_SEMI,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1506), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1166), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [13769] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1167), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [13845] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2469), 1,
      sym_identifier,
    ACTIONS(2471), 1,
      anon_sym_null,
    ACTIONS(2473), 1,
      anon_sym_this,
    STATE(315), 1,
      sym_function_call,
    STATE(720), 1,
      sym_map_item,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1408), 1,
      sym__type_identifier,
    STATE(1429), 1,
      sym_parenthesized_expression,
    STATE(1430), 1,
      sym__callable_expression,
    STATE(1437), 1,
      sym__primary_expression,
    STATE(1545), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(271), 2,
      sym_number_literal,
      sym_string,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(317), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1168), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(319), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(462), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [13981] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1169), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [14059] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2475), 1,
      sym_identifier,
    ACTIONS(2477), 1,
      anon_sym_null,
    ACTIONS(2479), 1,
      anon_sym_this,
    STATE(901), 1,
      sym_map_item,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1157), 1,
      sym_function_call,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1546), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1159), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1164), 2,
      sym_number_literal,
      sym_string,
    STATE(1170), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1184), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1411), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [14195] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2475), 1,
      sym_identifier,
    ACTIONS(2477), 1,
      anon_sym_null,
    ACTIONS(2479), 1,
      anon_sym_this,
    STATE(720), 1,
      sym_map_item,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1157), 1,
      sym_function_call,
    STATE(1385), 1,
      sym__type_identifier,
    STATE(1393), 1,
      sym_parenthesized_expression,
    STATE(1546), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1159), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1164), 2,
      sym_number_literal,
      sym_string,
    STATE(1171), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1184), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(1411), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [14331] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1172), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [14405] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1173), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [14481] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2282), 1,
      anon_sym_DOT,
    ACTIONS(2481), 1,
      anon_sym_LBRACK,
    STATE(660), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(2264), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1174), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 30,
      anon_sym_SEMI,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [14559] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2483), 1,
      sym_identifier,
    ACTIONS(2485), 1,
      anon_sym_null,
    ACTIONS(2487), 1,
      anon_sym_this,
    STATE(520), 1,
      sym_function_call,
    STATE(901), 1,
      sym_map_item,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1401), 1,
      sym__type_identifier,
    STATE(1432), 1,
      sym__callable_expression,
    STATE(1433), 1,
      sym_parenthesized_expression,
    STATE(1436), 1,
      sym__primary_expression,
    STATE(1547), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(460), 2,
      sym_number_literal,
      sym_string,
    STATE(525), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1175), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(526), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(593), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [14695] = 39,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(2489), 1,
      sym_identifier,
    ACTIONS(2491), 1,
      anon_sym_null,
    ACTIONS(2493), 1,
      anon_sym_this,
    STATE(567), 1,
      sym_function_call,
    STATE(926), 1,
      sym_map_item,
    STATE(1037), 1,
      sym__juxtable_expression,
    STATE(1415), 1,
      sym__type_identifier,
    STATE(1428), 1,
      sym_parenthesized_expression,
    STATE(1434), 1,
      sym__callable_expression,
    STATE(1439), 1,
      sym__primary_expression,
    STATE(1544), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(474), 2,
      sym_number_literal,
      sym_string,
    STATE(568), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1176), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(569), 3,
      sym_boolean_literal,
      sym_list,
      sym_map,
    STATE(592), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [14831] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(488), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1177), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [14958] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(616), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1178), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15085] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(620), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1179), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15212] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(643), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1180), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15339] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1189), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1181), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15466] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(644), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1182), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15593] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(397), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1183), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15720] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1184), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [15791] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1426), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1185), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [15918] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(642), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1186), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16045] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1400), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1187), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16172] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1490), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1188), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16299] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1189), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1511), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [16404] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(515), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1190), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16531] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(404), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1191), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16658] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(405), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1192), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16785] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(406), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1193), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [16912] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(407), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1194), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17039] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(408), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1195), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17166] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(409), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1196), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17293] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(410), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1197), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17420] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(411), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1198), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17547] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(412), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1199), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17674] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(413), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1200), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17801] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(414), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1201), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [17928] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(415), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1202), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18055] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(416), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1203), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18182] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1419), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1204), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18309] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1420), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1205), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18436] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(503), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1206), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18563] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1516), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1207), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18690] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1412), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1208), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18817] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1402), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1209), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [18944] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(607), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1210), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19071] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(538), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1211), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19198] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(551), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1212), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19325] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(478), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1213), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19452] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(485), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1214), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19579] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(487), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1215), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19706] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(489), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1216), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19833] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(490), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1217), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [19960] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(492), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1218), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20087] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(493), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1219), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20214] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(517), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1220), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20341] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(495), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1221), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20468] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(498), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1222), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20595] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(499), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1223), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20722] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(502), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1224), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20849] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(440), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1225), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [20976] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(422), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1226), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21103] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1533), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1227), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21230] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(626), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1228), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21357] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(628), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1229), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21484] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(529), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1230), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21611] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(530), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1231), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21738] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1531), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1232), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21865] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(531), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1233), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [21992] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(532), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1234), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22119] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(565), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1235), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22246] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(534), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1236), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22373] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(535), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1237), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22500] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1445), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1238), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22627] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(527), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1239), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22754] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(447), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1240), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [22881] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(448), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1241), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23008] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(449), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1242), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23135] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(450), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1243), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23262] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(451), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1244), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23389] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(452), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1245), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23516] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(453), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1246), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23643] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(454), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1247), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23770] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(455), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1248), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [23897] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(456), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1249), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24024] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(457), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1250), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24151] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(458), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1251), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24278] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(459), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1252), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24405] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(540), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1253), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24532] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(544), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1254), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24659] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(621), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1255), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24786] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(622), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1256), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [24913] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1510), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1257), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25040] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(465), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1258), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25167] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(516), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1259), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25294] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(549), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1260), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25421] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1444), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1261), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25548] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(550), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1262), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25675] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(633), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1263), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25802] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(641), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1264), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [25929] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(609), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1265), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26056] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(552), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1266), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26183] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(553), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1267), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26310] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1305), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1268), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26437] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1403), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1269), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26564] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1307), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1270), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26691] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(555), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1271), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26818] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(556), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1272), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [26945] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(461), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1273), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27072] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1397), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1274), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27199] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(436), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1275), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27326] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(504), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1276), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27453] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1085), 1,
      anon_sym_STAR,
    ACTIONS(1087), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1113), 1,
      anon_sym_null,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1149), 1,
      anon_sym_new,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(265), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    STATE(438), 1,
      sym__expression,
    ACTIONS(1095), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1099), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1147), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1277), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(475), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27580] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(573), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1278), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27707] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(590), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1279), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27834] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(591), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1280), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [27961] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(629), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1281), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28088] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(595), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1282), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28215] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1283), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [28284] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(418), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1284), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28411] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1308), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1285), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28538] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(392), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1286), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28665] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(560), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1287), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28792] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(997), 1,
      anon_sym_STAR,
    ACTIONS(999), 1,
      anon_sym_STAR_COLON,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1019), 1,
      anon_sym_LBRACE,
    ACTIONS(1027), 1,
      anon_sym_null,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1065), 1,
      anon_sym_new,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(273), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    STATE(394), 1,
      sym__expression,
    ACTIONS(1007), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1011), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(1063), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1288), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(387), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [28919] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(639), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1289), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29046] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(575), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1290), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29173] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(576), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1291), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29300] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1425), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1292), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29427] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(578), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1293), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29554] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(579), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1294), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29681] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(580), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1295), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29808] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(581), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1296), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [29935] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(582), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1297), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30062] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(583), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1298), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30189] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(584), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1299), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30316] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(585), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1300), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30443] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(586), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1301), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30570] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(587), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1302), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30697] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1303), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [30772] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(508), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1304), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [30899] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1305), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1449), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1451), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [30968] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(640), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1306), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31095] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    STATE(1307), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1455), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1457), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [31168] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1308), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1455), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1457), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [31237] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1542), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1309), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31364] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1486), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1310), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31491] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1493), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1311), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31618] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1500), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1312), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31745] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1417), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1313), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31872] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1357), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1314), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [31999] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1360), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1315), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32126] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1363), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1316), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32253] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1366), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1317), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32380] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1368), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1318), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32507] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1319), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [32582] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1370), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1320), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32709] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1372), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1321), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32836] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1373), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1322), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [32963] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1374), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1323), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33090] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(599), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1324), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33217] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1375), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1325), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33344] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1377), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1326), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33471] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1379), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1327), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33598] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1380), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1328), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [33725] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1329), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1280), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1278), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [33794] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1330), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [33867] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1331), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [33944] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(589), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1332), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34071] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1477), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1333), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34198] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1537), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1334), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34325] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1487), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1335), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34452] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1489), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1336), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34579] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(613), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1337), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34706] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1406), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1338), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34833] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(623), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1339), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [34960] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1508), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1340), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35087] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1509), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1341), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35214] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1511), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1342), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35341] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1513), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1343), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35468] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(635), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1344), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35595] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1526), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1345), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35722] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1527), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1346), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35849] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1479), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1347), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [35976] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1528), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1348), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36103] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1481), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1349), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36230] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(17), 1,
      anon_sym_STAR,
    ACTIONS(19), 1,
      anon_sym_STAR_COLON,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(49), 1,
      anon_sym_null,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(95), 1,
      anon_sym_new,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(374), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    STATE(572), 1,
      sym__expression,
    ACTIONS(29), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(33), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(93), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1350), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(480), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36357] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1504), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1351), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36484] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1482), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1352), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36611] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1522), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1353), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36738] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1530), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1354), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36865] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1495), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1355), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [36992] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1494), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1356), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37119] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1357), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1463), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1465), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [37188] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1505), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1358), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37315] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1488), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1359), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37442] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    STATE(1360), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1467), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1469), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [37515] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1524), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1361), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37642] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1483), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1362), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37769] = 14,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    STATE(1363), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1467), 10,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1469), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [37854] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1532), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1364), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [37981] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1523), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1365), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [38108] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    STATE(1366), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1467), 14,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1469), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [38185] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1502), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1367), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [38312] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    STATE(1368), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1467), 13,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1469), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [38393] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1484), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1369), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [38520] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    STATE(1370), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1467), 5,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(1469), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [38609] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1485), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1371), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [38736] = 18,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1372), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1467), 3,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 15,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [38829] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1467), 2,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1373), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 15,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [38924] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1467), 2,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1374), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 14,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [39021] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1467), 1,
      anon_sym_QMARK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1375), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 14,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [39120] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(605), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1376), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [39247] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1467), 1,
      anon_sym_QMARK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1377), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 13,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [39348] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(611), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1378), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [39475] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1467), 1,
      anon_sym_QMARK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1379), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1469), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [39578] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    STATE(1380), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1467), 16,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1469), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [39651] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(861), 1,
      anon_sym_STAR,
    ACTIONS(863), 1,
      anon_sym_STAR_COLON,
    ACTIONS(873), 1,
      anon_sym_SLASH,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(891), 1,
      anon_sym_null,
    ACTIONS(893), 1,
      anon_sym_this,
    ACTIONS(905), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_SQUOTE,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_DQUOTE,
    ACTIONS(919), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(921), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(927), 1,
      anon_sym_new,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1727), 1,
      sym_identifier,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(1030), 1,
      sym__callable_expression,
    STATE(1032), 1,
      sym__type_identifier,
    STATE(1045), 1,
      sym__primary_expression,
    STATE(1535), 1,
      sym__expression,
    ACTIONS(871), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(875), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(877), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(907), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(925), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(1048), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1076), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1381), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(909), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1037), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1072), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(1283), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [39778] = 35,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(586), 1,
      anon_sym_STAR,
    ACTIONS(588), 1,
      anon_sym_STAR_COLON,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(608), 1,
      anon_sym_LBRACE,
    ACTIONS(618), 1,
      anon_sym_null,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(658), 1,
      anon_sym_new,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(443), 1,
      sym__primary_expression,
    STATE(577), 1,
      sym__expression,
    ACTIONS(596), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(600), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(656), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1382), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
    STATE(597), 6,
      sym_access_op,
      sym_increment_op,
      sym_binary_op,
      sym_closure,
      sym_ternary_op,
      sym_unary_op,
  [39905] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1383), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [39981] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1384), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [40055] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2245), 1,
      anon_sym_DOT,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1031), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1385), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 30,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [40129] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1386), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40201] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1387), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40273] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1388), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40347] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1389), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40421] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1390), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [40492] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1692), 1,
      anon_sym_SEMI,
    STATE(1391), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1519), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40563] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1392), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(365), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(363), 30,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [40632] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1393), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 30,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [40701] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1690), 1,
      anon_sym_SEMI,
    STATE(1394), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [40772] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1290), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(2245), 1,
      anon_sym_DOT,
    STATE(1031), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1286), 2,
      anon_sym_LT,
      sym_identifier,
    STATE(1395), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 19,
      anon_sym_STAR,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_instanceof,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
    ACTIONS(97), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [40847] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1396), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [40920] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1397), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1501), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_RBRACK,
      anon_sym_catch,
      anon_sym_finally,
  [41020] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(1398), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 29,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [41088] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1399), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [41155] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1400), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1624), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [41254] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1370), 1,
      anon_sym_COLON,
    ACTIONS(1422), 1,
      anon_sym_DOT,
    STATE(328), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1401), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [41325] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1402), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1640), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [41424] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1403), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1602), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [41523] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1404), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1606), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [41622] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1405), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [41689] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1406), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1600), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [41788] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1407), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [41859] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1256), 1,
      anon_sym_COLON,
    ACTIONS(1282), 1,
      anon_sym_DOT,
    STATE(286), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1408), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [41930] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1409), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42029] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1410), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42098] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1411), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42165] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1412), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1636), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42264] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1413), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42331] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1414), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42398] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1441), 1,
      anon_sym_COLON,
    ACTIONS(1453), 1,
      anon_sym_DOT,
    STATE(432), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1415), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [42469] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1416), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42536] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1417), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1612), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42635] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1418), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [42702] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1419), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1628), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42801] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1420), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1632), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42900] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1421), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 32,
      anon_sym_SEMI,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [42965] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1264), 1,
      anon_sym_COLON,
    ACTIONS(1292), 1,
      anon_sym_DOT,
    STATE(269), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1422), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [43036] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1423), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1519), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [43103] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1424), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1519), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [43202] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1425), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1616), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [43301] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1426), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
    ACTIONS(1620), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [43400] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1346), 1,
      anon_sym_LPAREN,
    STATE(304), 1,
      sym_argument_list,
    STATE(1427), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [43468] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1441), 1,
      anon_sym_COLON,
    STATE(1428), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [43534] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1256), 1,
      anon_sym_COLON,
    STATE(1429), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [43600] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1358), 1,
      anon_sym_LPAREN,
    STATE(364), 1,
      sym_argument_list,
    STATE(1430), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [43668] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1264), 1,
      anon_sym_COLON,
    STATE(1431), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [43734] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1445), 1,
      anon_sym_LPAREN,
    STATE(466), 1,
      sym_argument_list,
    STATE(1432), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [43802] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1370), 1,
      anon_sym_COLON,
    STATE(1433), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(102), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(97), 27,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [43868] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    STATE(547), 1,
      sym_argument_list,
    STATE(1434), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1262), 18,
      anon_sym_STAR,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1260), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_LBRACK,
  [43936] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1435), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44005] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1422), 1,
      anon_sym_DOT,
    ACTIONS(1426), 1,
      anon_sym_LBRACK,
    STATE(328), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1436), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1436), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 23,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44076] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1282), 1,
      anon_sym_DOT,
    ACTIONS(1284), 1,
      anon_sym_LBRACK,
    STATE(286), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1312), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1437), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 23,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44147] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1438), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44214] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1453), 1,
      anon_sym_DOT,
    ACTIONS(1530), 1,
      anon_sym_LBRACK,
    STATE(432), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1528), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1439), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 23,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44285] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1292), 1,
      anon_sym_DOT,
    ACTIONS(1294), 1,
      anon_sym_LBRACK,
    STATE(269), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1326), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1440), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 17,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 23,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44356] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1692), 1,
      anon_sym_SEMI,
    STATE(1441), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1519), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44422] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1692), 1,
      anon_sym_SEMI,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1442), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1519), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [44520] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1690), 1,
      anon_sym_SEMI,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1443), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1515), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [44618] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1444), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2535), 3,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RPAREN,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [44715] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2503), 1,
      anon_sym_PLUS,
    ACTIONS(2505), 1,
      anon_sym_DASH,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1445), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1694), 3,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RPAREN,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [44812] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1445), 1,
      anon_sym_LPAREN,
    STATE(2426), 1,
      sym_argument_list,
    STATE(1446), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 26,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
      anon_sym_RPAREN,
  [44877] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1690), 1,
      anon_sym_SEMI,
    ACTIONS(1515), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1447), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [44942] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2537), 1,
      anon_sym_COMMA,
    ACTIONS(2539), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1448), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45038] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2541), 1,
      anon_sym_COMMA,
    ACTIONS(2543), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1449), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45134] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2545), 1,
      anon_sym_COMMA,
    ACTIONS(2547), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1450), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45230] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2549), 1,
      anon_sym_COMMA,
    ACTIONS(2551), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1451), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45326] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2553), 1,
      anon_sym_COMMA,
    ACTIONS(2555), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1452), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45422] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2557), 1,
      anon_sym_COMMA,
    ACTIONS(2559), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1453), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45518] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2561), 1,
      anon_sym_COMMA,
    ACTIONS(2563), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1454), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45614] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2565), 1,
      anon_sym_COMMA,
    ACTIONS(2567), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1455), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45710] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2569), 1,
      anon_sym_COMMA,
    ACTIONS(2571), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1456), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45806] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2573), 1,
      anon_sym_COMMA,
    ACTIONS(2575), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1457), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45902] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2577), 1,
      anon_sym_COMMA,
    ACTIONS(2579), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1458), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [45998] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2581), 1,
      anon_sym_COMMA,
    ACTIONS(2583), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1459), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46094] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2585), 1,
      anon_sym_COMMA,
    ACTIONS(2587), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1460), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46190] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1515), 1,
      anon_sym_while,
    ACTIONS(1690), 1,
      anon_sym_SEMI,
    STATE(1461), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1310), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_QMARK,
    ACTIONS(1308), 25,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
      anon_sym_as,
      anon_sym_PERCENT,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_COLON,
      anon_sym_STAR_STAR,
  [46254] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2307), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2589), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1462), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46350] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2383), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2591), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1463), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46446] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2385), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2593), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1464), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46542] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2595), 1,
      anon_sym_COMMA,
    ACTIONS(2597), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1465), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46638] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2599), 1,
      anon_sym_COMMA,
    ACTIONS(2601), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1466), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46734] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2603), 1,
      anon_sym_COMMA,
    ACTIONS(2605), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1467), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46830] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2607), 1,
      anon_sym_COMMA,
    ACTIONS(2609), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1468), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [46926] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2611), 1,
      anon_sym_COMMA,
    ACTIONS(2613), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1469), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47022] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2615), 1,
      anon_sym_COMMA,
    ACTIONS(2617), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1470), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47118] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2619), 1,
      anon_sym_COMMA,
    ACTIONS(2621), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1471), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47214] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2623), 1,
      anon_sym_COMMA,
    ACTIONS(2625), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1472), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47310] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2381), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2627), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1473), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47406] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2455), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2629), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1474), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47502] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2379), 1,
      anon_sym_RBRACK,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2631), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1475), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47598] = 24,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2633), 1,
      anon_sym_COMMA,
    ACTIONS(2635), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1476), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47694] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2637), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1477), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47787] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2639), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1478), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47880] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2641), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1479), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [47973] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2643), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1480), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48066] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2645), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1481), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48159] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2647), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1482), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48252] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2649), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1483), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48345] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2651), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1484), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48438] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2653), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1485), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48531] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2655), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1486), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48624] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_RBRACE,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1487), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48717] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2659), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1488), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48810] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2661), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1489), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48903] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2663), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1490), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [48996] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2665), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1491), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49089] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2667), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1492), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49182] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2669), 1,
      anon_sym_RBRACE,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1493), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49275] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2671), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1494), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49368] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2673), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1495), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49461] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2675), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1496), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49554] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2677), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1497), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49647] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2679), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1498), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49740] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2681), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1499), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49833] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2683), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1500), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [49926] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2685), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1501), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50019] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2687), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1502), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50112] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2689), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1503), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50205] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2691), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1504), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50298] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2693), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1505), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50391] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2695), 1,
      anon_sym_SEMI,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1506), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50484] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2697), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1507), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50577] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2699), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1508), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50670] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2701), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1509), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50763] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2703), 1,
      anon_sym_RBRACE,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1510), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50856] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2705), 1,
      anon_sym_RBRACE,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1511), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [50949] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2707), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1512), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51042] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2709), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1513), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51135] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2711), 1,
      anon_sym_SEMI,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1514), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51228] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2713), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1515), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51321] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2715), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1516), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51414] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2717), 1,
      anon_sym_SEMI,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1517), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51507] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1518), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51600] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2721), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1519), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51693] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2723), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1520), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51786] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2725), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1521), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51879] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2727), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1522), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [51972] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2729), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1523), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52065] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2731), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1524), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52158] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2733), 1,
      anon_sym_COMMA,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1525), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52251] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2735), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1526), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52344] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2737), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1527), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52437] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2739), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1528), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52530] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2741), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1529), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52623] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2743), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1530), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52716] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2745), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1531), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52809] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2747), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1532), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52902] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2749), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1533), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [52995] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2751), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1534), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53088] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2753), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1535), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53181] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2755), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1536), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53274] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2757), 1,
      anon_sym_RBRACK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1537), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53367] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2759), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1538), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53460] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2761), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1539), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53553] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2763), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1540), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53646] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2765), 1,
      anon_sym_RPAREN,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1541), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53739] = 23,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2767), 1,
      anon_sym_COLON,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1542), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53832] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1461), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1473), 1,
      anon_sym_PERCENT,
    ACTIONS(1489), 1,
      anon_sym_AMP,
    ACTIONS(1491), 1,
      anon_sym_CARET,
    ACTIONS(1493), 1,
      anon_sym_PIPE,
    ACTIONS(1495), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1497), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1503), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(1505), 1,
      anon_sym_QMARK,
    ACTIONS(1471), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2769), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1543), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1479), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(1459), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1477), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1483), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(1481), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(1487), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [53922] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1652), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1656), 1,
      anon_sym_PERCENT,
    ACTIONS(1672), 1,
      anon_sym_AMP,
    ACTIONS(1674), 1,
      anon_sym_CARET,
    ACTIONS(1676), 1,
      anon_sym_PIPE,
    ACTIONS(1678), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1680), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1682), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(1684), 1,
      anon_sym_QMARK,
    ACTIONS(1654), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1668), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2771), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1544), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1662), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(1650), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1660), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1666), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(1664), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(1670), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [54012] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1526), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1534), 1,
      anon_sym_PERCENT,
    ACTIONS(1550), 1,
      anon_sym_AMP,
    ACTIONS(1552), 1,
      anon_sym_CARET,
    ACTIONS(1554), 1,
      anon_sym_PIPE,
    ACTIONS(1556), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1558), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1560), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(1562), 1,
      anon_sym_QMARK,
    ACTIONS(1532), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1546), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2773), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1545), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1540), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(1524), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1538), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1544), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(1542), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(1548), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [54102] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2501), 1,
      anon_sym_PERCENT,
    ACTIONS(2517), 1,
      anon_sym_AMP,
    ACTIONS(2519), 1,
      anon_sym_CARET,
    ACTIONS(2521), 1,
      anon_sym_PIPE,
    ACTIONS(2523), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2525), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2527), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(2529), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2531), 1,
      anon_sym_QMARK,
    ACTIONS(2497), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2503), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2513), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    STATE(1546), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2509), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(2495), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(2499), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(2507), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(2511), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(2515), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [54192] = 22,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1568), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1572), 1,
      anon_sym_PERCENT,
    ACTIONS(1588), 1,
      anon_sym_AMP,
    ACTIONS(1590), 1,
      anon_sym_CARET,
    ACTIONS(1592), 1,
      anon_sym_PIPE,
    ACTIONS(1594), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1596), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1642), 1,
      anon_sym_QMARK_COLON,
    ACTIONS(1644), 1,
      anon_sym_QMARK,
    ACTIONS(1570), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1584), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2775), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(1547), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1578), 3,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DOT_DOT,
    ACTIONS(1566), 4,
      anon_sym_DOT_AMP,
      anon_sym_DOT_AT,
      anon_sym_QMARK_DOT,
      anon_sym_STAR_DOT,
    ACTIONS(1576), 4,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_DOT_DOT_LT,
      anon_sym_LT_DOT_DOT_LT,
    ACTIONS(1582), 4,
      anon_sym_as,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_BANGinstanceof,
    ACTIONS(1580), 5,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_in,
      anon_sym_BANGin,
    ACTIONS(1586), 5,
      anon_sym_LT_EQ_GT,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_EQ_EQ_TILDE,
  [54282] = 29,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2777), 1,
      sym_identifier,
    ACTIONS(2781), 1,
      anon_sym_LPAREN,
    STATE(383), 1,
      sym__callable_expression,
    STATE(470), 1,
      sym_function_call,
    STATE(1883), 1,
      sym__type_identifier,
    STATE(1986), 1,
      sym__primary_expression,
    STATE(2288), 1,
      sym_assignment,
    STATE(2298), 1,
      sym_increment_op,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    ACTIONS(2779), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1548), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1727), 2,
      sym_parenthesized_expression,
      sym__juxtable_expression,
    STATE(1728), 2,
      sym_dotted_identifier,
      sym_index,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54383] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2029), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1549), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54472] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1978), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1550), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54561] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(598), 1,
      anon_sym_SLASH,
    ACTIONS(620), 1,
      anon_sym_this,
    ACTIONS(632), 1,
      anon_sym_LBRACK,
    ACTIONS(640), 1,
      anon_sym_SQUOTE,
    ACTIONS(642), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(644), 1,
      anon_sym_DQUOTE,
    ACTIONS(646), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(648), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1721), 1,
      sym_identifier,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    STATE(389), 1,
      sym__callable_expression,
    STATE(393), 1,
      sym__type_identifier,
    STATE(468), 1,
      sym__primary_expression,
    ACTIONS(602), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(634), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(391), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(481), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1551), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(636), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(563), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(541), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54650] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2032), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1552), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54739] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1009), 1,
      anon_sym_SLASH,
    ACTIONS(1029), 1,
      anon_sym_this,
    ACTIONS(1041), 1,
      anon_sym_LBRACK,
    ACTIONS(1049), 1,
      anon_sym_SQUOTE,
    ACTIONS(1051), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    ACTIONS(1055), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1057), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1706), 1,
      sym_identifier,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    STATE(256), 1,
      sym__primary_expression,
    STATE(284), 1,
      sym__callable_expression,
    STATE(292), 1,
      sym__type_identifier,
    ACTIONS(1013), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1043), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(285), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(367), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1553), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1045), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(360), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(303), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54828] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2035), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1554), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [54917] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2037), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1555), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55006] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2783), 1,
      sym_identifier,
    ACTIONS(2785), 1,
      anon_sym_SLASH,
    ACTIONS(2789), 1,
      anon_sym_LPAREN,
    ACTIONS(2791), 1,
      anon_sym_this,
    ACTIONS(2793), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2795), 1,
      anon_sym_LBRACK,
    ACTIONS(2801), 1,
      anon_sym_SQUOTE,
    ACTIONS(2803), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(2805), 1,
      anon_sym_DQUOTE,
    ACTIONS(2807), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(2809), 1,
      anon_sym_DOLLAR_SLASH,
    STATE(1063), 1,
      sym__callable_expression,
    STATE(1078), 1,
      sym__type_identifier,
    STATE(1088), 1,
      sym__primary_expression,
    ACTIONS(2787), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2797), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(1082), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1109), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1556), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2799), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1093), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(1112), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55095] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1920), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1557), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55184] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1971), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1558), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55273] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1712), 1,
      sym_identifier,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(330), 1,
      sym__type_identifier,
    STATE(344), 1,
      sym__primary_expression,
    STATE(383), 1,
      sym__callable_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(395), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1559), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55362] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1924), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1560), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55451] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1941), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1561), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55540] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1942), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1562), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55629] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1945), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1563), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55718] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1946), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1564), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55807] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1950), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1565), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55896] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1951), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1566), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [55985] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1952), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1567), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56074] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1097), 1,
      anon_sym_SLASH,
    ACTIONS(1115), 1,
      anon_sym_this,
    ACTIONS(1127), 1,
      anon_sym_LBRACK,
    ACTIONS(1135), 1,
      anon_sym_SQUOTE,
    ACTIONS(1137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1139), 1,
      anon_sym_DQUOTE,
    ACTIONS(1141), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(1143), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1698), 1,
      sym_identifier,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    STATE(254), 1,
      sym__primary_expression,
    STATE(275), 1,
      sym__type_identifier,
    STATE(294), 1,
      sym__callable_expression,
    ACTIONS(1101), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1129), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(298), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(349), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(1568), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(312), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(329), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56163] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2034), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1569), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56252] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1991), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1570), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56341] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1995), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1571), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56430] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1998), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1572), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56519] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1977), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1573), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56608] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2001), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1574), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56697] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2003), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1575), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56786] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2004), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1576), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56875] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2005), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1577), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [56964] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2018), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1578), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57053] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2023), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1579), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57142] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2025), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1580), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57231] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2027), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1581), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57320] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2028), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1582), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57409] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2030), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1583), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57498] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2031), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1584), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57587] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1923), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1585), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57676] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1926), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1586), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57765] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1928), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1587), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57854] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1930), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1588), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [57943] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1931), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1589), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58032] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1932), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1590), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58121] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1933), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1591), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58210] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1934), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1592), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58299] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1948), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1593), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58388] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1954), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1594), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58477] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1957), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1595), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58566] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1959), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1596), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58655] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1960), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1597), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58744] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1961), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1598), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58833] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1962), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1599), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [58922] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(1963), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1600), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [59011] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2011), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1601), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [59100] = 25,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(51), 1,
      anon_sym_this,
    ACTIONS(63), 1,
      anon_sym_LBRACK,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(1224), 1,
      sym_identifier,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    STATE(383), 1,
      sym__callable_expression,
    STATE(1860), 1,
      sym__type_identifier,
    STATE(2000), 1,
      sym__primary_expression,
    ACTIONS(35), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(26), 2,
      sym_dotted_identifier,
      sym_index,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1602), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(470), 3,
      sym_parenthesized_expression,
      sym__juxtable_expression,
      sym_function_call,
    STATE(439), 5,
      sym_boolean_literal,
      sym_list,
      sym_map,
      sym_number_literal,
      sym_string,
  [59189] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2811), 1,
      sym_identifier,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1603), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59268] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2815), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1604), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59347] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2819), 1,
      anon_sym__,
    ACTIONS(2821), 1,
      anon_sym_def,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    STATE(1667), 1,
      aux_sym_class_definition_repeat2,
    STATE(1674), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2108), 1,
      sym__type,
    STATE(1605), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2817), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59424] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2831), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1606), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59503] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2833), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1607), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59582] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2835), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1608), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59661] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2819), 1,
      anon_sym__,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2839), 1,
      anon_sym_def,
    STATE(1665), 1,
      aux_sym_class_definition_repeat2,
    STATE(1682), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2050), 1,
      sym__type,
    STATE(1609), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2837), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59738] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2843), 1,
      anon_sym__,
    ACTIONS(2845), 1,
      anon_sym_def,
    STATE(1659), 1,
      aux_sym_class_definition_repeat2,
    STATE(1684), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2091), 1,
      sym__type,
    STATE(1610), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2841), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59815] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2849), 1,
      anon_sym__,
    ACTIONS(2851), 1,
      anon_sym_def,
    STATE(1663), 1,
      aux_sym_class_definition_repeat2,
    STATE(1675), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2056), 1,
      sym__type,
    STATE(1611), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2847), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59892] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2857), 1,
      anon_sym_def,
    STATE(1658), 1,
      aux_sym_class_definition_repeat2,
    STATE(1680), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2048), 1,
      sym__type,
    STATE(1612), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2853), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [59969] = 20,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2861), 1,
      anon_sym_def,
    STATE(1666), 1,
      aux_sym_class_definition_repeat2,
    STATE(1676), 1,
      sym_access_modifier,
    STATE(1686), 1,
      aux_sym_class_definition_repeat1,
    STATE(1711), 1,
      sym_annotation,
    STATE(1749), 1,
      sym_modifier,
    STATE(2093), 1,
      sym__type,
    STATE(1613), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2827), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2859), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60046] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2863), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1614), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60125] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2865), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1615), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60204] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2867), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1616), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60283] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2869), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1617), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60362] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2871), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1618), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60441] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2873), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1619), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60520] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2875), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1620), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60599] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2877), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1621), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60678] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2879), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1622), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60757] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2881), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1623), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60836] = 21,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(43), 1,
      anon_sym__,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(1731), 1,
      anon_sym_def,
    ACTIONS(2813), 1,
      anon_sym_SEMI,
    ACTIONS(2883), 1,
      sym_identifier,
    STATE(1655), 1,
      aux_sym_class_definition_repeat1,
    STATE(1695), 1,
      aux_sym_class_definition_repeat2,
    STATE(1704), 1,
      sym_access_modifier,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2103), 1,
      sym__type,
    STATE(2163), 1,
      sym_declaration,
    STATE(1624), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1735), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [60915] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2617), 1,
      anon_sym_RBRACK,
    STATE(1625), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [60965] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2583), 1,
      anon_sym_RPAREN,
    STATE(1626), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61015] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2579), 1,
      anon_sym_RPAREN,
    STATE(1627), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61065] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2539), 1,
      anon_sym_RBRACK,
    STATE(1628), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61115] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2891), 1,
      anon_sym_RBRACK,
    STATE(1629), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61165] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2625), 1,
      anon_sym_RBRACK,
    STATE(1630), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61215] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2893), 1,
      anon_sym_RPAREN,
    STATE(1631), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61265] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2621), 1,
      anon_sym_RPAREN,
    STATE(1632), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61315] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2587), 1,
      anon_sym_RPAREN,
    STATE(1633), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61365] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2597), 1,
      anon_sym_RBRACK,
    STATE(1634), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61415] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2895), 1,
      anon_sym_RBRACK,
    STATE(1635), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61465] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2897), 1,
      anon_sym_RPAREN,
    STATE(1636), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61515] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2575), 1,
      anon_sym_RPAREN,
    STATE(1637), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61565] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2899), 1,
      anon_sym_RPAREN,
    STATE(1638), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61615] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2901), 1,
      anon_sym_RPAREN,
    STATE(1639), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61665] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2903), 1,
      anon_sym_RBRACK,
    STATE(1640), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61715] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2601), 1,
      anon_sym_RBRACK,
    STATE(1641), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61765] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2905), 1,
      anon_sym_RBRACK,
    STATE(1642), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61815] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2907), 1,
      anon_sym_RPAREN,
    STATE(1643), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61865] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2551), 1,
      anon_sym_RPAREN,
    STATE(1644), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61915] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2909), 1,
      anon_sym_RBRACK,
    STATE(1645), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [61965] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2547), 1,
      anon_sym_RBRACK,
    STATE(1646), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62015] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2911), 1,
      anon_sym_RPAREN,
    STATE(1647), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62065] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2609), 1,
      anon_sym_RPAREN,
    STATE(1648), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62115] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1649), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 15,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62163] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2913), 1,
      anon_sym_RPAREN,
    STATE(1650), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62213] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2563), 1,
      anon_sym_RPAREN,
    STATE(1651), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62263] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2915), 1,
      anon_sym_RBRACK,
    STATE(1652), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2350), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2885), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62313] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2917), 1,
      anon_sym_RPAREN,
    STATE(1653), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62363] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1654), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2889), 14,
      anon_sym_STAR_COLON,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2887), 15,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_true,
      anon_sym_false,
      anon_sym_null,
      anon_sym_this,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      anon_sym_new,
  [62410] = 18,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1729), 1,
      anon_sym_AT,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2919), 1,
      anon_sym_def,
    STATE(1698), 1,
      aux_sym_class_definition_repeat2,
    STATE(1705), 1,
      sym_access_modifier,
    STATE(1719), 1,
      aux_sym_class_definition_repeat1,
    STATE(1740), 1,
      sym_annotation,
    STATE(1755), 1,
      sym_modifier,
    STATE(2067), 1,
      sym__type,
    STATE(1655), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2921), 3,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
    ACTIONS(2923), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62480] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2859), 1,
      anon_sym_ATinterface,
    ACTIONS(2925), 1,
      sym_identifier,
    ACTIONS(2929), 1,
      anon_sym__,
    ACTIONS(2931), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2093), 1,
      sym__type,
    ACTIONS(2927), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1656), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62543] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2853), 1,
      anon_sym_ATinterface,
    ACTIONS(2929), 1,
      anon_sym__,
    ACTIONS(2933), 1,
      sym_identifier,
    ACTIONS(2937), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2048), 1,
      sym__type,
    ACTIONS(2935), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1657), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62606] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2939), 1,
      sym_identifier,
    ACTIONS(2941), 1,
      anon_sym_ATinterface,
    ACTIONS(2945), 1,
      anon_sym__,
    ACTIONS(2947), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2049), 1,
      sym__type,
    ACTIONS(2943), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1658), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62669] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2949), 1,
      sym_identifier,
    ACTIONS(2951), 1,
      anon_sym_ATinterface,
    ACTIONS(2955), 1,
      anon_sym__,
    ACTIONS(2957), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2095), 1,
      sym__type,
    ACTIONS(2953), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1659), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62732] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2959), 1,
      sym_identifier,
    ACTIONS(2961), 1,
      anon_sym_ATinterface,
    ACTIONS(2965), 1,
      anon_sym__,
    ACTIONS(2967), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2052), 1,
      sym__type,
    ACTIONS(2963), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1660), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62795] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2817), 1,
      anon_sym_ATinterface,
    ACTIONS(2969), 1,
      sym_identifier,
    ACTIONS(2973), 1,
      anon_sym__,
    ACTIONS(2975), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2108), 1,
      sym__type,
    ACTIONS(2971), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1661), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62858] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2847), 1,
      anon_sym_ATinterface,
    ACTIONS(2977), 1,
      sym_identifier,
    ACTIONS(2981), 1,
      anon_sym__,
    ACTIONS(2983), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2056), 1,
      sym__type,
    ACTIONS(2979), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1662), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62921] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2985), 1,
      sym_identifier,
    ACTIONS(2987), 1,
      anon_sym_ATinterface,
    ACTIONS(2991), 1,
      anon_sym__,
    ACTIONS(2993), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2057), 1,
      sym__type,
    ACTIONS(2989), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1663), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [62984] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2995), 1,
      sym_identifier,
    ACTIONS(2997), 1,
      anon_sym_ATinterface,
    ACTIONS(3001), 1,
      anon_sym__,
    ACTIONS(3003), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2062), 1,
      sym__type,
    ACTIONS(2999), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1664), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63047] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3005), 1,
      sym_identifier,
    ACTIONS(3007), 1,
      anon_sym_ATinterface,
    ACTIONS(3011), 1,
      anon_sym__,
    ACTIONS(3013), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2058), 1,
      sym__type,
    ACTIONS(3009), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1665), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63110] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2945), 1,
      anon_sym__,
    ACTIONS(3015), 1,
      sym_identifier,
    ACTIONS(3017), 1,
      anon_sym_ATinterface,
    ACTIONS(3021), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2100), 1,
      sym__type,
    ACTIONS(3019), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1666), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63173] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3011), 1,
      anon_sym__,
    ACTIONS(3023), 1,
      sym_identifier,
    ACTIONS(3025), 1,
      anon_sym_ATinterface,
    ACTIONS(3029), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2043), 1,
      sym__type,
    ACTIONS(3027), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1667), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63236] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2841), 1,
      anon_sym_ATinterface,
    ACTIONS(3031), 1,
      sym_identifier,
    ACTIONS(3035), 1,
      anon_sym__,
    ACTIONS(3037), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2091), 1,
      sym__type,
    ACTIONS(3033), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1668), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63299] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2837), 1,
      anon_sym_ATinterface,
    ACTIONS(2973), 1,
      anon_sym__,
    ACTIONS(3039), 1,
      sym_identifier,
    ACTIONS(3043), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2050), 1,
      sym__type,
    ACTIONS(3041), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1669), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63362] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3045), 1,
      sym_identifier,
    ACTIONS(3047), 1,
      anon_sym_ATinterface,
    ACTIONS(3051), 1,
      anon_sym__,
    ACTIONS(3053), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2102), 1,
      sym__type,
    ACTIONS(3049), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1670), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63425] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3001), 1,
      anon_sym__,
    ACTIONS(3055), 1,
      sym_identifier,
    ACTIONS(3057), 1,
      anon_sym_ATinterface,
    ACTIONS(3061), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2120), 1,
      sym__type,
    ACTIONS(3059), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1671), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63488] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3063), 1,
      sym_identifier,
    ACTIONS(3065), 1,
      anon_sym_ATinterface,
    ACTIONS(3069), 1,
      anon_sym__,
    ACTIONS(3071), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2104), 1,
      sym__type,
    ACTIONS(3067), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1672), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63551] = 16,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2965), 1,
      anon_sym__,
    ACTIONS(3073), 1,
      sym_identifier,
    ACTIONS(3075), 1,
      anon_sym_ATinterface,
    ACTIONS(3079), 1,
      anon_sym_def,
    STATE(1718), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2085), 1,
      sym__type,
    ACTIONS(3077), 2,
      anon_sym_interface,
      anon_sym_class,
    STATE(1673), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(91), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63614] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3081), 1,
      anon_sym__,
    ACTIONS(3083), 1,
      anon_sym_def,
    STATE(1671), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2043), 1,
      sym__type,
    STATE(1674), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(3025), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63674] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3085), 1,
      anon_sym__,
    ACTIONS(3087), 1,
      anon_sym_def,
    STATE(1672), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2057), 1,
      sym__type,
    STATE(1675), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2987), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63734] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3089), 1,
      anon_sym__,
    ACTIONS(3091), 1,
      anon_sym_def,
    STATE(1673), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2100), 1,
      sym__type,
    STATE(1676), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(3017), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63794] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2849), 1,
      anon_sym__,
    ACTIONS(2851), 1,
      anon_sym_def,
    STATE(1663), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2056), 1,
      sym__type,
    STATE(1677), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2847), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63854] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2857), 1,
      anon_sym_def,
    STATE(1658), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2048), 1,
      sym__type,
    STATE(1678), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2853), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63914] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2843), 1,
      anon_sym__,
    ACTIONS(2845), 1,
      anon_sym_def,
    STATE(1659), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2091), 1,
      sym__type,
    STATE(1679), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2841), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [63974] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3089), 1,
      anon_sym__,
    ACTIONS(3093), 1,
      anon_sym_def,
    STATE(1660), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2049), 1,
      sym__type,
    STATE(1680), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2941), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64034] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2819), 1,
      anon_sym__,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2839), 1,
      anon_sym_def,
    STATE(1665), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2050), 1,
      sym__type,
    STATE(1681), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2837), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64094] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3081), 1,
      anon_sym__,
    ACTIONS(3095), 1,
      anon_sym_def,
    STATE(1664), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2058), 1,
      sym__type,
    STATE(1682), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(3007), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64154] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2819), 1,
      anon_sym__,
    ACTIONS(2821), 1,
      anon_sym_def,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    STATE(1667), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2108), 1,
      sym__type,
    STATE(1683), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2817), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64214] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3097), 1,
      anon_sym__,
    ACTIONS(3099), 1,
      anon_sym_def,
    STATE(1670), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2095), 1,
      sym__type,
    STATE(1684), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2951), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64274] = 15,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(87), 1,
      anon_sym_int,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2861), 1,
      anon_sym_def,
    STATE(1666), 1,
      aux_sym_class_definition_repeat2,
    STATE(1749), 1,
      sym_modifier,
    STATE(2093), 1,
      sym__type,
    STATE(1685), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2829), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    ACTIONS(2859), 3,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 7,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64334] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3101), 1,
      anon_sym_AT,
    ACTIONS(3106), 1,
      anon_sym_int,
    STATE(1711), 1,
      sym_annotation,
    STATE(1686), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_class_definition_repeat1,
    ACTIONS(3104), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64379] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(1706), 1,
      sym_closure,
    ACTIONS(1306), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1687), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1304), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64424] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(1710), 1,
      sym_closure,
    ACTIONS(1338), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1688), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1336), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64469] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(1713), 1,
      sym_closure,
    ACTIONS(1334), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1689), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1332), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64514] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3114), 1,
      anon_sym_LPAREN,
    STATE(1708), 1,
      sym_argument_list,
    ACTIONS(3110), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1690), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3112), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64559] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3108), 1,
      anon_sym_LBRACE,
    STATE(1702), 1,
      sym_closure,
    ACTIONS(1324), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1691), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1322), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [64604] = 18,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3116), 1,
      sym_identifier,
    ACTIONS(3119), 1,
      anon_sym_SLASH,
    ACTIONS(3122), 1,
      anon_sym_LPAREN,
    ACTIONS(3125), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3134), 1,
      anon_sym_SQUOTE,
    ACTIONS(3137), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(3140), 1,
      anon_sym_DQUOTE,
    ACTIONS(3143), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(3146), 1,
      anon_sym_DOLLAR_SLASH,
    STATE(2358), 1,
      sym_map_item,
    ACTIONS(3128), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    ACTIONS(3131), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(1692), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_map_repeat1,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [64668] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2263), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1693), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [64734] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2323), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1694), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [64800] = 14,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2929), 1,
      anon_sym__,
    ACTIONS(2933), 1,
      sym_identifier,
    ACTIONS(3151), 1,
      anon_sym_def,
    STATE(1751), 1,
      aux_sym_class_definition_repeat2,
    STATE(1755), 1,
      sym_modifier,
    STATE(2067), 1,
      sym__type,
    STATE(1695), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [64856] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2228), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1696), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [64922] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2325), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1697), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [64988] = 14,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2939), 1,
      sym_identifier,
    ACTIONS(2945), 1,
      anon_sym__,
    ACTIONS(3153), 1,
      anon_sym_def,
    STATE(1751), 1,
      aux_sym_class_definition_repeat2,
    STATE(1755), 1,
      sym_modifier,
    STATE(2071), 1,
      sym__type,
    STATE(1698), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [65044] = 14,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2959), 1,
      sym_identifier,
    ACTIONS(2965), 1,
      anon_sym__,
    ACTIONS(3155), 1,
      anon_sym_def,
    STATE(1751), 1,
      aux_sym_class_definition_repeat2,
    STATE(1755), 1,
      sym_modifier,
    STATE(2098), 1,
      sym__type,
    STATE(1699), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1737), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [65100] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2339), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1700), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [65166] = 19,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(31), 1,
      anon_sym_SLASH,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(73), 1,
      anon_sym_SQUOTE,
    ACTIONS(75), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(77), 1,
      anon_sym_DQUOTE,
    ACTIONS(79), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(81), 1,
      anon_sym_DOLLAR_SLASH,
    ACTIONS(3149), 1,
      sym_identifier,
    STATE(1692), 1,
      aux_sym_map_repeat1,
    STATE(2293), 1,
      sym_map_item,
    ACTIONS(65), 2,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
    STATE(346), 2,
      sym__plain_string,
      sym__interpolate_string,
    STATE(1701), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(67), 3,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
    STATE(2356), 4,
      sym_parenthesized_expression,
      sym__type_identifier,
      sym_number_literal,
      sym_string,
  [65232] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1338), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1702), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1336), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65271] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1412), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1703), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1410), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65310] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(2855), 1,
      anon_sym__,
    ACTIONS(2919), 1,
      anon_sym_def,
    STATE(1698), 1,
      aux_sym_class_definition_repeat2,
    STATE(1755), 1,
      sym_modifier,
    STATE(2067), 1,
      sym__type,
    STATE(1704), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2923), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [65363] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2823), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3089), 1,
      anon_sym__,
    ACTIONS(3157), 1,
      anon_sym_def,
    STATE(1699), 1,
      aux_sym_class_definition_repeat2,
    STATE(1755), 1,
      sym_modifier,
    STATE(2071), 1,
      sym__type,
    STATE(1705), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2923), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [65416] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1334), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1706), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1332), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65455] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1362), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1707), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1364), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65494] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3159), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1708), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3161), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65533] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1376), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1709), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1374), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65572] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1306), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1710), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1304), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65611] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3163), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1711), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3165), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65650] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1392), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1712), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1390), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65689] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1430), 2,
      anon_sym_AT,
      anon_sym_int,
    STATE(1713), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1428), 19,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65728] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3167), 1,
      anon_sym_LBRACE,
    STATE(1747), 1,
      sym_closure,
    STATE(1714), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1322), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65768] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3169), 1,
      anon_sym_LPAREN,
    STATE(1742), 1,
      sym_argument_list,
    STATE(1715), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3112), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65808] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3167), 1,
      anon_sym_LBRACE,
    STATE(1734), 1,
      sym_closure,
    STATE(1716), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1336), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65848] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1722), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1717), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(373), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [65886] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3173), 1,
      anon_sym_ATinterface,
    STATE(1749), 1,
      sym_modifier,
    ACTIONS(3175), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1718), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_class_definition_repeat2,
    ACTIONS(3171), 14,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [65928] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3178), 1,
      anon_sym_AT,
    STATE(1740), 1,
      sym_annotation,
    STATE(1719), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_class_definition_repeat1,
    ACTIONS(3104), 17,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [65968] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3167), 1,
      anon_sym_LBRACE,
    STATE(1739), 1,
      sym_closure,
    STATE(1720), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1332), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66008] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3167), 1,
      anon_sym_LBRACE,
    STATE(1735), 1,
      sym_closure,
    STATE(1721), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1304), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66048] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3181), 1,
      anon_sym_DOT,
    STATE(1722), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(377), 18,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66086] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1723), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(377), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66121] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1724), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(389), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66156] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    ACTIONS(3188), 1,
      anon_sym_RPAREN,
    STATE(1743), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2328), 1,
      sym_parameter,
    STATE(1725), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66207] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    ACTIONS(3190), 1,
      anon_sym_RPAREN,
    STATE(1750), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2250), 1,
      sym_parameter,
    STATE(1726), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66258] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1727), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(97), 5,
      anon_sym_DOT,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
    ACTIONS(1198), 14,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
  [66295] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1728), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(363), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66330] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    ACTIONS(3192), 1,
      anon_sym_RPAREN,
    STATE(1741), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2319), 1,
      sym_parameter,
    STATE(1729), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66381] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1730), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(391), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66416] = 13,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    ACTIONS(3194), 1,
      anon_sym_RPAREN,
    STATE(1744), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2304), 1,
      sym_parameter,
    STATE(1731), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66467] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1732), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(369), 19,
      anon_sym_DOT,
      anon_sym_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [66502] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1733), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1364), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66536] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1734), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1304), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66570] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1735), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1332), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66604] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1736), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1374), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66638] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    STATE(1738), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2294), 1,
      sym_parameter,
    STATE(1737), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66686] = 11,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3196), 1,
      sym_identifier,
    ACTIONS(3199), 1,
      anon_sym_def,
    ACTIONS(3202), 1,
      aux_sym__type_identifier_token1,
    STATE(2047), 1,
      sym__type,
    STATE(2396), 1,
      sym_parameter,
    STATE(1738), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__param_list_repeat1,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(3205), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66732] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1739), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1428), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66766] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1740), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3165), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66800] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    STATE(1738), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2223), 1,
      sym_parameter,
    STATE(1741), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66848] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1742), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3161), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [66882] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    STATE(1738), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2222), 1,
      sym_parameter,
    STATE(1743), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66930] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    STATE(1738), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2311), 1,
      sym_parameter,
    STATE(1744), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [66978] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1745), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1410), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67012] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1746), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1390), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67046] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1747), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1336), 18,
      anon_sym_AT,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_public,
      anon_sym_protected,
      anon_sym_private,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67080] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3210), 1,
      anon_sym_ATinterface,
    STATE(1748), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3208), 17,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67116] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3214), 1,
      anon_sym_ATinterface,
    STATE(1749), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3212), 17,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67152] = 12,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(901), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3184), 1,
      sym_identifier,
    ACTIONS(3186), 1,
      anon_sym_def,
    STATE(1738), 1,
      aux_sym__param_list_repeat1,
    STATE(2047), 1,
      sym__type,
    STATE(2266), 1,
      sym_parameter,
    STATE(1750), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(87), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67200] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1755), 1,
      sym_modifier,
    ACTIONS(3216), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
    STATE(1751), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_class_definition_repeat2,
    ACTIONS(3171), 12,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67237] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3221), 1,
      anon_sym_int,
    STATE(1752), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3219), 16,
      anon_sym_ATinterface,
      anon_sym_interface,
      anon_sym_class,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67272] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3227), 1,
      anon_sym_RBRACK,
    STATE(1753), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67307] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3229), 1,
      anon_sym_RBRACK,
    STATE(1754), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67342] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1755), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3212), 15,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67373] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3231), 1,
      anon_sym_RBRACK,
    STATE(1756), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67408] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3233), 1,
      anon_sym_RBRACK,
    STATE(1757), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67443] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3235), 1,
      anon_sym_RBRACK,
    STATE(1758), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67478] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3237), 1,
      anon_sym_RBRACK,
    STATE(1759), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67513] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3239), 1,
      anon_sym_RBRACK,
    STATE(1760), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67548] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1761), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3208), 15,
      anon_sym__,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67579] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3241), 1,
      aux_sym__type_identifier_token1,
    STATE(1765), 1,
      aux_sym_generics_repeat1,
    STATE(2008), 1,
      sym__type,
    STATE(1762), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67618] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3243), 1,
      anon_sym_RBRACK,
    STATE(1763), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67653] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3245), 1,
      anon_sym_RBRACK,
    STATE(1764), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67688] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3241), 1,
      aux_sym__type_identifier_token1,
    STATE(1768), 1,
      aux_sym_generics_repeat1,
    STATE(1968), 1,
      sym__type,
    STATE(1765), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67727] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3247), 1,
      anon_sym_RBRACK,
    STATE(1766), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67762] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3249), 1,
      anon_sym_RBRACK,
    STATE(1767), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67797] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3251), 1,
      aux_sym__type_identifier_token1,
    STATE(2081), 1,
      sym__type,
    STATE(1768), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_generics_repeat1,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(3254), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67834] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3257), 1,
      anon_sym_RBRACK,
    STATE(1769), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67869] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1770), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3219), 14,
      anon_sym__,
      anon_sym_def,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [67899] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1771), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3223), 7,
      anon_sym_SLASH,
      sym_identifier,
      aux_sym__type_identifier_token1,
      aux_sym_number_literal_token1,
      aux_sym_number_literal_token4,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
    ACTIONS(3225), 7,
      anon_sym_LPAREN,
      aux_sym_number_literal_token2,
      aux_sym_number_literal_token3,
      aux_sym_number_literal_token5,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DOLLAR_SLASH,
  [67931] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3259), 1,
      aux_sym__type_identifier_token1,
    STATE(1939), 1,
      sym__type,
    STATE(1772), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [67967] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3241), 1,
      aux_sym__type_identifier_token1,
    STATE(1939), 1,
      sym__type,
    STATE(1773), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1935), 4,
      sym__type_identifier,
      sym_builtintype,
      sym_array_type,
      sym_type_with_generics,
    ACTIONS(2825), 8,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68003] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3263), 1,
      anon_sym_RPAREN,
    STATE(1774), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68033] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3265), 1,
      anon_sym_DASH_GT,
    STATE(1775), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68063] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3267), 1,
      anon_sym_RPAREN,
    STATE(1776), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68093] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3269), 1,
      anon_sym_RPAREN,
    STATE(1777), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68123] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3271), 1,
      anon_sym_RPAREN,
    STATE(1778), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68153] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3273), 1,
      anon_sym_DASH_GT,
    STATE(1779), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68183] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3275), 1,
      anon_sym_RPAREN,
    STATE(1780), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68213] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3277), 1,
      anon_sym_RPAREN,
    STATE(1781), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68243] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3279), 1,
      anon_sym_RPAREN,
    STATE(1782), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68273] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3281), 1,
      anon_sym_RPAREN,
    STATE(1783), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68303] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1784), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3261), 11,
      anon_sym_def,
      sym_identifier,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68330] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3283), 1,
      anon_sym_EQ,
    STATE(1800), 1,
      sym_parameter_list,
    STATE(1785), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1979), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68362] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3285), 1,
      anon_sym_EQ,
    STATE(1800), 1,
      sym_parameter_list,
    STATE(1786), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2007), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68394] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3287), 1,
      anon_sym_GT,
    STATE(1787), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3289), 9,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68422] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3291), 1,
      anon_sym_EQ,
    STATE(1801), 1,
      sym_parameter_list,
    STATE(1788), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1763), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68454] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3293), 1,
      anon_sym_EQ,
    STATE(1801), 1,
      sym_parameter_list,
    STATE(1789), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1753), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68486] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3295), 1,
      anon_sym_EQ,
    STATE(1796), 1,
      sym_parameter_list,
    STATE(1790), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1771), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68518] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3297), 1,
      anon_sym_EQ,
    STATE(1795), 1,
      sym_parameter_list,
    STATE(1791), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1741), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68550] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3299), 1,
      anon_sym_EQ,
    STATE(1795), 1,
      sym_parameter_list,
    STATE(1792), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1892), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68582] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3301), 1,
      anon_sym_GT,
    STATE(1793), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3289), 9,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68610] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1777), 1,
      anon_sym_LPAREN,
    ACTIONS(3303), 1,
      anon_sym_EQ,
    STATE(1796), 1,
      sym_parameter_list,
    STATE(1794), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1790), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68642] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(702), 1,
      sym_closure,
    STATE(1795), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2053), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68671] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(687), 1,
      sym_closure,
    STATE(1796), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2049), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68700] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1797), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1757), 9,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68725] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3312), 1,
      anon_sym_STAR_SLASH,
    ACTIONS(3314), 1,
      anon_sym_ATparam,
    ACTIONS(3317), 1,
      anon_sym_ATthrows,
    STATE(1898), 2,
      sym_groovy_doc_param,
      sym_groovy_doc_throws,
    ACTIONS(3309), 3,
      aux_sym_groovy_doc_token3,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
    STATE(1798), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_groovy_doc_repeat1,
  [68758] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1799), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3289), 9,
      aux_sym__type_identifier_token1,
      anon_sym_int,
      anon_sym_boolean,
      anon_sym_char,
      anon_sym_short,
      anon_sym_long,
      anon_sym_float,
      anon_sym_double,
      anon_sym_void,
  [68783] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(724), 1,
      sym_closure,
    STATE(1800), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2057), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68812] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(738), 1,
      sym_closure,
    STATE(1801), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2041), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68841] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3320), 1,
      anon_sym_DOT,
    ACTIONS(3322), 1,
      anon_sym_as,
    STATE(1802), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1812), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68870] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3326), 1,
      anon_sym_STAR_SLASH,
    ACTIONS(3328), 1,
      anon_sym_ATparam,
    ACTIONS(3330), 1,
      anon_sym_ATthrows,
    STATE(1798), 1,
      aux_sym_groovy_doc_repeat1,
    STATE(1803), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1898), 2,
      sym_groovy_doc_param,
      sym_groovy_doc_throws,
    ACTIONS(3324), 3,
      aux_sym_groovy_doc_token3,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
  [68905] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3332), 1,
      anon_sym_DOT,
    ACTIONS(3334), 1,
      anon_sym_as,
    STATE(1804), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1868), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68934] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1805), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1779), 9,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [68959] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3328), 1,
      anon_sym_ATparam,
    ACTIONS(3330), 1,
      anon_sym_ATthrows,
    ACTIONS(3336), 1,
      anon_sym_STAR_SLASH,
    STATE(1803), 1,
      aux_sym_groovy_doc_repeat1,
    STATE(1806), 2,
      sym_comment,
      sym_groovy_doc,
    STATE(1898), 2,
      sym_groovy_doc_param,
      sym_groovy_doc_throws,
    ACTIONS(3324), 3,
      aux_sym_groovy_doc_token3,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
  [68994] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3283), 1,
      anon_sym_EQ,
    STATE(1841), 1,
      sym_parameter_list,
    STATE(1807), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1979), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69024] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3285), 1,
      anon_sym_EQ,
    STATE(1841), 1,
      sym_parameter_list,
    STATE(1808), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2007), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69054] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1805), 1,
      sym__type_identifier,
    STATE(1856), 1,
      sym__import_name,
    STATE(2107), 1,
      sym_modifier,
    STATE(1809), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69088] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3291), 1,
      anon_sym_EQ,
    STATE(1862), 1,
      sym_parameter_list,
    STATE(1810), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1763), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69118] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3293), 1,
      anon_sym_EQ,
    STATE(1862), 1,
      sym_parameter_list,
    STATE(1811), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1753), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69148] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3342), 1,
      anon_sym_DOT,
    STATE(1812), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2035), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69174] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3297), 1,
      anon_sym_EQ,
    STATE(1858), 1,
      sym_parameter_list,
    STATE(1813), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1741), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69204] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3344), 1,
      sym_identifier,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    STATE(691), 1,
      sym__import_name,
    STATE(719), 1,
      sym__type_identifier,
    STATE(2087), 1,
      sym_modifier,
    STATE(1814), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69238] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3295), 1,
      anon_sym_EQ,
    STATE(1828), 1,
      sym_parameter_list,
    STATE(1815), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1771), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69268] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3297), 1,
      anon_sym_EQ,
    STATE(1816), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1741), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69294] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1802), 1,
      sym__import_name,
    STATE(1805), 1,
      sym__type_identifier,
    STATE(2068), 1,
      sym_modifier,
    STATE(1817), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69328] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3299), 1,
      anon_sym_EQ,
    STATE(1858), 1,
      sym_parameter_list,
    STATE(1818), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1892), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69358] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3283), 1,
      anon_sym_EQ,
    STATE(1819), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1979), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69384] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3348), 1,
      sym_identifier,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    STATE(710), 1,
      sym__type_identifier,
    STATE(713), 1,
      sym__import_name,
    STATE(2084), 1,
      sym_modifier,
    STATE(1820), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69418] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3352), 1,
      anon_sym_COMMA,
    STATE(1821), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__juxt_argument_list_repeat1,
    ACTIONS(1268), 6,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69444] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3295), 1,
      anon_sym_EQ,
    STATE(1822), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1771), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69470] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3355), 1,
      sym_identifier,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    STATE(683), 1,
      sym__type_identifier,
    STATE(850), 1,
      sym__import_name,
    STATE(2105), 1,
      sym_modifier,
    STATE(1823), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69504] = 10,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3359), 1,
      sym_identifier,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    STATE(799), 1,
      sym__type_identifier,
    STATE(894), 1,
      sym__import_name,
    STATE(2088), 1,
      sym_modifier,
    STATE(1824), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3340), 3,
      anon_sym_static,
      anon_sym_final,
      anon_sym_synchronized,
  [69538] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2081), 1,
      anon_sym_LPAREN,
    ACTIONS(3303), 1,
      anon_sym_EQ,
    STATE(1828), 1,
      sym_parameter_list,
    STATE(1825), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1790), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69568] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3363), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    STATE(1851), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1826), 2,
      sym_comment,
      sym_groovy_doc,
  [69603] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3373), 1,
      anon_sym_SLASH,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1827), 2,
      sym_comment,
      sym_groovy_doc,
  [69636] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(796), 1,
      sym_closure,
    STATE(1828), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2049), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69663] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1837), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1829), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69690] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3381), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1830), 2,
      sym_comment,
      sym_groovy_doc,
  [69725] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3383), 1,
      anon_sym_catch,
    ACTIONS(3385), 1,
      anon_sym_finally,
    STATE(1831), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1830), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [69752] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3387), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1844), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1832), 2,
      sym_comment,
      sym_groovy_doc,
  [69787] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3373), 1,
      anon_sym_SLASH_DOLLAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1833), 2,
      sym_comment,
      sym_groovy_doc,
  [69820] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1869), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1834), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(373), 6,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [69845] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3387), 1,
      anon_sym_SLASH_DOLLAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    STATE(1833), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1835), 2,
      sym_comment,
      sym_groovy_doc,
  [69878] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1848), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1836), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69903] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2453), 1,
      anon_sym_COMMA,
    STATE(1821), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1837), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1898), 5,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [69930] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3395), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1849), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1838), 2,
      sym_comment,
      sym_groovy_doc,
  [69965] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3397), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1842), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1839), 2,
      sym_comment,
      sym_groovy_doc,
  [70000] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3395), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1872), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1840), 2,
      sym_comment,
      sym_groovy_doc,
  [70033] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(823), 1,
      sym_closure,
    STATE(1841), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2057), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [70060] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3399), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1842), 2,
      sym_comment,
      sym_groovy_doc,
  [70095] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3397), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1845), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1843), 2,
      sym_comment,
      sym_groovy_doc,
  [70128] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3373), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1844), 2,
      sym_comment,
      sym_groovy_doc,
  [70163] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3399), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1845), 2,
      sym_comment,
      sym_groovy_doc,
  [70196] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3401), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(3403), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3406), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3409), 1,
      sym_escape_sequence,
    ACTIONS(3412), 1,
      anon_sym_DOLLAR,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1846), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__interpolate_string_repeat2,
  [70229] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3418), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3421), 1,
      anon_sym_SLASH_DOLLAR,
    ACTIONS(3423), 1,
      anon_sym_DOLLAR,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3415), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1847), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__interpolate_string_repeat4,
  [70260] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1821), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1848), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1898), 6,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
      anon_sym_catch,
      anon_sym_finally,
  [70285] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3426), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1849), 2,
      sym_comment,
      sym_groovy_doc,
  [70320] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3428), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1850), 2,
      sym_comment,
      sym_groovy_doc,
  [70355] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3430), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1846), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1851), 2,
      sym_comment,
      sym_groovy_doc,
  [70390] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3432), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1830), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1852), 2,
      sym_comment,
      sym_groovy_doc,
  [70425] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3432), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1864), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1853), 2,
      sym_comment,
      sym_groovy_doc,
  [70458] = 11,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3365), 1,
      aux_sym__interpolate_string_token2,
    ACTIONS(3367), 1,
      aux_sym__interpolate_string_token3,
    ACTIONS(3369), 1,
      sym_escape_sequence,
    ACTIONS(3371), 1,
      anon_sym_DOLLAR,
    ACTIONS(3434), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(1850), 1,
      aux_sym__interpolate_string_repeat2,
    STATE(1997), 1,
      sym_interpolation,
    STATE(1854), 2,
      sym_comment,
      sym_groovy_doc,
  [70493] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3434), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1857), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1855), 2,
      sym_comment,
      sym_groovy_doc,
  [70526] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3436), 1,
      anon_sym_DOT,
    ACTIONS(3438), 1,
      anon_sym_as,
    STATE(1856), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1812), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [70553] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3428), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1857), 2,
      sym_comment,
      sym_groovy_doc,
  [70586] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(834), 1,
      sym_closure,
    STATE(1858), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2053), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [70613] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    ACTIONS(3381), 1,
      anon_sym_SLASH,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1859), 2,
      sym_comment,
      sym_groovy_doc,
  [70646] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1860), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(97), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [70673] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    ACTIONS(3426), 1,
      anon_sym_SLASH,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1861), 2,
      sym_comment,
      sym_groovy_doc,
  [70706] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(851), 1,
      sym_closure,
    STATE(1862), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2041), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [70733] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3430), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1863), 2,
      sym_comment,
      sym_groovy_doc,
  [70766] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3381), 1,
      anon_sym_SLASH_DOLLAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1864), 2,
      sym_comment,
      sym_groovy_doc,
  [70799] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    ACTIONS(3399), 1,
      anon_sym_SLASH,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1865), 2,
      sym_comment,
      sym_groovy_doc,
  [70832] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3442), 1,
      anon_sym_DOT,
    ACTIONS(3444), 1,
      anon_sym_as,
    STATE(1866), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1868), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [70859] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3363), 1,
      anon_sym_SLASH_DOLLAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    STATE(1863), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1867), 2,
      sym_comment,
      sym_groovy_doc,
  [70892] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3446), 1,
      anon_sym_SLASH,
    ACTIONS(3448), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3454), 1,
      anon_sym_DOLLAR,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3451), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1868), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__interpolate_string_repeat3,
  [70923] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3457), 1,
      anon_sym_DOT,
    STATE(1869), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(377), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [70948] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    ACTIONS(3430), 1,
      anon_sym_SLASH,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1870), 2,
      sym_comment,
      sym_groovy_doc,
  [70981] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    ACTIONS(3428), 1,
      anon_sym_SLASH,
    STATE(1868), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1871), 2,
      sym_comment,
      sym_groovy_doc,
  [71014] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3391), 1,
      aux_sym__interpolate_string_token6,
    ACTIONS(3393), 1,
      anon_sym_DOLLAR,
    ACTIONS(3426), 1,
      anon_sym_SLASH_DOLLAR,
    STATE(1847), 1,
      aux_sym__interpolate_string_repeat4,
    STATE(1955), 1,
      sym_interpolation,
    ACTIONS(3389), 2,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
    STATE(1872), 2,
      sym_comment,
      sym_groovy_doc,
  [71047] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3363), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1877), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1873), 2,
      sym_comment,
      sym_groovy_doc,
  [71079] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3373), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1874), 2,
      sym_comment,
      sym_groovy_doc,
  [71111] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3428), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1875), 2,
      sym_comment,
      sym_groovy_doc,
  [71143] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1871), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1876), 2,
      sym_comment,
      sym_groovy_doc,
  [71173] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3430), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1877), 2,
      sym_comment,
      sym_groovy_doc,
  [71205] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3468), 1,
      anon_sym_STAR_SLASH,
    STATE(1878), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3466), 5,
      aux_sym_groovy_doc_token3,
      anon_sym_ATparam,
      anon_sym_ATthrows,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
  [71229] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3470), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1879), 2,
      sym_comment,
      sym_groovy_doc,
  [71261] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3395), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1899), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1880), 2,
      sym_comment,
      sym_groovy_doc,
  [71293] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3480), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1897), 1,
      aux_sym__plain_string_repeat2,
    STATE(1881), 2,
      sym_comment,
      sym_groovy_doc,
  [71325] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3432), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1900), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1882), 2,
      sym_comment,
      sym_groovy_doc,
  [71357] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3482), 1,
      anon_sym_DOT,
    STATE(1717), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1883), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(97), 4,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [71383] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3484), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1884), 2,
      sym_comment,
      sym_groovy_doc,
  [71415] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3283), 1,
      anon_sym_EQ,
    STATE(1885), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1979), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [71439] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1861), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1886), 2,
      sym_comment,
      sym_groovy_doc,
  [71469] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3486), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1887), 2,
      sym_comment,
      sym_groovy_doc,
  [71501] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3297), 1,
      anon_sym_EQ,
    STATE(1888), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1741), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [71525] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1827), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1889), 2,
      sym_comment,
      sym_groovy_doc,
  [71555] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1865), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1890), 2,
      sym_comment,
      sym_groovy_doc,
  [71585] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3488), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1909), 1,
      aux_sym__plain_string_repeat2,
    STATE(1891), 2,
      sym_comment,
      sym_groovy_doc,
  [71617] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3490), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(3492), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3495), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3498), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3501), 1,
      sym_escape_sequence,
    STATE(1892), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__plain_string_repeat2,
  [71647] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3504), 1,
      anon_sym_DQUOTE,
    ACTIONS(3506), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3509), 1,
      sym_escape_sequence,
    ACTIONS(3512), 1,
      anon_sym_DOLLAR,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1893), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__interpolate_string_repeat1,
  [71677] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3515), 1,
      anon_sym_COMMA,
    STATE(1894), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__juxt_argument_list_repeat1,
    ACTIONS(1268), 4,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [71701] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3520), 1,
      anon_sym_STAR_SLASH,
    STATE(1895), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3518), 5,
      aux_sym_groovy_doc_token3,
      anon_sym_ATparam,
      anon_sym_ATthrows,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
  [71725] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3522), 1,
      anon_sym_else,
    STATE(1896), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1904), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_catch,
      anon_sym_finally,
  [71749] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3524), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1897), 2,
      sym_comment,
      sym_groovy_doc,
  [71781] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3528), 1,
      anon_sym_STAR_SLASH,
    STATE(1898), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3526), 5,
      aux_sym_groovy_doc_token3,
      anon_sym_ATparam,
      anon_sym_ATthrows,
      sym_groovy_doc_tag,
      sym_groovy_doc_at_text,
  [71805] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3426), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1899), 2,
      sym_comment,
      sym_groovy_doc,
  [71837] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3381), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1900), 2,
      sym_comment,
      sym_groovy_doc,
  [71869] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1859), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1901), 2,
      sym_comment,
      sym_groovy_doc,
  [71899] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3399), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1893), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1902), 2,
      sym_comment,
      sym_groovy_doc,
  [71931] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3530), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1908), 1,
      aux_sym__plain_string_repeat2,
    STATE(1903), 2,
      sym_comment,
      sym_groovy_doc,
  [71963] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3295), 1,
      anon_sym_EQ,
    STATE(1904), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1771), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [71987] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3532), 1,
      anon_sym_finally,
    STATE(1905), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1936), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [72011] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3534), 1,
      anon_sym_DOT,
    STATE(1906), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2035), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_while,
      anon_sym_else,
  [72035] = 9,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3375), 1,
      aux_sym__interpolate_string_token4,
    ACTIONS(3379), 1,
      anon_sym_DOLLAR,
    STATE(1870), 1,
      aux_sym__interpolate_string_repeat3,
    STATE(1927), 1,
      sym_interpolation,
    ACTIONS(3377), 2,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
    STATE(1907), 2,
      sym_comment,
      sym_groovy_doc,
  [72065] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3536), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1908), 2,
      sym_comment,
      sym_groovy_doc,
  [72097] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3538), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1892), 1,
      aux_sym__plain_string_repeat2,
    STATE(1909), 2,
      sym_comment,
      sym_groovy_doc,
  [72129] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3540), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1879), 1,
      aux_sym__plain_string_repeat2,
    STATE(1910), 2,
      sym_comment,
      sym_groovy_doc,
  [72161] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3387), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1874), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1911), 2,
      sym_comment,
      sym_groovy_doc,
  [72193] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3542), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1887), 1,
      aux_sym__plain_string_repeat2,
    STATE(1912), 2,
      sym_comment,
      sym_groovy_doc,
  [72225] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3397), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1902), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1913), 2,
      sym_comment,
      sym_groovy_doc,
  [72257] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3472), 1,
      aux_sym__plain_string_token2,
    ACTIONS(3474), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3476), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3478), 1,
      sym_escape_sequence,
    ACTIONS(3544), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    STATE(1884), 1,
      aux_sym__plain_string_repeat2,
    STATE(1914), 2,
      sym_comment,
      sym_groovy_doc,
  [72289] = 10,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3434), 1,
      anon_sym_DQUOTE,
    ACTIONS(3460), 1,
      aux_sym__interpolate_string_token1,
    ACTIONS(3462), 1,
      sym_escape_sequence,
    ACTIONS(3464), 1,
      anon_sym_DOLLAR,
    STATE(1875), 1,
      aux_sym__interpolate_string_repeat1,
    STATE(2099), 1,
      sym_interpolation,
    STATE(1915), 2,
      sym_comment,
      sym_groovy_doc,
  [72321] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(3546), 1,
      sym_identifier,
    ACTIONS(3548), 1,
      aux_sym__type_identifier_token1,
    STATE(28), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1916), 2,
      sym_comment,
      sym_groovy_doc,
  [72348] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1917), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3550), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [72369] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1111), 1,
      anon_sym_LPAREN,
    ACTIONS(3552), 1,
      sym_identifier,
    ACTIONS(3554), 1,
      aux_sym__type_identifier_token1,
    STATE(16), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1918), 2,
      sym_comment,
      sym_groovy_doc,
  [72396] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3558), 1,
      anon_sym_extends,
    STATE(904), 1,
      sym_closure,
    STATE(2143), 1,
      sym_generic_parameters,
    STATE(1919), 2,
      sym_comment,
      sym_groovy_doc,
  [72425] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(802), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1920), 2,
      sym_comment,
      sym_groovy_doc,
  [72454] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1894), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1921), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1898), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
  [72479] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3560), 1,
      anon_sym_extends,
    STATE(950), 1,
      sym_closure,
    STATE(2127), 1,
      sym_generic_parameters,
    STATE(1922), 2,
      sym_comment,
      sym_groovy_doc,
  [72508] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(939), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1923), 2,
      sym_comment,
      sym_groovy_doc,
  [72537] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(778), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1924), 2,
      sym_comment,
      sym_groovy_doc,
  [72566] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3562), 1,
      anon_sym_extends,
    STATE(954), 1,
      sym_closure,
    STATE(2169), 1,
      sym_generic_parameters,
    STATE(1925), 2,
      sym_comment,
      sym_groovy_doc,
  [72595] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(957), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1926), 2,
      sym_comment,
      sym_groovy_doc,
  [72624] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3566), 1,
      aux_sym__interpolate_string_token4,
    STATE(1927), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3564), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
      anon_sym_DOLLAR,
  [72647] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(965), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1928), 2,
      sym_comment,
      sym_groovy_doc,
  [72676] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3568), 1,
      anon_sym_extends,
    STATE(974), 1,
      sym_closure,
    STATE(2195), 1,
      sym_generic_parameters,
    STATE(1929), 2,
      sym_comment,
      sym_groovy_doc,
  [72705] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(987), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1930), 2,
      sym_comment,
      sym_groovy_doc,
  [72734] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(993), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1931), 2,
      sym_comment,
      sym_groovy_doc,
  [72763] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(998), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1932), 2,
      sym_comment,
      sym_groovy_doc,
  [72792] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(970), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1933), 2,
      sym_comment,
      sym_groovy_doc,
  [72821] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(984), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1934), 2,
      sym_comment,
      sym_groovy_doc,
  [72850] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1935), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1290), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [72871] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3570), 1,
      anon_sym_RBRACE,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    STATE(2019), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(1936), 2,
      sym_comment,
      sym_groovy_doc,
  [72900] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3576), 1,
      anon_sym_extends,
    STATE(826), 1,
      sym_closure,
    STATE(2130), 1,
      sym_generic_parameters,
    STATE(1937), 2,
      sym_comment,
      sym_groovy_doc,
  [72929] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3578), 1,
      anon_sym_extends,
    STATE(798), 1,
      sym_closure,
    STATE(2184), 1,
      sym_generic_parameters,
    STATE(1938), 2,
      sym_comment,
      sym_groovy_doc,
  [72958] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    STATE(2007), 1,
      sym_generics,
    ACTIONS(3582), 2,
      anon_sym_GT,
      anon_sym_COMMA,
    STATE(1939), 2,
      sym_comment,
      sym_groovy_doc,
  [72985] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3586), 1,
      anon_sym_extends,
    STATE(812), 1,
      sym_closure,
    STATE(2202), 1,
      sym_generic_parameters,
    STATE(1940), 2,
      sym_comment,
      sym_groovy_doc,
  [73014] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(802), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1941), 2,
      sym_comment,
      sym_groovy_doc,
  [73043] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(818), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1942), 2,
      sym_comment,
      sym_groovy_doc,
  [73072] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2789), 1,
      anon_sym_LPAREN,
    ACTIONS(2793), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3588), 1,
      sym_identifier,
    STATE(1084), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1943), 2,
      sym_comment,
      sym_groovy_doc,
  [73099] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3590), 1,
      anon_sym_extends,
    STATE(826), 1,
      sym_closure,
    STATE(2210), 1,
      sym_generic_parameters,
    STATE(1944), 2,
      sym_comment,
      sym_groovy_doc,
  [73128] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(845), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1945), 2,
      sym_comment,
      sym_groovy_doc,
  [73157] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(846), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1946), 2,
      sym_comment,
      sym_groovy_doc,
  [73186] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3592), 1,
      anon_sym_extends,
    STATE(854), 1,
      sym_closure,
    STATE(2180), 1,
      sym_generic_parameters,
    STATE(1947), 2,
      sym_comment,
      sym_groovy_doc,
  [73215] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(859), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1948), 2,
      sym_comment,
      sym_groovy_doc,
  [73244] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3596), 1,
      aux_sym__interpolate_string_token6,
    STATE(1949), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3594), 4,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_SLASH_DOLLAR,
      anon_sym_DOLLAR,
  [73267] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(853), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1950), 2,
      sym_comment,
      sym_groovy_doc,
  [73296] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(855), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1951), 2,
      sym_comment,
      sym_groovy_doc,
  [73325] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(864), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1952), 2,
      sym_comment,
      sym_groovy_doc,
  [73354] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3598), 1,
      anon_sym_extends,
    STATE(866), 1,
      sym_closure,
    STATE(2138), 1,
      sym_generic_parameters,
    STATE(1953), 2,
      sym_comment,
      sym_groovy_doc,
  [73383] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(869), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1954), 2,
      sym_comment,
      sym_groovy_doc,
  [73412] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3602), 1,
      aux_sym__interpolate_string_token6,
    STATE(1955), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3600), 4,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_SLASH_DOLLAR,
      anon_sym_DOLLAR,
  [73435] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3606), 1,
      aux_sym__interpolate_string_token4,
    STATE(1956), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3604), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
      anon_sym_DOLLAR,
  [73458] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(872), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1957), 2,
      sym_comment,
      sym_groovy_doc,
  [73487] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3608), 1,
      anon_sym_extends,
    STATE(875), 1,
      sym_closure,
    STATE(2165), 1,
      sym_generic_parameters,
    STATE(1958), 2,
      sym_comment,
      sym_groovy_doc,
  [73516] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(881), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1959), 2,
      sym_comment,
      sym_groovy_doc,
  [73545] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(882), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1960), 2,
      sym_comment,
      sym_groovy_doc,
  [73574] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(886), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1961), 2,
      sym_comment,
      sym_groovy_doc,
  [73603] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(887), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1962), 2,
      sym_comment,
      sym_groovy_doc,
  [73632] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(889), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1963), 2,
      sym_comment,
      sym_groovy_doc,
  [73661] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3612), 1,
      aux_sym__plain_string_token3,
    STATE(1964), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3610), 4,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      aux_sym__plain_string_token2,
      aux_sym__plain_string_token4,
      sym_escape_sequence,
  [73684] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1702), 1,
      anon_sym_LPAREN,
    ACTIONS(1704), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3614), 1,
      sym_identifier,
    STATE(351), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1965), 2,
      sym_comment,
      sym_groovy_doc,
  [73711] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1966), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3616), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [73732] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3618), 1,
      anon_sym_else,
    STATE(1967), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1904), 4,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_catch,
      anon_sym_finally,
  [73755] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3301), 1,
      anon_sym_GT,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3620), 1,
      anon_sym_COMMA,
    STATE(2007), 1,
      sym_generics,
    STATE(1968), 2,
      sym_comment,
      sym_groovy_doc,
  [73784] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3546), 1,
      sym_identifier,
    STATE(28), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1969), 2,
      sym_comment,
      sym_groovy_doc,
  [73811] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3622), 1,
      anon_sym_extends,
    STATE(782), 1,
      sym_closure,
    STATE(2121), 1,
      sym_generic_parameters,
    STATE(1970), 2,
      sym_comment,
      sym_groovy_doc,
  [73840] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(864), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1971), 2,
      sym_comment,
      sym_groovy_doc,
  [73869] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2781), 1,
      anon_sym_LPAREN,
    ACTIONS(3624), 1,
      sym_identifier,
    ACTIONS(3626), 1,
      aux_sym__type_identifier_token1,
    STATE(1723), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1972), 2,
      sym_comment,
      sym_groovy_doc,
  [73896] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3606), 1,
      aux_sym__interpolate_string_token6,
    STATE(1973), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3604), 4,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_SLASH_DOLLAR,
      anon_sym_DOLLAR,
  [73919] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(616), 1,
      anon_sym_LPAREN,
    ACTIONS(3628), 1,
      sym_identifier,
    ACTIONS(3630), 1,
      aux_sym__type_identifier_token1,
    STATE(38), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1974), 2,
      sym_comment,
      sym_groovy_doc,
  [73946] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1975), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3632), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [73967] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1723), 1,
      anon_sym_LPAREN,
    ACTIONS(1725), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3634), 1,
      sym_identifier,
    STATE(482), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1976), 2,
      sym_comment,
      sym_groovy_doc,
  [73994] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(845), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1977), 2,
      sym_comment,
      sym_groovy_doc,
  [74023] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(846), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1978), 2,
      sym_comment,
      sym_groovy_doc,
  [74052] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3638), 1,
      aux_sym__interpolate_string_token2,
    STATE(1979), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3636), 4,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      aux_sym__interpolate_string_token3,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [74075] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3640), 1,
      anon_sym_RBRACE,
    STATE(1936), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(1980), 2,
      sym_comment,
      sym_groovy_doc,
  [74104] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1992), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1981), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
  [74127] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3642), 1,
      anon_sym_extends,
    STATE(798), 1,
      sym_closure,
    STATE(2174), 1,
      sym_generic_parameters,
    STATE(1982), 2,
      sym_comment,
      sym_groovy_doc,
  [74156] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1708), 1,
      anon_sym_LPAREN,
    ACTIONS(1710), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3644), 1,
      sym_identifier,
    STATE(362), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1983), 2,
      sym_comment,
      sym_groovy_doc,
  [74183] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3646), 1,
      anon_sym_extends,
    STATE(684), 1,
      sym_closure,
    STATE(2173), 1,
      sym_generic_parameters,
    STATE(1984), 2,
      sym_comment,
      sym_groovy_doc,
  [74212] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    ACTIONS(3648), 1,
      sym_identifier,
    STATE(1057), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(1985), 2,
      sym_comment,
      sym_groovy_doc,
  [74239] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3482), 1,
      anon_sym_DOT,
    ACTIONS(3650), 1,
      anon_sym_LBRACK,
    STATE(1717), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1436), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1986), 2,
      sym_comment,
      sym_groovy_doc,
  [74266] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1987), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3652), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [74287] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3606), 1,
      aux_sym__interpolate_string_token2,
    STATE(1988), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3604), 4,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      aux_sym__interpolate_string_token3,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [74310] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3654), 1,
      anon_sym_RBRACE,
    STATE(1993), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(1989), 2,
      sym_comment,
      sym_groovy_doc,
  [74339] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3656), 1,
      anon_sym_extends,
    STATE(734), 1,
      sym_closure,
    STATE(2136), 1,
      sym_generic_parameters,
    STATE(1990), 2,
      sym_comment,
      sym_groovy_doc,
  [74368] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(705), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1991), 2,
      sym_comment,
      sym_groovy_doc,
  [74397] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(1894), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(1992), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1898), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_else,
  [74420] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3658), 1,
      anon_sym_RBRACE,
    STATE(2019), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(1993), 2,
      sym_comment,
      sym_groovy_doc,
  [74449] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3660), 1,
      anon_sym_extends,
    STATE(688), 1,
      sym_closure,
    STATE(2170), 1,
      sym_generic_parameters,
    STATE(1994), 2,
      sym_comment,
      sym_groovy_doc,
  [74478] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(698), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1995), 2,
      sym_comment,
      sym_groovy_doc,
  [74507] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3662), 1,
      anon_sym_RBRACE,
    STATE(2019), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(1996), 2,
      sym_comment,
      sym_groovy_doc,
  [74536] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3666), 1,
      aux_sym__interpolate_string_token2,
    STATE(1997), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3664), 4,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      aux_sym__interpolate_string_token3,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [74559] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(721), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(1998), 2,
      sym_comment,
      sym_groovy_doc,
  [74588] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3668), 1,
      anon_sym_extends,
    STATE(725), 1,
      sym_closure,
    STATE(2179), 1,
      sym_generic_parameters,
    STATE(1999), 2,
      sym_comment,
      sym_groovy_doc,
  [74617] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(730), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2000), 2,
      sym_comment,
      sym_groovy_doc,
  [74646] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(732), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2001), 2,
      sym_comment,
      sym_groovy_doc,
  [74675] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3672), 1,
      aux_sym__plain_string_token3,
    STATE(2002), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3670), 4,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      aux_sym__plain_string_token2,
      aux_sym__plain_string_token4,
      sym_escape_sequence,
  [74698] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(739), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2003), 2,
      sym_comment,
      sym_groovy_doc,
  [74727] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(743), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2004), 2,
      sym_comment,
      sym_groovy_doc,
  [74756] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(748), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2005), 2,
      sym_comment,
      sym_groovy_doc,
  [74785] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3674), 1,
      aux_sym__plain_string_token3,
    STATE(2006), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3490), 4,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      aux_sym__plain_string_token2,
      aux_sym__plain_string_token4,
      sym_escape_sequence,
  [74808] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(2007), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3676), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [74829] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3678), 1,
      anon_sym_GT,
    ACTIONS(3680), 1,
      anon_sym_COMMA,
    STATE(2007), 1,
      sym_generics,
    STATE(2008), 2,
      sym_comment,
      sym_groovy_doc,
  [74858] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3682), 1,
      anon_sym_extends,
    STATE(904), 1,
      sym_closure,
    STATE(2171), 1,
      sym_generic_parameters,
    STATE(2009), 2,
      sym_comment,
      sym_groovy_doc,
  [74887] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3684), 1,
      anon_sym_extends,
    STATE(684), 1,
      sym_closure,
    STATE(2154), 1,
      sym_generic_parameters,
    STATE(2010), 2,
      sym_comment,
      sym_groovy_doc,
  [74916] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    ACTIONS(1278), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(2011), 2,
      sym_comment,
      sym_groovy_doc,
  [74943] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3686), 1,
      anon_sym_extends,
    STATE(782), 1,
      sym_closure,
    STATE(2201), 1,
      sym_generic_parameters,
    STATE(2012), 2,
      sym_comment,
      sym_groovy_doc,
  [74972] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3688), 1,
      sym_identifier,
    ACTIONS(3690), 1,
      anon_sym_LPAREN,
    ACTIONS(3692), 1,
      aux_sym__type_identifier_token1,
    STATE(668), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(2013), 2,
      sym_comment,
      sym_groovy_doc,
  [74999] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3694), 1,
      aux_sym__interpolate_string_token2,
    STATE(2014), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3401), 4,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      aux_sym__interpolate_string_token3,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [75022] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3696), 1,
      anon_sym_RBRACE,
    STATE(2020), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(2015), 2,
      sym_comment,
      sym_groovy_doc,
  [75051] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3638), 1,
      aux_sym__interpolate_string_token6,
    STATE(2016), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3636), 4,
      anon_sym_DOLLAR_SLASH,
      anon_sym_DOLLAR_DOLLAR,
      anon_sym_SLASH_DOLLAR,
      anon_sym_DOLLAR,
  [75074] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3698), 1,
      anon_sym_extends,
    STATE(734), 1,
      sym_closure,
    STATE(2178), 1,
      sym_generic_parameters,
    STATE(2017), 2,
      sym_comment,
      sym_groovy_doc,
  [75103] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(705), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2018), 2,
      sym_comment,
      sym_groovy_doc,
  [75132] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3700), 1,
      anon_sym_RBRACE,
    ACTIONS(3702), 1,
      anon_sym_case,
    ACTIONS(3705), 1,
      anon_sym_default,
    STATE(2147), 1,
      sym_case,
    STATE(2019), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_switch_block_repeat1,
  [75159] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3708), 1,
      anon_sym_RBRACE,
    STATE(2019), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(2020), 2,
      sym_comment,
      sym_groovy_doc,
  [75188] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1714), 1,
      anon_sym_LPAREN,
    ACTIONS(1716), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3710), 1,
      sym_identifier,
    STATE(424), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(2021), 2,
      sym_comment,
      sym_groovy_doc,
  [75215] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3712), 1,
      anon_sym_extends,
    STATE(688), 1,
      sym_closure,
    STATE(2185), 1,
      sym_generic_parameters,
    STATE(2022), 2,
      sym_comment,
      sym_groovy_doc,
  [75244] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(698), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2023), 2,
      sym_comment,
      sym_groovy_doc,
  [75273] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2533), 1,
      anon_sym_COMMA,
    STATE(1921), 1,
      aux_sym__juxt_argument_list_repeat1,
    STATE(2024), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1252), 3,
      anon_sym_SEMI,
      anon_sym_while,
      anon_sym_else,
  [75298] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(721), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2025), 2,
      sym_comment,
      sym_groovy_doc,
  [75327] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3714), 1,
      anon_sym_extends,
    STATE(725), 1,
      sym_closure,
    STATE(2189), 1,
      sym_generic_parameters,
    STATE(2026), 2,
      sym_comment,
      sym_groovy_doc,
  [75356] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(730), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2027), 2,
      sym_comment,
      sym_groovy_doc,
  [75385] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(732), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2028), 2,
      sym_comment,
      sym_groovy_doc,
  [75414] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(739), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2029), 2,
      sym_comment,
      sym_groovy_doc,
  [75443] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(743), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2030), 2,
      sym_comment,
      sym_groovy_doc,
  [75472] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(748), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2031), 2,
      sym_comment,
      sym_groovy_doc,
  [75501] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(778), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2032), 2,
      sym_comment,
      sym_groovy_doc,
  [75530] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3638), 1,
      aux_sym__interpolate_string_token4,
    STATE(2033), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3636), 4,
      anon_sym_SLASH,
      anon_sym_BSLASH_SLASH,
      aux_sym__interpolate_string_token5,
      anon_sym_DOLLAR,
  [75553] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(818), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2034), 2,
      sym_comment,
      sym_groovy_doc,
  [75582] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(853), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2035), 2,
      sym_comment,
      sym_groovy_doc,
  [75611] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3572), 1,
      anon_sym_case,
    ACTIONS(3574), 1,
      anon_sym_default,
    ACTIONS(3716), 1,
      anon_sym_RBRACE,
    STATE(1996), 1,
      aux_sym_switch_block_repeat1,
    STATE(2147), 1,
      sym_case,
    STATE(2036), 2,
      sym_comment,
      sym_groovy_doc,
  [75640] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1648), 1,
      anon_sym_LBRACK,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3440), 1,
      anon_sym_DOT,
    STATE(855), 1,
      sym_closure,
    STATE(1834), 1,
      aux_sym_dotted_identifier_repeat1,
    STATE(2037), 2,
      sym_comment,
      sym_groovy_doc,
  [75669] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(2038), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3718), 5,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      sym_identifier,
      anon_sym_LBRACK_RBRACK,
  [75690] = 9,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3556), 1,
      anon_sym_LT,
    ACTIONS(3720), 1,
      anon_sym_extends,
    STATE(975), 1,
      sym_closure,
    STATE(2188), 1,
      sym_generic_parameters,
    STATE(2039), 2,
      sym_comment,
      sym_groovy_doc,
  [75719] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1025), 1,
      anon_sym_LPAREN,
    ACTIONS(3722), 1,
      sym_identifier,
    ACTIONS(3724), 1,
      aux_sym__type_identifier_token1,
    STATE(20), 2,
      sym_parenthesized_expression,
      sym__type_identifier,
    STATE(2040), 2,
      sym_comment,
      sym_groovy_doc,
  [75746] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3299), 1,
      anon_sym_EQ,
    STATE(2041), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1892), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [75768] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3484), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2042), 2,
      sym_comment,
      sym_groovy_doc,
  [75794] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3730), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2043), 2,
      sym_comment,
      sym_groovy_doc,
  [75820] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3732), 1,
      anon_sym_EQ,
    STATE(2044), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(401), 3,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RPAREN,
  [75842] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3734), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2045), 2,
      sym_comment,
      sym_groovy_doc,
  [75868] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3536), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2046), 2,
      sym_comment,
      sym_groovy_doc,
  [75894] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3736), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2047), 2,
      sym_comment,
      sym_groovy_doc,
  [75920] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3738), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2048), 2,
      sym_comment,
      sym_groovy_doc,
  [75946] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3740), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2049), 2,
      sym_comment,
      sym_groovy_doc,
  [75972] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3742), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2050), 2,
      sym_comment,
      sym_groovy_doc,
  [75998] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3744), 1,
      sym_identifier,
    ACTIONS(3746), 1,
      anon_sym_STAR2,
    STATE(754), 1,
      sym__type_identifier,
    STATE(2051), 2,
      sym_comment,
      sym_groovy_doc,
  [76024] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3748), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2052), 2,
      sym_comment,
      sym_groovy_doc,
  [76050] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3540), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2065), 1,
      aux_sym__plain_string_repeat1,
    STATE(2053), 2,
      sym_comment,
      sym_groovy_doc,
  [76076] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3750), 1,
      anon_sym_else,
    STATE(2054), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1904), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [76098] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3303), 1,
      anon_sym_EQ,
    STATE(2055), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1790), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [76120] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3752), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2056), 2,
      sym_comment,
      sym_groovy_doc,
  [76146] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3754), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2057), 2,
      sym_comment,
      sym_groovy_doc,
  [76172] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3756), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2058), 2,
      sym_comment,
      sym_groovy_doc,
  [76198] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3758), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2059), 2,
      sym_comment,
      sym_groovy_doc,
  [76224] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3344), 1,
      sym_identifier,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    STATE(719), 1,
      sym__type_identifier,
    STATE(758), 1,
      sym__import_name,
    STATE(2060), 2,
      sym_comment,
      sym_groovy_doc,
  [76250] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3762), 1,
      anon_sym_STAR2,
    STATE(1797), 1,
      sym__type_identifier,
    STATE(2061), 2,
      sym_comment,
      sym_groovy_doc,
  [76276] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3764), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2062), 2,
      sym_comment,
      sym_groovy_doc,
  [76302] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3488), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2101), 1,
      aux_sym__plain_string_repeat1,
    STATE(2063), 2,
      sym_comment,
      sym_groovy_doc,
  [76328] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3355), 1,
      sym_identifier,
    STATE(683), 1,
      sym__type_identifier,
    STATE(1906), 1,
      sym__import_name,
    STATE(2064), 2,
      sym_comment,
      sym_groovy_doc,
  [76354] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3470), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2065), 2,
      sym_comment,
      sym_groovy_doc,
  [76380] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3766), 1,
      sym_identifier,
    ACTIONS(3768), 1,
      anon_sym_STAR2,
    STATE(838), 1,
      sym__type_identifier,
    STATE(2066), 2,
      sym_comment,
      sym_groovy_doc,
  [76406] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3770), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2067), 2,
      sym_comment,
      sym_groovy_doc,
  [76432] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1804), 1,
      sym__import_name,
    STATE(1805), 1,
      sym__type_identifier,
    STATE(2068), 2,
      sym_comment,
      sym_groovy_doc,
  [76458] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3744), 1,
      sym_identifier,
    ACTIONS(3762), 1,
      anon_sym_STAR2,
    STATE(754), 1,
      sym__type_identifier,
    STATE(2069), 2,
      sym_comment,
      sym_groovy_doc,
  [76484] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3774), 1,
      anon_sym_STAR2,
    STATE(696), 1,
      sym__type_identifier,
    STATE(2070), 2,
      sym_comment,
      sym_groovy_doc,
  [76510] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3776), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2071), 2,
      sym_comment,
      sym_groovy_doc,
  [76536] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3778), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2072), 2,
      sym_comment,
      sym_groovy_doc,
  [76562] = 7,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3780), 1,
      anon_sym_SQUOTE,
    ACTIONS(3782), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3785), 1,
      sym_escape_sequence,
    STATE(2073), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym__plain_string_repeat1,
  [76586] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1805), 1,
      sym__type_identifier,
    STATE(1812), 1,
      sym__import_name,
    STATE(2074), 2,
      sym_comment,
      sym_groovy_doc,
  [76612] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3291), 1,
      anon_sym_EQ,
    STATE(2075), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1763), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [76634] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3542), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2096), 1,
      aux_sym__plain_string_repeat1,
    STATE(2076), 2,
      sym_comment,
      sym_groovy_doc,
  [76660] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3480), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2089), 1,
      aux_sym__plain_string_repeat1,
    STATE(2077), 2,
      sym_comment,
      sym_groovy_doc,
  [76686] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3606), 1,
      aux_sym__interpolate_string_token1,
    STATE(2078), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3604), 3,
      anon_sym_DQUOTE,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [76708] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3788), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2079), 2,
      sym_comment,
      sym_groovy_doc,
  [76734] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3293), 1,
      anon_sym_EQ,
    STATE(2080), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(1753), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [76756] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3790), 1,
      anon_sym_COMMA,
    STATE(2007), 1,
      sym_generics,
    STATE(2081), 2,
      sym_comment,
      sym_groovy_doc,
  [76782] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3792), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2082), 2,
      sym_comment,
      sym_groovy_doc,
  [76808] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3794), 1,
      sym_identifier,
    ACTIONS(3796), 1,
      anon_sym_STAR2,
    STATE(679), 1,
      sym__type_identifier,
    STATE(2083), 2,
      sym_comment,
      sym_groovy_doc,
  [76834] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3348), 1,
      sym_identifier,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    STATE(709), 1,
      sym__import_name,
    STATE(710), 1,
      sym__type_identifier,
    STATE(2084), 2,
      sym_comment,
      sym_groovy_doc,
  [76860] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3798), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2085), 2,
      sym_comment,
      sym_groovy_doc,
  [76886] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3638), 1,
      aux_sym__interpolate_string_token1,
    STATE(2086), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3636), 3,
      anon_sym_DQUOTE,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [76908] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3344), 1,
      sym_identifier,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    STATE(719), 1,
      sym__type_identifier,
    STATE(723), 1,
      sym__import_name,
    STATE(2087), 2,
      sym_comment,
      sym_groovy_doc,
  [76934] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3359), 1,
      sym_identifier,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    STATE(799), 1,
      sym__type_identifier,
    STATE(907), 1,
      sym__import_name,
    STATE(2088), 2,
      sym_comment,
      sym_groovy_doc,
  [76960] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3524), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2089), 2,
      sym_comment,
      sym_groovy_doc,
  [76986] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3800), 1,
      anon_sym_STAR2,
    STATE(1797), 1,
      sym__type_identifier,
    STATE(2090), 2,
      sym_comment,
      sym_groovy_doc,
  [77012] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3802), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2091), 2,
      sym_comment,
      sym_groovy_doc,
  [77038] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3746), 1,
      anon_sym_STAR2,
    ACTIONS(3760), 1,
      sym_identifier,
    STATE(1797), 1,
      sym__type_identifier,
    STATE(2092), 2,
      sym_comment,
      sym_groovy_doc,
  [77064] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3804), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2093), 2,
      sym_comment,
      sym_groovy_doc,
  [77090] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3285), 1,
      anon_sym_EQ,
    STATE(2094), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2007), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [77112] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3806), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2095), 2,
      sym_comment,
      sym_groovy_doc,
  [77138] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3486), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2096), 2,
      sym_comment,
      sym_groovy_doc,
  [77164] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3808), 1,
      anon_sym_EQ,
    STATE(2097), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(2143), 3,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_RPAREN,
  [77186] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3810), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2098), 2,
      sym_comment,
      sym_groovy_doc,
  [77212] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3814), 1,
      aux_sym__interpolate_string_token1,
    STATE(2099), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3812), 3,
      anon_sym_DQUOTE,
      sym_escape_sequence,
      anon_sym_DOLLAR,
  [77234] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3816), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2100), 2,
      sym_comment,
      sym_groovy_doc,
  [77260] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3538), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2073), 1,
      aux_sym__plain_string_repeat1,
    STATE(2101), 2,
      sym_comment,
      sym_groovy_doc,
  [77286] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3818), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2102), 2,
      sym_comment,
      sym_groovy_doc,
  [77312] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3820), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2103), 2,
      sym_comment,
      sym_groovy_doc,
  [77338] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3822), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2104), 2,
      sym_comment,
      sym_groovy_doc,
  [77364] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3355), 1,
      sym_identifier,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    STATE(683), 1,
      sym__type_identifier,
    STATE(781), 1,
      sym__import_name,
    STATE(2105), 2,
      sym_comment,
      sym_groovy_doc,
  [77390] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3348), 1,
      sym_identifier,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    STATE(710), 1,
      sym__type_identifier,
    STATE(771), 1,
      sym__import_name,
    STATE(2106), 2,
      sym_comment,
      sym_groovy_doc,
  [77416] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1805), 1,
      sym__type_identifier,
    STATE(1866), 1,
      sym__import_name,
    STATE(2107), 2,
      sym_comment,
      sym_groovy_doc,
  [77442] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3824), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2108), 2,
      sym_comment,
      sym_groovy_doc,
  [77468] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(2109), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(389), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_COMMA,
      anon_sym_LBRACK_RBRACK,
  [77488] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3530), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2046), 1,
      aux_sym__plain_string_repeat1,
    STATE(2110), 2,
      sym_comment,
      sym_groovy_doc,
  [77514] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3826), 1,
      anon_sym_STAR2,
    STATE(696), 1,
      sym__type_identifier,
    STATE(2111), 2,
      sym_comment,
      sym_groovy_doc,
  [77540] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3796), 1,
      anon_sym_STAR2,
    STATE(1797), 1,
      sym__type_identifier,
    STATE(2112), 2,
      sym_comment,
      sym_groovy_doc,
  [77566] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3359), 1,
      sym_identifier,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    STATE(799), 1,
      sym__type_identifier,
    STATE(919), 1,
      sym__import_name,
    STATE(2113), 2,
      sym_comment,
      sym_groovy_doc,
  [77592] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3794), 1,
      sym_identifier,
    ACTIONS(3800), 1,
      anon_sym_STAR2,
    STATE(679), 1,
      sym__type_identifier,
    STATE(2114), 2,
      sym_comment,
      sym_groovy_doc,
  [77618] = 8,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3544), 1,
      anon_sym_SQUOTE,
    ACTIONS(3726), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3728), 1,
      sym_escape_sequence,
    STATE(2042), 1,
      aux_sym__plain_string_repeat1,
    STATE(2115), 2,
      sym_comment,
      sym_groovy_doc,
  [77644] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3766), 1,
      sym_identifier,
    ACTIONS(3828), 1,
      anon_sym_STAR2,
    STATE(838), 1,
      sym__type_identifier,
    STATE(2116), 2,
      sym_comment,
      sym_groovy_doc,
  [77670] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3830), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2117), 2,
      sym_comment,
      sym_groovy_doc,
  [77696] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3355), 1,
      sym_identifier,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    STATE(683), 1,
      sym__type_identifier,
    STATE(922), 1,
      sym__import_name,
    STATE(2118), 2,
      sym_comment,
      sym_groovy_doc,
  [77722] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3832), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2119), 2,
      sym_comment,
      sym_groovy_doc,
  [77748] = 8,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3580), 1,
      anon_sym_LT,
    ACTIONS(3584), 1,
      anon_sym_LBRACK_RBRACK,
    ACTIONS(3834), 1,
      sym_identifier,
    STATE(2007), 1,
      sym_generics,
    STATE(2120), 2,
      sym_comment,
      sym_groovy_doc,
  [77774] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3836), 1,
      anon_sym_extends,
    STATE(816), 1,
      sym_closure,
    STATE(2121), 2,
      sym_comment,
      sym_groovy_doc,
  [77797] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3838), 1,
      sym_identifier,
    STATE(1937), 1,
      sym__type_identifier,
    STATE(2122), 2,
      sym_comment,
      sym_groovy_doc,
  [77820] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3842), 1,
      anon_sym_RPAREN,
    STATE(2198), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2123), 2,
      sym_comment,
      sym_groovy_doc,
  [77843] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3844), 1,
      sym_identifier,
    STATE(2009), 1,
      sym__type_identifier,
    STATE(2124), 2,
      sym_comment,
      sym_groovy_doc,
  [77866] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3848), 1,
      anon_sym_extends,
    ACTIONS(3846), 2,
      anon_sym_GT,
      anon_sym_COMMA,
    STATE(2125), 2,
      sym_comment,
      sym_groovy_doc,
  [77887] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3850), 1,
      sym_identifier,
    STATE(701), 1,
      sym__type_identifier,
    STATE(2126), 2,
      sym_comment,
      sym_groovy_doc,
  [77910] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3852), 1,
      anon_sym_extends,
    STATE(949), 1,
      sym_closure,
    STATE(2127), 2,
      sym_comment,
      sym_groovy_doc,
  [77933] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3854), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2128), 2,
      sym_comment,
      sym_groovy_doc,
  [77956] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3850), 1,
      sym_identifier,
    STATE(701), 1,
      sym__type_identifier,
    STATE(2129), 2,
      sym_comment,
      sym_groovy_doc,
  [77979] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3856), 1,
      anon_sym_extends,
    STATE(849), 1,
      sym_closure,
    STATE(2130), 2,
      sym_comment,
      sym_groovy_doc,
  [78002] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3858), 1,
      sym_identifier,
    STATE(2010), 1,
      sym__type_identifier,
    STATE(2131), 2,
      sym_comment,
      sym_groovy_doc,
  [78025] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3860), 1,
      anon_sym_RPAREN,
    STATE(2176), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2132), 2,
      sym_comment,
      sym_groovy_doc,
  [78048] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3862), 1,
      sym_identifier,
    STATE(814), 1,
      sym__type_identifier,
    STATE(2133), 2,
      sym_comment,
      sym_groovy_doc,
  [78071] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3864), 1,
      anon_sym_RPAREN,
    STATE(2177), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2134), 2,
      sym_comment,
      sym_groovy_doc,
  [78094] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3866), 1,
      sym_identifier,
    STATE(2017), 1,
      sym__type_identifier,
    STATE(2135), 2,
      sym_comment,
      sym_groovy_doc,
  [78117] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3868), 1,
      anon_sym_extends,
    STATE(755), 1,
      sym_closure,
    STATE(2136), 2,
      sym_comment,
      sym_groovy_doc,
  [78140] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3870), 1,
      sym_identifier,
    STATE(2022), 1,
      sym__type_identifier,
    STATE(2137), 2,
      sym_comment,
      sym_groovy_doc,
  [78163] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3872), 1,
      anon_sym_extends,
    STATE(873), 1,
      sym_closure,
    STATE(2138), 2,
      sym_comment,
      sym_groovy_doc,
  [78186] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3874), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2139), 2,
      sym_comment,
      sym_groovy_doc,
  [78209] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3876), 1,
      sym_identifier,
    STATE(2026), 1,
      sym__type_identifier,
    STATE(2140), 2,
      sym_comment,
      sym_groovy_doc,
  [78232] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3878), 1,
      sym_identifier,
    STATE(1944), 1,
      sym__type_identifier,
    STATE(2141), 2,
      sym_comment,
      sym_groovy_doc,
  [78255] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3880), 1,
      sym_identifier,
    STATE(858), 1,
      sym__type_identifier,
    STATE(2142), 2,
      sym_comment,
      sym_groovy_doc,
  [78278] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3882), 1,
      anon_sym_extends,
    STATE(794), 1,
      sym_closure,
    STATE(2143), 2,
      sym_comment,
      sym_groovy_doc,
  [78301] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3884), 1,
      anon_sym_RPAREN,
    STATE(2128), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2144), 2,
      sym_comment,
      sym_groovy_doc,
  [78324] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3886), 1,
      sym_identifier,
    STATE(777), 1,
      sym__type_identifier,
    STATE(2145), 2,
      sym_comment,
      sym_groovy_doc,
  [78347] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3888), 1,
      sym_identifier,
    STATE(1982), 1,
      sym__type_identifier,
    STATE(2146), 2,
      sym_comment,
      sym_groovy_doc,
  [78370] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    STATE(2147), 2,
      sym_comment,
      sym_groovy_doc,
    ACTIONS(3890), 3,
      anon_sym_RBRACE,
      anon_sym_case,
      anon_sym_default,
  [78389] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3892), 1,
      sym_identifier,
    STATE(2039), 1,
      sym__type_identifier,
    STATE(2148), 2,
      sym_comment,
      sym_groovy_doc,
  [78412] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3894), 1,
      sym_identifier,
    STATE(2192), 1,
      aux_sym_generic_parameters_repeat1,
    STATE(2258), 1,
      sym_generic_param,
    STATE(2149), 2,
      sym_comment,
      sym_groovy_doc,
  [78435] = 6,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3898), 1,
      aux_sym__plain_string_token1,
    ACTIONS(3896), 2,
      anon_sym_SQUOTE,
      sym_escape_sequence,
    STATE(2150), 2,
      sym_comment,
      sym_groovy_doc,
  [78456] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3900), 1,
      sym_identifier,
    STATE(999), 1,
      sym__type_identifier,
    STATE(2151), 2,
      sym_comment,
      sym_groovy_doc,
  [78479] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3902), 1,
      sym_identifier,
    STATE(1922), 1,
      sym__type_identifier,
    STATE(2152), 2,
      sym_comment,
      sym_groovy_doc,
  [78502] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3766), 1,
      sym_identifier,
    STATE(838), 1,
      sym__type_identifier,
    STATE(2153), 2,
      sym_comment,
      sym_groovy_doc,
  [78525] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3904), 1,
      anon_sym_extends,
    STATE(699), 1,
      sym_closure,
    STATE(2154), 2,
      sym_comment,
      sym_groovy_doc,
  [78548] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3906), 1,
      sym_identifier,
    STATE(1925), 1,
      sym__type_identifier,
    STATE(2155), 2,
      sym_comment,
      sym_groovy_doc,
  [78571] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3772), 1,
      sym_identifier,
    STATE(696), 1,
      sym__type_identifier,
    STATE(2156), 2,
      sym_comment,
      sym_groovy_doc,
  [78594] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3908), 1,
      sym_identifier,
    STATE(1929), 1,
      sym__type_identifier,
    STATE(2157), 2,
      sym_comment,
      sym_groovy_doc,
  [78617] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3886), 1,
      sym_identifier,
    STATE(777), 1,
      sym__type_identifier,
    STATE(2158), 2,
      sym_comment,
      sym_groovy_doc,
  [78640] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3910), 1,
      sym_identifier,
    STATE(1940), 1,
      sym__type_identifier,
    STATE(2159), 2,
      sym_comment,
      sym_groovy_doc,
  [78663] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3912), 1,
      sym_escape_sequence,
    ACTIONS(3914), 1,
      anon_sym_DOLLAR,
    STATE(2014), 1,
      sym_interpolation,
    STATE(2160), 2,
      sym_comment,
      sym_groovy_doc,
  [78686] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3916), 1,
      sym_identifier,
    STATE(694), 1,
      sym__type_identifier,
    STATE(2161), 2,
      sym_comment,
      sym_groovy_doc,
  [78709] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3918), 1,
      sym_identifier,
    STATE(1947), 1,
      sym__type_identifier,
    STATE(2162), 2,
      sym_comment,
      sym_groovy_doc,
  [78732] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3920), 1,
      anon_sym_SEMI,
    ACTIONS(3922), 1,
      anon_sym_COMMA,
    STATE(2175), 1,
      aux_sym_for_parameters_repeat1,
    STATE(2163), 2,
      sym_comment,
      sym_groovy_doc,
  [78755] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3924), 1,
      sym_identifier,
    STATE(1953), 1,
      sym__type_identifier,
    STATE(2164), 2,
      sym_comment,
      sym_groovy_doc,
  [78778] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3926), 1,
      anon_sym_extends,
    STATE(883), 1,
      sym_closure,
    STATE(2165), 2,
      sym_comment,
      sym_groovy_doc,
  [78801] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3928), 1,
      sym_identifier,
    STATE(1958), 1,
      sym__type_identifier,
    STATE(2166), 2,
      sym_comment,
      sym_groovy_doc,
  [78824] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1213), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3760), 1,
      sym_identifier,
    STATE(1797), 1,
      sym__type_identifier,
    STATE(2167), 2,
      sym_comment,
      sym_groovy_doc,
  [78847] = 7,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3930), 1,
      aux_sym__plain_string_token3,
    ACTIONS(3932), 1,
      aux_sym__plain_string_token4,
    ACTIONS(3934), 1,
      sym_escape_sequence,
    STATE(2168), 2,
      sym_comment,
      sym_groovy_doc,
  [78870] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3936), 1,
      anon_sym_extends,
    STATE(966), 1,
      sym_closure,
    STATE(2169), 2,
      sym_comment,
      sym_groovy_doc,
  [78893] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3938), 1,
      anon_sym_extends,
    STATE(722), 1,
      sym_closure,
    STATE(2170), 2,
      sym_comment,
      sym_groovy_doc,
  [78916] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3940), 1,
      anon_sym_extends,
    STATE(794), 1,
      sym_closure,
    STATE(2171), 2,
      sym_comment,
      sym_groovy_doc,
  [78939] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3357), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3794), 1,
      sym_identifier,
    STATE(679), 1,
      sym__type_identifier,
    STATE(2172), 2,
      sym_comment,
      sym_groovy_doc,
  [78962] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3942), 1,
      anon_sym_extends,
    STATE(699), 1,
      sym_closure,
    STATE(2173), 2,
      sym_comment,
      sym_groovy_doc,
  [78985] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    ACTIONS(3944), 1,
      anon_sym_extends,
    STATE(821), 1,
      sym_closure,
    STATE(2174), 2,
      sym_comment,
      sym_groovy_doc,
  [79008] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3922), 1,
      anon_sym_COMMA,
    ACTIONS(3946), 1,
      anon_sym_SEMI,
    STATE(2191), 1,
      aux_sym_for_parameters_repeat1,
    STATE(2175), 2,
      sym_comment,
      sym_groovy_doc,
  [79031] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3948), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2176), 2,
      sym_comment,
      sym_groovy_doc,
  [79054] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3950), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2177), 2,
      sym_comment,
      sym_groovy_doc,
  [79077] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3952), 1,
      anon_sym_extends,
    STATE(755), 1,
      sym_closure,
    STATE(2178), 2,
      sym_comment,
      sym_groovy_doc,
  [79100] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    ACTIONS(3954), 1,
      anon_sym_extends,
    STATE(735), 1,
      sym_closure,
    STATE(2179), 2,
      sym_comment,
      sym_groovy_doc,
  [79123] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3956), 1,
      anon_sym_extends,
    STATE(775), 1,
      sym_closure,
    STATE(2180), 2,
      sym_comment,
      sym_groovy_doc,
  [79146] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3958), 1,
      anon_sym_RPAREN,
    STATE(2139), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2181), 2,
      sym_comment,
      sym_groovy_doc,
  [79169] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3346), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3960), 1,
      sym_identifier,
    STATE(839), 1,
      sym__type_identifier,
    STATE(2182), 2,
      sym_comment,
      sym_groovy_doc,
  [79192] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3962), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2183), 2,
      sym_comment,
      sym_groovy_doc,
  [79215] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3964), 1,
      anon_sym_extends,
    STATE(821), 1,
      sym_closure,
    STATE(2184), 2,
      sym_comment,
      sym_groovy_doc,
  [79238] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3966), 1,
      anon_sym_extends,
    STATE(722), 1,
      sym_closure,
    STATE(2185), 2,
      sym_comment,
      sym_groovy_doc,
  [79261] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3968), 1,
      sym_identifier,
    STATE(1919), 1,
      sym__type_identifier,
    STATE(2186), 2,
      sym_comment,
      sym_groovy_doc,
  [79284] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3970), 1,
      anon_sym_COMMA,
    ACTIONS(3973), 1,
      anon_sym_RPAREN,
    STATE(2187), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_for_parameters_repeat2,
  [79305] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3975), 1,
      anon_sym_extends,
    STATE(983), 1,
      sym_closure,
    STATE(2188), 2,
      sym_comment,
      sym_groovy_doc,
  [79328] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    ACTIONS(3977), 1,
      anon_sym_extends,
    STATE(735), 1,
      sym_closure,
    STATE(2189), 2,
      sym_comment,
      sym_groovy_doc,
  [79351] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3979), 1,
      anon_sym_RPAREN,
    STATE(2183), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2190), 2,
      sym_comment,
      sym_groovy_doc,
  [79374] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3981), 1,
      anon_sym_SEMI,
    ACTIONS(3983), 1,
      anon_sym_COMMA,
    STATE(2191), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_for_parameters_repeat1,
  [79395] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3986), 1,
      sym_identifier,
    STATE(2367), 1,
      sym_generic_param,
    STATE(2192), 3,
      sym_comment,
      sym_groovy_doc,
      aux_sym_generic_parameters_repeat1,
  [79416] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3862), 1,
      sym_identifier,
    STATE(814), 1,
      sym__type_identifier,
    STATE(2193), 2,
      sym_comment,
      sym_groovy_doc,
  [79439] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3989), 1,
      sym_identifier,
    STATE(2012), 1,
      sym__type_identifier,
    STATE(2194), 2,
      sym_comment,
      sym_groovy_doc,
  [79462] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_extends,
    STATE(994), 1,
      sym_closure,
    STATE(2195), 2,
      sym_comment,
      sym_groovy_doc,
  [79485] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3993), 1,
      sym_identifier,
    STATE(1984), 1,
      sym__type_identifier,
    STATE(2196), 2,
      sym_comment,
      sym_groovy_doc,
  [79508] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3794), 1,
      sym_identifier,
    STATE(679), 1,
      sym__type_identifier,
    STATE(2197), 2,
      sym_comment,
      sym_groovy_doc,
  [79531] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3840), 1,
      anon_sym_COMMA,
    ACTIONS(3995), 1,
      anon_sym_RPAREN,
    STATE(2187), 1,
      aux_sym_for_parameters_repeat2,
    STATE(2198), 2,
      sym_comment,
      sym_groovy_doc,
  [79554] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3916), 1,
      sym_identifier,
    STATE(694), 1,
      sym__type_identifier,
    STATE(2199), 2,
      sym_comment,
      sym_groovy_doc,
  [79577] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3350), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(3744), 1,
      sym_identifier,
    STATE(754), 1,
      sym__type_identifier,
    STATE(2200), 2,
      sym_comment,
      sym_groovy_doc,
  [79600] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(3997), 1,
      anon_sym_extends,
    STATE(816), 1,
      sym_closure,
    STATE(2201), 2,
      sym_comment,
      sym_groovy_doc,
  [79623] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_extends,
    STATE(840), 1,
      sym_closure,
    STATE(2202), 2,
      sym_comment,
      sym_groovy_doc,
  [79646] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4001), 1,
      sym_identifier,
    STATE(1990), 1,
      sym__type_identifier,
    STATE(2203), 2,
      sym_comment,
      sym_groovy_doc,
  [79669] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4003), 1,
      sym_identifier,
    STATE(1970), 1,
      sym__type_identifier,
    STATE(2204), 2,
      sym_comment,
      sym_groovy_doc,
  [79692] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3894), 1,
      sym_identifier,
    STATE(2149), 1,
      aux_sym_generic_parameters_repeat1,
    STATE(2283), 1,
      sym_generic_param,
    STATE(2205), 2,
      sym_comment,
      sym_groovy_doc,
  [79715] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4005), 1,
      sym_identifier,
    STATE(1994), 1,
      sym__type_identifier,
    STATE(2206), 2,
      sym_comment,
      sym_groovy_doc,
  [79738] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4007), 1,
      anon_sym_else,
    ACTIONS(1904), 2,
      anon_sym_SEMI,
      anon_sym_while,
    STATE(2207), 2,
      sym_comment,
      sym_groovy_doc,
  [79759] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4009), 1,
      sym_identifier,
    STATE(1999), 1,
      sym__type_identifier,
    STATE(2208), 2,
      sym_comment,
      sym_groovy_doc,
  [79782] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(59), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4011), 1,
      sym_identifier,
    STATE(1938), 1,
      sym__type_identifier,
    STATE(2209), 2,
      sym_comment,
      sym_groovy_doc,
  [79805] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    ACTIONS(4013), 1,
      anon_sym_extends,
    STATE(849), 1,
      sym_closure,
    STATE(2210), 2,
      sym_comment,
      sym_groovy_doc,
  [79828] = 7,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3361), 1,
      aux_sym__type_identifier_token1,
    ACTIONS(4015), 1,
      sym_identifier,
    STATE(989), 1,
      sym__type_identifier,
    STATE(2211), 2,
      sym_comment,
      sym_groovy_doc,
  [79851] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2603), 1,
      anon_sym_COMMA,
    ACTIONS(2605), 1,
      anon_sym_RPAREN,
    STATE(2212), 2,
      sym_comment,
      sym_groovy_doc,
  [79871] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2577), 1,
      anon_sym_COMMA,
    ACTIONS(2579), 1,
      anon_sym_RPAREN,
    STATE(2213), 2,
      sym_comment,
      sym_groovy_doc,
  [79891] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4017), 1,
      anon_sym_COMMA,
    ACTIONS(4019), 1,
      anon_sym_DASH_GT,
    STATE(2214), 2,
      sym_comment,
      sym_groovy_doc,
  [79911] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4021), 1,
      anon_sym_LPAREN,
    STATE(134), 1,
      sym_for_parameters,
    STATE(2215), 2,
      sym_comment,
      sym_groovy_doc,
  [79931] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2573), 1,
      anon_sym_COMMA,
    ACTIONS(2575), 1,
      anon_sym_RPAREN,
    STATE(2216), 2,
      sym_comment,
      sym_groovy_doc,
  [79951] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(135), 1,
      sym_parenthesized_expression,
    STATE(2217), 2,
      sym_comment,
      sym_groovy_doc,
  [79971] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4025), 1,
      anon_sym_LPAREN,
    STATE(138), 1,
      sym_for_parameters,
    STATE(2218), 2,
      sym_comment,
      sym_groovy_doc,
  [79991] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    STATE(863), 1,
      sym_closure,
    STATE(2219), 2,
      sym_comment,
      sym_groovy_doc,
  [80011] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    STATE(935), 1,
      sym_closure,
    STATE(2220), 2,
      sym_comment,
      sym_groovy_doc,
  [80031] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(160), 1,
      sym_parenthesized_expression,
    STATE(2221), 2,
      sym_comment,
      sym_groovy_doc,
  [80051] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3269), 1,
      anon_sym_RPAREN,
    ACTIONS(4027), 1,
      anon_sym_COMMA,
    STATE(2222), 2,
      sym_comment,
      sym_groovy_doc,
  [80071] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3275), 1,
      anon_sym_RPAREN,
    ACTIONS(4029), 1,
      anon_sym_COMMA,
    STATE(2223), 2,
      sym_comment,
      sym_groovy_doc,
  [80091] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(863), 1,
      sym_closure,
    STATE(2224), 2,
      sym_comment,
      sym_groovy_doc,
  [80111] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2611), 1,
      anon_sym_COMMA,
    ACTIONS(2613), 1,
      anon_sym_RPAREN,
    STATE(2225), 2,
      sym_comment,
      sym_groovy_doc,
  [80131] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(190), 1,
      sym_parenthesized_expression,
    STATE(2226), 2,
      sym_comment,
      sym_groovy_doc,
  [80151] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4031), 1,
      anon_sym_LPAREN,
    STATE(191), 1,
      sym_for_parameters,
    STATE(2227), 2,
      sym_comment,
      sym_groovy_doc,
  [80171] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3229), 1,
      anon_sym_RBRACK,
    ACTIONS(4033), 1,
      anon_sym_COMMA,
    STATE(2228), 2,
      sym_comment,
      sym_groovy_doc,
  [80191] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2585), 1,
      anon_sym_COMMA,
    ACTIONS(2587), 1,
      anon_sym_RPAREN,
    STATE(2229), 2,
      sym_comment,
      sym_groovy_doc,
  [80211] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(170), 1,
      sym_parenthesized_expression,
    STATE(2230), 2,
      sym_comment,
      sym_groovy_doc,
  [80231] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(729), 1,
      sym_closure,
    STATE(2231), 2,
      sym_comment,
      sym_groovy_doc,
  [80251] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    STATE(948), 1,
      sym_closure,
    STATE(2232), 2,
      sym_comment,
      sym_groovy_doc,
  [80271] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(2355), 1,
      sym_closure,
    STATE(2233), 2,
      sym_comment,
      sym_groovy_doc,
  [80291] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(193), 1,
      sym_parenthesized_expression,
    STATE(2234), 2,
      sym_comment,
      sym_groovy_doc,
  [80311] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4035), 1,
      anon_sym_LPAREN,
    STATE(194), 1,
      sym_for_parameters,
    STATE(2235), 2,
      sym_comment,
      sym_groovy_doc,
  [80331] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(146), 1,
      sym_parenthesized_expression,
    STATE(2236), 2,
      sym_comment,
      sym_groovy_doc,
  [80351] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(196), 1,
      sym_parenthesized_expression,
    STATE(2237), 2,
      sym_comment,
      sym_groovy_doc,
  [80371] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4037), 1,
      anon_sym_LPAREN,
    STATE(197), 1,
      sym_for_parameters,
    STATE(2238), 2,
      sym_comment,
      sym_groovy_doc,
  [80391] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(198), 1,
      sym_parenthesized_expression,
    STATE(2239), 2,
      sym_comment,
      sym_groovy_doc,
  [80411] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(207), 1,
      sym_parenthesized_expression,
    STATE(2240), 2,
      sym_comment,
      sym_groovy_doc,
  [80431] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4039), 1,
      anon_sym_LPAREN,
    STATE(208), 1,
      sym_for_parameters,
    STATE(2241), 2,
      sym_comment,
      sym_groovy_doc,
  [80451] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(172), 1,
      sym_parenthesized_expression,
    STATE(2242), 2,
      sym_comment,
      sym_groovy_doc,
  [80471] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(200), 1,
      sym_parenthesized_expression,
    STATE(2243), 2,
      sym_comment,
      sym_groovy_doc,
  [80491] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4041), 1,
      anon_sym_LPAREN,
    STATE(201), 1,
      sym_for_parameters,
    STATE(2244), 2,
      sym_comment,
      sym_groovy_doc,
  [80511] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(203), 1,
      sym_parenthesized_expression,
    STATE(2245), 2,
      sym_comment,
      sym_groovy_doc,
  [80531] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    STATE(204), 1,
      sym_for_parameters,
    STATE(2246), 2,
      sym_comment,
      sym_groovy_doc,
  [80551] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4045), 1,
      anon_sym_LPAREN,
    STATE(174), 1,
      sym_for_parameters,
    STATE(2247), 2,
      sym_comment,
      sym_groovy_doc,
  [80571] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(175), 1,
      sym_parenthesized_expression,
    STATE(2248), 2,
      sym_comment,
      sym_groovy_doc,
  [80591] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(143), 1,
      sym_parenthesized_expression,
    STATE(2249), 2,
      sym_comment,
      sym_groovy_doc,
  [80611] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4047), 1,
      anon_sym_COMMA,
    ACTIONS(4049), 1,
      anon_sym_RPAREN,
    STATE(2250), 2,
      sym_comment,
      sym_groovy_doc,
  [80631] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2541), 1,
      anon_sym_COMMA,
    ACTIONS(2543), 1,
      anon_sym_RPAREN,
    STATE(2251), 2,
      sym_comment,
      sym_groovy_doc,
  [80651] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    STATE(2299), 1,
      sym_parenthesized_expression,
    STATE(2252), 2,
      sym_comment,
      sym_groovy_doc,
  [80671] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(152), 1,
      sym_parenthesized_expression,
    STATE(2253), 2,
      sym_comment,
      sym_groovy_doc,
  [80691] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4051), 1,
      anon_sym_LPAREN,
    STATE(154), 1,
      sym_for_parameters,
    STATE(2254), 2,
      sym_comment,
      sym_groovy_doc,
  [80711] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(41), 1,
      anon_sym_LBRACE,
    STATE(884), 1,
      sym_closure,
    STATE(2255), 2,
      sym_comment,
      sym_groovy_doc,
  [80731] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(155), 1,
      sym_parenthesized_expression,
    STATE(2256), 2,
      sym_comment,
      sym_groovy_doc,
  [80751] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    STATE(2300), 1,
      sym_parenthesized_expression,
    STATE(2257), 2,
      sym_comment,
      sym_groovy_doc,
  [80771] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4053), 1,
      anon_sym_GT,
    ACTIONS(4055), 1,
      anon_sym_COMMA,
    STATE(2258), 2,
      sym_comment,
      sym_groovy_doc,
  [80791] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(182), 1,
      sym_parenthesized_expression,
    STATE(2259), 2,
      sym_comment,
      sym_groovy_doc,
  [80811] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4057), 1,
      anon_sym_LPAREN,
    STATE(183), 1,
      sym_for_parameters,
    STATE(2260), 2,
      sym_comment,
      sym_groovy_doc,
  [80831] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(184), 1,
      sym_parenthesized_expression,
    STATE(2261), 2,
      sym_comment,
      sym_groovy_doc,
  [80851] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4059), 1,
      anon_sym_COMMA,
    ACTIONS(4061), 1,
      anon_sym_RBRACK,
    STATE(2262), 2,
      sym_comment,
      sym_groovy_doc,
  [80871] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3227), 1,
      anon_sym_RBRACK,
    ACTIONS(4063), 1,
      anon_sym_COMMA,
    STATE(2263), 2,
      sym_comment,
      sym_groovy_doc,
  [80891] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(168), 1,
      sym_parenthesized_expression,
    STATE(2264), 2,
      sym_comment,
      sym_groovy_doc,
  [80911] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    STATE(879), 1,
      sym_closure,
    STATE(2265), 2,
      sym_comment,
      sym_groovy_doc,
  [80931] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3277), 1,
      anon_sym_RPAREN,
    ACTIONS(4065), 1,
      anon_sym_COMMA,
    STATE(2266), 2,
      sym_comment,
      sym_groovy_doc,
  [80951] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4067), 1,
      anon_sym_COMMA,
    ACTIONS(4069), 1,
      anon_sym_RBRACK,
    STATE(2267), 2,
      sym_comment,
      sym_groovy_doc,
  [80971] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(212), 1,
      sym_parenthesized_expression,
    STATE(2268), 2,
      sym_comment,
      sym_groovy_doc,
  [80991] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4071), 1,
      anon_sym_LPAREN,
    STATE(213), 1,
      sym_for_parameters,
    STATE(2269), 2,
      sym_comment,
      sym_groovy_doc,
  [81011] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(162), 1,
      sym_parenthesized_expression,
    STATE(2270), 2,
      sym_comment,
      sym_groovy_doc,
  [81031] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4073), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      aux_sym_interpolation_token1,
    STATE(2271), 2,
      sym_comment,
      sym_groovy_doc,
  [81051] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    STATE(2281), 1,
      sym_parenthesized_expression,
    STATE(2272), 2,
      sym_comment,
      sym_groovy_doc,
  [81071] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4077), 1,
      anon_sym_LBRACE,
    ACTIONS(4079), 1,
      aux_sym_interpolation_token1,
    STATE(2273), 2,
      sym_comment,
      sym_groovy_doc,
  [81091] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(141), 1,
      sym_parenthesized_expression,
    STATE(2274), 2,
      sym_comment,
      sym_groovy_doc,
  [81111] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(163), 1,
      sym_parenthesized_expression,
    STATE(2275), 2,
      sym_comment,
      sym_groovy_doc,
  [81131] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4081), 1,
      anon_sym_LPAREN,
    STATE(142), 1,
      sym_for_parameters,
    STATE(2276), 2,
      sym_comment,
      sym_groovy_doc,
  [81151] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4083), 1,
      anon_sym_COMMA,
    ACTIONS(4085), 1,
      anon_sym_RBRACK,
    STATE(2277), 2,
      sym_comment,
      sym_groovy_doc,
  [81171] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(161), 1,
      sym_parenthesized_expression,
    STATE(2278), 2,
      sym_comment,
      sym_groovy_doc,
  [81191] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4087), 1,
      anon_sym_LBRACE,
    ACTIONS(4089), 1,
      aux_sym_interpolation_token1,
    STATE(2279), 2,
      sym_comment,
      sym_groovy_doc,
  [81211] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4091), 1,
      anon_sym_LBRACE,
    ACTIONS(4093), 1,
      aux_sym_interpolation_token1,
    STATE(2280), 2,
      sym_comment,
      sym_groovy_doc,
  [81231] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4095), 1,
      anon_sym_LBRACE,
    STATE(820), 1,
      sym_switch_block,
    STATE(2281), 2,
      sym_comment,
      sym_groovy_doc,
  [81251] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4097), 2,
      anon_sym_extends,
      anon_sym_LBRACE,
    STATE(2282), 2,
      sym_comment,
      sym_groovy_doc,
  [81269] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4099), 1,
      anon_sym_GT,
    ACTIONS(4101), 1,
      anon_sym_COMMA,
    STATE(2283), 2,
      sym_comment,
      sym_groovy_doc,
  [81289] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(836), 1,
      sym_parenthesized_expression,
    STATE(2284), 2,
      sym_comment,
      sym_groovy_doc,
  [81309] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4103), 1,
      anon_sym_LPAREN,
    STATE(218), 1,
      sym_for_parameters,
    STATE(2285), 2,
      sym_comment,
      sym_groovy_doc,
  [81329] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(153), 1,
      sym_parenthesized_expression,
    STATE(2286), 2,
      sym_comment,
      sym_groovy_doc,
  [81349] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    STATE(2306), 1,
      sym_parenthesized_expression,
    STATE(2287), 2,
      sym_comment,
      sym_groovy_doc,
  [81369] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3981), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(2288), 2,
      sym_comment,
      sym_groovy_doc,
  [81387] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(136), 1,
      sym_parenthesized_expression,
    STATE(2289), 2,
      sym_comment,
      sym_groovy_doc,
  [81407] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4105), 1,
      sym_identifier,
    ACTIONS(4107), 1,
      anon_sym_GT,
    STATE(2290), 2,
      sym_comment,
      sym_groovy_doc,
  [81427] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4109), 1,
      anon_sym_COMMA,
    ACTIONS(4111), 1,
      anon_sym_RBRACK,
    STATE(2291), 2,
      sym_comment,
      sym_groovy_doc,
  [81447] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4113), 1,
      anon_sym_LPAREN,
    STATE(844), 1,
      sym_parenthesized_expression,
    STATE(2292), 2,
      sym_comment,
      sym_groovy_doc,
  [81467] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3247), 1,
      anon_sym_RBRACK,
    ACTIONS(4115), 1,
      anon_sym_COMMA,
    STATE(2293), 2,
      sym_comment,
      sym_groovy_doc,
  [81487] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3273), 1,
      anon_sym_DASH_GT,
    ACTIONS(4117), 1,
      anon_sym_COMMA,
    STATE(2294), 2,
      sym_comment,
      sym_groovy_doc,
  [81507] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4119), 1,
      anon_sym_COMMA,
    ACTIONS(4121), 1,
      anon_sym_RBRACK,
    STATE(2295), 2,
      sym_comment,
      sym_groovy_doc,
  [81527] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2137), 1,
      anon_sym_LBRACE,
    STATE(884), 1,
      sym_closure,
    STATE(2296), 2,
      sym_comment,
      sym_groovy_doc,
  [81547] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2557), 1,
      anon_sym_COMMA,
    ACTIONS(2559), 1,
      anon_sym_RPAREN,
    STATE(2297), 2,
      sym_comment,
      sym_groovy_doc,
  [81567] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4123), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(2298), 2,
      sym_comment,
      sym_groovy_doc,
  [81585] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4125), 1,
      anon_sym_LBRACE,
    STATE(980), 1,
      sym_switch_block,
    STATE(2299), 2,
      sym_comment,
      sym_groovy_doc,
  [81605] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4127), 1,
      anon_sym_LBRACE,
    STATE(892), 1,
      sym_switch_block,
    STATE(2300), 2,
      sym_comment,
      sym_groovy_doc,
  [81625] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    STATE(836), 1,
      sym_parenthesized_expression,
    STATE(2301), 2,
      sym_comment,
      sym_groovy_doc,
  [81645] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4129), 2,
      anon_sym_extends,
      anon_sym_LBRACE,
    STATE(2302), 2,
      sym_comment,
      sym_groovy_doc,
  [81663] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4131), 1,
      anon_sym_COMMA,
    ACTIONS(4133), 1,
      anon_sym_RBRACK,
    STATE(2303), 2,
      sym_comment,
      sym_groovy_doc,
  [81683] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4135), 1,
      anon_sym_COMMA,
    ACTIONS(4137), 1,
      anon_sym_RPAREN,
    STATE(2304), 2,
      sym_comment,
      sym_groovy_doc,
  [81703] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2633), 1,
      anon_sym_COMMA,
    ACTIONS(2635), 1,
      anon_sym_RPAREN,
    STATE(2305), 2,
      sym_comment,
      sym_groovy_doc,
  [81723] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4139), 1,
      anon_sym_LBRACE,
    STATE(676), 1,
      sym_switch_block,
    STATE(2306), 2,
      sym_comment,
      sym_groovy_doc,
  [81743] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2135), 1,
      anon_sym_LBRACE,
    STATE(955), 1,
      sym_closure,
    STATE(2307), 2,
      sym_comment,
      sym_groovy_doc,
  [81763] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2561), 1,
      anon_sym_COMMA,
    ACTIONS(2563), 1,
      anon_sym_RPAREN,
    STATE(2308), 2,
      sym_comment,
      sym_groovy_doc,
  [81783] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2581), 1,
      anon_sym_COMMA,
    ACTIONS(2583), 1,
      anon_sym_RPAREN,
    STATE(2309), 2,
      sym_comment,
      sym_groovy_doc,
  [81803] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2619), 1,
      anon_sym_COMMA,
    ACTIONS(2621), 1,
      anon_sym_RPAREN,
    STATE(2310), 2,
      sym_comment,
      sym_groovy_doc,
  [81823] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3281), 1,
      anon_sym_RPAREN,
    ACTIONS(4141), 1,
      anon_sym_COMMA,
    STATE(2311), 2,
      sym_comment,
      sym_groovy_doc,
  [81843] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3846), 1,
      anon_sym_COMMA,
    ACTIONS(4143), 1,
      anon_sym_extends,
    STATE(2312), 2,
      sym_comment,
      sym_groovy_doc,
  [81863] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(167), 1,
      sym_parenthesized_expression,
    STATE(2313), 2,
      sym_comment,
      sym_groovy_doc,
  [81883] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3973), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(2314), 2,
      sym_comment,
      sym_groovy_doc,
  [81901] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2043), 1,
      anon_sym_LBRACE,
    STATE(933), 1,
      sym_closure,
    STATE(2315), 2,
      sym_comment,
      sym_groovy_doc,
  [81921] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(883), 1,
      anon_sym_LBRACE,
    STATE(1905), 1,
      sym_closure,
    STATE(2316), 2,
      sym_comment,
      sym_groovy_doc,
  [81941] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2045), 1,
      anon_sym_LBRACE,
    STATE(729), 1,
      sym_closure,
    STATE(2317), 2,
      sym_comment,
      sym_groovy_doc,
  [81961] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4145), 1,
      anon_sym_LPAREN,
    STATE(956), 1,
      sym_parenthesized_expression,
    STATE(2318), 2,
      sym_comment,
      sym_groovy_doc,
  [81981] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4147), 1,
      anon_sym_COMMA,
    ACTIONS(4149), 1,
      anon_sym_RPAREN,
    STATE(2319), 2,
      sym_comment,
      sym_groovy_doc,
  [82001] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(187), 1,
      sym_parenthesized_expression,
    STATE(2320), 2,
      sym_comment,
      sym_groovy_doc,
  [82021] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4151), 1,
      anon_sym_LPAREN,
    STATE(188), 1,
      sym_for_parameters,
    STATE(2321), 2,
      sym_comment,
      sym_groovy_doc,
  [82041] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3208), 2,
      sym_identifier,
      aux_sym__type_identifier_token1,
    STATE(2322), 2,
      sym_comment,
      sym_groovy_doc,
  [82059] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3231), 1,
      anon_sym_RBRACK,
    ACTIONS(4153), 1,
      anon_sym_COMMA,
    STATE(2323), 2,
      sym_comment,
      sym_groovy_doc,
  [82079] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(1733), 1,
      anon_sym_LPAREN,
    STATE(712), 1,
      sym_parenthesized_expression,
    STATE(2324), 2,
      sym_comment,
      sym_groovy_doc,
  [82099] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3237), 1,
      anon_sym_RBRACK,
    ACTIONS(4155), 1,
      anon_sym_COMMA,
    STATE(2325), 2,
      sym_comment,
      sym_groovy_doc,
  [82119] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2565), 1,
      anon_sym_COMMA,
    ACTIONS(2567), 1,
      anon_sym_RPAREN,
    STATE(2326), 2,
      sym_comment,
      sym_groovy_doc,
  [82139] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2607), 1,
      anon_sym_COMMA,
    ACTIONS(2609), 1,
      anon_sym_RPAREN,
    STATE(2327), 2,
      sym_comment,
      sym_groovy_doc,
  [82159] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4157), 1,
      anon_sym_COMMA,
    ACTIONS(4159), 1,
      anon_sym_RPAREN,
    STATE(2328), 2,
      sym_comment,
      sym_groovy_doc,
  [82179] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(156), 1,
      sym_parenthesized_expression,
    STATE(2329), 2,
      sym_comment,
      sym_groovy_doc,
  [82199] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4161), 1,
      anon_sym_LPAREN,
    STATE(157), 1,
      sym_for_parameters,
    STATE(2330), 2,
      sym_comment,
      sym_groovy_doc,
  [82219] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(158), 1,
      sym_parenthesized_expression,
    STATE(2331), 2,
      sym_comment,
      sym_groovy_doc,
  [82239] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2569), 1,
      anon_sym_COMMA,
    ACTIONS(2571), 1,
      anon_sym_RPAREN,
    STATE(2332), 2,
      sym_comment,
      sym_groovy_doc,
  [82259] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4163), 2,
      anon_sym_extends,
      anon_sym_LBRACE,
    STATE(2333), 2,
      sym_comment,
      sym_groovy_doc,
  [82277] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4165), 1,
      anon_sym_LPAREN,
    STATE(712), 1,
      sym_parenthesized_expression,
    STATE(2334), 2,
      sym_comment,
      sym_groovy_doc,
  [82297] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2553), 1,
      anon_sym_COMMA,
    ACTIONS(2555), 1,
      anon_sym_RPAREN,
    STATE(2335), 2,
      sym_comment,
      sym_groovy_doc,
  [82317] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2549), 1,
      anon_sym_COMMA,
    ACTIONS(2551), 1,
      anon_sym_RPAREN,
    STATE(2336), 2,
      sym_comment,
      sym_groovy_doc,
  [82337] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(133), 1,
      sym_parenthesized_expression,
    STATE(2337), 2,
      sym_comment,
      sym_groovy_doc,
  [82357] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4053), 1,
      anon_sym_GT,
    ACTIONS(4105), 1,
      sym_identifier,
    STATE(2338), 2,
      sym_comment,
      sym_groovy_doc,
  [82377] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3235), 1,
      anon_sym_RBRACK,
    ACTIONS(4167), 1,
      anon_sym_COMMA,
    STATE(2339), 2,
      sym_comment,
      sym_groovy_doc,
  [82397] = 6,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4023), 1,
      anon_sym_LPAREN,
    STATE(130), 1,
      sym_parenthesized_expression,
    STATE(2340), 2,
      sym_comment,
      sym_groovy_doc,
  [82417] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4169), 1,
      anon_sym_in,
    STATE(2341), 2,
      sym_comment,
      sym_groovy_doc,
  [82434] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4069), 1,
      anon_sym_RBRACK,
    STATE(2342), 2,
      sym_comment,
      sym_groovy_doc,
  [82451] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4171), 1,
      anon_sym_DASH_GT,
    STATE(2343), 2,
      sym_comment,
      sym_groovy_doc,
  [82468] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4173), 1,
      anon_sym_while,
    STATE(2344), 2,
      sym_comment,
      sym_groovy_doc,
  [82485] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2741), 1,
      anon_sym_RPAREN,
    STATE(2345), 2,
      sym_comment,
      sym_groovy_doc,
  [82502] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4175), 1,
      sym_identifier,
    STATE(2346), 2,
      sym_comment,
      sym_groovy_doc,
  [82519] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2643), 1,
      anon_sym_RPAREN,
    STATE(2347), 2,
      sym_comment,
      sym_groovy_doc,
  [82536] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4177), 1,
      anon_sym_DASH_GT,
    STATE(2348), 2,
      sym_comment,
      sym_groovy_doc,
  [82553] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2697), 1,
      anon_sym_RPAREN,
    STATE(2349), 2,
      sym_comment,
      sym_groovy_doc,
  [82570] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4179), 1,
      sym_escape_sequence,
    STATE(2350), 2,
      sym_comment,
      sym_groovy_doc,
  [82587] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4181), 1,
      anon_sym_in,
    STATE(2351), 2,
      sym_comment,
      sym_groovy_doc,
  [82604] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4121), 1,
      anon_sym_RBRACK,
    STATE(2352), 2,
      sym_comment,
      sym_groovy_doc,
  [82621] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(660), 1,
      ts_builtin_sym_end,
    STATE(2353), 2,
      sym_comment,
      sym_groovy_doc,
  [82638] = 5,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4183), 1,
      aux_sym_groovy_doc_token2,
    STATE(2354), 2,
      sym_comment,
      sym_groovy_doc,
  [82655] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4185), 1,
      ts_builtin_sym_end,
    STATE(2355), 2,
      sym_comment,
      sym_groovy_doc,
  [82672] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(2356), 2,
      sym_comment,
      sym_groovy_doc,
  [82689] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2713), 1,
      anon_sym_RPAREN,
    STATE(2357), 2,
      sym_comment,
      sym_groovy_doc,
  [82706] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4187), 1,
      anon_sym_COMMA,
    STATE(2358), 2,
      sym_comment,
      sym_groovy_doc,
  [82723] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4189), 1,
      anon_sym_DASH_GT,
    STATE(2359), 2,
      sym_comment,
      sym_groovy_doc,
  [82740] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2725), 1,
      anon_sym_RPAREN,
    STATE(2360), 2,
      sym_comment,
      sym_groovy_doc,
  [82757] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4191), 1,
      ts_builtin_sym_end,
    STATE(2361), 2,
      sym_comment,
      sym_groovy_doc,
  [82774] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4193), 1,
      anon_sym_while,
    STATE(2362), 2,
      sym_comment,
      sym_groovy_doc,
  [82791] = 5,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4195), 1,
      aux_sym_shebang_token1,
    STATE(2363), 2,
      sym_comment,
      sym_groovy_doc,
  [82808] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4197), 1,
      sym_identifier,
    STATE(2364), 2,
      sym_comment,
      sym_groovy_doc,
  [82825] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4199), 1,
      sym_identifier,
    STATE(2365), 2,
      sym_comment,
      sym_groovy_doc,
  [82842] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4201), 1,
      sym_identifier,
    STATE(2366), 2,
      sym_comment,
      sym_groovy_doc,
  [82859] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4203), 1,
      anon_sym_COMMA,
    STATE(2367), 2,
      sym_comment,
      sym_groovy_doc,
  [82876] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4205), 1,
      sym_identifier,
    STATE(2368), 2,
      sym_comment,
      sym_groovy_doc,
  [82893] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4207), 1,
      sym_identifier,
    STATE(2369), 2,
      sym_comment,
      sym_groovy_doc,
  [82910] = 5,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4209), 1,
      aux_sym_groovy_doc_token1,
    STATE(2370), 2,
      sym_comment,
      sym_groovy_doc,
  [82927] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4211), 1,
      anon_sym_COLON,
    STATE(2371), 2,
      sym_comment,
      sym_groovy_doc,
  [82944] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2721), 1,
      anon_sym_RPAREN,
    STATE(2372), 2,
      sym_comment,
      sym_groovy_doc,
  [82961] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4213), 1,
      sym_identifier,
    STATE(2373), 2,
      sym_comment,
      sym_groovy_doc,
  [82978] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4215), 1,
      anon_sym_DASH_GT,
    STATE(2374), 2,
      sym_comment,
      sym_groovy_doc,
  [82995] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4217), 1,
      sym_identifier,
    STATE(2375), 2,
      sym_comment,
      sym_groovy_doc,
  [83012] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4219), 1,
      anon_sym_DASH_GT,
    STATE(2376), 2,
      sym_comment,
      sym_groovy_doc,
  [83029] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4221), 1,
      anon_sym_while,
    STATE(2377), 2,
      sym_comment,
      sym_groovy_doc,
  [83046] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2689), 1,
      anon_sym_RPAREN,
    STATE(2378), 2,
      sym_comment,
      sym_groovy_doc,
  [83063] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2719), 1,
      anon_sym_RPAREN,
    STATE(2379), 2,
      sym_comment,
      sym_groovy_doc,
  [83080] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4223), 1,
      sym_identifier,
    STATE(2380), 2,
      sym_comment,
      sym_groovy_doc,
  [83097] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4225), 1,
      anon_sym_DASH_GT,
    STATE(2381), 2,
      sym_comment,
      sym_groovy_doc,
  [83114] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4227), 1,
      anon_sym_LPAREN,
    STATE(2382), 2,
      sym_comment,
      sym_groovy_doc,
  [83131] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4105), 1,
      sym_identifier,
    STATE(2383), 2,
      sym_comment,
      sym_groovy_doc,
  [83148] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2751), 1,
      anon_sym_RPAREN,
    STATE(2384), 2,
      sym_comment,
      sym_groovy_doc,
  [83165] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4229), 1,
      anon_sym_DASH_GT,
    STATE(2385), 2,
      sym_comment,
      sym_groovy_doc,
  [83182] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4133), 1,
      anon_sym_RBRACK,
    STATE(2386), 2,
      sym_comment,
      sym_groovy_doc,
  [83199] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4231), 1,
      ts_builtin_sym_end,
    STATE(2387), 2,
      sym_comment,
      sym_groovy_doc,
  [83216] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(777), 1,
      ts_builtin_sym_end,
    STATE(2388), 2,
      sym_comment,
      sym_groovy_doc,
  [83233] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4233), 1,
      anon_sym_DASH_GT,
    STATE(2389), 2,
      sym_comment,
      sym_groovy_doc,
  [83250] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4235), 1,
      sym_identifier,
    STATE(2390), 2,
      sym_comment,
      sym_groovy_doc,
  [83267] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4237), 1,
      sym_identifier,
    STATE(2391), 2,
      sym_comment,
      sym_groovy_doc,
  [83284] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3736), 1,
      sym_identifier,
    STATE(2392), 2,
      sym_comment,
      sym_groovy_doc,
  [83301] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4239), 1,
      anon_sym_DASH_GT,
    STATE(2393), 2,
      sym_comment,
      sym_groovy_doc,
  [83318] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2685), 1,
      anon_sym_RPAREN,
    STATE(2394), 2,
      sym_comment,
      sym_groovy_doc,
  [83335] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2677), 1,
      anon_sym_RPAREN,
    STATE(2395), 2,
      sym_comment,
      sym_groovy_doc,
  [83352] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4241), 1,
      anon_sym_COMMA,
    STATE(2396), 2,
      sym_comment,
      sym_groovy_doc,
  [83369] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4243), 1,
      anon_sym_DASH_GT,
    STATE(2397), 2,
      sym_comment,
      sym_groovy_doc,
  [83386] = 5,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(3305), 1,
      aux_sym_comment_token1,
    ACTIONS(3307), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4245), 1,
      aux_sym_comment_token2,
    STATE(2398), 2,
      sym_comment,
      sym_groovy_doc,
  [83403] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(3930), 1,
      sym_escape_sequence,
    STATE(2399), 2,
      sym_comment,
      sym_groovy_doc,
  [83420] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2759), 1,
      anon_sym_RPAREN,
    STATE(2400), 2,
      sym_comment,
      sym_groovy_doc,
  [83437] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4085), 1,
      anon_sym_RBRACK,
    STATE(2401), 2,
      sym_comment,
      sym_groovy_doc,
  [83454] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4247), 1,
      sym_identifier,
    STATE(2402), 2,
      sym_comment,
      sym_groovy_doc,
  [83471] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4249), 1,
      sym_identifier,
    STATE(2403), 2,
      sym_comment,
      sym_groovy_doc,
  [83488] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2763), 1,
      anon_sym_RPAREN,
    STATE(2404), 2,
      sym_comment,
      sym_groovy_doc,
  [83505] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4061), 1,
      anon_sym_RBRACK,
    STATE(2405), 2,
      sym_comment,
      sym_groovy_doc,
  [83522] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4251), 1,
      anon_sym_while,
    STATE(2406), 2,
      sym_comment,
      sym_groovy_doc,
  [83539] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2679), 1,
      anon_sym_RPAREN,
    STATE(2407), 2,
      sym_comment,
      sym_groovy_doc,
  [83556] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2765), 1,
      anon_sym_RPAREN,
    STATE(2408), 2,
      sym_comment,
      sym_groovy_doc,
  [83573] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4253), 1,
      sym_identifier,
    STATE(2409), 2,
      sym_comment,
      sym_groovy_doc,
  [83590] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4255), 1,
      sym_identifier,
    STATE(2410), 2,
      sym_comment,
      sym_groovy_doc,
  [83607] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2681), 1,
      anon_sym_RPAREN,
    STATE(2411), 2,
      sym_comment,
      sym_groovy_doc,
  [83624] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4257), 1,
      anon_sym_while,
    STATE(2412), 2,
      sym_comment,
      sym_groovy_doc,
  [83641] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4259), 1,
      sym_identifier,
    STATE(2413), 2,
      sym_comment,
      sym_groovy_doc,
  [83658] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4261), 1,
      anon_sym_in,
    STATE(2414), 2,
      sym_comment,
      sym_groovy_doc,
  [83675] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2639), 1,
      anon_sym_RPAREN,
    STATE(2415), 2,
      sym_comment,
      sym_groovy_doc,
  [83692] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4263), 1,
      sym_identifier,
    STATE(2416), 2,
      sym_comment,
      sym_groovy_doc,
  [83709] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2755), 1,
      anon_sym_RPAREN,
    STATE(2417), 2,
      sym_comment,
      sym_groovy_doc,
  [83726] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2707), 1,
      anon_sym_RPAREN,
    STATE(2418), 2,
      sym_comment,
      sym_groovy_doc,
  [83743] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4265), 1,
      sym_identifier,
    STATE(2419), 2,
      sym_comment,
      sym_groovy_doc,
  [83760] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4267), 1,
      anon_sym_while,
    STATE(2420), 2,
      sym_comment,
      sym_groovy_doc,
  [83777] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4269), 1,
      sym_identifier,
    STATE(2421), 2,
      sym_comment,
      sym_groovy_doc,
  [83794] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4271), 1,
      anon_sym_in,
    STATE(2422), 2,
      sym_comment,
      sym_groovy_doc,
  [83811] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4273), 1,
      sym_identifier,
    STATE(2423), 2,
      sym_comment,
      sym_groovy_doc,
  [83828] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4275), 1,
      sym_identifier,
    STATE(2424), 2,
      sym_comment,
      sym_groovy_doc,
  [83845] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4277), 1,
      anon_sym_in,
    STATE(2425), 2,
      sym_comment,
      sym_groovy_doc,
  [83862] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4279), 1,
      anon_sym_RPAREN,
    STATE(2426), 2,
      sym_comment,
      sym_groovy_doc,
  [83879] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4281), 1,
      sym_identifier,
    STATE(2427), 2,
      sym_comment,
      sym_groovy_doc,
  [83896] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4283), 1,
      anon_sym_in,
    STATE(2428), 2,
      sym_comment,
      sym_groovy_doc,
  [83913] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4285), 1,
      aux_sym_annotation_token1,
    STATE(2429), 2,
      sym_comment,
      sym_groovy_doc,
  [83930] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4287), 1,
      anon_sym_in,
    STATE(2430), 2,
      sym_comment,
      sym_groovy_doc,
  [83947] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4289), 1,
      anon_sym_in,
    STATE(2431), 2,
      sym_comment,
      sym_groovy_doc,
  [83964] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4291), 1,
      anon_sym_in,
    STATE(2432), 2,
      sym_comment,
      sym_groovy_doc,
  [83981] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4293), 1,
      anon_sym_in,
    STATE(2433), 2,
      sym_comment,
      sym_groovy_doc,
  [83998] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4295), 1,
      anon_sym_in,
    STATE(2434), 2,
      sym_comment,
      sym_groovy_doc,
  [84015] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4297), 1,
      anon_sym_in,
    STATE(2435), 2,
      sym_comment,
      sym_groovy_doc,
  [84032] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4299), 1,
      anon_sym_in,
    STATE(2436), 2,
      sym_comment,
      sym_groovy_doc,
  [84049] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4301), 1,
      anon_sym_in,
    STATE(2437), 2,
      sym_comment,
      sym_groovy_doc,
  [84066] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2665), 1,
      anon_sym_RPAREN,
    STATE(2438), 2,
      sym_comment,
      sym_groovy_doc,
  [84083] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4303), 1,
      anon_sym_in,
    STATE(2439), 2,
      sym_comment,
      sym_groovy_doc,
  [84100] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4305), 1,
      sym_identifier,
    STATE(2440), 2,
      sym_comment,
      sym_groovy_doc,
  [84117] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4307), 1,
      sym_identifier,
    STATE(2441), 2,
      sym_comment,
      sym_groovy_doc,
  [84134] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4309), 1,
      sym_identifier,
    STATE(2442), 2,
      sym_comment,
      sym_groovy_doc,
  [84151] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2667), 1,
      anon_sym_RPAREN,
    STATE(2443), 2,
      sym_comment,
      sym_groovy_doc,
  [84168] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4311), 1,
      sym_identifier,
    STATE(2444), 2,
      sym_comment,
      sym_groovy_doc,
  [84185] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4313), 1,
      anon_sym_LPAREN,
    STATE(2445), 2,
      sym_comment,
      sym_groovy_doc,
  [84202] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4315), 1,
      sym_identifier,
    STATE(2446), 2,
      sym_comment,
      sym_groovy_doc,
  [84219] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4317), 1,
      anon_sym_DASH_GT,
    STATE(2447), 2,
      sym_comment,
      sym_groovy_doc,
  [84236] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2723), 1,
      anon_sym_RPAREN,
    STATE(2448), 2,
      sym_comment,
      sym_groovy_doc,
  [84253] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2733), 1,
      anon_sym_COMMA,
    STATE(2449), 2,
      sym_comment,
      sym_groovy_doc,
  [84270] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4319), 1,
      sym_identifier,
    STATE(2450), 2,
      sym_comment,
      sym_groovy_doc,
  [84287] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4321), 1,
      sym_identifier,
    STATE(2451), 2,
      sym_comment,
      sym_groovy_doc,
  [84304] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4323), 1,
      sym_identifier,
    STATE(2452), 2,
      sym_comment,
      sym_groovy_doc,
  [84321] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4325), 1,
      anon_sym_RPAREN,
    STATE(2453), 2,
      sym_comment,
      sym_groovy_doc,
  [84338] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4327), 1,
      aux_sym_annotation_token1,
    STATE(2454), 2,
      sym_comment,
      sym_groovy_doc,
  [84355] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4329), 1,
      anon_sym_DASH_GT,
    STATE(2455), 2,
      sym_comment,
      sym_groovy_doc,
  [84372] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_RPAREN,
    STATE(2456), 2,
      sym_comment,
      sym_groovy_doc,
  [84389] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4331), 1,
      anon_sym_in,
    STATE(2457), 2,
      sym_comment,
      sym_groovy_doc,
  [84406] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4333), 1,
      anon_sym_LPAREN,
    STATE(2458), 2,
      sym_comment,
      sym_groovy_doc,
  [84423] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4111), 1,
      anon_sym_RBRACK,
    STATE(2459), 2,
      sym_comment,
      sym_groovy_doc,
  [84440] = 5,
    ACTIONS(3), 1,
      aux_sym_comment_token1,
    ACTIONS(5), 1,
      anon_sym_SLASH_STAR,
    ACTIONS(7), 1,
      anon_sym_SLASH_STAR_STAR,
    ACTIONS(4335), 1,
      sym_identifier,
    STATE(2460), 2,
      sym_comment,
      sym_groovy_doc,
  [84457] = 1,
    ACTIONS(4337), 1,
      ts_builtin_sym_end,
  [84461] = 1,
    ACTIONS(4339), 1,
      ts_builtin_sym_end,
  [84465] = 1,
    ACTIONS(4341), 1,
      ts_builtin_sym_end,
  [84469] = 1,
    ACTIONS(4343), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(1025)] = 0,
  [SMALL_STATE(1026)] = 138,
  [SMALL_STATE(1027)] = 218,
  [SMALL_STATE(1028)] = 298,
  [SMALL_STATE(1029)] = 378,
  [SMALL_STATE(1030)] = 458,
  [SMALL_STATE(1031)] = 537,
  [SMALL_STATE(1032)] = 614,
  [SMALL_STATE(1033)] = 693,
  [SMALL_STATE(1034)] = 770,
  [SMALL_STATE(1035)] = 916,
  [SMALL_STATE(1036)] = 1062,
  [SMALL_STATE(1037)] = 1136,
  [SMALL_STATE(1038)] = 1210,
  [SMALL_STATE(1039)] = 1288,
  [SMALL_STATE(1040)] = 1366,
  [SMALL_STATE(1041)] = 1440,
  [SMALL_STATE(1042)] = 1586,
  [SMALL_STATE(1043)] = 1660,
  [SMALL_STATE(1044)] = 1738,
  [SMALL_STATE(1045)] = 1812,
  [SMALL_STATE(1046)] = 1894,
  [SMALL_STATE(1047)] = 2040,
  [SMALL_STATE(1048)] = 2118,
  [SMALL_STATE(1049)] = 2192,
  [SMALL_STATE(1050)] = 2266,
  [SMALL_STATE(1051)] = 2340,
  [SMALL_STATE(1052)] = 2486,
  [SMALL_STATE(1053)] = 2560,
  [SMALL_STATE(1054)] = 2634,
  [SMALL_STATE(1055)] = 2708,
  [SMALL_STATE(1056)] = 2782,
  [SMALL_STATE(1057)] = 2856,
  [SMALL_STATE(1058)] = 2930,
  [SMALL_STATE(1059)] = 3004,
  [SMALL_STATE(1060)] = 3150,
  [SMALL_STATE(1061)] = 3223,
  [SMALL_STATE(1062)] = 3296,
  [SMALL_STATE(1063)] = 3369,
  [SMALL_STATE(1064)] = 3446,
  [SMALL_STATE(1065)] = 3519,
  [SMALL_STATE(1066)] = 3592,
  [SMALL_STATE(1067)] = 3665,
  [SMALL_STATE(1068)] = 3738,
  [SMALL_STATE(1069)] = 3811,
  [SMALL_STATE(1070)] = 3884,
  [SMALL_STATE(1071)] = 3957,
  [SMALL_STATE(1072)] = 4030,
  [SMALL_STATE(1073)] = 4103,
  [SMALL_STATE(1074)] = 4176,
  [SMALL_STATE(1075)] = 4249,
  [SMALL_STATE(1076)] = 4330,
  [SMALL_STATE(1077)] = 4403,
  [SMALL_STATE(1078)] = 4478,
  [SMALL_STATE(1079)] = 4555,
  [SMALL_STATE(1080)] = 4630,
  [SMALL_STATE(1081)] = 4770,
  [SMALL_STATE(1082)] = 4842,
  [SMALL_STATE(1083)] = 4914,
  [SMALL_STATE(1084)] = 5054,
  [SMALL_STATE(1085)] = 5126,
  [SMALL_STATE(1086)] = 5266,
  [SMALL_STATE(1087)] = 5406,
  [SMALL_STATE(1088)] = 5478,
  [SMALL_STATE(1089)] = 5556,
  [SMALL_STATE(1090)] = 5696,
  [SMALL_STATE(1091)] = 5768,
  [SMALL_STATE(1092)] = 5840,
  [SMALL_STATE(1093)] = 5912,
  [SMALL_STATE(1094)] = 5984,
  [SMALL_STATE(1095)] = 6124,
  [SMALL_STATE(1096)] = 6196,
  [SMALL_STATE(1097)] = 6268,
  [SMALL_STATE(1098)] = 6408,
  [SMALL_STATE(1099)] = 6480,
  [SMALL_STATE(1100)] = 6552,
  [SMALL_STATE(1101)] = 6692,
  [SMALL_STATE(1102)] = 6764,
  [SMALL_STATE(1103)] = 6836,
  [SMALL_STATE(1104)] = 6908,
  [SMALL_STATE(1105)] = 6980,
  [SMALL_STATE(1106)] = 7115,
  [SMALL_STATE(1107)] = 7248,
  [SMALL_STATE(1108)] = 7319,
  [SMALL_STATE(1109)] = 7396,
  [SMALL_STATE(1110)] = 7467,
  [SMALL_STATE(1111)] = 7598,
  [SMALL_STATE(1112)] = 7731,
  [SMALL_STATE(1113)] = 7802,
  [SMALL_STATE(1114)] = 7873,
  [SMALL_STATE(1115)] = 8006,
  [SMALL_STATE(1116)] = 8077,
  [SMALL_STATE(1117)] = 8212,
  [SMALL_STATE(1118)] = 8283,
  [SMALL_STATE(1119)] = 8354,
  [SMALL_STATE(1120)] = 8487,
  [SMALL_STATE(1121)] = 8558,
  [SMALL_STATE(1122)] = 8695,
  [SMALL_STATE(1123)] = 8830,
  [SMALL_STATE(1124)] = 8967,
  [SMALL_STATE(1125)] = 9038,
  [SMALL_STATE(1126)] = 9109,
  [SMALL_STATE(1127)] = 9246,
  [SMALL_STATE(1128)] = 9317,
  [SMALL_STATE(1129)] = 9452,
  [SMALL_STATE(1130)] = 9587,
  [SMALL_STATE(1131)] = 9724,
  [SMALL_STATE(1132)] = 9857,
  [SMALL_STATE(1133)] = 9992,
  [SMALL_STATE(1134)] = 10063,
  [SMALL_STATE(1135)] = 10198,
  [SMALL_STATE(1136)] = 10335,
  [SMALL_STATE(1137)] = 10470,
  [SMALL_STATE(1138)] = 10607,
  [SMALL_STATE(1139)] = 10686,
  [SMALL_STATE(1140)] = 10823,
  [SMALL_STATE(1141)] = 10958,
  [SMALL_STATE(1142)] = 11095,
  [SMALL_STATE(1143)] = 11230,
  [SMALL_STATE(1144)] = 11365,
  [SMALL_STATE(1145)] = 11500,
  [SMALL_STATE(1146)] = 11635,
  [SMALL_STATE(1147)] = 11770,
  [SMALL_STATE(1148)] = 11905,
  [SMALL_STATE(1149)] = 12040,
  [SMALL_STATE(1150)] = 12175,
  [SMALL_STATE(1151)] = 12310,
  [SMALL_STATE(1152)] = 12381,
  [SMALL_STATE(1153)] = 12456,
  [SMALL_STATE(1154)] = 12527,
  [SMALL_STATE(1155)] = 12660,
  [SMALL_STATE(1156)] = 12731,
  [SMALL_STATE(1157)] = 12803,
  [SMALL_STATE(1158)] = 12875,
  [SMALL_STATE(1159)] = 13011,
  [SMALL_STATE(1160)] = 13083,
  [SMALL_STATE(1161)] = 13213,
  [SMALL_STATE(1162)] = 13343,
  [SMALL_STATE(1163)] = 13419,
  [SMALL_STATE(1164)] = 13493,
  [SMALL_STATE(1165)] = 13567,
  [SMALL_STATE(1166)] = 13639,
  [SMALL_STATE(1167)] = 13769,
  [SMALL_STATE(1168)] = 13845,
  [SMALL_STATE(1169)] = 13981,
  [SMALL_STATE(1170)] = 14059,
  [SMALL_STATE(1171)] = 14195,
  [SMALL_STATE(1172)] = 14331,
  [SMALL_STATE(1173)] = 14405,
  [SMALL_STATE(1174)] = 14481,
  [SMALL_STATE(1175)] = 14559,
  [SMALL_STATE(1176)] = 14695,
  [SMALL_STATE(1177)] = 14831,
  [SMALL_STATE(1178)] = 14958,
  [SMALL_STATE(1179)] = 15085,
  [SMALL_STATE(1180)] = 15212,
  [SMALL_STATE(1181)] = 15339,
  [SMALL_STATE(1182)] = 15466,
  [SMALL_STATE(1183)] = 15593,
  [SMALL_STATE(1184)] = 15720,
  [SMALL_STATE(1185)] = 15791,
  [SMALL_STATE(1186)] = 15918,
  [SMALL_STATE(1187)] = 16045,
  [SMALL_STATE(1188)] = 16172,
  [SMALL_STATE(1189)] = 16299,
  [SMALL_STATE(1190)] = 16404,
  [SMALL_STATE(1191)] = 16531,
  [SMALL_STATE(1192)] = 16658,
  [SMALL_STATE(1193)] = 16785,
  [SMALL_STATE(1194)] = 16912,
  [SMALL_STATE(1195)] = 17039,
  [SMALL_STATE(1196)] = 17166,
  [SMALL_STATE(1197)] = 17293,
  [SMALL_STATE(1198)] = 17420,
  [SMALL_STATE(1199)] = 17547,
  [SMALL_STATE(1200)] = 17674,
  [SMALL_STATE(1201)] = 17801,
  [SMALL_STATE(1202)] = 17928,
  [SMALL_STATE(1203)] = 18055,
  [SMALL_STATE(1204)] = 18182,
  [SMALL_STATE(1205)] = 18309,
  [SMALL_STATE(1206)] = 18436,
  [SMALL_STATE(1207)] = 18563,
  [SMALL_STATE(1208)] = 18690,
  [SMALL_STATE(1209)] = 18817,
  [SMALL_STATE(1210)] = 18944,
  [SMALL_STATE(1211)] = 19071,
  [SMALL_STATE(1212)] = 19198,
  [SMALL_STATE(1213)] = 19325,
  [SMALL_STATE(1214)] = 19452,
  [SMALL_STATE(1215)] = 19579,
  [SMALL_STATE(1216)] = 19706,
  [SMALL_STATE(1217)] = 19833,
  [SMALL_STATE(1218)] = 19960,
  [SMALL_STATE(1219)] = 20087,
  [SMALL_STATE(1220)] = 20214,
  [SMALL_STATE(1221)] = 20341,
  [SMALL_STATE(1222)] = 20468,
  [SMALL_STATE(1223)] = 20595,
  [SMALL_STATE(1224)] = 20722,
  [SMALL_STATE(1225)] = 20849,
  [SMALL_STATE(1226)] = 20976,
  [SMALL_STATE(1227)] = 21103,
  [SMALL_STATE(1228)] = 21230,
  [SMALL_STATE(1229)] = 21357,
  [SMALL_STATE(1230)] = 21484,
  [SMALL_STATE(1231)] = 21611,
  [SMALL_STATE(1232)] = 21738,
  [SMALL_STATE(1233)] = 21865,
  [SMALL_STATE(1234)] = 21992,
  [SMALL_STATE(1235)] = 22119,
  [SMALL_STATE(1236)] = 22246,
  [SMALL_STATE(1237)] = 22373,
  [SMALL_STATE(1238)] = 22500,
  [SMALL_STATE(1239)] = 22627,
  [SMALL_STATE(1240)] = 22754,
  [SMALL_STATE(1241)] = 22881,
  [SMALL_STATE(1242)] = 23008,
  [SMALL_STATE(1243)] = 23135,
  [SMALL_STATE(1244)] = 23262,
  [SMALL_STATE(1245)] = 23389,
  [SMALL_STATE(1246)] = 23516,
  [SMALL_STATE(1247)] = 23643,
  [SMALL_STATE(1248)] = 23770,
  [SMALL_STATE(1249)] = 23897,
  [SMALL_STATE(1250)] = 24024,
  [SMALL_STATE(1251)] = 24151,
  [SMALL_STATE(1252)] = 24278,
  [SMALL_STATE(1253)] = 24405,
  [SMALL_STATE(1254)] = 24532,
  [SMALL_STATE(1255)] = 24659,
  [SMALL_STATE(1256)] = 24786,
  [SMALL_STATE(1257)] = 24913,
  [SMALL_STATE(1258)] = 25040,
  [SMALL_STATE(1259)] = 25167,
  [SMALL_STATE(1260)] = 25294,
  [SMALL_STATE(1261)] = 25421,
  [SMALL_STATE(1262)] = 25548,
  [SMALL_STATE(1263)] = 25675,
  [SMALL_STATE(1264)] = 25802,
  [SMALL_STATE(1265)] = 25929,
  [SMALL_STATE(1266)] = 26056,
  [SMALL_STATE(1267)] = 26183,
  [SMALL_STATE(1268)] = 26310,
  [SMALL_STATE(1269)] = 26437,
  [SMALL_STATE(1270)] = 26564,
  [SMALL_STATE(1271)] = 26691,
  [SMALL_STATE(1272)] = 26818,
  [SMALL_STATE(1273)] = 26945,
  [SMALL_STATE(1274)] = 27072,
  [SMALL_STATE(1275)] = 27199,
  [SMALL_STATE(1276)] = 27326,
  [SMALL_STATE(1277)] = 27453,
  [SMALL_STATE(1278)] = 27580,
  [SMALL_STATE(1279)] = 27707,
  [SMALL_STATE(1280)] = 27834,
  [SMALL_STATE(1281)] = 27961,
  [SMALL_STATE(1282)] = 28088,
  [SMALL_STATE(1283)] = 28215,
  [SMALL_STATE(1284)] = 28284,
  [SMALL_STATE(1285)] = 28411,
  [SMALL_STATE(1286)] = 28538,
  [SMALL_STATE(1287)] = 28665,
  [SMALL_STATE(1288)] = 28792,
  [SMALL_STATE(1289)] = 28919,
  [SMALL_STATE(1290)] = 29046,
  [SMALL_STATE(1291)] = 29173,
  [SMALL_STATE(1292)] = 29300,
  [SMALL_STATE(1293)] = 29427,
  [SMALL_STATE(1294)] = 29554,
  [SMALL_STATE(1295)] = 29681,
  [SMALL_STATE(1296)] = 29808,
  [SMALL_STATE(1297)] = 29935,
  [SMALL_STATE(1298)] = 30062,
  [SMALL_STATE(1299)] = 30189,
  [SMALL_STATE(1300)] = 30316,
  [SMALL_STATE(1301)] = 30443,
  [SMALL_STATE(1302)] = 30570,
  [SMALL_STATE(1303)] = 30697,
  [SMALL_STATE(1304)] = 30772,
  [SMALL_STATE(1305)] = 30899,
  [SMALL_STATE(1306)] = 30968,
  [SMALL_STATE(1307)] = 31095,
  [SMALL_STATE(1308)] = 31168,
  [SMALL_STATE(1309)] = 31237,
  [SMALL_STATE(1310)] = 31364,
  [SMALL_STATE(1311)] = 31491,
  [SMALL_STATE(1312)] = 31618,
  [SMALL_STATE(1313)] = 31745,
  [SMALL_STATE(1314)] = 31872,
  [SMALL_STATE(1315)] = 31999,
  [SMALL_STATE(1316)] = 32126,
  [SMALL_STATE(1317)] = 32253,
  [SMALL_STATE(1318)] = 32380,
  [SMALL_STATE(1319)] = 32507,
  [SMALL_STATE(1320)] = 32582,
  [SMALL_STATE(1321)] = 32709,
  [SMALL_STATE(1322)] = 32836,
  [SMALL_STATE(1323)] = 32963,
  [SMALL_STATE(1324)] = 33090,
  [SMALL_STATE(1325)] = 33217,
  [SMALL_STATE(1326)] = 33344,
  [SMALL_STATE(1327)] = 33471,
  [SMALL_STATE(1328)] = 33598,
  [SMALL_STATE(1329)] = 33725,
  [SMALL_STATE(1330)] = 33794,
  [SMALL_STATE(1331)] = 33867,
  [SMALL_STATE(1332)] = 33944,
  [SMALL_STATE(1333)] = 34071,
  [SMALL_STATE(1334)] = 34198,
  [SMALL_STATE(1335)] = 34325,
  [SMALL_STATE(1336)] = 34452,
  [SMALL_STATE(1337)] = 34579,
  [SMALL_STATE(1338)] = 34706,
  [SMALL_STATE(1339)] = 34833,
  [SMALL_STATE(1340)] = 34960,
  [SMALL_STATE(1341)] = 35087,
  [SMALL_STATE(1342)] = 35214,
  [SMALL_STATE(1343)] = 35341,
  [SMALL_STATE(1344)] = 35468,
  [SMALL_STATE(1345)] = 35595,
  [SMALL_STATE(1346)] = 35722,
  [SMALL_STATE(1347)] = 35849,
  [SMALL_STATE(1348)] = 35976,
  [SMALL_STATE(1349)] = 36103,
  [SMALL_STATE(1350)] = 36230,
  [SMALL_STATE(1351)] = 36357,
  [SMALL_STATE(1352)] = 36484,
  [SMALL_STATE(1353)] = 36611,
  [SMALL_STATE(1354)] = 36738,
  [SMALL_STATE(1355)] = 36865,
  [SMALL_STATE(1356)] = 36992,
  [SMALL_STATE(1357)] = 37119,
  [SMALL_STATE(1358)] = 37188,
  [SMALL_STATE(1359)] = 37315,
  [SMALL_STATE(1360)] = 37442,
  [SMALL_STATE(1361)] = 37515,
  [SMALL_STATE(1362)] = 37642,
  [SMALL_STATE(1363)] = 37769,
  [SMALL_STATE(1364)] = 37854,
  [SMALL_STATE(1365)] = 37981,
  [SMALL_STATE(1366)] = 38108,
  [SMALL_STATE(1367)] = 38185,
  [SMALL_STATE(1368)] = 38312,
  [SMALL_STATE(1369)] = 38393,
  [SMALL_STATE(1370)] = 38520,
  [SMALL_STATE(1371)] = 38609,
  [SMALL_STATE(1372)] = 38736,
  [SMALL_STATE(1373)] = 38829,
  [SMALL_STATE(1374)] = 38924,
  [SMALL_STATE(1375)] = 39021,
  [SMALL_STATE(1376)] = 39120,
  [SMALL_STATE(1377)] = 39247,
  [SMALL_STATE(1378)] = 39348,
  [SMALL_STATE(1379)] = 39475,
  [SMALL_STATE(1380)] = 39578,
  [SMALL_STATE(1381)] = 39651,
  [SMALL_STATE(1382)] = 39778,
  [SMALL_STATE(1383)] = 39905,
  [SMALL_STATE(1384)] = 39981,
  [SMALL_STATE(1385)] = 40055,
  [SMALL_STATE(1386)] = 40129,
  [SMALL_STATE(1387)] = 40201,
  [SMALL_STATE(1388)] = 40273,
  [SMALL_STATE(1389)] = 40347,
  [SMALL_STATE(1390)] = 40421,
  [SMALL_STATE(1391)] = 40492,
  [SMALL_STATE(1392)] = 40563,
  [SMALL_STATE(1393)] = 40632,
  [SMALL_STATE(1394)] = 40701,
  [SMALL_STATE(1395)] = 40772,
  [SMALL_STATE(1396)] = 40847,
  [SMALL_STATE(1397)] = 40920,
  [SMALL_STATE(1398)] = 41020,
  [SMALL_STATE(1399)] = 41088,
  [SMALL_STATE(1400)] = 41155,
  [SMALL_STATE(1401)] = 41254,
  [SMALL_STATE(1402)] = 41325,
  [SMALL_STATE(1403)] = 41424,
  [SMALL_STATE(1404)] = 41523,
  [SMALL_STATE(1405)] = 41622,
  [SMALL_STATE(1406)] = 41689,
  [SMALL_STATE(1407)] = 41788,
  [SMALL_STATE(1408)] = 41859,
  [SMALL_STATE(1409)] = 41930,
  [SMALL_STATE(1410)] = 42029,
  [SMALL_STATE(1411)] = 42098,
  [SMALL_STATE(1412)] = 42165,
  [SMALL_STATE(1413)] = 42264,
  [SMALL_STATE(1414)] = 42331,
  [SMALL_STATE(1415)] = 42398,
  [SMALL_STATE(1416)] = 42469,
  [SMALL_STATE(1417)] = 42536,
  [SMALL_STATE(1418)] = 42635,
  [SMALL_STATE(1419)] = 42702,
  [SMALL_STATE(1420)] = 42801,
  [SMALL_STATE(1421)] = 42900,
  [SMALL_STATE(1422)] = 42965,
  [SMALL_STATE(1423)] = 43036,
  [SMALL_STATE(1424)] = 43103,
  [SMALL_STATE(1425)] = 43202,
  [SMALL_STATE(1426)] = 43301,
  [SMALL_STATE(1427)] = 43400,
  [SMALL_STATE(1428)] = 43468,
  [SMALL_STATE(1429)] = 43534,
  [SMALL_STATE(1430)] = 43600,
  [SMALL_STATE(1431)] = 43668,
  [SMALL_STATE(1432)] = 43734,
  [SMALL_STATE(1433)] = 43802,
  [SMALL_STATE(1434)] = 43868,
  [SMALL_STATE(1435)] = 43936,
  [SMALL_STATE(1436)] = 44005,
  [SMALL_STATE(1437)] = 44076,
  [SMALL_STATE(1438)] = 44147,
  [SMALL_STATE(1439)] = 44214,
  [SMALL_STATE(1440)] = 44285,
  [SMALL_STATE(1441)] = 44356,
  [SMALL_STATE(1442)] = 44422,
  [SMALL_STATE(1443)] = 44520,
  [SMALL_STATE(1444)] = 44618,
  [SMALL_STATE(1445)] = 44715,
  [SMALL_STATE(1446)] = 44812,
  [SMALL_STATE(1447)] = 44877,
  [SMALL_STATE(1448)] = 44942,
  [SMALL_STATE(1449)] = 45038,
  [SMALL_STATE(1450)] = 45134,
  [SMALL_STATE(1451)] = 45230,
  [SMALL_STATE(1452)] = 45326,
  [SMALL_STATE(1453)] = 45422,
  [SMALL_STATE(1454)] = 45518,
  [SMALL_STATE(1455)] = 45614,
  [SMALL_STATE(1456)] = 45710,
  [SMALL_STATE(1457)] = 45806,
  [SMALL_STATE(1458)] = 45902,
  [SMALL_STATE(1459)] = 45998,
  [SMALL_STATE(1460)] = 46094,
  [SMALL_STATE(1461)] = 46190,
  [SMALL_STATE(1462)] = 46254,
  [SMALL_STATE(1463)] = 46350,
  [SMALL_STATE(1464)] = 46446,
  [SMALL_STATE(1465)] = 46542,
  [SMALL_STATE(1466)] = 46638,
  [SMALL_STATE(1467)] = 46734,
  [SMALL_STATE(1468)] = 46830,
  [SMALL_STATE(1469)] = 46926,
  [SMALL_STATE(1470)] = 47022,
  [SMALL_STATE(1471)] = 47118,
  [SMALL_STATE(1472)] = 47214,
  [SMALL_STATE(1473)] = 47310,
  [SMALL_STATE(1474)] = 47406,
  [SMALL_STATE(1475)] = 47502,
  [SMALL_STATE(1476)] = 47598,
  [SMALL_STATE(1477)] = 47694,
  [SMALL_STATE(1478)] = 47787,
  [SMALL_STATE(1479)] = 47880,
  [SMALL_STATE(1480)] = 47973,
  [SMALL_STATE(1481)] = 48066,
  [SMALL_STATE(1482)] = 48159,
  [SMALL_STATE(1483)] = 48252,
  [SMALL_STATE(1484)] = 48345,
  [SMALL_STATE(1485)] = 48438,
  [SMALL_STATE(1486)] = 48531,
  [SMALL_STATE(1487)] = 48624,
  [SMALL_STATE(1488)] = 48717,
  [SMALL_STATE(1489)] = 48810,
  [SMALL_STATE(1490)] = 48903,
  [SMALL_STATE(1491)] = 48996,
  [SMALL_STATE(1492)] = 49089,
  [SMALL_STATE(1493)] = 49182,
  [SMALL_STATE(1494)] = 49275,
  [SMALL_STATE(1495)] = 49368,
  [SMALL_STATE(1496)] = 49461,
  [SMALL_STATE(1497)] = 49554,
  [SMALL_STATE(1498)] = 49647,
  [SMALL_STATE(1499)] = 49740,
  [SMALL_STATE(1500)] = 49833,
  [SMALL_STATE(1501)] = 49926,
  [SMALL_STATE(1502)] = 50019,
  [SMALL_STATE(1503)] = 50112,
  [SMALL_STATE(1504)] = 50205,
  [SMALL_STATE(1505)] = 50298,
  [SMALL_STATE(1506)] = 50391,
  [SMALL_STATE(1507)] = 50484,
  [SMALL_STATE(1508)] = 50577,
  [SMALL_STATE(1509)] = 50670,
  [SMALL_STATE(1510)] = 50763,
  [SMALL_STATE(1511)] = 50856,
  [SMALL_STATE(1512)] = 50949,
  [SMALL_STATE(1513)] = 51042,
  [SMALL_STATE(1514)] = 51135,
  [SMALL_STATE(1515)] = 51228,
  [SMALL_STATE(1516)] = 51321,
  [SMALL_STATE(1517)] = 51414,
  [SMALL_STATE(1518)] = 51507,
  [SMALL_STATE(1519)] = 51600,
  [SMALL_STATE(1520)] = 51693,
  [SMALL_STATE(1521)] = 51786,
  [SMALL_STATE(1522)] = 51879,
  [SMALL_STATE(1523)] = 51972,
  [SMALL_STATE(1524)] = 52065,
  [SMALL_STATE(1525)] = 52158,
  [SMALL_STATE(1526)] = 52251,
  [SMALL_STATE(1527)] = 52344,
  [SMALL_STATE(1528)] = 52437,
  [SMALL_STATE(1529)] = 52530,
  [SMALL_STATE(1530)] = 52623,
  [SMALL_STATE(1531)] = 52716,
  [SMALL_STATE(1532)] = 52809,
  [SMALL_STATE(1533)] = 52902,
  [SMALL_STATE(1534)] = 52995,
  [SMALL_STATE(1535)] = 53088,
  [SMALL_STATE(1536)] = 53181,
  [SMALL_STATE(1537)] = 53274,
  [SMALL_STATE(1538)] = 53367,
  [SMALL_STATE(1539)] = 53460,
  [SMALL_STATE(1540)] = 53553,
  [SMALL_STATE(1541)] = 53646,
  [SMALL_STATE(1542)] = 53739,
  [SMALL_STATE(1543)] = 53832,
  [SMALL_STATE(1544)] = 53922,
  [SMALL_STATE(1545)] = 54012,
  [SMALL_STATE(1546)] = 54102,
  [SMALL_STATE(1547)] = 54192,
  [SMALL_STATE(1548)] = 54282,
  [SMALL_STATE(1549)] = 54383,
  [SMALL_STATE(1550)] = 54472,
  [SMALL_STATE(1551)] = 54561,
  [SMALL_STATE(1552)] = 54650,
  [SMALL_STATE(1553)] = 54739,
  [SMALL_STATE(1554)] = 54828,
  [SMALL_STATE(1555)] = 54917,
  [SMALL_STATE(1556)] = 55006,
  [SMALL_STATE(1557)] = 55095,
  [SMALL_STATE(1558)] = 55184,
  [SMALL_STATE(1559)] = 55273,
  [SMALL_STATE(1560)] = 55362,
  [SMALL_STATE(1561)] = 55451,
  [SMALL_STATE(1562)] = 55540,
  [SMALL_STATE(1563)] = 55629,
  [SMALL_STATE(1564)] = 55718,
  [SMALL_STATE(1565)] = 55807,
  [SMALL_STATE(1566)] = 55896,
  [SMALL_STATE(1567)] = 55985,
  [SMALL_STATE(1568)] = 56074,
  [SMALL_STATE(1569)] = 56163,
  [SMALL_STATE(1570)] = 56252,
  [SMALL_STATE(1571)] = 56341,
  [SMALL_STATE(1572)] = 56430,
  [SMALL_STATE(1573)] = 56519,
  [SMALL_STATE(1574)] = 56608,
  [SMALL_STATE(1575)] = 56697,
  [SMALL_STATE(1576)] = 56786,
  [SMALL_STATE(1577)] = 56875,
  [SMALL_STATE(1578)] = 56964,
  [SMALL_STATE(1579)] = 57053,
  [SMALL_STATE(1580)] = 57142,
  [SMALL_STATE(1581)] = 57231,
  [SMALL_STATE(1582)] = 57320,
  [SMALL_STATE(1583)] = 57409,
  [SMALL_STATE(1584)] = 57498,
  [SMALL_STATE(1585)] = 57587,
  [SMALL_STATE(1586)] = 57676,
  [SMALL_STATE(1587)] = 57765,
  [SMALL_STATE(1588)] = 57854,
  [SMALL_STATE(1589)] = 57943,
  [SMALL_STATE(1590)] = 58032,
  [SMALL_STATE(1591)] = 58121,
  [SMALL_STATE(1592)] = 58210,
  [SMALL_STATE(1593)] = 58299,
  [SMALL_STATE(1594)] = 58388,
  [SMALL_STATE(1595)] = 58477,
  [SMALL_STATE(1596)] = 58566,
  [SMALL_STATE(1597)] = 58655,
  [SMALL_STATE(1598)] = 58744,
  [SMALL_STATE(1599)] = 58833,
  [SMALL_STATE(1600)] = 58922,
  [SMALL_STATE(1601)] = 59011,
  [SMALL_STATE(1602)] = 59100,
  [SMALL_STATE(1603)] = 59189,
  [SMALL_STATE(1604)] = 59268,
  [SMALL_STATE(1605)] = 59347,
  [SMALL_STATE(1606)] = 59424,
  [SMALL_STATE(1607)] = 59503,
  [SMALL_STATE(1608)] = 59582,
  [SMALL_STATE(1609)] = 59661,
  [SMALL_STATE(1610)] = 59738,
  [SMALL_STATE(1611)] = 59815,
  [SMALL_STATE(1612)] = 59892,
  [SMALL_STATE(1613)] = 59969,
  [SMALL_STATE(1614)] = 60046,
  [SMALL_STATE(1615)] = 60125,
  [SMALL_STATE(1616)] = 60204,
  [SMALL_STATE(1617)] = 60283,
  [SMALL_STATE(1618)] = 60362,
  [SMALL_STATE(1619)] = 60441,
  [SMALL_STATE(1620)] = 60520,
  [SMALL_STATE(1621)] = 60599,
  [SMALL_STATE(1622)] = 60678,
  [SMALL_STATE(1623)] = 60757,
  [SMALL_STATE(1624)] = 60836,
  [SMALL_STATE(1625)] = 60915,
  [SMALL_STATE(1626)] = 60965,
  [SMALL_STATE(1627)] = 61015,
  [SMALL_STATE(1628)] = 61065,
  [SMALL_STATE(1629)] = 61115,
  [SMALL_STATE(1630)] = 61165,
  [SMALL_STATE(1631)] = 61215,
  [SMALL_STATE(1632)] = 61265,
  [SMALL_STATE(1633)] = 61315,
  [SMALL_STATE(1634)] = 61365,
  [SMALL_STATE(1635)] = 61415,
  [SMALL_STATE(1636)] = 61465,
  [SMALL_STATE(1637)] = 61515,
  [SMALL_STATE(1638)] = 61565,
  [SMALL_STATE(1639)] = 61615,
  [SMALL_STATE(1640)] = 61665,
  [SMALL_STATE(1641)] = 61715,
  [SMALL_STATE(1642)] = 61765,
  [SMALL_STATE(1643)] = 61815,
  [SMALL_STATE(1644)] = 61865,
  [SMALL_STATE(1645)] = 61915,
  [SMALL_STATE(1646)] = 61965,
  [SMALL_STATE(1647)] = 62015,
  [SMALL_STATE(1648)] = 62065,
  [SMALL_STATE(1649)] = 62115,
  [SMALL_STATE(1650)] = 62163,
  [SMALL_STATE(1651)] = 62213,
  [SMALL_STATE(1652)] = 62263,
  [SMALL_STATE(1653)] = 62313,
  [SMALL_STATE(1654)] = 62363,
  [SMALL_STATE(1655)] = 62410,
  [SMALL_STATE(1656)] = 62480,
  [SMALL_STATE(1657)] = 62543,
  [SMALL_STATE(1658)] = 62606,
  [SMALL_STATE(1659)] = 62669,
  [SMALL_STATE(1660)] = 62732,
  [SMALL_STATE(1661)] = 62795,
  [SMALL_STATE(1662)] = 62858,
  [SMALL_STATE(1663)] = 62921,
  [SMALL_STATE(1664)] = 62984,
  [SMALL_STATE(1665)] = 63047,
  [SMALL_STATE(1666)] = 63110,
  [SMALL_STATE(1667)] = 63173,
  [SMALL_STATE(1668)] = 63236,
  [SMALL_STATE(1669)] = 63299,
  [SMALL_STATE(1670)] = 63362,
  [SMALL_STATE(1671)] = 63425,
  [SMALL_STATE(1672)] = 63488,
  [SMALL_STATE(1673)] = 63551,
  [SMALL_STATE(1674)] = 63614,
  [SMALL_STATE(1675)] = 63674,
  [SMALL_STATE(1676)] = 63734,
  [SMALL_STATE(1677)] = 63794,
  [SMALL_STATE(1678)] = 63854,
  [SMALL_STATE(1679)] = 63914,
  [SMALL_STATE(1680)] = 63974,
  [SMALL_STATE(1681)] = 64034,
  [SMALL_STATE(1682)] = 64094,
  [SMALL_STATE(1683)] = 64154,
  [SMALL_STATE(1684)] = 64214,
  [SMALL_STATE(1685)] = 64274,
  [SMALL_STATE(1686)] = 64334,
  [SMALL_STATE(1687)] = 64379,
  [SMALL_STATE(1688)] = 64424,
  [SMALL_STATE(1689)] = 64469,
  [SMALL_STATE(1690)] = 64514,
  [SMALL_STATE(1691)] = 64559,
  [SMALL_STATE(1692)] = 64604,
  [SMALL_STATE(1693)] = 64668,
  [SMALL_STATE(1694)] = 64734,
  [SMALL_STATE(1695)] = 64800,
  [SMALL_STATE(1696)] = 64856,
  [SMALL_STATE(1697)] = 64922,
  [SMALL_STATE(1698)] = 64988,
  [SMALL_STATE(1699)] = 65044,
  [SMALL_STATE(1700)] = 65100,
  [SMALL_STATE(1701)] = 65166,
  [SMALL_STATE(1702)] = 65232,
  [SMALL_STATE(1703)] = 65271,
  [SMALL_STATE(1704)] = 65310,
  [SMALL_STATE(1705)] = 65363,
  [SMALL_STATE(1706)] = 65416,
  [SMALL_STATE(1707)] = 65455,
  [SMALL_STATE(1708)] = 65494,
  [SMALL_STATE(1709)] = 65533,
  [SMALL_STATE(1710)] = 65572,
  [SMALL_STATE(1711)] = 65611,
  [SMALL_STATE(1712)] = 65650,
  [SMALL_STATE(1713)] = 65689,
  [SMALL_STATE(1714)] = 65728,
  [SMALL_STATE(1715)] = 65768,
  [SMALL_STATE(1716)] = 65808,
  [SMALL_STATE(1717)] = 65848,
  [SMALL_STATE(1718)] = 65886,
  [SMALL_STATE(1719)] = 65928,
  [SMALL_STATE(1720)] = 65968,
  [SMALL_STATE(1721)] = 66008,
  [SMALL_STATE(1722)] = 66048,
  [SMALL_STATE(1723)] = 66086,
  [SMALL_STATE(1724)] = 66121,
  [SMALL_STATE(1725)] = 66156,
  [SMALL_STATE(1726)] = 66207,
  [SMALL_STATE(1727)] = 66258,
  [SMALL_STATE(1728)] = 66295,
  [SMALL_STATE(1729)] = 66330,
  [SMALL_STATE(1730)] = 66381,
  [SMALL_STATE(1731)] = 66416,
  [SMALL_STATE(1732)] = 66467,
  [SMALL_STATE(1733)] = 66502,
  [SMALL_STATE(1734)] = 66536,
  [SMALL_STATE(1735)] = 66570,
  [SMALL_STATE(1736)] = 66604,
  [SMALL_STATE(1737)] = 66638,
  [SMALL_STATE(1738)] = 66686,
  [SMALL_STATE(1739)] = 66732,
  [SMALL_STATE(1740)] = 66766,
  [SMALL_STATE(1741)] = 66800,
  [SMALL_STATE(1742)] = 66848,
  [SMALL_STATE(1743)] = 66882,
  [SMALL_STATE(1744)] = 66930,
  [SMALL_STATE(1745)] = 66978,
  [SMALL_STATE(1746)] = 67012,
  [SMALL_STATE(1747)] = 67046,
  [SMALL_STATE(1748)] = 67080,
  [SMALL_STATE(1749)] = 67116,
  [SMALL_STATE(1750)] = 67152,
  [SMALL_STATE(1751)] = 67200,
  [SMALL_STATE(1752)] = 67237,
  [SMALL_STATE(1753)] = 67272,
  [SMALL_STATE(1754)] = 67307,
  [SMALL_STATE(1755)] = 67342,
  [SMALL_STATE(1756)] = 67373,
  [SMALL_STATE(1757)] = 67408,
  [SMALL_STATE(1758)] = 67443,
  [SMALL_STATE(1759)] = 67478,
  [SMALL_STATE(1760)] = 67513,
  [SMALL_STATE(1761)] = 67548,
  [SMALL_STATE(1762)] = 67579,
  [SMALL_STATE(1763)] = 67618,
  [SMALL_STATE(1764)] = 67653,
  [SMALL_STATE(1765)] = 67688,
  [SMALL_STATE(1766)] = 67727,
  [SMALL_STATE(1767)] = 67762,
  [SMALL_STATE(1768)] = 67797,
  [SMALL_STATE(1769)] = 67834,
  [SMALL_STATE(1770)] = 67869,
  [SMALL_STATE(1771)] = 67899,
  [SMALL_STATE(1772)] = 67931,
  [SMALL_STATE(1773)] = 67967,
  [SMALL_STATE(1774)] = 68003,
  [SMALL_STATE(1775)] = 68033,
  [SMALL_STATE(1776)] = 68063,
  [SMALL_STATE(1777)] = 68093,
  [SMALL_STATE(1778)] = 68123,
  [SMALL_STATE(1779)] = 68153,
  [SMALL_STATE(1780)] = 68183,
  [SMALL_STATE(1781)] = 68213,
  [SMALL_STATE(1782)] = 68243,
  [SMALL_STATE(1783)] = 68273,
  [SMALL_STATE(1784)] = 68303,
  [SMALL_STATE(1785)] = 68330,
  [SMALL_STATE(1786)] = 68362,
  [SMALL_STATE(1787)] = 68394,
  [SMALL_STATE(1788)] = 68422,
  [SMALL_STATE(1789)] = 68454,
  [SMALL_STATE(1790)] = 68486,
  [SMALL_STATE(1791)] = 68518,
  [SMALL_STATE(1792)] = 68550,
  [SMALL_STATE(1793)] = 68582,
  [SMALL_STATE(1794)] = 68610,
  [SMALL_STATE(1795)] = 68642,
  [SMALL_STATE(1796)] = 68671,
  [SMALL_STATE(1797)] = 68700,
  [SMALL_STATE(1798)] = 68725,
  [SMALL_STATE(1799)] = 68758,
  [SMALL_STATE(1800)] = 68783,
  [SMALL_STATE(1801)] = 68812,
  [SMALL_STATE(1802)] = 68841,
  [SMALL_STATE(1803)] = 68870,
  [SMALL_STATE(1804)] = 68905,
  [SMALL_STATE(1805)] = 68934,
  [SMALL_STATE(1806)] = 68959,
  [SMALL_STATE(1807)] = 68994,
  [SMALL_STATE(1808)] = 69024,
  [SMALL_STATE(1809)] = 69054,
  [SMALL_STATE(1810)] = 69088,
  [SMALL_STATE(1811)] = 69118,
  [SMALL_STATE(1812)] = 69148,
  [SMALL_STATE(1813)] = 69174,
  [SMALL_STATE(1814)] = 69204,
  [SMALL_STATE(1815)] = 69238,
  [SMALL_STATE(1816)] = 69268,
  [SMALL_STATE(1817)] = 69294,
  [SMALL_STATE(1818)] = 69328,
  [SMALL_STATE(1819)] = 69358,
  [SMALL_STATE(1820)] = 69384,
  [SMALL_STATE(1821)] = 69418,
  [SMALL_STATE(1822)] = 69444,
  [SMALL_STATE(1823)] = 69470,
  [SMALL_STATE(1824)] = 69504,
  [SMALL_STATE(1825)] = 69538,
  [SMALL_STATE(1826)] = 69568,
  [SMALL_STATE(1827)] = 69603,
  [SMALL_STATE(1828)] = 69636,
  [SMALL_STATE(1829)] = 69663,
  [SMALL_STATE(1830)] = 69690,
  [SMALL_STATE(1831)] = 69725,
  [SMALL_STATE(1832)] = 69752,
  [SMALL_STATE(1833)] = 69787,
  [SMALL_STATE(1834)] = 69820,
  [SMALL_STATE(1835)] = 69845,
  [SMALL_STATE(1836)] = 69878,
  [SMALL_STATE(1837)] = 69903,
  [SMALL_STATE(1838)] = 69930,
  [SMALL_STATE(1839)] = 69965,
  [SMALL_STATE(1840)] = 70000,
  [SMALL_STATE(1841)] = 70033,
  [SMALL_STATE(1842)] = 70060,
  [SMALL_STATE(1843)] = 70095,
  [SMALL_STATE(1844)] = 70128,
  [SMALL_STATE(1845)] = 70163,
  [SMALL_STATE(1846)] = 70196,
  [SMALL_STATE(1847)] = 70229,
  [SMALL_STATE(1848)] = 70260,
  [SMALL_STATE(1849)] = 70285,
  [SMALL_STATE(1850)] = 70320,
  [SMALL_STATE(1851)] = 70355,
  [SMALL_STATE(1852)] = 70390,
  [SMALL_STATE(1853)] = 70425,
  [SMALL_STATE(1854)] = 70458,
  [SMALL_STATE(1855)] = 70493,
  [SMALL_STATE(1856)] = 70526,
  [SMALL_STATE(1857)] = 70553,
  [SMALL_STATE(1858)] = 70586,
  [SMALL_STATE(1859)] = 70613,
  [SMALL_STATE(1860)] = 70646,
  [SMALL_STATE(1861)] = 70673,
  [SMALL_STATE(1862)] = 70706,
  [SMALL_STATE(1863)] = 70733,
  [SMALL_STATE(1864)] = 70766,
  [SMALL_STATE(1865)] = 70799,
  [SMALL_STATE(1866)] = 70832,
  [SMALL_STATE(1867)] = 70859,
  [SMALL_STATE(1868)] = 70892,
  [SMALL_STATE(1869)] = 70923,
  [SMALL_STATE(1870)] = 70948,
  [SMALL_STATE(1871)] = 70981,
  [SMALL_STATE(1872)] = 71014,
  [SMALL_STATE(1873)] = 71047,
  [SMALL_STATE(1874)] = 71079,
  [SMALL_STATE(1875)] = 71111,
  [SMALL_STATE(1876)] = 71143,
  [SMALL_STATE(1877)] = 71173,
  [SMALL_STATE(1878)] = 71205,
  [SMALL_STATE(1879)] = 71229,
  [SMALL_STATE(1880)] = 71261,
  [SMALL_STATE(1881)] = 71293,
  [SMALL_STATE(1882)] = 71325,
  [SMALL_STATE(1883)] = 71357,
  [SMALL_STATE(1884)] = 71383,
  [SMALL_STATE(1885)] = 71415,
  [SMALL_STATE(1886)] = 71439,
  [SMALL_STATE(1887)] = 71469,
  [SMALL_STATE(1888)] = 71501,
  [SMALL_STATE(1889)] = 71525,
  [SMALL_STATE(1890)] = 71555,
  [SMALL_STATE(1891)] = 71585,
  [SMALL_STATE(1892)] = 71617,
  [SMALL_STATE(1893)] = 71647,
  [SMALL_STATE(1894)] = 71677,
  [SMALL_STATE(1895)] = 71701,
  [SMALL_STATE(1896)] = 71725,
  [SMALL_STATE(1897)] = 71749,
  [SMALL_STATE(1898)] = 71781,
  [SMALL_STATE(1899)] = 71805,
  [SMALL_STATE(1900)] = 71837,
  [SMALL_STATE(1901)] = 71869,
  [SMALL_STATE(1902)] = 71899,
  [SMALL_STATE(1903)] = 71931,
  [SMALL_STATE(1904)] = 71963,
  [SMALL_STATE(1905)] = 71987,
  [SMALL_STATE(1906)] = 72011,
  [SMALL_STATE(1907)] = 72035,
  [SMALL_STATE(1908)] = 72065,
  [SMALL_STATE(1909)] = 72097,
  [SMALL_STATE(1910)] = 72129,
  [SMALL_STATE(1911)] = 72161,
  [SMALL_STATE(1912)] = 72193,
  [SMALL_STATE(1913)] = 72225,
  [SMALL_STATE(1914)] = 72257,
  [SMALL_STATE(1915)] = 72289,
  [SMALL_STATE(1916)] = 72321,
  [SMALL_STATE(1917)] = 72348,
  [SMALL_STATE(1918)] = 72369,
  [SMALL_STATE(1919)] = 72396,
  [SMALL_STATE(1920)] = 72425,
  [SMALL_STATE(1921)] = 72454,
  [SMALL_STATE(1922)] = 72479,
  [SMALL_STATE(1923)] = 72508,
  [SMALL_STATE(1924)] = 72537,
  [SMALL_STATE(1925)] = 72566,
  [SMALL_STATE(1926)] = 72595,
  [SMALL_STATE(1927)] = 72624,
  [SMALL_STATE(1928)] = 72647,
  [SMALL_STATE(1929)] = 72676,
  [SMALL_STATE(1930)] = 72705,
  [SMALL_STATE(1931)] = 72734,
  [SMALL_STATE(1932)] = 72763,
  [SMALL_STATE(1933)] = 72792,
  [SMALL_STATE(1934)] = 72821,
  [SMALL_STATE(1935)] = 72850,
  [SMALL_STATE(1936)] = 72871,
  [SMALL_STATE(1937)] = 72900,
  [SMALL_STATE(1938)] = 72929,
  [SMALL_STATE(1939)] = 72958,
  [SMALL_STATE(1940)] = 72985,
  [SMALL_STATE(1941)] = 73014,
  [SMALL_STATE(1942)] = 73043,
  [SMALL_STATE(1943)] = 73072,
  [SMALL_STATE(1944)] = 73099,
  [SMALL_STATE(1945)] = 73128,
  [SMALL_STATE(1946)] = 73157,
  [SMALL_STATE(1947)] = 73186,
  [SMALL_STATE(1948)] = 73215,
  [SMALL_STATE(1949)] = 73244,
  [SMALL_STATE(1950)] = 73267,
  [SMALL_STATE(1951)] = 73296,
  [SMALL_STATE(1952)] = 73325,
  [SMALL_STATE(1953)] = 73354,
  [SMALL_STATE(1954)] = 73383,
  [SMALL_STATE(1955)] = 73412,
  [SMALL_STATE(1956)] = 73435,
  [SMALL_STATE(1957)] = 73458,
  [SMALL_STATE(1958)] = 73487,
  [SMALL_STATE(1959)] = 73516,
  [SMALL_STATE(1960)] = 73545,
  [SMALL_STATE(1961)] = 73574,
  [SMALL_STATE(1962)] = 73603,
  [SMALL_STATE(1963)] = 73632,
  [SMALL_STATE(1964)] = 73661,
  [SMALL_STATE(1965)] = 73684,
  [SMALL_STATE(1966)] = 73711,
  [SMALL_STATE(1967)] = 73732,
  [SMALL_STATE(1968)] = 73755,
  [SMALL_STATE(1969)] = 73784,
  [SMALL_STATE(1970)] = 73811,
  [SMALL_STATE(1971)] = 73840,
  [SMALL_STATE(1972)] = 73869,
  [SMALL_STATE(1973)] = 73896,
  [SMALL_STATE(1974)] = 73919,
  [SMALL_STATE(1975)] = 73946,
  [SMALL_STATE(1976)] = 73967,
  [SMALL_STATE(1977)] = 73994,
  [SMALL_STATE(1978)] = 74023,
  [SMALL_STATE(1979)] = 74052,
  [SMALL_STATE(1980)] = 74075,
  [SMALL_STATE(1981)] = 74104,
  [SMALL_STATE(1982)] = 74127,
  [SMALL_STATE(1983)] = 74156,
  [SMALL_STATE(1984)] = 74183,
  [SMALL_STATE(1985)] = 74212,
  [SMALL_STATE(1986)] = 74239,
  [SMALL_STATE(1987)] = 74266,
  [SMALL_STATE(1988)] = 74287,
  [SMALL_STATE(1989)] = 74310,
  [SMALL_STATE(1990)] = 74339,
  [SMALL_STATE(1991)] = 74368,
  [SMALL_STATE(1992)] = 74397,
  [SMALL_STATE(1993)] = 74420,
  [SMALL_STATE(1994)] = 74449,
  [SMALL_STATE(1995)] = 74478,
  [SMALL_STATE(1996)] = 74507,
  [SMALL_STATE(1997)] = 74536,
  [SMALL_STATE(1998)] = 74559,
  [SMALL_STATE(1999)] = 74588,
  [SMALL_STATE(2000)] = 74617,
  [SMALL_STATE(2001)] = 74646,
  [SMALL_STATE(2002)] = 74675,
  [SMALL_STATE(2003)] = 74698,
  [SMALL_STATE(2004)] = 74727,
  [SMALL_STATE(2005)] = 74756,
  [SMALL_STATE(2006)] = 74785,
  [SMALL_STATE(2007)] = 74808,
  [SMALL_STATE(2008)] = 74829,
  [SMALL_STATE(2009)] = 74858,
  [SMALL_STATE(2010)] = 74887,
  [SMALL_STATE(2011)] = 74916,
  [SMALL_STATE(2012)] = 74943,
  [SMALL_STATE(2013)] = 74972,
  [SMALL_STATE(2014)] = 74999,
  [SMALL_STATE(2015)] = 75022,
  [SMALL_STATE(2016)] = 75051,
  [SMALL_STATE(2017)] = 75074,
  [SMALL_STATE(2018)] = 75103,
  [SMALL_STATE(2019)] = 75132,
  [SMALL_STATE(2020)] = 75159,
  [SMALL_STATE(2021)] = 75188,
  [SMALL_STATE(2022)] = 75215,
  [SMALL_STATE(2023)] = 75244,
  [SMALL_STATE(2024)] = 75273,
  [SMALL_STATE(2025)] = 75298,
  [SMALL_STATE(2026)] = 75327,
  [SMALL_STATE(2027)] = 75356,
  [SMALL_STATE(2028)] = 75385,
  [SMALL_STATE(2029)] = 75414,
  [SMALL_STATE(2030)] = 75443,
  [SMALL_STATE(2031)] = 75472,
  [SMALL_STATE(2032)] = 75501,
  [SMALL_STATE(2033)] = 75530,
  [SMALL_STATE(2034)] = 75553,
  [SMALL_STATE(2035)] = 75582,
  [SMALL_STATE(2036)] = 75611,
  [SMALL_STATE(2037)] = 75640,
  [SMALL_STATE(2038)] = 75669,
  [SMALL_STATE(2039)] = 75690,
  [SMALL_STATE(2040)] = 75719,
  [SMALL_STATE(2041)] = 75746,
  [SMALL_STATE(2042)] = 75768,
  [SMALL_STATE(2043)] = 75794,
  [SMALL_STATE(2044)] = 75820,
  [SMALL_STATE(2045)] = 75842,
  [SMALL_STATE(2046)] = 75868,
  [SMALL_STATE(2047)] = 75894,
  [SMALL_STATE(2048)] = 75920,
  [SMALL_STATE(2049)] = 75946,
  [SMALL_STATE(2050)] = 75972,
  [SMALL_STATE(2051)] = 75998,
  [SMALL_STATE(2052)] = 76024,
  [SMALL_STATE(2053)] = 76050,
  [SMALL_STATE(2054)] = 76076,
  [SMALL_STATE(2055)] = 76098,
  [SMALL_STATE(2056)] = 76120,
  [SMALL_STATE(2057)] = 76146,
  [SMALL_STATE(2058)] = 76172,
  [SMALL_STATE(2059)] = 76198,
  [SMALL_STATE(2060)] = 76224,
  [SMALL_STATE(2061)] = 76250,
  [SMALL_STATE(2062)] = 76276,
  [SMALL_STATE(2063)] = 76302,
  [SMALL_STATE(2064)] = 76328,
  [SMALL_STATE(2065)] = 76354,
  [SMALL_STATE(2066)] = 76380,
  [SMALL_STATE(2067)] = 76406,
  [SMALL_STATE(2068)] = 76432,
  [SMALL_STATE(2069)] = 76458,
  [SMALL_STATE(2070)] = 76484,
  [SMALL_STATE(2071)] = 76510,
  [SMALL_STATE(2072)] = 76536,
  [SMALL_STATE(2073)] = 76562,
  [SMALL_STATE(2074)] = 76586,
  [SMALL_STATE(2075)] = 76612,
  [SMALL_STATE(2076)] = 76634,
  [SMALL_STATE(2077)] = 76660,
  [SMALL_STATE(2078)] = 76686,
  [SMALL_STATE(2079)] = 76708,
  [SMALL_STATE(2080)] = 76734,
  [SMALL_STATE(2081)] = 76756,
  [SMALL_STATE(2082)] = 76782,
  [SMALL_STATE(2083)] = 76808,
  [SMALL_STATE(2084)] = 76834,
  [SMALL_STATE(2085)] = 76860,
  [SMALL_STATE(2086)] = 76886,
  [SMALL_STATE(2087)] = 76908,
  [SMALL_STATE(2088)] = 76934,
  [SMALL_STATE(2089)] = 76960,
  [SMALL_STATE(2090)] = 76986,
  [SMALL_STATE(2091)] = 77012,
  [SMALL_STATE(2092)] = 77038,
  [SMALL_STATE(2093)] = 77064,
  [SMALL_STATE(2094)] = 77090,
  [SMALL_STATE(2095)] = 77112,
  [SMALL_STATE(2096)] = 77138,
  [SMALL_STATE(2097)] = 77164,
  [SMALL_STATE(2098)] = 77186,
  [SMALL_STATE(2099)] = 77212,
  [SMALL_STATE(2100)] = 77234,
  [SMALL_STATE(2101)] = 77260,
  [SMALL_STATE(2102)] = 77286,
  [SMALL_STATE(2103)] = 77312,
  [SMALL_STATE(2104)] = 77338,
  [SMALL_STATE(2105)] = 77364,
  [SMALL_STATE(2106)] = 77390,
  [SMALL_STATE(2107)] = 77416,
  [SMALL_STATE(2108)] = 77442,
  [SMALL_STATE(2109)] = 77468,
  [SMALL_STATE(2110)] = 77488,
  [SMALL_STATE(2111)] = 77514,
  [SMALL_STATE(2112)] = 77540,
  [SMALL_STATE(2113)] = 77566,
  [SMALL_STATE(2114)] = 77592,
  [SMALL_STATE(2115)] = 77618,
  [SMALL_STATE(2116)] = 77644,
  [SMALL_STATE(2117)] = 77670,
  [SMALL_STATE(2118)] = 77696,
  [SMALL_STATE(2119)] = 77722,
  [SMALL_STATE(2120)] = 77748,
  [SMALL_STATE(2121)] = 77774,
  [SMALL_STATE(2122)] = 77797,
  [SMALL_STATE(2123)] = 77820,
  [SMALL_STATE(2124)] = 77843,
  [SMALL_STATE(2125)] = 77866,
  [SMALL_STATE(2126)] = 77887,
  [SMALL_STATE(2127)] = 77910,
  [SMALL_STATE(2128)] = 77933,
  [SMALL_STATE(2129)] = 77956,
  [SMALL_STATE(2130)] = 77979,
  [SMALL_STATE(2131)] = 78002,
  [SMALL_STATE(2132)] = 78025,
  [SMALL_STATE(2133)] = 78048,
  [SMALL_STATE(2134)] = 78071,
  [SMALL_STATE(2135)] = 78094,
  [SMALL_STATE(2136)] = 78117,
  [SMALL_STATE(2137)] = 78140,
  [SMALL_STATE(2138)] = 78163,
  [SMALL_STATE(2139)] = 78186,
  [SMALL_STATE(2140)] = 78209,
  [SMALL_STATE(2141)] = 78232,
  [SMALL_STATE(2142)] = 78255,
  [SMALL_STATE(2143)] = 78278,
  [SMALL_STATE(2144)] = 78301,
  [SMALL_STATE(2145)] = 78324,
  [SMALL_STATE(2146)] = 78347,
  [SMALL_STATE(2147)] = 78370,
  [SMALL_STATE(2148)] = 78389,
  [SMALL_STATE(2149)] = 78412,
  [SMALL_STATE(2150)] = 78435,
  [SMALL_STATE(2151)] = 78456,
  [SMALL_STATE(2152)] = 78479,
  [SMALL_STATE(2153)] = 78502,
  [SMALL_STATE(2154)] = 78525,
  [SMALL_STATE(2155)] = 78548,
  [SMALL_STATE(2156)] = 78571,
  [SMALL_STATE(2157)] = 78594,
  [SMALL_STATE(2158)] = 78617,
  [SMALL_STATE(2159)] = 78640,
  [SMALL_STATE(2160)] = 78663,
  [SMALL_STATE(2161)] = 78686,
  [SMALL_STATE(2162)] = 78709,
  [SMALL_STATE(2163)] = 78732,
  [SMALL_STATE(2164)] = 78755,
  [SMALL_STATE(2165)] = 78778,
  [SMALL_STATE(2166)] = 78801,
  [SMALL_STATE(2167)] = 78824,
  [SMALL_STATE(2168)] = 78847,
  [SMALL_STATE(2169)] = 78870,
  [SMALL_STATE(2170)] = 78893,
  [SMALL_STATE(2171)] = 78916,
  [SMALL_STATE(2172)] = 78939,
  [SMALL_STATE(2173)] = 78962,
  [SMALL_STATE(2174)] = 78985,
  [SMALL_STATE(2175)] = 79008,
  [SMALL_STATE(2176)] = 79031,
  [SMALL_STATE(2177)] = 79054,
  [SMALL_STATE(2178)] = 79077,
  [SMALL_STATE(2179)] = 79100,
  [SMALL_STATE(2180)] = 79123,
  [SMALL_STATE(2181)] = 79146,
  [SMALL_STATE(2182)] = 79169,
  [SMALL_STATE(2183)] = 79192,
  [SMALL_STATE(2184)] = 79215,
  [SMALL_STATE(2185)] = 79238,
  [SMALL_STATE(2186)] = 79261,
  [SMALL_STATE(2187)] = 79284,
  [SMALL_STATE(2188)] = 79305,
  [SMALL_STATE(2189)] = 79328,
  [SMALL_STATE(2190)] = 79351,
  [SMALL_STATE(2191)] = 79374,
  [SMALL_STATE(2192)] = 79395,
  [SMALL_STATE(2193)] = 79416,
  [SMALL_STATE(2194)] = 79439,
  [SMALL_STATE(2195)] = 79462,
  [SMALL_STATE(2196)] = 79485,
  [SMALL_STATE(2197)] = 79508,
  [SMALL_STATE(2198)] = 79531,
  [SMALL_STATE(2199)] = 79554,
  [SMALL_STATE(2200)] = 79577,
  [SMALL_STATE(2201)] = 79600,
  [SMALL_STATE(2202)] = 79623,
  [SMALL_STATE(2203)] = 79646,
  [SMALL_STATE(2204)] = 79669,
  [SMALL_STATE(2205)] = 79692,
  [SMALL_STATE(2206)] = 79715,
  [SMALL_STATE(2207)] = 79738,
  [SMALL_STATE(2208)] = 79759,
  [SMALL_STATE(2209)] = 79782,
  [SMALL_STATE(2210)] = 79805,
  [SMALL_STATE(2211)] = 79828,
  [SMALL_STATE(2212)] = 79851,
  [SMALL_STATE(2213)] = 79871,
  [SMALL_STATE(2214)] = 79891,
  [SMALL_STATE(2215)] = 79911,
  [SMALL_STATE(2216)] = 79931,
  [SMALL_STATE(2217)] = 79951,
  [SMALL_STATE(2218)] = 79971,
  [SMALL_STATE(2219)] = 79991,
  [SMALL_STATE(2220)] = 80011,
  [SMALL_STATE(2221)] = 80031,
  [SMALL_STATE(2222)] = 80051,
  [SMALL_STATE(2223)] = 80071,
  [SMALL_STATE(2224)] = 80091,
  [SMALL_STATE(2225)] = 80111,
  [SMALL_STATE(2226)] = 80131,
  [SMALL_STATE(2227)] = 80151,
  [SMALL_STATE(2228)] = 80171,
  [SMALL_STATE(2229)] = 80191,
  [SMALL_STATE(2230)] = 80211,
  [SMALL_STATE(2231)] = 80231,
  [SMALL_STATE(2232)] = 80251,
  [SMALL_STATE(2233)] = 80271,
  [SMALL_STATE(2234)] = 80291,
  [SMALL_STATE(2235)] = 80311,
  [SMALL_STATE(2236)] = 80331,
  [SMALL_STATE(2237)] = 80351,
  [SMALL_STATE(2238)] = 80371,
  [SMALL_STATE(2239)] = 80391,
  [SMALL_STATE(2240)] = 80411,
  [SMALL_STATE(2241)] = 80431,
  [SMALL_STATE(2242)] = 80451,
  [SMALL_STATE(2243)] = 80471,
  [SMALL_STATE(2244)] = 80491,
  [SMALL_STATE(2245)] = 80511,
  [SMALL_STATE(2246)] = 80531,
  [SMALL_STATE(2247)] = 80551,
  [SMALL_STATE(2248)] = 80571,
  [SMALL_STATE(2249)] = 80591,
  [SMALL_STATE(2250)] = 80611,
  [SMALL_STATE(2251)] = 80631,
  [SMALL_STATE(2252)] = 80651,
  [SMALL_STATE(2253)] = 80671,
  [SMALL_STATE(2254)] = 80691,
  [SMALL_STATE(2255)] = 80711,
  [SMALL_STATE(2256)] = 80731,
  [SMALL_STATE(2257)] = 80751,
  [SMALL_STATE(2258)] = 80771,
  [SMALL_STATE(2259)] = 80791,
  [SMALL_STATE(2260)] = 80811,
  [SMALL_STATE(2261)] = 80831,
  [SMALL_STATE(2262)] = 80851,
  [SMALL_STATE(2263)] = 80871,
  [SMALL_STATE(2264)] = 80891,
  [SMALL_STATE(2265)] = 80911,
  [SMALL_STATE(2266)] = 80931,
  [SMALL_STATE(2267)] = 80951,
  [SMALL_STATE(2268)] = 80971,
  [SMALL_STATE(2269)] = 80991,
  [SMALL_STATE(2270)] = 81011,
  [SMALL_STATE(2271)] = 81031,
  [SMALL_STATE(2272)] = 81051,
  [SMALL_STATE(2273)] = 81071,
  [SMALL_STATE(2274)] = 81091,
  [SMALL_STATE(2275)] = 81111,
  [SMALL_STATE(2276)] = 81131,
  [SMALL_STATE(2277)] = 81151,
  [SMALL_STATE(2278)] = 81171,
  [SMALL_STATE(2279)] = 81191,
  [SMALL_STATE(2280)] = 81211,
  [SMALL_STATE(2281)] = 81231,
  [SMALL_STATE(2282)] = 81251,
  [SMALL_STATE(2283)] = 81269,
  [SMALL_STATE(2284)] = 81289,
  [SMALL_STATE(2285)] = 81309,
  [SMALL_STATE(2286)] = 81329,
  [SMALL_STATE(2287)] = 81349,
  [SMALL_STATE(2288)] = 81369,
  [SMALL_STATE(2289)] = 81387,
  [SMALL_STATE(2290)] = 81407,
  [SMALL_STATE(2291)] = 81427,
  [SMALL_STATE(2292)] = 81447,
  [SMALL_STATE(2293)] = 81467,
  [SMALL_STATE(2294)] = 81487,
  [SMALL_STATE(2295)] = 81507,
  [SMALL_STATE(2296)] = 81527,
  [SMALL_STATE(2297)] = 81547,
  [SMALL_STATE(2298)] = 81567,
  [SMALL_STATE(2299)] = 81585,
  [SMALL_STATE(2300)] = 81605,
  [SMALL_STATE(2301)] = 81625,
  [SMALL_STATE(2302)] = 81645,
  [SMALL_STATE(2303)] = 81663,
  [SMALL_STATE(2304)] = 81683,
  [SMALL_STATE(2305)] = 81703,
  [SMALL_STATE(2306)] = 81723,
  [SMALL_STATE(2307)] = 81743,
  [SMALL_STATE(2308)] = 81763,
  [SMALL_STATE(2309)] = 81783,
  [SMALL_STATE(2310)] = 81803,
  [SMALL_STATE(2311)] = 81823,
  [SMALL_STATE(2312)] = 81843,
  [SMALL_STATE(2313)] = 81863,
  [SMALL_STATE(2314)] = 81883,
  [SMALL_STATE(2315)] = 81901,
  [SMALL_STATE(2316)] = 81921,
  [SMALL_STATE(2317)] = 81941,
  [SMALL_STATE(2318)] = 81961,
  [SMALL_STATE(2319)] = 81981,
  [SMALL_STATE(2320)] = 82001,
  [SMALL_STATE(2321)] = 82021,
  [SMALL_STATE(2322)] = 82041,
  [SMALL_STATE(2323)] = 82059,
  [SMALL_STATE(2324)] = 82079,
  [SMALL_STATE(2325)] = 82099,
  [SMALL_STATE(2326)] = 82119,
  [SMALL_STATE(2327)] = 82139,
  [SMALL_STATE(2328)] = 82159,
  [SMALL_STATE(2329)] = 82179,
  [SMALL_STATE(2330)] = 82199,
  [SMALL_STATE(2331)] = 82219,
  [SMALL_STATE(2332)] = 82239,
  [SMALL_STATE(2333)] = 82259,
  [SMALL_STATE(2334)] = 82277,
  [SMALL_STATE(2335)] = 82297,
  [SMALL_STATE(2336)] = 82317,
  [SMALL_STATE(2337)] = 82337,
  [SMALL_STATE(2338)] = 82357,
  [SMALL_STATE(2339)] = 82377,
  [SMALL_STATE(2340)] = 82397,
  [SMALL_STATE(2341)] = 82417,
  [SMALL_STATE(2342)] = 82434,
  [SMALL_STATE(2343)] = 82451,
  [SMALL_STATE(2344)] = 82468,
  [SMALL_STATE(2345)] = 82485,
  [SMALL_STATE(2346)] = 82502,
  [SMALL_STATE(2347)] = 82519,
  [SMALL_STATE(2348)] = 82536,
  [SMALL_STATE(2349)] = 82553,
  [SMALL_STATE(2350)] = 82570,
  [SMALL_STATE(2351)] = 82587,
  [SMALL_STATE(2352)] = 82604,
  [SMALL_STATE(2353)] = 82621,
  [SMALL_STATE(2354)] = 82638,
  [SMALL_STATE(2355)] = 82655,
  [SMALL_STATE(2356)] = 82672,
  [SMALL_STATE(2357)] = 82689,
  [SMALL_STATE(2358)] = 82706,
  [SMALL_STATE(2359)] = 82723,
  [SMALL_STATE(2360)] = 82740,
  [SMALL_STATE(2361)] = 82757,
  [SMALL_STATE(2362)] = 82774,
  [SMALL_STATE(2363)] = 82791,
  [SMALL_STATE(2364)] = 82808,
  [SMALL_STATE(2365)] = 82825,
  [SMALL_STATE(2366)] = 82842,
  [SMALL_STATE(2367)] = 82859,
  [SMALL_STATE(2368)] = 82876,
  [SMALL_STATE(2369)] = 82893,
  [SMALL_STATE(2370)] = 82910,
  [SMALL_STATE(2371)] = 82927,
  [SMALL_STATE(2372)] = 82944,
  [SMALL_STATE(2373)] = 82961,
  [SMALL_STATE(2374)] = 82978,
  [SMALL_STATE(2375)] = 82995,
  [SMALL_STATE(2376)] = 83012,
  [SMALL_STATE(2377)] = 83029,
  [SMALL_STATE(2378)] = 83046,
  [SMALL_STATE(2379)] = 83063,
  [SMALL_STATE(2380)] = 83080,
  [SMALL_STATE(2381)] = 83097,
  [SMALL_STATE(2382)] = 83114,
  [SMALL_STATE(2383)] = 83131,
  [SMALL_STATE(2384)] = 83148,
  [SMALL_STATE(2385)] = 83165,
  [SMALL_STATE(2386)] = 83182,
  [SMALL_STATE(2387)] = 83199,
  [SMALL_STATE(2388)] = 83216,
  [SMALL_STATE(2389)] = 83233,
  [SMALL_STATE(2390)] = 83250,
  [SMALL_STATE(2391)] = 83267,
  [SMALL_STATE(2392)] = 83284,
  [SMALL_STATE(2393)] = 83301,
  [SMALL_STATE(2394)] = 83318,
  [SMALL_STATE(2395)] = 83335,
  [SMALL_STATE(2396)] = 83352,
  [SMALL_STATE(2397)] = 83369,
  [SMALL_STATE(2398)] = 83386,
  [SMALL_STATE(2399)] = 83403,
  [SMALL_STATE(2400)] = 83420,
  [SMALL_STATE(2401)] = 83437,
  [SMALL_STATE(2402)] = 83454,
  [SMALL_STATE(2403)] = 83471,
  [SMALL_STATE(2404)] = 83488,
  [SMALL_STATE(2405)] = 83505,
  [SMALL_STATE(2406)] = 83522,
  [SMALL_STATE(2407)] = 83539,
  [SMALL_STATE(2408)] = 83556,
  [SMALL_STATE(2409)] = 83573,
  [SMALL_STATE(2410)] = 83590,
  [SMALL_STATE(2411)] = 83607,
  [SMALL_STATE(2412)] = 83624,
  [SMALL_STATE(2413)] = 83641,
  [SMALL_STATE(2414)] = 83658,
  [SMALL_STATE(2415)] = 83675,
  [SMALL_STATE(2416)] = 83692,
  [SMALL_STATE(2417)] = 83709,
  [SMALL_STATE(2418)] = 83726,
  [SMALL_STATE(2419)] = 83743,
  [SMALL_STATE(2420)] = 83760,
  [SMALL_STATE(2421)] = 83777,
  [SMALL_STATE(2422)] = 83794,
  [SMALL_STATE(2423)] = 83811,
  [SMALL_STATE(2424)] = 83828,
  [SMALL_STATE(2425)] = 83845,
  [SMALL_STATE(2426)] = 83862,
  [SMALL_STATE(2427)] = 83879,
  [SMALL_STATE(2428)] = 83896,
  [SMALL_STATE(2429)] = 83913,
  [SMALL_STATE(2430)] = 83930,
  [SMALL_STATE(2431)] = 83947,
  [SMALL_STATE(2432)] = 83964,
  [SMALL_STATE(2433)] = 83981,
  [SMALL_STATE(2434)] = 83998,
  [SMALL_STATE(2435)] = 84015,
  [SMALL_STATE(2436)] = 84032,
  [SMALL_STATE(2437)] = 84049,
  [SMALL_STATE(2438)] = 84066,
  [SMALL_STATE(2439)] = 84083,
  [SMALL_STATE(2440)] = 84100,
  [SMALL_STATE(2441)] = 84117,
  [SMALL_STATE(2442)] = 84134,
  [SMALL_STATE(2443)] = 84151,
  [SMALL_STATE(2444)] = 84168,
  [SMALL_STATE(2445)] = 84185,
  [SMALL_STATE(2446)] = 84202,
  [SMALL_STATE(2447)] = 84219,
  [SMALL_STATE(2448)] = 84236,
  [SMALL_STATE(2449)] = 84253,
  [SMALL_STATE(2450)] = 84270,
  [SMALL_STATE(2451)] = 84287,
  [SMALL_STATE(2452)] = 84304,
  [SMALL_STATE(2453)] = 84321,
  [SMALL_STATE(2454)] = 84338,
  [SMALL_STATE(2455)] = 84355,
  [SMALL_STATE(2456)] = 84372,
  [SMALL_STATE(2457)] = 84389,
  [SMALL_STATE(2458)] = 84406,
  [SMALL_STATE(2459)] = 84423,
  [SMALL_STATE(2460)] = 84440,
  [SMALL_STATE(2461)] = 84457,
  [SMALL_STATE(2462)] = 84461,
  [SMALL_STATE(2463)] = 84465,
  [SMALL_STATE(2464)] = 84469,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2398),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [9] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(25),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(918),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1823),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2118),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2454),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1889),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1235),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(434),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2204),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(822),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2451),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(480),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(470),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2289),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2218),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2236),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2233),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2053),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1910),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1911),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2257),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1917),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1752),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1748),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1259),
  [97] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0),
  [99] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(249),
  [102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0),
  [104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1275),
  [107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1275),
  [110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1239),
  [112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1568),
  [117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1907),
  [120] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1277),
  [123] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(309),
  [126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(44),
  [129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1122),
  [132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(368),
  [135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(261),
  [138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1056),
  [141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1051),
  [144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(288),
  [147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(288),
  [150] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(2110),
  [153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1903),
  [156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1873),
  [159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1826),
  [162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1867),
  [165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1225),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1225),
  [171] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(251),
  [174] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1286),
  [177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1286),
  [180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1190),
  [182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [184] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1553),
  [187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1876),
  [190] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1288),
  [193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(356),
  [196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(51),
  [199] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(311),
  [202] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(268),
  [205] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1035),
  [208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(274),
  [211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(274),
  [214] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(2063),
  [217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1891),
  [220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1915),
  [223] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1854),
  [226] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1855),
  [229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1183),
  [232] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1183),
  [235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(382),
  [238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1280),
  [241] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1280),
  [244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [248] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1551),
  [251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1886),
  [254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1282),
  [257] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(536),
  [260] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(49),
  [263] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(548),
  [266] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(399),
  [269] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1059),
  [272] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(388),
  [275] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(388),
  [278] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(2115),
  [281] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1914),
  [284] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1880),
  [287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1838),
  [290] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1840),
  [293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1278),
  [296] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1278),
  [299] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(305),
  [302] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1304),
  [305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1304),
  [308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1210),
  [310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1559),
  [315] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1889),
  [318] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1235),
  [321] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(434),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(54),
  [327] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(558),
  [330] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(469),
  [333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1046),
  [336] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(340),
  [339] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(340),
  [342] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(2053),
  [345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1910),
  [348] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1911),
  [351] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1832),
  [354] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1835),
  [357] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1259),
  [360] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1259),
  [363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__juxtable_expression, 1, 0, 0),
  [365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__juxtable_expression, 1, 0, 0),
  [367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dotted_identifier, 2, 0, 0),
  [375] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dotted_identifier, 2, 0, 0),
  [377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0),
  [379] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0),
  [381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1918),
  [384] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2040),
  [387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_identifier, 1, 0, 0),
  [389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_identifier, 1, 0, 0),
  [391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index, 4, 0, 0),
  [393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_index, 4, 0, 0),
  [395] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1916),
  [398] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1974),
  [401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 1, 0, 3),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2460),
  [411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1707),
  [427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1733),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [455] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(32),
  [458] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1001),
  [461] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [464] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1280),
  [467] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1824),
  [470] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2113),
  [473] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2454),
  [476] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1281),
  [479] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1551),
  [482] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1886),
  [485] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1282),
  [488] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(536),
  [491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2148),
  [494] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2148),
  [497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(49),
  [500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [502] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(988),
  [505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2409),
  [508] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1145),
  [511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(597),
  [514] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(563),
  [517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2337),
  [523] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2215),
  [526] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(497),
  [529] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2217),
  [532] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1059),
  [535] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(388),
  [538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(388),
  [541] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(662),
  [544] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2115),
  [547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1914),
  [550] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1880),
  [553] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1838),
  [556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1840),
  [559] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2252),
  [562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(177),
  [567] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1917),
  [570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1752),
  [573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1748),
  [576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1278),
  [579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1278),
  [582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1280),
  [588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1824),
  [592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2113),
  [594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1281),
  [596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1886),
  [600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1282),
  [602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2148),
  [606] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2148),
  [608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 3, 0, 48),
  [612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(988),
  [614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2409),
  [616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2337),
  [626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2215),
  [628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2217),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(662),
  [640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2115),
  [642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1880),
  [646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1838),
  [648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1840),
  [650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2252),
  [652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 3, 0, 48),
  [654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1278),
  [660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [662] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(25),
  [665] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(918),
  [668] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1304),
  [671] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1304),
  [674] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1823),
  [677] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2118),
  [680] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1186),
  [683] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1559),
  [686] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1889),
  [689] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1235),
  [692] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(434),
  [695] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2204),
  [698] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2204),
  [701] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(54),
  [704] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(822),
  [707] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2451),
  [710] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1128),
  [713] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(480),
  [716] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(470),
  [719] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(150),
  [722] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2289),
  [725] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2218),
  [728] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [731] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2236),
  [734] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1046),
  [737] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(340),
  [740] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(340),
  [743] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(659),
  [746] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2053),
  [749] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1910),
  [752] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1911),
  [755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1832),
  [758] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1835),
  [761] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(2257),
  [764] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(149),
  [767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1259),
  [770] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1259),
  [773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 3, 0, 0),
  [775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 3, 0, 0),
  [777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 2, 0, 0),
  [779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 4, 0, 48),
  [781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 4, 0, 48),
  [783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 2, 0, 0),
  [785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 2, 0, 0),
  [787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1703),
  [823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1746),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(663),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1809),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2064),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1890),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1270),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1074),
  [879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2194),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2441),
  [887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1037),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2329),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2330),
  [901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1081),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2313),
  [905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1069),
  [909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2076),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1912),
  [917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1913),
  [919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1839),
  [921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1843),
  [923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(708),
  [941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1817),
  [943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2074),
  [945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2131),
  [949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(681),
  [951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2452),
  [953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(222),
  [955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2340),
  [957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2264),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2287),
  [963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2253),
  [967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2254),
  [969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2256),
  [971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2259),
  [975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2260),
  [977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2261),
  [979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(943),
  [983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2243),
  [985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2244),
  [987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(986),
  [993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1286),
  [999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [1001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1814),
  [1003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2060),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [1007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [1009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1876),
  [1011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1288),
  [1013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(356),
  [1015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [1017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2159),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [1021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [1023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2440),
  [1025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [1027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [1029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [1031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2274),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2276),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [1039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2249),
  [1041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [1045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2063),
  [1051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1891),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1915),
  [1055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [1057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [1059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2272),
  [1061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [1063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [1065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [1067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2245),
  [1069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2246),
  [1071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2275),
  [1073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(211),
  [1075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2320),
  [1077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2321),
  [1079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2286),
  [1081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1275),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [1089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1820),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2106),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [1095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1907),
  [1099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1277),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [1103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [1105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2196),
  [1107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [1109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2366),
  [1111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [1113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(475),
  [1115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(312),
  [1117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [1119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2242),
  [1121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2247),
  [1123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [1125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2248),
  [1127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [1129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(288),
  [1131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [1133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(657),
  [1135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2110),
  [1137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1903),
  [1139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1873),
  [1141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [1143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [1145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [1147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [1149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [1151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2226),
  [1153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2227),
  [1155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2331),
  [1157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [1159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2234),
  [1161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2235),
  [1163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2221),
  [1165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(171),
  [1167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2237),
  [1169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2238),
  [1171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2278),
  [1173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [1175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2240),
  [1177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2241),
  [1179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2230),
  [1181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [1183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2268),
  [1185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [1187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2239),
  [1189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [1191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1108),
  [1193] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1268),
  [1196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1313),
  [1198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [1200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1556),
  [1203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1890),
  [1206] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1270),
  [1209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [1211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1163),
  [1213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1056),
  [1215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), SHIFT(1041),
  [1218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(665),
  [1220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(704),
  [1222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(932),
  [1224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(26),
  [1226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(964),
  [1228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1138),
  [1230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1407),
  [1232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [1234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(15),
  [1236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [1238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [1240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(18),
  [1242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [1244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(937),
  [1246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1303),
  [1248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1438),
  [1250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1386),
  [1252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__juxt_argument_list, 1, 0, 0),
  [1254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__juxt_argument_list, 1, 0, 0),
  [1256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [1258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [1260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__primary_expression, 1, 0, 0),
  [1262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__primary_expression, 1, 0, 0),
  [1264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1284),
  [1266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [1268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1272] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__juxtable_expression, 1, 0, 0), REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1275] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__juxtable_expression, 1, 0, 0), REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_increment_op, 2, 0, 0),
  [1280] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_increment_op, 2, 0, 0),
  [1282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [1284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [1286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1, 0, 0),
  [1288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2040),
  [1290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1, 0, 0),
  [1292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1983),
  [1294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [1296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__plain_string, 2, 0, 0),
  [1298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__plain_string, 2, 0, 0),
  [1300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolate_string, 2, 0, 0),
  [1302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__interpolate_string, 2, 0, 0),
  [1304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [1306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [1308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [1310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [1312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interpolate_string, 3, 0, 0),
  [1316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__interpolate_string, 3, 0, 0),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__plain_string, 3, 0, 0),
  [1320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__plain_string, 3, 0, 0),
  [1322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [1328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number_literal, 1, 0, 0),
  [1330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number_literal, 1, 0, 0),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 5, 0, 0),
  [1334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 5, 0, 0),
  [1336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1331),
  [1342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [1344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1388),
  [1346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 1, 0, 0),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 0),
  [1352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1965),
  [1355] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1983),
  [1358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [1360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1918),
  [1362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure, 2, 0, 0),
  [1364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure, 2, 0, 0),
  [1366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 2, 0, 8),
  [1368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 2, 0, 8),
  [1370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [1372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [1374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure, 3, 0, 0),
  [1376] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure, 3, 0, 0),
  [1378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 3, 0, 0),
  [1380] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 3, 0, 0),
  [1382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 3, 0, 0),
  [1384] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 3, 0, 0),
  [1386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_literal, 1, 0, 0),
  [1388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_literal, 1, 0, 0),
  [1390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure, 4, 0, 0),
  [1392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure, 4, 0, 0),
  [1394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 4, 0, 0),
  [1396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 4, 0, 0),
  [1398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 4, 0, 0),
  [1400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 4, 0, 0),
  [1402] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__callable_expression, 1, 0, 0), REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1405] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__callable_expression, 1, 0, 0), REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0),
  [1408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [1410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure, 5, 0, 0),
  [1412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure, 5, 0, 0),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 5, 0, 0),
  [1416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 5, 0, 0),
  [1418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map, 5, 0, 0),
  [1420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map, 5, 0, 0),
  [1422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2021),
  [1424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [1426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [1428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 6, 0, 0),
  [1430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 6, 0, 0),
  [1432] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 2, 0, 0),
  [1434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 2, 0, 0),
  [1436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [1438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2021),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [1443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [1445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1097),
  [1449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_access_op, 2, 0, 0),
  [1451] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_op, 2, 0, 0),
  [1453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1976),
  [1455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_op, 2, 0, 0),
  [1457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_op, 2, 0, 0),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [1463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_access_op, 3, 0, 0),
  [1465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_op, 3, 0, 0),
  [1467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_op, 3, 0, 0),
  [1469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_op, 3, 0, 0),
  [1471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [1473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [1475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [1477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [1479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [1481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1193),
  [1483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [1485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1196),
  [1487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [1489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [1491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [1493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [1495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [1497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [1499] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_map_item, 3, 0, 24),
  [1501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_map_item, 3, 0, 24),
  [1503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [1505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1345),
  [1507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [1509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ternary_op, 5, 0, 34),
  [1511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ternary_op, 5, 0, 34),
  [1513] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 1, 0, 0),
  [1515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 1, 0, 0),
  [1517] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 2, 0, 0),
  [1519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 2, 0, 0),
  [1521] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1976),
  [1524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [1532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [1534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1243),
  [1538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [1540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1244),
  [1542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1242),
  [1544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [1546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1245),
  [1548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [1550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1246),
  [1552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [1554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [1556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [1558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [1560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [1562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1333),
  [1564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1974),
  [1566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [1568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [1570] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [1572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [1574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [1576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [1578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [1580] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1214),
  [1582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [1584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [1586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [1588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1217),
  [1590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [1592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1219),
  [1594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [1596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [1598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 21),
  [1600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 21),
  [1602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion, 2, 0, 0),
  [1604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assertion, 2, 0, 0),
  [1606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return, 2, 0, 0),
  [1608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return, 2, 0, 0),
  [1610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment, 3, 0, 0),
  [1612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment, 3, 0, 0),
  [1614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 26),
  [1616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 26),
  [1618] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 36),
  [1620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 36),
  [1622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 38),
  [1624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 38),
  [1626] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 6, 0, 51),
  [1628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 6, 0, 51),
  [1630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 6, 0, 53),
  [1632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 6, 0, 53),
  [1634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 7, 0, 66),
  [1636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 7, 0, 66),
  [1638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 7, 0, 68),
  [1640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 7, 0, 68),
  [1642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [1644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [1646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [1648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [1650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [1652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [1654] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1291),
  [1656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [1658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1293),
  [1660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [1662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [1664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1382),
  [1666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [1668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1295),
  [1670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [1672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [1674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [1676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [1678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [1680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [1682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [1684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [1686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [1688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [1690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [1692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 26),
  [1696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return, 1, 0, 0),
  [1698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [1700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return, 1, 0, 0),
  [1702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1140),
  [1704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(355),
  [1706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [1708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [1710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [1712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [1714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [1716] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [1718] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2013),
  [1721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(481),
  [1723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [1725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(500),
  [1727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [1729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [1731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2450),
  [1733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [1735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1770),
  [1737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1761),
  [1739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 2, 0, 4),
  [1741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 2, 0, 4),
  [1743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [1745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1731),
  [1747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 15),
  [1749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 15),
  [1751] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 42),
  [1753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 42),
  [1755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [1757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_name, 3, 0, 0),
  [1759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__import_name, 3, 0, 0),
  [1761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 5, 0, 41),
  [1763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 5, 0, 41),
  [1765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [1767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 1, 0, 0),
  [1769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 1, 0, 0),
  [1771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 3, 0, 16),
  [1773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 3, 0, 16),
  [1775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [1777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [1779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_name, 1, 0, 0),
  [1781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__import_name, 1, 0, 0),
  [1783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 3, 0, 11),
  [1785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 3, 0, 11),
  [1787] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1168),
  [1790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 3, 0, 17),
  [1792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 3, 0, 17),
  [1794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1262),
  [1796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 5, 0, 37),
  [1798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 5, 0, 37),
  [1800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 5, 0, 39),
  [1802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 5, 0, 39),
  [1804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 3, 0, 0),
  [1806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 3, 0, 0),
  [1808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [1810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 2, 0, 1),
  [1812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 2, 0, 1),
  [1814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2111),
  [1816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2182),
  [1818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 5, 0, 10),
  [1820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 5, 0, 10),
  [1822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 4, 0, 1),
  [1824] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 4, 0, 1),
  [1826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 4, 0, 18),
  [1828] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 4, 0, 18),
  [1830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 2, 0, 7),
  [1832] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 2, 0, 7),
  [1834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [1836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 6, 0, 44),
  [1838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 6, 0, 44),
  [1840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 19),
  [1842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 19),
  [1844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [1846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 2, 0, 0),
  [1848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 5, 0, 31),
  [1850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 5, 0, 31),
  [1852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 4, 0, 22),
  [1854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 4, 0, 22),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_loop, 3, 0, 13),
  [1858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_loop, 3, 0, 13),
  [1860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 5, 0, 32),
  [1862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 5, 0, 32),
  [1864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_loop, 3, 0, 14),
  [1866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_loop, 3, 0, 14),
  [1868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_import, 3, 0, 10),
  [1870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_import, 3, 0, 10),
  [1872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2069),
  [1874] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2129),
  [1876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [1878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 4, 0, 0),
  [1880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_while_loop, 4, 0, 23),
  [1882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_while_loop, 4, 0, 23),
  [1884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2051),
  [1886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2199),
  [1888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1253),
  [1890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 2, 0, 9),
  [1892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 2, 0, 9),
  [1894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [1896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__juxt_argument_list, 2, 0, 0),
  [1898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__juxt_argument_list, 2, 0, 0),
  [1900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 25),
  [1902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 25),
  [1904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 13),
  [1906] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 13),
  [1908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 6, 0, 49),
  [1910] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 6, 0, 49),
  [1912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 6, 0, 50),
  [1914] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 6, 0, 50),
  [1916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2070),
  [1918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2142),
  [1920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 6, 0, 52),
  [1922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 6, 0, 52),
  [1924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 6, 0, 54),
  [1926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 6, 0, 54),
  [1928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 5, 0, 0),
  [1930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 5, 0, 0),
  [1932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_in_loop, 7, 0, 58),
  [1934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_in_loop, 7, 0, 58),
  [1936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 7, 0, 62),
  [1938] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 7, 0, 62),
  [1940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 7, 0, 63),
  [1942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 7, 0, 63),
  [1944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [1946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_list, 3, 0, 0),
  [1948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 7, 0, 64),
  [1950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 7, 0, 64),
  [1952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 33),
  [1954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 33),
  [1956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 27),
  [1958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 27),
  [1960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 7, 0, 65),
  [1962] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 7, 0, 65),
  [1964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [1966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [1968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 7, 0, 67),
  [1970] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 7, 0, 67),
  [1972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 8, 0, 74),
  [1974] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 8, 0, 74),
  [1976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1158),
  [1979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 29),
  [1981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 29),
  [1983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [1985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_juxt_function_call, 2, 0, 8),
  [1987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_juxt_function_call, 2, 0, 8),
  [1989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 8, 0, 75),
  [1991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 8, 0, 75),
  [1993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 9, 0, 79),
  [1995] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 9, 0, 79),
  [1997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [1999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 2, 0, 0),
  [2001] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 2, 0, 0),
  [2003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 9, 0, 80),
  [2005] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 9, 0, 80),
  [2007] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 30),
  [2009] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 30),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [2013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 3, 0, 0),
  [2015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 3, 0, 0),
  [2017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [2019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [2021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 5, 0, 35),
  [2023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 5, 0, 35),
  [2025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 4, 0, 0),
  [2027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 4, 0, 0),
  [2029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_block, 2, 0, 0),
  [2031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_block, 2, 0, 0),
  [2033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_package, 2, 0, 2),
  [2035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_package, 2, 0, 2),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2156),
  [2039] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 6, 0, 55),
  [2041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 6, 0, 55),
  [2043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [2047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 4, 0, 28),
  [2049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 4, 0, 28),
  [2051] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 3, 0, 12),
  [2053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 3, 0, 12),
  [2055] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 40),
  [2057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 40),
  [2059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2200),
  [2061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [2063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2114),
  [2065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2158),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1725),
  [2071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [2073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [2075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [2077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [2079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [2081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1729),
  [2083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [2085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [2087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [2089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2083),
  [2091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2133),
  [2093] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1176),
  [2096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [2098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2382),
  [2100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2296),
  [2102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2445),
  [2104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2232),
  [2106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [2108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [2110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 3, 0, 0),
  [2112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 3, 0, 0),
  [2114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2066),
  [2116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2151),
  [2118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [2120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1175),
  [2123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [2125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [2127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2116),
  [2129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2211),
  [2131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [2133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [2135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [2139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [2141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1180),
  [2143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 2, 0, 9),
  [2145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2172),
  [2147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [2149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [2151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2219),
  [2153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [2155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [2157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [2159] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [2161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shebang, 2, 0, 0),
  [2163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shebang, 2, 0, 0),
  [2165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 6, 0, 57),
  [2167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 6, 0, 57),
  [2169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 5, 0, 47),
  [2171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 5, 0, 47),
  [2173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 7, 0, 73),
  [2175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 7, 0, 73),
  [2177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 6, 0, 60),
  [2179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 6, 0, 60),
  [2181] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 6, 0, 61),
  [2183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 6, 0, 61),
  [2185] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 8, 0, 77),
  [2187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 8, 0, 77),
  [2189] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 7, 0, 72),
  [2191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 7, 0, 72),
  [2193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 4, 0, 0),
  [2195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 4, 0, 0),
  [2197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 5, 0, 46),
  [2199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 5, 0, 46),
  [2201] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 9, 0, 81),
  [2203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 9, 0, 81),
  [2205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 8, 0, 76),
  [2207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 8, 0, 76),
  [2209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 8, 0, 78),
  [2211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 8, 0, 78),
  [2213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 7, 0, 69),
  [2215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 7, 0, 69),
  [2217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 6, 0, 59),
  [2219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 6, 0, 59),
  [2221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 6, 0, 56),
  [2223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 6, 0, 56),
  [2225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_label, 2, 0, 3),
  [2227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label, 2, 0, 3),
  [2229] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 7, 0, 71),
  [2231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 7, 0, 71),
  [2233] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 7, 0, 70),
  [2235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 7, 0, 70),
  [2237] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_parameters, 5, 0, 45),
  [2239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_parameters, 5, 0, 45),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1985),
  [2247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1985),
  [2250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1392),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2342),
  [2258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2352),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [2268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [2270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [2272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [2274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [2276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2386),
  [2278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [2280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [2282] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2013),
  [2284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1943),
  [2286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1943),
  [2289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [2293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [2297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [2299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [2301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [2303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [2305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [2309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [2311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1048),
  [2314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [2317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [2320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1556),
  [2323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1890),
  [2326] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1270),
  [2329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1074),
  [2332] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(43),
  [2335] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1122),
  [2338] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1283),
  [2341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1037),
  [2344] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1056),
  [2347] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1041),
  [2350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0),
  [2352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1069),
  [2355] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1069),
  [2358] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2076),
  [2361] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [2364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1913),
  [2367] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1839),
  [2370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1843),
  [2373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1285),
  [2376] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1285),
  [2379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [2381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [2383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [2385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [2387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1392),
  [2390] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [2393] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [2396] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1556),
  [2399] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1890),
  [2402] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1270),
  [2405] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1074),
  [2408] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(43),
  [2411] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1122),
  [2414] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1283),
  [2417] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1037),
  [2420] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1056),
  [2423] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1041),
  [2426] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1069),
  [2429] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1069),
  [2432] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2076),
  [2435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1912),
  [2438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1913),
  [2441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1839),
  [2444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1843),
  [2447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1285),
  [2450] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1285),
  [2453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [2455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [2457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [2459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [2461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [2463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [2465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [2467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [2469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [2471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [2473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(315),
  [2475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [2477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [2479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [2481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [2483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(446),
  [2485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(593),
  [2487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [2489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(473),
  [2491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [2493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [2495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [2497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [2499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [2501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [2503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [2505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1317),
  [2507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [2509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1318),
  [2511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [2513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [2515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [2517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1321),
  [2519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [2521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1323),
  [2523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [2525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [2527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [2529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [2531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1309),
  [2533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [2535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 3, 0, 20),
  [2537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1640),
  [2539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [2541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [2543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1688),
  [2545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [2547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1647),
  [2551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1687),
  [2553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [2555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [2557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [2559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [2561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1653),
  [2563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [2565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1637),
  [2567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [2569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1633),
  [2571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1716),
  [2573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1636),
  [2575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [2577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1631),
  [2579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [2581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1643),
  [2583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [2585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [2587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [2589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [2591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1634),
  [2593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [2595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [2597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [2599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1645),
  [2601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [2603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [2605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [2607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1650),
  [2609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [2611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [2613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [2615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1629),
  [2617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [2619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1639),
  [2621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1652),
  [2625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [2627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [2629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [2631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1625),
  [2633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1626),
  [2635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [2637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [2639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [2641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [2643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [2647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [2649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [2651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [2653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [2655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [2657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [2659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [2661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [2663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [2665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2231),
  [2667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [2669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [2671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [2673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [2675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [2677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [2679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2315),
  [2681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [2683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [2685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2317),
  [2687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [2689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [2691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [2693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [2695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [2697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [2699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [2701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [2705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [2711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [2715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [2717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [2719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [2721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [2723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [2725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [2727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [2731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [2733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1654),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [2739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [2745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [2747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [2749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [2753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [2755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2316),
  [2757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [2759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [2761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [2765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [2767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [2769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [2771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [2775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [2777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1728),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [2783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [2785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1901),
  [2787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1118),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [2791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1093),
  [2793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [2797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1133),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [2801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [2805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1882),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [2811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2422),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [2815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2437),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [2831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2457),
  [2833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2351),
  [2835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2439),
  [2837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [2839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2380),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2152),
  [2843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [2847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [2849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [2851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2413),
  [2853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2186),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2124),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [2863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2414),
  [2865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2341),
  [2867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2425),
  [2869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2428),
  [2871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2430),
  [2873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2431),
  [2875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2432),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2433),
  [2879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2434),
  [2881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2435),
  [2883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2436),
  [2885] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0),
  [2887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [2889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [2891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [2893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [2895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [2897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [2899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1720),
  [2901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [2903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [2905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [2907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [2909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1689),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [2917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [2919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [2921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [2923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [2925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(924),
  [2927] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2124),
  [2929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [2931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2419),
  [2933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1888),
  [2935] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2186),
  [2937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2365),
  [2939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1904),
  [2941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2209),
  [2943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2209),
  [2945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(807),
  [2947] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2375),
  [2949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [2951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [2953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2155),
  [2955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [2957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2368),
  [2959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1885),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2141),
  [2963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2141),
  [2965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(908),
  [2967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2391),
  [2969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [2971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2203),
  [2973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [2975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2402),
  [2977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [2979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2162),
  [2981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [2983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2413),
  [2985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [2987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2164),
  [2989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2164),
  [2991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(835),
  [2993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2403),
  [2995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1819),
  [2997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2140),
  [2999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2140),
  [3001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [3003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2346),
  [3005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1822),
  [3007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [3009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2137),
  [3011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [3013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2444),
  [3015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(921),
  [3017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2146),
  [3019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2146),
  [3021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2364),
  [3023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [3025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [3027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2206),
  [3029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2373),
  [3031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [3033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2152),
  [3035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(973),
  [3037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2442),
  [3039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1816),
  [3041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2135),
  [3043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2380),
  [3045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [3047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [3049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2157),
  [3051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(972),
  [3053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2390),
  [3055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(762),
  [3057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [3059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2208),
  [3061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2446),
  [3063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(772),
  [3065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2166),
  [3067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2166),
  [3069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(856),
  [3071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2427),
  [3073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [3075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2122),
  [3077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2122),
  [3079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2369),
  [3081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [3083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [3085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [3087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2403),
  [3089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [3091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [3093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [3095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2444),
  [3097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [3099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2368),
  [3101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2454),
  [3104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [3106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [3110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_annotation, 2, 0, 0),
  [3112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation, 2, 0, 0),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [3116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(2356),
  [3119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1889),
  [3122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1128),
  [3125] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [3128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(340),
  [3131] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(340),
  [3134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(2053),
  [3137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1910),
  [3140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1911),
  [3143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1832),
  [3146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0), SHIFT_REPEAT(1835),
  [3149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2356),
  [3151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2423),
  [3153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2416),
  [3155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2424),
  [3157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [3159] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_annotation, 3, 0, 0),
  [3161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation, 3, 0, 0),
  [3163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 1, 0, 0),
  [3165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 1, 0, 0),
  [3167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [3169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [3171] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat2, 2, 0, 0),
  [3173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat2, 2, 0, 0),
  [3175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat2, 2, 0, 0), SHIFT_REPEAT(1748),
  [3178] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(2429),
  [3181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1972),
  [3184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2044),
  [3186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2392),
  [3188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [3192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [3196] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__param_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2044),
  [3199] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__param_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2392),
  [3202] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__param_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1081),
  [3205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__param_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1917),
  [3208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_modifier, 1, 0, 0),
  [3210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_modifier, 1, 0, 0),
  [3212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat2, 1, 0, 0),
  [3214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat2, 1, 0, 0),
  [3216] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat2, 2, 0, 0), SHIFT_REPEAT(1761),
  [3219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_modifier, 1, 0, 0),
  [3221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_access_modifier, 1, 0, 0),
  [3223] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0),
  [3225] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_map_repeat1, 2, 0, 0),
  [3227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [3229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [3231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [3233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [3235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [3237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [3239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [3241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2109),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [3245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [3247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [3249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3251] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generics_repeat1, 2, 0, 0), SHIFT_REPEAT(248),
  [3254] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generics_repeat1, 2, 0, 0), SHIFT_REPEAT(1917),
  [3257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [3259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [3261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__param_list_repeat1, 2, 0, 0),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [3265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__param_list, 3, 0, 0),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [3269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [3273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__param_list, 2, 0, 0),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [3277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [3279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [3283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [3285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [3287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [3289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generics_repeat1, 2, 0, 0),
  [3291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [3293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2038),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [3305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2461),
  [3307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2370),
  [3309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_groovy_doc_repeat1, 2, 0, 0), SHIFT_REPEAT(1898),
  [3312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_groovy_doc_repeat1, 2, 0, 0),
  [3314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_groovy_doc_repeat1, 2, 0, 0), SHIFT_REPEAT(2421),
  [3317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_groovy_doc_repeat1, 2, 0, 0), SHIFT_REPEAT(2410),
  [3320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [3322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2161),
  [3324] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1898),
  [3326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [3328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2421),
  [3330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2410),
  [3332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2126),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2464),
  [3338] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [3340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2322),
  [3342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2167),
  [3344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [3346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(706),
  [3348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(710),
  [3350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [3352] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1171),
  [3355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [3357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [3359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(799),
  [3361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(913),
  [3363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [3365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [3367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2160),
  [3369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1997),
  [3371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2273),
  [3373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [3375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [3377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1927),
  [3379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [3381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1107),
  [3383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [3385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2255),
  [3387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [3389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [3391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1955),
  [3393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2271),
  [3395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(396),
  [3397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1068),
  [3399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1073),
  [3401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0),
  [3403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0), SHIFT_REPEAT(1997),
  [3406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2160),
  [3409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0), SHIFT_REPEAT(1997),
  [3412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2273),
  [3415] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat4, 2, 0, 0), SHIFT_REPEAT(1949),
  [3418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat4, 2, 0, 0), SHIFT_REPEAT(1955),
  [3421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat4, 2, 0, 0),
  [3423] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat4, 2, 0, 0), SHIFT_REPEAT(2271),
  [3426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [3428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [3430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(266),
  [3432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1117),
  [3434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [3436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2112),
  [3438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2193),
  [3440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [3442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [3444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [3446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat3, 2, 0, 0),
  [3448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat3, 2, 0, 0), SHIFT_REPEAT(1927),
  [3451] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat3, 2, 0, 0), SHIFT_REPEAT(1927),
  [3454] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat3, 2, 0, 0), SHIFT_REPEAT(2280),
  [3457] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dotted_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1969),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2099),
  [3462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2099),
  [3464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2279),
  [3466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_doc_throws, 2, 0, 0),
  [3468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_doc_throws, 2, 0, 0),
  [3470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [3472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2168),
  [3474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [3476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2399),
  [3478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2002),
  [3480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [3482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [3484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [3486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1066),
  [3488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(257),
  [3490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0),
  [3492] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2168),
  [3495] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2002),
  [3498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2399),
  [3501] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0), SHIFT_REPEAT(2002),
  [3504] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat1, 2, 0, 0),
  [3506] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2099),
  [3509] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2099),
  [3512] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2279),
  [3515] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__juxt_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1170),
  [3518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_groovy_doc_param, 2, 0, 0),
  [3520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_doc_param, 2, 0, 0),
  [3522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [3524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [3526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_groovy_doc_repeat1, 1, 0, 0),
  [3528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_groovy_doc_repeat1, 1, 0, 0),
  [3530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [3532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [3534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [3536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(267),
  [3538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(281),
  [3540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [3542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [3544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(386),
  [3546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(28),
  [3548] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [3550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_builtintype, 1, 0, 0),
  [3552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(16),
  [3554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(23),
  [3556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2205),
  [3558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [3562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [3564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat3, 1, 0, 0),
  [3566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat3, 1, 0, 0),
  [3568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [3570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2371),
  [3576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [3578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [3582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_param, 3, 0, 43),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [3588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [3590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [3594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat4, 1, 0, 6),
  [3596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat4, 1, 0, 6),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [3600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat4, 1, 0, 0),
  [3602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat4, 1, 0, 0),
  [3604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolation, 2, 0, 0),
  [3606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 2, 0, 0),
  [3608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [3610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 3, 0, 0),
  [3612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat2, 3, 0, 0),
  [3614] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [3616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generics, 3, 0, 0),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [3622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [3624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1723),
  [3626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1724),
  [3628] = {.entry = {.count = 1, .reusable = false}}, SHIFT(38),
  [3630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [3632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generics, 5, 0, 0),
  [3634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(482),
  [3636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolation, 4, 0, 0),
  [3638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 4, 0, 0),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [3642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [3644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(362),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [3648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1057),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [3652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type, 2, 0, 0),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [3656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [3660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [3662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [3664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat2, 1, 0, 0),
  [3666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat2, 1, 0, 0),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [3670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat2, 1, 0, 0),
  [3672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat2, 1, 0, 0),
  [3674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat2, 2, 0, 0),
  [3676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_with_generics, 2, 0, 0),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [3688] = {.entry = {.count = 1, .reusable = false}}, SHIFT(668),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [3692] = {.entry = {.count = 1, .reusable = false}}, SHIFT(666),
  [3694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat2, 2, 0, 0),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1580),
  [3700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0),
  [3702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1207),
  [3705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2371),
  [3708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [3710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(424),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [3716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [3718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generics, 4, 0, 0),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [3722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(20),
  [3724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2150),
  [3728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2150),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1261),
  [3734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1818),
  [3736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2097),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1808),
  [3742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1794),
  [3744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [3746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [3748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [3750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [3754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [3756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [3758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [3760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1797),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [3764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [3766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(838),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [3770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2055),
  [3772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(696),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [3780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat1, 2, 0, 0),
  [3782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2150),
  [3785] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2150),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [3794] = {.entry = {.count = 1, .reusable = false}}, SHIFT(679),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [3798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [3804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [3806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [3808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2080),
  [3812] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__interpolate_string_repeat1, 1, 0, 0),
  [3814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__interpolate_string_repeat1, 1, 0, 0),
  [3816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [3818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [3820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [3822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [3824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [3828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [3830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [3832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1792),
  [3834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [3836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [3838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1937),
  [3840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [3842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [3844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2009),
  [3846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_param, 1, 0, 3),
  [3848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [3850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(701),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [3858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2010),
  [3860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [3862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(814),
  [3864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [3866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2017),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [3870] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2022),
  [3872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [3874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [3876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2026),
  [3878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1944),
  [3880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(858),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [3886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [3888] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1982),
  [3890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_block_repeat1, 1, 0, 0),
  [3892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2039),
  [3894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [3896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__plain_string_repeat1, 1, 0, 0),
  [3898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__plain_string_repeat1, 1, 0, 0),
  [3900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [3902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [3904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1579),
  [3906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1925),
  [3908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1929),
  [3910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1940),
  [3912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [3914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2273),
  [3916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(694),
  [3918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1947),
  [3920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [3924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1953),
  [3926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [3928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1958),
  [3930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [3932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2350),
  [3934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2006),
  [3936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [3938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [3940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [3942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [3944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [3946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [3948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [3950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [3952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [3954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [3956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [3958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [3960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(839),
  [3962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [3964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [3966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [3968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1919),
  [3970] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_parameters_repeat2, 2, 0, 0), SHIFT_REPEAT(147),
  [3973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_parameters_repeat2, 2, 0, 0),
  [3975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [3977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [3979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_parameters_repeat1, 2, 0, 0),
  [3983] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(1548),
  [3986] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(2312),
  [3989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2012),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [3993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1984),
  [3995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [3997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [3999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [4001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1990),
  [4003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1970),
  [4005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1994),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [4009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1999),
  [4011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1938),
  [4013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1567),
  [4015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(989),
  [4017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [4019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__param_list, 1, 0, 0),
  [4021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1622),
  [4023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [4025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [4027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [4029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1782),
  [4031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [4033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1769),
  [4035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1618),
  [4037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [4039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [4041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [4043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1608),
  [4045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [4047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [4049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [4051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [4053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2282),
  [4055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2290),
  [4057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1616),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [4063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [4065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1774),
  [4067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [4069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [4071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1621),
  [4073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [4075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [4077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [4079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [4081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1624),
  [4083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [4085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [4087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [4089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [4091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [4093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [4097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_parameters, 4, 0, 0),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2333),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2338),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [4105] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_parameters_repeat1, 2, 0, 0),
  [4107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [4109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [4111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [4115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [4117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1775),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [4121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [4123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment, 1, 0, 0),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [4129] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_parameters, 5, 0, 0),
  [4131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [4133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [4137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1980),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1772),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1780),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [4153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [4155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [4157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [4159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [4161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [4163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_parameters, 3, 0, 0),
  [4165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [4167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [4169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [4171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [4173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2334),
  [4175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [4177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [4179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [4181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [4183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1806),
  [4185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline, 2, 0, 0),
  [4187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1771),
  [4189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [4191] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [4193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [4195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1005),
  [4197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [4199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [4201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [4203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [4205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [4207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [4209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [4211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [4213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [4215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [4217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1807),
  [4219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [4221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2292),
  [4223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1790),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [4231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 3, 0, 0),
  [4233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [4235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [4237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [4239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [4241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [4243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [4245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2463),
  [4247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [4251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [4253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [4255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [4257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2301),
  [4259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [4261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [4263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [4265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [4267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [4269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [4271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [4273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [4275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [4279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym___immediately_invoked_closure, 2, 0, 8),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [4283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1347),
  [4285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [4287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [4293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [4295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [4297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [4299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [4301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [4303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [4305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [4307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1813),
  [4309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [4311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [4313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [4315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [4317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [4319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1888),
  [4321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [4323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [4325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__immediately_invoked_closure, 1, 0, 5),
  [4327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1690),
  [4329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [4331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [4333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [4335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [4337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comment, 1, 0, 0),
  [4339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_doc, 5, 0, 0),
  [4341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comment, 2, 0, 0),
  [4343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_groovy_doc, 4, 0, 0),
};

#ifdef __cplusplus
extern "C" {
#endif
#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_groovy(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .primary_state_ids = ts_primary_state_ids,
    .name = "groovy",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 0,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
