/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 304
#define LARGE_STATE_COUNT 2
#define SYMBOL_COUNT 126
#define ALIAS_COUNT 4
#define TOKEN_COUNT 83
#define EXTERNAL_TOKEN_COUNT 3
#define FIELD_COUNT 1
#define MAX_ALIAS_SEQUENCE_LENGTH 10
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 17
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym__normal_bare_identifier = 1,
  anon_sym_SLASH_DASH = 2,
  anon_sym_LBRACE = 3,
  anon_sym_RBRACE = 4,
  anon_sym_SEMI = 5,
  sym__identifier_char = 6,
  sym___identifier_char_no_digit = 7,
  anon_sym_null = 8,
  anon_sym_i8 = 9,
  anon_sym_i16 = 10,
  anon_sym_i32 = 11,
  anon_sym_i64 = 12,
  anon_sym_u8 = 13,
  anon_sym_u16 = 14,
  anon_sym_u32 = 15,
  anon_sym_u64 = 16,
  anon_sym_isize = 17,
  anon_sym_usize = 18,
  anon_sym_f32 = 19,
  anon_sym_f64 = 20,
  anon_sym_decimal64 = 21,
  anon_sym_decimal128 = 22,
  anon_sym_date_DASHtime = 23,
  anon_sym_time = 24,
  anon_sym_date = 25,
  anon_sym_duration = 26,
  anon_sym_decimal = 27,
  anon_sym_currency = 28,
  anon_sym_country_DASH2 = 29,
  anon_sym_country_DASH3 = 30,
  anon_sym_country_DASHsubdivision = 31,
  anon_sym_email = 32,
  anon_sym_idn_DASHemail = 33,
  anon_sym_hostname = 34,
  anon_sym_idn_DASHhostname = 35,
  anon_sym_ipv4 = 36,
  anon_sym_ipv6 = 37,
  anon_sym_url = 38,
  anon_sym_url_DASHreference = 39,
  anon_sym_irl = 40,
  anon_sym_iri_DASHreference = 41,
  anon_sym_url_DASHtemplate = 42,
  anon_sym_uuid = 43,
  anon_sym_regex = 44,
  anon_sym_base64 = 45,
  anon_sym_EQ = 46,
  anon_sym_LPAREN = 47,
  anon_sym_RPAREN = 48,
  anon_sym_DQUOTE = 49,
  aux_sym__escaped_string_token1 = 50,
  sym_escape = 51,
  sym__hex_digit = 52,
  anon_sym_DOT = 53,
  anon_sym_e = 54,
  anon_sym_E = 55,
  anon_sym__ = 56,
  sym__digit = 57,
  anon_sym_PLUS = 58,
  anon_sym_DASH = 59,
  anon_sym_0x = 60,
  anon_sym_0o = 61,
  aux_sym__octal_token1 = 62,
  anon_sym_0b = 63,
  anon_sym_0 = 64,
  anon_sym_1 = 65,
  anon_sym_true = 66,
  anon_sym_false = 67,
  anon_sym_BSLASH = 68,
  aux_sym__newline_token1 = 69,
  aux_sym__newline_token2 = 70,
  aux_sym__newline_token3 = 71,
  aux_sym__newline_token4 = 72,
  aux_sym__newline_token5 = 73,
  aux_sym__newline_token6 = 74,
  aux_sym__newline_token7 = 75,
  sym__bom = 76,
  sym__unicode_space = 77,
  anon_sym_SLASH_SLASH = 78,
  aux_sym_single_line_comment_token1 = 79,
  sym__eof = 80,
  sym_multi_line_comment = 81,
  sym__raw_string = 82,
  sym_document = 83,
  sym_node = 84,
  sym_node_field = 85,
  sym__node_field_comment = 86,
  sym__node_field = 87,
  sym_node_children = 88,
  sym__node_space = 89,
  sym__node_terminator = 90,
  sym_identifier = 91,
  sym__bare_identifier = 92,
  sym_keyword = 93,
  sym_annotation_type = 94,
  sym_prop = 95,
  sym_value = 96,
  sym_type = 97,
  sym_string = 98,
  sym__escaped_string = 99,
  sym_number = 100,
  sym__decimal = 101,
  sym__exponent = 102,
  sym__integer = 103,
  sym__sign = 104,
  sym__hex = 105,
  sym__octal = 106,
  sym__binary = 107,
  sym_boolean = 108,
  sym__escline = 109,
  sym__linespace = 110,
  sym__newline = 111,
  sym__ws = 112,
  sym_single_line_comment = 113,
  aux_sym_document_repeat1 = 114,
  aux_sym_document_repeat2 = 115,
  aux_sym_node_repeat1 = 116,
  aux_sym_node_repeat2 = 117,
  aux_sym_node_repeat3 = 118,
  aux_sym__bare_identifier_repeat1 = 119,
  aux_sym__escaped_string_repeat1 = 120,
  aux_sym__integer_repeat1 = 121,
  aux_sym__hex_repeat1 = 122,
  aux_sym__octal_repeat1 = 123,
  aux_sym__binary_repeat1 = 124,
  aux_sym_single_line_comment_repeat1 = 125,
  alias_sym_decimal = 126,
  alias_sym_node_children_comment = 127,
  alias_sym_node_field_comment = 128,
  alias_sym_string_fragment = 129,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__normal_bare_identifier] = "_normal_bare_identifier",
  [anon_sym_SLASH_DASH] = "node_comment",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_SEMI] = ";",
  [sym__identifier_char] = "_identifier_char",
  [sym___identifier_char_no_digit] = "__identifier_char_no_digit",
  [anon_sym_null] = "null",
  [anon_sym_i8] = "i8",
  [anon_sym_i16] = "i16",
  [anon_sym_i32] = "i32",
  [anon_sym_i64] = "i64",
  [anon_sym_u8] = "u8",
  [anon_sym_u16] = "u16",
  [anon_sym_u32] = "u32",
  [anon_sym_u64] = "u64",
  [anon_sym_isize] = "isize",
  [anon_sym_usize] = "usize",
  [anon_sym_f32] = "f32",
  [anon_sym_f64] = "f64",
  [anon_sym_decimal64] = "decimal64",
  [anon_sym_decimal128] = "decimal128",
  [anon_sym_date_DASHtime] = "date-time",
  [anon_sym_time] = "time",
  [anon_sym_date] = "date",
  [anon_sym_duration] = "duration",
  [anon_sym_decimal] = "decimal",
  [anon_sym_currency] = "currency",
  [anon_sym_country_DASH2] = "country-2",
  [anon_sym_country_DASH3] = "country-3",
  [anon_sym_country_DASHsubdivision] = "country-subdivision",
  [anon_sym_email] = "email",
  [anon_sym_idn_DASHemail] = "idn-email",
  [anon_sym_hostname] = "hostname",
  [anon_sym_idn_DASHhostname] = "idn-hostname",
  [anon_sym_ipv4] = "ipv4",
  [anon_sym_ipv6] = "ipv6",
  [anon_sym_url] = "url",
  [anon_sym_url_DASHreference] = "url-reference",
  [anon_sym_irl] = "irl",
  [anon_sym_iri_DASHreference] = "iri-reference",
  [anon_sym_url_DASHtemplate] = "url-template",
  [anon_sym_uuid] = "uuid",
  [anon_sym_regex] = "regex",
  [anon_sym_base64] = "base64",
  [anon_sym_EQ] = "=",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym__escaped_string_token1] = "_escaped_string_token1",
  [sym_escape] = "escape",
  [sym__hex_digit] = "_hex_digit",
  [anon_sym_DOT] = ".",
  [anon_sym_e] = "e",
  [anon_sym_E] = "E",
  [anon_sym__] = "_",
  [sym__digit] = "_digit",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_0x] = "0x",
  [anon_sym_0o] = "0o",
  [aux_sym__octal_token1] = "_octal_token1",
  [anon_sym_0b] = "0b",
  [anon_sym_0] = "0",
  [anon_sym_1] = "1",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_BSLASH] = "\\",
  [aux_sym__newline_token1] = "_newline_token1",
  [aux_sym__newline_token2] = "_newline_token2",
  [aux_sym__newline_token3] = "_newline_token3",
  [aux_sym__newline_token4] = "_newline_token4",
  [aux_sym__newline_token5] = "_newline_token5",
  [aux_sym__newline_token6] = "_newline_token6",
  [aux_sym__newline_token7] = "_newline_token7",
  [sym__bom] = "_bom",
  [sym__unicode_space] = "_unicode_space",
  [anon_sym_SLASH_SLASH] = "//",
  [aux_sym_single_line_comment_token1] = "single_line_comment_token1",
  [sym__eof] = "_eof",
  [sym_multi_line_comment] = "multi_line_comment",
  [sym__raw_string] = "_raw_string",
  [sym_document] = "document",
  [sym_node] = "node",
  [sym_node_field] = "node_field",
  [sym__node_field_comment] = "_node_field_comment",
  [sym__node_field] = "_node_field",
  [sym_node_children] = "node_children",
  [sym__node_space] = "_node_space",
  [sym__node_terminator] = "_node_terminator",
  [sym_identifier] = "identifier",
  [sym__bare_identifier] = "_bare_identifier",
  [sym_keyword] = "keyword",
  [sym_annotation_type] = "annotation_type",
  [sym_prop] = "prop",
  [sym_value] = "value",
  [sym_type] = "type",
  [sym_string] = "string",
  [sym__escaped_string] = "_escaped_string",
  [sym_number] = "number",
  [sym__decimal] = "_decimal",
  [sym__exponent] = "exponent",
  [sym__integer] = "_integer",
  [sym__sign] = "_sign",
  [sym__hex] = "_hex",
  [sym__octal] = "_octal",
  [sym__binary] = "_binary",
  [sym_boolean] = "boolean",
  [sym__escline] = "_escline",
  [sym__linespace] = "_linespace",
  [sym__newline] = "_newline",
  [sym__ws] = "_ws",
  [sym_single_line_comment] = "single_line_comment",
  [aux_sym_document_repeat1] = "document_repeat1",
  [aux_sym_document_repeat2] = "document_repeat2",
  [aux_sym_node_repeat1] = "node_repeat1",
  [aux_sym_node_repeat2] = "node_repeat2",
  [aux_sym_node_repeat3] = "node_repeat3",
  [aux_sym__bare_identifier_repeat1] = "_bare_identifier_repeat1",
  [aux_sym__escaped_string_repeat1] = "_escaped_string_repeat1",
  [aux_sym__integer_repeat1] = "_integer_repeat1",
  [aux_sym__hex_repeat1] = "_hex_repeat1",
  [aux_sym__octal_repeat1] = "_octal_repeat1",
  [aux_sym__binary_repeat1] = "_binary_repeat1",
  [aux_sym_single_line_comment_repeat1] = "single_line_comment_repeat1",
  [alias_sym_decimal] = "decimal",
  [alias_sym_node_children_comment] = "node_children_comment",
  [alias_sym_node_field_comment] = "node_field_comment",
  [alias_sym_string_fragment] = "string_fragment",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__normal_bare_identifier] = sym__normal_bare_identifier,
  [anon_sym_SLASH_DASH] = anon_sym_SLASH_DASH,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [sym__identifier_char] = sym__identifier_char,
  [sym___identifier_char_no_digit] = sym___identifier_char_no_digit,
  [anon_sym_null] = anon_sym_null,
  [anon_sym_i8] = anon_sym_i8,
  [anon_sym_i16] = anon_sym_i16,
  [anon_sym_i32] = anon_sym_i32,
  [anon_sym_i64] = anon_sym_i64,
  [anon_sym_u8] = anon_sym_u8,
  [anon_sym_u16] = anon_sym_u16,
  [anon_sym_u32] = anon_sym_u32,
  [anon_sym_u64] = anon_sym_u64,
  [anon_sym_isize] = anon_sym_isize,
  [anon_sym_usize] = anon_sym_usize,
  [anon_sym_f32] = anon_sym_f32,
  [anon_sym_f64] = anon_sym_f64,
  [anon_sym_decimal64] = anon_sym_decimal64,
  [anon_sym_decimal128] = anon_sym_decimal128,
  [anon_sym_date_DASHtime] = anon_sym_date_DASHtime,
  [anon_sym_time] = anon_sym_time,
  [anon_sym_date] = anon_sym_date,
  [anon_sym_duration] = anon_sym_duration,
  [anon_sym_decimal] = anon_sym_decimal,
  [anon_sym_currency] = anon_sym_currency,
  [anon_sym_country_DASH2] = anon_sym_country_DASH2,
  [anon_sym_country_DASH3] = anon_sym_country_DASH3,
  [anon_sym_country_DASHsubdivision] = anon_sym_country_DASHsubdivision,
  [anon_sym_email] = anon_sym_email,
  [anon_sym_idn_DASHemail] = anon_sym_idn_DASHemail,
  [anon_sym_hostname] = anon_sym_hostname,
  [anon_sym_idn_DASHhostname] = anon_sym_idn_DASHhostname,
  [anon_sym_ipv4] = anon_sym_ipv4,
  [anon_sym_ipv6] = anon_sym_ipv6,
  [anon_sym_url] = anon_sym_url,
  [anon_sym_url_DASHreference] = anon_sym_url_DASHreference,
  [anon_sym_irl] = anon_sym_irl,
  [anon_sym_iri_DASHreference] = anon_sym_iri_DASHreference,
  [anon_sym_url_DASHtemplate] = anon_sym_url_DASHtemplate,
  [anon_sym_uuid] = anon_sym_uuid,
  [anon_sym_regex] = anon_sym_regex,
  [anon_sym_base64] = anon_sym_base64,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym__escaped_string_token1] = aux_sym__escaped_string_token1,
  [sym_escape] = sym_escape,
  [sym__hex_digit] = sym__hex_digit,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_e] = anon_sym_e,
  [anon_sym_E] = anon_sym_E,
  [anon_sym__] = anon_sym__,
  [sym__digit] = sym__digit,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_0x] = anon_sym_0x,
  [anon_sym_0o] = anon_sym_0o,
  [aux_sym__octal_token1] = aux_sym__octal_token1,
  [anon_sym_0b] = anon_sym_0b,
  [anon_sym_0] = anon_sym_0,
  [anon_sym_1] = anon_sym_1,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [aux_sym__newline_token1] = aux_sym__newline_token1,
  [aux_sym__newline_token2] = aux_sym__newline_token2,
  [aux_sym__newline_token3] = aux_sym__newline_token3,
  [aux_sym__newline_token4] = aux_sym__newline_token4,
  [aux_sym__newline_token5] = aux_sym__newline_token5,
  [aux_sym__newline_token6] = aux_sym__newline_token6,
  [aux_sym__newline_token7] = aux_sym__newline_token7,
  [sym__bom] = sym__bom,
  [sym__unicode_space] = sym__unicode_space,
  [anon_sym_SLASH_SLASH] = anon_sym_SLASH_SLASH,
  [aux_sym_single_line_comment_token1] = aux_sym_single_line_comment_token1,
  [sym__eof] = sym__eof,
  [sym_multi_line_comment] = sym_multi_line_comment,
  [sym__raw_string] = sym__raw_string,
  [sym_document] = sym_document,
  [sym_node] = sym_node,
  [sym_node_field] = sym_node_field,
  [sym__node_field_comment] = sym__node_field_comment,
  [sym__node_field] = sym__node_field,
  [sym_node_children] = sym_node_children,
  [sym__node_space] = sym__node_space,
  [sym__node_terminator] = sym__node_terminator,
  [sym_identifier] = sym_identifier,
  [sym__bare_identifier] = sym__bare_identifier,
  [sym_keyword] = sym_keyword,
  [sym_annotation_type] = sym_annotation_type,
  [sym_prop] = sym_prop,
  [sym_value] = sym_value,
  [sym_type] = sym_type,
  [sym_string] = sym_string,
  [sym__escaped_string] = sym__escaped_string,
  [sym_number] = sym_number,
  [sym__decimal] = sym__decimal,
  [sym__exponent] = sym__exponent,
  [sym__integer] = sym__integer,
  [sym__sign] = sym__sign,
  [sym__hex] = sym__hex,
  [sym__octal] = sym__octal,
  [sym__binary] = sym__binary,
  [sym_boolean] = sym_boolean,
  [sym__escline] = sym__escline,
  [sym__linespace] = sym__linespace,
  [sym__newline] = sym__newline,
  [sym__ws] = sym__ws,
  [sym_single_line_comment] = sym_single_line_comment,
  [aux_sym_document_repeat1] = aux_sym_document_repeat1,
  [aux_sym_document_repeat2] = aux_sym_document_repeat2,
  [aux_sym_node_repeat1] = aux_sym_node_repeat1,
  [aux_sym_node_repeat2] = aux_sym_node_repeat2,
  [aux_sym_node_repeat3] = aux_sym_node_repeat3,
  [aux_sym__bare_identifier_repeat1] = aux_sym__bare_identifier_repeat1,
  [aux_sym__escaped_string_repeat1] = aux_sym__escaped_string_repeat1,
  [aux_sym__integer_repeat1] = aux_sym__integer_repeat1,
  [aux_sym__hex_repeat1] = aux_sym__hex_repeat1,
  [aux_sym__octal_repeat1] = aux_sym__octal_repeat1,
  [aux_sym__binary_repeat1] = aux_sym__binary_repeat1,
  [aux_sym_single_line_comment_repeat1] = aux_sym_single_line_comment_repeat1,
  [alias_sym_decimal] = alias_sym_decimal,
  [alias_sym_node_children_comment] = alias_sym_node_children_comment,
  [alias_sym_node_field_comment] = alias_sym_node_field_comment,
  [alias_sym_string_fragment] = alias_sym_string_fragment,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__normal_bare_identifier] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SLASH_DASH] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [sym__identifier_char] = {
    .visible = false,
    .named = true,
  },
  [sym___identifier_char_no_digit] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_null] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_i8] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_i16] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_i32] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_i64] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u16] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u32] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u64] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_isize] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_usize] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_f32] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_f64] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_decimal64] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_decimal128] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_date_DASHtime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_time] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_date] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_duration] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_decimal] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_currency] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_country_DASH2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_country_DASH3] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_country_DASHsubdivision] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_email] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_idn_DASHemail] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_hostname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_idn_DASHhostname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ipv4] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ipv6] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_url] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_url_DASHreference] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_irl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_iri_DASHreference] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_url_DASHtemplate] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_uuid] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_regex] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_base64] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__escaped_string_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_escape] = {
    .visible = true,
    .named = true,
  },
  [sym__hex_digit] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_e] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_E] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [sym__digit] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_0x] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_0o] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__octal_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_0b] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_0] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__newline_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token4] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token5] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token6] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__newline_token7] = {
    .visible = false,
    .named = false,
  },
  [sym__bom] = {
    .visible = false,
    .named = true,
  },
  [sym__unicode_space] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SLASH_SLASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_single_line_comment_token1] = {
    .visible = false,
    .named = false,
  },
  [sym__eof] = {
    .visible = false,
    .named = true,
  },
  [sym_multi_line_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__raw_string] = {
    .visible = false,
    .named = true,
  },
  [sym_document] = {
    .visible = true,
    .named = true,
  },
  [sym_node] = {
    .visible = true,
    .named = true,
  },
  [sym_node_field] = {
    .visible = true,
    .named = true,
  },
  [sym__node_field_comment] = {
    .visible = false,
    .named = true,
  },
  [sym__node_field] = {
    .visible = false,
    .named = true,
  },
  [sym_node_children] = {
    .visible = true,
    .named = true,
  },
  [sym__node_space] = {
    .visible = false,
    .named = true,
  },
  [sym__node_terminator] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__bare_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym_keyword] = {
    .visible = true,
    .named = true,
  },
  [sym_annotation_type] = {
    .visible = true,
    .named = true,
  },
  [sym_prop] = {
    .visible = true,
    .named = true,
  },
  [sym_value] = {
    .visible = true,
    .named = true,
  },
  [sym_type] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym__escaped_string] = {
    .visible = false,
    .named = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym__decimal] = {
    .visible = false,
    .named = true,
  },
  [sym__exponent] = {
    .visible = true,
    .named = true,
  },
  [sym__integer] = {
    .visible = false,
    .named = true,
  },
  [sym__sign] = {
    .visible = false,
    .named = true,
  },
  [sym__hex] = {
    .visible = false,
    .named = true,
  },
  [sym__octal] = {
    .visible = false,
    .named = true,
  },
  [sym__binary] = {
    .visible = false,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym__escline] = {
    .visible = false,
    .named = true,
  },
  [sym__linespace] = {
    .visible = false,
    .named = true,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym__ws] = {
    .visible = false,
    .named = true,
  },
  [sym_single_line_comment] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_document_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_document_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_node_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_node_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_node_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__bare_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__escaped_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__integer_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hex_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__octal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__binary_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_single_line_comment_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_decimal] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_node_children_comment] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_node_field_comment] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_string_fragment] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_children = 1,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_children] = "children",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [7] = {.index = 2, .length = 1},
  [8] = {.index = 2, .length = 1},
  [11] = {.index = 3, .length = 1},
  [12] = {.index = 3, .length = 1},
  [14] = {.index = 4, .length = 1},
  [15] = {.index = 4, .length = 1},
  [16] = {.index = 5, .length = 1},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_children, 1},
  [1] =
    {field_children, 2},
  [2] =
    {field_children, 3},
  [3] =
    {field_children, 4},
  [4] =
    {field_children, 5},
  [5] =
    {field_children, 6},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [1] = alias_sym_string_fragment,
  },
  [4] = {
    [1] = anon_sym_SLASH_DASH,
  },
  [5] = {
    [0] = alias_sym_node_children_comment,
  },
  [6] = {
    [0] = alias_sym_node_field_comment,
    [1] = alias_sym_node_field_comment,
  },
  [8] = {
    [1] = anon_sym_SLASH_DASH,
  },
  [9] = {
    [0] = alias_sym_node_field_comment,
    [1] = alias_sym_node_field_comment,
    [2] = alias_sym_node_field_comment,
  },
  [10] = {
    [2] = alias_sym_decimal,
  },
  [12] = {
    [1] = anon_sym_SLASH_DASH,
  },
  [13] = {
    [3] = alias_sym_decimal,
  },
  [15] = {
    [1] = anon_sym_SLASH_DASH,
  },
  [16] = {
    [1] = anon_sym_SLASH_DASH,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__node_field, 2,
    sym__node_field,
    alias_sym_node_field_comment,
  sym__integer, 2,
    sym__integer,
    alias_sym_decimal,
  aux_sym_node_repeat1, 3,
    aux_sym_node_repeat1,
    alias_sym_node_field_comment,
    anon_sym_SLASH_DASH,
  aux_sym__escaped_string_repeat1, 2,
    aux_sym__escaped_string_repeat1,
    alias_sym_string_fragment,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 58,
  [60] = 56,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 15,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 80,
  [81] = 81,
  [82] = 82,
  [83] = 83,
  [84] = 84,
  [85] = 85,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 30,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 47,
  [123] = 52,
  [124] = 50,
  [125] = 125,
  [126] = 53,
  [127] = 33,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 132,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 138,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 143,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 147,
  [148] = 148,
  [149] = 149,
  [150] = 150,
  [151] = 151,
  [152] = 152,
  [153] = 153,
  [154] = 154,
  [155] = 155,
  [156] = 156,
  [157] = 157,
  [158] = 158,
  [159] = 159,
  [160] = 160,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 165,
  [166] = 166,
  [167] = 167,
  [168] = 168,
  [169] = 169,
  [170] = 170,
  [171] = 171,
  [172] = 172,
  [173] = 57,
  [174] = 174,
  [175] = 175,
  [176] = 61,
  [177] = 177,
  [178] = 178,
  [179] = 179,
  [180] = 180,
  [181] = 181,
  [182] = 55,
  [183] = 183,
  [184] = 184,
  [185] = 185,
  [186] = 186,
  [187] = 15,
  [188] = 188,
  [189] = 189,
  [190] = 190,
  [191] = 191,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 196,
  [197] = 30,
  [198] = 198,
  [199] = 53,
  [200] = 200,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 207,
  [208] = 33,
  [209] = 209,
  [210] = 210,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 47,
  [218] = 218,
  [219] = 50,
  [220] = 220,
  [221] = 52,
  [222] = 222,
  [223] = 223,
  [224] = 223,
  [225] = 223,
  [226] = 226,
  [227] = 226,
  [228] = 226,
  [229] = 223,
  [230] = 226,
  [231] = 53,
  [232] = 232,
  [233] = 233,
  [234] = 57,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 61,
  [243] = 55,
  [244] = 58,
  [245] = 245,
  [246] = 246,
  [247] = 56,
  [248] = 248,
  [249] = 249,
  [250] = 249,
  [251] = 251,
  [252] = 249,
  [253] = 251,
  [254] = 251,
  [255] = 249,
  [256] = 251,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 15,
  [262] = 262,
  [263] = 263,
  [264] = 30,
  [265] = 265,
  [266] = 50,
  [267] = 47,
  [268] = 52,
  [269] = 33,
  [270] = 270,
  [271] = 271,
  [272] = 188,
  [273] = 55,
  [274] = 56,
  [275] = 58,
  [276] = 276,
  [277] = 190,
  [278] = 61,
  [279] = 279,
  [280] = 57,
  [281] = 281,
  [282] = 189,
  [283] = 281,
  [284] = 276,
  [285] = 285,
  [286] = 286,
  [287] = 207,
  [288] = 288,
  [289] = 289,
  [290] = 196,
  [291] = 291,
  [292] = 292,
  [293] = 293,
  [294] = 191,
  [295] = 193,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 192,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
};

static const TSCharacterRange sym__normal_bare_identifier_character_set_2[] = {
  {'!', '!'}, {'#', '\''}, {'*', '+'}, {'-', '.'}, {'0', ':'}, {'?', 'Z'}, {'^', '_'}, {'a', 'z'},
  {'|', '|'}, {'~', '~'}, {0xa9, 0xaa}, {0xae, 0xae}, {0xb2, 0xb3}, {0xb5, 0xb5}, {0xb9, 0xba}, {0xbc, 0xbe},
  {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x300, 0x374},
  {0x376, 0x377}, {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf}, {0x5c1, 0x5c2},
  {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3}, {0x6d5, 0x6dc},
  {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa}, {0x7fd, 0x7fd},
  {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963}, {0x966, 0x96f},
  {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bc, 0x9c4},
  {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9f4, 0x9f9}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb77}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbf2}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc78, 0xc7e}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3},
  {0xcb5, 0xcb9}, {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef},
  {0xcf1, 0xcf3}, {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd63}, {0xd66, 0xd78},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf33}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x137c}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x17f0, 0x17f9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e},
  {0x1920, 0x192b}, {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b},
  {0x1a20, 0x1a5e}, {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x203c, 0x203c}, {0x2049, 0x2049}, {0x2070, 0x2071},
  {0x2074, 0x2079}, {0x207f, 0x2089}, {0x2090, 0x209c}, {0x20d0, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115},
  {0x2119, 0x211d}, {0x2122, 0x2122}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x212f, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2150, 0x2189}, {0x2194, 0x2199}, {0x21a9, 0x21aa}, {0x231a, 0x231b}, {0x2328, 0x2328}, {0x23cf, 0x23cf},
  {0x23e9, 0x23f3}, {0x23f8, 0x23fa}, {0x2460, 0x249b}, {0x24c2, 0x24c2}, {0x24ea, 0x24ff}, {0x25aa, 0x25ab}, {0x25b6, 0x25b6}, {0x25c0, 0x25c0},
  {0x25fb, 0x25fe}, {0x2600, 0x2604}, {0x260e, 0x260e}, {0x2611, 0x2611}, {0x2614, 0x2615}, {0x2618, 0x2618}, {0x261d, 0x261d}, {0x2620, 0x2620},
  {0x2622, 0x2623}, {0x2626, 0x2626}, {0x262a, 0x262a}, {0x262e, 0x262f}, {0x2638, 0x263a}, {0x2640, 0x2640}, {0x2642, 0x2642}, {0x2648, 0x2653},
  {0x265f, 0x2660}, {0x2663, 0x2663}, {0x2665, 0x2666}, {0x2668, 0x2668}, {0x267b, 0x267b}, {0x267e, 0x267f}, {0x2692, 0x2697}, {0x2699, 0x2699},
  {0x269b, 0x269c}, {0x26a0, 0x26a1}, {0x26a7, 0x26a7}, {0x26aa, 0x26ab}, {0x26b0, 0x26b1}, {0x26bd, 0x26be}, {0x26c4, 0x26c5}, {0x26c8, 0x26c8},
  {0x26ce, 0x26cf}, {0x26d1, 0x26d1}, {0x26d3, 0x26d4}, {0x26e9, 0x26ea}, {0x26f0, 0x26f5}, {0x26f7, 0x26fa}, {0x26fd, 0x26fd}, {0x2702, 0x2702},
  {0x2705, 0x2705}, {0x2708, 0x270d}, {0x270f, 0x270f}, {0x2712, 0x2712}, {0x2714, 0x2714}, {0x2716, 0x2716}, {0x271d, 0x271d}, {0x2721, 0x2721},
  {0x2728, 0x2728}, {0x2733, 0x2734}, {0x2744, 0x2744}, {0x2747, 0x2747}, {0x274c, 0x274c}, {0x274e, 0x274e}, {0x2753, 0x2755}, {0x2757, 0x2757},
  {0x2763, 0x2764}, {0x2776, 0x2793}, {0x2795, 0x2797}, {0x27a1, 0x27a1}, {0x27b0, 0x27b0}, {0x27bf, 0x27bf}, {0x2934, 0x2935}, {0x2b05, 0x2b07},
  {0x2b1b, 0x2b1c}, {0x2b50, 0x2b50}, {0x2b55, 0x2b55}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2cfd, 0x2cfd}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27},
  {0x2d2d, 0x2d2d}, {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe},
  {0x2dc0, 0x2dc6}, {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x2e2f, 0x2e2f}, {0x3005, 0x3007}, {0x3021, 0x3035},
  {0x3038, 0x303d}, {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e},
  {0x3192, 0x3195}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3220, 0x3229}, {0x3248, 0x324f}, {0x3251, 0x325f}, {0x3280, 0x3289}, {0x3297, 0x3297},
  {0x3299, 0x3299}, {0x32b1, 0x32bf}, {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa672},
  {0xa674, 0xa67d}, {0xa67f, 0xa6f1}, {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc},
  {0xa7f2, 0xa827}, {0xa82c, 0xa82c}, {0xa830, 0xa835}, {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb},
  {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c}, {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d},
  {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e},
  {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9},
  {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28},
  {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f},
  {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb}, {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a},
  {0xff41, 0xff5a}, {0xff66, 0xffbe}, {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026},
  {0x10028, 0x1003a}, {0x1003c, 0x1003d}, {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10107, 0x10133}, {0x10140, 0x10178}, {0x1018a, 0x1018b},
  {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102fb}, {0x10300, 0x10323}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d},
  {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527},
  {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9},
  {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba},
  {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10858, 0x10876}, {0x10879, 0x1089e},
  {0x108a7, 0x108af}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x108fb, 0x1091b}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109bc, 0x109cf}, {0x109d2, 0x10a03},
  {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a48}, {0x10a60, 0x10a7e}, {0x10a80, 0x10a9f},
  {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10aeb, 0x10aef}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b58, 0x10b72}, {0x10b78, 0x10b91}, {0x10ba9, 0x10baf},
  {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10cfa, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85},
  {0x10e60, 0x10e7e}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f27}, {0x10f30, 0x10f54}, {0x10f70, 0x10f85},
  {0x10fb0, 0x10fcb}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11052, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9},
  {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da},
  {0x111dc, 0x111dc}, {0x111e1, 0x111f4}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d},
  {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328},
  {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357},
  {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0},
  {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461},
  {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644},
  {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x1173b}, {0x11740, 0x11746},
  {0x11800, 0x1183a}, {0x118a0, 0x118f2}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938},
  {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47},
  {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40},
  {0x11c50, 0x11c6c}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a},
  {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98},
  {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x11fc0, 0x11fd4},
  {0x12000, 0x12399}, {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646},
  {0x16100, 0x16139}, {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4},
  {0x16b00, 0x16b36}, {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b5b, 0x16b61}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79},
  {0x16e40, 0x16e96}, {0x16f00, 0x16f4a}, {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7},
  {0x18800, 0x18cd5}, {0x18cff, 0x18d08}, {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152},
  {0x1b155, 0x1b155}, {0x1b164, 0x1b167}, {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e},
  {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d}, {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad},
  {0x1d242, 0x1d244}, {0x1d2c0, 0x1d2d3}, {0x1d2e0, 0x1d2f3}, {0x1d360, 0x1d378}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2},
  {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514},
  {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0},
  {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8},
  {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f},
  {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a},
  {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9},
  {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8c7, 0x1e8d6},
  {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ec71, 0x1ecab}, {0x1ecad, 0x1ecaf}, {0x1ecb1, 0x1ecb4}, {0x1ed01, 0x1ed2d}, {0x1ed2f, 0x1ed3d}, {0x1ee00, 0x1ee03},
  {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22}, {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b},
  {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47}, {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57},
  {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b}, {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72},
  {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c}, {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb},
  {0x1f004, 0x1f004}, {0x1f0cf, 0x1f0cf}, {0x1f100, 0x1f10c}, {0x1f170, 0x1f171}, {0x1f17e, 0x1f17f}, {0x1f18e, 0x1f18e}, {0x1f191, 0x1f19a}, {0x1f1e6, 0x1f1ff},
  {0x1f201, 0x1f202}, {0x1f21a, 0x1f21a}, {0x1f22f, 0x1f22f}, {0x1f232, 0x1f23a}, {0x1f250, 0x1f251}, {0x1f300, 0x1f321}, {0x1f324, 0x1f393}, {0x1f396, 0x1f397},
  {0x1f399, 0x1f39b}, {0x1f39e, 0x1f3f0}, {0x1f3f3, 0x1f3f5}, {0x1f3f7, 0x1f4fd}, {0x1f4ff, 0x1f53d}, {0x1f549, 0x1f54e}, {0x1f550, 0x1f567}, {0x1f56f, 0x1f570},
  {0x1f573, 0x1f57a}, {0x1f587, 0x1f587}, {0x1f58a, 0x1f58d}, {0x1f590, 0x1f590}, {0x1f595, 0x1f596}, {0x1f5a4, 0x1f5a5}, {0x1f5a8, 0x1f5a8}, {0x1f5b1, 0x1f5b2},
  {0x1f5bc, 0x1f5bc}, {0x1f5c2, 0x1f5c4}, {0x1f5d1, 0x1f5d3}, {0x1f5dc, 0x1f5de}, {0x1f5e1, 0x1f5e1}, {0x1f5e3, 0x1f5e3}, {0x1f5e8, 0x1f5e8}, {0x1f5ef, 0x1f5ef},
  {0x1f5f3, 0x1f5f3}, {0x1f5fa, 0x1f64f}, {0x1f680, 0x1f6c5}, {0x1f6cb, 0x1f6d2}, {0x1f6d5, 0x1f6d7}, {0x1f6dc, 0x1f6e5}, {0x1f6e9, 0x1f6e9}, {0x1f6eb, 0x1f6ec},
  {0x1f6f0, 0x1f6f0}, {0x1f6f3, 0x1f6fc}, {0x1f7e0, 0x1f7eb}, {0x1f7f0, 0x1f7f0}, {0x1f90c, 0x1f93a}, {0x1f93c, 0x1f945}, {0x1f947, 0x1f9ff}, {0x1fa70, 0x1fa7c},
  {0x1fa80, 0x1fa89}, {0x1fa8f, 0x1fac6}, {0x1face, 0x1fadc}, {0x1fadf, 0x1fae9}, {0x1faf0, 0x1faf8}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df}, {0x2a700, 0x2b739},
  {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym__identifier_char_character_set_1[] = {
  {'!', '!'}, {'#', '\''}, {'*', '+'}, {'-', '.'}, {'0', ':'}, {'?', 'Z'}, {'^', '_'}, {'a', 'z'},
  {'|', '|'}, {'~', '~'}, {0xaa, 0xaa}, {0xb2, 0xb3}, {0xb5, 0xb5}, {0xb9, 0xba}, {0xbc, 0xbe}, {0xc0, 0xd6},
  {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x300, 0x374}, {0x376, 0x377},
  {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481},
  {0x483, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf}, {0x5c1, 0x5c2}, {0x5c4, 0x5c5},
  {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3}, {0x6d5, 0x6dc}, {0x6df, 0x6e8},
  {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa}, {0x7fd, 0x7fd}, {0x800, 0x82d},
  {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963}, {0x966, 0x96f}, {0x971, 0x983},
  {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bc, 0x9c4}, {0x9c7, 0x9c8},
  {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9f4, 0x9f9}, {0x9fc, 0x9fc}, {0x9fe, 0x9fe},
  {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39},
  {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa66, 0xa75},
  {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabc, 0xac5},
  {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03}, {0xb05, 0xb0c},
  {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48}, {0xb4b, 0xb4d},
  {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb77}, {0xb82, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90},
  {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbbe, 0xbc2},
  {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbf2}, {0xc00, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28},
  {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc63},
  {0xc66, 0xc6f}, {0xc78, 0xc7e}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd63}, {0xd66, 0xd78}, {0xd7a, 0xd7f},
  {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca}, {0xdcf, 0xdd4},
  {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59}, {0xe81, 0xe82},
  {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xec8, 0xece},
  {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf33}, {0xf35, 0xf35}, {0xf37, 0xf37}, {0xf39, 0xf39},
  {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049}, {0x1050, 0x109d},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x137c}, {0x1380, 0x138f}, {0x13a0, 0x13f5},
  {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715}, {0x171f, 0x1734},
  {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd}, {0x17e0, 0x17e9},
  {0x17f0, 0x17f9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59}, {0x1b6b, 0x1b73},
  {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2},
  {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2070, 0x2071}, {0x2074, 0x2079}, {0x207f, 0x2089}, {0x2090, 0x209c},
  {0x20d0, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2119, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126},
  {0x2128, 0x2128}, {0x212a, 0x212d}, {0x212f, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2150, 0x2189}, {0x2460, 0x249b},
  {0x24ea, 0x24ff}, {0x2776, 0x2793}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2cfd, 0x2cfd}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x2e2f, 0x2e2f}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e},
  {0x3192, 0x3195}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3220, 0x3229}, {0x3248, 0x324f}, {0x3251, 0x325f}, {0x3280, 0x3289}, {0x32b1, 0x32bf},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa672}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa830, 0xa835}, {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953},
  {0xa960, 0xa97c}, {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76},
  {0xaa7a, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26},
  {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c},
  {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10107, 0x10133}, {0x10140, 0x10178}, {0x1018a, 0x1018b}, {0x101fd, 0x101fd}, {0x10280, 0x1029c},
  {0x102a0, 0x102d0}, {0x102e0, 0x102fb}, {0x10300, 0x10323}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf},
  {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a},
  {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3},
  {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808},
  {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10858, 0x10876}, {0x10879, 0x1089e}, {0x108a7, 0x108af}, {0x108e0, 0x108f2},
  {0x108f4, 0x108f5}, {0x108fb, 0x1091b}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109bc, 0x109cf}, {0x109d2, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13},
  {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a48}, {0x10a60, 0x10a7e}, {0x10a80, 0x10a9f}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6},
  {0x10aeb, 0x10aef}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b58, 0x10b72}, {0x10b78, 0x10b91}, {0x10ba9, 0x10baf}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2},
  {0x10cc0, 0x10cf2}, {0x10cfa, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e60, 0x10e7e}, {0x10e80, 0x10ea9},
  {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f27}, {0x10f30, 0x10f54}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fcb}, {0x10fe0, 0x10ff6},
  {0x11000, 0x11046}, {0x11052, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f},
  {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x111e1, 0x111f4},
  {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8},
  {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333},
  {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c},
  {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5},
  {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7},
  {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8},
  {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x1173b}, {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118f2},
  {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959},
  {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d},
  {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40}, {0x11c50, 0x11c6c}, {0x11c72, 0x11c8f},
  {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47},
  {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6},
  {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x11fc0, 0x11fd4}, {0x12000, 0x12399}, {0x12400, 0x1246e},
  {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139}, {0x16800, 0x16a38},
  {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36}, {0x16b40, 0x16b43},
  {0x16b50, 0x16b59}, {0x16b5b, 0x16b61}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e96}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d2c0, 0x1d2d3},
  {0x1d2e0, 0x1d2f3}, {0x1d360, 0x1d378}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac},
  {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539},
  {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa},
  {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb},
  {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e},
  {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f},
  {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa},
  {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8c7, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959},
  {0x1ec71, 0x1ecab}, {0x1ecad, 0x1ecaf}, {0x1ecb1, 0x1ecb4}, {0x1ed01, 0x1ed2d}, {0x1ed2f, 0x1ed3d}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1f100, 0x1f10c}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym___identifier_char_no_digit_character_set_1[] = {
  {'!', '!'}, {'#', '\''}, {'*', '+'}, {'-', '.'}, {':', ':'}, {'?', 'Z'}, {'^', '_'}, {'a', 'z'},
  {'|', '|'}, {'~', '~'}, {0xaa, 0xaa}, {0xb2, 0xb3}, {0xb5, 0xb5}, {0xb9, 0xba}, {0xbc, 0xbe}, {0xc0, 0xd6},
  {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee}, {0x300, 0x374}, {0x376, 0x377},
  {0x37a, 0x37d}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5}, {0x3f7, 0x481},
  {0x483, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf}, {0x5c1, 0x5c2}, {0x5c4, 0x5c5},
  {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3}, {0x6d5, 0x6dc}, {0x6df, 0x6e8},
  {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa}, {0x7fd, 0x7fd}, {0x800, 0x82d},
  {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963}, {0x966, 0x96f}, {0x971, 0x983},
  {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9}, {0x9bc, 0x9c4}, {0x9c7, 0x9c8},
  {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9f4, 0x9f9}, {0x9fc, 0x9fc}, {0x9fe, 0x9fe},
  {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36}, {0xa38, 0xa39},
  {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e}, {0xa66, 0xa75},
  {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9}, {0xabc, 0xac5},
  {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03}, {0xb05, 0xb0c},
  {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48}, {0xb4b, 0xb4d},
  {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb77}, {0xb82, 0xb83}, {0xb85, 0xb8a}, {0xb8e, 0xb90},
  {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9}, {0xbbe, 0xbc2},
  {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbf2}, {0xc00, 0xc0c}, {0xc0e, 0xc10}, {0xc12, 0xc28},
  {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d}, {0xc60, 0xc63},
  {0xc66, 0xc6f}, {0xc78, 0xc7e}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd63}, {0xd66, 0xd78}, {0xd7a, 0xd7f},
  {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca}, {0xdcf, 0xdd4},
  {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59}, {0xe81, 0xe82},
  {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6}, {0xec8, 0xece},
  {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf33}, {0xf35, 0xf35}, {0xf37, 0xf37}, {0xf39, 0xf39},
  {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049}, {0x1050, 0x109d},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256}, {0x1258, 0x1258},
  {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0}, {0x12c2, 0x12c5},
  {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x137c}, {0x1380, 0x138f}, {0x13a0, 0x13f5},
  {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715}, {0x171f, 0x1734},
  {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd}, {0x17e0, 0x17e9},
  {0x17f0, 0x17f9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59}, {0x1b6b, 0x1b73},
  {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1cd0, 0x1cd2},
  {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b},
  {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc}, {0x1fd0, 0x1fd3},
  {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x2070, 0x2071}, {0x2074, 0x2079}, {0x207f, 0x2089}, {0x2090, 0x209c},
  {0x20d0, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2119, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126},
  {0x2128, 0x2128}, {0x212a, 0x212d}, {0x212f, 0x2139}, {0x213c, 0x213f}, {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2150, 0x2189}, {0x2460, 0x249b},
  {0x24ea, 0x24ff}, {0x2776, 0x2793}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2cfd, 0x2cfd}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x2e2f, 0x2e2f}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035},
  {0x3038, 0x303c}, {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30fa}, {0x30fc, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e},
  {0x3192, 0x3195}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff}, {0x3220, 0x3229}, {0x3248, 0x324f}, {0x3251, 0x325f}, {0x3280, 0x3289}, {0x32b1, 0x32bf},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa672}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa830, 0xa835}, {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953},
  {0xa960, 0xa97c}, {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76},
  {0xaa7a, 0xaac2}, {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26},
  {0xab28, 0xab2e}, {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6},
  {0xd7cb, 0xd7fb}, {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c},
  {0xfb3e, 0xfb3e}, {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdfb},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe70, 0xfe74}, {0xfe76, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff41, 0xff5a}, {0xff66, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10107, 0x10133}, {0x10140, 0x10178}, {0x1018a, 0x1018b}, {0x101fd, 0x101fd}, {0x10280, 0x1029c},
  {0x102a0, 0x102d0}, {0x102e0, 0x102fb}, {0x10300, 0x10323}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf},
  {0x103d1, 0x103d5}, {0x10400, 0x1049d}, {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a},
  {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3},
  {0x10600, 0x10736}, {0x10740, 0x10755}, {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808},
  {0x1080a, 0x10835}, {0x10837, 0x10838}, {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10858, 0x10876}, {0x10879, 0x1089e}, {0x108a7, 0x108af}, {0x108e0, 0x108f2},
  {0x108f4, 0x108f5}, {0x108fb, 0x1091b}, {0x10920, 0x10939}, {0x10980, 0x109b7}, {0x109bc, 0x109cf}, {0x109d2, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13},
  {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a}, {0x10a3f, 0x10a48}, {0x10a60, 0x10a7e}, {0x10a80, 0x10a9f}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6},
  {0x10aeb, 0x10aef}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b58, 0x10b72}, {0x10b78, 0x10b91}, {0x10ba9, 0x10baf}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2},
  {0x10cc0, 0x10cf2}, {0x10cfa, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d}, {0x10d6f, 0x10d85}, {0x10e60, 0x10e7e}, {0x10e80, 0x10ea9},
  {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f27}, {0x10f30, 0x10f54}, {0x10f70, 0x10f85}, {0x10fb0, 0x10fcb}, {0x10fe0, 0x10ff6},
  {0x11000, 0x11046}, {0x11052, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8}, {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f},
  {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc}, {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x111e1, 0x111f4},
  {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d}, {0x1128f, 0x1129d}, {0x1129f, 0x112a8},
  {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328}, {0x1132a, 0x11330}, {0x11332, 0x11333},
  {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357}, {0x1135d, 0x11363}, {0x11366, 0x1136c},
  {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0}, {0x113c2, 0x113c2}, {0x113c5, 0x113c5},
  {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461}, {0x11480, 0x114c5}, {0x114c7, 0x114c7},
  {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644}, {0x11650, 0x11659}, {0x11680, 0x116b8},
  {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x1173b}, {0x11740, 0x11746}, {0x11800, 0x1183a}, {0x118a0, 0x118f2},
  {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938}, {0x1193b, 0x11943}, {0x11950, 0x11959},
  {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47}, {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d},
  {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40}, {0x11c50, 0x11c6c}, {0x11c72, 0x11c8f},
  {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a}, {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47},
  {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98}, {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6},
  {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x11fc0, 0x11fd4}, {0x12000, 0x12399}, {0x12400, 0x1246e},
  {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139}, {0x16800, 0x16a38},
  {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36}, {0x16b40, 0x16b43},
  {0x16b50, 0x16b59}, {0x16b5b, 0x16b61}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e96}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d2c0, 0x1d2d3},
  {0x1d2e0, 0x1d2f3}, {0x1d360, 0x1d378}, {0x1d400, 0x1d454}, {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac},
  {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539},
  {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa},
  {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb},
  {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c}, {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e},
  {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018}, {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f},
  {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149}, {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa},
  {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee}, {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8c7, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959},
  {0x1ec71, 0x1ecab}, {0x1ecad, 0x1ecaf}, {0x1ecb1, 0x1ecb4}, {0x1ed01, 0x1ed2d}, {0x1ed2f, 0x1ed3d}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1f100, 0x1f10c}, {0x1fbf0, 0x1fbf9},
  {0x20000, 0x2a6df}, {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a},
  {0x31350, 0x323af}, {0xe0100, 0xe01ef},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(20);
      ADVANCE_MAP(
        '\n', 32,
        '\f', 32,
        '\r', 32,
        '"', 31,
        '(', 29,
        ')', 30,
        '+', 46,
        '-', 47,
        '.', 37,
        '0', 52,
        '1', 53,
        ';', 24,
        '=', 28,
        'E', 41,
        '\\', 54,
        '_', 43,
        'e', 39,
        '{', 22,
        '}', 23,
        0x85, 32,
        0x2028, 32,
        0x2029, 32,
        0xfeff, 32,
        '8', 26,
        '9', 26,
      );
      if (('2' <= lookahead && lookahead <= '7')) ADVANCE(26);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(25);
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(32);
      if (lookahead == 0xa9 ||
          lookahead == 0xae ||
          lookahead == 0x203c ||
          lookahead == 0x2049 ||
          lookahead == 0x2122 ||
          (0x2194 <= lookahead && lookahead <= 0x2199) ||
          lookahead == 0x21a9 ||
          lookahead == 0x21aa ||
          lookahead == 0x231a ||
          lookahead == 0x231b ||
          lookahead == 0x2328 ||
          lookahead == 0x23cf ||
          (0x23e9 <= lookahead && lookahead <= 0x23f3) ||
          (0x23f8 <= lookahead && lookahead <= 0x23fa) ||
          lookahead == 0x24c2 ||
          lookahead == 0x25aa ||
          lookahead == 0x25ab ||
          lookahead == 0x25b6 ||
          lookahead == 0x25c0 ||
          (0x25fb <= lookahead && lookahead <= 0x25fe) ||
          (0x2600 <= lookahead && lookahead <= 0x2604) ||
          lookahead == 0x260e ||
          lookahead == 0x2611 ||
          lookahead == 0x2614 ||
          lookahead == 0x2615 ||
          lookahead == 0x2618 ||
          lookahead == 0x261d ||
          lookahead == 0x2620 ||
          lookahead == 0x2622 ||
          lookahead == 0x2623 ||
          lookahead == 0x2626 ||
          lookahead == 0x262a ||
          lookahead == 0x262e ||
          lookahead == 0x262f ||
          (0x2638 <= lookahead && lookahead <= 0x263a) ||
          lookahead == 0x2640 ||
          lookahead == 0x2642 ||
          (0x2648 <= lookahead && lookahead <= 0x2653) ||
          lookahead == 0x265f ||
          lookahead == 0x2660 ||
          lookahead == 0x2663 ||
          lookahead == 0x2665 ||
          lookahead == 0x2666 ||
          lookahead == 0x2668 ||
          lookahead == 0x267b ||
          lookahead == 0x267e ||
          lookahead == 0x267f ||
          (0x2692 <= lookahead && lookahead <= 0x2697) ||
          lookahead == 0x2699 ||
          lookahead == 0x269b ||
          lookahead == 0x269c ||
          lookahead == 0x26a0 ||
          lookahead == 0x26a1 ||
          lookahead == 0x26a7 ||
          lookahead == 0x26aa ||
          lookahead == 0x26ab ||
          lookahead == 0x26b0 ||
          lookahead == 0x26b1 ||
          lookahead == 0x26bd ||
          lookahead == 0x26be ||
          lookahead == 0x26c4 ||
          lookahead == 0x26c5 ||
          lookahead == 0x26c8 ||
          lookahead == 0x26ce ||
          lookahead == 0x26cf ||
          lookahead == 0x26d1 ||
          lookahead == 0x26d3 ||
          lookahead == 0x26d4 ||
          lookahead == 0x26e9 ||
          lookahead == 0x26ea ||
          (0x26f0 <= lookahead && lookahead <= 0x26f5) ||
          (0x26f7 <= lookahead && lookahead <= 0x26fa) ||
          lookahead == 0x26fd ||
          lookahead == 0x2702 ||
          lookahead == 0x2705 ||
          (0x2708 <= lookahead && lookahead <= 0x270d) ||
          lookahead == 0x270f ||
          lookahead == 0x2712 ||
          lookahead == 0x2714 ||
          lookahead == 0x2716 ||
          lookahead == 0x271d ||
          lookahead == 0x2721 ||
          lookahead == 0x2728 ||
          lookahead == 0x2733 ||
          lookahead == 0x2734 ||
          lookahead == 0x2744 ||
          lookahead == 0x2747 ||
          lookahead == 0x274c ||
          lookahead == 0x274e ||
          (0x2753 <= lookahead && lookahead <= 0x2755) ||
          lookahead == 0x2757 ||
          lookahead == 0x2763 ||
          lookahead == 0x2764 ||
          (0x2795 <= lookahead && lookahead <= 0x2797) ||
          lookahead == 0x27a1 ||
          lookahead == 0x27b0 ||
          lookahead == 0x27bf ||
          lookahead == 0x2934 ||
          lookahead == 0x2935 ||
          (0x2b05 <= lookahead && lookahead <= 0x2b07) ||
          lookahead == 0x2b1b ||
          lookahead == 0x2b1c ||
          lookahead == 0x2b50 ||
          lookahead == 0x2b55 ||
          lookahead == 0x3030 ||
          lookahead == 0x303d ||
          lookahead == 0x3297 ||
          lookahead == 0x3299 ||
          lookahead == 0x1f004 ||
          lookahead == 0x1f0cf ||
          lookahead == 0x1f170 ||
          lookahead == 0x1f171 ||
          lookahead == 0x1f17e ||
          lookahead == 0x1f17f ||
          lookahead == 0x1f18e ||
          (0x1f191 <= lookahead && lookahead <= 0x1f19a) ||
          (0x1f1e6 <= lookahead && lookahead <= 0x1f1ff) ||
          lookahead == 0x1f201 ||
          lookahead == 0x1f202 ||
          lookahead == 0x1f21a ||
          lookahead == 0x1f22f ||
          (0x1f232 <= lookahead && lookahead <= 0x1f23a) ||
          lookahead == 0x1f250 ||
          lookahead == 0x1f251 ||
          (0x1f300 <= lookahead && lookahead <= 0x1f321) ||
          (0x1f324 <= lookahead && lookahead <= 0x1f393) ||
          lookahead == 0x1f396 ||
          lookahead == 0x1f397 ||
          (0x1f399 <= lookahead && lookahead <= 0x1f39b) ||
          (0x1f39e <= lookahead && lookahead <= 0x1f3f0) ||
          (0x1f3f3 <= lookahead && lookahead <= 0x1f3f5) ||
          (0x1f3f7 <= lookahead && lookahead <= 0x1f4fd) ||
          (0x1f4ff <= lookahead && lookahead <= 0x1f53d) ||
          (0x1f549 <= lookahead && lookahead <= 0x1f54e) ||
          (0x1f550 <= lookahead && lookahead <= 0x1f567) ||
          lookahead == 0x1f56f ||
          lookahead == 0x1f570 ||
          (0x1f573 <= lookahead && lookahead <= 0x1f57a) ||
          lookahead == 0x1f587 ||
          (0x1f58a <= lookahead && lookahead <= 0x1f58d) ||
          lookahead == 0x1f590 ||
          lookahead == 0x1f595 ||
          lookahead == 0x1f596 ||
          lookahead == 0x1f5a4 ||
          lookahead == 0x1f5a5 ||
          lookahead == 0x1f5a8 ||
          lookahead == 0x1f5b1 ||
          lookahead == 0x1f5b2 ||
          lookahead == 0x1f5bc ||
          (0x1f5c2 <= lookahead && lookahead <= 0x1f5c4) ||
          (0x1f5d1 <= lookahead && lookahead <= 0x1f5d3) ||
          (0x1f5dc <= lookahead && lookahead <= 0x1f5de) ||
          lookahead == 0x1f5e1 ||
          lookahead == 0x1f5e3 ||
          lookahead == 0x1f5e8 ||
          lookahead == 0x1f5ef ||
          lookahead == 0x1f5f3 ||
          (0x1f5fa <= lookahead && lookahead <= 0x1f64f) ||
          (0x1f680 <= lookahead && lookahead <= 0x1f6c5) ||
          (0x1f6cb <= lookahead && lookahead <= 0x1f6d2) ||
          (0x1f6d5 <= lookahead && lookahead <= 0x1f6d7) ||
          (0x1f6dc <= lookahead && lookahead <= 0x1f6e5) ||
          lookahead == 0x1f6e9 ||
          lookahead == 0x1f6eb ||
          lookahead == 0x1f6ec ||
          lookahead == 0x1f6f0 ||
          (0x1f6f3 <= lookahead && lookahead <= 0x1f6fc) ||
          (0x1f7e0 <= lookahead && lookahead <= 0x1f7eb) ||
          lookahead == 0x1f7f0 ||
          (0x1f90c <= lookahead && lookahead <= 0x1f93a) ||
          (0x1f93c <= lookahead && lookahead <= 0x1f945) ||
          (0x1f947 <= lookahead && lookahead <= 0x1f9ff) ||
          (0x1fa70 <= lookahead && lookahead <= 0x1fa7c) ||
          (0x1fa80 <= lookahead && lookahead <= 0x1fa89) ||
          (0x1fa8f <= lookahead && lookahead <= 0x1fac6) ||
          (0x1face <= lookahead && lookahead <= 0x1fadc) ||
          (0x1fadf <= lookahead && lookahead <= 0x1fae9) ||
          (0x1faf0 <= lookahead && lookahead <= 0x1faf8)) ADVANCE(25);
      if (set_contains(sym___identifier_char_no_digit_character_set_1, 826, lookahead)) ADVANCE(25);
      if (lookahead != 0) ADVANCE(32);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        ')', 30,
        '/', 10,
        '0', 45,
        ';', 24,
        '=', 28,
        '\\', 54,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(44);
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      if (set_contains(sym___identifier_char_no_digit_character_set_1, 826, lookahead)) ADVANCE(27);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        ')', 30,
        '/', 10,
        ';', 24,
        '=', 28,
        '\\', 54,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      if (set_contains(sym__identifier_char_character_set_1, 826, lookahead)) ADVANCE(26);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        '.', 36,
        '/', 10,
        ';', 24,
        'E', 40,
        '\\', 54,
        '_', 42,
        'e', 38,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(44);
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        '/', 10,
        '0', 52,
        '1', 53,
        ';', 24,
        '\\', 54,
        '_', 42,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        '/', 10,
        ';', 24,
        '\\', 54,
        '_', 42,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(50);
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        '/', 10,
        ';', 24,
        '\\', 54,
        '_', 42,
        '{', 22,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(35);
      END_STATE();
    case 7:
      if (lookahead == '\n') ADVANCE(56);
      if (lookahead == '\f') ADVANCE(59);
      if (lookahead == '\r') ADVANCE(8);
      if (lookahead == 0x85) ADVANCE(58);
      if (lookahead == 0x2028) ADVANCE(60);
      if (lookahead == 0x2029) ADVANCE(61);
      if (lookahead != 0) ADVANCE(65);
      END_STATE();
    case 8:
      if (lookahead == '\n') ADVANCE(57);
      if (lookahead == '\'') ADVANCE(55);
      END_STATE();
    case 9:
      if (lookahead == '"') ADVANCE(31);
      if (lookahead == '\\') ADVANCE(33);
      if (lookahead != 0) ADVANCE(32);
      END_STATE();
    case 10:
      if (lookahead == '-') ADVANCE(21);
      if (lookahead == '/') ADVANCE(64);
      END_STATE();
    case 11:
      if (lookahead == '{') ADVANCE(18);
      END_STATE();
    case 12:
      if (lookahead == '}') ADVANCE(34);
      END_STATE();
    case 13:
      if (lookahead == '}') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(12);
      END_STATE();
    case 14:
      if (lookahead == '}') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(13);
      END_STATE();
    case 15:
      if (lookahead == '}') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(14);
      END_STATE();
    case 16:
      if (lookahead == '}') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(15);
      END_STATE();
    case 17:
      if (lookahead == '}') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(16);
      END_STATE();
    case 18:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(17);
      END_STATE();
    case 19:
      if (eof) ADVANCE(20);
      ADVANCE_MAP(
        '\n', 56,
        '\f', 59,
        '\r', 8,
        '"', 31,
        '(', 29,
        ')', 30,
        '+', 46,
        '-', 47,
        '/', 10,
        '0', 45,
        ';', 24,
        '=', 28,
        '\\', 54,
        '{', 22,
        '}', 23,
        0x85, 58,
        0x2028, 60,
        0x2029, 61,
        0xfeff, 62,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(44);
      if (lookahead == '\t' ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x1680 ||
          (0x2000 <= lookahead && lookahead <= 0x200a) ||
          lookahead == 0x202f ||
          lookahead == 0x205f ||
          lookahead == 0x3000) ADVANCE(63);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 20:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 21:
      ACCEPT_TOKEN(anon_sym_SLASH_DASH);
      END_STATE();
    case 22:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 23:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 24:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 25:
      ACCEPT_TOKEN(sym__normal_bare_identifier);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 26:
      ACCEPT_TOKEN(sym__identifier_char);
      END_STATE();
    case 27:
      ACCEPT_TOKEN(sym___identifier_char_no_digit);
      END_STATE();
    case 28:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 29:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 30:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(aux_sym__escaped_string_token1);
      END_STATE();
    case 33:
      ACCEPT_TOKEN(aux_sym__escaped_string_token1);
      ADVANCE_MAP(
        'u', 11,
        '"', 34,
        '/', 34,
        '\\', 34,
        'b', 34,
        'f', 34,
        'n', 34,
        'r', 34,
        't', 34,
      );
      END_STATE();
    case 34:
      ACCEPT_TOKEN(sym_escape);
      END_STATE();
    case 35:
      ACCEPT_TOKEN(sym__hex_digit);
      END_STATE();
    case 36:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 37:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 38:
      ACCEPT_TOKEN(anon_sym_e);
      END_STATE();
    case 39:
      ACCEPT_TOKEN(anon_sym_e);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 40:
      ACCEPT_TOKEN(anon_sym_E);
      END_STATE();
    case 41:
      ACCEPT_TOKEN(anon_sym_E);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 42:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 43:
      ACCEPT_TOKEN(anon_sym__);
      if (set_contains(sym__normal_bare_identifier_character_set_2, 968, lookahead)) ADVANCE(25);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(sym__digit);
      END_STATE();
    case 45:
      ACCEPT_TOKEN(sym__digit);
      if (lookahead == 'b') ADVANCE(51);
      if (lookahead == 'o') ADVANCE(49);
      if (lookahead == 'x') ADVANCE(48);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 48:
      ACCEPT_TOKEN(anon_sym_0x);
      END_STATE();
    case 49:
      ACCEPT_TOKEN(anon_sym_0o);
      END_STATE();
    case 50:
      ACCEPT_TOKEN(aux_sym__octal_token1);
      END_STATE();
    case 51:
      ACCEPT_TOKEN(anon_sym_0b);
      END_STATE();
    case 52:
      ACCEPT_TOKEN(anon_sym_0);
      END_STATE();
    case 53:
      ACCEPT_TOKEN(anon_sym_1);
      END_STATE();
    case 54:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      END_STATE();
    case 55:
      ACCEPT_TOKEN(aux_sym__newline_token1);
      END_STATE();
    case 56:
      ACCEPT_TOKEN(aux_sym__newline_token2);
      END_STATE();
    case 57:
      ACCEPT_TOKEN(aux_sym__newline_token3);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(aux_sym__newline_token4);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(aux_sym__newline_token5);
      END_STATE();
    case 60:
      ACCEPT_TOKEN(aux_sym__newline_token6);
      END_STATE();
    case 61:
      ACCEPT_TOKEN(aux_sym__newline_token7);
      END_STATE();
    case 62:
      ACCEPT_TOKEN(sym__bom);
      END_STATE();
    case 63:
      ACCEPT_TOKEN(sym__unicode_space);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      END_STATE();
    case 65:
      ACCEPT_TOKEN(aux_sym_single_line_comment_token1);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'b', 1,
        'c', 2,
        'd', 3,
        'e', 4,
        'f', 5,
        'h', 6,
        'i', 7,
        'n', 8,
        'r', 9,
        't', 10,
        'u', 11,
      );
      END_STATE();
    case 1:
      if (lookahead == 'a') ADVANCE(12);
      END_STATE();
    case 2:
      if (lookahead == 'o') ADVANCE(13);
      if (lookahead == 'u') ADVANCE(14);
      END_STATE();
    case 3:
      if (lookahead == 'a') ADVANCE(15);
      if (lookahead == 'e') ADVANCE(16);
      if (lookahead == 'u') ADVANCE(17);
      END_STATE();
    case 4:
      if (lookahead == 'm') ADVANCE(18);
      END_STATE();
    case 5:
      if (lookahead == '3') ADVANCE(19);
      if (lookahead == '6') ADVANCE(20);
      if (lookahead == 'a') ADVANCE(21);
      END_STATE();
    case 6:
      if (lookahead == 'o') ADVANCE(22);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '1', 23,
        '3', 24,
        '6', 25,
        '8', 26,
        'd', 27,
        'p', 28,
        'r', 29,
        's', 30,
      );
      END_STATE();
    case 8:
      if (lookahead == 'u') ADVANCE(31);
      END_STATE();
    case 9:
      if (lookahead == 'e') ADVANCE(32);
      END_STATE();
    case 10:
      if (lookahead == 'i') ADVANCE(33);
      if (lookahead == 'r') ADVANCE(34);
      END_STATE();
    case 11:
      if (lookahead == '1') ADVANCE(35);
      if (lookahead == '3') ADVANCE(36);
      if (lookahead == '6') ADVANCE(37);
      if (lookahead == '8') ADVANCE(38);
      if (lookahead == 'r') ADVANCE(39);
      if (lookahead == 's') ADVANCE(40);
      if (lookahead == 'u') ADVANCE(41);
      END_STATE();
    case 12:
      if (lookahead == 's') ADVANCE(42);
      END_STATE();
    case 13:
      if (lookahead == 'u') ADVANCE(43);
      END_STATE();
    case 14:
      if (lookahead == 'r') ADVANCE(44);
      END_STATE();
    case 15:
      if (lookahead == 't') ADVANCE(45);
      END_STATE();
    case 16:
      if (lookahead == 'c') ADVANCE(46);
      END_STATE();
    case 17:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 18:
      if (lookahead == 'a') ADVANCE(48);
      END_STATE();
    case 19:
      if (lookahead == '2') ADVANCE(49);
      END_STATE();
    case 20:
      if (lookahead == '4') ADVANCE(50);
      END_STATE();
    case 21:
      if (lookahead == 'l') ADVANCE(51);
      END_STATE();
    case 22:
      if (lookahead == 's') ADVANCE(52);
      END_STATE();
    case 23:
      if (lookahead == '6') ADVANCE(53);
      END_STATE();
    case 24:
      if (lookahead == '2') ADVANCE(54);
      END_STATE();
    case 25:
      if (lookahead == '4') ADVANCE(55);
      END_STATE();
    case 26:
      ACCEPT_TOKEN(anon_sym_i8);
      END_STATE();
    case 27:
      if (lookahead == 'n') ADVANCE(56);
      END_STATE();
    case 28:
      if (lookahead == 'v') ADVANCE(57);
      END_STATE();
    case 29:
      if (lookahead == 'i') ADVANCE(58);
      if (lookahead == 'l') ADVANCE(59);
      END_STATE();
    case 30:
      if (lookahead == 'i') ADVANCE(60);
      END_STATE();
    case 31:
      if (lookahead == 'l') ADVANCE(61);
      END_STATE();
    case 32:
      if (lookahead == 'g') ADVANCE(62);
      END_STATE();
    case 33:
      if (lookahead == 'm') ADVANCE(63);
      END_STATE();
    case 34:
      if (lookahead == 'u') ADVANCE(64);
      END_STATE();
    case 35:
      if (lookahead == '6') ADVANCE(65);
      END_STATE();
    case 36:
      if (lookahead == '2') ADVANCE(66);
      END_STATE();
    case 37:
      if (lookahead == '4') ADVANCE(67);
      END_STATE();
    case 38:
      ACCEPT_TOKEN(anon_sym_u8);
      END_STATE();
    case 39:
      if (lookahead == 'l') ADVANCE(68);
      END_STATE();
    case 40:
      if (lookahead == 'i') ADVANCE(69);
      END_STATE();
    case 41:
      if (lookahead == 'i') ADVANCE(70);
      END_STATE();
    case 42:
      if (lookahead == 'e') ADVANCE(71);
      END_STATE();
    case 43:
      if (lookahead == 'n') ADVANCE(72);
      END_STATE();
    case 44:
      if (lookahead == 'r') ADVANCE(73);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(74);
      END_STATE();
    case 46:
      if (lookahead == 'i') ADVANCE(75);
      END_STATE();
    case 47:
      if (lookahead == 'a') ADVANCE(76);
      END_STATE();
    case 48:
      if (lookahead == 'i') ADVANCE(77);
      END_STATE();
    case 49:
      ACCEPT_TOKEN(anon_sym_f32);
      END_STATE();
    case 50:
      ACCEPT_TOKEN(anon_sym_f64);
      END_STATE();
    case 51:
      if (lookahead == 's') ADVANCE(78);
      END_STATE();
    case 52:
      if (lookahead == 't') ADVANCE(79);
      END_STATE();
    case 53:
      ACCEPT_TOKEN(anon_sym_i16);
      END_STATE();
    case 54:
      ACCEPT_TOKEN(anon_sym_i32);
      END_STATE();
    case 55:
      ACCEPT_TOKEN(anon_sym_i64);
      END_STATE();
    case 56:
      if (lookahead == '-') ADVANCE(80);
      END_STATE();
    case 57:
      if (lookahead == '4') ADVANCE(81);
      if (lookahead == '6') ADVANCE(82);
      END_STATE();
    case 58:
      if (lookahead == '-') ADVANCE(83);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(anon_sym_irl);
      END_STATE();
    case 60:
      if (lookahead == 'z') ADVANCE(84);
      END_STATE();
    case 61:
      if (lookahead == 'l') ADVANCE(85);
      END_STATE();
    case 62:
      if (lookahead == 'e') ADVANCE(86);
      END_STATE();
    case 63:
      if (lookahead == 'e') ADVANCE(87);
      END_STATE();
    case 64:
      if (lookahead == 'e') ADVANCE(88);
      END_STATE();
    case 65:
      ACCEPT_TOKEN(anon_sym_u16);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_u32);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_u64);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_url);
      if (lookahead == '-') ADVANCE(89);
      END_STATE();
    case 69:
      if (lookahead == 'z') ADVANCE(90);
      END_STATE();
    case 70:
      if (lookahead == 'd') ADVANCE(91);
      END_STATE();
    case 71:
      if (lookahead == '6') ADVANCE(92);
      END_STATE();
    case 72:
      if (lookahead == 't') ADVANCE(93);
      END_STATE();
    case 73:
      if (lookahead == 'e') ADVANCE(94);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(anon_sym_date);
      if (lookahead == '-') ADVANCE(95);
      END_STATE();
    case 75:
      if (lookahead == 'm') ADVANCE(96);
      END_STATE();
    case 76:
      if (lookahead == 't') ADVANCE(97);
      END_STATE();
    case 77:
      if (lookahead == 'l') ADVANCE(98);
      END_STATE();
    case 78:
      if (lookahead == 'e') ADVANCE(99);
      END_STATE();
    case 79:
      if (lookahead == 'n') ADVANCE(100);
      END_STATE();
    case 80:
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'h') ADVANCE(102);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_ipv4);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_ipv6);
      END_STATE();
    case 83:
      if (lookahead == 'r') ADVANCE(103);
      END_STATE();
    case 84:
      if (lookahead == 'e') ADVANCE(104);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_null);
      END_STATE();
    case 86:
      if (lookahead == 'x') ADVANCE(105);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_time);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 89:
      if (lookahead == 'r') ADVANCE(106);
      if (lookahead == 't') ADVANCE(107);
      END_STATE();
    case 90:
      if (lookahead == 'e') ADVANCE(108);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_uuid);
      END_STATE();
    case 92:
      if (lookahead == '4') ADVANCE(109);
      END_STATE();
    case 93:
      if (lookahead == 'r') ADVANCE(110);
      END_STATE();
    case 94:
      if (lookahead == 'n') ADVANCE(111);
      END_STATE();
    case 95:
      if (lookahead == 't') ADVANCE(112);
      END_STATE();
    case 96:
      if (lookahead == 'a') ADVANCE(113);
      END_STATE();
    case 97:
      if (lookahead == 'i') ADVANCE(114);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_email);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 100:
      if (lookahead == 'a') ADVANCE(115);
      END_STATE();
    case 101:
      if (lookahead == 'm') ADVANCE(116);
      END_STATE();
    case 102:
      if (lookahead == 'o') ADVANCE(117);
      END_STATE();
    case 103:
      if (lookahead == 'e') ADVANCE(118);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_isize);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_regex);
      END_STATE();
    case 106:
      if (lookahead == 'e') ADVANCE(119);
      END_STATE();
    case 107:
      if (lookahead == 'e') ADVANCE(120);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_usize);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_base64);
      END_STATE();
    case 110:
      if (lookahead == 'y') ADVANCE(121);
      END_STATE();
    case 111:
      if (lookahead == 'c') ADVANCE(122);
      END_STATE();
    case 112:
      if (lookahead == 'i') ADVANCE(123);
      END_STATE();
    case 113:
      if (lookahead == 'l') ADVANCE(124);
      END_STATE();
    case 114:
      if (lookahead == 'o') ADVANCE(125);
      END_STATE();
    case 115:
      if (lookahead == 'm') ADVANCE(126);
      END_STATE();
    case 116:
      if (lookahead == 'a') ADVANCE(127);
      END_STATE();
    case 117:
      if (lookahead == 's') ADVANCE(128);
      END_STATE();
    case 118:
      if (lookahead == 'f') ADVANCE(129);
      END_STATE();
    case 119:
      if (lookahead == 'f') ADVANCE(130);
      END_STATE();
    case 120:
      if (lookahead == 'm') ADVANCE(131);
      END_STATE();
    case 121:
      if (lookahead == '-') ADVANCE(132);
      END_STATE();
    case 122:
      if (lookahead == 'y') ADVANCE(133);
      END_STATE();
    case 123:
      if (lookahead == 'm') ADVANCE(134);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_decimal);
      if (lookahead == '1') ADVANCE(135);
      if (lookahead == '6') ADVANCE(136);
      END_STATE();
    case 125:
      if (lookahead == 'n') ADVANCE(137);
      END_STATE();
    case 126:
      if (lookahead == 'e') ADVANCE(138);
      END_STATE();
    case 127:
      if (lookahead == 'i') ADVANCE(139);
      END_STATE();
    case 128:
      if (lookahead == 't') ADVANCE(140);
      END_STATE();
    case 129:
      if (lookahead == 'e') ADVANCE(141);
      END_STATE();
    case 130:
      if (lookahead == 'e') ADVANCE(142);
      END_STATE();
    case 131:
      if (lookahead == 'p') ADVANCE(143);
      END_STATE();
    case 132:
      if (lookahead == '2') ADVANCE(144);
      if (lookahead == '3') ADVANCE(145);
      if (lookahead == 's') ADVANCE(146);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_currency);
      END_STATE();
    case 134:
      if (lookahead == 'e') ADVANCE(147);
      END_STATE();
    case 135:
      if (lookahead == '2') ADVANCE(148);
      END_STATE();
    case 136:
      if (lookahead == '4') ADVANCE(149);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_duration);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_hostname);
      END_STATE();
    case 139:
      if (lookahead == 'l') ADVANCE(150);
      END_STATE();
    case 140:
      if (lookahead == 'n') ADVANCE(151);
      END_STATE();
    case 141:
      if (lookahead == 'r') ADVANCE(152);
      END_STATE();
    case 142:
      if (lookahead == 'r') ADVANCE(153);
      END_STATE();
    case 143:
      if (lookahead == 'l') ADVANCE(154);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_country_DASH2);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_country_DASH3);
      END_STATE();
    case 146:
      if (lookahead == 'u') ADVANCE(155);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_date_DASHtime);
      END_STATE();
    case 148:
      if (lookahead == '8') ADVANCE(156);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_decimal64);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_idn_DASHemail);
      END_STATE();
    case 151:
      if (lookahead == 'a') ADVANCE(157);
      END_STATE();
    case 152:
      if (lookahead == 'e') ADVANCE(158);
      END_STATE();
    case 153:
      if (lookahead == 'e') ADVANCE(159);
      END_STATE();
    case 154:
      if (lookahead == 'a') ADVANCE(160);
      END_STATE();
    case 155:
      if (lookahead == 'b') ADVANCE(161);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_decimal128);
      END_STATE();
    case 157:
      if (lookahead == 'm') ADVANCE(162);
      END_STATE();
    case 158:
      if (lookahead == 'n') ADVANCE(163);
      END_STATE();
    case 159:
      if (lookahead == 'n') ADVANCE(164);
      END_STATE();
    case 160:
      if (lookahead == 't') ADVANCE(165);
      END_STATE();
    case 161:
      if (lookahead == 'd') ADVANCE(166);
      END_STATE();
    case 162:
      if (lookahead == 'e') ADVANCE(167);
      END_STATE();
    case 163:
      if (lookahead == 'c') ADVANCE(168);
      END_STATE();
    case 164:
      if (lookahead == 'c') ADVANCE(169);
      END_STATE();
    case 165:
      if (lookahead == 'e') ADVANCE(170);
      END_STATE();
    case 166:
      if (lookahead == 'i') ADVANCE(171);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_idn_DASHhostname);
      END_STATE();
    case 168:
      if (lookahead == 'e') ADVANCE(172);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_url_DASHtemplate);
      END_STATE();
    case 171:
      if (lookahead == 'v') ADVANCE(174);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_iri_DASHreference);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_url_DASHreference);
      END_STATE();
    case 174:
      if (lookahead == 'i') ADVANCE(175);
      END_STATE();
    case 175:
      if (lookahead == 's') ADVANCE(176);
      END_STATE();
    case 176:
      if (lookahead == 'i') ADVANCE(177);
      END_STATE();
    case 177:
      if (lookahead == 'o') ADVANCE(178);
      END_STATE();
    case 178:
      if (lookahead == 'n') ADVANCE(179);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_country_DASHsubdivision);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 19, .external_lex_state = 2},
  [2] = {.lex_state = 19, .external_lex_state = 1},
  [3] = {.lex_state = 19, .external_lex_state = 1},
  [4] = {.lex_state = 19, .external_lex_state = 1},
  [5] = {.lex_state = 19, .external_lex_state = 1},
  [6] = {.lex_state = 19, .external_lex_state = 1},
  [7] = {.lex_state = 19, .external_lex_state = 1},
  [8] = {.lex_state = 19, .external_lex_state = 1},
  [9] = {.lex_state = 19, .external_lex_state = 2},
  [10] = {.lex_state = 19, .external_lex_state = 2},
  [11] = {.lex_state = 19, .external_lex_state = 2},
  [12] = {.lex_state = 19, .external_lex_state = 2},
  [13] = {.lex_state = 19, .external_lex_state = 2},
  [14] = {.lex_state = 19, .external_lex_state = 2},
  [15] = {.lex_state = 19, .external_lex_state = 1},
  [16] = {.lex_state = 19, .external_lex_state = 2},
  [17] = {.lex_state = 19, .external_lex_state = 2},
  [18] = {.lex_state = 19, .external_lex_state = 2},
  [19] = {.lex_state = 19, .external_lex_state = 2},
  [20] = {.lex_state = 19, .external_lex_state = 2},
  [21] = {.lex_state = 19, .external_lex_state = 2},
  [22] = {.lex_state = 19, .external_lex_state = 2},
  [23] = {.lex_state = 19, .external_lex_state = 2},
  [24] = {.lex_state = 19, .external_lex_state = 2},
  [25] = {.lex_state = 19, .external_lex_state = 2},
  [26] = {.lex_state = 19, .external_lex_state = 2},
  [27] = {.lex_state = 19, .external_lex_state = 2},
  [28] = {.lex_state = 19, .external_lex_state = 2},
  [29] = {.lex_state = 19, .external_lex_state = 2},
  [30] = {.lex_state = 19, .external_lex_state = 1},
  [31] = {.lex_state = 19, .external_lex_state = 2},
  [32] = {.lex_state = 19, .external_lex_state = 2},
  [33] = {.lex_state = 19, .external_lex_state = 1},
  [34] = {.lex_state = 19, .external_lex_state = 2},
  [35] = {.lex_state = 19, .external_lex_state = 2},
  [36] = {.lex_state = 19, .external_lex_state = 2},
  [37] = {.lex_state = 19, .external_lex_state = 2},
  [38] = {.lex_state = 19, .external_lex_state = 2},
  [39] = {.lex_state = 19, .external_lex_state = 2},
  [40] = {.lex_state = 19, .external_lex_state = 2},
  [41] = {.lex_state = 19, .external_lex_state = 2},
  [42] = {.lex_state = 19, .external_lex_state = 2},
  [43] = {.lex_state = 19, .external_lex_state = 2},
  [44] = {.lex_state = 19, .external_lex_state = 2},
  [45] = {.lex_state = 19, .external_lex_state = 2},
  [46] = {.lex_state = 19, .external_lex_state = 2},
  [47] = {.lex_state = 19, .external_lex_state = 1},
  [48] = {.lex_state = 19, .external_lex_state = 2},
  [49] = {.lex_state = 19, .external_lex_state = 2},
  [50] = {.lex_state = 19, .external_lex_state = 1},
  [51] = {.lex_state = 19, .external_lex_state = 2},
  [52] = {.lex_state = 19, .external_lex_state = 1},
  [53] = {.lex_state = 19, .external_lex_state = 1},
  [54] = {.lex_state = 19, .external_lex_state = 2},
  [55] = {.lex_state = 19, .external_lex_state = 1},
  [56] = {.lex_state = 19, .external_lex_state = 2},
  [57] = {.lex_state = 19, .external_lex_state = 1},
  [58] = {.lex_state = 19, .external_lex_state = 1},
  [59] = {.lex_state = 19, .external_lex_state = 2},
  [60] = {.lex_state = 19, .external_lex_state = 1},
  [61] = {.lex_state = 19, .external_lex_state = 1},
  [62] = {.lex_state = 19, .external_lex_state = 3},
  [63] = {.lex_state = 19, .external_lex_state = 2},
  [64] = {.lex_state = 19, .external_lex_state = 3},
  [65] = {.lex_state = 19, .external_lex_state = 3},
  [66] = {.lex_state = 19, .external_lex_state = 3},
  [67] = {.lex_state = 19, .external_lex_state = 3},
  [68] = {.lex_state = 19, .external_lex_state = 3},
  [69] = {.lex_state = 19, .external_lex_state = 3},
  [70] = {.lex_state = 19, .external_lex_state = 3},
  [71] = {.lex_state = 19, .external_lex_state = 3},
  [72] = {.lex_state = 19, .external_lex_state = 3},
  [73] = {.lex_state = 19, .external_lex_state = 2},
  [74] = {.lex_state = 19, .external_lex_state = 2},
  [75] = {.lex_state = 19, .external_lex_state = 2},
  [76] = {.lex_state = 19, .external_lex_state = 2},
  [77] = {.lex_state = 19, .external_lex_state = 2},
  [78] = {.lex_state = 19, .external_lex_state = 2},
  [79] = {.lex_state = 19, .external_lex_state = 2},
  [80] = {.lex_state = 19, .external_lex_state = 3},
  [81] = {.lex_state = 19, .external_lex_state = 3},
  [82] = {.lex_state = 3, .external_lex_state = 3},
  [83] = {.lex_state = 3, .external_lex_state = 3},
  [84] = {.lex_state = 19, .external_lex_state = 3},
  [85] = {.lex_state = 19, .external_lex_state = 3},
  [86] = {.lex_state = 19, .external_lex_state = 3},
  [87] = {.lex_state = 19, .external_lex_state = 3},
  [88] = {.lex_state = 19, .external_lex_state = 3},
  [89] = {.lex_state = 19, .external_lex_state = 3},
  [90] = {.lex_state = 19, .external_lex_state = 3},
  [91] = {.lex_state = 19, .external_lex_state = 3},
  [92] = {.lex_state = 19, .external_lex_state = 3},
  [93] = {.lex_state = 19, .external_lex_state = 3},
  [94] = {.lex_state = 19, .external_lex_state = 2},
  [95] = {.lex_state = 19, .external_lex_state = 3},
  [96] = {.lex_state = 19, .external_lex_state = 3},
  [97] = {.lex_state = 19, .external_lex_state = 3},
  [98] = {.lex_state = 19, .external_lex_state = 3},
  [99] = {.lex_state = 19, .external_lex_state = 3},
  [100] = {.lex_state = 19, .external_lex_state = 3},
  [101] = {.lex_state = 19, .external_lex_state = 3},
  [102] = {.lex_state = 19, .external_lex_state = 3},
  [103] = {.lex_state = 19, .external_lex_state = 3},
  [104] = {.lex_state = 19, .external_lex_state = 3},
  [105] = {.lex_state = 19, .external_lex_state = 3},
  [106] = {.lex_state = 3, .external_lex_state = 3},
  [107] = {.lex_state = 19, .external_lex_state = 3},
  [108] = {.lex_state = 19, .external_lex_state = 3},
  [109] = {.lex_state = 19, .external_lex_state = 3},
  [110] = {.lex_state = 19, .external_lex_state = 3},
  [111] = {.lex_state = 19, .external_lex_state = 3},
  [112] = {.lex_state = 19, .external_lex_state = 3},
  [113] = {.lex_state = 19, .external_lex_state = 3},
  [114] = {.lex_state = 19, .external_lex_state = 3},
  [115] = {.lex_state = 19, .external_lex_state = 3},
  [116] = {.lex_state = 19, .external_lex_state = 3},
  [117] = {.lex_state = 19, .external_lex_state = 3},
  [118] = {.lex_state = 19, .external_lex_state = 3},
  [119] = {.lex_state = 19, .external_lex_state = 3},
  [120] = {.lex_state = 19, .external_lex_state = 3},
  [121] = {.lex_state = 19, .external_lex_state = 2},
  [122] = {.lex_state = 19, .external_lex_state = 2},
  [123] = {.lex_state = 19, .external_lex_state = 2},
  [124] = {.lex_state = 19, .external_lex_state = 2},
  [125] = {.lex_state = 19, .external_lex_state = 2},
  [126] = {.lex_state = 19, .external_lex_state = 2},
  [127] = {.lex_state = 19, .external_lex_state = 2},
  [128] = {.lex_state = 19, .external_lex_state = 2},
  [129] = {.lex_state = 4, .external_lex_state = 3},
  [130] = {.lex_state = 19, .external_lex_state = 2},
  [131] = {.lex_state = 19, .external_lex_state = 2},
  [132] = {.lex_state = 19, .external_lex_state = 2},
  [133] = {.lex_state = 19, .external_lex_state = 2},
  [134] = {.lex_state = 19, .external_lex_state = 2},
  [135] = {.lex_state = 19, .external_lex_state = 2},
  [136] = {.lex_state = 19, .external_lex_state = 2},
  [137] = {.lex_state = 3, .external_lex_state = 3},
  [138] = {.lex_state = 4, .external_lex_state = 3},
  [139] = {.lex_state = 4, .external_lex_state = 3},
  [140] = {.lex_state = 19, .external_lex_state = 2},
  [141] = {.lex_state = 19, .external_lex_state = 2},
  [142] = {.lex_state = 19, .external_lex_state = 2},
  [143] = {.lex_state = 19, .external_lex_state = 2},
  [144] = {.lex_state = 19, .external_lex_state = 2},
  [145] = {.lex_state = 19, .external_lex_state = 2},
  [146] = {.lex_state = 19, .external_lex_state = 2},
  [147] = {.lex_state = 3, .external_lex_state = 3},
  [148] = {.lex_state = 19, .external_lex_state = 2},
  [149] = {.lex_state = 19, .external_lex_state = 2},
  [150] = {.lex_state = 4, .external_lex_state = 3},
  [151] = {.lex_state = 4, .external_lex_state = 3},
  [152] = {.lex_state = 19, .external_lex_state = 2},
  [153] = {.lex_state = 19, .external_lex_state = 2},
  [154] = {.lex_state = 19, .external_lex_state = 2},
  [155] = {.lex_state = 19, .external_lex_state = 2},
  [156] = {.lex_state = 19, .external_lex_state = 2},
  [157] = {.lex_state = 19, .external_lex_state = 2},
  [158] = {.lex_state = 19, .external_lex_state = 2},
  [159] = {.lex_state = 19, .external_lex_state = 2},
  [160] = {.lex_state = 19, .external_lex_state = 2},
  [161] = {.lex_state = 19, .external_lex_state = 2},
  [162] = {.lex_state = 19, .external_lex_state = 2},
  [163] = {.lex_state = 19, .external_lex_state = 2},
  [164] = {.lex_state = 19, .external_lex_state = 2},
  [165] = {.lex_state = 19, .external_lex_state = 2},
  [166] = {.lex_state = 19, .external_lex_state = 2},
  [167] = {.lex_state = 19, .external_lex_state = 2},
  [168] = {.lex_state = 19, .external_lex_state = 2},
  [169] = {.lex_state = 19, .external_lex_state = 2},
  [170] = {.lex_state = 19, .external_lex_state = 2},
  [171] = {.lex_state = 19, .external_lex_state = 2},
  [172] = {.lex_state = 5, .external_lex_state = 3},
  [173] = {.lex_state = 19, .external_lex_state = 2},
  [174] = {.lex_state = 6, .external_lex_state = 3},
  [175] = {.lex_state = 5, .external_lex_state = 3},
  [176] = {.lex_state = 19, .external_lex_state = 2},
  [177] = {.lex_state = 6, .external_lex_state = 3},
  [178] = {.lex_state = 3, .external_lex_state = 3},
  [179] = {.lex_state = 3, .external_lex_state = 3},
  [180] = {.lex_state = 6, .external_lex_state = 3},
  [181] = {.lex_state = 5, .external_lex_state = 3},
  [182] = {.lex_state = 19, .external_lex_state = 2},
  [183] = {.lex_state = 6, .external_lex_state = 3},
  [184] = {.lex_state = 6, .external_lex_state = 3},
  [185] = {.lex_state = 5, .external_lex_state = 3},
  [186] = {.lex_state = 5, .external_lex_state = 3},
  [187] = {.lex_state = 19, .external_lex_state = 3},
  [188] = {.lex_state = 2, .external_lex_state = 3},
  [189] = {.lex_state = 2, .external_lex_state = 3},
  [190] = {.lex_state = 2, .external_lex_state = 3},
  [191] = {.lex_state = 19, .external_lex_state = 3},
  [192] = {.lex_state = 19, .external_lex_state = 3},
  [193] = {.lex_state = 19, .external_lex_state = 3},
  [194] = {.lex_state = 19, .external_lex_state = 3},
  [195] = {.lex_state = 19, .external_lex_state = 3},
  [196] = {.lex_state = 1, .external_lex_state = 3},
  [197] = {.lex_state = 19, .external_lex_state = 3},
  [198] = {.lex_state = 19, .external_lex_state = 3},
  [199] = {.lex_state = 19, .external_lex_state = 3},
  [200] = {.lex_state = 19, .external_lex_state = 3},
  [201] = {.lex_state = 19, .external_lex_state = 3},
  [202] = {.lex_state = 19, .external_lex_state = 3},
  [203] = {.lex_state = 19, .external_lex_state = 3},
  [204] = {.lex_state = 19, .external_lex_state = 3},
  [205] = {.lex_state = 19, .external_lex_state = 3},
  [206] = {.lex_state = 19, .external_lex_state = 3},
  [207] = {.lex_state = 19, .external_lex_state = 3},
  [208] = {.lex_state = 19, .external_lex_state = 3},
  [209] = {.lex_state = 19, .external_lex_state = 3},
  [210] = {.lex_state = 19, .external_lex_state = 3},
  [211] = {.lex_state = 19, .external_lex_state = 3},
  [212] = {.lex_state = 19, .external_lex_state = 3},
  [213] = {.lex_state = 19, .external_lex_state = 3},
  [214] = {.lex_state = 19, .external_lex_state = 3},
  [215] = {.lex_state = 19, .external_lex_state = 3},
  [216] = {.lex_state = 19, .external_lex_state = 3},
  [217] = {.lex_state = 19, .external_lex_state = 3},
  [218] = {.lex_state = 19, .external_lex_state = 3},
  [219] = {.lex_state = 19, .external_lex_state = 3},
  [220] = {.lex_state = 19, .external_lex_state = 3},
  [221] = {.lex_state = 19, .external_lex_state = 3},
  [222] = {.lex_state = 19, .external_lex_state = 3},
  [223] = {.lex_state = 19, .external_lex_state = 4},
  [224] = {.lex_state = 19, .external_lex_state = 4},
  [225] = {.lex_state = 19, .external_lex_state = 4},
  [226] = {.lex_state = 19, .external_lex_state = 4},
  [227] = {.lex_state = 19, .external_lex_state = 4},
  [228] = {.lex_state = 19, .external_lex_state = 4},
  [229] = {.lex_state = 19, .external_lex_state = 4},
  [230] = {.lex_state = 19, .external_lex_state = 4},
  [231] = {.lex_state = 19, .external_lex_state = 4},
  [232] = {.lex_state = 19, .external_lex_state = 3},
  [233] = {.lex_state = 19, .external_lex_state = 3},
  [234] = {.lex_state = 19, .external_lex_state = 3},
  [235] = {.lex_state = 19, .external_lex_state = 3},
  [236] = {.lex_state = 19, .external_lex_state = 3},
  [237] = {.lex_state = 19, .external_lex_state = 3},
  [238] = {.lex_state = 19, .external_lex_state = 3},
  [239] = {.lex_state = 19, .external_lex_state = 3},
  [240] = {.lex_state = 19, .external_lex_state = 3},
  [241] = {.lex_state = 19, .external_lex_state = 3},
  [242] = {.lex_state = 19, .external_lex_state = 3},
  [243] = {.lex_state = 19, .external_lex_state = 3},
  [244] = {.lex_state = 19, .external_lex_state = 3},
  [245] = {.lex_state = 19, .external_lex_state = 3},
  [246] = {.lex_state = 19, .external_lex_state = 3},
  [247] = {.lex_state = 19, .external_lex_state = 3},
  [248] = {.lex_state = 19, .external_lex_state = 2},
  [249] = {.lex_state = 7, .external_lex_state = 3},
  [250] = {.lex_state = 7, .external_lex_state = 3},
  [251] = {.lex_state = 7, .external_lex_state = 3},
  [252] = {.lex_state = 7, .external_lex_state = 3},
  [253] = {.lex_state = 7, .external_lex_state = 3},
  [254] = {.lex_state = 7, .external_lex_state = 3},
  [255] = {.lex_state = 7, .external_lex_state = 3},
  [256] = {.lex_state = 7, .external_lex_state = 3},
  [257] = {.lex_state = 19, .external_lex_state = 2},
  [258] = {.lex_state = 19, .external_lex_state = 2},
  [259] = {.lex_state = 19, .external_lex_state = 2},
  [260] = {.lex_state = 7, .external_lex_state = 3},
  [261] = {.lex_state = 19, .external_lex_state = 4},
  [262] = {.lex_state = 19, .external_lex_state = 4},
  [263] = {.lex_state = 19, .external_lex_state = 4},
  [264] = {.lex_state = 19, .external_lex_state = 4},
  [265] = {.lex_state = 1, .external_lex_state = 4},
  [266] = {.lex_state = 19, .external_lex_state = 4},
  [267] = {.lex_state = 19, .external_lex_state = 4},
  [268] = {.lex_state = 19, .external_lex_state = 4},
  [269] = {.lex_state = 19, .external_lex_state = 4},
  [270] = {.lex_state = 19, .external_lex_state = 4},
  [271] = {.lex_state = 19, .external_lex_state = 4},
  [272] = {.lex_state = 2, .external_lex_state = 4},
  [273] = {.lex_state = 19, .external_lex_state = 4},
  [274] = {.lex_state = 19, .external_lex_state = 4},
  [275] = {.lex_state = 19, .external_lex_state = 4},
  [276] = {.lex_state = 9, .external_lex_state = 4},
  [277] = {.lex_state = 2, .external_lex_state = 4},
  [278] = {.lex_state = 19, .external_lex_state = 4},
  [279] = {.lex_state = 9, .external_lex_state = 4},
  [280] = {.lex_state = 19, .external_lex_state = 4},
  [281] = {.lex_state = 9, .external_lex_state = 4},
  [282] = {.lex_state = 2, .external_lex_state = 4},
  [283] = {.lex_state = 9, .external_lex_state = 4},
  [284] = {.lex_state = 9, .external_lex_state = 4},
  [285] = {.lex_state = 9, .external_lex_state = 4},
  [286] = {.lex_state = 19, .external_lex_state = 4},
  [287] = {.lex_state = 19, .external_lex_state = 4},
  [288] = {.lex_state = 4, .external_lex_state = 4},
  [289] = {.lex_state = 19, .external_lex_state = 4},
  [290] = {.lex_state = 1, .external_lex_state = 4},
  [291] = {.lex_state = 4, .external_lex_state = 4},
  [292] = {.lex_state = 19, .external_lex_state = 4},
  [293] = {.lex_state = 6, .external_lex_state = 4},
  [294] = {.lex_state = 19, .external_lex_state = 4},
  [295] = {.lex_state = 19, .external_lex_state = 4},
  [296] = {.lex_state = 19, .external_lex_state = 4},
  [297] = {.lex_state = 0, .external_lex_state = 4},
  [298] = {.lex_state = 19, .external_lex_state = 4},
  [299] = {.lex_state = 19, .external_lex_state = 4},
  [300] = {.lex_state = 6, .external_lex_state = 4},
  [301] = {.lex_state = 5, .external_lex_state = 4},
  [302] = {.lex_state = 5, .external_lex_state = 4},
  [303] = {.lex_state = 19, .external_lex_state = 4},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__normal_bare_identifier] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [sym__identifier_char] = ACTIONS(1),
    [sym___identifier_char_no_digit] = ACTIONS(1),
    [anon_sym_null] = ACTIONS(1),
    [anon_sym_i8] = ACTIONS(1),
    [anon_sym_i16] = ACTIONS(1),
    [anon_sym_i32] = ACTIONS(1),
    [anon_sym_i64] = ACTIONS(1),
    [anon_sym_u8] = ACTIONS(1),
    [anon_sym_u16] = ACTIONS(1),
    [anon_sym_u32] = ACTIONS(1),
    [anon_sym_u64] = ACTIONS(1),
    [anon_sym_isize] = ACTIONS(1),
    [anon_sym_usize] = ACTIONS(1),
    [anon_sym_f32] = ACTIONS(1),
    [anon_sym_f64] = ACTIONS(1),
    [anon_sym_decimal64] = ACTIONS(1),
    [anon_sym_decimal128] = ACTIONS(1),
    [anon_sym_date_DASHtime] = ACTIONS(1),
    [anon_sym_time] = ACTIONS(1),
    [anon_sym_date] = ACTIONS(1),
    [anon_sym_duration] = ACTIONS(1),
    [anon_sym_decimal] = ACTIONS(1),
    [anon_sym_currency] = ACTIONS(1),
    [anon_sym_country_DASH2] = ACTIONS(1),
    [anon_sym_country_DASH3] = ACTIONS(1),
    [anon_sym_country_DASHsubdivision] = ACTIONS(1),
    [anon_sym_email] = ACTIONS(1),
    [anon_sym_idn_DASHemail] = ACTIONS(1),
    [anon_sym_hostname] = ACTIONS(1),
    [anon_sym_idn_DASHhostname] = ACTIONS(1),
    [anon_sym_ipv4] = ACTIONS(1),
    [anon_sym_ipv6] = ACTIONS(1),
    [anon_sym_url] = ACTIONS(1),
    [anon_sym_url_DASHreference] = ACTIONS(1),
    [anon_sym_irl] = ACTIONS(1),
    [anon_sym_iri_DASHreference] = ACTIONS(1),
    [anon_sym_url_DASHtemplate] = ACTIONS(1),
    [anon_sym_uuid] = ACTIONS(1),
    [anon_sym_regex] = ACTIONS(1),
    [anon_sym_base64] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym__escaped_string_token1] = ACTIONS(1),
    [sym__hex_digit] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_e] = ACTIONS(1),
    [anon_sym_E] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [sym__digit] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [aux_sym__octal_token1] = ACTIONS(1),
    [anon_sym_0] = ACTIONS(1),
    [anon_sym_1] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [aux_sym__newline_token2] = ACTIONS(1),
    [aux_sym__newline_token4] = ACTIONS(1),
    [aux_sym__newline_token5] = ACTIONS(1),
    [aux_sym__newline_token6] = ACTIONS(1),
    [aux_sym__newline_token7] = ACTIONS(1),
    [sym__bom] = ACTIONS(1),
    [sym__unicode_space] = ACTIONS(1),
    [aux_sym_single_line_comment_token1] = ACTIONS(1),
    [sym__eof] = ACTIONS(1),
    [sym_multi_line_comment] = ACTIONS(3),
    [sym__raw_string] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_document] = STATE(297),
    [sym_node] = STATE(31),
    [sym_identifier] = STATE(65),
    [sym__bare_identifier] = STATE(207),
    [sym_type] = STATE(258),
    [sym_string] = STATE(207),
    [sym__escaped_string] = STATE(191),
    [sym__sign] = STATE(196),
    [sym__linespace] = STATE(40),
    [sym__newline] = STATE(40),
    [sym__ws] = STATE(40),
    [sym_single_line_comment] = STATE(40),
    [aux_sym_document_repeat1] = STATE(40),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym__normal_bare_identifier] = ACTIONS(7),
    [anon_sym_SLASH_DASH] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(11),
    [anon_sym_DQUOTE] = ACTIONS(13),
    [anon_sym_PLUS] = ACTIONS(15),
    [anon_sym_DASH] = ACTIONS(15),
    [aux_sym__newline_token1] = ACTIONS(17),
    [aux_sym__newline_token2] = ACTIONS(17),
    [aux_sym__newline_token3] = ACTIONS(17),
    [aux_sym__newline_token4] = ACTIONS(17),
    [aux_sym__newline_token5] = ACTIONS(17),
    [aux_sym__newline_token6] = ACTIONS(17),
    [aux_sym__newline_token7] = ACTIONS(17),
    [sym__bom] = ACTIONS(17),
    [sym__unicode_space] = ACTIONS(17),
    [anon_sym_SLASH_SLASH] = ACTIONS(19),
    [sym_multi_line_comment] = ACTIONS(17),
    [sym__raw_string] = ACTIONS(21),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(112), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(142), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(29), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [130] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(97), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(171), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(49), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [260] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(116), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(135), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(51), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [390] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(88), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(144), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(53), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [520] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(98), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(133), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(55), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [650] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(92), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(148), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(57), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [780] = 36,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(25), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(79), 1,
      sym_type,
    STATE(105), 1,
      sym_node_children,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(158), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(59), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [910] = 10,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(63), 1,
      anon_sym_DQUOTE,
    ACTIONS(67), 1,
      sym__raw_string,
    STATE(290), 1,
      sym__sign,
    STATE(294), 1,
      sym__escaped_string,
    ACTIONS(65), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(287), 2,
      sym__bare_identifier,
      sym_string,
    STATE(298), 2,
      sym_identifier,
      sym_annotation_type,
    ACTIONS(61), 37,
      anon_sym_i8,
      anon_sym_i16,
      anon_sym_i32,
      anon_sym_i64,
      anon_sym_u8,
      anon_sym_u16,
      anon_sym_u32,
      anon_sym_u64,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_decimal64,
      anon_sym_decimal128,
      anon_sym_date_DASHtime,
      anon_sym_time,
      anon_sym_date,
      anon_sym_duration,
      anon_sym_decimal,
      anon_sym_currency,
      anon_sym_country_DASH2,
      anon_sym_country_DASH3,
      anon_sym_country_DASHsubdivision,
      anon_sym_email,
      anon_sym_idn_DASHemail,
      anon_sym_hostname,
      anon_sym_idn_DASHhostname,
      anon_sym_ipv4,
      anon_sym_ipv6,
      anon_sym_url,
      anon_sym_url_DASHreference,
      anon_sym_irl,
      anon_sym_iri_DASHreference,
      anon_sym_url_DASHtemplate,
      anon_sym_uuid,
      anon_sym_regex,
      anon_sym_base64,
  [980] = 31,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(69), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    STATE(74), 1,
      aux_sym_node_repeat1,
    STATE(79), 1,
      sym_type,
    STATE(122), 1,
      sym__escline,
    STATE(137), 1,
      sym__integer,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(198), 1,
      sym_node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    STATE(222), 2,
      sym__node_field_comment,
      sym__node_field,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [1085] = 30,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    ACTIONS(75), 1,
      anon_sym_LBRACE,
    STATE(12), 1,
      aux_sym_node_repeat1,
    STATE(79), 1,
      sym_type,
    STATE(122), 1,
      sym__escline,
    STATE(137), 1,
      sym__integer,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(211), 1,
      sym__node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [1186] = 30,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    ACTIONS(77), 1,
      anon_sym_LBRACE,
    STATE(74), 1,
      aux_sym_node_repeat1,
    STATE(79), 1,
      sym_type,
    STATE(122), 1,
      sym__escline,
    STATE(137), 1,
      sym__integer,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(212), 1,
      sym__node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [1287] = 29,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    STATE(14), 1,
      aux_sym_node_repeat1,
    STATE(79), 1,
      sym_type,
    STATE(122), 1,
      sym__escline,
    STATE(137), 1,
      sym__integer,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(211), 1,
      sym__node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [1385] = 29,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(23), 1,
      sym__normal_bare_identifier,
    ACTIONS(31), 1,
      anon_sym_null,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    STATE(74), 1,
      aux_sym_node_repeat1,
    STATE(79), 1,
      sym_type,
    STATE(122), 1,
      sym__escline,
    STATE(137), 1,
      sym__integer,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(194), 1,
      sym_string,
    STATE(212), 1,
      sym__node_field,
    STATE(213), 1,
      sym_boolean,
    STATE(265), 1,
      sym__sign,
    STATE(287), 1,
      sym__bare_identifier,
    STATE(296), 1,
      sym_identifier,
    ACTIONS(35), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(43), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(209), 2,
      sym_keyword,
      sym_number,
    STATE(220), 2,
      sym_prop,
      sym_value,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [1483] = 8,
    ACTIONS(83), 1,
      anon_sym_BSLASH,
    STATE(15), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(86), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(79), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(81), 20,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [1534] = 16,
    ACTIONS(91), 1,
      sym__normal_bare_identifier,
    ACTIONS(94), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(97), 1,
      anon_sym_LPAREN,
    ACTIONS(100), 1,
      anon_sym_DQUOTE,
    ACTIONS(109), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(112), 1,
      sym__raw_string,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(89), 2,
      ts_builtin_sym_end,
      anon_sym_RBRACE,
    ACTIONS(103), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(54), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(106), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1600] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(115), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(32), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(117), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1665] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(119), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(41), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(121), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1730] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(123), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(22), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(48), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(125), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1795] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(115), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(24), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(32), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(117), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1860] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(127), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(18), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(43), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(129), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1925] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(131), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(35), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(133), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [1990] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(131), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(26), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(35), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(133), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2055] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(135), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(42), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(137), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2120] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(139), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(17), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(49), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(141), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2185] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(143), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(44), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(145), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2250] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(143), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(29), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(44), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(145), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2315] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(127), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(43), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(129), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2380] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(147), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(16), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(46), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(149), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2445] = 6,
    ACTIONS(155), 1,
      anon_sym_BSLASH,
    STATE(52), 1,
      sym__escline,
    STATE(53), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(158), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(151), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(153), 20,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [2490] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(161), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(28), 2,
      sym_node,
      aux_sym_document_repeat2,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(45), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(163), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2555] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(135), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2619] = 4,
    STATE(53), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(171), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(167), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(169), 21,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [2659] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(131), 1,
      anon_sym_RBRACE,
    STATE(27), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(77), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(174), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2723] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(143), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2787] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(123), 1,
      anon_sym_RBRACE,
    STATE(23), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(76), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(176), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2851] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(178), 1,
      anon_sym_RBRACE,
    STATE(19), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(36), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(180), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2915] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(123), 1,
      anon_sym_RBRACE,
    STATE(23), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(34), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(182), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [2979] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(139), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(75), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(184), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3043] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(161), 1,
      ts_builtin_sym_end,
    STATE(21), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(78), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(186), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3107] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(188), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3171] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(190), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3235] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(119), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3299] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(147), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3363] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(127), 1,
      ts_builtin_sym_end,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3427] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(192), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3491] = 4,
    STATE(50), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(194), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(151), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(153), 21,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [3531] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(131), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3595] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(115), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3659] = 4,
    STATE(53), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(201), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(197), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(199), 21,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [3699] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(204), 1,
      anon_sym_RBRACE,
    STATE(25), 1,
      sym_node,
    STATE(65), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(39), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(206), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3763] = 4,
    STATE(33), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(208), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(197), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(199), 21,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [3803] = 4,
    STATE(53), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(215), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(211), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(213), 21,
      sym__eof,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [3843] = 15,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(9), 1,
      anon_sym_SLASH_DASH,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(21), 1,
      sym__raw_string,
    STATE(65), 1,
      sym_identifier,
    STATE(159), 1,
      sym_node,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(258), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(165), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [3904] = 2,
    ACTIONS(218), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(220), 24,
      sym__eof,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [3938] = 2,
    ACTIONS(224), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(222), 24,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [3972] = 2,
    ACTIONS(226), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(228), 24,
      sym__eof,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [4006] = 2,
    ACTIONS(230), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(232), 24,
      sym__eof,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [4040] = 2,
    ACTIONS(230), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(232), 24,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [4074] = 2,
    ACTIONS(224), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(222), 24,
      sym__eof,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [4108] = 2,
    ACTIONS(234), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(236), 24,
      sym__eof,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [4142] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(8), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(92), 1,
      sym_node_children,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(148), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(57), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4195] = 19,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(240), 1,
      anon_sym_null,
    STATE(79), 1,
      sym_type,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(200), 1,
      sym_value,
    STATE(213), 1,
      sym_boolean,
    STATE(270), 1,
      sym__sign,
    ACTIONS(242), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(244), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(209), 3,
      sym_keyword,
      sym_string,
      sym_number,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [4260] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(3), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(71), 1,
      aux_sym_node_repeat2,
    STATE(112), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(142), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(29), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4313] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(2), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(66), 1,
      aux_sym_node_repeat2,
    STATE(118), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(166), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(246), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4366] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(3), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(112), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(142), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(29), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4419] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(7), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(116), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(135), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(51), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4472] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(5), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(98), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(133), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(55), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4525] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(7), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(62), 1,
      aux_sym_node_repeat2,
    STATE(116), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(135), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(51), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4578] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(6), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(68), 1,
      aux_sym_node_repeat2,
    STATE(97), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(171), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(49), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4631] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(6), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(97), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(171), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(49), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4684] = 13,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(27), 1,
      anon_sym_LBRACE,
    ACTIONS(45), 1,
      anon_sym_BSLASH,
    ACTIONS(238), 1,
      anon_sym_SLASH_DASH,
    STATE(4), 1,
      aux_sym_node_repeat1,
    STATE(47), 1,
      sym__escline,
    STATE(57), 1,
      sym__node_space,
    STATE(67), 1,
      aux_sym_node_repeat2,
    STATE(107), 1,
      sym_node_children,
    STATE(30), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(47), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(149), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(248), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [4737] = 4,
    ACTIONS(255), 1,
      anon_sym_SLASH_SLASH,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(250), 9,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(252), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [4771] = 8,
    ACTIONS(258), 1,
      anon_sym_BSLASH,
    STATE(74), 1,
      aux_sym_node_repeat1,
    STATE(122), 1,
      sym__escline,
    STATE(173), 1,
      sym__node_space,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(261), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(79), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(81), 10,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
  [4812] = 5,
    ACTIONS(255), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(264), 1,
      anon_sym_RBRACE,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(250), 7,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(252), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [4847] = 5,
    ACTIONS(255), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(267), 1,
      anon_sym_RBRACE,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(250), 7,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(252), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [4882] = 5,
    ACTIONS(255), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(270), 1,
      anon_sym_RBRACE,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(250), 7,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(252), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [4917] = 5,
    ACTIONS(255), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(273), 1,
      ts_builtin_sym_end,
    STATE(73), 5,
      sym__linespace,
      sym__newline,
      sym__ws,
      sym_single_line_comment,
      aux_sym_document_repeat1,
    ACTIONS(250), 7,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(252), 10,
      sym_multi_line_comment,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
  [4952] = 16,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(37), 1,
      anon_sym_0x,
    ACTIONS(39), 1,
      anon_sym_0o,
    ACTIONS(41), 1,
      anon_sym_0b,
    ACTIONS(240), 1,
      anon_sym_null,
    STATE(137), 1,
      sym__integer,
    STATE(191), 1,
      sym__escaped_string,
    STATE(213), 1,
      sym_boolean,
    STATE(270), 1,
      sym__sign,
    ACTIONS(242), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(244), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(215), 3,
      sym_keyword,
      sym_string,
      sym_number,
    STATE(205), 4,
      sym__decimal,
      sym__hex,
      sym__octal,
      sym__binary,
  [5008] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(131), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(276), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5049] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(140), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(282), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5090] = 3,
    STATE(82), 1,
      aux_sym__integer_repeat1,
    ACTIONS(286), 2,
      anon_sym__,
      sym__digit,
    ACTIONS(284), 19,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_e,
      anon_sym_E,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [5119] = 3,
    STATE(106), 1,
      aux_sym__integer_repeat1,
    ACTIONS(291), 2,
      anon_sym__,
      sym__digit,
    ACTIONS(289), 19,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_e,
      anon_sym_E,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [5148] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(152), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(293), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5189] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(141), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(295), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5230] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(153), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(297), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5271] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(120), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(153), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(297), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5312] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(100), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(101), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(303), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(154), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(301), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5353] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(155), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(305), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5394] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(156), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(307), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5435] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(102), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(156), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(307), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5476] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(103), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(104), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(311), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(157), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(309), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5517] = 8,
    ACTIONS(315), 1,
      anon_sym_BSLASH,
    STATE(10), 1,
      aux_sym_node_repeat1,
    STATE(93), 1,
      aux_sym_node_repeat2,
    STATE(122), 1,
      sym__escline,
    STATE(173), 1,
      sym__node_space,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(318), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(313), 12,
      sym__eof,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [5556] = 6,
    ACTIONS(321), 1,
      anon_sym_BSLASH,
    STATE(123), 1,
      sym__escline,
    STATE(126), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(324), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(151), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(153), 10,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
  [5591] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(81), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(131), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(276), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5632] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(84), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(141), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(295), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5673] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(85), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(96), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(329), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(132), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(327), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5714] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(86), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(87), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(333), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(128), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(331), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5755] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(117), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(165), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(335), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5796] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(161), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(337), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5837] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(108), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(161), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(337), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5878] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(162), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(339), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5919] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(163), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(341), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [5960] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(109), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(163), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(341), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6001] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(110), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(111), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(345), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(164), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(343), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6042] = 3,
    STATE(82), 1,
      aux_sym__integer_repeat1,
    ACTIONS(349), 2,
      anon_sym__,
      sym__digit,
    ACTIONS(347), 19,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_e,
      anon_sym_E,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [6071] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(114), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(115), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(353), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(134), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(351), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6112] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(167), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(355), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6153] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(168), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(357), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6194] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(169), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(359), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6235] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(113), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(169), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(359), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6276] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(80), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(95), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(363), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(143), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(361), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6317] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(170), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(365), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6358] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(145), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(367), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6399] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(89), 1,
      aux_sym_node_repeat1,
    STATE(204), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(195), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(299), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(145), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(367), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6440] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(90), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(91), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(371), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(146), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(369), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6481] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(136), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(373), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6522] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(119), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(99), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(377), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(130), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(375), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6563] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(165), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(335), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6604] = 9,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    ACTIONS(278), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(280), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    STATE(160), 3,
      sym__node_terminator,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(379), 9,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [6645] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    STATE(64), 1,
      sym_identifier,
    STATE(122), 1,
      sym__escline,
    STATE(125), 1,
      aux_sym_node_repeat1,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(257), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [6699] = 4,
    STATE(124), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(381), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(151), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(153), 11,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
  [6729] = 4,
    STATE(127), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(384), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(197), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(199), 11,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
  [6759] = 4,
    STATE(126), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(387), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(197), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(199), 11,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
  [6789] = 16,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    ACTIONS(71), 1,
      anon_sym_BSLASH,
    STATE(72), 1,
      sym_identifier,
    STATE(74), 1,
      aux_sym_node_repeat1,
    STATE(122), 1,
      sym__escline,
    STATE(173), 1,
      sym__node_space,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    STATE(259), 1,
      sym_type,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(94), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
    ACTIONS(73), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [6843] = 4,
    STATE(126), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(390), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(211), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(213), 11,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
  [6873] = 4,
    STATE(126), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(393), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(167), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(169), 11,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
  [6903] = 1,
    ACTIONS(396), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [6926] = 3,
    STATE(138), 1,
      aux_sym__binary_repeat1,
    ACTIONS(400), 3,
      anon_sym__,
      anon_sym_0,
      anon_sym_1,
    ACTIONS(398), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [6953] = 1,
    ACTIONS(402), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [6976] = 1,
    ACTIONS(404), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [6999] = 1,
    ACTIONS(406), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7022] = 1,
    ACTIONS(408), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7045] = 1,
    ACTIONS(410), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7068] = 1,
    ACTIONS(412), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7091] = 1,
    ACTIONS(414), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7114] = 4,
    ACTIONS(418), 1,
      anon_sym_DOT,
    STATE(216), 1,
      sym__exponent,
    ACTIONS(420), 2,
      anon_sym_e,
      anon_sym_E,
    ACTIONS(416), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7143] = 3,
    STATE(150), 1,
      aux_sym__binary_repeat1,
    ACTIONS(424), 3,
      anon_sym__,
      anon_sym_0,
      anon_sym_1,
    ACTIONS(422), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7170] = 3,
    STATE(151), 1,
      aux_sym__binary_repeat1,
    ACTIONS(426), 3,
      anon_sym__,
      anon_sym_0,
      anon_sym_1,
    ACTIONS(422), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7197] = 1,
    ACTIONS(428), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7220] = 1,
    ACTIONS(430), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7243] = 1,
    ACTIONS(432), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7266] = 1,
    ACTIONS(434), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7289] = 1,
    ACTIONS(436), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7312] = 1,
    ACTIONS(438), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7335] = 1,
    ACTIONS(440), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7358] = 4,
    ACTIONS(444), 1,
      anon_sym_DOT,
    STATE(201), 1,
      sym__exponent,
    ACTIONS(420), 2,
      anon_sym_e,
      anon_sym_E,
    ACTIONS(442), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7387] = 1,
    ACTIONS(446), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7410] = 1,
    ACTIONS(448), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7433] = 3,
    STATE(150), 1,
      aux_sym__binary_repeat1,
    ACTIONS(452), 3,
      anon_sym__,
      anon_sym_0,
      anon_sym_1,
    ACTIONS(450), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7460] = 3,
    STATE(150), 1,
      aux_sym__binary_repeat1,
    ACTIONS(424), 3,
      anon_sym__,
      anon_sym_0,
      anon_sym_1,
    ACTIONS(455), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7487] = 1,
    ACTIONS(457), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7510] = 1,
    ACTIONS(459), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7533] = 1,
    ACTIONS(461), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7556] = 1,
    ACTIONS(463), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7579] = 1,
    ACTIONS(465), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7602] = 1,
    ACTIONS(467), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7625] = 1,
    ACTIONS(469), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7648] = 1,
    ACTIONS(89), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7671] = 1,
    ACTIONS(471), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7694] = 1,
    ACTIONS(473), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7717] = 1,
    ACTIONS(475), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7740] = 1,
    ACTIONS(477), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7763] = 1,
    ACTIONS(479), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7786] = 1,
    ACTIONS(481), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7809] = 1,
    ACTIONS(483), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7832] = 1,
    ACTIONS(485), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7855] = 1,
    ACTIONS(487), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7878] = 1,
    ACTIONS(489), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7901] = 1,
    ACTIONS(491), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7924] = 1,
    ACTIONS(493), 20,
      sym_multi_line_comment,
      sym__raw_string,
      ts_builtin_sym_end,
      anon_sym_SLASH_DASH,
      anon_sym_RBRACE,
      sym__normal_bare_identifier,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7947] = 3,
    STATE(172), 1,
      aux_sym__octal_repeat1,
    ACTIONS(497), 2,
      anon_sym__,
      aux_sym__octal_token1,
    ACTIONS(495), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [7973] = 2,
    ACTIONS(226), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(228), 14,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [7997] = 3,
    STATE(174), 1,
      aux_sym__hex_repeat1,
    ACTIONS(502), 2,
      sym__hex_digit,
      anon_sym__,
    ACTIONS(500), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8023] = 3,
    STATE(172), 1,
      aux_sym__octal_repeat1,
    ACTIONS(507), 2,
      anon_sym__,
      aux_sym__octal_token1,
    ACTIONS(505), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8049] = 2,
    ACTIONS(234), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(236), 14,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [8073] = 3,
    STATE(174), 1,
      aux_sym__hex_repeat1,
    ACTIONS(511), 2,
      sym__hex_digit,
      anon_sym__,
    ACTIONS(509), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8099] = 3,
    STATE(210), 1,
      sym__exponent,
    ACTIONS(420), 2,
      anon_sym_e,
      anon_sym_E,
    ACTIONS(513), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8125] = 3,
    STATE(202), 1,
      sym__exponent,
    ACTIONS(420), 2,
      anon_sym_e,
      anon_sym_E,
    ACTIONS(515), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8151] = 3,
    STATE(177), 1,
      aux_sym__hex_repeat1,
    ACTIONS(519), 2,
      sym__hex_digit,
      anon_sym__,
    ACTIONS(517), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8177] = 3,
    STATE(175), 1,
      aux_sym__octal_repeat1,
    ACTIONS(523), 2,
      anon_sym__,
      aux_sym__octal_token1,
    ACTIONS(521), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8203] = 2,
    ACTIONS(218), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(220), 14,
      sym_multi_line_comment,
      sym__raw_string,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [8227] = 3,
    STATE(174), 1,
      aux_sym__hex_repeat1,
    ACTIONS(511), 2,
      sym__hex_digit,
      anon_sym__,
    ACTIONS(517), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8253] = 3,
    STATE(183), 1,
      aux_sym__hex_repeat1,
    ACTIONS(527), 2,
      sym__hex_digit,
      anon_sym__,
    ACTIONS(525), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8279] = 3,
    STATE(186), 1,
      aux_sym__octal_repeat1,
    ACTIONS(531), 2,
      anon_sym__,
      aux_sym__octal_token1,
    ACTIONS(529), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8305] = 3,
    STATE(172), 1,
      aux_sym__octal_repeat1,
    ACTIONS(507), 2,
      anon_sym__,
      aux_sym__octal_token1,
    ACTIONS(521), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8331] = 7,
    ACTIONS(533), 1,
      anon_sym_BSLASH,
    STATE(187), 1,
      aux_sym_node_repeat1,
    STATE(217), 1,
      sym__escline,
    STATE(234), 1,
      sym__node_space,
    STATE(197), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(536), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(81), 10,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8365] = 3,
    ACTIONS(541), 1,
      sym__identifier_char,
    STATE(188), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(539), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8390] = 3,
    ACTIONS(546), 1,
      sym__identifier_char,
    STATE(188), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(544), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8415] = 3,
    ACTIONS(550), 1,
      sym__identifier_char,
    STATE(189), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(548), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8440] = 1,
    ACTIONS(552), 17,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8460] = 1,
    ACTIONS(554), 17,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8480] = 1,
    ACTIONS(556), 17,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_EQ,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8500] = 2,
    ACTIONS(560), 1,
      anon_sym_EQ,
    ACTIONS(558), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8522] = 5,
    ACTIONS(565), 1,
      anon_sym_BSLASH,
    STATE(221), 1,
      sym__escline,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(569), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(562), 10,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8550] = 2,
    ACTIONS(575), 1,
      sym___identifier_char_no_digit,
    ACTIONS(573), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8572] = 5,
    ACTIONS(577), 1,
      anon_sym_BSLASH,
    STATE(221), 1,
      sym__escline,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(580), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(153), 10,
      sym__eof,
      anon_sym_SEMI,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8600] = 1,
    ACTIONS(313), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8619] = 3,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(583), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(213), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8642] = 1,
    ACTIONS(586), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8661] = 1,
    ACTIONS(588), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8680] = 1,
    ACTIONS(590), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8699] = 1,
    ACTIONS(592), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8718] = 3,
    STATE(206), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(597), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(594), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8741] = 1,
    ACTIONS(601), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8760] = 3,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(606), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(603), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8783] = 1,
    ACTIONS(560), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8802] = 3,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(610), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(169), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [8825] = 1,
    ACTIONS(558), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8844] = 1,
    ACTIONS(613), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8863] = 1,
    ACTIONS(615), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8882] = 1,
    ACTIONS(617), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8901] = 1,
    ACTIONS(619), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8920] = 1,
    ACTIONS(621), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8939] = 1,
    ACTIONS(623), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8958] = 1,
    ACTIONS(442), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [8977] = 3,
    STATE(219), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(625), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(153), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [9000] = 1,
    ACTIONS(628), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9019] = 3,
    STATE(199), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(630), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(199), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [9042] = 1,
    ACTIONS(633), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9061] = 3,
    STATE(208), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(635), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(199), 11,
      sym__eof,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [9084] = 1,
    ACTIONS(638), 16,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SLASH_DASH,
      anon_sym_LBRACE,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9103] = 5,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    STATE(182), 2,
      sym__newline,
      sym_single_line_comment,
    STATE(226), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(642), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(640), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9129] = 5,
    ACTIONS(648), 1,
      anon_sym_SLASH_SLASH,
    STATE(227), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(243), 2,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(646), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(644), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9155] = 5,
    ACTIONS(654), 1,
      anon_sym_SLASH_SLASH,
    STATE(55), 2,
      sym__newline,
      sym_single_line_comment,
    STATE(228), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(652), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(650), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9181] = 5,
    ACTIONS(19), 1,
      anon_sym_SLASH_SLASH,
    STATE(176), 2,
      sym__newline,
      sym_single_line_comment,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(658), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(656), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9207] = 5,
    ACTIONS(648), 1,
      anon_sym_SLASH_SLASH,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(242), 2,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(658), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(660), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9233] = 5,
    ACTIONS(654), 1,
      anon_sym_SLASH_SLASH,
    STATE(61), 2,
      sym__newline,
      sym_single_line_comment,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(658), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(662), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9259] = 5,
    ACTIONS(668), 1,
      anon_sym_SLASH_SLASH,
    STATE(230), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(273), 2,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(666), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(664), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9285] = 5,
    ACTIONS(668), 1,
      anon_sym_SLASH_SLASH,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    STATE(278), 2,
      sym__newline,
      sym_single_line_comment,
    ACTIONS(658), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(670), 7,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9311] = 3,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(672), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
    ACTIONS(213), 10,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      anon_sym_SLASH_SLASH,
  [9333] = 1,
    ACTIONS(675), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9350] = 1,
    ACTIONS(677), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9367] = 1,
    ACTIONS(228), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9384] = 1,
    ACTIONS(679), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9401] = 1,
    ACTIONS(681), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9418] = 1,
    ACTIONS(683), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9435] = 1,
    ACTIONS(685), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9452] = 1,
    ACTIONS(687), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9469] = 1,
    ACTIONS(689), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9486] = 1,
    ACTIONS(691), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9503] = 1,
    ACTIONS(236), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9520] = 1,
    ACTIONS(220), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9537] = 1,
    ACTIONS(232), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9554] = 1,
    ACTIONS(693), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9571] = 1,
    ACTIONS(695), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9588] = 1,
    ACTIONS(222), 14,
      sym__eof,
      sym_multi_line_comment,
      anon_sym_SEMI,
      anon_sym_BSLASH,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
      sym__bom,
      sym__unicode_space,
      anon_sym_SLASH_SLASH,
  [9605] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(697), 5,
      sym__normal_bare_identifier,
      anon_sym_null,
      sym__digit,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(699), 7,
      sym__raw_string,
      anon_sym_DQUOTE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_0x,
      anon_sym_0o,
      anon_sym_0b,
  [9625] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(703), 1,
      aux_sym_single_line_comment_token1,
    STATE(244), 1,
      sym__newline,
    STATE(256), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(701), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9648] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(707), 1,
      aux_sym_single_line_comment_token1,
    STATE(251), 1,
      aux_sym_single_line_comment_repeat1,
    STATE(275), 1,
      sym__newline,
    ACTIONS(705), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9671] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(711), 1,
      aux_sym_single_line_comment_token1,
    STATE(260), 1,
      aux_sym_single_line_comment_repeat1,
    STATE(274), 1,
      sym__newline,
    ACTIONS(709), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9694] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(715), 1,
      aux_sym_single_line_comment_token1,
    STATE(58), 1,
      sym__newline,
    STATE(254), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(713), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9717] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(711), 1,
      aux_sym_single_line_comment_token1,
    STATE(56), 1,
      sym__newline,
    STATE(260), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(717), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9740] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(711), 1,
      aux_sym_single_line_comment_token1,
    STATE(60), 1,
      sym__newline,
    STATE(260), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(719), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9763] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(723), 1,
      aux_sym_single_line_comment_token1,
    STATE(59), 1,
      sym__newline,
    STATE(253), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(721), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9786] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(711), 1,
      aux_sym_single_line_comment_token1,
    STATE(247), 1,
      sym__newline,
    STATE(260), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(725), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9809] = 9,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    STATE(70), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
  [9839] = 9,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    STATE(64), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
  [9869] = 9,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(7), 1,
      sym__normal_bare_identifier,
    ACTIONS(13), 1,
      anon_sym_DQUOTE,
    ACTIONS(21), 1,
      sym__raw_string,
    STATE(69), 1,
      sym_identifier,
    STATE(191), 1,
      sym__escaped_string,
    STATE(196), 1,
      sym__sign,
    ACTIONS(15), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    STATE(207), 2,
      sym__bare_identifier,
      sym_string,
  [9899] = 4,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(729), 1,
      aux_sym_single_line_comment_token1,
    STATE(260), 1,
      aux_sym_single_line_comment_repeat1,
    ACTIONS(727), 8,
      sym__eof,
      aux_sym__newline_token1,
      aux_sym__newline_token2,
      aux_sym__newline_token3,
      aux_sym__newline_token4,
      aux_sym__newline_token5,
      aux_sym__newline_token6,
      aux_sym__newline_token7,
  [9919] = 7,
    ACTIONS(81), 1,
      anon_sym_LBRACE,
    ACTIONS(732), 1,
      anon_sym_BSLASH,
    STATE(261), 1,
      aux_sym_node_repeat1,
    STATE(267), 1,
      sym__escline,
    STATE(280), 1,
      sym__node_space,
    STATE(264), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(735), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [9944] = 7,
    ACTIONS(77), 1,
      anon_sym_LBRACE,
    ACTIONS(738), 1,
      anon_sym_BSLASH,
    STATE(261), 1,
      aux_sym_node_repeat1,
    STATE(267), 1,
      sym__escline,
    STATE(280), 1,
      sym__node_space,
    STATE(264), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(740), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [9969] = 7,
    ACTIONS(75), 1,
      anon_sym_LBRACE,
    ACTIONS(738), 1,
      anon_sym_BSLASH,
    STATE(262), 1,
      aux_sym_node_repeat1,
    STATE(267), 1,
      sym__escline,
    STATE(280), 1,
      sym__node_space,
    STATE(264), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(740), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [9994] = 5,
    ACTIONS(153), 1,
      anon_sym_LBRACE,
    ACTIONS(742), 1,
      anon_sym_BSLASH,
    STATE(268), 1,
      sym__escline,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(745), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [10013] = 8,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(573), 1,
      anon_sym_EQ,
    ACTIONS(748), 1,
      sym___identifier_char_no_digit,
    ACTIONS(750), 1,
      anon_sym_0x,
    ACTIONS(752), 1,
      anon_sym_0o,
    ACTIONS(754), 1,
      anon_sym_0b,
    STATE(147), 1,
      sym__integer,
  [10038] = 3,
    ACTIONS(199), 2,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(756), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [10052] = 3,
    ACTIONS(153), 2,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
    STATE(266), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(759), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [10066] = 3,
    ACTIONS(199), 2,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
    STATE(269), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(762), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [10080] = 3,
    ACTIONS(169), 2,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
    STATE(231), 2,
      sym__ws,
      aux_sym_node_repeat3,
    ACTIONS(765), 3,
      sym_multi_line_comment,
      sym__bom,
      sym__unicode_space,
  [10094] = 6,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(33), 1,
      sym__digit,
    ACTIONS(750), 1,
      anon_sym_0x,
    ACTIONS(752), 1,
      anon_sym_0o,
    ACTIONS(754), 1,
      anon_sym_0b,
    STATE(147), 1,
      sym__integer,
  [10113] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(768), 1,
      sym__digit,
    STATE(218), 1,
      sym__integer,
    STATE(292), 1,
      sym__sign,
    ACTIONS(770), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
  [10130] = 4,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(772), 1,
      sym__identifier_char,
    STATE(272), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(539), 2,
      anon_sym_EQ,
      anon_sym_RPAREN,
  [10144] = 1,
    ACTIONS(220), 5,
      sym_multi_line_comment,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [10152] = 1,
    ACTIONS(222), 5,
      sym_multi_line_comment,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [10160] = 1,
    ACTIONS(232), 5,
      sym_multi_line_comment,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [10168] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(775), 1,
      anon_sym_DQUOTE,
    ACTIONS(777), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(779), 1,
      sym_escape,
    STATE(279), 1,
      aux_sym__escaped_string_repeat1,
  [10184] = 4,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(781), 1,
      sym__identifier_char,
    STATE(282), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(548), 2,
      anon_sym_EQ,
      anon_sym_RPAREN,
  [10198] = 1,
    ACTIONS(236), 5,
      sym_multi_line_comment,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [10206] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(783), 1,
      anon_sym_DQUOTE,
    ACTIONS(785), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(788), 1,
      sym_escape,
    STATE(279), 1,
      aux_sym__escaped_string_repeat1,
  [10222] = 1,
    ACTIONS(228), 5,
      sym_multi_line_comment,
      anon_sym_LBRACE,
      anon_sym_BSLASH,
      sym__bom,
      sym__unicode_space,
  [10230] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(777), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(779), 1,
      sym_escape,
    ACTIONS(791), 1,
      anon_sym_DQUOTE,
    STATE(276), 1,
      aux_sym__escaped_string_repeat1,
  [10246] = 4,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(793), 1,
      sym__identifier_char,
    STATE(272), 1,
      aux_sym__bare_identifier_repeat1,
    ACTIONS(544), 2,
      anon_sym_EQ,
      anon_sym_RPAREN,
  [10260] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(777), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(779), 1,
      sym_escape,
    ACTIONS(795), 1,
      anon_sym_DQUOTE,
    STATE(284), 1,
      aux_sym__escaped_string_repeat1,
  [10276] = 5,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(777), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(779), 1,
      sym_escape,
    ACTIONS(797), 1,
      anon_sym_DQUOTE,
    STATE(279), 1,
      aux_sym__escaped_string_repeat1,
  [10292] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(801), 1,
      aux_sym__escaped_string_token1,
    ACTIONS(799), 2,
      anon_sym_DQUOTE,
      sym_escape,
  [10303] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(768), 1,
      sym__digit,
    STATE(178), 1,
      sym__integer,
  [10313] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(560), 2,
      anon_sym_EQ,
      anon_sym_RPAREN,
  [10321] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(803), 2,
      anon_sym_0,
      anon_sym_1,
  [10329] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(768), 1,
      sym__digit,
    STATE(179), 1,
      sym__integer,
  [10339] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(573), 1,
      anon_sym_RPAREN,
    ACTIONS(748), 1,
      sym___identifier_char_no_digit,
  [10349] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(805), 2,
      anon_sym_0,
      anon_sym_1,
  [10357] = 3,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(768), 1,
      sym__digit,
    STATE(203), 1,
      sym__integer,
  [10367] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(807), 1,
      sym__hex_digit,
  [10374] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(552), 1,
      anon_sym_RPAREN,
  [10381] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(556), 1,
      anon_sym_RPAREN,
  [10388] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(809), 1,
      anon_sym_EQ,
  [10395] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(811), 1,
      ts_builtin_sym_end,
  [10402] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(813), 1,
      anon_sym_RPAREN,
  [10409] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(554), 1,
      anon_sym_RPAREN,
  [10416] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(815), 1,
      sym__hex_digit,
  [10423] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(817), 1,
      aux_sym__octal_token1,
  [10430] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(819), 1,
      aux_sym__octal_token1,
  [10437] = 2,
    ACTIONS(3), 1,
      sym_multi_line_comment,
    ACTIONS(821), 1,
      anon_sym_RPAREN,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(2)] = 0,
  [SMALL_STATE(3)] = 130,
  [SMALL_STATE(4)] = 260,
  [SMALL_STATE(5)] = 390,
  [SMALL_STATE(6)] = 520,
  [SMALL_STATE(7)] = 650,
  [SMALL_STATE(8)] = 780,
  [SMALL_STATE(9)] = 910,
  [SMALL_STATE(10)] = 980,
  [SMALL_STATE(11)] = 1085,
  [SMALL_STATE(12)] = 1186,
  [SMALL_STATE(13)] = 1287,
  [SMALL_STATE(14)] = 1385,
  [SMALL_STATE(15)] = 1483,
  [SMALL_STATE(16)] = 1534,
  [SMALL_STATE(17)] = 1600,
  [SMALL_STATE(18)] = 1665,
  [SMALL_STATE(19)] = 1730,
  [SMALL_STATE(20)] = 1795,
  [SMALL_STATE(21)] = 1860,
  [SMALL_STATE(22)] = 1925,
  [SMALL_STATE(23)] = 1990,
  [SMALL_STATE(24)] = 2055,
  [SMALL_STATE(25)] = 2120,
  [SMALL_STATE(26)] = 2185,
  [SMALL_STATE(27)] = 2250,
  [SMALL_STATE(28)] = 2315,
  [SMALL_STATE(29)] = 2380,
  [SMALL_STATE(30)] = 2445,
  [SMALL_STATE(31)] = 2490,
  [SMALL_STATE(32)] = 2555,
  [SMALL_STATE(33)] = 2619,
  [SMALL_STATE(34)] = 2659,
  [SMALL_STATE(35)] = 2723,
  [SMALL_STATE(36)] = 2787,
  [SMALL_STATE(37)] = 2851,
  [SMALL_STATE(38)] = 2915,
  [SMALL_STATE(39)] = 2979,
  [SMALL_STATE(40)] = 3043,
  [SMALL_STATE(41)] = 3107,
  [SMALL_STATE(42)] = 3171,
  [SMALL_STATE(43)] = 3235,
  [SMALL_STATE(44)] = 3299,
  [SMALL_STATE(45)] = 3363,
  [SMALL_STATE(46)] = 3427,
  [SMALL_STATE(47)] = 3491,
  [SMALL_STATE(48)] = 3531,
  [SMALL_STATE(49)] = 3595,
  [SMALL_STATE(50)] = 3659,
  [SMALL_STATE(51)] = 3699,
  [SMALL_STATE(52)] = 3763,
  [SMALL_STATE(53)] = 3803,
  [SMALL_STATE(54)] = 3843,
  [SMALL_STATE(55)] = 3904,
  [SMALL_STATE(56)] = 3938,
  [SMALL_STATE(57)] = 3972,
  [SMALL_STATE(58)] = 4006,
  [SMALL_STATE(59)] = 4040,
  [SMALL_STATE(60)] = 4074,
  [SMALL_STATE(61)] = 4108,
  [SMALL_STATE(62)] = 4142,
  [SMALL_STATE(63)] = 4195,
  [SMALL_STATE(64)] = 4260,
  [SMALL_STATE(65)] = 4313,
  [SMALL_STATE(66)] = 4366,
  [SMALL_STATE(67)] = 4419,
  [SMALL_STATE(68)] = 4472,
  [SMALL_STATE(69)] = 4525,
  [SMALL_STATE(70)] = 4578,
  [SMALL_STATE(71)] = 4631,
  [SMALL_STATE(72)] = 4684,
  [SMALL_STATE(73)] = 4737,
  [SMALL_STATE(74)] = 4771,
  [SMALL_STATE(75)] = 4812,
  [SMALL_STATE(76)] = 4847,
  [SMALL_STATE(77)] = 4882,
  [SMALL_STATE(78)] = 4917,
  [SMALL_STATE(79)] = 4952,
  [SMALL_STATE(80)] = 5008,
  [SMALL_STATE(81)] = 5049,
  [SMALL_STATE(82)] = 5090,
  [SMALL_STATE(83)] = 5119,
  [SMALL_STATE(84)] = 5148,
  [SMALL_STATE(85)] = 5189,
  [SMALL_STATE(86)] = 5230,
  [SMALL_STATE(87)] = 5271,
  [SMALL_STATE(88)] = 5312,
  [SMALL_STATE(89)] = 5353,
  [SMALL_STATE(90)] = 5394,
  [SMALL_STATE(91)] = 5435,
  [SMALL_STATE(92)] = 5476,
  [SMALL_STATE(93)] = 5517,
  [SMALL_STATE(94)] = 5556,
  [SMALL_STATE(95)] = 5591,
  [SMALL_STATE(96)] = 5632,
  [SMALL_STATE(97)] = 5673,
  [SMALL_STATE(98)] = 5714,
  [SMALL_STATE(99)] = 5755,
  [SMALL_STATE(100)] = 5796,
  [SMALL_STATE(101)] = 5837,
  [SMALL_STATE(102)] = 5878,
  [SMALL_STATE(103)] = 5919,
  [SMALL_STATE(104)] = 5960,
  [SMALL_STATE(105)] = 6001,
  [SMALL_STATE(106)] = 6042,
  [SMALL_STATE(107)] = 6071,
  [SMALL_STATE(108)] = 6112,
  [SMALL_STATE(109)] = 6153,
  [SMALL_STATE(110)] = 6194,
  [SMALL_STATE(111)] = 6235,
  [SMALL_STATE(112)] = 6276,
  [SMALL_STATE(113)] = 6317,
  [SMALL_STATE(114)] = 6358,
  [SMALL_STATE(115)] = 6399,
  [SMALL_STATE(116)] = 6440,
  [SMALL_STATE(117)] = 6481,
  [SMALL_STATE(118)] = 6522,
  [SMALL_STATE(119)] = 6563,
  [SMALL_STATE(120)] = 6604,
  [SMALL_STATE(121)] = 6645,
  [SMALL_STATE(122)] = 6699,
  [SMALL_STATE(123)] = 6729,
  [SMALL_STATE(124)] = 6759,
  [SMALL_STATE(125)] = 6789,
  [SMALL_STATE(126)] = 6843,
  [SMALL_STATE(127)] = 6873,
  [SMALL_STATE(128)] = 6903,
  [SMALL_STATE(129)] = 6926,
  [SMALL_STATE(130)] = 6953,
  [SMALL_STATE(131)] = 6976,
  [SMALL_STATE(132)] = 6999,
  [SMALL_STATE(133)] = 7022,
  [SMALL_STATE(134)] = 7045,
  [SMALL_STATE(135)] = 7068,
  [SMALL_STATE(136)] = 7091,
  [SMALL_STATE(137)] = 7114,
  [SMALL_STATE(138)] = 7143,
  [SMALL_STATE(139)] = 7170,
  [SMALL_STATE(140)] = 7197,
  [SMALL_STATE(141)] = 7220,
  [SMALL_STATE(142)] = 7243,
  [SMALL_STATE(143)] = 7266,
  [SMALL_STATE(144)] = 7289,
  [SMALL_STATE(145)] = 7312,
  [SMALL_STATE(146)] = 7335,
  [SMALL_STATE(147)] = 7358,
  [SMALL_STATE(148)] = 7387,
  [SMALL_STATE(149)] = 7410,
  [SMALL_STATE(150)] = 7433,
  [SMALL_STATE(151)] = 7460,
  [SMALL_STATE(152)] = 7487,
  [SMALL_STATE(153)] = 7510,
  [SMALL_STATE(154)] = 7533,
  [SMALL_STATE(155)] = 7556,
  [SMALL_STATE(156)] = 7579,
  [SMALL_STATE(157)] = 7602,
  [SMALL_STATE(158)] = 7625,
  [SMALL_STATE(159)] = 7648,
  [SMALL_STATE(160)] = 7671,
  [SMALL_STATE(161)] = 7694,
  [SMALL_STATE(162)] = 7717,
  [SMALL_STATE(163)] = 7740,
  [SMALL_STATE(164)] = 7763,
  [SMALL_STATE(165)] = 7786,
  [SMALL_STATE(166)] = 7809,
  [SMALL_STATE(167)] = 7832,
  [SMALL_STATE(168)] = 7855,
  [SMALL_STATE(169)] = 7878,
  [SMALL_STATE(170)] = 7901,
  [SMALL_STATE(171)] = 7924,
  [SMALL_STATE(172)] = 7947,
  [SMALL_STATE(173)] = 7973,
  [SMALL_STATE(174)] = 7997,
  [SMALL_STATE(175)] = 8023,
  [SMALL_STATE(176)] = 8049,
  [SMALL_STATE(177)] = 8073,
  [SMALL_STATE(178)] = 8099,
  [SMALL_STATE(179)] = 8125,
  [SMALL_STATE(180)] = 8151,
  [SMALL_STATE(181)] = 8177,
  [SMALL_STATE(182)] = 8203,
  [SMALL_STATE(183)] = 8227,
  [SMALL_STATE(184)] = 8253,
  [SMALL_STATE(185)] = 8279,
  [SMALL_STATE(186)] = 8305,
  [SMALL_STATE(187)] = 8331,
  [SMALL_STATE(188)] = 8365,
  [SMALL_STATE(189)] = 8390,
  [SMALL_STATE(190)] = 8415,
  [SMALL_STATE(191)] = 8440,
  [SMALL_STATE(192)] = 8460,
  [SMALL_STATE(193)] = 8480,
  [SMALL_STATE(194)] = 8500,
  [SMALL_STATE(195)] = 8522,
  [SMALL_STATE(196)] = 8550,
  [SMALL_STATE(197)] = 8572,
  [SMALL_STATE(198)] = 8600,
  [SMALL_STATE(199)] = 8619,
  [SMALL_STATE(200)] = 8642,
  [SMALL_STATE(201)] = 8661,
  [SMALL_STATE(202)] = 8680,
  [SMALL_STATE(203)] = 8699,
  [SMALL_STATE(204)] = 8718,
  [SMALL_STATE(205)] = 8741,
  [SMALL_STATE(206)] = 8760,
  [SMALL_STATE(207)] = 8783,
  [SMALL_STATE(208)] = 8802,
  [SMALL_STATE(209)] = 8825,
  [SMALL_STATE(210)] = 8844,
  [SMALL_STATE(211)] = 8863,
  [SMALL_STATE(212)] = 8882,
  [SMALL_STATE(213)] = 8901,
  [SMALL_STATE(214)] = 8920,
  [SMALL_STATE(215)] = 8939,
  [SMALL_STATE(216)] = 8958,
  [SMALL_STATE(217)] = 8977,
  [SMALL_STATE(218)] = 9000,
  [SMALL_STATE(219)] = 9019,
  [SMALL_STATE(220)] = 9042,
  [SMALL_STATE(221)] = 9061,
  [SMALL_STATE(222)] = 9084,
  [SMALL_STATE(223)] = 9103,
  [SMALL_STATE(224)] = 9129,
  [SMALL_STATE(225)] = 9155,
  [SMALL_STATE(226)] = 9181,
  [SMALL_STATE(227)] = 9207,
  [SMALL_STATE(228)] = 9233,
  [SMALL_STATE(229)] = 9259,
  [SMALL_STATE(230)] = 9285,
  [SMALL_STATE(231)] = 9311,
  [SMALL_STATE(232)] = 9333,
  [SMALL_STATE(233)] = 9350,
  [SMALL_STATE(234)] = 9367,
  [SMALL_STATE(235)] = 9384,
  [SMALL_STATE(236)] = 9401,
  [SMALL_STATE(237)] = 9418,
  [SMALL_STATE(238)] = 9435,
  [SMALL_STATE(239)] = 9452,
  [SMALL_STATE(240)] = 9469,
  [SMALL_STATE(241)] = 9486,
  [SMALL_STATE(242)] = 9503,
  [SMALL_STATE(243)] = 9520,
  [SMALL_STATE(244)] = 9537,
  [SMALL_STATE(245)] = 9554,
  [SMALL_STATE(246)] = 9571,
  [SMALL_STATE(247)] = 9588,
  [SMALL_STATE(248)] = 9605,
  [SMALL_STATE(249)] = 9625,
  [SMALL_STATE(250)] = 9648,
  [SMALL_STATE(251)] = 9671,
  [SMALL_STATE(252)] = 9694,
  [SMALL_STATE(253)] = 9717,
  [SMALL_STATE(254)] = 9740,
  [SMALL_STATE(255)] = 9763,
  [SMALL_STATE(256)] = 9786,
  [SMALL_STATE(257)] = 9809,
  [SMALL_STATE(258)] = 9839,
  [SMALL_STATE(259)] = 9869,
  [SMALL_STATE(260)] = 9899,
  [SMALL_STATE(261)] = 9919,
  [SMALL_STATE(262)] = 9944,
  [SMALL_STATE(263)] = 9969,
  [SMALL_STATE(264)] = 9994,
  [SMALL_STATE(265)] = 10013,
  [SMALL_STATE(266)] = 10038,
  [SMALL_STATE(267)] = 10052,
  [SMALL_STATE(268)] = 10066,
  [SMALL_STATE(269)] = 10080,
  [SMALL_STATE(270)] = 10094,
  [SMALL_STATE(271)] = 10113,
  [SMALL_STATE(272)] = 10130,
  [SMALL_STATE(273)] = 10144,
  [SMALL_STATE(274)] = 10152,
  [SMALL_STATE(275)] = 10160,
  [SMALL_STATE(276)] = 10168,
  [SMALL_STATE(277)] = 10184,
  [SMALL_STATE(278)] = 10198,
  [SMALL_STATE(279)] = 10206,
  [SMALL_STATE(280)] = 10222,
  [SMALL_STATE(281)] = 10230,
  [SMALL_STATE(282)] = 10246,
  [SMALL_STATE(283)] = 10260,
  [SMALL_STATE(284)] = 10276,
  [SMALL_STATE(285)] = 10292,
  [SMALL_STATE(286)] = 10303,
  [SMALL_STATE(287)] = 10313,
  [SMALL_STATE(288)] = 10321,
  [SMALL_STATE(289)] = 10329,
  [SMALL_STATE(290)] = 10339,
  [SMALL_STATE(291)] = 10349,
  [SMALL_STATE(292)] = 10357,
  [SMALL_STATE(293)] = 10367,
  [SMALL_STATE(294)] = 10374,
  [SMALL_STATE(295)] = 10381,
  [SMALL_STATE(296)] = 10388,
  [SMALL_STATE(297)] = 10395,
  [SMALL_STATE(298)] = 10402,
  [SMALL_STATE(299)] = 10409,
  [SMALL_STATE(300)] = 10416,
  [SMALL_STATE(301)] = 10423,
  [SMALL_STATE(302)] = 10430,
  [SMALL_STATE(303)] = 10437,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(287),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(83),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(214),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [79] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0),
  [81] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0),
  [83] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(225),
  [86] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(30),
  [89] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0),
  [91] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(207),
  [94] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(121),
  [97] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(9),
  [100] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(281),
  [103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(196),
  [106] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(54),
  [109] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(255),
  [112] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat2, 2, 0, 0), SHIFT_REPEAT(191),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 3, 0, 0),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 2, 0, 0),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [151] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__node_space, 1, 0, 0),
  [153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0),
  [155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(225),
  [158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(53),
  [161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 1, 0, 0),
  [163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [167] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__node_space, 3, 0, 0),
  [169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_space, 3, 0, 0),
  [171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 3, 0, 0), SHIFT(53),
  [174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_document, 4, 0, 0),
  [190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(50),
  [197] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__node_space, 2, 0, 0),
  [199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0),
  [201] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(53),
  [204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(33),
  [211] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0),
  [213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0),
  [215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT_REPEAT(53),
  [218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__escline, 2, 0, 0),
  [220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__escline, 2, 0, 0),
  [222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_single_line_comment, 3, 0, 0),
  [224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_single_line_comment, 3, 0, 0),
  [226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_node_repeat1, 1, 0, 0),
  [228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 1, 0, 0),
  [230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_single_line_comment, 2, 0, 0),
  [232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_single_line_comment, 2, 0, 0),
  [234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__escline, 3, 0, 0),
  [236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__escline, 3, 0, 0),
  [238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0),
  [252] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(73),
  [255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT_REPEAT(255),
  [258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(94),
  [264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT(240),
  [267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT(238),
  [270] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_document_repeat1, 2, 0, 0), SHIFT(241),
  [273] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_document, 2, 0, 0), REDUCE(aux_sym_document_repeat1, 2, 0, 0),
  [276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__integer_repeat1, 2, 0, 0),
  [286] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__integer_repeat1, 2, 0, 0), SHIFT_REPEAT(82),
  [289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__integer, 1, 0, 0),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_node_repeat2, 2, 0, 0),
  [315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat2, 2, 0, 0), SHIFT_REPEAT(223),
  [318] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat2, 2, 0, 0), SHIFT_REPEAT(94),
  [321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(223),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(126),
  [327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__integer, 2, 0, 0),
  [349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [381] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(124),
  [384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(127),
  [387] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(126),
  [390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT_REPEAT(126),
  [393] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 3, 0, 0), SHIFT(126),
  [396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 11),
  [398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binary, 2, 0, 0),
  [400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 3, 0, 2),
  [404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 3),
  [406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 7),
  [408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 0),
  [410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 8),
  [412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 4),
  [414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 5, 0, 2),
  [416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 1, 0, 0),
  [418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binary, 3, 0, 0),
  [424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 3),
  [430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 7),
  [432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 3, 0, 0),
  [434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 4, 0, 3),
  [436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 0),
  [438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 8),
  [440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 12),
  [442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 2, 0, 0),
  [444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 6, 0, 4),
  [448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 4, 0, 4),
  [450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__binary_repeat1, 2, 0, 0),
  [452] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__binary_repeat1, 2, 0, 0), SHIFT_REPEAT(150),
  [455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binary, 4, 0, 0),
  [457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 7),
  [459] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 11),
  [461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 14),
  [463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 8),
  [465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 12),
  [467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 15),
  [469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 7, 0, 4),
  [471] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 8, 0, 11),
  [473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 8, 0, 14),
  [475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 8, 0, 12),
  [477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 8, 0, 15),
  [479] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 8, 0, 16),
  [481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 4, 0, 2),
  [483] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 2, 0, 0),
  [485] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 9, 0, 14),
  [487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 9, 0, 15),
  [489] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 9, 0, 16),
  [491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 10, 0, 16),
  [493] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node, 4, 0, 0),
  [495] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__octal_repeat1, 2, 0, 0),
  [497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__octal_repeat1, 2, 0, 0), SHIFT_REPEAT(172),
  [500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hex_repeat1, 2, 0, 0),
  [502] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hex_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [505] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__octal, 4, 0, 0),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hex, 4, 0, 0),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 4, 0, 13),
  [515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 3, 0, 10),
  [517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hex, 3, 0, 0),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__octal, 3, 0, 0),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hex, 2, 0, 0),
  [527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__octal, 2, 0, 0),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [533] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(224),
  [536] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(197),
  [539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__bare_identifier_repeat1, 2, 0, 0),
  [541] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__bare_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(188),
  [544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__bare_identifier, 3, 0, 0),
  [546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__bare_identifier, 2, 0, 0),
  [550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 0),
  [554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__escaped_string, 3, 0, 1),
  [556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__escaped_string, 2, 0, 0),
  [558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value, 1, 0, 0),
  [560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier, 1, 0, 0),
  [562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), REDUCE(aux_sym_node_repeat3, 2, 0, 0),
  [565] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT(224),
  [569] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT(199),
  [573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__bare_identifier, 1, 0, 0),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [577] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(224),
  [580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(199),
  [583] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT_REPEAT(199),
  [586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_prop, 3, 0, 0),
  [588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 3, 0, 0),
  [590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 4, 0, 10),
  [592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__exponent, 3, 0, 0),
  [594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), REDUCE(sym__node_space, 2, 0, 0),
  [597] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), REDUCE(sym__node_space, 2, 0, 0), SHIFT(206),
  [601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1, 0, 0),
  [603] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), REDUCE(sym__node_space, 3, 0, 0),
  [606] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), REDUCE(sym__node_space, 3, 0, 0), SHIFT(199),
  [610] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 3, 0, 0), SHIFT(199),
  [613] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__decimal, 5, 0, 13),
  [615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_field_comment, 2, 0, 6),
  [617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_field_comment, 3, 0, 9),
  [619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword, 1, 0, 0),
  [621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value, 2, 0, 0),
  [625] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(219),
  [628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__exponent, 2, 0, 0),
  [630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(199),
  [633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__node_field, 1, 0, 0),
  [635] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(208),
  [638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_field, 1, 0, 0),
  [640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [672] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat3, 2, 0, 0), SHIFT_REPEAT(231),
  [675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 6, 0, 0),
  [677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 7, 0, 5),
  [679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 8, 0, 5),
  [681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 3, 0, 5),
  [683] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 2, 0, 0),
  [685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 5, 0, 5),
  [687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 3, 0, 0),
  [689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 4, 0, 0),
  [691] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 6, 0, 5),
  [693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 5, 0, 0),
  [695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_node_children, 4, 0, 5),
  [697] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type, 3, 0, 0),
  [699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 3, 0, 0),
  [701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_single_line_comment_repeat1, 2, 0, 0),
  [729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_single_line_comment_repeat1, 2, 0, 0), SHIFT_REPEAT(260),
  [732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_node_repeat1, 2, 0, 0), SHIFT_REPEAT(264),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(229),
  [745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(231),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(231),
  [759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 1, 0, 0), SHIFT(266),
  [762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 2, 0, 0), SHIFT(269),
  [765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__node_space, 3, 0, 0), SHIFT(231),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__bare_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(272),
  [775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(285),
  [779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__escaped_string_repeat1, 2, 0, 0),
  [785] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__escaped_string_repeat1, 2, 0, 0), SHIFT_REPEAT(285),
  [788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__escaped_string_repeat1, 2, 0, 0), SHIFT_REPEAT(285),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__escaped_string_repeat1, 1, 0, 0),
  [801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__escaped_string_repeat1, 1, 0, 0),
  [803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [811] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_annotation_type, 1, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__eof = 0,
  ts_external_token_multi_line_comment = 1,
  ts_external_token__raw_string = 2,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__eof] = sym__eof,
  [ts_external_token_multi_line_comment] = sym_multi_line_comment,
  [ts_external_token__raw_string] = sym__raw_string,
};

static const bool ts_external_scanner_states[5][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__eof] = true,
    [ts_external_token_multi_line_comment] = true,
    [ts_external_token__raw_string] = true,
  },
  [2] = {
    [ts_external_token_multi_line_comment] = true,
    [ts_external_token__raw_string] = true,
  },
  [3] = {
    [ts_external_token__eof] = true,
    [ts_external_token_multi_line_comment] = true,
  },
  [4] = {
    [ts_external_token_multi_line_comment] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_kdl_external_scanner_create(void);
void tree_sitter_kdl_external_scanner_destroy(void *);
bool tree_sitter_kdl_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_kdl_external_scanner_serialize(void *, char *);
void tree_sitter_kdl_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_kdl(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym__normal_bare_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_kdl_external_scanner_create,
      tree_sitter_kdl_external_scanner_destroy,
      tree_sitter_kdl_external_scanner_scan,
      tree_sitter_kdl_external_scanner_serialize,
      tree_sitter_kdl_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "kdl",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 0,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
